#ifndef BatchLab_DataPointSet_h
#define BatchLab_DataPointSet_h

// Inheritance :
#include <BatchLab/Core/BaseObject.h>
#include <AIDA/IDataPointSet.h>
#include <Slash/Store/IStorable.h>
#include <Slash/Store/IArray.h>
#include <Slash/Meta/IObject.h>

#include <BatchLab/Core/Annotation.h>
#include <BatchLab/Core/DataPoint.h>

namespace BatchLab {

class DataPointSet 
:public BaseObject
,public virtual AIDA::IDataPointSet
,public virtual Slash::Store::IStorable
,public virtual Slash::Store::IArray //For the IO.
,public virtual Slash::Meta::IObject
{
public: //AIDA/IManagedObject
  virtual void* cast(const std::string&) const;
  virtual std::string type() const; //AIDA-v3r3p0
public: //AIDA/IDataPointSet
  virtual std::string title() const;
  virtual bool setTitle(const std::string&);
  virtual AIDA::IAnnotation& annotation();
  virtual const AIDA::IAnnotation& annotation() const;
  virtual int dimension() const;
  virtual void clear();
  virtual int size() const;
  virtual AIDA::IDataPoint* point(int);
  virtual const AIDA::IDataPoint* point(int) const;
  virtual AIDA::IDataPoint* addPoint();
  virtual bool addPoint(const AIDA::IDataPoint&);
  virtual bool removePoint(int);
  virtual double lowerExtent(int) const;
  virtual double upperExtent(int) const;
  virtual bool scale(double);
  virtual bool scaleValues(double);
  virtual bool scaleErrors(double);
  virtual bool setCoordinate(int coord, 
                             const std::vector<double>&,
                             const std::vector<double>&);
  virtual bool setCoordinate(int coord, 
                             const std::vector<double>&,
                             const std::vector<double>&,
                             const std::vector<double>&);
public: //Slash::Store::IObject
  virtual Slash::Store::IObject* copy() const;
  virtual std::string storeClassName() const;
public: //IVisited
  virtual bool visit(Slash::Store::IConstVisitor&) const;
  virtual bool read(Slash::Store::IVisitor&);
public: //IArray
  virtual Slash::Store::IArray::Orders orders() const;
  virtual bool write(char*&,char*) const;
  virtual bool read(const Slash::Store::IArray::Orders&,char*&,char*);
public:
  DataPointSet(Slash::Core::IWriter&,const std::string& = "");
  DataPointSet(Slash::Core::IWriter&,
               const std::string&,
               const std::string&,
               int);
  DataPointSet(const DataPointSet&);
  virtual ~DataPointSet();
  Annotation& getAnnotation();
  void setAnnotation(const Annotation&);
  std::vector<DataPoint>& points(); //for the IO.
  bool setPoints(unsigned int,const double*); //for the IO.
  SLASH_STORE_HEADER
private:
  Annotation fAnnotation;
  std::string fTitle;
  int fDimension;
  std::vector<DataPoint> fPoints;
};

}

#endif
