#ifndef BatchLab_MemoryTree_h
#define BatchLab_MemoryTree_h

// Inheritance :
#include <BatchLab/Core/BaseTree.h>

#include <BatchLab/Core/Folder.h>

namespace BatchLab {

// ITree should come first due to BatchLab::castTree function.

class MemoryTree : public BaseTree {
public: //Slash::Core::IManager
  virtual void* cast(const std::string&) const;
public: //AIDA/ITree
  virtual AIDA::IManagedObject* find(const std::string&);
  virtual bool cd(const std::string&);
  virtual bool mkdir(const std::string&);
  virtual bool mkdirs ( const std::string&);
  virtual bool commit();
  virtual std::string findPath(const AIDA::IManagedObject&) const;
  virtual std::string storeName() const;
  virtual bool rm(const std::string&);
  virtual std::string pwd() const;
  virtual bool ls(const std::string & = ".", bool = false,std::ostream& = std::cout) const;
  virtual bool rmdir ( const std::string&);
  virtual bool close();
  virtual std::vector<std::string> listObjectNames(const std::string& = ".", bool = false) const;
  virtual std::vector<std::string> listObjectTypes(const std::string& = ".", bool = false) const;
  virtual void setOverwrite(bool = true);
  virtual bool mv(const std::string&,const std::string&);
  virtual bool cp(const std::string&,const std::string&,bool = false );
  virtual bool symlink(const std::string&,const std::string&);
  virtual bool mount(const std::string&,AIDA::ITree&,const std::string&);
  virtual bool unmount(const std::string&);
  virtual AIDA::ITree* findTree(const std::string&);
  virtual std::string storeType() const; //AIDA-v3r3p0
  virtual bool isReadOnly() const; //AIDA-v3r3p0
public: //Slash::Store::ITree.
  virtual bool addObject(Slash::Store::IObject*,const std::string&);
  virtual bool isObjectValid(const Slash::Store::IObject*) const;
  virtual Slash::Store::IObject* findObject(const std::string&);
public: //Slash::Core::IHierarchy
  virtual std::string hierarchy(const std::string& = "") const; 
public: 
  virtual Slash::Store::IFolder* currentFolder() const;
  virtual void setCurrentFolder(Slash::Store::IFolder*);
  virtual Slash::Store::IFolder& rootFolder();
  virtual std::vector<AIDA::IManagedObject*> allObjects() const;
  virtual bool isADirectory(const std::string&) const;
public:
  MemoryTree(Slash::Core::ISession&,const std::string&);
  virtual ~MemoryTree();
protected:
  Slash::Store::IFolder* findFolder(const std::string&,bool = false) const;
  Slash::Store::IFolder* makeFolders(const std::string&); 
private:
  Folder fRoot;
  Slash::Store::IFolder* fCurrentFolder;
};

}

#endif
