// this :
#include <BatchLab/Core/AnalysisFactory.h>

#include <Lib/LibraryManager.h>
#include <Lib/Debug.h>
#include <Lib/FunctionManager.h>
#include <Lib/Cast.h>
#include <Lib/Out.h>
#include <Lib/Dictionary.h>

#include <Slash/Core/ILibrary.h>
#include <Slash/Tools/Managers.h>

#include <BatchLab/Core/TreeFactory.h>
#include <BatchLab/Core/TupleFactory.h>
#include <BatchLab/Core/HistogramFactory.h>
#include <BatchLab/Core/FunctionFactory.h>
#include <BatchLab/Core/DataPointSetFactory.h>
#include <BatchLab/Core/FitFactory.h>
#include <BatchLab/Core/HistogramAllocator.h>

#include <AIDA/ITree.h>

//FIXME : is it the right place ?
#include <BatchLab/Core/Histogram.h> //FIXME
#include <BatchLab/Core/Cloud.h>
#include <BatchLab/Core/DataPointSet.h>
#include <BatchLab/Core/Function.h>
#include <BatchLab/Core/MemoryTuple.h>

namespace BatchLab {
  template <class T>
  Slash::Meta::IObject* allocate(Slash::Core::IWriter& aWriter){
    return new T(aWriter);
  }
}

#define ADD_ALLOCATOR(aDataClass) \
 {typedef Slash::Meta::IObject* (*F)(Slash::Core::IWriter&);\
  F f = BatchLab::allocate<aDataClass>;\
  manager->add(new Lib::Class(fSession.printer(),#aDataClass,1,f));}

//////////////////////////////////////////////////////////////////////////////
BatchLab::AnalysisFactory::AnalysisFactory(
 Slash::Core::ISession& aSession
)
:fSession(aSession)
,fName("AnalysisFactory")
,fFunctionFactory(0)
,fTupleFactory(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("BatchLab::AnalysisFactory");

  if(!fSession.findManager("LibraryManager")) {
    fSession.addManager(new Lib::LibraryManager(fSession));
  }
  if(!fSession.findManager("FunctionManager")) {
    fSession.addManager(new Lib::FunctionManager(fSession));
  }
  if(!fSession.findManager("FunctionManager")) {
    fSession.addManager(new Lib::FunctionManager(fSession));
  }
  if(!fSession.findManager("BatchLab::HistogramAllocator")) {
    fSession.addManager(new BatchLab::HistogramAllocator(fSession.printer()));
  }
  // For the IO :
  if(!fSession.findManager("Lib::Dictionary")) {
    Lib::Dictionary* manager = new Lib::Dictionary(fSession.printer());
    //FIXME : is it the right place ?
    ADD_ALLOCATOR(BatchLab::Histogram1D);
    ADD_ALLOCATOR(BatchLab::Histogram2D);
    ADD_ALLOCATOR(BatchLab::Histogram3D);
    ADD_ALLOCATOR(BatchLab::Profile1D);
    ADD_ALLOCATOR(BatchLab::Profile2D);
    ADD_ALLOCATOR(BatchLab::Cloud1D);
    ADD_ALLOCATOR(BatchLab::Cloud2D);
    ADD_ALLOCATOR(BatchLab::Cloud3D);
    ADD_ALLOCATOR(BatchLab::DataPointSet);

    ADD_ALLOCATOR(BatchLab::MemoryTuple);

    ADD_ALLOCATOR(BatchLab::Function::Gaussian);
    ADD_ALLOCATOR(BatchLab::Function::BreitWigner);
    ADD_ALLOCATOR(BatchLab::Function::ExpHBOOK);
    ADD_ALLOCATOR(BatchLab::Function::ExpAIDA);
    ADD_ALLOCATOR(BatchLab::Function::Polynomial);
    ADD_ALLOCATOR(BatchLab::Function::Scripted);
    ADD_ALLOCATOR(BatchLab::Function::Composite);
    ADD_ALLOCATOR(BatchLab::Function::Compiled);

    fSession.addManager(manager);
  }
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::AnalysisFactory::~AnalysisFactory(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  delete fFunctionFactory;
  delete fTupleFactory;
  Lib::Debug::decrement("BatchLab::AnalysisFactory");
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

std::string BatchLab::AnalysisFactory::name() const {return fName;}

void* BatchLab::AnalysisFactory::cast(const std::string& aClass) const {
  // inheritance :
  if_Lib_SCast(AIDA::IAnalysisFactory)
  else if_Lib_SCast(Slash::AIDA::IAnalysisFactory)
  else if_Lib_SCast(Slash::Core::IManager)
  // this :
  else if_Lib_SCast(BatchLab::AnalysisFactory)
  else return 0;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::ITreeFactory* BatchLab::AnalysisFactory::createTreeFactory(
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
 const std::string& //aOptions
#endif
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return new TreeFactory(fSession);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ITupleFactory* BatchLab::AnalysisFactory::createTupleFactory(
 AIDA::ITree& aTree
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aOptions
#endif
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Store::ITree* baseTree = Lib_Cast(aTree,Slash::Store::ITree);
  if(!baseTree) {
    Lib::Out out(fSession.printer());
    out << "BatchLab::AnalysisFactory::createTupleFactory :"
        << " received tree is not a Slash::Store::ITree." << Lib::endl;
    return 0;
  }
  return new BatchLab::TupleFactory(fSession,*baseTree);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IHistogramFactory* BatchLab::AnalysisFactory::createHistogramFactory(
 AIDA::ITree& aTree
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aOptions
#endif
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Store::ITree* baseTree = Lib_Cast(aTree,Slash::Store::ITree);
  if(!baseTree) {
    Lib::Out out(fSession.printer());
    out << "BatchLab::AnalysisFactory::createHistogramFactory :"
        << " received tree is not a Slash::Store::ITree." << Lib::endl;
    return 0;
  }
  return new BatchLab::HistogramFactory(fSession,*baseTree);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFunctionFactory* BatchLab::AnalysisFactory::createFunctionFactory(
 AIDA::ITree& aTree
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aOptions
#endif
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Store::ITree* baseTree = Lib_Cast(aTree,Slash::Store::ITree);
  if(!baseTree) {
    Lib::Out out(fSession.printer());
    out << "BatchLab::AnalysisFactory::createFunctionFactory :"
        << " received tree is not a Slash::Store::ITree." << Lib::endl;
    return 0;
  }
  return new BatchLab::FunctionFactory(fSession,*baseTree);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IPlotterFactory* BatchLab::AnalysisFactory::createPlotterFactory(
 int aArgc
,char** aArgv
,const std::string& aImplementation
,const std::string& aLibrary
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aOptions
#endif
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if( (aImplementation.size()) && (aLibrary.size()) ) {
    // Get a plotter factory from another implementation :
    std::string entry = aImplementation + "_createPlotterFactory";

    Slash::Core::ILibraryManager* libraryManager = 
      Slash::libraryManager(fSession);
    if(!libraryManager) {
      Lib::Out out(fSession.printer());
      out << "BatchLab::AnalysisFactory::createPlotterFactory :"
          << " LibraryManager not found." << Lib::endl;
      return 0;
    }
    Slash::Core::ILibrary* dll = 
      libraryManager->addLibrary(aLibrary,aLibrary,"");
    if(!dll) {
      Lib::Out out(fSession.printer());
      out << "BatchLab::AnalysisFactory::createPlotterFactory :"
          << " unable to add a DLL" << Lib::endl;
      return 0;
    }
    typedef AIDA::IPlotterFactory*(*GetFactoryProc)(int,char**);
    GetFactoryProc p = (GetFactoryProc)dll->findProcedure(entry);
    if(!p) {
      Lib::Out out(fSession.printer());
      out << "BatchLab::AnalysisFactory::createPlotterFactory :"
          << " cannot get the entry point \"" << entry << "\"" 
          << "in \"" << aLibrary << "\"." << Lib::endl;
      return 0;
    }
    return p(aArgc,aArgv);
  } else {
    return 0;
  }
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IDataPointSetFactory* BatchLab::AnalysisFactory::createDataPointSetFactory(
 AIDA::ITree& aTree
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,const std::string& //aOptions
#endif
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Store::ITree* baseTree = Lib_Cast(aTree,Slash::Store::ITree);
  if(!baseTree) {
    Lib::Out out(fSession.printer());
    out << "BatchLab::AnalysisFactory::createDataPointSetFactory :"
        << " received tree is not a Slash::Store::ITree." << Lib::endl;
    return 0;
  }
  return new BatchLab::DataPointSetFactory(fSession.printer(),*baseTree);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFitFactory* BatchLab::AnalysisFactory::createFitFactory(
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
 const std::string& //aOptions
#endif
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Store::ITree* memoryTree = Slash::memoryTree(fSession);
  if(!memoryTree) {
    TreeFactory tf(fSession);
    tf.create();
    memoryTree = Slash::memoryTree(fSession);
  }
  if(!memoryTree) {
    Lib::Out out(fSession.printer());
    out << "BatchLab::AnalysisFactory::createFitFactory :"
        << " MemoryTree not found." 
        << Lib::endl;
    return 0;
  }
  if(!fFunctionFactory) {
    fFunctionFactory = new BatchLab::FunctionFactory(fSession,*memoryTree);
  }
  if(!fFunctionFactory) {
    Lib::Out out(fSession.printer());
    out << "BatchLab::AnalysisFactory::createFitFactory :"
        << " can't get a FunctionFactory." 
        << Lib::endl;
    return 0;
  }
  if(!fTupleFactory) {
    fTupleFactory = new BatchLab::TupleFactory(fSession,*memoryTree);
  }
  if(!fTupleFactory) {
    Lib::Out out(fSession.printer());
    out << "BatchLab::AnalysisFactory::createFitFactory :"
        << " can't get a TupleFactory." 
        << Lib::endl;
    return 0;
  }
  return new BatchLab::FitFactory
               (fSession.printer(),*fFunctionFactory,*fTupleFactory);
}
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
AIDA::IGenericFactory* BatchLab::AnalysisFactory::createGenericFactory(
 const std::string& //aFactoryType
,const std::string& //aOptions
)
{
  return 0;
}
AIDA::IGenericFactory* BatchLab::AnalysisFactory::createManagedObjectGenericFactory(
 const std::string& //factoryType
,AIDA::ITree& //tree
,const std::string& //options
)
{
  return 0;
}
#endif
