// this :
#include <BatchLab/Core/AsciiTree.h>

#include <AIDA/ITuple.h>

#include <Slash/Core/ISession.h>
#include <Slash/Store/IObject.h>
#include <Slash/Store/IFolderIterator.h>
#include <Slash/Store/IFolderVisitor.h>

#include <Slash/Tools/smanip.h>
#include <Slash/Tools/file.h>

#include <amanip/tuple.h>

#include <Lib/fmanip.h>
#include <Lib/FileWriter.h>
#include <Lib/Cast.h>
#include <Lib/StreamBuffer.h>
#include <iostream>

#include <BatchLab/Core/AsciiTuple.h> //FIXME

BatchLab::AsciiTree::AsciiTree(
 Slash::Core::ISession& aSession
,const std::string& aName
,const std::string& aStoreName
,bool aReadOnly
,bool aCreateNew
,const std::string& //aOptions
)
:MemoryTree(aSession,aName)
,fStoreName(aStoreName)
,fReadOnly(aReadOnly)
,fCreateNew(aCreateNew)
,fHippoStyle(false)
{
  if(fCreateNew) {
    // Handled in the commit.
  } else {
    // Read the existing file :
    if(verboseLevel()) {
      std::ostream& out = session().cout();
      out << "BatchLab::AsciiTree::AsciiTree :"
          << " open " << Slash::sout(aStoreName) 
          << "..." << std::endl;
    }

    std::string path,name,suffix;
    if(!Slash::path_name_suffix(aStoreName,path,name,suffix)) return;
    std::string name_only = Slash::nosuffix(name,false);
    
    BatchLab::AsciiTuple* object = 
      new BatchLab::AsciiTuple(printer(),name_only,"",aStoreName);
  //if(!MemoryTree::mkdirs("/")) {
  //  delete object;
  //}
    if(!manageObject(object,"/")) {
      delete object;
    }

    if(verboseLevel()) {
      std::ostream& out = session().cout();
      out << "BatchLab::AsciiTree::AsciiTree :"
          << " load of " << Slash::sout(aStoreName) << " done."
          << std::endl;
    }

  }
}

BatchLab::AsciiTree::~AsciiTree(){}

void* BatchLab::AsciiTree::cast(const std::string& aClass) const {
  if_Lib_SCast(BatchLab::AsciiTree)
  else return MemoryTree::cast(aClass);
}

std::string BatchLab::AsciiTree::storeName() const {return fStoreName;}
std::string BatchLab::AsciiTree::storeType() const {return "BatchLab::AsciiTree";}
bool BatchLab::AsciiTree::isReadOnly() const {return fReadOnly;}
bool BatchLab::AsciiTree::createNew() const {return fCreateNew;}
void BatchLab::AsciiTree::setHippoStyle(bool aValue){fHippoStyle = aValue;}
bool BatchLab::AsciiTree::hippoStyle() const {return fHippoStyle;}

namespace BatchLab {
  class AsciiStoreVisitor : public virtual Slash::Store::IFolderVisitor {
  public:
    AsciiStoreVisitor(AsciiTree& aTree,
                      std::ostream& a_io,
                      std::ostream& a_out)
    :fTree(aTree),f_io(a_io),f_out(a_out),f_done(false) {}
  public:
    virtual bool visit(const Slash::Store::IFolder& aFolder,
                       const Slash::Store::IObject& aObject){
      if(fTree.verboseLevel()) {
        f_out << "BatchLab::AsciiStoreVisitor::visit :" 
              << " path : " << amanip::sout(aFolder.path())
              << " name : " << amanip::sout(aObject.name())
              << std::endl;
      }

      if(const AIDA::ITuple* object = Lib_ConstCast(aObject,AIDA::ITuple)) {

        if(f_done) {
          f_out << "BatchLab::AsciiStoreVisitor::visit :" 
                << " path : " << amanip::sout(aFolder.path())
                << " name : " << amanip::sout(aObject.name())
                << " object not written. A tuple had already been written."
                << std::endl;
          return false;
        }

        if(fTree.hippoStyle()) {
          if(!amanip::write(*(const_cast<AIDA::ITuple*>(object)),
                            f_io,true,true,
                            f_out)) return false;
        } else {
          if(!amanip::write(*(const_cast<AIDA::ITuple*>(object)),
                            f_io,false,false,
                            f_out)) return false;
        }

        f_done = true;

      } else {

        f_out << "BatchLab::AsciiStoreVisitor::visit :" 
              << " path : " << amanip::sout(aFolder.path())
              << " name : " << amanip::sout(aObject.name())
              << " object not written (only tuples are supported)."
              << std::endl;

      }
        
      return true;
    }
    virtual bool beginVisit(const Slash::Store::IFolder&,const Slash::Store::IFolder&){ return true;}
    virtual bool endVisit(const Slash::Store::IFolder&,const Slash::Store::IFolder&){ return true;}
    virtual bool isRecursive() const { return true;}
  private:
    AsciiTree& fTree;
    std::ostream& f_io;
    std::ostream& f_out;
    bool f_done;
};
}

bool BatchLab::AsciiTree::commit(){
  if(fReadOnly) return false;

  std::string fileName = name(); //!!!
  if(fCreateNew) {
    Slash::file::remove(fileName);
  } else {
    //Can't update a file !
    return false;
  }

  // The name may contain env variables.
  std::string fullName = Lib::fmanip::name(fileName);
  Slash::Core::IWriter* writer = new Lib::FileWriter(fullName,printer());
  if(!writer->enabled()) {
    delete writer;
    return false;
  }

  Lib::StreamBuffer io_sbuf(*writer);
  std::ostream io(&io_sbuf); //it is used to write strings only.
  
  Lib::StreamBuffer out_sbuf(printer());
  std::ostream out(&out_sbuf);

  AsciiStoreVisitor visitor(*this,io,out);
  rootFolder().visit(visitor);

  delete writer;

  return true;
}
