// this :
#include <BatchLab/Core/BasePlotterFactory.h>

#include <Slash/Tools/Managers.h>
#include <Slash/Core/ILibrary.h>

#include <Slash/UI/IPlotterManager.h>
#include <Slash/UI/IPlotter.h>

#include <Lib/Out.h>
#include <Lib/Debug.h>
#include <Lib/Cast.h>
#include <Lib/Manager.h>

#include <AIDA/IPlotter.h>

//////////////////////////////////////////////////////////////////////////////
BatchLab::BasePlotterFactory::BasePlotterFactory(
 Slash::Core::ISession& aSession
,const std::string& aImplementation
,const std::string& aLibrary
,int aArgc
,char** aArgv
)
:fSession(aSession)
,fImplementation(aImplementation)
,fLibrary(aLibrary)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("BatchLab::BasePlotterFactory");
  for(int index=0;index<aArgc;index++) fArgs.push_back(aArgv[index]);
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::BasePlotterFactory::~BasePlotterFactory(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::decrement("BatchLab::BasePlotterFactory");
}
AIDA::IMarkerStyle* BatchLab::BasePlotterFactory::createMarkerStyle() {return 0;}
AIDA::ITextStyle* BatchLab::BasePlotterFactory::createTextStyle() {return 0;}
AIDA::ILineStyle* BatchLab::BasePlotterFactory::createLineStyle() {return 0;}
AIDA::IFillStyle* BatchLab::BasePlotterFactory::createFillStyle() {return 0;}
AIDA::IDataStyle* BatchLab::BasePlotterFactory::createDataStyle() {return 0;}
AIDA::IAxisStyle* BatchLab::BasePlotterFactory::createAxisStyle() {return 0;}
AIDA::ITitleStyle* BatchLab::BasePlotterFactory::createTitleStyle() {return 0;}
AIDA::IPlotterStyle* BatchLab::BasePlotterFactory::createPlotterStyle() {return 0;}

//////////////////////////////////////////////////////////////////////////////
AIDA::IPlotter* BatchLab::BasePlotterFactory::loadPlotter(
 const std::string& aLibrary
,const std::string& aPath
,const std::string& aPackage
,const std::string& aPlotterManager
,const std::string& aName
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Core::ILibraryManager* libraryManager = 
    Slash::libraryManager(fSession);
  if(!libraryManager) {
    Lib::Out out(fSession.printer());
    out << "BatchLab::BasePlotterFactory::load :"
        << " LibraryManager not found." << Lib::endl;
    return 0;
  }
  Slash::Core::ILibrary* dll = 
    libraryManager->addLibrary(aLibrary,aPath,aPackage);
  if(!dll) {
    Lib::Out out(fSession.printer());
    out << "BatchLab::BasePlotterFactory::load :"
        << " unable to add a DLL" << Lib::endl;
    return 0;
  }
  Slash::UI::IPlotterManager* plotterManager = 
    Lib_findManager(fSession,aPlotterManager,Slash::UI::IPlotterManager);
  if(!plotterManager) {
    Lib::Out out(fSession.printer());
    out << "BatchLab::BasePlotterFactory::load :"
        << " unable to find " << aPlotterManager << " plotter manager." 
        << Lib::endl;
    return 0;
  }
  plotterManager->setArguments(fArgs);
  Slash::UI::IPlotter* plotter = plotterManager->create(aName);
  if(!plotter) return 0;
  return Lib_Cast(*plotter,AIDA::IPlotter);
}
