// this :
#include <BatchLab/Core/BaseTree.h>

#include <Slash/Core/IRelationVisitor.h>
#include <Slash/Core/IRelation.h>

#include <Slash/Tools/Managers.h>
#include <Slash/Tools/Relation.h>

#include <Lib/Out.h>
#include <Lib/fmanip.h>
#include <Lib/smanip.h>
#include <Lib/Cast.h>
#include <Lib/Debug.h>

#include <Slash/Store/IObject.h>

BatchLab::BaseTree::BaseTree(Slash::Core::ISession& aSession,const std::string& aName)
:fSession(aSession) //For Validers
,fName(aName)
,fVerboseLevel(0)
,fRelationManager(0)
{
  Lib::Debug::increment("BatchLab::BaseTree");
  //printf("debug : %lu %lu :%lu\n",this,(AIDA::ITree*)this,&fKey);

  fRelationManager = Slash::relationManager(fSession);

  //NOTE : Relation to UI done in TreeFactory to avoid a dependency to IUI.
  //       Can't emitUpdate here since the construction is not finished.
}

BatchLab::BaseTree::~BaseTree() {
  if(fVerboseLevel) {
    Lib::Out out(fSession.printer());
    out << "BatchLab::BaseTree::~Tree :"
        << " called for " << this << " " << fName 
        << Lib::endl;
  }

  // Notify UI.
  //emitUpdate(); //virtual !
  if(fRelationManager) {
    Slash::RV_First_Schedule visitor(this);
    fRelationManager->visit(visitor);
    // Remove relations concerning this tree :
    fRelationManager->removeWith(static_cast<AIDA::ITree*>(this));

    //NOTE : if, when visiting, the relation does a schedule
    //       that indirectly uses this object with some method
    //       of a super class, we are dead. Then we should do a :
    //         RV_Or_Remove visitor(this,this);
    //         fRelationManager->remove(visitor,true);
    //       as in ~BaseObject.
  }

  fSession.removeManager(this);  //To support the AIDA delete tree.
  Lib::Debug::decrement("BatchLab::BaseTree");
}

std::string BatchLab::BaseTree::name() const {return fName;}

void* BatchLab::BaseTree::cast(const std::string& aClass) const {
  if_Lib_SCast(AIDA::ITree)
  else if_Lib_SCast(Slash::Store::ITree) 
  else if_Lib_SCast(BatchLab::BaseTree) 
  else if_Lib_SCast(Slash::Core::IManager)
  else if_Lib_SCast(Slash::Core::IHierarchy)
  else return 0;
}

Slash::Core::ISession& BatchLab::BaseTree::session() const {return fSession;}

Slash::Core::IWriter& BatchLab::BaseTree::printer() const {return fSession.printer();}

void BatchLab::BaseTree::setVerboseLevel(int aVerboseLevel) {fVerboseLevel = aVerboseLevel;}

int BatchLab::BaseTree::verboseLevel() const {return fVerboseLevel;}

void BatchLab::BaseTree::emitUpdate() {
  // To update UI.
  if(!fRelationManager) return;
  Slash::RV_First_Schedule visitor(this);
  fRelationManager->visit(visitor);
}

bool BatchLab::BaseTree::manageObject(Slash::Store::IObject* aObject,const std::string& aPath) {
  return addObject(aObject,aPath);
}


