// this :
#include <BatchLab/Core/DataPoint.h>

#include <Lib/Debug.h>

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
BatchLab::DataPoint::DataPoint(
 int aDimension
)
:fDimension(aDimension)
,fMeasurements(aDimension)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("BatchLab::DataPoint");
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::DataPoint::DataPoint(
 const DataPoint& aPoint
)
:AIDA::IDataPoint(aPoint)
,fDimension(aPoint.dimension())
,fMeasurements(aPoint.dimension())
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("BatchLab::DataPoint");
  initialize(aPoint);
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::DataPoint& BatchLab::DataPoint::operator=(
 const DataPoint& aPoint
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fDimension = aPoint.dimension();
  fMeasurements.resize(aPoint.dimension());
  initialize(aPoint);
  return *this;
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::DataPoint::DataPoint(
 const AIDA::IDataPoint& aPoint
)
:fDimension(aPoint.dimension())
,fMeasurements(aPoint.dimension())
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("BatchLab::DataPoint");
  initialize(aPoint);
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::DataPoint::~DataPoint(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::decrement("BatchLab::DataPoint");
}
//////////////////////////////////////////////////////////////////////////////
void BatchLab::DataPoint::initialize(
 const AIDA::IDataPoint& aPoint
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  for(int index=0;index<fDimension;index++) {
    const AIDA::IMeasurement* measurement = aPoint.coordinate(index);
    if(measurement) {
      fMeasurements[index] = Measurement(measurement->value(),
                                         measurement->errorPlus(),
                                         measurement->errorMinus());
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
int BatchLab::DataPoint::dimension(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fDimension;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IMeasurement* BatchLab::DataPoint::coordinate(
 int aCoordinate
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aCoordinate<0) return 0;
  if(aCoordinate>=fDimension) return 0;
  return &(fMeasurements[aCoordinate]);
}
//////////////////////////////////////////////////////////////////////////////
const AIDA::IMeasurement* BatchLab::DataPoint::coordinate(
 int aCoordinate
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aCoordinate<0) return 0;
  if(aCoordinate>=fDimension) return 0;
  return &(fMeasurements[aCoordinate]);
}
