// this :
#include <BatchLab/Core/MemoryTree.h>

#include <AIDA/IManagedObject.h>
#include <AIDA/IConstants.h>

#include <Slash/Core/ISession.h>
#include <Slash/Store/IObject.h>
#include <Slash/Store/IFolderIterator.h>
#include <Slash/Store/IFolderVisitor.h>

#include <Slash/Data/IArray.h>
#include <Slash/Tools/cast.h>

#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#include <amanip/mo.h>
#endif

#include <Lib/Out.h>
#include <Lib/smanip.h>
#include <Lib/sout.h>
#include <Lib/Debug.h>
#include <Lib/Cast.h>

#include <BatchLab/Core/Folder.h>
#include <BatchLab/Core/LinkFolder.h>

#ifdef WIN32
// disable the warning about the usage of "this" in the constructor.
#pragma warning(disable:4355)
#endif

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
BatchLab::MemoryTree::MemoryTree(
 Slash::Core::ISession& aSession
,const std::string& aName
)
:BaseTree(aSession,aName)
,fRoot(aSession.printer(),0,"",*this)
,fCurrentFolder(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("BatchLab::MemoryTree");
  fCurrentFolder = &fRoot;
  //Note : for UI notification; the below is useless since the 
  //       UI relation is not already set !
  emitUpdate();
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::MemoryTree::~MemoryTree(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //Lib::Out out(fPrinter);
  //out << "BatchLab::MemoryTree::~Tree : called for " << this << " " << fName << Lib::endl;
  Lib::Debug::decrement("BatchLab::MemoryTree");
}
//////////////////////////////////////////////////////////////////////////////
void* BatchLab::MemoryTree::cast(
 const std::string& aClass
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if_Lib_SCast(BatchLab::MemoryTree)
  else return BaseTree::cast(aClass);
}
//////////////////////////////////////////////////////////////////////////////
Slash::Store::IObject* BatchLab::MemoryTree::findObject(
 const std::string& aPath
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aPath.empty()) return 0;
  std::string path,name,suffix;
  Lib::smanip::pathNameSuffix(aPath,path,name,suffix);
  Slash::Store::IFolder* folder = findFolder(path);
  if(!folder) return 0;
  return folder->findObject(name);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IManagedObject* BatchLab::MemoryTree::find(
 const std::string& aPath
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Store::IObject* baseObject = findObject(aPath);
  if(!baseObject) return 0;
  return Lib_Cast(*baseObject,AIDA::IManagedObject);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::mkdir(
 const std::string& aPath
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aPath.empty()) return false;
  std::string path,name,suffix;
  Lib::smanip::pathNameSuffix(aPath,path,name,suffix);
  //printf("debug : mkdir \"%s\" : \"%s\" \"%s\" \"%s\"\n",
    //aPath.c_str(),path.c_str(),name.c_str(),suffix.c_str());
  // name is the last folder to create.
  Slash::Store::IFolder* folder = findFolder(path);
  if(!folder) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::mkdir :" 
        << " can't find folder " << Lib::sout(path) << "." 
        << Lib::endl;
    return false;
  }
  if(!folder->mkdir(name)) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::mkdir :" 
        << " can't create folder " << Lib::sout(name) 
        << " in folder " << Lib::sout(folder->path()) << "." 
        << Lib::endl;
    return false;
  }
  if(verboseLevel()) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::mkdir :" 
        << " folder " << Lib::sout(name) 
        << " created in folder " << Lib::sout(folder->path()) << "." 
        << Lib::endl;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::mkdirs( 
 const std::string& aPath
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return (makeFolders(aPath) ? true : false);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::ls(
 const std::string& aPath
,bool aRecursive
,std::ostream&
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Store::IFolder* folder = findFolder(aPath);
  if(!folder) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::ls :" 
        << " can't find folder " << Lib::sout(aPath) << "." 
        << Lib::endl;
    return false;
  }
  //Lib::Out out(printer());
  //out << "MemoryTree (" << name() << ") : " << Lib::endl;
  folder->ls(aRecursive);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::commit(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return true;
}
namespace BatchLab {
  class FindFolder : public virtual Slash::Store::IFolderVisitor {
  public:
    virtual bool visit(const Slash::Store::IFolder& aFolder,
                       const Slash::Store::IObject& aObject){
      if(&aObject!=fObject) return true;
      fFolder = (Slash::Store::IFolder*)&aFolder; 
      return false; //Stop visit.
    }
    virtual bool beginVisit(const Slash::Store::IFolder&,const Slash::Store::IFolder&){ return true;}
    virtual bool endVisit(const Slash::Store::IFolder&,const Slash::Store::IFolder&){ return true;}
    virtual bool isRecursive() const { return true;}
  public:
    const Slash::Store::IObject* fObject;
    Slash::Store::IFolder* fFolder;
  };
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::MemoryTree::findPath(
 const AIDA::IManagedObject& aObject
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const Slash::Store::IObject* baseObject = 
    Lib_ConstCast(aObject,Slash::Store::IObject);
  if(!baseObject) return "";
  FindFolder visitor;
  visitor.fObject = baseObject;
  visitor.fFolder = 0;
  fRoot.visit(visitor);
  Slash::Store::IFolder* folder = visitor.fFolder; //Found folder.
  if(!folder) return "";
  std::string path = folder->linkingPath();
  if(path=="/") return path + aObject.name();
  return path + "/" + aObject.name();
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::MemoryTree::pwd(
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fCurrentFolder->linkingPath();
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::rmdir(
 const std::string& aPath
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aPath=="/") {
    fRoot.clear();
    fCurrentFolder = &fRoot;
    return true;
  }
  // Find folder :
  Slash::Store::IFolder* folder = findFolder(aPath);
  if(!folder) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::rmdir :" 
        << " can't find folder " << Lib::sout(aPath) << "." 
        << Lib::endl;
    return false;
  }
  //FIXME : case of linked.
  // Case the found folder is the root (of everything) :
  if(folder==&fRoot) {
    fRoot.clear();
    fCurrentFolder = &fRoot;
    return true;
  }
  // Take care of the current folder :
  if(folder==fCurrentFolder) {
    fCurrentFolder = folder->parent();
  }
  // Then remove the folder :
  folder->clear();
  Slash::Store::IFolder* parent = folder->parent();
  parent->remove(folder);
  delete folder;
  if(verboseLevel()) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::rmdir :" 
        << " folder " << Lib::sout(aPath) << " removed."
        << Lib::endl;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::cd(
 const std::string& aPath
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Store::IFolder* folder = findFolder(aPath);
  if(!folder) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::cd :" 
        << " can't find folder " << Lib::sout(aPath) << "." 
        << Lib::endl;
    return false;
  }
  fCurrentFolder = folder;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::isADirectory(
 const std::string& aPath
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return findFolder(aPath)?true:false;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::addObject(
 Slash::Store::IObject* aObject
,const std::string& aPath
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aObject->name().empty()) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::addObject :"
        << " WARNING : add object with empty name." << Lib::endl;
  }
  // Find folder :
  Slash::Store::IFolder* folder = findFolder(aPath);
  if(!folder) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::addObject :" 
        << " can't find folder " << Lib::sout(aPath) << "." 
        << Lib::endl;
    return false;
  }
  folder->add(aObject); //will do a fTree.emitUpdate.
  aObject->setFolder(folder);
  if(verboseLevel()) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::addObject :" 
        << " " << Lib::sout(aObject->name())
        << " in folder " << Lib::sout(folder->path()) << "." 
        << Lib::endl;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::rm(
 const std::string& aPath
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aPath.empty()) return false;
  std::string path,name,suffix;
  Lib::smanip::pathNameSuffix(aPath,path,name,suffix);
  // Find folder :
  Slash::Store::IFolder* folder = findFolder(path);
  if(!folder) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::rm :" 
        << " can't find folder " << Lib::sout(path) << "." 
        << Lib::endl;
    return false;
  }
  Slash::Store::IObject* object = folder->findObject(name);
  if(!object) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::rm :" 
        << " can't find object " << Lib::sout(name)
        << " in folder " << Lib::sout(folder->path()) << "." 
        << Lib::endl;
    return false;
  }
  if(verboseLevel()) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::rm :" 
        << " " << Lib::sout(object->name())
        << " in folder " << Lib::sout(folder->path()) << "..."
        << Lib::endl;
  }
  //folder->remove(object);
  delete object; //It remove the object from the folder (See ~BaseObject)
  emitUpdate();
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::isObjectValid(
 const Slash::Store::IObject* aObject
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fRoot.isValid(aObject);
}
//////////////////////////////////////////////////////////////////////////////
Slash::Store::IFolder* BatchLab::MemoryTree::currentFolder(
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fCurrentFolder;
}
//////////////////////////////////////////////////////////////////////////////
void BatchLab::MemoryTree::setCurrentFolder(
 Slash::Store::IFolder* aFolder
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fCurrentFolder = aFolder;
}
//////////////////////////////////////////////////////////////////////////////
Slash::Store::IFolder& BatchLab::MemoryTree::rootFolder(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fRoot;
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::MemoryTree::hierarchy(
 const std::string&
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string sout;
  sout += "<tree>";
  sout += fRoot.hierarchy("all");
  sout += "</tree>";
  return sout;
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::MemoryTree::storeName(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return "";
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::MemoryTree::storeType(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return "BatchLab::MemoryTree";
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::isReadOnly(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return false;
}
namespace BatchLab {
  class AllObjectsVisitor : public virtual Slash::Store::IFolderVisitor {
  public:
    AllObjectsVisitor(std::vector<AIDA::IManagedObject*>& aVector)
      :fVector(aVector) {}
  public:
    virtual bool visit(const Slash::Store::IFolder&,const Slash::Store::IObject& aObject){
      const AIDA::IManagedObject* aobject = 
        Lib_ConstCast(aObject,AIDA::IManagedObject);
      if(!aobject) return true;
      fVector.push_back(const_cast<AIDA::IManagedObject*>(aobject));
      return true;
    }
    virtual bool beginVisit(const Slash::Store::IFolder&,const Slash::Store::IFolder&){ return true;}
    virtual bool endVisit(const Slash::Store::IFolder&,const Slash::Store::IFolder&){ return true;}
    virtual bool isRecursive() const { return true;}
  private:
    std::vector<AIDA::IManagedObject*>& fVector;
  };
}
//////////////////////////////////////////////////////////////////////////////
std::vector<AIDA::IManagedObject*> BatchLab::MemoryTree::allObjects(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<AIDA::IManagedObject*> vec;
  AllObjectsVisitor visitor(vec);
  fRoot.visit(visitor);
  return vec;
}
//////////////////////////////////////////////////////////////////////////////
Slash::Store::IFolder* BatchLab::MemoryTree::makeFolders( 
 const std::string& aPath
) 
//////////////////////////////////////////////////////////////////////////////
// Leaf in the path is returned.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Store::IFolder* start = 0;
  std::string path;  
  if(aPath.size() && (aPath[0]=='/')) {
    start = (Slash::Store::IFolder*)&fRoot;
    path = aPath.substr(1,aPath.size()-1);
  } else {
    start = fCurrentFolder;
    path = aPath;
  }
  std::vector<std::string> words;
  Lib::smanip::words(path,"/",words);
  Slash::Store::IFolder* folder = start;
  for(unsigned int index=0;index<words.size();index++) {
    if(words[index]==".") {
    } else if(words[index]=="..") {
      folder = folder->parent();
      if(!folder) {
        Lib::Out out(printer());
        out << "BatchLab::MemoryTree::makeFolders :" 
            << " input path " << Lib::sout(aPath) << " :"
            << " case .. ; folder without parent."
            << Lib::endl;
        return 0;
      }
    } else {
      Slash::Store::IFolder* f = folder->findFolder(words[index]);
      if(f) {
        folder = f;
      } else {
        folder = folder->mkdir(words[index]);
        if(!folder) {
          Lib::Out out(printer());
          out << "BatchLab::MemoryTree::makeFolders :" 
              << " input path " << Lib::sout(aPath) << " :"
              << " can't mkdir " << Lib::sout(words[index])
              << Lib::endl;
          return 0;
        }
      }
    }
  }
  return folder;
}
//////////////////////////////////////////////////////////////////////////////
Slash::Store::IFolder* BatchLab::MemoryTree::findFolder(
 const std::string& aPath
,bool aVerbose
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Store::IFolder* start = 0;
  std::string path;  
  if(aPath.size() && (aPath[0]=='/')) {
    start = (Slash::Store::IFolder*)&fRoot;
    path = aPath.substr(1,aPath.size()-1);
  } else {
    start = fCurrentFolder;
    path = aPath;
  }
  std::vector<std::string> words;
  Lib::smanip::words(path,"/",words);
  Slash::Store::IFolder* folder = start;
  for(unsigned int index=0;index<words.size();index++) {
    if(words[index]==".") {
    } else if(words[index]=="..") {
      if(Slash::Store::IFolder* p = folder->linkingFolder()) {
        if(&(p->tree())!=&(folder->tree())) { //Mounting.
          folder = p->parent();
        } else { //Symlink.
          folder = folder->parent();
        }
      } else {
        folder = folder->parent();
      }
      if(!folder) {
        if(aVerbose) {
          Lib::Out out(printer());
          out << "BatchLab::MemoryTree::findFolder :" 
              << " input path " << Lib::sout(aPath) << " :"
              << " case .. ; folder without parent."
              << Lib::endl;
        }
        return 0;
      }
    } else {
      folder = folder->findFolder(words[index]);
      if(!folder) {
        if(aVerbose) {
          Lib::Out out(printer());
          out << "BatchLab::MemoryTree::findFolder :" 
              << " input path " << Lib::sout(aPath) << " :"
              << " can't findFolder " << Lib::sout(words[index])
              << Lib::endl;
        }
        return 0;
      }
      const BatchLab::LinkFolder* link = 
        Lib_ConstCast(*folder,BatchLab::LinkFolder);
      if(link) {
        folder = link->linkedFolder();
        if(!folder) {
          if(aVerbose) {
            Lib::Out out(printer());
            out << "BatchLab::MemoryTree::findFolder :" 
                << " input path " << Lib::sout(aPath) << " :"
                << " can't find linkedFolder for " << Lib::sout(words[index])
                << Lib::endl;
          }
          return 0;
        }
      }
    }
  }
  return folder;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::close(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
   return true;
}

#include <AIDA/ITuple.h> //FIXME
#include <BatchLab/Core/TupleFactory.h> //FIXME

static bool copy_object(Slash::Core::ISession& aSession,const Slash::Store::IObject& aObject,Slash::Store::IFolder& aFolder) {
  // target tree.
  Slash::Store::ITree& baseTree = aFolder.tree();
  if(const AIDA::ITuple* tuple = Lib_ConstCast(aObject,AIDA::ITuple)) {
    BatchLab::TupleFactory tupleFactory(aSession,baseTree);
    if(!tupleFactory.createCopy(aObject.name(),*tuple)) {
      Lib::Out out(aSession.printer());
      out << "BatchLab::MemoryTree::copy_object :"
          << " can't create a tuple copy of " << Lib::sout(aObject.name()) 
          << Lib::endl;
      return false;
    }
  } else {
    Slash::Store::IObject* copy = aObject.copy();
    if(!copy) {
      Lib::Out out(aSession.printer());
      out << "BatchLab::MemoryTree::copy_object :"
          << " unable to copy " << Lib::sout(aObject.name()) 
          << Lib::endl;
      return false;
    } 
    if(!baseTree.manageObject(copy,aFolder.path())) {
      Lib::Out out(aSession.printer());
      out << "BatchLab::MemoryTree::copy_object :"
          << " unable to manage object " << Lib::sout(aObject.name()) 
          << Lib::endl;
      return false;
    }
  }
  if(aSession.verboseLevel()) {
    Lib::Out out(aSession.printer());
    out << "BatchLab::MemoryTree::copy_object :"
        << " copied " << Lib::sout(aObject.name())
        << " in folder " << Lib::sout(aFolder.name())
        << " of tree " << Lib::sout(baseTree.storeName())
        << Lib::endl;
  }
  return true;
}

namespace BatchLab {
  class CopyVisitor : public virtual Slash::Store::IFolderVisitor {
  public:
    CopyVisitor(Slash::Core::ISession& aSession,Slash::Core::IWriter& aPrinter,Slash::Store::IFolder& aFolder,int aVerboseLevel)
    :fSession(aSession),fPrinter(aPrinter),fFolder(aFolder),fStatus(true)
    ,fVerboseLevel(aVerboseLevel){}
    bool status() const { return fStatus;}
  public:
    virtual bool visit(const Slash::Store::IFolder& /*aFolder*/,
                       const Slash::Store::IObject& aObject){
      if(!copy_object(fSession,aObject,fFolder)) {
        fStatus = false;
        return false;
      }
      return true;
    }
    virtual bool beginVisit(const Slash::Store::IFolder&,const Slash::Store::IFolder& /*aSubFolder*/){ 
      return false; //Do not process subfolders.
    }
    virtual bool endVisit(const Slash::Store::IFolder&,const Slash::Store::IFolder&){ 
      return false;
    }
    virtual bool isRecursive() const { return true;}
  private:
    Slash::Core::ISession& fSession;
    Slash::Core::IWriter& fPrinter;
    Slash::Store::IFolder& fFolder; //Target folder.
    bool fStatus;
    int fVerboseLevel;
  };
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::cp(
 const std::string& aFrom
,const std::string& aTo
,bool //aRecursive
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aTo==aFrom) return true;
  // Is aFrom an object or a folder ?
  Slash::Store::IFolder* fromFolder = findFolder(aFrom);
  if(fromFolder) { //aFrom is a folder.
    Slash::Store::IFolder* toFolder = findFolder(aTo);
    if(!toFolder) {
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::cp :" 
          << " target path " << Lib::sout(aTo)
          << " is not a folder in tree " << Lib::sout(name()) << "."
          << Lib::endl;
      return false;
    }
    if(toFolder==fromFolder) return true;
    CopyVisitor visitor(session(),printer(),*toFolder,verboseLevel());
    if(verboseLevel()) {
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::cp :" 
          << " folders found. Start copying..."
          << Lib::endl;
    }
    fromFolder->visit(visitor);
    return visitor.status();
  } else { //is it an object ?
    Slash::Store::IObject* fromObject = findObject(aFrom);
    if(!fromObject) {
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::cp :" 
          << " from path " << Lib::sout(aFrom)
          << " not found in tree " << Lib::sout(name()) << "."
          << Lib::endl;
      return false;
    }
    Slash::Store::IFolder* toFolder = findFolder(aTo);
    if(!toFolder) {
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::cp :" 
          << " target path " << Lib::sout(aTo)
          << " is not a folder in tree " << Lib::sout(name()) << "."
          << Lib::endl;
      return false;
    }
    return copy_object(session(),*fromObject,*toFolder);
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::mount(
 const std::string& aLink
,AIDA::ITree& aTree
,const std::string& aPath
) 
//////////////////////////////////////////////////////////////////////////////
// aTree is mounted within "this".
// The aPath path within aTree is mounted at aLink within "this".
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aLink.empty()) return false;
  //FIXME : have to install a notification if aTree is deleted.
  
  std::string lpath,lname,suffix;
  Lib::smanip::pathNameSuffix(aLink,lpath,lname,suffix);
  // lname is the linking point.
  Slash::Store::IFolder* parent = findFolder(lpath);
  if(!parent) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::mount :" 
        << " can't find parent path " << Lib::sout(lpath)
        << " in mounting tree " << Lib::sout(name())
        << Lib::endl;
    return false;
  }

  BatchLab::MemoryTree* batchlabTree = Lib_Cast(aTree,BatchLab::MemoryTree);
  if(!batchlabTree) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::mount :" 
        << " tree is not a BatchLab::MemoryTree."
        << Lib::endl;
    return false;
  }

  //AIDA/TestTree::testMountAndUnmount : aPath=="" means "/" !
  Slash::Store::IFolder* tfolder = 
    batchlabTree->findFolder(aPath.empty()?"/":aPath);
  if(!tfolder) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::mount :" 
        << " can't find path to mount " << Lib::sout(aPath)
        << " in tree " << Lib::sout(aTree.storeName())
        << Lib::endl;
    return false;
  }

  bool status = parent->link(tfolder,lname);
  if(verboseLevel()) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::mount :" 
        <<  " mount of " << Lib::sout(aPath)
        <<  " on " << Lib::sout(aLink) 
        << (status?" is a success.":" is a failure.") 
        << Lib::endl;
  }
  return status;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::unmount(
 const std::string& aLink
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Store::IFolder* folder = findFolder(aLink);
  if(!folder) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::unmount :" 
        << " can't find path " << Lib::sout(aLink)
        << " in tree " << Lib::sout(name())
        << Lib::endl;
    return false;
  }
  folder = folder->linkingFolder();
  if(folder) {
    BatchLab::LinkFolder* link = Lib_Cast(*folder,BatchLab::LinkFolder);
    if(link) {
      if(link->isSymlink()) { //Symlink
      } else { //Mounting
        Slash::Store::IFolder* parent = folder->parent();
        if(!parent) return false;
        bool status = parent->unlink(folder);
        if(verboseLevel()) {
          Lib::Out out(printer());
          out << "BatchLab::MemoryTree::unmount :" 
              <<  " unmount of " << Lib::sout(aLink)
              << (status?" is a success.":" is a failure.") 
              << Lib::endl;
        }
        return status;
      }
    }
  }
  Lib::Out out(printer());
  out << "BatchLab::MemoryTree::unmount :" 
      << " folder " << Lib::sout(aLink)
      << " is not a mouting point."
      << Lib::endl;
  return false;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ITree* BatchLab::MemoryTree::findTree(
 const std::string& aPath
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Store::IFolder* folder = findFolder(aPath);
  if(!folder) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::findTree :" 
        << " path " << Lib::sout(aPath)
        << " in tree " << Lib::sout(name())
        << " not found."
        << Lib::endl;
    return 0;
  }
  AIDA::ITree* aidaTree = Lib_Cast(folder->tree(),AIDA::ITree);
  if(!aidaTree) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::findTree :" 
        << " folder tree not an AIDA::ITree."
        << Lib::endl;
  }
  return aidaTree;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::mv(
 const std::string& aWhat
,const std::string& aTo
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{

  std::string wpath,wname,wsuffix;
  Lib::smanip::pathNameSuffix(aWhat,wpath,wname,wsuffix);
  bool wobj = false;
  Slash::Store::IFolder* folder = findFolder(aWhat);
  if(!folder) {
    folder = findFolder(wpath);
    wobj = true;
  }
  if(!folder) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::mv :" 
        << " from path " << Lib::sout(aWhat)
        << " not found in tree " << Lib::sout(name()) << "."
        << Lib::endl;
    return false;
  }

  std::string topath,toname,tosuffix;
  Lib::smanip::pathNameSuffix(aTo,topath,toname,tosuffix);
  bool toobj = false;
  Slash::Store::IFolder* toFolder = findFolder(aTo);
  if(!toFolder) {
    toFolder = findFolder(topath);
    toobj = true;
  }
  if(!toFolder) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::mv :" 
        << " target path " << Lib::sout(aTo)
        << " not found in tree " << Lib::sout(name()) << "."
        << Lib::endl;
    return false;
  }


  if(!wobj && !toobj) {
    //Move aWhat directory into aTo directory :
    if(folder==toFolder) { //Can't be moved in itself.
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::mv :" 
          << " can't move a folder into itself."
          << Lib::endl;
      return false;
    }
    // Check that toFolder does not have folder as ancestor :
    Slash::Store::IFolder* parent = toFolder->parent();
    while(parent) {
      if(parent==folder) {
        Lib::Out out(printer());
        out << "BatchLab::MemoryTree::mv :" 
            << " Folder " << Lib::sout(aTo)
            << " has folder " << Lib::sout(aWhat)
            << " as ancestor."
            << Lib::endl;
        return false;
      }
      parent = parent->parent();
    }
    // Seems ok.
    parent = folder->parent();
    if(!parent) return false; //The root directory can't be moved.
    parent->remove(folder);
    toFolder->add(folder);
    folder->setParent(toFolder);
    if(verboseLevel()) {
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::mv :" 
            << " Folder " << Lib::sout(aWhat)
	  << " moved into folder " << Lib::sout(aTo) << "."
            << Lib::endl;
    }
    return true;

  } else if(wobj && !toobj) {
    // Move aWhat object into aTo directory :
    Slash::Store::IObject* object = folder->findObject(wname);
    if(!object) {
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::mv :" 
          << " Object " << Lib::sout(wname)
          << " not found in folder " << Lib::sout(wpath) << "."
          << Lib::endl;
      return false;
    }  
    folder->remove(object);
    toFolder->add(object);
    if(verboseLevel()) {
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::mv :" 
          << " Object " << Lib::sout(aWhat)
          << " moved into folder " << Lib::sout(aTo) << "."
          << Lib::endl;
    }
    return true;

  } else if(!wobj && toobj) {
    // Move aWhat directory into aTo directory :
    if(toFolder->findObject(toname)) {
      // Move aWhat directory into aTo object : no sense.
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::mv :" 
          << " Can't move " << Lib::sout(aWhat)
          << " directory into " << Lib::sout(aTo) << " object."
          << Lib::endl;
      return false;
    }
    //toname not an existing object or directory.
    if(folder==toFolder) { //Can't be moved in itself.
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::mv :" 
          << " can't move a folder into itself."
          << Lib::endl;
      return false;
    }
    // Check that toFolder does not have folder as ancestor :
    Slash::Store::IFolder* parent = toFolder->parent();
    while(parent) {
      if(parent==folder) {
        Lib::Out out(printer());
        out << "BatchLab::MemoryTree::mv :" 
            << " Folder " << Lib::sout(aTo)
            << " has folder " << Lib::sout(aWhat)
            << " as ancestor."
            << Lib::endl;
        return false;
      }
      parent = parent->parent();
    }
    // Seems ok.
    parent = folder->parent();
    if(!parent) return false; //The root directory can't be moved.
    parent->remove(folder);
    toFolder->add(folder);
    folder->setParent(toFolder);
    folder->setName(toname);
    if(verboseLevel()) {
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::mv :" 
            << " Folder " << Lib::sout(aWhat)
	  << " moved in folder " << Lib::sout(aTo) << "."
            << Lib::endl;
    }
    return true;

  } else if(wobj && toobj) {
    // Move aWhat object into aTo object (the renaming it) :
    Slash::Store::IObject* object = folder->findObject(wname);
    if(!object) {
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::mv :" 
          << " Object " << Lib::sout(wname)
          << " not found in folder " << Lib::sout(wpath) << "."
          << Lib::endl;
      return false;
    }  
    if(toFolder->findObject(toname)) {
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::mv :" 
          << " Object " << Lib::sout(aTo)
          << " already exists."
          << Lib::endl;
      return false;
    }  
    folder->remove(object);
    toFolder->add(object);
    object->setName(toname);
    if(verboseLevel()) {
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::mv :" 
          << " Object " << Lib::sout(aWhat)
          << " moved to " << Lib::sout(aTo) << "."
          << Lib::endl;
    }
    return true;
  }
  return false;
}

namespace BatchLab {
  class ListObjectsVisitor : public virtual Slash::Store::IFolderVisitor {
  public:
    ListObjectsVisitor
     (bool aRecursive
     ,const std::string& aPath
     ,std::vector<std::string>& aVector
     ,bool aNames)
     :fRecursive(aRecursive)
     ,fPath(aPath)
     ,fVector(aVector)
     ,fNames(aNames)
     ,fDone(false) {}
  public:
    virtual bool visit(const Slash::Store::IFolder& aFolder,
                       const Slash::Store::IObject& aObject){
      if(fNames) {
        if(fPath==".") {
          fVector.push_back(aObject.name());
        } else {
          fVector.push_back(aFolder.linkingPath()+"/"+aObject.name());
        }
      } else {
        const AIDA::IManagedObject* aobject = 
          Lib_ConstCast(aObject,AIDA::IManagedObject);
        if(!aobject) return true; //FIXME : warning.

        std::string sclass;
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
        if(!amanip::type(*aobject,sclass)) {
          if(Slash_safe_cast(*aobject,Slash::Data::IArray)) {
            sclass = "Slash::Data::IArray";;
          }
        }
#else
        sclass = aobject->type();
#endif
        Lib::smanip::replace(sclass,"AIDA::","");
        fVector.push_back(sclass);
      }
      return true;
    }
    virtual bool beginVisit(const Slash::Store::IFolder& aFolder,const Slash::Store::IFolder& aSubFolder){ 
      if(fNames) {
        if(fPath==".") {
          fVector.push_back(aSubFolder.name()+"/");
        } else {
          fVector.push_back(aFolder.linkingPath()+"/"+aSubFolder.name()+"/");
        }
      } else {
        fVector.push_back("dir");
      }
      return true;
    }
    virtual bool endVisit(const Slash::Store::IFolder&,const Slash::Store::IFolder&){ 
      return true;
    }
    virtual bool isRecursive() const { return fRecursive;}
  private:
    bool fRecursive;
    std::string fPath;
    std::vector<std::string>& fVector;
    bool fNames;
    bool fDone;
  };
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> BatchLab::MemoryTree::listObjectNames(
 const std::string& aPath
,bool aRecursive 
) const 
//////////////////////////////////////////////////////////////////////////////
// aPath could be a directory path or an object path
// (to pass test AIDA/TestTree::testListObjectNames_ListObjectTypes)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> vect;
  std::string path,name,suffix;
  Lib::smanip::pathNameSuffix(aPath,path,name,suffix);
  Slash::Store::IFolder* folder = findFolder(path);
  if(!folder) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::listObjectNames :" 
        << " path " << Lib::sout(path)
        << " not found in tree " << Lib::sout(storeName()) << "."
        << Lib::endl;
    return vect;
  }
  if(folder->findObject(name)) {
    vect.push_back(aPath);
    return vect;
  }
  if(name.size()) {
    folder = folder->findFolder(name);
    if(!folder) {
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::listObjectNames :" 
          << " folder " << Lib::sout(name)
          << " not found in directory " << Lib::sout(path)
          << " of tree " << Lib::sout(storeName()) << "."
          << Lib::endl;
      return vect;
    }
  }
  ListObjectsVisitor visitor(aRecursive,aPath,vect,true);
  folder->visit(visitor);
  return vect;
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> BatchLab::MemoryTree::listObjectTypes(
 const std::string& aPath
,bool aRecursive
) const 
//////////////////////////////////////////////////////////////////////////////
// aPath could be a directory path or an object path
// (to pass test AIDA/TestTree::testListObjectNames_ListObjectTypes)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> vect;
  std::string path,name,suffix;
  Lib::smanip::pathNameSuffix(aPath,path,name,suffix);
  Slash::Store::IFolder* folder = findFolder(path);
  if(!folder) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTree::listObjectTypes :" 
        << " path " << Lib::sout(path)
        << " not found in tree " << Lib::sout(storeName()) << "."
        << Lib::endl;
    return vect;
  }
  Slash::Store::IObject* baseObject = folder->findObject(name);
  if(baseObject) {
    AIDA::IManagedObject* aobject = 
      Lib_Cast(*baseObject,AIDA::IManagedObject);
    if(aobject) {
      std::string sclass;
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
      if(!amanip::type(*aobject,sclass)) {
        if(Slash_safe_cast(*aobject,Slash::Data::IArray)) {
          sclass = "Slash::Data::IArray";;
        }
      }
#else
      sclass = aobject->type();
#endif
      Lib::smanip::replace(sclass,"AIDA::","");
      vect.push_back(sclass);
      return vect;
    }
  }
  if(name.size()) {
    folder = folder->findFolder(name);
    if(!folder) {
      Lib::Out out(printer());
      out << "BatchLab::MemoryTree::listObjectTypes:" 
          << " folder " << Lib::sout(name)
          << " not found in directory " << Lib::sout(path)
          << " of tree " << Lib::sout(storeName()) << "."
          << Lib::endl;
      return vect;
    }
  }
  ListObjectsVisitor visitor(aRecursive,aPath,vect,false);
  folder->visit(visitor);
  return vect;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTree::symlink(
 const std::string& aPath
,const std::string& aLink
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return mount(aLink,*this,aPath);
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void BatchLab::MemoryTree::setOverwrite(
 bool
)
{
}
