// this :
#include <BatchLab/Core/TreeFactory.h>

#include <Slash/Tools/Managers.h>
#include <Slash/Store/IStoreManager.h>

#include <Lib/Out.h>
#include <Lib/Debug.h>
#include <Lib/smanip.h>
#include <Lib/sout.h>
#include <Lib/Cast.h>
#include <Lib/Manager.h>

#include <BatchLab/Core/MemoryTree.h>
#include <BatchLab/Core/AsciiTree.h>

BatchLab::TreeFactory::TreeFactory(Slash::Core::ISession& aSession)
:fSession(aSession)
,fPrinter(aSession.printer())
{
  Lib::Debug::increment("BatchLab::TreeFactory");
}

BatchLab::TreeFactory::~TreeFactory(){
  Lib::Debug::decrement("BatchLab::TreeFactory");
}

AIDA::ITree* BatchLab::TreeFactory::createTree(const std::string&){
  AIDA::ITree* memoryTree = Lib_findManager(fSession,"MemoryTree",AIDA::ITree);
  if(memoryTree) return memoryTree;
  MemoryTree* tree = new MemoryTree(fSession,"MemoryTree");
  tree->setVerboseLevel(fSession.verboseLevel());
  fSession.addManager(tree);
  return tree;
}

AIDA::ITree* BatchLab::TreeFactory::create(){return BatchLab::TreeFactory::createTree("");}

//////////////////////////////////////////////////////////////////////////////
AIDA::ITree* BatchLab::TreeFactory::create(
 const std::string& aStoreName
,const std::string& aStoreType
,bool aReadOnly
,bool aCreateNew
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return create_tree(aStoreName,
                     aStoreName,aStoreType,aReadOnly,aCreateNew,aOptions);
}
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
//////////////////////////////////////////////////////////////////////////////
AIDA::ITree* BatchLab::TreeFactory::createTree(
 const std::string& aStoreName
,const std::string& aStoreType
,int aMode
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return BatchLab::TreeFactory::createNamedTree
           (aStoreName,aStoreName,aStoreType,aMode,aOptions);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ITree* BatchLab::TreeFactory::createNamedTree(
 const std::string& aName
,const std::string& aStoreName
,const std::string& aStoreType
,int aMode
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aMode==AIDA::ITreeFactory::AUTO) {
    bool readOnly = false;
    bool createNew = true;
    return 
      create_tree(aName,aStoreName,aStoreType,readOnly,createNew,aOptions);
  } else if(aMode==AIDA::ITreeFactory::CREATE) {
    bool readOnly = false;
    bool createNew = true;
    return 
      create_tree(aName,aStoreName,aStoreType,readOnly,createNew,aOptions);
  } else if(aMode==AIDA::ITreeFactory::RECREATE) {
    bool readOnly = false;
    bool createNew = true;
    return 
      create_tree(aName,aStoreName,aStoreType,readOnly,createNew,aOptions);
  } else if(aMode==AIDA::ITreeFactory::READONLY) {
    bool readOnly = true;
    bool createNew = false;
    return 
      create_tree(aName,aStoreName,aStoreType,readOnly,createNew,aOptions);
  } else if(aMode==AIDA::ITreeFactory::UPDATE) {
    bool readOnly = false;
    bool createNew = false;
    return 
      create_tree(aName,aStoreName,aStoreType,readOnly,createNew,aOptions);
  } else {
    return 0;
  }
}
#endif
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::ITree* BatchLab::TreeFactory::create_tree(
 const std::string& aName
,const std::string& aStoreName
,const std::string& aStoreType
,bool aReadOnly
,bool aCreateNew
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string storeType = aStoreType;
  Lib::smanip::tolowercase(storeType);

  if(storeType=="memory") {
    //FIXME : should protect against the creation of two "MemoryTree".
    // Memory tree :
    MemoryTree* tree = new MemoryTree(fSession,aName);
    tree->setVerboseLevel(fSession.verboseLevel());
    fSession.addManager(tree);
    return tree;
  } else if(storeType=="ascii") {
    AsciiTree* tree = 
      new AsciiTree(fSession,aName,aStoreName,aReadOnly,aCreateNew,aOptions);
    tree->setVerboseLevel(fSession.verboseLevel());
    fSession.addManager(tree);
    return tree;
  } else if( (storeType=="hippodraw") ||
             (storeType=="tnt") ) {
    AsciiTree* tree = 
      new AsciiTree(fSession,aName,aStoreName,aReadOnly,aCreateNew,aOptions);
    tree->setVerboseLevel(fSession.verboseLevel());
    tree->setHippoStyle(true);
    fSession.addManager(tree);
    return tree;
  } else if(storeType=="mysql") {
    return loadTree("MySQL","BatchLabMySQL","BatchLabMySQL",
                    "MySQLManager",aStoreName,aReadOnly,aCreateNew,aOptions);
//  } else if(storeType=="rio") {
//    return loadTree("Rio","BatchLabRio","BatchLabRio",
//                    "Rio_Manager",aStoreName,aReadOnly,aCreateNew,aOptions);
  } else if(storeType=="root") {
    return loadTree("Rio","BatchLabRio","BatchLabRio",
                    "Rio_Manager",aStoreName,aReadOnly,aCreateNew,aOptions);
  } else if( (storeType.empty())     ||
             (storeType=="xml")  || //Supported by SLAC.
             (storeType=="aida") ){
    return loadTree("XML","BatchLabXML","BatchLabXML",
                    "XML_Manager",aStoreName,aReadOnly,aCreateNew,aOptions);
  } else if(storeType=="zip") {
    return loadTree("Zip","BatchLabZip","BatchLabZip",
                    "Zip_Manager",aStoreName,aReadOnly,aCreateNew,aOptions);
  } else if(storeType=="fs") {
    return loadTree("FS","BatchLabFS","BatchLabFS",
                    "FS_Manager",aStoreName,aReadOnly,aCreateNew,aOptions);
  } else if(storeType=="hdf5") {
    return loadTree("HDF5","BatchLabHDF5","BatchLabHDF5",
                    "HDF5_Manager",aStoreName,aReadOnly,aCreateNew,aOptions);
  } else if((storeType=="hbook")||(storeType=="zebra")) {
    AIDA::ITree* tree = loadTree("Zebra","BatchLabZebra","BatchLabZebra",
                    "Zebra_Manager",aStoreName,aReadOnly,aCreateNew,aOptions);
    if(tree) {
      return tree;
    } else if(aCreateNew) {
      Lib::Out out(fPrinter);
      out << "BatchLab::TreeFactory::create_tree :"
          << " can't create " << Lib::sout(aStoreName) << "." 
          << Lib::endl;
      out << " This is probably because the BatchLab Zebra"
          << " driver had not been reconstructed on this installation." 
          << Lib::endl;
      std::string path,name,suffix;
      Lib::smanip::pathNameSuffix(aStoreName,path,name,suffix);
      std::string xmlName = path+Lib::smanip::nosuffix(name)+".aida";
      out << " We try to proceed by creating the AIDA compressed xml file " 
          << Lib::sout(xmlName) << "..." << Lib::endl;
      return loadTree("XML","BatchLabXML","BatchLabXML",
               "XML_Manager",xmlName,aReadOnly,aCreateNew,"compress=yes");
    } else {
      Lib::Out out(fPrinter);
      out << "BatchLab::TreeFactory::create_tree :"
          << " failed to create zebra tree " << Lib::sout(aStoreName) << "." 
          << Lib::endl;
      return 0;
    }
  }

  Lib::Out out(fPrinter);
  out << "BatchLab::TreeFactory::create_tree :"
      << " unknown type \"" << aStoreType << "\"." << Lib::endl;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ITree* BatchLab::TreeFactory::loadTree(
 const std::string& aLibrary
,const std::string& aPath
,const std::string& aPackage
,const std::string& aStoreManager
,const std::string& aName
,bool aReadOnly
,bool aCreateNew
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Core::ILibraryManager* libraryManager = 
    Slash::libraryManager(fSession);
  if(!libraryManager) {
    Lib::Out out(fPrinter);
    out << "BatchLab::TreeFactory::loadTree :"
        << " LibraryManager not found." << Lib::endl;
    return 0;
  }
  Slash::Core::ILibrary* dll = 
    libraryManager->addLibrary(aLibrary,aPath,aPackage);
  if(!dll) {
    Lib::Out out(fPrinter);
    out << "BatchLab::TreeFactory::loadTree :"
        << " unable to add a DLL" << Lib::endl;
    return 0;
  }
  Slash::Store::IStoreManager* storeManager = 
    Lib_findManager(fSession,aStoreManager,Slash::Store::IStoreManager);
  if(!storeManager) {
    Lib::Out out(fPrinter);
    out << "BatchLab::TreeFactory::loadTree :"
        << " unable to find " << aStoreManager << " store manager." 
        << Lib::endl;
    return 0;
  }
  Slash::Store::ITree* baseTree = 
    storeManager->createTree(aName,aReadOnly,aCreateNew,aOptions);
  if(!baseTree) {
    Lib::Out out(fPrinter);
    out << "BatchLab::TreeFactory::loadTree :"
        << " can't create tree " << Lib::sout(aName) << "."
        << Lib::endl;
    return 0;
  }
  AIDA::ITree* aidaTree = Lib_Cast(*baseTree,AIDA::ITree);
  if(!aidaTree) {
    Lib::Out out(fPrinter);
    out << "BatchLab::TreeFactory::loadTree :"
        << " tree " << Lib::sout(aName) << " is not an AIDA::ITree."
        << Lib::endl;
    return 0;
  }
  return aidaTree;
}
