// this :
#include <BatchLab/XML/AIDA_ML.h>

#include <Slash/Store/IObject.h>

#include <Lib/ItemML.h>
#include <Lib/Out.h>
#include <Lib/sout.h>

#include <BatchLab/Core/BaseTree.h>
#include <BatchLab/XML/ReadContext.h>
#include <BatchLab/XML/WriteContext.h>

BatchLab::AIDA_ML::AIDA_ML(Lib::ItemMLFactory& aFactory,BaseTree& aTree)
:Lib::BaseML(aFactory,aTree.verboseLevel())
,fTree(aTree)
{
  addDefaultReaders();
  addDefaultWriters();

  std::vector<std::string> tags;
  tags.push_back("aida");
  tags.push_back("histogram1d");
  tags.push_back("histogram2d");
  tags.push_back("histogram3d");
  tags.push_back("profile1d");
  tags.push_back("profile2d");
  tags.push_back("axis");
  tags.push_back("statistics");
  tags.push_back("data1d");
  tags.push_back("data2d");
  tags.push_back("data3d");
  tags.push_back("tuple");
  tags.push_back("columns");
  tags.push_back("rows");
  tags.push_back("row");
  tags.push_back("entryTuple");
  tags.push_back("cloud1d");
  tags.push_back("cloud2d");
  tags.push_back("cloud3d");
  tags.push_back("entries1d");
  tags.push_back("entries2d");
  tags.push_back("entries3d");
  tags.push_back("dataPointSet");
  tags.push_back("dataPoint");
  tags.push_back("function");
  tags.push_back("arguments");
  tags.push_back("argument");
  tags.push_back("parameters");
  tags.push_back("annotation");
  setTags(tags);
}

BatchLab::AIDA_ML::~AIDA_ML(){}

bool BatchLab::AIDA_ML::visitEndElement(Lib::ItemML& aItemML,bool& aKeep) {
  const std::string& tagName = aItemML.tagName();
  if(tagName=="aida") {
    aKeep = false;
  } else if(
            (tagName=="histogram1d") ||
            (tagName=="histogram2d") ||
            (tagName=="histogram3d") ||
            (tagName=="profile1d") ||
            (tagName=="profile2d") ||
            (tagName=="cloud1d") ||
            (tagName=="cloud2d") ||
            (tagName=="cloud3d") ||
            (tagName=="tuple") ||
            (tagName=="dataPointSet") ||
            (tagName=="function")
     ){

    if(fTree.verboseLevel()) {
      Lib::Out out(fTree.printer());
      out << "AIDA_ML::visitEndElement :" 
          << " visit " << Lib::sout(tagName) << "..."
          << Lib::endl;
    }

    BatchLab::XML::Reader reader = findReader(tagName);
    if(!reader) {
      Lib::Out out(fTree.printer());
      out << "AIDA_ML::visitEndElement :" 
          << " reader not found for class " << Lib::sout(tagName)
          << Lib::endl;
    } else {
      BatchLab::XML::ReadContext ctx
        (fTree.verboseLevel(),fTree.session(),aItemML);
      std::string spath;
      Slash::Store::IObject* object = reader(ctx,spath);
      if(object) {
        if(!fTree.mkdirs(spath) || !fTree.manageObject(object,spath)) {
          delete object;
        }
      } else {
        fErrors++;
        //return false; //Continue anyway.
      }
    }

    aKeep = false;
  } else {
    aKeep = true;
  }
  return true;
}

void BatchLab::AIDA_ML::addReader(const std::string& aClass,BatchLab::XML::Reader aReader,bool aCheck){
  if(aCheck==true) {if(findReader(aClass)) return;}
  fReaders.push_back(std::pair<std::string,XML::Reader>(aClass,aReader));
}

BatchLab::XML::Reader BatchLab::AIDA_ML::findReader(const std::string& aClass){
  unsigned int number = fReaders.size();
  for(unsigned int index=0;index<number;index++) {
    if(aClass==fReaders[index].first) return fReaders[index].second;
  }
  return 0;
}

#include <BatchLab/XML/DataReader.h>

void BatchLab::AIDA_ML::addDefaultReaders(){
  addReader("histogram1d",(XML::Reader)BatchLab::XML::read_Histogram1D);
  addReader("histogram2d",(XML::Reader)BatchLab::XML::read_Histogram2D);
  addReader("histogram3d",(XML::Reader)BatchLab::XML::read_Histogram3D);
  addReader("profile1d",(XML::Reader)BatchLab::XML::read_Profile1D);
  addReader("profile2d",(XML::Reader)BatchLab::XML::read_Profile2D);
  addReader("cloud1d",(XML::Reader)BatchLab::XML::read_Cloud1D);
  addReader("cloud2d",(XML::Reader)BatchLab::XML::read_Cloud2D);
  addReader("cloud3d",(XML::Reader)BatchLab::XML::read_Cloud3D);
  addReader("tuple",(XML::Reader)BatchLab::XML::read_Tuple);
  addReader("dataPointSet",(XML::Reader)BatchLab::XML::read_DataPointSet);
  addReader("function",(XML::Reader)BatchLab::XML::read_Function);
}

//////////////////////////////////////////////////////////////////////////////
/// Writing //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// Readers had been displaced on the AIDA_ML object 
// to be usable by the Zip driver.
//////////////////////////////////////////////////////////////////////////////

void BatchLab::AIDA_ML::addWriter(const std::string& aClass,BatchLab::XML::Writer aWriter,bool aCheck){
  if(aCheck==true) {if(findWriter(aClass)) return;}
  fWriters.push_back(std::pair<std::string,XML::Writer>(aClass,aWriter));
}

BatchLab::XML::Writer BatchLab::AIDA_ML::findWriter(const std::string& aClass){
  unsigned int number = fWriters.size();
  for(unsigned int index=0;index<number;index++) {
    if(aClass==fWriters[index].first) return fWriters[index].second;
  }
  return 0;
}

bool BatchLab::AIDA_ML::write(Slash::Core::IWriter& aWriter,const std::string& aPath,const Slash::Store::IObject& aObject){
  std::string s = aObject.storeClassName();
  BatchLab::XML::Writer writer = findWriter(s);
  if(!writer) {
    Lib::Out out(fTree.printer());
    out << "BatchLab::AIDA_XML::write :" 
        << " writer not found for class " << Lib::sout(s)
        << Lib::endl;
    return false;
  }

  Lib::Out out(aWriter);
  BatchLab::XML::WriteContext ctx(fTree.printer(),
                                  out,
                                  aPath,aObject.name());
  if(!writer(aObject,ctx)) {
    Lib::Out out(fTree.printer());
    out << "BatchLab::AIDA_XML::write :" 
        << " writing failed for " << Lib::sout(aObject.name())
        << Lib::endl;
    return false;
  }
  return true;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#include <BatchLab/XML/DataWriter.h>

void BatchLab::AIDA_ML::addDefaultWriters(){
  addWriter("BatchLab::Histogram1D",(XML::Writer)BatchLab::XML::write_Histogram1D);
  addWriter("BatchLab::Histogram2D",(XML::Writer)BatchLab::XML::write_Histogram2D);
  addWriter("BatchLab::Histogram3D",(XML::Writer)BatchLab::XML::write_Histogram3D);
  addWriter("BatchLab::Profile1D",(XML::Writer)BatchLab::XML::write_Profile1D);
  addWriter("BatchLab::Profile2D",(XML::Writer)BatchLab::XML::write_Profile2D);
  addWriter("BatchLab::MemoryTuple",(XML::Writer)BatchLab::XML::write_Tuple);
  addWriter("BatchLab::Cloud1D",(XML::Writer)BatchLab::XML::write_Cloud1D);
  addWriter("BatchLab::Cloud2D",(XML::Writer)BatchLab::XML::write_Cloud2D);
  addWriter("BatchLab::Cloud3D",(XML::Writer)BatchLab::XML::write_Cloud3D);
  addWriter("BatchLab::DataPointSet",(XML::Writer)BatchLab::XML::write_DataPointSet);

  addWriter("BatchLab::Function::Gaussian",(XML::Writer)BatchLab::XML::write_Function);
  addWriter("BatchLab::Function::Scripted",(XML::Writer)BatchLab::XML::write_Function);
  addWriter("BatchLab::Function::Compiled",(XML::Writer)BatchLab::XML::write_Function);
}

