// this :
#include <BatchLab/XML/DataWriter.h>

#include <AIDA/IConstants.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
#include <AIDA/IHistogram3D.h>
#include <AIDA/ICloud1D.h>
#include <AIDA/ICloud2D.h>
#include <AIDA/ICloud3D.h>
#include <AIDA/IProfile1D.h>
#include <AIDA/IProfile2D.h>
#include <AIDA/ITuple.h>
#include <AIDA/ITupleEntry.h>
#include <AIDA/IAxis.h>
#include <AIDA/IDataPointSet.h>
#include <AIDA/IMeasurement.h>
#include <AIDA/IDataPoint.h>
#include <AIDA/IFunction.h>

#include <amanip/smanip.h>
#include <amanip/axis.h>
#include <amanip/tuple.h>

#include <Lib/Out.h>
#include <Lib/smanip.h>
#include <Lib/sout.h>
#include <Lib/Cast.h>
#include <Lib/StreamBuffer.h>
#include <iostream>

#include <BatchLab/Version.h>
#include <BatchLab/Core/Histogram.h> //FIXME
#include <BatchLab/Core/BaseTuple.h> //FIXME
#include <BatchLab/XML/WriteContext.h>

static bool get_booking(std::ostream& a_out,const AIDA::ITuple& a_o,std::string& a_string) {
  a_string += "{";
  int coln = a_o.columns();
  for(int index=0;index<coln;index++) {
    if(index) a_string += ",";
    std::string type = a_o.columnType(index);
    a_string += type + " ";
    std::string name = a_o.columnName(index);
    a_string += name + " = ";
    if(type=="ITuple") {
      const AIDA::ITuple* tupleTemplate = a_o.getTuple(index);
      if(!tupleTemplate) {
        a_out << "BatchLab::amanip::tuple_getBooking : "
              << " tuple column without a template tuple."
              << std::endl;
        a_string = "";
        return false;
      }
      if(!get_booking(a_out,*tupleTemplate,a_string)) return false;
    } else {
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
      const BatchLab::BaseTuple* baseTuple =  
        Lib_ConstCast(a_o,BatchLab::BaseTuple);
      if(!baseTuple) {
        a_out << "BatchLab::amanip::tuple_getBooking : "
              << " tuple not a BatchLab::BaseTuple."
              << std::endl;
        a_string = "";
        return false;
      }
      std::string def = baseTuple->columnDefaultValue(index);
      a_string += def;
#else
      std::string def = a_o.columnDefaultValue(index);
      a_string += def;
#endif
    }  
       
  }  
  a_string += "}";
  return true;
}

static void writeHistogramAxis(const AIDA::IAxis&,
                               const std::string&,Lib::Out&,int=0);
static void writeHistogramBin(Lib::Out&,
			      const AIDA::IHistogram1D&,
                              const std::string&,
                              int);
static void writeHistogramBin(Lib::Out&,
                              const AIDA::IHistogram2D&,
			      const std::string&,
                              int,int);
static void writeHistogramBin(Lib::Out&,
			      const AIDA::IHistogram3D&,
                              const std::string&,
                              int,int,int);
static void writeHistogramBin(Lib::Out&,
                              const AIDA::IProfile1D&,
                              const std::string&,int);
static void writeHistogramBin(Lib::Out&,
                              const AIDA::IProfile2D&,
                              const std::string&,int,int);

static void writeTupleRows(Lib::Out&,AIDA::ITuple&,const std::string&,
                           std::ostream&);

//////////////////////////////////////////////////////////////////////////////
bool BatchLab::XML::write_Histogram1D(
 const Slash::Store::IObject& aIObject
,BatchLab::XML::WriteContext& aContext
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const AIDA::IHistogram1D* aObject = 
    Lib_ConstCast(aIObject,AIDA::IHistogram1D);
  if(!aObject) return false;

  Lib::Out& aOut = aContext.out();
  std::string aPath = aContext.path();
  std::string aName = aContext.name();
  int aShift = aContext.shift();

  std::string spaces;
  for(int i=0;i<aShift;i++) spaces += " ";

  // <histogram1d> :
  aOut << spaces << "  <histogram1d"
       << " path=" << Lib::sout(aPath)
       << " name=" << Lib::sout(aName)
       << " title=" << Lib::sout(aObject->title())
       << ">" << Lib::endl;

  // <axis> :
  writeHistogramAxis(aObject->axis(),"x",aOut,aShift);

  // <statistics> : 
  aOut << spaces << "    <statistics"
       << " entries=" << Lib::sout(aObject->entries())
       << ">" << Lib::endl;
  aOut << spaces << "      <statistic"
       << " direction=" << Lib::sout("x") 
       << " mean=" << Lib::sout(amanip::d2s(aObject->mean())) 
       << " rms=" << Lib::sout(amanip::d2s(aObject->rms())) 
       << "/>" << Lib::endl;
  aOut << spaces << "    </statistics>" << Lib::endl;
  
  // bins :
  aOut << spaces << "    <data1d>" << Lib::endl;

  int xbins = aObject->axis().bins();
  for(int index=0;index<xbins;index++)
    writeHistogramBin(aOut,*aObject,spaces,index);

  writeHistogramBin(aOut,*aObject,spaces,AIDA::IAxis::UNDERFLOW_BIN);
  writeHistogramBin(aOut,*aObject,spaces,AIDA::IAxis::OVERFLOW_BIN);

  aOut << spaces << "    </data1d>" << Lib::endl;
  aOut << spaces << "  </histogram1d>" << Lib::endl;

  return true;  
}
//////////////////////////////////////////////////////////////////////////////
static void writeHistogramBin(
 Lib::Out& aOut
,const AIDA::IHistogram2D& aObject
,const std::string& aSpaces
,int aIndexX
,int aIndexY
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int entries = aObject.binEntries(aIndexX,aIndexY);
  if(entries) {
    aOut << aSpaces << "      <bin2d"
         << " binNumX=" << Lib::sout(amanip::bin_to_string(aIndexX))
         << " binNumY=" << Lib::sout(amanip::bin_to_string(aIndexY))
         << " entries=" << Lib::sout(entries)
         << " height=" << Lib::sout(amanip::d2s(aObject.binHeight(aIndexX,aIndexY)))
         << " error=" << Lib::sout(amanip::d2s(aObject.binError(aIndexX,aIndexY)));

    double meanX = aObject.binMeanX(aIndexX,aIndexY);
    if(meanX!=0) {
      aOut << " weightedMeanX=" << Lib::sout(amanip::d2s(meanX));
    }
    double meanY = aObject.binMeanY(aIndexX,aIndexY);
    if(meanY!=0) {
      aOut << " weightedMeanY=" << Lib::sout(amanip::d2s(meanY));
    }

    //FIXME : use AIDA only.
    //FIXME : the below cast slow down things.
    if(const BatchLab::Histogram2D* object = 
         Lib_ConstCast(aObject,BatchLab::Histogram2D)) {
      double stddevX = object->binStdDevX(aIndexX,aIndexY);
      if(stddevX!=0) {
        aOut << " weightedRmsX=" << Lib::sout(amanip::d2s(stddevX));
      }
      double stddevY = object->binStdDevY(aIndexX,aIndexY);
      if(stddevY!=0) {
        aOut << " weightedRmsY=" << Lib::sout(amanip::d2s(stddevY));
      }
    }

    aOut << "/>" << Lib::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::XML::write_Histogram2D(
 const Slash::Store::IObject& aIObject
,BatchLab::XML::WriteContext& aContext
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const AIDA::IHistogram2D* aObject = 
    Lib_ConstCast(aIObject,AIDA::IHistogram2D);
  if(!aObject) return false;

  Lib::Out& aOut = aContext.out();
  std::string aPath = aContext.path();
  std::string aName = aContext.name();
  int aShift = aContext.shift();

  std::string spaces;
  for(int i=0;i<aShift;i++) spaces += " ";

  // <histogram2d> :
  aOut << spaces << "  <histogram2d"
       << " path=" << Lib::sout(aPath)
       << " name=" << Lib::sout(aName)
       << " title=" << Lib::sout(aObject->title())
       << ">" << Lib::endl;


  // <axis> :
  writeHistogramAxis(aObject->xAxis(),"x",aOut,aShift);
  writeHistogramAxis(aObject->yAxis(),"y",aOut,aShift);

  // <statistics> : 
  aOut << spaces << "    <statistics"
       << " entries=" << Lib::sout(aObject->entries())
       << ">" << Lib::endl;
  aOut << spaces << "      <statistic"
       << " direction=" << Lib::sout("x")
       << " mean=" << Lib::sout(amanip::d2s(aObject->meanX()))
       << " rms=" << Lib::sout(amanip::d2s(aObject->rmsX())) 
       << "/>" << Lib::endl;
  aOut << spaces << "      <statistic"
       << " direction=" << Lib::sout("y")
       << " mean=" << Lib::sout(amanip::d2s(aObject->meanY()))
       << " rms=" << Lib::sout(amanip::d2s(aObject->rmsY())) 
       << "/>" << Lib::endl;
  aOut << spaces << "    </statistics>" << Lib::endl;
  
  // bins :
  aOut << spaces << "    <data2d>" << Lib::endl;

  int xbins = aObject->xAxis().bins();
  int ybins = aObject->yAxis().bins();
  int indexX,indexY;
  for(indexX=0;indexX<xbins;indexX++) {
    for(indexY=0;indexY<ybins;indexY++) {
      writeHistogramBin(aOut,*aObject,spaces,indexX,indexY);
    }
  }
  
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::UNDERFLOW_BIN,AIDA::IAxis::UNDERFLOW_BIN);
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::OVERFLOW_BIN,AIDA::IAxis::UNDERFLOW_BIN);
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::UNDERFLOW_BIN,AIDA::IAxis::OVERFLOW_BIN);
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::OVERFLOW_BIN,AIDA::IAxis::OVERFLOW_BIN);

  for(indexX=0;indexX<xbins;indexX++){
    writeHistogramBin(aOut,*aObject,spaces,indexX,AIDA::IAxis::UNDERFLOW_BIN);
    writeHistogramBin(aOut,*aObject,spaces,indexX,AIDA::IAxis::OVERFLOW_BIN);
  }

  for(indexY=0;indexY<ybins;indexY++){
    writeHistogramBin(aOut,*aObject,spaces,AIDA::IAxis::UNDERFLOW_BIN,indexY);
    writeHistogramBin(aOut,*aObject,spaces,AIDA::IAxis::OVERFLOW_BIN,indexY);
  }

  aOut << spaces << "    </data2d>" << Lib::endl;
  aOut << spaces << "  </histogram2d>" << Lib::endl;

  return true;
}
//////////////////////////////////////////////////////////////////////////////
static void writeHistogramBin(
 Lib::Out& aOut
,const AIDA::IHistogram3D& aObject
,const std::string& aSpaces
,int aIndexX
,int aIndexY
,int aIndexZ
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int entries = aObject.binEntries(aIndexX,aIndexY,aIndexZ);
  if(entries) {
    aOut << aSpaces << "      <bin3d"
         << " binNumX=" << Lib::sout(amanip::bin_to_string(aIndexX))
         << " binNumY=" << Lib::sout(amanip::bin_to_string(aIndexY))
         << " binNumZ=" << Lib::sout(amanip::bin_to_string(aIndexZ))
         << " entries=" << Lib::sout(entries)
         << " height=" << Lib::sout(amanip::d2s(aObject.binHeight(aIndexX,aIndexY,aIndexZ)))
         << " error=" << Lib::sout(amanip::d2s(aObject.binError(aIndexX,aIndexY,aIndexZ)));


    double meanX = aObject.binMeanX(aIndexX,aIndexY,aIndexZ);
    if(meanX!=0) {
      aOut << " weightedMeanX=" << Lib::sout(amanip::d2s(meanX));
    }
    double meanY = aObject.binMeanY(aIndexX,aIndexY,aIndexZ);
    if(meanY!=0) {
      aOut << " weightedMeanY=" << Lib::sout(amanip::d2s(meanY));
    }
    double meanZ = aObject.binMeanZ(aIndexX,aIndexY,aIndexZ);
    if(meanY!=0) {
      aOut << " weightedMeanZ=" << Lib::sout(amanip::d2s(meanZ));
    }

    //FIXME : use AIDA only.
    //FIXME : the below cast slow down things.
    if(const BatchLab::Histogram3D* object = 
         Lib_ConstCast(aObject,BatchLab::Histogram3D)) {
      double stddevX = object->binStdDevX(aIndexX,aIndexY,aIndexZ);
      if(stddevX!=0) {
        aOut << " weightedRmsX=" << Lib::sout(amanip::d2s(stddevX));
      }
      double stddevY = object->binStdDevY(aIndexX,aIndexY,aIndexZ);
      if(stddevY!=0) {
        aOut << " weightedRmsY=" << Lib::sout(amanip::d2s(stddevY));
      }
      double stddevZ = object->binStdDevZ(aIndexX,aIndexY,aIndexZ);
      if(stddevZ!=0) {
        aOut << " weightedRmsZ=" << Lib::sout(amanip::d2s(stddevZ));
      }
    }

    aOut << "/>" << Lib::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::XML::write_Histogram3D(
 const Slash::Store::IObject& aIObject
,BatchLab::XML::WriteContext& aContext
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const AIDA::IHistogram3D* aObject = 
    Lib_ConstCast(aIObject,AIDA::IHistogram3D);
  if(!aObject) return false;

  Lib::Out& aOut = aContext.out();
  std::string aPath = aContext.path();
  std::string aName = aContext.name();
  int aShift = aContext.shift();

  std::string spaces;
  for(int i=0;i<aShift;i++) spaces += " ";

  // <histogram3d> :
  aOut << spaces << "  <histogram3d"
       << " path=" << Lib::sout(aPath)
       << " name=" << Lib::sout(aName)
       << " title=" << Lib::sout(aObject->title())
       << ">" << Lib::endl;

  // <axis> :
  writeHistogramAxis(aObject->xAxis(),"x",aOut,aShift);
  writeHistogramAxis(aObject->yAxis(),"y",aOut,aShift);
  writeHistogramAxis(aObject->zAxis(),"z",aOut,aShift);

  // <statistics> : 
  aOut << spaces << "    <statistics"
       << " entries=" << Lib::sout(aObject->entries())
       << ">" << Lib::endl;
  aOut << spaces << "      <statistic"
       << " direction=" << Lib::sout("x")
       << " mean=" << Lib::sout(amanip::d2s(aObject->meanX()))
       << " rms=" << Lib::sout(amanip::d2s(aObject->rmsX())) 
       << "/>" << Lib::endl;
  aOut << spaces << "      <statistic"
       << " direction=" << Lib::sout("y")
       << " mean=" << Lib::sout(amanip::d2s(aObject->meanY()))
       << " rms=" << Lib::sout(amanip::d2s(aObject->rmsY())) 
       << "/>" << Lib::endl;
  aOut << spaces << "      <statistic"
       << " direction=" << Lib::sout("z")
       << " mean=" << Lib::sout(amanip::d2s(aObject->meanZ()))
       << " rms=" << Lib::sout(amanip::d2s(aObject->rmsZ())) 
       << "/>" << Lib::endl;
  aOut << spaces << "    </statistics>" << Lib::endl;
  
  // bins :
  aOut << spaces << "    <data3d>" << Lib::endl;
  int xbins = aObject->xAxis().bins();
  int ybins = aObject->yAxis().bins();
  int zbins = aObject->zAxis().bins();
  int indexX,indexY,indexZ;
  for(indexX=0;indexX<xbins;indexX++) {
    for(indexY=0;indexY<ybins;indexY++) {
      for(indexZ=0;indexZ<zbins;indexZ++) {
        writeHistogramBin(aOut,*aObject,spaces,indexX,indexY,indexZ);
      }
    }
  }

  // Corners :
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::UNDERFLOW_BIN,
                    AIDA::IAxis::UNDERFLOW_BIN,
                    AIDA::IAxis::UNDERFLOW_BIN);
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::OVERFLOW_BIN,
                    AIDA::IAxis::UNDERFLOW_BIN,
                    AIDA::IAxis::UNDERFLOW_BIN);
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::UNDERFLOW_BIN,
                    AIDA::IAxis::OVERFLOW_BIN,
                    AIDA::IAxis::UNDERFLOW_BIN);
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::OVERFLOW_BIN,
                    AIDA::IAxis::OVERFLOW_BIN,
                    AIDA::IAxis::UNDERFLOW_BIN);
  
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::UNDERFLOW_BIN,
                    AIDA::IAxis::UNDERFLOW_BIN,
                    AIDA::IAxis::OVERFLOW_BIN);
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::OVERFLOW_BIN,
                    AIDA::IAxis::UNDERFLOW_BIN,
                    AIDA::IAxis::OVERFLOW_BIN);
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::UNDERFLOW_BIN,
                    AIDA::IAxis::OVERFLOW_BIN,
                    AIDA::IAxis::OVERFLOW_BIN);
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::OVERFLOW_BIN,
                    AIDA::IAxis::OVERFLOW_BIN,
                    AIDA::IAxis::OVERFLOW_BIN);


  // Edges :
  for(indexX=0;indexX<xbins;indexX++){
    writeHistogramBin(aOut,*aObject,spaces,
                      indexX,
                      AIDA::IAxis::UNDERFLOW_BIN,
                      AIDA::IAxis::UNDERFLOW_BIN);
    writeHistogramBin(aOut,*aObject,spaces,
                      indexX,
                      AIDA::IAxis::OVERFLOW_BIN,
                      AIDA::IAxis::UNDERFLOW_BIN);
    writeHistogramBin(aOut,*aObject,spaces,
                      indexX,
                      AIDA::IAxis::UNDERFLOW_BIN,
                      AIDA::IAxis::OVERFLOW_BIN);
    writeHistogramBin(aOut,*aObject,spaces,
                      indexX,
                      AIDA::IAxis::OVERFLOW_BIN,
                      AIDA::IAxis::OVERFLOW_BIN);
  }

  for(indexY=0;indexY<ybins;indexY++){
    writeHistogramBin(aOut,*aObject,spaces,
                      AIDA::IAxis::UNDERFLOW_BIN,
                      indexY,
                      AIDA::IAxis::UNDERFLOW_BIN);
    writeHistogramBin(aOut,*aObject,spaces,
                      AIDA::IAxis::OVERFLOW_BIN,
                      indexY,
                      AIDA::IAxis::UNDERFLOW_BIN);
    writeHistogramBin(aOut,*aObject,spaces,
                      AIDA::IAxis::UNDERFLOW_BIN,
                      indexY,
                      AIDA::IAxis::OVERFLOW_BIN);
    writeHistogramBin(aOut,*aObject,spaces,
                      AIDA::IAxis::OVERFLOW_BIN,
                      indexY,
                      AIDA::IAxis::OVERFLOW_BIN);
  }

  for(indexZ=0;indexZ<zbins;indexZ++){
    writeHistogramBin(aOut,*aObject,spaces,
                      AIDA::IAxis::UNDERFLOW_BIN,
                      AIDA::IAxis::UNDERFLOW_BIN,
                      indexZ);
    writeHistogramBin(aOut,*aObject,spaces,
                      AIDA::IAxis::OVERFLOW_BIN,
                      AIDA::IAxis::UNDERFLOW_BIN,
                      indexZ);
    writeHistogramBin(aOut,*aObject,spaces,
                      AIDA::IAxis::UNDERFLOW_BIN,
                      AIDA::IAxis::OVERFLOW_BIN,
                      indexZ);
    writeHistogramBin(aOut,*aObject,spaces,
                      AIDA::IAxis::OVERFLOW_BIN,
                      AIDA::IAxis::OVERFLOW_BIN,
                      indexZ);
  }


  // Faces :
  for(indexX=0;indexX<xbins;indexX++) {
    for(indexY=0;indexY<ybins;indexY++) {
      writeHistogramBin(aOut,*aObject,spaces,
                        indexX,indexY,AIDA::IAxis::UNDERFLOW_BIN);
      writeHistogramBin(aOut,*aObject,spaces,
                        indexX,indexY,AIDA::IAxis::OVERFLOW_BIN);
    }
  }
  for(indexY=0;indexY<ybins;indexY++) {
    for(indexZ=0;indexZ<zbins;indexZ++) {
      writeHistogramBin(aOut,*aObject,spaces,
                        AIDA::IAxis::UNDERFLOW_BIN,indexY,indexZ);
      writeHistogramBin(aOut,*aObject,spaces,
                        AIDA::IAxis::OVERFLOW_BIN,indexY,indexZ);
    }
  }
  for(indexX=0;indexX<xbins;indexX++) {
    for(indexZ=0;indexZ<zbins;indexZ++) {
      writeHistogramBin(aOut,*aObject,spaces,
                        indexX,AIDA::IAxis::UNDERFLOW_BIN,indexZ);
      writeHistogramBin(aOut,*aObject,spaces,
                        indexX,AIDA::IAxis::OVERFLOW_BIN,indexZ);
    }
  }

  aOut << spaces << "    </data3d>" << Lib::endl;
  aOut << spaces << "  </histogram3d>" << Lib::endl;

  return true;
}
//////////////////////////////////////////////////////////////////////////////
static void writeHistogramBin(
 Lib::Out& aOut
,const AIDA::IProfile1D& aObject
,const std::string& aSpaces
,int aIndex
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aObject.binEntries(aIndex)) {
    aOut << aSpaces << "      <bin1d"
         << " binNum=" << Lib::sout(amanip::bin_to_string(aIndex))
         << " entries=" << Lib::sout(aObject.binEntries(aIndex))
         << " height=" << Lib::sout(amanip::d2s(aObject.binHeight(aIndex)))
         << " error=" << Lib::sout(amanip::d2s(aObject.binError(aIndex)))
         << " weightedMean=" << Lib::sout(amanip::d2s(aObject.binMean(aIndex)));

    //FIXME : should use AIDA only.
    //FIXME : the below cast slow down things.
    if(const BatchLab::Profile1D* object = 
         Lib_ConstCast(aObject,BatchLab::Profile1D)) {
      double stddev = object->binStdDev(aIndex);
      if(stddev!=0) {
        aOut << " weightedRms=" << Lib::sout(amanip::d2s(stddev));
      }
    }

    aOut << " rms=" << Lib::sout(amanip::d2s(aObject.binRms(aIndex)));
    aOut << "/>" << Lib::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::XML::write_Profile1D(
 const Slash::Store::IObject& aIObject
,BatchLab::XML::WriteContext& aContext
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const AIDA::IProfile1D* aObject = 
    Lib_ConstCast(aIObject,AIDA::IProfile1D);
  if(!aObject) return false;

  Lib::Out& aOut = aContext.out();
  std::string aPath = aContext.path();
  std::string aName = aContext.name();
  int aShift = aContext.shift();

  std::string spaces;
  for(int i=0;i<aShift;i++) spaces += " ";

  // <profile1d> :
  aOut << spaces << "  <profile1d"
       << " path=" << Lib::sout(aPath)
       << " name=" << Lib::sout(aName)
       << " title=" << Lib::sout(aObject->title())
       << ">" << Lib::endl;

  // <axis> :
  writeHistogramAxis(aObject->axis(),"x",aOut,aShift);

  // <statistics> : 
  aOut << spaces << "    <statistics"
       << " entries=" << Lib::sout(aObject->entries())
       << ">" << Lib::endl;
  aOut << spaces << "      <statistic"
       << " direction=" << Lib::sout("x") 
       << " mean=" << Lib::sout(amanip::d2s(aObject->mean())) 
       << " rms=" << Lib::sout(amanip::d2s(aObject->rms())) 
       << "/>" << Lib::endl;
  aOut << spaces << "    </statistics>" << Lib::endl;
        
  // bins :
  aOut << spaces << "    <data1d>" << Lib::endl;
  int xbins = aObject->axis().bins();
  for(int index=0;index<xbins;index++) {
    writeHistogramBin(aOut,*aObject,spaces,index);
  }

  writeHistogramBin(aOut,*aObject,spaces,AIDA::IAxis::UNDERFLOW_BIN);
  writeHistogramBin(aOut,*aObject,spaces,AIDA::IAxis::OVERFLOW_BIN);
  
  aOut << spaces << "    </data1d>" << Lib::endl;
  aOut << spaces << "  </profile1d>" << Lib::endl;

  return true;
}
//////////////////////////////////////////////////////////////////////////////
static void writeHistogramBin(
 Lib::Out& aOut
,const AIDA::IProfile2D& aObject
,const std::string& aSpaces
,int aIndexX
,int aIndexY
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aObject.binEntries(aIndexX,aIndexY)) {
    aOut << aSpaces << "      <bin2d"
         << " binNumX=" << Lib::sout(amanip::bin_to_string(aIndexX))
         << " binNumY=" << Lib::sout(amanip::bin_to_string(aIndexY))
         << " entries=" << Lib::sout(aObject.binEntries(aIndexX,aIndexY))
         << " height=" << Lib::sout(amanip::d2s(aObject.binHeight(aIndexX,aIndexY)))
         << " error=" << Lib::sout(amanip::d2s(aObject.binError(aIndexX,aIndexY)))
         << " weightedMeanX=" << Lib::sout(amanip::d2s(aObject.binMeanX(aIndexX,aIndexY)))
         << " weightedMeanY=" << Lib::sout(amanip::d2s(aObject.binMeanY(aIndexX,aIndexY)));

    //FIXME : should use AIDA only.
    //FIXME : the below cast slow down things.
    if(const BatchLab::Profile2D* object = 
         Lib_ConstCast(aObject,BatchLab::Profile2D)) {
      double stddevX = object->binStdDevX(aIndexX,aIndexY);
      if(stddevX!=0) {
        aOut << " weightedRmsX=" << Lib::sout(amanip::d2s(stddevX));
      }
      double stddevY = object->binStdDevY(aIndexX,aIndexY);
      if(stddevY!=0) {
        aOut << " weightedRmsY=" << Lib::sout(amanip::d2s(stddevY));
      }
    }

    aOut << " rms=" << Lib::sout(amanip::d2s(aObject.binRms(aIndexX,aIndexY)));
    aOut << "/>" << Lib::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::XML::write_Profile2D(
 const Slash::Store::IObject& aIObject
,BatchLab::XML::WriteContext& aContext
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const AIDA::IProfile2D* aObject = 
    Lib_ConstCast(aIObject,AIDA::IProfile2D);
  if(!aObject) return false;

  Lib::Out& aOut = aContext.out();
  std::string aPath = aContext.path();
  std::string aName = aContext.name();
  int aShift = aContext.shift();

  std::string spaces;
  for(int i=0;i<aShift;i++) spaces += " ";

  // <profile2d> :
  aOut << spaces << "  <profile2d"
       << " path=" << Lib::sout(aPath)
       << " name=" << Lib::sout(aName)
       << " title=" << Lib::sout(aObject->title())
       << ">" << Lib::endl;

  // <axis> :
  writeHistogramAxis(aObject->xAxis(),"x",aOut,aShift);
  writeHistogramAxis(aObject->yAxis(),"y",aOut,aShift);
  
  // <statistics> : 
  aOut << spaces << "    <statistics"
       << " entries=" << Lib::sout(aObject->entries())
       << ">" << Lib::endl;
  aOut << spaces << "      <statistic"
       << " direction=" << Lib::sout("x")
       << " mean=" << Lib::sout(amanip::d2s(aObject->meanX()))
       << " rms=" << Lib::sout(amanip::d2s(aObject->rmsX())) 
       << "/>" << Lib::endl;
  aOut << spaces << "      <statistic"
       << " direction=" << Lib::sout("y")
       << " mean=" << Lib::sout(amanip::d2s(aObject->meanY()))
       << " rms=" << Lib::sout(amanip::d2s(aObject->rmsY())) 
       << "/>" << Lib::endl;
  aOut << spaces << "    </statistics>" << Lib::endl;
  
  // bins :
  aOut << spaces << "    <data2d>" << Lib::endl;
 {int xbins = aObject->xAxis().bins();
  int ybins = aObject->yAxis().bins();
  for(int indexX=0;indexX<xbins;indexX++) {
    for(int indexY=0;indexY<ybins;indexY++) {
      writeHistogramBin(aOut,*aObject,spaces,indexX,indexY);
    }
  }}
  
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::UNDERFLOW_BIN,AIDA::IAxis::UNDERFLOW_BIN);
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::OVERFLOW_BIN,AIDA::IAxis::UNDERFLOW_BIN);
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::UNDERFLOW_BIN,AIDA::IAxis::OVERFLOW_BIN);
  writeHistogramBin(aOut,*aObject,spaces,
                    AIDA::IAxis::OVERFLOW_BIN,AIDA::IAxis::OVERFLOW_BIN);

  for(int indexX=0;indexX<aObject->xAxis().bins();indexX++){
    writeHistogramBin(aOut,*aObject,spaces,indexX,AIDA::IAxis::UNDERFLOW_BIN);
    writeHistogramBin(aOut,*aObject,spaces,indexX,AIDA::IAxis::OVERFLOW_BIN);
  }

  for(int indexY=0;indexY<aObject->yAxis().bins();indexY++){
    writeHistogramBin(aOut,*aObject,spaces,AIDA::IAxis::UNDERFLOW_BIN,indexY);
    writeHistogramBin(aOut,*aObject,spaces,AIDA::IAxis::OVERFLOW_BIN,indexY);
  }

  aOut << spaces << "    </data2d>" << Lib::endl;
  aOut << spaces << "  </profile2d>" << Lib::endl;

  return true;
}
//////////////////////////////////////////////////////////////////////////////
void writeHistogramAxis(
 const AIDA::IAxis& aAxis
,const std::string& aDirection
,Lib::Out& aOut
,int aShift
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string spaces;
  for(int i=0;i<aShift;i++) spaces += " ";

  if(aAxis.isFixedBinning()) {
    aOut << spaces << "    <axis"
         << " direction=" << Lib::sout(aDirection) 
         << " numberOfBins=" << Lib::sout(aAxis.bins()) 
         << " min=" << Lib::sout(amanip::d2s(aAxis.lowerEdge())) 
         << " max=" << Lib::sout(amanip::d2s(aAxis.upperEdge())) 
         << "/>" << Lib::endl;
  } else {
    aOut << spaces << "    <axis"
         << " direction=" << Lib::sout(aDirection) 
         << " numberOfBins=" << Lib::sout(aAxis.bins()) 
         << " min=" << Lib::sout(amanip::d2s(aAxis.lowerEdge())) 
         << " max=" << Lib::sout(amanip::d2s(aAxis.upperEdge())) 
         << ">" << Lib::endl;
    int number = aAxis.bins()-1;
    for(int index=0;index<number;index++) {
      aOut << spaces << "      <binBorder"
           << " value=" << Lib::sout(amanip::d2s(aAxis.binUpperEdge(index))) 
           << "/>" << Lib::endl;
    }
    aOut << spaces << "    </axis>" << Lib::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::XML::write_Tuple(
 const Slash::Store::IObject& aIObject
,BatchLab::XML::WriteContext& aContext
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const AIDA::ITuple* aObject = 
    Lib_ConstCast(aIObject,AIDA::ITuple);
  if(!aObject) return false;

  Lib::Out& aOut = aContext.out();
  std::string aPath = aContext.path();
  std::string aName = aContext.name();
  //int aShift = aContext.shift();
  Slash::Core::IWriter& printer = aContext.printer();
  Lib::StreamBuffer sbuf(printer);
  std::ostream out(&sbuf);

  int aShift = 0;
  std::string spaces;
  for(int i=0;i<aShift;i++) spaces += " ";

  // <tuple> :
  aOut << spaces << "  <tuple"
       << " path=" << Lib::sout(aPath)
       << " name=" << Lib::sout(aName)
       << " title=" << Lib::sout(aObject->title())
       << ">" << Lib::endl;

  // <columns> :
  aOut << spaces << "    <columns>" << Lib::endl;
  int coln = aObject->columns();
  for(int index=0;index<coln;index++) {
    std::string type = aObject->columnType(index);
    if(type=="ITuple") {
      std::string booking;
      AIDA::ITuple* tupleTemplate = 
        const_cast<AIDA::ITuple*>(aObject->getTuple(index));
      if(!tupleTemplate) {
        out << "BatchLab::XML::writeTuple : "
            << " tuple column without a template tuple."
            << std::endl;
      } else {
        get_booking(out,*tupleTemplate,booking);
      }
      aOut << spaces << "      <column"
           << " name=" << Lib::sout(aObject->columnName(index)) 
           << " type=" << Lib::sout(aObject->columnType(index))
           << " booking=" << Lib::sout(booking)
           << "/>" << Lib::endl;
    } else {
      bool isDefaultDefault = true;
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
      std::string sdefault;
      const BatchLab::BaseTuple* baseTuple = 
        Lib_ConstCast(*aObject,BatchLab::BaseTuple);
      if(!baseTuple) {
          out << "BatchLab::XML::writeTuple : "
              << " tuple not a BatchLab::BaseTuple !"
              << std::endl;
      } else { 
        sdefault = baseTuple->columnDefaultValue(index);
        isDefaultDefault = baseTuple->isDefaultDefault(index);
      }
#else
      std::string sdefault = aObject->columnDefaultValue(index);
      const BatchLab::BaseTuple* baseTuple = 
        Lib_ConstCast(*aObject,BatchLab::BaseTuple);
      if(!baseTuple) {
          out << "BatchLab::XML::writeTuple : "
              << " tuple not a BatchLab::BaseTuple !"
              << std::endl;
      } else { 
        isDefaultDefault = baseTuple->isDefaultDefault(index);
      }
#endif
      if((isDefaultDefault)||(sdefault.empty())) {
        aOut << spaces << "      <column"
             << " name=" << Lib::sout(aObject->columnName(index)) 
             << " type=" << Lib::sout(aObject->columnType(index))
             << "/>" << Lib::endl;
      } else {
        aOut << spaces << "      <column"
             << " name=" << Lib::sout(aObject->columnName(index)) 
             << " type=" << Lib::sout(aObject->columnType(index))
             << " default=" << Lib::sout(sdefault)
             << "/>" << Lib::endl;
      }
    }
  }
  aOut << spaces << "    </columns>" << Lib::endl;
  
  // rows :
  aOut << spaces << "    <rows>" << Lib::endl;
  writeTupleRows(aOut,const_cast<AIDA::ITuple&>(*aObject),
                 spaces+"      ",out);
  aOut << spaces << "    </rows>" << Lib::endl;
  
  aOut << spaces << "  </tuple>" << Lib::endl;

  return true;
}
//////////////////////////////////////////////////////////////////////////////
void writeTupleRows(
 Lib::Out& aOut
,AIDA::ITuple& aObject
,const std::string& aSpaces
,std::ostream& a_out
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int coln = aObject.columns();
  aObject.start();
  while(aObject.next()) {
    aOut << aSpaces << "<row>" << Lib::endl;
    for(int index=0;index<coln;index++) {
      std::string type = aObject.columnType(index);
      if(type=="ITuple") {
        const AIDA::ITupleEntry* tupleEntry = aObject.getObject(index);
        if(!tupleEntry) {        
          a_out << "BatchLab::XML::writeTupleRows : "
                << " no tuple entry."
                << std::endl;
        } else {
          AIDA::ITuple* tuple = 
            dynamic_cast<AIDA::ITuple*>((AIDA::ITupleEntry*)tupleEntry);
          if(!tuple) {
            a_out << "BatchLab::XML::writeTupleRows : "
                  << " can't cast ITupleEntry to ITuple."
                  << std::endl;
          } else {
            aOut << aSpaces << "  <entryTuple>" << Lib::endl;
            writeTupleRows(aOut,*tuple,aSpaces+"    ",a_out);
            aOut << aSpaces << "  </entryTuple>" << Lib::endl;
          }
        }
      } else {
        std::string value;
        if(!amanip::get_string(a_out,aObject,index,value)) {
        }
        //FIXME : and if the column type is string and that the
        //        string value contains double quote ?
        // > : &lt;
        // < : &gt;
        // & : &amp;
        // " : &quote;
        // ' : &apos;
        aOut << aSpaces << "  <entry" 
             << " value=" << Lib::sout(value)
             << "/>" << Lib::endl;
      }
    }
    aOut << aSpaces << "</row>" << Lib::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::XML::write_Cloud1D(
 const Slash::Store::IObject& aIObject
,BatchLab::XML::WriteContext& aContext
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const AIDA::ICloud1D* aObject = 
    Lib_ConstCast(aIObject,AIDA::ICloud1D);
  if(!aObject) return false;

  Lib::Out& aOut = aContext.out();
  std::string aPath = aContext.path();
  std::string aName = aContext.name();
  //int aShift = aContext.shift();

  // <cloud1d> :
  aOut << "  <cloud1d"
       << " path=" << Lib::sout(aPath)
       << " name=" << Lib::sout(aName)
       << " title=" << Lib::sout(aObject->title())
       << " maxEntries=" << Lib::sout(aObject->maxEntries())
       << ">" << Lib::endl;

  if(aObject->isConverted()) {

    const AIDA::IHistogram1D& object = aObject->histogram();
    //FIXME : aName ?
    const Slash::Store::IObject* obj = 
      Lib_ConstCast(object,Slash::Store::IObject);
    if(!obj) return false;
    write_Histogram1D(*obj,aContext);

  } else {
    // entries :
    aOut << "    <entries1d>" << Lib::endl;
    int number = aObject->entries();
    for(int index=0;index<number;index++) {
      double weight = aObject->weight(index);
      if(weight==1.) {
        aOut << "        <entry1d" 
             << " valueX=" << Lib::sout(amanip::d2s(aObject->value(index)))
             << "/>" << Lib::endl;
      } else {
        aOut << "        <entry1d" 
             << " valueX=" << Lib::sout(amanip::d2s(aObject->value(index)))
             << " weight=" << Lib::sout(amanip::d2s(weight))
             << "/>" << Lib::endl;
      }
    }
    aOut << "    </entries1d>" << Lib::endl;
  }

  aOut << "  </cloud1d>" << Lib::endl;

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::XML::write_Cloud2D(
 const Slash::Store::IObject& aIObject
,BatchLab::XML::WriteContext& aContext
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const AIDA::ICloud2D* aObject = 
    Lib_ConstCast(aIObject,AIDA::ICloud2D);
  if(!aObject) return false;

  Lib::Out& aOut = aContext.out();
  std::string aPath = aContext.path();
  std::string aName = aContext.name();
  //int aShift = aContext.shift();

  // <cloud2d> :
  aOut << "  <cloud2d"
       << " path=" << Lib::sout(aPath)
       << " name=" << Lib::sout(aName)
       << " title=" << Lib::sout(aObject->title())
       << " maxEntries=" << Lib::sout(aObject->maxEntries())
       << ">" << Lib::endl;

  if(aObject->isConverted()) {

    const AIDA::IHistogram2D& object = aObject->histogram();
    //FIXME : aName ?
    const Slash::Store::IObject* obj = 
      Lib_ConstCast(object,Slash::Store::IObject);
    if(!obj) return false;
    write_Histogram2D(*obj,aContext);

  } else {
    // entries :
    aOut << "    <entries2d>" << Lib::endl;
    int number = aObject->entries();
    for(int index=0;index<number;index++) {
      double weight = aObject->weight(index);
      if(weight==1.) {
        aOut << "        <entry2d" 
             << " valueX=" << Lib::sout(amanip::d2s(aObject->valueX(index)))
             << " valueY=" << Lib::sout(amanip::d2s(aObject->valueY(index)))
             << "/>" << Lib::endl;
      } else {
        aOut << "        <entry2d" 
             << " valueX=" << Lib::sout(amanip::d2s(aObject->valueX(index)))
             << " valueY=" << Lib::sout(amanip::d2s(aObject->valueY(index)))
             << " weight=" << Lib::sout(amanip::d2s(weight))
             << "/>" << Lib::endl;
      }
    }
    aOut << "    </entries2d>" << Lib::endl;
  }
  
  aOut << "  </cloud2d>" << Lib::endl;

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::XML::write_Cloud3D(
 const Slash::Store::IObject& aIObject
,BatchLab::XML::WriteContext& aContext
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const AIDA::ICloud3D* aObject = 
    Lib_ConstCast(aIObject,AIDA::ICloud3D);
  if(!aObject) return false;

  Lib::Out& aOut = aContext.out();
  std::string aPath = aContext.path();
  std::string aName = aContext.name();
  //int aShift = aContext.shift();

  // <cloud3d> :
  aOut << "  <cloud3d"
       << " path=" << Lib::sout(aPath)
       << " name=" << Lib::sout(aName)
       << " title=" << Lib::sout(aObject->title())
       << " maxEntries=" << Lib::sout(aObject->maxEntries())
       << ">" << Lib::endl;

  if(aObject->isConverted()) {    

    const AIDA::IHistogram3D& object = aObject->histogram();
    //FIXME : aName ?
    //FIXME : test Lib_Cast
    const Slash::Store::IObject* obj = 
      Lib_ConstCast(object,Slash::Store::IObject);
    if(!obj) return false;
    write_Histogram3D(*obj,aContext);

  } else {
    // entries :
    aOut << "    <entries3d>" << Lib::endl;
    int number = aObject->entries();
    for(int index=0;index<number;index++) {
      double weight = aObject->weight(index);
      if(weight==1.) {
        aOut << "        <entry3d" 
             << " valueX=" << Lib::sout(amanip::d2s(aObject->valueX(index)))
             << " valueY=" << Lib::sout(amanip::d2s(aObject->valueY(index)))
             << " valueZ=" << Lib::sout(amanip::d2s(aObject->valueZ(index)))
             << "/>" << Lib::endl;
      } else {
        aOut << "        <entry3d" 
             << " valueX=" << Lib::sout(amanip::d2s(aObject->valueX(index)))
             << " valueY=" << Lib::sout(amanip::d2s(aObject->valueY(index)))
             << " valueZ=" << Lib::sout(amanip::d2s(aObject->valueZ(index)))
             << " weight=" << Lib::sout(amanip::d2s(weight))
             << "/>" << Lib::endl;
      }
    }
    aOut << "    </entries3d>" << Lib::endl;
  }

  aOut << "  </cloud3d>" << Lib::endl;

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::XML::write_DataPointSet(
 const Slash::Store::IObject& aIObject
,BatchLab::XML::WriteContext& aContext
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const AIDA::IDataPointSet* aObject = 
    Lib_ConstCast(aIObject,AIDA::IDataPointSet);
  if(!aObject) return false;

  Lib::Out& aOut = aContext.out();
  std::string aPath = aContext.path();
  std::string aName = aContext.name();
  //int aShift = aContext.shift();

  // <dataPointSet> :
  aOut << "  <dataPointSet"
       << " path=" << Lib::sout(aPath)
       << " name=" << Lib::sout(aName)
       << " title=" << Lib::sout(aObject->title())
       << " dimension=" << Lib::sout(aObject->dimension())
       << ">" << Lib::endl;

  // dataPoints :
  for(int index=0;index<aObject->size();index++) {
    const AIDA::IDataPoint* point = aObject->point(index);
    if(point) {
      aOut << "    <dataPoint>" << Lib::endl;
      for(int icoord=0;icoord<aObject->dimension();icoord++) {
        const AIDA::IMeasurement* measurement = point->coordinate(icoord);
        if(measurement) {
          aOut << "      <measurement" 
               << " value=" << Lib::sout(amanip::d2s(measurement->value()))
               << " errorMinus=" << Lib::sout(amanip::d2s(measurement->errorMinus()))
               << " errorPlus=" << Lib::sout(amanip::d2s(measurement->errorPlus()))
               << "/>" << Lib::endl;
        }
      }
      aOut << "    </dataPoint>" << Lib::endl;
    }
  }
  
  aOut << "  </dataPointSet>" << Lib::endl;

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::XML::write_Function(
 const Slash::Store::IObject& aIObject
,BatchLab::XML::WriteContext& aContext
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const AIDA::IFunction* aObject = 
    Lib_ConstCast(aIObject,AIDA::IFunction);
  if(!aObject) return false;

  Lib::Out& aOut = aContext.out();
  std::string aPath = aContext.path();
  std::string aName = aContext.name();
  //int aShift = aContext.shift();

  // <function> :
  aOut << "  <function"
       << " path=" << Lib::sout(aPath)
       << " name=" << Lib::sout(aName)
       << " title=" << Lib::sout(aObject->title())
       << ">" << Lib::endl;

  aOut << "    <codelet>" << Lib::endl;
  aOut << "      " << aObject->codeletString() << Lib::endl;
  aOut << "    </codelet>" << Lib::endl;

  aOut << "    <arguments>" << Lib::endl;
  std::vector<std::string> variables = aObject->variableNames();
 {for(unsigned int index=0;index<variables.size();index++) {
    aOut << "      <argument" 
         << " name=" << Lib::sout(variables[index]) 
         << "/>" << Lib::endl;
  }}
  aOut << "    </arguments>" << Lib::endl;

  std::vector<std::string> names = aObject->parameterNames();
  std::vector<double> values = aObject->parameters();
  if(names.size()!=values.size()) return true; //FIXME

  aOut << "    <parameters>" << Lib::endl;
 {for(unsigned int index=0;index<names.size();index++) {
    aOut << "      <parameter" 
         << " name=" << Lib::sout(names[index]) 
         << " value=" << Lib::sout(amanip::d2s(values[index])) 
         << "/>" << Lib::endl;
  }}
  aOut << "    </parameters>" << Lib::endl;
  
  aOut << "  </function>" << Lib::endl;

  return true;
}

//////////////////////////////////////////////////////////////////////////////
static void writeHistogramBin(
 Lib::Out& aOut
,const AIDA::IHistogram1D& aObject
,const std::string& aSpaces
,int aIndex
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int entries = aObject.binEntries(aIndex);
  if(entries) {
    aOut << aSpaces << "      <bin1d"
         << " binNum=" << Lib::sout(amanip::bin_to_string(aIndex))
         << " entries=" << Lib::sout(entries)
         << " height=" << Lib::sout(amanip::d2s(aObject.binHeight(aIndex)))
         << " error=" << Lib::sout(amanip::d2s(aObject.binError(aIndex)));

    double mean = aObject.binMean(aIndex);
    if(mean!=0) {
      aOut << " weightedMean=" << Lib::sout(amanip::d2s(mean));
    }

    //FIXME : use AIDA only.
    //FIXME : the below cast slow down things.
    if(const BatchLab::Histogram1D* object = 
         Lib_ConstCast(aObject,BatchLab::Histogram1D)) {
      double stddev = object->binStdDev(aIndex);
      if(stddev!=0) {
        aOut << " weightedRms=" << Lib::sout(amanip::d2s(stddev));
      }
    }

    aOut << "/>" << Lib::endl;
  }
}
