// this :
#include <BatchLab/Core/AsciiTuple.h>

#include <Lib/Out.h>
#include <Lib/smanip.h>
#include <Lib/fmanip.h>
#include <Lib/sout.h>
#include <Lib/Columns.h>
#include <Lib/Cast.h>
#include <Lib/cstring.h>

#include <Lib/FileReader.h>
#include <Lib/GZipReader.h>

#include <cstdio> //stdin

BatchLab::AsciiTuple::AsciiTuple(
 Slash::Core::IWriter& aPrinter
,const std::string& aName
,const std::string& aTitle
,const std::string& aFile
)
:MemoryTuple(aPrinter,aName)
{
  setTitle(aTitle);
  if(aFile=="stdin") {
    Slash::Core::IReader* reader = new Lib::FileReader((void*)stdin,aPrinter);
    if(!read(*reader)) clear();
    delete reader;
  } else {

    std::string suffix = Lib::smanip::suffix(aFile);
    bool compressed = (suffix=="gz"?true:false);

    std::string name = Lib::fmanip::name(aFile);
    Slash::Core::IReader* reader = 0;

    if(compressed) {
      if(!Lib::GZipReader::available()) {
        Lib::Out out(aPrinter);
        out << "BatchLab::AsciiTuple::AsciiTuple :" 
            << " compression not available. " 
            << " check installation of Lib package "
            << " and Lib::GZipReader in particular. "
            << Lib::endl;
      } else {
        reader = new Lib::GZipReader(name,aPrinter);
      }
    } else {
      reader = new Lib::FileReader(name,aPrinter);
    }
    if(!reader || !reader->enabled()) {
      Lib::Out out(aPrinter);
      out << "BatchLab::AsciiTuple::AsciiTuple :" 
          << " can't open file " << aFile << Lib::endl;
      delete reader;
    } else {
      if(!read(*reader)) clear();
      delete reader;
    }
  }
}

BatchLab::AsciiTuple::~AsciiTuple(){}

void* BatchLab::AsciiTuple::cast(const std::string& aClass) const {
  if_Lib_SCast(BatchLab::AsciiTuple)
  else return BatchLab::MemoryTuple::cast(aClass);
}

std::string BatchLab::AsciiTuple::storeClassName() const {
  return "BatchLab::AsciiTuple";
}

bool BatchLab::AsciiTuple::read(Slash::Core::IReader& aReader) {
#define BUFSIZE 65536
  char buffer[BUFSIZE+1];

  // Read header :

  bool hippoStyle = false;
  bool readFirstRow = false;
  bool readSecondRow = false;

  // First line :
  unsigned int l;
  if(!aReader.sread(buffer,sizeof(buffer),l)) return false; //EOF.

  //printf("debug : BatchLab::AsciiTuple::read : line \"%s\"\n",buffer);

  std::vector<double> ds;
  // Guess the style of the file (ascii or hippodraw) :
  Lib_strtods(buffer,ds);
  unsigned int dn = ds.size();
  if(dn) {
    // First line is numbers. The style is assumed to be "ascii".
    std::string sbooking;
    for(unsigned int index=0;index<dn;index++) {
      std::string s;
      Lib::smanip::printf(s,32,"v%d",index);
      if(index) sbooking += " ";
      sbooking += s;
    }    
    std::list<Lib::Value> variables;
    if(!booking(printer(),sbooking,variables)) return false;
    bool status = initialize(variables);
    Lib::Columns::deleteColumns(variables);
    if(!status) return false;

  } else {
    // Expect hippodraw style, then a title line followed
    // with a line describing columns and then data.
    setTitle(buffer);

    // Read second line :
    if(!aReader.sread(buffer,sizeof(buffer),l)) return false; //EOF.
    // Second line is the columns describer :
    std::list<Lib::Value> variables;
    if(!booking(printer(),buffer,variables)) return false;
    bool status = initialize(variables);
    Lib::Columns::deleteColumns(variables);
    if(!status) return false;
  }

  // At this point, the tuple is booked, then read data ; full speed...
  ds.resize(columns());

  while(true) {
    if(!aReader.sread(buffer,sizeof(buffer),l)) break; //EOF.

    //printf("debug : BatchLab::AsciiTuple::read : line \"%s\"\n",buffer);

    Lib_strtods(buffer,ds,true); //the last true is important.

    if(!fill(ds)) {
      Lib::Out out(printer());
      out << "BatchLab::AsciiTuple::read :"
          << " can't fill." << Lib::endl;
      return false;
    }

  }

  return true;
}
