// this :
#include <BatchLab/Core/BaseTuple.h>

#include <amanip/filters.h>
#include <amanip/project.h>
#include <amanip/tuple.h>

#include <Lib/Out.h>
#include <Lib/Debug.h>
#include <Lib/Processor.h>
#include <Lib/Columns.h>
#include <Lib/smanip.h>
#include <Lib/sout.h>
#include <Lib/Cast.h>
#include <Lib/valuemanip.h>

#include <Lib/StreamBuffer.h>
#include <iostream>

#include <BatchLab/Core/Filter.h>
#include <BatchLab/Core/Evaluator.h>

#define NotFound (-1)
#define MINIMUM(a,b) ((a)<(b)?a:b)
#define MAXIMUM(a,b) ((a)>(b)?a:b)

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
BatchLab::BaseColumn::BaseColumn(
 Slash::Core::IWriter& aPrinter
,const std::string& aLabel
,const std::string& aType
)
:fPrinter(aPrinter)
,fLabel(aLabel)
,fType(aType)
,fRow(-1)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("BatchLab::BaseColumn");
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::BaseColumn::BaseColumn(
 const BaseColumn& aFrom
)
:fPrinter(aFrom.fPrinter)
,fLabel(aFrom.fLabel)
,fType(aFrom.fType)
,fRow(-1)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("BatchLab::BaseColumn");
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::BaseColumn::~BaseColumn(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::decrement("BatchLab::BaseColumn");
}
//////////////////////////////////////////////////////////////////////////////
void BatchLab::BaseColumn::setRow(
 int aRow
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fRow = aRow;
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::BaseColumn::name(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fLabel;
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::BaseColumn::type(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fType;
}
//////////////////////////////////////////////////////////////////////////////
void* BatchLab::BaseColumn::cast(
 const std::string& aClass
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if_Lib_SCast(BatchLab::BaseColumn)
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
  else if_Lib_SCast(AIDA::IBaseTupleColumn)
#endif
  else return 0;
}
//////////////////////////////////////////////////////////////////////////////
double BatchLab::BaseColumn::minimum(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = size();
  if(!number) return 0; //FIXME : throw.
  Lib::Value val;
  double d;
  if(!value(0,val)) {} //FIXME :throw
  if(!Lib::valuemanip::to_double(val,d)) {} //FIXME : throw
  double mn = d;
  for(unsigned int index=1;index<number;index++) {
    if(!value(index,val)) {} //FIXME :throw
    if(!Lib::valuemanip::to_double(val,d)) {} //FIXME : throw
    mn = MINIMUM(d,mn);
  }
  return mn;
}
//////////////////////////////////////////////////////////////////////////////
double BatchLab::BaseColumn::maximum(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = size();
  if(!number) return 0; //FIXME : throw.
  Lib::Value val;
  double d;
  if(!value(0,val)) {} //FIXME :throw
  if(!Lib::valuemanip::to_double(val,d)) {} //FIXME : throw
  double mx = d;
  for(unsigned int index=1;index<number;index++) {
    if(!value(index,val)) {} //FIXME :throw
    if(!Lib::valuemanip::to_double(val,d)) {} //FIXME : throw
    mx = MAXIMUM(d,mx);
  }
  return mx;
}
//////////////////////////////////////////////////////////////////////////////
double BatchLab::BaseColumn::mean(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = size();
  if(!number) return 0; //FIXME : throw.
  Lib::Value val;
  double d;
  if(!value(0,val)) {} //FIXME :throw
  if(!Lib::valuemanip::to_double(val,d)) {} //FIXME : throw
  double sum = d;
  for(unsigned int index=1;index<number;index++) {
    if(!value(index,val)) {} //FIXME :throw
    if(!Lib::valuemanip::to_double(val,d)) {} //FIXME : throw
    sum += d;
  }
  return sum/double(number);
}
//////////////////////////////////////////////////////////////////////////////
double BatchLab::BaseColumn::rms(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = size();
  if(!number) return 0; //FIXME : throw.
  Lib::Value val;
  double d;
  if(!value(0,val)) {} //FIXME :throw
  if(!Lib::valuemanip::to_double(val,d)) {} //FIXME : throw
  double sum = d;
  double sum2 = d*d;
  for(unsigned int index=1;index<number;index++) {
    if(!value(index,val)) {} //FIXME :throw
    if(!Lib::valuemanip::to_double(val,d)) {} //FIXME : throw
    sum += d;
    sum2 += d*d;
  }
  double mean = sum/double(number);
  return ::sqrt(::fabs(sum2/double(number) - mean * mean));
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
BatchLab::BaseTuple::BaseTuple(
 Slash::Core::IWriter& aPrinter
,const std::string& aName
,const std::string& aTitle
)
:BaseObject(aPrinter,aName)
,fRow(-1)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setTitle(aTitle);
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::BaseTuple::BaseTuple(
 const BaseTuple& aFrom
)
:AIDA::IManagedObject(aFrom)
,Slash::Store::IObject(aFrom)
,AIDA::ITuple(aFrom)
,AIDA::ITupleEntry(aFrom)
,BaseObject(aFrom)
,fAnnotation(aFrom.fAnnotation)
,fTitle(aFrom.fTitle)
,fRow(-1)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = aFrom.fColumns.size();
  for(unsigned int index=0;index<number;index++) {
    BaseColumn* column = aFrom.fColumns[index]->copy();
    if(!column) {
      Lib::Out out(printer());
      out << "BatchLab::BaseTuple::BaseTuple :"
          << " can't copy column."
          << Lib::endl;
      fColumns.clear();
      return;
    }
    fColumns.push_back(column);
  }
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::BaseTuple::~BaseTuple(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  clear();
}
//////////////////////////////////////////////////////////////////////////////
void* BatchLab::BaseTuple::cast(
 const std::string& aClass
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if_Lib_SCast(AIDA::ITuple)
  else if_Lib_SCast(AIDA::ITupleEntry)
  else if_Lib_SCast(BatchLab::BaseTuple)
  else return BaseObject::cast(aClass);
}

static bool is_column_type(const std::string& aString) {
  std::string type = aString;
  Lib::smanip::tolowercase(type);
  if( (type=="float") ||
      (type=="double") ||
      (type=="char") ||
      (type=="short") ||
      (type=="int") ||
      (type=="long") ||
      (type=="boolean") ||
      (type=="string") ||
      (type=="ITuple") ) return true;
  return false;
}

//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::booking(
 Slash::Core::IWriter& aPrinter
,const std::string& aString
,std::list<Lib::Value>& aVariables
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aVariables.clear();

  if(aString.empty()) return true; //This happens when reading a tuple from a file.

  int verboseLevel = 0;
  if(verboseLevel) {
    Lib::Out out(aPrinter);
    out << "BatchLab::BaseTuple::booking :"
        << " parse " << Lib::sout(aString) << "."
        << Lib::endl;
  }

  //printf("debug : BatchLab::BaseTuple::booking : \"%s\".\n",aString.c_str());

  if(aString.find(";")!=std::string::npos) {

    std::string snew = aString;
    Lib::smanip::replace(snew,";",",");

    Lib::Columns proc(snew,aPrinter,verboseLevel);
    if(!proc.findVariables()) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseTuple::booking :"
          << " Bad syntax in " << Lib::sout(aString) << "."
          << Lib::endl;
      return false;
    }
    aVariables = proc.result();

  } else if(aString.find(",")==std::string::npos) {
    if(aString.find("\t")!=std::string::npos) {
      // Handle hippodraw format "col1\tcol2\tcol3".
      // "double" will be the type by default.
      std::vector<std::string> names;
      Lib::smanip::words(aString,"\t",names);
      Lib::smanip::strip(names);
      Lib::smanip::replace(names," ","_");

      Lib::Value v;
      v.setType(Slash::Core::IValue::DOUBLE);
      unsigned int namen = names.size();
      for(unsigned int index=0;index<namen;index++) {
        //printf("debug : %d \"%s\"\n",index,aNames[index].c_str());
        v.setLabel(names[index]);
        aVariables.push_back(v);
      }
    } else {
      std::vector<std::string> names;
      Lib::smanip::words(aString," ",names);
      if( ( (names.size()==2) || (names.size()==4) ) && 
          (is_column_type(names[0])) ) {
        // Could be one column at new format :
        //   <type> <name>
        //   <type> <name> = <value>
        Lib::Columns proc(aString,aPrinter,verboseLevel);
        if(!proc.findVariables()) {
          Lib::Out out(aPrinter);
          out << "BatchLab::BaseTuple::booking :"
              << " Bad syntax in " << Lib::sout(aString) << "."
              << Lib::endl;
          return false;
        }
        aVariables = proc.result();
      } else {
        // Old format is "col1 col2 col3".
        // "double" will be the type by default.
        Lib::Value v;
        v.setType(Slash::Core::IValue::DOUBLE);
        unsigned int namen = names.size();
        for(unsigned int index=0;index<namen;index++) {
          //printf("debug : %d \"%s\"\n",index,aNames[index].c_str());
          v.setLabel(names[index]);
          aVariables.push_back(v);
        }
      }
    }

  } else {
    Lib::Columns proc(aString,aPrinter,verboseLevel);
    if(!proc.findVariables()) {
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseTuple::booking :"
          << " Bad syntax in " << Lib::sout(aString) << "."
          << Lib::endl;
      return false;
    }
    aVariables = proc.result();
  }
  if(!aVariables.size()) {
    Lib::Out out(aPrinter);
    out << "BatchLab::BaseTuple::booking :"
        << " Bad syntax in " << Lib::sout(aString) << "."
        << Lib::endl;
    return false;
  }

  if(verboseLevel) {
    int index = 0;
    std::list<Lib::Value>::iterator it;
    for(it=aVariables.begin();it!=aVariables.end();++it,index++){
      Lib::Value& var = (*it);
      Lib::Out out(aPrinter);
      out << "BatchLab::BaseTuple::booking :"
          << " variable " << index 
          <<  " (" << (unsigned int)aVariables.size() << ")"
          << " name = " << Lib::sout(var.label())
          << " type = " << Lib::sout(var.stype()) << "."
          << Lib::endl;
    }
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::BaseTuple::convertType(
 Slash::Core::IValue::Type aType
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  switch(aType) {
  case Slash::Core::IValue::SHORT: return "short";
  case Slash::Core::IValue::INT: return "int";
  case Slash::Core::IValue::INT64: return "long"; //and not int64.
  case Slash::Core::IValue::DOUBLE:  return "double";
  //AIDA3.3 : case BOOLEAN: return "bool";
  case Slash::Core::IValue::BOOL: return "boolean";
  case Slash::Core::IValue::FLOAT: return "float";
  case Slash::Core::IValue::UNSIGNED_CHAR: return "byte";
  case Slash::Core::IValue::CHAR: return "char";
  case Slash::Core::IValue::STRING: return "string";
  case Slash::Core::IValue::VOID_STAR: return "ITuple";

  case Slash::Core::IValue::ARRAY_DOUBLE: return "double[]";

  case Slash::Core::IValue::UNSIGNED_INT:
  case Slash::Core::IValue::UNSIGNED_INT64:
  default:
    Lib::Out out(printer());
    out << "BatchLab::BaseTuple::convertType :"
        << " unhandle Lib::Variable type " << ((int)aType) << "."
        << Lib::endl;
    return "";
  }
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::BaseTuple::type(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return "AIDA::ITuple";
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::BaseTuple::title(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fTitle;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::setTitle(
 const std::string& aTitle
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fTitle = aTitle;
  fAnnotation.setValue("Title",aTitle);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IAnnotation& BatchLab::BaseTuple::annotation(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fAnnotation;
}
//////////////////////////////////////////////////////////////////////////////
const AIDA::IAnnotation& BatchLab::BaseTuple::annotation(
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fAnnotation;
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::Annotation& BatchLab::BaseTuple::getAnnotation(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fAnnotation;
}
//////////////////////////////////////////////////////////////////////////////
void BatchLab::BaseTuple::setAnnotation(
 const BatchLab::Annotation& aFrom
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fAnnotation = aFrom;
}
//////////////////////////////////////////////////////////////////////////////
int BatchLab::BaseTuple::findColumn(
 const std::string& aName
) const
//////////////////////////////////////////////////////////////////////////////
// Not case sensitive (for OpenPAW).
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = fColumns.size();
  // case sensitive :
  //for(unsigned int index=0;index<number;index++) {
  //  if(fColumns[index]->label()==aName) return index;
  //}
  // Not case sensitive :
  std::string as = aName;
  Lib::smanip::touppercase(as);
  for(unsigned int index=0;index<number;index++) {
    std::string s = fColumns[index]->name();
    Lib::smanip::touppercase(s);
    if(s==as) return index;
  }
  return NotFound;
}
//////////////////////////////////////////////////////////////////////////////
int BatchLab::BaseTuple::rowIndex(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  return fRow;
}
bool BatchLab::BaseTuple::close() {return true;}
//////////////////////////////////////////////////////////////////////////////
void BatchLab::BaseTuple::start(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  fRow = -1;
  setColumnsRow(fRow);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::next(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  if((fRow+1)>=rows()) return false;
  fRow++;
  setColumnsRow(fRow);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::skip(
 int aRows
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  if((fRow+aRows)>=rows()) return false;
  fRow += aRows;
  setColumnsRow(fRow);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::setRow(
 int aRow
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  if(aRow<0) return false;
  if(aRow>=rows()) return false;
  fRow = aRow;
  setColumnsRow(fRow);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
int BatchLab::BaseTuple::columns(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fColumns.size();
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::BaseTuple::columnName(
 int aIndex
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  BaseColumn* column = findColumn1(aIndex);
  if(!column) return "";
  return column->name();
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> BatchLab::BaseTuple::columnNames(
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = fColumns.size();
  std::vector<std::string> ss(number);
  for(unsigned int index=0;index<number;index++) {
    ss[index] = fColumns[index]->name();
  }
  return ss;
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::BaseTuple::columnType(
 int aIndex
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  BaseColumn* column = findColumn1(aIndex);
  if(!column) return "";
  return column->type();
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> BatchLab::BaseTuple::columnTypes(
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = fColumns.size();
  std::vector<std::string> ss(number);
  for(unsigned int index=0;index<number;index++) {
    ss[index] = fColumns[index]->type();
  }
  return ss;
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::BaseColumn* BatchLab::BaseTuple::findColumn1(
 int aIndex
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int number = fColumns.size();
  if((aIndex<0)||(aIndex>=number)) return 0;
  return fColumns[aIndex];
}
//////////////////////////////////////////////////////////////////////////////
void BatchLab::BaseTuple::clear(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = fColumns.size();
  for(unsigned int index=0;index<number;index++) delete fColumns[index];
  fColumns.clear();
  fRow = -1;
  setColumnsRow(fRow);
}
//////////////////////////////////////////////////////////////////////////////
void BatchLab::BaseTuple::resetRow(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = fColumns.size();
  for(unsigned int index=0;index<number;index++) 
    fColumns[index]->resetFilled();
}
//////////////////////////////////////////////////////////////////////////////
void BatchLab::BaseTuple::reset(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  unsigned int number = fColumns.size();
  for(unsigned int index=0;index<number;index++) fColumns[index]->reset();
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::BaseTuple::columnDefaultValue(
 int aIndex
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  BaseColumn* column = findColumn1(aIndex);
  if(!column) return "";
  return column->stringDefaultValue();
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::isDefaultDefault(
 int aIndex
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  BaseColumn* column = findColumn1(aIndex);
  if(!column) return false;
  return column->isDefaultDefault();
}
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
//////////////////////////////////////////////////////////////////////////////
AIDA::IBaseTupleColumn* BatchLab::BaseTuple::column(
 const std::string& aName
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<BaseColumn*>::const_iterator it;
  for(it=fColumns.begin();it!=fColumns.end();++it) {
    if(aName==(*it)->name()) {
      return Lib_Cast(*(*it),AIDA::IBaseTupleColumn);
    }   
  }
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IBaseTupleColumn* BatchLab::BaseTuple::column(
 int aIndex
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  BaseColumn* column = findColumn1(aIndex);
  if(!column) return 0;
  return Lib_Cast(*column,AIDA::IBaseTupleColumn);
}
#endif
//////////////////////////////////////////////////////////////////////////////
/// Projections //////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
#include <AIDA/IHistogram3D.h>

#include <AIDA/ICloud1D.h>
#include <AIDA/ICloud2D.h>
#include <AIDA/ICloud3D.h>

#include <AIDA/IProfile1D.h>
#include <AIDA/IProfile2D.h>
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::IHistogram1D& aFillable
,AIDA::IEvaluator& aEvaluator
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  ::amanip::accept_all filter;
  return amanip::project<AIDA::IHistogram1D>
           (out,*this,aFillable,aEvaluator,filter);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::IHistogram1D& aFillable
,AIDA::IEvaluator& aEvaluator
,AIDA::IFilter& aFilter
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  return amanip::project<AIDA::IHistogram1D>
           (out,*this,aFillable,aEvaluator,aFilter);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::IHistogram2D& aFillable
,AIDA::IEvaluator& aEvaluatorX
,AIDA::IEvaluator& aEvaluatorY
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  ::amanip::accept_all filter;
  return amanip::project<AIDA::IHistogram2D>
           (out,*this,aFillable,aEvaluatorX,aEvaluatorY,filter);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::IHistogram2D& aFillable
,AIDA::IEvaluator& aEvaluatorX
,AIDA::IEvaluator& aEvaluatorY
,AIDA::IFilter& aFilter
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  return amanip::project<AIDA::IHistogram2D>
           (out,*this,aFillable,aEvaluatorX,aEvaluatorY,aFilter);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::IHistogram3D& aFillable
,AIDA::IEvaluator& aEvaluatorX
,AIDA::IEvaluator& aEvaluatorY
,AIDA::IEvaluator& aEvaluatorZ
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  ::amanip::accept_all filter;
  return amanip::project<AIDA::IHistogram3D>
           (out,*this,aFillable,
              aEvaluatorX,aEvaluatorY,aEvaluatorZ,filter);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::IHistogram3D& aFillable
,AIDA::IEvaluator& aEvaluatorX
,AIDA::IEvaluator& aEvaluatorY
,AIDA::IEvaluator& aEvaluatorZ
,AIDA::IFilter& aFilter
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  return amanip::project<AIDA::IHistogram3D>
           (out,*this,aFillable,
              aEvaluatorX,aEvaluatorY,aEvaluatorZ,aFilter);
}
bool BatchLab::BaseTuple::project(AIDA::IHistogram1D&,
                             AIDA::IEvaluator&,
                             AIDA::IFilter&,
                             AIDA::IEvaluator&) {return false;}
bool BatchLab::BaseTuple::project(AIDA::IHistogram1D&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&) {return false;}
bool BatchLab::BaseTuple::project(AIDA::IHistogram2D&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IFilter&,
                             AIDA::IEvaluator&) {return false;}
bool BatchLab::BaseTuple::project(AIDA::IHistogram2D&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&) {return false;}
bool BatchLab::BaseTuple::project(AIDA::IHistogram3D&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&) {return false;}
bool BatchLab::BaseTuple::project(AIDA::IHistogram3D&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IFilter&,
                             AIDA::IEvaluator&) {return false;}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::ICloud1D& aFillable
,AIDA::IEvaluator& aEvaluator
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  ::amanip::accept_all filter;
  return amanip::project<AIDA::ICloud1D>
           (out,*this,aFillable,aEvaluator,filter);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::ICloud1D& aFillable
,AIDA::IEvaluator& aEvaluator
,AIDA::IFilter& aFilter
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  return amanip::project<AIDA::ICloud1D>
           (out,*this,aFillable,aEvaluator,aFilter);
}
bool BatchLab::BaseTuple::project(AIDA::ICloud1D&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&) {return false;}
bool BatchLab::BaseTuple::project(AIDA::ICloud1D&,
                             AIDA::IEvaluator&,
                             AIDA::IFilter&,
                             AIDA::IEvaluator&) {return false;}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::ICloud2D& aFillable
,AIDA::IEvaluator& aEvaluatorX
,AIDA::IEvaluator& aEvaluatorY
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  ::amanip::accept_all filter;
  return amanip::project<AIDA::ICloud2D>
           (out,*this,aFillable,aEvaluatorX,aEvaluatorY,filter);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::ICloud2D& aFillable
,AIDA::IEvaluator& aEvaluatorX
,AIDA::IEvaluator& aEvaluatorY
,AIDA::IFilter& aFilter
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  return amanip::project<AIDA::ICloud2D>
           (out,*this,aFillable,aEvaluatorX,aEvaluatorY,aFilter);
}
bool BatchLab::BaseTuple::project(AIDA::ICloud2D&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&) {return false;}
bool BatchLab::BaseTuple::project(AIDA::ICloud2D&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IFilter&,
                             AIDA::IEvaluator&) {return false;}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::ICloud3D& aFillable
,AIDA::IEvaluator& aEvaluatorX
,AIDA::IEvaluator& aEvaluatorY
,AIDA::IEvaluator& aEvaluatorZ
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  ::amanip::accept_all filter;
  return amanip::project<AIDA::ICloud3D>
           (out,*this,aFillable,
             aEvaluatorX,aEvaluatorY,aEvaluatorZ,filter);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::ICloud3D& aFillable
,AIDA::IEvaluator& aEvaluatorX
,AIDA::IEvaluator& aEvaluatorY
,AIDA::IEvaluator& aEvaluatorZ
,AIDA::IFilter& aFilter
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  return amanip::project<AIDA::ICloud3D>
           (out,*this,aFillable,
             aEvaluatorX,aEvaluatorY,aEvaluatorZ,aFilter);
}
bool BatchLab::BaseTuple::project(AIDA::ICloud3D&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&) {return false;}
bool BatchLab::BaseTuple::project(AIDA::ICloud3D&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IFilter&,
                             AIDA::IEvaluator&) {return false;}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::IProfile1D& aFillable
,AIDA::IEvaluator& aEvaluatorX
,AIDA::IEvaluator& aEvaluatorY
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  ::amanip::accept_all filter;
  return amanip::project<AIDA::IProfile1D>
           (out,*this,aFillable,aEvaluatorX,aEvaluatorY,filter);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::IProfile1D& aFillable
,AIDA::IEvaluator& aEvaluatorX
,AIDA::IEvaluator& aEvaluatorY
,AIDA::IFilter& aFilter
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  return amanip::project<AIDA::IProfile1D>
           (out,*this,aFillable,aEvaluatorX,aEvaluatorY,aFilter);
}
bool BatchLab::BaseTuple::project(AIDA::IProfile1D&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&) {return false;}
bool BatchLab::BaseTuple::project(AIDA::IProfile1D&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IFilter&,
                             AIDA::IEvaluator&) {return false;}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::IProfile2D& aFillable
,AIDA::IEvaluator& aEvaluatorX
,AIDA::IEvaluator& aEvaluatorY
,AIDA::IEvaluator& aEvaluatorZ
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  ::amanip::accept_all filter;
  return amanip::project<AIDA::IProfile2D>
           (out,*this,aFillable,
              aEvaluatorX,aEvaluatorY,aEvaluatorZ,filter);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::BaseTuple::project(
 AIDA::IProfile2D& aFillable
,AIDA::IEvaluator& aEvaluatorX
,AIDA::IEvaluator& aEvaluatorY
,AIDA::IEvaluator& aEvaluatorZ
,AIDA::IFilter& aFilter
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  return amanip::project<AIDA::IProfile2D>
           (out,*this,aFillable,
              aEvaluatorX,aEvaluatorY,aEvaluatorZ,aFilter);
}
bool BatchLab::BaseTuple::project(AIDA::IProfile2D&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&) {return false;}
bool BatchLab::BaseTuple::project(AIDA::IProfile2D&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IEvaluator&,
                             AIDA::IFilter&,
                             AIDA::IEvaluator&) {return false;}
//////////////////////////////////////////////////////////////////////////////
double BatchLab::BaseTuple::evaluateMin(
 AIDA::IEvaluator& aEval
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  double xmin,xmax;
  ::amanip::accept_all filter;
  amanip::min_max(out,*this,aEval,filter,xmin,xmax);
  return xmin;
} 
//////////////////////////////////////////////////////////////////////////////
double BatchLab::BaseTuple::evaluateMin(
 AIDA::IEvaluator& aEval
,AIDA::IFilter& aFilter
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  double xmin,xmax;
  amanip::min_max(out,*this,aEval,aFilter,xmin,xmax);
  return xmin;
} 
//////////////////////////////////////////////////////////////////////////////
double BatchLab::BaseTuple::evaluateMax(
 AIDA::IEvaluator& aEval
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  double xmin,xmax;
  ::amanip::accept_all filter;
  amanip::min_max(out,*this,aEval,filter,xmin,xmax);
  return xmax;
} 
//////////////////////////////////////////////////////////////////////////////
double BatchLab::BaseTuple::evaluateMax(
 AIDA::IEvaluator& aEval
,AIDA::IFilter& aFilter
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(printer());
  std::ostream out(&sbuf);
  double xmin,xmax;
  amanip::min_max(out,*this,aEval,aFilter,xmin,xmax);
  return xmax;
} 
//////////////////////////////////////////////////////////////////////////////
void BatchLab::BaseTuple::setColumnsRow(
 int aRow
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<BaseColumn*>::iterator it;
  for(it=fColumns.begin();it!=fColumns.end();++it) {
    (*it)->setRow(aRow);   
  }
}
