// this :
#include <BatchLab/Core/MemoryTuple.h>

#include <amanip/tuple.h>

#include <Lib/Out.h>
#include <Lib/smanip.h>
#include <Lib/sout.h>
#include <Lib/Columns.h>
#include <Lib/Cast.h>
#include <Lib/StreamBuffer.h>
#include <iostream>

#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#define AIDA_int64 Slash::int64
#else
#define AIDA_int64 AIDA::int64
#endif

#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
#include <AIDA/ITupleColumn.h>
#endif

#include <math.h>

#define NotFound (-1)
#define MINIMUM(a,b) ((a)<(b)?a:b)
#define MAXIMUM(a,b) ((a)>(b)?a:b)

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

typedef unsigned char byte;

namespace BatchLab {

template <class T> 
class MemoryColumn 
:public BaseColumn 
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
,public virtual AIDA::ITupleColumn<T>
#endif
{
public: //BaseColumn
  virtual unsigned int size() const { return fData.size();}
  virtual void add() {
    fData.push_back(*fFilled);
    *fFilled = fDefault;
  }
  virtual bool append(void* aData) {
    T* data = (T*)aData;
    fData.push_back(*data);
    return true;
  }
  virtual void reset() { clear(); }
  virtual void resetFilled() {
    *fFilled = fDefault;
  }
  virtual std::string stringDefaultValue() const { 
    return Lib::smanip::tostring(fDefault);
  }
  virtual bool isDefaultDefault() const { 
    return (fDefault==fDefaultDefault?true:false);
  }
  virtual bool value(int aIndex,Lib::Value& aVariable) const {
    T v;
    bool status = MemoryColumn::value(aIndex,v);
    aVariable.set(v);
    return status;
  }
public: //AIDA::ITupleColumn
  virtual bool fill(const T& aValue) {
    if(!fBinded) {
      *fFilled = aValue;
      return true;
    } else { 
      return false;
    }
  }
  virtual T& value() const { //FIXME : should not be const
    Lib_self(MemoryColumn);
    if((fRow<0)||(fRow>=fData.size())) {
      return self.fDefault;
    } else {
      //FIXME : error with [T=bool]
      //return self.fData[fRow];
      self.fTmp = self.fData[fRow]; //FIXME
      return self.fTmp;
    }
  }
  virtual bool bind(T* aVariable) {
    if(!fBinded) delete fFilled;
    fFilled = aVariable;
    fBinded = true;
    return true;
  }
  virtual const T& defaultValue() const {
    return fDefault;
  }
  virtual T* fillableObject() const {
    return fFilled;
  }
public:
  bool visit_base(Slash::Store::IConstVisitor& aVisitor) const {
    int version = 1;
    if(!aVisitor.visit("fVersion",version)) return false;
    if(!aVisitor.visit("fLabel",fLabel)) return false;
    if(!aVisitor.visit("fType",fType)) return false;
    if(!aVisitor.visit("fData",fData)) return false;
    return true;
  }
  bool read_base(Slash::Store::IVisitor& aVisitor) {
    int version;
    if(!aVisitor.visit(version)) return false;
    if(!aVisitor.visit(fLabel)) return false;
    if(!aVisitor.visit(fType)) return false;
    if(!aVisitor.visit(fData)) return false;
    return true;
  }
public:
  MemoryColumn(Slash::Core::IWriter& aPrinter,
               const std::string& aLabel,
               const std::string& aType,
               const T& aDefaultDefault,
               const T& aDefault)
  :BaseColumn(aPrinter,aLabel,aType)
  ,fDefaultDefault(aDefaultDefault)
  ,fDefault(aDefault)
  ,fFilled(0)
  ,fBinded(false){
    fFilled = new T;
    *fFilled = fDefault;    
  }
  MemoryColumn(const MemoryColumn<T>& aFrom)
  :BaseColumn(aFrom)
  ,fData(aFrom.fData)
  ,fDefaultDefault(aFrom.fDefaultDefault)
  ,fDefault(aFrom.fDefault)
  ,fFilled(0)
  ,fBinded(false){
    fFilled = new T;
    *fFilled = aFrom.fDefault;
  }
  virtual ~MemoryColumn() { 
    clear();
    if(!fBinded) delete fFilled;
  }
public:
  virtual bool value(int aIndex,T& aValue) const {
    if((aIndex<0)||(aIndex>=fData.size())) {
      aValue = fDefault;
      return false;
    } else {
      //FIXME : g++ warning with [T=bool] :  aValue = fData[aIndex];
      aValue = fData[aIndex];
      return true;
    }
  }
private:
  void clear() { fData.clear();}
protected:
  std::vector<T> fData;
  T fDefaultDefault;
  T fDefault;
  T* fFilled;
  bool fBinded;
  T fTmp; //FIXME
};


#define VISIT_MEMBERS(aClass)\
private:\
  static bool visit(const Slash::Store::IStorable& aObject,\
                         Slash::Store::IConstVisitor& aVisitor){\
    const aClass* local = dynamic_cast<const aClass*>(&aObject);\
    if(!local) return false;\
    return local->aClass::visit(aVisitor);\
  }\
public:\
  virtual bool visit(\
    Slash::Store::IConstVisitor& aVisitor) const {\
    if(!aVisitor.begin(*this,#aClass,aClass::visit)) return false;\
    if(!visit_base(aVisitor)) return false;\
    if(!aVisitor.end(*this)) return false;\
    return true;\
  }

#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2)

#define MEMORY_COLUMN_SOURCE(aClassName,aType,aSType) \
class aClassName\
:public MemoryColumn< aType >\
,public virtual Slash::Store::IStorable\
{\
public:\
  virtual void* cast(const std::string& aClass) const {\
    if_Lib_SCast(Slash::Store::IStorable)\
    else if_Lib_SCast(BatchLab::aClassName)\
    else return BatchLab::BaseColumn::cast(aClass);\
  }\
  virtual std::string stringDefaultValue() const {\
    return Lib::smanip::tostring(fDefault);\
  }\
  virtual bool isDefaultDefault() const {\
    return (fDefault==fDefaultDefault?true:false);\
  }\
  virtual BaseColumn* copy() const { return new aClassName(*this);}\
  virtual bool read(Slash::Store::IVisitor& aVisitor) {\
    if(!aVisitor.begin(*this)) return false;\
    if(!read_base(aVisitor)) return false;\
    if(!aVisitor.end(*this)) return false;\
    return true;\
  }\
public:\
  aClassName(Slash::Core::IWriter& aPrinter,const std::string& aLabel,\
             aType aDefaultDefault,aType aDefault)\
  :MemoryColumn< aType >(aPrinter,aLabel,#aSType,aDefaultDefault,aDefault){}\
  aClassName(const aClassName& aFrom)\
  :Slash::Store::IStorable(aFrom)\
  ,MemoryColumn< aType >(aFrom){}\
  virtual ~aClassName() {}\
private:\
  VISIT_MEMBERS(aClassName)\
};

#else //AIDA_VERSION
// The only difference with the upper is in the cast
// method to cast on AIDA::ITupleColumn.

#define MEMORY_COLUMN_SOURCE(aClassName,aType,aSType) \
class aClassName\
:public MemoryColumn< aType >\
,public virtual Slash::Store::IStorable\
{\
public:\
  virtual void* cast(const std::string& aClass) const {\
    if_Lib_SCast(Slash::Store::IStorable)\
    else if(aClass==std::string("AIDA::ITupleColumn<")+#aType+">") {\
      typedef AIDA::ITupleColumn< aType > ColType;\
      return (void*)static_cast<const ColType*>(this);\
    }\
    else if_Lib_SCast(BatchLab::aClassName)\
    else return BatchLab::BaseColumn::cast(aClass);\
  }\
  virtual std::string stringDefaultValue() const {\
    return Lib::smanip::tostring(fDefault);\
  }\
  virtual bool isDefaultDefault() const {\
    return (fDefault==fDefaultDefault?true:false);\
  }\
  virtual BaseColumn* copy() const { return new aClassName(*this);}\
  virtual bool read(Slash::Store::IVisitor& aVisitor) {\
    if(!aVisitor.begin(*this)) return false;\
    if(!read_base(aVisitor)) return false;\
    if(!aVisitor.end(*this)) return false;\
    return true;\
  }\
public:\
  aClassName(Slash::Core::IWriter& aPrinter,const std::string& aLabel,\
             aType aDefaultDefault,aType aDefault)\
  :MemoryColumn< aType >(aPrinter,aLabel,#aSType,aDefaultDefault,aDefault){}\
  aClassName(const aClassName& aFrom)\
  :Slash::Store::IStorable(aFrom)\
  ,MemoryColumn< aType >(aFrom){}\
  virtual ~aClassName() {}\
private:\
  VISIT_MEMBERS(aClassName)\
};

#endif

MEMORY_COLUMN_SOURCE(MemoryColumn_float,float,float)
MEMORY_COLUMN_SOURCE(MemoryColumn_double,double,double)
MEMORY_COLUMN_SOURCE(MemoryColumn_char,char,char)
MEMORY_COLUMN_SOURCE(MemoryColumn_byte,byte,byte)
MEMORY_COLUMN_SOURCE(MemoryColumn_bool,bool,boolean)
MEMORY_COLUMN_SOURCE(MemoryColumn_short,short,short)
MEMORY_COLUMN_SOURCE(MemoryColumn_int,int,int)
MEMORY_COLUMN_SOURCE(MemoryColumn_int64,AIDA_int64,long)
MEMORY_COLUMN_SOURCE(MemoryColumn_string,std::string,string)
MEMORY_COLUMN_SOURCE(MemoryColumn_array_double,std::vector<double>,double[])

class MemoryColumn_tuple 
:public BaseColumn 
,public virtual Slash::Store::IStorable
{
public: //BaseColumn
  virtual unsigned int size() const {return fData.size();}
  virtual void add() { 
    if(!fTemplate) return;
    fData.push_back(fTemplate);
    fTemplate = 0;
  }
  virtual bool append(void* aData) { //Used in HDF5_DataReader.
    AIDA::ITuple* data = (AIDA::ITuple*)aData;
    BatchLab::MemoryTuple* tuple = Lib_Cast(*data,BatchLab::MemoryTuple);
    if(!tuple) return false;
    fData.push_back(tuple);
    return true;
  }
  virtual void reset() {
    clear();
    if(fTemplate) delete fTemplate;
    fTemplate = 0;
  }
  virtual BaseColumn* copy() const {return new MemoryColumn_tuple(*this);}
  virtual void resetFilled() {
    if(!fTemplate) return;
    fTemplate->reset();
  }
  virtual void* cast(const std::string& aClass) const {
    if_Lib_SCast(Slash::Store::IStorable)
    else if_Lib_SCast(BatchLab::MemoryColumn_tuple)
    else return BatchLab::BaseColumn::cast(aClass);
  }
  virtual std::string stringDefaultValue() const { return "";}  
  virtual bool isDefaultDefault() const { return true;}
  virtual bool value(int aIndex,Lib::Value& aVariable) const {
    AIDA::ITuple* v;
    bool status = MemoryColumn_tuple::value(aIndex,v);
    aVariable.set((void*)v);
    return status;
  }
public: //IVisited
  static bool visit(const Slash::Store::IStorable& aObject,
                         Slash::Store::IConstVisitor& aVisitor){
    const BatchLab::MemoryColumn_tuple* local = 
      dynamic_cast<const BatchLab::MemoryColumn_tuple*>(&aObject);
    if(!local) return false;
    return local->BatchLab::MemoryColumn_tuple::visit(aVisitor);
  }
  virtual bool visit(
    Slash::Store::IConstVisitor& aVisitor) const {
    SLASH_STORE_BEGIN(BatchLab::MemoryColumn_tuple)

    int version = 1;
    if(!aVisitor.visit("fVersion",version)) return false;

    if(!aVisitor.visit("fLabel",fLabel)) return false;
    if(!aVisitor.visit("fType",fType)) return false;    

    unsigned int number = fData.size();
    if(!aVisitor.visit("fSize",number)) return false;
    //FIXME : optimize for HDF5.
    for(unsigned int index=0;index<number;index++) {
      MemoryTuple* tuple = fData[index];
      std::string sindex = Lib::smanip::tostring(index);
      if(!aVisitor.visit(sindex,*tuple)) return false;        
    }

    if(!aVisitor.end(*this)) return false;
    return true;
  }
  virtual bool read(Slash::Store::IVisitor& aVisitor) {
    if(!aVisitor.begin(*this)) return false;

    int version;
    if(!aVisitor.visit(version)) return false;

    if(!aVisitor.visit(fLabel)) return false;
    if(!aVisitor.visit(fType)) return false;

    unsigned int sz;
    if(!aVisitor.visit(sz)) return false;
    fData.resize(sz,0);
    for(unsigned int index=0;index<sz;index++) {
      MemoryTuple* tuple = new MemoryTuple(fPrinter,"");
      if(!tuple->read(aVisitor)) return false;        
      fData[index] = tuple;
    }

    if(!aVisitor.end(*this)) return false;
    return true;
  }
public:
  MemoryColumn_tuple(Slash::Core::IWriter& aPrinter,
                     const std::string& aLabel,
                     const std::list<Lib::Value>& aVariables)
  :BaseColumn(aPrinter,aLabel,"ITuple")
  ,fVariables(aVariables)
  ,fTemplate(0){}
  MemoryColumn_tuple(const MemoryColumn_tuple& aFrom)
  :Slash::Store::IStorable(aFrom)
  ,BaseColumn(aFrom)
  ,fVariables(aFrom.fVariables)
  ,fTemplate(0) {
    // Copy data :
    unsigned int nfrom = aFrom.fData.size();
    fData.resize(nfrom,0);
    for(unsigned int ifrom=0;ifrom<nfrom;ifrom++) {  
      MemoryTuple* tuple = aFrom.fData[ifrom];
      fData[ifrom] = new BatchLab::MemoryTuple(*tuple);
      fData[ifrom]->setName(name());
    }
  }
  virtual ~MemoryColumn_tuple(){
    clear();
    if(fTemplate) {
      // This may happen if someone do a getTuple to query
      // (booking) infos of the subtuple. See for exa the 
      // <store>_DataWriter code to get sub tuples booking.
      // Then someone do a getTuple without doing addRow.
      delete fTemplate;
    }
    Lib::Columns::deleteColumns(fVariables);
  }
  AIDA::ITuple* getTuple() {
    if(fTemplate) {
      Lib::Out out(fPrinter);
      out << "BatchLab::MemoryColumn_tuple::getTuple :"
          << " can't be done twice before the addRow."
          << Lib::endl;
      return 0;
    }
    fTemplate = new MemoryTuple(fPrinter,name(),"",fVariables);
    return fTemplate;
  }
  virtual bool fill(const AIDA::ITuple& aValue) {
    if(fTemplate) {
      Lib::Out out(fPrinter);
      out << "BatchLab::MemoryColumn_tuple::fill :"
          << " can't be done whilst a getTuple without addRow had been done."
          << Lib::endl;
      return false;
    }
    fTemplate = new MemoryTuple(fPrinter,name(),"",fVariables);
    Lib::StreamBuffer sbuf(fPrinter);
    std::ostream out(&sbuf);
    if(!amanip::copy
      (out,const_cast<AIDA::ITuple&>(aValue),*fTemplate)) {
      out << "BatchLab::MemoryColumn_tuple::fill(ITupleEntry) :"
          << " cannot copy tuple content."
          << std::endl;
      delete fTemplate;
      fTemplate = 0;
      return false;
    }
    return true;
  }
  virtual bool value(int aIndex,AIDA::ITuple*& aValue) const {
    if((aIndex<0)||(aIndex>=fData.size())) {
      aValue = 0;
      return false;
    } else {
      aValue = fData[aIndex];
      return true;
    }
  }
  const std::list<Lib::Value>& variables() const { return fVariables;}
private:
  void clear() {
    std::vector<MemoryTuple*>::iterator it;
    for(it=fData.begin();it!=fData.end();++it) delete (*it);
    fData.clear();
  }
private:
  std::list<Lib::Value> fVariables;
  std::vector<MemoryTuple*> fData;
  MemoryTuple* fTemplate;
};

}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
BatchLab::MemoryTuple::MemoryTuple(
 Slash::Core::IWriter& aPrinter
,const std::string& aName
)
:BaseTuple(aPrinter,aName,"")
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::MemoryTuple::MemoryTuple(
 Slash::Core::IWriter& aPrinter
,const std::string& aName
,const std::string& aTitle
,const std::string& aBooking
)
:BaseTuple(aPrinter,aName,aTitle)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::list<Lib::Value> variables;
  if(!booking(printer(),aBooking,variables)) return; //FIXME : throw
  initialize(variables);
  Lib::Columns::deleteColumns(variables);
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::MemoryTuple::MemoryTuple(
 Slash::Core::IWriter& aPrinter
,const std::string& aName
,const std::string& aTitle
,const std::list<Lib::Value>& aVariables
)
:BaseTuple(aPrinter,aName,aTitle)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  initialize(aVariables);
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::MemoryTuple::MemoryTuple(
 const MemoryTuple& aFrom
)
:AIDA::IManagedObject(aFrom)
,Slash::Store::IObject(aFrom)
,AIDA::ITuple(aFrom)
,AIDA::ITupleEntry(aFrom)
,Slash::Store::IStorable(aFrom)
,Slash::Meta::IObject(aFrom)
,BaseTuple(aFrom)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::MemoryTuple::~MemoryTuple(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTuple::initialize(
 const std::list<Lib::Value>& aVariables
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //printf("debug : BatchLab::MemoryTuple::initialize : %ld\n",this);
  std::list<Lib::Value>::const_iterator it;
  for(it=aVariables.begin();it!=aVariables.end();++it) {
    std::string name = (*it).label();
    if(findColumn(name)!=NotFound) {
      Lib::Out out(printer());
      out << "BatchLab::MemoryTuple::MemoryTuple :"
          << " column " << name 
          << " already exists" << Lib::endl;
      clear();
      return false;
    }
    //printf("debug : BatchLab::MemoryTuple::initialize : var : %ld \"%s\" \"%s\"\n",this,name.c_str(),stype.c_str());
    //FIXME : template the below in some way.
    switch((*it).type()) {
    case Slash::Core::IValue::CHAR:
      fColumns.push_back
        (new BatchLab::MemoryColumn_char
           (printer(),name,0,(*it).get_char()));
      break;
    case Slash::Core::IValue::SHORT:
      fColumns.push_back
        (new BatchLab::MemoryColumn_short
           (printer(),name,0,(*it).get_short()));
      break;
    case Slash::Core::IValue::INT:
      fColumns.push_back 
        (new BatchLab::MemoryColumn_int
          (printer(),name,0,(*it).get_int()));
      break;
    case Slash::Core::IValue::INT64:
      fColumns.push_back
        (new BatchLab::MemoryColumn_int64
           (printer(),name,0,(*it).get_int64()));
      break;
    case Slash::Core::IValue::DOUBLE:
      fColumns.push_back
        (new BatchLab::MemoryColumn_double
           (printer(),name,0,(*it).get_double()));
      break;
    case Slash::Core::IValue::FLOAT:
      fColumns.push_back
        (new BatchLab::MemoryColumn_float
           (printer(),name,0,(*it).get_float()));
      break;
    case Slash::Core::IValue::BOOL:
      fColumns.push_back
        (new BatchLab::MemoryColumn_bool
          (printer(),name,0,(*it).get_bool()));
      break;
    case Slash::Core::IValue::UNSIGNED_CHAR:
      fColumns.push_back
        (new BatchLab::MemoryColumn_byte
           (printer(),name,0,(*it).get_unsigned_char()));
      break;
    case Slash::Core::IValue::STRING:
      fColumns.push_back
        (new BatchLab::MemoryColumn_string
           (printer(),name,"",(*it).get_string()));
      break;
    case Slash::Core::IValue::ARRAY_DOUBLE:{
      std::vector<unsigned int> is;
      const std::vector<double>& v = (*it).get_array_double(is);
      fColumns.push_back
        (new BatchLab::MemoryColumn_array_double
           (printer(),name,std::vector<double>(),v));
      }break;
    case Slash::Core::IValue::VOID_STAR:{ //Assume AIDA::ITuple.
      std::list<Lib::Value>* pvars = 
        (std::list<Lib::Value>*)(*it).get_void_star();

      std::list<Lib::Value> vars;
      Lib::Columns::copyColumns(*pvars,vars);

      fColumns.push_back
        (new BatchLab::MemoryColumn_tuple(printer(),name,vars));
     }break;

    //case UNSIGNED_INTEGER:
    //case UNSIGNED_INT64:
    //case VECTOR_DOUBLE:
    default:
      Lib::Out out(printer());
      out << "BatchLab::MemoryTuple::initialize :" 
          << " unhandle type " << Lib::sout((*it).stype()) 
          << " for variable with name " << Lib::sout(name) 
          << "." 
          << Lib::endl;
      clear();
      return false;
    }
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
Slash::Store::IObject* BatchLab::MemoryTuple::copy(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return new BatchLab::MemoryTuple(*this);
}
//////////////////////////////////////////////////////////////////////////////
std::string BatchLab::MemoryTuple::storeClassName(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return "BatchLab::MemoryTuple";
}
//////////////////////////////////////////////////////////////////////////////
void* BatchLab::MemoryTuple::cast(
 const std::string& aClass
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if_Lib_SCast(Slash::Store::IStorable)
  else if_Lib_SCast(Slash::Meta::IObject)
  else if_Lib_SCast(BatchLab::MemoryTuple)
  else return BaseTuple::cast(aClass);
}
//////////////////////////////////////////////////////////////////////////////
int BatchLab::MemoryTuple::rows(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fColumns.size()<=0) return 0;
  return fColumns[0]->size();
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTuple::addRow(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<BaseColumn*>::iterator it;
  for(it=fColumns.begin();it!=fColumns.end();++it) {
    (*it)->add();
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#define GWRITE(T,ST,GT,MT) \
T BatchLab::MemoryTuple::GT(int aColumn) const{\
  BaseColumn* column = findColumn1(aColumn);\
  if(!column) {\
    Lib::Out out(printer());\
    out << "BatchLab::MemoryTuple::get(" << ST << ") :"\
        << " column " << aColumn << " not found."\
        << Lib::endl;\
    return 0;\
  }\
  MT* col = Lib_Cast(*column,MT);\
  if((column->type()!=ST)||(!col)) {\
    Lib::Out out(printer());\
    out << "BatchLab::MemoryTuple::get(" << ST << ") :"\
        << " type mismatch (" << Lib::sout(column->type()) << ") "\
        << " for column " << aColumn\
        << " " << Lib::sout(column->name()) << "."\
        << Lib::endl;\
    return 0;\
  }\
  T value;\
  if(!col->value(fRow,value)) {\
    Lib::Out out(printer());\
    out << "BatchLab::MemoryTuple::get(" << ST << ") :"\
        << " can't get value."\
        << Lib::endl;\
    return 0;\
  }\
  return value;\
}

GWRITE(float,"float",getFloat,BatchLab::MemoryColumn_float)
GWRITE(double,"double",getDouble,BatchLab::MemoryColumn_double)
GWRITE(char,"char",getChar,BatchLab::MemoryColumn_char)
GWRITE(byte,"byte",getByte,BatchLab::MemoryColumn_byte)
GWRITE(short,"short",getShort,BatchLab::MemoryColumn_short)
GWRITE(int,"int",getInt,BatchLab::MemoryColumn_int)
GWRITE(std::string,"string",getString,BatchLab::MemoryColumn_string)
GWRITE(bool,"boolean",getBoolean,BatchLab::MemoryColumn_bool)

#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
long BatchLab::MemoryTuple::getLong(int aColumn) const{
  BaseColumn* column = findColumn1(aColumn);
  if(!column) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::getLong :"
        << " column " << aColumn << " not found."
        << Lib::endl;
    return 0;
  }
  BatchLab::MemoryColumn_int64* col = 
    Lib_Cast(*column,BatchLab::MemoryColumn_int64);
  if((column->type()!="long")||(!col)) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::getLong :"
        << " type mismatch (" << Lib::sout(column->type()) << ") "
        << " for column " << aColumn
        << " " << Lib::sout(column->name()) << "."
        << Lib::endl;
    return 0;
  }
  AIDA_int64 value;
  if(!col->value(fRow,value)) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::getLong :"
        << " can't get value."
        << Lib::endl;
    return 0;
  }
  return (long)value;
}
#else
GWRITE(AIDA_int64,"long",getLong,BatchLab::MemoryColumn_int64)
#endif
//////////////////////////////////////////////////////////////////////////////
const AIDA::ITupleEntry* BatchLab::MemoryTuple::getObject(
 int aColumn
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  BaseColumn* column = findColumn1(aColumn);
  if(!column) {\
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::getObject :"
        << " column " << aColumn << " not found."
        << Lib::endl;
    return 0;
  }
  MemoryColumn_tuple* col = Lib_Cast(*column,BatchLab::MemoryColumn_tuple);
  if((column->type()!="ITuple")||(!col)) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::getObject :"
        << " type mismatch (" << Lib::sout(column->type()) << ") "
        << " for column " << aColumn
        << " " << Lib::sout(column->name()) << "."
        << Lib::endl;
    return 0;
  }
  AIDA::ITuple* value = 0;
  if(!col->value(fRow,value)) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::getObject :"
        << " can't get value."
        << Lib::endl;
    return 0;
  }
  if(!value) return 0;
  return (const AIDA::ITupleEntry*)Lib_Cast(*value,AIDA::ITupleEntry);
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
AIDA::ITuple* BatchLab::MemoryTuple::findTuple(
 int aColumn
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  BaseColumn* column = findColumn1(aColumn);
  if(!column) {\
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::findTuple :"
        << " column " << aColumn << " not found."
        << Lib::endl;
    return 0;
  }
  MemoryColumn_tuple* col = Lib_Cast(*column,BatchLab::MemoryColumn_tuple);
  if((column->type()!="ITuple")||(!col)) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::findTuple :"
        << " type mismatch (" << Lib::sout(column->type()) << ") "
        << " for column " << aColumn
        << " " << Lib::sout(column->name()) << "."
        << Lib::endl;
    return 0;
  }
  return col->getTuple();
}
//////////////////////////////////////////////////////////////////////////////
const AIDA::ITuple* BatchLab::MemoryTuple::findTuple(
 int aColumn
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  BaseColumn* column = findColumn1(aColumn);
  if(!column) {\
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::findTuple :"
        << " column " << aColumn << " not found."
        << Lib::endl;
    return 0;
  }
  MemoryColumn_tuple* col = Lib_Cast(*column,BatchLab::MemoryColumn_tuple);
  if((column->type()!="ITuple")||(!col)) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::findTuple :"
        << " type mismatch (" << Lib::sout(column->type()) << ") "
        << " for column " << aColumn
        << " " << Lib::sout(column->name()) << "."
        << Lib::endl;
    return 0;
  }
  return col->getTuple();
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ITuple* BatchLab::MemoryTuple::getTuple(
 int aColumn
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return findTuple(aColumn);
}
//////////////////////////////////////////////////////////////////////////////
const AIDA::ITuple* BatchLab::MemoryTuple::getTuple(
 int aColumn
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return findTuple(aColumn);
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#define FWRITE(T,ST,FT,MT) \
bool BatchLab::MemoryTuple::FT(int aColumn,T aValue){\
  BaseColumn* column = findColumn1(aColumn);\
  if(!column) {\
    Lib::Out out(printer());\
    out << "BatchLab::MemoryTuple::fill(" << ST << ") :"\
        << " column " << aColumn << " not found."\
        << Lib::endl;\
    return false;\
  }\
  MT* col = Lib_Cast(*column,MT);\
  if((column->type()!=ST)||(!col)) {\
    Lib::Out out(printer());\
    out << "BatchLab::MemoryTuple::fill(" << ST << ") :"\
        << " type mismatch (" << Lib::sout(column->type()) << ") "\
        << " for column " << aColumn\
        << " " << Lib::sout(column->name()) << "."\
        << Lib::endl;\
    return false;\
  }\
  return col->fill(aValue);\
}

FWRITE(float,"float",fill,BatchLab::MemoryColumn_float)
FWRITE(double,"double",fill,BatchLab::MemoryColumn_double)
FWRITE(char,"char",fill,BatchLab::MemoryColumn_char)
FWRITE(short,"short",fill,BatchLab::MemoryColumn_short)
FWRITE(int,"int",fill,BatchLab::MemoryColumn_int)
FWRITE(bool,"boolean",fill,BatchLab::MemoryColumn_bool)

#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
bool BatchLab::MemoryTuple::fill(int aColumn,long aValue){
  BaseColumn* column = findColumn1(aColumn);
  if(!column) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::fill(long) :"
        << " column " << aColumn << " not found."
        << Lib::endl;
    return false;
  }
  BatchLab::MemoryColumn_int64* col = 
    Lib_Cast(*column,BatchLab::MemoryColumn_int64);
  if((column->type()!="long")||(!col)) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::fill(long) :"
        << " type mismatch (" << Lib::sout(column->type()) << ") "
        << " for column " << aColumn
        << " " << Lib::sout(column->name()) << "."
        << Lib::endl;
    return false;
  }
  return col->fill((AIDA_int64)aValue);
}
#else
FWRITE(AIDA_int64,"long",fill,BatchLab::MemoryColumn_int64)
FWRITE(byte,"byte",fill,BatchLab::MemoryColumn_byte)
#endif

//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTuple::fill(
 int aColumn
,const std::string& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  BaseColumn* column = findColumn1(aColumn);
  if(!column) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::fill(string) :"
        << " column " << aColumn << " not found."
        << Lib::endl;
    return false;
  }
  MemoryColumn_string* col = Lib_Cast(*column,BatchLab::MemoryColumn_string);
  if((column->type()!="string")||(!col)) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::fill(string) :"
        << " type mismatch (" << Lib::sout(column->type()) << ") "
        << " for column " << aColumn
        << " " << Lib::sout(column->name()) << "."
        << Lib::endl;
    return false;
  }
  return col->fill(aValue);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTuple::fill(
 int aColumn
,const AIDA::ITupleEntry& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  BaseColumn* column = findColumn1(aColumn);
  if(!column) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::fill(ITuple) :"
        << " column " << aColumn << " not found."
        << Lib::endl;
    return false;
  }
  MemoryColumn_tuple* col = Lib_Cast(*column,BatchLab::MemoryColumn_tuple);
  if((column->type()!="ITuple")||(!col)) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::fill(ITuple) :"
        << " type mismatch (" << Lib::sout(column->type()) << ") "
        << " for column " << aColumn
        << " " << Lib::sout(column->name()) << "."
        << Lib::endl;
    return false;
  }
  const AIDA::ITuple* tuple = dynamic_cast<const AIDA::ITuple*>(&aValue);
  if(!tuple) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::fill :"
        << " ITupleEntry not an ITuple."
        << Lib::endl;
    return false;
  }
  return col->fill(*tuple);
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTuple::fill(
 const std::vector<double>& aRow
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  int n = aRow.size();
  if(n!=columns()) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::fill :"
        << " number of column mismatch (expected = " 
        << columns() << ", given = " << n << ")" << Lib::endl;
    return false;
  }
  for(int count=0;count<n;count++) {
    if(!fill(count,aRow[count])) return false;
  }
  return addRow();
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTuple::fill(
 const std::vector<float>& aRow
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  int n = aRow.size();
  if(n!=columns()) {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::fill :"
        << " number of column mismatch (expected = "
        << columns() << ", given = " << n << ")" << Lib::endl;
    return false;
  }
  for(int count=0;count<n;count++) {
    if(!fill(count,aRow[count])) return false;
  }
  return addRow();
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#define COLMINMAX(T,MT,OP) \
    MT* col = Lib_Cast(*column,MT);\
    T value;\
    col->value(0,value);\
    T v = value;\
    for(int row=1;row<rown;row++) {\
      col->value(row,value);\
      v = OP(v,value);\
    }\
    return (double)v

//////////////////////////////////////////////////////////////////////////////
double BatchLab::MemoryTuple::columnMin(
 int aColumn
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  BaseColumn* column = findColumn1(aColumn);
  if(!column) return 0;
  int rown = rows();
  if(rown<=0) return 0;
  if(column->type()=="float") {
    COLMINMAX(float,BatchLab::MemoryColumn_float,MINIMUM);
  } else if(column->type()=="double") {
    COLMINMAX(double,BatchLab::MemoryColumn_double,MINIMUM);
  } else if(column->type()=="char") {
    COLMINMAX(char,BatchLab::MemoryColumn_char,MINIMUM);
  } else if(column->type()=="short") {
    COLMINMAX(short,BatchLab::MemoryColumn_short,MINIMUM);
  } else if(column->type()=="int") {
    COLMINMAX(int,BatchLab::MemoryColumn_int,MINIMUM);
  } else if(column->type()=="long") {
    COLMINMAX(AIDA_int64,BatchLab::MemoryColumn_int64,MINIMUM);
  } else if(column->type()=="boolean") {
    COLMINMAX(bool,BatchLab::MemoryColumn_bool,MINIMUM);
  } else {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::columnMin :"
        << " make no sense on column type " << column->type() << "."
        << Lib::endl;
    return 0;
  }
}
//////////////////////////////////////////////////////////////////////////////
double BatchLab::MemoryTuple::columnMax(
 int aColumn
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  BaseColumn* column = findColumn1(aColumn);
  if(!column) return 0;
  int rown = rows();
  if(rown<=0) return 0;
  if(column->type()=="float") {
    COLMINMAX(float,BatchLab::MemoryColumn_float,MAXIMUM);
  } else if(column->type()=="double") {
    COLMINMAX(double,BatchLab::MemoryColumn_double,MAXIMUM);
  } else if(column->type()=="char") {
    COLMINMAX(char,BatchLab::MemoryColumn_char,MAXIMUM);
  } else if(column->type()=="short") {
    COLMINMAX(short,BatchLab::MemoryColumn_short,MAXIMUM);
  } else if(column->type()=="int") {
    COLMINMAX(int,BatchLab::MemoryColumn_int,MAXIMUM);
  } else if(column->type()=="long") {
    COLMINMAX(AIDA_int64,BatchLab::MemoryColumn_int64,MAXIMUM);
  } else if(column->type()=="boolean") {
    COLMINMAX(bool,BatchLab::MemoryColumn_bool,MAXIMUM);
  } else {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::columnMin :"
        << " make no sense on column type " << column->type() << "."
        << Lib::endl;
    return 0;
  }
}
//////////////////////////////////////////////////////////////////////////////
double BatchLab::MemoryTuple::columnMean(
 int aColumn
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
#define COLMEAN(T,MT) \
    MT* col = Lib_Cast(*column,MT);\
    double sv = 0;\
    T value = 0;\
    for(int row=0;row<rown;row++) {\
      col->value(row,value);\
      sv += value;\
    }\
    return sv/rown

  BaseColumn* column = findColumn1(aColumn);
  if(!column) return 0;
  int rown = rows();
  if(rown<=0) return 0;
  if(column->type()=="float") {
    COLMEAN(float,BatchLab::MemoryColumn_float);
  } else if(column->type()=="double") {
    COLMEAN(double,BatchLab::MemoryColumn_double);
  } else if(column->type()=="char") {
    COLMEAN(char,BatchLab::MemoryColumn_char);
  } else if(column->type()=="short") {
    COLMEAN(short,BatchLab::MemoryColumn_short);
  } else if(column->type()=="int") {
    COLMEAN(int,BatchLab::MemoryColumn_int);
  } else if(column->type()=="long") {
    COLMEAN(AIDA_int64,BatchLab::MemoryColumn_int64);
  } else if(column->type()=="boolean") {
    COLMEAN(bool,BatchLab::MemoryColumn_bool);
  } else {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::columnMean :"
        << " make no sense on column type " << column->type() << "."
        << Lib::endl;
    return 0;
  }
}
//////////////////////////////////////////////////////////////////////////////
double BatchLab::MemoryTuple::columnRms(
 int aColumn
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
#define COLRMS(T,MT) \
    MT* col = Lib_Cast(*column,MT);\
    double sv = 0;\
    double sv2 = 0;\
    T value = 0;\
    for(int row=0;row<rown;row++) {\
      col->value(row,value);\
      sv += (double)value;\
      sv2 += (double)value * (double)value;\
    }\
    double mean = sv/((double)rown);\
    double rms = ::sqrt(::fabs(sv2/((double)rown) - mean * mean));\
    return rms

  BaseColumn* column = findColumn1(aColumn);
  if(!column) return 0;
  int rown = rows();
  if(rown<=0) return 0;
  if(column->type()=="double") {
    COLRMS(double,BatchLab::MemoryColumn_double);
  } else if(column->type()=="float") {
    COLRMS(float,BatchLab::MemoryColumn_float);
  } else if(column->type()=="int") {
    COLRMS(int,BatchLab::MemoryColumn_int);
  } else if(column->type()=="short") {
    COLRMS(short,BatchLab::MemoryColumn_short);
  } else if(column->type()=="long") {
    COLRMS(AIDA_int64,BatchLab::MemoryColumn_int64);
  } else if(column->type()=="char") {
    COLRMS(char,BatchLab::MemoryColumn_char);
  } else if(column->type()=="boolean") {
    COLRMS(bool,BatchLab::MemoryColumn_bool);
  } else {
    Lib::Out out(printer());
    out << "BatchLab::MemoryTuple::columnMean :"
        << " make no sense on column type " << column->type() << "."
        << Lib::endl;
    return 0;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTuple::getBooking(
 std::string& aString
) const
//////////////////////////////////////////////////////////////////////////////
// Faster version than amanip::tuple_getBooking
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int coln = fColumns.size();
  for(unsigned int index=0;index<coln;index++) {
    if(index) aString += ",";
    BaseColumn* column = fColumns[index];
    std::string type = column->type();
    aString += type + " ";
    aString += column->name() + " = ";
    if(type=="ITuple") {
      BatchLab::MemoryColumn_tuple* col = 
        Lib_Cast(*column,BatchLab::MemoryColumn_tuple);
      MemoryTuple subTuple(printer(),col->name(),"",col->variables());
      std::string s;
      if(!subTuple.getBooking(s)) {
        aString = "";
        return false;
      }
      aString += "{"+s+"}";
    } else {
      aString += column->stringDefaultValue();
    }
  }
  return true;
}
SLASH_STORE_SOURCE(BatchLab::MemoryTuple)
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTuple::visit(
 Slash::Store::IConstVisitor& aVisitor
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SLASH_STORE_BEGIN(BatchLab::MemoryTuple)

  int version = 1;
  if(!aVisitor.visit("fVersion",version)) return false;

  if(!aVisitor.visit("fAnnotation",fAnnotation)) return false;
  if(!aVisitor.visit("fTitle",fTitle)) return false;

  std::string sbooking;
  if(!getBooking(sbooking)) return false;
  if(!aVisitor.visit("fBooking",sbooking)) return false;

  unsigned int coln = fColumns.size();
  for(unsigned int index=0;index<coln;index++) {
    std::string sindex = Lib::smanip::tostring(index);
    BaseColumn* column = fColumns[index];
    Slash::Store::IStorable* v = Lib_Cast(*column,Slash::Store::IStorable);
    if(!aVisitor.visit("fColumn_"+sindex,*v)) return false;
  }

  if(!aVisitor.end(*this)) return false;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool BatchLab::MemoryTuple::read(
 Slash::Store::IVisitor& aVisitor
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  clear();

  if(!aVisitor.begin(*this)) return false;

  int version;
  if(!aVisitor.visit(version)) return false;

  if(!fAnnotation.read(aVisitor)) return false;

  if(!aVisitor.visit(fTitle)) return false;

  std::string sbooking;
  if(!aVisitor.visit(sbooking)) return false;

  std::list<Lib::Value> variables;
  if(!booking(printer(),sbooking,variables)) return false;
  initialize(variables);
  Lib::Columns::deleteColumns(variables);

  unsigned int coln = fColumns.size();
  for(unsigned int index=0;index<coln;index++) {
    std::string sindex = Lib::smanip::tostring(index);
    BaseColumn* column = fColumns[index];
    if(!column->read(aVisitor)) return false;
  }

  if(!aVisitor.end(*this)) return false;
  return true;
}
