// this :
#include <BatchLab/Core/TupleFactory.h>

#include <AIDA/ITree.h>

#include <Slash/Core/ISession.h>
#include <Slash/Store/IStoreManager.h>

#include <amanip/tuple.h>

#include <Lib/smanip.h>
#include <Lib/Out.h>
#include <Lib/Manager.h>
#include <Lib/Columns.h>
#include <Lib/sout.h>
#include <Lib/Cast.h>
#include <Lib/Debug.h>
#include <Lib/StreamBuffer.h>
#include <iostream>

#include <BatchLab/Core/AsciiTree.h> //FIXME
#include <BatchLab/Core/MemoryTuple.h>
#include <BatchLab/Core/ChainedTuple.h>
#include <BatchLab/Core/Filter.h>
#include <BatchLab/Core/Evaluator.h>

#define NotFound (-1)

//////////////////////////////////////////////////////////////////////////////
BatchLab::TupleFactory::TupleFactory(
 Slash::Core::ISession& aSession
,Slash::Store::ITree& aTree
)
:BaseFactory(aSession.printer(),aTree)
,fSession(aSession)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("BatchLab::TupleFactory");
}
//////////////////////////////////////////////////////////////////////////////
BatchLab::TupleFactory::~TupleFactory(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::decrement("BatchLab::TupleFactory");
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ITuple* BatchLab::TupleFactory::create( 
 const std::string& aPath
,const std::string& aTitle
,const std::vector<std::string>& aColumnNames
,const std::vector<std::string>& aColumnTypes
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aColumnNames.size()) return 0;
  if(aColumnNames.size()!=aColumnTypes.size()) return 0;
  std::string s; 
  for(unsigned int index=0;index<aColumnNames.size();index++) {
    if(index) s += ",";
    s += aColumnTypes[index] + " " + aColumnNames[index];
  }
  return create(aPath,aTitle,s,aOptions);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ITuple* BatchLab::TupleFactory::create(
 const std::string& aPath
,const std::string& aTitle
,const std::string& aBooking
,const std::string& aOptions
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;

  Slash::Store::IObject* object = 0;
  if(fTree.storeType()=="BatchLab::Rio::Tree") {
    //FIXME : check if an "aPath" object already exists !
    Slash::Store::IStoreManager* storeManager = 
      Lib_findManager(fSession,"Rio_Manager",Slash::Store::IStoreManager);
    if(!storeManager) return 0;
    std::vector<std::string> args;
    args.push_back(aBooking);
    object = storeManager->createObject
               (fTree,"AIDA::ITuple",path,name,aTitle,args,aOptions);

  } else if(fTree.storeType()=="BatchLab::Zebra::Tree") {
    //FIXME : check if an "aPath" object already exists !
    Slash::Store::IStoreManager* storeManager = 
      Lib_findManager(fSession,"Zebra_Manager",Slash::Store::IStoreManager);
    if(!storeManager) return 0;
    std::vector<std::string> args;
    args.push_back(aBooking);
    object = storeManager->createObject
               (fTree,"AIDA::ITuple",path,name,aTitle,args,aOptions);

  } else if(fTree.storeType()=="BatchLab::AsciiTree") {
    //FIXME : check if an "aPath" object already exists !
    //NOTE : at commit, only one tuple put in file.
    object= 
      new BatchLab::MemoryTuple(fSession.printer(),name,aTitle,aBooking);
  } else if( (fTree.storeType()=="BatchLab::MemoryTree") || 
             (fTree.storeType()=="BatchLab::XML::Tree")  ){
    //FIXME : check if an "aPath" object already exists !
    object = 
      new BatchLab::MemoryTuple(fSession.printer(),name,aTitle,aBooking);
  } else if( (fTree.storeType()=="BatchLab::HDF5::Tree") || 
             (fTree.storeType()=="BatchLab::Zip::Tree")  ||
             (fTree.storeType()=="BatchLab::FS::Tree")  ){
    //FIXME : check if an "aPath" object already exists !
    //FIXME : optimize : have a store-paginated tuple.
    object = 
      new BatchLab::MemoryTuple(fSession.printer(),name,aTitle,aBooking);
  } else {
    Lib::Out out(fPrinter);
    out << "BatchLab::TupleFactory::create :" 
        << " Can't create tuple " << Lib::sout(aPath) 
        << " in tree with store " << Lib::sout(fTree.storeName())
        << Lib::endl;
  }
  if(!object) return 0;
  AIDA::ITuple* aobject = Lib_Cast(*object,AIDA::ITuple);
  if(!aobject) {
    Lib::Out out(fPrinter);
    out << "BatchLab::TupleFactory::create :" 
        << " not a AIDA::ITuple." << Lib::endl;
    delete object;
    return 0;
  }
  if(!manage(aOptions)) return aobject;
  if(manageObject(object,path)) return aobject;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ITuple* BatchLab::TupleFactory::createCopy(
 const std::string& aPath
,const AIDA::ITuple& aFrom
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(fPrinter);
  std::ostream out(&sbuf);
  AIDA::ITuple* tuple = 
    create(aPath,aFrom.title(),aFrom.columnNames(),aFrom.columnTypes(),"");
  if(!tuple) {
    out << "BatchLab::TupleFactory::createCopy :" 
        << " Can't create tuple " << amanip::sout(aPath) 
        << " in tree with store " << amanip::sout(fTree.storeName())
        << std::endl;
    return 0;
  }
  if(!amanip::copy(out,const_cast<AIDA::ITuple&>(aFrom),*tuple)) {
    out << "BatchLab::TupleFactory::createCopy :" 
        << " cannot copy tuple content."
        << std::endl;
    tuple->reset();
  }
  return tuple;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFilter* BatchLab::TupleFactory::createFilter(
 const std::string& aScript
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  AIDA::ITree* atree = Lib_Cast(fTree,AIDA::ITree);
  if(!atree) return 0;
  return new BatchLab::Filter(fSession.printer(),aScript,*atree);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IFilter* BatchLab::TupleFactory::createFilter(
 const std::string& aScript
,int
,int
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  AIDA::ITree* atree = Lib_Cast(fTree,AIDA::ITree);
  if(!atree) return 0;
  //FIXME
  return new BatchLab::Filter(fSession.printer(),aScript,*atree);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::IEvaluator* BatchLab::TupleFactory::createEvaluator(
 const std::string& aScript
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  AIDA::ITree* atree = Lib_Cast(fTree,AIDA::ITree);
  if(!atree) return 0;
  return new BatchLab::Evaluator(fSession.printer(),aScript,*atree);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ITuple* BatchLab::TupleFactory::createChained(
 const std::string& aPath
,const std::string& aTitle
,const std::vector<AIDA::ITuple*>& aTuples
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  std::string path,name,suffix;
  if(!Lib::smanip::pathNameSuffix(aPath,path,name,suffix)) return 0;
  BatchLab::ChainedTuple* object = new BatchLab::ChainedTuple(fSession.printer(),
                                                    name,aTitle,
                                                    aTuples);
  //if(!manage(aOptions)) return object;
  if(manageObject(object,path)) return object;
  delete object;
  return 0;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ITuple* BatchLab::TupleFactory::createChained(
 const std::string& aPath
,const std::string& aTitle
,const std::vector<std::string>& aTuples
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<AIDA::ITuple*> tuples;
  unsigned tuplen = aTuples.size();
  for(unsigned int index=0;index<tuplen;index++) {
    Slash::Store::IObject* bo = fTree.findObject(aTuples[index]);
    if(!bo) {
      Lib::Out out(fPrinter);
      out << "BatchLab::TupleFactory::createChained :" 
          << " object " << Lib::sout(aTuples[index]) 
          << " not found in tree with store " << Lib::sout(fTree.storeName())
          << Lib::endl;
      return 0;
    }
    AIDA::ITuple* tuple = Lib_Cast(*bo,AIDA::ITuple);
    if(!tuple) {
      Lib::Out out(fPrinter);
      out << "BatchLab::TupleFactory::createChained :" 
          << " object " << Lib::sout(aTuples[index]) 
          << " not a tuple."
          << Lib::endl;
      return 0;
    }
    tuples.push_back(tuple);
  }
  return createChained(aPath,aTitle,tuples);
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ITuple* BatchLab::TupleFactory::createCopy(
 const std::string& aPath
,AIDA::ITuple& aFrom
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(fPrinter);
  std::ostream out(&sbuf);
  AIDA::ITuple* tuple = 
    create(aPath,aFrom.title(),aFrom.columnNames(),aFrom.columnTypes(),"");
  if(!tuple) {
    out << "BatchLab::TupleFactory::createCopy :" 
        << " Can't create tuple " << amanip::sout(aPath) 
        << " in tree with store " << amanip::sout(fTree.storeName())
        << std::endl;
    return 0;
  }
  if(!amanip::copy(out,const_cast<AIDA::ITuple&>(aFrom),*tuple)) {
    out << "BatchLab::TupleFactory::createCopy :" 
        << " cannot copy tuple content."
        << std::endl;
    tuple->reset();
  }
  return tuple;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ITuple* BatchLab::TupleFactory::createFiltered(
 const std::string& aPath
,AIDA::ITuple& aFrom
,AIDA::IFilter& aFilter
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::StreamBuffer sbuf(fPrinter);
  std::ostream out(&sbuf);
  AIDA::ITuple* tuple = 
    create(aPath,aFrom.title(),aFrom.columnNames(),aFrom.columnTypes(),"");
  if(!tuple) {
    out << "BatchLab::TupleFactory::createFiltered :" 
        << " Can't create tuple " << amanip::sout(aPath) 
        << " in tree with store " << amanip::sout(fTree.storeName())
        << std::endl;
    return 0;
  }
  if(!amanip::copy(out,const_cast<AIDA::ITuple&>(aFrom),*tuple,aFilter)) {
    out << "BatchLab::TupleFactory::createFiltered :" 
        << " cannot copy tuple content."
        << std::endl;
    tuple->reset();
  }
  return tuple;
}
//////////////////////////////////////////////////////////////////////////////
AIDA::ITuple* BatchLab::TupleFactory::createFiltered(
 const std::string& aPath
,AIDA::ITuple& aFrom
,AIDA::IFilter& aFilter
,const std::vector<std::string>& aColumns
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int coln = aColumns.size();
  std::vector<int> colis;
  std::vector<std::string> types;
  for(unsigned int index=0;index<coln;index++) {
    int coli = aFrom.findColumn(aColumns[index]);
    if(coli==NotFound) {
      Lib::Out out(fPrinter);
      out << "BatchLab::TupleFactory::createFiltered :" 
          << " " << Lib::sout(aColumns[index]) 
          << " not a column of tuple with title " 
          << Lib::sout(aFrom.title())
          << Lib::endl;
      return 0;
    }
    colis.push_back(coli);
    types.push_back(aFrom.columnTypes()[coli]);
  }  
  AIDA::ITuple* tuple = create(aPath,aFrom.title(),aColumns,types,"");
  if(!tuple) {
    Lib::Out out(fPrinter);
    out << "BatchLab::TupleFactory::createFiltered :" 
        << " Can't create tuple " << Lib::sout(aPath) 
        << " in tree with store " << Lib::sout(fTree.storeName())
        << Lib::endl;
    return 0;
  }
  aFrom.start();
  while(aFrom.next()) {
    if(!aFilter.accept()) continue;
    bool status = true;
    for(int index=0;index<(int)coln;index++) {
      int fromi = colis[index];
      std::string type = aFrom.columnType(fromi);
      if(type=="float") {
        if(!tuple->fill(index,aFrom.getFloat(fromi))) {status = false;break;}
      } else if(type=="double") {
        if(!tuple->fill(index,aFrom.getDouble(fromi))) {status = false;break;}
      } else if(type=="char") {
        if(!tuple->fill(index,aFrom.getChar(fromi))) {status = false;break;}
      } else if(type=="short") {
        if(!tuple->fill(index,aFrom.getShort(fromi))) {status = false;break;}
      } else if(type=="int") {
        if(!tuple->fill(index,aFrom.getInt(fromi))) {status = false;break;}
      } else if(type=="long") {
        if(!tuple->fill(index,aFrom.getLong(fromi))) {status = false;break;}
      } else if(type=="boolean") {
        if(!tuple->fill(index,aFrom.getBoolean(fromi))) {status = false;break;}
      } else if(type=="string") {
        if(!tuple->fill(index,aFrom.getString(fromi))) {status = false;break;}
      } else if(type=="ITuple") {
        if(!tuple->fill(index,*aFrom.getObject(fromi))) {status = false;break;}
      } else {
        Lib::Out out(fPrinter);
        out << "BatchLab::TupleFactory::createFiltered :" 
            << " column type " << type << " not handled."
            << Lib::endl;
        status = false;
        break;
      }
    }
    if(!status) {
      Lib::Out out(fPrinter);
      out << "BatchLab::TupleFactory::createFiltered :" 
          << " problem when filling tuple."
          << Lib::endl;
      tuple->reset();
      break;
    }
    tuple->addRow();
  }
  return tuple;
}
