/**************************************************************************
 *
 *  This file is part of the Coin SoQt GUI binding library.
 *  Copyright (C) 1998-2001 by Systems in Motion.  All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License version
 *  2.1 as published by the Free Software Foundation.  See the file
 *  LICENSE.LGPL at the root directory of the distribution for all the
 *  details.
 *
 *  If you want to use Coin SoQt for applications not compatible with the
 *  LGPL, please contact SIM to acquire a Professional Edition License.
 *
 *  Systems in Motion, Prof Brochs gate 6, N-7030 Trondheim, NORWAY
 *  http://www.sim.no/ support@sim.no Voice: +47 22114160 Fax: +47 22207097
 *
 **************************************************************************/

/*  */

#ifndef SOWIN_INTERNALDEFS_H
#define SOWIN_INTERNALDEFS_H

/* ********************************************************************** */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

/* ********************************************************************** */

#ifdef __FILE__
#define SOWIN_STUB_FILE __FILE__
#else
/* FIXME: avoid this case by redefining the compilation rules to define the
   current filename in some variable.  [20010919 larsa] */
#define SOWIN_STUB_FILE ((char *)0L)
#endif

#ifdef __LINE__
#define SOWIN_STUB_LINE __LINE__
#else
#define SOWIN_STUB_LINE 0
#endif

#if HAVE_VAR___PRETTY_FUNCTION__
#define SOWIN_STUB_FUNC __PRETTY_FUNCTION__
#else
#if HAVE_VAR___FUNCTION__
#define SOWIN_STUB_FUNC __FUNCTION__
#else
#if HAVE_VAR___func__
#define SOWIN_STUB_FUNC __func__
#else
#define SOWIN_STUB_FUNC ((char *)0L)
#endif
#endif
#endif

#if SOWIN_DEBUG

#include <Inventor/errors/SoDebugError.h>

/*
  This stupid thing is here to silence some compilers that complain on
  constant if-expressions.
*/

inline int sowin_not_null(const void * arg) { return (arg != NULL) ? 1 : 0; }


/* Collects common code from SOWIN_STUB(), SOWIN_STUB()_ONCE and
   SOWIN_OBSOLETED(). */

#define SOWIN_FUNCTIONIDENTIFIER(SOWINSTR) \
    SbString SOWINSTR(sowin_not_null(SOWIN_STUB_FILE) ? SOWIN_STUB_FILE : ""); \
    if (SOWINSTR.getLength() > 0) { SOWINSTR += ':'; } \
    SOWINSTR += SbString((int)SOWIN_STUB_LINE); SOWINSTR += ':'; \
    SOWINSTR += sowin_not_null(SOWIN_STUB_FUNC) ? SOWIN_STUB_FUNC : ""

/*
  SOWIN_STUB(): a macro that prints out stub information.
  Used where there is functionality missing.
*/

#define SOWIN_STUB() \
  do { \
    SOWIN_FUNCTIONIDENTIFIER(s); \
    SoDebugError::postWarning(s.getString(), \
                              "STUB: functionality not yet completed"); \
  } while (0)

/*
  SOWIN_STUB_ONCE(): a macro that prints out stub information, but only on
  the first pass.  Used where there is functionality missing that may be
  called hundreds of times.
*/

#define SOWIN_STUB_ONCE() \
  do { \
    static int first = 1; \
    if (first) { \
      SOWIN_FUNCTIONIDENTIFIER(s); \
      SoDebugError::postWarning(s.getString(), \
                                "STUB: functionality not yet completed " \
                                "(first and last warning)"); \
      first = 0; \
    } \
  } while (0)

/*
  SOWIN_OBSOLETED(): a macro that prints out information about an obsoleted
  function still being in use.
*/

#define SOWIN_OBSOLETED() \
  do { \
    SOWIN_FUNCTIONIDENTIFIER(s); \
    SoDebugError::post(s.getString(), \
                       "OBSOLETED: functionality no longer supported"); \
  } while (0)

#else /* !SOWIN_DEBUG */

/* snub the stubs */
#define SOWIN_STUB()      do { } while (0)
#define SOWIN_STUB_ONCE() do { } while (0)
#define SOWIN_OBSOLETED() do { } while (0)

#endif /* !SOWIN_DEBUG */

/*
  GETTEXT support
*/

#if defined(ENABLE_NLS) && defined(HAVE_GETTEXT)
#include <libintl.h>
#define _(String) dgettext (PACKAGE, String)
#define N_(String) (String)
#else
#define _(String) (String)
#define N_(String) (String)
#define textdomain(Domain)
#define bindtextdomain(Package, Directory)
#endif /* HAVE_GETTEXT && ENABLE_NLS */

/* ********************************************************************** */

#endif /* ! SOWIN_INTERNALDEFS_H */
