// this :
#include <G3Lab/SoF77.h>

// Slash :
#include <Slash/Core/ISession.h>
#include <Slash/Tools/FString.h>

// Inventor :
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoLineSet.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoLightModel.h>

// HEPVis :
#include <HEPVis/SbRotation.h>
#include <HEPVis/SbPolyhedron.h>
#include <HEPVis/SbStyle.h>
#include <HEPVis/SbString.h>
#include <HEPVis/SbProjector.h>
#include <HEPVis/nodes/SoPolyhedron.h>
#include <HEPVis/nodes/SoMarkerSet.h>
#include <HEPVis/nodes/SoHighlightMaterial.h>
#include <HEPVis/misc/SoGC.h>
typedef HEPVis_SoMarkerSet SoMarkerSet;

static Slash::Core::ISession* sSession = 0; //FIXME
static SoGC* sGC = 0; //FIXME

//////////////////////////////////////////////////////////////////////////////
void SoF77_SetSession(
 Slash::Core::ISession* aSession
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  sSession = aSession;
}
//////////////////////////////////////////////////////////////////////////////
void SoF77_SetGC(
 SoGC* aGC
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  sGC = aGC;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void Slash_stdcall soprnt(
 void* aF77
#ifdef WIN32
,unsigned int
#endif
,int* aLength
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!sSession) return;
  Slash::FString fs(aF77,*aLength);
  //printf("debug : soprnt \"%s\"\n",fs.std_string().c_str());
  std::ostream& out = sSession->cout();
  out << fs.std_string() << std::endl;
}
//////////////////////////////////////////////////////////////////////////////
void Slash_stdcall sopmar(
 int* aColor
,int* //aMarkerStyle
,int* aNumber
,float* aXs
,float* aYs
,float* aZs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoSeparator* separator = sGC->getSeparator();

  //printf("debug : sopmar : begin : %d\n",*aNumber);
  int pointn = *aNumber;
  if(pointn<=0) return;

  if(sGC->getLocalSetup()==FALSE) { 
    // Take color given by modeling.color :
  } else { // Take geant3 colors (aColor) :
    //   blue 3, red 47, black 2, green 23, grey 37 
         if(*aColor==3)  sGC->setFromString("color blue");
    else if(*aColor==47) sGC->setFromString("color red");
    else if(*aColor==2)  sGC->setFromString("color black");
    else if(*aColor==23) sGC->setFromString("color green");
    else if(*aColor==37) sGC->setFromString("color grey");
  }

  separator->addChild(sGC->getHighlightMaterial());

  SbVec3f* points = new SbVec3f[pointn];
  for(int index=0;index<pointn;index++) {
    points[index].setValue(aXs[index],aYs[index],aZs[index]);
  }

  SoCoordinate3* coordinate3 = new SoCoordinate3;
  coordinate3->point.setValues(0,pointn,points);
  separator->addChild(coordinate3);

  delete [] points;

  SoMarkerSet* markerSet = new SoMarkerSet;
  markerSet->numPoints = pointn;
  markerSet->markerIndex = SoMarkerSet::CROSS_5_5;
  separator->addChild(markerSet);
}
//////////////////////////////////////////////////////////////////////////////
void Slash_stdcall soplin(
 int* aColor
,int* aLineStyle
,int* aNumber
,float* aCoords
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoSeparator* separator = sGC->getSeparator();

  //printf("debug : soplin : begin : %d\n",*aNumber);
  int pointn = *aNumber;
  if(pointn<=0) return;

  if(sGC->getLocalSetup()==FALSE) { 
    // Take color given by modeling.color :
  } else { // Take geant3 colors (aColor) :
    //   blue 3, red 47, black 2, green 23, grey 37 
         if(*aColor==3)  sGC->setFromString("color blue");
    else if(*aColor==47) sGC->setFromString("color red");
    else if(*aColor==2)  sGC->setFromString("color black");
    else if(*aColor==23) sGC->setFromString("color green");
    else if(*aColor==37) sGC->setFromString("color grey");

         if(*aLineStyle==0) sGC->setLinePattern(SbLinePattern_solid);
    else if(*aLineStyle==1) sGC->setLinePattern(SbLinePattern_dashed);
    else if(*aLineStyle==2) sGC->setLinePattern(SbLinePattern_dotted);
    else if(*aLineStyle==3) sGC->setLinePattern(SbLinePattern_dash_dotted);
  }

  separator->addChild(sGC->getHighlightMaterial());
  separator->addChild(sGC->getDrawStyle());
  separator->addChild(sGC->getLightModel());

  SbVec3f* points = new SbVec3f[pointn];

  float* coords = aCoords;
  for(int index=0;index<pointn;index++) {
    float x = *coords;coords++;
    float y = *coords;coords++;
    float z = *coords;coords++;
    points[index].setValue(x,y,z);
    //printf("debug : soplin : begin :  %d : %g %g %g\n",index,x,y,z);
  }

  SoCoordinate3* coordinate3 = new SoCoordinate3;
  coordinate3->point.setValues(0,pointn,points);
  separator->addChild(coordinate3);

  delete [] points;

  SoLineSet* lineSet = new SoLineSet;
  lineSet->numVertices.setValues(0,1,&pointn);
  separator->addChild(lineSet);
}
//////////////////////////////////////////////////////////////////////////////
void Slash_stdcall soshap(
 int* aShape
,float* aPars
,float* aRotation
,float* aTranslation
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  //printf("debug : soshap %d\n",*aShape);

  SbPolyhedron* result = 0;

  switch(*aShape) {
  case 1:{
    double dx = aPars[0];
    double dy = aPars[1];
    double dz = aPars[2];
    result = new SbPolyhedronTrd2(dx,dx,dy,dy,dz);
  }break;
  case 2:{
    double dxd = aPars[0];
    double dxu = aPars[1];
    double dy = aPars[2];
    double dz = aPars[3];
    result = new SbPolyhedronTrd2(dxd,dxu,dy,dy,dz);
  }break;
  case 3:{
    double dxd = aPars[0];
    double dxu = aPars[1];
    double dyd = aPars[2];
    double dyu = aPars[3];
    double dz = aPars[4];
    result = new SbPolyhedronTrd2(dxd,dxu,dyd,dyu,dz);
  }break; 
  case 4:{
    double DZ = aPars[0];
    double TX = aPars[1] * M_PI/180;
    double TY = aPars[2] * M_PI/180;
    double H1 = aPars[3];
    double BL1 = aPars[4];
    double TL1 = aPars[5];
    double TTH1 = aPars[6] * M_PI/180;
    double H2 = aPars[7];
    double BL2 = aPars[8];
    double TL2 = aPars[9];
    double TTH2 = aPars[10] * M_PI/180;
    result = new SbPolyhedronTrap(DZ,
				  TX,TY,
				  H1,BL1,TL1,TTH1,
				  H2,BL2,TL2,TTH2);
  }break;
  case 5:{
    double rmin = aPars[0];
    double rmax = aPars[1];
    double dz = aPars[2];
    result = new SbPolyhedronTube(rmin,rmax,dz);
  }break;
  case 6:{
    double rmin = aPars[0];
    double rmax = aPars[1];
    double dz = aPars[2];
    double phmin = aPars[3] * M_PI/180;
    double phmax = aPars[4] * M_PI/180;
    result = new SbPolyhedronTubs(rmin,rmax,dz,phmin,phmax-phmin);
  }break;
  case 7:{
    double dz = aPars[0];
    double rmin1 = aPars[1];
    double rmax1 = aPars[2];
    double rmin2 = aPars[3];
    double rmax2 = aPars[4];
    result = new SbPolyhedronCone(rmin1,rmax1,rmin2,rmax2,dz);
  }break;
  case 8:{
    double dz = aPars[0];
    double rmin1 = aPars[1];
    double rmax1 = aPars[2];
    double rmin2 = aPars[3];
    double rmax2 = aPars[4];
    double phmin = aPars[5] * M_PI/180;
    double phmax = aPars[6] * M_PI/180;
    //FIXME : should give phstart, dphi
    result = new SbPolyhedronCons(rmin1,rmax1,rmin2,rmax2,dz,phmin,phmax);
  }break;
  case 9:{
    //FIXME : to be checked.
    double rmin = aPars[0];
    double rmax = aPars[1];
    double phi = aPars[2] * M_PI/180;
    double phmax = aPars[3] * M_PI/180;
    double the = aPars[4] * M_PI/180;
    double thmax = aPars[5] * M_PI/180;
    double dphi = phmax - phi;
    double dthe = thmax - the;
    result = new SbPolyhedronSphere(rmin,rmax,phi,dphi,the,dthe);
  }break;
  case 10:{
    // 'PARA' is a parallelpiped.   It has 6 parameters,  the half length   
    //        in x, the half length in y, the half length in z, the angle   
    //        w.r.t. the y axis from the centre  of the low y edge to the   
    //        centre of the  high y edge, and the theta  phi polar angles   
    //        from the centre of the low z face to the centre of the high   
    //        z face.                                                       
    double DX   = aPars[0];
    double DY   = aPars[1];
    double DZ   = aPars[2];
    double TXY  = aPars[3];
    double TXZ  = aPars[4];
    double TYZ  = aPars[5];
    double TX   = TXZ * M_PI/180;
    double TY   = TYZ * M_PI/180;
    double H1   = DY;
    double BL1  = DX;
    double TL1  = DX;
    double TTH1 = TXY * M_PI/180;
    double H2   = DY;
    double BL2  = DX;
    double TL2  = DX;
    double TTH2 = TXY * M_PI/180;
    result = new SbPolyhedronTrap(DZ,
				  TX,TY,
				  H1,BL1,TL1,TTH1,
				  H2,BL2,TL2,TTH2);
  }break;
  case 11:{
    double phi = aPars[0] * M_PI/180; 
    double dphi = aPars[1] * M_PI/180; 
    int npdv = (int)aPars[2];
    int nz = (int)aPars[3];
    double* zs = new double[nz];
    double* rmins = new double[nz];
    double* rmaxs = new double[nz];
    for(int index=0;index<nz;index++) {
      zs[index] = aPars[4+3*index];
      rmins[index] = aPars[4+3*index+1];
      rmaxs[index] = aPars[4+3*index+2];
    }	
    result = new SbPolyhedronPgon(phi,dphi,npdv,nz,zs,rmins,rmaxs);
    delete [] zs;
    delete [] rmins;
    delete [] rmaxs;
  }break;
  case 12:{
    double phi = aPars[0] * M_PI/180; 
    double dphi = aPars[1] * M_PI/180; 
    int nz = (int)aPars[2]; //At least = 1.
    double* zs = new double[nz];
    double* rmins = new double[nz];
    double* rmaxs = new double[nz];
    for(int index=0;index<nz;index++) {
      zs[index] = aPars[3+3*index];
      rmins[index] = aPars[3+3*index+1];
      rmaxs[index] = aPars[3+3*index+2];
    }	
    result = new SbPolyhedronPcon(phi,dphi,nz,zs,rmins,rmaxs);
    delete [] zs;
    delete [] rmins;
    delete [] rmaxs;
  }break;
  /*
  //case 13:
  //  ELTU
  //  break;

  case 14:{
    double RMIN = aPars[0]; 
    double RMAX = aPars[1]; 
    double DZ   = aPars[2]; 
    double PHI  = aPars[3]; 
    GoAddGEANT_HYPE_ToNode (Class.onode,Class.matrix,RMIN,RMAX,DZ,PHI);
  }break;
  */
  default:
    std::ostream& out = sSession->cout();
    out << "soshap : " 
	<< " " << *aShape << " is a bad shape type." << std::endl;
    break;
  }

  if(!result) return;

  if(aRotation[9]!=0) {
    // Rotated then placed :
    //SbMatrix matrix(aRotation[0],aRotation[1],aRotation[2],0,
    //                aRotation[3],aRotation[4],aRotation[5],0,
    //                aRotation[6],aRotation[7],aRotation[8],0,
    //                           0,           0,           0,1);
    //result->Transform(HEPVis::SbRotation(matrix),SbVec3d(0, 0, 0));

    //NOTE : the below Transform had not been tested.
    result->Transform(
      HEPVis::SbRotation(aRotation[0],aRotation[1],aRotation[2],0,
                         aRotation[3],aRotation[4],aRotation[5],0,
                         aRotation[6],aRotation[7],aRotation[8],0,
                                    0,           0,           0,1),
      SbVec3d(0, 0, 0));

    result->Transform(HEPVis::SbRotation(SbVec3d(1, 0, 0),0),
	      SbVec3d(aTranslation[0],aTranslation[1],aTranslation[2]));
  } else {
    result->Transform(HEPVis::SbRotation(SbVec3d(1, 0, 0),0),
	      SbVec3d(aTranslation[0],aTranslation[1],aTranslation[2]));
  }

  SoSeparator* separator = sGC->getSeparator();

  separator->addChild(sGC->getLightModel());
  separator->addChild(sGC->getDrawStyle());

  SoPolyhedron* soPolyhedron = sGC->getPolyhedron(*result);
  separator->addChild(soPolyhedron);

  delete result;

}
