// this :
#include <G3Lab/Types.h>

// Slash :
#include <Slash/Data/IAccessor.h>
#include <Slash/Data/IIterator.h>
#include <Slash/Tools/smanip.h>
#include <Slash/Tools/FString.h>

// Lib :
#include <Lib/Value.h>

// Inventor :
#include <Inventor/nodes/SoSeparator.h>

// HEPVis :
#include <HEPVis/SbProjector.h>
#include <HEPVis/SbStyle.h>
#include <HEPVis/nodes/SoSceneGraph.h>
#include <HEPVis/nodes/SoHighlightMaterial.h>
#include <HEPVis/nodes/SoPolyhedron.h>
#include <HEPVis/nodekits/SoRegion.h>
#include <HEPVis/misc/SoStyleCache.h>

// G3Lab :
#include <G3Lab/GF77.h>
#include <G3Lab/gmanip.h>
#include <G3Lab/SoF77.h>
#include <G3Lab/Iterator.h>

#include <cstdio>

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

G3Lab::VOLU_Type::VOLU_Type(Slash::Core::ISession& aSession)
:OnX::InventorAccessor(aSession)
,fSession(aSession)
,fType("VOLU")
{
  addProperty("id",Lib::Property::INTEGER);
  addProperty("name",Lib::Property::STRING);
  addProperty("numed",Lib::Property::INTEGER);
  addProperty("shape",Lib::Property::STRING);
}

G3Lab::VOLU_Type::~VOLU_Type() {G3Lab::clear_ids(f_ids);}

void* G3Lab::VOLU_Type::cast(const std::string& aClass) const {
  if_Lib_SCast(G3Lab::VOLU_Type)
  else return OnX::InventorAccessor::cast(aClass);
}

std::string G3Lab::VOLU_Type::name() const {return fType;}

void G3Lab::VOLU_Type::fill_ids() {G3Lab::fill_ids(ogvoln(),f_ids);}

Slash::Data::IIterator* G3Lab::VOLU_Type::iterator() {
  return new G3Lab::Iterator(f_ids);
}

Slash::Core::IValue* G3Lab::VOLU_Type::findValue(Slash::Data::IAccessor::Data aIdentifier,const std::string& aName,void*) {
  int* obj = (int*)aIdentifier;
  if(aName=="id") {
    return new Lib::Value(*obj);
  } else if(aName=="name") {
    return new Lib::Value(gmanip::volumeName(*obj));
  } else if(aName=="shape") {
    return new Lib::Value(gmanip::volumeShape(*obj));
  } else if(aName=="numed") {
    return new Lib::Value(gmanip::volumeNumed(*obj));
  } else {
    return new Lib::Value();
  }
}

void G3Lab::VOLU_Type::visualize(Slash::Data::IAccessor::Data aIdentifier,void*){
  if(!fSoRegion) return;

  SoF77_SetGC(&fSoGC);

  int* obj = (int*)aIdentifier;

  SoSceneGraph* separator = new SoSceneGraph;
  // Build name (for picking) :
  char s[64];
  ::sprintf(s,"VOLU/0x%lx",(unsigned long)obj); //and not *obj
  separator->setString(s);

  fSoGC.setSeparator(separator);
  
  separator->addChild(fSoGC.getHighlightMaterial());

  std::string name = gmanip::volumeName(*obj);

  //Get a style corresponding to VOLU type.
  fillSoGC(std::string("VOLU_")+name);

  gmanip::setVolumeDrawingAttribute("*   ","SEEN",GVOL_INV_SON_INV);
  gmanip::setVolumeVisibility(name);

  std::vector<int> vols;
  std::vector<int> dvols;
  gmanip::representVolume(gmanip::topVolume(),vols,dvols);

  // Send scene graph to the viewing region :
  fSoRegion->doIt(SbAddNode(separator,"staticScene"));

  SoF77_SetGC(0);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

G3Lab::VERT_Type::VERT_Type(Slash::Core::ISession& aSession)
:OnX::InventorAccessor(aSession)
,fSession(aSession)
,fType("VERT")
{
  addProperty("id",Lib::Property::INTEGER);
  addProperty("x",Lib::Property::DOUBLE);
  addProperty("y",Lib::Property::DOUBLE);
  addProperty("z",Lib::Property::DOUBLE);
}

G3Lab::VERT_Type::~VERT_Type(){G3Lab::clear_ids(f_ids);}

void* G3Lab::VERT_Type::cast(const std::string& aClass) const {
  if_Lib_SCast(G3Lab::VERT_Type)
  else return OnX::InventorAccessor::cast(aClass);
}

std::string G3Lab::VERT_Type::name() const {return fType;}
  
void G3Lab::VERT_Type::fill_ids() {G3Lab::fill_ids(ogvern(),f_ids);}

Slash::Data::IIterator* G3Lab::VERT_Type::iterator() {
  return new G3Lab::Iterator(f_ids);
}

Slash::Core::IValue* G3Lab::VERT_Type::findValue(Slash::Data::IAccessor::Data aIdentifier,const std::string& aName,void*) {
  int* obj = (int*)aIdentifier;
  float rval;
  if(aName=="id") {
    return new Lib::Value(*obj);
  } else if(aName=="x") {
    int iprop = 1;
    ogverv(obj,&iprop,&rval);
    return new Lib::Value((double)rval);
  } else if(aName=="y") {
    int iprop = 2;
    ogverv(obj,&iprop,&rval);
    return new Lib::Value((double)rval);
  } else if(aName=="z") {
    int iprop = 3;
    ogverv(obj,&iprop,&rval);
    return new Lib::Value((double)rval);
  } else {
    return new Lib::Value();
  }
}

void G3Lab::VERT_Type::visualize(Slash::Data::IAccessor::Data aIdentifier,void*){
  if(!fSoRegion) return;

  SoF77_SetGC(&fSoGC);

  int* obj = (int*)aIdentifier;

  // Build name (for picking) :
  char s[64];
  ::sprintf(s,"VERT/0x%lx",(unsigned long)obj); //and not *obj

  SoSceneGraph* separator = new SoSceneGraph;
  separator->setString(s);

  fSoGC.setSeparator(separator);
  
  ogverd(obj);

  // Send scene graph to the viewing region :
  fSoRegion->doIt(SbAddNode(separator,"dynamicScene"));

  SoF77_SetGC(0);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

G3Lab::KINE_Type::KINE_Type(Slash::Core::ISession& aSession)
:OnX::InventorAccessor(aSession)
,fSession(aSession)
,fType("KINE")
{
  addProperty("id",Lib::Property::INTEGER);
  addProperty("type",Lib::Property::STRING);
  addProperty("ipart",Lib::Property::INTEGER);
  addProperty("energy",Lib::Property::DOUBLE);
  addProperty("tracking",Lib::Property::INTEGER);
}

G3Lab::KINE_Type::~KINE_Type() {G3Lab::clear_ids(f_ids);}

void* G3Lab::KINE_Type::cast(const std::string& aClass) const {
  if_Lib_SCast(G3Lab::KINE_Type)
  else return OnX::InventorAccessor::cast(aClass);
}

std::string G3Lab::KINE_Type::name() const {return fType;}

void G3Lab::KINE_Type::fill_ids() {G3Lab::fill_ids(ogkinn(),f_ids);}

Slash::Data::IIterator* G3Lab::KINE_Type::iterator() {
  return new G3Lab::Iterator(f_ids);
}

Slash::Core::IValue* G3Lab::KINE_Type::findValue(Slash::Data::IAccessor::Data aIdentifier,const std::string& aName,void*) {
  int* obj = (int*)aIdentifier;
  if(aName=="id") {
    return new Lib::Value(*obj);
  } else if(aName=="type") {
    int iname = ogkinm(obj);
    std::string s = gmanip::h2s(iname,20);
    Slash::strip(s);
    return new Lib::Value(s);
  } else if(aName=="ipart") {
    int ipart = ogkipa(obj);
    return new Lib::Value(ipart);
  } else if(aName=="energy") {
    float e = ogkien(obj);
    return new Lib::Value((double)e);
  } else if(aName=="tracking") {
    int itk = ogkitk(obj);
    return new Lib::Value(itk);
  } else {
    return new Lib::Value();
  }
}

void G3Lab::KINE_Type::visualize(Slash::Data::IAccessor::Data aIdentifier,void*) {
  if(!fSoRegion) return;

  int* obj = (int*)aIdentifier;

  //Get a style corresponding to KINE type.
 {int iname = ogkinm(obj);
  std::string s = gmanip::h2s(iname,20);
  Slash::strip(s);
  std::string style = "KINE_"+s;
  fillSoGC(style);}

  SoF77_SetGC(&fSoGC);

  // Build name (for picking) :
  char s[64];
  ::sprintf(s,"KINE/0x%lx",(unsigned long)obj); //and not *obj

  SoSceneGraph* separator = new SoSceneGraph;
  separator->setString(s);

  fSoGC.setSeparator(separator);
  
  int idscnd = 0; //Do not draw secondaries.
  ogkind(obj,&idscnd);

  // Send scene graph to the viewing region :
  fSoRegion->doIt(SbAddNode(separator,"dynamicScene"));

  SoF77_SetGC(0);
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
/*
//////////////////////////////////////////////////////////////////////////////
void G3Lab::declareUDETsHitsAndDIGIs (
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int objn = ogdetn();
  if(objn<=0) return;


  int ival;
  float rval;
  Slash::FString sval(20);

  for(int index=0;index<objn;index++) {
    int obj = index+1;
    int iprop = 1;
    ogdetv(&obj,&iprop,&ival,&rval,sval,iarr);
    std::string NAME = sval.std_string();
    Slash::strip(NAME);

    int iprop = 2;
    ogdetv(&obj,&iprop,&ival,&rval,sval,iarr);
    std::string USET = sval.std_string();
    Slash::strip(USET);

    std::string name = "DIGI_" + USET + "_" + NAME;
  }
}
*/
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

G3Lab::DIGI_Type::DIGI_Type(Slash::Core::ISession& aSession)
:OnX::InventorAccessor(aSession)
,fSession(aSession)
,fType("DIGI")
{
  addProperty("id",Lib::Property::INTEGER);
  addProperty("voln",Lib::Property::INTEGER);
  addProperty("parn",Lib::Property::INTEGER);
}

G3Lab::DIGI_Type::~DIGI_Type() {G3Lab::clear_ids(f_ids);}

void* G3Lab::DIGI_Type::cast(const std::string& aClass) const {
  if_Lib_SCast(G3Lab::DIGI_Type)
  else return OnX::InventorAccessor::cast(aClass);
}

std::string G3Lab::DIGI_Type::name() const {return fType;}

void G3Lab::DIGI_Type::fill_ids() {
  if(!uset_udet(fUSET,fUDET)) return;
  int objn = gmanip::detectorNumberOfDIGIs(fUSET,fUDET);
  G3Lab::fill_ids(objn,f_ids);  
}

Slash::Data::IIterator* G3Lab::DIGI_Type::iterator() {
  return new G3Lab::Iterator(f_ids);
}

Slash::Core::IValue* G3Lab::DIGI_Type::findValue(Slash::Data::IAccessor::Data aIdentifier,const std::string& aName,void*) {
  int* obj = (int*)aIdentifier;
  if(aName=="id") {
    return new Lib::Value(*obj);
  } else if(aName=="voln") {
    std::vector<int> vols;
    std::vector<int> digs;
    if(!gmanip::detectorIthDIGI(fUSET,fUDET,*obj,vols,digs)) {
      return new Lib::Value();
    }
    return new Lib::Value((int)vols.size());
  } else if(aName=="parn") {
    std::vector<int> vols;
    std::vector<int> digs;
    if(!gmanip::detectorIthDIGI(fUSET,fUDET,*obj,vols,digs)) {
      return new Lib::Value();
    }
    return new Lib::Value((int)digs.size());
  } else {
    return new Lib::Value();
  }
}

void G3Lab::DIGI_Type::visualize(Slash::Data::IAccessor::Data aIdentifier,void*) {
  if(!fSoRegion) return;

  SoF77_SetGC(&fSoGC);

  int* obj = (int*)aIdentifier;

  std::vector<int> vols;
  std::vector<int> digs;
  if(!gmanip::detectorIthDIGI(fUSET,fUDET,*obj,vols,digs)) return;

  // Build name (for picking) :
  char s[64];
  ::sprintf(s,"DIGI/0x%lx",(unsigned long)obj); //and not *obj

  SoSceneGraph* separator = new SoSceneGraph;
  separator->setString(s);

  fSoGC.setSeparator(separator);
  
  separator->addChild(fSoGC.getHighlightMaterial());

  std::vector<int> dvols;
  gmanip::detectorVolumes(fUSET,fUDET,dvols);

  // Assume that last volume is the hit volume !!!
  std::string name = gmanip::h2s(dvols[vols.size()-1]); 

  gmanip::setVolumeDrawingAttribute("*   ","SEEN",GVOL_INV_SON_INV);
  gmanip::setVolumeVisibility(name);

  if(!gmanip::representVolume(gmanip::topVolume(),vols,dvols)) {
    gmanip::setVolumeDrawingAttribute("*   ","SEEN",GVOL_VIS_SON_VIS);
    if(!gmanip::representVolume(gmanip::topVolume(),vols,dvols)) {
      //FIXME : CWarnF ("Can't draw volume:\n");
      for(unsigned int index=0;index<vols.size();index++) {
	/*
	name          = gmanip::h2s(dvols[count]);
	CWarnF        ("%s %d\n",name,vols[count]);
	CStringDelete (name);
	*/
      }
    }
  }

  // Send scene graph to the viewing region :
  fSoRegion->doIt(SbAddNode(separator,"dynamicScene"));

  SoF77_SetGC(0);
}

bool G3Lab::DIGI_Type::uset_udet(std::string& aUSET,std::string& aUDET) {
  aUSET = "";
  aUDET = "";
  std::string location;
  if(!fSession.parameterValue("digi.location",location)) {
    std::ostream& out = fSession.cout();
    out << "G3Lab::DIGI_Type::iterator : "
	<< " no location given." << std::endl;
    return false;
  }
  std::vector<std::string> words;
  Slash::words(location,"_",false,words);
  if(words.size()!=2) {
    std::ostream& out = fSession.cout();
    out << "G3Lab::DIGI_Type::iterator : "
	<< " bad location format." << std::endl;
    return false;
  }
  aUSET = words[0];
  aUDET = words[1];
  return true;
}
