// this :
#include <G3Lab/gmanip.h>

#include <Lib/fmanip.h>

#include <G3Lab/GF77.h>

#define MINIMUM(a,b) ((a)<(b)?a:b)

void G3Lab::gmanip::initialize(){oginit();}

void G3Lab::gmanip::loadGeometry(int aFormat,const std::string& aName,int aUnit){
  // Get full name :
  std::string name = Lib::fmanip::name(aName);

  if( (aFormat==GFILE_FZ) ||
      (aFormat==GFILE_FA) ||
      (aFormat==GFILE_FX)) {
    Lib::fmanip::bindToFortranUnit(name,aUnit);
  } else if( (aFormat==GFILE_RZ) ||
	     (aFormat==GFILE_RA) ||
	     (aFormat==GFILE_RX) ) {
    Lib::fmanip::fortranLink(name);
  }
  int fmt = aFormat;
  int unit = aUnit;
  int size = name.size();
  char* buffer = new char[size+1];
  for(int index=0;index<size;index++) buffer[index] = name[index];
#ifdef WIN32
  ogrgeo(&fmt,buffer,size,&size,&unit);
#else
  ogrgeo(&fmt,buffer,&size,&unit);
#endif
  delete [] buffer;
}

void G3Lab::gmanip::saveGeometry(int aFormat,const std::string& aName,int aUnit){
  std::string name = aName;

  if( (aFormat==GFILE_FZ) || 
      (aFormat==GFILE_FA) || 
      (aFormat==GFILE_FX) ) {
    Lib::fmanip::bindToFortranUnit(name,aUnit);
  } else if( (aFormat==GFILE_RZ) ||
	     (aFormat==GFILE_RA) ||
	     (aFormat==GFILE_RX) ) {
    Lib::fmanip::fortranLink(name);
  }

  int fmt = aFormat;
  int unit = aUnit;
  int size = name.size();
  char* buffer = new char[size+1];
  for(int index=0;index<size;index++) buffer[index] = name[index];
#ifdef WIN32
  ogwgeo(&fmt,buffer,size,&size,&unit);
#else
  ogwgeo(&fmt,buffer,&size,&unit);
#endif
  delete [] buffer;
}

void G3Lab::gmanip::readEvent(int aUnit){
  int unit = aUnit;
  ogrevt(&unit);
}

int G3Lab::gmanip::produceEvent(){return ogevbg();}

void G3Lab::gmanip::clearEvent(){gtrigc();}

std::vector<std::string> G3Lab::gmanip::volumeNames(){
  std::vector<std::string> names;
  int number = ogvoln ();
  if(number==0) return names;
  for(int index=0;index<number;index++) { 
    int ivo = index+1;
    int iname = ogvonm(&ivo);
    names.push_back(h2s(iname));
  }
  return names;
}

void G3Lab::gmanip::printVolumes(){
  int mode = 0;
  gpvolu(&mode);
}

std::string G3Lab::gmanip::topVolume(){
  int iname = 0;
  int stat = ogvolt(&iname);
  if(!stat) return "";
  return h2s(iname);
}

std::string G3Lab::gmanip::volumeName(int aVolume){
  // aVolume is in [1,ogvoln()]
  int number = ogvoln ();
  if(aVolume<=0) return "";
  if(aVolume>number) return "";
  int ivo = aVolume;
  int iname = ogvonm(&ivo); 
  return h2s(iname);
}

std::string G3Lab::gmanip::volumeShape(int aVolume){
  // aVolume is in [1,ogvoln()]
  static std::string shapes[30] = {
    "BOX ","TRD1","TRD2","TRAP","TUBE","TUBS","CONE",
    "CONS","SPHE","PARA","PGON","PCON","ELTU","HYPE",
    "    ","    ","    ","    ","    ","    ",
    "    ","    ","    ","    ","    ","    ",
    "    ",
    "GTRA","CTUB","    "};
  int number = ogvoln ();
  if(aVolume<=0) return "";
  if(aVolume>number) return "";
  int ivo = aVolume;
  int index = ogvosh(&ivo);
  if(index<=0) return "    ";
  if(index>=31) return "    ";
  return shapes[index-1];
}

int G3Lab::gmanip::volumeNumed(int aVolume){
  // aVolume is in [1,ogvoln()]
  int number = ogvoln ();
  if(aVolume<=0) return 0;
  if(aVolume>number) return 0;
  int ivo = aVolume;
  return ogvomd(&ivo);
}

int G3Lab::gmanip::volumeNumberOfSons(const std::string& aName){
  if(aName.empty()) return 0;
  int iname = s2h(aName);
  int nson,ndiv;
  ogvons(&iname,&nson,&ndiv);
  return nson;
}

std::string G3Lab::gmanip::volumeIthSon(const std::string& aName,int aSon){
  if(aName.empty()) return "";
  int iname = s2h(aName);
  int son = aSon + 1;
  int ison = 0;
  ogvois(&son,&iname,&ison);
  return h2s(ison);
}

std::vector<std::string> G3Lab::gmanip::volumeUpperVolumes(const std::string& aName){
  std::vector<std::string> list;
  int iname = s2h(aName);
  int mothn = 0;
  int moths[64];
  int mothx = 64;
  ogvomo(&iname,&mothn,moths,&mothx);
  if(mothn==0) return list;
  for(int index=0;index<mothn;index++) { 
    int ivo = moths[index];
    iname = ogvonm(&ivo);
    list.push_back(h2s(iname));
  }
  return list;  
}

void G3Lab::gmanip::setVolumeVisibility(const std::string& aName){
  if(aName.empty()) return;
  int seen = volumeDrawingAttribute(aName,"SEEN");
  //if(seen==GVOL_VIS_SON_VIS) 
  //  setVolumeDrawingAttribute(aName,"SEEN", GVOL_VIS_SON_VIS);
  //else if(seen==GVOL_VIS_SON_INV) 
  //  setVolumeDrawingAttribute(aName,"SEEN", GVOL_VIS_SON_INV);
  //else
  if(seen==GVOL_INV_SON_VIS) 
    setVolumeDrawingAttribute(aName,"SEEN", GVOL_VIS_SON_VIS);
  else if(seen==GVOL_INV_SON_INV) 
    setVolumeDrawingAttribute(aName,"SEEN", GVOL_VIS_SON_INV);
  enableMothers(aName);
}

void G3Lab::gmanip::setVolumeDrawingAttribute(const std::string& aName,const std::string& aAttr,int aValue){
  //  *    1:  'WORK'   0=volume inactive   1=active volume            *
  //  *    2:  'SEEN'   0=unseen 1=seen -1,-2=tree unseen -3=tree opt. *
  //  *    3:  'LSTY'   line style 1,2,3,...                           *
  //  *    4:  'LWID'   line width 1,2,3,...                           *
  //  *    5:  'COLO'   color code 1,2,3,...                           *
  //  *    6:  'FILL'   fill area  0,1,2,...                           *
  //  *    7:  'SET '   set number associated to this volume           *
  //  *    8:  'DET '   detector number associated to this volume      *
  //  *    9:  'DTYP'   detector type (1,2)                            *
  //  *   10:  'NODE'   <>0=a node is created for 3D (PIONS,GMR,etc.)  *
  // For 'SEEN':
  //  1  vis vis  (default)
  //  0  inv vis
  // -1  inv inv
  // -2  vis inv
  if(aName.empty()) return;
  if(aAttr.empty()) return;
  int iname = s2h(aName);
  int iatt = s2h(aAttr);
  int ival = aValue;
  ogsatt(&iname,&iatt,&ival);
}

int G3Lab::gmanip::volumeDrawingAttribute(const std::string& aName,const std::string& aAttr){
  if(aName.empty()) return 0;
  if(aAttr.empty()) return 0;
  int iname = s2h(aName);
  int iatt = s2h(aAttr);
  int ival = ogfatt(&iname,&iatt);
  return ival;
}

bool G3Lab::gmanip::representVolume(const std::string& aName,const std::vector<int>& aVols,const std::vector<int>& aDVols){
  if(aName.empty()) return false;
  int iname = s2h(aName);
  int voln = MINIMUM(aVols.size(),aDVols.size());
  int vols[50];
  int dvols[50];
  int index;
  for(index=0;index<voln;index++) vols[index] = aVols[index];
  for(index=0;index<voln;index++) dvols[index] = aDVols[index];
  int status = ogvold(&iname,&voln,vols,dvols);
  return status==1 ? true : false;
}

int G3Lab::gmanip::detectorNumberOfHits(const std::string& aSet,const std::string& aDet){
  if(aSet.empty()) return 0;
  if(aDet.empty()) return 0;
  int iuset = s2h(aSet);
  int iudet = s2h(aDet);
  return oghitn(&iuset,&iudet);
}

bool G3Lab::gmanip::detectorIthHit(const std::string& aSet,const std::string& aDet,int aHit,std::vector<int>& aVols,std::vector<float>& aHits){
  aVols.clear();
  aHits.clear();
  if(aSet.empty()) return false;
  if(aDet.empty()) return false;
  int iuset = s2h(aSet);
  int iudet = s2h(aDet);
  int voln,hitn,itra;
  int hit = aHit;
  int vols[50];
  float hits[50];
  int status = oghitv(&iuset,&iudet,
		      &hit,&voln,vols,&hitn,hits,&itra);
  int index;
  for(index=0;index<voln;index++) aVols.push_back(vols[index]);
  for(index=0;index<hitn;index++) aHits.push_back(hits[index]);
  return status==1 ? true : false;
}

int G3Lab::gmanip::detectorNumberOfDIGIs(const std::string& aSet,const std::string& aDet){
  if(aSet.empty()) return 0;
  if(aDet.empty()) return 0;
  int iuset = s2h(aSet);
  int iudet = s2h(aDet);
  return ogdign(&iuset,&iudet);
}

bool G3Lab::gmanip::detectorIthDIGI(const std::string& aSet,const std::string& aDet,int aDigi,std::vector<int>& aVols,std::vector<int>& aDigs){
  aVols.clear();
  aDigs.clear();
  if(aSet.empty()) return false;
  if(aDet.empty()) return false;
  int iuset = s2h(aSet);
  int iudet = s2h(aDet);
  int vols[50];
  int digs[50];
  int voln,dign;
  int dig = aDigi;
  int status = ogdigv(&iuset,&iudet,&dig,&voln,vols,&dign,digs);
  int index;
  for(index=0;index<voln;index++) aVols.push_back(vols[index]);
  for(index=0;index<dign;index++) aDigs.push_back(digs[index]);
  return status==1 ? true : false;
}

void G3Lab::gmanip::detectorVolumes(const std::string& aSet,const std::string& aDet,std::vector<int>& aVols){
  aVols.clear();
  if(aSet.empty()) return;
  if(aDet.empty()) return;
  int iuset = s2h(aSet);
  int iudet = s2h(aDet);
  int voln;
  int vols[50];
  ogdeto(&iuset,&iudet,&voln,vols);
  for(int index=0;index<voln;index++) aVols.push_back(vols[index]);
}

void G3Lab::gmanip::representTrack(int aItra,bool aDrawSecondary){
  int itra = aItra;
  int idscnd = aDrawSecondary?1:0;
  ogkind(&itra,&idscnd);
}

void G3Lab::gmanip::representVertex(int aIvert){
  int ivert = aIvert;
  ogverd(&ivert);
}

void G3Lab::gmanip::enableMothers(const std::string& aName){
  if(aName.empty()) return;
  std::vector<std::string> moths  = volumeUpperVolumes(aName);
  for(unsigned int index=0;index<moths.size();index++) { 
    std::string& mother = moths[index];
    int seen = volumeDrawingAttribute(mother,"SEEN");
    //if(seen==GVOL_VIS_SON_VIS) 
    //  setVolumeDrawingAttribute(mother,"SEEN", GVOL_VIS_SON_VIS);
    //else if(seen==GVOL_INV_SON_VIS) 
    //  setVolumeDrawingAttribute(mother,"SEEN", GVOL_INV_SON_VIS);
    //else
    if(seen==GVOL_VIS_SON_INV) 
      setVolumeDrawingAttribute(mother,"SEEN", GVOL_VIS_SON_VIS);
    else if(seen==GVOL_INV_SON_INV) 
      setVolumeDrawingAttribute(mother,"SEEN", GVOL_INV_SON_VIS);
    enableMothers(mother);
  }
}

int G3Lab::gmanip::s2h(const std::string& aString){
  int size = aString.size();
  int ival = 0;
  char* buffer = new char[size+1]; //Add one, just in case size is 0.
  for(int index=0;index<size;index++) buffer[index] = aString[index];
  int bsize = size;
  int vsize = 4;
#ifdef WIN32
  uctoh(buffer,bsize,&ival,&bsize,&vsize);
#else
  uctoh(buffer,&ival,&bsize,&vsize);
#endif
  delete [] buffer;
  return ival;
}

std::string G3Lab::gmanip::h2s(int aHollerith,int aSize){
  int iname = aHollerith;
  char* buffer = new char[aSize+1]; //Add one, just in case aSize is 0.
  int vsize = 4;
  int bsize = aSize;
#ifdef WIN32
  uhtoc(&iname,&vsize,buffer,bsize,&bsize);
#else
  uhtoc(&iname,&vsize,buffer,&bsize);
#endif
  std::string s;
  for(int index=0;index<aSize;index++) {
    char c = buffer[index];
    if(c<=31) break;
    //if(c=='\0') break;
    s += c;
  }
  delete [] buffer;
  return s;
}


