//
//  All functions here should be OnX callbacks, that is to say
// functions with signature :
//   extern "C" {
//     void callback_without_arguments(Slash::UI::IUI&);
//     void callback_with_arguments(Slash::UI::IUI&,
//                                  const std::vector<std::string>&);
//   }
//

// Slash :
#include <Slash/Core/ISession.h>
#include <Slash/UI/IUI.h>

extern "C" {

//////////////////////////////////////////////////////////////////////////////
// Example of a callback without graphics.
//////////////////////////////////////////////////////////////////////////////

void @@package@@_hello(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  std::ostream& out = aUI.session().cout();
  out << "@@package@@_hello :"
      << " hello !"
      << std::endl;
}

} // extern "C"


#include <OnX/Helpers/Inventor.h>

// Inventor :
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSphere.h>

// HEPVis :
#include <HEPVis/nodekits/SoDisplayRegion.h>

extern "C" {

//////////////////////////////////////////////////////////////////////////////
// Example of a callback with Inventor graphics.
//////////////////////////////////////////////////////////////////////////////

void @@package@@_sphere(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  if(!ui_check_args(aUI,aArgs,1)) return; //defined in OnX/Helpers/OnX.

  //find_SoRegion is a helper defined in OnX/Helpers/OnX.h
  SoRegion* soRegion = find_SoRegion(aUI,aArgs[0]);
  if(!soRegion) return;

  // Check if it is a visualization region :
  if(!soRegion->isOfType(SoDisplayRegion::getClassTypeId())) return;
  SoDisplayRegion* displayRegion = (SoDisplayRegion*)soRegion;

  // Get "static scene" separator :
  SoSeparator* scene = (SoSeparator*)displayRegion->getStaticScene();
  if(!scene) return;

  SoSphere* soSphere = new SoSphere;
  scene->addChild(soSphere);

}

}

//////////////////////////////////////////////////////////////////////////////
// Example of a callback with G4 code.
//////////////////////////////////////////////////////////////////////////////

#include <Slash/Tools/Manager.h>

// G4Lab :
#include <G4Lab/Interfaces/IGeant4Manager.h>

// Geant4 :
#include <G4RunManager.hh>

inline IGeant4Manager* find_Geant4Manager(Slash::Core::ISession& aSession) {
  IGeant4Manager* g4Manager = 
    Slash_findManager(aSession,"Geant4Manager",IGeant4Manager);
  if(!g4Manager) {
    std::ostream& out = aUI.session().cout();
    out << "Geant4Manager not found." << std::endl;
    return 0;
  }
  return g4Manager;
}

extern "C" {

void @@package@@_beam_on(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  if(!ui_check_args(aUI,aArgs,1)) return; //defined in OnX/Helpers/OnX.

  IGeant4Manager* g4Manager = find_Geant4Manager(aUI.session());
  if(!g4Manager) return;
  G4RunManager* runManager = g4Manager->runManager();
  if(!runManager) return;

  int n_event;
  if(!ui_check(aUI,aArgs[0],n_event)) return; //helper to convert to an int.
  runManager->BeamOn(n_event);

}

} // extern "C"
