// this :
#include "Detector.hh"

// Geant4 :
#include "G4Material.hh"
#include "G4Box.hh"
#include "G4LogicalVolume.hh"
#include "G4PVPlacement.hh"
#include "G4GeometryManager.hh"
#include "G4UserLimits.hh"

#include "MagneticField.hh"

Detector::Detector()
:stepLimit(0),fpMagField(0) {
  fpMagField = new MagneticField();
}

Detector::~Detector() {
  delete stepLimit;
  delete fpMagField;
}

G4VPhysicalVolume* Detector::Construct() {
  //--------- Material definition ---------

  G4double a, z;
  G4double density;
  G4int nel;

  //Air
  G4Element* N = new G4Element("Nitrogen", "N", z=7., a= 14.01*g/mole);
  G4Element* O = new G4Element("Oxygen"  , "O", z=8., a= 16.00*g/mole);
   
  G4Material* Air = new G4Material("Air", density= 1.0*1.29*mg/cm3, nel=2);
  Air->AddElement(N, 70*perCent);
  Air->AddElement(O, 30*perCent);

  //vacuum
  G4Material* Vacuum = 
    new G4Material("Vacuum", density= 0.0000000001 * 1.29*mg/cm3, nel=2);
  Vacuum->AddElement(N, 70*perCent);
  Vacuum->AddElement(O, 30*perCent);


  // Print all the materials defined.
  //
  //G4cout << G4endl << "The materials defined are : " << G4endl << G4endl;
  //G4cout << *(G4Material::GetMaterialTable()) << G4endl;

  //------------------------------ 
  // World
  //------------------------------ 

  G4double world_length = 4*m;
 
  G4GeometryManager::GetInstance()->SetWorldMaximumExtent(world_length);

  G4double hwl = 0.5*world_length;
  G4Box* world_geom = new G4Box("world",hwl,hwl,hwl);
  G4LogicalVolume* world_LV = 
    //new G4LogicalVolume( world_geom, Vacuum, "World", 0, 0, 0);
    new G4LogicalVolume( world_geom, Air, "World", 0, 0, 0);
  
  G4VPhysicalVolume* world_PV = 
    new G4PVPlacement(0,               // no rotation
                      G4ThreeVector(), // at (0,0,0)
                      world_LV,        // its logical volume
		      "World",         // its name
                      0,               // its mother  volume
                      false,           // no boolean operations
                      0);              // copy number


  //------------------------------ 
  // Target
  //------------------------------
  
  G4Box* target_geom = new G4Box("target",25.*cm,25*cm,25*cm);
  G4LogicalVolume* target_LV = 
    new G4LogicalVolume(target_geom,Vacuum,"Target",0,0,0);
  new G4PVPlacement(0,                // no rotation
		    G4ThreeVector(),  // at (0,0,0)
                    target_LV,        // its logical volume
                    "Target",         // its name
                    world_LV,         // its mother  volume
                    false,            // no boolean operations
                    0);               // copy number 

  //G4double maxStep = 0.01*mm;
  G4double maxStep = 0.1*mm;
  //G4double maxStep = 0.000005*mm;
  //G4double maxLength = 300.0*cm; 
  G4double maxLength = 10.0*m; 
  //G4double maxTime = 10.0*ns;
  G4double maxTime = 1.0*s;
  G4double minEkin = 0.001*MeV;
  G4double minRange = 0.001*mm;
  G4UserLimits *stepLimit = 
    new G4UserLimits(maxStep,maxLength,maxTime,minEkin,minRange);

  world_LV->SetUserLimits(stepLimit);
  target_LV->SetUserLimits(stepLimit);

  return world_PV;
}
