// this :
#include <G4Lab/ParticleAccessor.h>

// Geant4 :
#include <G4ParticleTable.hh>
#include <G4ParticleDefinition.hh>

// Slash :
#include <Slash/Core/ISession.h>
#include <Slash/Data/IIterator.h>

// Lib :
#include <Lib/Value.h>

std::string G4Lab::ParticleAccessor::accessor_name() {return "Particle";}

//////////////////////////////////////////////////////////////////////////////
G4Lab::ParticleAccessor::ParticleAccessor(
 Slash::Core::ISession& aSession
)
:Lib::BaseAccessor(aSession.printer())
,fSession(aSession)
,fType(accessor_name())
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //Automatically done from G4ParticleDefinition.h :

#define PROP_G4String Lib::Property::STRING
#define PROP_G4double Lib::Property::DOUBLE
#define PROP_G4int    Lib::Property::INTEGER
#define PROP_G4bool   Lib::Property::BOOLEAN


  addProperty("ParticleName",PROP_G4String);
  addProperty("PDGMass",PROP_G4double);
  addProperty("PDGWidth",PROP_G4double);
  addProperty("PDGCharge",PROP_G4double);
  addProperty("PDGSpin",PROP_G4double);
  addProperty("PDGiSpin",PROP_G4int);
  addProperty("PDGiParity",PROP_G4int);
  addProperty("PDGiConjugation",PROP_G4int);
  addProperty("PDGIsospin",PROP_G4double);
  addProperty("PDGIsospin3",PROP_G4double);
  addProperty("PDGiIsospin",PROP_G4int);
  addProperty("PDGiIsospin3",PROP_G4int);
  addProperty("PDGiGParity",PROP_G4int);
  addProperty("PDGMagneticMoment",PROP_G4double);
  addProperty("ParticleType",PROP_G4String);
  addProperty("ParticleSubType",PROP_G4String);
  addProperty("LeptonNumber",PROP_G4int);
  addProperty("BaryonNumber",PROP_G4int);
  addProperty("PDGEncoding",PROP_G4int);
  addProperty("AntiPDGEncoding",PROP_G4int);
  addProperty("QuarkContent",Lib::Property::VECTOR_INTEGER);
  addProperty("AntiQuarkContent",Lib::Property::VECTOR_INTEGER);
  addProperty("ShortLived",PROP_G4bool);
  addProperty("PDGStable",PROP_G4bool);
  addProperty("PDGLifeTime",PROP_G4double);
  addProperty("AtomicNumber",PROP_G4int);
  addProperty("AtomicMass",PROP_G4int);
}
//////////////////////////////////////////////////////////////////////////////
G4Lab::ParticleAccessor::~ParticleAccessor(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fDefs.clear();
}
//////////////////////////////////////////////////////////////////////////////
void* G4Lab::ParticleAccessor::cast(
 const std::string& aClass
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return Lib::BaseAccessor::cast(aClass);
}
//////////////////////////////////////////////////////////////////////////////
std::string G4Lab::ParticleAccessor::name(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fType;
}
namespace G4Lab {
  class ParticleIterator : public virtual Slash::Data::IIterator {
  public: //Slash::Data::IIterator
    virtual Slash::Data::IAccessor::Data object() {
      if(fIterator==fVector.end()) return 0;
      return *fIterator;
    }
    virtual void next() { ++fIterator; }
    virtual void* tag() { return 0;}
  public:
    ParticleIterator(std::vector<G4ParticleDefinition*>& aVector)
      :fVector(aVector) {
      fIterator = fVector.begin();
    }
    virtual ~ParticleIterator() {}
  private:
    std::vector<G4ParticleDefinition*>& fVector;
    std::vector<G4ParticleDefinition*>::iterator fIterator;
  };
}
//////////////////////////////////////////////////////////////////////////////
Slash::Data::IIterator* G4Lab::ParticleAccessor::iterator(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fDefs.size()) {
    if(!buildDefs()) return 0;
  }
  return new ParticleIterator(fDefs);
}
//////////////////////////////////////////////////////////////////////////////
Slash::Core::IValue* G4Lab::ParticleAccessor::findValue(
 Slash::Data::IAccessor::Data aData
,const std::string& aName
,void*
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  G4ParticleDefinition* obj = (G4ParticleDefinition*)aData;

  //Automatically done from G4ParticleDefinition.h :

  if(aName=="ParticleName") {
    return new Lib::Value(obj->GetParticleName());}
  else if(aName=="PDGMass") {
    return new Lib::Value(obj->GetPDGMass());}
  else if(aName=="PDGWidth") {
    return new Lib::Value(obj->GetPDGWidth());}
  else if(aName=="PDGCharge") {
    return new Lib::Value(obj->GetPDGCharge());}
  else if(aName=="PDGSpin") {
    return new Lib::Value(obj->GetPDGSpin());}
  else if(aName=="PDGiSpin") {
    return new Lib::Value(obj->GetPDGiSpin());}
  else if(aName=="PDGiParity") {
    return new Lib::Value(obj->GetPDGiParity());}
  else if(aName=="PDGiConjugation") {
    return new Lib::Value(obj->GetPDGiConjugation());}
  else if(aName=="PDGIsospin") {
    return new Lib::Value(obj->GetPDGIsospin());}
  else if(aName=="PDGIsospin3") {
    return new Lib::Value(obj->GetPDGIsospin3());}
  else if(aName=="PDGiIsospin") {
    return new Lib::Value(obj->GetPDGiIsospin());}
  else if(aName=="PDGiIsospin3") {
    return new Lib::Value(obj->GetPDGiIsospin3());}
  else if(aName=="PDGiGParity") {
    return new Lib::Value(obj->GetPDGiGParity());}
  else if(aName=="PDGMagneticMoment") {
    return new Lib::Value(obj->GetPDGMagneticMoment());}
  else if(aName=="ParticleType") {
    return new Lib::Value(obj->GetParticleType());}
  else if(aName=="ParticleSubType") {
    return new Lib::Value(obj->GetParticleSubType());}
  else if(aName=="LeptonNumber") {
    return new Lib::Value(obj->GetLeptonNumber());}
  else if(aName=="BaryonNumber") {
    return new Lib::Value(obj->GetBaryonNumber());}
  else if(aName=="PDGEncoding") {
    return new Lib::Value(obj->GetPDGEncoding());}
  else if(aName=="AntiPDGEncoding") {
    return new Lib::Value(obj->GetAntiPDGEncoding());}
  else if(aName=="QuarkContent") {
    //  The value of flavor is assigned as follows 
    //  1:d, 2:u, 3:s, 4:c, 5:b, 6:t
    std::vector<unsigned int> orders(1);
    orders[0] = 6;
    std::vector<int> vec(6);
    for(unsigned int index=0;index<6;index++) {
      vec[index] = obj->GetQuarkContent(index+1);
    }
    return new Lib::Value(orders,vec);}
  else if(aName=="AntiQuarkContent") {
    //  The value of flavor is assigned as follows 
    //  1:d, 2:u, 3:s, 4:c, 5:b, 6:t
    std::vector<unsigned int> orders(1);
    orders[0] = 6;
    std::vector<int> vec(6);
    for(unsigned int index=0;index<6;index++) {
      vec[index] = obj->GetAntiQuarkContent(index+1);
    }
    return new Lib::Value(orders,vec);
  }
  else if(aName=="ShortLived") {
    return new Lib::Value(obj->IsShortLived());}
  else if(aName=="PDGStable") {
    return new Lib::Value(obj->GetPDGStable());}
  else if(aName=="PDGLifeTime") {
    return new Lib::Value(obj->GetPDGLifeTime());}
  else if(aName=="AtomicNumber") {
    return new Lib::Value(obj->GetAtomicNumber());}
  else if(aName=="AtomicMass") {
    return new Lib::Value(obj->GetAtomicMass());}
  else {   
    return new Lib::Value();}
}
//////////////////////////////////////////////////////////////////////////////
bool G4Lab::ParticleAccessor::buildDefs(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fDefs.size()) return true; //done.

/*
  if(!fManager.isRunning()) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::PhysicsTableAccessor::buildProcessTables :"
        << " It is needed to have started a run to get physics tables." 
        << std::endl;
    return false;
  }
*/

  G4ParticleTable* particleTable = G4ParticleTable::GetParticleTable();
  if(!particleTable) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::ParticleAccessor::buildDefs :"
        << " Can't get particle table." 
        << std::endl;
    return false;
  }

  //FIXME : the iterator is shared ! How to get our own ?
  //        And then avoid our fDefs std::vector !

  G4ParticleTable::G4PTblDicIterator* particleTableIterator = 
    particleTable->GetIterator();
  if(!particleTableIterator) {
    std::ostream& out = fSession.cout();
    out << "G4Lab::ParticleAccessor::buildDefs :"
        << " Can't get particle table iterator." 
        << std::endl;
    return false;
  }

  particleTableIterator->reset();
  while( (*particleTableIterator)() ){
    G4ParticleDefinition* particle = particleTableIterator->value();
    if(!particle) {
      std::ostream& out = fSession.cout();
      out << "G4Lab::ParticleAccessor::buildDefs :"
          << " Can't get particle table particle." 
          << std::endl;
      fDefs.clear();
      return false;
    }
    fDefs.push_back(particle);
  }

  return true;
}
