// this :
#include <G4Lab/UIOnX.h>

// Geant4 :
#include <G4UImanager.hh>

// Slash :
#include <Slash/Core/ISession.h>
#include <Slash/UI/IUI.h>
#include <Slash/Tools/UI.h>
#include <Slash/Tools/system.h>
#include <Slash/Tools/args.h>

// G4Lab :
#include <G4Lab/Manager.h>

#include <iostream>

G4Lab::UIOnX::UIOnX(
 G4RunManager& aRunManager
,const std::string& aGUI
,int aArgc
,char** aArgv
)
:fSession(0)
,fManager(0)
,fGUI(aGUI)
{
  //FIXME : deprecated : G4LABROOT in fact not needed.

  // setup env G4LABROOT which is needed for examples 
  //to find some custom .onx files :
  if(!Slash::isenv("G4LABROOT")) {
    std::string OSC_HOME_DIR;
    if(!Slash::getenv("OSC_HOME_DIR",OSC_HOME_DIR)) {
      std::cout << "G4Lab::UIOnX::UIOnX :"
                << " can't guess where OpenScientist had been installed."
                << std::endl;
    } else {
      std::string G4LABROOT = OSC_HOME_DIR+"/Resources/G4Lab";
      if(!Slash::putenv("G4LABROOT",G4LABROOT)) {
        std::cout << "G4Lab::UIOnX::UIOnX :"
                  << " can't putenv G4LABROOT."
                  << std::endl;
      }
      // Setup ONXROOT, used to fetch some common .onx files :
      std::string ONXROOT = OSC_HOME_DIR+"/Resources/OnX";
      if(!Slash::putenv("ONXROOT",ONXROOT)) {
        std::cout << "G4Lab::UIOnX::UIOnX :"
                  << " can't putenv ONXROOT."
                  << std::endl;
      }
      // Setup ONXLABROOT, used around plotting :
      std::string ONXLABROOT = OSC_HOME_DIR+"/Resources/OnXLab";
      if(!Slash::putenv("ONXLABROOT",ONXLABROOT)) {
        std::cout << "G4Lab::UIOnX::UIOnX :"
                  << " can't putenv ONXLABROOT."
                  << std::endl;
      }
    }
  }

  std::vector<std::string> args;
  for(int index=0;index<aArgc;index++) args.push_back(aArgv[index]);

  fSession = Slash_create_session(args); 

  // An IGeant4Manager to pass the G4RunManager to the scripting.
  fManager = new Manager(*fSession,"Geant4Manager",&aRunManager);
  // Declare our Manager to the OnX session.
  // It must be done prior the creation of the UI (in order that the 
  // initialize callbacks script execute properly).
  fSession->addManager(fManager);

  G4UImanager* uiManager = G4UImanager::GetUIpointer();
  if(uiManager) {
    uiManager->SetSession(this);
    //uiManager->SetCoutDestination(this);
  }
}

G4Lab::UIOnX::UIOnX(const UIOnX&){}
G4Lab::UIOnX& G4Lab::UIOnX::operator=(const UIOnX&){return *this;}

G4Lab::UIOnX::~UIOnX() { 
  G4UImanager* uiManager = G4UImanager::GetUIpointer();
  if(uiManager) {
    uiManager->SetSession(NULL);
    //uiManager->SetCoutDestination(NULL);
  }
  delete fSession;
}

bool G4Lab::UIOnX::isValid() const {return fSession ? true:false;}

Slash::UI::IUI* G4Lab::UIOnX::createUI(const std::string& aDriver,const std::string& aFile,bool aShow) {
  if(!fSession) return 0;
  Slash::UI::IUI* ui = Slash::find_UI(*fSession);
  if(ui) return ui; //done.
  Slash::UI::IUI_Manager* uim = Slash::uiManager(*fSession);
  if(!uim) return 0;
  Slash::args args(fSession->arguments());
  std::string driver = aDriver;
  if(driver.empty()) driver = gui_toolkit(args,true);
  return uim->create(driver,aFile.size()?aFile:fGUI,args.tovector(),aShow);
}

G4UIsession* G4Lab::UIOnX::SessionStart(){
  Slash::UI::IUI* ui = createUI("","",true);
  if(!ui) return 0;
  ui->steer();
  return this;
}

void G4Lab::UIOnX::PauseSessionStart(G4String aState){
  if(!fSession) return;

  //std::ostream& out = fSession->cout();
  //out << "G4UIOnX::PauseSessionStart : debug : " << aState << std::endl;

  if(aState=="G4_pause> ") { 
    //SecondaryLoop("Pause, type continue to exit this state");
  }

  if(aState=="EndOfEvent") {
    // Picking with feed back in event data Done here !!!
    //SecondaryLoop("End of event, type continue to exit this state");
  }
}

Slash::Core::ISession* G4Lab::UIOnX::session(){return fSession;}

Slash::Core::IWriter* G4Lab::UIOnX::printer(){
  if(!fSession) return 0;
  return &(fSession->printer());
}

void G4Lab::UIOnX::addManager(Slash::Core::IManager* aManager) {
  if(!fSession) return;
  fSession->addManager(aManager);
}

void G4Lab::UIOnX::addAccessor(Slash::Data::IAccessor* aAccessor) {
  if(!fManager) return;
  fManager->addAccessor(aAccessor);
}

AIDA::IAnalysisFactory* G4Lab::UIOnX::findAIDA() {
  if(!fManager) return 0;
  return fManager->findAIDA();
}
