
// AIDA :
#include <AIDA/AIDA.h>

// std :
#include <iostream>

//////////////////////////////////////////////////////////////////////////////
bool plot(
 AIDA::IAnalysisFactory& aAIDA
,AIDA::IHistogram1D& a_z
,AIDA::IHistogram1D& a_z_dE
,AIDA::IHistogram1D& a_r
,AIDA::IHistogram1D& a_r_dE
,AIDA::IHistogram2D& a_z_dE_r_dE
,const std::string& aFile
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string header = "G4LabTemplateFour_analysis : ";

  AIDA::IPlotterFactory* plotterFactory = 
    aAIDA.createPlotterFactory(0,0);
  if(!plotterFactory) {
    std::cout << header << "can't get a PlotterFactory." << std::endl;
    return false;
  }
    
  AIDA::IPlotter* plotter = plotterFactory->create();
  if(!plotter) {
    std::cout << header << "can't get a plotter." << std::endl;
    return false;
  }
    
  plotter->clearRegions();
  plotter->createRegions(3,2,0);

 {plotter->setCurrentRegionNumber(0);
  AIDA::IPlotterRegion& region = plotter->currentRegion();
  //set_region_style(region);
  region.style().xAxisStyle().setLabel("z(X0)");
  region.style().yAxisStyle().setLabel("Entries");
  region.plot(a_z);}

 {plotter->setCurrentRegionNumber(1);
  AIDA::IPlotterRegion& region = plotter->currentRegion();
  //set_region_style(region);
  region.style().xAxisStyle().setLabel("z(X0)");
  region.style().yAxisStyle().setLabel("E");
  region.plot(a_z_dE);}

 {plotter->setCurrentRegionNumber(3);
  AIDA::IPlotterRegion& region = plotter->currentRegion();
  //set_region_style(region);
  region.style().xAxisStyle().setLabel("r(X0)");
  region.style().yAxisStyle().setLabel("Entries");
  region.plot(a_r);}

 {plotter->setCurrentRegionNumber(4);
  AIDA::IPlotterRegion& region = plotter->currentRegion();
  //set_region_style(region);
  region.style().xAxisStyle().setLabel("r(X0)");
  region.style().yAxisStyle().setLabel("E");
  region.plot(a_r_dE);}

 {plotter->setCurrentRegionNumber(5);
  AIDA::IPlotterRegion& region = plotter->currentRegion();
  //set_region_style(region);
  region.style().xAxisStyle().setLabel("z_dE");
  region.style().yAxisStyle().setLabel("r_dE");
  region.plot(a_z_dE_r_dE);}

  plotter->show();

  /* FIXME : try to have multiple plotters in the tab stack :
 {AIDA::IPlotter* plotter = plotterFactory->create("Viewer_2");
  if(!plotter) {
    std::cout << header << "can't get a plotter." << std::endl;
    return false;
  }
  plotter->clearRegions();
  plotter->createRegions(1,1,0);}
  */

  plotter->interact();

  plotter->writeToFile(aFile+".ps");

  delete plotter;
  delete plotterFactory;

  return true;
}

//////////////////////////////////////////////////////////////////////////////
int main(int aArgc,char** aArgv) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string program = "G4LabTemplateFour_analysis";
  std::string header = program + " : ";

  AIDA::IAnalysisFactory* aida = AIDA_createAnalysisFactory();
  if(!aida) {
    std::cout << header << "AIDA not found." << std::endl;
    return 1;
  }

  AIDA::ITreeFactory* treeFactory = aida->createTreeFactory();
  if(!treeFactory) {
    std::cout << header << "can't get a TreeFactory." << std::endl;
    return 1;
  }

  ////////////////////////////////////////////////////////
  /// Read data //////////////////////////////////////////
  ////////////////////////////////////////////////////////

  std::vector<std::string> files;
  files.push_back("copper_el_E_10_GeV");
  files.push_back("aluminium_el_E_10_GeV");
  files.push_back("lead_el_E_10_GeV");

  files.push_back("lead_el_E_1_GeV");
  files.push_back("lead_el_E_10_GeV");
  files.push_back("lead_el_E_100_GeV");
  files.push_back("lead_el_E_1000_GeV");

  for(unsigned int i=0;i<files.size();i++) {

  std::string file = files[i];

  AIDA::ITree* tree = treeFactory->create(file+".root","root",true,false);
  if(!tree) {
    std::cout << header << "can't open data file." << std::endl;
    return 1;
  }

  AIDA::IHistogram1D* z = find_histogram1D(*tree,"z");
  if(!z) {
    std::cout << header 
              << "object z not found in tree." 
              << std::endl;
    return 1;
  }

  AIDA::IHistogram1D* z_dE = find_histogram1D(*tree,"z_dE");
  if(!z_dE) {
    std::cout << header 
              << "object z_dE not found in tree." 
              << std::endl;
    return 1;
  }

  AIDA::IHistogram1D* r = find_histogram1D(*tree,"r");
  if(!r) {
    std::cout << header 
              << "object r not found in tree." 
              << std::endl;
    return 1;
  }

  AIDA::IHistogram1D* r_dE = find_histogram1D(*tree,"r_dE");
  if(!r_dE) {
    std::cout << header 
              << "object r_dE not found in tree." 
              << std::endl;
    return 1;
  }

  AIDA::IHistogram2D* z_dE_r_dE = find_histogram2D(*tree,"z_dE_r_dE");
  if(!z_dE_r_dE) {
    std::cout << header 
              << "object z_dE_r_dE not found in tree." 
              << std::endl;
    return 1;
  }

  plot(*aida,*z,*z_dE,*r,*r_dE,*z_dE_r_dE,file);

  delete tree;
  }

  delete treeFactory;
  delete aida;

  return 0;
}
