/*
  OnX is started through the G4Lab::UIOnX (which inherits G4UISession).
 */

// Geant4 :
#include <G4RunManager.hh>
#include <G4UserSteppingAction.hh>
#include <QGSP_EMV.hh>

// G4Lab :
#include <G4Lab/UIOnX.h>
#include <G4Lab/TrackingAction.h>

// G4LabTemplateFour :
#include <G4LabTemplateFour/Detector.h>
#include <G4LabTemplateFour/Analysis.h>
#include <G4LabTemplateFour/Primary.h>
//#include <G4LabTemplateFour/PhysicsList.h>
#include <G4LabTemplateFour/RunAction.h>
#include <G4LabTemplateFour/EventAction.h>
#include <G4LabTemplateFour/SteppingAction.h>

#include <Slash/Tools/args.h>
#include <Slash/Tools/osc_env.h>

int main(int argc,char** argv) {

  Slash::args args(argc,argv);

  unsigned int dump_rate = 100;

  // E in GeV :
  double E;
  if(!args.find("-E",E)) E = 1;

  std::string smat;
  args.find("-mat",smat);

  //name,z,a,density.
  G4Material* material = 0;
  double RhoMoliere = 0;
  if(smat=="alu") {
    material = new G4Material("aluminium",13.,26.98*g/mole,2.7*g/cm3);
    RhoMoliere = 44*mm;
  } else if(smat=="copper") {
    material = new G4Material("copper",29.,63.55*g/mole,8.96*g/cm3);
    RhoMoliere = 15.2*mm;
    //X0 = 1.43568 cm
  } else if(smat=="lead") {
    material = new G4Material("lead",82.,207.19*g/mole,11.35*g/cm3);
    RhoMoliere = 16*mm;
    //X0 = 0.56118 cm
  } else {
    std::cout << "G4LabTemplateFour_vis :" 
              << std::endl
              << " unknown material \"" << smat << "\""
              << std::endl
              << " give one with :" 
              << std::endl
              << "   OS> G4LabTemplateFour_vis -mat=[alu,copper,lead]"
              << std::endl
              << " for example :" 
              << std::endl
              << "   UNIX> ./G4LabTemplateFour_vis -mat=alu"
              << std::endl;
    return 1;
  }

  E *= 1*GeV;

  G4RunManager* runManager = new G4RunManager;

  G4LabTemplateFour::Detector* det = 
    new G4LabTemplateFour::Detector(*material);

  runManager->SetUserInitialization(det);
  //runManager->SetUserInitialization(new G4LabTemplateFour::PhysicsList); 

  G4VUserPhysicsList* pl = new QGSP_EMV();
  double cut = 30*um;
  pl->SetCutValue(cut,"gamma");
  pl->SetCutValue(cut,"e-");
  pl->SetCutValue(cut,"e+");
  //pl->DumpCutValuesTable();

  runManager->SetUserInitialization(pl); // ext QGSP_EMV (EM_7.0)

  runManager->SetUserAction
    (new G4LabTemplateFour::PrimaryGeneratorAction(-1*cm,E));
  runManager->SetUserAction(new G4LabTemplateFour::EventAction(dump_rate));
  runManager->SetUserAction(new G4LabTemplateFour::RunAction());

  // The G4Lab::TrackingAction permits to use the G4Lab::Trajectory
  // that permits to have G4Track infos on picking.
  runManager->SetUserAction(new G4Lab::TrackingAction);

  // Need to pass the G4RunManager at UIOnX creation
  // so that the G4Lab::Manager get it properly.
  std::string gui = "../scripts/OnX/G4LabTemplateFour.onx";
  G4Lab::UIOnX* session = new G4Lab::UIOnX(*runManager,gui,1,argv);

  AIDA::IAnalysisFactory* aida = session->findAIDA();

  G4LabTemplateFour::Analysis* analysis = 
    new G4LabTemplateFour::Analysis(*aida,*det);

  runManager->SetUserAction
    (new G4LabTemplateFour::SteppingAction(*analysis,*det,RhoMoliere));

  runManager->Initialize();

  // Main UI loop :
  // GUI is constructed here.
  // Create callbacks are executed here.
  session->SessionStart();

  delete runManager;
  //the upper deletes :
  // Detector, PhysicsList, PrimaryGeneratorAction, EventAction, RunAction

  delete analysis;

  delete session;

  return 0;
}

