
// this :
#include <G4LabTemplateFour/Analysis.h>

// AIDA :
#include <AIDA/IAnalysisFactory.h>
#include <AIDA/ITreeFactory.h>
#include <AIDA/ITree.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
#include <AIDA/IProfile1D.h>
#include <AIDA/ITupleFactory.h>
#include <AIDA/ITuple.h>

#include <G4LabTemplateFour/Detector.h>

G4LabTemplateFour::Analysis::Analysis(
 AIDA::IAnalysisFactory& aAIDA
,Detector& aDetector
,const std::string& aFileName
)
:fAIDA(aAIDA)
,f_run(0)
,fTree(0)
,f_z(0)
,f_z_dE(0)
,f_r(0)
,f_r_dE(0)
,f_z_dE_r_dE(0)
,f_tuple(0)
,f_dE(0)
{
  AIDA::ITreeFactory* tf = fAIDA.createTreeFactory();
  if(aFileName.size()) {
    std::string opts = "compress=true";
    //std::string opts = "compress=true";
    fTree = tf->create(aFileName+".root","root",false,true,opts);
  } else {
    fTree = tf->create();
  }
  delete tf;
  if(!fTree) return;
  AIDA::IHistogramFactory* hf = fAIDA.createHistogramFactory(*fTree);
  if(!hf) return;

  //double radlen = aDetector.radiation_length()/cm;
  double mx = 40; //in radlen units
  int nbin = 80;

  f_z = hf->createHistogram1D("z",aFileName+" z",nbin,0,mx);
  f_z_dE = hf->createHistogram1D("z_dE",aFileName+" z_dE",nbin,0,mx);

  double mxr = 8;
  f_r = hf->createHistogram1D("r",aFileName+" r",nbin,0,mxr);
  f_r_dE = hf->createHistogram1D("r_dE",aFileName+" r_dE",nbin,0,mxr);

  f_z_dE_r_dE = 
    hf->createHistogram2D("z_dE_r_dE",aFileName+" r_dE/z_dE",
      nbin,0,mx,nbin,0,mxr);

  delete hf;

  AIDA::ITupleFactory* tpf = fAIDA.createTupleFactory(*fTree);
  std::string columns = "string particle,double E,int numberOfEvents"; 
  f_run  = tpf->create("run","run",columns,"");
  delete tpf;

  //AIDA::ITupleFactory* tpf = fAIDA.createTupleFactory(*fTree);
  //std::string columns = "double z,double E"; 
  //f_tuple  = tpf->create("tuple","tuple",columns,"");
  //delete tpf;
}

G4LabTemplateFour::Analysis::~Analysis(){
  std::cout << "dE = " << f_dE/GeV << " GeV." << std::endl;
  if(fTree) {
    fTree->commit();  
    fTree->ls();
    delete fTree;
  }
}
