//
//  All functions here should be OnX callbacks, that is to say
// functions with signature :
//   extern "C" {
//     void callback_without_arguments(Slash::UI::IUI&);
//     void callback_with_arguments(Slash::UI::IUI&,
//                                  const std::vector<std::string>&);
//   }
//

// Slash :
#include <Slash/Core/ISession.h>
#include <Slash/UI/IUI.h>
#include <Slash/UI/ISoViewer.h>

#include <Slash/Tools/Manager.h>
#include <Slash/Tools/Data.h>
#include <Slash/Tools/UI.h>
#include <Slash/Tools/AIDA.h>

#include <amanip/find.h>

// G4Lab :
#include <G4Lab/Interfaces/IGeant4Manager.h>

// AIDA :
#include <AIDA/IPlotterFactory.h>
#include <AIDA/IPlotter.h>
#include <AIDA/IPlotterRegion.h>

extern "C" {

//////////////////////////////////////////////////////////////////////////////
/// G4LabTemplateFour/scripts/OnX/G4LabTemplateFour.onx
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void G4LabTemplateFour_Initialize(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
// Should be executed (in a widget create callback) 
// when the GUI is constructed 
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aUI.setCallback("OnX_File_exit","activate","DLD","G4Lab G4Lab_exit");

  Slash::Core::ISession& session = aUI.session();

  session.setParameter("modeling.highlightColor","white");
  session.setParameter("modeling.trajectories","immediate_all");
  session.setParameter("event.scan","yes");
  session.setParameter("g4Lab.nextEventAction","Trajectories_all_immediate");
    
  IGeant4Manager* g4Manager = 
    Slash_findManager(session,"Geant4Manager",IGeant4Manager);
  if(!g4Manager) {
    aUI.echo("G4LabTemplateFour_Initialize : Geant4Manager not found.");
  } else {
    // Declare a callback executed at each end of event :
    g4Manager->setEventEndScript("DLD","G4LabTemplateFour G4LabTemplateFour_eventEnd");
    //Declare G4Lab types (PV, Trajectory, HitsCollections, etc...).
    g4Manager->initialize(); 
  }    
  
  // Create some extra histos used in the callbacks :
  //AIDA::IAnalysisFactory* aida = Slash::find_AIDA(session);

  aUI.executeScript("G4","/tracking/storeTrajectory 1");
}
//////////////////////////////////////////////////////////////////////////////
void G4LabTemplateFour_Finalize(
 Slash::Core::ISession&
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void G4LabTemplateFour_InputNextEventAction_ok(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string action;
  if(!aUI.parameterValue("G4LabTemplateFour_InputNextEventAction_input_action.value",action)) return;
  aUI.session().setParameter("g4Lab.nextEventAction",action);
}
//////////////////////////////////////////////////////////////////////////////
void G4LabTemplateFour_execute_eventEnd(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Core::ISession& session = aUI.session();

  std::string value;
  if(!session.parameterValue("event.scan",value)) return;
  // To avoid falling in the ui.steer();
  session.setParameter("event.scan","no");

  aUI.executeScript("DLD","G4LabTemplateFour G4LabTemplateFour_eventEnd");

  // Restore previous event.scan :
  session.setParameter("event.scan",value);
}
//////////////////////////////////////////////////////////////////////////////
/// scripts/OnX/Geant.onx ////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void G4LabTemplateFour_Geant4_detector(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Core::ISession& session = aUI.session();
  
  Slash::Data::collect(session,"PV","name==\"chunk_PV\"");
  Slash::Data::visualize(session,Slash::s_current());

  Slash::UI::ISoViewer* soViewer = 
    Slash::find_SoViewer(aUI,Slash::s_current());
  if(soViewer) soViewer->viewAll();
}

//////////////////////////////////////////////////////////////////////////////
void G4LabTemplateFour_plot_histos(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Core::ISession& session = aUI.session();

  // Find AIDA :
  AIDA::IAnalysisFactory* aida = Slash::find_AIDA(session);
  if(!aida) return;

  // Find tree :
  std::string stree = "G4LabTemplateFour.root";
  AIDA::ITree* tree = Slash_findManager(session,stree,AIDA::ITree);
  if(!tree) {
    std::ostream& out = session.cout();
    out << "AIDA::ITree " << Slash::sout(stree) << " not found." << std::endl;
  }

  AIDA::IPlotterFactory* plotterFactory = aida->createPlotterFactory();
  if(!plotterFactory) return;
  AIDA::IPlotter* plotter = plotterFactory->create("");
  if(!plotter) return;

  plotter->createRegions(1,2,0);

  {
    AIDA::IHistogram1D* h_z = amanip::find_histogram_1D(*tree,"z");
    if(h_z) {
      AIDA::IPlotterRegion& region = plotter->currentRegion();
      region.plot(*h_z);
    }
  }

  {
    AIDA::IProfile1D* h_z_dE = amanip::find_profile_1D(*tree,"z_dE");
    if(h_z_dE) {
      AIDA::IPlotterRegion& region = plotter->next();
      region.plot(*h_z_dE);
    }
  }

  delete plotter;
  delete plotterFactory;
}

//////////////////////////////////////////////////////////////////////////////
void G4LabTemplateFour_reset_histos(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Core::ISession& session = aUI.session();

  // Find AIDA :
  AIDA::IAnalysisFactory* aida = Slash::find_AIDA(session);
  if(!aida) return;

  // Find tree :
  std::string stree = "G4LabTemplateFour.root";
  AIDA::ITree* tree = Slash_findManager(session,stree,AIDA::ITree);
  if(!tree) {
    std::ostream& out = session.cout();
    out << "AIDA::ITree " << Slash::sout(stree) << " not found." << std::endl;
  }

  {
    AIDA::IHistogram1D* h_z = amanip::find_histogram_1D(*tree,"z");
    if(h_z) {
      h_z->reset();
    }
  }

  {
    AIDA::IProfile1D* h_z_dE = amanip::find_profile_1D(*tree,"z_dE");
    if(h_z_dE) {
      h_z_dE->reset();
    }
  }

}

} // extern "C"
