@ECHO OFF
SETLOCAL

REM # Check that osc-config is available :
osc-config 1> NUL 2> NUL
IF ERRORLEVEL 1 (
  ECHO osc-config program not found.
  GOTO build_return
)

REM # Check that aida-config is available :
aida-config 1> NUL 2> NUL
IF ERRORLEVEL 1 (
  ECHO aida-config program not found.
  GOTO build_return
)

SET bin=.

REM # /GR to enable RTTI and dynamic cast.
REM # /GX to enable exception handling (for STL).
REM # /O2 optimize
REM # /W3 warmings
SET def_cppflags=/nologo /DWIN32 /MD /GX /GR /O2 /W3 /I. /I..

FOR /F "usebackq delims=;" %%s IN (`osc-config --cppflags`) DO SET onx_cppflags=%%s
FOR /F "usebackq delims=;" %%s IN (`osc-config --libs`) DO SET onx_libs=%%s
FOR /F "usebackq delims=;" %%s IN (`osc-config --iv_cppflags`) DO SET onx_iv_cppflags=%%s
FOR /F "usebackq delims=;" %%s IN (`osc-config --iv_libs`) DO SET onx_iv_libs=%%s
FOR /F "usebackq delims=;" %%s IN (`osc-config --core_libs`) DO SET onx_core_libs=%%s
FOR /F "usebackq delims=;" %%s IN (`aida-config --cppflags`) DO SET aida_cppflags=%%s
FOR /F "usebackq delims=;" %%s IN (`osc-config --g4lab_cppflags`) DO SET onx_g4lab_cppflags=%%s
FOR /F "usebackq delims=;" %%s IN (`osc-config --g4lab_libs`) DO SET onx_g4lab_libs=%%s

REM #//////////////////////////////////////////////////
REM #// library with Geant4 user code :
REM #//////////////////////////////////////////////////
SET cppflags=%def_cppflags%
SET cppflags=%cppflags% %onx_g4lab_cppflags%
SET cppflags=%cppflags% %aida_cppflags%
cl.exe /c %cppflags% /Fo%bin%\CellHit.obj /Tp..\source\CellHit.cxx
cl.exe /c %cppflags% /Fo%bin%\CellSD.obj /Tp..\source\CellSD.cxx
cl.exe /c %cppflags% /Fo%bin%\Detector.obj /Tp..\source\Detector.cxx
cl.exe /c %cppflags% /Fo%bin%\PhysicsList.obj /Tp..\source\PhysicsList.cxx
cl.exe /c %cppflags% /Fo%bin%\Primary.obj /Tp..\source\Primary.cxx
cl.exe /c %cppflags% /Fo%bin%\EventAction.obj /Tp..\source\EventAction.cxx
cl.exe /c %cppflags% /Fo%bin%\RunAction.obj /Tp..\source\RunAction.cxx
cl.exe /c %cppflags% /Fo%bin%\SteppingAction.obj /Tp..\source\SteppingAction.cxx

SET G4_objs=%bin%\CellHit.obj %bin%\CellSD.obj %bin%\Detector.obj %bin%\PhysicsList.obj %bin%\Primary.obj %bin%\EventAction.obj  %bin%\RunAction.obj %bin%\SteppingAction.obj

REM #//////////////////////////////////////////////////
REM #// dll G4LabTemplateOne :
REM #//////////////////////////////////////////////////
SET cppflags=%def_cppflags%
SET cppflags=%cppflags% %onx_cppflags%
SET cppflags=%cppflags% %onx_iv_cppflags%
SET cppflags=%cppflags% %aida_cppflags%
SET cppflags=%cppflags% %onx_g4lab_cppflags%
cl.exe /c %cppflags% /Fo%bin%\Callbacks.obj /Tp..\source\Callbacks.cxx

REM # Build the archive lib :
SET objs=%bin%\Callbacks.obj

lib.exe /nologo /machine:ix86 /out:%bin%\G4LabTemplateOne.arc %objs%

DEL %objs%

REM # Build the win32def.exe utility program :
cl.exe %def_cppflags% /Fe%bin%\win32def.exe /Tpwin32def.c

REM # Building the .def file :
dumpbin /out:%bin%\tmp /symbols %bin%\G4LabTemplateOne.arc
%bin%\win32def.exe G4LabTemplateOne < %bin%\tmp > %bin%\G4LabTemplateOne.def

REM # Build the DLL :
lib.exe /nologo /machine:ix86 /def:%bin%\G4LabTemplateOne.def /out:%bin%\G4LabTemplateOne.lib
IF EXIST %bin%\G4LabTemplateOne.dll DEL %bin%\G4LabTemplateOne.dll
SET libs=
SET libs=%libs% %onx_iv_libs%
SET libs=%libs% %onx_libs%
link.exe /nologo /dll /out:%bin%\G4LabTemplateOne.dll %bin%\G4LabTemplateOne.exp %bin%\G4LabTemplateOne.arc %libs%

REM #//////////////////////////////////////////////////
REM #// application G4LabTemplateOne_vis :
REM #//////////////////////////////////////////////////

REM # application G4LabTemplateOne_vis :
SET cppflags=%def_cppflags%
SET cppflags=%cppflags% %onx_cppflags%
SET cppflags=%cppflags% %onx_iv_cppflags%
SET cppflags=%cppflags% %aida_cppflags%
SET cppflags=%cppflags% %onx_g4lab_cppflags%
cl.exe /c %cppflags% /Fo%bin%\G4LabTemplateOne_vis.obj /Tp..\applications\G4LabTemplateOne_vis.cxx

SET objs=%bin%\G4LabTemplateOne_vis.obj %G4_objs%

IF EXIST %bin%\G4LabTemplateOne_vis.exe DEL %bin%\G4LabTemplateOne_vis.exe
SET libs=%onx_iv_libs%
SET libs=%libs% %onx_core_libs%
SET libs=%libs% %onx_libs%
SET libs=%libs% %onx_g4lab_libs%
link.exe /nologo /out:%bin%\G4LabTemplateOne_vis.exe %objs% %libs%

DEL %bin%\tmp
DEL %objs%

DEL %bin%\win32def.obj
DEL %bin%\G4LabTemplateOne.arc
DEL %bin%\G4LabTemplateOne.def
DEL %bin%\G4LabTemplateOne.exp

:build_return

ENDLOCAL
@ECHO ON
