// this :
#include <G4LabTemplateThree/Detector.h>

// GDML :
#include <Saxana/SAXProcessor.h>
#include <Saxana/ProcessingConfigurator.h>
#include <G4Processor/GDMLProcessor.h>

//////////////////////////////////////////////////////////////////////////////
G4LabTemplateThree::Detector::Detector(
 const std::string& aFile
)
:fSXP(0)
,fConfig(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fSXP = new SAXProcessor();
  fSXP->Initialize();

  fConfig = new ProcessingConfigurator();
  fConfig->SetURI(aFile);
  fConfig->SetSetupName("Default");
  fSXP->Configure(fConfig);
}

//////////////////////////////////////////////////////////////////////////////
G4LabTemplateThree::Detector::~Detector(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fSXP->Finalize();
  delete fSXP;
  delete fConfig;
}

//////////////////////////////////////////////////////////////////////////////
G4VPhysicalVolume* G4LabTemplateThree::Detector::Construct(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  if(!fSXP) {
    G4cout << "G4LabTemplateThree::Detector::Construct :"
           << " no SAXProcessor."
           << " Return a dummy world."
           << G4endl;
    return dummyWorld();
  }
  if(!fSXP->Run().IsOk()) {
    G4cout << "G4LabTemplateThree::Detector::Construct :"
           << " SAXProcessor problem."
           << " Return a dummy world."
           << G4endl;
    return dummyWorld();
  }
  G4VPhysicalVolume* world = 
    (G4VPhysicalVolume*)GDMLProcessor::GetInstance()->GetWorldVolume();
  if(!world) {
    G4cout << "G4LabTemplateThree::Detector::Construct :"
           << " can't get a world volume from GDML."
           << " Return a dummy world."
           << G4endl;
    return dummyWorld();
  }
  return world;
}
#include <G4Material.hh>
#include <G4Box.hh>
#include <G4LogicalVolume.hh>
#include <G4PVPlacement.hh>
//////////////////////////////////////////////////////////////////////////////
G4VPhysicalVolume* G4LabTemplateThree::Detector::dummyWorld(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  G4Material* vacuum = 
    new G4Material("Galactic Vacuum",
                   1.,1.01*g/mole,universe_mean_density, 
                   kStateGas,0.1*kelvin,1.e-19*pascal);
  G4LogicalVolume* worldLV = 
    new G4LogicalVolume(new G4Box("WorldBox",1.* m,1.* m,1.* m),
                        vacuum,"WorldLV");
  return  new G4PVPlacement(G4Transform3D(),"WorldPV",worldLV,0,false,0);
}
