#ifndef G4LabTemplateTwo_DetectorHit_h
#define G4LabTemplateTwo_DetectorHit_h 1

//This code is inspired from G4 ExN04

//Geant4
#include "G4VHit.hh"
#include "G4THitsCollection.hh"
#include "G4Allocator.hh"
#include "G4ThreeVector.hh"
#include "G4Transform3D.hh"

class G4LogicalVolume;

namespace G4LabTemplateTwo {
  
  class DetectorHit : public G4VHit {
  public:
    
    DetectorHit();
    ~DetectorHit();
    DetectorHit(const DetectorHit &right);
    const DetectorHit& operator=(const DetectorHit &right);
    G4int operator==(const DetectorHit &right) const;
    
    inline void *operator new(size_t);
    inline void operator delete(void *aHit);
    
    void Draw();
    void Print();
    
    virtual const std::map<G4String,G4AttDef>* GetAttDefs() const;
    virtual std::vector<G4AttValue>* CreateAttValues() const;
    
  public:
    //Set & Get
    void SetTrackPdgID (G4int trackPdgID) { m_trackPdgID = trackPdgID; }
    G4int GetTrackPdgID () { return m_trackPdgID; }

    void SetTrackID  (G4int trackID)  { m_trackID = trackID; }
    G4int GetTrackID()              { return m_trackID; }
    
    void SetEdep(G4double edep) { m_edep = edep; }
    G4double GetEdep()        { return m_edep; }
    
    void SetPos(G4ThreeVector pos) { m_pos = pos; }
    G4ThreeVector GetPos() { return m_pos; }

    void SetMomentum (G4ThreeVector momentum) { m_momentum = momentum; }
    G4ThreeVector GetMomentum() { return m_momentum; }
    
    void SetTotalE (G4double totalE) { m_totalE = totalE; }
    G4double GetTotalE() { return m_totalE; } 
    
    void SetLV (G4LogicalVolume* aLV) { m_LV = aLV; }
    void SetTSF (const G4Transform3D& aTSF) { m_TSF = aTSF; }
  private:
    G4int         m_trackID;       // G4 track IDentifier
    G4int         m_trackPdgID;    // PDG code of the Particle
    G4double      m_edep;          // Energy deposit in the step
    G4ThreeVector m_pos;           // position before the step
    G4ThreeVector m_momentum;      // particle 3-momentum
    G4double      m_totalE;        // particle total energy
    G4LogicalVolume* m_LV;
    G4Transform3D    m_TSF;
  };
  
}


namespace G4LabTemplateTwo {
  typedef G4THitsCollection<DetectorHit> DetectorHitsCollection;
  extern G4Allocator<DetectorHit> DetectorHitAllocator;
}




inline void* G4LabTemplateTwo::DetectorHit::operator new(size_t)
{
  void *aHit;
  aHit = (void *) G4LabTemplateTwo::DetectorHitAllocator.MallocSingle();
  return aHit;
}

inline void G4LabTemplateTwo::DetectorHit::operator delete(void *aHit)
{
  G4LabTemplateTwo::DetectorHitAllocator.FreeSingle((G4LabTemplateTwo::DetectorHit*) aHit);
}


#endif
