#ifndef G4LabTemplateTwo_EventAction_h
#define G4LabTemplateTwo_EventAction_h

#include <fstream>

//Geant4
#include <G4UserEventAction.hh>
#include <G4Types.hh>
#include <G4ThreeVector.hh>

class G4SDManager;

namespace AIDA {
  class IHistogram1D;
  class IHistogram2D;
  class ITuple;
}

namespace G4LabTemplateTwo {

class Analysis;

class EventAction : public G4UserEventAction {
  public:
    EventAction(Analysis&);
    virtual ~EventAction();
    void BeginOfEventAction(const G4Event*);
    void EndOfEventAction(const G4Event*);

  private:
    //trigger and solid angle computation function
    G4double isAccecpted(const G4ThreeVector& dir, 
			 const G4ThreeVector& pos);
      
    //Data member
    G4int m_verboseLevel; //dump
    G4double  m_triggerToTargetDistance; // distance of the trigger station
    G4double  m_triggerRadius2; //radius squarred of the trigger station
    
    Analysis& fAnalysis; //AIDA Manager

    std::ofstream m_output; //Ascii output for Data

    G4int m_detectorCollID; //Identifier of the Detector Hit Collection
    G4SDManager* m_SDman; //pointer to the Sensitive Detector Manager

    //User Histo
    AIDA::IHistogram2D* m_histoXY;
    AIDA::IHistogram1D* m_histoTotalE;
    AIDA::IHistogram2D* m_histoPiPlus; 
    AIDA::IHistogram2D* m_histoPiMinus;
    AIDA::IHistogram2D* m_histoKPlus;  
    AIDA::IHistogram2D* m_histoKMinus; 
    AIDA::IHistogram2D* m_histoProton; 
    AIDA::IHistogram2D* m_histoOther;  
    AIDA::ITuple*       m_tuple;
};

}

#endif
