/*
 * Batch program over Geant4 and AIDA.
 */

// Geant4 :
#include <G4RunManager.hh>
#include <G4UImanager.hh>
#include <QGSP.hh> //JEC 27/2/05
// G4LabTemplateTwo :
#include <G4LabTemplateTwo/Detector.h>
#include <G4LabTemplateTwo/Primary.h>

//#include <G4LabTemplateTwo/PhysicsList.h> JEC 27/2/05
#include <G4LabTemplateTwo/StepLimiterBuilder.h>


#include <G4LabTemplateTwo/RunAction.h>
#include <G4LabTemplateTwo/EventAction.h>
#include <G4LabTemplateTwo/SteppingAction.h>
#include <G4LabTemplateTwo/Analysis.h>

// AIDA :
#include <AIDA/IAnalysisFactory.h>

// std :
#include <iostream>

// Lib :
#include <Lib/System.h>

//////////////////////////////////////////////////////////////////////////////
int main(
 int //aArgc
,char** //aArgv
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  AIDA::IAnalysisFactory* aida = AIDA_createAnalysisFactory();
  if(!aida) {
    std::cout << "G4LabTemplateTwo_batch :"
              << " AIDA not found." << std::endl;
    return 0;
  }

  G4LabTemplateTwo::Analysis* analysis = new G4LabTemplateTwo::Analysis(*aida);
  G4RunManager* runManager = new G4RunManager;

  runManager->SetUserInitialization(new G4LabTemplateTwo::Detector());

  //Use Hans-Peter Wellisch 27/2/05 
  QGSP* aQGSP = new QGSP();
  StepLimiterBuilder* theStepLimiter = new StepLimiterBuilder();
  aQGSP->RegisterPhysics(theStepLimiter);
  runManager->SetUserInitialization(aQGSP);
  

  //  runManager->SetUserInitialization(new G4LabTemplateTwo::PhysicsList);


  runManager->SetUserAction(new G4LabTemplateTwo::PrimaryGeneratorAction);
  runManager->SetUserAction(new G4LabTemplateTwo::RunAction());
  runManager->SetUserAction(new G4LabTemplateTwo::EventAction(*analysis));
  runManager->SetUserAction(new G4LabTemplateTwo::SteppingAction(*analysis));

  runManager->Initialize();

  G4UImanager* UI = G4UImanager::GetUIpointer();
  if(UI) {
    std::string file = 
      Lib::System::getenv("G4LABTEMPLATETWOROOT")+"/scripts/Geant4/init.g4m";
    UI->ApplyCommand("/control/execute "+file);
  }

  //runManager->BeamOn(1000000);
  //runManager->BeamOn(10);
  runManager->BeamOn(100);

  delete runManager;
  delete analysis;
  delete aida;

  return 0;
}

