/*
   Example of a program started "a la Geant4 examples way".
  OnX is attached through the G4Lab::UIOnX (which inherits G4USlash::Core::ISession).
 */

// Geant4 :
#include <G4RunManager.hh>
#include <G4UImanager.hh>
#include <QGSP.hh> //JEC 27/2/05

// AIDA :
#include <AIDA/IAnalysisFactory.h>

// Lib :
#include <Slash/Core/ISession.h>
#include <Lib/Manager.h>
#include <Lib/System.h>
#include <Lib/Cast.h>

// G4Lab :
#include <G4Lab/UIOnX.h>
#include <G4Lab/TrackingAction.h>
#include <G4Lab/Helpers.h>

// std :
#include <iostream>

// G4LabTemplateTwo :
#include <G4LabTemplateTwo/Detector.h>
#include <G4LabTemplateTwo/Primary.h>

//#include <G4LabTemplateTwo/PhysicsList.h> JEC 27/2/05
#include <G4LabTemplateTwo/StepLimiterBuilder.h>

#include <G4LabTemplateTwo/RunAction.h>
#include <G4LabTemplateTwo/EventAction.h>
#include <G4LabTemplateTwo/SteppingAction.h>
#include <G4LabTemplateTwo/Analysis.h>

// Create a manager to access the G4LabTemplateTwo::Analysis 
// from the callbacks :
namespace G4LabTemplateTwo {
class AnalysisManager 
:public virtual Slash::Core::IManager
,public Analysis  {
public: //Slash::Core::IManager
  virtual std::string name() const { return fName;}
  virtual void* cast(const std::string& aClass) const {
    if_Lib_SCast(G4LabTemplateTwo::AnalysisManager)
    else if_Lib_SCast(G4LabTemplateTwo::IAnalysis)
    else if_Lib_SCast(G4LabTemplateTwo::Analysis)
    else if_Lib_SCast(Slash::Core::IManager)
    else return 0;
  }
public:
  AnalysisManager(AIDA::IAnalysisFactory& aAIDA,bool aBatch)
  :Analysis(aAIDA,aBatch),fName("G4LabTemplateTwo::AnalysisManager") {}
  virtual ~AnalysisManager(){}
private:
  std::string fName;
};
}

//////////////////////////////////////////////////////////////////////////////
int main(
 int aArgc
,char** aArgv
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //Lib::Debug::checkByClass(true);

  G4RunManager* runManager = new G4RunManager;

  // Need to pass the G4RunManager at UIOnX creation
  std::string gui = 
    "$G4LABTEMPLATETWOROOT/scripts/OnX/G4LabTemplateTwo_session.onx";
  G4Lab::UIOnX* session = new G4Lab::UIOnX(*runManager,gui,aArgc,aArgv);

  G4LabTemplateTwo::AnalysisManager* analysisManager = 0;

  AIDA::IAnalysisFactory* aida = session->findAIDA();
  if(!aida) {
    std::cout << "G4LabTemplateTwo_session :"
              << " AnalysisFactory not found." 
              << std::endl;
  } else {
    analysisManager = new G4LabTemplateTwo::AnalysisManager(*aida,false);
    session->addManager(analysisManager);
  }

  runManager->SetUserInitialization(new G4LabTemplateTwo::Detector());
  //Use Hans-Peter Wellisch 27/2/05 
  QGSP* aQGSP = new QGSP();
  StepLimiterBuilder* theStepLimiter = new StepLimiterBuilder();
  aQGSP->RegisterPhysics(theStepLimiter);
  runManager->SetUserInitialization(aQGSP);
  //runManager->SetUserInitialization(new G4LabTemplateTwo::PhysicsList);

  runManager->SetUserAction(new G4LabTemplateTwo::PrimaryGeneratorAction);
  runManager->SetUserAction(new G4LabTemplateTwo::RunAction());
  runManager->SetUserAction(new G4LabTemplateTwo::EventAction(*analysisManager));
  runManager->SetUserAction(new G4Lab::TrackingAction);
  runManager->SetUserAction(new G4LabTemplateTwo::SteppingAction(*analysisManager));

  runManager->Initialize();
    
  G4UImanager* UI = G4UImanager::GetUIpointer();
  if(UI) {
    std::string file = 
      Lib::System::getenv("G4LABTEMPLATETWOROOT")+"/scripts/Geant4/init.g4m";
    UI->ApplyCommand("/control/execute "+file);
  }


  // Main UI loop :
  session->SessionStart();

  delete session; //It will delete the analysisManager.

  delete runManager;

  return 0;
}

