// this :
#include <G4LabTemplateTwo/Analysis.h>

#include <AIDA/IAnalysisFactory.h>
#include <AIDA/ITreeFactory.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/ITree.h>
#include <AIDA/ITupleFactory.h>

#include <iostream>
#include <vector>
#include <string>

G4LabTemplateTwo::Analysis::Analysis(AIDA::IAnalysisFactory& aAIDA,bool aBatch)
:fAIDA(aAIDA)
,fBatch(aBatch)
,fTree(0)
{
  AIDA::ITreeFactory* treeFactory = fAIDA.createTreeFactory();
  if(fBatch) {
    //ROOT tree :
    std::string opts = "export=root";
    fTree = treeFactory->create("G4LabTemplateTwo.root","root",false,true,opts);

    //HBOOK
    //std::string opts = "";
    //fTree = treeFactory->create("g4labtemplatetwo.hbook","hbook",false,true,opts);

    // AIDA XML tree :
    //std::string opts = "compress=yes";
    // fTree = treeFactory->create("G4LabTemplateTwo.xml","xml",false,true,opts);
  } else {
    // Memory tree :
    fTree = treeFactory->create();
  }
  delete treeFactory;
  if(!fTree) {
    std::cout << "G4LabTemplateTwo::Analysis : FATAL"
              << " tree not created." << std::endl;
    exit(0);
  }
  AIDA::IHistogramFactory* hf = fAIDA.createHistogramFactory(*fTree);

  //User Part: Simple Histogram Booking
  hf->createHistogram1D("histo_DE","StepDE",100,0,10);
  hf->createHistogram1D("histo_DL","StepLength",100,0,40);

  hf->createHistogram2D("histo_HitXY","Hit Y vs X (cm)",
				      100,-50,50,100,-50,50);
  hf->createHistogram1D("histo_HitTotalE","Hit total E",100,0,10);
  
  int MaxNumberOfBins = 8;
  std::vector<double> pBin(MaxNumberOfBins);
  double lowEdges[] = { 5.5, 8.5, 12.5, 17.5, 25.0, 35.0, 53.75, 81.25 }; //low edges GeV
  for (int i=0; i<MaxNumberOfBins; ++i) {pBin[i] = lowEdges[i];}
  hf->createHistogram2D("histoPiPlus"," PT vs Ptot  for pi+ (GeV)",
			10,0.,100.,10,0.,0.5);
  //    hf->createHistogram1D("histo_PiPlus"," Ptot for pi+ (GeV)",pBin);
  hf->createHistogram2D("histoPiMinus"," PT vs Ptot  for pi- (GeV)",
			10,0.,10.,10,0.,0.5);
  //   = hf->createHistogram1D("histo_PiMinus"," Ptot for pi- (GeV)",pBin);
  hf->createHistogram2D("histoKPlus"," PT vs Ptot  for K+ (GeV)",
			10,0.,10.,10,0.,0.5);
  //    hf->createHistogram1D("histo_KPlus"," Ptot for K+ (GeV)",pBin);
  hf->createHistogram2D("histoKMinus"," PT vs Ptot  for K- (GeV)",
			10,0.,10.,10,0.,0.5);
  //    = hf->createHistogram1D("histo_KMinus"," Ptot for K- (GeV)",pBin);
  hf->createHistogram2D("histoProton"," PT vs Ptot  for p (GeV)",
			10,0.,10.,10,0.,0.5);
  //    = hf->createHistogram1D("histo_Proton"," Ptot for p (GeV)",pBin);
  hf->createHistogram2D("histoOther"," PT vs Ptot  for other (GeV)",
			10,0.,10.,10,0.,0.5);
  //    = hf->createHistogram1D("histo_Other"," Ptot for Other (GeV)",pBin);

  //Booking Tuple
  AIDA::ITupleFactory* tf = fAIDA.createTupleFactory(*fTree);

  int NumberOfColumn = 7;
  std::vector<std::string> column(NumberOfColumn);
  const char* c_column[] = {"pdg", 
			    "x", "y", "z", 
			    "px", "py", "pz"};
  std::vector<std::string> coltype(NumberOfColumn);
  const char* c_coltype[] = {"int", 
			     "double","double","double",
			     "double","double","double"};
  for (int icol = 0; icol<NumberOfColumn; ++icol) {
    column[icol]  = c_column[icol];
    coltype[icol] = c_coltype[icol];
  }

  tf->create("tupleOFFTarget","Out of the Target",
	     column,coltype);

  //End of User Part
  delete tf;
  delete hf;
}//Ctor

//----------------------------------------------------------------------
G4LabTemplateTwo::Analysis::~Analysis(){
  //Not to be changed
  if(!fTree) return;
  if(fBatch) {
    fTree->commit();
    delete fTree;
    fTree = 0;
  }
}//Dtor

//----------------------------------------------------------------------

void G4LabTemplateTwo::Analysis::closeTree(){
  if(!fTree) return;
  if(!fBatch) return;
  fTree->commit();
  delete fTree;
  fTree = 0;
}
