//
//  All functions here should be OnX callbacks, that is to say
// functions with signature :
//   extern "C" {
//     void callback_without_arguments(Slash::UI::IUI&);
//     void callback_with_arguments(Slash::UI::IUI&,const std::vector<std::string>&);
//   }
//

#include <OnX/Helpers/OnX.h>
#include <OnX/Helpers/Inventor.h>

// Slash :
#include <Slash/Core/ISession.h>
#include <Slash/UI/IUI.h>

// Lib :
#include <Lib/Manager.h>
#include <Lib/Out.h>

// G4Lab :
#include <G4Lab/Interfaces/IGeant4Manager.h>

// G4LabTemplateTwo :
#include <G4LabTemplateTwo/IAnalysis.h>

extern "C" {

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void G4LabTemplateTwo_Initialize(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
// Should be executed (in a widget create callback) 
// when the GUI is constructed 
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::UI::ISoViewer* soViewer = ui_SoViewer(aUI);
  if(soViewer) {
    soViewer->setDecoration(true);
    soViewer->setViewing(true);
  }

  Slash::Core::ISession& session = aUI.session();

  IGeant4Manager* g4Manager = 
    Lib_findManager(session,"Geant4Manager",IGeant4Manager);
  if(!g4Manager) {
    aUI.echo("G4LabTemplateTwo_Initialize : Geant4Manager not found.");
    return;
  }
  g4Manager->initialize(); 

  //Used in G4LabTemplateTwo_eventEnd callback.
  session.setParameter("event.scan","yes"); 
    
  aUI.setCallback("OnX_File_exit","activate","DLD","G4Lab G4Lab_exit");

  // Declare a callback executed at each end of event :
  g4Manager->setEventEndScript
    ("DLD","G4LabTemplateTwo G4LabTemplateTwo_eventEnd");
      
  aUI.executeScript("G4","/tracking/storeTrajectory 1");
}
//////////////////////////////////////////////////////////////////////////////
void G4LabTemplateTwo_Finalize(
 Slash::UI::IUI&
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
void G4LabTemplateTwo_Geant4_detector(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Core::ISession& session = aUI.session();
  
  data_collect(session,"PV","name==\"TargetPhys\"");
  data_visualize(session);

  data_collect(session,"PV","name==\"DetectorPhys\"");
  data_visualize(session);
}
//////////////////////////////////////////////////////////////////////////////
void G4LabTemplateTwo_analysis_closeTree(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Core::ISession& session = aUI.session();
  
  G4LabTemplateTwo::IAnalysis* analysisManager = 
    Lib_findManager(session,"G4LabTemplateTwo::AnalysisManager",G4LabTemplateTwo::IAnalysis);
  if(!analysisManager) {
    aUI.echo("G4LabTemplateTwo_analysis_closeTree : G4LabTemplateTwo::AnalysisManager not found.");
    return;
  }

  analysisManager->closeTree();
}


} // extern "C"
