
// this :
#include "G4LabTemplateTwo/Detector.h"

// Geant4 :
#include <G4Element.hh>
#include <G4Material.hh>
#include <G4Box.hh>
#include <G4Tubs.hh>
#include <G4LogicalVolume.hh>
#include <G4PVPlacement.hh>
#include <G4Transform3D.hh>
#include <G4ThreeVector.hh>
#include <G4SDManager.hh>
#include <G4UserLimits.hh>
#include <G4ios.hh>
#include <G4Types.hh>

//G4LabTemplateTwo
#include "G4LabTemplateTwo/SDetector.h"


G4LabTemplateTwo::Detector::Detector() {}

G4LabTemplateTwo::Detector::~Detector() {}

//----------------------------------------------------------------------------------------
G4VPhysicalVolume* G4LabTemplateTwo::Detector::Construct() {        

  //------------
  // Materials 
  //------------
  //Vacuum defined in G4 User Guide::Material source listing 4.2.1
  //universe_mean_density is a static cast from CLHEP...
  G4Material* Vacuum = new 
    G4Material("Galactic Vacuum",
	       1.,
	       1.01*g/mole,
	       universe_mean_density, 
	       kStateGas,
	       0.1*kelvin,
	       1.e-19*pascal);
  //Liquid Mercury: default STP temperature and pressure
  G4Material* LiqHg = new 
    G4Material("LiquidMercury",
	       80.,
	       200.59*g/mole,
	       13.546*g/cm3,
	       kStateLiquid);
  //Get the Radiation Length and Nuclear Interaction Length
  G4cout << "Liq. Hg X0  : " << LiqHg->GetRadlen() / cm << "(cm)" << G4endl;
  G4cout << "Liq. Hg Lhad: " << LiqHg->GetNuclearInterLength() / cm << "(cm)" 
	<< G4endl;
  //Solid Beryllium
  G4Material* SolBe = new 
    G4Material("Beryllium",
	       4.,
	       9.01*g/mole,
	       1.848*g/cm3,
	       kStateSolid);
  //Get the Radiation Length and Nuclear Interaction Length
  G4cout << "Solid Be X0  : " << SolBe->GetRadlen() / cm << "(cm)" << G4endl;
  G4cout << "Solid Be Lhad: " << SolBe->GetNuclearInterLength() / cm << "(cm)" 
	 << G4endl;

  //------------
  // Volumes
  //------------
  //Target Parameters
  G4double targetHalfLength = 50.* mm;
  G4double targetHalfHeight =  1.* mm;
  G4double targetHalfWide   = 80.* mm;
				
  //Detector Parameters 10% larger than the Target
  G4double detectorHalfLength = targetHalfLength  * 1.10;
  G4double detectorHalfHeight = targetHalfHeight  * 1.10;
  G4double detectorHalfWide   = targetHalfWide    * 1.10;

  //World Parameters
  G4double worldHalfLength =  1.* m;
  G4double worldHalfHeight =  1.* m;
  G4double worldHalfWide   =  1.* m;
  
  //World (big box) : Logical & Physical Volumes
  G4LogicalVolume* worldLV = 
    new G4LogicalVolume(new G4Box(
				  "WorldBox",
				  worldHalfWide,
				  worldHalfHeight,
				  worldHalfLength),
                        Vacuum,
			"WorldLog");
  G4PVPlacement* worldPV = 
    new G4PVPlacement(G4Transform3D(),"WorldPhys",worldLV,0,false,0);

  //The detector volume
  G4LogicalVolume* detectorLV = 
    new G4LogicalVolume(new G4Box(
				  "Detector",
				  detectorHalfWide,
				  detectorHalfHeight,
				  detectorHalfLength
				  ),
			Vacuum,
			"DetectorLog");

  G4Transform3D detectorPosition; // default: Identity matrix and 0,0,0 translation.
  detectorPosition = G4TranslateZ3D(targetHalfLength); // translation along Z govern by the Target Length
  G4PVPlacement* detectorPV = 
    new G4PVPlacement(detectorPosition,"DetectorPhys",
		      detectorLV,worldPV,false,0);

  //The Target Volume
  G4LogicalVolume* targetLV = 
    new G4LogicalVolume(new G4Box(
				   "TargetBe",
				   targetHalfWide,
				   targetHalfHeight,
				   targetHalfLength
				   ),
			SolBe,
			"TargerLog");
  /*G4PVPlacement* targetPV = */
    new G4PVPlacement(G4Transform3D(),"TargetBePhys",
		      targetLV,detectorPV,false,0);

  //------------
  // Sensitive Detector
  //------------
  G4SDManager* SDman = G4SDManager::GetSDMpointer();

  G4String sensorName = "SDetector";
  SDetector* aSD = new SDetector( sensorName );

  aSD->SetVerboseLevel(0);

  SDman->AddNewDetector( aSD ); //register this Sensitive detector (SD)
  detectorLV->SetSensitiveDetector( aSD ); // attache this SD to a Logical Vol

  //------------
  // User Limit
  //------------
  //Use the default ctor just to trigger the fact that there exists a 
  //User Limites for the given Logical Volume (JEC)
  G4UserLimits* theUserLimitForDetector = new G4UserLimits();
  detectorLV->SetUserLimits(theUserLimitForDetector);

  //usage:
  //G4UserLimits: DBL_MAX, /max allowed Step size in this volume
  //              DBL_MAX, //max total track length
  //              DBL_MAX, //max time
  //                   0., // min kinetic energy  (only for charged particles)
  //		       0.  // min remaining range (only for charged particles)
		       
  
  return worldPV;
}//Construct
