
//This code is inspired from G4 ExN04

//This
#include "G4LabTemplateTwo/DetectorHit.h"

//Geant4
#include <G4UnitsTable.hh>
#include <G4ios.hh>
#include <G4AttDefStore.hh>
#include <G4AttDef.hh>
#include <G4AttValue.hh>
 
namespace G4LabTemplateTwo {
  G4Allocator<DetectorHit> DetectorHitAllocator;
}

G4LabTemplateTwo::DetectorHit::DetectorHit() { m_LV = 0; }

G4LabTemplateTwo::DetectorHit::~DetectorHit() {;}

G4LabTemplateTwo::DetectorHit::DetectorHit(const G4LabTemplateTwo::DetectorHit &right) : G4VHit() {
  m_trackID      = right.m_trackID;
  m_trackPdgID   = right.m_trackPdgID;
  m_edep         = right.m_edep;
  m_pos          = right.m_pos;
  m_momentum     = right. m_momentum;
  m_totalE       = right.m_totalE;
  m_LV           = right.m_LV;
  m_TSF          = right.m_TSF;
}//Copy Ctor

const G4LabTemplateTwo::DetectorHit& 
G4LabTemplateTwo::DetectorHit::operator=(const G4LabTemplateTwo::DetectorHit &right) {

  m_trackID      = right.m_trackID;
  m_trackPdgID   = right.m_trackPdgID;
  m_edep         = right.m_edep;
  m_pos          = right.m_pos;
  m_momentum     = right.m_momentum;
  m_totalE       = right.m_totalE;
  m_TSF          = right.m_TSF;

  return *this;
}//= op

G4int G4LabTemplateTwo::DetectorHit::operator==(const G4LabTemplateTwo::DetectorHit &right) const {
  return (this==&right) ? 1 : 0;
}// == op

void G4LabTemplateTwo::DetectorHit::Draw() {;}

void G4LabTemplateTwo::DetectorHit::Print() {
  G4cout << "TrackID " << m_trackID 
	 << " PDG code " << m_trackPdgID
	 << " Edep " << G4BestUnit(m_edep,"Energy")
	 << " R " << G4BestUnit(m_pos.perp(),"Length")
	 << " Z " << G4BestUnit(m_pos.z(),"Length")
         << " PT " << G4BestUnit(m_momentum.perp(),"Energy")
	 << " PL " << G4BestUnit(m_momentum.z(),"Energy")
	 << " Etot " << G4BestUnit(m_totalE,"Energy")
	 << G4endl;
}//Print


const std::map<G4String,G4AttDef>* G4LabTemplateTwo::DetectorHit::GetAttDefs() const {
  G4bool isNew;
  std::map<G4String,G4AttDef>* store = 
    G4AttDefStore::GetInstance("Hit",isNew);

  if (isNew) {    
    G4String TrackID("TrackID");
    (*store)[TrackID] = G4AttDef(TrackID,"TrackID","Dump","","G4int");
    
    G4String Edep("Edep");
    (*store)[Edep] = G4AttDef(Edep,"Edep","Dump","","G4double");    

    G4String LV("LV");
    (*store)[LV] = 
      G4AttDef(LV,"Logical Volume","Display","","G4LogicalVolume");

    G4String TSF("TSF");
    (*store)[TSF] = 
      G4AttDef(TSF,"3D Transformation","Display","","G4Transform3D");
 
  }
  return store;
}//GetAttDefs

std::vector<G4AttValue>* G4LabTemplateTwo::DetectorHit::CreateAttValues() const {

  std::ostringstream s;
  
  std::vector<G4AttValue>* values = new std::vector<G4AttValue>;

  s.str("");
  s << m_trackID;
  values->push_back(G4AttValue("TrackID",s.str(),""));
  
  s.str("");
  s << G4BestUnit(m_edep,"Energy");
  values->push_back(G4AttValue("Edep",s.str(),""));
    
  s.str("");
  s << m_LV;
  values->push_back(G4AttValue("LV",s.str(),""));

  s.str("");
  s << &m_TSF;
  values->push_back(G4AttValue("TSF",s.str(),""));

  return values;
}//CreateAttValues
