
// this :
#include <G4LabTemplateTwo/EventAction.h>

//Geant4
#include <G4SDManager.hh>
#include <G4Event.hh>
#include <G4ios.hh>

// AIDA :
#include <AIDA/ITree.h>
#include <AIDA/IManagedObject.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
#include <AIDA/ITuple.h>

//G4LabTemplateTwo
#include "G4LabTemplateTwo/Cast.h"
#include "G4LabTemplateTwo/DetectorHit.h"
#include "G4LabTemplateTwo/Analysis.h"

G4LabTemplateTwo::EventAction::EventAction(G4LabTemplateTwo::Analysis& aAnalysis) 
:fAnalysis(aAnalysis)
,m_output("G4LabTemplateTwo400.dat")
,m_detectorCollID(0)
{



  //Get User Histo pointers
  AIDA::ITree* usrTree = aAnalysis.tree(); //get the tree
  if (!usrTree) {
    G4cout << "EventAction: FATAL: cannot get Analysis Tree" << G4endl;
    exit(0);
  }
  //see AIDA::IManagedObject a comment on dynamic_cast
  AIDA::IManagedObject* obj; //generic Object managed by a Tree

  obj = usrTree->find("histo_HitXY");
  if(!obj) {
    G4cout << "EventAction: FATAL: no histo_HitXY" << G4endl;
    usrTree->ls();
    exit(0);
  }
  //  m_histoXY = obj->cast("AIDA::IHistogram2D");
  m_histoXY = CAST(obj,AIDA::IHistogram2D);
  if(!m_histoXY){
    G4cout << "EventAction: FATAL: histoXY not a 2D" << G4endl;
    exit(0);
  }

  obj = usrTree->find("histo_HitTotalE");
  if(!obj) {
    G4cout << "EventAction: FATAL: no histo_HitTotalE" << G4endl;
    usrTree->ls();
    exit(0);
  }
  m_histoTotalE=  CAST(obj,AIDA::IHistogram1D);
  if(!m_histoTotalE) {
    G4cout << "EventAction: FATAL: histoTotalE not a 1D" << G4endl;
    usrTree->ls();
    exit(0);
  }


  obj = usrTree->find("histoPiPlus");
  if(!obj) {
    G4cout << "EventAction: FATAL: no histoPiPlus" << G4endl;
    usrTree->ls();
    exit(0);
  }
  m_histoPiPlus = CAST(obj,AIDA::IHistogram2D);
  if (!m_histoPiPlus) {
    G4cout << "EventAction: FATAL: histoPiPlus not a 2D " << G4endl;
    usrTree->ls();
    exit(0);
  }


  obj = usrTree->find("histoPiMinus");
  if(!obj) {
    G4cout << "EventAction: FATAL: no histoPiMinus" << G4endl;
    usrTree->ls();
    exit(0);
  }  
  m_histoPiMinus = CAST(obj,AIDA::IHistogram2D);
  if (!m_histoPiMinus) {
    G4cout << "EventAction: FATAL: histoPiMinus not a 2D" << G4endl;
    usrTree->ls();
    exit(0);
  }

  obj = usrTree->find("histoKPlus");
  if(!obj) {
    G4cout << "EventAction: FATAL: no histoKPlus " << G4endl;
    usrTree->ls();
    exit(0);
  }  
  m_histoKPlus = CAST(obj,AIDA::IHistogram2D);
  if (!m_histoKPlus) {
    G4cout << "EventAction: FATAL: histoKPlus not a 2D" << G4endl;
    usrTree->ls();
    exit(0);
  }

  obj = usrTree->find("histoKMinus");
  if(!obj) {
    G4cout << "EventAction: FATAL: no histoKMinus " << G4endl;
    usrTree->ls();
    exit(0);
  }  
  m_histoKMinus = CAST(obj,AIDA::IHistogram2D);
  if (!m_histoKMinus) {
    G4cout << "EventAction: FATAL: histoKMinus not a 2D" << G4endl;
    usrTree->ls();
    exit(0);
  }


  obj = usrTree->find("histoProton");
  if(!obj) {
    G4cout << "EventAction: FATAL: no histoProton" << G4endl;
    usrTree->ls();
    exit(0);
  }  
  m_histoProton = CAST(obj,AIDA::IHistogram2D);
  if (!m_histoProton) {
    G4cout << "EventAction: FATAL: histoProton not a 2D" << G4endl;
    usrTree->ls();
  }

  obj = usrTree->find("histoOther");
  if(!obj) {
    G4cout << "EventAction: FATAL: no histoProton" << G4endl;
    usrTree->ls();
    exit(0);
  }  
  m_histoOther =  CAST(obj,AIDA::IHistogram2D);
  if (!m_histoOther) {
    G4cout << "EventAction: FATAL: histoOther not a 2D" << G4endl;
    usrTree->ls();
    exit(0);
  }


  obj = usrTree->find("tupleOFFTarget");
  if(!obj) {
    G4cout << "EventAction: FATAL: no tupleOFFTarget" << G4endl;
    usrTree->ls();
    exit(0);
  }  
  m_tuple =  CAST(obj,AIDA::ITuple);
  if (!m_tuple) {
    G4cout << "EventAction: FATAL: tuple not a Tuple" << G4endl;
    usrTree->ls();
    exit(0);
  }

  
  //fix geometrical values for isAccepted functions
  m_triggerToTargetDistance = 10.0 * m;
  G4double triggerRadius    =  1.0 * m;
  m_triggerRadius2 = triggerRadius * triggerRadius; 
  //  G4cout << " (JEC) EventAction::Ctor" << G4endl;

  m_SDman = G4SDManager::GetSDMpointer();

  //m_verboseLevel = m_SDMan->GetVerboseLevel(); //one cannot retreive SDManager level !!! not symmetric to the EventManager...
  m_verboseLevel = 0;
}//Ctor

//-----------------------------------------------------------------------------------

G4LabTemplateTwo::EventAction::~EventAction() {
  m_output.close();
}//Dtor

//-----------------------------------------------------------------------------------

void G4LabTemplateTwo::EventAction::BeginOfEventAction(const G4Event* /* ievt */) {
  /*
    G4cout << " (JEC) EventAction::Begin EventAction" << G4endl;
    if (!ievt) ievt->Print();
  */
}//BeginOfEventAction

//-----------------------------------------------------------------------------------

void G4LabTemplateTwo::EventAction::EndOfEventAction(const G4Event* ievt) {
  /* 
     G4cout << " (JEC) EventAction::End   EventAction" << G4endl;
     if (!ievt) ievt->Print();
  */
  m_detectorCollID = m_SDman->GetCollectionID("hitCollection");

  G4HCofThisEvent* HCE = ievt->GetHCofThisEvent();
  DetectorHitsCollection* DetHC = 0;
  if (HCE) {
    DetHC = (DetectorHitsCollection*)(HCE->GetHC(m_detectorCollID));   
  }
  if (DetHC) {
    G4int NbHits = DetHC->entries();
    DetectorHit* iHit = 0;
    if (m_verboseLevel>0) { 
      G4cout << "\n-------->Hits Collection: in this event they are " << NbHits
	     << " hits " << G4endl;
    }
    
    G4double ptot    = 0.; //Total Momentum GeV
    G4double pT      = 0.; //Transverse Momentum GeV
    G4int    tkPdgId = 0;  //PDG Id of the particle
    G4int    tkId    = 0;  //Geant Track Id
    G4ThreeVector direction;
    
    G4double weight = 0.; //default weight for filling plots

    G4ThreeVector initPosition;
    G4double x,y,z = 0.;    //particle position at Exit of the Target (mm)
    G4double px,py,pz = 0.; //particle momentum at Exit of the Target (GeV)
    
    for (G4int i=0;i<NbHits;i++) {
      iHit = (*DetHC)[i];
      if (m_verboseLevel>0) { iHit->Print(); }

      //Tuple filling
      tkPdgId = iHit->GetTrackPdgID();
      if ( 22 != tkPdgId ) { //not a Gamma
	px = (iHit->GetMomentum()).x()/GeV;
	py = (iHit->GetMomentum()).y()/GeV;
	pz = (iHit->GetMomentum()).z()/GeV;
	initPosition = iHit->GetPos();
	x  = (initPosition.x())/mm;
	y  = (initPosition.y())/mm;
	z  = (initPosition.z())/mm;
	
// 	m_output << tkPdgId << " " 
// 		 << x << " "
// 		 << y << " "
// 		 << z << " "
// 		 << px << " "
// 		 << py << " "
// 		 << pz << std::endl;
	m_tuple->fill(0,tkPdgId);
	m_tuple->fill(1,x);
	m_tuple->fill(2,y);
	m_tuple->fill(3,z);
	m_tuple->fill(4,px);
	m_tuple->fill(5,py);
	m_tuple->fill(6,pz);
	m_tuple->addRow();
      }
      //Fill User histograms
      m_histoTotalE->fill( iHit->GetTotalE() );
      /* TrackId: PDG encoding
	 p    : 2212
	 pbar: -2212
	 n    : 2112
	 pi+:    211
	 pi-:   -211
	 pi0:    111
	 K+:     321
	 K-:    -321
	 K0S:    310
	 K0L:    130
	 e+:     -11
	 e-:     +11
	 mu+:    -13
	 mu-:    +13
	 gamma:   22
      */
      ptot    = (iHit->GetMomentum().mag())/GeV; //in GeV units
      pT      = (iHit->GetMomentum().perp())/GeV;

      tkId    = iHit->GetTrackID();
      direction = iHit->GetMomentum().unit();
      
      if (m_verboseLevel>1) {
	G4cout << " ptot: " << ptot << G4endl;
      }
      //Look if the particle is accepted by solid angle arguments
      
      if ( (weight = isAccecpted(direction,initPosition)) > 0 ) {
	if (m_verboseLevel>1) {
	  G4cout << " accepted PDG part.: " << tkPdgId << G4endl;
	}
	if ( 211 == tkPdgId ) { 	        //pi+
	  m_histoPiPlus->fill(ptot,pT,weight);
	} else  if ( -211 == tkPdgId ) {	//pi-
	  m_histoPiMinus->fill(ptot,pT,weight);
	} else  if ( 321 == tkPdgId ) {	//K+
	  m_histoKPlus->fill(ptot,pT,weight);
	} else  if ( -321 == tkPdgId ) { 	//K-
	  m_histoKMinus->fill(ptot,pT,weight);
	} else  if ( (2212 == tkPdgId) && ( 1 != tkId) ) { //secondary proton
	  m_histoProton->fill(ptot,pT,weight);
	} else {
	  m_histoOther->fill(ptot,pT,weight);
	}
      }
    }//eof loop on hits 
  }//Detector Hit Collection
}//BeginOfEventAction

//---------------------------------------------------------------------------
G4double G4LabTemplateTwo::EventAction::isAccecpted(const G4ThreeVector& dir, 
					 const G4ThreeVector& pos) {
  
  G4double wSolidAngle = -1.;
  
  if (m_verboseLevel>1) { 
    G4cout 
      << "(X,Y,Z): " << pos.x() << " " << pos.y() << " " << pos.z() 
      << G4endl;
    G4cout 
      << "e(X,Y,Z):" << dir.x() << " " << dir.y() << " " << dir.z() 
      << G4endl;
  }

  if(dir.z() <= 0.) return wSolidAngle; //backward direction not accepted

   
  G4double deltaZ = m_triggerToTargetDistance - pos.z();
  G4double cosX   = dir.x()/dir.z();
  G4double cosY   = dir.y()/dir.z();
  G4double xTrig  = pos.x() + cosX * deltaZ;
  G4double yTrig  = pos.y() + cosY * deltaZ;
  
  m_histoXY->fill( xTrig/cm, yTrig/cm );

  G4double r2 = pow(xTrig,2.0) + pow(yTrig,2.0);
  if (m_verboseLevel>1) { 
    G4cout << "trig->target : " << m_triggerToTargetDistance 
	   << " trig radius**2 : " << m_triggerRadius2 << G4endl; 
    G4cout << " xTrig , yTrig " << xTrig << " " << yTrig << G4endl;
  } 
  if ( r2 <  m_triggerRadius2 ) { 
    wSolidAngle = 
      pi * r2/(m_triggerToTargetDistance*m_triggerToTargetDistance);
  }

  return 1.;
}//isAccecpted



