//This code is inspired from ExN02TrackerSD

//This
#include "G4LabTemplateTwo/SDetector.h"

//Geant4
#include <G4HCofThisEvent.hh>
#include <G4Step.hh>
#include <G4ThreeVector.hh>
#include <G4SDManager.hh>
#include <G4ios.hh>


// CLHEP:
#include <CLHEP/Units/PhysicalConstants.h>


G4LabTemplateTwo::SDetector::SDetector(G4String name)
:G4VSensitiveDetector(name)
,hitCollection(0)
,HCID(-1)
{
  G4String HCname;
  collectionName.insert(HCname="hitCollection");
}//Ctor

//----------------------------------------------------------------------------

G4LabTemplateTwo::SDetector::~SDetector(){ }

//----------------------------------------------------------------------------

void G4LabTemplateTwo::SDetector::Initialize(G4HCofThisEvent*)
{
  hitCollection = new DetectorHitsCollection
                      (SensitiveDetectorName,collectionName[0]); 
}//Initialize

//-----------------------------------------------------------------------

G4bool G4LabTemplateTwo::SDetector::ProcessHits(G4Step* aStep,G4TouchableHistory*)
{
 
  //which track ?
  G4Track* currentTk = aStep->GetTrack();
  if (!currentTk) return false;

  //Get PreStep point to test boundary
  G4StepPoint* point1 = aStep->GetPreStepPoint();
  if (!point1) return false;
  
  G4StepPoint* preStepPoint = aStep->GetPreStepPoint();
  G4TouchableHandle touchable = preStepPoint->GetTouchableHandle();
  G4VPhysicalVolume* pv = touchable->GetVolume();
  G4AffineTransform trans = touchable->GetHistory()->GetTopTransform();
  G4RotationMatrix rot = trans.NetRotation();
  trans.Invert();
  G4ThreeVector pos = trans.NetTranslation();
  G4Transform3D tsf(rot,pos);

  if (point1->GetStepStatus() == fGeomBoundary) {
    //particle enter the volume => new hit
    DetectorHit* newHit = new DetectorHit();
    newHit->SetTrackID    (currentTk->GetTrackID());
    newHit->SetTrackPdgID (currentTk->GetDefinition()->GetPDGEncoding());
    newHit->SetMomentum   (currentTk->GetMomentum());
    newHit->SetTotalE     (currentTk->GetTotalEnergy());
    
    newHit->SetEdep       (aStep->GetTotalEnergyDeposit());
    newHit->SetPos        (aStep->GetPreStepPoint()->GetPosition());
    
    newHit->SetLV(pv->GetLogicalVolume());
    newHit->SetTSF(tsf);

    hitCollection->insert( newHit );
  }
  
  return true;
}//ProcessHits

//----------------------------------------------------------------------------

void G4LabTemplateTwo::SDetector::EndOfEvent(G4HCofThisEvent* HCE)
{
  
  if(verboseLevel>0) PrintAll();

  if(HCID<0)
  { HCID = G4SDManager::GetSDMpointer()->GetCollectionID(collectionName[0]); }
  HCE->AddHitsCollection( HCID, hitCollection ); 
}//EndOfEvent

void G4LabTemplateTwo::SDetector::PrintAll() {
  
  G4int NbHits = hitCollection->entries();
  G4cout << "\n--------<"<< GetName() << "> Hits Collection contains " << NbHits
	 << " hits " << G4endl;
  DetectorHit* iHit = 0;
  for (G4int i=0;i<NbHits;i++) {
      iHit = (*hitCollection)[i];
      iHit->Print();
  }
}//PrintAll

