//This
#include "G4LabTemplateTwo/SpecialCuts.h"

//Geant4
#include "G4Step.hh"
#include "G4UserLimits.hh"
#include "G4ParticleChange.hh"

//G4LabTemplateTwo
#include "G4LabTemplateTwo/StepLimiterMessenger.h"



G4LabTemplateTwo::SpecialCuts::SpecialCuts(const G4String& aName)
  : G4VProcess(aName),
    m_verboseLevel(1)
{

  pParticleChange = new G4ParticleChange;
  if (!pParticleChange) {
    G4Exception("ERROR:TiaraEnergyCutProcess::SpecialCuts new failed to create G4ParticleChange!");
  }
 
  pMessenger = new StepLimiterMessenger(this);

  if (m_verboseLevel>0) {
    G4cout << "(JEC) SpecialCuts " << GetProcessName() << " is created "<< G4endl;
  }
}//Ctor

//----------------------------------------------------------------------

G4LabTemplateTwo::SpecialCuts::~SpecialCuts() {
  delete pParticleChange;
  delete pMessenger;
}//Dtor

//----------------------------------------------------------------------


void G4LabTemplateTwo::SpecialCuts::SetVerbose(G4int verboseLevel) {
  m_verboseLevel = verboseLevel;
}

//----------------------------------------------------------------------

G4LabTemplateTwo::SpecialCuts::SpecialCuts(SpecialCuts& right)
  : G4VProcess(right)
{
}

//----------------------------------------------------------------------
 

G4double 
  G4LabTemplateTwo::SpecialCuts::PostStepGetPhysicalInteractionLength( 
				       const G4Track& aTrack,
				       G4double, // previousStepSize
				       G4ForceCondition* condition  ) {

  if (m_verboseLevel>1) {
    G4cout << "(JEC): " << GetProcessName() 
	   << " called for " 
	   << aTrack.GetDefinition()->GetParticleName()
	   << " in volume "
	   << aTrack.GetVolume()->GetLogicalVolume()->GetName()
	   << G4endl;
  }
  
  // condition is set to "Not Forced" by default
  *condition = NotForced;
  
  G4double proposedStep = DBL_MAX;
  //Look if the volume as a User Limits
  G4UserLimits* pUserLimits =
    aTrack.GetVolume()->GetLogicalVolume()->GetUserLimits();
  if (pUserLimits) {
    // max allowed step length + force it
    //
    *condition = Forced; 
    proposedStep = pUserLimits->GetMaxAllowedStep(aTrack);
    if (m_verboseLevel>1) {
      G4cout << "(JEC): " << GetProcessName() 
	     << " active & step proposed = "
	     << proposedStep
	     << G4endl;
    }
    if (proposedStep < 0.) proposedStep =0.; 
  }
  return proposedStep;
}//PostStepGetPhysicalInteractionLength

//----------------------------------------------------------------------

G4VParticleChange*
  G4LabTemplateTwo::SpecialCuts::PostStepDoIt( const G4Track& aTrack,
				    const G4Step&  aStep) {

   pParticleChange->Initialize(aTrack);
   //code inspired from http://geant4.web.cern.ch/geant4/support/faq.html
   //check if the particle is entering the volume
   G4StepPoint* point1 = aStep.GetPreStepPoint();
   if (m_verboseLevel>1) {
     G4cout << "(JEC): " << GetProcessName() 
	    << " PostStepDoIt:: preStep (R,Z) "
	    << (point1->GetPosition()).perp()
	    << " "
	    << (point1->GetPosition()).z();
   }
   
   if (point1->GetStepStatus() == fGeomBoundary) {
     if (m_verboseLevel>1) {
       G4cout << " Enter Volume: Kill";
     }
     //Stop the current particle IF entering in the Volume considered
     //and required by the UserLimits.
     //The following lines are from N05 novice example
     pParticleChange->ProposeTrackStatus(fStopAndKill);
   }
   if (m_verboseLevel>1) {
     G4cout << G4endl;
   }
   return pParticleChange;
}//PostStepDoIt













