
// this :
#include <G4LabTemplateTwo/SteppingAction.h>

// AIDA :
#include <AIDA/IHistogram1D.h>
#include <AIDA/ITree.h>
#include <AIDA/IManagedObject.h>

// Geant4 :
#include <G4Step.hh>

//G4LabTemplateTwo
#include <G4LabTemplateTwo/Analysis.h>
#include "G4LabTemplateTwo/Cast.h"

G4LabTemplateTwo::SteppingAction::SteppingAction(Analysis& aAnalysis)
:fAnalysis(aAnalysis){
  //Get User Histo pointers
  AIDA::ITree* usrTree = aAnalysis.tree(); //get the tree

  //see AIDA::IManagedObject a comment on dynamic_cast
  AIDA::IManagedObject* obj; //generic Object managed by a Tree

  obj = usrTree->find("histo_DE");
  if(!obj) {
    G4cout << "SteppingAction: FATAL: no histo_DE" << G4endl;
    usrTree->ls();
    exit(0);
  }
  m_histoDE = CAST(obj,AIDA::IHistogram1D);
  if(!m_histoDE){
    G4cout << "SteppingAction: FATAL: histoDE not a 1D" << G4endl;
    exit(0);
  }

  obj = usrTree->find("histo_DL");
  if(!obj) {
    G4cout << "SteppingAction: FATAL: no histo_DL" << G4endl;
    usrTree->ls();
    exit(0);
  }
  m_histoDL = CAST(obj,AIDA::IHistogram1D);
  if(!m_histoDL){
    G4cout << "SteppingAction: FATAL: histoDL not a 1D" << G4endl;
    exit(0);
  }

}//Ctor
//--------------------------------------------------------------------------

G4LabTemplateTwo::SteppingAction::~SteppingAction(){}

//--------------------------------------------------------------------------

void G4LabTemplateTwo::SteppingAction::UserSteppingAction(const G4Step* aStep) {
  // Get out MeV and mm :
  G4double delta = aStep->GetDeltaEnergy();
  //delta *= 10000.;
  delta = ::fabs(delta);
  if(delta!=0.) m_histoDE->fill(fabs(delta),1.);
  G4double length = aStep->GetStepLength();
  length *= 10.;
  if(length!=0.) m_histoDL->fill(length,1.);
}//UserSteppingAction

