#include <HEPVis/Gui/SoGui.h>

#include <Inventor/errors/SoDebugError.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/nodes/SoCube.h>

#include <HEPVis/SbArguments.h>
#include <HEPVis/SbColors.h>
#include <HEPVis/misc/SoTools.h>
#include <HEPVis/nodekits/SoAxis.h>
#include <HEPVis/nodekits/SoAxes2D.h>
#include <HEPVis/nodes/SoTextHershey.h>
#include <HEPVis/nodes/SoTextTTF.h>
#include <HEPVis/nodes/SoGrid.h>
#include <HEPVis/nodes/SoTextStyle.h>
#include <HEPVis/nodes/SoLineStyle.h>
#include <HEPVis/actions/SoAlternateRepAction.h>

#include <stdlib.h>

int main(int aArgc, char** aArgv) {

  SbArguments args(aArgc,aArgv);

  // Set up the main window
  SoGuiWidget topWidget=SoGui::init(aArgv[0]);
  if (!topWidget) exit(1);

  // initialize HEPVis used classes :
  SoTextHershey::initClass();
  SoTextTTF::initClass();
  SoTextStyle::initClass();
  SoLineStyle::initClass();
  SoAxis::initClass();
  SoAxes2D::initClass();
  SoGrid::initClass();
  SoAlternateRepAction::initClass();

  SoSeparator* root=new SoSeparator;
  root->ref();

  SoLightModel* lightModel = new SoLightModel;
  //lightModel->model.setValue(SoLightModel::BASE_COLOR);
  root->addChild(lightModel);

  SoAxes2D* axes = new SoAxes2D;
  axes->getXAxis()->getLabelsStyle()->color.setValue(SbColor_yellow);
  axes->getYAxis()->getLabelsStyle()->color.setValue(SbColor_yellow);
  
  axes->gridVisible.setValue(TRUE);
  axes->wallVisible.setValue(TRUE);
  axes->wallColor.setValue(SbColor(0.5f,0.5f,0.5f));

  axes->getXAxis()->getLineStyle()->pattern.setValue(SbLinePattern_dashed);
  axes->getYAxis()->getLineStyle()->pattern.setValue(SbLinePattern_dotted);

  axes->getXAxis()->minimumValue.setValue(-13);
  axes->getXAxis()->maximumValue.setValue(20);
  axes->getXAxis()->title.setValue("AxisX2");

  axes->getYAxis()->minimumValue.setValue(0);
  axes->getYAxis()->maximumValue.setValue(5.5f);
  axes->getYAxis()->title.setValue("AxisY2");
  root->addChild(axes);  

  SoTranslation* translation = new SoTranslation;
  translation->translation.setValue(SbVec3f(0.5F,0.5F,0.5F));
  root->addChild(translation);

  SoCube* cube = new SoCube();
  cube->width = 0.1f;
  cube->height = 0.1f;
  cube->depth = 0.1f;
  root->addChild(cube);

  if (args.isAnArgument("-write")) {
    SbBool altRep = args.isAnArgument("-altrep")?TRUE:FALSE;
    SbString file;
    args.find("-file",file);
    if(file=="") file = "out.iv";
    SoTools::write(*root,file,"",FALSE,altRep);
  }

  SoGuiExaminerViewer* examinerViewer = new SoGuiExaminerViewer(topWidget);
  examinerViewer->setSceneGraph(root);
  examinerViewer->viewAll();
  examinerViewer->show();

  SoGui::show(topWidget);
  SoGui::mainLoop();

  root->unref();
  delete examinerViewer;

#ifdef SoFreePackage //Cleanup
  SoGui::clear();  
#endif

  return 0; 
}


