#include <math.h>

#include <stdlib.h>

#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/actions/SoWriteAction.h>

#include <HEPVis/SbMath.h>
#include <HEPVis/nodes/SoCons.h>

//
// Test of SoCons class
// Joe Boudreau Nov. 11, 1996
//
// This demo draws a G4Cons
//

int main(int, char ** argv) {

  // Set up the main window
 
  SoGuiWidget myWindow=SoGui::init(argv[0]);
  if (!myWindow) exit(1);

  // You must initialize a new class
  SoCons::initClass();

  // Set up the root of the scene database and
  // reference it

  SoSeparator *root=new SoSeparator;
  root->ref();

  // Set up the material for the object to be drawn

  SoMaterial    *RedPlastic = new SoMaterial;

  RedPlastic->diffuseColor.setValue(1, 0, 0);
  root->addChild(RedPlastic);

  // Users will expect a new node to change appearance when a field
  // is changed.  Here, we connect one of the fields to an engine
  // which should animate the object if this feature is properly
  // implemented.

  SoTimeCounter *TheCounter = new SoTimeCounter;
  TheCounter->max  = 360; 
  TheCounter->step = 4;
  TheCounter->frequency=0.2F;
  
  SoCalculator *TheCalculator = new SoCalculator();
  TheCalculator->a.connectFrom(&TheCounter->output);
  TheCalculator->expression.setValue("oa=a*M_PI/180.0");
  
  // Make the G4Cons

  SoCons *TheG4Cons   = new SoCons();
  TheG4Cons->fRmin1 = 0.5F;
  TheG4Cons->fRmax1 = 1;
  TheG4Cons->fRmin2 = 4.5F;
  TheG4Cons->fRmax2 = 5;
  TheG4Cons->fDz   = 1.0;
  TheG4Cons->fDPhi = FM_PI;
  if (getenv("HEPVIS_ALTERNATEREP")||getenv("ALTERNATEREP")) TheG4Cons->generateAlternateRep(0);
  TheG4Cons->fDPhi.connectFrom(&TheCalculator->oa);
  root->addChild(TheG4Cons);

  SoWriteAction *WriteAction= new SoWriteAction();
  WriteAction->getOutput()->setBinary(FALSE);
  WriteAction->apply(root);

  // View the new object in an Examiner Viewer

  SoGuiExaminerViewer *myViewer=new SoGuiExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  myViewer->viewAll();
  myViewer->show();

  SoGui::show(myWindow);
  SoGui::mainLoop();

  root->unref();
  delete WriteAction;
  delete myViewer;

#ifdef SoFreePackage //Cleanup
  SoCons::clearClass();
  SoGui::clear();  
#endif

  return 0; 

}

