#include <math.h>
#include <stdlib.h>

#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/actions/SoWriteAction.h>

#include <HEPVis/nodes/SoSiStrips.h>
#include <HEPVis/nodekits/SoDblSiStripDetKit.h>

// Test of SoDblSiStripDet class
// Joe Boudreau Nov. 11, 1996
//
// This demo draws a SiStripDet
//
short Strip[8]  = {100, 101, 102, 103, 130, 131, 132, 133};
short ADC[8]    = {50,   80, 110,  70,  20,  34,  18,   5};
short Strip2[8] = {30,   31,  32,  33,  42,  43,  44,  45};
short ADC2[8]   = {53,   60,  94,  20,  19,  34,  53,  25};
 
int main(int, char ** argv) {

 // Put everything in a block for SunOS/CC memory decount :
 {
  // Set up the main window
 
  SoGuiWidget myWindow=SoGui::init(argv[0]);
  if (!myWindow) exit(1);

  // You must initialize a new class

  SoDblSiStripDetKit::initClass();
  SoSiStrips::initClass();
  // Set up the root of the scene database and
  // reference it

  SoSeparator *root=new SoSeparator;
  root->ref();

  SoDblSiStripDetKit *TheSiStripDet = new SoDblSiStripDetKit();

  SoMaterial* RedPlastic = 
    (SoMaterial*)TheSiStripDet->getPart("waferAppearance.material",TRUE);
  RedPlastic->diffuseColor.setValue(1, 0, 0);

  SoMaterial* GreenPlastic = 
    (SoMaterial*)TheSiStripDet->getPart("stripsAppearance.material",TRUE);
  GreenPlastic->diffuseColor.setValue(0,1,0);  

  SoSiStrips* ThePhiStrips = 
    (SoSiStrips*)TheSiStripDet->getPart("phiStrips",TRUE);
  SoSiStrips* TheZStrips = 
    (SoSiStrips*)TheSiStripDet->getPart("zStrips",TRUE);
  ThePhiStrips->strip.setValues(0,8,Strip);
  ThePhiStrips->adc.setValues(0,8,ADC);

  // Make the Z  Strips
  TheZStrips->strip.setValues(0,8,Strip2);
  TheZStrips->adc.setValues(0,8,ADC2);

  // This is probably a good idea

  ThePhiStrips->length.setValue(4.0);
  TheZStrips->length.connectFrom(&ThePhiStrips->width);
  TheZStrips->width.connectFrom(&ThePhiStrips->length);

  root->addChild(TheSiStripDet);
  
  if (getenv("HEPVIS_ALTERNATEREP")||getenv("ALTERNATEREP")) TheSiStripDet->generateAlternateRep(0);
  SoWriteAction *WriteAction= new SoWriteAction();
  WriteAction->getOutput()->setBinary(FALSE);
  WriteAction->apply(root);

  // View the new object in an Examiner Viewer
  TheSiStripDet->setAlignable(TRUE);
  SoGuiExaminerViewer *myViewer=new SoGuiExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  myViewer->show();

  SoGui::show(myWindow);
  SoGui::mainLoop();

  root->unref();
  delete WriteAction;
  delete myViewer;}

#ifdef SoFreePackage //Cleanup
  SoDblSiStripDetKit::clearClass();
  SoSiStrips::clearClass();
  SoGui::clear();  
#endif

  return 0; 

}

