// EllipsodSegment.C
// ===================================================================
//
//    Main Program for Testing SoEllipsoidSegment Class  
//    
//    This Demo Draws a Segment of Ellipsoid 
//
//    Xiaoling Fan, Northwestern University
//
//    Sept., 1996
//
// ===================================================================
#include <math.h>
#include <stdlib.h>

#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoFaceSet.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/actions/SoWriteAction.h>

#include <HEPVis/SbMath.h>
#include <HEPVis/nodes/SoEllipsoidSegment.h>

int main(int, char **argv)
{
  SoGuiWidget myWindow = SoGui::init(argv[0]);
  if (myWindow == NULL) exit(1);

  SoEllipsoidSegment::initClass();  

  SoSeparator* root = new SoSeparator;
  root->ref();

  // Create a light

  root->addChild(new SoDirectionalLight);

  // Color

  SoMaterial* myMaterial = new SoMaterial;
  myMaterial->diffuseColor.setValue(1.0,0.0,1.0);
  root->addChild(myMaterial);

  SoTimeCounter* timeCounter = new SoTimeCounter;
  timeCounter->max  = 360; 
  timeCounter->step = 4;
  timeCounter->frequency=0.2F;
  
  SoCalculator* calculator = new SoCalculator();
  calculator->a.connectFrom(&timeCounter->output);
  calculator->expression.setValue("oa=a*M_PI/180.0");
  
  SoEllipsoidSegment* ellipsoid = new SoEllipsoidSegment();
  ellipsoid->ax.setValue(45);
  ellipsoid->by.setValue(60);
  ellipsoid->cz.setValue(90);
  //Start at theta null to see if there is no anomaly on top.
  //ellipsoid->fSTheta.setValue(FM_PI_2);
  ellipsoid->fDTheta.setValue(FM_PI_2);
  ellipsoid->fSPhi.setValue(0);
  ellipsoid->fDPhi.setValue(FM_PI/4);
  ellipsoid->nSides.setValue(32);
  //ellipsoid->nSides.setValue(4);
  if (getenv("HEPVIS_ALTERNATEREP")||getenv("ALTERNATEREP")) ellipsoid->generateAlternateRep(0);
  ellipsoid->fDPhi.connectFrom(&calculator->oa);
  root->addChild(ellipsoid);
  
  // Write out log file 

  SoWriteAction* writeAction= new SoWriteAction();
  writeAction->getOutput()->setBinary(FALSE);
  writeAction->apply(root);

  // Adding the Examiner Viewer:

  SoGuiExaminerViewer *myViewer = new SoGuiExaminerViewer (myWindow);  
  myViewer->setSceneGraph(root);

  myViewer->setTitle("EllipsoidSegment");
  myViewer->show();

  SoGui::show(myWindow);
  SoGui::mainLoop();
 
  root->unref();
  delete writeAction;
  delete myViewer;

#ifdef SoFreePackage //Cleanup
  SoEllipsoidSegment::clearClass();
  SoGui::clear();  
#endif

  return 0;
}









