#include <math.h>
#include <stdlib.h>

#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/actions/SoWriteAction.h>

#include <HEPVis/nodes/SoHadEmCrystalHit.h>
#include <HEPVis/nodes/SoCoordinateAxis.h>

// Test of SoHadEmCrystalHit class
// Joe Boudreau Nov. 2000

// This demo draws a HadEmCrystalHit
 
int main(int, char ** argv) {

  // Set up the main window
 
  SoGuiWidget myWindow=SoGui::init(argv[0]);
  if (!myWindow) exit(1);

  // You must initialize a new class

  SoHadEmCrystalHit::initClass();
  SoCoordinateAxis::initClass();

  // Set up the root of the scene database and
  // reference it

  SoSeparator *root=new SoSeparator;
  root->ref();

  SoCoordinateAxis *TheAxis   = new SoCoordinateAxis();
  TheAxis->fNDivision = 4;
  TheAxis->fDivisionLength = 10;
  root->addChild(TheAxis);

  // Set up the material for the object to be drawn

  SoMaterial    *RedPlastic = new SoMaterial;
  RedPlastic->diffuseColor.setValue(1, 0, 0);
  root->addChild(RedPlastic);

  // Users will expect a new node to change appearance when a field
  // is changed.  Here, we connect one of the fields to an engine
  // which should animate the object if this feature is properly
  // implemented.

  SoTimeCounter *TheCounter = new SoTimeCounter;
  TheCounter->max  = 360; 
  TheCounter->step = 4;
  TheCounter->frequency=0.2F;
  
  SoCalculator *TheCalculator = new SoCalculator();
  TheCalculator->a.connectFrom(&TheCounter->output);
  TheCalculator->expression.set1Value(0,"oa=4*sin(a*M_PI/180.0)*sin(a*M_PI/180.0)");
  TheCalculator->expression.set1Value(1,"ob=4*cos(a*M_PI/180.0)*cos(a*M_PI/180.0)");

  // Make the HadEmCrystalHit

  SoHadEmCrystalHit *TheHadEmCrystalHit    = new SoHadEmCrystalHit();
  TheHadEmCrystalHit->EnergyHad              = 4.0;
  TheHadEmCrystalHit->EnergyEm               = 4.0;
  TheHadEmCrystalHit->MinimumEnergyTot       = 0.0;
  TheHadEmCrystalHit->Scale                  = 5.0;
  TheHadEmCrystalHit->Front1                 =SbVec3f(0,20,0);
  TheHadEmCrystalHit->Front2                 =SbVec3f(2,20,0);
  TheHadEmCrystalHit->Front3                 =SbVec3f(2,20,2);
  TheHadEmCrystalHit->Front4                 =SbVec3f(0,20,2);
  if (getenv("HEPVIS_ALTERNATEREP")||getenv("ALTERNATEREP")) TheHadEmCrystalHit->generateAlternateRep(0);
  TheHadEmCrystalHit->EnergyEm.connectFrom(&TheCalculator->oa);
  TheHadEmCrystalHit->EnergyHad.connectFrom(&TheCalculator->ob);
  root->addChild(TheHadEmCrystalHit);

  SoWriteAction *WriteAction= new SoWriteAction();
  WriteAction->getOutput()->setBinary(FALSE);
  WriteAction->apply(root);

  // View the new object in an Examiner Viewer


  SoGuiExaminerViewer *myViewer=new SoGuiExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  myViewer->viewAll();
  myViewer->show();

  SoGui::show(myWindow);
  SoGui::mainLoop();

  root->unref();
  delete WriteAction;
  delete myViewer;

#ifdef SoFreePackage //Cleanup
  SoHadEmCrystalHit::clearClass();
  SoCoordinateAxis::clearClass();
  SoGui::clear();  
#endif

  return 0; 

}

