//*****************************************************************************
//
// Test the jet class
//
//*****************************************************************************
#include <math.h>
#include <stdlib.h>

#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/actions/SoWriteAction.h>

#include <HEPVis/SbMath.h>
#include <HEPVis/nodes/SoJet.h>

int main(int, char **argv)
{

   // Initialize Inventor and Xt
   SoGuiWidget myWindow = SoGui::init(argv[0]);
   if (myWindow == NULL) exit(1);

   SoJet::initClass();

   SoSeparator *root = new SoSeparator;
   root->ref();

   SoJet* theJet = new SoJet;
   theJet->fTheta.setValue(FM_PI/6);
   theJet->fPhi.setValue(3*FM_PI/4);
   theJet->fEnergy.setValue(35);

   SoJet *anotherJet = new SoJet;
   anotherJet->fTheta.setValue(2);

   anotherJet->fPhi.setValue(1);
   anotherJet->fTheta.setValue(1);
   anotherJet->fMaxZ.setValue(3);
   anotherJet->fMaxR.setValue(1);
   anotherJet->fEnergy.setValue(20);

   SoElapsedTime *myCounter = new SoElapsedTime;
   myCounter->speed.setValue(0.1F);

   anotherJet->fTheta.connectFrom(&myCounter->timeOut);


   root->addChild(theJet);

   if (getenv("HEPVIS_ALTERNATEREP")||getenv("ALTERNATEREP"))  theJet->generateAlternateRep(0);

   SoWriteAction *WriteAction= new SoWriteAction();
   WriteAction->getOutput()->setBinary(FALSE);
   WriteAction->apply(root);

   theJet->fPhi.connectFrom(&myCounter->timeOut);
   theJet->fTheta.connectFrom(&myCounter->timeOut);

   root->addChild(anotherJet);

   SoGuiExaminerViewer *myViewer = 
            new SoGuiExaminerViewer(myWindow);
   myViewer->setSceneGraph(root);
   myViewer->setTitle("testSoJet");
   myViewer->viewAll();
   myViewer->show();

   SoGui::show(myWindow);
   SoGui::mainLoop();

   root->unref();
   delete WriteAction;
   delete myViewer;

#ifdef SoFreePackage //Cleanup
   SoJet::clearClass();
   SoGui::clear();  
#endif

   return 0;
}




