#include <HEPVis/Gui/SoGui.h>

// Inventor :
#include <Inventor/errors/SoDebugError.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoEventCallback.h>
//#include <Inventor/nodes/SoMarkerSet.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/actions/SoWriteAction.h>

// HEPVisUtils :
#include <HEPVis/SbArguments.h>
#include <HEPVis/misc/SoTools.h>
#include <HEPVis/nodes/SoMarkerSet.h>
#include <HEPVis/actions/SoGL2PSAction.h>
#include <HEPVis/actions/SoAlternateRepAction.h>

// Else :
#include <stdlib.h>
#include <iostream>

SoGuiExaminerViewer* examinerViewer = 0; 

#ifdef HEPVisWin
#include "MenuBar.ic"
HEPVis_menuBar_def
#endif

void keystroke(void*,SoEventCallback*);

//////////////////////////////////////////////////////////////////////////////
int main (
 int aArgc
,char** aArgv
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SbArguments args(aArgc,aArgv);

  if(args.isAnArgument("-h")) {
    std::cout << "HEPVis::MarkerSet :" << std::endl;
    std::cout << "Usage :" << std::endl;
    std::cout << " OS> MarkerSet<GUI> <options>" << std::endl;
    std::cout << "Options :" << std::endl;
    std::cout << " -h : to get help." << std::endl;
    std::cout << " -altrep : to produce an alternate representation." 
              << std::endl;
    std::cout << " -write : to apply SoWriteAction." << std::endl;
    std::cout << " -size=[5,7,9] : marker size." << std::endl;
    std::cout << "For example :" << std::endl;
    std::cout << " OS> MarkerSetXt -size=9" << std::endl;
    std::cout << " OS> MarkerSetXt -altrep -write" << std::endl;
    std::cout << " UNIX> MarkerSetXt -altrep -write > out.iv" << std::endl;
    std::cout << " UNIX> ivview out.iv" << std::endl;
    return 0;
  }

  SoGuiWidget topWidget = SoGui::init(aArgv[0]);
  if (!topWidget) ::exit(1);

  HEPVis_SoMarkerSet::initClass();
  SoGL2PSAction::initClass();
  SoAlternateRepAction::initClass();

  SoSeparator* sceneGraph = new SoSeparator;
  sceneGraph->ref();

  SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue(1, 0, 0);
  sceneGraph->addChild(material);

  int style = 900;
 {SbString value;
  if(args.find("-size",value)==TRUE) {
    if(value=="5") style = 500;
    else if(value=="7") style = 700;
    else if(value=="9") style = 900;
  }}

  int mn = style;
  int mx = style+20;

  float step = 0.1F;
  float d = step;

  for(int index = mn;index<mx;index++) {
    SbVec3f points[4];
    points[0].setValue( d, d,0);
    points[1].setValue(-d, d,0);
    points[2].setValue(-d,-d,0);
    points[3].setValue( d,-d,0);
    d += step;

    SoCoordinate3* coordinate3 = new SoCoordinate3;
    int32_t pointn = 4;
    coordinate3->point.setValues(0,pointn,points);
    sceneGraph->addChild(coordinate3);
 
    HEPVis_SoMarkerSet* markerSet = new HEPVis_SoMarkerSet;
    markerSet->markerIndex.setValue(index);
    //markerSet->markerIndex.setValue(HEPVis_SoMarkerSet::CIRCLE_FILLED_5_5);
    markerSet->numPoints.setValue(pointn);
    sceneGraph->addChild(markerSet);

    //SoMarkerSet* markerSet = new SoMarkerSet;
    ////markerSet->markerIndex.setValue(SoMarkerSet::CIRCLE_FILLED_5_5);
    //markerSet->markerIndex.setValue(SoMarkerSet::TRIANGLE_FILLED_5_5);
    //markerSet->numPoints.setValue(pointn);
    //sceneGraph->addChild(markerSet);
  }

  if (args.isAnArgument("-write")) {
    SbBool altRep = args.isAnArgument("-altrep")?TRUE:FALSE;
    SbString file;
    args.find("-file",file);
    if(file=="") file = "out.iv";
    SoTools::write(*sceneGraph,file,"",FALSE,altRep);
  }

  SoEventCallback* eventCallback = new SoEventCallback;
  eventCallback->addEventCallback(SoKeyboardEvent::getClassTypeId(),
                                  keystroke,0);
  sceneGraph->addChild(eventCallback);                                  

  examinerViewer = new SoGuiExaminerViewer(topWidget);
  const SbViewportRegion& vpRegion = examinerViewer->getViewportRegion();
  SoGL2PSAction* action = new SoGL2PSAction(vpRegion);
#ifdef HEPVisWin
  HEPVis_menuBar_set
#endif
  action->setFileName("out.ps");
  examinerViewer->setGLRenderAction(action);
  examinerViewer->setSceneGraph(sceneGraph);
  examinerViewer->viewAll();
  examinerViewer->show();

  SoDebugError::postInfo("DetectorTreeKit",
    "type P key in picking mode to produce a postscript file.");

  SoGui::show(topWidget);
  SoGui::mainLoop();

  sceneGraph->unref();
  delete examinerViewer;

  return 0; 
}

void keystroke(void*,SoEventCallback* eventCB){
  if (eventCB->isHandled()) return;

  const SoKeyboardEvent* event = (SoKeyboardEvent*)eventCB->getEvent();

  if (SO_KEY_PRESS_EVENT(event,P)) {
    if(examinerViewer) {
      SoDebugError::postInfo("DetectorTreeKit","produce out.ps...");

      SoGL2PSAction* action = 
        (SoGL2PSAction*)examinerViewer->getGLRenderAction();
      action->enableFileWriting();
      examinerViewer->render();
      action->disableFileWriting();

      eventCB->setHandled();
    }
  }
}
