#include <math.h>

#include <stdlib.h>

#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/actions/SoWriteAction.h>

#include <HEPVis/SbMath.h>
#include <HEPVis/nodekits/SoMeterStickKit.h>

//
// Test of SoMeterStickKit class
// Joe Boudreau Nov. 2000
//
// This demo draws a Meter Stick.  You can pick it up and move it
// around.

int main(int, char ** argv) {

  // Set up the main window
 
  SoGuiWidget myWindow = SoGui::init(argv[0]);
  if (!myWindow) exit(1);

  // You must initialize a new class
  SoMeterStickKit::initClass();

  // Set up the root of the scene database and
  // reference it

  SoSeparator *root=new SoSeparator;
  root->ref();

  // Set up the material for the object to be drawn

  SoMaterial    *RedPlastic = new SoMaterial;

  RedPlastic->diffuseColor.setValue(1, 0, 0);
  root->addChild(RedPlastic);

  // Users will expect a new node to change appearance when a field
  // is changed.  Here, we connect one of the fields to an engine
  // which should animate the object if this feature is properly
  // implemented.

  // Make the MeterStickKit

  SoMeterStickKit *TheMeterStickKit   = new SoMeterStickKit();
  TheMeterStickKit->rotationSnap.setValue(TRUE);
  if (getenv("HEPVIS_ALTERNATEREP")||getenv("ALTERNATEREP")) TheMeterStickKit->generateAlternateRep(0);
  root->addChild(TheMeterStickKit);

  SoWriteAction *WriteAction= new SoWriteAction();
  WriteAction->getOutput()->setBinary(FALSE);
  WriteAction->apply(root);

  // View the new object in an Examiner Viewer

  SoGuiExaminerViewer *myViewer=new SoGuiExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  myViewer->viewAll();
  myViewer->show();

  SoGui::show(myWindow);
  SoGui::mainLoop();

  root->unref();
  delete WriteAction;
  delete myViewer;

#ifdef SoFreePackage //Cleanup
  SoMeterStickKit::clearClass();
  SoGui::clear();  
#endif

  return 0; 

}

