#include <HEPVis/Gui/SoGui.h>

// Inventor :
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoTransform.h>

// HEPVis :
#include <HEPVis/actions/SoGL2PSAction.h>
#include <HEPVis/nodekits/SoPage.h>
#include <HEPVis/nodekits/SoDisplayRegion.h>

#include <HEPVis/SbTestPlotter.h>
#include <HEPVis/nodekits/SoPlotterRegion.h>
#include <HEPVis/nodekits/SoAxis.h>
#include <HEPVis/nodekits/SoPlotter.h>

#include "SetRegion.icc"

extern "C" {
  void HEPVisUtils_initClasses();
  void HEPVisPlotter_initClasses();
}

//////////////////////////////////////////////////////////////////////////////
int main (
 int //aArgc
,char** aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SoGuiWidget myWindow = SoGui::init(aArgs[0]);
  if (myWindow==NULL) exit(1);

  HEPVisUtils_initClasses();
  HEPVisPlotter_initClasses();

  SoPage* page = new SoPage;
  page->ref();
  page->title.setValue("My display");
  page->deleteRegions();
  page->createRegions("SoDisplayRegion",2,2,0);

  SoCamera* camera = 0;
 {SoDisplayRegion* displayRegion = (SoDisplayRegion*)page->getRootRegion(0);
  SoSeparator* soStatic = displayRegion->getStaticScene();
  SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue(SbColor(1,0,0));
  soStatic->addChild(material);
  SoTransform* transform = new SoTransform;
  transform->rotation.setValue(SbRotation(SbVec3f(1,1,0),1.5));
  soStatic->addChild(transform);
  soStatic->addChild(new SoCylinder);
  camera = displayRegion->getCamera();}

 {SoDisplayRegion* displayRegion = (SoDisplayRegion*)page->getRootRegion(1);
  SoSeparator* soStatic = displayRegion->getStaticScene();
  SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue(SbColor(0,1,0));
  soStatic->addChild(material);
  SoTransform* transform = new SoTransform;
  transform->rotation.setValue(SbRotation(SbVec3f(1,1,0),1.5));
  soStatic->addChild(transform);
  soStatic->addChild(new SoCone);}

 {SoPlotterRegion* soPlotterRegion = 
    (SoPlotterRegion*)page->createGridRegion("SoPlotterRegion",2,2,2);
  SbTestHistogram1D* histo1D = new SbTestHistogram1D;
  SoPlotter* plotter = soPlotterRegion->getPlotter();
  plotter->setName("plotter_1");
  plotter->title.setValue("Title");
  plotter->setPlottableObject(new SbTestPlottableBins1D(*histo1D));
  //plotter->set("binsStyle[0]","modeling points color 0 0 1 markerStyle PLUS markerSize 9");
  plotter->getXAxis()->title.setValue("x axis title");
  plotter->getYAxis()->title.setValue("y axis title");
  plotter->getZAxis()->title.setValue("z axis title");}

  SoGuiExaminerViewer* examinerViewer = 
    new SoGuiExaminerViewer(myWindow);
  const SbViewportRegion& vpRegion = examinerViewer->getViewportRegion();
  SoGL2PSAction* action = new SoGL2PSAction(vpRegion);
  action->setFileName("out.ps");
  examinerViewer->setGLRenderAction(action);
  examinerViewer->setSize(SbVec2s(400,400));
  examinerViewer->setSceneGraph(page);
  examinerViewer->setCamera(camera);
  examinerViewer->viewAll();
  examinerViewer->show();
  examinerViewer->setBackgroundColor(SbColor(0.6F,0.6F,0.6F));
  //examinerViewer->setDecoration(FALSE);
  //examinerViewer->setViewing(FALSE);
  page->setSetRegionCallback(new SetRegionCallback(examinerViewer));

  SoGui::show(myWindow);
  SoGui::mainLoop();

  page->unref();
  delete examinerViewer;
  delete action;

  return EXIT_SUCCESS;
}
