#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/actions/SoWriteAction.h>

#include <HEPVis/nodes/SoPyramid.h>
#include <HEPVis/actions/SoAlternateRepAction.h>

int main(int, char ** argv) {

  SoGuiWidget topWidget = SoGui::init(argv[0]);
  if (!topWidget) exit(1);

  SoPyramid::initClass();
  SoAlternateRepAction::initClass();

  SoSeparator* sceneGraph = new SoSeparator;
  sceneGraph->ref();

  SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue(1, 0, 0);
  sceneGraph->addChild(material);

  // Have some animation :
  SoTimeCounter* timeCounter = new SoTimeCounter;
  timeCounter->min =  50; 
  timeCounter->max = 450; 
  timeCounter->frequency= 0.1F;
  
  // Make the G4Trd

  SoPyramid* pyramid = new SoPyramid();
  pyramid->backWidth = 900;
  pyramid->backHeight = 900;
  pyramid->frontWidth = 100;
  pyramid->frontHeight = 100;
  pyramid->depth = 200;
  //if (getenv("HEPVIS_ALTERNATEREP")||
  //    getenv("ALTERNATEREP"))pyramid->generateAlternateRep(0);
  pyramid->backWidth.connectFrom(&timeCounter->output);
  pyramid->backHeight.connectFrom(&timeCounter->output);
  sceneGraph->addChild(pyramid);

  SoWriteAction* writeAction= new SoWriteAction();
  writeAction->getOutput()->setBinary(FALSE);
  writeAction->apply(sceneGraph);

  // View the new object in an Examiner Viewer

  SoGuiExaminerViewer* viewer = new SoGuiExaminerViewer(topWidget);
  viewer->setSceneGraph(sceneGraph);
  viewer->viewAll();
  viewer->show();

  SoGui::show(topWidget);
  SoGui::mainLoop();

  sceneGraph->unref();
  delete writeAction;
  delete viewer;

#ifdef HEPVisSDL
  SoGui::clear();  
#endif

  return 0; 

}

