   
#include <SDL.h>

static void OpenGL_cube(unsigned int width,unsigned int height);

static void mainLoop(SDL_Surface& aScreen) {
  int done = 0;
  int thenTicks = -1;
    
  while ( !done ) {

    /* Check for events */
    SDL_Event event;
    while ( SDL_PollEvent (&event) ) {
      switch (event.type) {
      case SDL_MOUSEMOTION:
        break;
      case SDL_MOUSEBUTTONDOWN:
        break;
      case SDL_KEYDOWN:
        /* Any keypress quits the app... */
      case SDL_QUIT:
        done = 1;
        break;
      default:
        break;			}
    }
    
    // Draw at 24 hz
    //     This approach is not normally recommended - it is better to
    //     use time-based animation and run as fast as possible
    OpenGL_cube(aScreen.w,aScreen.h);
    SDL_GL_SwapBuffers ();

    // Time how long each draw-swap-delay cycle takes
    // and adjust delay to get closer to target framerate
    int fps = 24;
    int delay = 1000/fps;
    if (thenTicks > 0) {
      int nowTicks = SDL_GetTicks ();
      delay += (1000/fps - (nowTicks-thenTicks));
      thenTicks = nowTicks;
      if (delay < 0) delay = 1000/fps;
    } else {
      thenTicks = SDL_GetTicks ();
    }

    SDL_Delay (delay);
  }
}

int main(int /*argc*/, char* /*argv*/ []) {
  if (SDL_Init(SDL_INIT_VIDEO)<0) { // Init SDL video subsystem
    fprintf(stderr, "Couldn't initialize SDL: %s\n",SDL_GetError());
    return 1;
  }

  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,16);
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
    
  Uint32 flags = SDL_OPENGL;
  flags |= SDL_FULLSCREEN;

  SDL_Surface* screen = SDL_SetVideoMode(640,480,0,flags);  // Create window
  if (screen == NULL) {
    fprintf (stderr, "Couldn't set 640x480 OpenGL video mode: %s\n",
      SDL_GetError());
    SDL_Quit();
    return 1;
  }
    
  mainLoop(*screen);
    
  SDL_Quit();
	
  return 0;
}

#include <HEPVis/SbGL.h>

static void OpenGL_DrawCube();
//////////////////////////////////////////////////////////////////////////////
void OpenGL_cube(
 unsigned int width
,unsigned int height
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  /*
  std::string component = aUI.callbackData().component();
  unsigned int width,height;
  aUI.size(component,width,height);
  //printf("debug : %s : %d %d\n",component.c_str(),width,height);
  */

  glEnable(GL_DEPTH_TEST);
  //glEnable(GL_SCISSOR_TEST);

  glEnable(GL_LINE_STIPPLE);
  glPolygonMode(GL_FRONT,GL_FILL);
  //  So that lightning is correctly computed 
  // when some scaling is applied.
  glEnable(GL_NORMALIZE);

  glShadeModel(GL_FLAT);
  //glEnable(GL_LIGHTING);
  //glDisable(GL_LIGHT0);

  glViewport(0,0,width,height);
  //glScissor(0,0,width,height);

  // In OpenGL, by default,
  // camera is placed at origin, looking at -z, up along y.
  GLint viewport[4];
  glGetIntegerv(GL_VIEWPORT,viewport);
  double aspect = ((double)viewport[2])/((double)viewport[3]);
  glMatrixMode(GL_PROJECTION); 
  glLoadIdentity();

  double half_height = 1;
  glOrtho(-aspect * half_height,aspect*half_height,
          -half_height,half_height,-100,100);

  // Better to clear after glViewport and glScissor (else problems with Mesa).
  glClearColor(0.8F,0.8F,0.8F,0);
  glClear(GL_COLOR_BUFFER_BIT);
  glClear(GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  //glScaled(1,1,1);
  // Rotate then translate :
  glTranslated(0,0,-10);
  glRotated(30,0,1,0);
  glRotated(30,1,0,0);

  OpenGL_DrawCube();

  glFinish();
}
//////////////////////////////////////////////////////////////////////////////
static void OpenGL_DrawCube(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  glLineWidth(4);

  // Front :
  glColor3d(1,0,0);    //red
  glBegin(GL_POLYGON);
  glNormal3d(0,0,1);
  glVertex3d( 0.5,-0.5,0.5);
  glVertex3d( 0.5, 0.5,0.5);
  glVertex3d(-0.5, 0.5,0.5);
  glVertex3d(-0.5,-0.5,0.5);
  glEnd();

  // Back :
  glColor3d(0,1,0);   //green
  glBegin(GL_POLYGON);
  glNormal3d(0,0,-1);
  glVertex3d( 0.5,-0.5,-0.5);
  glVertex3d(-0.5,-0.5,-0.5);
  glVertex3d(-0.5, 0.5,-0.5);
  glVertex3d( 0.5, 0.5,-0.5);
  glEnd();

  glColor3d(1,1,0);   //yellow
  glBegin(GL_LINES);
  glVertex3d( 0.5,-0.5, 0.5);
  glVertex3d( 0.5,-0.5,-0.5);
  glVertex3d( 0.5, 0.5, 0.5);
  glVertex3d( 0.5, 0.5,-0.5);
  glVertex3d(-0.5, 0.5, 0.5);
  glVertex3d(-0.5, 0.5,-0.5);
  glVertex3d(-0.5,-0.5, 0.5);
  glVertex3d(-0.5,-0.5,-0.5);
  glEnd();
}
