//*****************************************************************************
//
// Test the spline track class
//
//*****************************************************************************

#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/engines/SoElapsedTime.h>

#include <HEPVis/nodes/SoSplineCurve.h>
#include <HEPVis/actions/SoAlternateRepAction.h>

#include <math.h>
#include <stdlib.h>

// 
// Ianna Osborne, Jan. 22, 2002
//
// This demo draws a spline track
//
const float FM_SQRT3_2 =  0.8660254037844F; // sqrt(3)/2 for sin(60 degrees)

int main( int, char ** argv )
{
  SoGuiWidget myWindow=SoGui::init(argv[0]);
  if (!myWindow) exit(1);
    
  SoSplineCurve::initClass();
  SoAlternateRepAction::initClass();

  SoSeparator* root = new SoSeparator;
  root->ref();

  // Set up the material for the object to be drawn

  SoMaterial* redPlastic = new SoMaterial;
  redPlastic->diffuseColor.setValue( 1, 0, 0 );
  root->addChild( redPlastic );

  SoSplineCurve* theTrackOne = new SoSplineCurve;
  SbVec3f defPointsOne[2] = { 
    SbVec3f (1,0,0), 
    SbVec3f (-0.5,FM_SQRT3_2,0)
  };
  SbVec3f defTangentsOne[2] = { 
    SbVec3f (0,1,0), 
    SbVec3f (-FM_SQRT3_2,-0.5F,0)
  };
  theTrackOne->points.setValues(0,2,defPointsOne);
  theTrackOne->tangents.setValues(0,2,defTangentsOne);
  root->addChild( theTrackOne );
 
  //  perfect planar helix
  SoSplineCurve* theTrackTwo = new SoSplineCurve;
  SbVec3f defPointsTwo[3] = { 
    SbVec3f (1,0,0), 
    SbVec3f (0,1,0), 
    SbVec3f (-1,0,0) 
  };
  SbVec3f defTangentsTwo[3] = { 
    SbVec3f (0,1,0), 
    SbVec3f (-1,0,0), 
    SbVec3f (0,-1,0) 
  };
  theTrackTwo->points.setValues( 0, 3, defPointsTwo );
  theTrackTwo->tangents.setValues( 0, 3, defTangentsTwo );
  root->addChild( theTrackTwo );
    
  //  reversed planar  helix
  SoSplineCurve* theTrackThree = new SoSplineCurve;
  SbVec3f defPointsThree[5] = { 
    SbVec3f (1,0,0), 
    SbVec3f (0,1,0), 
    SbVec3f (-1,0,0),
    SbVec3f (-2,-1,0), 
    SbVec3f (-3,0,0) };
  SbVec3f defTangentsThree[5] = { 
    SbVec3f (0,1,0), 
    SbVec3f (-1,0,0), 
    SbVec3f (0,-1,0),
    SbVec3f (-1,0,0), 
    SbVec3f (0,1,0) 
  };
  theTrackThree->points.setValues( 0, 5, defPointsThree );
  theTrackThree->tangents.setValues( 0, 5, defTangentsThree );
  root->addChild( theTrackThree );

  //  non-helix version   
  SoSplineCurve* theTrackFour = new SoSplineCurve;
  const float rtan = 0.982246F;
  const float ztan = 0.187595F;
  SbVec3f defPointsFour[5] = { 
    SbVec3f (1,0,0), 
    SbVec3f (0,1,0), 
    SbVec3f (-1,0,0.6F),
    SbVec3f (-2,-1,0.9F), 
    SbVec3f (-3,0,1.2F) 
  };
  SbVec3f defTangentsFour[5] = { 
    SbVec3f (0, rtan,ztan), 
    SbVec3f (-rtan,0,ztan),
    SbVec3f (0,-rtan,ztan), 
    SbVec3f (-rtan,0,ztan),
    SbVec3f (0, rtan,ztan) 
  };
  theTrackFour->points.setValues(0,5,defPointsFour);
  theTrackFour->tangents.setValues(0,5,defTangentsFour);
  root->addChild( theTrackFour );

  //  straight line
  SoSplineCurve* theTrackFive = new SoSplineCurve;
  SbVec3f defPointsFive[3] = { 
    SbVec3f (1,0,0), 
    SbVec3f (1.5,0,0), 
    SbVec3f (2,0,0) 
  };
  SbVec3f defTangentsFive[3] = { 
    SbVec3f (1,0,0), 
    SbVec3f (1,0,0), 
    SbVec3f (1,0,0) 
  };
  theTrackFive->points.setValues(0,3,defPointsFive);
  theTrackFive->tangents.setValues(0,3,defTangentsFive);
  root->addChild( theTrackFive );

  // View the new object in an Examiner Viewer
  SoGuiExaminerViewer* examinerViewer = new SoGuiExaminerViewer(myWindow);
  examinerViewer->setSize(SbVec2s(400,400));
  examinerViewer->setSceneGraph(root);
  examinerViewer->setTitle("HEPVis/SoSplineCurve");
  examinerViewer->show();
    
  SoGui::show(myWindow);
  SoGui::mainLoop();

  root->unref();
  delete examinerViewer;

  return 0;
}
