//
// To test / debug VRML2 production with nodekits.
//   Guy Barrand 12 Oct 2006.
//

#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>

#include <HEPVis/SbArguments.h>
#include <HEPVis/misc/SoTools.h>
#include <HEPVis/nodekits/SoTestKit.h>
#include <HEPVis/actions/SoAlternateRepAction.h>

#include <stdlib.h> //exit
#include <iostream>

//////////////////////////////////////////////////////////////////////////////
int main (
 int aArgc
,char** aArgv
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SbArguments args(aArgc,aArgv);

  if(args.isAnArgument("-h")) {
    std::cout << "HEPVis::TestKit :" << std::endl;
    std::cout << "Usage :" << std::endl;
    std::cout << " OS> TestKit<GUI> <options>" << std::endl;
    std::cout << "Options :" << std::endl;
    std::cout << " -h : to get help." << std::endl;
    std::cout << " -altrep : to produce an alternate representation." 
              << std::endl;
    std::cout << " -write : to write scene in a file (default is out.iv)." 
              << std::endl;
    std::cout << " -file : name of the file (with .iv or .wrl extension)." 
              << std::endl;
    std::cout << "For example :" << std::endl;
    std::cout << " OS> TestKitXt -altrep -write" << std::endl;
    std::cout << " UNIX> ivview out.iv" << std::endl;
    return 0;
  }

  // Set up the main window
  SoGuiWidget topWidget=SoGui::init(aArgv[0]);
  if (!topWidget) exit(1);

  // initialize HEPVis used classes :
  SoTestKit::initClass();
  SoAlternateRepAction::initClass();

  SoSeparator* sceneGraph = new SoSeparator;
  sceneGraph->ref();

  SoTestKit* axis = new SoTestKit;
  sceneGraph->addChild(axis);

  if (args.isAnArgument("-write")) {
    SbBool altRep = args.isAnArgument("-altrep")?TRUE:FALSE;
    SbString file;
    args.find("-file",file);
    if(file=="") file = "out.iv";
    SoTools::write(*sceneGraph,file,"",FALSE,altRep);
  }

  SoGuiExaminerViewer* examinerViewer = new SoGuiExaminerViewer(topWidget);
  examinerViewer->setBackgroundColor(SbColor(1,1,1));
  examinerViewer->setSceneGraph(sceneGraph);
  examinerViewer->show();

  SoGui::show(topWidget);
  SoGui::mainLoop();

  sceneGraph->unref();
  delete examinerViewer;

  return 0; 
}
