#include <math.h>

#include <HEPVis/Gui/SoGui.h>

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoCalculator.h>
#include <math.h>
#include <stdlib.h>
#include <HEPVis/nodes/SoSiWedgeStrips.h>
#include <HEPVis/nodekits/SoDblSiStripWedgeDetKit.h>
#include <HEPVis/nodes/SoWedge.h>
//
// Test of SoDblSiStripWedgeDetKit class
// Peter Hamel June 10, 1999
//

short Strip[8]  = {0, 51, 132, 231, 400, 551, 630, 700};
short ADC[8]    = {50,   80, 110,  70,  20,  34,  18,   5};
short Strip2[8] = {0, 54, 135, 296, 412, 530, 600, 700};
short ADC2[8]   = {53,   60,  94,  20,  19,  34,  53,  25};
 
int main(int, char ** argv) {

  // Set up the main window
 
  SoGuiWidget myWindow=SoGui::init(argv[0]);
  if (!myWindow) exit(1);

  // You must initialize a new class

  SoDblSiStripWedgeDetKit::initClass();
  SoSiWedgeStrips::initClass();
  SoWedge::initClass();

  // Set up the root of the scene database and
  // reference it

  SoSeparator *root=new SoSeparator;
  root->ref();

  // Set up the material for the object to be drawn

  SoDblSiStripWedgeDetKit *TheSiStripDet = new SoDblSiStripWedgeDetKit();

  // Set the size of the wedge

  TheSiStripDet->halfHeight.setValue(3);
  TheSiStripDet->halfWidth1.setValue(3);
  TheSiStripDet->halfWidth2.setValue(2);
  TheSiStripDet->thickness.setValue(0.03F);

  // Now set up an engine to make the wedge pulsate. This isn't intended to be
  // practical of course, but the hepvis documentation asked for an engine
  // of some sort.

  SoElapsedTime *myCounter = new SoElapsedTime;
  SoCalculator *myCalc = new SoCalculator;
  myCalc->a.connectFrom(&myCounter->timeOut);
  myCalc->expression.set1Value(0, "oa = 2.0+.5*sin(a)");
  TheSiStripDet->halfWidth1.connectFrom(&myCalc->oa);

  SoMaterial    *RedPlastic   = (SoMaterial *) TheSiStripDet->getPart("wedgeAppearance.material",TRUE);
  RedPlastic->diffuseColor.setValue(1, 0, 0);
  SoMaterial    *GreenPlastic = (SoMaterial *) TheSiStripDet->getPart("phiStripsAppearance.material",TRUE);
  GreenPlastic->diffuseColor.setValue(0,1,0);
  SoMaterial    *BluePlastic = (SoMaterial *) TheSiStripDet->getPart("zStripsAppearance.material",TRUE);
  BluePlastic->diffuseColor.setValue(0,1,1);

  SoSiWedgeStrips *ThePhiStrips = (SoSiWedgeStrips *) TheSiStripDet->getPart("phiStrips",TRUE);
  SoSiWedgeStrips *TheZStrips   = (SoSiWedgeStrips *) TheSiStripDet->getPart("zStrips",TRUE);
  ThePhiStrips->strip.setValues(0,8,Strip);
  ThePhiStrips->adc.setValues(0,8,ADC);
  ThePhiStrips->offset.setValue(0.1F);

  // Make the Z  Strips
  TheZStrips->strip.setValues(0,8,Strip2);
  TheZStrips->adc.setValues(0,8,ADC2);
  TheZStrips->offset.setValue(0.2F);

  root->addChild(TheSiStripDet);
  
  if (getenv("HEPVIS_ALTERNATEREP")||getenv("ALTERNATEREP")) TheSiStripDet->generateAlternateRep(0);

  SoWriteAction *WriteAction= new SoWriteAction();
  WriteAction->getOutput()->setBinary(FALSE);
  WriteAction->apply(root);

  // View the new object in an Examiner Viewer

  TheSiStripDet->setAlignable(TRUE);
  SoGuiExaminerViewer *myViewer=new SoGuiExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  myViewer->show();

  SoGui::show(myWindow);
  SoGui::mainLoop();

  return 0; 

}
