// this :
#include <HEPVis/SbColormap.h>

SbBaseColormap::SbBaseColormap(){}

SbBaseColormap::SbBaseColormap(const SbBaseColormap& aFrom)
:fValues(aFrom.fValues),fColors(aFrom.fColors){}

SbBaseColormap& SbBaseColormap::operator=(const SbBaseColormap& aFrom){ 
  fValues = aFrom.fValues;
  fColors = aFrom.fColors;
  return *this;
}
SbBaseColormap::~SbBaseColormap(){}

unsigned int SbBaseColormap::colorn() const {return fColors.size();}
unsigned int SbBaseColormap::valn() const {return fValues.size();}

SbColor SbBaseColormap::color(unsigned int a_index) const {
  unsigned int n = fColors.size();
  if(a_index>=n) return SbColor(0.5F,0.5F,0.5F);
  return fColors[a_index];
}
float SbBaseColormap::value(unsigned int a_index) const {
  unsigned int n = fValues.size();
  if(a_index>=n) return 0;
  return fValues[a_index];
}

void SbBaseColormap::setColors(SbColor(*aGet)(float),unsigned int aCelln) {
  fColors.resize(aCelln);
  if(!aCelln) return;
  float d = 1.0F/(aCelln-1);
  for(unsigned int index=0;index<aCelln;index++) {
    fColors[index] = aGet(d*index);
  }
}

#include <float.h>
#include <HEPVis/SbMath.h>

inline float TakeLog(float aVal){
  if(aVal<=0) {
    return -FLT_MAX;
  } else {
    return FLOG10(aVal);
  }
}

void SbBaseColormap::set_PAW_coloring() {
  unsigned int valn = fValues.size();
  if(valn==1) {
    fValues[0] = TakeLog(fValues[0]);
  } else if(valn>=2) {
    //CERN/PAW coloring :
    if(fValues[0]==0) fValues[0] = 10e-5F;
    float vmin = TakeLog(fValues[0]);
    float vmax = TakeLog(fValues[valn-1]);
    float dv = (vmax-vmin)/(valn-1);
    for(unsigned int count=0;count<valn;count++) {
      fValues[count] = vmin + dv * count;
    }
  }
}
