//  20/08/1998 : G.Barrand : creation.
//  30/01/2004 : L.Garnier : addition of methods for hatch parameters


// this :
#include <HEPVis/SbStyle.h>

//Inventor :
#include <Inventor/SbString.h>
#include <Inventor/SbPList.h>
#include <Inventor/errors/SoDebugError.h>

#include <HEPVis/SbString.h>
#include <HEPVis/SbColors.h>
#include <HEPVis/SbMath.h>

static SbBool stringToLinePattern(const SbString&,SbLinePattern&);
static SbMarkerStyle stringToMarkerStyle(const SbString&);
static SbAreaStyle stringToAreaStyle(const SbString&);
static SbPaintingPolicy stringToPaintingPolicy(const SbString&);
static SbHatchingPolicy stringToHatchingPolicy(const SbString&);
static SbProjectionType stringToProjectionType(const SbString&);

#define NoLimit (-1)

SbStyle::SbStyle()
:fColor(0,0,0)
,fHighlightColor(0,0,0)
,fLineWidth(1)
,fMarkerSize(1)
,fFontSize(10)
,fLinePattern(SbLinePattern_solid)
,fMarkerStyle(SbMarkerDot)
,fAreaStyle(SbAreaSolid)
,fModeling(SbModeling_boxes)
,fLightModel(SbLightModel_phong)
,fTickModeling(SbTickModeling_hippo)
,fEncoding(SbEncoding_none)
,fSmoothing(FALSE)
,fHinting(FALSE)
,fCut("")
,fPaintingPolicy(SbPaintingUniform)
,fHatchingPolicy(SbHatchingNone)
,fProjectionType(SbProjectionNone)
,fFontName("defaultFont")
,fMultiNodeLimit(NoLimit)
,fDivisions(510) //Same as ROOT/TAttAxis
,fRotationSteps(24) //Same as SbPolyhedron default.
,fTransparency(0)
,fSpacing(0.05F)
,fAngle(FM_PI/4)
,fScale(1)
,fOffset(0)
,fStripWidth(0)
,fVisible(TRUE)
,fBarOffset(0.25F)
,fBarWidth(0.5F)
,fEditable(FALSE)
,fAutomated(TRUE)
,fOptions("")
,fColorMapping("")
,fEnforced(FALSE)
{}

SbStyle::SbStyle(const SbStyle& aFrom)
:fColor(aFrom.fColor)
,fHighlightColor(aFrom.fHighlightColor)
,fLineWidth(aFrom.fLineWidth)
,fMarkerSize(aFrom.fMarkerSize)
,fFontSize(aFrom.fFontSize)
,fLinePattern(aFrom.fLinePattern)
,fMarkerStyle(aFrom.fMarkerStyle)
,fAreaStyle(aFrom.fAreaStyle)
,fModeling(aFrom.fModeling)
,fLightModel(aFrom.fLightModel)
,fTickModeling(aFrom.fTickModeling)
,fEncoding(aFrom.fEncoding)
,fSmoothing(aFrom.fSmoothing)
,fHinting(aFrom.fHinting)
,fCut(aFrom.fCut)
,fPaintingPolicy(aFrom.fPaintingPolicy)
,fHatchingPolicy(aFrom.fHatchingPolicy)
,fProjectionType(aFrom.fProjectionType)
,fFontName(aFrom.fFontName)
,fMultiNodeLimit(aFrom.fMultiNodeLimit)
,fDivisions(aFrom.fDivisions)
,fRotationSteps(aFrom.fRotationSteps)
,fTransparency(aFrom.fTransparency)
,fSpacing(aFrom.fSpacing)
,fAngle(aFrom.fAngle)
,fScale(aFrom.fScale)
,fOffset(aFrom.fOffset)
,fStripWidth(aFrom.fStripWidth)
,fVisible(aFrom.fVisible)
,fBarOffset(aFrom.fBarOffset)
,fBarWidth(aFrom.fBarWidth)
,fEditable(aFrom.fEditable)
,fAutomated(aFrom.fAutomated)
,fOptions(aFrom.fOptions)
,fColorMapping(aFrom.fColorMapping)
,fEnforced(aFrom.fEnforced)
{}

SbStyle::~SbStyle(){}

SbStyle& SbStyle::operator=(const SbStyle& aFrom){copy(aFrom);return *this;}

void SbStyle::copy(const SbStyle& aFrom){
  fColor = aFrom.fColor;
  fHighlightColor = aFrom.fHighlightColor;
  fLineWidth = aFrom.fLineWidth;
  fMarkerSize = aFrom.fMarkerSize;
  fFontSize = aFrom.fFontSize;
  fLinePattern = aFrom.fLinePattern;
  fMarkerStyle = aFrom.fMarkerStyle;
  fAreaStyle = aFrom.fAreaStyle;
  fModeling = aFrom.fModeling;
  fLightModel = aFrom.fLightModel;
  fTickModeling = aFrom.fTickModeling;
  fEncoding = aFrom.fEncoding;
  fSmoothing = aFrom.fSmoothing;
  fHinting = aFrom.fHinting;
  fCut = aFrom.fCut;
  fPaintingPolicy = aFrom.fPaintingPolicy;
  fHatchingPolicy = aFrom.fHatchingPolicy;
  fProjectionType = aFrom.fProjectionType;
  fFontName = aFrom.fFontName;
  fMultiNodeLimit = aFrom.fMultiNodeLimit;
  fDivisions = aFrom.fDivisions;
  fRotationSteps = aFrom.fRotationSteps;
  fTransparency = aFrom.fTransparency;
  fSpacing = aFrom.fSpacing;
  fAngle = aFrom.fAngle;
  fScale = aFrom.fScale;
  fOffset = aFrom.fOffset;
  fStripWidth = aFrom.fStripWidth;
  fVisible = aFrom.fVisible;
  fBarOffset = aFrom.fBarOffset;
  fBarWidth = aFrom.fBarWidth;
  fEditable = aFrom.fEditable;
  fAutomated = aFrom.fAutomated;
  fOptions = aFrom.fOptions;
  fColorMapping = aFrom.fColorMapping;
  fEnforced = aFrom.fEnforced;
}

SbBool SbStyle::isEqual(const SbStyle& aFrom){
#define IS_EQUAL(aField) \
  if(aField!=aFrom.aField) return FALSE;

/*
#define IS_EQUAL(aField) \
  if(aField!=aFrom.aField) {\
    SoDebugError::postInfo("SbStyle::isEqual","failed for %s",#aField);\
    return FALSE;\
  }*/

  IS_EQUAL(fLineWidth)
  IS_EQUAL(fMarkerSize)
  IS_EQUAL(fFontSize)
  IS_EQUAL(fLinePattern)
  IS_EQUAL(fMarkerStyle)
  IS_EQUAL(fAreaStyle)
  IS_EQUAL(fSmoothing)
  IS_EQUAL(fHinting)
  IS_EQUAL(fPaintingPolicy)
  IS_EQUAL(fHatchingPolicy)
  IS_EQUAL(fProjectionType)
  IS_EQUAL(fMultiNodeLimit)
  IS_EQUAL(fDivisions)
  IS_EQUAL(fRotationSteps)
  IS_EQUAL(fTransparency)
  IS_EQUAL(fSpacing)
  IS_EQUAL(fAngle)
  IS_EQUAL(fScale)
  IS_EQUAL(fOffset)
  IS_EQUAL(fStripWidth)
  IS_EQUAL(fVisible)
  IS_EQUAL(fBarOffset)
  IS_EQUAL(fBarWidth)  
  IS_EQUAL(fEditable)
  IS_EQUAL(fAutomated)
  IS_EQUAL(fEnforced)

  //SbColor
  IS_EQUAL(fColor)
  IS_EQUAL(fHighlightColor)

  //SbString
  IS_EQUAL(fModeling)
  IS_EQUAL(fLightModel)
  IS_EQUAL(fTickModeling)
  IS_EQUAL(fEncoding)
  IS_EQUAL(fCut)
  IS_EQUAL(fFontName)
  IS_EQUAL(fOptions)
  IS_EQUAL(fColorMapping)

  return TRUE;
}

void SbStyle::reset(){
  fColor = SbColor_grey;
  fHighlightColor = SbColor_lightgrey;
  fLineWidth = 1;
  fMarkerSize = 1;
  fFontSize = 10;
  fLinePattern = SbLinePattern_solid;
  fMarkerStyle = SbMarkerDot;
  fAreaStyle = SbAreaSolid;
  fModeling = SbModeling_boxes;
  fLightModel = SbLightModel_phong;
  fTickModeling = SbTickModeling_hippo;
  fEncoding = SbEncoding_none;
  fSmoothing = FALSE;
  fHinting = FALSE;
  fCut = "";
  fPaintingPolicy = SbPaintingUniform;
  fHatchingPolicy = SbHatchingNone;
  fProjectionType = SbProjectionNone;
  fFontName = "defaultFont";
  fMultiNodeLimit = NoLimit;
  fTransparency = 0;
  fDivisions = 510;
  fRotationSteps = 24;
  fSpacing = 0.05F;
  fAngle = FM_PI/4;
  fScale = 1;
  fOffset = 0;
  fStripWidth = 0;
  fVisible = TRUE;
  fBarOffset = 0.25F;
  fBarWidth = 0.5F;  
  fEditable = FALSE;
  fAutomated = TRUE;
  fOptions = "";
  fColorMapping = "";
  fEnforced = FALSE;
}

void SbStyle::dump(){
  SoDebugError::postInfo("SbStyle::dump","\n\
color : %g %g %g\n\
highlightColor : %g %g %g\n\
lineWidth : %d\n\
markerSize : %d\n\
fontSize : %d\n\
linePattern : %x\n\
markerStyle : \"%s\"\n\
areaStyle : \"%s\"\n\
modeling : \"%s\"\n\
lightModel : \"%s\"\n\
tickModeling : \"%s\"\n\
encoding : \"%s\"\n\
smoothing : %d\n\
hinting : %d\n\
cut : \"%s\"\n\
paintingPolicy : \"%s\"\n\
hatchingPolicy : \"%s\"\n\
projectionType : \"%s\"\n\
fontName : \"%s\"\n\
multiNodeLimit : %d\n\
transparency : %g\n\
divisions: %d\n\
rotationSteps: %d\n\
angle: %g\n\
scale: %g\n\
offset: %g\n\
stripWidth: %g\n\
spacing: %g\n\
visible: %d\n\
barOffset: %g\n\
barWidth: %g\n\
editable: %d\n\
automated: %d\n\
options: %s\n\
colorMapping: %s\n\
enforced: %d\n\
"
,fColor[0],fColor[1],fColor[2]
,fHighlightColor[0],fHighlightColor[1],fHighlightColor[2]
,fLineWidth
,fMarkerSize
,fFontSize
,fLinePattern
,markerStyleToString(fMarkerStyle)
,areaStyleToString(fAreaStyle)
,fModeling.getString()
,fLightModel.getString()
,fTickModeling.getString()
,fEncoding.getString()
,fSmoothing
,fHinting
,fCut.getString()
,paintingPolicyToString(fPaintingPolicy)
,hatchingPolicyToString(fHatchingPolicy)
,projectionTypeToString(fProjectionType)
,fFontName.getString()
,fMultiNodeLimit
,fTransparency
,fDivisions
,fRotationSteps
,fAngle
,fScale
,fOffset
,fStripWidth
,fSpacing
,fVisible
,fBarOffset
,fBarWidth
,fEditable
,fAutomated
,fOptions.getString()
,fColorMapping.getString()
,fEnforced
);
}

#define CHECK_2 \
      if(wordn!=2) {\
        SoDebugError::postInfo("SbStyle::setFromString",\
          "%s: %s has a bad word count. Two expected.\n",\
          aString.getString(),line.getString());\
        SbStringDelete(words);\
        SbStringDelete(lines);\
        return FALSE;\
      }

#define WORD1_bool \
      SbString& word1 = *((SbString*)words[1]);\
      SbBool value;\
      if(SbStringConvertToBool(word1,value)==FALSE)  {\
        SoDebugError::postInfo("SbStyle::setFromString",\
          "%s: %s not a boolean.\n",aString.getString(),word1.getString());\
        SbStringDelete(words);\
        SbStringDelete(lines);\
        return FALSE;\
      }

#define WORD1_int \
      SbString& word1 = *((SbString*)words[1]);\
      int value;\
      if(SbStringConvertToInt(word1,value)==FALSE)  {\
        SoDebugError::postInfo("SbStyle::setFromString",\
          "%s: %s not a number.\n",aString.getString(),word1.getString());\
        SbStringDelete(words);\
        SbStringDelete(lines);\
        return FALSE;\
      }

#define WORD1_float \
      SbString& word1 = *((SbString*)words[1]);\
      float value;\
      if(SbStringConvertToFloat(word1,value)==FALSE)  {\
        SoDebugError::postInfo("SbStyle::setFromString",\
          "%s: %s not a number.\n",aString.getString(),word1.getString());\
        SbStringDelete(words);\
        SbStringDelete(lines);\
        return FALSE;\
      }

SbBool SbStyle::setFromString(const SbString& aString){
  // aString = list of "name value" separated by \n
  //printf("debug : SbStyle::setFromString : \"%s\"\n",aString.getString());
  SbPList lines = SbStringGetLines(aString);
  int linen = lines.getLength();
  for(int count=0;count<linen;count++) {
    SbString& line = *((SbString*)lines[count]);
    if(!line.getLength()) continue;
    if(line=="setDefaults") {
      reset();
      continue;
    } else if(line=="reset") {
      reset();
      continue;
    }
    SbPList words = SbStringGetWords(line," ");
    int wordn = words.getLength();
    if(wordn<=0) {
      SoDebugError::postInfo("SbStyle::setFromString",
        "%s: %s has a bad word count (at least two expected).\n",
        aString.getString(),line.getString());
      SbStringDelete(words);
      SbStringDelete(lines);
      return FALSE;
    }
    SbString& word0 = *((SbString*)words[0]);
    if(word0=="color") {
      if(wordn==2) {
        SbString& word1 = *((SbString*)words[1]);
        if(stringToColor(word1,fColor)==FALSE) {
          SoDebugError::postInfo("SbStyle::setFromString",
            "%s: %s not a color.\n",aString.getString(),word1.getString());
          SbStringDelete(words);
          SbStringDelete(lines);
          return FALSE;
        } else if( (word1.getString()[0]=='#') && word1.getLength()==7) {
          // #RRGGBB format :
          //  1 3 5
          unsigned long rr,gg,bb;
         {SbString ss("0x");
          ss += word1.getString()[1];
          ss += word1.getString()[2];
          if(!SbStringConvertToUnsignedLong(ss,rr)) {
            SoDebugError::postInfo("SbStyle::setFromString",
              "%s: %s not a number.\n",aString.getString(),ss.getString());
            SbStringDelete(words);
            SbStringDelete(lines);
            return FALSE;
          }}
         {SbString ss("0x");
          ss += word1.getString()[3];
          ss += word1.getString()[4];
          if(!SbStringConvertToUnsignedLong(ss,gg)) {
            SoDebugError::postInfo("SbStyle::setFromString",
              "%s: %s not a number.\n",aString.getString(),ss.getString());
            SbStringDelete(words);
            SbStringDelete(lines);
            return FALSE;
          }}
         {SbString ss("0x");
          ss += word1.getString()[5];
          ss += word1.getString()[6];
          if(!SbStringConvertToUnsignedLong(ss,bb)) {
            SoDebugError::postInfo("SbStyle::setFromString",
              "%s: %s not a number.\n",aString.getString(),ss.getString());
            SbStringDelete(words);
            SbStringDelete(lines);
            return FALSE;
          }}
          fColor.setValue(((float)rr)/255,((float)gg)/255,((float)bb)/255);
        } 
      } else if (wordn==4) {
        SbString& rs = *((SbString*)words[1]);
        float r;
        if(SbStringConvertToFloat(rs,r)==FALSE)  {
          SoDebugError::postInfo("SbStyle::setFromString",
            "%s: %s not a number.\n",aString.getString(),rs.getString());
          SbStringDelete(words);
          SbStringDelete(lines);
          return FALSE;
        }
        SbString& gs = *((SbString*)words[2]);
        float g;
        if(SbStringConvertToFloat(gs,g)==FALSE)  {
          SoDebugError::postInfo("SbStyle::setFromString",
            "%s: %s not a number.\n",aString.getString(),gs.getString());
          SbStringDelete(words);
          SbStringDelete(lines);
          return FALSE;
        }
        SbString& bs = *((SbString*)words[3]);
        float b;
        if(SbStringConvertToFloat(bs,b)==FALSE)  {
          SoDebugError::postInfo("SbStyle::setFromString",
            "%s: %s not a number.\n",aString.getString(),bs.getString());
          SbStringDelete(words);
          SbStringDelete(lines);
          return FALSE;
        }
        fColor.setValue(r,g,b);
      } else {
        SoDebugError::postInfo("SbStyle::setFromString",
          "%s: %s has a bad word count (two or four expected).\n",
          aString.getString(),line.getString());
        SbStringDelete(words);
        SbStringDelete(lines);
        return FALSE;
      }
    } else if(word0=="highlightColor") {
      if(wordn==2) {
        SbString& word1 = *((SbString*)words[1]);
        if(stringToColor(word1,fHighlightColor)==FALSE) {
          SoDebugError::postInfo("SbStyle::setFromString",
            "%s: %s not a color.\n",aString.getString(),word1.getString());
          SbStringDelete(words);
          SbStringDelete(lines);
          return FALSE;
        } else if( (word1.getString()[0]=='#') && word1.getLength()==7) {
          // #RRGGBB format :
          //  1 3 5
          unsigned long rr,gg,bb;
         {SbString ss("0x");
          ss += word1.getString()[1];
          ss += word1.getString()[2];
          if(!SbStringConvertToUnsignedLong(ss,rr)) {
            SoDebugError::postInfo("SbStyle::setFromString",
              "%s: %s not a number.\n",aString.getString(),ss.getString());
            SbStringDelete(words);
            SbStringDelete(lines);
            return FALSE;
          }}
         {SbString ss("0x");
          ss += word1.getString()[3];
          ss += word1.getString()[4];
          if(!SbStringConvertToUnsignedLong(ss,gg)) {
            SoDebugError::postInfo("SbStyle::setFromString",
              "%s: %s not a number.\n",aString.getString(),ss.getString());
            SbStringDelete(words);
            SbStringDelete(lines);
            return FALSE;
          }}
         {SbString ss("0x");
          ss += word1.getString()[5];
          ss += word1.getString()[6];
          if(!SbStringConvertToUnsignedLong(ss,bb)) {
            SoDebugError::postInfo("SbStyle::setFromString",
              "%s: %s not a number.\n",aString.getString(),ss.getString());
            SbStringDelete(words);
            SbStringDelete(lines);
            return FALSE;
          }}
          fHighlightColor.setValue
            (((float)rr)/255,((float)gg)/255,((float)bb)/255);
        } 
      } else if (wordn==4) {
        SbString& rs = *((SbString*)words[1]);
        float r;
        if(SbStringConvertToFloat(rs,r)==FALSE)  {
          SoDebugError::postInfo("SbStyle::setFromString",
            "%s: %s not a number.\n",aString.getString(),rs.getString());
          SbStringDelete(words);
          SbStringDelete(lines);
          return FALSE;
        }
        SbString& gs = *((SbString*)words[2]);
        float g;
        if(SbStringConvertToFloat(gs,g)==FALSE)  {
          SoDebugError::postInfo("SbStyle::setFromString",
            "%s: %s not a number.\n",aString.getString(),gs.getString());
          SbStringDelete(words);
          SbStringDelete(lines);
          return FALSE;
        }
        SbString& bs = *((SbString*)words[3]);
        float b;
        if(SbStringConvertToFloat(bs,b)==FALSE)  {
          SoDebugError::postInfo("SbStyle::setFromString",
            "%s: %s not a number.\n",aString.getString(),bs.getString());
          SbStringDelete(words);
          SbStringDelete(lines);
          return FALSE;
        }
        fHighlightColor.setValue(r,g,b);
      } else {
        SoDebugError::postInfo("SbStyle::setFromString",
          "%s: %s has a bad word count (two or four expected).\n",
          aString.getString(),line.getString());
        SbStringDelete(words);
        SbStringDelete(lines);
        return FALSE;
      }
    } else if( (word0=="linePattern") || (word0=="lineStyle") ) {
      CHECK_2
      SbString& word1 = *((SbString*)words[1]);
      unsigned long ul;
      if(stringToLinePattern(word1,fLinePattern)==FALSE) {
        if(SbStringConvertToUnsignedLong(word1,ul)==FALSE) {
          SoDebugError::postInfo("SbStyle::setFromString",
            "%s: %s not a line pattern.\n",
          aString.getString(),word1.getString());
          fLinePattern = SbLinePattern_solid;
          SbStringDelete(words);
          SbStringDelete(lines);
          return FALSE;
        }
        fLinePattern = (SbLinePattern)ul;
      }
    } else if(word0=="markerStyle") {
      CHECK_2
      SbString& word1 = *((SbString*)words[1]);
      fMarkerStyle = stringToMarkerStyle(word1);
    } else if(word0=="areaStyle") {
      CHECK_2
      SbString& word1 = *((SbString*)words[1]);
      fAreaStyle = stringToAreaStyle(word1);
    } else if(word0=="modeling") {
      CHECK_2
      SbString& word1 = *((SbString*)words[1]);
      fModeling = word1;
    } else if(word0=="lightModel") {
      CHECK_2
      SbString& word1 = *((SbString*)words[1]);
      fLightModel = word1;
    } else if(word0=="tickModeling") {
      CHECK_2
      SbString& word1 = *((SbString*)words[1]);
      fTickModeling = word1;
    } else if(word0=="encoding") {
      CHECK_2
      SbString& word1 = *((SbString*)words[1]);
      fEncoding = word1;
    } else if(word0=="smoothing") {
      CHECK_2
      WORD1_bool
      fSmoothing = value;
    } else if(word0=="hinting") {
      CHECK_2
      WORD1_bool
      fHinting = value;
    } else if(word0=="enforced") {
      CHECK_2
      WORD1_bool
      fEnforced = value;
    } else if(word0=="cut") {
      fCut = "";
      for(int wordi=1;wordi<wordn;wordi++) {
        if(wordi!=1) fCut += " ";
        fCut += *((SbString*)words[wordi]);
      }
    } else if(word0=="options") {
      fOptions = "";
      for(int wordi=1;wordi<wordn;wordi++) {
        if(wordi!=1) fOptions += " ";
        fOptions += *((SbString*)words[wordi]);
      }
    } else if(word0=="colorMapping") {
      fColorMapping = "";
      for(int wordi=1;wordi<wordn;wordi++) {
        if(wordi!=1) fColorMapping += " ";
        fColorMapping += *((SbString*)words[wordi]);
      }
    } else if(word0=="painting") {
      CHECK_2
      SbString& word1 = *((SbString*)words[1]);
      fPaintingPolicy = stringToPaintingPolicy(word1);
    } else if(word0=="hatching") {
      CHECK_2
      SbString& word1 = *((SbString*)words[1]);
      fHatchingPolicy = stringToHatchingPolicy(word1);
    } else if(word0=="projection") {
      CHECK_2
      SbString& word1 = *((SbString*)words[1]);
      fProjectionType = stringToProjectionType(word1);
    } else if(word0=="fontName") {
      CHECK_2
      SbString& word1 = *((SbString*)words[1]);
      setFontName(word1);
    } else if(word0=="lineWidth") {
      CHECK_2
      WORD1_int
      fLineWidth = value;
    } else if(word0=="markerSize") {
      CHECK_2
      WORD1_int
      fMarkerSize = value;
    } else if(word0=="fontSize") {
      CHECK_2
      WORD1_int
      fFontSize = value;
    } else if(word0=="transparency") {
      CHECK_2
      WORD1_float
      if(value<0.F) value = 0.F;
      fTransparency = value;
    } else if(word0=="multiNodeLimit") {
      CHECK_2
      WORD1_int
      fMultiNodeLimit = value;
    } else if(word0=="divisions") {
      CHECK_2
      WORD1_int
      fDivisions = value;
    } else if(word0=="rotationSteps") {
      CHECK_2
      WORD1_int
      fRotationSteps = value;
    } else if(word0=="angle") {
      CHECK_2
      WORD1_float
      fAngle = value;
    } else if(word0=="scale") {
      CHECK_2
      WORD1_float
      fScale = value;
    } else if(word0=="offset") {
      CHECK_2
      WORD1_float
      fOffset = value;
    } else if(word0=="stripWidth") {
      CHECK_2
      WORD1_float
      fStripWidth = value;
    } else if(word0=="spacing") {
      CHECK_2
      WORD1_float
      fSpacing = value;
    } else if(word0=="visible") {
      CHECK_2
      WORD1_bool
      fVisible = value;
    } else if(word0=="editable") {
      CHECK_2
      WORD1_bool
      fEditable = value;
    } else if(word0=="automated") {
      CHECK_2
      WORD1_bool
      fAutomated = value;
    } else if(word0=="barOffset") {
      CHECK_2
      WORD1_float
      fBarOffset = value;
    } else if(word0=="barWidth") {
      CHECK_2
      WORD1_float
      fBarWidth = value;
    } else {
      SoDebugError::postInfo("SbStyle::setFromString",
        "SbStyle::setFromString: bad option %s.\n",word0.getString());
      SbStringDelete(words);
      SbStringDelete(lines);
      return FALSE;
    }      
    SbStringDelete(words);
  } 
  SbStringDelete(lines);
  return TRUE;
}

SbString SbStyle::getString() const{
  char ss[23 * 32];
  ::sprintf(ss,"\
color %g %g %g\n\
highlightColor %g %g %g\n\
lineWidth %d\n\
markerSize %d\n\
fontSize %d\n\
linePattern %x\n\
multiNodeLimit %d\n\
transparency %g\n\
divisions %d\n\
rotationSteps %d\n\
angle %g\n\
scale %g\n\
offset %g\n\
stripWidth %g\n\
spacing %g\n\
barOffset %g\n\
barWidth %g"
,fColor[0],fColor[1],fColor[2]
,fHighlightColor[0],fHighlightColor[1],fHighlightColor[2]
,fLineWidth
,fMarkerSize
,fFontSize
,fLinePattern
,fMultiNodeLimit
,fTransparency
,fDivisions
,fRotationSteps
,fAngle
,fScale
,fOffset
,fStripWidth
,fSpacing
,fBarOffset
,fBarWidth);

  SbString lf("\n");

  SbString s(ss);  

  s += lf;  
  s += "smoothing ";
  s += (fSmoothing?"true":"false");

  s += lf;  
  s += "hinting ";
  s += (fHinting?"true":"false");

  s += lf;  
  s += "enforced ";
  s += (fEnforced?"true":"false");

  s += lf;  
  s += "visible ";
  s += (fVisible?"true":"false");

  s += lf;  
  s += "editable ";
  s += (fEditable?"true":"false");

  s += lf;  
  s += "automated ";
  s += (fAutomated?"true":"false");

  s += lf;  
  s += "markerStyle ";
  s += markerStyleToString(fMarkerStyle);

  s += lf;  
  s += "areaStyle ";
  s += areaStyleToString(fAreaStyle);

  s += lf;
  s += "modeling ";
  s += fModeling;

  s += lf;
  s += "lightModel ";
  s += fLightModel;

  s += lf;
  s += "tickModeling ";
  s += fTickModeling;

  s += lf;
  s += "encoding ";
  s += fEncoding;

  s += lf;
  s += "cut ";
  s += fCut;

  s += lf;
  s += "painting ";
  s += paintingPolicyToString(fPaintingPolicy);

  s += lf;
  s += "hatching ";
  s += hatchingPolicyToString(fHatchingPolicy);

  s += lf;
  s += "projection ";
  s += projectionTypeToString(fProjectionType);

  s += lf;
  s += "fontName ";
  s += fFontName;

  s += lf;
  s += "options ";
  s += fOptions;

  s += lf;
  s += "colorMapping ";
  s += fColorMapping;

  return s;
}

void SbStyle::setColor(const SbColor& aValue){fColor = aValue;}
SbColor SbStyle::getColor() const {return fColor;}

void SbStyle::setHighlightColor(const SbColor& aValue){fHighlightColor = aValue;}
SbColor SbStyle::getHighlightColor() const {return fHighlightColor;}

void SbStyle::setMultiNodeLimit(int aValue){fMultiNodeLimit = aValue;}
int SbStyle::getMultiNodeLimit() const {return fMultiNodeLimit;}

void SbStyle::setModeling(const SbString& aValue){fModeling = aValue;}
const SbString& SbStyle::getModeling() const {return fModeling;}

void SbStyle::setLightModel(const SbString& aValue){fLightModel = aValue;}
const SbString& SbStyle::getLightModel() const {return fLightModel;}

void SbStyle::setTickModeling(const SbString& aValue){fTickModeling = aValue;}
const SbString& SbStyle::getTickModeling() const {return fTickModeling;}

void SbStyle::setEncoding(const SbString& aValue){fEncoding = aValue;}
const SbString& SbStyle::getEncoding() const {return fEncoding;}

void SbStyle::setSmoothing(SbBool aValue){fSmoothing = aValue;}
SbBool SbStyle::getSmoothing() const {return fSmoothing;}

void SbStyle::setHinting(SbBool aValue){fHinting = aValue;}
SbBool SbStyle::getHinting() const {return fHinting;}

SbPaintingPolicy SbStyle::getPaintingPolicy() const {return fPaintingPolicy;}
void SbStyle::setPaintingPolicy(SbPaintingPolicy aValue){fPaintingPolicy = aValue;}

SbHatchingPolicy SbStyle::getHatchingPolicy() const {return fHatchingPolicy;}
void SbStyle::setHatchingPolicy(SbHatchingPolicy aValue) {fHatchingPolicy = aValue;}

SbProjectionType SbStyle::getProjectionType() const {return fProjectionType;}
void SbStyle::setProjectionType(SbProjectionType aValue){fProjectionType = aValue;}

void SbStyle::setLinePattern(SbLinePattern aValue){fLinePattern = aValue;}
SbLinePattern SbStyle::getLinePattern() const {return fLinePattern;}

void SbStyle::setLineWidth(int aValue){fLineWidth = aValue;}
int SbStyle::getLineWidth() const {return fLineWidth;}

void SbStyle::setMarkerStyle(SbMarkerStyle aValue){fMarkerStyle = aValue;}
SbMarkerStyle SbStyle::getMarkerStyle() const {return fMarkerStyle;}

void SbStyle::setMarkerSize(int aValue){fMarkerSize = aValue;}
int SbStyle::getMarkerSize() const {return fMarkerSize;}

void SbStyle::setFontSize(int aValue){fFontSize = aValue;}
int SbStyle::getFontSize() const {return fFontSize;}

float SbStyle::getTransparency() const {return fTransparency;}
void SbStyle::setTransparency(float aValue) {fTransparency = aValue;}

void SbStyle::setAreaStyle(SbAreaStyle aValue){fAreaStyle = aValue;}
SbAreaStyle SbStyle::getAreaStyle() const {return fAreaStyle;}

void SbStyle::setFontName(const SbString& aValue){fFontName = aValue;}
const SbString& SbStyle::getFontName() const {return fFontName;}

void SbStyle::setDivisions(int aValue){fDivisions = aValue;}
int SbStyle::getDivisions() const {return fDivisions;}

void SbStyle::setRotationSteps(int aValue){fRotationSteps = aValue;}
int SbStyle::getRotationSteps() const {return fRotationSteps;}

void SbStyle::setOffset(float aValue){fOffset = aValue;}
float SbStyle::getOffset() const {return fOffset;}

void SbStyle::setStripWidth(float aValue){fStripWidth = aValue;}
float SbStyle::getStripWidth() const {return fStripWidth;}

void SbStyle::setAngle(float aValue){fAngle = aValue;}
float SbStyle::getAngle() const {return fAngle;}

void SbStyle::setScale(float aValue){fScale = aValue;}
float SbStyle::getScale() const {return fScale;}

void SbStyle::setBarOffset(float aValue){fBarOffset = aValue;}
float SbStyle::getBarOffset() const {return fBarOffset;}

void SbStyle::setBarWidth(float aValue){fBarWidth = aValue;}
float SbStyle::getBarWidth() const {return fBarWidth;}

void SbStyle::setSpacing(float aValue){fSpacing = aValue;}
float SbStyle::getSpacing() const {return fSpacing;}

void SbStyle::setVisible(SbBool aValue){fVisible = aValue;}
SbBool SbStyle::getVisible() const{return fVisible;}

void SbStyle::setEditable(SbBool aValue){fEditable = aValue;}
SbBool SbStyle::getEditable() const{return fEditable;}

void SbStyle::setAutomated(SbBool aValue){fAutomated = aValue;}
SbBool SbStyle::getAutomated() const{return fAutomated;}

void SbStyle::setCut(const SbString& aValue){fCut = aValue;}
SbString SbStyle::getCut() const {return fCut;}

void SbStyle::setOptions(const SbString& aValue){fOptions = aValue;}
SbString SbStyle::getOptions() const{return fOptions;}

void SbStyle::setColorMapping(const SbString& aValue){fColorMapping = aValue;}
SbString SbStyle::getColorMapping() const{return fColorMapping;}

void SbStyle::setEnforced(SbBool aValue){fEnforced = aValue;}
SbBool SbStyle::getEnforced() const{return fEnforced;}

void SbStyle::setByValue(const SbString& aString,std::vector<float>& aValues,std::vector<SbColor>& aColors){
  //  The given string is of the format :
  //      [<color name> <value>] <color name>
  //  or :
  //      [<value> <color name>] <value>
  //  For example :
  //      black 10 cyan 50 green 100 orange 200 blue 300 pink 400 red
  SbPList words = SbStringGetWords(aString," ");
  int wordn = words.getLength();
  int number = wordn/2;
  if(number<=0) {
    aValues.clear();
    aColors.clear();
  } else if((2*number+1)!=wordn) {
    SoDebugError::postInfo("SbStyle::getValueColorMapping",
      "An odd number (%d given) of words is expected in :\n%s\n",
      wordn,aString.getString());
    aValues.clear();
    aColors.clear();
  } else {
    // look if :
    //   <col> <num> <col> ... <num> <col>
    // or :
    //   <num> <col> <num> ... <col> <num>
    // FIXME : case of <col> being three floats ?
    SbString* word = 0;
    word = ((SbString*)words[0]);
    SbColor c;
    if(stringToColor(*word,c)==TRUE) {
      // <col> <num> <col> ... <num> <col>
      aValues.resize(number);
      aColors.resize(number+1);
      for(int count=0;count<number;count++) {
        word = ((SbString*)words[2*count]);
        if(stringToColor(*word,aColors[count])==FALSE) {
          SoDebugError::postInfo("SbStyle::getValueColorMapping",
            "%s: %s not a color.\n",aString.getString(),word->getString());
          SbStringDelete(words);
          aValues.clear();
          aColors.clear();
          return;
        }
        word = ((SbString*)words[2*count+1]);
        if(SbStringConvertToFloat(*word,aValues[count])==FALSE)  {
          SoDebugError::postInfo("SbStyle::getValueColorMapping",
            "%s: %s not a number.\n",aString.getString(),word->getString());
          SbStringDelete(words);
          aValues.clear();
          aColors.clear();
          return;
        }
      }
      word = ((SbString*)words[wordn-1]);
      if(stringToColor(*word,aColors[number])==FALSE) {
        SoDebugError::postInfo("SbStyle::getValueColorMapping",
          "%s: %s not a color.\n",aString.getString(),word->getString());
        SbStringDelete(words);
        aValues.clear();
        aColors.clear();
      }
    } else {
      // <num> <col> <num> ... <col> <num>
      aValues.resize(number+1);
      aColors.resize(number);
      for(int count=0;count<number;count++) {
        word = ((SbString*)words[2*count]);
        if(SbStringConvertToFloat(*word,aValues[count])==FALSE)  {
          SoDebugError::postInfo("SbStyle::getValueColorMapping",
            "%s: %s not a number.\n",aString.getString(),word->getString());
          SbStringDelete(words);
          aValues.clear();
          aColors.clear();
          return;
        }
        word = ((SbString*)words[2*count+1]);
        if(stringToColor(*word,aColors[count])==FALSE) {
          SoDebugError::postInfo("SbStyle::getValueColorMapping",
            "%s: %s not a color.\n",aString.getString(),word->getString());
          SbStringDelete(words);
          aValues.clear();
          aColors.clear();
          return;
        }
      }
      word = ((SbString*)words[wordn-1]);
      if(SbStringConvertToFloat(*word,aValues[number])==FALSE)  {
        SoDebugError::postInfo("SbStyle::getValueColorMapping",
          "%s: %s not a number.\n",aString.getString(),word->getString());
        SbStringDelete(words);
        aValues.clear();
        aColors.clear();
        return;
      }
    }
  } 
  SbStringDelete(words);
}

SbColor SbStyle::getByValue(float aValue,const std::vector<float>& aValues,const std::vector<SbColor>& aColors){
  // There are aValuen (n) entries in aValues and (n+1) aColors
  //  aColors[0] aValues[0] aColors[1] aValues[1]...
  //                 aValues[n-2] aColors[n-1] aValues[n-1] aColors[n]
  //      black 10 cyan 50 green 100 orange 200 blue 300 pink 400 red
  // valuen = 6
  // values[0] 10 
  // values[1] 50 
  // values[2] 100 
  // values[3] 200 
  // values[4] 300 
  // values[5] 400 
  //
  // colors[0] black
  // colors[1] cyan
  // colors[2] green
  // colors[3] orange
  // colors[4] blue
  // colors[5] pink
  // colors[6] red
  unsigned valn = aValues.size();
  if(!valn) return SbColor_black;
  if(aColors.size()==(valn+1)) {
    // col0 val0 col1 val1 col2 val2 col3
    if(aValue<aValues[0]) {
      return aColors[0];
    } else {
      for(unsigned int count=0;count<=(valn-2);count++) {
        if( (aValues[count]<=aValue) && (aValue<aValues[count+1]) ) {
          return aColors[count+1];
        }
      }
      return aColors[valn];
    }
  } else if((aColors.size()+1)==valn) {
    // val0 col0 val1 col1 val2 col2 val3
    for(unsigned int count=0;count<=(valn-2);count++) {
      if( (aValues[count]<=aValue) && (aValue<aValues[count+1]) ) {
        return aColors[count];
      }
    }
    if(aValue<aValues[0]) return aColors[0];
    if(aValue>=aValues[valn-1]) return aColors[aColors.size()-1];
    return SbColor_black;
  } else {
    return SbColor_black;
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
static struct {
  char string[32];
  SbMarkerStyle value;
} MarkerStyleList [] = {
  {"dot",         SbMarkerDot},
  {"plus",        SbMarkerPlus},
  {"asterisk",    SbMarkerAsterisk},
  {"cross",       SbMarkerCross},
  {"star",        SbMarkerStar},
  {"circle_line",         SbMarkerCircleLine},
  {"circle_filled",       SbMarkerCircleFilled},
  {"triangle_up_line",    SbMarkerTriangleUpLine},
  {"triangle_up_filled",  SbMarkerTriangleUpFilled},
  {"triangle_down_line",  SbMarkerTriangleDownLine},
  {"triangle_down_filled",SbMarkerTriangleDownFilled},
  {"david_star_line",     SbMarkerDavidStarLine},
  {"david_star_filled",   SbMarkerDavidStarFilled},
  {"swiss_cross_line",    SbMarkerSwissCrossLine},
  {"swiss_cross_filled",  SbMarkerSwissCrossFilled},
  {"diamond_line",        SbMarkerDiamondLine},
  {"diamond_filled",      SbMarkerDiamondFilled},
  {"square_line",         SbMarkerSquareLine},
  {"square_filled",       SbMarkerSquareFilled},
  {"penta_star_line",     SbMarkerPentaStarLine},
  {"penta_star_filled",   SbMarkerPentaStarFilled},
};  
static struct {
  char string[16];
  SbAreaStyle value;
} AreaStyleList [] = {
  {"solid",         SbAreaSolid},
  {"hatched",       SbAreaHatched},
  {"checker",       SbAreaChecker},
  {"edged",         SbAreaEdged}
};  
static struct {
  char string[20];
  SbPaintingPolicy value;
} PaintingPolicyList [] = {
  {"uniform",      SbPaintingUniform},
  {"by_value",     SbPaintingByValue},
  {"by_level",     SbPaintingByLevel},
  {"grey_scale",   SbPaintingGreyScale},
  {"violet_to_red",SbPaintingVioletToRed},
  {"grey_scale_inverse",SbPaintingGreyScaleInverse}
};  
static struct {
  char string[16];
  SbHatchingPolicy value;
} HatchingPolicyList [] = {
  {"none",           SbHatchingNone},
  {"right",          SbHatchingRight},
  {"left",           SbHatchingLeft},
  {"left_and_right", SbHatchingLeftAndRight}
};  
static struct {
  char string[16];
  SbProjectionType value;
} ProjectionTypeList [] = {
  {"none",     SbProjectionNone},
  {"rz",       SbProjectionRZ},
  {"phiz",     SbProjectionPHIZ},
  {"zr",       SbProjectionZR},
  {"zphi",     SbProjectionZPHI}
};  
/*
static struct {
  char*       string;
  SbStyle::FontType    value;
} FontTypeList [] = {
  {"latin"   ,    SbStyle::FontLatin},
  {"greek"   ,    SbStyle::FontGreek}
};  
*/
#define Number(arr)  (sizeof(arr) / sizeof(arr[0]))

SbMarkerStyle stringToMarkerStyle(const SbString& aString){
  for(unsigned int count=0;count<Number(MarkerStyleList);count++) {
    if(aString==MarkerStyleList[count].string) 
      return MarkerStyleList[count].value;
  }
  SoDebugError::postInfo("stringToMarkerStyle",
    "%s not a SbMarkerStyle.\n",aString.getString());
  return SbMarkerDot;
}

char* SbStyle::markerStyleToString(SbMarkerStyle aValue){
  for(unsigned int count=0;count<Number(MarkerStyleList);count++) {
    if(aValue==MarkerStyleList[count].value) 
      return MarkerStyleList[count].string;
  }
  return 0;
}

SbAreaStyle stringToAreaStyle (const SbString& aString){
  for(unsigned int count=0;count<Number(AreaStyleList);count++) {
    if(aString==AreaStyleList[count].string) 
      return AreaStyleList[count].value;
  }
  SoDebugError::postInfo("stringToAreaStyle",
    "%s not a SbAreaStyle.\n",aString.getString());
  return SbAreaSolid;
}

char* SbStyle::areaStyleToString(SbAreaStyle aValue){
  for(unsigned int count=0;count<Number(AreaStyleList);count++) {
    if(aValue==AreaStyleList[count].value) 
      return AreaStyleList[count].string;
  }
  return 0;
}

SbPaintingPolicy stringToPaintingPolicy(const SbString& aString){
  for(unsigned int count=0;count<Number(PaintingPolicyList);count++) {
    if(aString==PaintingPolicyList[count].string) 
      return PaintingPolicyList[count].value;
  }
  SoDebugError::postInfo("stringToPaintingPolicy",
    "%s not a SbPaintingPolicy.\n",aString.getString());
  return SbPaintingUniform;
}

char* SbStyle::paintingPolicyToString(SbPaintingPolicy aValue){
  for(unsigned int count=0;count<Number(PaintingPolicyList);count++) {
    if(aValue==PaintingPolicyList[count].value) 
      return PaintingPolicyList[count].string;
  }
  return 0;
}

SbHatchingPolicy stringToHatchingPolicy(const SbString& aString){
  for(unsigned int count=0;count<Number(HatchingPolicyList);count++) {
    if(aString==HatchingPolicyList[count].string) 
      return HatchingPolicyList[count].value;
  }
  SoDebugError::postInfo("stringToHatchingPolicy",
    "%s not a SbHatchingPolicy.\n",aString.getString());
  return SbHatchingNone;
}

char* SbStyle::hatchingPolicyToString(SbHatchingPolicy aValue){
  for(unsigned int count=0;count<Number(HatchingPolicyList);count++) {
    if(aValue==HatchingPolicyList[count].value) 
      return HatchingPolicyList[count].string;
  }
  return 0;
}

SbProjectionType stringToProjectionType(const SbString& aString){
  for(unsigned int count=0;count<Number(ProjectionTypeList);count++) {
    if(aString==ProjectionTypeList[count].string) 
      return ProjectionTypeList[count].value;
  }
  SoDebugError::postInfo("stringToProjectionType",
    "%s not a SbProjectionType.\n",aString.getString());
  return SbProjectionNone;
}

char* SbStyle::projectionTypeToString(SbProjectionType aValue){
  for(unsigned int count=0;count<Number(ProjectionTypeList);count++) {
    if(aValue==ProjectionTypeList[count].value) 
      return ProjectionTypeList[count].string;
  }
  return 0;
}

#include <vector>

SbBool stringToLinePattern(const SbString& aString,SbLinePattern& aPattern){
  typedef std::pair<SbString,SbLinePattern> Pattern;
  static std::vector<Pattern> sPatterns;
  if(!sPatterns.size()) {
    sPatterns.push_back(Pattern("solid",SbLinePattern_solid));
    sPatterns.push_back(Pattern("dashed",SbLinePattern_dashed));
    sPatterns.push_back(Pattern("dotted",SbLinePattern_dotted));
    sPatterns.push_back(Pattern("dash_dotted",SbLinePattern_dash_dotted));
  }

  std::vector<Pattern>::const_iterator it;
  for(it=sPatterns.begin();it!=sPatterns.end();++it) {
    if((*it).first==aString) {
      aPattern = (*it).second;
      return TRUE;
    }
  }

  return FALSE;
}

SbBool SbStyle::stringToColor(const SbString& aString,SbColor& aColor){
  if( (aString.getLength()==7) && (aString.getString()[0]=='#') ) {
    // #RRGGBB format :
    //  1 3 5
    unsigned long rr,gg,bb;

   {SbString s("0x");
    s += aString.getSubString(1,2);
    if(SbStringConvertToUnsignedLong(s,rr)==FALSE) return FALSE;}

   {SbString s("0x");
    s += aString.getSubString(3,4);
    if(SbStringConvertToUnsignedLong(s,gg)==FALSE) return FALSE;}

   {SbString s("0x");
    s += aString.getSubString(5,6);
    if(SbStringConvertToUnsignedLong(s,bb)==FALSE) return FALSE;}

    aColor = SbColor(((float)rr)/255,((float)gg)/255,((float)bb)/255);
    return TRUE;
  } 

  // Look in the named color table :
 {SbStyleColormap& defaultColormap = getDefaultColormap();
  std::vector<SbStyleColor>::const_iterator it;
  for(it=defaultColormap.begin();it!=defaultColormap.end();++it) {
    if((*it).first==aString) {
      aColor = (*it).second;
      return TRUE;
    }
  }}

  // Look if three floats :
 {SbPList words = SbStringGetWords(aString," ");
  if(words.getLength()==3) {
    SbString& word_0 = *((SbString*)words[0]);
    SbString& word_1 = *((SbString*)words[1]);
    SbString& word_2 = *((SbString*)words[2]);
    float r,g,b;
    if( SbStringConvertToFloat(word_0,r) &&
        SbStringConvertToFloat(word_1,g) &&
        SbStringConvertToFloat(word_2,b) ){
      aColor = SbColor(r,g,b);
      SbStringDelete(words);
      return TRUE;
    }
  }
  SbStringDelete(words);}

  return FALSE;
}

SbBool SbStyle::colorToString(const SbColor& aColor,SbString& aString){
  return SbStringPrintF(aString,3*32,"%g %g %g",aColor[0],aColor[1],aColor[2]);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

SbStyleColor::SbStyleColor()
:std::pair<SbString,SbColor>("",SbColor())
{}

SbStyleColor::SbStyleColor(const SbString& aName,const SbColor& aColor)
:std::pair<SbString,SbColor>(aName,aColor)
{}

SbStyleColor::SbStyleColor(
 const SbString& aName
,float aRed
,float aGreen
,float aBlue
)
:std::pair<SbString,SbColor>(aName,SbColor(aRed,aGreen,aBlue))
{}

SbStyleColor::SbStyleColor(const SbStyleColor& aFrom)
:std::pair<SbString,SbColor>(aFrom.first,aFrom.second)
{}

SbStyleColor::~SbStyleColor(){}

SbStyleColor& SbStyleColor::operator=(const SbStyleColor& aFrom){
  this->first = aFrom.first;
  this->second = aFrom.second;
  return *this;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

SbColor SbStyleColormap::getColor(int aIndex) const{
  if((aIndex<0)||(aIndex>=(int)size())) return SbColor(0.5F,0.5F,0.5F);
  return (*this)[aIndex].second;
}

SbString SbStyleColormap::getStringColor(int aIndex) const{
  if((aIndex<0)||(aIndex>=(int)size())) return "";
  SbColor c = (*this)[aIndex].second;
  char ss[3 * 32];
  ::sprintf(ss,"%g %g %g",c[0],c[1],c[2]);
  return SbString(ss);  
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

SbColor SbStyle::getGrey(float a_ratio){
  if(a_ratio<0.0F) a_ratio = 0;
  if(a_ratio>1.0F) a_ratio = 1;
  return SbColor(a_ratio,a_ratio,a_ratio);
}

SbColor SbStyle::getGreyInverse(float a_ratio){
  if(a_ratio<0.0F) a_ratio = 0;
  if(a_ratio>1.0F) a_ratio = 1;
  a_ratio = 1 - a_ratio;
  return SbColor(a_ratio,a_ratio,a_ratio);
}

static float hls_to_rgb(float rn1, float rn2, float huei) {
  // taken from ROOT/TColor.
  float hue = huei;
  if (hue > 360) hue = hue - 360;
  if (hue < 0)   hue = hue + 360;
  if (hue < 60 ) return rn1 + (rn2-rn1)*hue/60;
  if (hue < 180) return rn2;
  if (hue < 240) return rn1 + (rn2-rn1)*(240-hue)/60;
  return rn1;
}


void SbStyle::hls2rgb(float hue, float light, float satur,
                      float &r, float &g, float &b){
  // taken from ROOT/TColor.
  // Static method to compute RGB from HLS. The l and s are between [0,1]
  // and h is between [0,360]. The returned r,g,b triplet is between [0,1].

  float rh, rl, rs, rm1, rm2;
  rh = rl = rs = 0;
  if (hue   > 0) rh = hue;   if (rh > 360) rh = 360;
  if (light > 0) rl = light; if (rl > 1)   rl = 1;
  if (satur > 0) rs = satur; if (rs > 1)   rs = 1;

  if (rl <= 0.5) {
    rm2 = rl*(1.0 + rs);
  } else {
    rm2 = rl + rs - rl*rs;
  }
  rm1 = 2.0*rl - rm2;

  if (!rs) { r = rl; g = rl; b = rl; return; }
  r = hls_to_rgb(rm1, rm2, rh+120);
  g = hls_to_rgb(rm1, rm2, rh);
  b = hls_to_rgb(rm1, rm2, rh-120);
}

SbColor SbStyle::getVioletToRed(float a_ratio){
  if(a_ratio<0.0F) a_ratio = 0;
  if(a_ratio>1.0F) a_ratio = 1;
  // a_ratio in [0,1]
  // From ROOT pretty palette.
  // Initialize with the spectrum Violet->Red
  // The color model used here is based on the HLS model which
  // is much more suitable for creating palettes than RGB.
  // Fixing the saturation and lightness we can scan through the
  // spectrum of visible light by using "hue" alone.
  // In Root hue takes values from 0 to 360.
  float saturation = 1;
  float lightness = 0.5;
  float hue_mn = 0;
  float hue_mx = 280;
  float hue = hue_mx - a_ratio * (hue_mx-hue_mn);
  float r,g,b;
  hls2rgb(hue,lightness,saturation,r,g,b);
  return SbColor(r,g,b);
}


void SbStyle::setColors(SbColor(*aGet)(float),unsigned int aCelln,std::vector<SbColor>& aColors) {
  aColors.resize(aCelln);
  if(!aCelln) return;
  float d = 1.0F/(aCelln-1);
  for(unsigned int index=0;index<aCelln;index++) {
    aColors[index] = aGet(d*index);
  }
}

void SbStyle::setGreyScale(unsigned int aCelln,std::vector<SbColor>& aColors){setColors(getGrey,aCelln,aColors);}

void SbStyle::setVioletToRed(unsigned int aCelln,std::vector<SbColor>& aColors){setColors(getVioletToRed,aCelln,aColors);}

static SbStyleColormap* sDefaultColormap = 0;
SbStyleColormap& SbStyle::getDefaultColormap(){
  if(!sDefaultColormap) {
    sDefaultColormap = new SbStyleColormap;

    // SbColor_xxx defined in SbStyles.h
    sDefaultColormap->push_back(SbStyleColor("aquamarine",SbColor_aquamarine));
    sDefaultColormap->push_back(SbStyleColor("mediumaquamarine",SbColor_mediumaquamarine));
    sDefaultColormap->push_back(SbStyleColor("black",SbColor_black));
    sDefaultColormap->push_back(SbStyleColor("blue",SbColor_blue));
    sDefaultColormap->push_back(SbStyleColor("cadetblue",SbColor_cadetblue));
    sDefaultColormap->push_back(SbStyleColor("cornflowerblue",SbColor_cornflowerblue));
    sDefaultColormap->push_back(SbStyleColor("darkslateblue",SbColor_darkslateblue));
    sDefaultColormap->push_back(SbStyleColor("lightblue",SbColor_lightblue));
    sDefaultColormap->push_back(SbStyleColor("lightsteelblue",SbColor_lightsteelblue));
    sDefaultColormap->push_back(SbStyleColor("mediumblue",SbColor_mediumblue));

    sDefaultColormap->push_back(SbStyleColor("mediumslateblue",SbColor_mediumslateblue));
    sDefaultColormap->push_back(SbStyleColor("midnightblue",SbColor_midnightblue));
    sDefaultColormap->push_back(SbStyleColor("navyblue",SbColor_navyblue));
    sDefaultColormap->push_back(SbStyleColor("navy",SbColor_navy));
    sDefaultColormap->push_back(SbStyleColor("skyblue",SbColor_skyblue));
    sDefaultColormap->push_back(SbStyleColor("slateblue",SbColor_slateblue));
    sDefaultColormap->push_back(SbStyleColor("steelblue",SbColor_steelblue));
    sDefaultColormap->push_back(SbStyleColor("coral",SbColor_coral));
    sDefaultColormap->push_back(SbStyleColor("cyan",SbColor_cyan));
    sDefaultColormap->push_back(SbStyleColor("firebrick",SbColor_firebrick));
      
    sDefaultColormap->push_back(SbStyleColor("brown",SbColor_brown));
    sDefaultColormap->push_back(SbStyleColor("gold",SbColor_gold));
    sDefaultColormap->push_back(SbStyleColor("goldenrod",SbColor_goldenrod));
    sDefaultColormap->push_back(SbStyleColor("green",SbColor_green));
    sDefaultColormap->push_back(SbStyleColor("darkgreen",SbColor_darkgreen));
    sDefaultColormap->push_back(SbStyleColor("darkolivegreen",SbColor_darkolivegreen));
    sDefaultColormap->push_back(SbStyleColor("forestgreen",SbColor_forestgreen));
    sDefaultColormap->push_back(SbStyleColor("limegreen",SbColor_limegreen));
    sDefaultColormap->push_back(SbStyleColor("mediumseagreen",SbColor_mediumseagreen));
    sDefaultColormap->push_back(SbStyleColor("mediumspringgreen",SbColor_mediumspringgreen));
      
    sDefaultColormap->push_back(SbStyleColor("palegreen",SbColor_palegreen));
    sDefaultColormap->push_back(SbStyleColor("seagreen",SbColor_seagreen));
    sDefaultColormap->push_back(SbStyleColor("springgreen",SbColor_springgreen));
    sDefaultColormap->push_back(SbStyleColor("yellowgreen",SbColor_yellowgreen));
    sDefaultColormap->push_back(SbStyleColor("darkslategrey",SbColor_darkslategrey));
    sDefaultColormap->push_back(SbStyleColor("dimgrey",SbColor_dimgrey));
    sDefaultColormap->push_back(SbStyleColor("lightgrey",SbColor_lightgrey));
    sDefaultColormap->push_back(SbStyleColor("grey",SbColor_grey));
    sDefaultColormap->push_back(SbStyleColor("khaki",SbColor_khaki));
    sDefaultColormap->push_back(SbStyleColor("magenta",SbColor_magenta));
      
    sDefaultColormap->push_back(SbStyleColor("maroon",SbColor_maroon));
    sDefaultColormap->push_back(SbStyleColor("orange",SbColor_orange));
    sDefaultColormap->push_back(SbStyleColor("orchid",SbColor_orchid));
    sDefaultColormap->push_back(SbStyleColor("darkorchid",SbColor_darkorchid));
    sDefaultColormap->push_back(SbStyleColor("mediumorchid",SbColor_mediumorchid));
    sDefaultColormap->push_back(SbStyleColor("pink",SbColor_pink));
    sDefaultColormap->push_back(SbStyleColor("plum",SbColor_plum));
    sDefaultColormap->push_back(SbStyleColor("red",SbColor_red));
    sDefaultColormap->push_back(SbStyleColor("indianred",SbColor_indianred));
    sDefaultColormap->push_back(SbStyleColor("mediumvioletred",SbColor_mediumvioletred));
      
    sDefaultColormap->push_back(SbStyleColor("orangered",SbColor_orangered));
    sDefaultColormap->push_back(SbStyleColor("violetred",SbColor_violetred));
    sDefaultColormap->push_back(SbStyleColor("salmon",SbColor_salmon));
    sDefaultColormap->push_back(SbStyleColor("sienna",SbColor_sienna));
    sDefaultColormap->push_back(SbStyleColor("tan",SbColor_tan));
    sDefaultColormap->push_back(SbStyleColor("thistle",SbColor_thistle));
    sDefaultColormap->push_back(SbStyleColor("turquoise",SbColor_turquoise));
    sDefaultColormap->push_back(SbStyleColor("darkturquoise",SbColor_darkturquoise));
    sDefaultColormap->push_back(SbStyleColor("mediumturquoise",SbColor_mediumturquoise));
    sDefaultColormap->push_back(SbStyleColor("violet",SbColor_violet));
      
    sDefaultColormap->push_back(SbStyleColor("blueviolet",SbColor_blueviolet));
    sDefaultColormap->push_back(SbStyleColor("wheat",SbColor_wheat));
    sDefaultColormap->push_back(SbStyleColor("white",SbColor_white));
    sDefaultColormap->push_back(SbStyleColor("yellow",SbColor_yellow));
    sDefaultColormap->push_back(SbStyleColor("greenyellow",SbColor_greenyellow));
  }

  return *sDefaultColormap;
}


static SbStyleColormap* sROOT_Colormap = 0;
SbStyleColormap& SbStyle::getROOT_Colormap(){
  // ROOT-4.00.08/TApplication.cxx/InitializeColors
  if(!sROOT_Colormap) {
    sROOT_Colormap = new SbStyleColormap;
    sROOT_Colormap->resize(51);

    //base/inc/Gtypes.h
    enum EColor {kWhite,kBlack,kRed,kGreen,kBlue,kYellow,kMagenta,kCyan};
                 //0    1      2    3      4     5       6        7

#define new_TColor(i,r,g,b,n) \
  (*sROOT_Colormap)[i].operator=(SbStyleColor((n),(float)(r),(float)(g),(float)(b)));

    new_TColor(kWhite,1,1,1,"background");
    new_TColor(kBlack,0,0,0,"black");
    new_TColor(kRed,1,0,0,"red");
    new_TColor(kGreen,0,1,0,"green");
    new_TColor(kBlue,0,0,1,"blue");
    new_TColor(kYellow,1,1,0,"yellow");
    new_TColor(kMagenta,1,0,1,"magenta");
    new_TColor(kCyan,0,1,1,"cyan");
    new_TColor(10,0.999,0.999,0.999,"white");
    new_TColor(11,0.754,0.715,0.676,"editcol");

    // The color white above is defined as being nearly white.
    // Sets the associated dark color also to white.
    //new_TColor(110,0.999,0.999,.999,"Color110");

    // Initialize Custom colors
    new_TColor(20,0.8,0.78,0.67,"Color20");
    new_TColor(31,0.54,0.66,0.63,"Color31");
    new_TColor(41,0.83,0.81,0.53,"Color41");
    new_TColor(30,0.52,0.76,0.64,"Color30");
    new_TColor(32,0.51,0.62,0.55,"Color32");
    new_TColor(24,0.70,0.65,0.59,"Color24");
    new_TColor(21,0.8,0.78,0.67,"Color21");
    new_TColor(47,0.67,0.56,0.58,"Color47");
    new_TColor(35,0.46,0.54,0.57,"Color35");
    new_TColor(33,0.68,0.74,0.78,"Color33");
    new_TColor(39,0.5,0.5,0.61,"Color39");
    new_TColor(37,0.43,0.48,0.52,"Color37");
    new_TColor(38,0.49,0.6,0.82,"Color38");
    new_TColor(36,0.41,0.51,0.59,"Color36");
    new_TColor(49,0.58,0.41,0.44,"Color49");
    new_TColor(43,0.74,0.62,0.51,"Color43");
    new_TColor(22,0.76,0.75,0.66,"Color22");
    new_TColor(45,0.75,0.51,0.47,"Color45");
    new_TColor(44,0.78,0.6,0.49,"Color44");
    new_TColor(26,0.68,0.6,0.55,"Color26");
    new_TColor(28,0.53,0.4,0.34,"Color28");
    new_TColor(25,0.72,0.64,0.61,"Color25");
    new_TColor(27,0.61,0.56,0.51,"Color27");
    new_TColor(23,0.73,0.71,0.64,"Color23");
    new_TColor(42,0.87,0.73,0.53,"Color42");
    new_TColor(46,0.81,0.37,0.38,"Color46");
    new_TColor(48,0.65,0.47,0.48,"Color48");
    new_TColor(34,0.48,0.56,0.6,"Color34");
    new_TColor(40,0.67,0.65,0.75,"Color40");
    new_TColor(29,0.69,0.81,0.78,"Color29");

    // Initialize some additional greyish non saturated colors
    new_TColor(8, 0.35,0.83,0.33,"Color8");
    new_TColor(9, 0.35,0.33,0.85,"Color9");
    new_TColor(12,.3,.3,.3,"grey12");
    new_TColor(13,.4,.4,.4,"grey13");
    new_TColor(14,.5,.5,.5,"grey14");
    new_TColor(15,.6,.6,.6,"grey15");
    new_TColor(16,.7,.7,.7,"grey16");
    new_TColor(17,.8,.8,.8,"grey17");
    new_TColor(18,.9,.9,.9,"grey18");
    new_TColor(19,.95,.95,.95,"grey19");
    new_TColor(50, 0.83,0.35,0.33,"Color50");

    // Initialize the Pretty Palette Spectrum Violet->Red
    //   The color model used here is based on the HLS model which
    //   is much more suitable for creating palettes than RGB.
    //   Fixing the saturation and lightness we can scan through the
    //   spectrum of visible light by using "hue" alone.
    //   In Root hue takes values from 0 to 360.
    /*FIXME
    float  saturation = 1;
    float  lightness = 0.5;
    float  MaxHue = 280;
    float  MinHue = 0;
    Int_t  MaxPretty = 50;
    float  hue;

    for (i=0 ; i<MaxPretty ; i++) {
      hue = MaxHue-(i+1)*((MaxHue-MinHue)/MaxPretty);
      TColor::HLStoRGB(hue, lightness, saturation, r, g, b);
      new_TColor(i+51, r, g, b);
    }*/

  }
  return *sROOT_Colormap;

#undef new_TColor
}
