#include <stdio.h> //debug

//#define DEBUG

#include <HEPVis/SbBasic.h>

#ifdef HEPVisWin

#include <windows.h>
#include <windowsx.h>
#include <wingdi.h>

#include <HEPVis/Win/SbPainterWin.h>

#define FLOAT_LONG (LONG)(float)
#define DBL_INT (int)(float)

//////////////////////////////////////////////////////////////////////////////
SbPainterWin::SbPainterWin(
)
:fHWND(NULL)
,fHDC(NULL)
,fPointn(0)
,fPoints(NULL)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
SbPainterWin::~SbPainterWin(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  delete [] fPoints;
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterWin::setWindow(
 HWND aWindow
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fHWND = aWindow;
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterWin::setHDC(
 HDC aHDC
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fHDC = aHDC;
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterWin::beginTraversal(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterWin::clearColorBuffer(
 float aRed
,float aGreen
,float aBlue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fRed = aRed;
  fGreen = aGreen;
  fBlue = aBlue;

  POINT* points = new POINT[4];
  points[0].x = 0; 
  points[0].y = 0; 

  points[1].x = 0; 
  points[1].y = fWindowHeight; 

  points[2].x = fWindowWidth; 
  points[2].y = fWindowHeight; 

  points[3].x = fWindowWidth; 
  points[3].y = 0; 

  drawPolygon(4,points,aRed,aGreen,aBlue,SbAreaSolid);
  delete [] points;

}
//////////////////////////////////////////////////////////////////////////////
void SbPainterWin::endTraversal(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterWin::drawPrimitive (
 SbPrimitiveType aType
,int aPointn
,float* aXs
,float* aYs
,float* aZs
,const SbPainterContext& aAtb
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(allocatePoints(aType,aPointn/*,aAtb*/)==FALSE) return;
  switch(aType) {
  case SbPrimitivePoints:
    if(convertLineToPoints(aPointn,aXs,aYs,aZs)==TRUE) {
      drawMarkers(aPointn,fPoints,
		  aAtb.fRed,aAtb.fGreen,aAtb.fBlue,
		  aAtb.fMarkerStyle,aAtb.fMarkerSize);
    }
    break;
  case SbPrimitiveLineStrip:
  case SbPrimitiveLineLoop:
    if(convertLineToPoints (aPointn,aXs,aYs,aZs)==TRUE) {
      drawLines(aPointn,fPoints,
		aAtb.fRed,aAtb.fGreen,aAtb.fBlue,
		aAtb.fLinePattern,aAtb.fLineWidth);
    }
    break;
  case SbPrimitivePolygon:
    if(convertLineToPoints(aPointn,aXs,aYs,aZs)==TRUE) {
      drawPolygon(aPointn,fPoints,
		  aAtb.fRed,aAtb.fGreen,aAtb.fBlue,
		  aAtb.fAreaStyle);
    }
    break;
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
SbBool SbPainterWin::allocatePoints(
 int aMode
,int aPointn
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aPointn==0) return FALSE;
  if(aPointn>fPointn) {
    if(fPoints!=NULL) delete [] fPoints;
    fPoints = new POINT[aPointn];
    if(fPoints==NULL) {
      fPoints = NULL;
      fPointn = 0;
      return FALSE;
    }
    fPointn = aPointn;
  }
  return TRUE;
}
//////////////////////////////////////////////////////////////////////////////
SbBool SbPainterWin::convertLineToPoints (
 int aNumber
,float* aXs
,float* aYs
,float* /*aZs*/
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aNumber<=0) return FALSE;
  for(int count=0;count<aNumber;count++) { 
    fPoints[count].x = FLOAT_LONG aXs[count];
    fPoints[count].y = FLOAT_LONG (fWindowHeight - aYs[count]);
  }
  return TRUE;
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterWin::drawPolygon(
 int aPointn
,POINT* aPoints
,float aRed
,float aGreen
,float aBlue
,const SbAreaStyle& /*aStyle*/
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fHWND==NULL) return;
  if(fHDC==NULL) return;
  COLORREF color = RGB(aRed * 255,aGreen * 255,aBlue * 255);
  COLORREF colorEdge;
  if(fEdges==TRUE) {
    colorEdge = RGB(fRed,fGreen,fBlue);
  } else {
    colorEdge = color;
  }
  HBRUSH hBrush = CreateSolidBrush(color); // Brush for the interior.
  HPEN hPen = CreatePen(PS_SOLID,1,colorEdge); // Pen for the outline. 
  HBRUSH hBrushPrev = SelectBrush(fHDC,hBrush);
  HPEN hPenPrev = SelectPen(fHDC,hPen);
  Polygon(fHDC,aPoints,aPointn);
  SelectBrush(fHDC,hBrushPrev);
  SelectPen(fHDC,hPenPrev);
  DeleteBrush(hBrush);
  DeletePen(hPen);
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterWin::drawLines(
 int aPointn
,POINT* aPoints
,float aRed
,float aGreen
,float aBlue
,SbLinePattern /*aPattern*/
,int /*aWidth*/
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fHWND==NULL) return;
  if(fHDC==NULL) return;
  COLORREF color = RGB(aRed * 255,aGreen * 255,aBlue * 255);
  HPEN hPen = CreatePen(PS_SOLID,1,color);
  HPEN hPenPrev = SelectPen(fHDC,hPen);
  Polyline(fHDC,aPoints,aPointn);
  SelectPen(fHDC,hPenPrev);
  DeletePen(hPen);
  /*
  if( (aPointn==2) &&
      (aPoints[0].x == (int)aPoints[1].x) &&
      (aPoints[0].y == (int)aPoints[1].y) ) {
    XDrawPoint(display,drawable,gc,(int)aPoints[0].x,(int)aPoints[0].y);
  } else {
    XDrawLines(display,drawable,gc,aPoints,aPointn,CoordModeOrigin);
  }
  */
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterWin::drawMarkers (
 int aPointn
,POINT* aPoints
,float aRed
,float aGreen
,float aBlue
,const SbMarkerStyle& aStyle 
,int aSize
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fHWND==NULL) return;
  if(fHDC==NULL) return;
  int mark_size  = aSize <=1 ? 4 : aSize;   
  COLORREF color = RGB(aRed * 255,aGreen * 255,aBlue * 255);
  HPEN hPen = CreatePen(PS_SOLID,1,color);
  HPEN hPenPrev = SelectPen(fHDC,hPen);
  if(aStyle==SbMarkerCircleLine) {
    int half = mark_size/2;
    int ulx,uly;
    int brx,bry;
    int begx,begy;
    int endx,endy;
    for(int count=0;count<aPointn;count++) {
      ulx = aPoints[count].x - half;
      uly = aPoints[count].y - half;
      brx = aPoints[count].x + half;
      bry = aPoints[count].y + half;
      begx = aPoints[count].x;
      begy = aPoints[count].y + half;
      endx = aPoints[count].x;
      endy = aPoints[count].y + half;
      Arc(fHDC,ulx,uly,brx,bry,begx,begy,endx,endy);
    }
  } else {
    for(int count=0;count<aPointn;count++) {
      drawMarkerSegments(aStyle,mark_size,
			 aPoints[count].x,aPoints[count].y);
    }
  }
  SelectPen(fHDC,hPenPrev);
  DeletePen(hPen);
}
//////////////////////////////////////////////////////////////////////////////
void SbPainterWin::drawMarkerSegments(
 SbMarkerStyle aStyle 
,int aSize 
,int aCenter_x 
,int aCenter_y 
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aSize /= 2;
  aSize = aSize==0 ? 1 : aSize;
  POINT points[2];
  switch(aStyle) {
  case SbMarkerPlus:
    points[0].x = aCenter_x - aSize;
    points[0].y = aCenter_y ;
    points[1].x = aCenter_x + aSize;
    points[1].y = aCenter_y ;
    Polyline(fHDC,points,2);
    
    points[0].x = aCenter_x ;
    points[0].y = aCenter_y - aSize;
    points[1].x = aCenter_x ;
    points[1].y = aCenter_y + aSize;
    Polyline(fHDC,points,2);

    break;
  case SbMarkerAsterisk:
    points[0].x = aCenter_x - aSize;
    points[0].y = aCenter_y ;
    points[1].x = aCenter_x + aSize;
    points[1].y = aCenter_y ;
    Polyline(fHDC,points,2);

    points[0].x = aCenter_x ;
    points[0].y = aCenter_y - aSize;
    points[1].x = aCenter_x ;
    points[1].y = aCenter_y + aSize;
    Polyline(fHDC,points,2);
    
    points[0].x = aCenter_x - DBL_INT (aSize * .707);
    points[0].y = aCenter_y - DBL_INT (aSize * .707);
    points[1].x = aCenter_x + DBL_INT (aSize * .707);
    points[1].y = aCenter_y + DBL_INT (aSize * .707);
    Polyline(fHDC,points,2);
    
    points[0].x = aCenter_x + DBL_INT (aSize * .707);
    points[0].y = aCenter_y - DBL_INT (aSize * .707);
    points[1].x = aCenter_x - DBL_INT (aSize * .707);
    points[1].y = aCenter_y + DBL_INT (aSize * .707);
    Polyline(fHDC,points,2);
    break;
  case SbMarkerCross:
    points[0].x = aCenter_x - aSize;
    points[0].y = aCenter_y - aSize;
    points[1].x = aCenter_x + aSize;
    points[1].y = aCenter_y + aSize;
    Polyline(fHDC,points,2);
    
    points[0].x = aCenter_x + aSize;
    points[0].y = aCenter_y - aSize;
    points[1].x = aCenter_x - aSize;
    points[1].y = aCenter_y + aSize;
    Polyline(fHDC,points,2);
    break;
  case SbMarkerStar:
    points[0].x = aCenter_x;
    points[0].y = aCenter_y + aSize;
    points[1].x = aCenter_x - DBL_INT (aSize * .6);
    points[1].y = aCenter_y - DBL_INT (aSize * .5);
    Polyline(fHDC,points,2);
    
    points[0].x = aCenter_x - DBL_INT (aSize * .6);
    points[0].y = aCenter_y - DBL_INT (aSize * .5);
    points[1].x = aCenter_x + DBL_INT (aSize * .6);
    points[1].y = aCenter_y - DBL_INT (aSize * .5);
    Polyline(fHDC,points,2);

    points[0].x = aCenter_x + DBL_INT (aSize * .6);
    points[0].y = aCenter_y - DBL_INT (aSize * .5);
    points[1].x = aCenter_x;
    points[1].y = aCenter_y + aSize;
    Polyline(fHDC,points,2);
    break;
  default:
    points[0].x = aCenter_x;
    points[0].y = aCenter_y;
    points[1].x = aCenter_x + 1;
    points[1].y = aCenter_y;
    Polyline(fHDC,points,2);
    break;
  }
}
#endif






