
#include <Lib/Arguments.h>
#include <Lib/Printer.h>
#include <Lib/Out.h>
#include <Lib/fmanip.h>
#include <Lib/smanip.h>

#include <stdlib.h>

#ifdef WIN32
static bool Substitute(const std::vector< std::pair<std::string,std::string> >&,std::string&,bool,Slash::Core::IWriter&);
#endif

#ifdef __alpha
extern "C" { extern int putenv(char *);}
#endif

//////////////////////////////////////////////////////////////////////////////
int main (
 int aArgc
,char** aArgv
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::Arguments args(aArgc,aArgv);

  if(args.number()<3) {
    out << "Usage :" << Lib::endl;
    out << "  OS> Lib_run -putenv=<file name> -command=<command> [-debug]" 
        << Lib::endl;
    return EXIT_FAILURE;
  }

  std::string fileName;
  if(!args.find("-putenv",fileName)) {
    out << "Usage :" << Lib::endl;
    out << "  OS> Lib_run -putenv=<file name> -command=<command> [-debug]" 
        << Lib::endl;
    return EXIT_FAILURE;
  }

  std::string command;
  if(!args.find("-command",command) || (command.empty())) {
    out << "Usage :" << Lib::endl;
    out << "  OS> Lib_run -putenv=<file name> -command=<command> [-debug]" 
        << Lib::endl;
    return EXIT_FAILURE;
  }


  bool debug = args.isAnArgument("-debug");

  std::vector<std::string> text;
  if(!Lib::fmanip::read(fileName,text)) {
    out << "Can't read \"" << fileName << "\"" << Lib::endl;
    return EXIT_FAILURE;
  }

  std::vector< std::pair<std::string,std::string> > envs;
  unsigned int index;
  for(index=0;index<text.size();index++) {

    if(debug)
      out << "debug : treat \"" << text[index] << "\"" << Lib::endl;

    std::vector<std::string> words;
    Lib::smanip::words(text[index]," ",words);
    if(words.size() && words[0]=="set") {

      std::string rest = "";
      unsigned int number = words.size();
      for(unsigned int i=1;i<number;i++) {
        rest += words[i];
        if(i!=number-1) rest += " ";
      }
      Lib::smanip::words(rest,"=",words);
      if(words.size()==1) {
        envs.push_back(std::pair<std::string,std::string>(words[0],""));
        if(debug)
          out << " debug : set \"" << words[0]
              << "\" to value \"\"" << Lib::endl; 
      } else if(words.size()==2) {
        envs.push_back(std::pair<std::string,std::string>(words[0],words[1]));
        if(debug)
          out << " debug : set \"" << words[0]
              << "\" to value \"" << words[1] << "\"" << Lib::endl; 
      }

    } else if((words.size()>=3) && words[0]=="setenv") {

      std::string rest = "";
      unsigned int number = words.size();
      for(unsigned int i=2;i<number;i++) {
        rest += words[i];
        if(i!=number-1) rest += " ";
      }
      envs.push_back(std::pair<std::string,std::string>(words[1],rest));

    }

  }

#ifdef WIN32
  std::vector< std::pair<std::string,std::string> > subenvs;
  for(index=0;index<envs.size();index++) {
    const std::pair<std::string,std::string>& p = envs[index];
    std::string value = p.second;
    if(value.size() && value[0]=='"') 
      Lib::smanip::strip(value,Lib::smanip::both,'"');
    if(value.size() && value[0]=='\'') 
      Lib::smanip::strip(value,Lib::smanip::both,'\'');
    Substitute(subenvs,value,debug,printer);
    subenvs.push_back(std::pair<std::string,std::string>(p.first,value));
  }
  envs = subenvs;
#endif

  std::vector<std::string> putenvs;

  for(index=0;index<envs.size();index++) {
    const std::pair<std::string,std::string>& p = envs[index];
    std::string value = p.second;
    if(value.size() && value[0]=='"') 
      Lib::smanip::strip(value,Lib::smanip::both,'"');
    if(value.size() && value[0]=='\'') 
      Lib::smanip::strip(value,Lib::smanip::both,'\'');
    std::string pvalue = p.first + "=" + value;
    if(debug)
      out << "debug : putenv \"" << pvalue << "\"" << Lib::endl; 
    putenvs.push_back(pvalue);
    if(::putenv((char*)putenvs[putenvs.size()-1].c_str())) {
      out << "putenv failure for \"" << pvalue << "\"" << Lib::endl; 
      return EXIT_FAILURE;
    }
    //check:
    char* s = ::getenv((char*)p.first.c_str());
    std::string sget = s?s:"";
    if(sget!=value) {
      out << "putenv badly set for \"" << p.first << "\"" << Lib::endl; 
      out << " getenv says \"" << sget << "\" whilst it should be \"" 
          << value << "\""<< Lib::endl; 
      return EXIT_FAILURE;
    } else if(debug) {
      out << "debug : getenv ok." << Lib::endl; 
    }
  }

  if(debug)
    out << "debug : command \"" << command << "\"" << Lib::endl; 

  if(::system((char*)command.c_str())==-1) {
    out << "::system failed for command \"" << command << "\"" << Lib::endl; 
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;
}
#ifdef WIN32
//////////////////////////////////////////////////////////////////////////////
bool Substitute(
 const std::vector< std::pair<std::string,std::string> >& aEnvs
,std::string& aString
,bool aDebug
,Slash::Core::IWriter& aPrinter
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string::size_type percent;
  std::string::size_type pos = 0;
  while((percent=aString.find('%',pos))!=std::string::npos){
    std::string::size_type percent2 = aString.find('%',percent+1);
    if(percent2==std::string::npos) {
      Lib::Out out(aPrinter);
      out << "Unmatched % in \"" << aString << "\"" << Lib::endl;
      return false;
    } else {
      std::string env = aString.substr(percent+1,percent2-(percent+1));
      bool found = false;
      std::string value;
      for(unsigned int index=0;index<aEnvs.size();index++) {
        const std::pair<std::string,std::string>& p = aEnvs[index];
        if(p.first==env) {
          value = p.second;
          found = true;
          break;
        }
      }
      if(!found) {
        char* s = ::getenv((char*)env.c_str());
        if(s) {
          value = s;
          found = true;
        }
      }
      if(!found) {
        // May be correct. The env variable does not
        // come from the input file but may exist in the environment.
        //pos = percent2+1;
        Lib::Out out(aPrinter);
        out << "Variable \"" << env << "\" not found." << Lib::endl;
        return false;
      } else {
        if(aDebug) {
          Lib::Out out(aPrinter);
          out << "debug : --------------------------" << Lib::endl;
          out << "debug : old \"" << aString << "\"." << Lib::endl;
        }
        std::string snew = aString.substr(0,percent);
        snew += value;
        snew += aString.substr(percent2+1,aString.length()-(percent2+1));
        aString = snew;
        if(aDebug) {
          Lib::Out out(aPrinter);
          out << "debug : --------------------------" << Lib::endl;
          out << "debug : new \"" << aString << "\"." << Lib::endl;
        }
      }
    }
  }
  return true;
}
#endif

// embed code :
#include <Printer.cxx>
#include <Out.cxx>
#include <Arguments.cxx>
#include <smanip.cxx>
#include <sout.cxx>
#include <cmanip.cxx>
#include <fmanip.cxx>
#include <System.cxx>
