
#include <Lib/Printer.h>
#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/Arguments.h>
#include <Lib/dirmanip.h>
#include <Lib/smanip.h>
#include <Lib/fmanip.h>

#include <stdlib.h>
#include <stdio.h>

class Finder : public Slash::Core::IDirectoryVisitor {
public:
  virtual bool directory(const std::string& aDirectory,bool& aProcess) {
    if(fVerbose) {
      Lib::Out out(fPrinter);
      out << "Lib_subs : treat directory " 
          << Lib::sout(aDirectory) << Lib::endl;
    }
    aProcess = true;
    return true;
  }
  virtual bool file(const std::string& aFile) {
    if(fVerbose) {
      Lib::Out out(fPrinter);
      out << "Lib_subs : treat " << Lib::sout(aFile) << Lib::endl;
    }
    if(fOld.size()) {
      std::vector<std::string> text;
      if(!Lib::fmanip::read(aFile,text)) {
	Lib::Out out(fPrinter);
	out << "Can't read " << Lib::sout(aFile) << Lib::endl;
	return false;
      }
      Lib::fmanip::remove(aFile);
      FILE* file = ::fopen(aFile.c_str(),"wb");
      if(file) {
	for(unsigned int index=0;index<text.size();index++) {
	  std::string s = text[index];
	  if(Lib::smanip::replace(s,fOld,fNew)) {
            if(fVerbose) {
              Lib::Out out(fPrinter);
              out << index << Lib::endl;
              out << "<   " << text[index] << Lib::endl;
              out << "---" << Lib::endl;
              out << ">   " << s << Lib::endl;
            }
          }
	  ::fprintf(file,"%s\n",s.c_str());
	}
	::fclose(file);
      }
    }
    return true;
  }
public:
  Finder(Slash::Core::IWriter& aPrinter,
	 const std::string& aOld,
	 const std::string& aNew,bool aVerbose)
    :fPrinter(aPrinter),fOld(aOld),fNew(aNew),fVerbose(aVerbose){}
  virtual ~Finder() {}
private:
  Slash::Core::IWriter& fPrinter;  
  std::string fOld;
  std::string fNew;
  bool fVerbose;
};

//////////////////////////////////////////////////////////////////////////////
int main (
 int aArgc
,char** aArgv
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Printer printer;
  Lib::Out out(printer);

  Lib::Arguments args(aArgc,aArgv);

  std::string shelp = 
    "  OS> Lib_subs [-verbose] [-path=<starting path>] [-file=<file pattern>] -old=<string to replace> -new=<replacing string>";
  shelp += "\n";
  shelp += "\n";
  shelp += " Take care of current directory. If so files are ./<name>.";
  shelp += "\n";
  shelp += " Then file pattern should be of the form ./<pattern> ";
  shelp += "\n";
  shelp += " or start with *";

  bool verbose = false;
  std::string sverbose;
  if(args.find("-verbose",sverbose)) {
    if(sverbose.empty()) verbose = true;
  }

  std::string path;
  if(!args.find("-path",path)) path = ".";

  std::string filter;
  if(!args.find("-file",filter)) filter = "*";

  std::string sold;
  if(!args.find("-old",sold)) {
    out << shelp << Lib::endl;
    return EXIT_FAILURE;
  }

  std::string snew;
  if(!args.find("-new",snew)) {
    out << shelp << Lib::endl;
    return EXIT_FAILURE;
  }

  if(verbose) {
    out << "Lib_subs :" << Lib::endl;
    out << "   path = " << Lib::sout(path) << Lib::endl;
    out << " filter = " << Lib::sout(filter) << Lib::endl;
    out << "    old = " << Lib::sout(sold) << Lib::endl;
    out << "    new = " << Lib::sout(snew) << Lib::endl;
  }


  Lib::Directory tree;
  tree.setPath(path);
  if(!Lib::dirmanip::content(tree,filter)) {
    out << "Can't get files of " << Lib::sout(path) << Lib::endl;
    return EXIT_FAILURE;
  }

  Finder finder(printer,sold,snew,verbose);
  tree.visit(finder);

  return EXIT_SUCCESS;
}

// embed code :
#include <Printer.cxx>
#include <Out.cxx>
#include <sout.cxx>
#include <Arguments.cxx>
#include <dirmanip.cxx>
#include <smanip.cxx>
#include <cmanip.cxx>
#include <fmanip.cxx>
#include <System.cxx>

