#ifndef Lib_ExpFunc_h
#define Lib_ExpFunc_h

#include <Lib/IExpFunc.h>

namespace Lib {namespace Function {class InsidePolygon;}}

namespace Lib {

class ExpFunc : public virtual IExpFunc {
public: //IExpFunc
public:
  virtual std::string name() const;
  virtual unsigned int numberOfArguments() const;
public:
  ExpFunc(const std::string&,unsigned int);
  ExpFunc(const ExpFunc&);
  virtual ~ExpFunc();
  ExpFunc& operator=(const ExpFunc&);
protected:
  typedef double(*Func0)(double);
  bool eval0(const Lib::Value&,Lib::Value&,std::string&,Func0);
  typedef bool(*InDomain)(double);
  bool eval0(const Lib::Value&,Lib::Value&,std::string&,Func0,InDomain);
  typedef double(*Func1)(double,double);
  bool eval1(const Lib::Value&,Lib::Value&,std::string&,const Lib::Value&,Func1);
protected:
  std::string fName;
  unsigned int fNumArgs;
};

#define LIB_EXPFUNC_HH(a_func) \
  public:\
    virtual bool eval(const Args&,Lib::Value&,std::string&);\
    virtual IExpFunc* copy() const;\
  public:\
    ExpFunc_##a_func();\
    ExpFunc_##a_func(const ExpFunc_##a_func&);\
    virtual ~ExpFunc_##a_func();\
    ExpFunc_##a_func& operator=(const ExpFunc_##a_func&);

#define LIB_EXPFUNC_H(a_func) \
  class ExpFunc_##a_func : public ExpFunc {\
    LIB_EXPFUNC_HH(a_func)\
  };

LIB_EXPFUNC_H(cos)
LIB_EXPFUNC_H(sin)
LIB_EXPFUNC_H(cosh)
LIB_EXPFUNC_H(sinh)
LIB_EXPFUNC_H(tanh)
LIB_EXPFUNC_H(atan)
LIB_EXPFUNC_H(sqrt)
LIB_EXPFUNC_H(exp)
LIB_EXPFUNC_H(fabs)
LIB_EXPFUNC_H(abs)
LIB_EXPFUNC_H(rndm)
LIB_EXPFUNC_H(toint)
LIB_EXPFUNC_H(int)

LIB_EXPFUNC_H(tan)
LIB_EXPFUNC_H(acos)
LIB_EXPFUNC_H(asin)
LIB_EXPFUNC_H(log)
LIB_EXPFUNC_H(log10)

LIB_EXPFUNC_H(nco)
LIB_EXPFUNC_H(vsum)

LIB_EXPFUNC_H(pow)
LIB_EXPFUNC_H(mod)

LIB_EXPFUNC_H(array)
LIB_EXPFUNC_H(expo)
LIB_EXPFUNC_H(pol1)

LIB_EXPFUNC_H(gauss)
LIB_EXPFUNC_H(bw)
LIB_EXPFUNC_H(breit)
LIB_EXPFUNC_H(cauchy)
LIB_EXPFUNC_H(pol2)

LIB_EXPFUNC_H(pol3)

LIB_EXPFUNC_H(pol4)

class ExpFuncs {
public:
  ExpFuncs();
  virtual ~ExpFuncs();
  ExpFuncs(const ExpFuncs&);
  ExpFuncs& operator=(const ExpFuncs&);
public:
  IExpFunc* find(const std::string&) const;
  void add(IExpFunc*);
private:
  std::vector<IExpFunc*> fFuncs;
};

}

#endif



