// this :
#include <Lib/A4s.h>

#include <Lib/mmanip.h>

typedef std::complex<double> z;

// A4 Helpers :
//////////////////////////////////////////////////////////////////////////////
bool Lib::l2(
 const A4d& aArray
,const A44d& aMetric
,double& aL2
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aL2 = 0;
  Lib::Array<double> tmp;
  if(!Lib::contract(aMetric,0,aArray,0,tmp)) return false;
  return tmp.contract(aArray,aL2);
}

//////////////////////////////////////////////////////////////////////////////
bool Lib::x0normalize(
 A4d& aArray
,const A44d& aMetric
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  // 1 = g00 * v0 * v0 + 2 * v0 * (g0i * vi) + gjk * vj * vk;
  // 0 = a * x * x + b * x + c;
  //

  double a = aMetric.value00();

  double b = 0;
  b += aMetric.value01() * aArray.value1();
  b += aMetric.value02() * aArray.value2();
  b += aMetric.value03() * aArray.value3();
  b *= 2;

  double c = 0;
  c += aMetric.value11() * aArray.value1() * aArray.value1();
  c += aMetric.value12() * aArray.value1() * aArray.value2();
  c += aMetric.value13() * aArray.value1() * aArray.value3();

  c += aMetric.value21() * aArray.value2() * aArray.value1();
  c += aMetric.value22() * aArray.value2() * aArray.value2();
  c += aMetric.value23() * aArray.value2() * aArray.value3();

  c += aMetric.value31() * aArray.value3() * aArray.value1();
  c += aMetric.value32() * aArray.value3() * aArray.value2();
  c += aMetric.value33() * aArray.value3() * aArray.value3();
  c -= 1;

  double xm,xp;
  int n = mmanip::solve_equation_2_degree(a,b,c,xm,xp);
  if(n==0) return false;

  if(n==1) {
    aArray.setValue0(xm);
  } else {
    aArray.setValue0(xp);
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
/// Some common matrices /////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Lib::A44d_I::A44d_I(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue00(1);
  setValue11(1);
  setValue22(1);
  setValue33(1);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44d_eta::A44d_eta(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue00( 1);
  setValue11(-1);
  setValue22(-1);
  setValue33(-1);
}
//   0  0  0  1
//   0  0 -1  0
//   0  1  0  0
//  -1  0  0  0
//////////////////////////////////////////////////////////////////////////////
Lib::A44d_J::A44d_J(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue03( 1);
  setValue12(-1);
  setValue21( 1);
  setValue30(-1);
}
//   0 -1  0  0
//   1  0  0  0
//   0  0  0  1
//   0  0 -1  0
//////////////////////////////////////////////////////////////////////////////
Lib::A44d_K::A44d_K(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue01(-1);
  setValue10( 1);
  setValue23( 1);
  setValue32(-1);
}
//  0  0  1  0       0 -1  0  0       1  0  0  0
//  0  0  0  1      -1  0  0  0       0 -1  0  0
//  1  0  0  0       0  0  0  1       0  0 -1  0
//  0  1  0  0       0  0  1  0       0  0  0  1
//////////////////////////////////////////////////////////////////////////////
Lib::A44d_T1::A44d_T1(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue02(1);
  setValue13(1);
  setValue20(1);
  setValue31(1);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44d_T2::A44d_T2(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue01(-1);
  setValue10(-1);
  setValue23( 1);
  setValue32( 1);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44d_T3::A44d_T3(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue00( 1);
  setValue11(-1);
  setValue22(-1);
  setValue33( 1);
}

// E1              E2                E3
//  0  0  0  0       0  0  0  0       0  0  0  0
//  0  0  0  0       0  0  0 -1       0  0  1  0
//  0  0  0  1       0  0  0  0       0 -1  0  0
//  0  0 -1  0       0  1  0  0       0  0  0  0

// E4              E5                E6
//  0  0  0  1       0  0  1  0       0  1  0  0
//  0  0  0  0       0  0  0  0      -1  0  0  0
//  0  0  0  0      -1  0  0  0       0  0  0  0
// -1  0  0  0       0  0  0  0       0  0  0  0
//////////////////////////////////////////////////////////////////////////////
Lib::A44d_E1::A44d_E1(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue23( 1);
  setValue32(-1);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44d_E2::A44d_E2(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue13(-1);
  setValue31( 1);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44d_E3::A44d_E3(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue12( 1);
  setValue21(-1);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44d_E4::A44d_E4(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue03( 1);
  setValue30(-1);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44d_E5::A44d_E5(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue02( 1);
  setValue20(-1);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44d_E6::A44d_E6(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue01( 1);
  setValue10(-1);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Lib::A44z_I::A44z_I(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue00(z(1,0));
  setValue11(z(1,0));
  setValue22(z(1,0));
  setValue33(z(1,0));
}
/*
  Standard Dirac matrices :
  G0               G1               G2              G3
    1  0  0  0       0  0  0  1       0  0  0 -i      0  0  1  0
    0  1  0  0       0  0  1  0       0  0  i  0      0  0  0 -1
    0  0 -1  0       0 -1  0  0       0  i  0  0     -1  0  0  0
    0  0  0 -1      -1  0  0  0      -i  0  0  0      0  1  0  0
*/
//////////////////////////////////////////////////////////////////////////////
Lib::A44z_Dirac0::A44z_Dirac0(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue00( 1);
  setValue11( 1);
  setValue22(-1);
  setValue33(-1);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44z_Dirac1::A44z_Dirac1(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue03( 1);
  setValue12( 1);
  setValue21(-1);
  setValue30(-1);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44z_Dirac2::A44z_Dirac2(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  z i(0,1);
  setValue03(-i);
  setValue12( i);
  setValue21( i);
  setValue30(-i);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44z_Dirac3::A44z_Dirac3(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue02( 1);
  setValue13(-1);
  setValue20(-1);
  setValue31( 1);
}

// Lorentz rep (1/2,1/2)
// HH1             HH2              HH3
//  0  1  1  0       0 -i -i  0       2  0  0  0
//  1  0  0  1       i  0  0 -i       0  0  0  0
//  1  0  0  1       i  0  0 -i       0  0  0  0
//  0  1  1  0       0  i  i  0       0  0  0 -2
// the uppers must be multiplied by : -i*0.5

// HH4             HH5              HH6
//  0  0  0  0       0  i -i  0       0 -1  1  0
//  0  2  0  0      -i  0  0 -i      -1  0  0  1
//  0  0 -2  0       i  0  0  i       1  0  0 -1
//  0  0  0  0       0  i -i  0       0  1 -1  0
// the uppers must be multiplied by : 0.5

//////////////////////////////////////////////////////////////////////////////
Lib::A44z_HH1::A44z_HH1(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  z i(0,1);
  setValue01( 1);
  setValue02( 1);
  setValue10( 1);
  setValue13( 1);
  setValue20( 1);
  setValue23( 1);
  setValue31( 1);
  setValue32( 1);
  multiply(-i*0.5);
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44z_HH2::A44z_HH2(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  z i(0,1);
  setValue01(-i);
  setValue02(-i);
  setValue10( i);
  setValue13(-i);
  setValue20( i);
  setValue23(-i);
  setValue31( i);
  setValue32( i);
  multiply(-i*0.5);  
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44z_HH3::A44z_HH3(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  z i(0,1);
  setValue00( 2);
  setValue11( 0);
  setValue22( 0);
  setValue33(-2);
  multiply(-i*0.5);  
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44z_HH4::A44z_HH4(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  setValue00( 0);
  setValue11( 2);
  setValue22(-2);
  setValue33( 0);
  multiply(0.5);  
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44z_HH5::A44z_HH5(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  z i(0,1);
  setValue01( i);
  setValue02(-i);
  setValue10(-i);
  setValue13(-i);
  setValue20( i);
  setValue23( i);
  setValue31( i);
  setValue32(-i);
  multiply(0.5);  
}
//////////////////////////////////////////////////////////////////////////////
Lib::A44z_HH6::A44z_HH6(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  z i(0,1);
  setValue01(-1);
  setValue02( 1);
  setValue10(-1);
  setValue13( 1);
  setValue20( 1);
  setValue23(-1);
  setValue31( 1);
  setValue32(-1);
  multiply(0.5);  
}

//////////////////////////////////////////////////////////////////////////////
/// To test compilation.
//////////////////////////////////////////////////////////////////////////////

void Lib_A4_test() {
  Lib::A4d a4;
  Lib::A44d a44;
  Lib::A444d a444;

  Lib::A44<double> aa1;
  Lib::A44<double> aa2;
  aa1.invert(aa2);
  Lib::A44<double> tmp;
  aa1.mxmul(aa2,tmp);

  Lib::Array<double> a;
  Lib::contract(a44,0,a4,0,a);
  Lib::swap(a44,0,1,tmp);
  Slash::Core::IWriter* printer = 0;
  Lib::diff(*printer,a44,tmp,0.);
}


