%{

// this :
#include <Lib/ColumnsLexer.h>

#include <Slash/Core/IWriter.h>
#include <Lib/Value.h>
#include <Lib/ColumnsToken.h>
#include <Lib/Out.h>
#include <Lib/sout.h>

#include <string.h>

#ifdef WIN32
#include <io.h>
#define isatty _isatty
#else
#include <unistd.h> //isatty
#endif

#define _UNISTD_H_
#define YY_NEVER_INTERACTIVE 1
#undef YY_INPUT
#define YY_INPUT(aBuffer,aResult,aMaxSize) (aResult=Lib::ColumnsLexer::input(aBuffer,aMaxSize)) 
#define NotFound (-1)

static Lib::ColumnsLexer* sLexer = 0;

%}

/* signus handled in by the parser :
number [-+]?([0-9]+|(([0-9]+\.[0-9]*)|([0-9]*\.[0-9]+))([eE][-+]?[0-9]+)?)
integer [0-9]+
*/
integer [0-9]+
hexaint 0x[0-9a-fA-F]+
number (([0-9]+\.[0-9]*)|([0-9]*\.[0-9]+))([eE][-+]?[0-9]+)?
name   [a-zA-Z_][a-zA-Z0-9_\-.]*
string (\"[^\"\n]*\")|('[^'\n]*')

%%


[\t ]+ /* ignore whitespace */;

{hexaint} { 
  if(sLexer->verboseLevel()) {
    Lib::Out out(sLexer->printer());
    out << "Lib::ColumnLexers::parse : "
        << " hexaint : " << Lib::sout(yytext) 
        << Lib::endl;
  }
  if(sscanf(yytext,"0x%lx",&(Columns_lval.fInteger))!=1) {
    Columns_lval.fInteger = 0;
  }
  return NUMBER_INTEGER;
}

{integer} { 
  char* s;
  if(sLexer->verboseLevel()) {
    Lib::Out out(sLexer->printer());
    out << "Lib::ColumnLexers::parse : "
        << " integer : " << Lib::sout(yytext) 
        << Lib::endl;
  }
  Columns_lval.fInteger = strtol(yytext,&s,10);
  if(s==yytext) {
    Columns_lval.fInteger = 0;
  }
  return NUMBER_INTEGER;
}

{number} {
  char* s;
  if(sLexer->verboseLevel()) {
    Lib::Out out(sLexer->printer());
    out << "Lib::ColumnLexers::parse : "
        << " number : " << Lib::sout(yytext) 
        << Lib::endl;
  }
  Columns_lval.fNumber = strtod(yytext,&s);
  if(s==yytext) {
    Columns_lval.fNumber = 0.;
  }
  return NUMBER_REAL;
}

true return T_TRUE;
TRUE return T_TRUE;
True return T_TRUE;
false return T_FALSE;
FALSE return T_FALSE;
False return T_FALSE;

or return T_OR;
OR return T_OR;
and return T_AND;
AND return T_AND;

short return T_SHORT;
int return T_INT;
long return T_INT64;
float return T_FLOAT;
double return T_DOUBLE;
boolean return T_BOOLEAN;
byte return T_BYTE;
char return T_CHAR;
string return T_STRING;
java.lang.String return T_JAVA_STRING;
ITuple return T_TUPLE;

{name} {
  if(sLexer->verboseLevel()) {
    Lib::Out out(sLexer->printer());
    out << "Lib::ColumnLexers::parse : "
        << " name : " << Lib::sout(yytext) 
        << Lib::endl;
  }
  sLexer->variables().push_back(yytext);
  Columns_lval.fIndex = sLexer->variables().size()-1;
  return T_NAME;
}

{string} {
  if(sLexer->verboseLevel()) {
    Lib::Out out(sLexer->printer());
    out << "Lib::ColumnLexers::parse : "
        << " string : " << Lib::sout(yytext) 
        << Lib::endl;
  }
  // Remove leading and trailing double quote :
  // "abcd"
  // 012345
  std::string s = yytext;
  sLexer->variables().push_back(s.substr(1,s.length()-2)); 
  int index = sLexer->variables().size()-1;
  Columns_lval.fIndex = index;
  return STRG;
}

. return yytext[0];

%%

Lib::ColumnsLexer::ColumnsLexer(Slash::Core::IWriter& aPrinter,int aVerboseLevel)
:fVerboseLevel(aVerboseLevel)
,fCurrent(0)
,fEnd(0)
,fPrinter(aPrinter)
{}

int Lib::ColumnsLexer::verboseLevel() const { return fVerboseLevel;}

Lib::ColumnsLexer& Lib::ColumnsLexer::operator=(const ColumnsLexer& aColumnsLexer){
  fVerboseLevel = aColumnsLexer.fVerboseLevel;
  // Reset other things :
  fString = "";
  fCurrent = 0;
  fEnd = 0;
  fVariables.clear();
  return *this;
}

void Lib::ColumnsLexer::set(const std::string& aString) {
  fString = aString;
  sLexer = 0;
  fCurrent = (char*)fString.c_str();
  fEnd = fCurrent + ::strlen(fCurrent);
  sLexer = this;
}

Slash::Core::IWriter& Lib::ColumnsLexer::printer() const {return fPrinter;}

std::vector<std::string>& Lib::ColumnsLexer::variables(){ return fVariables;}

std::string Lib::ColumnsLexer::variable(unsigned int aIndex) const {
  return fVariables[aIndex];
}

int Lib::ColumnsLexer::input(char* aBuffer,int aMaxSize) {
  if(!sLexer) return 0;
  if(aMaxSize<=0) return 0;
  int rest = sLexer->fEnd - sLexer->fCurrent;
  int n = aMaxSize>rest ? rest : aMaxSize;
  if(n>0) {
    memcpy(aBuffer,sLexer->fCurrent,n);
    sLexer->fCurrent += n;
  }
  return n;
}

int yywrap(){return 1;}

