// this :
#include <Lib/Dictionary.h>

#include <Slash/Meta/IClass.h>

#include <Lib/Debug.h>
#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/Cast.h>

namespace Lib {

class None_Class : public virtual Slash::Meta::IClass {
public://Slash::Meta::IClass
  virtual Slash::Meta::IObject* create() {return 0;}
  virtual std::string name() const { return fName;}
  virtual unsigned int version() const { return fVersion;}
public:
  None_Class():fName("Lib::None"),fVersion(1){}
  virtual ~None_Class(){}
private:
  std::string fName;
  unsigned int fVersion;
};

}

//////////////////////////////////////////////////////////////////////////////
Lib::Class::Class(
 Slash::Core::IWriter& aPrinter
,const std::string& aName
,unsigned int aVersion
,Allocator aAllocator
) 
:fPrinter(aPrinter)
,fName(aName)
,fVersion(aVersion)
,fAllocator(aAllocator)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("Lib::Class");
}
//////////////////////////////////////////////////////////////////////////////
Lib::Class::~Class(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::decrement("Lib::Class");
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::Class::name(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fName;
}
//////////////////////////////////////////////////////////////////////////////
unsigned int Lib::Class::version(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fVersion;
}
//////////////////////////////////////////////////////////////////////////////
Slash::Meta::IObject* Lib::Class::create(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fAllocator) return 0;
  return (Slash::Meta::IObject*)fAllocator(fPrinter);
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Lib::Dictionary::Dictionary(
 Slash::Core::IWriter& aPrinter
) 
:fName("Lib::Dictionary")
,fPrinter(aPrinter)
,fNoneClass(new None_Class())
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("Lib::Dictionary");
}
//////////////////////////////////////////////////////////////////////////////
Lib::Dictionary::~Dictionary(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<Slash::Meta::IClass*>::iterator it;
  for(it=fClasses.begin();it!=fClasses.end();++it) delete (*it);
  fClasses.clear();

  delete fNoneClass;

  Lib::Debug::decrement("Lib::Dictionary");
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::Dictionary::name(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fName;
}
//////////////////////////////////////////////////////////////////////////////
void* Lib::Dictionary::cast(
 const std::string& aClass
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  if_Lib_SCast(Slash::Core::IManager)
  else if_Lib_SCast(Slash::Meta::IDictionary)
  else if_Lib_SCast(Lib::Dictionary)
  else return 0;
}
//////////////////////////////////////////////////////////////////////////////
Slash::Meta::IClass* Lib::Dictionary::find(
 const std::string& aName
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<Slash::Meta::IClass*>::const_iterator it;
  for(it=fClasses.begin();it!=fClasses.end();++it) {
    if((*it)->name()==aName) return *it;
  }
  Lib::Out out(fPrinter);
  out << "Lib::Dictionary::findClass :" 
      << " " << Lib::sout(aName) << " not found." 
      << Lib::endl;
  return fNoneClass;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Dictionary::add(
 Slash::Meta::IClass* aClass                                
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fClasses.push_back(aClass);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
unsigned int Lib::Dictionary::numberOfClasses(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fClasses.size();
}
//////////////////////////////////////////////////////////////////////////////
const Slash::Meta::IClass* Lib::Dictionary::getClass(
 unsigned int aIndex
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fClasses[aIndex];
}
