// this :
#include <Lib/FileReader.h>

#include <Lib/Out.h>
#include <Lib/Debug.h>

#include <stdio.h>

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Lib::FileReader::FileReader(
 const std::string& aFileName
,Slash::Core::IWriter& aPrinter
)
:fPrinter(aPrinter)
,fFileName(aFileName)
,fFILE(0)
,fOwner(true)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("Lib::FileReader");
  fFILE = ::fopen(aFileName.c_str(),"rb");
}
//////////////////////////////////////////////////////////////////////////////
Lib::FileReader::FileReader(
 void* aFILE //FILE*
,Slash::Core::IWriter& aPrinter
)
:fPrinter(aPrinter)
,fFILE(aFILE)
,fOwner(false)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("Lib::FileReader");
}
//////////////////////////////////////////////////////////////////////////////
Lib::FileReader::~FileReader(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fOwner && fFILE) ::fclose((FILE*)fFILE);
  Lib::Debug::decrement("Lib::FileReader");
}
//////////////////////////////////////////////////////////////////////////////
void Lib::FileReader::disable(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{  
  if(!fFILE) return; // Already closed.
  if(fOwner) ::fclose((FILE*)fFILE);
  fFILE = 0;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::FileReader::enable(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fFILE) return; // Already open.
  if(fOwner) fFILE = ::fopen(fFileName.c_str(),"rb");
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::FileReader::enabled(
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fFILE ? true : false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::FileReader::read(
 void* aBuffer
,unsigned int aLength
,int& aReadLength
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aReadLength = 0;
  if(!fFILE) return false;
  size_t len = ::fread(aBuffer,1,aLength,(FILE*)fFILE);
  aReadLength = (size_t)len;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::FileReader::rewind(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fFILE) return false;
  ::rewind((FILE*)fFILE);
  return true;
}

#include <string.h>

//////////////////////////////////////////////////////////////////////////////
bool Lib::FileReader::read(
 void* aBuffer
,unsigned int aLength
,std::string& aString
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fFILE) {
    aString.resize(0);
    return false;
  }

  char* buffer = (char*)aBuffer;
  if(::fgets(buffer,aLength,(FILE*)fFILE)==NULL) {
    disable();
    aString.resize(0);
    return false; //EOF
  }

  size_t l = ::strlen(buffer);
  //  On Windows, editors when saving binary files,
  // put \r\n at place of \n ; we then look for \r\n.
  if( (l>=2) && (buffer[l-2]=='\r') && (buffer[l-1]=='\n') ) {
    buffer[l-2] = '\0';
  } else if( (l>=1) && (buffer[l-1]=='\n') ) {
    buffer[l-1] = '\0';
  }

  aString = buffer;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
bool Lib::FileReader::sread(
 void* aBuffer
,unsigned int aLength
,unsigned int& aStringLength
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fFILE) {
    aStringLength = 0;
    return false;
  }

  char* buffer = (char*)aBuffer;
  if(::fgets(buffer,aLength,(FILE*)fFILE)==NULL) {
    disable();
    aStringLength = 0;
    return false; //EOF
  }

  size_t l = ::strlen(buffer);
  //  On Windows, editors when saving binary files,
  // put \r\n at place of \n ; we then look for \r\n.
  if( (l>=2) && (buffer[l-2]=='\r') && (buffer[l-1]=='\n') ) {
    buffer[l-2] = '\0';
    l -= 2;
  } else if( (l>=1) && (buffer[l-1]=='\n') ) {
    buffer[l-1] = '\0';
    l -= 1;
  }

  aStringLength = l;
  return true;
}

