%{

// this :
#include <Lib/FortranFormatLexer.h>

#include <Slash/Core/IWriter.h>
#include <Lib/FortranFormatToken.h>
#include <Lib/Out.h>
#include <Lib/sout.h>

#include <string.h>

#ifdef WIN32
#include <io.h>
#define isatty _isatty
#else
#include <unistd.h> //isatty
#endif

#define _UNISTD_H_
#define YY_NEVER_INTERACTIVE 1
#undef YY_INPUT
#define YY_INPUT(aBuffer,aResult,aMaxSize) (aResult=Lib::FortranFormatLexer::input(aBuffer,aMaxSize)) 
#define NotFound (-1)

static Lib::FortranFormatLexer* sLexer = 0;

%}

integer [0-9]+

exponent E[0-9]+

%%

[\t ]+ /* ignore whitespace */;

{integer} { 
  char* s;
  if(sLexer->verboseLevel()) {
    Lib::Out out(sLexer->printer());
    out << "Lib::FortranFormatLexer::parse :"
        << " integer : " << Lib::sout(yytext) 
        << Lib::endl;
  }
  FortranFormat_lval.fInteger = ::strtol(yytext,&s,10);
  if(s==yytext) {
    FortranFormat_lval.fInteger = 0;
  }
  return NUMBER_INTEGER;
}

{exponent} { 
  char* s;
  if(sLexer->verboseLevel()) {
    Lib::Out out(sLexer->printer());
    out << "Lib::FortranFormatLexer::parse :"
        << " exponent : " << Lib::sout(yytext) 
        << Lib::endl;
  }
  FortranFormat_lval.fInteger = ::strtol(yytext+1,&s,10);
  if(s==yytext) {
    FortranFormat_lval.fInteger = 0;
  }
  return NUMBER_EXPONENT;
}

X return T_X;
x return T_X;
I return T_I;
i return T_I;
F return T_F;
f return T_F;
E return T_E;
e return T_E;
G return T_G;
g return T_G;
A return T_A;
a return T_A;
L return T_L;
l return T_L;
\/ return T_SLASH;

. return yytext[0];

%%
//////////////////////////////////////////////////////////////////////////////
Lib::FortranFormatLexer::FortranFormatLexer(
 Slash::Core::IWriter& aPrinter
,int aVerboseLevel
)
:fVerboseLevel(aVerboseLevel)
,fCurrent(0)
,fEnd(0)
,fPrinter(aPrinter)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
int Lib::FortranFormatLexer::verboseLevel(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fVerboseLevel;
}
//////////////////////////////////////////////////////////////////////////////
Lib::FortranFormatLexer::FortranFormatLexer(
 const FortranFormatLexer& aFrom
)
:fPrinter(aFrom.fPrinter)
//////////////////////////////////////////////////////////////////////////////
// private.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
Lib::FortranFormatLexer& Lib::FortranFormatLexer::operator=(
 const FortranFormatLexer&
)
//////////////////////////////////////////////////////////////////////////////
// private.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return *this;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::FortranFormatLexer::set(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fString = aString;
  sLexer = 0;
  fCurrent = (char*)fString.c_str();
  fEnd = fCurrent + ::strlen(fCurrent);
  sLexer = this;
}
//////////////////////////////////////////////////////////////////////////////
Slash::Core::IWriter& Lib::FortranFormatLexer::printer(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fPrinter;
}
//////////////////////////////////////////////////////////////////////////////
int Lib::FortranFormatLexer::input(
 char* aBuffer
,int aMaxSize
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!sLexer) return 0;
  if(aMaxSize<=0) return 0;
  int rest = sLexer->fEnd - sLexer->fCurrent;
  int n = aMaxSize>rest ? rest : aMaxSize;
  if(n>0) {
    ::memcpy(aBuffer,sLexer->fCurrent,n);
    sLexer->fCurrent += n;
  }
  return n;
}
//////////////////////////////////////////////////////////////////////////////
int yywrap(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return 1;
}

