// this :
#include <Lib/FunctionManager.h>

#include <Slash/Core/ISession.h>

#include <Lib/Out.h>
#include <Lib/Debug.h>
#include <Lib/smanip.h>
#include <Lib/Cast.h>
#include <Lib/Functions.h>
#include <Lib/fmanip.h>
#include <Lib/sout.h>
#include <Lib/valuemanip.h>

#include <stdio.h>

#define NotFound (-1)

//////////////////////////////////////////////////////////////////////////////
bool Lib::FunctionManager::find_f77_functions(
 const std::string& aFile
,std::vector<std::string>& aNames
,std::vector<FunctionArgs>& aArgs
,Slash::Core::IWriter& aPrinter
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aNames.clear();
  aArgs.clear();
  std::vector<std::string> lines;
  if(!Lib::fmanip::read(aFile,lines)) {
    Lib::Out out(aPrinter);
    out << "Lib::FunctionManager::find_f77_functions :" 
        << " can't read file " << Lib::sout(aFile) << "."
        << Lib::endl;
    return false;
  }
  unsigned int linen = lines.size();
  bool inFunction = false;
  for(unsigned int linei = 0;linei<linen;linei++) {
    std::string line = lines[linei];
    Lib::smanip::touppercase(line);
    std::vector<std::string> words;
    Lib::smanip::words(line," ",words);
    unsigned int index;
    if( Lib::smanip::lineIndex(words,"FUNCTION",index)   ||
        Lib::smanip::lineIndex(words,"SUBROUTINE",index) ){
      if(inFunction) { //Syntax error.
        Lib::Out out(aPrinter);
        out << "Lib::FunctionManager::find_f77_functions :" 
            << " In " << Lib::sout(line) << ","
            << " [FUNCTION,SUBROUTINE] / END mismatch."
            << Lib::endl;
        aNames.clear();
        aArgs.clear();
        return false;
      } 
      // For example : REAL FUNCTION FUNC(I)
      // For example : FUNCTION FUNC(I)
      if((index!=0)&&(index!=1)) { //Syntax error. 
        Lib::Out out(aPrinter);
        out << "Lib::FunctionManager::find_f77_functions :" 
            << " In " << Lib::sout(line) << ","
            << " [FUNCTION,SUBROUTINE] is not the first or second word."
            << Lib::endl;
        aNames.clear();
        aArgs.clear();
        return false;
      }
      if((index+1)>=words.size()) { //Syntax error.
        Lib::Out out(aPrinter);
        out << "Lib::FunctionManager::find_f77_functions :" 
            << " In " << Lib::sout(line) << ","
            << " No space expected between func name and arguments."
            << Lib::endl;
        aNames.clear();
        aArgs.clear();
        return false;
      }
      std::string name;
      std::vector<std::string> sargs;
      if(!Lib::smanip::nameArgs(words[index+1],name,sargs)) {
        Lib::Out out(aPrinter);
        out << "Lib::FunctionManager::find_f77_functions :" 
            << " In " << Lib::sout(line) << ","
            << " syntax error in arguments."
            << Lib::endl;
        aNames.clear();
        aArgs.clear();
        return false;
      }
      unsigned int sargn = sargs.size();
      std::vector<Lib::Value> args(sargn);
      for(unsigned int argi=0;argi<sargn;argi++) {
	std::string s = sargs[argi];
        Lib::smanip::strip(s);
        if(!Lib::smanip::isVariable(s)) {
          Lib::Out out(aPrinter);
          out << "Lib::FunctionManager::find_f77_functions :" 
              << " In " << Lib::sout(line) << ","
              << " " << Lib::sout(s) << " has not the syntax of a variable."
              << Lib::endl;
          aNames.clear();
          aArgs.clear();
          return false;
        }
        Lib::Value& v = args[argi];
        v.setLabel(s);
        v.setDimension(1);
        char c = s[0];
        if( (c>='I')&&(c<='N') ) {
          v.setType(Slash::Core::IValue::INT_STAR);
        } else {
          v.setType(Slash::Core::IValue::FLOAT_STAR);
        }
        //WARNING : can't have a DOUBLE_STAR here, else
        //          semantic clash with the VECTOR logic.
      }
      aNames.push_back(name);
      aArgs.push_back(args);
      inFunction = true;

    } else if(Lib::smanip::lineIndex(words,"END",index)) {
      if(!inFunction) { //Syntax error.
        Lib::Out out(aPrinter);
        out << "Lib::FunctionManager::find_f77_functions :" 
            << " In " << Lib::sout(line) << ","
            << " [FUNCTION,SUBROUTINE] / END mismatch."
            << Lib::endl;
        aNames.clear();
        aArgs.clear();
        return false;
      }
      if(index!=0) { //Syntax error.
        Lib::Out out(aPrinter);
        out << "Lib::FunctionManager::find_f77_functions :" 
            << " In " << Lib::sout(line) << ","
            << " END is not the first word."
            << Lib::endl;
        aNames.clear();
        aArgs.clear();
        return false;
      }
      inFunction = false;
    } else if(inFunction) {
      // Search argument declaration :
      if( Lib::smanip::lineIndex(words,"DIMENSION",index) ||
          Lib::smanip::lineIndex(words,"VECTOR",index) ){
        if((index+1)>=words.size()) { //Syntax error.
          Lib::Out out(aPrinter);
          out << "Lib::FunctionManager::find_f77_functions :" 
              << " In " << Lib::sout(line) << ","
              << " [DIMENSION,VECTOR] is not the previous last word."
              << Lib::endl;
          aNames.clear();
          aArgs.clear();
          return false;
        }
        if(index!=0) { //Syntax error.
          Lib::Out out(aPrinter);
          out << "Lib::FunctionManager::find_f77_functions :" 
              << " In " << Lib::sout(line) << ","
              << " [DIMENSION,VECTOR] is not the first word."
              << Lib::endl;
          aNames.clear();
          aArgs.clear();
          return false;
        }
        bool isVector = (words[index]=="VECTOR");
        std::string s = words[index+1];
        std::vector<std::string> words2;
        Lib::smanip::words(s,",",words2);
        std::vector<Lib::Value>& lastArgs = aArgs[aArgs.size()-1];
        unsigned int wordn2 = words2.size();
        for(unsigned int vari=0;vari<wordn2;vari++) {
          const std::string& VNAME = words2[vari];
   	  std::string NAME;
          int dim;
          if(!Lib::smanip::parseArray(VNAME,NAME,dim)) {
            Lib::Out out(aPrinter);
            out << "Lib::FunctionManager::find_f77_functions :" 
                << " In " << Lib::sout(line) << ","
                << " array syntax error in " << Lib::sout(VNAME) << "."
                << Lib::endl;
            aNames.clear();
            aArgs.clear();
            return false;
          }
          Lib::smanip::strip(NAME);
          if(!Lib::smanip::isVariable(NAME)) {
            Lib::Out out(aPrinter);
            out << "Lib::FunctionManager::find_f77_functions :" 
                << " In " << Lib::sout(line) << ","
                << " " << Lib::sout(NAME) 
                << " has not the syntax of a variable."
                << Lib::endl;
            aNames.clear();
            aArgs.clear();
            return false;
          }
          if(isVector) {
  	    Lib::Value v;
            v.setType(Slash::Core::IValue::DOUBLE_STAR);
            v.setLabel(NAME);
            if(dim!=NotFound) v.setDimension(dim); //Else 0. Used in pafort.f.
            lastArgs.push_back(v);
          } else {
            if(dim==NotFound) {
              Lib::Out out(aPrinter);
              out << "Lib::FunctionManager::find_f77_functions :" 
                  << " In " << Lib::sout(line) << ","
                  << " variable " << Lib::sout(VNAME) << " is not an array."
                  << Lib::endl;
              aNames.clear();
              aArgs.clear();
              return false;
            }

            //printf("debug : %s %s\n",NAME.c_str(),DIM.c_str());
            bool found = false;
            unsigned int argn = lastArgs.size();
            for(unsigned int argi=0;argi<argn;argi++) {
              if(lastArgs[argi].label()==NAME) {
     	        Lib::Value& v = lastArgs[argi];
                v.setDimension((int)dim);
                found = true;
                break;
              }
            }
            if(!found) {
              Lib::Out out(aPrinter);
              out << "Lib::FunctionManager::find_f77_functions :" 
                  << " In " << Lib::sout(line) << ","
                  << " " << Lib::sout(NAME) 
                  << " not found in argument list."
                  << Lib::endl;
              aNames.clear();
              aArgs.clear();
              return false;
            }
          }
        }
      }
    }
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::FunctionManager::edit_f77_file(
 const std::string& aFile
,const std::string& aNewFile
,const std::vector<std::string>& aNames
,const std::vector<FunctionArgs>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
// Create  modified FORTRAN file :
//   REAL FUNCTION FUNC(X)
//   VECTOR A,B(1) 
// changed to :
//   REAL FUNCTION FUNC(X,A,B)
//   REAL A,B(1)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> lines;
  if(!Lib::fmanip::read(aFile,lines)) return false;

  FILE* out_file = ::fopen(aNewFile.c_str(),"wb");
  if(!out_file) return false;

  unsigned funci = 0;

  unsigned int linen = lines.size();
  bool inFunction = false;
  for(unsigned int linei = 0;linei<linen;linei++) {
    std::string line = lines[linei];

    Lib::smanip::touppercase(line);

    std::vector<std::string> words;
    Lib::smanip::words(line," ",words);

    unsigned int index;
    if( Lib::smanip::lineIndex(words,"FUNCTION",index)   ||
        Lib::smanip::lineIndex(words,"SUBROUTINE",index) ){

      // Assume that the line syntax is correct :
      //    <REAL,INTEGER> FUNCTION <name>(<args>)
      // Add extra arguments (coming from VECTOR) at end
      // of arguments.

      std::string new_line;

      std::string::size_type lb = line.find('(');
      if(lb==std::string::npos) {
        new_line = lines[linei];
      } else {
        new_line = lines[linei].substr(0,lb);
      }
      new_line += "(";

     {const std::vector<Lib::Value>& args = aArgs[funci];
      unsigned int argn = args.size();
      for(unsigned int argi=0;argi<argn;argi++) {
        if(argi) new_line += ",";
        new_line += args[argi].label();
      }}
 
      new_line += ")";

      ::fprintf(out_file,"%s\n",new_line.c_str());
      funci++;

    } else if(Lib::smanip::lineIndex(words,"VECTOR",index)) {

      if((index+1)>=words.size()) {
        return false;
      }

                           //01234567
      std::string new_line = "      REAL*8 ";

      std::string s = words[index+1];
      std::vector<std::string> words2;
      Lib::smanip::words(s,",",words2);
      unsigned int wordn2 = words2.size();
      for(unsigned int vari=0;vari<wordn2;vari++) {
        const std::string& VNAME = words2[vari];
        std::string NAME;
        int dim;
        if(!Lib::smanip::parseArray(VNAME,NAME,dim)) {
          return false;
        }
        Lib::smanip::strip(NAME);
        if(vari) new_line += ",";
        new_line += NAME;
        new_line += "(*)";
      }

      ::fprintf(out_file,"%s\n",new_line.c_str());

    } else {

      ::fprintf(out_file,"%s\n",lines[linei].c_str());

    }
  }

  ::fclose(out_file);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Lib::FunctionManager::FunctionManager(
 Slash::Core::ISession& aSession
)
:fSession(aSession)
,fName("FunctionManager")
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
Lib::FunctionManager::~FunctionManager(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  // Delete in reverse order :
  std::list<Slash::Data::IFunction*>::iterator it;
  while(!fList.empty()) {
    it = fList.end();
    it--;
    Slash::Data::IFunction* entry  = *it;
    fList.erase(it);
    delete entry;
  }
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::FunctionManager::name(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fName;
}
//////////////////////////////////////////////////////////////////////////////
void* Lib::FunctionManager::cast(
 const std::string& aClass
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if_Lib_SCast(Slash::Data::IFunctionManager)
  else if_Lib_SCast(Slash::Core::IManager)
  else return 0;
}
//////////////////////////////////////////////////////////////////////////////
Slash::Data::IFunction* Lib::FunctionManager::addFunction(
 const std::string& aFile
,const std::string& aModel
,void* aAddress
,const std::vector<Lib::Value>& aArgs
,unsigned int aParn
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aModel.empty()) return 0;
  Slash::Data::IFunction* entry = findFunction(aModel);
  if(entry) return entry;
  entry = 
    new Lib::Function::Compiled
      (fSession.printer(),aFile,aModel,aAddress,aArgs,aParn);
  fList.push_back(entry);
  return entry;
}
//////////////////////////////////////////////////////////////////////////////
Slash::Data::IFunction* Lib::FunctionManager::findFunction(
 const std::string& aModel
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::list<Slash::Data::IFunction*>::iterator it;
  for(it=fList.begin();it!=fList.end();++it) {
    if(aModel==(*it)->model()) return (*it);
  }
  return 0;
}

//////////////////////////////////////////////////////////////////////////////
bool Lib::FunctionManager::load(
 const std::string& aFile
,const std::string& aBuilder
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Core::IWriter& printer = fSession.printer();

  if(!fmanip::exists(aFile)) {
    Lib::Out out(printer);
    out << "Lib::FunctionManager::load :" 
        << " file " << Lib::sout(aFile) << " not found."
        << Lib::endl;
    return false;
  }

  if(fmanip::is_f77(aFile)) return loadF77(aFile,aBuilder);

  if(fmanip::is_cpp(aFile)) return loadC(aFile,aBuilder);

  return false;
}

#include <Lib/System.h>
#include <Lib/DLL.h>
#include <Lib/Manager.h>
#include <Slash/Core/ILibraryManager.h>
#include <Slash/Core/ILibrary.h>

#define NotFound (-1)

typedef bool (*DLLBuilder)(const std::string&,const std::string&,const std::vector<std::string>&,const std::vector<Lib::FunctionManager::FunctionArgs>&,const std::string&,Slash::Core::IWriter&);

//////////////////////////////////////////////////////////////////////////////
static bool get_procs(
 Slash::Core::ISession& aSession
,const std::string& aDLL
,const std::string& aSuffix
,const std::string& aPrefix
,const std::vector<std::string>& aFuncs
,const std::vector<Lib::FunctionManager::FunctionArgs>& aArgs
,std::vector<Slash::Core::ILibrary::Procedure>& aProcs
,const std::string& aBuilder
,DLLBuilder aDLLBuilder
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aProcs.clear();

  Slash::Core::IWriter& printer = aSession.printer();
  Lib::Out out(printer);

  Slash::Core::ILibraryManager* libraryManager = 
    Lib_findManager(aSession,"LibraryManager",Slash::Core::ILibraryManager);
  if(!libraryManager) {
    out << "get_procs :"
        << " can't find manager " << Lib::sout("LibraryManager") 
        << Lib::endl;
    return false;
  }

  Slash::Core::ILibrary* library = libraryManager->findLibrary(aDLL);
  if(!library) {
    library = libraryManager->addLibrary(aDLL,aDLL,"",true); //true = quiet.
    if(!library) { //DLL not found.
      if(!aDLLBuilder(aDLL,aSuffix,aFuncs,aArgs,aBuilder,printer)) 
        return false;
      library = libraryManager->addLibrary(aDLL,aDLL,"");
    }
  }
  if(!library) return false;

  unsigned int funcn = aFuncs.size();
  for(unsigned int funci=0;funci<funcn;funci++) {
    const std::string& funcName = aFuncs[funci];
    std::string CNAME = aPrefix + funcName;
    Slash::Core::ILibrary::Procedure proc = library->findProcedure(CNAME);
    if(proc) {
      aProcs.push_back(proc);
      if(aSession.verboseLevel())
        out << "get_C_function :"
            << " function " << Lib::sout(funcName) 
            << " found in " << Lib::sout(aDLL) 
            << Lib::endl;
    } else {
      if(aSession.verboseLevel())
        out << "get_C_function :"
            << " function " << Lib::sout(funcName) 
            << " not found in " << Lib::sout(aDLL) 
            << Lib::endl;
      aProcs.clear();
      return false;
    }    
  }

  return true;
}

static bool build_C_dll(const std::string&,const std::string&,const std::vector<std::string>&,const std::vector<Lib::FunctionManager::FunctionArgs>&,const std::string&,Slash::Core::IWriter&);

//////////////////////////////////////////////////////////////////////////////
bool Lib::FunctionManager::loadC(
 const std::string& aFile
,const std::string& aBuilder
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Core::IWriter& printer = fSession.printer();
  std::vector<std::string> funcs;
  std::vector<FunctionArgs> args;
  std::vector<unsigned int> parns;
  if(!find_cpp_functions(aFile,funcs,args,parns)) {
    Lib::Out out(printer);
    out << "Lib::FunctionManager::loadC :" 
        << " syntax problem in file " << Lib::sout(aFile) << "."
        << Lib::endl;
    return false;
  }
  if(!funcs.size()) {
    Lib::Out out(printer);
    out << "Lib::FunctionManager::loadC :" 
        << " file " << Lib::sout(aFile) << " has not function."
        << Lib::endl;
    return false;
  }

  if(fSession.verboseLevel()) {
    Lib::Out out(printer);
    out << "Lib::FunctionManager::loadC :" 
        << " functions in file " << Lib::sout(aFile) 
        << Lib::endl;
    unsigned int funcn = funcs.size();
    for(unsigned int funci=0;funci<funcn;funci++) {
      std::string sargs;
      unsigned int argn = args[funci].size();
      for(unsigned int argi=0;argi<argn;argi++) {
        if(argi) sargs += ",";
        sargs += args[funci][argi].stype();
        sargs += " ";
        sargs += args[funci][argi].label();
        sargs += "(";
        sargs += smanip::tostring((int)args[funci][argi].dimension());
        sargs += ")";
      }
      out << "Lib::FunctionManager::loadC :" 
          << "   name : " << Lib::sout(funcs[funci]) 
	  << " args : " << Lib::sout(sargs) 
          << Lib::endl;
    }
  }

  std::string path,name,suffix;
  smanip::pathNameSuffix(aFile,path,name,suffix);
  std::string dll = smanip::nosuffix(name);
 
  std::string fullname = fmanip::name(aFile);
  fmanip::copy(fullname,name); //FIXME : avoid this copy.

  std::vector<Slash::Core::ILibrary::Procedure> procs;
  if(!get_procs(fSession,dll,suffix,"Lib_wrap_",funcs,args,procs,aBuilder,
                build_C_dll)) {
    Lib::Out out(printer);
    out << "Lib::FunctionManager::loadC :" 
        << " unable to get some C function in " << Lib::sout(dll)
        << Lib::endl;
    return false;
  }
  if(fSession.verboseLevel()) {
    Lib::Out out(printer);
    out << "Lib::FunctionManager::loadC :" 
        << " found C function in " << Lib::sout(dll)
        << Lib::endl;
  }

  unsigned int funcn = funcs.size();
  for(unsigned int funci=0;funci<funcn;funci++) {
    addFunction
      (aFile,funcs[funci],(void*)procs[funci],args[funci],parns[funci]);    
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::FunctionManager::find_cpp_functions(
 const std::string& aFile
,std::vector<std::string>& aNames
,std::vector<FunctionArgs>& aArgs
,std::vector<unsigned int>& aParns
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aNames.clear();
  aArgs.clear();
  aParns.clear();
  std::vector<std::string> lines;
  Lib::fmanip::read(aFile,lines);
  unsigned int linen = lines.size();
  for(unsigned int linei = 0;linei<linen;linei++) {
    std::string& line = lines[linei];
    std::string sfunc;

    unsigned int parn = 0;

   {std::string::size_type pos_func = line.find("/*function(");
    if(pos_func!=std::string::npos) {
      // header of the form /*function(<unsigned int>)*/
      std::string::size_type end_pos_func = line.find(")*/");
      if( (end_pos_func==std::string::npos) || (pos_func>=end_pos_func) ) {
        aNames.clear();
        aArgs.clear();
        aParns.clear();
        return false;
      }
      sfunc = line.substr(end_pos_func+3,line.size()-(end_pos_func+3));
      std::string sparn = 
        line.substr(pos_func+11,(end_pos_func-(pos_func+11)));
      if(!Lib::smanip::touint(sparn,parn)) { //Syntax error.
        aNames.clear();
        aArgs.clear();
        aParns.clear();
        return false;
      }
    }}

   {std::string::size_type pos_func = line.find("/*function*/");
    if(pos_func!=std::string::npos) {
      sfunc = line.substr(pos_func+12,line.size()-(pos_func+12));
    }}

    if(!sfunc.size()) continue; //not a header.

    std::string head;
    std::vector<std::string> sargs;
    if(!Lib::smanip::nameArgs(sfunc,head,sargs)) {
      aNames.clear();
      aArgs.clear();
      aParns.clear();
      return false;
    }
    Lib::smanip::strip(head);

    // head is of the form <type> <name>.
   {std::vector<std::string> words;
    Lib::smanip::words(head," ",words);
    if(words.size()!=2) {
      aNames.clear();
      aArgs.clear();
      aParns.clear();
      return false;
    }
    std::string name = words[1];
    Lib::smanip::strip(name);
    if(!Lib::smanip::isVariable(name)) {
      aNames.clear();
      aArgs.clear();
      aParns.clear();
      return false;
    }
    aNames.push_back(name);}

    unsigned int argn = sargs.size();

    // Arguments :
    std::vector<Lib::Value> vargs;
    for(unsigned int argi=0;argi<argn;argi++) {
      std::vector<std::string> words;
      Lib::smanip::words(sargs[argi]," ",words);
      if(words.size()!=2) { //Syntax error
        aNames.clear();
        aArgs.clear();
        aParns.clear();
        return false;
      }
      Lib::Value v;

      std::string type = words[0];
      std::string s = words[1];
      Lib::smanip::strip(s);
      // Look for <atring>[<number>]
     {std::string::size_type lbrace = s.find('[');
      std::string::size_type rbrace = s.find(']');
      if((lbrace==std::string::npos)&&(rbrace==std::string::npos)){
        if(!Lib::smanip::isVariable(s)) {
          aNames.clear();
          aArgs.clear();
          aParns.clear();
          return false;
        }
        v.setLabel(s);
        v.setDimension(0);
        if(type=="int") {
          v.setType(Slash::Core::IValue::INT);
        } else if(type=="float") {
          v.setType(Slash::Core::IValue::FLOAT);
        } else if(type=="double") {
          v.setType(Slash::Core::IValue::DOUBLE);
        } else { //Type not supported.
          aNames.clear();
          aArgs.clear();
          aParns.clear();
          return false;
        }
      } else if( (lbrace!=std::string::npos) &&
                 (rbrace!=std::string::npos) && 
                 (lbrace<rbrace)             ){
        std::string NAME = s.substr(0,lbrace);
        if(!Lib::smanip::isVariable(NAME)) {
          aNames.clear();
          aArgs.clear();
          aParns.clear();
          return false;
        }
        std::string DIM = s.substr(lbrace+1,rbrace-(lbrace+1));
        unsigned int dim = 0;
        if(DIM.empty()) { //xxx[] syntax
          dim = 0; 
        } else {
          if(!Lib::smanip::touint(DIM,dim)) { //Syntax error.
            aNames.clear();
            aArgs.clear();
            aParns.clear();
            return false;
          }
        }
        v.setLabel(NAME);
        v.setDimension(dim);
        if(type=="int") {
          v.setType(Slash::Core::IValue::INT_STAR);
        } else if(type=="float") {
          v.setType(Slash::Core::IValue::FLOAT_STAR);
        } else if(type=="double") { //VECTOR for OpenPAW.
          v.setType(Slash::Core::IValue::DOUBLE_STAR);
          //dim could be 0 to indicate OpenPAW/pafort.f 
          //to search for an existing VECTOR
        } else { //Type not supported.
          aNames.clear();
          aArgs.clear();
          aParns.clear();
          return false;
        }
      } else {  //Syntax error.
        aNames.clear();
        aArgs.clear();
        aParns.clear();
        return false;
      }}          
      vargs.push_back(v);

    } 
    aArgs.push_back(vargs);
    aParns.push_back(parn);
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool build_C_dll(
 const std::string& aName
,const std::string& aSuffix
,const std::vector<std::string>& aFuncs
,const std::vector<Lib::FunctionManager::FunctionArgs>& aArgs
,const std::string& aBuilder
,Slash::Core::IWriter& aPrinter
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string CNAME = aName;
  CNAME = std::string("Lib_wrap_") + CNAME;
  
  // Produce C frontend file :
  std::string fileName = CNAME+".cpp";
  FILE* file = ::fopen(fileName.c_str(),"wb");
  if(!file) {
    Lib::Out out(aPrinter);
    out << "build_C_dll :" 
        << " can't create file " << Lib::sout(fileName)
        << Lib::endl;
    return false;
  }

  ::fprintf(file,"\
#include <Slash/Core/IValue.h>\n\
extern \"C\" {\n\
");


#ifdef WIN32
  std::string s_declspec = "_declspec(dllexport) ";
#else 
  std::string s_declspec = "";
#endif

  std::string sh = "    ";

  unsigned int funcn = aFuncs.size();
  for(unsigned int funci=0;funci<funcn;funci++) {
    std::string funcName = aFuncs[funci];
    unsigned int argn = aArgs[funci].size();

    std::string sargs;
    std::string sargs2;
    std::string svals;
    svals = "if(aArgs.size()!="+
      Lib::smanip::tostring(argn)+") return false;\n";

    for(unsigned int argi=0;argi<argn;argi++) {
      if(argi) {
        sargs  += ",";
        sargs2 += ",";
        svals  += "\n";
      }
      const Lib::Value& v = aArgs[funci][argi]; //FIXME : use Value.
      std::string scxx;
      if(Lib::valuemanip::cxx_type(v,scxx)) {
        //FIXME : put code checking received type.
        sargs += scxx+" ";
        std::string what = scxx;
        Lib::smanip::replace(what,"*","_star");
        svals += sh+scxx+" "+v.label()+
          " = aArgs["+Lib::smanip::tostring(argi)+"]->get_"+what+"();";
      } else {
        Lib::Out out(aPrinter);
        out << "build_c_dll :" 
            << " for file " << Lib::sout(aName) 
            << " can't handle variable type " << Lib::sout(v.stype()) 
            << "."
            << Lib::endl;
        ::fclose(file);
        Lib::fmanip::remove(fileName);
        return false;
      }
      sargs  += v.label();
      sargs2 += v.label();
    }

    ::fprintf(file,"\
  double %s(%s);\n\
  %sbool Lib_wrap_%s(const std::vector<Slash::Core::IValue*>& aArgs,Slash::Core::IValue& aReturn) {\n\
    %s\n\
    aReturn.set(%s(%s));\n\
    return true;\n\
  }\n\
",
funcName.c_str(),sargs.c_str(),
s_declspec.c_str(),aFuncs[funci].c_str(),
svals.c_str(),
funcName.c_str(),sargs2.c_str()
);

  }

  ::fprintf(file,"\
}\n\
");

  ::fclose(file);

  std::string builder;
  if(aBuilder.size()) {
    builder = aBuilder;
  } else {
#ifdef WIN32
    builder = "$LIBROOT/scripts/DOS/makcdll.bat";
#else
    builder = "$LIBROOT/scripts/sh/makcdll.sh";
#endif
  }
  if(!Lib::fmanip::exists(builder)) {
    Lib::Out out(aPrinter);
    out << "build_c_dll :" 
        << " file " << Lib::sout(builder) << " not found."
        << Lib::endl;
    return false;
  }

  std::string sys_build_proc = Lib::fmanip::name(builder);
  sys_build_proc = Lib::System::quote(sys_build_proc);

  //FIXME : check ret status.
  Lib::System::execute(sys_build_proc+" "+aName+" "+aSuffix+" "+CNAME);

  return true;
}
/*
//////////////////////////////////////////////////////////////////////////////
bool build_C_dll(
 const std::string& aName
,const std::string& aSuffix
#ifdef WIN32
,const std::vector<std::string>& aFuncs
#else
,const std::vector<std::string>&
#endif
,const std::vector<Lib::FunctionManager::FunctionArgs>&
,const std::string& aBuilder
,Slash::Core::IWriter& aPrinter
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //printf("debug : build_C_dll : \"%s\" \"%s\"\n",
  //  aName.c_str(),aSuffix.c_str());

#ifdef WIN32
  // Produce .def file :
  std::string file_def = aName+".def";
  FILE* file = ::fopen(file_def.c_str(),"wb");
  if(!file) {
    Lib::Out out(aPrinter);
    out << "build_C_dll :" 
        << " can't create file " << Lib::sout(file_def)
        << Lib::endl;
    return false;
  }
  ::fprintf(file,"\
LIBRARY %s\n\
EXPORTS\n\
",aName.c_str());
  unsigned int funcn = aFuncs.size();
  for(unsigned int funci=0;funci<funcn;funci++) {
    std::string funcName = aFuncs[funci];
    ::fprintf(file," %s\n",funcName.c_str());
  }
  ::fclose(file);
#endif

  std::string builder;
  if(aBuilder.size()) {
    builder = aBuilder;
  } else {
#ifdef WIN32
    builder = "$LIBROOT/scripts/DOS/makcdll.bat";
#else
    builder = "$LIBROOT/scripts/sh/makcdll.sh";
#endif
  }
  if(!Lib::fmanip::exists(builder)) {
    Lib::Out out(aPrinter);
    out << "build_C_dll :" 
        << " file " << Lib::sout(builder) << " not found."
        << Lib::endl;
    return false;
  }

  std::string sys_build_proc = Lib::fmanip::name(builder);
  sys_build_proc = Lib::System::quote(sys_build_proc);

  //FIXME : check ret status.
  Lib::System::execute(sys_build_proc+" "+aName+" "+aSuffix);
  //Lib::System::execute
  //  (sys_build_proc+" "+aName+" "+aSuffix+" >& /Users/barrand/out.log");

#ifdef WIN32
#else
  std::string dll = 
    Lib::System::dllPrefix()+aName+"."+Lib::System::dllSuffix();
  if(!Lib::fmanip::exists(dll)) {
    Lib::Out out(aPrinter);
    out << "build_C_dll :" 
        << " DLL " << Lib::sout(dll) << " not created."
        << Lib::endl;
    return false;
  }  
#endif

  return true;
}
*/

static bool build_f77_dll(const std::string&,const std::string&,const std::vector<std::string>&,const std::vector<Lib::FunctionManager::FunctionArgs>&,const std::string&,Slash::Core::IWriter&);

//////////////////////////////////////////////////////////////////////////////
bool Lib::FunctionManager::loadF77(
 const std::string& aFile
,const std::string& aBuilder
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Core::IWriter& printer = fSession.printer();
  std::vector<std::string> funcs;
  std::vector<FunctionArgs> args;
  if(!find_f77_functions(aFile,funcs,args,printer)) {
    Lib::Out out(printer);
    out << "Lib::FunctionManager::loadF77 :" 
        << " syntax problem in file " << Lib::sout(aFile) << "."
        << Lib::endl;
    return false;
  }
  if(!funcs.size()) {
    Lib::Out out(printer);
    out << "Lib::FunctionManager::loadF77 :" 
        << " file " << Lib::sout(aFile) << " has not function."
        << Lib::endl;
    return false;
  }

  std::string new_file = "Lib_new_"+Lib::smanip::nosuffix(aFile)+".f";

  if(!edit_f77_file(aFile,new_file,funcs,args)) {
    Lib::Out out(printer);
    out << "Lib::FunctionManager::loadF77 :" 
        << " can't produce new file " << Lib::sout(new_file)
        << " from file " << Lib::sout(aFile) << "."
        << Lib::endl;
    return false;
  }

  if(fSession.verboseLevel()) {
    Lib::Out out(printer);
    out << "Lib::FunctionManager::loadF77 :" 
        << " functions in file " << Lib::sout(new_file) 
        << Lib::endl;
    unsigned int funcn = funcs.size();
    for(unsigned int funci=0;funci<funcn;funci++) {
      std::string sargs;
      unsigned int argn = args[funci].size();
      for(unsigned int argi=0;argi<argn;argi++) {
        if(argi) sargs += ",";
        sargs += args[funci][argi].label();
        sargs += "(";
        sargs += smanip::tostring((int)args[funci][argi].dimension());
        sargs += ")";
      }
      out << "Lib::FunctionManager::loadF77 :" 
          << " name : " << Lib::sout(funcs[funci]) 
	  << " args : " << Lib::sout(sargs) 
          << Lib::endl;
    }
  }

  std::string path,name,suffix;
  smanip::pathNameSuffix(new_file,path,name,suffix);
  std::string dll = smanip::nosuffix(name);
 
  std::string fullname = fmanip::name(new_file);
  fmanip::copy(fullname,name);

  std::vector<Slash::Core::ILibrary::Procedure> procs;
  if(!get_procs(fSession,dll,suffix,"Lib_wrap_",funcs,args,procs,aBuilder,
                build_f77_dll)) {
    Lib::Out out(printer);
    out << "Lib::FunctionManager::loadF77 :" 
        << " unable to get some f77 function in " << Lib::sout(dll)
        << Lib::endl;
    return false;
  }
  if(fSession.verboseLevel()) {
    Lib::Out out(printer);
    out << "Lib::FunctionManager::loadF77 :" 
        << " found f77 function in " << Lib::sout(dll)
        << Lib::endl;
  }

  unsigned int funcn = funcs.size();
  for(unsigned int funci=0;funci<funcn;funci++) {
    addFunction(new_file,funcs[funci],(void*)procs[funci],args[funci],0);
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool build_f77_dll(
 const std::string& aName
,const std::string& aSuffix
,const std::vector<std::string>& aFuncs
,const std::vector<Lib::FunctionManager::FunctionArgs>& aArgs
,const std::string& aBuilder
,Slash::Core::IWriter& aPrinter
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //printf("debug : build_f77_dll : \"%s\"\n",aName.c_str());

  std::string CNAME = aName;
  CNAME = std::string("Lib_wrap_") + CNAME;
  
  // Produce C frontend file :
  std::string fileName = CNAME+".cpp";
  FILE* file = ::fopen(fileName.c_str(),"wb");
  if(!file) {
    Lib::Out out(aPrinter);
    out << "build_f77_dll :" 
        << " can't create file " << Lib::sout(fileName)
        << Lib::endl;
    return false;
  }

  ::fprintf(file,"\
#include <Slash/Core/IValue.h>\n\
extern \"C\" {\n\
");


#ifdef WIN32
  std::string s_stdcall = "__stdcall ";              //space at end.
  std::string s_under = "";
  std::string s_declspec = "_declspec(dllexport) ";  //idem
#else 
  std::string s_stdcall = "";
  std::string s_under = "_";
  std::string s_declspec = "";
#endif

  std::string sh = "    ";

  unsigned int funcn = aFuncs.size();
  for(unsigned int funci=0;funci<funcn;funci++) {
    std::string funcName = aFuncs[funci];
    unsigned int argn = aArgs[funci].size();

    std::string sargs;
    std::string sargs2;
    std::string svals;
    svals = "if(aArgs.size()!="+
      Lib::smanip::tostring(argn)+") return false;\n";

    for(unsigned int argi=0;argi<argn;argi++) {
      if(argi) {
        sargs  += ",";
        sargs2 += ",";
        svals  += "\n";
      }
      const Lib::Value& v = aArgs[funci][argi]; //FIXME : use Value.
      std::string scxx;
      if(Lib::valuemanip::cxx_type(v,scxx)) {
        //FIXME : put code checking received type.
        sargs += scxx+" ";
        std::string what = scxx;
        Lib::smanip::replace(what,"*","_star");
        svals += sh+scxx+" "+v.label()+
          " = aArgs["+Lib::smanip::tostring(argi)+"]->get_"+what+"();";
      } else {
        Lib::Out out(aPrinter);
        out << "build_f77_dll :" 
            << " for file " << Lib::sout(aName) 
            << " can't handle variable type " << Lib::sout(v.stype()) 
            << "."
            << Lib::endl;
        ::fclose(file);
        Lib::fmanip::remove(fileName);
        return false;
      }
      sargs  += v.label();
      sargs2 += v.label();
    }

#ifdef WIN32
    Lib::smanip::touppercase(funcName);
#else
    Lib::smanip::tolowercase(funcName);
#endif

    ::fprintf(file,"\
  float %s%s%s(%s);\n\
  %sbool Lib_wrap_%s(const std::vector<Slash::Core::IValue*>& aArgs,Slash::Core::IValue& aReturn) {\n\
    %s\n\
    aReturn.set(%s%s(%s));\n\
    return true;\n\
  }\n\
",
s_stdcall.c_str(),funcName.c_str(),s_under.c_str(),sargs.c_str(),
s_declspec.c_str(),aFuncs[funci].c_str(),
svals.c_str(),
funcName.c_str(),s_under.c_str(),sargs2.c_str()
);

  }

  ::fprintf(file,"\
}\n\
");

  ::fclose(file);

  std::string builder;
  if(aBuilder.size()) {
    builder = aBuilder;
  } else {
#ifdef WIN32
    builder = "$LIBROOT/scripts/DOS/makf77dll.bat";
#else
    builder = "$LIBROOT/scripts/sh/makf77dll.sh";
#endif
  }
  if(!Lib::fmanip::exists(builder)) {
    Lib::Out out(aPrinter);
    out << "build_f77_dll :" 
        << " file " << Lib::sout(builder) << " not found."
        << Lib::endl;
    return false;
  }

  std::string sys_build_proc = Lib::fmanip::name(builder);
  sys_build_proc = Lib::System::quote(sys_build_proc);

  //FIXME : check ret status.
  Lib::System::execute(sys_build_proc+" "+aName+" "+aSuffix+" "+CNAME);

  return true;
}
