// this :
#include <Lib/Functions.h>


// Slash :
#include <Slash/Core/IValue.h>

// Lib :
#include <Lib/SpaceAlgebra.h>
#include <Lib/Processor.h>
#include <Lib/Out.h>
#include <Lib/smanip.h>
#include <Lib/sout.h>
#include <Lib/nmanip.h>
#include <Lib/mmanip.h>
#include <Lib/Debug.h>
#include <Lib/Value.h>
#include <Lib/Cast.h>

// Else :
#include <math.h>

#define MAXIMUM(a,b) ((a)>(b)?a:b)

Lib::Function::Gaussian::Gaussian(double aScale,double aMean,double aSigma){
  set(aScale,aMean,aSigma);
}

Lib::Function::Gaussian::Gaussian(const Gaussian& aFrom)
:fScale(aFrom.fScale)
,fMean(aFrom.fMean)
,fSigma(aFrom.fSigma)
{}

Lib::Function::Gaussian& Lib::Function::Gaussian::operator=(const Gaussian& aFrom){
  fScale = aFrom.fScale;
  fMean = aFrom.fMean;
  fSigma = aFrom.fSigma;
  return *this;
}

void Lib::Function::Gaussian::set(double aScale,double aMean,double aSigma){
  fScale = aScale;
  fMean = aMean;
  fSigma = aSigma;
  if(fSigma<=0) fSigma = 1;
}

double Lib::Function::Gaussian::value(double aX) const {
  double value = (aX - fMean)/fSigma;
  return fScale * ::exp(-0.5 * value * value);
}

void Lib::Function::Gaussian::setScale(double aScale){fScale = aScale;}

void Lib::Function::Gaussian::setMean(double aMean){fMean = aMean;}

void Lib::Function::Gaussian::setSigma(double aSigma){
  fSigma = aSigma;
  if(fSigma<=0) fSigma = 1;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

Lib::Function::BreitWigner::BreitWigner(double aHeight,double aCenter,double aWidth){
  set(aHeight,aCenter,aWidth);
}

Lib::Function::BreitWigner::BreitWigner(const BreitWigner& aFrom)
:fHeight(aFrom.fHeight)
,fCenter(aFrom.fCenter)
,fWidth(aFrom.fWidth)
{}

Lib::Function::BreitWigner& Lib::Function::BreitWigner::operator=(const BreitWigner& aFrom){
  fHeight = aFrom.fHeight;
  fCenter = aFrom.fCenter;
  fWidth = aFrom.fWidth;
  return *this;
}

void Lib::Function::BreitWigner::set(double aHeight,double aCenter,double aWidth){
  fHeight = aHeight;
  fCenter = aCenter;
  fWidth = aWidth;
  if(fWidth<=0) fWidth = 1;
}

double Lib::Function::BreitWigner::value(double aX) const {
  double value = 2.*(aX - fCenter)/fWidth;
  return fHeight/(1. + value * value);
}

void Lib::Function::BreitWigner::setHeight(double aHeight){fHeight = aHeight;}

void Lib::Function::BreitWigner::setCenter(double aCenter){fCenter = aCenter;}

void Lib::Function::BreitWigner::setWidth(double aWidth){
  fWidth = aWidth;
  if(fWidth<=0) fWidth = 1;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

Lib::Function::Exponential::Exponential(double aA,double aB){
  // exp(aA + aB * aX)  //Same as in PAW fitting definition.
  set(aA,aB);
}

Lib::Function::Exponential::Exponential(const Exponential& aFrom)
:fA(aFrom.fA)
,fB(aFrom.fB)
{}

Lib::Function::Exponential& Lib::Function::Exponential::operator=(const Exponential& aFrom){
  fA = aFrom.fA;
  fB = aFrom.fB;
  return *this;
}

void Lib::Function::Exponential::set(double aA,double aB){
  fA = aA;
  fB = aB;
}

void Lib::Function::Exponential::setA(double aA){fA = aA;}
void Lib::Function::Exponential::setB(double aB){fB = aB;}

double Lib::Function::Exponential::value(double aX) const {return ::exp(fA + fB * aX);}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

Lib::Function::Polynomial::Polynomial(const std::vector<double>& aParams){
  // P[0] + P[1] x + P[2] x**2 + ...
  setParameters(aParams);
}

Lib::Function::Polynomial::Polynomial(const Polynomial& aFrom)
:fPs(aFrom.fPs)
{}

Lib::Function::Polynomial& Lib::Function::Polynomial::operator=(const Polynomial& aFrom){
  fPs = aFrom.fPs;
  return *this;
}

bool Lib::Function::Polynomial::setParameters(const std::vector<double>& aParams){
  fPs = aParams;
  return true;
}

void Lib::Function::Polynomial::setParameter(unsigned int aIndex,double aValue){
  if(aIndex>=fPs.size()) return;
  fPs[aIndex] = aValue;
}

double Lib::Function::Polynomial::value(double aX) const {
  double value = 0;
  double xxx = 1;
  unsigned int pn = fPs.size();
  for(unsigned int i=0;i<pn;i++) {
    value += fPs[i] * xxx;
    xxx = xxx * aX;
  }
  return value;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

Lib::Function::Cauchy::Cauchy(double aHeight,double aCenter,double aWidth){
  set(aHeight,aCenter,aWidth);
}

Lib::Function::Cauchy::Cauchy(const Cauchy& aFrom)
:fHeight(aFrom.fHeight)
,fCenter(aFrom.fCenter)
,fWidth(aFrom.fWidth)
{}

Lib::Function::Cauchy& Lib::Function::Cauchy::operator=(const Cauchy& aFrom){
  fHeight = aFrom.fHeight;
  fCenter = aFrom.fCenter;
  fWidth = aFrom.fWidth;
  return *this;
}

void Lib::Function::Cauchy::set(double aHeight,double aCenter,double aWidth){
  fHeight = aHeight;
  fCenter = aCenter;
  fWidth = aWidth;
  if(fWidth<=0) fWidth = 1;
}

double Lib::Function::Cauchy::value(double aX) const {
  // From Wikipedia.
  double b = fWidth/2.;
  double value = (aX - fCenter)/b;
  return fHeight/((value * value + 1)*b*Lib::mmanip::pi());
}

void Lib::Function::Cauchy::setHeight(double aHeight){fHeight = aHeight;}

void Lib::Function::Cauchy::setCenter(double aCenter){fCenter = aCenter;}

void Lib::Function::Cauchy::setWidth(double aWidth){
  fWidth = aWidth;
  if(fWidth<=0) fWidth = 1;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

Lib::Function::Scripted::Scripted(const std::string& aString,Slash::Core::IWriter& aPrinter)
:fScript(aString)
,fIsValid(false)
,fDimension(0)
,fPrinter(aPrinter)
,fProcessor("",aPrinter)
{
  set(aString);
}

Lib::Function::Scripted::Scripted(const Scripted& aFrom)
:fScript(aFrom.fScript)
,fIsValid(false)
,fDimension(aFrom.fDimension)
,fParameterNames(aFrom.fParameterNames)
,fParameterValues(aFrom.fParameterValues)
,fVariableNames(aFrom.fVariableNames)
,fPrinter(aFrom.fPrinter)
,fProcessor("",aFrom.fPrinter)
{
  set(aFrom.fScript);
}

Lib::Function::Scripted::~Scripted(){}

Lib::Function::Scripted& Lib::Function::Scripted::operator=(const Scripted& aFrom){
  fScript = aFrom.fScript;
  fDimension = aFrom.fDimension;
  fParameterNames = aFrom.fParameterNames;
  fParameterValues = aFrom.fParameterValues;
  fVariableNames = aFrom.fVariableNames;

  //fPrinter = aFrom.fPrinter;
  fIsValid = false;

  set(aFrom.fScript);
  return *this;
}

bool Lib::Function::Scripted::isValid() const{return fIsValid;}

unsigned int Lib::Function::Scripted::dimension() const {return fDimension;}

std::string Lib::Function::Scripted::script() const {return fScript;}

std::vector<std::string> Lib::Function::Scripted::parameterNames() const {
  return fParameterNames;
}

std::vector<double> Lib::Function::Scripted::parameterValues() const {
  return fParameterValues;
}

std::vector<std::string> Lib::Function::Scripted::variableNames() const {
  return fVariableNames;
}

bool Lib::Function::Scripted::set(const std::string& aString){
  reset();
  //Check for AIDA/SLAC variable syntax convention x[<int>] :
  std::string ss = aString;
  std::string ns;
  bool slac_syntax = false;
  while(true) {
    std::string::size_type lb = ss.find("x[");
    if(lb==std::string::npos){
      ns += ss;
      break;
    }
    std::string::size_type rb = ss.find(']',lb+2);
    if(rb==std::string::npos) {
      Lib::Out out(fPrinter);
      out << "Lib::Function::Scripted::set :"
          << " bad [] balance in " << Lib::sout(aString)
          << Lib::endl;
      reset();
      return false;
    }
    //     abx[012]ef
    //     0 2    7 9   size=10
    std::string snum = ss.substr(lb+2,rb-lb-2);
    int coord;
    if(!Lib::smanip::toint(snum,coord)) {
      Lib::Out out(fPrinter);
      out << "Lib::Function::Scripted::set :"
          << " integer expected within [] in " << Lib::sout(aString)
          << ". Found " << Lib::sout(snum) << "."
          << Lib::endl;
      reset();
      return false;
    }
    slac_syntax = true;
    ns += ss.substr(0,lb)+ "x"+ snum;
    ss = ss.substr(rb+1,ss.size()-rb);
  }
  //printf("debug : ns \"%s\"\n",ns.c_str());

  // Find variables in the script :
  fProcessor.setScript(ns);
  std::vector<std::string> vars;
  if(!fProcessor.findVariables(vars)) {
    Lib::Out out(fPrinter);
    out << "Lib::Function::Scripted::set :"
        << " Lib::Processor can't find variables in " << Lib::sout(aString)
        << Lib::endl;
    reset();
    return false;
  }

  //vars had been passed to Lib::smanip::unique.

  //  In vars, some are expected to be of the form x<int>.
  // The x<int> are the variables of the scripted function. 
  // The rest are the parameters of the scripted function. 
  // The variables are used to find the dimension of the function.
  //  If x0 only is found, then it is a one dim function.
  // If x0,x1 only are found, then it is a two dim function, etc...

  std::vector<int> coords;
  bool found_x = false;
  bool found_y = false;
  bool found_z = false;
 {for(unsigned int index=0;index<vars.size();index++) {
    int coord;
    if(vars[index]=="x") {
      found_x = true;
    } else if(vars[index]=="y") {
      found_y = true;
    } else if(vars[index]=="z") {
      found_z = true;
    } else if(isVariable(vars[index],coord)) {
      coords.push_back(coord);
    } else {
      fParameterNames.push_back(vars[index]);
    }
  }}


  vars.clear(); //Reorder with params first and variables last.
 {for(unsigned int index=0;index<fParameterNames.size();index++) {
    vars.push_back(fParameterNames[index]);
  }}

  // Sort coords :
  if(coords.size()) {
    if(found_x||found_y||found_z) {      
      Lib::Out out(fPrinter);
      out << "Lib::Function::Scripted::set :"
          << " inconsistent variable specification in " << Lib::sout(aString)
          << Lib::endl;
      reset();
      return false;
    }    
    Lib::nmanip::sort(coords);
    for(unsigned int index=0;index<coords.size();index++) {
      std::string s;
      Lib::smanip::printf(s,32,"x%d",coords[index]);
      vars.push_back(s);
      if(slac_syntax) {
        Lib::smanip::printf(s,32,"x[%d]",coords[index]);
        fVariableNames.push_back(s);
      } else {
        fVariableNames.push_back(s);
      }
    }
  } else {
    if(found_x) {
      vars.push_back("x");
      fVariableNames.push_back("x");
    }
    if(found_y) {
      vars.push_back("y");
      fVariableNames.push_back("y");
    }
    if(found_z) {
      vars.push_back("z");
      fVariableNames.push_back("z");
    }
  }

  if(!fProcessor.compile(vars)) {
    Lib::Out out(fPrinter);
    out << "Lib::Function::Scripted::set :"
        << " Lib::Processor can't compile " << Lib::sout(aString)
        << Lib::endl;
    reset();
    return false;
  }

  fParameterValues.resize(fParameterNames.size(),0);

  fInput.resize
    (fParameterNames.size()+fVariableNames.size(),Lib::Value((double)0));

  fScript = aString;
  fDimension = fVariableNames.size();
  fIsValid = true;
  return true;
}

bool Lib::Function::Scripted::setParameters(const std::vector<double>& aParams){
  if(aParams.size()!=fParameterValues.size()) return false;
  fParameterValues = aParams;
  return true;
}

bool Lib::Function::Scripted::value(
 const std::vector<double>& aXs
,double& aValue
,std::string& aError
) const {
  Lib_self(Scripted);

  if(!isValid()) {
    aValue = 0;
    aError = "Lib::Function::Scripted::value : is not valid.";
    return false;
  }
  if(aXs.size()!=fDimension) {
    aValue = 0;
    aError = "Lib::Function::Scripted::value : xs.size()/dimension mismatch.";
    return false;
  }

  unsigned int ii = 0;
 {unsigned int number = fParameterValues.size(); 
  for(unsigned int index=0;index<number;index++) {
    self.fInput[ii].set(fParameterValues[index]);
    ii++;
  }}
 {unsigned int number = aXs.size(); 
  for(unsigned int index=0;index<number;index++) {
    self.fInput[ii].set(aXs[index]);
    ii++;
  }}

  Lib::Value var;
  if(!self.fProcessor.execute(fInput,var,aError)) {
    aValue = 0;
    return false;
  }

  if(var.type()!=Slash::Core::IValue::DOUBLE) {
    aError = std::string("Lib::Function::Scripted::value :")
           + " result is not a double Variable.";
    aValue = 0;
    return false;
  }

  //aValue = var.toDouble(); //FIXME
  aValue = var.get_double();
  return true;
}

bool Lib::Function::Scripted::isVariable(const std::string& aVar,int& aIndex){
  // If aVar is not of the form x<int> return false.
  // If yes, return true with aIndex=<int>.
  if(aVar.size()<=1) {
    aIndex = 0;
    return false;
  }
  if(aVar[0]!='x') {
    aIndex = 0;
    return false;
  }
  if(!Lib::smanip::toint(aVar.substr(1,aVar.size()-1),aIndex)) {
    aIndex = 0;
    return false;;
  }
  return true;
}

void Lib::Function::Scripted::reset(){
  // Reset :
  fIsValid = false;
  fScript = "";
  fDimension = 0;
  fParameterNames.clear();
  fParameterValues.clear();
  fVariableNames.clear();
  fProcessor.clear();
  fInput.clear();
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

Lib::Function::Composite::Composite(const std::string& aString,Slash::Core::IWriter& aPrinter)
:fScript(aString)
,fScripted("0",aPrinter)
,fPrinter(aPrinter)
{
  set(aString);
}

Lib::Function::Composite::Composite(const Composite& aFrom)
:fScript(aFrom.fScript)
,fScripted(aFrom.fScripted)
,fPrinter(aFrom.fPrinter)
{
  set(aFrom.fScript);
}

Lib::Function::Composite::~Composite(){}

Lib::Function::Composite& Lib::Function::Composite::operator=(const Composite& aFrom){
  fScript = aFrom.fScript;
  fScripted = aFrom.fScripted;
  //fPrinter = aFrom.fPrinter;
  set(aFrom.fScript);
  return *this;
}

bool Lib::Function::Composite::isValid() const {return fScripted.isValid();}

unsigned int Lib::Function::Composite::dimension() const {return fScripted.dimension();}

std::string Lib::Function::Composite::script() const {return fScript;}

std::vector<std::string> Lib::Function::Composite::parameterNames() const {
  return fScripted.parameterNames();
}

std::vector<double> Lib::Function::Composite::parameterValues() const {
  return fScripted.parameterValues();
}

std::vector<std::string> Lib::Function::Composite::variableNames() const {
  return fScripted.variableNames();
}

bool Lib::Function::Composite::setParameters(const std::vector<double>& aParams){return fScripted.setParameters(aParams);}

bool Lib::Function::Composite::value(
 const std::vector<double>& aXs
,double& aValue
,std::string& aError
) const {
  return fScripted.value(aXs,aValue,aError);
}

bool Lib::Function::Composite::set(const std::string& aString){
  // exa : G+E
  // FIXME : have more operations.
  std::string s;
  std::string script; 
  std::vector<std::string> words;
  Lib::smanip::words(aString,"+",words);
  unsigned int wordn = words.size();
  for(unsigned int index=0;index<wordn;index++) {
    if(index) script += "+";
    std::string& word = words[index];
    Lib::smanip::strip(word);
    Lib::smanip::touppercase(word);
    std::string p; //variable prefix.
#define pc p.c_str()
    if(word=="E") {
      if(wordn>=2) Lib::smanip::printf(p,32,"E_%d_",index);
      Lib::smanip::printf(s,132,"%samplitude*exp(x*%sexponent)",pc,pc);
    } else if(word=="EHBOOK") {
      if(wordn>=2) Lib::smanip::printf(p,32,"EHBOOK_%d_",index);
      Lib::smanip::printf(s,132,"expo(x,%sA,%sB)",pc,pc);
    } else if(word=="G") {
      if(wordn>=2) Lib::smanip::printf(p,32,"G_%d_",index);
      Lib::smanip::printf
        (s,132,"gauss(x,%samplitude,%smean,%ssigma)",pc,pc,pc);
    } else if(word=="BW") {
      if(wordn>=2) Lib::smanip::printf(p,32,"BW_%d_",index);
      Lib::smanip::printf
        (s,132,"breit(x,%samplitude,%sorigin,%swidth)",pc,pc,pc);
    } else if(word=="P1") {
      if(wordn>=2) Lib::smanip::printf(p,32,"P1_%d_",index);
      Lib::smanip::printf(s,132,"pol1(x,%sp0,%sp1)",pc,pc);
    } else if(word=="P2") {
      if(wordn>=2) Lib::smanip::printf(p,32,"P2_%d_",index);
      Lib::smanip::printf(s,132,"pol2(x,%sp0,%sp1,%sp2)",pc,pc,pc);
    } else if(word=="P3") {
      if(wordn>=2) Lib::smanip::printf(p,32,"P3_%d_",index);
      Lib::smanip::printf
        (s,132,"pol3(x,%sp0,%sp1,%sp2,%sp3)",pc,pc,pc,pc);
    } else if(word=="P4") {
      if(wordn>=2) Lib::smanip::printf(p,32,"P4_%d_",index);
      Lib::smanip::printf
        (s,132,"pol4(x,%sp0,%sp1,%sp2,%sp3,%sp4)",pc,pc,pc,pc,pc);
#undef pc
    } else {
      Lib::Out out(fPrinter);
      out << "Lib::Function::Composite::set :" 
          << " in " << Lib::sout(aString)
          << ", unknown function " << Lib::sout(word)
          << Lib::endl;
      fScript = "";
      return false;
    }
    script += s;
  }

  //printf("debug : composite : \"%s\"\n",script.c_str());

  if(!fScripted.set(script)) {
    fScript = "";
    return false;
  }
  fScript = aString;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

Lib::Function::Compiled::Compiled(
 Slash::Core::IWriter& aPrinter
,const std::string& aFile
,const std::string& aModel
,void* aAddress
,const std::vector<Lib::Value>& aArgs
,unsigned int aParn
)
:fPrinter(aPrinter)
,fFile(aFile)
,fModel(aModel)
,fAddress(aAddress)
,fParn(aParn)
,fReturn(0)
{
  Lib::Debug::increment("Lib::Function::Compiled");

  unsigned int argn = aArgs.size();
  for(unsigned int argi=0;argi<argn;argi++) {
    const Lib::Value& v = aArgs[argi];
    if( (v.type()==Slash::Core::IValue::DOUBLE_STAR) ||
        (v.type()==Slash::Core::IValue::INT_STAR)    ||
        (v.type()==Slash::Core::IValue::FLOAT_STAR)  ||
        (v.type()==Slash::Core::IValue::DOUBLE)      ||
        (v.type()==Slash::Core::IValue::INT)         ||
        (v.type()==Slash::Core::IValue::FLOAT)       ){
      //ok
    } else {
      //unknown type :
      Lib::Out out(fPrinter);
      out << "Lib::Function::Compiled::Compiled :"
          << " " << Lib::sout(v.stype()) << " not handled."
          << Lib::endl;
      fAddress = 0;
      fDims.clear();
      return;
    }
    fDims.push_back(v.dimension());
    //printf("debug :  arg \"%s\" %d\n",NAME.c_str(),dim);
  }

  // to optimize :
 {fValues.resize(argn,0);
  for(unsigned int argi=0;argi<argn;argi++) {
    const Lib::Value& v = aArgs[argi];
    if(v.type()==Slash::Core::IValue::DOUBLE) {
      fValues[argi] = new Lib::Value(double(0));
    } else if(v.type()==Slash::Core::IValue::FLOAT) {
      fValues[argi] = new Lib::Value(float(0));
    } else if(v.type()==Slash::Core::IValue::INT) {
      fValues[argi] = new Lib::Value(int(0));
    } else if(v.type()==Slash::Core::IValue::DOUBLE_STAR) {
      unsigned int dim = fDims[argi];
      double* array = dim ? new double[dim] : 0;
      fValues[argi] = new Lib::Value(array);
    } else if(v.type()==Slash::Core::IValue::INT_STAR) {
      unsigned int dim = fDims[argi];
      int* array = dim ? new int[dim] : 0;
      fValues[argi] = new Lib::Value(array);
    } else if(v.type()==Slash::Core::IValue::FLOAT_STAR) {
      unsigned int dim = fDims[argi];
      float* array = dim? new float[dim] : 0;
      fValues[argi] = new Lib::Value(array);
    }
  }}
  fReturn = new Lib::Value;

}

Lib::Function::Compiled::~Compiled(){
  delete fReturn;
 {unsigned int argn = fValues.size();
  for(unsigned int argi=0;argi<argn;argi++) {
    Slash::Core::IValue* ivalue = fValues[argi];
    if(ivalue->type()==Slash::Core::IValue::FLOAT_STAR) {
      if(ivalue->get_float_star()) delete [] ivalue->get_float_star();
    } else if(ivalue->type()==Slash::Core::IValue::DOUBLE_STAR) {
      if(ivalue->get_double_star()) delete [] ivalue->get_double_star();
    } else if(ivalue->type()==Slash::Core::IValue::INT_STAR) {
      if(ivalue->get_int_star()) delete [] ivalue->get_int_star();
    }
    delete ivalue;
  }}
  Lib::Debug::decrement("Lib::Function::Compiled");
}

std::string Lib::Function::Compiled::file() const {return fFile;}

std::string Lib::Function::Compiled::model() const {return fModel;}

void* Lib::Function::Compiled::address() const {return fAddress;}

unsigned int Lib::Function::Compiled::dimension() const {
  unsigned int dim = 0;
  unsigned int dimn = fDims.size();
  for(unsigned int i=0;i<dimn;i++) dim += (fDims[i]?fDims[i]:1);
  return dim;
}

unsigned int Lib::Function::Compiled::numberOfParameters() const {return fParn;}

bool Lib::Function::Compiled::value(const std::vector<double>& aArgs,double& aValue) const {
  if(!fAddress) {
    aValue = 0;
    return false;
  }

  if(aArgs.size()!=dimension()) {
    aValue = 0;
    return false;
  }

  bool status = true;

  unsigned int argn = fValues.size();
  unsigned int aArgn = aArgs.size();

 {unsigned int argc = 0;
  for(unsigned int argi=0;argi<argn;argi++) {
    Slash::Core::IValue* ivalue = fValues[argi];

    if(ivalue->type()==Slash::Core::IValue::DOUBLE) {
      if((argc+1)>aArgn) {
        status = false;
        break;
      } else {
        ivalue->set(aArgs[argc]);
        argc++;
      }
    } else if(ivalue->type()==Slash::Core::IValue::FLOAT) {
      if((argc+1)>aArgn) {
        status = false;
        break;
      } else {
        ivalue->set((float)aArgs[argc]);
        argc++;
      }
    } else if(ivalue->type()==Slash::Core::IValue::INT) {
      if((argc+1)>aArgn) {
        status = false;
        break;
      } else {
        ivalue->set((int)aArgs[argc]);
        argc++;
      }
    } else if(ivalue->type()==Slash::Core::IValue::DOUBLE_STAR) {
      unsigned int dim = fDims[argi];
      if((argc+dim)>aArgn) {
        status = false;
        break;
      } else {
        double* array = ivalue->get_double_star();
        for(unsigned int i=0;i<dim;i++) {
          array[i] = aArgs[argc];
          argc++;
        }
      }

    } else if(ivalue->type()==Slash::Core::IValue::INT_STAR) {
      unsigned int dim = fDims[argi];
      if((argc+dim)>aArgn) {
        status = false;
        break;
      } else {
        int* array = ivalue->get_int_star();
        for(unsigned int i=0;i<dim;i++) {
          array[i] = (int)aArgs[argc];
          argc++;
        }
      }

    } else if(ivalue->type()==Slash::Core::IValue::FLOAT_STAR) {
      unsigned int dim = fDims[argi];
      if((argc+dim)>aArgn) {
        status = false;
        break;
      } else {
        float* array = ivalue->get_float_star();
        for(unsigned int i=0;i<dim;i++) {
          array[i] = (float)aArgs[argc];
          argc++;
        }
      }

    } else {
      status = false;
      break;
    }
  }
  if(argc!=dimension()) {
    status = false;
  }}
  //if(!status) {
    //printf("debug : Lib::Function::Compiled::value : arg setup failed.\n");
  //}

  if(status) {
    typedef bool (*Function)(const std::vector<Slash::Core::IValue*>&,
                             Slash::Core::IValue&);
    status = ((Function)fAddress)(fValues,*fReturn);
  }
 
  if(status) {
    status = true;
    if(fReturn->type()==Slash::Core::IValue::FLOAT) {
      aValue = fReturn->get_float();
    } else if(fReturn->type()==Slash::Core::IValue::DOUBLE) {
      aValue = fReturn->get_double();
    } else {
      status = false;    
    }
  }

  return status;  
}

bool Lib::Function::Compiled::value(
 const std::vector<Slash::Core::IValue*>& aArgs
,Slash::Core::IValue& aValue
) const {
  if(!fAddress) {
    aValue.setNone();
    return false;
  }

  typedef bool (*Function)(const std::vector<Slash::Core::IValue*>&,
                           Slash::Core::IValue&);

  return ((Function)fAddress)(aArgs,aValue);
}

std::vector<Slash::Core::IValue*> Lib::Function::Compiled::signature(Slash::Core::IValue*& aReturn) const {
  // args + return value.
  aReturn = fReturn;
  return fValues;
}

/*
bool Lib::Function::Compiled::build_signature(
 const std::vector<Lib::Value>& aArgs
,std::string& aString
) const {
  // signature is a list of "," separated of :
  //    <string> [INTEGER,FLOAT,DOUBLE,..] <int>
  std::string sargs;
  unsigned int argn = aArgs.size();
  for(unsigned int argi=0;argi<argn;argi++) {
    const Lib::Value& v = aArgs[argi];
    if(!v.label().size()) {
      Lib::Out out(fPrinter);
      out << "Lib::Function::Compiled::build_signature :" 
          << " empty variable name."
          << Lib::endl;
      aString = "";
      return false;
    }
    if(argi) sargs += ",";
    sargs += v.label();
    sargs += " ";
    sargs += v.stype();
    sargs += " ";
    sargs += smanip::tostring((int)aArgs[argi].dimension());
  }
  aString = sargs;
  return true;
}
*/


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Lib::Function::InsidePolygon::InsidePolygon(){}
//Lib::Function::InsidePolygon::InsidePolygon(const std::vector<Point>& aPoints){set(aPoints);}
Lib::Function::InsidePolygon::InsidePolygon(const InsidePolygon& aFrom)
:fPoints(aFrom.fPoints)
{}
Lib::Function::InsidePolygon& Lib::Function::InsidePolygon::operator=(const InsidePolygon& aFrom){
  fPoints = aFrom.fPoints;
  return *this;
}
const std::vector<Lib::Function::InsidePolygon::Point>& Lib::Function::InsidePolygon::points() const {return fPoints;}

bool Lib::Function::InsidePolygon::set(const std::vector<Point>& aPoints){
  fPoints = aPoints;
  if(!fPoints.size()) return false;
  //check closure :
  if((fPoints.size()>=2) && (fPoints[fPoints.size()-1]!=fPoints[0]) ) {
    fPoints.push_back(fPoints[0]);
  }
  return true;
}
bool Lib::Function::InsidePolygon::set(unsigned int aIndex,const Point& aPoint){
  if(aIndex>=fPoints.size()) return false;
  fPoints[aIndex] = aPoint;
  return true;
}
//bool Lib::Function::InsidePolygon::isValid() const {
//  if(!fPoints.size()) return false; //no points given.
//  if(fPoints[fPoints.size()-1]!=fPoints[0]) return false; //Not closed polygon.
//  return true;
//}

// From :
//   http://softsurfer.com/Archive/algorithm_0103/algorithm_0103.htm


// Copyright 2001, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.

// V[] = vertex points of a polygon V[n+1] with V[n]=V[0]
bool Lib::Function::InsidePolygon::inside(const Point& a_P,const std::vector<Point>& a_V) {

  unsigned int n = a_V.size()-1;

  int    wn = 0;    // the winding number counter

  // loop through all edges of the polygon
  for (unsigned int i=0; i<n; i++) {   // edge from V[i] to V[i+1]
    if (a_V[i].second <= a_P.second) {         // start y <= P[1]
      if (a_V[i+1].second > a_P.second)      // an upward crossing
        if (is_left( a_V[i], a_V[i+1], a_P) > 0)  // P left of edge
          ++wn;            // have a valid up intersect
    } else {                       // start y > P[1] (no test needed)
      if (a_V[i+1].second <= a_P.second)     // a downward crossing
        if (is_left( a_V[i], a_V[i+1], a_P) < 0)  // P right of edge
          --wn;            // have a valid down intersect
    }
  }

  return ((wn!=0)?true:false);
}

double Lib::Function::InsidePolygon::value(const Point& aPoint) const {
  if(!fPoints.size()) return 0;
  return (inside(aPoint,fPoints)?1:0);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Lib::Function::InsideEllipse::InsideEllipse():f_a(1),f_b(1){}
Lib::Function::InsideEllipse::InsideEllipse(const InsideEllipse& aFrom):f_a(aFrom.f_a),f_b(aFrom.f_b){}
Lib::Function::InsideEllipse& Lib::Function::InsideEllipse::operator=(const InsideEllipse& aFrom){f_a = aFrom.f_a;f_b = aFrom.f_b;return *this;}

bool Lib::Function::InsideEllipse::set_from_a_b(double a_a,double a_b) {
  if(a_a<=0) return false;
  if(a_b<=0) return false;
  f_a = a_a;
  f_b = a_b;
  return true;
}

double Lib::Function::InsideEllipse::a() const {return f_a;}
double Lib::Function::InsideEllipse::b() const {return f_b;}

double Lib::Function::InsideEllipse::value(double a_x,double a_y) const {
  if(f_a>=f_b) {
    double f = ::sqrt(f_a*f_a-f_b*f_b);
    double x1  = a_x-f;
    double x2  = a_x+f;
    double y2 = a_y*a_y;
    double d = ::sqrt(x1*x1+y2) + ::sqrt(x2*x2+y2);
    return (d>(2*f_a)?false:true); //border is outside.
  } else {
    double f = ::sqrt(f_b*f_b-f_a*f_a);
    double y1  = a_y-f;
    double y2  = a_y+f;
    double x2 = a_x*a_x;
    double d = ::sqrt(x2+y1*y1) + ::sqrt(x2+y2*y2);
    return (d>(2*f_b)?false:true); //border is outside.
  }
}
