// this :
#include <Lib/GZipWriter.h>

#include <Lib/Out.h>
#include <Lib/Debug.h>

#ifdef LIB_HAS_ZLIB
#include <zlib.h>
#endif

//////////////////////////////////////////////////////////////////////////////
bool Lib::GZipWriter::available(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  return true;
#else
  return false;
#endif
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Lib::GZipWriter::GZipWriter(
 const std::string& aFileName
,Slash::Core::IWriter& aPrinter
)
:fPrinter(aPrinter)
,fFileName(aFileName)
,fFile(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("Lib::GZipWriter");
#ifdef LIB_HAS_ZLIB
  fFile = ::gzopen(aFileName.c_str(),"wb");
#else
  Lib::Out out(fPrinter);
  out << "Lib::GZipWriter::GZipWriter :"
      << " Lib package not reconstructed with zlib." << Lib::endl;
#endif
}
//////////////////////////////////////////////////////////////////////////////
Lib::GZipWriter::~GZipWriter(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  if(fFile) ::gzclose((gzFile*)fFile);
#endif
  Lib::Debug::decrement("Lib::GZipWriter");
}
//////////////////////////////////////////////////////////////////////////////
void Lib::GZipWriter::disable(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  if(!fFile) return; // Already closed.
  ::gzclose((gzFile*)fFile);
  fFile = 0;
#endif
}
//////////////////////////////////////////////////////////////////////////////
void Lib::GZipWriter::enable(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  if(fFile) return; // Already open.
  fFile = ::gzopen(fFileName.c_str(),"wb");
#endif
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::GZipWriter::enabled(
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fFile ? true : false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::GZipWriter::flush(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::GZipWriter::write(
 const std::string& aString
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  if(!fFile) return false;
  if(::gzprintf((gzFile*)fFile,"%s",aString.c_str())<0) {
    Lib::Out out(fPrinter);
    out << "Lib::GZipWriter::write :"
        << " error writing in file \"" << fFileName 
        << "\"." << Lib::endl;
    disable();
    return false;
  } else {
    return true;
  }
#else
  return false;
#endif
}
