// this :
#include <Lib/Mutex.h>

// private :
Lib::Mutex::Mutex(const Mutex&){}
Lib::Mutex& Lib::Mutex::operator=(const Mutex&){return *this;}

#ifdef WIN32
#include <windows.h>

Lib::Mutex::Mutex():f_imp(0){}

Lib::Mutex::~Mutex(){
  if(f_imp) {
    CRITICAL_SECTION* cs = (CRITICAL_SECTION*)f_imp;
    ::DeleteCriticalSection(cs);
    delete cs;
  }
}

bool Lib::Mutex::initialize(){
  if(f_imp) return true; //done.
  CRITICAL_SECTION* cs = new CRITICAL_SECTION;
  ::InitializeCriticalSection(cs);
  f_imp = cs;
  return true;
}

bool Lib::Mutex::lock(){
  if(!f_imp) return false;
  CRITICAL_SECTION* cs = (CRITICAL_SECTION*)f_imp;
  ::EnterCriticalSection(cs);
  return true;
}

bool Lib::Mutex::unlock(){
  if(!f_imp) return false;
  CRITICAL_SECTION* cs = (CRITICAL_SECTION*)f_imp;
  ::LeaveCriticalSection(cs);
  return true;
}

//bool Lib::Mutex::trylock(){
//  if(!f_imp) return false;
//  CRITICAL_SECTION* cs = (CRITICAL_SECTION*)f_imp;
//  if(::TryEnterCriticalSection(cs)) return true;
//  return false;
//}

#else

#include <pthread.h>

Lib::Mutex::Mutex():f_imp(0){}

bool Lib::Mutex::initialize(){
  if(f_imp) return true; //done.
  pthread_mutex_t* mutex = new pthread_mutex_t;
  int status = ::pthread_mutex_init(mutex,0);
  if(status) {
    delete mutex;
    return false;
  }
  f_imp = mutex;
  return true;
}

Lib::Mutex::~Mutex(){
  if(f_imp) {
    pthread_mutex_t* mutex = (pthread_mutex_t*)f_imp;
    ::pthread_mutex_destroy(mutex);
    delete mutex;
  }
}

bool Lib::Mutex::lock(){
  if(!f_imp) return false;
  pthread_mutex_t* mutex = (pthread_mutex_t*)f_imp;
  int status = ::pthread_mutex_lock(mutex);
  return status?false:true;
}

bool Lib::Mutex::unlock(){
  if(!f_imp) return false;
  pthread_mutex_t* mutex = (pthread_mutex_t*)f_imp;
  int status = ::pthread_mutex_unlock(mutex);
  return status?false:true;
}

//bool Lib::Mutex::trylock(){
//  if(!f_imp) return false;
//  pthread_mutex_t* mutex = (pthread_mutex_t*)f_imp;
//  int status = ::pthread_mutex_trylock(mutex);
//  return status?false:true;
//}

#endif
