// this :
#include <Lib/Variable.h>

#include <Slash/Core/IWriter.h>

#include <Lib/Debug.h>
#include <Lib/Out.h>
#include <Lib/smanip.h>

//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = NONE;
  u.f_unsigned_int64 = 0;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 const Variable& aVariable
)
:fPrinter(aVariable.fPrinter)
,fName(0)
,fDimension(aVariable.fDimension)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");

  if(aVariable.fName) fName = new std::string(*aVariable.fName);

  fType = aVariable.fType;

  if(aVariable.fType==STRING) {
    u.f_std_string = new std::string(*aVariable.u.f_std_string);
  } else if(aVariable.fType==VECTOR_DOUBLE) {
    u.f_Array_double = new Array<double>(*aVariable.u.f_Array_double);
  } else {
    u = aVariable.u;
  }

}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,char aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = CHAR;
  u.f_char = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,short aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = SHORT;
  u.f_short = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,int aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = INTEGER;
  u.f_int = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,float aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = FLOAT;
  u.f_float = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,double aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = DOUBLE;
  u.f_double = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,unsigned char aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = UNSIGNED_CHAR;
  u.f_unsigned_char = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,unsigned short aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = UNSIGNED_SHORT;
  u.f_unsigned_short  = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,unsigned int aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = UNSIGNED_INTEGER;
  u.f_unsigned_int  = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,Lib::int64 aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = INT64;
  u.f_int64 = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,Lib::uint64 aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = UNSIGNED_INT64;
  u.f_unsigned_int64 = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,const char* aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = STRING;
  u.f_std_string = new std::string(aValue?aValue:"");
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,const std::string& aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = STRING;
  u.f_std_string = new std::string(aValue);
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,void* aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = POINTER;
  u.f_void_star = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,double* aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = DOUBLE_STAR;
  u.f_double_star = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,float* aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = FLOAT_STAR;
  u.f_float_star = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,int* aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = INT_STAR;
  u.f_int_star = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,bool aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = BOOLEAN;
  u.f_bool = aValue;
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::Variable(
 Slash::Core::IWriter& aPrinter
,const Array<double>& aValue
)
:fPrinter(aPrinter)
,fName(0)
,fDimension(0)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Debug::increment("Lib::Variable");
  fType = VECTOR_DOUBLE;
  u.f_Array_double = new Array<double>(aValue);
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::~Variable(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  delete fName;
  reset();
  Debug::decrement("Lib::Variable");
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::reset(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fType==STRING) {
    delete u.f_std_string;
    u.f_std_string = 0;
  } else if(fType==VECTOR_DOUBLE) {
    delete u.f_Array_double;
    u.f_Array_double = 0;
  } else {
    u.f_unsigned_int64 = 0;
  }
}
//////////////////////////////////////////////////////////////////////////////
Lib::Variable& Lib::Variable::operator=(
 const Lib::Variable& aVariable
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  delete fName;
  fName = 0;
  if(aVariable.fName) fName = new std::string(*aVariable.fName);

  fDimension = aVariable.fDimension;

  reset();

  fType = aVariable.fType;
  if(aVariable.fType==STRING) {
    u.f_std_string = new std::string(*aVariable.u.f_std_string);
  } else if(aVariable.fType==VECTOR_DOUBLE) {
    u.f_Array_double = new Array<double>(*aVariable.u.f_Array_double);
  } else {
    u = aVariable.u;
  }

  return *this;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::setName(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  delete fName;
  fName = new std::string(aString);
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::Variable::name(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fName?*fName:"";
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::setDimension(
 unsigned int aDimension
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fDimension = aDimension;
}
//////////////////////////////////////////////////////////////////////////////
unsigned int Lib::Variable::dimension(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fDimension;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::setType(
 eType aType
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = aType;
  switch(aType) {
  case NONE:     u.f_unsigned_int64 = 0;break;
  case INTEGER:  u.f_int = 0;break;
  case DOUBLE:   u.f_double = 0;break;
  case POINTER:  u.f_void_star = 0;break;
  case BOOLEAN:  u.f_bool = false;break;
  case SHORT:    u.f_short = 0;break;
  case INT64:    u.f_int64 = 0;break;
  case FLOAT:    u.f_float = 0;break;
  case CHAR:     u.f_char = 0;break;
  case UNSIGNED_CHAR:    u.f_unsigned_char = 0;break;
  case UNSIGNED_SHORT:   u.f_unsigned_short = 0;break;
  case UNSIGNED_INTEGER: u.f_unsigned_int = 0;break;
  case UNSIGNED_INT64:   u.f_unsigned_int64 =0;break;
  case DOUBLE_STAR:      u.f_double_star = 0;break;
  case FLOAT_STAR:       u.f_float_star = 0;break;
  case INT_STAR:         u.f_int_star = 0;break;
  case STRING:           u.f_std_string = new std::string("");break;
  case VECTOR_DOUBLE:    u.f_Array_double = new Array<double>();break;
  }
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::setNone(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = NONE;
  u.f_unsigned_int64 = 0;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::is_none(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return (fType==NONE ? true : false);
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::is_vector_double(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return (fType==VECTOR_DOUBLE ? true : false);
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::is_bool(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return (fType==BOOLEAN ? true : false);
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::is_double(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return (fType==DOUBLE ? true : false);
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Lib::Variable::eType Lib::Variable::type(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fType;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::Variable::stype(
 eType aType
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#define scase(a) case a:return #a;
  switch(aType) {
  scase(NONE)
  scase(INTEGER)
  scase(INT64)
  scase(DOUBLE)
  scase(STRING)
  scase(POINTER)
  scase(BOOLEAN)
  scase(SHORT)
  scase(FLOAT)
  scase(CHAR)
  scase(UNSIGNED_CHAR)
  scase(UNSIGNED_SHORT)
  scase(UNSIGNED_INTEGER)
  scase(UNSIGNED_INT64)
  scase(DOUBLE_STAR)
  scase(FLOAT_STAR)
  scase(INT_STAR)
  scase(VECTOR_DOUBLE)
  }
#undef scase
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::Variable::stype(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return stype(fType);
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::cxx_type(
 std::string& aValue 
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  switch(fType) {
  case INTEGER: 
    aValue = "int";
    return true;
  case DOUBLE: 
    aValue = "double";
    return true;
  case STRING: 
    aValue = "std::string";
    return true;
  case POINTER: 
    aValue = "void*";
    return true;
  case BOOLEAN: 
    aValue = "bool";
    return true;
  case SHORT: 
    aValue = "short";
    return true;
  case FLOAT: 
    aValue = "float";
    return true;
  case CHAR: 
    aValue = "char";
    return true;
  case UNSIGNED_CHAR: 
    aValue = "unsigned char";
    return true;
  case UNSIGNED_SHORT: 
    aValue = "unsigned short";
    return true;
  case UNSIGNED_INTEGER: 
    aValue = "unsigned int";
    return true;
  case DOUBLE_STAR: 
    aValue = "double*";
    return true;
  case FLOAT_STAR: 
    aValue = "float*";
    return true;
  case INT_STAR: 
    aValue = "int*";
    return true;
  case VECTOR_DOUBLE:
  case NONE:
  case INT64:
  case UNSIGNED_INT64:
    aValue = "";
    return false;
  }
  return false;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
std::string Lib::Variable::toString(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
  switch(fType) {
  case NONE:
    smanip::printf(s,5,"(nil)");
    return s;
  case INTEGER:
    smanip::printf(s,16,"%d",u.f_int);
    return s;
  case DOUBLE:
    smanip::printf(s,16,"%g",u.f_double);
    return s;
  case POINTER:
    smanip::printf(s,16,"0x%lx",(unsigned long)u.f_void_star);
    return s;
  case UNSIGNED_SHORT:
    smanip::printf(s,16,"%u",u.f_unsigned_short);
    return s;
  case UNSIGNED_INTEGER:
    smanip::printf(s,16,"%u",u.f_unsigned_int);
    return s;
  case BOOLEAN:
    smanip::printf(s,5,"%s",u.f_bool?"true":"false");
    return s;
  case SHORT:
    smanip::printf(s,16,"%d",u.f_short);
    return s;
  case INT64:
    smanip::printf(s,16,LIB_INT64_FORMAT,u.f_int64);
    return s;
  case UNSIGNED_INT64:
    smanip::printf(s,16,LIB_UINT64_FORMAT,u.f_unsigned_int64);
    return s;
  case FLOAT:
    smanip::printf(s,16,"%g",u.f_float);
    return s;
  case UNSIGNED_CHAR:
    smanip::printf(s,16,"%c",u.f_unsigned_char);
    return s;
  case CHAR:
    smanip::printf(s,16,"%c",u.f_char);
    return s;
  case DOUBLE_STAR:
    smanip::printf(s,16,"0x%lx",(unsigned long)u.f_double_star);
    return s;
  case FLOAT_STAR:
    smanip::printf(s,16,"0x%lx",(unsigned long)u.f_float_star);
    return s;
  case INT_STAR:
    smanip::printf(s,16,"0x%lx",(unsigned long)u.f_int_star);
    return s;
  case STRING:
    return *u.f_std_string;
  case VECTOR_DOUBLE:
   {const std::vector<double>& stdv = u.f_Array_double->vector();
    unsigned int number = stdv.size();
    for(unsigned int index=0;index<number;index++) {
      if(index) s+= "\n";
      s += smanip::tostring(stdv[index]);
    }}
    return s;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::to_double(
 double& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  switch(fType) {
  case INTEGER:
    aValue = u.f_int;
    return true;
  case DOUBLE:
    aValue = u.f_double;
    return true;
  case UNSIGNED_SHORT:
    aValue = u.f_unsigned_short;
    return true;
  case UNSIGNED_INTEGER:
    aValue = u.f_unsigned_int;
    return true;
  case SHORT:
    aValue = u.f_short;
    return true;
  case INT64:
    aValue = (double)u.f_int64;
    return true;
  case UNSIGNED_INT64:
    aValue = (double)u.f_unsigned_int64;
    return true;
  case FLOAT:
    aValue = u.f_float;
    return true;
  case UNSIGNED_CHAR:
    aValue = u.f_unsigned_char;
    return true;
  case CHAR:
    aValue = u.f_char;
    return true;
  case BOOLEAN:
    aValue = u.f_bool?1:0;
    return true;
  case VECTOR_DOUBLE:
  case NONE:
  case STRING:
  case POINTER:
  case DOUBLE_STAR:
  case FLOAT_STAR:
  case INT_STAR:
    break;
  }
  setNone();
  return false;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::assign(
 const Lib::Variable& aVariable
,std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fType==aVariable.fType) {
    reset();
    if(aVariable.fType==STRING) {
      u.f_std_string = new std::string(*aVariable.u.f_std_string);
    } else if(aVariable.fType==VECTOR_DOUBLE) {
      u.f_Array_double = new Array<double>(*aVariable.u.f_Array_double);
    } else {
      u = aVariable.u;
    }
    return true;
  } else {
    switch(fType) {
    case INTEGER:
      break;
    case DOUBLE:
      switch(aVariable.fType) {
      case INTEGER:
        u.f_double = aVariable.u.f_int;
        return true;
      case UNSIGNED_SHORT:
        u.f_double = aVariable.u.f_unsigned_short;
        return true;
      case UNSIGNED_INTEGER:
        u.f_double = aVariable.u.f_unsigned_int;
        return true;
      case NONE:
      case DOUBLE:
      case POINTER:
      case BOOLEAN:
      case SHORT:
      case INT64:
      case FLOAT:
      case UNSIGNED_CHAR:
      case CHAR:
      case UNSIGNED_INT64:
      case DOUBLE_STAR:
      case FLOAT_STAR:
      case INT_STAR:
      case STRING:
      case VECTOR_DOUBLE:
        break;
      }
      break;
    case STRING:
      break;
    case POINTER:
      break;
    case UNSIGNED_INTEGER:
      break;
    case BOOLEAN:
      break;
    case SHORT:
      switch(aVariable.fType) {
      case INTEGER:
        u.f_short = (short)aVariable.u.f_int;
        return true;
      case NONE:
      case DOUBLE:
      case POINTER:
      case BOOLEAN:
      case SHORT:
      case INT64:
      case FLOAT:
      case UNSIGNED_CHAR:
      case CHAR:
      case UNSIGNED_SHORT:
      case UNSIGNED_INTEGER:
      case UNSIGNED_INT64:
      case DOUBLE_STAR:
      case FLOAT_STAR:
      case INT_STAR:
      case STRING:
      case VECTOR_DOUBLE:
        break;
      }
      break;
    case FLOAT:
      switch(aVariable.fType) {
      case INTEGER:
        u.f_float = (float)aVariable.u.f_int;
        return true;
      case DOUBLE:
        u.f_float = (float)aVariable.u.f_double;
        return true;
      case NONE:
      case POINTER:
      case BOOLEAN:
      case SHORT:
      case INT64:
      case FLOAT:
      case UNSIGNED_CHAR:
      case CHAR:
      case UNSIGNED_SHORT:
      case UNSIGNED_INTEGER:
      case UNSIGNED_INT64:
      case DOUBLE_STAR:
      case FLOAT_STAR:
      case INT_STAR:
      case STRING:
      case VECTOR_DOUBLE:
        break;
      }
      break;
    case UNSIGNED_CHAR:
      switch(aVariable.fType) {
      case INTEGER:
        u.f_unsigned_char = (unsigned char)aVariable.u.f_int;
        return true;
      case NONE:
      case DOUBLE:
      case POINTER:
      case BOOLEAN:
      case SHORT:
      case INT64:
      case FLOAT:
      case UNSIGNED_CHAR:
      case CHAR:
      case UNSIGNED_SHORT:
      case UNSIGNED_INTEGER:
      case UNSIGNED_INT64:
      case DOUBLE_STAR:
      case FLOAT_STAR:
      case INT_STAR:
      case STRING:
      case VECTOR_DOUBLE:
        break;
      }
      break;
    case CHAR:
      switch(aVariable.fType) {
      case INTEGER:
        u.f_char = (char)aVariable.u.f_int;
        return true;
      case STRING:
        if(aVariable.u.f_std_string->size()!=1) break;
        u.f_char = (*aVariable.u.f_std_string)[0];
        return true;
      case NONE:
      case DOUBLE:
      case POINTER:
      case BOOLEAN:
      case SHORT:
      case INT64:
      case FLOAT:
      case UNSIGNED_CHAR:
      case CHAR:
      case UNSIGNED_SHORT:
      case UNSIGNED_INTEGER:
      case UNSIGNED_INT64:
      case DOUBLE_STAR:
      case FLOAT_STAR:
      case INT_STAR:
      case VECTOR_DOUBLE:
        break;
      }
      break;
    case NONE:
    case INT64:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case VECTOR_DOUBLE:
      break;
    }
    aError = std::string("Lib::Variable::assign :")
           + " unable to assign " + aVariable.stype() 
           + " to " + stype();
    setNone();
    return false;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::minus(
 std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  eType oType = fType;
  switch(fType) {
  case INTEGER:
    u.f_int = -u.f_int;
    return true;
  case DOUBLE:
    u.f_double = -u.f_double;
    return true;
  case SHORT:
    u.f_short = -u.f_short;
    return true;
  case INT64:
    u.f_int64 = -u.f_int64;
    return true;
  case CHAR:
    u.f_char = -u.f_char;
    return true;
  case FLOAT:
    u.f_float = -1.0F * u.f_float;
    return true;
  case VECTOR_DOUBLE:{
    std::vector<double>& stdv = u.f_Array_double->vector();
    unsigned int number = stdv.size();
    for(unsigned int index=0;index<number;index++) stdv[index] = -stdv[index];
    return true;}
  case BOOLEAN:
  case NONE:
  case STRING:
  case POINTER:
  case UNSIGNED_CHAR:
  case UNSIGNED_SHORT:
  case UNSIGNED_INT64:
  case DOUBLE_STAR:
  case FLOAT_STAR:
  case INT_STAR:
  case UNSIGNED_INTEGER:
    break;
  }
  aError = std::string("Lib::Variable::minus :")
         + " can't apply on type " + stype(oType);
  setNone();
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::do_not(
 std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  eType oType = fType;
  switch(fType) {
  case BOOLEAN:
    u.f_bool = u.f_bool?false:true;
    return true;
  case INTEGER:
  case DOUBLE:
  case STRING:
  case POINTER:
  case UNSIGNED_SHORT:
  case UNSIGNED_INTEGER:
  case SHORT:
  case FLOAT:
  case UNSIGNED_CHAR:
  case CHAR:
  case NONE:
  case INT64:
  case UNSIGNED_INT64:
  case DOUBLE_STAR:
  case FLOAT_STAR:
  case INT_STAR:
  case VECTOR_DOUBLE:
    break;
  }
  aError = std::string("Lib::Variable::do_not :")
         + " can't apply on type " + stype(oType);
  setNone();
  return false;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::add(
 const Variable& aV
,std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  eType oType = fType;
  switch(fType) {
  case INTEGER:
    switch(aV.fType) {
    case INTEGER:
      u.f_int += aV.u.f_int;
      return true;
    case VECTOR_DOUBLE:{
      int tmp = u.f_int;
      set(*aV.u.f_Array_double); //this becomes a VECTOR_DOUBLE.
      u.f_Array_double->add(tmp);
      return true;}
    case NONE:
    case DOUBLE:
    case STRING:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case UNSIGNED_INTEGER:
      break;
    }
    break;
  case DOUBLE:
    switch(aV.fType) {
    case DOUBLE:
      u.f_double += aV.u.f_double;
      return true;
    case UNSIGNED_SHORT:
      u.f_double += aV.u.f_unsigned_short;
      return true;
    case UNSIGNED_INTEGER:
      u.f_double += aV.u.f_unsigned_int;
      return true;
    case VECTOR_DOUBLE:{
      double tmp = u.f_double;
      set(*aV.u.f_Array_double); //this becomes a VECTOR_DOUBLE.
      u.f_Array_double->add(tmp);
      return true;}
    case NONE:
    case INTEGER:
    case STRING:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
      break;
    }
    break;
  case UNSIGNED_INTEGER:
    switch(aV.fType) {
    case UNSIGNED_INTEGER:
      u.f_unsigned_int += aV.u.f_unsigned_int;
      return true;
/*  case DOUBLE:{
      double d = u.f_unsigned_int + aV.u.f_double;
      if(d>=0)  {
        u.f_unsigned_int = (unsigned int)d;
        return true;
      }}*/
    case VECTOR_DOUBLE:{
      unsigned int tmp = u.f_unsigned_int;
      set(*aV.u.f_Array_double); //this becomes a VECTOR_DOUBLE.
      u.f_Array_double->add(tmp);
      return true;}
    case DOUBLE:{
      double d = double(u.f_unsigned_int) + aV.u.f_double;
      set(d); //this becomes a DOUBLE.
      return true;}
    case NONE:
    case INTEGER:
    case STRING:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
      break;
    }
    break;
  case STRING:
    switch(aV.fType) {
    case STRING:
      set((*u.f_std_string)+(*aV.u.f_std_string));
      return true;
    case NONE:
    case INTEGER:
    case DOUBLE:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INTEGER:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case VECTOR_DOUBLE:
      break;
    }
    break;
  case VECTOR_DOUBLE:
    switch(aV.fType) {
    case INTEGER:
      u.f_Array_double->add(aV.u.f_int);
      return true;
    case UNSIGNED_INTEGER:
      u.f_Array_double->add(aV.u.f_unsigned_int);
      return true;
    case DOUBLE:
      u.f_Array_double->add(aV.u.f_double);
      return true;
    case VECTOR_DOUBLE:
      if(u.f_Array_double->orders()!=aV.u.f_Array_double->orders()) {
        aError = "Variable::add : vector orders mismatch.";
        setNone();
        return false;
      }
      u.f_Array_double->add(*aV.u.f_Array_double);
      return true;
    case NONE:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case NONE:
  case POINTER:
  case BOOLEAN:
  case SHORT:
  case INT64:
  case FLOAT:
  case UNSIGNED_CHAR:
  case CHAR:
  case UNSIGNED_SHORT:
  case UNSIGNED_INT64:
  case DOUBLE_STAR:
  case FLOAT_STAR:
  case INT_STAR:
    break;
  }
  aError = std::string("Variable::add :" )
         + " can't apply on type " + stype(oType)
         + " and type " + aV.stype();
  setNone();
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::subtract(
 const Variable& aV
,std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  eType oType = fType;
  switch(fType) {
  case INTEGER:
    switch(aV.fType) {
    case INTEGER:
      u.f_int -= aV.u.f_int;
      return true;
    case VECTOR_DOUBLE:{
      int tmp = u.f_int;
      set(*aV.u.f_Array_double); //this becomes a VECTOR_DOUBLE.
      u.f_Array_double->multiply(-1);
      u.f_Array_double->add(tmp);
      return true;}
    case NONE:
    case DOUBLE:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INTEGER:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case DOUBLE:
    switch(aV.fType) {
    case DOUBLE:
      u.f_double -= aV.u.f_double;
      return true;
    case UNSIGNED_SHORT:
      u.f_double -= aV.u.f_unsigned_short;
      return true;
    case UNSIGNED_INTEGER:
      u.f_double -= aV.u.f_unsigned_int;
      return true;
    case VECTOR_DOUBLE:{
      double tmp = u.f_double;
      set(*aV.u.f_Array_double); //this becomes a VECTOR_DOUBLE.
      u.f_Array_double->multiply(-1);
      u.f_Array_double->add(tmp);
      return true;}
    case NONE:
    case INTEGER:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case UNSIGNED_INTEGER:
    switch(aV.fType) {
//uuu
/*
    case UNSIGNED_INTEGER:
      u.f_unsigned_int -= aV.u.f_unsigned_int;
      return true;
*/
    case VECTOR_DOUBLE:{
      unsigned int tmp = u.f_unsigned_int;
      set(*aV.u.f_Array_double); //this becomes a VECTOR_DOUBLE.
      u.f_Array_double->multiply(-1);
      u.f_Array_double->add(tmp);
      return true;}
    case DOUBLE:{
      double d = double(u.f_unsigned_int) - aV.u.f_double;
      set(d); //this becomes a DOUBLE.
      return true;}
    case NONE:
    case INTEGER:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INTEGER:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case VECTOR_DOUBLE:
    switch(aV.fType) {
    case INTEGER:
      u.f_Array_double->add(-aV.u.f_int);
      return true;
    case UNSIGNED_INTEGER:
      u.f_Array_double->add(-((double)aV.u.f_unsigned_int));
      return true;
    case DOUBLE:
      u.f_Array_double->add(-aV.u.f_double);
      return true;
    case VECTOR_DOUBLE:{
      if(u.f_Array_double->orders()!=aV.u.f_Array_double->orders()) {
        aError = "Variable::subtract : vector orders mismatch.";
        setNone();
        return false;
      }
      std::vector<double>& stdv = u.f_Array_double->vector();
      const std::vector<double>& stdv2 = aV.u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) { 
        stdv[index] -= stdv2[index];
      }
      return true;}
    case NONE:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case NONE:
  case POINTER:
  case BOOLEAN:
  case SHORT:
  case INT64:
  case FLOAT:
  case UNSIGNED_CHAR:
  case CHAR:
  case UNSIGNED_SHORT:
  case UNSIGNED_INT64:
  case DOUBLE_STAR:
  case FLOAT_STAR:
  case INT_STAR:
  case STRING:
    break;
  }
  aError = std::string("Variable::subtract :" )
         + " can't apply on type " + stype(oType)
         + " and type " + aV.stype();
  setNone();
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::multiply(
 const Variable& aV
,std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  eType oType = fType;
  switch(fType) {
  case INTEGER:
    switch(aV.fType) {
    case INTEGER:
      u.f_int *= aV.u.f_int;
      return true;
    case DOUBLE:
      u.f_int *= (int)aV.u.f_double;
      return true;
    case FLOAT:
      u.f_int *= (int)aV.u.f_float;
      return true;
    case VECTOR_DOUBLE:{
      int tmp = u.f_int;
      set(*aV.u.f_Array_double); //this becomes a VECTOR_DOUBLE.
      u.f_Array_double->multiply(tmp);
      return true;}
    case NONE:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case UNSIGNED_INTEGER:
    case STRING:
      break;
    }
    break;
  case DOUBLE:
    switch(aV.fType) {
    case INTEGER:
      u.f_double *= aV.u.f_int;
      return true;
    case DOUBLE:
      u.f_double *= aV.u.f_double;
      return true;
    case UNSIGNED_SHORT:
      u.f_double *= aV.u.f_unsigned_short;
      return true;
    case UNSIGNED_INTEGER:
      u.f_double *= aV.u.f_unsigned_int;
      return true;
    case VECTOR_DOUBLE:{
      double tmp = u.f_double;
      set(*aV.u.f_Array_double); //this becomes a VECTOR_DOUBLE.
      u.f_Array_double->multiply(tmp);
      return true;}
    case NONE:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case UNSIGNED_INTEGER:
    switch(aV.fType) {
    case DOUBLE:{
      double tmp = double(u.f_unsigned_int) * aV.u.f_double;
      set(tmp); //this becomes a DOUBLE.
      return true;}

      //u.f_unsigned_int *= (unsigned int)aV.u.f_double;
      //return true;

    case UNSIGNED_INTEGER:
      u.f_unsigned_int *= aV.u.f_unsigned_int;
      return true;
    case VECTOR_DOUBLE:{
      unsigned int tmp = u.f_unsigned_int;
      set(*aV.u.f_Array_double); //this becomes a VECTOR_DOUBLE.
      u.f_Array_double->multiply(tmp);
      return true;}
    case NONE:
    case POINTER:
    case BOOLEAN:
    case INTEGER:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case VECTOR_DOUBLE:
    switch(aV.fType) {
    case INTEGER:
      u.f_Array_double->multiply(aV.u.f_int);
      return true;
    case UNSIGNED_INTEGER:
      u.f_Array_double->multiply(aV.u.f_unsigned_int);
      return true;
    case DOUBLE:
      u.f_Array_double->multiply(aV.u.f_double);
      return true;
    case VECTOR_DOUBLE:{
      if(u.f_Array_double->orders()!=aV.u.f_Array_double->orders()) {
        aError = "Variable::multiply : vector orders mismatch.";
        setNone();
        return false;
      }
      std::vector<double>& stdv = u.f_Array_double->vector();
      const std::vector<double>& stdv2 = aV.u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) { 
        stdv[index] *= stdv2[index];
      }
      return true;}
    case NONE:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case NONE:
  case STRING:
  case POINTER:
  case BOOLEAN:
  case SHORT:
  case INT64:
  case FLOAT:
  case UNSIGNED_CHAR:
  case CHAR:
  case UNSIGNED_SHORT:
  case UNSIGNED_INT64:
  case DOUBLE_STAR:
  case FLOAT_STAR:
  case INT_STAR:
    break;
  }
  aError = std::string("Variable::multiply :" )
         + " can't apply on type " + stype(oType)
         + " and type " + aV.stype();
  setNone();
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::divide(
 const Variable& aV
,std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  eType oType = fType;
  switch(fType) {
  case INTEGER:
    switch(aV.fType) {
    case VECTOR_DOUBLE:{
      int tmp = u.f_int;
      set(*aV.u.f_Array_double); //this becomes a VECTOR_DOUBLE.
      if(!u.f_Array_double->invert()) {
        aError = Lib_Variable_error_div_zero;
        setNone();
        return false;
      }
      u.f_Array_double->multiply(tmp);
      return true;}
    case NONE:
    case INTEGER:
    case DOUBLE:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INTEGER:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case DOUBLE:
    switch(aV.fType) {
    case DOUBLE:
      if(aV.u.f_double==0) {
        aError = Lib_Variable_error_div_zero;
        setNone();
        return false;
      } else {
        u.f_double /= aV.u.f_double;
        return true;
      }
    case UNSIGNED_INTEGER:
      if(aV.u.f_unsigned_int==0) {
        aError = Lib_Variable_error_div_zero;
        setNone();
        return false;
      } else {
        u.f_double /= aV.u.f_unsigned_int;
        return true;
      }
    case VECTOR_DOUBLE:{
      double tmp = u.f_double;
      set(*aV.u.f_Array_double); //this becomes a VECTOR_DOUBLE.
      if(!u.f_Array_double->invert()) {
        aError = Lib_Variable_error_div_zero;
        setNone();
        return false;
      }
      u.f_Array_double->multiply(tmp);
      return true;}

    case NONE:
    case INTEGER:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case UNSIGNED_INTEGER:
    switch(aV.fType) {
/*  case UNSIGNED_INTEGER:
      if(aV.u.f_unsigned_int==0) {
        aError = Lib_Variable_error_div_zero;
        setNone();
        return false;
      } else {
        //FIXME : ???
        set((double)(u.f_unsigned_int  / aV.u.f_unsigned_int ));
        return true;
      }*/
    case VECTOR_DOUBLE:{
      unsigned int tmp = u.f_unsigned_int;
      set(*aV.u.f_Array_double); //this becomes a VECTOR_DOUBLE.
      if(!u.f_Array_double->invert()) {
        aError = Lib_Variable_error_div_zero;
        setNone();
        return false;
      }
      u.f_Array_double->multiply(tmp);
      return true;}

    case DOUBLE:
      if(aV.u.f_double==0) {
        aError = Lib_Variable_error_div_zero;
        setNone();
        return false;
      } else {
        double d = double(u.f_unsigned_int)/aV.u.f_double;
        set(d); //this becomes a DOUBLE.
        return true;
      }
    case NONE:
    case INTEGER:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INTEGER:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case VECTOR_DOUBLE:
    switch(aV.fType) {
    case INTEGER:
      if(aV.u.f_int==0) {
        aError = Lib_Variable_error_div_zero;
        setNone();
        return false;
      } else {
        double value = 1./aV.u.f_int;
        u.f_Array_double->multiply(value);
        return true;
      }
    case UNSIGNED_INTEGER:
      if(aV.u.f_unsigned_int==0) {
        aError = Lib_Variable_error_div_zero;
        setNone();
        return false;
      } else {
        double value = 1./aV.u.f_unsigned_int;
        u.f_Array_double->multiply(value);
        return true;
      }
    case DOUBLE:
      if(aV.u.f_double==0) {
        aError = Lib_Variable_error_div_zero;
        setNone();
        return false;
      } else {
        u.f_Array_double->multiply(1./aV.u.f_double);
        return true;
      }
    case VECTOR_DOUBLE:{
      if(u.f_Array_double->orders()!=aV.u.f_Array_double->orders()) {
        aError = Lib_Variable_error_div_zero;
        setNone();
        return false;
      }
      std::vector<double>& stdv = u.f_Array_double->vector();
      const std::vector<double>& stdv2 = aV.u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) { 
        if(stdv2[index]==0) {
          aError = std::string
            ("Lib::Variable::divide : VECTOR : divide by 0 at index")
            + Lib::smanip::tostring(index) 
            + ". Let it and continue.";
          //continue;
          aError = Lib_Variable_error_div_zero;
          setNone();
          return false;
        } else {
          stdv[index] /= stdv2[index];
        }
      }
      return true;}

    case NONE:
    case STRING:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case INT64:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
      break;
    }
    break;

  case NONE:
  case STRING:
  case POINTER:
  case BOOLEAN:
  case SHORT:
  case FLOAT:
  case UNSIGNED_CHAR:
  case CHAR:
  case INT64:
  case UNSIGNED_SHORT:
  case UNSIGNED_INT64:
  case DOUBLE_STAR:
  case FLOAT_STAR:
  case INT_STAR:
    break;
  }
  aError = std::string("Variable::divide :")
         + " can't apply on type " + stype(oType)
         + " and type " + aV.stype();
  setNone();
  return false;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::if_gt(
 const Lib::Variable& aV
,std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  eType oType = fType;
  switch(fType) {
  case INTEGER:
    switch(aV.fType) {
    case INTEGER:
      set((bool)(u.f_int>aV.u.f_int?true:false));
      return true;
    case FLOAT:
      set((bool)(u.f_int>aV.u.f_float?true:false));
      return true;
    case DOUBLE:
      set((bool)(u.f_int>aV.u.f_double?true:false));
      return true;
    case UNSIGNED_SHORT:
      set((bool)(u.f_int>(int)aV.u.f_unsigned_short?true:false));
      return true;
    case UNSIGNED_INTEGER:
      set((bool)(u.f_int>(int)aV.u.f_unsigned_int?true:false));
      return true;
    case VECTOR_DOUBLE:{
      const std::vector<double>& stdv = aV.u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(u.f_int<=stdv[index]) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case NONE:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case DOUBLE:
    switch(aV.fType) {
    case DOUBLE:
      set((bool)(u.f_double>aV.u.f_double?true:false));
      return true;
    case FLOAT:
      set((bool)(u.f_double>aV.u.f_float?true:false));
      return true;
    case INTEGER:
      set((bool)(u.f_double>aV.u.f_int?true:false));
      return true;
    case UNSIGNED_SHORT:
      set((bool)(u.f_double>aV.u.f_unsigned_short?true:false));
      return true;
    case UNSIGNED_INTEGER:
      set((bool)(u.f_double>aV.u.f_unsigned_int?true:false));
      return true;
    case VECTOR_DOUBLE:{
      const std::vector<double>& stdv = aV.u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(u.f_double<=stdv[index]) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case NONE:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case FLOAT:
    switch(aV.fType) {
    case DOUBLE:
      set((bool)(u.f_float>aV.u.f_double?true:false));
      return true;
    case FLOAT:
      set((bool)(u.f_float>aV.u.f_float?true:false));
      return true;
    case INTEGER:
      set((bool)(u.f_float>aV.u.f_int?true:false));
      return true;
    case UNSIGNED_SHORT:
      set((bool)(u.f_float>aV.u.f_unsigned_short?true:false));
      return true;
    case UNSIGNED_INTEGER:
      set((bool)(u.f_float>aV.u.f_unsigned_int?true:false));
      return true;
    case VECTOR_DOUBLE:{
      const std::vector<double>& stdv = aV.u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(u.f_float<=stdv[index]) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case NONE:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case POINTER:
    switch(aV.fType) {
    case POINTER:
      set((bool)(u.f_void_star>aV.u.f_void_star?true:false));
      return true;
    case NONE:
    case INTEGER:
    case DOUBLE:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INTEGER:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
    case VECTOR_DOUBLE:
      break;
    }
    break;
  case UNSIGNED_INTEGER:
    switch(aV.fType) {
    case UNSIGNED_INTEGER:
      set((bool)(u.f_unsigned_int >aV.u.f_unsigned_int ?true:false));
      return true;
    case INTEGER:
      set((bool)((int)u.f_unsigned_int >aV.u.f_int?true:false));
      return true;
    case NONE:
    case POINTER:
    case DOUBLE:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
    case VECTOR_DOUBLE:
      break;
    }
    break;
  case VECTOR_DOUBLE:
    switch(aV.fType) {
    case INTEGER:{
      const std::vector<double>& stdv = u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(stdv[index]<=aV.u.f_int) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case FLOAT:{
      const std::vector<double>& stdv = u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(stdv[index]<=aV.u.f_float) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case DOUBLE:{
      const std::vector<double>& stdv = u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(stdv[index]<=aV.u.f_double) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case UNSIGNED_INTEGER:{
      const std::vector<double>& stdv = u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(stdv[index]<=aV.u.f_unsigned_int ) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case VECTOR_DOUBLE:{
      const std::vector<double>& stdv = u.f_Array_double->vector();
      const std::vector<double>& stdv2 = aV.u.f_Array_double->vector();
      if(stdv.size()!=stdv2.size()) break;
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(stdv[index]<=stdv2[index]) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case NONE:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case NONE:
  case BOOLEAN:
  case SHORT:
  case UNSIGNED_CHAR:
  case CHAR:
  case STRING:
  case INT64:
  case UNSIGNED_SHORT:
  case UNSIGNED_INT64:
  case DOUBLE_STAR:
  case FLOAT_STAR:
  case INT_STAR:
    break;
  }
  aError = std::string("Variable::if_gt() :")
         + " can't compare type " + stype(oType) 
         + " with type " + aV.stype();
  setNone();
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::if_eq(
 const Variable& aV
,std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  eType oType = fType;
  switch(fType) {
  case INTEGER:
    switch(aV.fType) {
    case INTEGER:
      set((bool)(u.f_int==aV.u.f_int?true:false));
      return true;
    case DOUBLE:
      set((bool)(u.f_int==aV.u.f_double?true:false));
      return true;
    case FLOAT:
      set((bool)(u.f_int==aV.u.f_float?true:false));
      return true;
    case UNSIGNED_SHORT:
      set((bool)(u.f_int==(int)aV.u.f_unsigned_short?true:false));
      return true;
    case UNSIGNED_INTEGER:
      set((bool)(u.f_int==(int)aV.u.f_unsigned_int ?true:false));
      return true;
    case VECTOR_DOUBLE:{
      const std::vector<double>& stdv = aV.u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(u.f_int!=stdv[index]) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }

    case NONE:
    case STRING:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
      break;
    }
    break;
  case FLOAT:
    switch(aV.fType) {
    case FLOAT:
      set((bool)(u.f_float==aV.u.f_float?true:false));
      return true;
    case DOUBLE:
      set((bool)(u.f_float==aV.u.f_double?true:false));
      return true;
    case INTEGER:
      set((bool)(u.f_float==aV.u.f_int?true:false));
      return true;
    case UNSIGNED_SHORT:
      set((bool)(u.f_float==aV.u.f_unsigned_short?true:false));
      return true;
    case UNSIGNED_INTEGER:
      set((bool)(u.f_float==aV.u.f_unsigned_int?true:false));
      return true;
    case VECTOR_DOUBLE:{
      const std::vector<double>& stdv = aV.u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(u.f_float!=stdv[index]) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case NONE:
    case STRING:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case UNSIGNED_CHAR:
    case CHAR:
    case INT64:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
      break;
    }
    break;
  case DOUBLE:
    switch(aV.fType) {
    case DOUBLE:
      set((bool)(u.f_double==aV.u.f_double?true:false));
      return true;
    case FLOAT:
      set((bool)(u.f_double==aV.u.f_float?true:false));
      return true;
    case INTEGER:
      set((bool)(u.f_double==aV.u.f_int?true:false));
      return true;
    case UNSIGNED_SHORT:
      set((bool)(u.f_double==aV.u.f_unsigned_short?true:false));
      return true;
    case UNSIGNED_INTEGER:
      set((bool)(u.f_double==aV.u.f_unsigned_int ?true:false));
      return true;
    case VECTOR_DOUBLE:{
      const std::vector<double>& stdv = aV.u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(u.f_double!=stdv[index]) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case NONE:
    case STRING:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case UNSIGNED_CHAR:
    case CHAR:
    case INT64:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
      break;
    }
    break;
  case STRING:
    switch(aV.fType) {
    case STRING:
      set((bool)((*u.f_std_string)==(*aV.u.f_std_string)?true:false));
      return true;
    case INTEGER:
     {int v;
      if(!Lib::smanip::toint(*u.f_std_string,v)) break;
      set((bool)(v==aV.u.f_int?true:false));
      return true;}
    case FLOAT:
     {float v;
      if(!Lib::smanip::tofloat(*u.f_std_string,v)) break;
      set((bool)(v==aV.u.f_float?true:false));
      return true;}
    case DOUBLE:
     {double v;
      if(!Lib::smanip::todouble(*u.f_std_string,v)) break;
      set((bool)(v==aV.u.f_double?true:false));
      return true;}
    case UNSIGNED_INTEGER:
     {unsigned int v;
      if(!Lib::smanip::touint(*u.f_std_string,v)) break;
      set((bool)(v==aV.u.f_unsigned_int ?true:false));
      return true;}
    case BOOLEAN:
     {bool v;
      if(!Lib::smanip::tobool(*u.f_std_string,v)) break;
      set((bool)(v==aV.u.f_bool?true:false));
      return true;}
    case NONE:
    case POINTER:
    case SHORT:
    case UNSIGNED_CHAR:
    case CHAR:
    case INT64:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case VECTOR_DOUBLE:
      break;
    }
    break;
  case POINTER:
    switch(aV.fType) {
    case POINTER:
      set((bool)(u.f_void_star==aV.u.f_void_star?true:false));
      return true;
    case NONE:
    case INTEGER:
    case DOUBLE:
    case BOOLEAN:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INTEGER:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
    case VECTOR_DOUBLE:
      break;
    }
    break;
  case UNSIGNED_INTEGER:
    switch(aV.fType) {
    case UNSIGNED_INTEGER:
      set((bool)(u.f_unsigned_int ==aV.u.f_unsigned_int ?true:false));
      return true;
    case INTEGER:
      set((bool)((int)u.f_unsigned_int ==aV.u.f_int?true:false));
      return true;
    case FLOAT:
      set((bool)((int)u.f_unsigned_int ==aV.u.f_float?true:false));
      return true;
    case DOUBLE:
      set((bool)((int)u.f_unsigned_int ==aV.u.f_double?true:false));
      return true;
    case VECTOR_DOUBLE:{
      const std::vector<double>& stdv = aV.u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(u.f_unsigned_int !=stdv[index]) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case NONE:
    case BOOLEAN:
    case POINTER:
    case SHORT:
    case INT64:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
      break;
    }
    break;
  case BOOLEAN:
    switch(aV.fType) {
    case BOOLEAN:
      u.f_bool = ((u.f_bool==aV.u.f_bool)?true:false);
      return true;
    case NONE:
    case INTEGER:
    case DOUBLE:
    case POINTER:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INTEGER:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case STRING:
    case VECTOR_DOUBLE:
      break;
    }
    break;
  case VECTOR_DOUBLE:
    switch(aV.fType) {
    case INTEGER:{
      const std::vector<double>& stdv = u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(stdv[index]!=aV.u.f_int) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case FLOAT:{
      const std::vector<double>& stdv = u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(stdv[index]!=aV.u.f_float) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case DOUBLE:{
      const std::vector<double>& stdv = u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(stdv[index]!=aV.u.f_double) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case UNSIGNED_INTEGER:{
      const std::vector<double>& stdv = u.f_Array_double->vector();
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(stdv[index]!=aV.u.f_unsigned_int ) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case VECTOR_DOUBLE:{
      const std::vector<double>& stdv = u.f_Array_double->vector();
      const std::vector<double>& stdv2 = aV.u.f_Array_double->vector();
      if(stdv.size()!=stdv2.size()) {
        set(false);
        return true;
      }
      unsigned int number = stdv.size();
      for(unsigned int index=0;index<number;index++) {
        if(stdv[index]!=stdv2[index]) {
          set(false);
          return true;
        }
      }
      set(true);
      return true;
    }
    case NONE:
    case POINTER:
    case BOOLEAN:
    case SHORT:
    case UNSIGNED_CHAR:
    case CHAR:
    case STRING:
    case INT64:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
      break;
    }
    break;
  case NONE:
  case SHORT:
  case INT64:
  case UNSIGNED_CHAR:
  case CHAR:
  case UNSIGNED_SHORT:
  case UNSIGNED_INT64:
  case DOUBLE_STAR:
  case FLOAT_STAR:
  case INT_STAR:
    break;
  }
  aError = std::string("Variable::if_eq() :")
         + " can't compare type " + stype(oType) 
         + " with type " + aV.stype();
  setNone();
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::if_ne(
 const Variable& aV
,std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!if_eq(aV,aError)) return false;
  u.f_bool = (u.f_bool?false:true);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::if_ge(
 const Variable& aV
,std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Variable tmp(*this);
  if(!if_eq(aV,aError)) return false;
  if(u.f_bool) return true;

  // then not equal, check gt :
  if(!tmp.if_gt(aV,aError)) return false;

  u.f_bool = tmp.u.f_bool;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::if_lt(
 const Variable& aV
,std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!if_ge(aV,aError)) return false;
  u.f_bool = (u.f_bool?false:true);  
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::if_le(
 const Variable& aV
,std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!if_gt(aV,aError)) return false;
  u.f_bool = (u.f_bool?false:true);  
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::if_and(
 const Variable& aV
,std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  eType oType = fType;
  switch(fType) {
  case BOOLEAN:
    switch(aV.fType) {
    case BOOLEAN:
      u.f_bool = ((u.f_bool && aV.u.f_bool)?true:false);
      return true;
    case INTEGER:
    case VECTOR_DOUBLE:
    case NONE:
    case DOUBLE:
    case STRING:
    case POINTER:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case UNSIGNED_INTEGER:
      break;
    }
    break;
  case NONE:
  case INTEGER:
  case DOUBLE:
  case POINTER:
  case SHORT:
  case INT64:
  case FLOAT:
  case UNSIGNED_CHAR:
  case CHAR:
  case UNSIGNED_SHORT:
  case UNSIGNED_INTEGER:
  case UNSIGNED_INT64:
  case DOUBLE_STAR:
  case FLOAT_STAR:
  case INT_STAR:
  case STRING:
  case VECTOR_DOUBLE:
    break;
  }
  aError = std::string("Variable::if_and :")
         + " can't apply on type " + stype(oType) 
         + " and type " + aV.stype();
  setNone();
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::if_or(
 const Variable& aV
,std::string& aError
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  eType oType = fType;
  switch(fType) {
  case BOOLEAN:
    switch(aV.fType) {
    case BOOLEAN:
      u.f_bool = ((u.f_bool || aV.u.f_bool)?true:false);
      return true;
    case INTEGER:
    case VECTOR_DOUBLE:
    case NONE:
    case DOUBLE:
    case STRING:
    case POINTER:
    case SHORT:
    case INT64:
    case FLOAT:
    case UNSIGNED_CHAR:
    case CHAR:
    case UNSIGNED_SHORT:
    case UNSIGNED_INT64:
    case DOUBLE_STAR:
    case FLOAT_STAR:
    case INT_STAR:
    case UNSIGNED_INTEGER:
      break;
    }
    break;
  case NONE:
  case INTEGER:
  case DOUBLE:
  case POINTER:
  case SHORT:
  case INT64:
  case FLOAT:
  case UNSIGNED_CHAR:
  case CHAR:
  case UNSIGNED_SHORT:
  case UNSIGNED_INTEGER:
  case UNSIGNED_INT64:
  case DOUBLE_STAR:
  case FLOAT_STAR:
  case INT_STAR:
  case STRING:
  case VECTOR_DOUBLE:
    break;
  }
  aError = std::string("Variable::if_or :")
         + " can't apply on type " + stype(oType) 
         + " and type " + aV.stype();
  setNone();
  return false;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
unsigned char Lib::Variable::get_unsigned_char() const {
  return u.f_unsigned_char;
}
char Lib::Variable::get_char() const {
  return u.f_char;
}
unsigned short Lib::Variable::get_unsigned_short() const {
  return u.f_unsigned_short;
}
unsigned int Lib::Variable::get_unsigned_int() const {
  return u.f_unsigned_int;
}
int Lib::Variable::get_int() const {
  return u.f_int;
}
Lib::int64 Lib::Variable::get_int64() const {
  return u.f_int64;
}
short Lib::Variable::get_short() const {
  return u.f_short;
}
float Lib::Variable::get_float() const {
  return u.f_float;
}
double Lib::Variable::get_double() const {
  return u.f_double;
}
void* Lib::Variable::get_void_star() const {
  return u.f_void_star;
}
double* Lib::Variable::get_double_star() const {
  return u.f_double_star;
}
float* Lib::Variable::get_float_star() const {
  return u.f_float_star;
}
int* Lib::Variable::get_int_star() const {
  return u.f_int_star;
}
bool Lib::Variable::get_bool() const {
  return u.f_bool;
}
std::string Lib::Variable::get_std_string() const {
  return *u.f_std_string;
}
Lib::Array<double>& Lib::Variable::get_Array_double() {
  return *u.f_Array_double;
}

//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 char aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = CHAR;
  u.f_char = aValue;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 short aValue
)
//////////////////////////////////////////////////////////////////////////////
{
  reset();
  fType = SHORT;
  u.f_short = aValue;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 int aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = INTEGER;
  u.f_int = aValue;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 float aValue
)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = FLOAT;
  u.f_float = aValue;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 double aValue
)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = DOUBLE;
  u.f_double = aValue;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 unsigned char aValue
)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = UNSIGNED_CHAR;
  u.f_unsigned_char = aValue;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 unsigned short aValue
)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = UNSIGNED_SHORT;
  u.f_unsigned_short  = aValue;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 unsigned int aValue
)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = UNSIGNED_INTEGER;
  u.f_unsigned_int  = aValue;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 Lib::int64 aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = INT64;
  u.f_int64 = aValue;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 Lib::uint64 aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = UNSIGNED_INT64;
  u.f_unsigned_int64 = aValue;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 const std::string& aValue
)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = STRING;
  u.f_std_string = new std::string(aValue);
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 const char* aValue
)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = STRING;
  u.f_std_string = new std::string(aValue);
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 const Array<double>& aValue
)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = VECTOR_DOUBLE;
  u.f_Array_double = new Array<double>(aValue);
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 void* aValue
)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = POINTER;
  u.f_void_star = aValue;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 double* aValue
)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = DOUBLE_STAR;
  u.f_double_star = aValue;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 float* aValue
)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = FLOAT_STAR;
  u.f_float_star = aValue;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 int* aValue
)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = INT_STAR;
  u.f_int_star = aValue;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::Variable::set(
 bool aValue
)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();
  fType = BOOLEAN;
  u.f_bool = aValue;
}

#include <Slash/Core/IValue.h>
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool Lib::Variable::set(
 const Slash::Core::IValue& aValue
)
/////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  reset();

  switch(aValue.type()) {
  case Slash::Core::IValue::NONE:
    fType = Variable::NONE;
    u.f_unsigned_int64 = 0;
    return true;
  case Slash::Core::IValue::CHAR:
    fType = Variable::CHAR;
    u.f_char = aValue.get_char();
    return true;
  case Slash::Core::IValue::SHORT:
    fType = Variable::SHORT;
    u.f_short = aValue.get_short();
    return true;
  case Slash::Core::IValue::INT:
    fType = Variable::INTEGER;
    u.f_int = aValue.get_int();
    return true;
  case Slash::Core::IValue::INT64:
    fType = Variable::INT64;
    u.f_int64 = aValue.get_int64();
    return true;
  case Slash::Core::IValue::UNSIGNED_INT64:
    fType = Variable::UNSIGNED_INT64;
    u.f_unsigned_int64 = aValue.get_unsigned_int64();
    return true;
  case Slash::Core::IValue::UNSIGNED_CHAR:
    fType = Variable::UNSIGNED_CHAR;
    u.f_unsigned_char = aValue.get_unsigned_char();
    return true;
  case Slash::Core::IValue::UNSIGNED_SHORT:
    fType = Variable::UNSIGNED_SHORT;
    u.f_unsigned_short  = aValue.get_unsigned_short();
    return true;
  case Slash::Core::IValue::UNSIGNED_INT:
    fType = Variable::UNSIGNED_INTEGER;
    u.f_unsigned_int  = aValue.get_unsigned_int();
    return true;
  case Slash::Core::IValue::DOUBLE_STAR:
    fType = Variable::DOUBLE_STAR;
    u.f_double_star  = aValue.get_double_star();
    return true;
  case Slash::Core::IValue::FLOAT_STAR:
    fType = Variable::FLOAT_STAR;
    u.f_float_star  = aValue.get_float_star();
    return true;
  case Slash::Core::IValue::INT_STAR:
    fType = Variable::INT_STAR;
    u.f_int_star  = aValue.get_int_star();
    return true;
  case Slash::Core::IValue::FLOAT:
    fType = Variable::FLOAT;
    u.f_float = aValue.get_float();
    return true;
  case Slash::Core::IValue::DOUBLE:
    fType = Variable::DOUBLE;
    u.f_double = aValue.get_double();
    return true;
  case Slash::Core::IValue::VOID_STAR:
    fType = Variable::POINTER;
    u.f_void_star = aValue.get_void_star();
    return true;
  case Slash::Core::IValue::BOOL:
    fType = Variable::BOOLEAN;
    u.f_bool = aValue.get_bool();
    return true;
  case Slash::Core::IValue::STRING:
    fType = Variable::STRING;
    u.f_std_string = new std::string(aValue.get_string());
    return true;
/*
  case Slash::Core::IValue::STD_VECTOR_DOUBLE:
   {fType = Variable::VECTOR_DOUBLE;
    u.f_Array_double = new Array<double>();
    std::vector<double> vd = aValue.get_std_vector_double();
    unsigned int n = vd.size();
    std::vector<unsigned int> is(1);
    is[0] = n;         
    if(!u.f_Array_double->configure(is)) { //FIXME
    }
    for(unsigned int index=0;index<n;index++) {
      is[0] = index;         
      if(!u.f_Array_double->setValue(is,vd[index])) { //FIXME
      }
    }}
    return true;
  }
*/
  case Slash::Core::IValue::ARRAY_DOUBLE:
   {fType = Variable::VECTOR_DOUBLE;
    std::vector<unsigned int> is;
    const std::vector<double>& vd = aValue.get_array_double(is);
    u.f_Array_double = new Array<double>();
    if(!u.f_Array_double->configure(is)) {
      setNone();
      return false;
    }
    if(!u.f_Array_double->fill(vd)) {
      setNone();
      return false;
    }
    return true;}
  case Slash::Core::IValue::ARRAY_UNSIGNED_CHAR:
  case Slash::Core::IValue::ARRAY_CHAR:
  case Slash::Core::IValue::ARRAY_UNSIGNED_SHORT:
  case Slash::Core::IValue::ARRAY_SHORT:
  case Slash::Core::IValue::ARRAY_UNSIGNED_INT:
  case Slash::Core::IValue::ARRAY_INT:
  case Slash::Core::IValue::ARRAY_UNSIGNED_INT64:
  case Slash::Core::IValue::ARRAY_INT64:
  case Slash::Core::IValue::ARRAY_FLOAT:
  //case Slash::Core::IValue::ARRAY_DOUBLE:
  case Slash::Core::IValue::ARRAY_BOOL:
  case Slash::Core::IValue::ARRAY_STRING:
    break;
  }
  setNone();
  return false;
}

#include <Lib/Value.h>
//////////////////////////////////////////////////////////////////////////////
Slash::Core::IValue* Lib::Variable::toValue(
 const Lib::Variable& aVariable
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  switch(aVariable.fType) {
  case NONE:
    return new Lib::Value();
  case INTEGER:
    return new Lib::Value(aVariable.u.f_int);
  case DOUBLE:
    return new Lib::Value(aVariable.u.f_double);
  case POINTER:
    return new Lib::Value(aVariable.u.f_void_star);
  case UNSIGNED_SHORT:
    return new Lib::Value(aVariable.u.f_unsigned_short);
  case UNSIGNED_INTEGER:
    return new Lib::Value(aVariable.u.f_unsigned_int);
  case DOUBLE_STAR:
    return new Lib::Value(aVariable.u.f_double_star);
  case FLOAT_STAR:
    return new Lib::Value(aVariable.u.f_float_star);
  case INT_STAR:
    return new Lib::Value(aVariable.u.f_int_star);
  case BOOLEAN:
    return new Lib::Value(aVariable.u.f_bool);
  case SHORT:
    return new Lib::Value(aVariable.u.f_short);
  case FLOAT:
    return new Lib::Value(aVariable.u.f_float);
  case UNSIGNED_CHAR:
    return new Lib::Value(aVariable.u.f_unsigned_char);
  case CHAR:
    return new Lib::Value(aVariable.u.f_char);
  case STRING:
    return new Lib::Value(*aVariable.u.f_std_string);
  case VECTOR_DOUBLE:
    return new Lib::Value(aVariable.u.f_Array_double->orders(),
                          aVariable.u.f_Array_double->vector());
  //FIXME
  case INT64:
  case UNSIGNED_INT64:
    return 0;
  }
}

/*
  case NONE:
  case INTEGER:
  case DOUBLE:
  case POINTER:
  case BOOLEAN:
  case SHORT:
  case INT64:
  case FLOAT:
  case CHAR:
  case UNSIGNED_CHAR:
  case UNSIGNED_SHORT:
  case UNSIGNED_INTEGER:
  case UNSIGNED_INT64:
  case STRING:
  case VECTOR_DOUBLE:
*/

