// this :
#include <Lib/ZipWriter.h>

#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/Debug.h>

#ifdef LIB_HAS_ZLIB
#include <zip.h>
#endif

//////////////////////////////////////////////////////////////////////////////
bool Lib::ZipWriter::available(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  return true;
#else
  return false;
#endif
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Lib::ZipWriter::ZipWriter(
 const std::string& aFileName
,Slash::Core::IWriter& aPrinter
)
:fPrinter(aPrinter)
,fFileName(aFileName)
,fFile(0)
,fEntryOpened(false)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("Lib::ZipWriter");
#ifdef LIB_HAS_ZLIB
  //fFile is a zipFile.
  fFile = zipOpen(fFileName.c_str(),APPEND_STATUS_CREATE);
  if(!fFile) {
    Lib::Out out(fPrinter);
    out << "Lib::ZipWriter::ZipWriter :"
        << "error opening " << Lib::sout(fFileName)
        << Lib::endl;
  }
#else
  Lib::Out out(fPrinter);
  out << "Lib::ZipWriter::ZipWriter :"
      << " Lib package not reconstructed with zlib." 
      << Lib::endl;
#endif
}
//////////////////////////////////////////////////////////////////////////////
Lib::ZipWriter::~ZipWriter(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  if(fFile) {
    if(zipClose((zipFile)fFile,NULL)!=ZIP_OK) {
      Lib::Out out(fPrinter);
      out << "Lib::ZipWriter::~ZipWriter :"
          << " closing zip file " << Lib::sout(fFileName) << " failed."
          << Lib::endl;
    }
  }
#endif
  Lib::Debug::decrement("Lib::ZipWriter");
}
//////////////////////////////////////////////////////////////////////////////
void Lib::ZipWriter::disable(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  if(!fFile) return; // Already closed.
  if(zipClose((zipFile)fFile,NULL)!=ZIP_OK) {
    Lib::Out out(fPrinter);
    out << "Lib::ZipWriter::disable :"
        << " closing zip file " << Lib::sout(fFileName) << " failed."
        << Lib::endl;
  }
  fFile = 0;
  fEntryOpened = false;
#endif
}
//////////////////////////////////////////////////////////////////////////////
void Lib::ZipWriter::enable(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  if(fFile) return; // Already open.
  fFile = zipOpen(fFileName.c_str(),APPEND_STATUS_CREATE);
  if(!fFile) {
    Lib::Out out(fPrinter);
    out << "Lib::ZipWriter::enable :"
        << "error opening " << Lib::sout(fFileName)
        << Lib::endl;
  }
  fEntryOpened = false;
#endif
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::ZipWriter::enabled(
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fFile ? true : false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::ZipWriter::flush(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::ZipWriter::write(
 const std::string& aString
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  if(!fFile) return false;

  int size_write = aString.size();
  if(zipWriteInFileInZip((zipFile)fFile,aString.c_str(),size_write)<0) {
    Lib::Out out(fPrinter);
    out << "Lib::ZipWriter::close :"
        << " error writing in file " << Lib::sout(fFileName) 
        << Lib::endl;
    disable();
    return false;
  }
  return true;
#else
  return false;
#endif
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::ZipWriter::openEntry(
 const std::string& aString
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fFile) return false;

  if(fEntryOpened) return true; //Already opened. Do a closeEntry first.

  zip_fileinfo zi;

  zi.tmz_date.tm_sec = 0;
  zi.tmz_date.tm_min = 0;
  zi.tmz_date.tm_hour = 0;
  zi.tmz_date.tm_mday = 0;
  zi.tmz_date.tm_mon = 0;
  zi.tmz_date.tm_year = 0;
  //filetime(filenameinzip,&zi.tmz_date,&zi.dosDate);

  zi.dosDate = 0;
  zi.internal_fa = 0;
  zi.external_fa = 0;

  int err = 
    //zipOpenNewFileInZip3(zf,file_in_zip.c_str(),&zi,
    zipOpenNewFileInZip((zipFile)fFile,aString.c_str(),&zi,
      NULL,0,NULL,0,NULL /* comment*/,
      Z_DEFLATED,Z_DEFAULT_COMPRESSION);
  if(err!=ZIP_OK) {
    Lib::Out out(fPrinter);
    out << "Lib::ZipWriter::open :"
        << " can't open entry " << Lib::sout(aString)
        << " in file " << Lib::sout(fFileName) 
        << Lib::endl;
    disable();
    return false;
  }

  fEntryOpened = true;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::ZipWriter::closeEntry(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fFile) return false;
  if(!fEntryOpened) return true;
  if(zipCloseFileInZip((zipFile)fFile)!=ZIP_OK) {
    Lib::Out out(fPrinter);
    out << "Lib::ZipWriter::close :"
        << " can't close entry in file" << Lib::sout(fFileName) 
        << Lib::endl;
    disable();
    return false;
  }
  fEntryOpened = false;
  return true;
}
