// this :
#include <Lib/fmanip.h>

#include <Lib/smanip.h>
#include <Lib/System.h>
#include <Lib/vmanip.h>

#include <stdio.h>

//////////////////////////////////////////////////////////////////////////////
std::string Lib::fmanip::name(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string fname = aString;
  // Check for env variable in name :
  smanip::getenv(fname);
  if(System::isKindOf("Windows")) {
    smanip::replace(fname,"/","\\");
    smanip::replace(fname,"\"","");
  } else if(System::isKindOf("VMS")) {
    smanip::replace(fname,"/",":");
  }
  return fname;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::exists(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s = name(aString);
  FILE* file = ::fopen(s.c_str(),"rb");
  if(!file) return false;
  ::fclose(file);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::remove(
 const std::string& aName
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return (::remove(aName.c_str()) ==0 ? true : false);
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::rename(
 const std::string& aName
,const std::string& aNewName
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //NOTE : aName must not be a path !
  //       Darwin is ok with a path but not Linux !
  //       For example : 
  //         ::rename("/tmp/tmp01"."x");
  //       return -1 on Linux.
  //       To do the upper then someone must use move. 
  //       But there is no move in the standard lib C !
  return (::rename(aName.c_str(),aNewName.c_str()) == 0 ? true : false);
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::copy(
 const std::string& aFrom
,const std::string& aTo
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aTo==aFrom) return true;
  if(System::isKindOf("UNIX")) {
    std::string cmd = "/bin/cp ";
    cmd += Lib::System::quote(aFrom);
    cmd += " ";
    cmd += Lib::System::quote(aTo);
    System::execute(cmd);
    return true;
  } else if(System::isKindOf("Windows")) {
    std::string cmd = "COPY /Y ";
    cmd += Lib::System::quote(aFrom);
    cmd += " ";
    cmd += Lib::System::quote(aTo);
    System::execute(cmd);
    return true;
  } else {
    return false;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::move(
 const std::string& aFrom
,const std::string& aTo
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aTo==aFrom) return true;
  if(System::isKindOf("UNIX")) {
    std::string cmd = "/bin/mv ";
    cmd += Lib::System::quote(aFrom);
    cmd += " ";
    cmd += Lib::System::quote(aTo);
    System::execute(cmd);
    return true;
  } else if(System::isKindOf("Windows")) {
    std::string cmd = "MOVE /Y ";
    cmd += Lib::System::quote(aFrom);
    cmd += " ";
    cmd += Lib::System::quote(aTo);
    System::execute(cmd);
    return true;
  } else {
    return false;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::rm(
 const std::string& aWhat
)
//////////////////////////////////////////////////////////////////////////////
// WARNING : do not confuse with fmanip::remove which is 
//           an encapsulation of the remove of stdio.h that
//           does not support wildcards.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(System::isKindOf("UNIX")) {
    std::string cmd = "/bin/rm -f ";
    cmd += Lib::System::quote(aWhat);
    System::execute(cmd);
    return true;
  } else if(System::isKindOf("Windows")) {
    std::string cmd = "DEL /Q ";
    cmd += Lib::System::quote(aWhat);
    System::execute(cmd);
    return true;
  } else {
    return false;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::is_f77(
 const std::string& aName
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string suffix = Lib::smanip::suffix(aName);
  Lib::smanip::tolowercase(suffix);
  if(suffix=="f") return true;       //the standard.
  if(suffix=="for") return true;     //for opaw. Known by g77.
  if(suffix=="ftn") return true;     //for opaw.
  if(suffix=="fortran") return true; //for opaw.
  if(suffix=="f77") return true;
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::is_cpp(
 const std::string& aName
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string suffix = Lib::smanip::suffix(aName);
  Lib::smanip::tolowercase(suffix);
  if(suffix=="c") return true;
  if(suffix=="cxx") return true;
  if(suffix=="cpp") return true;
  if(suffix=="C") return true;
  return false;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::fmanip::fortranLink(
 std::string& aName
)
//////////////////////////////////////////////////////////////////////////////
// If aName contains upper case RZOPEN is unable to open.
// We do a temporary lower case named link into /tmp
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aName.empty()) return;
  if(System::isKindOf("UNIX")) {
    std::string lowername = aName;
    smanip::tolowercase(lowername);
    if(lowername==aName) return; 
    // Contain upper case :
    
    std::string cmd = "ln -f -s " + aName + " ";
    
    std::string path,name,suffix;
    smanip::pathNameSuffix(aName,path,name,suffix);
    
    std::string fname;
    if(path.size()) { // Some directory path in aName :
      smanip::tolowercase(name);
      fname = "/tmp/" + name;
    } else {
      // No directory in path name but some upper case in file name.
      smanip::tolowercase(name);
      fname = name;
    }
    
    cmd += fname;
    
    ::remove(fname.c_str());
    System::execute(cmd);

    aName = fname;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::bindToFortranUnit(
 const std::string& aName
,int aUnit
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aUnit<=0) return false;
  if(System::isKindOf("UNIX")) {
    std::string sf77 = System::fortranUnitFormat();
    std::string s;
    smanip::printf(s,64,sf77.c_str(),aUnit);
    ::remove(s.c_str());

    std::string cmd = "ln -f -s ";
    cmd += Lib::System::quote(aName);
    cmd += " ";
    cmd += Lib::System::quote(s);
    System::execute(cmd);

    FILE* file = ::fopen(s.c_str(),"rb");
    if(!file) return false; // Can't bind file to fortran unit.
    ::fclose(file);
    return true;
  } else if(System::isKindOf("Windows")) {
    // Symbolic link does not exsit, do a copy !
    std::string sf77 = System::fortranUnitFormat();
    std::string s;
    smanip::printf(s,64,sf77.c_str(),aUnit);
    ::remove(s.c_str());
    
    std::string cmd = "COPY /Y ";
    cmd += Lib::System::quote(aName);
    cmd += " ";
    cmd += Lib::System::quote(s);
    System::execute(cmd);
    
    FILE* file = ::fopen(s.c_str(),"rb");
    if(!file) return false; // Can't bind file to fortran unit.
    ::fclose(file);
    return true;
  } else if(System::isKindOf("VMS")) {
    std::string cmd;
    smanip::printf(cmd,64,"$ define/job for%3.3d ",aUnit);
    cmd += aName;
    System::execute(cmd);
    return true;
  } else {
    return false;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::readBytes(
 const std::string& aName
,char*& aBuffer
,uint64& aLength
)
//////////////////////////////////////////////////////////////////////////////
// Return buffer should be deleted with delete [].
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string name = fmanip::name(aName);
  FILE* file = ::fopen(name.c_str(),"rb");
  if(!file) {
    aBuffer = 0;
    aLength = 0;
    return false;
  }
  // Get file size :
  ::fseek(file,0L,SEEK_END);
  long filesize = ::ftell(file);
  if(!filesize) {
    ::fclose(file);
    aBuffer = new char[1];
    aLength = 0;
    return true; //empty file.
  }
  // Add one byte to be able to add \n if file contain lines.
  aBuffer = new char[filesize+1];
  if(!aBuffer) {
    ::fclose(file);
    aBuffer = 0;
    aLength = 0;
    return false;
  }
  ::rewind(file);
  size_t nitem = ::fread(aBuffer,(size_t)filesize,(size_t)1,file);
  if(nitem!=1){
    ::fclose(file);
    delete [] aBuffer;
    aBuffer = 0;
    aLength = 0;
    return false;
  }
  ::fclose(file);
  aBuffer[filesize] = 0;
  aLength = filesize;
  return true;
}  
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::writeBytes(
 const std::string& aName
,const char* aBuffer
,uint64 aLength
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string name = fmanip::name(aName);
  FILE* file = ::fopen(name.c_str(),"wb");
  if(!file) return false;
  if(!aLength) {
    ::fclose(file);
    return true;
  }
  size_t nitem = ::fwrite((char*)aBuffer,(size_t)aLength,(size_t)1,file);
  ::fclose(file);
  return (nitem!=1?false:true);
}  
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::replace(
 const std::string& aName
,const std::string& aOld
,const std::string& aNew
,std::vector<std::string>& aChanges
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aChanges.clear();
  std::vector<std::string> text;
  if(!read(aName,text)) return false;
  // Replace :
  bool touched = false;
 {unsigned int textn = text.size();
  for(unsigned index=0;index<textn;index++) {
    std::string old = text[index];
    if(smanip::replace(text[index],aOld,aNew)) {
      //printf("debug :\"%s\"\n",text[index].c_str());
      aChanges.push_back(old);
      aChanges.push_back(text[index]);
      touched = true;
    }
  }}
  if(!touched) return true;
  // Create temporary file :
  //NOTE : have a tmpName in current directory so that the below 
  //       rename works. (Else we should do a move).
  std::string tmpName;
  if(!tmpname(tmpName,".","Lib_")) return false;
  FILE* file = ::fopen(tmpName.c_str(),"wb");
  if(!file) return false;
 {unsigned int textn = text.size();
  for(unsigned index=0;index<textn;index++) {
    if(::fprintf(file,"%s\n",text[index].c_str())<0) {
      ::fclose(file);
      remove(tmpName);
      return false;
    }
  }}
  ::fclose(file);
  if(!remove(aName)) return false;
  if(!rename(tmpName,aName)) return false;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::removeLines(
 const std::string& aName
,const std::string& aToRemove
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> text;
  if(!read(aName,text)) return false;
  bool touched = false;
  std::vector<std::string>::iterator it;
  for(it = text.begin();it!=text.end();) {
    std::string line = (*it);
    smanip::strip(line);
    if(line==aToRemove) {
      text.erase(it); 
      touched = true;
    } else {
      it++;
    }
  }
  if(!touched) return true;
  // Create temporary file :
  //NOTE : have a tmpName in current directory so that the below 
  //       rename works. (Else we should do a move).
  std::string tmpName;
  if(!tmpname(tmpName,".","Lib_")) return false;
  FILE* file = ::fopen(tmpName.c_str(),"wb");
  if(!file) return false;
 {unsigned int textn = text.size();
  for(unsigned index=0;index<textn;index++) {
    if(::fprintf(file,"%s\n",text[index].c_str())<0) {
      ::fclose(file);
      remove(tmpName);
      return false;
    }
  }}
  ::fclose(file);
  if(!rename(tmpName,aName)) return false;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::if_rm_eol(
 const std::string& aName
,const std::string& aIF
,const std::string& aRM
,std::vector<std::string>& aChanges
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aChanges.clear();
  std::vector<std::string> text;
  if(!read(aName,text)) return false;
  bool touched = false;
  std::vector<std::string>::iterator it;
  for(it = text.begin();it!=text.end();++it) {
    std::string old = (*it);
    std::string& line = (*it);
    if(line.find(aIF)!=std::string::npos) {
      std::string::size_type pos = line.find(aRM);
      if(pos!=std::string::npos) {
        line = line.substr(0,pos);
        aChanges.push_back(old);
        aChanges.push_back(line);
        touched = true;
      }
    }
  }
  if(!touched) return true;
  // Create temporary file :
  //NOTE : have a tmpName in current directory so that the below 
  //       rename works. (Else we should do a move).
  std::string tmpName;
  if(!tmpname(tmpName,".","Lib_")) return false;
  FILE* file = ::fopen(tmpName.c_str(),"wb");
  if(!file) return false;
 {unsigned int textn = text.size();
  for(unsigned index=0;index<textn;index++) {
    if(::fprintf(file,"%s\n",text[index].c_str())<0) {
      ::fclose(file);
      remove(tmpName);
      return false;
    }
  }}
  ::fclose(file);
  if(!rename(tmpName,aName)) return false;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::found(
 const std::string& aName
,const std::string& aWhat
,std::vector<std::string>& aFound
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aFound.clear();
  std::vector<std::string> text;
  if(!read(aName,text)) return false;
  unsigned int textn = text.size();
  for(unsigned int index=0;index<textn;index++) {
    const std::string& line = text[index];
    if(line.find(aWhat)!=std::string::npos) {
      aFound.push_back(line);
    }
  }
  return true;
}
#include <string.h> //strlen
#define BUFSIZE 65536
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::read(
 const std::string& aName
,std::vector<double>& aDoubles
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aDoubles.clear();
  std::string name = fmanip::name(aName);
  FILE* file = ::fopen(name.c_str(),"rb");
  if(!file) return false;
  std::string sep = " ";
  char buffer[BUFSIZE+1];
  while(true) {
    if(::fgets(buffer,BUFSIZE,(FILE*)file)==NULL) break; // EOF.
    int l = ::strlen(buffer);
    //  On Windows, editors when saving binary files,
    // put \r\n at place of \n ; we then look for \r\n.
    if( (l>=2) && (buffer[l-2]=='\r') && (buffer[l-1]=='\n') ) {
      buffer[l-2] = '\0';
    } else if( (l>=1) && (buffer[l-1]=='\n') ) {
      buffer[l-1] = '\0';
    }
    std::string sbuffer = buffer;
    std::vector<double> ds;
    if(!smanip::todoubles(sbuffer,ds,sep)) {
      aDoubles.clear();
      ::fclose(file);
      return false;
    }
    Lib::append<double>(aDoubles,ds);
  }
  ::fclose(file);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::write(
 const std::string& aName
,const std::vector<std::string>& aText
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string name = fmanip::name(aName);
  FILE* file = ::fopen(name.c_str(),"wb");
  if(!file) return false;
  unsigned int textn = aText.size();
  for(unsigned index=0;index<textn;index++) {
    if(::fprintf(file,"%s\n",aText[index].c_str())<0) {
      ::fclose(file);
      return false;
    }
  }
  ::fclose(file);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::write(
 const std::string& aName
,const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
// aString could contain \n separated lines.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string name = fmanip::name(aName);
  FILE* file = ::fopen(name.c_str(),"wb");
  if(!file) return false;
  if(::fprintf(file,"%s",aString.c_str())<0) {
    ::fclose(file);
    return false;
  }
  ::fclose(file);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::read(
 const std::string& aName
,std::vector<std::string>& aText
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aText.clear();
  //FIXME : should read without a limited buffer.
  std::string name = fmanip::name(aName);
  FILE* file = ::fopen(name.c_str(),"rb");
  if(!file) return false;
  char buffer[BUFSIZE+1];
  while(true) {
    if(::fgets(buffer,BUFSIZE,(FILE*)file)==NULL) break; // EOF.
    int l = ::strlen(buffer);
    //  On Windows, editors when saving binary files,
    // put \r\n at place of \n ; we then look for \r\n.
    if( (l>=2) && (buffer[l-2]=='\r') && (buffer[l-1]=='\n') ) {
      buffer[l-2] = '\0';
    } else if( (l>=1) && (buffer[l-1]=='\n') ) {
      buffer[l-1] = '\0';
    }
    aText.push_back(buffer);
  }
  ::fclose(file);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::fmanip::find(
 const std::vector<std::string>& aPaths
,const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(Lib::fmanip::exists(aString)) return aString;

  unsigned int number = aPaths.size();
  for(unsigned int index=0;index<number;index++) {
    std::string s = aPaths[index]+"/"+aString;    
    if(Lib::fmanip::exists(s)) return s;
  }
  return "";
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::fmanip::find(
 const std::string& aPaths
,const std::string& aSep
,const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return Lib::fmanip::find(smanip::words(aPaths,aSep),aString);
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::fmanip::towin(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string fname = aString;

  // string may contain ${xxx}
  //Lib::smanip::replace(fname,"${","%");
  //Lib::smanip::replace(fname,"}","%");
    
  Lib::smanip::replace(fname,"/","\\");
  return fname;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::findResource(
 const std::string& aFile
,const std::string& aWhat
,std::string& aValue
) 
//////////////////////////////////////////////////////////////////////////////
// Used by the xxx-config programs.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> txt;
  if(!Lib::fmanip::read(aFile,txt)) {
    ::fprintf(stderr,"Can't read the file :\n");
    ::fprintf(stderr,"%s\n",aFile.c_str());
    ::fprintf(stderr,
              "containing the platform dependant incs and libs infos.\n");
    aValue = "";
    return false;
  }
  if(!txt.size()) {
    ::fprintf(stderr,"The file :\n");
    ::fprintf(stderr,"%s\n",aFile.c_str());
    ::fprintf(stderr,"is empty. We expect to find in it\n");
    ::fprintf(stderr,"the platform dependant incs and libs infos.\n");
    aValue = "";
    return false;
  }
 {unsigned int number = txt.size();
  for(unsigned int index=0;index<number;index++){
    std::vector<std::string> words = Lib::smanip::words(txt[index],"=");
    unsigned int wordn = words.size();
    if((wordn==1)&&(words[0]==aWhat)){
      aValue = "";
      return true;
    } else if((wordn>=2)&&(words[0]==aWhat)){
      aValue = words[1];
      return true;
    }
  }}
  ::fprintf(stderr,"The file :\n");
  ::fprintf(stderr,"%s\n",aFile.c_str());
  ::fprintf(stderr,"does not contain the %s infos.\n",aWhat.c_str());
  aValue = "";
  return false;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#include <sys/stat.h>
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::tmpname(
 std::string& aName
,const std::string& aDir
,const std::string& aPrefix
,const std::string& aSuffix
)
//////////////////////////////////////////////////////////////////////////////
//  For UNIX aDir must be a string ended with '/'. Exa : /tmp/ .
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  for(int count=0;count<1000;count++) {
    std::string name;
    smanip::printf(name,3,"%03d",count);
    std::string tmpnm;
    if(aDir==".") {
      tmpnm += aPrefix;
      tmpnm += name;
      tmpnm += aSuffix;
    } else if(aDir.size()) {
      tmpnm += aDir;
      tmpnm += aPrefix;
      tmpnm += name;
      tmpnm += aSuffix;
    } else {
#ifdef WIN32
      std::string env = System::getenv("HOMEDRIVE");
      std::string drive = (env!="" ? env : std::string("C:"));
      tmpnm += drive;
      tmpnm += "\\";
      tmpnm += aPrefix;
      tmpnm += name;
      tmpnm += aSuffix;
#else
      tmpnm += "/tmp/";
      tmpnm += aPrefix;
      tmpnm += name;
      tmpnm += aSuffix;
#endif
    }
    struct stat finfo;
    if(::stat(tmpnm.c_str(),&finfo) < 0) {
      // No directory or file found :
      aName = tmpnm; // Good file name.
      return true; 
    }
  }
  // No good name found :
  aName = "";
  return false;
}
#ifdef WIN32
#include <windows.h>
#include <io.h>
#endif
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::chmod(
 const std::string& aName
,unsigned int aMode
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef WIN32
  // Set the file permission bits. Returns -1 in case or error, 0 otherwise.
  // On windows mode can only be a combination of "user read" (0400),
  // "user write" (0200) or "user read | user write" (0600). Any other value
  // for mode are ignored.
  if(_chmod(aName.c_str(),aMode)<0) return false;
#else
  if(::chmod(aName.c_str(),aMode)<0) return false;
#endif
  return true;
}
#include <stdlib.h>
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::chmod(
 const std::string& aName
,eWho aWho
,eOperation aOp
,eWhat aWhat
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef WIN32
  //FIXME
  return false;
#else
  std::string cmd = "chmod ";
  switch(aWho){
  case USER: cmd += "u";break;
  case GROUP: cmd += "g";break;
  case OTHER: cmd += "o";break;
  case ALL: cmd += "a";break;
  //default:break;
  }
  switch(aOp){
  case ADD: cmd += "+";break;
  case SUBTRACT: cmd += "-";break;
  //default:break;
  }
  switch(aWhat){
  case READ: cmd += "r";break;
  case WRITE: cmd += "w";break;
  case EXECUTE: cmd += "x";break;
  //default:break;
  }
  cmd += " "+aName;
  int ret = ::system(cmd.c_str());
  return true;
#endif  
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::fmanip::header(
 const std::string& aString
,unsigned char aHeader[4]
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s = name(aString);
  FILE* file = ::fopen(s.c_str(),"rb");
  if(!file) return false;
  size_t nitem = ::fread(aHeader,1,4,file);
  ::fclose(file);
  if(nitem!=4) return false;
  return true;
}
