// this :
#include <Lib/smanip.h>

#include <Lib/Out.h>
#include <Lib/cmanip.h>
#include <Lib/cstring.h>

// else :
#include <stdio.h>
#include <stdarg.h>

#include <map>

static const int NotFound = -1;

#define MINIMUM(a,b) ((a)<(b)?a:b)

//////////////////////////////////////////////////////////////////////////////
void Lib::smanip::strip(
 std::string& aString
,what aType
,char aChar
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string::size_type l = aString.length();;
  if(l==0) return;

  switch ( aType ) {
  case leading:{
    std::string::size_type i;
    char* pos = (char*)aString.c_str();
    for(i=0;i<l;i++,pos++) {
      if(*pos!=aChar) {
        aString = aString.substr(i,l-i);
        return;
      }
    }
    }break;
  case trailing:{
    std::string::size_type i;
    char* pos = (char*)aString.c_str();
    pos += (l-1);
    for(i=l-1;;i--,pos--) {
      if(*pos!=aChar) {
        aString = aString.substr(0,i+1);
        return;
      }
    }
    }break;
  case both:
    Lib::smanip::strip(aString,leading,aChar);
    Lib::smanip::strip(aString,trailing,aChar);
    break;
  //default:break;
  }
}

std::string Lib::smanip::strp(const std::string& aString,what aType,char aChar){
  std::string s(aString);
  strip(s,aType,aChar);
  return s;
}

//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::match(
 const std::string& aString   
,const std::string& aPattern 
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int lpattern = aPattern.length();
  int lstring  = aString.length();
  if ((lpattern==0)&&(lstring==0)) return true;
  if ((lpattern==0)&&(lstring!=0)) return true;
  if ((lpattern!=0)&&(lstring==0)) return false;
  // pattern is * :
  if(aPattern=="*") return true;
  int wcount = 0;
  int count;
  for(count=0;count<lpattern;count++) {
    if(aPattern[count]=='*') wcount++;
  }
  // no wildcard :
  if(wcount==0) {
    return (aPattern==aString ? true : false );
  }

  // complex pattern :
#define PACKET 32
  char** words = (char**)::malloc(PACKET * sizeof(char*));
  if(words==NULL) return false;
  int wmax = PACKET;
  int wnumber= 0;
  char* s = Lib_strdup(aPattern.c_str());
  char* token = s;
  while(true) { 
    char* pos = ::strstr (token,"*");
    if(pos) {
      *pos = '\0';
      if(*token!='\0') {
        if(wnumber>=wmax) {
          words = (char**)::realloc(words,(wmax+PACKET) * sizeof(char*));
          if(words==NULL) {
            Lib_strdel(s);
            return false;
          }
          wmax += PACKET;
        }
        words[wnumber] = token;
        wnumber++;
      }
      token = pos + 1;
    } else { // last word :
      if(*token!='\0') {
        if(wnumber>=wmax) {
          words = (char**)::realloc(words,(wmax+PACKET) * sizeof(char*));
          if(words==NULL) {
            Lib_strdel(s);
            return false;
          }
          wmax += PACKET;
        }
        words[wnumber] = token;
        wnumber++;
      }
      break;
    }
  }
  // check that at least one word is not empty :
  bool ok = false;
  for(count=0;count<wnumber;count++) { 
    if(*(words[count])!='\0') {
      ok = true;
      break;
    }
  }
  if(!ok) { // only wildcards :
    Lib_strdel(s);
    if(words) ::free(words);
    return true;
  } 

  // loop on words :
  bool match = true;
  token = (char*)aString.c_str();
  for(count=0;count<wnumber;count++) { 
    int   lword;
    lword = ::strlen(words[count]); 
    if(lword>0) { 
      char* pos;
      if(count==0) {
        if(aPattern[0]!='*') {
          // Begin of pattern (words[0]) and aString must match :
          if(::strncmp(token,words[count],lword)!=0) {
            match = false; // Different.
            break;      
          }
          token = token + lword;
          continue;
        }
      }
      pos = ::strstr (token,words[count]);
      if(!pos) {
        match = false;
        break;
      }
      if((count==(wnumber-1)) && (aPattern[lpattern-1]!='*') ) {
        // Last word :
        if(::strcmp(aString.c_str()+lstring-lword,words[count])!=0) 
          match = false; // Compare last word and end of aString.
        break;
      } else {
        token = pos + lword;
      }
    }
  }
  Lib_strdel(s);
  if(words) ::free(words);
  return match;
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> Lib::smanip::prepend(
 const std::string& aWhat
,const std::vector<std::string>& aText
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> text;
  unsigned int number = aText.size();
  for(unsigned int index=0;index<number;index++) {
    text.push_back(aWhat+aText[index]);
  }
  return text;
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> Lib::smanip::append(
 const std::vector<std::string>& aText
,const std::string& aWhat
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> text;
  unsigned int number = aText.size();
  for(unsigned int index=0;index<number;index++) {
    text.push_back(aText[index]+aWhat);
  }
  return text;
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> Lib::smanip::match(
 const std::vector<std::string>& aText
,const std::string& aPattern 
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> text;
  unsigned int number = aText.size();
  for(unsigned int index=0;index<number;index++) {
    if(match(aText[index],aPattern)) text.push_back(aText[index]);
  }
  return text;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::getenv(
 std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string::size_type dollar;
  while((dollar=aString.find('$'))!=std::string::npos){
    std::string::size_type slash = aString.find('/',dollar+1);
    std::string::size_type back_slash = aString.find('\\',dollar+1);
    std::string::size_type pos = std::string::npos; 
    if(slash!=std::string::npos) {
      if(back_slash!=std::string::npos) {
        pos = MINIMUM(slash,back_slash);
      } else {
        pos = slash;
      }
    } else {
      if(back_slash!=std::string::npos) {
        pos = back_slash;
      } else {
        pos = std::string::npos; 
      }
    }
    std::string env;
    if(pos==std::string::npos) {
      env = aString.substr(dollar+1,aString.length()-(dollar+1));
    } else {
      //     abc$xxx/ef
      //     0  3   7 9
      env = aString.substr(dollar+1,pos-(dollar+1));
    }
    char* val = ::getenv(env.c_str());
    if(!val) return false; // Stop looping if env variable not found.
    std::string value = aString.substr(0,dollar);
    value += val;
    if(pos!=std::string::npos)
      value += aString.substr(pos,aString.length()-pos);
    aString = value;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::replace(
 std::string& aString
,const std::string& aOld     
,const std::string& aNew     
)
//////////////////////////////////////////////////////////////////////////////
// return true : some replacement done.
// return false : nothing replaced.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aOld.empty()) return false;
  std::string snew;
  std::string::size_type lold = aOld.length();
  bool status = false;
  std::string stmp = aString;
  while(true) {
    std::string::size_type pos = stmp.find(aOld);
    if(pos==std::string::npos){
      snew += stmp;
      break;
    } else {
      snew += stmp.substr(0,pos);
      snew += aNew;
      stmp = stmp.substr(pos+lold,stmp.length()-(pos+lold));
      status = true;
    }
  }
  aString = snew;
  return status;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::printf(
 std::string& aString
,int aLength
,const char* aFormat
,...
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aString = "";
  if(aLength<0) return false;
  if(!aFormat) return false;
  char* s = new char[aLength+1];
  if(!s) return false;
  s[aLength] = '\0';
  va_list args;
  va_start (args,aFormat);
  vsprintf (s,aFormat,args);
  va_end   (args);
  if(s[aLength]!='\0') {
    delete [] s;
    return false;
  }
  aString = s;
  delete [] s;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::justify(
 std::string& aString
,unsigned int aSize
,side aSide
)
//////////////////////////////////////////////////////////////////////////////
// aSize is the final string length.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  strip(aString);
  if(aSize<=aString.size()) {
    aString.resize(aSize);
    return false;
  } else {
    if(aSide==left) {
      aString = aString + std::string(aSize-aString.size(),' ');
    } else if(aSide==right) {
      aString = std::string(aSize-aString.size(),' ') + aString;
    } else if(aSide==middle) {
      int l = aSize - aString.size();
      int h = l/2;
      if(h*2==l) { //even number of spaces :
        aString = std::string(h,' ') + aString + std::string(h,' ');
      } else { // odd number of spaces :
        aString = std::string(h,' ') + aString + std::string(h+1,' ');
      }
    }
    return false;
  }
}
/*
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::scanf(
 const std::string& aString
,const char* aFormat
,...
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aFormat) return false;
  va_list args;
  va_start (args,aFormat);
  int status = vsscanf  (aString.c_str(),aFormat,args);
  va_end   (args);
  return (status!=1 ? false : true);
}
*/
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::touint64(
 const std::string& aString
 ,Lib::uint64& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //FIXME : notation n[eE][+]m not supported !
#ifdef WIN32
  char* s;
  unsigned __int64 v = ::_strtoui64(aString.c_str(),&s,10);
#else
#ifdef LIB_INT64_AS_LONG
  char* s;
  unsigned long v = ::strtoul(aString.c_str(),&s,10);
#else //32bits platform :
  char* s;
  unsigned long long v = ::strtoull(aString.c_str(),&s,10);
#endif
#endif
  if( (s==aString.c_str()) || (s!=(aString.c_str()+aString.size())) ) {
    aValue = 0;
    return false;
  }
  aValue = (Lib::uint64)v;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::toint64(
 const std::string& aString
 ,Lib::int64& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //FIXME : notation n[eE][+]m not supported !
#ifdef WIN32
  char* s;
  __int64 v = ::_strtoi64(aString.c_str(),&s,10);
#else
#ifdef LIB_INT64_AS_LONG
  char* s;
  long v = ::strtol(aString.c_str(),&s,10);
#else //32bits platform :
  char* s;
  long long v = ::strtoll(aString.c_str(),&s,10);
#endif
#endif
  if( (s==aString.c_str()) || (s!=(aString.c_str()+aString.size())) ) {
    aValue = 0;
    return false;
  }
  aValue = (Lib::int64)v;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::topointer(
 const std::string& aString
,void*& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_INT64_AS_LONG
  unsigned long v = 0L;
  if(::sscanf(aString.c_str(),"0x%lx",&v)!=1) {
    if(::sscanf(aString.c_str(),"%lu",&v)!=1) {
      aValue = 0;
      return false;
    }
  }
#else //32bits platform :
  unsigned int v = 0;
  if(::sscanf(aString.c_str(),"0x%x",&v)!=1) {
    if(::sscanf(aString.c_str(),"%u",&v)!=1) {
      aValue = 0;
      return false;
    }
  }
#endif
  aValue = (void*)v;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::touint(
 const std::string& aString
,unsigned int& aValue
)
//////////////////////////////////////////////////////////////////////////////
// Notation n[eE][+]m not supported !
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(::sscanf(aString.c_str(),"0x%x",&aValue)==1) return true;
  if(::sscanf(aString.c_str(),"%u",&aValue)==1) return true;
  aValue = 0;
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::touintx(
 const std::string& aString
,unsigned int& aValue
)
//////////////////////////////////////////////////////////////////////////////
// Notation n[eE][+]m not supported !
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(::sscanf(aString.c_str(),"%x",&aValue)==1) return true;
  aValue = 0;
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::toint(
 const std::string& aString
,int& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //FIXME : notation n[eE][+]m not supported !
  char* s;
  long l = ::strtol(aString.c_str(),&s,10);
  if( (s==aString.c_str()) || (s!=(aString.c_str()+aString.size())) ) {
    aValue = 0;
    return false;
  }
  aValue = (int)l;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::tobool(
 const std::string& aString
,bool& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string slower = aString;
  tolowercase(slower);
  if((slower=="1")||(slower=="true")||(slower=="yes")||(slower=="on")) {
    aValue = true;
    return true;
  } else if((slower=="0")||(slower=="false")||(slower=="no")||(slower=="off")){
    aValue = false;
    return true;
  } else {
    aValue = false;
    return false;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::tofloat(
 const std::string& aString
,float& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  char* s;
  aValue = (float)::strtod(aString.c_str(),&s);
  if( (s==aString.c_str()) || (s!=(aString.c_str()+aString.size())) ) {
    aValue = 0;
    return false;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::todouble(
 const std::string& aString
,double& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  char* s;
  aValue = ::strtod(aString.c_str(),&s);
  if( (s==aString.c_str()) || (s!=(aString.c_str()+aString.size())) ) {
    aValue = 0;
    return false;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::toints(
 const std::string& aString
,std::vector<int>& aValues
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> words;
  smanip::words(aString," ",words);
  aValues.resize(words.size());
  int valuei = 0;
  for(std::vector<std::string>::iterator it = words.begin();
      it!=words.end();++it) {
    int value;
    if(!toint(*it,value)) {
      aValues.clear();
      return false;
    }
    aValues[valuei] = value;
    valuei++;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::todoubles(
 const std::vector<std::string>& aText
,std::vector<double>& aValues
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = aText.size();
  aValues.resize(number,0);
  for(unsigned int index=0;index<number;index++) {
    double value;
    if(!todouble(aText[index],value)) {
      aValues.clear();
      return false;
    }
    aValues[index] = value;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::todoubles(
 const std::string& aString
,std::vector<double>& aValues
,const std::string& aSeparator
)
//////////////////////////////////////////////////////////////////////////////
// Support PAW syntax : 3*2.4 is understood as three times 2.4.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aValues.clear();
  std::vector<std::string> words;
  smanip::words(aString,aSeparator,words);
  bool seekMultiplier = (aSeparator!="*"?true:false);
  if(!seekMultiplier) aValues.resize(words.size());
  int valuei = 0;
  std::vector<std::string> subwords;
  std::vector<std::string>::iterator it;
  for(it = words.begin();it!=words.end();++it) {
    if(seekMultiplier) {
      std::string::size_type pos = (*it).find('*');
      if(pos==std::string::npos){
        double value;
        if(!todouble(*it,value)) {
          aValues.clear();
          return false;
        }
        aValues.push_back(value);
      } else {
        //3*2.4
        //01234
        std::string sn = (*it).substr(0,pos);
        std::string sv = (*it).substr(pos+1,aString.size()-(pos+1));
        int multiplier;
        if(!toint(sn,multiplier)) {
          aValues.clear();
          return false;
        }
        double value;
        if(!todouble(sv,value)) {
          aValues.clear();
          return false;
        }
        for(int index=0;index<multiplier;index++) aValues.push_back(value);
      }
    } else {
      double value;
      if(!todouble(*it,value)) {
        aValues.clear();
        return false;
      }
      aValues[valuei] = value;
      valuei++;
    }    
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::smanip::tolowercase(
 std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string::size_type l = aString.length();;
  std::string::size_type i;
  std::string s;
  for(i=0;i<l;i++) {
    char c = aString[i];
    s += ((c) >= 'A' && (c) <= 'Z' ?  c - 'A' + 'a' : c);
  }
  aString = s;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::smanip::touppercase(
 std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string::size_type l = aString.length();;
  std::string::size_type i;
  std::string s;
  for(i=0;i<l;i++) {
    char c = aString[i];
    s += ((c) >= 'a' && (c) <= 'z' ?  c - 'a' + 'A' : c);
  }
  aString = s;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::togeometry(
 const std::string& aString
,int& aX
,int& aY
,int& aW
,int& aH
)
//////////////////////////////////////////////////////////////////////////////
// Decode string : <int:w>x<int:h>[+<int:x>+<int:y>] 
// (which is the Xt geometry resource format).
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aX = 0;
  aY = 0;
  aW = 0;
  aH = 0;
  int x,y,w,h;        
  if(::sscanf(aString.c_str(),"%dx%d+%d+%d",&w,&h,&x,&y)==4) {
    aX = x;
    aY = y;
    aW = w;
    aH = h;
    return true;
  } else if(::sscanf(aString.c_str(),"%dx%d",&w,&h)==2) {
    aW = w;
    aH = h;
    return true;
  } else {
    return false;
  }
}
//////////////////////////////////////////////////////////////////////////////
char* Lib::smanip::tocstring(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
// Return value to be deleted with delete [].
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  char* cstring = new char[aString.length()+1];
  if(!cstring) return 0;
  strcpy(cstring,aString.c_str());
  return cstring;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::polynomialDegree(
 const std::string& aString
,int& aDegree
)
//////////////////////////////////////////////////////////////////////////////
// aString = [p,P,Polynomial]<integer degree>
// We extract <degree> from the upper.
// The degree is the maximum "x" power.
// For example P1 is ax+b which of degree 1.
// For example P2 is ax*x+bx+c which of degree 2.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s = aString;
  tolowercase(s);
  replace(s,"polynomial","p");
  aDegree = 0;
  if(s.size()<=0) return false;
  if(s[0]!='p') return false;
  int number = 0;
  if(!toint(s.substr(1,s.size()-1),number)) return false;;
  aDegree = number;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
int Lib::smanip::axtoi(
 const std::string& aString
) 
//////////////////////////////////////////////////////////////////////////////
// convert from ASCII hex to int
// Exa : "FF" -> 256.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int x = 0;
  int n = aString.size();
  // convert n nibbles
  for (int i = 0; i < n; i++) {
    char c = aString[i];
    // numbers 0 - 9
    if ((c > 0x2F) && (c < 0x3A))
      x += ((c - 0x30) << ((n - i - 1) * 4));

    // capital letters A - F
    if ((c > 0x40) && (c < 0x47))
      x += ((c - 0x41 + 0x0A) << ((n - i - 1) * 4));

    // lower case letters a - f
    if ((c > 0x60) && (c < 0x67))
      x += ((c - 0x61 + 0x0A) << ((n - i - 1) * 4));
  }
  return x;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::pathNameSuffix(
 const std::string& aString
,std::string& aPath
,std::string& aName
,std::string& aSuffix
)
//////////////////////////////////////////////////////////////////////////////
// If aString = dir0/dir1/dir2/dir3/name.xxx
//   aPath = dir0/dir1/dir2/dir3
//   aName = name.xxx
//   aSuffix = xxx
// If aString = dir0/name.xxx
//   aPath = dir0
//   aName = name.xxx
//   aSuffix = xxx
// If aString = name.xxx
//   aPath = ""
//   aName = name.xxx
//   aSuffix = xxx
// If aString = /name.xxx
//   aPath = "/"
//   aName = name.xxx
//   aSuffix = xxx
// If aString = .
//   aPath = "."
//   aName = ""
//   aSuffix = ""
// If aString = ..
//   aPath = ".."
//   aName = ""
//   aSuffix = ""
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aString==".") {
    aPath = ".";
    aName = "";
    aSuffix = "";
    return true;
  } else if(aString=="..") {
    aPath = "..";
    aName = "";
    aSuffix = "";
    return true;
  }

  std::string::size_type pos_slash = aString.rfind('/');
  std::string::size_type pos_bslash = aString.rfind('\\');
  std::string::size_type pos = 0;
  if(pos_slash==std::string::npos) {
    if(pos_bslash==std::string::npos) {
      pos = std::string::npos;
    } else {
      pos = pos_bslash;
    }
  } else {
    if(pos_bslash==std::string::npos) {
      pos = pos_slash;
    } else {
      if(pos_slash<=pos_bslash) {
        pos = pos_bslash;
      } else {
        pos = pos_slash;
      }
    }
  }

  if(pos==std::string::npos) {
    aPath = "";
    pos = 0;
  } else if(pos==0) {
    aPath = "/";
    pos++;
  } else {
    aPath = aString.substr(0,pos);
    pos++;
  }
  std::string s = aString.substr(pos,aString.size()-pos);
  pos = s.rfind('.');
  if(pos==std::string::npos) {
    aName = s;
    aSuffix = "";
  } else {
    aName = s;
    pos++;
    aSuffix = s.substr(pos,s.size()-pos);
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::pathWords(
 const std::string& aPath
,std::vector<std::string>& aWords
,bool& aAbsolute
,bool& aWinPath
,std::string& aDrive
)
//////////////////////////////////////////////////////////////////////////////
//  aPath could be at the Windows syntax : <drive>:\dir1\dir2...
// If so, only dir1, dirn are put in aWords.
//  It is assumed that the last item is a file name, then
// returned words is a list without the last item.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string dir = aPath;

  std::string::size_type pos = dir.find('\\');
  aWinPath = (pos==std::string::npos?false:true);
  if(aWinPath) {
    Lib::smanip::replace(dir,"\\","/");
    pos = dir.find(':');
    if(pos!=std::string::npos) { //drive 
      aDrive = dir.substr(0,pos);
      dir = dir.substr(pos+1,dir.size()-(pos+1));
    } else {
      aDrive = "";
    }
  } else {
    aDrive = "";
  }

  std::string path,name,suffix;
  Lib::smanip::pathNameSuffix(dir,path,name,suffix);

  aAbsolute = (path.size()&&(path[0]=='/')?true:false);

  aWords = Lib::smanip::words(path,"/");

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::isAbsolutePath(
 const std::string& aPath
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aPath.find('\\')!=std::string::npos) { //Windows path.
    if(aPath.find(':')!=std::string::npos) return true;
    return (aPath.size()&&(aPath[0]=='\\')?true:false);
  } else { //UNIX path
    return (aPath.size()&&(aPath[0]=='/')?true:false);
  }
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::suffix(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
// If aString = dir0/dir1/dir2/dir3/name.xxx
//   return xxx
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string::size_type pos = aString.rfind('.');
  if(pos==std::string::npos) return "";
  pos++;
  return aString.substr(pos,aString.size()-pos);
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::nosuffix(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
// If aString = dir0/dir1/dir2/dir3/name.xxx
//   return name
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  // Start searching after the last / (or last \ for Windows).
  std::string::size_type pos = aString.rfind('/');
  if(pos==std::string::npos) pos = aString.rfind('\\');
  if(pos==std::string::npos) pos = 0;
  else pos++;
  std::string s = aString.substr(pos,aString.size()-pos);
  std::string::size_type dot_pos = s.rfind('.');
  if(dot_pos==std::string::npos) return s;
  return s.substr(0,dot_pos);
}
//////////////////////////////////////////////////////////////////////////////
void Lib::smanip::lines( //DEPRECATED.
 const std::string& aString
,std::vector<std::string>& aLines
)
//////////////////////////////////////////////////////////////////////////////
// aString is a list separated by "\n" or "\\n".
// For "xxx\n\nxxx", {"xxx","","xxx"} will be created.
// WARNING : if aString is a Windows file name, it may 
//  contains a \n which is not a delimiter ; like ..\data\ntuples.hbook.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aLines.clear();
  int length = aString.length();
  if(length<=0) return;
  char* cstring = Lib_strdup(aString.c_str());
  if(!cstring) return;
  int pos = 0;
  length++;
  for(int count=0;count<length;count++) {
    if( (cstring[count]=='\n') || 
        (cstring[count]=='\0') || 
        ( (cstring[count]=='\\') && (cstring[count+1]=='n') ) ) { 
      char shift_one = (cstring[count]=='\n' ? 1 : 0);
      cstring[count] = '\0';
      aLines.push_back(cstring+pos);
      if(shift_one==1) {
        pos = count+1;
      } else {
        pos = count+2;
        count++;
      }
    }
  }
  Lib_strdel(cstring);
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> Lib::smanip::words(
 const std::string& aString
,const std::string& aLimiter 
,bool aTakeEmpty //=false
)
//////////////////////////////////////////////////////////////////////////////
//  If aLimiter is for exa "|" and for "xxx||xxx" :
//  - aTakeEmpty false : {"xxx","xxx"} will be created 
//    (and NOT {"xxx","","xxx"}).
//  - aTakeEmpty true : {"xxx","","xxx"} will be created.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> words;
  if(aString.empty()) return words;
  std::string::size_type lim = (aTakeEmpty?0:1);
  if(aLimiter.empty()) {
    words.push_back(aString);
  } else {
    std::string::size_type l = aString.length();
    std::string::size_type llimiter = aLimiter.length();
    std::string::size_type pos = 0;
    while(true) {
      std::string::size_type index = aString.find(aLimiter,pos);
      if(index==std::string::npos){ // Last word.
        if((l-pos)>=lim) words.push_back(aString.substr(pos,l-pos));
        break;
      } else {
        //     abcxxxef
        //     0  3  67
        if((index-pos)>=lim) words.push_back(aString.substr(pos,index-pos));
        pos = index + llimiter;
      }
    }
  }
  return words;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::smanip::words(
 const std::string& aString
,const std::string& aLimiter 
,std::vector<std::string>&  aWords
,bool aTakeEmpty //false
)
//////////////////////////////////////////////////////////////////////////////
// Deprecated.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aWords = words(aString,aLimiter,aTakeEmpty);
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> Lib::smanip::text(const std::string& aString){
  return words(aString,"\n",true);
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::remove(
 std::vector<std::string>& aStrings
,const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  bool found_some = false;
  std::vector<std::string>::iterator it;
  for(it=aStrings.begin();it!=aStrings.end();) {
    if(*it==aString) {
      it = aStrings.erase(it);
      found_some = true;
    } else {
      ++it;
    }
  }
  return found_some;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::removeLinesWith(
 std::vector<std::string>& aStrings
,const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  bool found_some = false;
  std::vector<std::string>::iterator it;
  for(it=aStrings.begin();it!=aStrings.end();) {
    if((*it).find(aString)!=std::string::npos) {
      it = aStrings.erase(it);
      found_some = true;
    } else {
      ++it;
    }
  }
  return found_some;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::smanip::unique(
 std::vector<std::string>& aStrings
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string>::iterator it, it2;
  for(it=aStrings.begin();it!=aStrings.end();++it) {
    for(it2=it+1;it2!=aStrings.end();) {
      if(*it2==*it) {
        it2 = aStrings.erase(it2);
      } else { 
        ++it2;
      }
    }
  }
}
/*
//////////////////////////////////////////////////////////////////////////////
int Lib::smanip::lineIndex(
 const std::vector<std::string>& aStrings
,const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string>::const_iterator it;
  int index = 0;
  for(it=aStrings.begin();it!=aStrings.end();++it,index++) {
    if(*it==aString) return index;
  }
  return NotFound;
}
*/
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::lineIndex(
 const std::vector<std::string>& aStrings
,const std::string& aString
,unsigned int& aIndex
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aIndex = 0;
  std::vector<std::string>::const_iterator it;
  for(it=aStrings.begin();it!=aStrings.end();++it,aIndex++) {
    if(*it==aString) return true;
  }
  aIndex = 0;
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::belong(
 const std::vector<std::string>& aStrings
,const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string>::const_iterator it;
  for(it=aStrings.begin();it!=aStrings.end();++it) {
    if(*it==aString) return true;
  }
  return false;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 const std::vector<std::string>& aVector
,const std::string& aSeparator
,bool a_sep_at_end
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = aVector.size();
  if(number<=0) return "";
  std::string result;
  number--;
  for(unsigned int index=0;index<number;index++) {
    result += aVector[index];
    result += aSeparator;
  }
  result += aVector[number];
  if(a_sep_at_end) result += aSeparator;
  return result;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::smanip::strip(
 std::vector<std::string>& aStrings
,what aType
,char aChar
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = aStrings.size();
  for(unsigned int index=0;index<number;index++)
    strip(aStrings[index],aType,aChar);
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::replace(
 std::vector<std::string>& aStrings
,const std::string& aOld     
,const std::string& aNew     
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = aStrings.size();
  for(unsigned int index=0;index<number;index++) {
    if(!replace(aStrings[index],aOld,aNew)) return false;
  }
  return true;
}
extern "C" {
typedef int (*SortFunction)(const void*,const void*);
static int smanip_Greater(char** a1,char** a2){
  size_t l1 = ::strlen(*a1);
  size_t l2 = ::strlen(*a2);
  if(l1==l2) return 0;
  return (l1>l2)?1:-1;
}
static int smanip_Lesser(char** a1,char** a2){
  size_t l1 = ::strlen(*a1);
  size_t l2 = ::strlen(*a2);
  if(l1==l2) return 0;
  return (l1<l2)?1:-1;
}
}
//////////////////////////////////////////////////////////////////////////////
void Lib::smanip::sort(
 std::vector<std::string>& aStrings
,bool aIncreasing
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int linen = aStrings.size();
  if(linen<=1) return;
  char** lines = new char*[linen];
 {for(unsigned int i=0;i<linen;i++) 
    lines[i] = Lib_strdup(aStrings[i].c_str());}
  if(aIncreasing) 
    ::qsort((void*)lines,(size_t)linen,(size_t)sizeof(char*),
            (SortFunction)smanip_Greater);
  else
    ::qsort((void*)lines,(size_t)linen,(size_t)sizeof(char*),
            (SortFunction)smanip_Lesser);
  for(unsigned int i=0;i<linen;i++) {
    aStrings[i] = lines[i];
    Lib_strdel(lines[i]);
  }
  delete [] lines;
}
extern "C" {
static int str_cmp(char** a1,char** a2){ return ::strcmp(*a1,*a2);}
}
//////////////////////////////////////////////////////////////////////////////
void Lib::smanip::cmpsort(
 std::vector<std::string>& aStrings
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int linen = aStrings.size();
  if(linen<=1) return;
  char** lines = new char*[linen];
 {for(unsigned int i=0;i<linen;i++) 
    lines[i] = Lib_strdup(aStrings[i].c_str());}
  ::qsort((void*)lines,(size_t)linen,(size_t)sizeof(char*),
          (SortFunction)str_cmp);
  for(unsigned int i=0;i<linen;i++) {
    aStrings[i] = lines[i];
    Lib_strdel(lines[i]);
  }
  delete [] lines;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 unsigned char aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
  if(cmanip::isPrintable(aValue)) 
    printf(s,32,"%c",aValue);
  else
    printf(s,32,"%d",aValue);
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 char aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
  if(cmanip::isPrintable(aValue)) 
    printf(s,32,"%c",aValue);
  else
    printf(s,32,"%d",aValue);
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 unsigned short aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
  printf(s,32,"%d",aValue);
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 short aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
  printf(s,32,"%d",aValue);
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 int aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
  printf(s,32,"%d",aValue);
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 unsigned int aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
  printf(s,32,"%u",aValue);
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostringx(
 unsigned int aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
  printf(s,32,"%x",aValue);
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 Lib::int64 aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
  printf(s,32,LIB_INT64_FORMAT,aValue);
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 Lib::uint64 aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
  printf(s,32,LIB_UINT64_FORMAT,aValue);
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 float aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
  printf(s,32,"%g",aValue);
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 double aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
  printf(s,32,"%g",aValue);
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 bool aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return std::string(aValue?"true":"false");
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 const std::string& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return aValue;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 const char* aValue
)
//////////////////////////////////////////////////////////////////////////////
// To avoid the [bool, const std::string&, passing "text"] problem.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return std::string(aValue);
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 const char* aBuffer
,unsigned int aLength
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
  s.resize(aLength,' ');
  for(unsigned int index=0;index<aLength;index++) s[index] = aBuffer[index];
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::tostring(
 unsigned int aLinen 
,const char* aLines[]
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
  for(unsigned int index=0;index<aLinen;index++) {
    s += std::string(aLines[index]);
    s += "\n";
  }
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::p2s(
 void* aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
#ifdef LIB_INT64_AS_LONG
  printf(s,32,"%lu",(unsigned long)aValue);
#else //32bits platform :
  printf(s,32,"%u",(unsigned int)aValue);
#endif
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::p2sx(
 void* aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s;
#ifdef LIB_INT64_AS_LONG
  printf(s,32,"0x%lx",(unsigned long)aValue);
#else //32bits platform :
  printf(s,32,"0x%x",(unsigned int)aValue);
#endif
  return s;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::smanip::dump(
 const std::string& aString
,Slash::Core::IWriter& aWriter
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Out out(aWriter);
  out << "\"" << aString << "\"" << Lib::endl;
  unsigned int number = aString.size();
  for(unsigned int index=0;index<number;index++) {
    out << aString[index] << " " << (int)aString[index] << Lib::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
void Lib::smanip::write(
 const std::vector<std::string>& aStrings
,Slash::Core::IWriter& aWriter
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string>::const_iterator it;
  Out out(aWriter);
  for(it=aStrings.begin();it!=aStrings.end();++it) {
    out << *it << endl;
  }
}
namespace Lib {
class Color {
public:
  Color(double aRed = 0,double aGreen = 0,double aBlue = 0)
  :fRed(aRed),fGreen(aGreen),fBlue(aBlue){}
  Color(const Color& aFrom)
  :fRed(aFrom.fRed),fGreen(aFrom.fGreen),fBlue(aFrom.fBlue){}
  Color& operator=(const Color& aFrom) {
    fRed = aFrom.fRed;
    fGreen = aFrom.fGreen;
    fBlue = aFrom.fBlue;
    return *this;
  }
public:
  double fRed,fGreen,fBlue;
};
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::torgbs(
 const std::string& aString
,std::string& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  double r,g,b;
  bool status = torgb(aString,r,g,b);
  printf(aValue,128,"%g %g %g",r,g,b);
  return status;
}

static std::map<std::string,Lib::Color>* sColors = 0;
//////////////////////////////////////////////////////////////////////////////
static void initColors(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(sColors) return; //done.

  sColors = new std::map<std::string,Lib::Color>(); //FIXME : who delete ?

  (*sColors)["aquamarine"] = Lib::Color(0.496101,0.996109,0.828138);//1
  (*sColors)["mediumaquamarine"] = Lib::Color(0.398444,0.800793,0.664073);
  (*sColors)["black"] = Lib::Color(0,0,0);
  (*sColors)["blue"] = Lib::Color(0,0,1);
  (*sColors)["cadetblue"] = Lib::Color(0.371099,0.617197,0.62501);
  (*sColors)["cornflowerblue"] = Lib::Color(0.390631,0.58204,0.925795);
  (*sColors)["darkslateblue"] = Lib::Color(0.281254,0.238285,0.542977);
  (*sColors)["lightblue"] = Lib::Color(0.675792,0.843763,0.898451);
  (*sColors)["lightsteelblue"] = Lib::Color(0.68751,0.765637,0.867201);
  (*sColors)["mediumblue"] = Lib::Color(0,0,0.800793); //10

  (*sColors)["mediumslateblue"] = Lib::Color(0.480476,0.406256,0.929702);
  (*sColors)["midnightblue"] = Lib::Color(0.0976577,0.0976577,0.437507);
  (*sColors)["navyblue"] = Lib::Color(0,0,0.500008);
  (*sColors)["navy"] = Lib::Color(0,0,0.500008);
  (*sColors)["skyblue"] = Lib::Color(0.527352,0.8047,0.917983);
  (*sColors)["slateblue"] = Lib::Color(0.414069,0.351568,0.800793);
  (*sColors)["steelblue"] = Lib::Color(0.273442,0.50782,0.703136);
  (*sColors)["coral"] = Lib::Color(0.996109,0.496101,0.312505);
  (*sColors)["cyan"] = Lib::Color(0,1,1);
  (*sColors)["firebrick"] = Lib::Color(0.695323,0.132815,0.132815);//20

  (*sColors)["brown"] = Lib::Color(0.644541,0.164065,0.164065);
  (*sColors)["gold"] = Lib::Color(0.996109,0.839857,0);
  (*sColors)["goldenrod"] = Lib::Color(0.851575,0.644541,0.125002);
  (*sColors)["green"] = Lib::Color(0,1,0);
  (*sColors)["darkgreen"] = Lib::Color(0,0.390631,0);
  (*sColors)["darkolivegreen"] = Lib::Color(0.332036,0.417975,0.183597);
  (*sColors)["forestgreen"] = Lib::Color(0.132815,0.542977,0.132815);
  (*sColors)["limegreen"] = Lib::Color(0.195315,0.800793,0.195315);
  (*sColors)["mediumseagreen"] = Lib::Color(0.234379,0.699229,0.441413);
  (*sColors)["mediumspringgreen"] = Lib::Color(0,0.976577,0.601572);//30

  (*sColors)["palegreen"] = Lib::Color(0.593759,0.980484,0.593759);
  (*sColors)["seagreen"] = Lib::Color(0.17969,0.542977,0.339849);
  (*sColors)["springgreen"] = Lib::Color(0,0.996109,0.496101);
  (*sColors)["yellowgreen"] = Lib::Color(0.601572,0.800793,0.195315);
  (*sColors)["darkslategrey"] = Lib::Color(0.183597,0.308598,0.308598);
  (*sColors)["dimgrey"] = Lib::Color(0.410163,0.410163,0.410163);
  (*sColors)["lightgrey"] = Lib::Color(0.824231,0.824231,0.824231);
  (*sColors)["grey"] = Lib::Color(0.750011,0.750011,0.750011);
  (*sColors)["khaki"] = Lib::Color(0.937514,0.898451,0.546883);
  (*sColors)["magenta"] = Lib::Color(1,0,1);//40

  (*sColors)["maroon"] = Lib::Color(0.68751,0.187503,0.375006);
  (*sColors)["orange"] = Lib::Color(0.996109,0.644541,0);
  (*sColors)["orchid"] = Lib::Color(0.851575,0.437507,0.83595);
  (*sColors)["darkorchid"] = Lib::Color(0.597665,0.195315,0.796887);
  (*sColors)["mediumorchid"] = Lib::Color(0.726574,0.332036,0.824231);
  (*sColors)["pink"] = Lib::Color(0.996109,0.750011,0.792981);
  (*sColors)["plum"] = Lib::Color(0.863294,0.62501,0.863294);
  (*sColors)["red"] = Lib::Color(1,0,0);
  (*sColors)["indianred"] = Lib::Color(0.800793,0.35938,0.35938);
  (*sColors)["mediumvioletred"] = Lib::Color(0.777356,0.0820325,0.519539);//50
      
  (*sColors)["orangered"] = Lib::Color(0.996109,0.269535,0);
  (*sColors)["violetred"] = Lib::Color(0.812512,0.125002,0.562509);
  (*sColors)["salmon"] = Lib::Color(0.976577,0.500008,0.445319);
  (*sColors)["sienna"] = Lib::Color(0.62501,0.320317,0.175784);
  (*sColors)["tan"] = Lib::Color(0.820325,0.703136,0.546883);
  (*sColors)["thistle"] = Lib::Color(0.843763,0.746105,0.843763);
  (*sColors)["turquoise"] = Lib::Color(0.250004,0.875013,0.812512);
  (*sColors)["darkturquoise"] = Lib::Color(0,0.8047,0.816419);
  (*sColors)["mediumturquoise"] = Lib::Color(0.281254,0.816419,0.796887);
  (*sColors)["violet"] = Lib::Color(0.929702,0.50782,0.929702);//60

  (*sColors)["blueviolet"] = Lib::Color(0.539071F,0.167971F,0.882826F);
  (*sColors)["wheat"] = Lib::Color(0.957046F,0.867201F,0.699229F);
  (*sColors)["white"] = Lib::Color(1,1,1);
  (*sColors)["yellow"] = Lib::Color(1,1,0);
  (*sColors)["greenyellow"] = Lib::Color(0.675792F,0.996109F,0.18359);
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::torgb(
 const std::string& aString
,double& aRed
,double& aGreen
,double& aBlue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aRed = 0.5;
  aGreen = 0.5;
  aBlue = 0.5;
  if(aString.empty()) return false;
  std::vector<double> ds;
  if(smanip::todoubles(aString,ds) && (ds.size()>=3)) {
    if( ds[0]<0 || ds[0]>1 ||
        ds[1]<0 || ds[1]>1 ||
        ds[2]<0 || ds[2]>1 ) return false;
    aRed = ds[0];
    aGreen = ds[1];
    aBlue = ds[2];
    return true;
  } else if( (aString.size()==7) && (aString[0]=='#') ) {
    // #RRGGBB format :
    //  1 3 5
    unsigned int rr,gg,bb;
    if(!touint(std::string("0x")+aString.substr(1,2),rr)) return false;
    if(!touint(std::string("0x")+aString.substr(3,2),gg)) return false;
    if(!touint(std::string("0x")+aString.substr(5,2),bb)) return false;
    aRed = double(rr)/255.0;
    aGreen = double(gg)/255.0;
    aBlue = double(bb)/255.0;
    return true;
  } else {
    initColors();    
    std::map<std::string,Color>::const_iterator it = sColors->find(aString);
    if(it!=sColors->end()) {
      aRed = (*it).second.fRed;
      aGreen = (*it).second.fGreen;
      aBlue = (*it).second.fBlue;
      return true;
      
    }
  }
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::torgba(
 const std::string& aString
,double& aRed
,double& aGreen
,double& aBlue
,double& aAlpha
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aRed = 0.5;
  aGreen = 0.5;
  aBlue = 0.5;
  aAlpha = 0;
  if(aString.empty()) return false;
  std::vector<double> ds;
  if(smanip::todoubles(aString,ds) && (ds.size()>=4)) {
    if( ds[0]<0 || ds[0]>1 ||
        ds[1]<0 || ds[1]>1 ||
        ds[2]<0 || ds[2]>1 ||
        ds[3]<0 || ds[3]>1 ) return false;
    aRed = ds[0];
    aGreen = ds[1];
    aBlue = ds[2];
    aAlpha = ds[3];
    return true;
  } else if( (aString.size()==7) && (aString[0]=='#') ) {
    // #RRGGBB format :
    //  1 3 5
    unsigned int rr,gg,bb;
    if(!touint(std::string("0x")+aString.substr(1,2),rr)) return false;
    if(!touint(std::string("0x")+aString.substr(3,2),gg)) return false;
    if(!touint(std::string("0x")+aString.substr(5,2),bb)) return false;
    aRed = rr/255;
    aGreen = gg/255;
    aBlue = bb/255;
    return true;
  } else {
    initColors();    
    std::map<std::string,Color>::const_iterator it = sColors->find(aString);
    if(it!=sColors->end()) {
      aRed = (*it).second.fRed;
      aGreen = (*it).second.fGreen;
      aBlue = (*it).second.fBlue;
      return true;      
    }
  }
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::torgb(
 const std::string& aString
,float& aRed
,float& aGreen
,float& aBlue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  double r,g,b;
  bool status = torgb(aString,r,g,b);
  aRed = float(r);
  aGreen = float(g);
  aBlue = float(b);
  return status;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::torgba(
 const std::string& aString
,float& aRed
,float& aGreen
,float& aBlue
,float& aAlpha
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  double r,g,b,a;
  bool status = torgba(aString,r,g,b,a);
  aRed = float(r);
  aGreen = float(g);
  aBlue = float(b);
  aAlpha = float(a);
  return status;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::smanip::justifylines(
 const std::vector<std::string>& aText
,std::vector<std::string>& aLines
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = aText.size();
  int begin = NotFound;
  unsigned int index;
  for(index=0;index<number;index++) {
    const std::string& line = aText[index];
    std::string::size_type not_space = line.find_first_not_of(' ');
    if(not_space==std::string::npos) {
      // Empty or blank line :
      // skip.
    } else {
      // Some text is in the line :
      if(begin==NotFound) {
        begin = not_space;
      } else {
        begin = MINIMUM(begin,(int)not_space);
      }
    }
  }
  if(begin==NotFound) return;
  aLines.clear();
  for(index=0;index<number;index++) {
    const std::string& line = aText[index];
    std::string::size_type not_space = line.find_first_not_of(' ');
    if(not_space==std::string::npos) {
      // Empty or blank line :
      aLines.push_back("");
    } else {
      aLines.push_back(line.substr(begin,line.length()-begin));
    }
  }
}

/*
bool Lib::smanip::options(
 const std::string& aString
,std::vector< std::pair<std::string,std::string> >& aOptions
,bool a_lower //true by default
){
  // aString : a1 = v1; a2 = v2, .....
  // Return stripped pairs (ai,vi)
  std::vector<std::string> lines;
  smanip::words(aString,";",lines);
  aOptions.resize(lines.size());
  unsigned int index = 0;
  for(std::vector<std::string>::iterator it = lines.begin();
      it!=lines.end();++it) {
    std::vector<std::string> words;
    smanip::words(*it,"=",words);
    if(words.size()!=2) {
      aOptions.clear();
      return false;
    }
    std::string option = words[0];
    std::string value = words[1];
    smanip::strip(option);
    if(a_lower) smanip::tolowercase(option);
    smanip::strip(value);
    aOptions[index] = std::pair<std::string,std::string>(option,value); 
    index++;
  }
  return true;
}

bool Lib::smanip::option(const std::string& aString,const std::string& aOption,std::string& aValue)
{
  // aString : a1 = v1; a2 = v2, .....
  // Search for aOption in aString and return the value.
  // The search is not case sensitive.
  aValue = "";
  std::string aoption = aOption;
  smanip::tolowercase(aoption);
  std::vector<std::string> lines;
  smanip::words(aString,";",lines);
  for(std::vector<std::string>::iterator it = lines.begin();
      it!=lines.end();++it) {
    std::vector<std::string> words;
    smanip::words(*it,"=",words);
    if(words.size()!=2) return false;
    std::string option = words[0];
    std::string value = words[1];
    smanip::strip(option);
    smanip::tolowercase(option);
    smanip::strip(value);
    if(aoption==option) {
      aValue = value;
      return true;
    }
  }
  return false;
}
*/
/*
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::dict(
 const std::string& aString
,const std::string& aSeparator
,std::vector< std::pair<std::string,std::string> >& aDict
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> lines;
  smanip::words(aString,aSeparator,lines);
  aDict.resize(lines.size());
  unsigned int index = 0;
  for(std::vector<std::string>::iterator it = lines.begin();
      it!=lines.end();++it) {
    std::vector<std::string> words;
    smanip::words(*it,"=",words);
    if(words.size()!=2) {
      aDict.clear();
      return false;
    }
    std::string key = words[0];
    std::string value = words[1];
    smanip::strip(key);
    smanip::strip(value);
    std::string s;
    if(!bracketed(value,'[',']',s)) {
      aDict.clear();
      return false;
    }
    smanip::strip(s);
    aDict[index] = std::pair<std::string,std::string>(key,s); 
    index++;
  }
  return true;
} 
*/
typedef std::vector< std::pair<std::string,std::string> > Dict;
void Lib::smanip::dump(const Dict& aDict,Slash::Core::IWriter& aWriter){
  Out out(aWriter);
  out << "number of entries : " << int(aDict.size()) << Lib::endl;
  for(Dict::const_iterator it = aDict.begin();it!=aDict.end();++it) {
    out << (*it).first << " " << (*it).second << Lib::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::version(
 const std::string& aString
,std::vector<unsigned int>& aVersion
)
//////////////////////////////////////////////////////////////////////////////
// aString is of the form [v,V][<number><char>]<number>
// Examples : 0 1.2 v0 v1r2 V1r2p3 v0.1.3
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int number = aString.size();
  if(!number) {
    aVersion.clear();
    return false;
  }
  char* pos = (char*)aString.c_str();
  if( ((*pos)=='v') || ((*pos)=='V') ) {
    if(number==1) {
      aVersion.clear();
      return false; //aString is only 'v' or 'V' !
    }
    pos++;
  }
  while(*pos!='\0') {
    char* s;
    long v = ::strtol(pos,&s,10);
    if(s==pos) {
      aVersion.clear();
      return false;
    }
    if(v<0) {
      aVersion.clear();
      return false;
    }
    aVersion.push_back((unsigned int)v);
    pos = s; //Could be at end.
    if(*pos=='\0') break;
    // Next char is then not a digit. Skip it.
    pos++;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::version_to_dot(
 const std::string& aString
,std::string& aVersion
)
//////////////////////////////////////////////////////////////////////////////
// aString is of the form [v,V][<number><char>]<number>
// Have it at the format : number.number
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<unsigned int> verss;
  if(!Lib::smanip::version(aString,verss)) {aVersion = "";return false;}
  if(!verss.size()) {aVersion = "";return false;}
  aVersion = tostring(verss[0]);
  for(unsigned int index=1;index<verss.size();index++) {
    aVersion += ".";   
    aVersion += tostring(verss[index]);
  }
  return true;  
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::isVariable(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
// Check if aString is of the form of a programmation variable.
// - not empty.
// - not starting by a number.
// - being composed of alphanumeric, number and '_' characters.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aString.size()) return false;
  char* pos = (char*)aString.c_str();
  if(cmanip::isDigit(*pos)) return false;
  pos++;
  while(*pos) {
    if( !cmanip::isAlphanumeric(*pos) && 
        !cmanip::isDigit(*pos) && 
        ((*pos)!='_') ) return false;
    pos++;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::binRoot(
 const std::string& aBinary
,std::string& aRoot
)
//////////////////////////////////////////////////////////////////////////////
// If aBinary :
// - dir0/.../dirn/bin/name   (words are : dir0,...,dirn,bin
//   aRoot = dir0/.../dirn
// - dir0/.../dirn/Contents/MacOS/name
//   aRoot = dir0/.../dirn/Contents
// - dir0/.../dirn/../<bin>/name
//   aRoot = dir0/.../dirn/Contents
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  bool absolute;
  bool winPath;
  std::string drive;
  std::vector<std::string> words;
  if(!Lib::smanip::pathWords(aBinary,
                             words,absolute,winPath,drive)) return false;

  unsigned int wordn = words.size();
  if((wordn>=1)&&(words[wordn-1]=="bin")) {
    // rm bin
    aRoot = absolute?"/":"";
    for(unsigned int index=0;index<(wordn-1);index++){
      if(index) aRoot += "/";
      aRoot += words[index];
    }
    //aRoot = path;
    if(winPath) {
      if(drive.size()) aRoot = drive + ":" + aRoot;
      replace(aRoot,"/","\\");
    }
    return true;
  } else if((wordn>=2)&&
            (words[wordn-1]=="MacOS")&&
            (words[wordn-2]=="Contents")) {
    // It seems to be a MacOS binary in a .app bundle.
    aRoot = absolute?"/":"";
    for(unsigned int index=0;index<(wordn-1);index++){
      if(index) aRoot += "/";
      aRoot += words[index];
    }
    //if(winPath) {
    //if(drive.size()) aRoot = drive + ":" + aRoot; //!
    //replace(aRoot,"/","\\");
    //}
    return true;
  } else if((wordn>=2)&&(words[wordn-2]=="..")) {
    aRoot = absolute?"/":"";
    for(unsigned int index=0;index<(wordn-1);index++){
      if(index) aRoot += "/";
      aRoot += words[index];
    }
    if(winPath) {
      if(drive.size()) aRoot = drive + ":" + aRoot;
      replace(aRoot,"/","\\");
    }
    return true;
  } else if(wordn>=3) {
    // Is it a binary in a CMT package ? That is to say under :
    //   <package>/<version>/<tag>
    //tag = words[wordn-1];
    //vers = words[wordn-2];
    //pack = words[wordn-3];
    std::vector<unsigned int> vers;
    if(version(words[wordn-2],vers)) {
      aRoot = absolute?"/":"";
      for(unsigned int index=0;index<(wordn-1);index++){
        if(index) aRoot += "/";
        aRoot += words[index];
      }
      if(winPath) {
        if(drive.size()) aRoot = drive + ":" + aRoot;
        replace(aRoot,"/","\\");
      }
      return true;
    } else {
      // Other cases ?
    }
  }
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::nameArgs(
 const std::string& aString
,std::string& aName
,std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
// Look if aName is of the form <string>(<string>,...,<string>)
// return true if so.
// return true if <string> without args
// return false if any bad syntax.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string::size_type lb = aString.find('(');
  std::string::size_type rb = aString.rfind(')');
  if( (lb==std::string::npos) && (rb==std::string::npos) ) {
    aName = aString;
    aArgs.clear();
    return true;
  } else if( (lb==std::string::npos) && (rb!=std::string::npos) ) {
    aName = "";
    aArgs.clear();
    return false;
  } else if( (lb!=std::string::npos) && (rb==std::string::npos) ) {
    aName = "";
    aArgs.clear();
    return false;
  } else if(lb>=rb) {
    aName = "";
    aArgs.clear();
    return false;
  }
  std::string s = aString.substr(lb+1,rb-lb-1);
  words(s,",",aArgs,true);
  aName = aString.substr(0,lb);
  return true; 
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::parseArray(
 const std::string& aString
,std::string& aName
,int& aIndex
)
//////////////////////////////////////////////////////////////////////////////
// Look if aName is of the form <string>(<integer>)
// return true if <string>(<integer>)
// return true if <string>. aName = <string> and aIndex = NotFound (-1).
// return false if bad syntax or <integer> not an integer.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string::size_type lb = aString.find('(');
  std::string::size_type rb = aString.rfind(')');
  if( (lb==std::string::npos) && (rb==std::string::npos) ) {
    aName = aString;
    aIndex = NotFound;
    return true;
  } else if( (lb==std::string::npos) && (rb!=std::string::npos) ) {
    aName = "";
    aIndex = 0;
    return false;
  } else if( (lb!=std::string::npos) && (rb==std::string::npos) ) {
    aName = "";
    aIndex = 0;
    return false;
  } else if(lb>=rb) {
    aName = "";
    aIndex = 0;
    return false;
  }
  //012345678
  //Slash::Core::ILibraryManager(123)
  std::string s = aString.substr(lb+1,rb-lb-1);
  if(!toint(s,aIndex)) {
    aName = "";
    aIndex = 0;
    return false;
  }
  aName = aString.substr(0,lb);
  return true; 
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::int2coord(
 int aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
       if(aValue==0) return "x";
  else if(aValue==1) return "y";
  else if(aValue==2) return "z";
  else if(aValue==3) return "t";
  else return tostring(aValue);
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::coord2int(
 const std::string& aString
,int& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if((aString=="x")||(aString=="X")) {
    aValue = 0;
    return true;
  } else if((aString=="y")||(aString=="Y")) {
    aValue = 1;
    return true;
  } else if((aString=="z")||(aString=="Z")) {
    aValue = 2;
    return true;
  } else if((aString=="t")||(aString=="T")) {
    aValue = 3;
    return true;
  } else {
    return toint(aString,aValue);
  }
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::toxml(
 const std::string& aString
)
///////////////////////////////////////////////////////////////////////////////
// > : &lt;
// < : &gt;
// & : &amp;
// " : &quot;
// ' : &apos;
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string s = aString;
  replace(s,"<","&lt;");
  replace(s,">","&gt;");
  replace(s,"&","&amp;");
  replace(s,"\"","&quot;");
  replace(s,"'","&apos;");
  return s;
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> Lib::smanip::tovector(
 int aArgc
,char** aArgv
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aArgc<=0) return std::vector<std::string>();
  std::vector<std::string> vec(aArgc);
  for(int index=0;index<aArgc;index++) vec[index] = aArgv[index];
  return vec;
}
//////////////////////////////////////////////////////////////////////////////
void Lib::smanip::newArgcv(
 const std::vector<std::string>& aArgs
,int& aArgc
,char**& aArgv
)
//////////////////////////////////////////////////////////////////////////////
// If using with :
//   int argc;
//   char** argv;
//   Lib::smanip::argcv(argc,argv);  
// delete with :
//  for(int index=0;index<argc;index++) delete [] argv[index];
//  delete [] argv;
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aArgc = (int)aArgs.size();
  if(aArgc<=0) {
    aArgc = 0;
    aArgv = 0;
    return;
  }
  typedef char* cstring;
  aArgv = new cstring[aArgc];
  for(int index=0;index<aArgc;index++) {
    const std::string& arg = aArgs[index];
    aArgv[index] = new char[arg.size()+1];
    ::strcpy(aArgv[index],arg.c_str());
  }
}
//////////////////////////////////////////////////////////////////////////////
void Lib::smanip::delArgcv(
 int& aArgc
,char**& aArgv
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  for(int index=0;index<aArgc;index++) delete [] aArgv[index];
  delete [] aArgv;
  aArgc = 0;
  aArgv = 0;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::beginWith(
 const std::string& aString
,const std::string& aBegin
,std::string& aRest
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  unsigned int sz = aBegin.size();  
  if(aString.substr(0,sz)==aBegin) {
    aRest = aString.substr(sz,aString.size()-sz);
    return true;
  } else {
    aRest = "";
    return false;
  }
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::removeVersions(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
// Used in the xxx-config programs.
// replace "@@top@@/pack/vers" -> "@@top@@/pack".
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef WIN32
  std::string sep = "\\";
#else
  std::string sep = "/";
#endif
  std::string sold = aString;
  Lib::smanip::replace(sold,"\"",sep+"@quote@"); //To treat AIDA/v3r2p1"/src
  std::vector<std::string> words = Lib::smanip::words(sold,sep);
  unsigned int wordn = words.size();
  std::string snew;
  for(unsigned int index=0;index<wordn;index++) {
    std::vector<unsigned int> vers;
    std::string word = words[index];
    if(!Lib::smanip::version(word,vers)) {
      if(snew.size()) snew += sep;
      snew += words[index];
    }
  }
  Lib::smanip::replace(snew,sep+"@quote@","\"");
  return snew;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::dirname(
 const std::string& aPath
,unsigned int aNumber
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path = aPath;
  for(unsigned int index=0;index<aNumber;index++) {
    std::string p,n,s;
    Lib::smanip::pathNameSuffix(path,p,n,s);
    path = p;
  }
  return path;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::basename(
 const std::string& aPath
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string path,name,suffix;
  Lib::smanip::pathNameSuffix(aPath,path,name,suffix);
  return name;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::equal(
 const std::string& aFrom
,const char* aTo
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  // Can save a lot of time compared to if(aFrom==aTo) if
  // aTo is of the form "xxx". (Because the below avoids
  // the copy of aTo in an temporary std::string.
  return (::strcmp(aFrom.c_str(),aTo)==0?true:false);
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::smanip::bracketed(
 const std::string& aString
,char a_l
,char a_r
,std::string& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aString.size()<2) return false;
  if(aString[0]!=a_l) return false;
  if(aString[aString.size()-1]!=a_r) return false;
  aValue = aString.substr(1,aString.size()-2); 
  return true;
}
//////////////////////////////////////////////////////////////////////////////
std::string Lib::smanip::quote(
 const std::string& aString
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return std::string("\"")+aString+"\"";
}
