#include <string>
#include <vector>

#include <Slash/Core/IWriter.h>
#include <Lib/Out.h>

namespace Lib {

class Particle;

class Decay : public std::vector<Particle> {
public:
  Decay();
  virtual ~Decay();
  void dump(Slash::Core::IWriter&);
};

class Particle {
public:
  Particle(const std::string& aName,const double aMass) 
  :fName(aName),fMass(aMass){}
  Particle(const Particle& aParticle)
  :fName(aParticle.fName)
  ,fMass(aParticle.fMass)
  ,fDecays(aParticle.fDecays){}
  virtual ~Particle() {}
  Particle& operator=(const Particle& aParticle){
    fName = aParticle.fName; 
    fMass = aParticle.fMass;
    fDecays = aParticle.fDecays;
    return *this;
  }
  void addDecay(const Decay& aDecay) {
    fDecays.push_back(aDecay);
  }
  std::string name() const { return fName;}
  void dump(Slash::Core::IWriter& aPrinter) {
    Lib::Out out(aPrinter);
    out << fName << " -> " << Lib::endl;
    unsigned int number = fDecays.size();
    for(unsigned int index=0;index<number;index++) {
      Decay& decay = fDecays[index];
      out << "     ";
      decay.dump(aPrinter);
    }
  }
private:
  std::string fName;
  double fMass;
  std::vector<Decay> fDecays;
};

}

Lib::Decay::Decay(){}
Lib::Decay::~Decay(){}
void Lib::Decay::dump(Slash::Core::IWriter& aPrinter) {
  Lib::Out out(aPrinter);
  unsigned int number = size();
  for(unsigned int index=0;index<number;index++) {
    const Lib::Particle& particle = (*this)[index];
    if(index) out << " + ";
    out << particle.name();
  }
  out << Lib::endl;
}


