/*+---------------------- Copyright notice ----------------------------+*/
/*| Copyright (C) 1995, Open Scientist project, LAL Orsay.             |*/
/*|  Permission to use, copy, modify, and distribute this software     |*/
/*| and its documentation for any purpose and without fee is hereby    |*/
/*| granted, provided that the above copyright notice appear in all    |*/
/*| copies and that both that copyright notice and this permission     |*/
/*| notice appear in supporting documentation.  This software is       |*/
/*| provided "as is" without express or implied warranty.              |*/
/*+---------------------- Copyright notice ----------------------------+*/

#include <stdlib.h>
#include <math.h>

#include <Midnight/Fitter.h>
#include <Midnight/Printer.h>

static void function(int&,double*,double&,double*,int,void*);
static double polynomial(double,int,double*);

static double ShootRandomFlat();

static int aParamn = 2;

#define fNpt  100
static double fXxs[fNpt];
static double fVals[fNpt];
static double fErrs[fNpt];

//////////////////////////////////////////////////////////////////////////////
int main(
)
//////////////////////////////////////////////////////////////////////////////
//  Book and fill some histos.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{

  // Fit with A + B * x :
  static const char* aNames[2] = {"A","B"};
  double aParams[2] = {0.5,2.};
  double aSteps[2] = {0.1,0.1};

  // Data to fit :
  int i;
  for(i=0;i<fNpt;i++) {
    fXxs[i] = i + 0.1 * ShootRandomFlat();
    fVals[i] = 2. * ( i  + 0.1 * ShootRandomFlat());
    fErrs[i] = 0.1;
  }

  Midnight::Fitter minuit(0);
  minuit.SetFCN(function,0);
  Midnight::Printer printer; // Example of output control.
  minuit.SetPrinter(&printer);

  int status;

  for(i=0;i<aParamn;i++) {
    minuit.mnparm(i,aNames[i],aParams[i],aSteps[i],0.,0.,status);
  }

  double args[10];
  args[0] = 0;
  minuit.mnexcm("SET PRI",args,1,status);
  minuit.mnexcm("MIGRAD",args,0,status);
  //minuit.mnexcm("MINOS",args,0,status);
  //
  double bmin,bmax;
  std::string s;
  for(i=0;i<aParamn;i++) {
    minuit.mnpout(i,s,aParams[i],aSteps[i],bmin,bmax,status);
  }

  return 0;
}
//////////////////////////////////////////////////////////////////////////////
void function(
 int& //aParamn
,double* //aGrad
,double& aValue
,double* aParams
,int //aStatus
,void*
)
//////////////////////////////////////////////////////////////////////////////
//  aParamn, aParams, aStatus : INPUT
//  aGrad, aValue : OUTPUT
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  double value;
  double chinew;
  double chisqr = 0.;
  for(int i=0;i<fNpt;i++) {
    value = polynomial(fXxs[i],aParamn,aParams); 
    chinew = (fVals[i] - value)/fErrs[i];
    chisqr = chisqr + chinew * chinew;
  }
  aValue = chisqr;
}
//////////////////////////////////////////////////////////////////////////////
double polynomial(
 double aX
,int aParamn
,double* aParams
)
//////////////////////////////////////////////////////////////////////////////
// P[0] + P[1] x + P[2] x**2 + ...
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  double value = 0.;
  double xxx = 1.;
  for(int i=0;i<aParamn;i++) {
    value += aParams[i] * xxx;
    xxx = xxx * aX;
  }
  return value;
}
//////////////////////////////////////////////////////////////////////////////
double ShootRandomFlat(
)
//////////////////////////////////////////////////////////////////////////////
// Shoot random numbers according a flat distribution.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  double value  = (double)rand();
  value /= (double)RAND_MAX;
  return value;
}
