//  The below code is a modified version of the ROOT/TMinuit class
// in order to have a stand alone C++ version of the Minuit package.
//    G.Barrand

#ifndef Midnight_Fitter_h
#define Midnight_Fitter_h

namespace Slash {namespace Core {class IWriter;}}

#include <string>

namespace Midnight {

typedef void(*Function_t)(int&,double*,double&,double*,int,void*);

class Fitter {
private:
        int        fEmpty;            //Initialization flag (1 = initialized)
        int        fMaxpar;           //Maximum number of parameters
        std::string      *fCpnam;           //Array of parameters names
        double     *fU;               //External (visible to user in FCN) value of parameters
        double     *fAlim;            //Lower limits for parameters. If zero no limits
        double     *fBlim;            //Upper limits for parameters
        double     *fErp;             //Positive Minos errors if calculated
        double     *fErn;             //Negative Minos errors if calculated
        double     *fWerr;            //External parameters error (standard deviation, defined by UP)
        double     *fGlobcc;          //Global Correlation Coefficients
           int     *fNvarl;           //parameters flag (-1=undefined, 0=constant..)
           int     *fNiofex;          //Internal parameters number, or zero if not currently variable
           int     *fNexofi;          //External parameters number for currently variable parameters
        double     *fX;               //Internal parameters values
        double     *fXt;              //Internal parameters values X saved as Xt
        double     *fDirin;           //(Internal) step sizes for current step
        double     *fXs;              //Internal parameters values saved for fixed params
        double     *fXts;             //Internal parameters values X saved as Xt for fixed params
        double     *fDirins;          //(Internal) step sizes for current step for fixed params
        double     *fGrd;             //First derivatives
        double     *fG2;              //
        double     *fGstep;           //Step sizes
        double     *fGin;             //
        double     *fDgrd;            //Uncertainties
        double     *fGrds;            //
        double     *fG2s;             //
        double     *fGsteps;          //
        int        *fIpfix;           //List of fixed parameters
        int        fNpfix;            //Number of fixed parameters
        double     *fVhmat;           //(Internal) error matrix stored as Half MATrix, since it is symmetric
        double     *fVthmat;          //VHMAT is sometimes saved in VTHMAT, especially in MNMNOT
        double     *fP;               //
        double     *fPstar;           //
        double     *fPstst;           //
        double     *fPbar;            //
        double     *fPrho;            //Minimum point of parabola
        int        fMaxint;           //Maximum number of internal parameters
        int        fNpar;             //Number of parameters
        int        fMaxext;           //Maximum number of external parameters
        int        fNu;               //
        int        fIsysrd;           //standardInput unit
        int        fIsyswr;           //standard output unit
        int        fIsyssa;           //
        int        fNpagwd;           //Page width
        int        fNpagln;           //Number of lines per page
        int        fNewpag;           //
        int        fIstkrd[10];       //
        int        fNstkrd;           //
        int        fIstkwr[10];       //
        int        fNstkwr;           //
        std::string      fCfrom;            //
        std::string      fCstatu;           //
        std::string      fCtitl;            //
        std::string      fCword;            //
        std::string      fCundef;           //
        std::string      fCvrsn;            //
        std::string      fCovmes[4];        //
        int        fISW[7];           //Array of switches
        int        fIdbg[11];         //Array of internal debug switches
        int        fNblock;           //Number of Minuit data blocks
        int        fIcomnd;           //Number of commands
        double     fAmin;             //Minimum value found for FCN
        double     fUp;               //FCN+-UP defines errors (for chisquare fits UP=1)
        double     fEDM;              //Estimated vertical distance to the minimum
        double     fFval3;            //
        double     fEpsi;             //
        double     fApsi;             //
        double     fDcovar;           //Relative change in covariance matrix
        int        fNfcn;             //Number of calls to FCN
        int        fNfcnmx;           //Maximum number of calls to FCN
        int        fNfcnlc;           //
        int        fNfcnfr;           //
        int        fItaur;            //
        int        fIstrat;           //
        int        fNwrmes[2];        //
        double     *fWord7;           //
        bool       fLwarn;            //true if warning messges are to be put out (default=true)
        bool       fLrepor;           //true if exceptional conditions are put out (default=false)
        bool       fLimset;           //true if a parameter is up against limits (for MINOS)
        bool       fLnolim;           //true if there are no limits on any parameters (not yet used)
        bool       fLnewmn;           //true if the previous process has unexpectedly improved FCN
        bool       fLphead;           //true if a heading should be put out for the next parameter definition
        double     fEpsmac;           //machine precision for floating points:
        double     fEpsma2;           //sqrt(fEpsmac)
        double     fVlimlo;           //
        double     fVlimhi;           //
        double     fUndefi;           //Undefined number = -54321
        double     fBigedm;           //Big EDM = 123456
        double     fUpdflt;           //
        double     *fXpt;             //X array of points for contours
        double     *fYpt;             //Y array of points for contours
        std::string      *fChpt;            //Character to be plotted at the X,Y contour positions
        double     fXmidcr;           //
        double     fYmidcr;           //
        double     fXdircr;           //
        double     fYdircr;           //
        int        fKe1cr;            //
        int        fKe2cr;            //
        std::string      *fOrigin;          //
        std::string      *fWarmes;          //
        int        fNfcwar[20];       //
        int        fIcirc[2];         //
        Function_t fFCN;
        void*      fTag;
	Slash::Core::IWriter* fPrinter;


// methods performed on Fitter class
private:
	   Fitter(const Fitter &m);
 void      BuildArrays(int maxpar=15);
 void      DeleteArrays();
public:
           Fitter();
           Fitter(int maxpar);
 virtual   ~Fitter();
 void      SetFCN(Function_t,void*);
 void      SetPrinter(Slash::Core::IWriter*);
 //
 void      mnamin();
 void      mnbins(double a1, double a2, int naa, double &bl, double &bh, int &nb, double &bwid);
 void      mncalf(double *pvec, double &ycalf);
 void      mncler();
 void      mncntr(int ke1, int ke2, int &ierrf);
 void      mncomd(std::string crdbin, int &icondn);
 void      mncont(int ke1, int ke2, int nptu, double *xptu, double *yptu, int &ierrf);
 void      mncrck(std::string &crdbuf, int maxcwd, std::string &comand, int &lnc
                ,  int mxp, double *plist, int &llist, int &ierr, int isyswr);
 void      mncros(double &aopt, int &iercr);
 void      mncuve();
 void      mnderi();
 void      mndxdi(double pint, int ipar, double &dxdi);
 void      mneig(double *a, int ndima, int n, int mits, double *work, double precis, int &ifault);
 void      mnemat(double *emat, int ndim);
 void      mnerrs(int number, double &eplus, double &eminus, double &eparab, double &gcc);
 void      mneval(double anext, double &fnext, int &ierev);
 void      mnexcm(std::string comand, double *plist, int llist, int &ierflg) ;
 void      mnexin(double *pint);
 void      mnfixp(int iint, int &ierr);
 void      mnfree(int k);
 void      mngrad();
 void      mnhelp(std::string comd);
 void      mnhess();
 void      mnhes1();
 void      mnimpr();
 void      mninex(double *pint);
 void      mninit(int i1, int i2, int i3);
 void      mnlims();
 void      mnline(double *start, double fstart, double *step, double slope, double toler);
 void      mnmatu(int kode);
 void      mnmigr();
 void      mnmnos();
 void      mnmnot(int ilax, int ilax2, double &val2pl, double &val2mi);
 void      mnparm(int k, std::string cnamj, double uk, double wk, double a, double b, int &ierflg);
 void      mnpars(std::string &crdbuf, int &icondn);
 void      mnpfit(double *parx2p, double *pary2p, int npar2p, double *coef2p, double &sdev2p);
 void      mnpint(double &pexti, int i, double &pinti);
 void      mnplot(double *xpt, double *ypt, std::string *chpt, int nxypt, int npagwd, int npagln);
 void      mnpout(int iuext, std::string& chnam, double &val, double &err, double &xlolim, double &xuplim, int &iuint);
 void      mnprin(int inkode, double fval);
 void      mnpsdf();
 void      mnrazz(double ynew, double *pnew, double *y, int &jh, int &jl);
 void      mnrn15(double &val, int &inseed);
 void      mnrset(int iopt);
 void      mnsave();
 void      mnscan();
 void      mnseek();
 void      mnset();
 void      mnsimp();
 void      mnstat(double &fmin, double &fedm, double &errdef, int &npari, int &nparx, int &istat);
 void      mntiny(double epsp1, double &epsbak);
 bool    mnunpt(std::string &cfname);
 void      mnvert(double *a, int l, int m, int n, int &ifail);
 void      mnwarn(const char *copt, const char *corg, const char *cmes);
 void      mnwerr();
};

}

#endif








