#include <Slash/Core/ISession.h>
#include <Slash/UI/IUI.h>
#include <Slash/Tools/UI.h>
#include <Slash/Tools/args.h>
#include <Slash/Tools/app.h>
#include <Slash/Tools/osc_env.h>

#include <Lib/Printer.h>
#include <Lib/InternetSocket.h>

#include <OnX/Net/NetManager.h>
#include <OnX/Net/Protocol.h>

#include <iostream>
#include <cstdlib>

int main(int argc,char** argv){

  if(argc>0) {
    std::string arg0(argv[0]);
    if(!Slash::if_mouse_startup(arg0,"OnX",std::cout)) {
      std::cout << "OnX_onx :"
                << " unable to set mouse startup." << std::endl;
      return EXIT_FAILURE;
    }
  }

  if(!Slash::osc::set_env(std::cout)) {
    std::cout << "OnX_server :"
              << " unable to set OnX environment." << std::endl;
    return EXIT_FAILURE;
  }

  Slash::args args(argc,argv);

  int verboseLevel = 0;
 {std::string sverbose;
  if(args.find("-verbose",sverbose)) {
    if(sverbose.empty()) verboseLevel = 1;
  }}

  int port = 50701;
  Lib::InternetSocket* server = 
    new Lib::InternetSocket(std::cout,"",port,verboseLevel);
  if(!server->accept()) {
    delete server;
    return EXIT_FAILURE;
  }

  std::string driver;
  if(!server->wait_string(driver)) {
    delete server;
    return EXIT_FAILURE;
  }

  std::string onx_file;
  if(!server->wait_string(onx_file)) {
    delete server;
    return EXIT_FAILURE;
  }

  //if(!server->sendString(ONX_PROTOCOL_MSG+":hello from server")) {
  //  delete server;
  //  return EXIT_FAILURE;
  //}

  Slash::Core::ISession* session = Slash_create_session(args.tovector());
  if(!session) {
    std::cout << "OnX_server :"
              << " session not found." << std::endl;
    delete server;
    return EXIT_FAILURE;
  }

  OnX::NetManager* netManager = 0; 
  if(!session->findManager("NetManager")) {
    netManager = new OnX::NetManager(*session,*server,false);
    session->addManager(netManager);
  }

  Slash::UI::IUI_Manager* uim = Slash::uiManager(*session);
  if(!uim) {
    std::cout << "OnX_server :"
              << " UI_Manager not found." << std::endl;
    delete session;
    delete server;
    return EXIT_FAILURE;
  }

  Slash::UI::IUI* ui = 
    uim->create(driver,onx_file,std::vector<std::string>(),true);
  if(!ui) {
    delete session;
    delete server;
    return EXIT_FAILURE;
  } 

  ui->steer();
  netManager->sendExit(0);

  delete session;
  delete server; //Must come after session.

  // Some libraries (for exa Coin, gtk) may have declared 
  // C functions with ::atexit. To ease debugging we call
  // explicitly ::exit here.
  ::exit(EXIT_SUCCESS);

  return EXIT_SUCCESS;
}
