//
// Example of declaration of some "user" data (here class Shape) 
// to the AccessorManager.
//


#include "Shape.h" // "user" class to declare to the AccessorManager.

#include <Slash/Tools/IvNet.h>

#include <Lib/Cast.h>

//////////////////////////////////////////////////////////////////////////////
// Code to declare the user class to OnX.  ///////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// The OnX_ShapeManager manages data (here instances of class OnX::Shape)
// or permits to access the entity that manages the data.
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// Inheritance :
#include <Slash/Core/IManager.h>

#include <vector>
#include <Slash/Core/ISession.h>
#include <Slash/Tools/Managers.h>
#include <Slash/Tools/UI.h>
#include <Slash/Tools/system.h>

#include <Lib/Randoms.h>

namespace OnX {
class ShapeManager : public virtual Slash::Core::IManager {
public: //Slash::Core::IManager
  virtual std::string name() const;
  virtual void* cast(const std::string&) const;
public:
  ShapeManager(Slash::Core::ISession&);
  virtual ~ShapeManager();
  std::vector<Shape*>& shapes();
private:
  std::string fName;
  std::vector<Shape*> fShapes;
};
}

OnX::ShapeManager::ShapeManager(Slash::Core::ISession& aSession)
:fName("OnX_ShapeManager"){
  // Create some data.
  fShapes.push_back(new Shape("Cube",1, 10, 10,0));
  fShapes.push_back(new Shape("Cube",2,-10, 10,0));
  fShapes.push_back(new Shape("Cube",3,-10,-10,0));
  fShapes.push_back(new Shape("Cube",4, 10,-10,0));
  //Have an intersection with the previous one :
  fShapes.push_back(new Shape("Cube",4, 13,-10,0));

  fShapes.push_back(new Shape("Sphere",1,0, 10, 10));
  fShapes.push_back(new Shape("Sphere",2,0,-10, 10));
  fShapes.push_back(new Shape("Sphere",3,0,-10,-10));
  fShapes.push_back(new Shape("Sphere",4,0, 10,-10));
  //Have an intersection with the previous one :
  fShapes.push_back(new Shape("Sphere",4,0, 17.9,-10));
  fShapes.push_back(new Shape("Cylinder",1, 10,0, 10));
  fShapes.push_back(new Shape("Cylinder",2,-10,0, 10));
  fShapes.push_back(new Shape("Cylinder",3,-10,0,-10));
  fShapes.push_back(new Shape("Cylinder",5, 10,0,-10));
  fShapes.push_back(new Shape("Cone",1,0,0,0));
  fShapes.push_back(new Shape("Hole",1,10,10,10));
  fShapes.push_back(new Shape("Hole",1,-10,-10,-10));
  // Have one centered at origin to test shape cuts :
  fShapes.push_back(new Shape("Hole",2,0,0,0));

  // To test line drawing :
  Lib::RandomFlat r;
  double s = 100; //scale.
  fShapes.push_back(new Shape("Track1",0,s*r.shoot(),s*r.shoot(),s*r.shoot()));
  fShapes.push_back(new Shape("Track2",0,s*r.shoot(),s*r.shoot(),s*r.shoot()));
  fShapes.push_back(new Shape("Track3",0,s*r.shoot(),s*r.shoot(),s*r.shoot()));
  fShapes.push_back(new Shape("Track4",0,s*r.shoot(),s*r.shoot(),s*r.shoot()));
  fShapes.push_back(new Shape("Track5",0,s*r.shoot(),s*r.shoot(),s*r.shoot()));

  Slash::UI::IStyleManager* styleManager = Slash::styleManager(aSession);
  if(styleManager) {
    std::string file;
    if(!Slash::check_getenv(aSession.cout(),
                            "ONX_SHAPE_STYLES","ONXSHAPESTYLES",file)) {
      std::string OSC_HOME_DIR;
      if(!Slash::getenv("OSC_HOME_DIR",OSC_HOME_DIR)) {
        aSession.cout() << "OnX::ShapeManager::ShapeManager :"
                        << " environment variable OSC_HOME_DIR not defined."
                        << std::endl;
      } else {
        file = OSC_HOME_DIR + "/Resources/OnX/examples/OnX/shapes.style";
      }
    }
    if(file.size()) styleManager->loadFile(file);
  }
}

OnX::ShapeManager::~ShapeManager(){
  for(unsigned int index=0;index<fShapes.size();index++) delete fShapes[index];
  fShapes.clear();
}

std::string OnX::ShapeManager::name() const {return fName;}

void* OnX::ShapeManager::cast(const std::string& aClass) const{
  if_Lib_SCast(OnX::ShapeManager)
  else if_Lib_SCast(Slash::Core::IManager)
  else return 0;
}

std::vector<OnX::Shape*>& OnX::ShapeManager::shapes(){return fShapes;}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// The Shape Accessor permits to manipulate the data 
// (with data_collect) and permits to order visualization (with data_visualize)
// of them. See OnX/source/Callbacks/Inventor.cxx/OnX_Examples_Shapes callback.
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// Inheritance :
#include <Slash/Data/IIterator.h>
#include <OnX/Inventor/InventorAccessor.h>

#include <Lib/Property.h>
#include <Lib/Value.h>
#include <Lib/smanip.h>
#include <Lib/cstring.h>
namespace OnX {
class ShapeAccessor : public OnX::InventorAccessor {
public:
  ShapeAccessor(Slash::Core::ISession&,std::vector<Shape*>&);
  virtual ~ShapeAccessor();
public: //Slash::Data::IAccessor
  virtual std::string name() const;
  virtual Slash::Data::IIterator* iterator();
  virtual Slash::Core::IValue* findValue(Slash::Data::IAccessor::Data,const std::string&,void*);
public: //OnX::IAccessor
  virtual void visualize(Slash::Data::IAccessor::Data,void*);
private:
  std::string fName;
  std::vector<Shape*>& fShapes;
};
}

OnX::ShapeAccessor::ShapeAccessor(Slash::Core::ISession& aSession,std::vector<Shape*>& aShapes)
:InventorAccessor(aSession)
,fName("OnX_Shape")
,fShapes(aShapes){
  addProperty("type",Lib::Property::STRING);
  addProperty("size",Lib::Property::DOUBLE);
  addProperty("x",Lib::Property::DOUBLE);
  addProperty("y",Lib::Property::DOUBLE);
  addProperty("z",Lib::Property::DOUBLE);
  addProperty("hits",Lib::Property::VECTOR_DOUBLE);
  addProperty("address",Lib::Property::POINTER);
}
OnX::ShapeAccessor::~ShapeAccessor() {}
std::string OnX::ShapeAccessor::name() const {return fName;}

namespace OnX {
class ShapeAccessorIterator : public virtual Slash::Data::IIterator {
public: //Slash::Data::IIterator
  virtual Slash::Data::IAccessor::Data object() {
    if(fIndex>=fNumber) return 0;
    return fList[fIndex];
  }
  virtual void next() { fIndex++;}
  virtual void* tag() { return 0;}
public:
  ShapeAccessorIterator(std::vector<OnX::Shape*>& aList)
  :fIndex(0),fList(aList),fNumber(aList.size()) {
    Lib::Debug::increment("OnX::ShapeAccessorIterator");
  }
  virtual ~ShapeAccessorIterator() {
    Lib::Debug::decrement("OnX::ShapeAccessorIterator");
  }
private:
  unsigned int fIndex;
  std::vector<OnX::Shape*>& fList;
  unsigned int fNumber;
};
}

Slash::Data::IIterator* OnX::ShapeAccessor::iterator() {
  // Returned object should be deleted by the receiver.
  // Emulated a HEP "Event" with tracks :
  Lib::RandomFlat r;
  double s = 100; //scale.
  for(unsigned int index=0;index<fShapes.size();index++) {
    if(Lib::smanip::match(fShapes[index]->type(),"Track*")) {
      fShapes[index]->setSize(s*r.shoot(),s*r.shoot(),s*r.shoot());
    }
  }
  return new ShapeAccessorIterator(fShapes);
}

Slash::Core::IValue* OnX::ShapeAccessor::findValue(
 Slash::Data::IAccessor::Data aData
,const std::string& aName
,void*
){
  Shape* obj = (Shape*)aData;
  if(Lib::eq(aName,"address")) {
    return new Lib::Value((void*)obj);
  } else if(Lib::eq(aName,"type")) {
    return new Lib::Value(obj->type());
  } else if(Lib::eq(aName,"size")) {
    return new Lib::Value(obj->size());
  } else if(Lib::eq(aName,"x")) {
    return new Lib::Value(obj->x());
  } else if(Lib::eq(aName,"y")) {
    return new Lib::Value(obj->y());
  } else if(Lib::eq(aName,"z")) {
    return new Lib::Value(obj->z());
  } else if(Lib::eq(aName,"hits")) {
    unsigned int order = 10;
    std::vector<double> vec(order);
    for(unsigned int index=0;index<order;index++) vec[index] = double(index);
    return new Lib::Value(vec);
  } else {
    return new Lib::Value();
  }
}

// Inventor :
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoLineSet.h>

// HEPVis :
#include <HEPVis/SbPolyhedron.h>
#include <HEPVis/nodes/SoSceneGraph.h>
#include <HEPVis/nodes/SoPolyhedron.h>
#include <HEPVis/nodes/SoTextTTF.h>
#include <HEPVis/nodes/SoHighlightMaterial.h>

#include <OnX/Helpers/Net.h>

void OnX::ShapeAccessor::visualize(Slash::Data::IAccessor::Data aData,void*) {
  if(!fSoRegion) return;
  Slash::UI::IUI* ui = Slash::find_UI(fSession);
  if(!ui) return;

  Shape* obj = (Shape*)aData;

  std::string style = "OnX_Shape_"+obj->type();
  fillSoGC(style); //Get a style corresponding to shape type.

  // Build scene graph :
  SoSceneGraph* separator = new SoSceneGraph;

  separator->addChild(fSoGC.getHighlightMaterial());
  separator->addChild(fSoGC.getDrawStyle());
  separator->addChild(fSoGC.getLightModel());

  // Build name (for picking) :
  std::string s;
  Lib::smanip::printf(s,128,"OnX_Shape/0x%lx",(unsigned long)obj);
  separator->setString(s.c_str());

  SoTransform* transform = new SoTransform;
  transform->translation.setValue(SbVec3f((float)obj->x(),
                                          (float)obj->y(),
                                          (float)obj->z()));
  const SbString& modeling = fSoGC.getModeling();
  if(obj->type()=="Cube") {
    separator->addChild(transform);
    if(modeling==SbModeling_texts) {
      SoTextTTF* text = new SoTextTTF;
      //text->rescale.setValue(TRUE);
      text->size.setValue(5);
      text->string.set1Value(0,"Cube");
      separator->addChild(text);
    } else {
/*
      SoCube* cube = new SoCube;
      cube->width = (float)obj->size();
      cube->height = (float)obj->size();
      separator->addChild(cube);
      cube->depth = (float)obj->size();
*/
      //To have shape cut.
      float sz2 = (float)obj->size()/2.F;
      SbPolyhedronBox box(sz2,sz2,sz2);
      //FIXME : with SoSceneGraph we can try cache of polyhedron.
      //SoPolyhedron* soPolyhedron = fSoGC.getPolyhedron(box);
      SoPolyhedron* soPolyhedron = new SoPolyhedron(box);
      separator->addChild(soPolyhedron);
    }
  } else if(obj->type()=="Sphere") {
    separator->addChild(transform);
    if(modeling==SbModeling_texts) {
      SoTextTTF* text = new SoTextTTF;
      text->string.set1Value(0,"Sphere");
      text->size.setValue(20);
      // Ask to the node to rescale the text according the viewport size.
      // 200 is the viewport size for the wanted upper size.
      text->rescale.setValue(200);
      separator->addChild(text);
    } else {
/*
      SoSphere* sphere = new SoSphere;
      sphere->radius = (float)obj->size();
      separator->addChild(sphere);
*/
      //To have shape cut.
      float radius = (float)obj->size();
      SbPolyhedronSphere sphere(0,radius,0,2*M_PI,0,M_PI);
      //SoPolyhedron* soPolyhedron = fSoGC.getPolyhedron(sphere);
      SoPolyhedron* soPolyhedron = new SoPolyhedron(sphere);
      separator->addChild(soPolyhedron);
    }
  } else if(obj->type()=="Cone") {
    separator->addChild(transform);
    //WARNING : shape cut does not apply on SoCone.
    SoCone* cone = new SoCone;
    cone->bottomRadius = (float)obj->size();
    cone->height = (float)obj->size();
    separator->addChild(cone);
  } else if(obj->type()=="Cylinder") {
    separator->addChild(transform);
    //WARNING : shape cut does not apply on SoCone.
    SoCylinder* cylinder = new SoCylinder;
    cylinder->radius = (float)obj->size();
    cylinder->height = (float)obj->size();
    separator->addChild(cylinder);
  } else if(obj->type()=="Hole") {
    separator->addChild(transform);

    /* A flat box with a hole :
    SbPolyhedronBox box(1,1,0.1);
    SbPolyhedronTubs tubs(0,1,1,0,2*M_PI);
    SbPolyhedron op = box.subtract(tubs);
    */
    float scale = (float)obj->size();

    // A Tube with a transverse hole :
    SbPolyhedronTubs tubs_1(scale*0.7,scale*1.5,scale*2,0,2*M_PI);
    SbPolyhedronTubs tubs_2(scale*0,scale*0.5,scale*4,0,2*M_PI);
    tubs_2.Transform
      (HEPVis::SbRotation(SbVec3d(0,1,0),M_PI/2.),SbVec3d(0,0,0));
    SbPolyhedron op = tubs_1.subtract(tubs_2);

    //SoPolyhedron* soPolyhedron = fSoGC.getPolyhedron(op);
    SoPolyhedron* soPolyhedron = new SoPolyhedron(op);
    if(modeling==SbModeling_wire_frame) {
      soPolyhedron->solid.setValue(FALSE);
      soPolyhedron->reducedWireFrame.setValue(TRUE);
    } else {
      soPolyhedron->solid.setValue(TRUE);
    }

    separator->addChild(soPolyhedron);

  } else if(obj->type().substr(0,5)=="Track") {
    transform->unref();

    int32_t pointn = 2;
    SbVec3f points[2];
    points[0].setValue(0,0,0);
    points[1].setValue((float)obj->x(),(float)obj->y(),(float)obj->z());
 
    SoCoordinate3* coordinate3 = new SoCoordinate3;
    coordinate3->point.setValues(0,pointn,points);
    separator->addChild(coordinate3);

    SoLineSet* lineSet = new SoLineSet;
    lineSet->numVertices.setValues(0,1,&pointn);
    separator->addChild(lineSet);
  }
  
  if(!Slash::add_scene(*ui,*fSoRegion,separator,
                       fWhere,"dynamicScene")) separator->unref();
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// OnX callback to declare the ShapeManager to the Slash::Core::ISession 
// and declare ShapeAccessor to the accessor manager (access through 
// the Slash::Core::ISession).
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#include <Slash/Data/IProcessor.h>
#include <OnX/Helpers/OnX.h>

//extern "C" {
//  void OnX_shape_type_initialize(Slash::UI::IUI&); // To check old "Types".
//}

extern "C" {

void OnXExas_shape_initialize_sess(Slash::Core::ISession& aSession) {
  // Should be executed before any use of data_collect("OnX_Shape").

  OnX::ShapeManager* shapeManager = 
    Slash_findManager(aSession,"OnX_ShapeManager",OnX::ShapeManager);
  if(!shapeManager) {
    shapeManager = new OnX::ShapeManager(aSession);
    aSession.addManager(shapeManager);
  }

  Slash::Data::IProcessor* accessorManager = Slash::accessorManager(aSession);
  if(accessorManager && !accessorManager->findAccessor("OnX_Shape")) {
    accessorManager->addAccessor
      (new OnX::ShapeAccessor(aSession,shapeManager->shapes()));
  }
}

void OnXExas_shape_initialize(Slash::UI::IUI& aUI) {
  OnXExas_shape_initialize_sess(aUI.session());
}

void OnXExas_shape_visualize(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  // args[0] where
  if(!ui_check_args(aUI,aArgs,1)) return;
  const std::string& where = aArgs[0];

  Slash::Core::ISession& session = aUI.session();

  session_setColor(session,"red");
  session.setParameter("modeling.modeling","solid");
  session.setParameter("modeling.transparency","0.4");
  data_collect(session,"OnX_Shape","type==\"Cube\"");
  data_visualize(session,where);

  session_setColor(session,"magenta");
  session.setParameter("modeling.modeling","wire_frame");
  session.setParameter("modeling.transparency","0");
  data_collect(session,"OnX_Shape","(type==\"Hole\")&&(size<=1)");
  data_visualize(session,where);

  session_setColor(session,"yellow");
  session.setParameter("modeling.modeling","none"); //From style file.
  session.setParameter("modeling.transparency","0");
  data_collect(session,"OnX_Shape","(type==\"Hole\")&&(size>1)");
  data_visualize(session,where);

  session_setColor(session,"blue");
  session.setParameter("modeling.modeling","texts");
  data_collect(session,"OnX_Shape","(type==\"Sphere\")&&(size<=2)");
  data_visualize(session,where);

  session_setColor(session,"none");
  session.setParameter("modeling.modeling","none"); //From Shape.styles
  data_collect(session,"OnX_Shape","(type==\"Sphere\")&&(size>2)");
 {Slash::Data::IProcessor* accessorManager = 
    Slash::accessorManager(aUI.session());
  if(accessorManager) {
    std::vector<std::string> ss;
    ss.push_back("where="+where);
    ss.push_back("style=OnX_Shape_Sphere_big");
    accessorManager->execute("visualize",ss);
  }}

  // For Track[1,2], take OnX/scripts/Styles/Shape.styles styles.
  session_modeling_none(session);
  data_collect(session,"OnX_Shape","type==\"Track*\"");
  data_visualize(session,where);

}

}
