
// Have a manager to delete histograms when quiting.
#include <Slash/Core/IManager.h>

#include <HEPVis/SbTestPlotter.h>

namespace OnXExas {
class TestHistogramManager : public virtual Slash::Core::IManager {
public: //Slash::Core::IManager
  virtual std::string name() const { return fName;}
  virtual void* cast(const std::string& aClass) const {
    if(aClass=="OnXExas::TestHistogramManager") {
      return (void*)static_cast<const OnXExas::TestHistogramManager*>(this);
    } else if(aClass=="Slash::Core::IManager") {
      return (void*)static_cast<const Slash::Core::IManager*>(this);
    } else {return 0;}
  }
public:
  TestHistogramManager() 
  :fName("OnXExas_TestHistogramManager"),fHistogram(0){}
  virtual ~TestHistogramManager(){ delete fHistogram;}
private:
  std::string fName;
public:
  SbTestHistogram1D* fHistogram;
};
}

#include <Slash/Core/ISession.h>
#include <Slash/Tools/UI.h>
#include <Slash/Tools/args.h>
#include <Slash/Tools/Inventor.h>

#include <HEPVis/nodekits/SoPage.h>
#include <HEPVis/nodekits/SoPlotterRegion.h>
#include <HEPVis/nodekits/SoPlotter.h>
#include <HEPVis/nodekits/SoAxis.h>

#ifdef OnXExas_include
//FIXME : arrange so that the below be local.

#include <Slash/Tools/smanip.h>
#include <HEPVis/nodes/SoViewportRegion.h>

extern "C" {

void OnXExas_plotter_rectangular_collect(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  // args[0] widget name (or @current@)

  std::ostream& out = aUI.session().cout();
  out << "OnXExas_plotter_rectangular_collect :" 
      << " callback value : " << aUI.callbackValue()
      << std::endl;

  std::vector<std::string> words;
  Slash::words(aUI.callbackValue()," ",false,words);
  if(words.size()!=4) return;
  int x;
  if(!Slash::to<int>(words[0],x)) return;
  int y;
  if(!Slash::to<int>(words[1],y)) return;
  unsigned int w;
  if(!Slash::to<unsigned int>(words[2],w)) return;
  unsigned int h;
  if(!Slash::to<unsigned int>(words[3],h)) return;

  int xmn = x;
  int ymn = y;
  int xmx = x+w;
  int ymx = y+h;

  if(!Slash::check_args(aArgs,1,aUI.session().cout())) return;

  SoPage* soPage = Slash::find_SoPage(aUI,aArgs[0]);
  if(!soPage) return;

  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;

  SoViewportRegion* vpr = soRegion->getViewportRegion();
  int vx,vy;
  vpr->getPositionPixels(vx,vy);
  unsigned int vw,vh;
  vpr->getSizePixels(vw,vh);

  int vxmn = vx;
  int vymn = vy;
  int vxmx = vx+vw;
  int vymx = vy+vh;

  if(xmn<vxmn) xmn = vxmn;
  if(ymn<vymn) ymn = vymn;
  if(xmx>vxmx) xmx = vxmx;
  if(ymx>vymx) ymx = vymx;

  // size in pixels in the region viewport coordinates :
  w = xmx-xmn;
  if(w<=0) return;
  h = ymx-ymn;
  if(h<=0) return;
  x = xmn-vx;
  y = ymn-vy;

  //::printf("OnXExas_ploter_rectangular_collect : x %d y %d w %d h%d\n",
  //  x,y,w,h);

  if(!soRegion->isOfType(SoPlotterRegion::getClassTypeId())) return;
  SoPlotterRegion* soPlotterRegion = (SoPlotterRegion*)soRegion;

  SoPlotter* soPlotter = soPlotterRegion->getPlotter();

#define PT(a_x,a_y) SbVec3f(float(a_x)/float(vw),float(a_y)/float(vh),0)

  SbVec3f p[4];
  soPlotter->vp2axis(PT(x,y),p[0]);
  soPlotter->vp2axis(PT(x+w,y),p[1]);
  soPlotter->vp2axis(PT(x+w,y+h),p[2]);
  soPlotter->vp2axis(PT(x,y+h),p[3]);

  out << "OnXExas_plotter_rectangular_collect :" 
      << " xmn " << p[0][0]
      << " ymn " << p[0][1]
      << " xmx " << p[2][0]
      << " ymx " << p[2][1]
      << " (in axis coords)."
      << std::endl;

  //if(soPlotter->shape.getValue()==SoPlotter::XYZ) {}

}

} // extern "C"

#endif //OnXExas_include

static void plotter_example(Slash::Core::ISession& a_session,std::ostream& a_out,const std::string& a_viewer,bool a_ui) {

  Slash::UI::IUI* ui = Slash::find_UI(a_session);     
  if(!ui) {
    a_out << "Slash::UI::IUI not found." << std::endl;
    return;
  }

  SoPage* soPage = Slash::find_SoPage(*ui,a_viewer);
  if(!soPage) return;

  SoRegion* soRegion = soPage->currentRegion();
  if(!soRegion) return;
  if(!soRegion->isOfType(SoPlotterRegion::getClassTypeId())) return;
  SoPlotterRegion* soPlotterRegion = (SoPlotterRegion*)soRegion;

  Slash::Core::ISession& session = ui->session();
  OnXExas::TestHistogramManager* histogramManager = 
    Slash_findManager(session,"OnXExas_TestHistogramManager",
                    OnXExas::TestHistogramManager);
  if(!histogramManager) {
    histogramManager = new OnXExas::TestHistogramManager();
    session.addManager(histogramManager);
  }

  SbTestHistogram1D* histo1D = new SbTestHistogram1D;
  histogramManager->fHistogram = histo1D;

  SoPlotter* plotter = soPlotterRegion->getPlotter();
  plotter->setName("plotter_1");
  plotter->title.setValue("Title");
  plotter->setPlottableObject(new SbTestPlottableBins1D(*histo1D));
  plotter->set("binsStyle[0]","modeling points color 0 0 1 markerStyle PLUS markerSize 9");
  plotter->getXAxis()->title.setValue("x axis title");
  plotter->getYAxis()->title.setValue("y axis title");
  plotter->getZAxis()->title.setValue("z axis title");

  if(a_ui) ui->steer();

}

#include <iostream>
#include <Slash/Tools/osc_env.h> //For TTFPATH

int main(int,char*[]) {

  if(!Slash::osc::set_env(std::cout)) {
    std::cout << "create_ui :"
              << " unable to set OSC environment." << std::endl;
    return 1;
  }

  int argc = 1;
  static char* argv[] = {(char*)"-onx_file=plotter.onx"};  

  Slash::Core::ISession* session = Slash_create_session(Slash::to(argc,argv));
  if(!session) {
    std::cout << "plotter_example :"
              << " session not found." << std::endl;
    return 1;
  }

  plotter_example(*session,std::cout,"@current@",true);
  delete session;

  return 0;
}
