#ifndef OnX_InventorAccessor_h
#define OnX_InventorAccessor_h

// Should be an embedded class.

// Inheritance :
#include <Lib/BaseAccessor.h>
#include <Slash/Data/IVisualizer.h>
#include <Slash/Core/ISession.h>


#include <Slash/UI/IUI.h>
#include <Slash/UI/IWidget.h>
#include <Slash/UI/ISoViewer.h>
#include <Slash/Tools/Managers.h>
#include <Slash/Tools/UI.h>

#include <Lib/smanip.h>
#include <Lib/Cast.h>
#include <Lib/Arguments.h>

#include <OnX/Helpers/Inventor.h>

#include <HEPVis/misc/SoGC.h>

namespace OnX {

class InventorAccessor 
:public Lib::BaseAccessor
,public virtual Slash::Data::IVisualizer {
public: //Slash::Data::IAccessor.
  virtual void* cast(const std::string& aClass) const {
    if_Lib_SCast(OnX::InventorAccessor)
    else if_Lib_SCast(Slash::Data::IVisualizer)
    else return Lib::BaseAccessor::cast(aClass);
  }
public: //Slash::Data::IVisualizer.
  virtual void beginVisualize() {
    fSoGC.clear(); //NOTE : see endVisualize().
    fSoRegion = 0;
    fWhere.clear();
    fStyle.clear();
    Slash::UI::IUI* ui = Slash::find_UI(fSession);
    if(!ui) {
      fSoGC.setDefaultStyleCache();
    } else {
     {const std::vector<std::string>& eargs = executeArguments();
      if(eargs.size()) {
        Lib::Arguments args(eargs);
        args.find("where",fWhere);
      }}
      fSoRegion = find_SoRegion(*ui,fWhere);
      if(!fSoRegion) {
        fSoGC.setDefaultStyleCache();
      } else {
        fSoGC.setStyleCache(fSoRegion->styleCache());
        fillSoGC(name()); //By default get a style with same name as the type.
      }
    }
   {const std::vector<std::string>& eargs = executeArguments();
    if(eargs.size()) {
      Lib::Arguments args(eargs);
      args.find("style",fStyle);
    }}
  }
  //virtual void visualize(Slash::Data::IAccessor::Data,void*){
  //  Still pure virtual.
  //}
  virtual void endVisualize() {
    //FIXME : if removing the fSoGC.clear in beginVisualise
    // we should have to revisit all derived class having a endVisualize.
    // to call InventorAccessor::endVisualize()
    fSoGC.clear(); 
  }
public:
  InventorAccessor(Slash::Core::ISession& aSession,
                   const std::string& aStyleFile = "")
  :Lib::BaseAccessor(aSession.printer())
  ,fSession(aSession)
  ,fStyleManager(0)
  ,fSoRegion(0){
    fStyleManager = Slash::styleManager(fSession);
    if(fStyleManager && (aStyleFile.size()) ) {
      fStyleManager->loadFile(aStyleFile);
    }
  }
  virtual ~InventorAccessor() {}
  bool loadStyleFile(const std::string& aStyleFile) {
    if(!fStyleManager) return false;
    return fStyleManager->loadFile(aStyleFile);
  }
  bool isStyle(const std::string& aStyle) const {
    if(!fStyleManager) return false;
    return fStyleManager->isStyle(aStyle);
  }
protected:
  void fillSoGC(const std::string& aStyle = "") {
    fSoGC.reset();

    // Style priority : we look first for an "exec style" coming
    // as an argument to the data_visualize. If no found then we
    // look for aStyle. And anyway, if not "none", 
    // session modeling.[color,...] overrides file style.

   {std::string value;
    if(fStyleManager && 
         fStyleManager->findStyle(fStyle.size()?fStyle:aStyle,value)) { 
      fSoGC.setFromString(SbString(value.c_str()));
    }}

    // If not "none", session modeling.[color,...] overrides file style.

    std::string sstyle;
    std::string value;
    fSession.parameterValue("modeling.color",value);
    if((value.size())&&(value!="none")) sstyle += "\ncolor "+value;

    fSession.parameterValue("modeling.highlightColor",value);
    if((value.size())&&(value!="none")) sstyle += "\nhighlightColor "+value;

    fSession.parameterValue("modeling.transparency",value);
    if((value.size())&&(value!="none")) sstyle += "\ntransparency "+value;
  
    fSession.parameterValue("modeling.lineStyle",value); // Deprecated :
    if((value.size())&&(value!="none")) sstyle += "\nlineStyle "+value;

    fSession.parameterValue("modeling.linePattern",value);
    if((value.size())&&(value!="none")) sstyle += "\nlinePattern "+value;

    fSession.parameterValue("modeling.lineWidth",value);
    if((value.size())&&(value!="none")) sstyle += "\nlineWidth "+value;
  
    //Polyhedron :
    fSession.parameterValue("modeling.projection",value);
    if((value.size())&&(value!="none")) sstyle += "\nprojection "+value;

    fSession.parameterValue("modeling.cut",value);
    if((value.size())&&(value!="none")) sstyle += "\ncut "+value;

    fSession.parameterValue("modeling.rotationSteps",value);
    if((value.size())&&(value!="none")) sstyle += "\nrotationSteps "+value;

    fSession.parameterValue("modeling.modeling",value);
    if((value.size())&&(value!="none")) sstyle += "\nmodeling "+value;
  
    //Backward compatibility.
    fSession.parameterValue("modeling.type",value);
    if((value.size())&&(value!="none")) {
      if(value=="text") {
        sstyle += "\nmodeling texts";
      } else {
        sstyle += "\nmodeling "+value;
      }
    }
  
    fSoGC.setFromString(SbString(sstyle.c_str()));

    fSession.parameterValue("modeling.localSetup",value);
    if((value.size())&&(value!="none")) {
      bool b;
      if(Lib::smanip::tobool(value,b)) {
        fSoGC.setLocalSetup(b?TRUE:FALSE);
      }
    }


  }
  SoGC& soGC() { return fSoGC;}
  SoRegion* soRegion() const { return fSoRegion;}
  std::string where() const { return fWhere;}
protected:
  Slash::Core::ISession& fSession;
  Slash::UI::IStyleManager* fStyleManager;
  SoRegion* fSoRegion;
  SoGC fSoGC;
  std::string fWhere;
  std::string fStyle;
};

}

#endif
