//
//  All functions here should be OnX callbacks, that is to say
// functions with signature :
//   extern "C" {
//     void callback_without_arguments(Slash::UI::IUI&);
//     void callback_with_arguments(Slash::UI::IUI&,
//                                  const std::vector<std::string>&);
//   }
//

//
//  This file does not need HEPVis and Inventor to compile.
//

// Helpers :
#include <OnX/Helpers/OnX.h>

//Slash:
#include <Slash/Core/IWriter.h>
#include <Slash/UI/ICallbackData.h>
#include <Slash/UI/ICyclic.h>
#include <Slash/Tools/Managers.h>
#include <Slash/Tools/system.h>

// Lib : 
#include <Lib/Printer.h>
#include <Lib/System.h>
#include <Lib/smanip.h>
#include <Lib/dirmanip.h>
#include <Lib/fmanip.h>
#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/sjust.h>
#include <Lib/Cloner.h>
#include <Lib/Arguments.h>
#include <Lib/nmanip.h>

// OnX :
#include <OnX/Version.h>
#include <OnX/Core/Messages.h>

#include "AccessorAccessor.h"
#include "SceneGraphAccessor.h"

#include <stdlib.h> //exit

#define MAXIMUM(a,b) ((a)>(b)?a:b)

inline std::string format_suffix(const std::string& a_format) {
  if(a_format.substr(0,3)=="gl2") {
    return a_format.substr(3,a_format.size()-3);    
  } else {
    return a_format;
  }
}


extern "C" {

void OnX_declare_accessors(Slash::UI::IUI& aUI); 

//////////////////////////////////////////////////////////////////////////////
//DEPRECATED :
// use : 
//    OnX OnX_declare_accessors
//    OnX session_set_output @{session.output}@

void OnX_Initialize(Slash::UI::IUI& aUI) {
  OnX_declare_accessors(aUI);
  aUI.executeScript("DLD","OnX session_set_output @{session.output}@");
}

//////////////////////////////////////////////////////////////////////////////
/// sys //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void OnX_sys_os(Slash::UI::IUI& aUI) {
  aUI.echo(Lib::System::os());
}

void OnX_sys_copy(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  if(!ui_check_args(aUI,aArgs,2)) return;
  std::string from = Lib::fmanip::name(aArgs[0]);
  std::string to = Lib::fmanip::name(aArgs[1]);
  Lib::fmanip::copy(from,to);
}

void OnX_sys_remove(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  if(!ui_check_args(aUI,aArgs,1)) return;
  Lib::fmanip::remove(Lib::fmanip::name(aArgs[0]));
}

void OnX_sys_rm(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  if(!ui_check_args(aUI,aArgs,1)) return;
  Lib::fmanip::rm(aArgs[0]);
}

void OnX_sys_pwd(Slash::UI::IUI& aUI) {
  Lib::Out out(aUI.session().printer());
  std::string s;
  if(!Lib::dirmanip::directory(s)) {
    out << "sys_pwd :" 
        << "Can't get current directory."
        << Lib::endl;
    return;
  }
  out << s << Lib::endl;
}

void OnX_sys_rmdir(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  if(!ui_check_args(aUI,aArgs,1)) return;
  if(!Lib::dirmanip::remove(Lib::fmanip::name(aArgs[0]))) {
    Lib::Out out(aUI.session().printer());
    out << "sys_remove_directory :" 
        << "Can't remove directory " << Lib::sout(aArgs[0]) << "."
        << Lib::endl;
    return;
  }
}

void OnX_sys_env(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  if(!ui_check_args(aUI,aArgs,1)) return;
  Lib::Out out(aUI.session().printer());
  std::string s;
  if(!Slash::getenv(aArgs[0],s)) {
    out << "sys_env :" 
        << " environment variable " << Lib::sout(aArgs[0]) << " not defined."
        << Lib::endl;
    return;
  }
  out << s << Lib::endl;
}

void OnX_sys_putenv(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  std::string last;
  if(!ui_check_min(aUI,aArgs,2,last)) return;
  if(!Slash::putenv(aArgs[0],last)) {
    Lib::Out out(aUI.session().printer());
    out << "sys_putenv :" 
        << " can't set environment variable " << Lib::sout(aArgs[0])
        << " with value " << Lib::sout(last) << "."
        << Lib::endl;
    return;
  }
}

void OnX_sys_execute(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  // args[0,n] command
  std::string last;
  if(!ui_check_min(aUI,aArgs,1,last)) return;
  Slash::execute(last);
}

//////////////////////////////////////////////////////////////////////////////
/// lib //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
void OnX_lib_write_file(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  // aArgs[0] : file name
  // aArgs[1,n] : text : list of lines separated with \n
  std::string last;
  if(!ui_check_min(aUI,aArgs,2,last)) return;
  const std::string& file = aArgs[0];
  if(!Lib::fmanip::write(file,last)) {
    Lib::Out out(aUI.session().printer());
    out << "lib_write_file :"
        << " can't write in " << Lib::sout(file)
        << Lib::endl;
    return;
  }
  //Lib::Out out(aUI.session().printer());
  //out << "lib_write_file :"
  //    << " file " << Lib::sout(file) << " written."
  //    << Lib::endl;
}
//////////////////////////////////////////////////////////////////////////////
void OnX_lib_clone(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,2)) return;
  const std::string& stemplate = aArgs[0];
  const std::string& pack = aArgs[1];
  std::string dir = Lib::fmanip::name(stemplate);
  std::string path,name,suffix;
  Lib::smanip::pathNameSuffix(dir,path,name,suffix);
  if(Lib::Cloner::clone(dir,name,pack,aUI.session().printer())) 
    aUI.echo("Cloning success.");
  else 
    aUI.echo("Cloning failure.");
}
//////////////////////////////////////////////////////////////////////////////
/// session //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void OnX_session_exit(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  // The below exit may call some atexit function (for exa with gtk)
  // that may do some print. We have to switch off OnX output.
  aUI.session().redirectOutput(false);

  int code = 0;
  if(aArgs.size()==1) {
    if(!Lib::smanip::toint(aArgs[0],code)) code = 0;
  }
  ::exit(code);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_session_set_parameter(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] what
// args[1,n] value
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,2,last)) return;
  aUI.session().setParameter(aArgs[0],last);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_session_remove_parameter(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] what
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  aUI.session().removeParameter(aArgs[0]);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_session_printer(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  const std::string& a_0 = aArgs[0];
  if(a_0=="enable") {
    aUI.session().printer().enable();
  } else if(a_0=="disable") {
    aUI.session().printer().disable();
  } else if(a_0=="flush") {
    aUI.session().printer().flush();
  } else {
    Lib::Out out(aUI.session().printer());
    out << "session_printer :" 
        << " " << Lib::sout(a_0) << " is a bad keyword."
        << Lib::endl;
    return;
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX_session_set_output(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] [UI,terminal,defout]
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //  The defaults in OnX::Session constructor are such that
  // session.printer is enabled (on the UI) and
  // std[out,err],std::[cout,cerr] are not redirected 
  // to session.printer (then they still write on terminal).
  //  Then :
  // - session.output=UI redirect all on UI.
  // - session.output=terminal redirect all on the terminal.
  // - session.output=defout redirect only session.printer to UI
  //   which is then the default of OnX::Session constructor.  

  if(!ui_check_args(aUI,aArgs,1)) return;
  const std::string& arg0 = aArgs[0];
  Slash::Core::ISession& session = aUI.session();
  if(arg0=="UI") { // OnX::Session default
    if(!aUI.isComponent("consoleText")) { 
      ::printf("OnX : session_set_output :\n");
      ::printf(" output wanted to be \"UI\" but\n");
      ::printf(" there is no \"consoleText\" widget around.\n");
      ::printf(" Then output are going to be send on termminal.\n");
      session.redirectOutput(false); //it does a session.printer().disable()
    } else {
      session.redirectOutput(true); //it does a session.printer().enable()
    }
  } else if(arg0=="terminal") {
    session.redirectOutput(false); //it does a session.printer().disable()
  } else if(arg0=="defout") {
    session.redirectOutput(false); //it does a session.printer().disable()
    session.printer().enable();
  } else {
    ::printf("session_set_output : unknown keyword \"%s\"\n",arg0.c_str());
    // hoping that the upper ::printf still work !
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX_session_redirect_output(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  bool value;
  if(!ui_check(aUI,aArgs[0],value)) return;
  aUI.session().redirectOutput(value);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_session_toggle_parameter(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] what
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  Slash::Core::ISession& session = aUI.session();
  bool value;
  if(!session_parameterValue(session,aArgs[0],value)) return;
  if(value) session.setParameter(aArgs[0],"false");
  else session.setParameter(aArgs[0],"true");
}
//////////////////////////////////////////////////////////////////////////////
void OnX_session_toggle_verbose(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Core::ISession& session = aUI.session();
  session.setVerboseLevel(session.verboseLevel()==0?1:0);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_session_toggle_helpOutput(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Core::ISession& session = aUI.session();
  std::string svalue;
  if(!session.parameterValue("session.helpOnDialog",svalue)) return;
  session.setParameter("session.helpOnDialog",svalue=="yes"?"no":"yes");
}
//////////////////////////////////////////////////////////////////////////////
void OnX_session_load(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] library
// args[1] package
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aArgs.size()!=2) return;
  aUI.session().load(aArgs[0],aArgs[1]);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_session_load_styles(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] style file
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  Slash::UI::IStyleManager* styleManager = Slash::styleManager(aUI.session());
  if(!styleManager) return;
  styleManager->loadFile(aArgs[0]);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_session_modeling_none(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  session_modeling_none(aUI.session());
}
//////////////////////////////////////////////////////////////////////////////
void OnX_session_dump_parameters(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Out out(aUI.session().printer());
  std::vector<std::string> keys = aUI.session().availableParameters();
  unsigned int l = 0;
  for(unsigned int index=0;index<keys.size();index++) {
    const std::string& key = keys[index];
    l = MAXIMUM(l,key.size());
  }
  for(unsigned int index=0;index<keys.size();index++) {
    const std::string& key = keys[index];
    std::string value;
    aUI.session().parameterValue(key,value);
    out << Lib::sjust(key,l) << " = " << value << Lib::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX_session_flush(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aUI.session().flush();
}
//////////////////////////////////////////////////////////////////////////////
void OnX_session_destroy_manager(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] manager
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  aUI.session().destroyManager(aArgs[0]);
}
//////////////////////////////////////////////////////////////////////////////
/// ui ///////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_do_nothing(Slash::UI::IUI&) {}
void OnX_ui_steer(Slash::UI::IUI& aUI) {aUI.steer();}
void OnX_ui_synchronize(Slash::UI::IUI& aUI) {aUI.synchronize();}

void OnX_ui_exit(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  int exitCode = 0;
  if(aArgs.size()==1) {
    int l;
    if(Lib::smanip::toint(aArgs[0],l)) {
      exitCode = l;
    } else {
      aUI.echo("ui_exit : "+aArgs[0]+" not a number.");
    }
  }
  aUI.exit(exitCode);
}

void OnX_ui_break_modal(Slash::UI::IUI& aUI) {aUI.exit(ONX_BREAK_MODAL);}

void OnX_ui_echo(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  std::string last;
  if(!ui_check_min(aUI,aArgs,0,last)) return;
  aUI.echo(last);
}

void OnX_ui_echo_file(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  // aArgs[0] : file name.
  if(!aArgs.size()) return;
  ui_echo_file(aUI,aArgs[0]);
}

void OnX_ui_warn(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  std::string last;
  if(!ui_check_min(aUI,aArgs,0,last)) return;
  aUI.warn(last);
}

void OnX_ui_ask(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  std::string last;
  if(!ui_check_min(aUI,aArgs,0,last)) return;
  aUI.echo(aUI.ask(last));
}

//////////////////////////////////////////////////////////////////////////////
void OnX_ui_set_parameter(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] what
// args[1,n] value
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,2,last)) return;
  aUI.setParameter(aArgs[0],last);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_toggle_parameter(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] what
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  bool value;
  if(!ui_parameterValue(aUI,aArgs[0],value)) return;
  if(value) aUI.setParameter(aArgs[0],"false");
  else aUI.setParameter(aArgs[0],"true");
}
/* try to use @tree_items@<list>@
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_set_tree(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
// args[0] <widget>.items
// args[1,n] list of items.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_min_args(aUI,aArgs,1)) return;
  std::string sout = "<tree>";
  for(unsigned int index=1;index<aArgs.size();index++) {
    sout += "<treeItem><label>"+aArgs[index]+"</label></treeItem>";
  }
  sout += "</tree>";
  aUI.setParameter(aArgs[0],sout);
}
*/
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_parameterValue(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] widget.parameter
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  std::string value;
  if(!aUI.parameterValue(aArgs[0],value)) return;
  aUI.echo(value);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_execute_file(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
// aArgs[0] interpreter
// aArgs[1] file
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aArgs.size()!=2) return;
  ui_execute_file(aUI,aArgs[0],aArgs[1]);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_load(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] parent
// args[1] file
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,2)) return;
  aUI.load(aArgs[0],aArgs[1]);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_show_dialog( //used in SWIG/Slash.i
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] file
// args[1] dialog
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,2)) return;
  Slash::UI::IWidget* main = aUI.findWidgetByAttribute("class","Window");
  if(!main) {
    Lib::Out out(aUI.session().printer());
    out << "ui_show_dialog :" 
        << " Can't find a Window to attach the dialog."
        << Lib::endl;
    return;
  }
  if(!aUI.load(*main,aArgs[0])) return;
  Slash::UI::IWidget* w = aUI.findWidget(aArgs[1]);
  if(w) w->show();
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_hide(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  Slash::UI::IWidget* w = aUI.findWidget(aArgs[0]);
  if(w) w->hide();
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_show(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  Slash::UI::IWidget* w = aUI.findWidget(aArgs[0]);
  if(w) w->show();
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_show_shell(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  const std::string& component = aArgs[0];
  std::vector<std::string> found = aUI.findParent(component,"shell","");
  if(found.size()) {
    const std::string& shell = found[0];
    Slash::UI::IWidget* w = aUI.findWidget(shell);
    if(w) w->show();
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_help_text( //TO_REVISIT
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// aArgs : text to print in the help dialog (or terminal).
//         It is a list of strings separated by \n.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,0,last)) return;
  Lib::smanip::replace(last,"\\n","\n");
  ui_help_text(aUI,last);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_help_file( //TO_REVISIT
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// aArgs[0] : file name.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  ui_help_file(aUI,aArgs[0]);
}

/////////////////////////////////////////////////////////////////////////////
void OnX_ui_set_file_chooser(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] default directory
// args[1] dirmask
// args[2] [open,save]
// args[3] interp
// args[4,n] script passed to the ok callback.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,5,last)) return;

  aUI.setParameter("fileChooser.dirMask",aArgs[1]); //Must come first.
  aUI.setParameter("fileChooser.directory",aArgs[0]);
  aUI.setParameter("fileChooser.mode",aArgs[2]);

  aUI.setCallback("fileChooser","ok","DLD","OnX ui_hide fileChooser",0);
  aUI.setCallback("fileChooser","ok",aArgs[3],last,1);
  aUI.setCallback("fileChooser","cancel","DLD","OnX ui_hide fileChooser");
}

//////////////////////////////////////////////////////////////////////////////
void OnX_ui_set_callback(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] widget
// args[1] callback
// args[2] interpreter
// args[3,n] script
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,4,last)) return;
  if(!aUI.setCallback(aArgs[0],aArgs[1],aArgs[2],last)) {
    Lib::Out out(aUI.session().printer());
    out << "ui_set_callback :" 
        << " Can't set callback " << Lib::sout(aArgs[1])
        << " on widget " << Lib::sout(aArgs[0])
        << " with interp " << Lib::sout(aArgs[2])
        << " and value " << Lib::sout(last)
        << Lib::endl;
  }
}
/*
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_add_callback(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] widget
// args[1] callback
// args[2] interpreter
// args[4] : script
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,4)) return;
  if(!aUI.setCallback(aArgs[0],aArgs[1],aArgs[2],aArgs[3],-1)) {
    Lib::Out out(aUI.session().printer());
    out << "ui_set_callback :" 
        << " Can't set callback " << Lib::sout(aArgs[1])
        << " on widget " << Lib::sout(aArgs[0])
        << " with interp " << Lib::sout(aArgs[2])
        << " and value " << Lib::sout(aArgs[3])
        << Lib::endl;
  }
}
*/
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_execute_callbacks(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,2)) return;
  aUI.executeCallbacks(aArgs[0],aArgs[1]);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_execute_script(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,2,last)) return;
  aUI.executeScript(aArgs[0],last);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_save(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aUI.save();
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_add_prompt(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
// Add a prompt in the consolePrompt list.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aArgs.size()) return;
  ui_add_prompt(aUI,aArgs[0]);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_lock_screen(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] password
// args[1] interp when unlocking
// args[2] script when unlocking
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,3)) return;
  aUI.lockScreen(aArgs[0],aArgs[1],aArgs[2]);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_set_current_widget(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  const std::string& name = aArgs[0];
  Slash::UI::IWidget* widget = aUI.findWidget(name);
  if(!widget) {
    Lib::Out out(aUI.session().printer());
    out << "ui_set_current_widget :" 
        << "Can't find widget " << Lib::sout(name)
        << Lib::endl;
    return;
  }
  aUI.setCurrentWidget(widget);
}

//////////////////////////////////////////////////////////////////////////////
void OnX_ui_create_component(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] name
// args[1] class
// args[2] parent
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,3)) return;
  aUI.createComponent(aArgs[0],aArgs[1],aArgs[2]);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_copy_component(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] name
// args[1] from
// args[2] parent
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,3)) return;
  aUI.copyComponent(aArgs[0],aArgs[1],aArgs[2]);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_destroy_component(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  aUI.destroyComponent(aArgs[0]);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_destroy_widget(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  Slash::UI::IWidget* widget = Slash::find_widget(aUI,aArgs[0]);
  if(!widget) return;
  aUI.destroyWidget(widget);
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_enable_terminal(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] interp
// args[1,n] prompt
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,2,last)) return;
  aUI.disableTerminal();
  aUI.enableTerminal(last,aArgs[0]);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_disable_terminal(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aUI.disableTerminal();
}
//////////////////////////////////////////////////////////////////////////////
/// widget ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void OnX_widget_find(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] widget name (or @current@)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  // @current@ then not treated.
  Slash::UI::IWidget* widget = aUI.findWidget(aArgs[0]);
  Lib::Out out(aUI.session().printer());
  if(widget) {
    out << "widget "<< Lib::sout(aArgs[0]) << " found." << Lib::endl;
  } else {
    out << "widget "<< Lib::sout(aArgs[0]) << " not found." << Lib::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX_widget_write(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] widget name (or @current@)
// args[1] file
// args[2] format
// args[3] options
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,4)) return;
  Slash::UI::IWidget* widget = Slash::find_widget(aUI,aArgs[0]);
  if(!widget) return;
  if(!widget->write(aArgs[1],aArgs[2],aArgs[3])) {
    Lib::Out out(aUI.session().printer());
    out << "widget_write : "
        << Lib::sout(aArgs[0]) << " failed."
        << Lib::endl;
    return;
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX_widget_manage(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] widget name (or @current@)
// args[1] yes/no
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,2)) return;
  Slash::UI::IWidget* widget = Slash::find_widget(aUI,aArgs[0]);
  if(!widget) return;
  bool value;
  if(!ui_check(aUI,aArgs[1],value)) return;
  if(!widget->manage(value)) {
    Lib::Out out(aUI.session().printer());
    out << "widget_manage : "
        << Lib::sout(aArgs[1]) << " failed."
        << Lib::endl;
    return;
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX_widget_set_parameter(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] widget name (or @current@)
// args[1] parameter
// args[2] value
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,3,last)) return;
  Slash::UI::IWidget* widget = Slash::find_widget(aUI,aArgs[0]);
  if(!widget) return;
  if(!widget->setParameter(aArgs[1],last)) {
    Lib::Out out(aUI.session().printer());
    out << "widget_set_parameter : "
        << " for widget " << Lib::sout(aArgs[0]) 
        << " set parameter " << Lib::sout(aArgs[1])
        << " with value " << Lib::sout(last) << " failed."
        << Lib::endl;
    return;
  }
}
//////////////////////////////////////////////////////////////////////////////
/// data /////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

// used in oroot_root.cxx :
void OnX_declare_accessors_sess(Slash::Core::ISession& aSession) {
  Slash::Data::IProcessor* accessorManager = Slash::accessorManager(aSession);
  if(accessorManager) {
    if(!accessorManager->findAccessor("Lib_Accessor"))
      accessorManager->addAccessor(new OnX::AccessorAccessor(aSession));
    if(!accessorManager->findAccessor("SceneGraph"))
      accessorManager->addAccessor(new OnX::SceneGraphAccessor(aSession));
  }
}

void OnX_declare_accessors(Slash::UI::IUI& aUI) {
  OnX_declare_accessors_sess(aUI.session());
}

//////////////////////////////////////////////////////////////////////////////
void OnX_data_exec(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_min_args(aUI,aArgs,1)) return;

  Slash::Data::IProcessor* da = Slash::accessorManager(aUI.session());
  if(!da) return;

  std::vector<std::string> args;
  for(unsigned int index=1;index<aArgs.size();index++) {
    args.push_back(aArgs[index]);
  }

  da->execute(aArgs[0],args);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_data_collect(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_or_args(aUI,aArgs,1,2)) return;
  if(aArgs.size()==1) {
    data_collect(aUI.session(),aArgs[0]);
  } else {
    data_collect(aUI.session(),aArgs[0],aArgs[1]);
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX_data_visualize(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
// args[0] where to visualize in.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  data_visualize(aUI.session(),aArgs[0]);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_data_filter(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  data_filter(aUI.session(),aArgs[0]);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_data_number(
 Slash::UI::IUI& aUI
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  data_number(aUI.session());
}
//////////////////////////////////////////////////////////////////////////////
void OnX_data_dump(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aArgs.size()==0) {
    data_dump(aUI.session());
  } else if(aArgs.size()==1) {
    data_dump(aUI.session(),aArgs[0]);
  } else {
    Lib::Out out(aUI.session().printer());
    out << "data_dump :" 
        << "zero or one argument expected."
        << Lib::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX_data_set(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,2,last)) return;
  data_set(aUI.session(),aArgs[0],last);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_data_destroy(
 Slash::UI::IUI& aUI
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  data_destroy(aUI.session());
}
//////////////////////////////////////////////////////////////////////////////
void OnX_data_clear(
 Slash::UI::IUI& aUI
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::Data::IProcessor* mgr = Slash::accessorManager(aUI.session());
  if(!mgr) return;
  mgr->eraseHandlers();
}
//////////////////////////////////////////////////////////////////////////////
/// cyclic ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void OnX_cyclic_create(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] name
// args[1] time (milisec)
// args[2] start
// args[3] interp
// args[4,n] script
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,5,last)) return;

  int delay = 0;
  if(!ui_check(aUI,aArgs[1],delay)) return;
  bool start;
  if(!ui_check(aUI,aArgs[2],start)) return;

  Slash::UI::ICyclic* cyclic = aUI.findCyclic(aArgs[0]);
  if(cyclic) {
    Lib::Out out(aUI.session().printer());
    out << "cyclic_create :" 
        << " cyclic " << Lib::sout(aArgs[0]) << " already exists."
        << Lib::endl;
    return;
  }
  cyclic = aUI.createCyclic(aArgs[0],delay,aArgs[3],last);
  if(!cyclic) {
    Lib::Out out(aUI.session().printer());
    out << "cyclic_create :" 
        << " can't create " << Lib::sout(aArgs[0]) << "."
        << Lib::endl;
    return;
  }
  if(start) cyclic->start();
}
//////////////////////////////////////////////////////////////////////////////
void OnX_cyclic_destroy(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] name
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  Slash::UI::ICyclic* cyclic = aUI.findCyclic(aArgs[0]);
  if(!cyclic) return;
  aUI.removeCyclic(cyclic);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_cyclic_stop(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] name
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  Slash::UI::ICyclic* cyclic = aUI.findCyclic(aArgs[0]);
  if(!cyclic) return;
  cyclic->stop();
}
//////////////////////////////////////////////////////////////////////////////
void OnX_cyclic_start(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] name
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  Slash::UI::ICyclic* cyclic = aUI.findCyclic(aArgs[0]);
  if(!cyclic) return;
  cyclic->start();
}
//////////////////////////////////////////////////////////////////////////////
void OnX_cyclic_toggle(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] name
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  Slash::UI::ICyclic* cyclic = aUI.findCyclic(aArgs[0]);
  if(!cyclic) return;
  if(cyclic->started()) {
    cyclic->stop();
  } else {
    cyclic->start();
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void OnX_tree_set(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
// args[0] <widget>.items
// args[1] list of items separated by \n
// Example : tree_set mainTree.items aa\nbb
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,2,last)) return;
  tree_set(aUI,aArgs[0],last);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_tree_set_with_file(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,2)) return;
  tree_set_with_file(aUI,aArgs[0],aArgs[1]);
}
//////////////////////////////////////////////////////////////////////////////
/// currents /////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_add_in_currents(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
)
//////////////////////////////////////////////////////////////////////////////
// args[0] widget to declare in the "currents" list.
// args[1] pdm to maintain the list of potential current widgets.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,2)) return;
  const std::string& name = aArgs[0];
  const std::string& pdm = aArgs[1];
  if(!aUI.isComponent(pdm)) return;
  std::string s = pdm+"_item_" + name;
  aUI.createComponent(s,"MenuItem",pdm);
  aUI.setParameter(s+".label",name);
  aUI.setCallback(s,"activate","DLD","OnX ui_set_current_widget "+name,0);
  aUI.setCallback(s,"activate","DLD","OnX ui_show_shell "+name,1);
}
void OnX_ui_change_current(Slash::UI::IUI&,const std::vector<std::string>&);
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_remove_from_currents(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
// args[0] widget to remove from the "currents" list.
// args[1] pdm to maintain the list of potential current widgets.
  if(!ui_check_args(aUI,aArgs,2)) return;
  const std::string& name = aArgs[0];
  const std::string& pdm = aArgs[1];
  if(aUI.isComponent(pdm)) aUI.destroyComponent(pdm+"_item_" + name);
  OnX_ui_change_current(aUI,aArgs);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_change_current(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
// args[0] widget
// optional :
// args[1] pdm to maintain the list of potential current widgets.
  if(!ui_check_or_args(aUI,aArgs,1,2)) return;
  Slash::UI::IWidget* w = aUI.currentWidget();
  if(!w) return;
  const std::string& name = aArgs[0];
  if(w->name()!=name) return;
  aUI.setCurrentWidget(0);
  //The widget was the current widget.
  //We have to set the current widget to another one.
  //We take the last entry in the passed viewers pdm.
  if(!aArgs.size()) return;
  const std::string& pdm = aArgs[1];
  std::vector<std::string> items = aUI.findChildren(pdm,"name","");
  if(!items.size()) return;
  std::string s = items[items.size()-1];
  Lib::smanip::replace(s,pdm+"_item_","");
  w = aUI.findWidget(s);
  aUI.setCurrentWidget(w);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_add_in_tab_stack(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  if(!ui_check_args(aUI,aArgs,1)) return;
  const std::string& name = aArgs[0];
  std::vector<std::string> found = aUI.findParent(name,"name","");
  if(!found.size()||!found[0].size()) return;
  const std::string& parent = found[0];
  std::string sclass;
  if(!aUI.parameterValue(parent+".class",sclass)) return;
  if(sclass!="TabStack") return;
  aUI.setParameter(name+".tabLabel",name);  
}

void OnX_ui_tab_stack_write(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  // args[0] tab stack name
  // args[1] file prefix
  // args[2] format
  // args[3] options
  if(!ui_check_args(aUI,aArgs,4)) return;

  const std::string& tab_stack = aArgs[0];  
  //FIXME : check we have a TabStack.

  std::string selection;
  if(!aUI.parameterValue(tab_stack+".selection",selection)) {
    std::ostream& out = aUI.session().cout();
    out << "OnX_ui_tab_stack_write : "
        << " can't get selection of " << Slash::sout(tab_stack)
        << std::endl;
  }

  std::vector<std::string> items = aUI.findChildren(tab_stack,"name","");

  std::vector<std::string>::iterator it;
  for(it=items.begin();it!=items.end();++it) {
    Slash::UI::IWidget* widget = aUI.findWidget(*it);
    if(!widget) continue; //ItemML being not a widget.
    
    //to have the widget visible.
    if(!aUI.setParameter(tab_stack+".selection",*it)) {
      std::ostream& out = aUI.session().cout();
      out << "OnX_ui_tab_stack_write : "
          << " can't select " << Slash::sout(*it)
          << std::endl;
      continue;
    }
    aUI.synchronize();

    const std::string& format = aArgs[2];
    std::string file;
    if(aArgs[1].size()) {
      file = aArgs[1]+(*it)+"."+format_suffix(format);
    } else {
      file = (*it)+"."+format_suffix(format);
    }
    if(!widget->write(file,format,aArgs[3])) {
      std::ostream& out = aUI.session().cout();
      out << "OnX_ui_tab_stack_write : "
          << " writing " << Slash::sout(*it)
          << " in " << Slash::sout(file) << " failed."
          << std::endl;
    } else if(aUI.session().verboseLevel()){      
      std::ostream& out = aUI.session().cout();
      out << "OnX_ui_tab_stack_write : "
          << " writing " << Slash::sout(*it)
          << " in " << Slash::sout(file) << " success."
          << std::endl;
    }
  } 

  aUI.setParameter(tab_stack+".selection",selection);  
}

//////////////////////////////////////////////////////////////////////////////
void OnX_ui_find_write(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] top
// args[1] what
// args[2] parameter
// args[3] file prefix
// args[4] format
// args[5] options
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,6)) return;
  std::vector<std::string> items = 
    aUI.findChildren(aArgs[0],aArgs[1],aArgs[2]);

  std::vector<std::string>::iterator it;
  for(it=items.begin();it!=items.end();++it) {
    Slash::UI::IWidget* widget = aUI.findWidget(*it);
    if(!widget) continue; //ItemML being not a widget.
    const std::string& format = aArgs[4];
    std::string file;
    if(aArgs[3].size()) {
      file = aArgs[3]+(*it)+"."+format_suffix(format);
    } else {
      file = (*it)+"."+format_suffix(format);
    }
    if(!widget->write(file,format,aArgs[5])) {
      Lib::Out out(aUI.session().printer());
      out << "OnX_ui_find_write : "
          << " writing " << Lib::sout(*it)
          << " in " << Lib::sout(file) << " failed."
          << Lib::endl;
    } else if(aUI.session().verboseLevel()){      
      Lib::Out out(aUI.session().printer());
      out << "OnX_ui_find_write : "
          << " writing " << Lib::sout(*it)
          << " in " << Lib::sout(file) << " success."
          << Lib::endl;
    }
  } 
}

//////////////////////////////////////////////////////////////////////////////
/// etc //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// to test @cat@
void OnX_dump_args(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  out_args(aUI,aArgs);
}
//////////////////////////////////////////////////////////////////////////////
void OnX_dld_exec_file(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// aArgs[0] file
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  std::vector<std::string> text;
  if(!Lib::fmanip::read(aArgs[0],text)) {
    Lib::Out out(aUI.session().printer());
    out << "dld_exec_file : "
        << " problem reading file " << Lib::sout(aArgs[0]) << "."
        << Lib::endl;
    return;
  }
  for(std::vector<std::string>::iterator it=text.begin();it!=text.end();++it) {
    Lib::smanip::strip(*it,Lib::smanip::trailing);
    if((*it).empty()) continue;
    if(*it=="return") break;
    aUI.executeScript("DLD",*it);
  } 
}
//////////////////////////////////////////////////////////////////////////////
void OnX_dld_exec_text(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// aArgs[0] lines (string separated with \n)
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_args(aUI,aArgs,1)) return;
  std::vector<std::string> text = Lib::smanip::words(aArgs[0],"\n");
  for(std::vector<std::string>::iterator it=text.begin();it!=text.end();++it) {
    Lib::smanip::strip(*it,Lib::smanip::trailing);
    if((*it).empty()) continue;
    if(*it=="return") break;
    aUI.executeScript("DLD",*it);
  } 
}

} // extern "C"

#include <Lib/calc.h>

extern "C" {

/*FIXME : TimeManager ?
//////////////////////////////////////////////////////////////////////////////
void OnX_time_begin(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Time begin = Lib::Time::now();
}
//////////////////////////////////////////////////////////////////////////////
void OnX_time_end(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Time end = Lib::Time::elapsed(begin);
  Lib::Out out(aUI.session().printer());
  out << "time_end :" 
      << elapsed : " << end.seconds() << " s. "
      << Lib::endl;
}
*/
//////////////////////////////////////////////////////////////////////////////
void OnX_lib_calc(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] formula. For exa : 2*x
// args[1,n-1] : variable values
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!ui_check_min_args(aUI,aArgs,1)) return;
  std::vector<std::string> vals;
  for(unsigned int index=1;index<aArgs.size();index++) {
    vals.push_back(aArgs[index]);
  }
  std::string v;
  if(!Lib::calc(aUI.session().printer(),aArgs[0],vals,v)) return;
  Lib::Out out(aUI.session().printer());
  out << v << Lib::endl;
}

}

#include <stdio.h>
#include <iostream>

extern "C" {

//////////////////////////////////////////////////////////////////////////////
void OnX_std_printf(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// To test the stdout redirection.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,0,last)) return;
  ::printf("%s",last.c_str());
  ::printf("\n");
}
//////////////////////////////////////////////////////////////////////////////
void OnX_std_cout(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// To test the std::cout redirection.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,0,last)) return;
  std::cout << last;
  std::cout << std::endl;
}

}

#include <Lib/Processor.h>

extern "C" {

//////////////////////////////////////////////////////////////////////////////
void OnX_ui_execute_script_if(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] condition
// args[1] interpreter
// args[2,n] script
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,3,last)) return;

  //::printf("debug : \"%s\"\n",aArgs[0].c_str());

  bool result;

 {Lib::Processor proc(aArgs[0],aUI.session().printer());
  std::vector<std::string> vars;
  proc.findVariables(vars);
  //Lib::smanip::write(vars,printer);
  proc.compile(vars);
  std::vector<Lib::Value> vs;
  std::string serror;
  proc.cut(vs,result,serror);}

  //::printf("debug : if %d\n",result);

  if(result) {
    aUI.executeScript(aArgs[1],last);
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX_ui_execute_script_loop(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
// args[0] name of the item variable
// args[1] \n list
// args[2] interpreter
// args[3,n] script
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string last;
  if(!ui_check_min(aUI,aArgs,4,last)) return;

  std::vector<std::string> text = Lib::smanip::text(aArgs[1]);
  for(std::vector<std::string>::iterator it=text.begin();it!=text.end();++it) {
    aUI.session().setParameter(aArgs[0],*it);
    aUI.executeScript(aArgs[2],last);
  } 
}

}

extern "C" {

#define ALIAS(a_alias,a_new) \
void a_alias(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {\
 a_new(aUI,aArgs);\
}


ALIAS(OnX_session_redirectOutput,OnX_session_redirect_output)
ALIAS(OnX_session_setParameter,OnX_session_set_parameter)
ALIAS(OnX_set_variable,OnX_session_set_parameter)
ALIAS(OnX_ui_setCurrentWidget,OnX_ui_set_current_widget)
ALIAS(OnX_ui_executeCallbacks,OnX_ui_execute_callbacks)
ALIAS(OnX_ui_executeScript,OnX_ui_execute_script)
ALIAS(OnX_ui_setParameter,OnX_ui_set_parameter)
ALIAS(OnX_ui_setCallback,OnX_ui_set_callback)
ALIAS(OnX_ui_toggleParameter,OnX_ui_toggle_parameter)
ALIAS(OnX_ui_destroyComponent,OnX_ui_destroy_component)
ALIAS(OnX_ui_set_fileChooser,OnX_ui_set_file_chooser)
ALIAS(OnX_widget_set_current,OnX_ui_set_current_widget)

}
