//
//  All functions here should be OnX callbacks, that is to say
// functions with signature :
//   extern "C" {
//     void callback_without_arguments(Slash::UI::IUI&);
//     void callback_with_arguments(Slash::UI::IUI&,
//                                  const std::vector<std::string>&);
//   }
//

//
//  This file does not need HEPVis and Inventor to compile.
//

// Helpers :
#include <OnX/Helpers/OnX.h>

// Lib : 
#include <Lib/Out.h>
#include <Lib/sout.h>

extern "C" {

/////////////////////////////////////////////////////////////////////////////
void OnX_EditExec_setFile(
 Slash::UI::IUI& aUI
,const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aArgs.size()!=1) return;

  std::string file = aArgs[0];

  std::vector<std::string> text;
  if(!Lib::fmanip::read(file,text)) return;
  aUI.setParameter("OnX_EditExec_text.value",Lib::smanip::tostring(text,"\n"));
  // Set interpreter :
  std::string suffix = Lib::smanip::suffix(file);
  std::string interp = session_interpreterFromSuffix(aUI.session(),suffix);
  if(interp.empty()) interp = "DLD";
  aUI.setParameter("OnX_EditExec_interpreter.default",interp);

  // Save the name of the file on some session parameter :
  aUI.session().setParameter("OnX_EditExec.file",file);
}
/////////////////////////////////////////////////////////////////////////////
void OnX_EditExec_ok(
 Slash::UI::IUI& aUI
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string svalue;
  if(!aUI.parameterValue("OnX_EditExec_text.value",svalue)) return;
  std::string sinterp;
  if(!aUI.parameterValue("OnX_EditExec_interpreter.value",sinterp)) return;

  Slash::UI::IScriptManager* sm = Slash::scriptManager(aUI.session());
  if(!sm) return;
  Slash::UI::IInterpreter* interp = sm->findInterpreter(sinterp);
  if(!interp) return;
  std::string ssuf = interp->suffix();
  std::string scmd = interp->execFileCommand();
  if(!scmd.size()) {
    Lib::Out out(aUI.session().printer());
    out << "OnX_EditExec_ok :"
        << " no file command given for " << Lib::sout(sinterp) << "." 
        << Lib::endl;
    return;
  }

  // Save text in a temporary file :
  std::string name;
  if(Lib::fmanip::tmpname(name,".","OnX_",ssuf.size()?("."+ssuf):"")) {
    if(Lib::fmanip::write(name,svalue)) {

      std::string file = Lib::fmanip::name(name);
      Lib::smanip::replace(scmd,"@onx_file@",file);

      sm->record(sinterp,svalue); //history
      Slash::UI::IInterpreter::Options params;
      interp->executeScript(scmd,params);

      Lib::fmanip::remove(file);
    }
  } else {
    Lib::Out out(aUI.session().printer());
    out << "OnX_EditExec_ok : can't open a temporary file." << Lib::endl;
  }
}

} // extern "C"
