// this :
#include "SceneGraphAccessor.h"

// Slash :
#include <Slash/Data/IIterator.h>
#include <Slash/Tools/UI.h>

// Lib :
#include <Lib/Value.h>
#include <Lib/Debug.h>
#include <Lib/cstring.h>

// Inventor :
#include <Inventor/nodes/SoShape.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/manips/SoTransformManip.h>

// HEPVis :
#include <HEPVis/SbColors.h>
#include <HEPVis/misc/SoStyleCache.h>
#include <HEPVis/misc/SoTools.h>
#include <HEPVis/nodes/SoSceneGraph.h>
#include <HEPVis/nodes/SoHighlightMaterial.h>
#include <HEPVis/nodekits/SoRegion.h>

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
OnX::SceneGraphAccessor::SceneGraphAccessor(
 Slash::Core::ISession& aSession
)
:BaseAccessor(aSession.printer())
,fSession(aSession)
,fName("SceneGraph")
,fSceneGraphs(0)
,fNumber(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  addProperty("name",Lib::Property::STRING);
  addProperty("highlight",Lib::Property::BOOLEAN);
  addProperty("with_manip",Lib::Property::BOOLEAN);
  addProperty("address",Lib::Property::POINTER);
}
//////////////////////////////////////////////////////////////////////////////
OnX::SceneGraphAccessor::~SceneGraphAccessor(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  clear();
}
//////////////////////////////////////////////////////////////////////////////
std::string OnX::SceneGraphAccessor::name(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fName;
}
namespace OnX {
class SceneGraphAccessorIterator : public virtual Slash::Data::IIterator {
public: //Slash::Data::IIterator
  virtual Slash::Data::IAccessor::Data object() {
    if(fIndex>=fNumber) return 0;
    return &(fList[fIndex]);
  }
  virtual void next() {fIndex++;}
  virtual void* tag() { return 0;}
public:
  SceneGraphAccessorIterator(OnX::SceneGraph* aList,unsigned int aNumber)
    :fIndex(0),fList(aList),fNumber(aNumber) {
    Lib::Debug::increment("OnX::SceneGraphAccessorIterator");
  }
  virtual ~SceneGraphAccessorIterator() {
    Lib::Debug::decrement("OnX::SceneGraphAccessorIterator");
  }
private:
  unsigned int fIndex;
  OnX::SceneGraph* fList;
  unsigned int fNumber;
};
}
//////////////////////////////////////////////////////////////////////////////
Slash::Data::IIterator* OnX::SceneGraphAccessor::iterator(
) 
//////////////////////////////////////////////////////////////////////////////
// Returned object should be deleted by the receiver.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  getSceneGraphs(iteratorArguments());
  return new SceneGraphAccessorIterator(fSceneGraphs,fNumber);
}
//////////////////////////////////////////////////////////////////////////////
Slash::Core::IValue* OnX::SceneGraphAccessor::findValue(
 Slash::Data::IAccessor::Data aData
,const std::string& aName
,void*
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SceneGraph* obj = (SceneGraph*)aData;
  if(Lib::eq(aName,"address")) {
    return new Lib::Value((void*)obj);
  } else if(Lib::eq(aName,"name")) {
    if(obj->fNode->isOfType(SoSceneGraph::getClassTypeId())==TRUE) {
      SoSceneGraph* soSG = (SoSceneGraph*)obj->fNode;
      return new Lib::Value(soSG->getString());
    } else {
      SoSearchAction searchAction;
      searchAction.setFind(SoSearchAction::TYPE);
      searchAction.setType(SoShape::getClassTypeId());
      searchAction.apply(obj->fNode);
      SoPath* path = searchAction.getPath();
      std::string spath = path ? path->getTail()->getName().getString():"nil";
      //printf("debug : SceneGraphAccessor::getProp: %ld name:%s|\n",
      //   obj->fNode,spath.c_str());
      return new Lib::Value(spath);
    }
  } else if(Lib::eq(aName,"highlight")) {
    SoSearchAction searchAction;
    searchAction.setFind(SoSearchAction::TYPE);
    searchAction.setType(SoHighlightMaterial::getClassTypeId());
    searchAction.apply(obj->fNode);
    SoPath* path = searchAction.getPath();
    bool value = false;
    if(path) {
      SoHighlightMaterial* material = (SoHighlightMaterial*)path->getTail();
      if(material->getSavedMaterial()) {
        value = true;
      }
    }
    return new Lib::Value(value);
  } else if(Lib::eq(aName,"with_manip")) {
    SoSearchAction searchAction;
    searchAction.setFind(SoSearchAction::TYPE);
    searchAction.setType(SoTransformManip::getClassTypeId());
    searchAction.apply(obj->fNode);
    SoPath* path = searchAction.getPath();
    bool value = path ? true : false;
    return new Lib::Value(value);
  } else {
    return new Lib::Value();
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX::SceneGraphAccessor::destroy(
 Slash::Data::IAccessor::Data aData
,void*
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  SceneGraph* obj = (SceneGraph*)aData;
  //printf("debug : destroy %lx\n",aData);
  if(!obj->fParent) return;
  if(obj->fParent->isOfType(SoGroup::getClassTypeId())==FALSE) return;
  ((SoGroup*)obj->fParent)->removeChild(obj->fNode);
}
//////////////////////////////////////////////////////////////////////////////
void OnX::SceneGraphAccessor::set(
 Slash::Data::IAccessor::Data aData
,const std::string& aWhat
,const std::string& aValue
,void*
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //::printf("debug : OnX::SceneGraphAccessor::set : %lx %s %s\n",
  //    aData,aWhat.c_str(),aValue.c_str());

  SceneGraph* obj = (SceneGraph*)aData;
  if(!obj->fNode) return;

  // Look first for a name :
  SoSearchAction searchAction;
  searchAction.setFind(SoSearchAction::NAME);
  searchAction.setName(aWhat.c_str());
  searchAction.setInterest(SoSearchAction::ALL);
  searchAction.apply(obj->fNode);
  SoPathList pathList = searchAction.getPaths();
  int number = pathList.getLength();
  if(!number) {
    // Not found ; look for a type :
    SoType type = SoType::fromName(aWhat.c_str());
    if(type==SoType::badType()) {
      Lib::Out out(fSession.printer());
      out << "Unknown type " << aWhat << Lib::endl;
      return;
    }
    SoSearchAction searchAction;
    searchAction.setFind(SoSearchAction::TYPE);
    searchAction.setType(type);
    searchAction.setInterest(SoSearchAction::ALL);
    searchAction.apply(obj->fNode);
    pathList = searchAction.getPaths();
    number = pathList.getLength();
  }

  std::vector<std::string> words;
  Lib::smanip::words(aValue," ",words);

  if( (words.size()==2) && (words[0]=="manip") && (words[1]=="off") ) { 
    for(int index=0;index<number;index++) {
      SoFullPath* path = (SoFullPath*)pathList[index];
      SoNode* node = path->getTail();
      if(!node) continue;
      if(node->isOfType(SoTransformManip::getClassTypeId())) { 
        ((SoTransformManip*)node)->replaceManip(path,NULL);
        ((SoTransformManip*)node)->unref();
      }
    }
  } else if( (words.size()==2) && (words[0]=="manip") ) { 
    SoType type = SoType::fromName(words[1].c_str());
    if(type==SoType::badType()) {
      Lib::Out out(fSession.printer());
      out << "Unknown type \"" << words[1] << "\"" << Lib::endl;
      return;
    }
    if(!type.isDerivedFrom(SoTransformManip::getClassTypeId())) {
      Lib::Out out(fSession.printer());
      out << "Accessor \"" << words[1] 
          << "\" not a SoTransformManip." << Lib::endl;
      return;
    }
    for(int index=0;index<number;index++) {
      SoFullPath* path = (SoFullPath*)pathList[index];
      SoNode* node = path->getTail();
      if(!node) continue;
      if(node->isOfType(SoTransform::getClassTypeId())) { 
        SoTransformManip* manip = (SoTransformManip*)type.createInstance();
        if(manip) {
          manip->ref();
          manip->replaceNode(path);
        }
      }
    }

  } else if( (words.size()==4) && (words[0]=="diffuseColor") ) {

    //FIXME : do not do it if color is the same !

    // Have to look for an SoHighlightMaterial 
    // (that may come from a SoStyleCache) :
    for(int index=0;index<number;index++) {
      SoFullPath* path = (SoFullPath*)pathList[index];
      SoNode* node = path->getTail();
      if(!node) continue;
      //::printf("debug : data_set : name : %lu \"%s\" \"%s\"\n",node,
      //  node->getName().getString(),node->getTypeId().getName().getString());
      if(node->getName()==SoTools::highlightName()) {
        SoHighlightMaterial* material = (SoHighlightMaterial*)node;
        SoMaterial* savedMaterial = material->getSavedMaterial();
        if(savedMaterial) {
          if(savedMaterial->getName()==SoStyleCache::cachedName()) {
            SoNode* cpy = savedMaterial->copy(); //then not cached
            if(cpy) {
              cpy->ref();
              cpy->setName("");
              cpy->set(aValue.c_str());
              material->setSavedMaterial((SoMaterial*)cpy);
              savedMaterial->unref();
            }      
          } else {
            savedMaterial->set(aValue.c_str());
          }
        } else {
          // Anomaly.
          Lib::Out out(fSession.printer());
          out << " \"highlight\" node without a savedMaterial !" << Lib::endl;
        }
      } else {
        if(node->getName()==SoStyleCache::cachedName()) {
          if(path->getLength()>=2) {
            SoNode* parent = path->getNodeFromTail(1);
            if(parent->isOfType(SoGroup::getClassTypeId())) { 
              SoGroup* grp = (SoGroup*)parent;
              SoNode* cpy = node->copy(); //then not cached.
              if(cpy) {
                grp->replaceChild(node,cpy);     
                cpy->setName("");
                cpy->set(aValue.c_str());
              }
            }        
          }      
        } else {
          node->set(aValue.c_str());
        }
      }
    }

  } else if( (words.size()==2) && (words[0]=="solid") ) {

    //FIXME : do not do it if solid is the same !

    bool value = false;
    if(!Lib::smanip::tobool(words[1],value)) { 
      Lib::Out out(fSession.printer());
      out << "\"" << words[1] << "\" not a boolean." << Lib::endl;
    }
    
    for(int index=0;index<number;index++) {
      SoFullPath* path = (SoFullPath*)pathList[index];
      SoNode* node = path->getTail();
      if(node) node->set(aValue.c_str());
    }

    // Have to set the light model :
   {SoSearchAction searchAction;
    searchAction.setFind(SoSearchAction::TYPE);
    searchAction.setType(SoLightModel::getClassTypeId());
    searchAction.setInterest(SoSearchAction::ALL);
    searchAction.apply(obj->fNode);
    pathList = searchAction.getPaths();
    number = pathList.getLength();
    for(int index=0;index<number;index++) {
      SoFullPath* path = (SoFullPath*)pathList[index];
      SoNode* node = path->getTail();
      if(!node) continue;
      if(node->getName()==SoStyleCache::cachedName()) {
        if(path->getLength()>=2) {
          SoNode* parent = path->getNodeFromTail(1);
          if(parent->isOfType(SoGroup::getClassTypeId())) { 
            SoGroup* grp = (SoGroup*)parent;
            SoNode* cpy = node->copy(); //then not cached.
            if(cpy) {
              grp->replaceChild(node,cpy);     
              cpy->setName("");
              ((SoLightModel*)cpy)->model.setValue(value ?
                                              SoLightModel::PHONG:
                                              SoLightModel::BASE_COLOR);
            }
          }        
        }      
      } else {
        ((SoLightModel*)node)->model.setValue(value ?
                                              SoLightModel::PHONG:
                                              SoLightModel::BASE_COLOR);
      }
    }}

    // Have to set the draw style :
   {SoSearchAction searchAction;
    searchAction.setFind(SoSearchAction::TYPE);
    searchAction.setType(SoDrawStyle::getClassTypeId());
    searchAction.setInterest(SoSearchAction::ALL);
    searchAction.apply(obj->fNode);
    pathList = searchAction.getPaths();
    number = pathList.getLength();
    for(int index=0;index<number;index++) {
      SoFullPath* path = (SoFullPath*)pathList[index];
      SoNode* node = path->getTail();
      if(!node) continue;
      if(node->getName()==SoStyleCache::cachedName()) {
        if(path->getLength()>=2) {
          SoNode* parent = path->getNodeFromTail(1);
          if(parent->isOfType(SoGroup::getClassTypeId())) { 
            SoGroup* grp = (SoGroup*)parent;
            SoNode* cpy = node->copy(); //then not cached.
            if(cpy) {
              grp->replaceChild(node,cpy);     
              cpy->setName("");
              ((SoDrawStyle*)cpy)->style.setValue(value ?
                                             SoDrawStyle::FILLED:
                                             SoDrawStyle::LINES);
            }
          }        
        }      
      } else {
        ((SoDrawStyle*)node)->style.setValue(value ?
                                             SoDrawStyle::FILLED:
                                             SoDrawStyle::LINES);
      }
    }}

    // Have to set the material emissiveColor (because of VRML) :
   {SoSearchAction searchAction;
    searchAction.setFind(SoSearchAction::TYPE);
    searchAction.setType(SoMaterial::getClassTypeId());
    searchAction.setInterest(SoSearchAction::ALL);
    searchAction.apply(obj->fNode);
    pathList = searchAction.getPaths();
    number = pathList.getLength();
    for(int index=0;index<number;index++) {
      SoFullPath* path = (SoFullPath*)pathList[index];
      SoNode* node = path->getTail();
      if(!node) continue;
      //::printf("debug : data_set : name : %lu \"%s\" \"%s\"\n",node,
      //  node->getName().getString(),node->getTypeId().getName().getString());
      if(node->getName()==SoTools::highlightName()) {
        SoHighlightMaterial* material = (SoHighlightMaterial*)node;
        if(value) { //want solid 
          ((SoMaterial*)material)->emissiveColor.setValue(SbColor_black);
        } else {
          ((SoMaterial*)material)->emissiveColor = 
             ((SoMaterial*)material)->diffuseColor;
        }
        SoMaterial* savedMaterial = material->getSavedMaterial();
        if(savedMaterial) {
          if(savedMaterial->getName()==SoStyleCache::cachedName()) {
            SoNode* cpy = savedMaterial->copy(); //then not cached
            if(cpy) {
              cpy->ref();
              cpy->setName("");
              if(value) { //want solid 
                ((SoMaterial*)cpy)->emissiveColor.setValue(SbColor_black);
              } else {
                ((SoMaterial*)cpy)->emissiveColor = 
                  ((SoMaterial*)cpy)->diffuseColor;
              }
              material->setSavedMaterial((SoMaterial*)cpy);
              savedMaterial->unref();
            }      
          } else {
            if(value) { //want solid 
              ((SoMaterial*)savedMaterial)->emissiveColor.setValue
                  (SbColor_black);
            } else {
              ((SoMaterial*)savedMaterial)->emissiveColor = 
                ((SoMaterial*)savedMaterial)->diffuseColor;
            }
          }
        } else {
          // Anomaly.
          Lib::Out out(fSession.printer());
          out << " \"highlight\" node without a savedMaterial !" << Lib::endl;
        }
      } else {
        if(node->getName()==SoStyleCache::cachedName()) {
          if(path->getLength()>=2) {
            SoNode* parent = path->getNodeFromTail(1);
            if(parent->isOfType(SoGroup::getClassTypeId())) { 
              SoGroup* grp = (SoGroup*)parent;
              SoNode* cpy = node->copy(); //then not cached.
              if(cpy) {
                grp->replaceChild(node,cpy);     
                cpy->setName("");

                if(value) { //want solid 
                  ((SoMaterial*)cpy)->emissiveColor.setValue(SbColor_black);
                } else {
                  ((SoMaterial*)cpy)->emissiveColor = 
                    ((SoMaterial*)cpy)->diffuseColor;
                }
              }
            }        
          }      
        } else {
          if(value) { //want solid 
            ((SoMaterial*)node)->emissiveColor.setValue(SbColor_black);
          } else {
            ((SoMaterial*)node)->emissiveColor = 
              ((SoMaterial*)node)->diffuseColor;
          }
        }
      }
    }}

  } else {
    for(int index=0;index<number;index++) {
      SoFullPath* path = (SoFullPath*)pathList[index];
      SoNode* node = path->getTail();
      if(node) node->set(aValue.c_str());
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
void OnX::SceneGraphAccessor::getSceneGraphs(
 const std::vector<std::string>& aArgs
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  clear();

  Slash::UI::IUI* ui = Slash::find_UI(fSession);
  if(!ui) return;

  if(!aArgs.size()) {
    Lib::Out out(fSession.printer());
    out << "OnX::SceneGraphAccessor::getSceneGraphs :" 
        << " no viewer specified." 
        << Lib::endl;
    return;
  }
  SoRegion* soRegion = find_SoRegion(*ui,aArgs[0]);
  if(!soRegion) return;

  SoNode* node = soRegion->getTopSeparator();
  if(!node) return;
 
  SoPathList* list = SoTools::getSceneGraphs(node);
  if(!list) return;

  unsigned int number = list->getLength();
  if(number) {
    fSceneGraphs = new SceneGraph[number];
    if(fSceneGraphs) {
      for(unsigned int index=0;index<number;index++) {
        SoFullPath* path = (SoFullPath*)(*list)[index];
        SceneGraph* sg = &(fSceneGraphs[index]);
        sg->fNode = path->getTail();
        int l = path->getLength();
        sg->fParent = l>=2 ? path->getNodeFromTail(1) : 0;
      }
      fNumber = number;
    }
  }

  delete list;
}
//////////////////////////////////////////////////////////////////////////////
void OnX::SceneGraphAccessor::clear(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  delete [] fSceneGraphs;
  fSceneGraphs = 0;
  fNumber = 0;
}
