//
//  All functions here should be OnX callbacks, that is to say
// functions with signature :
//   extern "C" {
//     void callback_without_arguments(Slash::UI::IUI&);
//     void callback_with_arguments(Slash::UI::IUI&,
//                                  const std::vector<std::string>&);
//   }
//

// Helpers :
#include <OnX/Helpers/OnX.h>

#include <Slash/UI/ICallbackData.h>
#include <Slash/UI/ISoViewer.h>
#include <Slash/Tools/Managers.h>

#include <OnX/Net/Protocol.h>

extern "C" {

void OnX_viewer_read_scene(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  // args[0] widget name (or @current@)
  // args[1] file
  // args[2] format
  // args[3] placement
  if(!ui_check_args(aUI,aArgs,4)) return;
  Slash::UI::ISoViewer* soViewer = find_SoViewer(aUI,aArgs[0]);
  if(!soViewer) return;
  std::string file = Lib::fmanip::name(aArgs[1]);
  if(!soViewer->readScene(file,aArgs[2],aArgs[3])) {
    aUI.echo("Can't read file "+file+".");
  }
}

void OnX_viewer_clear(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  // args[0] widget name (or @current@)
  // args[1] what. For example : all, staticScene, dynamicScene.
  if(!ui_check_args(aUI,aArgs,2)) return;
  Slash::UI::ISoViewer* soViewer = find_SoViewer(aUI,aArgs[0]);
  if(!soViewer) return;
  soViewer->clear(aArgs[1]);
}

void OnX_viewer_view_all(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  if(!ui_check_args(aUI,aArgs,1)) return;
  Slash::UI::ISoViewer* soViewer = find_SoViewer(aUI,aArgs[0]);
  if(!soViewer) return;
  soViewer->viewAll();
}

void OnX_viewer_stop_animating(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  if(!ui_check_args(aUI,aArgs,1)) return;
  Slash::UI::ISoViewer* soViewer = find_SoViewer(aUI,aArgs[0]);
  if(!soViewer) return;
  soViewer->stopAnimating();
}

void OnX_viewer_set_title(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  if(!ui_check_args(aUI,aArgs,2)) return;
  Slash::UI::ISoViewer* soViewer = find_SoViewer(aUI,aArgs[0]);
  if(!soViewer) return;
  soViewer->setTitle(aArgs[1]);
}

void OnX_viewer_remove_manips(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  if(!ui_check_args(aUI,aArgs,1)) return;
  Slash::UI::ISoViewer* soViewer = find_SoViewer(aUI,aArgs[0]);
  if(!soViewer) return;
  soViewer->removeManips();
}

void OnX_viewer_detect_intersections(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  // args[0] : viewer
  // args[1] : stop at first (boolean).
  // args[2] : what (nil, scene, staticScene, dynmicScene, region)
  if(!ui_check_args(aUI,aArgs,3)) return;
  Slash::UI::ISoViewer* soViewer = find_SoViewer(aUI,aArgs[0]);
  if(!soViewer) return;
  bool b;
  if(!ui_check(aUI,aArgs[1],b)) return;
  soViewer->detectIntersections(b,aArgs[2]);
}

void OnX_viewer_set_background_color(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  //  args[0] widget
  //  args[1] color name.
  // or :
  //  args[0] widget
  //  args[1] red
  //  args[2] green
  //  args[3] blue
  if(!ui_check_or_args(aUI,aArgs,2,4)) return;
  Slash::UI::ISoViewer* soViewer = find_SoViewer(aUI,aArgs[0]);
  if(!soViewer) return;
  double r,g,b;
  if(aArgs.size()==2) {
    if(!ui_check_rgb(aUI,aArgs[1],r,g,b)) return;
  } else {
    if(!ui_check(aUI,aArgs[1],r)) return;
    if(!ui_check(aUI,aArgs[2],g)) return;
    if(!ui_check(aUI,aArgs[3],b)) return;
  }
  soViewer->setBackgroundColor(r,g,b);
}

#define VIEWER_SET(a_x,a_X) \
void OnX_viewer_set_##a_x(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {\
  if(!ui_check_args(aUI,aArgs,2)) return;\
  Slash::UI::ISoViewer* soViewer = find_SoViewer(aUI,aArgs[0]);\
  if(!soViewer) return;\
  bool v;\
  if(!ui_check(aUI,aArgs[1],v)) return;\
  soViewer->set##a_X(v);\
}

VIEWER_SET(decoration,Decoration)
VIEWER_SET(viewing,Viewing)
VIEWER_SET(head_light,Headlight)
VIEWER_SET(feedback_visibility,FeedbackVisibility)
VIEWER_SET(auto_clipping,AutoClipping)
VIEWER_SET(smoothing,Smoothing)
VIEWER_SET(animation,Animation)

void OnX_viewer_render(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){
  if(!ui_check_args(aUI,aArgs,1)) return;
  Slash::UI::ISoViewer* soViewer = find_SoViewer(aUI,aArgs[0]);
  if(!soViewer) return;
  soViewer->render();
}

void OnX_viewer_collect( Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs){

//    <collect exec="DLD">OnX data_collect SceneGraph( @cat@ @this@ @cat@ ) highlight==true</collect>
//    <collect exec="DLD">OnX data_dump</collect>
//    <collect exec="DLD">OnX data_filter name</collect>
//    <collect exec="DLD">OnX data_dump @{AccessorManager.dumpMode}@</collect>

  //if(!ui_check_args(aUI,aArgs,1)) return;

  Slash::Core::ISession& session = aUI.session();

  std::string what("SceneGraph");
  if(aArgs.size()) {
    what += "("+aArgs[0]+")";  //For exa : SceneGraph(@current@)
  }

  data_collect(session,what,"highlight==true");

  Slash::Net::INetManager* netManager = Slash::netManager(session);
  if(netManager) {
    if(!netManager->isClient()) { //server asks data to the client.
      netManager->sendString
        (ONX_PROTOCOL_DATA+":"+data_values(session,"name"));

      Slash::UI::IUI::Options opts;
      opts.push_back(Slash::UI::IUI::Option("where","remote"));
      std::string svalue;
      session.parameterValue("AccessorManager.dumpMode",svalue);
      aUI.executeScript("DLD","OnX data_dump "+svalue,opts);
    }
  } else {
    data_filter(session,"name");
    std::string svalue;
    session.parameterValue("AccessorManager.dumpMode",svalue);
    data_dump(session,svalue);
  }

}
void OnX_viewer_rectangular_collect(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  // args[0] widget name (or @current@)

  //Lib::Out out(aUI.session().printer());
  //out << "OnX_viewer_rectangular_collect :" 
  //    << " callback value : " << aUI.callbackValue()
  //    << Lib::endl;

  std::vector<std::string> words;
  Lib::smanip::words(aUI.callbackValue()," ",words);
  if(words.size()!=4) return;
  int x;
  if(!Lib::smanip::toint(words[0],x)) return;
  int y;
  if(!Lib::smanip::toint(words[1],y)) return;
  unsigned int w;
  if(!Lib::smanip::touint(words[2],w)) return;
  unsigned int h;
  if(!Lib::smanip::touint(words[3],h)) return;

  if(!ui_check_args(aUI,aArgs,1)) return;
  Slash::UI::ISoViewer* soViewer = find_SoViewer(aUI,aArgs[0]);
  if(!soViewer) return;

  soViewer->collect(x,y,w,h);
}

void OnX_viewer_set_parameter(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {
  // args[0] what
  // args[1,n-1] : partial value
  //  value is built with : args[1] + " " + args[2] + ..... + args[n-1]
  std::string last;
  if(!ui_check_min(aUI,aArgs,2,last)) return;
  Slash::UI::IWidget* widget = aUI.currentWidget();
  if(!widget) return;
  if(!Lib_Cast(*widget,Slash::UI::ISoViewer)) return;
  widget->setParameter(aArgs[0],last);
}

}

extern "C" {

#define ALIAS(a_alias,a_new) \
void a_alias(Slash::UI::IUI& aUI,const std::vector<std::string>& aArgs) {\
 a_new(aUI,aArgs);\
}


ALIAS(OnX_viewer_setParameter,OnX_viewer_set_parameter)
ALIAS(OnX_viewer_set_color,OnX_viewer_set_background_color)

}
