// this :
#include <OnX/Core/BaseUI.h>

// Slash :
#include <Slash/Core/ISession.h>
#include <Slash/Core/IWriter.h>
#include <Slash/Core/IRelation.h>
#include <Slash/UI/IUI.h>
#include <Slash/UI/ICyclic.h>
#include <Slash/UI/IWidget.h>
#include <Slash/UI/IWidgetFactory.h>
#include <Slash/Tools/Managers.h>

// Lib :
#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/smanip.h>
#include <Lib/fmanip.h>
#include <Lib/Messages.h>
#include <Lib/System.h>
#include <Lib/FileWriter.h>
#include <Lib/Debug.h>
#include <Lib/Element.h>
#include <Lib/ItemML.h>
#include <Lib/BaseML.h>
#include <Lib/Cast.h>
#include <Lib/Arguments.h>
#include <Lib/cstring.h>

// OnX :
#include <OnX/Version.h>
#include <OnX/Core/CallbackData.h>
#include <OnX/Core/Messages.h>
#include <OnX/Core/WidgetClass.h>
#include <OnX/Core/Widget.h>
#include <OnX/Core/WidgetFactory.h>

#include <algorithm> //std::find

namespace OnX {

class UIML : public Lib::BaseML {
public:
  UIML(WidgetFactory& aFactory,int aVerbose):Lib::BaseML(aFactory,aVerbose){
    std::vector<std::string> tags(5);
    tags[0] = "OnX";
    tags[1] = "widget";
    tags[2] = "reference";
    tags[3] = "comment";
    tags[4] = "aliases";
    setTags(tags);
  }
};

class ExecuteMessage : public virtual Slash::Core::IMessage {
public:
  virtual void* cast(const std::string& aClass) const {
    if_Lib_SCast(OnX::ExecuteMessage)
    else if_Lib_SCast(Slash::Core::IMessage)
    else return 0;
  }
  virtual std::string message() const {
    return std::string("OnX::ExecuteMessage : \"") 
           + fAction->name() + std::string("\".");}
public:
  ExecuteMessage(OnX::Action* aAction):fAction(aAction) { 
    Lib::Debug::increment("OnX::ExecuteMessage"); 
  }
  virtual ~ExecuteMessage() {
    Lib::Debug::decrement("OnX::ExecuteMessage"); 
  }
  OnX::Action* action() const {return fAction;}
private:
  OnX::Action* fAction; //we may have a copy if needed.
};

}

static bool saveItemML(Lib::ItemML&,const std::string&,int,Slash::Core::IWriter&);
static bool saveItemML(Lib::ItemML&,Lib::Out&,int,int,Slash::Core::IWriter&);
static bool isGeneratedName(const std::string&);


OnX::Execute::Execute(BaseUI& aUI,const std::string& aWhat)
:fUI(aUI),fWhat(aWhat){}
const std::string& OnX::Execute::what() const {return fWhat;}
OnX::BaseUI& OnX::Execute::baseUI(){return fUI;}

#ifdef WIN32
// disable the warning about the usage of "this" in the constructor.
#pragma warning(disable:4355)
#endif

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
OnX::BaseUI::BaseUI(Slash::Core::ISession& aSession)
:fSession(aSession)
,fTopItemML(0)
,fScriptManager(0)
,fNetManager(0)
,fWidgetNameCounter(0)
,fBeingDestroyed(false)
,fVerboseLevel(0)
,fSteerLevel(0)
,fRelationManager(0)
,fCurrentWidget(0)
,fAskToTerminal(false)
//IB
,fWidgetFactory(*this,fSession.printer(),fWidgetClasses)
{
  Lib::Debug::increment("OnX::BaseUI");

  fScriptManager = Slash::scriptManager(fSession);
  fRelationManager = Slash::relationManager(fSession);
  fNetManager = Slash::netManager(fSession);

  std::string svalue;
  if(fSession.parameterValue("UI_Manager.verbose",svalue)) {
    int value;
    if(Lib::smanip::toint(svalue,value)) fVerboseLevel = value;
  } else {
    fVerboseLevel = fSession.verboseLevel();
  }

  // IB :
  createWidgetTable();
  createLanguageList();
  createPropertyList();
}

OnX::BaseUI::BaseUI(const BaseUI& aFrom)
:fSession(aFrom.fSession)
,fTopItemML(0)
,fScriptManager(0)
,fNetManager(0)
,fWidgetNameCounter(0)
,fBeingDestroyed(false)
,fVerboseLevel(0)
,fSteerLevel(0)
,fRelationManager(0)
,fCurrentWidget(0)
,fAskToTerminal(false)
//IB
,fWidgetFactory(*this,aFrom.fSession.printer(),fWidgetClasses)
{
  Lib::Debug::increment("OnX::BaseUI");
}

OnX::BaseUI::~BaseUI() { 
  //FIXME : Python : on a Python callback, someone
  //                 may do a sys.exit(0) that will
  //                 delete the ui object whilst in ui.steer() !
  //if(fSteerLevel) {
  //}

  fCurrentWidget = 0;

  // Remove relations concerning this UI :
  if(fRelationManager) fRelationManager->removeWith(this);

 {for(unsigned int index=0;index<fFactories.size();index++) {
    delete fFactories[index];
  }
  fFactories.clear();}

  // IB :
 {for(unsigned int index=0;index<fWidgetClasses.size();index++) {
    delete fWidgetClasses[index];
  }
  fWidgetClasses.clear();}

 {for(std::list<Action*>::iterator it=fActions.begin();it!=fActions.end();) {
    Action* action = *it;
    it = fActions.erase(it);
    delete action;
  }}

  delete fTopItemML;
  Lib::Debug::decrement("OnX::BaseUI");
}

OnX::BaseUI& OnX::BaseUI::operator=(const BaseUI& aFrom){ return *this;}

int OnX::BaseUI::steerLevel() const {return fSteerLevel;}

int OnX::BaseUI::steer(){
  if(!fTopItemML) return 0;
  executeCallbacks(*fTopItemML,"enterSteering");
  int value = nativeSteer();
  executeCallbacks(*fTopItemML,"leaveSteering");
  return value;
}

static bool dld_dummy_Function(const std::string&,const Slash::UI::IInterpreter::Options&,const Slash::UI::IInterpreter::Aliases&,void*){
  //::printf("debug : dummy\n");
  return true;
}

bool OnX::BaseUI::finalize(){
  if(!fTopItemML) return true;

  // Disbale terminal here (and not in <Xxx>UI)
  // and then before the destroy callbacks in order 
  // to avoid : DLD> OnX_Finalize.
  disableTerminal(); 

  if(fVerboseLevel) {
    Lib::Out out(fSession.printer());
    out << "OnX::BaseUI::finalize :"
        << " execute destroyCallbacks..." << Lib::endl;
  }

  //NOTE : at this point the UI_Manager is no more in the session
  //       and then a Slash::find_UI will fail, and then
  //       the DLD destroy callbacks will fail.
 {Slash::UI::IInterpreter* interpreter = 
    fScriptManager->findInterpreter("DLD");
  if(interpreter) {
    interpreter->setFunction(dld_dummy_Function,0);
  }}

  Execute tag(*this,"destroy");
  fTopItemML->postExecuteBackward(executeItemMLCallbacks,&tag);

  if(fVerboseLevel) {
    Lib::Out out(fSession.printer());
    out << "OnX::BaseUI::finalize :"
        << " execute destroyCallbacks : end" << Lib::endl;
  }

  return true;
}

bool OnX::BaseUI::beingDestroyed() const {return fBeingDestroyed;}

Slash::Core::ISession& OnX::BaseUI::session(){return fSession;}
Slash::Core::IWriter& OnX::BaseUI::printer() const {return fSession.printer();}
bool OnX::BaseUI::isValid() const {
  if(!fTopItemML) return false;
  return true;
}

void* OnX::BaseUI::cast(const std::string& aClass) const {
  if_Lib_SCast(Slash::UI::IUI)
  else if_Lib_SCast(IIB)
  else if_Lib_SCast(OnX::BaseUI)
  else return 0;
}

//NOTE : keyword use too in Action.cxx

#define S_BACK "background"

bool OnX::BaseUI::executeScript(
 const std::string& aInterp
,const std::string& aScript
,const Slash::UI::IUI::Options& aParams
){
  if(!fScriptManager) return false;

  //{for(unsigned int index=0;index<aParams.size();index++) {
  //   ::printf("debug : BaseUI : %d : %s %s\n",index,
  //      aParams[index].first.c_str(),aParams[index].second.c_str());
  //}}

  Lib::Arguments params(aParams);

  std::string where = "local";
  if(!params.find("where",where)) where = "local";

  bool bg = false;
  if(params.find(S_BACK,bg)) params.remove(S_BACK);

  //old syntax :
  std::string interp = aInterp;
  std::vector<std::string> words;
  Lib::smanip::words(aInterp," ",words);
  if(words.size()==1) {
  } else if(words.size()==2) {
    if(((words[1]=="bg")||(words[1]==S_BACK)) ){
      //FIXME : correct Panoramix and then let the below message.
      //Lib::Out out(fSession.printer());
      //out << "OnX::BaseUI::executeScript :"
      //    << " deprecated bg syntax."
      //    << " Use the background=\"yes\" XML attribute."
      //    << Lib::endl;
      bg = true;
      interp = words[0];
    } else {
      Lib::Out out(fSession.printer());
      out << "OnX::BaseUI::executeScript :"
          << " bad syntax for interp " << Lib::sout(aInterp)
          << Lib::endl;
      return false;
    }
  } else {
    Lib::Out out(fSession.printer());
    out << "OnX::BaseUI::executeScript :"
        << " bad syntax for interp " << Lib::sout(aInterp)
        << Lib::endl;
    return false;
  }

  if(bg) {
    if(fVerboseLevel) {
      Lib::Out out(fSession.printer());
      out << "OnX::BaseUI::executeScript :"
          << " post execution of "
          << " " << Lib::sout(interp)
          << " " << Lib::sout(aScript)
          << Lib::endl;
    }
    // Process in background :
    // Have a copy of fCallbackData in Action.

    Action* action = 
      new Action(*this,0,0,interp,aScript,params.args(),"post",0);
    action->setCallbackData(fCallbackData);
    notify(new OnX::ExecuteMessage(action)); //ExecuteMessage takes ownershup.
    return true; 
  } 

  if(where=="remote"){
    if(fVerboseLevel) {
      Lib::Out out(fSession.printer());
      out << "OnX::BaseUI::executeScript :"
          << " remote execution of "
          << " " << Lib::sout(interp)
          << " " << Lib::sout(aScript)
          << Lib::endl;
    }

    if(!fNetManager) {
      return fScriptManager->executeScript(interp,aScript,aParams);
    }

    //FIXME : pass fCallbackData
    //FIXME : pass aParams
    if(!fNetManager->executeScript(interp,aScript,aParams)) {
      Lib::Out out(fSession.printer());
      out << "OnX::BaseUI::executeScript :"
          << " net execute script faild for "
          << " " << Lib::sout(interp)
          << " " << Lib::sout(aScript)
          << Lib::endl;
      return false;
    } 

    // process Ends :
    while(true) {
      std::string what;
      std::vector<std::string> args;
      if(!fNetManager->waitEnd(what,args)) {
        Lib::Out out(fSession.printer());
        out << "OnX::BaseUI::executeScript :"
            << " net waitEnd failed for "
            << " " << Lib::sout(interp)
            << " " << Lib::sout(aScript)
            << Lib::endl;
        return false;
      } 
  
      if(fVerboseLevel) {
        Lib::Out out(fSession.printer());
        out << "OnX::BaseUI::executeScript :"
            << " receive End " << Lib::sout(what) << "."
            << Lib::endl;
      }

      if(what=="end") {
        break;

      } else if(what=="graphics") {
        std::string file;
        if(!Lib::fmanip::tmpname(file,".","OnX_",".hiv")) {
          Lib::Out out(fSession.printer());
          out << "OnX::BaseUI::executeScript :"
              << " can't get tmp file name."
              << Lib::endl;
          return false;
        }
        if(!fNetManager->waitFile(file)) {
           Lib::Out out(fSession.printer());
           out << "OnX::BaseUI::executeScript :"
               << " net waitFile failed for "
               << " " << Lib::sout(interp)
               << " " << Lib::sout(aScript)
               << Lib::endl;
           return false;
        } 
        //bool stat = 
        receiveScene(args[0],args[1],file);
        Lib::fmanip::remove(file);
        //if(!stat) return false; //not fatal.
    
      } else {
        Lib::Out out(fSession.printer());
        out << "OnX::BaseUI::executeScript :"
            << " bad keyword " << Lib::sout(what)
            << " for script " << Lib::sout(aScript)
            << Lib::endl;
        return false;
      }

    } //end while Ends
    return true;
 
  } else if(where=="local"){
    return fScriptManager->executeScript(interp,aScript,aParams);
  } else {
    Lib::Out out(fSession.printer());
    out << "OnX::BaseUI::executeScript :"
        << " bad where keyword " << Lib::sout(where)
        << " for script " << Lib::sout(aScript)
        << Lib::endl;
    return false;
  }
}

void OnX::BaseUI::echo(const std::string& aString){
  Lib::Out out(fSession.printer());
  out << aString << Lib::endl;
}

void OnX::BaseUI::addDispatcher(Slash::UI::IUI::Dispatcher aDispatcher){
  if(std::find(fDispatchers.begin(),
               fDispatchers.end(),
               aDispatcher)!=fDispatchers.end()) return;
  fDispatchers.push_back(aDispatcher);
}

void OnX::BaseUI::removeDispatcher(Slash::UI::IUI::Dispatcher aDispatcher){
  fDispatchers.remove(aDispatcher);
}

std::string OnX::BaseUI::callbackValue(){return fCallbackData.value();}
const Slash::UI::ICallbackData& OnX::BaseUI::callbackData(){
  return fCallbackData;
}

bool OnX::BaseUI::setCallbackData(const Slash::UI::ICallbackData& aFrom){
  return fCallbackData.copy(aFrom);
}

void OnX::BaseUI::setCallbackData(const CallbackData& aValue){
  fCallbackData = aValue;
}

void OnX::BaseUI::deleteChildren(const std::string&){}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void OnX::BaseUI::addAction(Action* aAction){fActions.push_back(aAction);}

void OnX::BaseUI::removeActions(OnX::Action::NativeWidget aComponent){
  for(std::list<Action*>::iterator it=fActions.begin();it!=fActions.end();) {
    Action* action = *it;
    if(action->nativeWidget()==aComponent) {
      it = fActions.erase(it);
      delete action;
    } else {
      ++it;
    }
  }
}

OnX::Action* OnX::BaseUI::getAction(OnX::Action::NativeWidget aComponent,const std::string& aName,int aIndex){
  int index = 0;
  for(std::list<Action*>::iterator it=fActions.begin();
      it!=fActions.end();++it) {
    Action* action = *it;
    if( (action->nativeWidget()==aComponent) &&
        (action->name()==aName) ) {
      if(index==aIndex) {
        return action;
      } else {
        index++;
      }
    }
  }
  return 0;
}

std::vector<std::string> OnX::BaseUI::disableCallbackScript(OnX::Action::NativeWidget aComponent,const std::string& aName){
  // Disable the scripts of the aName callback.
  // It can be used to avoid recursive loop in case
  // the same callback is indirectly triggered in the scripts.
  // Usage :
  //   std::vector<std::string> tmp = disableCallbackScript(...)
  //   // call something which does a callback
  //   enableCallbackScript(...,tmp)
  std::vector<std::string> scripts;
  for(std::list<Action*>::iterator it=fActions.begin();
      it!=fActions.end();++it) {
    Action* action = *it;
    if( (action->nativeWidget()==aComponent) &&
        (action->name()==aName) ) {
      scripts.push_back(action->script());
      action->setScript("");
    }
  }
  return scripts;
}

bool OnX::BaseUI::enableCallbackScript(OnX::Action::NativeWidget aComponent,const std::string& aName,const std::vector<std::string>& aScripts){
  unsigned int index = 0;
  for(std::list<Action*>::iterator it=fActions.begin();
      it!=fActions.end();++it) {
    Action* action = *it;
    if( (action->nativeWidget()==aComponent) &&
        (action->name()==aName) ) {
      if(index>=aScripts.size()) return false;
      action->setScript(aScripts[index]);
      index++;
    }
  }
  return true;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::load(Slash::UI::IWidget& aParent,const std::string& aToLoad,bool aIsFile,bool aShow){
  // Load a file or a string and add its items in the fTopItemML tree.
  Lib::ItemML* itemML = Lib_Cast(aParent,Lib::ItemML);
  if(!itemML) {
    Lib::Out out(printer());
    out << "OnX::BaseUI::load :"
        << " IWidget parent is not a Lib::ItemML."
        << Lib::endl;
    return false;
  }
  Lib::ItemML* top;
  return load(itemML,aToLoad,aIsFile,aShow,true,top);
}

bool OnX::BaseUI::load(const std::string& aParent,const std::string& aToLoad,bool aIsFile,bool aShow) {
  // Load a file or a string and add its items in the fTopItemML tree.
  if(fTopItemML) {
    Lib::ItemML* itemML = 0;
    if(aParent.empty()) {
      itemML = fTopItemML;
    } else {
      itemML = fTopItemML->findItem(aParent);
    }
    if(!itemML) {
      Lib::Out out(printer());
      out << "OnX::BaseUI::load :"
          << " parent widget " << Lib::sout(aParent) << " not found."
          << Lib::endl;
      return false;
    }
    Lib::ItemML* top;
    if(!load(itemML,aToLoad,aIsFile,aShow,true,top)) return false;
  } else {
    if(aParent.size()) {
      Lib::Out out(printer());
      out << "OnX::BaseUI::load :"
          << " can't specify a parent whilst no top."
          << Lib::endl;
      return false;
    } else {
      //startup (from UI_Manager::create).
      if(aToLoad=="@empty@") { //See OnX/examples/Python/gui.py
        // Someone may want to start a GUI from scratch
        // for example from Python. 
        // (See OnX/examples/Python/OnX_exa_gui_py.py)
        fTopItemML = new Lib::ItemML("OnX",fWidgetFactory,0);
      } else {
        Lib::ItemML* top;
        if(!load(0,aToLoad,aIsFile,aShow,false,top)) return false;
        fTopItemML = top;
        Execute tag(*this,"create");
        top->postExecute(executeItemMLCallbacks,&tag);
      }
    }
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::save(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fTopItemML) return false;
  return saveItemML(*fTopItemML,fTopItemML->file(),
                    fVerboseLevel,fSession.printer());
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::toWidgetAndResource(
 const std::string& aString
,std::string& aWidget
,std::string& aResource
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aWidget = "";
  aResource = "";
  std::vector<std::string> words;
  Lib::smanip::words(aString,".",words);
  if(words.size()!=2) {
    Lib::Out out(printer());
    out << "OnX::BaseUI::toWidgetAndResource :"
        << " Bad syntax in " << Lib::sout(aString) << "."
        << " <widget>.<parameter> expected." << Lib::endl;
    return false;
  }
  if( (words[0]=="currentViewer") || (words[0]=="viewer") ) {
    Lib::Out out(printer());
    out << "OnX::BaseUI::toWidgetAndResource :"
        << " words currentViewer and viewer are deprecated."
        << " Use ui_SoViewer helper." << Lib::endl;
    //FIXME : deprecate that. No reason that BaseUI contains
    //        any viewer logic.
    //aWidget = currentViewer();  
    //if(aWidget.empty()) return false; //No viewer around.
    return false;
  } else if(words[0]=="this") {
    Lib::Out out(printer());
    out << "OnX::BaseUI::toWidgetAndResource :"
        << " words this is deprecated."
        << " Use @this@ instead." << Lib::endl;
    aWidget = fCallbackData.component();
  } else {
    aWidget = words[0];
  }
  aResource = words[1];
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::propertyValue(
 const std::string& aWidget
,const std::string& aResource
,std::string& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aValue = "";
  if(!fTopItemML) return false;
  Lib::ItemML* itemML = fTopItemML->findItem(aWidget);
  if(!itemML) return false;
  itemML->propertyValue(aResource,aValue);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
Slash::UI::IWidget* OnX::BaseUI::findWidget(
 const std::string& aName
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fTopItemML) return 0;
  Lib::ItemML* itemML = fTopItemML->findItem(aName);
  if(!itemML) return 0;
  return Lib_Cast(*itemML,Slash::UI::IWidget);
}
//////////////////////////////////////////////////////////////////////////////
Slash::UI::IWidget* OnX::BaseUI::findWidgetByAttribute(
 const std::string& aAttribute
,const std::string& aValue
,bool aUpDown //true
,bool aLeftRight //true
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fTopItemML) return 0;
  Lib::ItemML* itemML = 
    fTopItemML->findByAttribute(aAttribute,aValue,aUpDown,aLeftRight);
  if(!itemML) return 0;
  return Lib_Cast(*itemML,Slash::UI::IWidget); //May be not an IWidget !
}

void* OnX::BaseUI::findNativeWidget(const std::string& aName){
  if(!fTopItemML) return 0;
  Lib::ItemML* itemML = fTopItemML->findItem(aName);
  if(!itemML) return 0;
  return itemML->getData1();
}

Lib::ItemML* OnX::BaseUI::findItemML(const std::string& aName) {
  if(!fTopItemML) return 0;
  return fTopItemML->findItem(aName);
}

bool OnX::BaseUI::isComponent(const std::string& aName){
  return (findNativeWidget(aName) ? true : false);
}

bool OnX::BaseUI::createNativeWidget(Lib::ItemML& aItemML){
  if(!createIf(aItemML)) return true;

  const std::string& tagName = aItemML.tagName();

  if(tagName=="comment") {

    return true;

  } else if(tagName=="aliases") {

    return true;

  } else if( (tagName=="OnX") ||
             (tagName=="reference") ) {

    return createChildren(aItemML);

  } else if(tagName=="widget") {

    std::string sclass;
    if(!aItemML.attributeValue("","class",sclass)) {
      Lib::Out out(fSession.printer());
      out << "BaseUI::createNativeWidget :" 
          << " widget without a class." << Lib::endl;
      aItemML.printXML(fSession.printer(),"  ");
      return false;
    }

    if(sclass=="InputDialog") {
      return createChildren(aItemML);
    }
    
    std::string name = buildName(aItemML);
    Lib::ItemML* parent = parentItemML(aItemML); 

    if(fVerboseLevel) {
      Lib::Out out(fSession.printer());
      out << "BaseUI::createNativeWidget :" 
          << " class " << Lib::sout(sclass)
          << " name " << Lib::sout(name)
          << "..." << Lib::endl;
    }

    return createNativeWidget(aItemML,sclass,name,parent);

  } else {

    Lib::Out out(fSession.printer());
    out << "BaseUI::createNativeWidget :"
        << " unknown XML tag " << Lib::sout(tagName) 
        << Lib::endl;
    return false;

  }
}

Slash::UI::IWidget* OnX::BaseUI::createWidget(
 const Slash::UI::IWidgetClass& aClass
,const std::string& aName
,Slash::UI::IWidget& aParent
,const std::vector<std::string>& aArgs
){
  // Look for a plugin factory :
  for(unsigned int index=0;index<fFactories.size();index++) {
    Slash::UI::IWidgetFactory* fac = fFactories[index];
    Slash::UI::IWidget* w = fac->createWidget(aClass,aName,aParent,aArgs);
    if(w) return w;
  }
  // Look to the predefined widgets :
  if(!fTopItemML) {
    Lib::Out out(printer());
    out << "OnX::BaseUI::createWidget :"
        << " topItemML not created." << Lib::endl;
    return 0;
  }
  Lib::ItemML* parentItemML = Lib_Cast(aParent,Lib::ItemML);
  if(!parentItemML) {
    Lib::Out out(printer());
    out << "OnX::BaseUI::createWidget :"
        << " IWidget parent is not a Lib::ItemML."
        << Lib::endl;
    return 0;
  }
  Lib::ItemML* itemML = createItemML(aName,aClass.name(),parentItemML);
  if(!itemML) return 0;
  if(!createNativeWidget(*itemML)) return 0;
  Slash::UI::IWidget* w = Lib_Cast(*itemML,Slash::UI::IWidget);
  if(!w) {
    Lib::Out out(printer());
    out << "OnX::BaseUI::createWidget :"
        << " Lib::ItemML not a IWidget."
        << Lib::endl;
    return 0;
  }
  Execute tag(*this,"create");
  itemML->postExecute(executeItemMLCallbacks,&tag);
  return w;
}

void OnX::BaseUI::showShells(Lib::ItemML& aItemML){
  // Show the shell widgets.
  Slash::UI::IWidget* widget = Lib_Cast(aItemML,Slash::UI::IWidget);
  if(widget && widget->nativeWidget()) {
    std::string sshow;
    // If show property not here, we show the widget.
    // If show is here and is not "no", we show the widget.
    if( !aItemML.propertyValue("show",sshow) || (sshow!="no") ) {
      widget->show();
    }
  } else { // Look children :
    const std::list<Lib::ItemML*>& list = aItemML.children();
    std::list<Lib::ItemML*>::const_iterator it;
    for(it=list.begin();it!=list.end();++it) showShells(*(*it));
  }
}

bool OnX::BaseUI::show(){
  if(!fTopItemML) return false;
  showShells(*fTopItemML);
  return true;
}

bool OnX::BaseUI::show(const std::string& aName){
  Slash::UI::IWidget* w = findWidget(aName);
  if(!w) return false;
  w->show();
  return true;
}

bool OnX::BaseUI::createComponent(
 const std::string& aName
,const std::string& aClass
,const std::string& aParent
){
  // Look for a plugin factory :
  for(unsigned int index=0;index<fFactories.size();index++) {
    Slash::UI::IWidgetFactory* fac = fFactories[index];
    if(fac->createComponent(aName,aClass,aParent)) return true;
  }
  // Look to the predefined widgets :
  if(!fTopItemML) {
    Lib::Out out(printer());
    out << "OnX::BaseUI::createComponent :"
        << " topItemML not created." << Lib::endl;
    return false;
  }
  Lib::ItemML* parentItemML = 0;
  if(aParent.size()) {
    parentItemML = fTopItemML->findItem(aParent);
    if(!parentItemML) {
      Lib::Out out(printer());
      out << "OnX::BaseUI::createComponent :"
          << " component " << Lib::sout(aParent) << " not found." << Lib::endl;
      return false;
    }
  } else {
    parentItemML = fTopItemML;
  }
  Lib::ItemML* itemML = createItemML(aName,aClass,parentItemML);
  if(!itemML) return false;
  if(!createNativeWidget(*itemML)) return false;
  Execute tag(*this,"create");
  itemML->postExecute(executeItemMLCallbacks,&tag);
  return true;
}

bool OnX::BaseUI::destroyWidget(Slash::UI::IWidget*& aWidget){
  Lib::ItemML* itemML = Lib_Cast(*aWidget,Lib::ItemML);
  if(!itemML) return false;
  bool status = destroyItemML(itemML);
  aWidget = 0;
  return status;
}

bool OnX::BaseUI::destroyComponent(const std::string& aName){
  if(!fTopItemML) return false;
  Lib::ItemML* itemML = fTopItemML->findItem(aName);
  if(!itemML) return false;
  return destroyItemML(itemML);
}

bool OnX::BaseUI::copyComponent(
 const std::string& aName
,const std::string& aComponent
,const std::string& aParent
){
  if(!fTopItemML) {
    Lib::Out out(printer());
    out << "OnX::BaseUI::copyComponent :"
        << " topItemML not created." << Lib::endl;
    return false;
  }
  Lib::ItemML* fromItemML = fTopItemML->findItem(aComponent);
  if(!fromItemML) {
    Lib::Out out(printer());
    out << "OnX::BaseUI::copyComponent :"
        << " component " << Lib::sout(aComponent) << " not found." << Lib::endl;
    return false;
  }
  Lib::ItemML* parentItemML = fTopItemML->findItem(aParent);
  if(!parentItemML) {
    Lib::Out out(printer());
    out << "OnX::BaseUI::copyComponent :"
        << " parent component " << Lib::sout(aParent) << " not found." 
        << Lib::endl;
    return false;
  }
  Lib::ItemML* itemML = fromItemML->createCopy(parentItemML);
  if(!itemML) return false;
  parentItemML->addChild(itemML);
  itemML->setAttributeValue("","name",aName);
  if(!createNativeWidget(*itemML)) return false;
  Execute tag(*this,"create");
  itemML->postExecute(executeItemMLCallbacks,&tag);
  return true;
}

Lib::ItemML* OnX::BaseUI::parentItemML(Lib::ItemML& aItemML){
  Lib::ItemML* parent = aItemML.parent(); 
  while(parent) {
    void* tkParent = parent->getData1();
    if(tkParent) return parent;
    parent = parent->parent();
  }
  return 0;
}

//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> OnX::BaseUI::findChildren(
 const std::string& aComponent
,const std::string& aWhat
,const std::string& aParameter
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> found;
  if(!fTopItemML) return found;

  Lib::ItemML* itemML = 0;
  if(aComponent.empty()) {
    itemML = fTopItemML;
  } else {
    itemML = fTopItemML->findItem(aComponent);
  }
  if(!itemML) return found;

  const std::list<Lib::ItemML*>& list = itemML->children();

  std::list<Lib::ItemML*>::const_iterator it;
  for(it=list.begin();it!=list.end();++it) {
    if(searchChild(*(*it),aWhat,aParameter,found)) break;
  }

  return found;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::searchChild(
 Lib::ItemML& aItemML
,const std::string& aWhat
,const std::string& aParameter
,std::vector<std::string>& aFound
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aWhat=="class") {
    // aParameter not used.
    std::string sclass;
    aItemML.attributeValue("","class",sclass);
    aFound.push_back(sclass);
  } else if(aWhat=="name") {
    // aParameter not used.
    std::string value;
    aItemML.attributeValue("","name",value);
    aFound.push_back(value);
  } else if(aWhat=="nameFromClass") {
    std::string sclass;
    aItemML.attributeValue("","class",sclass);
    if(sclass==aParameter) {
      std::string value;
      aItemML.attributeValue("","name",value);
      if(value.size()) {
        aFound.push_back(value);
      } else {
        Lib::Out out(printer());
        out << "OnX::BaseUI::searchChild :" 
            << " found a component with matching class " << Lib::sout(sclass)
            << " but without name." << Lib::endl;
      }
    }
  } else if(aWhat=="labelFromClass") {
    std::string sclass;
    aItemML.attributeValue("","class",sclass);
    if(sclass==aParameter) {
      std::string value;
      aItemML.propertyValue("label",value);
      if(value.size()) {
        aFound.push_back(value);
      } else {
        Lib::Out out(printer());
        out << "OnX::BaseUI::searchChild :" 
            << " found a component with matching class " << Lib::sout(sclass)
            << " but without label." << Lib::endl;
      }
    }
  } else if(aWhat=="nameFromLabel") {
    std::string label;
    aItemML.propertyValue("label",label);
    if(label==aParameter) {
      std::string value;
      aItemML.attributeValue("","name",value);
      if(value.size()) {
        aFound.push_back(value);
        return true; //stop.
      } else {
        Lib::Out out(printer());
        out << "OnX::BaseUI::searchChild :" 
            << " found a component with matching label " << Lib::sout(label)
            << " but without name." << Lib::endl;
      }
    }
  }

  const std::list<Lib::ItemML*>& list = aItemML.children();
  std::list<Lib::ItemML*>::const_iterator it;
  for(it=list.begin();it!=list.end();++it) {
    if(searchChild(*(*it),aWhat,aParameter,aFound)) return true;
  }

  return false;
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> OnX::BaseUI::findParent(
 const std::string& aComponent
,const std::string& aWhat
,const std::string& aParameter
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> found;
  if(!fTopItemML) return found;
  Lib::ItemML* itemML = fTopItemML->findItem(aComponent);
  if(!itemML) return found;
  searchParent(*itemML,aWhat,aParameter,found);
  return found;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::searchParent(
 Lib::ItemML& aItemML
,const std::string& aWhat
,const std::string& aParameter
,std::vector<std::string>& aFound
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::ItemML* parent = aItemML.parent(); 
  while(parent) {

    if(aWhat=="name") {

      std::string name;
      parent->attributeValue("","name",name);
      aFound.push_back(name);
      return true;

    } else if(aWhat=="class") {

      std::string sclass;
      if(!parent->attributeValue("","class",sclass)){
        Lib::Out out(printer());
        out << "BaseUI::searchParent :" 
            << " ItemML without a class." << Lib::endl;
        return false;
      }
      aFound.push_back(sclass);
      return true;

    } else if(aWhat=="shell") {
      // aParameter not used.

      if(parent->tagName()=="widget") {
        std::string sclass;
        if(!parent->attributeValue("","class",sclass)){
          Lib::Out out(printer());
          out << "BaseUI::searchParent :" 
              << " ItemML without a class." << Lib::endl;
          return false;
        }
        if((sclass=="Window")||(sclass=="Dialog")) {
          std::string name;
          parent->attributeValue("","name",name);
          aFound.push_back(name);
          return true;
        }
      }

    }
 
    parent = parent->parent();
  }

  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::executeCallbacks(
 const std::string& aComponent
,const std::string& aCallback
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fTopItemML) return false;
  if(aComponent=="*") {
    Execute tag(*this,aCallback);
    fTopItemML->postExecute(executeItemMLCallbacks,&tag);
    return true;
  } else {
    Lib::ItemML* itemML = fTopItemML->findItem(aComponent);
    if(!itemML) return false;
    return executeCallbacks(*itemML,aCallback);
  }
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::executeCallbacks(
 Lib::ItemML& aItemML
,const std::string& aCallback
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Slash::UI::IWidget* widget = Lib_Cast(aItemML,Slash::UI::IWidget);
  if(widget) {
    return widget->executeCallbacks(aCallback);
  } else {
    // Callback of an ItemML which is not a IWidget. //FIXME.
    //FIXME : the below duplicate OnX::Widget::executeCallbacks code.
    std::string name;
    aItemML.attributeValue("","name",name);
    const std::list<Lib::Element*> props = aItemML.properties();
    bool found = false;
    std::list<Lib::Element*>::const_iterator it;
    for(it=props.begin();it!=props.end();++it) {
      if(aCallback==(*it)->name()) {
  
        std::string sexec;
        (*it)->attributeValue("exec",sexec);
  
        fCallbackData = CallbackData(0,name,"","",0,0);
        executeScript(sexec,(*it)->value(),(*it)->attributes()); //FIXME : ret val
        fCallbackData = CallbackData();
  
        found = true;
      }
    }
    return found;
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::destroyItemML(
 Lib::ItemML*& aItemML
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Execute tag(*this,"destroy");
  aItemML->postExecuteBackward(executeItemMLCallbacks,&tag);

  /*FIXME
  std::string sclass;
  aItemML->attributeValue("","class",sclass);
  if(sclass=="OnX") {
  } else if(sclass=="comment") {
  } else if(sclass=="aliases") {
  } else if(sclass=="Reference") {    
    // Do it with a better implementation of OnX::Widget.
  } else if(sclass=="InputDialog") {
  }*/

  // Remove in parent before calling destroyNativeWidget :
  Lib::ItemML* parentItemML = aItemML->parent();
  if(parentItemML) {
    parentItemML->removeChild(aItemML,false); //false to not delete aItemML.
  } else { 
    std::string name;
    aItemML->attributeValue("","name",name);
    Lib::Out out(printer());
    out << "OnX::BaseUI::destroyItemML :"
        << " WARNING : itemML " << Lib::sout(name) << " with no parent !" 
        << Lib::endl;
  }

  //Destroy native widgets.
  bool status = destroyNativeWidget(*aItemML);

  delete aItemML;
  aItemML = 0;
  return status;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool OnX::BaseUI::load(
 Lib::ItemML* aParent // !aParent = startup
,const std::string& aToLoad
,bool aIsFile
,bool aShow
,bool aExecCreate
,Lib::ItemML*& aTop
){
  aTop = 0;
  // Load a file or a string and add its items in the fTopItemML tree.
  UIML uiml(fWidgetFactory,fVerboseLevel);
  if(fNetManager) {
    uiml.setFileGetter(fNetManager->fileGetter());
  } else {
    uiml.setFileGetter(0);
  }
  if(aIsFile) {
    if(!uiml.loadFile(aToLoad,false,fSession.printer())) return false;
  } else {
    if(!uiml.loadString(aToLoad,fSession.printer())) return false;
  }
  Lib::ItemML* top = uiml.topItem();
  if(!top) return true; //it is ok if the file is empty.
  if(!solveInputDialog(*top)) return false;
  if(!solveCopy(top)) return false;
  if(!setSessionParameters(*top)) return false;
  if(!setAliases(*top)) return false;
  if(aParent) {
    // Look if already loaded !
    Lib::ItemML* itemML = findFirstWidget(*top);
    if(itemML) {
      std::string name;
      itemML->attributeValue("","name",name);
      if(name.empty()) {
        // create a component with a OnX_Component_<number> name.
      } else if(aParent->findItem(name)) {
        return true; // Already loaded.
      }
    }
  }
  // Create widgets, set hierarchy.
  if(aParent) top->setParent(aParent);
  if(!createNativeWidget(*top)) return false;
  uiml.empty();
  if(aParent) aParent->addChild(top);
  if(aShow) showShells(*top);

  if(aExecCreate) {
    Execute tag(*this,"create");
    top->postExecute(executeItemMLCallbacks,&tag);
  }

  aTop = top;

  return true;
}

Lib::ItemML* OnX::BaseUI::createItemML(
 const std::string& aName
,const std::string& aClass
,Lib::ItemML* aParent
){
  std::vector<Lib::ItemML::Attribute> atts(2);
  atts[0] = Lib::ItemML::Attribute("class",aClass);
  atts[1] = Lib::ItemML::Attribute("name",aName);
  Lib::ItemML* itemML = fWidgetFactory.create("widget",atts,aParent);
  if(!itemML) return 0;
  if(aParent) aParent->addChild(itemML);
  return itemML;
}

//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::executeItemMLCallbacks(
 Lib::ItemML& aItemML
,void* aTag
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aItemML.tagName()=="comment") return false;
  if(aItemML.tagName()=="aliases") return false;
  OnX::Execute* tag = (OnX::Execute*)aTag;
  tag->baseUI().executeCallbacks(aItemML,tag->what());
  return true;
}
//////////////////////////////////////////////////////////////////////////////
std::string OnX::BaseUI::buildName(
 Lib::ItemML& aItemML                      
)
//////////////////////////////////////////////////////////////////////////////
// "OnX_Component_" is filtered in Lib::ItemML::save.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string name;
  if(!aItemML.attributeValue("","name",name)) { 
    name = createWidgetName();
    //FIXME : do it here ?
    aItemML.setAttributeValue("","name",name);
  } else {
    // used in solveCopy. See ViewerWindow.onx.
    std::string sat;
    if(!aItemML.attributeValue("","at",sat)) sat = "yes";
    if(sat=="yes") {
      std::string s;
      if(!atat(name,s)) {
        Lib::Out out(printer());
        out << "OnX::BaseUI::buildName :"
            << " atat on " << Lib::sout(name) << " failed."
            << Lib::endl;
      }
      name = s;
      aItemML.setAttributeValue("","name",name);
    }
    if(!name.size()) {
      Lib::Out out(printer());
      out << "OnX::BaseUI::buildName :"
          << " empty widget name. Create a non empty one."
          << Lib::endl;
      name = createWidgetName();
      aItemML.setAttributeValue("","name",name);
    }
  }
  return name;
}

bool OnX::BaseUI::createChildren(Lib::ItemML& aItemML){
  // Create children :
  const std::list<Lib::ItemML*>& list = aItemML.children();
  std::list<Lib::ItemML*>::const_iterator it;
  for(it=list.begin();it!=list.end();++it) {
    if(!createNativeWidget(*(*it))) return false;
  }
  return true;
}

bool OnX::BaseUI::setSessionParameters(Lib::ItemML& aItemML){
  const std::list<Lib::Element*>& properties = aItemML.properties();
  std::list<Lib::Element*>::const_iterator it;
  for(it=properties.begin();it!=properties.end();++it) {
    if( Lib::eq((*it)->name(),"session_parameter") ||
        Lib::eq((*it)->name(),"variable") ){
      std::string name;
      if(!(*it)->attributeValue("name",name)) {
        Lib::Out out(printer());
        out << "BaseUI::setSessionParameters :" 
            << " <session_parameter>/<variable> without name attribute." 
            << Lib::endl;
        return false;
      }
      // by default we look also for a \n separated list of items.
      bool look_for_list = true;
     {std::string s;
      if((*it)->attributeValue("list",s)) {
        if(!Lib::smanip::tobool(s,look_for_list)) {
          Lib::Out out(printer());
          out << "BaseUI::setSessionParameters :" 
              << " for <variable> attribuet list not a boolean." 
              << Lib::endl;
          return false;
        }
      }}
      if(look_for_list) {
        std::string s = (*it)->value();
        Lib::smanip::replace(s,"\\n","\n");
        fSession.setParameter(name,s);
      } else {
        fSession.setParameter(name,(*it)->value());
      }
    }
  }

  // children :
 {const std::list<Lib::ItemML*>& list = aItemML.children();
  std::list<Lib::ItemML*>::const_iterator it;
  for(it=list.begin();it!=list.end();++it) {
    if(!setSessionParameters(*(*it))) return false;
  }}

  return true;
}

//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::setAliases(
 Lib::ItemML& aItemML                      
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fScriptManager) return false;

  if(aItemML.tagName()=="aliases") {
    std::string sexec;
    if(!aItemML.attributeValue("","exec",sexec)){
      Lib::Out out(printer());
      out << "BaseUI::setAliases :" 
          << " <aliases> without a exec attribute." << Lib::endl;
      //aItemML.printXML(printer(),"  ");
      return false;
    }

    Slash::UI::IInterpreter* interpreter = 
      fScriptManager->findInterpreter(sexec);
    if(!interpreter) {
      Lib::Out out(fSession.printer());
      out << "OnX::BaseUI::setAliases :"
          << " Unknown interpreter : " << sexec << Lib::endl;
      return false;
    }

    const std::list<Lib::Element*>& properties = aItemML.properties();
    std::list<Lib::Element*>::const_iterator it;
    for(it=properties.begin();it!=properties.end();++it) {
      if((*it)->name()=="alias") {

        std::string name;
        if(!(*it)->attributeValue("name",name)) {
          Lib::Out out(fSession.printer());
          out << "OnX::BaseUI::setAliases :"
              << " <alias> without a name attribute." 
              << Lib::endl;
          return false;
        }

        std::string value = (*it)->value();

        if(!interpreter->addAlias(name,value)) return false;

        if(fVerboseLevel) {
          Lib::Out out(fSession.printer());
          out << "OnX::BaseUI::setAliases :"
              << " for interpreter " << Lib::sout(sexec)
              << " set alias " << Lib::sout(name)
              << " with value " << Lib::sout(value)
              << Lib::endl;
        }
   
      }
    }       
  }



  // children :
 {const std::list<Lib::ItemML*>& list = aItemML.children();
  std::list<Lib::ItemML*>::const_iterator it;
  for(it=list.begin();it!=list.end();++it) {
    if(!setAliases(*(*it))) return false;
  }}

  return true;

}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::solveInputDialog(
 Lib::ItemML& aItemML                      
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aItemML.tagName()=="widget") {
    std::string sclass;
    if(!aItemML.attributeValue("","class",sclass)){
      Lib::Out out(printer());
      out << "BaseUI::solveInputDialog :" 
          << " ItemML without a class." << Lib::endl;
      aItemML.printXML(printer(),"  ");
      return false;
    }
    if(sclass=="InputDialog") {
      if(!createInputDialog(aItemML)) return false;
    }
  }

  // Solve children :
 {const std::list<Lib::ItemML*>& list = aItemML.children();
  std::list<Lib::ItemML*>::const_iterator it;
  for(it=list.begin();it!=list.end();++it) {
    if(!solveInputDialog(*(*it))) return false;
  }}

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::solveCopy(
 Lib::ItemML*& aItemML                      
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string scopy;
  if(aItemML->attributeValue("","copy",scopy)) {
    if(!fTopItemML) {
      Lib::Out out(printer());
      out << "OnX::BaseUI::solveCopy :"
          << " topItemML not created." << Lib::endl;
      return false;
    }    
    std::string sat;
    if(!aItemML->attributeValue("","at",sat)) sat = "yes";
    if(sat=="yes") {
      std::string s;
      if(!atat(scopy,s)) {
        Lib::Out out(printer());
        out << "OnX::BaseUI::solveCopy :"
            << " atat on " << Lib::sout(scopy) << " failed."
            << Lib::endl;
        return false;
      }
      scopy = s;
    }
    Lib::ItemML* fromItemML = fTopItemML->findItem(scopy);
    if(!fromItemML) {
      Lib::Out out(printer());
      out << "OnX::BaseUI::solveCopy :"
          << " component " << Lib::sout(scopy) << " not found." 
          << Lib::endl;
      return false;
    }
    std::string sclass;
    if(!fromItemML->attributeValue("","class",sclass)) {
      Lib::Out out(printer());
      out << "OnX::BaseUI::solveCopy :"
          << " ItemML to copy without a class." << Lib::endl;
      return false;
    }

    Lib::ItemML* parent = aItemML->parent(); 
    if(!parent) {
      Lib::Out out(printer());
      out << "OnX::BaseUI::solveCopy :"
          << " ItemML without a parent." << Lib::endl;
      return false;
    }

    std::string name = buildName(*aItemML);

    Lib::ItemML* new_itemML = createItemML(name,sclass,0);
    if(!new_itemML) {
      Lib::Out out(printer());
      out << "OnX::BaseUI::solveCopy :"
          << " can't create an ItemML of class " << Lib::sout(sclass) << "." 
          << Lib::endl;
      return false;
    }

    if(!new_itemML->copy(*fromItemML)) return false;
    new_itemML->setAttributeValue("","name",name);

    parent->replaceChild(aItemML,new_itemML);
    new_itemML->setParent(parent);

    delete aItemML;
    aItemML = new_itemML;

    return true; //a copy must not have children.
  }

  // Solve children :
 {const std::list<Lib::ItemML*>& list = aItemML->children();
  std::list<Lib::ItemML*>::const_iterator it;
  for(it=list.begin();it!=list.end();++it) {
    Lib::ItemML* item = const_cast<Lib::ItemML*>(*it);
    if(!solveCopy(item)) return false;
  }}

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::notification(
 Slash::Core::IMessage* aMessage
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fVerboseLevel) {
    Lib::Out out(fSession.printer());
    out << "OnX::BaseUI::notification :"
        << " receive message " << aMessage->message() << Lib::endl;
  }
  if(OnX::ExecuteMessage* mess = Lib_Cast(*aMessage,OnX::ExecuteMessage)) {
    Action* action = mess->action();
    fCallbackData = action->callbackData();
    action->execute();
    fCallbackData = CallbackData();
    delete action;
  } else if(Lib::UpdateMessage* mess = 
              Lib_Cast(*aMessage,Lib::UpdateMessage)) {
    std::string s = mess->what();
    //::printf("debug : UpdateMessage : \"%s\"\n",s.c_str());
    if(s=="*") {
      executeCallbacks("*","update");
    } else if(fTopItemML) {
      Lib::ItemML* itemML = fTopItemML->findItem(s);
      if(!itemML) {
        Lib::Out out(fSession.printer());
        out << "OnX::BaseUI::notification :"
            << " UpdateMessage :"
            << " non existing component " << Lib::sout(s) << "."
            << Lib::endl;
      } else {
        executeCallbacks(*itemML,"update");
      }
    }
    Slash::Core::IRelation* relation = mess->relation();
    if(relation && fRelationManager && fRelationManager->isValid(*relation)) {
      //::printf("debug : UpdateMessage : relation unschedule\n");
      relation->unSchedule();
    }
  } else if(Lib::RelationMessage* mess = 
              Lib_Cast(*aMessage,Lib::RelationMessage)) {
    Slash::Core::IRelation& relation = mess->relation();
    if(fRelationManager && fRelationManager->isValid(relation)) {
      relation.unSchedule();
    }
  }
  delete aMessage;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::removeCallback(
 const std::string& aName
,const std::string& aWhich
,int aIndex
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fTopItemML) return false;
  Lib::ItemML* itemML = fTopItemML->findItem(aName);
  if(!itemML) return false;
  OnX::Action::NativeWidget widget = itemML->getData1();
  if(!widget) return false;
  Action* action = getAction(widget,aWhich,aIndex);
  if(!action) return false;
  deleteCallback(*itemML,aWhich,action);
  fActions.remove(action);
  delete action;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::setCallback(
 const std::string& aName
,const std::string& aWhich
,const std::string& aInterp
,const std::string& aScript
,int aIndex
)
//////////////////////////////////////////////////////////////////////////////
// If aIndex found override script.
// If not found add a new callback.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!fTopItemML) return false;
  Lib::ItemML* itemML = fTopItemML->findItem(aName);
  if(!itemML) return false;

  if(aWhich=="destroy") { 
    // particular case of destroy which is not handled on Actions.
    bool found = false;
   {const std::list<Lib::Element*> props = itemML->properties();
    int index = 0;
    std::list<Lib::Element*>::const_iterator it;
    for(it=props.begin();it!=props.end();++it) {
      if(aWhich==(*it)->name()) {
        if(aIndex==index) {
          (*it)->setAttributeValue("exec",aInterp);
          (*it)->setValue(aScript);
          found = true;
          break;
        }
        index++;
      }
    }}
    if(!found) { // new callback.
      std::vector<Lib::ItemML::Attribute> att;
      att.push_back(Lib::ItemML::Attribute("exec",aInterp));
      itemML->addProperty(aWhich,att,aScript);
    } else {
      itemML->setAttributeValue(aWhich,"exec",aInterp,aIndex);
      itemML->setPropertyValue(aWhich,aScript,aIndex);
    }
    return true;
  }

  OnX::Action::NativeWidget widget = itemML->getData1();
  if(!widget) return false;
  bool newCallback = false;
  Action* action = getAction(widget,aWhich,aIndex);
  if(!action) {
    newCallback = true;
    action = addCallback(*itemML,aWhich,aInterp,aScript);
  }
  if(!action) return false;
  action->setInterpreter(aInterp);
  action->setScript(aScript);
  // Update itemML :
  if (newCallback) {
    std::vector<Lib::ItemML::Attribute> att;
    att.push_back(Lib::ItemML::Attribute("exec",aInterp));
    itemML->addProperty(aWhich,att,aScript);
  } else {
    itemML->setAttributeValue(aWhich,"exec",aInterp,aIndex);
    itemML->setPropertyValue(aWhich,aScript,aIndex);
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::createIf(
 Lib::ItemML& aItemML                      
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string svalue;
  if(!aItemML.attributeValue("","if",svalue)) return true;
  std::vector<std::string> words = Lib::smanip::words(svalue,".");
  if((words.size()==2) && (words[0]=="env")) {
    // if="env.<env>"           return true if <env> exists.
    // if="env.<env>=<value>"   return true if <env>==<value>
    std::vector<std::string> words2 = Lib::smanip::words(words[1],"=");
    if(words2.size()==1) { //if <env> exist.
      std::string value;
      return Lib::System::getenv(words2[0],value);
    } else if(words2.size()==2) { //if <env> == <value>
      std::string value;
      if(!Lib::System::getenv(words2[0],value)) return false;
      return (value==words2[1]);
    }
  } else if((words.size()==2) && (words[0]=="exec")) {
    // if="exec.<interp>"  return true if <interp> is operational.
    std::string sexec = words[1];
    if(!fScriptManager) return false;
    Slash::UI::IInterpreter* interpreter = 
      fScriptManager->findInterpreter(sexec);
    if(!interpreter) return false;
    return interpreter->isOperational();
  } else {
    // if="<param>"           return true if <param> is a session parameter.
    // if="<param>=<value>"   return true if <param>==<value>
    std::vector<std::string> words = Lib::smanip::words(svalue,"=");
    if(words.size()==1) {
      std::string value;
      return fSession.parameterValue(words[0],value);
    } else if(words.size()==2) {
      std::string value;
      if(!fSession.parameterValue(words[0],value)) return false;
      return (words[1]==value);
    }
  }
  return false;
}    
//////////////////////////////////////////////////////////////////////////////
Lib::ItemML* OnX::BaseUI::findFirstWidget(
 Lib::ItemML& aItemML
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //FIXME : stop descending under comment ?
  if(aItemML.tagName()=="widget") return &aItemML;

  // Look children :
 {const std::list<Lib::ItemML*>& list = aItemML.children();
  std::list<Lib::ItemML*>::const_iterator it;
  for(it=list.begin();it!=list.end();++it) {
    Lib::ItemML* itemML = findFirstWidget(*(*it));
    if(itemML) return itemML;
  }}
  return 0;
}

void OnX::BaseUI::removeInRelations(Lib::ItemML& aItemML) {
  if(!fRelationManager) return;
  Slash::UI::IWidget* widget = Lib_Cast(aItemML,Slash::UI::IWidget);
  fRelationManager->removeWith(widget);
}

void OnX::BaseUI::openFile(const std::string& aFile){
  // Used by the "desktop open file" mechanism.
  fSession.setParameter("session.openFile",aFile);
  executeCallbacks("*","openFile");
}

bool OnX::BaseUI::is_in(const Lib::ItemML& aHead,Lib::ItemML* aItemML){
  if(aItemML==&aHead) return true;
  const std::list<Lib::ItemML*>& list = aHead.children();
  for(std::list<Lib::ItemML*>::const_iterator it=
      list.begin();it!=list.end();++it) {
    if(is_in(*(*it),aItemML)) return true; //found
  }
  return false;
}

static bool saveItemML(
 Lib::ItemML& aItemML
,const std::string& aFile
,int aVerboseLevel
,Slash::Core::IWriter& aPrinter
){
  if(aFile.empty()) return false;
  if(!aItemML.saveFlag()) return true;
  std::string sfile = Lib::fmanip::name(aFile);
  std::string name =  sfile + "_new";
  Lib::FileWriter fp(name,aPrinter);
  if(!fp.enabled()) {
    Lib::Out out(aPrinter);
    out << "OnX::BaseUI::save :" 
        << " can't open file " << name << Lib::endl;
    return false;
  }
  Lib::Out out(fp);
  if(aVerboseLevel) {
    Lib::Out out(aPrinter);
    out << "OnX::BaseUI::save :" 
        << " save file " << Lib::sout(sfile) << "..." << Lib::endl;
  }
  bool status = saveItemML(aItemML,out,0,aVerboseLevel,aPrinter);
  fp.disable();
  if(!status) {
    Lib::fmanip::remove(name);
    if(aVerboseLevel) {
      Lib::Out out(aPrinter);
      out << "OnX::BaseUI::save :" 
          << " unable to save file " << Lib::sout(sfile) << "..." << Lib::endl;
    }
  } else {
    // Replace file :
    Lib::fmanip::move(name,sfile);
  }
  return status;
}

static bool saveItemML(
 Lib::ItemML& aItemML
,Lib::Out& aOut
,int aShift
,int aVerboseLevel
,Slash::Core::IWriter& aPrinter
){
  if(!aItemML.saveFlag()) return true;
  
  std::string spaces;
  for(int i=0;i<aShift;i++) spaces += " ";

  const std::string& tagName = aItemML.tagName();

  std::string sname;
  aItemML.attributeValue("","name",sname);

  const std::list<Lib::ItemML*>& childs = aItemML.children();

  if(tagName=="reference") {
    std::string sfile;
    aItemML.attributeValue("","file",sfile);
    aOut << spaces 
         << "<reference"
         << " file="  << Lib::sout(sfile) << "/>" 
         << Lib::endl;
    
    // Save GUI associated to the reference :
    // Element should have one child :
    if(childs.size()!=1) return false;
    return saveItemML(*(childs.front()),sfile,aVerboseLevel,aPrinter);
  }

  if (aItemML.file() != "") {
    // If file field is present, we have to take care that file
    // is not already open by a parent.
    // (case of reference file (and others ?))
    Lib::ItemML* parentItemML = aItemML.parent();
    if (parentItemML) {
      std::string pfile;
      parentItemML->attributeValue("","file",pfile);
      if( (parentItemML->tagName() == "reference") && 
          (pfile == aItemML.file()) ) {
      } else if (aItemML.file() == parentItemML->file()) {
      } else { 
        std::string file = aItemML.file();
        aItemML.setFile(""); //to avoid recursivity.
        bool status = saveItemML(aItemML,file,aVerboseLevel,aPrinter);
        aItemML.setFile(file);
        return status;
      }
    }
  }
  
  if(tagName=="OnX") { 
    // Is in principle at shift=0. 
    aOut << spaces << "<OnX version=" << Lib::sout(ONX_VERSION) ;
  } else if(tagName=="aliases") {
    aOut << spaces << "<aliases";
  } else if(tagName=="comment") {
    aOut << spaces << "<comment";

    const std::vector<Lib::ItemML::Attribute>& atbs = aItemML.attributes();
    unsigned int atbn = atbs.size();
    for(unsigned int index=0;index<atbn;index++) {
      const std::string& atb = atbs[index].first;
      const std::string& atbv = atbs[index].second;
      if( (atb=="name") && isGeneratedName(atbv)) {
        // Do not save generated names.
      } else if(atb!="class") {
        aOut << " " << atb << "=" << Lib::sout(atbv);
      }
    }

  } else { //FIXME : check that tagName is "widget".
    aOut << spaces << "<widget";

    const std::vector<Lib::ItemML::Attribute>& atbs = aItemML.attributes();
    unsigned int atbn = atbs.size();
    for(unsigned int index=0;index<atbn;index++) {
      const std::string& atb = atbs[index].first;
      const std::string& atbv = atbs[index].second;
      if( (atb=="name") && (atbv.substr(0,14)=="OnX_Component_")) {
        // Do not save generated names.
      } else {
        aOut << " " << atb << "=" << Lib::sout(atbv);
      }
    }

  }

  const std::list<Lib::Element*>& props = aItemML.properties();

  if(!props.size() && !childs.size() ) {

    aOut << "/>" << Lib::endl;

  } else {

    aOut << ">" << Lib::endl;

    std::list<Lib::Element*>::const_iterator itp;
    for(itp=props.begin();itp!=props.end();++itp) {
      const std::vector<Lib::ItemML::Attribute>& atbs = (*itp)->attributes();

      bool isCallback = false;
      aOut << spaces << "  <" << (*itp)->name();
      unsigned int atbn = atbs.size();
      for(unsigned int index=0;index<atbn;index++) {
        const std::string& atb = atbs[index].first;
        const std::string& atbv = atbs[index].second;
        aOut << " " << atb << "=" << Lib::sout(atbv);
        if(atb=="exec") isCallback = true;
      }
      if((*itp)->value().empty()) {
        aOut << "/>" << Lib::endl;
      } else {
        aOut << ">";
        std::string value = Lib::smanip::toxml((*itp)->value());
        if(isCallback) {
          if(value.find("\\n\\")==std::string::npos) {
            aOut << value;
          } else {
            aOut << Lib::endl;
            Lib::smanip::replace(value,"\\n\\","@OnX@");
            Lib::smanip::replace(value,"@OnX@","\\n\\\n");
            Lib::smanip::strip(value,Lib::smanip::trailing,' ');
            aOut << value;
            unsigned int l = value.size();
            if(l && value[l-1]!='\n') aOut << Lib::endl;
            aOut << spaces << "  ";
          }
        } else {
          aOut << value;
        }
        aOut << "</" << (*itp)->name() << ">" << Lib::endl;
      }
    }

    if(tagName!="InputDialog") {
      //Children :
      std::list<Lib::ItemML*>::const_iterator it;
      for(it=childs.begin();it!=childs.end();++it) {
        if(!saveItemML(*(*it),aOut,aShift+2,aVerboseLevel,aPrinter)) 
          return false;
      }
    }  
  
    aOut << spaces << "</" << tagName << ">" << Lib::endl;

  }

  return true;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

Slash::UI::ICyclic* OnX::BaseUI::findCyclic(const std::string& aName) const {
  std::list<Slash::UI::ICyclic*>::const_iterator it;
  for(it=fCyclics.begin();it!=fCyclics.end();++it) {
    if((*it)->name()==aName) return *it;
  }
  return 0;
}

bool OnX::BaseUI::removeCyclic(Slash::UI::ICyclic*& aCyclic){
  fCyclics.remove(aCyclic);
  delete aCyclic;
  aCyclic = 0;
  return true;
}

void OnX::BaseUI::removeCyclics(){
  std::list<Slash::UI::ICyclic*>::iterator it;
  for(it=fCyclics.begin();it!=fCyclics.end();it = fCyclics.erase(it)) 
    delete (*it);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void OnX::BaseUI::steerModal(){while(steer()!=ONX_BREAK_MODAL) {}}

std::string OnX::BaseUI::ask(const std::string& aPrompt){
  if(!fTopItemML) return "";
  Lib::ItemML* parent = fTopItemML->findByAttribute("class","Window");
  if(!parent) {
    Lib::Out out(printer());
    out << "OnX::BaseUI::ask :"
        << " can't find a Window widget to attach the AskDialog."
        << Lib::endl;
    return "";
  }
  Lib::ItemML* top;
  if(!load(parent,"$ONXROOT/scripts/OnX/AskDialog.onx",true,true,true,top)){
    Lib::Out out(printer());
    out << "OnX::BaseUI::ask :"
        << " can't load the AskDialog."
        << Lib::endl;
    return "";
  }
  setParameter("OnX_AskDialog_prompt.label",aPrompt);
  setParameter("OnX_AskDialog_entry.value","");
  Slash::UI::IWidget* w = findWidget("OnX_AskDialog_dialog");
  if(w) w->show();
  steerModal();
  //echo("debug : exit steer loop.");
  std::string value;
  parameterValue("OnX_AskDialog_entry.value",value);
  if(fVerboseLevel) {
    Lib::Out out(fSession.printer());
    out << "OnX::BaseUI::askString :"
        << value
        << Lib::endl;
  }
  return value;
}

//////////////////////////////////////////////////////////////////////////////
// Laurent ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
std::vector< std::pair<bool,std::string> > OnX::BaseUI::parseDecorations(
 const std::string& aDecoration
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector< std::pair<bool,std::string> > decos;
  std::vector<std::string> items;
  Lib::smanip::lines(aDecoration,items); //FIXME
  unsigned int itemn = items.size();

  //look the first, if it is not all or none, we go to the next
  bool close_button = true;
  bool iconify_button = true;
  bool min_max_button = true;
  bool resize_handle = true;
  bool border = true;
  bool caption = true;
  bool system_menu = true;
  if (itemn >0) {
    if (items[0]=="all"){ // do nothing 
      close_button = true;
    } else if (items[0]=="none"){ // remove all
      close_button = false;
      iconify_button = false;
      min_max_button = false;
      resize_handle = false;
      border = false;
      caption = false;           
      system_menu = false;
    }

    // To have a icon, close, or minimize-maximize button, you should
    // have a caption
    // If you have a system menu, you will also have a close button

    Lib::Out out(printer());
    for(unsigned int index=0;index<itemn;index++){
      if (items[index]=="no_close_button")        close_button = false;
      else if (items[index]=="close_button")      close_button = true;
      else if (items[index]=="no_iconify_button") iconify_button = false;
      else if (items[index]=="iconify_button")    iconify_button = true;
      else if (items[index]=="no_system_menu")    system_menu = false;
      else if (items[index]=="system_menu")       system_menu = true;
      else if (items[index]=="no_min_max_button") min_max_button = false;
      else if (items[index]=="min_max_button")    min_max_button = true;
      else if (items[index]=="no_resize_handle")  resize_handle = false;
      else if (items[index]=="resize_handle")     resize_handle = true;
      else if (items[index]=="no_border")         border = false;
      else if (items[index]=="border")            border = true;
      else if (items[index]=="no_caption")        caption = false;
      else if (items[index]=="caption")           caption = true;
      else if (items[index]=="all") {}
      else if (items[index]=="none") {}
      else if (items[index].size()) 
        out << "OnX::BaseUI::parseDecorations :"
            << " WARNING : The decoration " << Lib::sout(items[index]) 
            << " is not recognised." 
            << " Legal options are "
            << " none"
            << " [no_]close_button"
            << " [no_]iconify_button" 
            << " [no_]system_menu" 
            << " [no_]min_max_button [no_]resize_handle [no_]border"
            << " [no_]caption all none." 
            << Lib::endl; 
    }
    if (!caption && 
        ((close_button )|| (iconify_button) || 
         (min_max_button) || (system_menu)))
      out << "OnX::BaseUI::parseDecorations :" 
          << " WARNING : You want to create a widget whithout caption, " 
          << "but you should have a caption to have a"
          << " close,iconify,system menu " 
          << "or min max button" 
          << Lib::endl;
  }
  decos.push_back(std::pair<bool,std::string>(close_button,"close_button"));
  decos.push_back(std::pair<bool,std::string>(iconify_button,"iconify_button"));
  decos.push_back(std::pair<bool,std::string>(min_max_button,"min_max_button"));
  decos.push_back(std::pair<bool,std::string>(resize_handle,"resize_handle"));
  decos.push_back(std::pair<bool,std::string>(border,"border"));
  decos.push_back(std::pair<bool,std::string>(caption,"caption"));
  decos.push_back(std::pair<bool,std::string>(system_menu,"system_menu"));
  return decos;
}
//////////////////////////////////////////////////////////////////////////////
// current widget ////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Slash::UI::IWidget* OnX::BaseUI::currentWidget(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fCurrentWidget;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::setCurrentWidget(
 Slash::UI::IWidget* aWidget
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fVerboseLevel) {
    Lib::Out out(fSession.printer());
    if(aWidget) {
      out << "OnX::BaseUI::setCurrentWidget :"
          << " " << Lib::sout(aWidget->name())
          << Lib::endl;
    } else {
      out << "OnX::BaseUI::setCurrentWidget :"
          << " to null."
          << Lib::endl;
    }
  }
  fCurrentWidget = aWidget;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
// terminal //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::enableTerminal(
 const std::string& //aPrompt
,const std::string& //aInterp
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::disableTerminal(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::redirectAskToTerminal(
 bool aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  fAskToTerminal = aValue;
  return true;
} 
//////////////////////////////////////////////////////////////////////////////
bool OnX::BaseUI::lockScreen(
 const std::string&
,const std::string&
,const std::string&
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return false;
} 

#include "InputDialog.cxx"
#include "IB.cxx"

bool treat_atat(Slash::UI::IUI&,const std::string&,std::string&);

bool OnX::BaseUI::atat(const std::string& aString,std::string& aValue) const{
  Slash::UI::IUI* ui = (Slash::UI::IUI*)this;
  return treat_atat(*ui,aString,aValue);
}
