// this :
#include <OnX/Core/Interpreter.h>

#include <Slash/Core/ISession.h>
#include <Slash/Core/IWriter.h>
#include <Slash/Tools/Managers.h>

#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/System.h>

//////////////////////////////////////////////////////////////////////////////
OnX::Interpreter::Interpreter(
 Slash::Core::ISession& aSession
,const std::string& aName
,const std::string& aLibrary
,const std::string& aPackage
,const std::string& aSuffix
,const std::string& aExecFileCmd
,const std::string& aMendatoryEnvironmentVariable
,Slash::UI::IInterpreter::Function aFunction
,void* aTag
,Slash::UI::IInterpreter::Completer aCompleter
,void* aCompleterTag
)
:fName(aName)
,fLibrary(aLibrary)
,fPackage(aPackage)
,fSuffix(aSuffix)
,fExecFileCmd(aExecFileCmd)
,fMendatoryEnvironmentVariable(aMendatoryEnvironmentVariable)
,fLoaded(false)
,fFunction(aFunction)
,fTag(aTag)
,fCompleter(aCompleter)
,fCompleterTag(aCompleterTag)
,fSession(aSession)
,fOperationalChecked(false)
,fOperationalStatus(false)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
OnX::Interpreter::~Interpreter(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
std::string OnX::Interpreter::name(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fName;
}
//////////////////////////////////////////////////////////////////////////////
std::string OnX::Interpreter::package(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fPackage;
}
//////////////////////////////////////////////////////////////////////////////
std::string OnX::Interpreter::suffix(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fSuffix;
}
//////////////////////////////////////////////////////////////////////////////
std::string OnX::Interpreter::execFileCommand(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fExecFileCmd;
}
//////////////////////////////////////////////////////////////////////////////
void OnX::Interpreter::setFunction(
 Slash::UI::IInterpreter::Function aFunction
,void* aTag
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aFunction) return;
  fFunction = aFunction;
  fTag = aTag;
}
//////////////////////////////////////////////////////////////////////////////
Slash::UI::IInterpreter::Function OnX::Interpreter::function(
 void*& aTag
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aTag = fTag;
  return fFunction;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::Interpreter::load(
 bool aQuiet
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fLoaded) return true;
  if(fLibrary.size()) {
    Slash::Core::ILibraryManager* lm = Slash::libraryManager(fSession);
    if(!lm) return false;
    if(!lm->addLibrary(fName,fLibrary,fPackage,aQuiet)) return false;
  }
  fLoaded = true;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::Interpreter::executeScript(
 const std::string& aScript
,const Slash::UI::IInterpreter::Options& aParams
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aScript.empty()) return true;
  if(!load()) {
    Lib::Out out(fSession.printer());
    out << "OnX::Interpreter::executeScript :"
        << " interpreter " << Lib::sout(fName) 
        << " loading failed." << Lib::endl;
    out << aScript << Lib::endl;
    return false;
  } else {
    if(fFunction) {
      if(fSession.verboseLevel()) {
        Lib::Out out(fSession.printer());
        out << "OnX::Interpreter::executeScript :"
            << " with interpreter " << Lib::sout(fName) 
            << " : " << Lib::endl;
        out << aScript << Lib::endl;
      }
      return fFunction(aScript,aParams,fAliases,fTag);
    } else {
      Lib::Out out(fSession.printer());
      out << "OnX::Interpreter::executeScript :"
          << " Interpreter (" << fName << ") " 
          << "native function not found for script :" << Lib::endl;
      out << aScript << Lib::endl;
      return false;
    }
  }

  //FIXME fSession.flush(); do it here ?

  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::Interpreter::isOperational(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(fOperationalChecked) return fOperationalStatus;

  fOperationalChecked = true;

  fOperationalStatus = false;
  if(fMendatoryEnvironmentVariable.size()) {
    bool status = Lib::System::isenv(fMendatoryEnvironmentVariable);
    if(fSession.verboseLevel()) {
      Lib::Out out(fSession.printer());
      out << "OnX::Interpreter::isOperational :"
          << " mendatoryEnvironmentVariable " 
          << Lib::sout(fMendatoryEnvironmentVariable) 
          << " defined : " << status
          << Lib::endl;
    }
    if(!status) return false;
  }
  if(!load(true)) return false;
  if(!fFunction) return false;
  fOperationalStatus = true;
  return true;
}
bool OnX::Interpreter::addAlias(const std::string& aAlias,const std::string& aValue){
  //FIXME : should check that it is already here :
  fAliases.push_back(std::pair<std::string,std::string>(aAlias,aValue));
  return true;
}
void OnX::Interpreter::setCompleter(Completer aFunc,void* aTag) {
 fCompleter = aFunc;
 fCompleterTag = aTag;
}
Slash::UI::IInterpreter::Completer OnX::Interpreter::completer(
 void*& aTag) const {aTag = fCompleterTag; return fCompleter;}
bool OnX::Interpreter::complete(const std::string& aString,
std::string& aValue) {
  if(!fCompleter) {
    Lib::Out out(fSession.printer());
    out << "OnX::Interpreter::complete :"
        << " interpreter " << Lib::sout(fName) 
        << " has no completion." 
        << Lib::endl;
    return false;
  }
  return fCompleter(aString,aValue,fCompleterTag);
}
