// this :
#include <OnX/Core/Session.h>

#include <Slash/Core/ISession.h>
#include <Slash/Core/IManager.h>
#include <Slash/Core/IWriter.h>
#include <Slash/Core/ILibrary.h>
#include <Slash/Tools/Managers.h>

#include <Slash/UI/IUI.h>
#include <Slash/UI/ICallbackData.h>
#include <Slash/Tools/UI.h>

#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/DLL.h>
#include <Lib/smanip.h>
#include <Lib/fmanip.h>
#include <Lib/dirmanip.h>
#include <Lib/StreamBuffer.h>
#include <Lib/LibraryManager.h>
#include <Lib/RelationManager.h>
#include <Lib/AccessorManager.h>
#include <Lib/System.h>
#include <Lib/ItemML.h>
#include <Lib/Element.h>
#include <Lib/BaseML.h>
#include <Lib/Arguments.h>
#include <Lib/Cast.h>

#include <OnX/Core/ScriptManager.h>
#include <OnX/Core/StyleManager.h>
#include <OnX/Core/UI_Manager.h>
#include <OnX/Version.h>

// Embedded code :
#include <OnX/Core/Session_onx.h>

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

static bool system_Function(const std::string&,const Slash::UI::IInterpreter::Options&,const Slash::UI::IInterpreter::Aliases&,void*);
static bool sh_Function(const std::string&,const Slash::UI::IInterpreter::Options&,const Slash::UI::IInterpreter::Aliases&,void*);
static bool DOS_Function(const std::string&,const Slash::UI::IInterpreter::Options&,const Slash::UI::IInterpreter::Aliases&,void*);
static bool session_Function(const std::string&,const Slash::UI::IInterpreter::Options&,const Slash::UI::IInterpreter::Aliases&,void*);
static bool dld_Function(const std::string&,const Slash::UI::IInterpreter::Options&,const Slash::UI::IInterpreter::Aliases&,void*);

OnX::Session::Session(bool aBalance,bool aDebugCheckClass)
: Lib::Session(*((Slash::Core::IWriter*)this),aBalance,aDebugCheckClass)
,fOutOld(0)
,fOutNew(0)
,fErrOld(0)
,fErrNew(0)
,fOutFile(0)
,fErrFile(0)
{
  OnX::Printer::setSession(this);
}

OnX::Session::~Session(){
  stdoutToDefault();
  OnX::Printer::disable();
  coutToDefault();
}

void* OnX::Session::cast(const std::string& aClass) const {
  // inheritance :
  if_Lib_SCast(OnX::Printer)
  else if(void* p = Lib::Session::cast(aClass)) return p;
  // this :
  else if_Lib_SCast(OnX::Session)
  else return 0;
}

//private:
OnX::Session::Session(const Session& aFrom)
: Lib::Session(aFrom)
,fOutOld(0)
,fOutNew(0)
,fErrOld(0)
,fErrNew(0)
,fOutFile(0)
,fErrFile(0)
{}
//private:
OnX::Session& OnX::Session::operator=(const Session&){return *this;}

bool OnX::Session::setDefaults(const std::string& aFile){
  setParameter("session.output","UI");
  setParameter("AccessorManager.dumpMode","table");
  setParameter("session.helpOnDialog","yes");
  // Used in the OnX/NextDialog dialog :
  //setParameter("ui.exitSecondaryLoop","no");

  // Session XML file :
  Lib::ItemML* top = 0;
  Lib::ItemMLFactory factory;
  Lib::BaseML sessionML(factory);
  std::vector<std::string> tags;
  tags.push_back("session");
  tags.push_back("manager");
  tags.push_back("interpreter");
  sessionML.setTags(tags);
  if(aFile.size()) {
    sessionML.loadFile(aFile,false,*this);
    top = sessionML.topItem();
  } else {
    std::string s = Lib::smanip::tostring(Session_onx_linen,
                                          Session_onx_lines);
    Lib::smanip::replace(s,"@@double_quote@@","\"");
    Lib::smanip::replace(s,"@@back_slash@@","\\");
    sessionML.loadString(s,*this);
    top = sessionML.topItem();
  }
  std::string svalue;
  int ivalue;
  if(top) {

   {const std::list<Lib::Element*>& properties = top->properties();
    std::list<Lib::Element*>::const_iterator it;
    for(it=properties.begin();it!=properties.end();++it) {
      std::string name = (*it)->name();
      std::string value = (*it)->value();
      //printf("debug : \"%s\" \"%s\"\n",name.c_str(),value.c_str());
      if(name=="verbose") {
        if(Lib::smanip::toint(value,ivalue)) fVerboseLevel = ivalue;
      } else if(name=="output") {
        setParameter("session.output",value);
      } else {
        setParameter(name,value);
      }
    }}

    // Set some manager attribute (verbose) in the session parameters :
    // <manager>.<attribute>:<value>
   {const std::list<Lib::ItemML*> itemMLs = top->children();
    std::list<Lib::ItemML*>::const_iterator it;
    for(it=itemMLs.begin();it!=itemMLs.end();++it) {
      if((*it)->tagName()=="manager") {
        std::string sname;
        if((*it)->attributeValue("","name",sname)) {
          std::string param = sname + ".verbose";
          if((*it)->propertyValue("verbose",svalue) && 
             Lib::smanip::toint(svalue,ivalue) ) {
            setParameter(param,Lib::smanip::tostring(ivalue));
          } else {
            setParameter(param,Lib::smanip::tostring(fVerboseLevel));
          }
        }
      }
    }}

  }

  // Should be the first.
  if(!findManager("LibraryManager")) {
    Lib::LibraryManager* libraryManager = new Lib::LibraryManager(*this);
    addManager(libraryManager);
  } else {
    //already inited.
  }

  if(!findManager("AccessorManager")) {
    Lib::AccessorManager* accessorManager = 
      new Lib::AccessorManager(*this);
    addManager(accessorManager);

    Lib::ItemML* item = 0;
    if(top && (item = top->findItem("AccessorManager")) ) {
      if(item->propertyValue("verbose",svalue)) {
        if(Lib::smanip::toint(svalue,ivalue))
          accessorManager->setVerboseLevel(ivalue);
      }
    } else {
      accessorManager->setVerboseLevel(fVerboseLevel);
    }

    accessorManager->addAction(new Lib::CollectAction(*accessorManager));
    accessorManager->addAction(new Lib::FilterAction(*accessorManager));
    accessorManager->addAction(new Lib::DumpAction(*accessorManager));
    accessorManager->addAction(new Lib::NumberAction(*accessorManager));
    accessorManager->addAction(new Lib::GetValuesAction(*accessorManager));
    accessorManager->addAction(new Lib::DestroyAction(*accessorManager));
    accessorManager->addAction(new Lib::SetAction(*accessorManager));
    accessorManager->addAction(new Lib::CalculateAction(*accessorManager));
    accessorManager->addAction(new Lib::InitAction(*accessorManager));
    accessorManager->addAction(new Lib::VisualizeAction(*accessorManager));
  }

  if(!findManager("RelationManager")) {
    addManager(new Lib::RelationManager(*this));
  }

  if(!findManager("ScriptManager")) {
    OnX::ScriptManager* scriptManager = new OnX::ScriptManager(*this);
    addManager(scriptManager);

    // Default interpreters :
#ifdef WIN32
    scriptManager->addInterpreter("system","","","bat","call @onx_file@","",
                                  system_Function,0);
#else
    scriptManager->addInterpreter("system","","","","chmod a+x @onx_file@;./@onx_file@","",
                                  system_Function,0);
#endif
    scriptManager->addInterpreter("sh","","","sh","chmod a+x @onx_file@;./@onx_file@","",
                                  sh_Function,0);
    scriptManager->addInterpreter("DOS","","","bat","call @onx_file@","",
                                  DOS_Function,0);

    scriptManager->addInterpreter("Session","","","session","OnX session_exec_file","",session_Function,(Slash::Core::ISession*)this);

    scriptManager->addInterpreter("DLD","","","dld","OnX dld_exec_file @onx_file@","",dld_Function,(Slash::Core::ISession*)this);

    Lib::ItemML* item = 0;
    if(top && (item = top->findItem("ScriptManager")) ) {
      if(item->propertyValue("recordExecutedScripts",svalue)) {
        bool bvalue;
        if(Lib::smanip::tobool(svalue,bvalue))
          scriptManager->recordExecutedScripts(bvalue);
      }
      const std::list<Lib::ItemML*> itemMLs = item->children();
      std::list<Lib::ItemML*>::const_iterator it;
      for(it=itemMLs.begin();it!=itemMLs.end();++it) {
        if((*it)->tagName()!="interpreter") continue;

        std::string sname;
        if((*it)->attributeValue("","name",sname)) {
          std::string slibrary;
          if((*it)->propertyValue("library",slibrary)) {
            std::string spackage;
            if((*it)->propertyValue("package",spackage)) {
              std::string ssuffix;
              if((*it)->propertyValue("suffix",ssuffix)) {
                std::string scmd;
                if((*it)->propertyValue("execFileCommand",scmd)) {
                  std::string senv;
                  (*it)->propertyValue("mendatoryEnvironmentVariable",senv);

                  if(fVerboseLevel) {
                    Lib::Out out(*this);
                    out << "OnX::Session::Session :"
                        << " from file " << Lib::sout(aFile)
                        << " addInterpreter :" 
                        << " name " << Lib::sout(sname)
                        << " library " << Lib::sout(slibrary)
                        << " package " << Lib::sout(spackage)
                        << " suffix " << Lib::sout(ssuffix)
                        << " execFileCommand " << Lib::sout(scmd)
                        << " mendatoryEnvironmentVariable " << Lib::sout(senv)
                        << Lib::endl;
                  }
                  scriptManager->addInterpreter(sname,slibrary,spackage,
                                                ssuffix,scmd,senv,0,0);
                }
              }
            }
          }
        }

      }
    }
  }

  if(!findManager("StyleManager")) {
    addManager(new OnX::StyleManager(*this));
  }

  
  if(!findManager("UI_Manager")) {
    OnX::UI_Manager* uiManager = new OnX::UI_Manager(*this);
    addManager(uiManager);
    //FIXME : put the below params on the UI_Manager.
    Lib::ItemML* item = 0;
    if(top && (item = top->findItem("UI_Manager")) ) {
      if(item->propertyValue("toolkit",svalue) && (svalue.size())) {
        setParameter("UI_Manager.toolkit",svalue);
      }
      if(item->propertyValue("threaded",svalue) && (svalue.size())) {
        setParameter("UI_Manager.thread",svalue);
      }
      if(item->propertyValue("style",svalue) && (svalue.size())) {
        setParameter("UI_Manager.style",svalue);
      }
    }
  }

  return true;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// overloaded :

bool OnX::Session::load(const std::string& aFile,const std::string& aPackage){
  Slash::Core::ILibraryManager* libraryManager = Slash::libraryManager(*this);
  if(!libraryManager) return false;
  return libraryManager->addLibrary(aPackage,aFile,aPackage) ? true : false;
}

#include <stdio.h>
void OnX::Session::redirectOutput(bool aYesNo){
  if(aYesNo) {
    coutToOnX();    // cout to fPrinter.
    stdoutToOnX();  // stdout to fPrinter.
    OnX::Printer::enable(); // Slash::Core::IWriter to Slash::UI::IUI.
  } else {
    stdoutToDefault(); // stdout to terminal.
    coutToDefault();   // cout to terminal.
    OnX::Printer::disable();   // Slash::Core::IWriter to stdout.
  }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

// Should be after any inclusion of string :
#ifdef HasAnsiStream
#include <iostream>
#else
#include <iostream.h>
#endif

// To redirect stdout :
#include <stdio.h>
#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#endif

void OnX::Session::coutToOnX(){
//#define NO_COUT
#if !defined(NO_COUT) && !defined(__hpux)
  if(!fOutNew) {
    fOutNew = new Lib::StreamBuffer(*this);
#ifdef HasAnsiStream
    fOutOld = std::cout.rdbuf(fOutNew);
#else
    cout = (streambuf*)fOutNew;
#endif
  }
  if(!fErrNew) {
    fErrNew = new Lib::StreamBuffer(*this);
#ifdef HasAnsiStream
    fErrOld = std::cerr.rdbuf(fErrNew);
#else
    cerr = (streambuf*)fErrNew;
#endif
  }
#endif
}

void OnX::Session::coutToDefault(){
  //  On HP the cout = oldStreamBuffer induces 
  // a crash at return of the main() of oxm program.
#if !defined(NO_COUT) && !defined(__hpux)
  if(fOutNew) {
#ifdef HasAnsiStream
    std::cout.rdbuf((std::streambuf*)fOutOld);
#else
    cout = (streambuf*)fOutOld;
#endif
    delete fOutNew;
    fOutNew = 0;
  }
  if(fErrNew) {
#ifdef HasAnsiStream
    std::cerr.rdbuf((std::streambuf*)fErrOld);
#else
    cout = (streambuf*)fErrOld;
#endif
    delete fErrNew;
    fErrNew = 0;
  }
#endif
}

void OnX::Session::stdoutToOnX(){
  // Avoid to redirect stdout, stderr.
  // On Darwin and Windows we loose the Python output !
  //return;

  //if ((::dup2(::fileno(stdout), ::fileno(stderr))) >= 0) {
  //} 
  /*FIXME :
    Linux : if starting with the mouse, we loose output 
     (redirected to /dev/null ???)
  */

  if(!fOutFile) {
    if(Lib::fmanip::tmpname(fOutName,".","OnX_",".out")) {
#ifdef WIN32
      fOutDefault = "CONOUT$";
#else
      char* stdoutName = ttyname(STDOUT_FILENO);
      // Could be NULL if stdout had been redirected on the command line.
      fOutDefault = stdoutName ? std::string(stdoutName) : std::string("");
#endif
      if(fOutDefault.size()) {
        if (::freopen(fOutName.c_str(), "w", stdout)) {
          fOutFile = ::fopen(fOutName.c_str(), "r");
        } else {
          Lib::fmanip::remove(fOutName);
        }
      }
    }
  }
  if(!fErrFile) {
    if(Lib::fmanip::tmpname(fErrName,".","OnX_",".err")) {
#ifdef WIN32
      fErrDefault = "CONOUT$";
#else
      char* stderrName = ttyname(STDERR_FILENO);
      // Could be NULL if stderr had been redirected on the command line.
      fErrDefault = stderrName ? std::string(stderrName) : std::string("");
#endif
      if(fErrDefault.size()) {
        if (::freopen(fErrName.c_str(), "w", stderr)) {
          fErrFile = ::fopen(fErrName.c_str(), "r");
        } else {
          Lib::fmanip::remove(fErrName);
        }
      }
    }
  }
}

void OnX::Session::stdoutToDefault(){
  if(fOutFile) {
    if (::freopen(fOutDefault.c_str(), "w", stdout)) {}
    ::fclose((FILE*)fOutFile);
    fOutFile = 0;
    Lib::fmanip::remove(fOutName);
  }
  if(fErrFile) {
    if (::freopen(fErrDefault.c_str(), "w", stderr)) {}
    ::fclose((FILE*)fErrFile);
    fErrFile = 0;
    Lib::fmanip::remove(fErrName);
  }
}
#define READBUFSIZE 32768  //16384  //65536;
//#define READBUFSIZE 16

bool OnX::Session::flush(){
  char buf[READBUFSIZE];
  int len;
  if(fOutFile) {
    ::fflush(stdout);
    do {
      len = ::fread(buf,1,READBUFSIZE-1,(FILE*)fOutFile);
      if(len<=0) break;
      buf[len] = '\0';
      OnX::Printer::write(buf);
    } while (len > 0);
  }
  if(fErrFile) {
    ::fflush(stderr);
    do {
      len = ::fread(buf,1,READBUFSIZE-1,(FILE*)fErrFile);
      if(len<=0) break;
      buf[len] = '\0';
      OnX::Printer::write(buf);
    } while (len > 0);
  }
  return true;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool system_Function(
 const std::string& aString
,const Slash::UI::IInterpreter::Options&
,const Slash::UI::IInterpreter::Aliases&
,void*
){
  Lib::System::execute(aString);
  return true;  
}

bool sh_Function(
 const std::string& aString
,const Slash::UI::IInterpreter::Options&
,const Slash::UI::IInterpreter::Aliases&
,void*
){
  if(Lib::System::isKindOf("UNIX")) Lib::System::execute(aString);
  return true;  
}

bool DOS_Function(
 const std::string& aString
,const Slash::UI::IInterpreter::Options&
,const Slash::UI::IInterpreter::Aliases&
,void*
){
  if(Lib::System::isKindOf("Windows")) Lib::System::execute(aString);
  return true;  
}

bool dld_Function(
 const std::string& aString
,const Slash::UI::IInterpreter::Options& aOptions
,const Slash::UI::IInterpreter::Aliases& aAliases
,void* aTag
){
  Slash::Core::ISession* session = (Slash::Core::ISession*)aTag;
  Lib::Out out(session->printer());

  if(session->verboseLevel()) {
    out << "OnX::Session::dld_Function :"
        << " execute " << Lib::sout(aString) << "..."
        << Lib::endl;
  }

  Slash::UI::IUI* ui = Slash::find_UI(*session);
  if(!ui) {
    //if(session->verboseLevel()) {
      out << "OnX::Session::dld_Function :"
          << " Slash::UI::IUI not found." 
          << " Can't execute script " << Lib::sout(aString) << "."
          << Lib::endl;
    //}
    return false;
  }

  std::vector<std::string> words;
  Lib::smanip::words(aString," ",words);
  if(words.size()<2) {
    out << "OnX::Session::dld_Function :"
        << " bad syntax in " << Lib::sout(aString) 
        << Lib::endl;
    return false;
  }

  std::string sdll;
  std::string scbk;
  sdll = words[0];
  scbk = words[1];
  //  if(!treat_atat(*ui,words[0],sdll)) return false;
  //  if(!treat_atat(*ui,words[1],scbk)) return false;
  //  Lib::smanip::strip(sdll);
  //  Lib::smanip::strip(scbk);

  std::string at("yes"); //default
  std::string alias("yes"); //default
  if(aOptions.size()) {
    Lib::Arguments opts(aOptions);
    if(!opts.find("at",at)) at = "yes";
    if(!opts.find("alias",alias)) alias = "yes";
  }

  std::vector<std::string> args;
  if(at=="no") {
    for(unsigned int index=2;index<words.size();index++) {
      args.push_back(words[index]);
    }
  } else if(at=="yes") {
    for(unsigned int index=2;index<words.size();index++) {
      bool b_cat_empty = (words[index]=="@cat@");
      bool b_cat_space = (words[index]=="@cat_space@");
      if(b_cat_empty||b_cat_space) {
        if(args.empty()) {
          if(index==(words.size()-1)) {
          } else {
            std::string s;
            if(!ui->atat(words[index+1],s)) {
              out << "OnX::Session::dld_Function :"
                  << " atat(1) failure for " 
                  << Lib::sout(words[index+1]) << "." 
                  << Lib::endl;
              return false;
            }
            args.push_back(s);
            index++;
          }     
        } else {
          if(index==(words.size()-1)) {
          } else {
            std::string prev = args.back();
            args.pop_back();
            std::string s;
            if(!ui->atat(words[index+1],s)) {
              out << "OnX::Session::dld_Function :"
                  << " atat(2) failure for " 
                  << Lib::sout(words[index+1]) << "." 
                  << Lib::endl;
              return false;
            }
            if(b_cat_empty) {
              args.push_back(prev+s);
            } else if(b_cat_space) {
              args.push_back(prev+" "+s);
            }
            index++;
          }     
        }
      } else {
        std::string s;
        if(!ui->atat(words[index],s)) {
          out << "OnX::Session::dld_Function :"
              << " atat(3) failure for " << Lib::sout(words[index]) << "." 
              << Lib::endl;
          return false;
        }
        args.push_back(s);
      }
    }
  } else {
    out << "OnX::Session::dld_Function :"
        << " at option " << Lib::sout(at) << " unknown." 
        << Lib::endl;
    return false;
  }

  if(sdll.empty()) {
    out << "OnX::Session::dld_Function :"
        << " dll not specified in " << Lib::sout(aString) << "." 
        << Lib::endl;
    return false;
  }

  // treat aliases :
  if(alias=="yes") {
    for(Slash::UI::IInterpreter::Aliases::const_iterator it=aAliases.begin();
      it!=aAliases.end();++it) {
    if(scbk==(*it).first) {
      scbk = (*it).second;
      break;
    }
  }}

  if(scbk.empty()) {
    out << "OnX::Session::dld_Function :"
        << " cbk not specified in " << Lib::sout(aString) << "." 
        << Lib::endl;
    return false;
  }

 {Slash::Core::ILibraryManager* lm = Slash::libraryManager(*session);
  if(lm) {
    Slash::Core::ILibrary* library = lm->addLibrary(sdll,sdll,"");
    if(library) {
      typedef void(*Callback)(Slash::UI::IUI&,const std::vector<std::string>&);
      Callback cbk = (Callback)library->findProcedure(scbk,true);
      if(cbk) {  
        cbk(*ui,args);
      } else {
        // try with a sdll namespace : sdll_scbk.
        Callback cbk = (Callback)library->findProcedure(sdll+"_"+scbk,true);
        if(cbk) {  
          cbk(*ui,args);
        } else {
          out << "OnX::Session::dld_Function :"
              << " can't find callback " << Lib::sout(scbk)
              << " or " << Lib::sout(sdll+"_"+scbk) 
              << " in " << Lib::sout(sdll) << "."
              << " Command was " << Lib::sout(aString) << "." 
              << Lib::endl;
          return false;
        }
      }
    }
  }}

  return true;  
}

bool session_Function(
 const std::string& aString
,const Slash::UI::IInterpreter::Options&
,const Slash::UI::IInterpreter::Aliases&
,void* aTag
){
  Slash::Core::ISession* session = (Slash::Core::ISession*)aTag;
  Lib::Out out(session->printer());

  if(session->verboseLevel()) {
    out << "OnX::Session::session_Function :"
        << " execute " << Lib::sout(aString) << "..."
        << Lib::endl;
  }

  std::string sdll;
  std::string scbk;

  std::vector<std::string> words;
  Lib::smanip::words(aString," ",words);
  if(words.size()<2) {
    out << "OnX::Session::session_Function :"
        << " bad syntax in " << Lib::sout(aString) << "." 
        << Lib::endl;
    return false;
  }
  sdll = words[0];
  scbk = words[1];
  Lib::smanip::strip(sdll);
  Lib::smanip::strip(scbk);

  std::vector<std::string> args;
  for(unsigned int index=2;index<words.size();index++) 
    args.push_back(words[index]);

  if(sdll.empty()) {
    out << "OnX::Session::session_Function :"
        << " dll not specified in " << Lib::sout(aString) << "." 
        << Lib::endl;
    return false;
  }
  if(scbk.empty()) {
    out << "OnX::Session::session_Function :"
        << " cbk not specified in " << Lib::sout(aString) << "." 
        << Lib::endl;
    return false;
  }

 {Slash::Core::ILibraryManager* lm = Slash::libraryManager(*session);
  if(lm) {
    Slash::Core::ILibrary* library = lm->addLibrary(sdll,sdll,"");
    if(library) {
      typedef void(*Callback)(Slash::Core::ISession&,const std::vector<std::string>&);
      Callback cbk = (Callback)library->findProcedure(scbk,true);
      if(cbk) {
        cbk(*session,args);
      } else {
        // try with a sdll namespace : sdll_scbk.
        Callback cbk = (Callback)library->findProcedure(sdll+"_"+scbk,true);
        if(cbk) {  
          cbk(*session,args);
        } else {
          out << "OnX::Session::session_Function :"
              << " can't find callback " << Lib::sout(scbk)
              << " or " << Lib::sout(sdll+"_"+scbk)
              << " in " << Lib::sout(sdll) << "."
              << " Command was " << Lib::sout(aString) << "." 
              << Lib::endl;
          return false;
        }
      }
    }
  }}

  return true;
}
