// this :
#include <OnX/Core/StyleManager.h>

#include <Slash/Core/ISession.h>
#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/BaseML.h>
#include <Lib/ItemML.h>
#include <Lib/Element.h>
#include <Lib/Cast.h>
#include <Lib/smanip.h>

//////////////////////////////////////////////////////////////////////////////
OnX::StyleManager::StyleManager(
 Slash::Core::ISession& aSession
)
:fSession(aSession)
,fName("StyleManager")
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
OnX::StyleManager::~StyleManager(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
std::string OnX::StyleManager::name(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fName;
}
//////////////////////////////////////////////////////////////////////////////
void* OnX::StyleManager::cast(
 const std::string& aClass
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if_Lib_SCast(OnX::StyleManager)
  else if_Lib_SCast(Slash::UI::IStyleManager)
  else if_Lib_SCast(Slash::Core::IManager)
  else return 0;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::StyleManager::isStyle(
 const std::string& aString
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector< std::pair<std::string,std::string> >::const_iterator it;
  for(it=fList.begin();it!=fList.end();++it) {
    //if(aString==(*it).first) return true;
    if(Lib::smanip::match(aString,(*it).first)) return true;
  }
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::StyleManager::findStyle(
 const std::string& aString
,std::string& aValue
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector< std::pair<std::string,std::string> >::const_iterator it;
  for(it=fList.begin();it!=fList.end();++it) {
    //if(aString==(*it).first) {
    if(Lib::smanip::match(aString,(*it).first)) {
      aValue = (*it).second;
      return true;
    }
  }
  aValue = "";
  return false;
}
//////////////////////////////////////////////////////////////////////////////
std::vector<std::string> OnX::StyleManager::styles(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::vector<std::string> v;
  std::vector< std::pair<std::string,std::string> >::const_iterator it;
  for(it=fList.begin();it!=fList.end();++it) v.push_back((*it).first);
  return v;
}
//////////////////////////////////////////////////////////////////////////////
bool OnX::StyleManager::loadFile(
 const std::string& aFile
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::ItemMLFactory factory;
  Lib::BaseML styleML(factory,fSession.verboseLevel());
  std::vector<std::string> tags;
  tags.push_back("styles");
  tags.push_back("style");
  styleML.setTags(tags);
  if(!styleML.loadFile(aFile,false,fSession.printer())) return false;
  Lib::ItemML* top = styleML.topItem();
  if(!top) return true; //File could be empty.

  const std::list<Lib::ItemML*>& list = top->children();
  std::list<Lib::ItemML*>::const_iterator it;

  for(it=list.begin();it!=list.end();++it) {
    // Find the field with name "name" :
    std::string sname;
    (*it)->attributeValue("","name",sname);
    if(sname.empty()) break;

    std::string s;
    const std::list<Lib::Element*> props = (*it)->properties();
    std::list<Lib::Element*>::const_iterator pit;
    for(pit=props.begin();pit!=props.end();++pit) {
      s += (*pit)->name();
      s += " ";
      s += (*pit)->value();
      s += "\n";
    }
    //FIXME : rm last \n.

    if(fSession.verboseLevel()) {
      Lib::Out out(fSession.printer());
      out << "OnX::StyleManager::loadFile :"
          << " style " << Lib::sout(sname) << " : " 
          << s
          << Lib::endl;
    }

    bool found = false;
   {std::vector< std::pair<std::string,std::string> >::iterator it2;
    for(it2=fList.begin();it2!=fList.end();++it2) {
      if(sname==(*it2).first) {
        //Existing style, override :
        (*it2).second = s;
        found = true;
        break;
      }
    }}

    if(!found)
      fList.push_back( std::pair<std::string,std::string>(sname,s));
  }

  return true;
}
