// this :
#include <OnX/Core/UI_Manager.h>

#include <Slash/Core/ISession.h>
#include <Slash/Core/ILibrary.h>
#include <Slash/UI/IUI.h>

#include <Slash/Tools/Managers.h>
#include <Slash/Tools/system.h>
#include <Slash/Tools/smanip.h>

#include <Lib/Cast.h>
#include <Lib/DLL.h>

OnX::UI_Manager::UI_Manager(Slash::Core::ISession& aSession)
:fSession(aSession)
,fName("UI_Manager")
,fUI(0)
{}

OnX::UI_Manager::~UI_Manager(){
  if(fUI) {
    fUI->finalize();
    delete fUI;
  }
  fUI = 0;
}

std::string OnX::UI_Manager::name() const {return fName;}

void* OnX::UI_Manager::cast(const std::string& aClass) const {
  if_Lib_SCast(Slash::UI::IUI_Manager)
  else if_Lib_SCast(Slash::Core::IManager)
  else return 0;
}

bool OnX::UI_Manager::add(Slash::UI::IUI* aUI) {
  if(fUI) {
    std::ostream& out = fSession.cout();
    out << "OnX::UI_Manager::add :"
        << " a UI already exist." 
        << std::endl;
    return false;
  }
  fUI = aUI; //take ownership.
  return true;
}

Slash::UI::IUI* OnX::UI_Manager::find(const std::string&) const {return fUI;}

#include <Slash/Tools/app.h>

Slash::UI::IUI* OnX::UI_Manager::create(
 const std::string& aDriver
,const std::string& aFile
,const std::vector<std::string>& aArgs
,bool aShow
){
  // For the moment only one Slash::UI::IUI handled.
  if(fUI) return fUI; //Done.

  std::string driver = aDriver;
  if(driver.empty()) {
    Slash::getenv("OSC_GUI_TOOLKIT",driver);
    if(driver.empty()) {
      Slash::getenv("ONX_DRIVER",driver);
    }
  }

  if((driver.empty())||((driver=="default"))) { // Guess something :
#if defined(WIN32)
    driver = "Win";
#else
#ifdef __APPLE__
    bool is_mac_app = false;
   {typedef std::pair<std::string,std::string> arg_t;
    std::vector<arg_t> args = fSession.arguments();
    if(args.size()) {
      Slash::is_mac_app(args[0].first,fSession.cout(),is_mac_app);
    }}

    if(is_mac_app) {

      if(Lib::DLL::isValid("OnXNextStep",fSession.cout())) {
        driver = "NextStep";
      } else if(Lib::DLL::isValid("OnXXt",fSession.cout())) {
        driver = "Xt";
      } else if(Lib::DLL::isValid("OnXQt",fSession.cout())) {
        driver = "Qt";
      } else if(Lib::DLL::isValid("OnXGtk",fSession.cout())) {
        driver = "Gtk";
      } else if(Lib::DLL::isValid("OnXSDL",fSession.cout())) {
        driver = "SDL";
      } else {
        driver = "Xt";
      }

    } else {

      if(Lib::DLL::isValid("OnXXt",fSession.cout())) {
        driver = "Xt";
      } else if(Lib::DLL::isValid("OnXQt",fSession.cout())) {
        driver = "Qt";
      } else if(Lib::DLL::isValid("OnXNextStep",fSession.cout())) {
        driver = "NextStep";
      } else {
        driver = "Xt";
      }

    }

#else //Linux
    if(Lib::DLL::isValid("OnXGtk",fSession.cout())) {
      driver = "Gtk";
    } else if(Lib::DLL::isValid("OnXXt",fSession.cout())) {
      driver = "Xt";
    } else if(Lib::DLL::isValid("OnXQt",fSession.cout())) {
      driver = "Qt";
    } else {
      driver = "Gtk";
    }
#endif
#endif
  }

  std::string library = "OnX"+driver;

  Slash::Core::ILibraryManager* lm = Slash::libraryManager(fSession);
  if(!lm) {
    std::ostream& out = fSession.cout();
    out << "OnX::UI_Manager::create :"
        << " LibraryManager not found." << std::endl;
    return 0;
  }
  Slash::Core::ILibrary* dll = 
    lm->addLibrary("OnX_driver_"+driver,library,"");
  if(!dll) {
    std::ostream& out = fSession.cout();
    out << "OnX::UI_Manager::create :"
        << " Unable to add DLL " << Slash::sout(library) 
        << std::endl;
    return 0;
  }

  if(fSession.verboseLevel()) {
    std::ostream& out = fSession.cout();
    out << "OnX::UI_Manager::create :"
        << " Load " << Slash::sout(driver) << " driver..." 
        << std::endl;
  }

  std::string function = "OnX_create"+driver+"UI";
  Factory uiFactory = (Factory)dll->findProcedure(function);
  if(!uiFactory) {
    std::ostream& out = fSession.cout();
    out << "OnX::UI_Manager::create :"
        << " Function " << Slash::sout(function) << " not found" 
        << " in library " << Slash::sout(library) << "." 
        << std::endl;
    return 0;
  }

  if(fSession.verboseLevel()) {
    std::ostream& out = fSession.cout();
    out << "OnX::UI_Manager::create :"
        << " Load file " << Slash::sout(aFile) << "."
        << std::endl;
  }

  fUI = uiFactory(fSession,aArgs);
  if(!fUI) {
    std::ostream& out = fSession.cout();
    out << "OnX::UI_Manager::create :"
        << " Creation of UI for " << Slash::sout(aFile) << " failed." 
        << std::endl;
    return 0;
  }

  if(!fUI->load("",aFile,true,aShow)) {
    delete fUI;
    fUI = 0; // To redirect output to stdout before the below print.
    std::ostream& out = fSession.cout();
    out << "OnX::UI_Manager::create :"
        << " Problem when creating the UI for file " 
        << Slash::sout(aFile) << "." 
        << std::endl;
    return 0;
  }

  if(fSession.verboseLevel()) {
    std::ostream& out = fSession.cout();
    out << "OnX::UI_Manager::create :"
        << " loading " << Slash::sout(aFile) << " ok." 
        << std::endl;
  }

  return fUI;
}

