// this :
#include <OnX/Net/NetUI.h>

#include <Slash/Core/ISession.h>
#include <Slash/UI/IUI.h>
#include <Slash/UI/IScriptManager.h>
#include <Slash/Data/IProcessor.h>
#include <Slash/Tools/system.h>
#include <Slash/Tools/file.h>
#include <Slash/Tools/args.h>

#include <Lib/Manager.h>
#include <Lib/Cast.h>

#include <OnX/Net/Protocol.h>
#include <OnX/Net/NetManager.h>

#include <stdlib.h> //exit
#include <stdio.h> //printf

#define ONX_HAS_INVENTOR
#ifdef ONX_HAS_INVENTOR
#include <Inventor/SoDB.h>
#include <Inventor/nodekits/SoNodeKit.h>
#include <OnX/Inventor/Inventor.h>
#include <OnX/Inventor/HEPVis.h>
#include <HEPVis/nodekits/SoPage.h>
#endif

#include <Slash/UI/IWidget.h>
#include <Slash/UI/ISoViewer.h>
#include <OnX/Core/WidgetClass.h>

namespace OnX {
namespace Net {

class SoViewer
:public virtual Slash::UI::IWidget 
,public virtual Slash::UI::ISoViewer
{
public: //Slash::UI::IWidget
  virtual void* cast(const std::string& aClass) const {
    if_Lib_SCast(Slash::UI::IWidget)
    else if_Lib_SCast(Slash::UI::ISoViewer)
    else return 0;
  }
  virtual const Slash::UI::IWidgetClass& widgetClass() const {
    return fWidgetClass;
  }
  virtual std::string name() const {return fName;}
  virtual void* nativeWidget() const {return 0;}
  virtual bool executeCallbacks(const std::string&) {return false;}
  virtual bool size(unsigned int&,unsigned int&) {return false;}
  virtual bool position(int&,int&) {return false;}
  virtual bool write(const std::string& file,
                     const std::string& format,
                     const std::string& options = "") {return false;}
  virtual bool setParameter(const std::string&,const std::string&) {return false;}
  virtual bool parameterValue(const std::string&,std::string&) {return false;}
  virtual bool manage(bool) {return false;}
  virtual bool refresh() {return false;}
  virtual bool show() {return false;}
  virtual bool hide() {return false;}

public: //Slash::UI::ISoViewer
  virtual void setBackgroundColor(double red,double green,double blue) {}
  virtual void setDecoration(bool) {}
  virtual void setViewing(bool) {}
  virtual void setAutoClipping(bool) {}
  virtual void setHeadlight(bool) {}
  virtual void setSmoothing(bool) {}
  virtual void setTitle(const std::string&) {}
  virtual void render() {}
  virtual void viewAll() {}
  virtual bool readScene(const std::string&,
                         const std::string&,
                         const std::string&){
    return false;
  }
  virtual bool clear(const std::string&){return false;}
  virtual bool removeManips(const std::string&) {return false;}
  virtual bool detectIntersections(bool stopAtFirst,
                                   const std::string& aOptions) {return false;}
  virtual void setFeedbackVisibility(bool) {}
  virtual void setAnimation(bool) {}
  virtual void stopAnimating() {}
  virtual SoNode* nodeOfInterest(const std::string& aWhat) const {return 0;}
  virtual SoNode* sceneGraph() const {return fSceneGraph;}
  virtual SbViewportRegion* viewportRegion() const {return 0;}
  virtual bool collect(int,int,unsigned int,unsigned int) {return false;}
  virtual SoGLRenderAction* getGLRenderAction() const {return 0;}
  virtual bool setTransparencyType(int) {return false;}
  virtual bool transparencyType(int& a_type) const {a_type = 0;return false;}
public:
  SoViewer(const std::string& aName = "")
  :fName(aName)
  ,fWidgetClass("PageViewer",
                std::vector<std::string>(),
                std::vector<std::string>(),
                std::vector<std::string>(),
                std::vector<std::string>())
  ,fSceneGraph(0)
  {}
  virtual ~SoViewer(){
    if(fSceneGraph) fSceneGraph->unref();
  }
public:
  void setSceneGraph(SoNode* aNode) {
    if(fSceneGraph) return; //done.
    fSceneGraph = aNode;
  }
protected:
  std::string fName;
  OnX::WidgetClass fWidgetClass;
  SoNode* fSceneGraph;
};

}}

OnX::NetUI::NetUI(Slash::Core::ISession& aSession,const std::vector<std::string>& /*aArgs*/)
:fSession(aSession)
,fScriptManager(0)
,fNetManager(0)
,fVerboseLevel(0)
,fIsValid(false)
,fClient(aSession.cout(),"",50701,fVerboseLevel)
,fCurrentWidget(0)
,fBeingDestroyed(false)
{
  fScriptManager = Lib_findManager(fSession,"ScriptManager",Slash::UI::IScriptManager);

  std::string svalue;
  if(fSession.parameterValue("UI_Manager.verbose",svalue)) {
    int value;
    if(Slash::to<int>(svalue,value)) fVerboseLevel = value;
  } else {
    fVerboseLevel = fSession.verboseLevel();
  }

  std::string host;
  if(!Slash::getenv("OSC_DISPLAY",host)) {
    std::ostream& out = fSession.cout();
    out << "OnX::NetUI::NetUI : "
        << " env variable OSC_DISPLAY not defined."
        << std::endl;
    return;
  }

  fClient.setHost(host);

  if(!fClient.connect()) {
    std::ostream& out = fSession.cout();
    out << "OnX::NetUI::NetUI : "
        << " can't connect to server."
        << std::endl;
    return;
  }

 {std::string driver;
  Slash::args args(fSession.arguments());
  args.find("-server_toolkit",driver);
  if(!fClient.send_string(driver)) {
    std::ostream& out = fSession.cout();
    out << "OnX::NetUI::NetUI : "
        << " can't send driver string to server."
        << std::endl;
    return;
  }}

  if(!aSession.findManager("NetManager")) {
    OnX::NetManager* netManager = new OnX::NetManager(aSession,fClient);
    aSession.addManager(netManager);
    fNetManager = netManager;
  }

#ifdef ONX_HAS_INVENTOR
  if(!SoDB::isInitialized()) {SoDB::init();}
  SoNodeKit::init();
  //SoInteraction::init();

  HEPVis::initClasses();
  OnX::Inventor::initialize(aSession);
#endif

  // Create a current widget :
  OnX::Net::SoViewer* soViewer = new OnX::Net::SoViewer(); //no name
#ifdef ONX_HAS_INVENTOR
  SoPage* soPage = new SoPage();
  soPage->ref();
  soPage->createRegions("SoDisplayRegion",1,1,0);
  soViewer->setSceneGraph(soPage); //soViewer becomes the owner.
#endif

  fCurrentWidget = soViewer;

  fIsValid = true;
}

OnX::NetUI::~NetUI() { 
  fBeingDestroyed = true;

  delete fCurrentWidget;

  fClient.disconnect();

#ifdef ONX_HAS_INVENTOR
  OnX::Inventor::finalize(fSession);
#endif
}

int OnX::NetUI::steerLevel() const {return 0;}

Slash::Core::IWriter& OnX::NetUI::printer() const {return fSession.printer();}

void* OnX::NetUI::cast(const std::string& aClass) const {
  if_Lib_SCast(Slash::UI::IUI)
  else return 0;
}

void OnX::NetUI::put(const std::string& aString){
  // Append line at end of report region.
  ::printf("%s",aString.c_str());
}

bool OnX::NetUI::executeScript(
 const std::string& aInterp
,const std::string& aScript
,const Slash::UI::IUI::Options& aParams
){
  if(!fScriptManager) return false;
  return fScriptManager->executeScript(aInterp,aScript,aParams);
}

int OnX::NetUI::steer(){
  if(!fIsValid) return 0;
  // Treat events coming from the OnX server :

  std::string s;
  std::string what;
  std::string cmd;

  while(true) {

    if(fVerboseLevel) {
      std::ostream& out = fSession.cout();
      out << "OnX::NetUI::steer : "
          << " go to waitString..."
          << std::endl;
    }

    if(!fClient.wait_string(s)) return 0;

    if(fVerboseLevel) {
      std::ostream& out = fSession.cout();
      out << "OnX::NetUI::steer : "
          << " from server " << Slash::sout(s)
          << std::endl;
    }

    if(!OnX_protocol(s,what,cmd)) {
      std::ostream& out = fSession.cout();
      out << "OnX::NetUI::steer : "
          << " bad syntax in " << Slash::sout(s) 
          << std::endl;
      fClient.disconnect();
      return 0;
    }

    if(what==ONX_PROTOCOL_FILE) {
      // Exa <reference> .onx file 
      if(!fNetManager->sendFile(Slash::file_name(cmd))) return 0;

    } else if(what==ONX_PROTOCOL_FILE_EXISTS) {      
      bool value = Slash::file::exists(Slash::file_name(cmd));
      if(!fNetManager->sendString(value?"true":"false")) return 0;

    } else if(what==ONX_PROTOCOL_EXIT) {
      int exit_code;
      if(!Slash::to<int>(cmd,exit_code)) return 0;
      return exit_code;

    } else if(what==ONX_PROTOCOL_EXEC) {
      std::string::size_type pos = cmd.find(":");
      if(pos==std::string::npos){
        std::ostream& out = fSession.cout();
        out << "OnX::NetUI::steer : "
            << " bad command syntax " << Slash::sout(cmd) << "."
            << std::endl;
        return 0;
      }
      std::string interp = cmd.substr(0,pos);
      std::string script = cmd.substr(pos+1,cmd.length()-(pos+1));

      if(fVerboseLevel) {
        std::ostream& out = fSession.cout();
        out << "OnX::NetUI::steer : "
            << " interp " << Slash::sout(interp)
            << " script " << Slash::sout(script)
            << std::endl;
      }

      if(fScriptManager) {
        Slash::UI::IInterpreter::Options params;
        fScriptManager->executeScript(interp,script,params);
      }

      fNetManager->pushEnd("end",std::vector<std::string>());

      if(!fNetManager->flushEnds()) return 0;

    } else if(what==ONX_PROTOCOL_DATA) {

      Slash::Data::IProcessor* accessorManager = 
        Lib_findManager(fSession,"AccessorManager",Slash::Data::IProcessor);
      if(accessorManager) {
        std::vector<std::string> ss;
        ss.push_back(cmd);
        accessorManager->execute("init",ss);
      }

    } else {
       std::ostream& out = fSession.cout();
       out << "OnX::NetUI::steer : "
           << " receive event : " << Slash::sout(s) << " from server."
           << std::endl;
    }
  }
  return 0;
}

void OnX::NetUI::synchronize(){}

void OnX::NetUI::exit(int){::exit(0);}

void OnX::NetUI::echo(const std::string& aString){
  std::ostream& out = fSession.cout();
  out << aString << std::endl;
}

void OnX::NetUI::warn(const std::string& aString){
  std::ostream& out = fSession.cout();
  out << "WARNING : " << aString << std::endl;
}

std::string OnX::NetUI::ask(const std::string&){return "";}

void OnX::NetUI::addDispatcher(Slash::UI::IUI::Dispatcher){}

bool OnX::NetUI::removeCallback(const std::string&,const std::string&,int){
  return false;
}

bool OnX::NetUI::setCallback(
 const std::string&
,const std::string&
,const std::string&
,const std::string&
,int
){
  return false;
}

bool OnX::NetUI::show(){return false;}

std::string OnX::NetUI::callbackValue(){return fValue;}

const std::string& OnX::NetUI::thisComponent(){return fValue;}

Slash::UI::IWidget* OnX::NetUI::findWidget(const std::string&) const {return 0;}

Slash::UI::IWidget* OnX::NetUI::findWidgetByAttribute(const std::string& aAttribute,const std::string& aValue,bool,bool) const {return 0;}

bool OnX::NetUI::load(Slash::UI::IWidget& aParent,const std::string& aToLoad,bool aIsFile,bool aShow) {return false;}

Slash::UI::IWidget* OnX::NetUI::createWidget(
 const Slash::UI::IWidgetClass&
,const std::string&
,Slash::UI::IWidget&
,const std::vector<std::string>&
){
  return 0;
}

bool OnX::NetUI::parameterValue(const std::string&,std::string& aValue) {
  aValue = "";
  return false;
}

bool OnX::NetUI::setParameter(const std::string&,const std::string&){return false;}

bool OnX::NetUI::load(const std::string& aParent,const std::string& aFile,bool,bool){
  if(!fClient.send_string(aFile)) {
    std::ostream& out = fSession.cout();
    out << "OnX::NetUI::NetUI : "
        << " can't send OnX file string to server."
        << std::endl;
    return false;
  }
  return true;
}

void OnX::NetUI::deleteChildren(const std::string&){}

Slash::UI::ICyclic* OnX::NetUI::createCyclic(const std::string&,int,const std::string&,const std::string&){return 0;}

bool OnX::NetUI::createComponent(const std::string&,const std::string&,const std::string&){return false;}

std::vector<std::string> OnX::NetUI::findChildren(const std::string&,const std::string&,const std::string&) const {return std::vector<std::string>();}

std::vector<std::string> OnX::NetUI::findParent(const std::string&,const std::string&,const std::string&) const {return std::vector<std::string>();}

bool OnX::NetUI::destroyComponent(const std::string&){return false;}

Slash::Core::ISession& OnX::NetUI::session(){return fSession;}

bool OnX::NetUI::executeCallbacks(const std::string&,const std::string&){return false;}

bool OnX::NetUI::notify(Slash::Core::IMessage*){return false;}

bool OnX::NetUI::copyComponent(const std::string&,const std::string&,const std::string&){return false;}

bool OnX::NetUI::save(){return false;}

const Slash::UI::ICallbackData& OnX::NetUI::callbackData(){return fCallbackData;}

Slash::UI::ICyclic* OnX::NetUI::findCyclic(const std::string&) const {return 0;}

bool OnX::NetUI::removeCyclic(Slash::UI::ICyclic*&){return false;}
bool OnX::NetUI::isComponent(const std::string&){return false;}

bool OnX::NetUI::lock(){return false;}

bool OnX::NetUI::unlock(){return false;}

bool OnX::NetUI::finalize(){return true;}

bool OnX::NetUI::show(const std::string&){return false;}

bool OnX::NetUI::beingDestroyed() const {return fBeingDestroyed;}

bool OnX::NetUI::isValid() const {return fIsValid;}

bool OnX::NetUI::setCallbackData(const Slash::UI::ICallbackData&){return true;}

std::string OnX::NetUI::createWidgetName(){return "";}

Slash::UI::IWidgetClass* OnX::NetUI::findWidgetClass(const std::string&) const {return 0;}

Slash::UI::IWidget* OnX::NetUI::currentWidget() const {return fCurrentWidget;}
bool OnX::NetUI::setCurrentWidget(Slash::UI::IWidget*){return false;} 

bool OnX::NetUI::enableTerminal(const std::string& aPrompt,const std::string& aInterp){return false;}

bool OnX::NetUI::disableTerminal(){return false;}
bool OnX::NetUI::redirectAskToTerminal(bool aValue){return false;} 

bool OnX::NetUI::lockScreen(const std::string&,const std::string&,const std::string&){return false;} 

bool OnX::NetUI::destroyWidget(Slash::UI::IWidget*&){return false;}

bool treat_atat(Slash::UI::IUI&,const std::string&,std::string&);

bool OnX::NetUI::atat(const std::string& aString,std::string& aValue) const{
  Slash::UI::IUI* ui = (Slash::UI::IUI*)this;
  return treat_atat(*ui,aString,aValue);
}
