// this :
#include <OnX/Qt/QTk.h>

#include <Lib/Debug.h>
#include <Lib/smanip.h>

#include <OnX/Qt/QCallbackData.h>

//#define DEBUG_DESTROY

#include <QtGui/qcolordialog.h>
#include <QtGui/qcheckbox.h>
#include <QtGui/qslider.h>
#include <QtGui/qtabwidget.h>
#include <QtGui/qcombobox.h>
#include <QtGui/qmenu.h>

#include <QtGui/qfiledialog.h>
#include <QtGui/qevent.h>
#include <QtGui/qmenudata.h> // for QMenuItem definition

#define S_OnX_QFileDialog     "OnX_QFileDialog"
#define S_OnX_QToggleMenuItem "OnX_QToggleMenuItem"
#define S_OnX_QMenuItem       "OnX_QMenuItem"
#define S_OnX_QMenu           "OnX_QMenu"
#define S_OnX_QColorDialog    "OnX_QColorDialog"
#define S_OnX_QOpenGLArea     "OnX_QOpenGLArea"
#define S_OnX_QSplitter       "OnX_QSplitter"

//FIXME : OnX_ -> OnX::

OnX_QApplication::OnX_QApplication(int& a_argc,char** a_argv)
:QApplication(a_argc,a_argv)
,fCB(0),fTag(0)
{}
OnX_QApplication::~OnX_QApplication(){}
bool OnX_QApplication::event(QEvent* a_event) {
  if(a_event->type()==QEvent::FileOpen) {
    if(fCB) {
      return fCB(static_cast<QFileOpenEvent*>(a_event)->file(),fTag);
    }
  }
  return QApplication::event(a_event);
}
void OnX_QApplication::setOpenFileCallback(OpenFileCB aCB,void* aTag){
  fCB = aCB;
  fTag = aTag;
}

OnX_QSplitter::OnX_QSplitter(
 Qt::Orientation aOrientation
,QWidget* aParent
)
:QSplitter(aOrientation,aParent)
,fSplitPosition(0){
  Lib::Debug::increment(S_OnX_QSplitter);
}
OnX_QSplitter::~OnX_QSplitter(){
  Lib::Debug::decrement(S_OnX_QSplitter);
}
void OnX_QSplitter::setSplitPosition(int aPosition){
  fSplitPosition = aPosition;
}
void OnX_QSplitter::showEvent(QShowEvent* aEvent){
  QList<int> vals = sizes();
  if(vals.size()==2) {
    vals[0] = fSplitPosition;
    vals[1] = (orientation()==Qt::Vertical ? height() : width());
    vals[1] -= fSplitPosition;
    setSizes(vals);
  }
  QSplitter::showEvent(aEvent);
}

OnX_QColorDialog::~OnX_QColorDialog(){
  Lib::Debug::decrement(S_OnX_QColorDialog);
}
std::string OnX_QColorDialog::value() const {return fValue;}

OnX_QColorDialog_Base::OnX_QColorDialog_Base()
#if QT_VERSION < 0x040600
:QWidget(0){}
#else
:QColorDialog(0){}
#endif
OnX_QColorDialog_Base::~OnX_QColorDialog_Base(){}

#if QT_VERSION < 0x040600
OnX_QColorDialog::OnX_QColorDialog() {
  Lib::Debug::increment(S_OnX_QColorDialog);
}
void OnX_QColorDialog::map(){
  fValue = "";
  QColor color = QColorDialog::getColor();
  if(!color.isValid()) return;
  Lib::smanip::printf(fValue,3*32,"%g %g %g",
  ((double)color.red())/255,
  ((double)color.green())/255,
  ((double)color.blue())/255);
  emit ok();
  fValue = "";
}
void OnX_QColorDialog::setOkHide(bool) {}
void OnX_QColorDialog::color_selected(){}
#else
OnX_QColorDialog::OnX_QColorDialog()
:f_ok_hide(true) //Qt default
,f_do_done(true)
{
  Lib::Debug::increment(S_OnX_QColorDialog);
  //setWindowTitle(title);
  //setOptions(options);
  setCurrentColor(Qt::white);
  setModal(false);
  QObject::connect(this,SIGNAL(colorSelected(const QColor&)),
                   this,SLOT(color_selected()));
}
void OnX_QColorDialog::color_selected(){
  QColor color = selectedColor();
  if(!color.isValid()) return;
  //printf("debug : colorSelected\n");
  Lib::smanip::printf(fValue,3*32,"%g %g %g",
  ((double)color.red())/255,
  ((double)color.green())/255,
  ((double)color.blue())/255);
  emit ok();
  fValue = "";
  if(f_ok_hide) return;
  //To have an "Apply" instead of a "Ok" :
  show();
}
void OnX_QColorDialog::setOkHide(bool aValue) {f_ok_hide = aValue;}

//NOTE : with modeless/show, an "Ok" induces a
//       recursive done()/setVisible().
//       We overload done() and setVisible() to avoid
//       that.
void OnX_QColorDialog::done(int r){
  if(!f_do_done) return;
  QColorDialog::done(r);
}
void OnX_QColorDialog::setVisible(bool aVisible){
  f_do_done = false;
  QColorDialog::setVisible(aVisible);
  f_do_done = true;
}
#endif

OnX_QFileDialog::OnX_QFileDialog()
:QWidget(0)
,fMode(AcceptOpen){
  Lib::Debug::increment(S_OnX_QFileDialog);
}
OnX_QFileDialog::~OnX_QFileDialog(){
  Lib::Debug::decrement(S_OnX_QFileDialog);
}
void OnX_QFileDialog::map(){
  if(fMode==AcceptSave) {
    QString s = QFileDialog::getSaveFileName(0,
                      "Choose a filename to save under",
                      fDirectory.c_str(),fFilter.c_str());
    fValue = s.toStdString();
  } else { //AcceptOpen
    QString s = QFileDialog::getOpenFileName(0,
                      "Choose a filename to open",
                      fDirectory.c_str(),fFilter.c_str());
    fValue = s.toStdString();
  }
  emit ok();
  fValue = "";
}
void OnX_QFileDialog::setAcceptMode(AcceptMode aMode) {
  fMode = aMode;
}
void OnX_QFileDialog::setFilter(const std::string& aFilter) {
  fFilter = aFilter;
}
void OnX_QFileDialog::setDirectory(const std::string& aDirectory) {
  fDirectory = aDirectory;
}
std::string OnX_QFileDialog::value() const {
  return fValue;
}

OnX_QMenu::OnX_QMenu(QMenu& aPopup)
:QWidget(0)
,fPopup(aPopup){
  Lib::Debug::increment(S_OnX_QMenu);
}
OnX_QMenu::~OnX_QMenu(){Lib::Debug::decrement(S_OnX_QMenu);}
QMenu& OnX_QMenu::popupMenu() const{ return fPopup;}

OnX_QMenuItem::OnX_QMenuItem(QMenu& aPopup,const std::string& aLabel)
:QWidget(0)
,fPopup(aPopup)
,fAction(0)
{
  fAction = fPopup.addAction(aLabel.c_str(),this,SLOT(execute()));
  Lib::Debug::increment(S_OnX_QMenuItem);
}
OnX_QMenuItem::~OnX_QMenuItem(){
  fPopup.removeAction(fAction);
  delete fAction;
  Lib::Debug::decrement(S_OnX_QMenuItem);
}
void OnX_QMenuItem::setAccel(const std::string& aValue) {
  //FIXME
}
QMenu& OnX_QMenuItem::popupMenu() const{ return fPopup;}

//int OnX_QMenuItem::id() const{ return fID;}
void OnX_QMenuItem::execute(){ emit activated();}
bool OnX_QMenuItem::label(std::string& aValue) const{ 
  aValue = fAction->text().toStdString();
  return true;
}
bool OnX_QMenuItem::setLabel(const std::string& aValue){ 
  fAction->setText(aValue.c_str());
  return true;
}

OnX_QToggleMenuItem::OnX_QToggleMenuItem(QMenu& aPopup,const std::string& aLabel)
:OnX_QMenuItem(aPopup,aLabel){
  fAction->setCheckable(true);
}
OnX_QToggleMenuItem::~OnX_QToggleMenuItem(){}
bool OnX_QToggleMenuItem::isChecked(bool& aValue) const{ 
  aValue = fAction->isChecked()?true:false;
  return true;
}
bool OnX_QToggleMenuItem::setChecked(bool aValue){ 
  fAction->setChecked(aValue);
  return true;
}

OnX_QMenuSeparator::OnX_QMenuSeparator(QMenu& aPopup)
:QWidget(0)
,fPopup(aPopup)
,fAction(0)
{
  fAction = new QAction(&fPopup);
  fAction->setSeparator(true);
  fPopup.addAction(fAction);
  Lib::Debug::increment(S_OnX_QMenuItem);
}
OnX_QMenuSeparator::~OnX_QMenuSeparator(){
  fPopup.removeAction(fAction);
  delete fAction;  
  Lib::Debug::decrement(S_OnX_QMenuItem);
}

OnX_QMainWindow::OnX_QMainWindow(QWidget* aParent,Qt::WFlags aFlags)
:QMainWindow(aParent,aFlags){
}
OnX_QMainWindow::~OnX_QMainWindow(){}
void OnX_QMainWindow::closeEvent(QCloseEvent* aEvent) {
  emit close_window();
  QMainWindow::closeEvent(aEvent);
}

OnX_QOpenGLArea::OnX_QOpenGLArea(QWidget* aParent)
:QGLWidget(aParent){
  Lib::Debug::increment(S_OnX_QOpenGLArea);
}
OnX_QOpenGLArea::~OnX_QOpenGLArea(){
  Lib::Debug::decrement(S_OnX_QOpenGLArea);
}
void OnX_QOpenGLArea::initialieGL() {}
void OnX_QOpenGLArea::resizeGL(int,int){}
void OnX_QOpenGLArea::paintGL() {
  QCallbackData data;
  emit paint(data);
}
void OnX_QOpenGLArea::mousePressEvent(QMouseEvent* aEvent) {
  QCallbackData data;
  data.event = "ButtonPress";
  data.x = aEvent->x();
  data.y = aEvent->y();
  emit eventHandler(data);
}
void OnX_QOpenGLArea::mouseReleaseEvent(QMouseEvent* aEvent) {
  QCallbackData data;
  data.event = "ButtonRelease";
  data.x = aEvent->x();
  data.y = aEvent->y();
  emit eventHandler(data);
}
void OnX_QOpenGLArea::mouseMoveEvent(QMouseEvent* aEvent) {
  QCallbackData data;
  data.event = "MotionNotify";
  data.x = aEvent->x();
  data.y = aEvent->y();
  emit eventHandler(data);
}


#include "../source/Core/OnX_gl2ps.h"

bool OnX_QOpenGLArea::write_gl2ps(const std::string& aFileName,const std::string&) { 
  FILE* file = ::fopen(aFileName.c_str(),"w");
  if(!file) return false;

  int options = GL2PS_OCCLUSION_CULL 
    | GL2PS_BEST_ROOT 
    | GL2PS_SILENT
    | GL2PS_DRAW_BACKGROUND;
  int sort = GL2PS_BSP_SORT;
  //sort = GL2PS_SIMPLE_SORT;
    
  GLint vp[4];
  vp[0] = 0;
  vp[1] = 0;
  vp[2] = width();
  vp[3] = height();

  int bufsize = 0;
  gl2psBeginPage("title","OnX_gl2ps", 
                 vp,GL2PS_EPS,sort,options, 
                 GL_RGBA,0, NULL,0,0,0,bufsize, 
                 file,aFileName.c_str());
    
  paintGL();

  gl2psEndPage();
  ::fclose(file);

  return true;
}
