// this :
#include <OnX/SDL/SDL_UI.h>

#include <Slash/Core/ISession.h>
#include <Slash/UI/IUI.h>
#include <Slash/UI/IScriptManager.h>
#include <Slash/UI/IWidget.h>
#include <Slash/UI/ISoViewer.h>
#include <Slash/Tools/system.h>

#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/smanip.h>
#include <Lib/Cast.h>
#include <Lib/Arguments.h>
#include <Lib/Element.h>

#include <stdlib.h> //exit
#include <stdio.h> //printf

#include <OnX/SDL/SDL_Cyclic.h>

#ifdef __APPLE__
#include <OnX/SDL/NS.h>
#endif

#include "../source/Core/Widgets.cxx"

//
#define ONX_HAS_INVENTOR

#ifdef ONX_HAS_INVENTOR
#include "SoSDL_Viewer.h"

#include <Inventor/SDL/SoSDL.h> //WIN32 : Before coin include.

#include <OnX/Inventor/Inventor.h>
#include <OnX/Inventor/HEPVis.h>
#include <OnX/Inventor/GuiViewer.h> //viewer_proxy

// Inventor :
#include <Inventor/nodes/SoSeparator.h>

// SDL meta scene :
#include <HEPVis/SbMath.h>
#include <HEPVis/SbColors.h>
#include <HEPVis/nodes/SoAreaStyle.h>
#include <HEPVis/nodes/SoTextStyle.h>
#include <HEPVis/nodekits/SoFormRegion.h>
#include <HEPVis/nodekits/SoSimpleButton.h>

#include <HEPVis/SDL/viewers/SoSDL_Viewer.h>

class GameViewer : public HEPVis::SoSDL_Viewer {
public:
  GameViewer(Slash::UI::IUI& aIUI,SoSDL_Widget* aWidget)
  : HEPVis::SoSDL_Viewer(aWidget)
  ,fIUI(aIUI)
  {}
  virtual ~GameViewer(){}
  void lock(const SbString& aPass,SoSDL_UnlockCB aCB,void* aData) {}
public:
  /*
  virtual SbBool processEvent(SDL_Event& aEvent,SbBool& aRender){
    if(aEvent.type==SDL_QUIT) { //close window in non-fullscreen mode.
      fIUI.exit();
      return TRUE;
    } else if(aEvent.type==SDL_KEYDOWN) {
      if(aEvent.key.keysym.sym==SDLK_q) {
        fIUI.exit();
        return TRUE;
      }
    }
    return SoSDL_Viewer::processEvent(aEvent,aRender);
  }
  */
private:
  Slash::UI::IUI& fIUI;
};

#include <HEPVis/SDL/app.h>

#endif

OnX::SDL_UI::SDL_UI(Slash::Core::ISession& aSession,const std::vector<std::string>& aArgs)
:BaseUI(aSession)
,fTopWidget(0)
,fJoyTag(0)
,fSceneGraph(0)
,fSoViewer(0)
,fUI_Region(0)
#ifdef __APPLE__
,fNS(0)
#endif
{
#ifdef __APPLE__
  fNS = new NS();
#endif

  //fScriptManager = 
  //  Lib_findManager(fSession,"ScriptManager",Slash::UI::IScriptManager);

  //////////////////////////////////////////
  // Initialize SDL and related packages :
  //////////////////////////////////////////

 {Lib::Arguments args(aArgs);

  int video_width = 0;
  if(!args.find("-video_width",video_width)) video_width = 640;
  int video_height = 0;
  if(!args.find("-video_height",video_height)) video_height = 480;

  Uint32 flags = SDL_FULLSCREEN;
  if(args.isAnArgument("-not_full_screen")) {
    flags = 0;
  } else if(Slash::isenv("OSC_NOT_FULL_SCREEN")) {
    flags = 0;
  }
  //flags = 0;
  fTopWidget = SoSDL::init(video_width,video_height,flags);
  if(!fTopWidget) return;}

  SDL_Surface* screen = SoSDL::screen();
  if(!screen) return;
  HEPVis::initClasses();
  OnX::Inventor::initialize(aSession);

  //////////////////////////////////////////
  /// Meta scene : /////////////////////////
  //////////////////////////////////////////

  fSceneGraph = new SoSeparator;
  fSceneGraph->ref();

  fJoyTag = new JoyTag;
 {SoEventCallback* eventCallback = new SoEventCallback;
  eventCallback->addEventCallback(SoKeyboardEvent::getClassTypeId(),
                                  hepvis_keyboardCB,0);
  eventCallback->addEventCallback(SoJoyButtonEvent::getClassTypeId(),
                                  joyButtonCB,fJoyTag);
  eventCallback->addEventCallback(SoJoyAxisEvent::getClassTypeId(),
                                  hepvis_joyAxisCB,fJoyTag);
  fSceneGraph->addChild(eventCallback);}

  ///////////////////////////////////////////////
  // GUI :
  ///////////////////////////////////////////////
  fUI_Region = new SoFormRegion; //And not from soPage->createRegion.
  fUI_Region->origin.setValue(SbVec2f(0,0));
  fUI_Region->size.setValue(SbVec2f(0.1F,0.05F));
  fUI_Region->columns.setValue(0);
  fUI_Region->rows.setValue(1);
  fUI_Region->visible.setValue(FALSE);
  fSceneGraph->addChild(fUI_Region);

  ///////////////////////////////////////////////
  // Shooting cross :
  ///////////////////////////////////////////////
  SoRegion* shootRegion = new SoRegion;
  float w = 0.008F;
  float h = 0.008F;
  shootRegion->origin.setValue(SbVec2f(0.5F*(1.0F-w),0.5F*(1.0F-h)));
  shootRegion->size.setValue(SbVec2f(w,h));
  shootRegion->visible.setValue(FALSE);
  fSceneGraph->addChild(shootRegion);

  // Viewer :
  fSoViewer = new GameViewer(*this,fTopWidget);
  if(fSoViewer->setSceneGraph(fSceneGraph)==FALSE) {
    fTopWidget->setViewer(0);
    fSceneGraph->unref();
    fSceneGraph = 0;
    delete fSoViewer;
    fSoViewer = 0;
    fUI_Region = 0;
    Lib::Out out(fSession.printer());
    out << "OnX::SDL_UI::SDL_UI :"
        << " SoSDL_Viewer::setSceneGraph failure."
        << Lib::endl;
    return;
  }

  fSoViewer->setHeadlight(FALSE);
  fSoViewer->setShootRegion(shootRegion);
  fSoViewer->setOptionsRegion(fUI_Region);
}

OnX::SDL_UI::SDL_UI(const SDL_UI& aFrom)
:BaseUI(aFrom.fSession)
,fTopWidget(0)
,fSceneGraph(0)
,fSoViewer(0)
,fUI_Region(0)
#ifdef __APPLE__
,fNS(0)
#endif
{}

OnX::SDL_UI::~SDL_UI() { 
  fBeingDestroyed = true;

  removeCyclics();

  delete fJoyTag;

  fTopWidget = 0; //deleted by the SoSDL::clear.

  fSceneGraph->unref();

  delete fSoViewer;

  SoSDL::clear();  

#ifdef __APPLE__
  delete fNS;
#endif
}

OnX::SDL_UI& OnX::SDL_UI::operator=(const SDL_UI&){return *this;}

void* OnX::SDL_UI::cast(const std::string& aClass) const {
  if_Lib_SCast(OnX::SDL_UI)
  else return BaseUI::cast(aClass);
}

void OnX::SDL_UI::put(const std::string& aString){
  // Append line at end of report region.
#ifdef __APPLE__
  fNS->put(aString.c_str());
#else
  ::printf("%s",aString.c_str());
#endif
}

void OnX::SDL_UI::synchronize(){}

void OnX::SDL_UI::exit(int){
  SoSDL::toQuit();
#ifdef __APPLE__
  fNS->exit();
#endif
}

void OnX::SDL_UI::warn(const std::string& aString){
  Lib::Out out(fSession.printer());
  out << "WARNING : " << aString << Lib::endl;
}

bool OnX::SDL_UI::parameterValue(const std::string&,std::string& aValue){
  aValue = "";
  return false;
}

bool OnX::SDL_UI::setParameter(const std::string& aWhat,const std::string& aValue){
  std::string aWidget;
  std::string aResource;
  if(!toWidgetAndResource(aWhat,aWidget,aResource)) return false;

  if(fVerboseLevel) {
    Lib::Out out(fSession.printer());
    out << "OnX::SDL_UI::setParameter :"
        << " widget " << Lib::sout(aWidget)
        << " resource " << Lib::sout(aResource)
        << " value " << Lib::sout(aValue) << "."
        << Lib::endl;
  }

  Lib::ItemML* itemML = findItemML(aWidget);
  if(!itemML) {
    Lib::Out out(fSession.printer());
    out << "OnX::SDL_UI::setParameter :"
        << " widget " << Lib::sout(aWidget)
        << " not found." 
        << Lib::endl;
    return false;
  }

  //std::string svalue;

  if(aResource=="foreground") {
    SoNode* soNode = (SoNode*)itemML->getData1();
    if(!soNode) return false;
    if(soNode->isOfType(SoSimpleButton::getClassTypeId())==TRUE) {
      SoSimpleButton* button = (SoSimpleButton*)soNode;
      SbColor sbColor;
      if(SbStyle::stringToColor(aValue.c_str(),sbColor)==FALSE) {
        Lib::Out out(fSession.printer());
        out << "OnX::SDL_UI::setParameter :"
            << " can't convert " << Lib::sout(aValue)
            << " to a SbColor." 
            << Lib::endl;
        return false;
      }
      button->getLabelStyle()->color.setValue(sbColor);
      return true;
    }
  } else if(aResource=="background") {
    SoNode* soNode = (SoNode*)itemML->getData1();
    if(!soNode) return false;
    if(soNode->isOfType(SoSimpleButton::getClassTypeId())==TRUE) {
      SoSimpleButton* button = (SoSimpleButton*)soNode;
      SbColor sbColor;
      if(SbStyle::stringToColor(aValue.c_str(),sbColor)==FALSE) {
        Lib::Out out(fSession.printer());
        out << "OnX::SDL_UI::setParameter :"
            << " can't convert " << Lib::sout(aValue)
            << " to a SbColor." 
            << Lib::endl;
        return false;
      }
      button->getBackStyle()->color.setValue(sbColor);
      return true;
    }
  } else {

    PageViewer* widget = (PageViewer*)itemML->getData1();
    if(!widget) return false;
    if(widget->setParameter(aResource,aValue)) return true;

  }

  Lib::Out out(fSession.printer());
  out << "OnX::SDL_UI::setParameter :"
      << " unable to set parameter for : widget " << Lib::sout(aWidget)
      << " widget " << Lib::sout(aWidget)
      << " resource " << Lib::sout(aResource)
      << " value " << Lib::sout(aValue) << "."
      << Lib::endl;

  return false;
}

Slash::UI::ICyclic* OnX::SDL_UI::createCyclic(
 const std::string& aName
,int aDelay
,const std::string& aInterp
,const std::string& aScript
){
  Slash::UI::ICyclic* cyclic = 
    new SDL_Cyclic(fSession,aName,aDelay,aInterp,aScript);
  fCyclics.push_back(cyclic);
  return cyclic;
}

bool OnX::SDL_UI::notify(Slash::Core::IMessage*){return false;}

bool OnX::SDL_UI::lock(){return false;}
bool OnX::SDL_UI::unlock(){return false;}

//////////////////////////////////////////////////////////////////////////////
/// BaseUI ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

int OnX::SDL_UI::nativeSteer() {
  if(!fUI_Region) return 0;
  //::printf("debug : OnX::SDL_UI::nativeSteer : begin\n");
#ifdef __APPLE__
 {std::string file = fNS->openFile();
  fNS->resetOpenFile();
  if(file.size()) {
    openFile(file); //Should be done after startup()
  }}
  fNS->steer(); //will do the SoSDL::mainLoop();
#else
  SoSDL::mainLoop();    
#endif
  //::printf("debug : OnX::SDL_UI::nativeSteer : end\n");
  return 0;
}

bool OnX::SDL_UI::createNativeWidget(
 Lib::ItemML& aItemML                      
,const std::string& aClass
,const std::string& aName
,Lib::ItemML* aParent
){
  if(!fUI_Region) return false;

  std::string svalue;

  if(aClass=="Window") {
  } else if(aClass=="MenuBar") {
  } else if(aClass=="CascadeButton") {
  } else if(aClass=="Menu") {

  } else if(aClass=="MenuItem") {

    int32_t columns = fUI_Region->columns.getValue();

    float h = 0.05F;
    float w = 0.1F;

    fUI_Region->columns.setValue(columns+1);
    fUI_Region->size.setValue(SbVec2f((columns+1)*w,h));

    SoSimpleButton* button = new SoSimpleButton;
    button->getBackStyle()->color.setValue(SbColor_white);
    //button->getLabelStyle()->color.setValue(SbColor(0.7F,0,7F,0.7F));
    //button->getLabelStyle()->fontName.setValue("Text3/defaultFont");
    button->getLabelStyle()->fontName.setValue("Hershey");
    //button->background.setValue(0.9F,0,0);
    //button->shadowThickness.setValue(0.1F);
    //button->marginWidth.setValue(0.1F);
    //button->marginHeight.setValue(0.1F);

    button->setName(aName.c_str());
    if(aItemML.propertyValue("label",svalue)){
      button->label.setValue(svalue.c_str());
    }
    addCallbacks(aItemML,button,"activate");

    fUI_Region->shapes.set1Value(columns,button);

    aItemML.setData((SoNode*)button,(SoNode*)fUI_Region,0);

  } else if(aClass=="PageViewer") {

    OnX::PageViewer* viewer = new OnX::PageViewer(*fSoViewer);

    SoNode* soNode = viewer->getSceneGraph();

    OnX::SoViewer* soViewer = Lib_Cast(aItemML,OnX::SoViewer);
    if(!soViewer) {
      Lib::Out out(fSession.printer());
      out << "SDL_UI::createNativeWidget :"
          << " ItemML not a OnX::SoViewer." 
          << Lib::endl;
      aItemML.printXML(fSession.printer(),"  ");
      return false;
    }

    soViewer->setViewerProxy(viewer,viewer_proxy< OnX::SoSDL_Viewer >);
    soViewer->setSceneGraph(soNode);

    fSceneGraph->insertChild(soNode,0); //something not in the other UI.

    aItemML.setData(viewer,viewer,0);

  } else {
    Lib::Out out(fSession.printer());
    out << "OnX::SDL_UI::createNativeWidget :"
        << " unknown widget type : " 
        << aClass 
        << Lib::endl;
    return false;
  }

  return createChildren(aItemML);
}

bool OnX::SDL_UI::lockScreen(
 const std::string& aPassword
,const std::string& aInterp
,const std::string& aScript
){
  if(!fSoViewer) return false;
  fUnlockInterp = aInterp;
  fUnlockScript = aScript;
  fSoViewer->lock(SbString(aPassword.c_str()),unlockCB,this);
  return true;
} 

void OnX::SDL_UI::unlockCB(void* aData){
  OnX::SDL_UI* This = (OnX::SDL_UI*)aData;
  This->executeScript(This->fUnlockInterp,This->fUnlockScript);
} 

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool OnX::SDL_UI::destroyNativeWidget(Lib::ItemML&) {return false;}

OnX::Action* OnX::SDL_UI::addCallback(Lib::ItemML&,
                           const std::string&,
                           const std::string&,
                           const std::string&) {
  return 0;
}
void OnX::SDL_UI::deleteCallback(Lib::ItemML&,const std::string&,Action*) {}
std::string OnX::SDL_UI::widgetName(Lib::ItemML&) {return "";}
void OnX::SDL_UI::showAll(Lib::ItemML&) {}
std::string OnX::SDL_UI::ask(const std::string&) {return "";}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void OnX::SDL_UI::addCallbacks(
 Lib::ItemML& aItemML
,SoSimpleButton* aNativeWidget
,const std::string& aNameXML
){
  Slash::UI::IWidget* widget = Lib_Cast(aItemML,Slash::UI::IWidget);
  const std::list<Lib::Element*> props = aItemML.properties();
  std::list<Lib::Element*>::const_iterator it;
  for(it=props.begin();it!=props.end();++it) {
    if(aNameXML==(*it)->name()) {
      std::string sexec;
      (*it)->attributeValue("exec",sexec);
      Action* action = new Action(*this,widget,aNativeWidget,
                                  sexec,(*it)->value(),(*it)->attributes(),
                                  aNameXML,0);
      addAction(action);
      aNativeWidget->addCallback(callback,(void*)action);
    }
  }
}

void OnX::SDL_UI::callback(void* aTag,SoNode*){
  // The generic callback.
  Action* action =(Action*)aTag;
  if(!action) return;

  SDL_UI* This = Lib_Cast(action->ui(),OnX::SDL_UI);
  if(!This) return;
  
  std::string svalue;
  std::string sevent;
  int cbkx = 0;
  int cbky = 0;

  SoSimpleButton* soWidget = (SoSimpleButton*)action->nativeWidget();

  // Retreive "value" :
  // Nothing for the moment.

  std::string name(soWidget->getName().getString());

  if(This->session().verboseLevel()) {
    Lib::Out out(This->session().printer());
    out << "OnX::SDL_UI::callback :"
        << " component " << Lib::sout(name)
     //   << " type " << Lib::sout(XClassGetName(xtClass))
     //   << " value " << Lib::sout(svalue)
     //   << " event " << sevent 
     //   << " x " << cbkx
     //   << " y " << cbky
        << "."        << Lib::endl;
  }

  This->fCallbackData = 
    CallbackData(action->widget(),name,svalue,sevent,cbkx,cbky);

  OnX::Action tmp(*action);
  tmp.execute();

  This->fCallbackData = CallbackData();
}

#ifdef ONX_HAS_INVENTOR
#include "../Core/receive_scene.cxx"
#endif

bool OnX::SDL_UI::receiveScene(
 const std::string& aWhere
,const std::string& aPlacement
,const std::string& aFile
)
{
#ifdef ONX_HAS_INVENTOR
  return receive_scene(*this,aWhere,aPlacement,aFile);
#else
  return false;
#endif
}
