#ifndef OnX_SoSDL_Viewer_h
#define OnX_SoSDL_Viewer_h

#include <Lib/smanip.h>

#include <Inventor/SDL/SoSDL.h> //WIN32 : Before coin include.

#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/actions/SoGLRenderAction.h>

#include <HEPVis/SDL/viewers/SoSDL_Viewer.h>

#include <HEPVis/SbColors.h>
#include <HEPVis/SbMath.h>
#include <HEPVis/nodekits/SoPage.h>
#include <HEPVis/nodekits/SoDisplayRegion.h>

namespace OnX {

class SoSDL_Viewer {
public: //Metods needed in OnX/Inventor/GuiViewer.h

  // The returned node is NOT the fSoSDLViewer.getSceneGraph ! 
  // (In general it is an SoPage).
  SoNode* getSceneGraph() const { return fSoNode;} 

  //virtual void setSceneGraph(SoNode* aNode) { fSoNode = aNode;}
  virtual void render() {fHEPVis_SoSDL_Viewer.render();}
  virtual void viewAll() {fHEPVis_SoSDL_Viewer.viewAll();}

  const SbColor& getBackgroundColor() const {
    return fHEPVis_SoSDL_Viewer.getBackgroundColor();
  }
  SoGLRenderAction* getGLRenderAction() const {
    return fHEPVis_SoSDL_Viewer.getGLRenderAction();
  }
  const SbViewportRegion& getViewportRegion() const {
    return fHEPVis_SoSDL_Viewer.getViewportRegion();
  }

  //virtual void setFeedbackVisibility(bool aValue) {}
  virtual void setAutoClipping(bool aValue) {}
  virtual void setDecoration(bool aValue) {}
  virtual void setViewing(bool aValue) {}
  virtual void setHeadlight(bool aValue) {}

  SbBool isDecoration(void) const {return FALSE;}
  SbBool isAutoClipping(void) const {return FALSE;}
  SbBool isHeadlight(void) const {return FALSE;}
  SbBool isViewing(void) const {return FALSE;}
  void scheduleRedraw(){}
  void setSize(const SbVec2s){}
  void setTitle(const char*){}
  void setBackgroundColor(const SbColor & color){
    if(!fSoNode) return;    
    if(fSoNode->isOfType(SoPage::getClassTypeId())==FALSE) return;
    SoPage* soPage = (SoPage*)fSoNode;
    SoRegion* soRegion = soPage->getHighlightedRegion();
    if(!soRegion) return;
    soRegion->color.setValue(color);
  }

  void setTransparencyType(SoGLRenderAction::TransparencyType a_type){
    SoGLRenderAction* action = fHEPVis_SoSDL_Viewer.getGLRenderAction();
    if(!action) return;
    action->setTransparencyType(a_type);
  }

  SoGLRenderAction::TransparencyType getTransparencyType() const {
    SoGLRenderAction* action = fHEPVis_SoSDL_Viewer.getGLRenderAction();
    if(!action) return SoGLRenderAction::SCREEN_DOOR;
    return action->getTransparencyType();
  }

public:
  SoSDL_Viewer( ::HEPVis::SoSDL_Viewer& aViewer)
  :fHEPVis_SoSDL_Viewer(aViewer)
  ,fSoNode(0){}
  virtual ~SoSDL_Viewer(){}
protected:
  SoSDL_Viewer(const SoSDL_Viewer& aFrom)
  :fHEPVis_SoSDL_Viewer(aFrom.fHEPVis_SoSDL_Viewer)
  ,fSoNode(0){}
  SoSDL_Viewer& operator=(const SoSDL_Viewer&) { return *this;}
protected:
  ::HEPVis::SoSDL_Viewer& fHEPVis_SoSDL_Viewer;
  SoNode* fSoNode; //should be a SoPage.
};

class SetRegionCallback : public virtual SbSetRegionCallback {
public:
  SetRegionCallback( ::SoSDL_Viewer& aViewer):fViewer(aViewer){}
  virtual void setCamera(SoCamera* aCamera) {
    fViewer.setCamera(aCamera);
  }
private:
  ::SoSDL_Viewer& fViewer;
};


class PageViewer : public SoSDL_Viewer {
public:
  PageViewer( ::HEPVis::SoSDL_Viewer& aViewer)
  : SoSDL_Viewer(aViewer) {

    SoPage* soPage = new SoPage;

    soPage->setSetRegionCallback
      (new SetRegionCallback(fHEPVis_SoSDL_Viewer));

    fSoNode = soPage;

    soPage->titleVisible.setValue(FALSE);
   // Disconnect SoCanvas mouse and keyboard defaults :
   {SoEventCallback* eventCallback = new SoEventCallback;
    if(soPage->setPart("callbackList[0]",eventCallback)==FALSE) 
      eventCallback->unref();}

    ///////////////////////////////////////////////
    // Scene :
    ///////////////////////////////////////////////
    SoDisplayRegion* land = 
      (SoDisplayRegion*)soPage->createRegion("SoDisplayRegion",0,0,1,1);
    //SoDisplayRegion* land = new SoDisplayRegion;
    land->color.setValue(SbColor_white);
    soPage->setCurrentRegion(land);

    ///////////////////////////////////////////////
    // Camera (us) :
    ///////////////////////////////////////////////
    land->setCameraType(SoPerspectiveCamera::getClassTypeId());

    SoPerspectiveCamera* soCamera = (SoPerspectiveCamera*)land->getCamera();
    soCamera->viewportMapping.setValue(SoCamera::ADJUST_CAMERA);
    float d = 100;
    soCamera->position.setValue(0,2,d); //y = 2 (meter)
    soCamera->focalDistance.setValue(d);
    soCamera->orientation.setValue(SbRotation(SbVec3f(0,1,0),0));
    soCamera->nearDistance.setValue(d*0.01F);
    soCamera->farDistance.setValue(100*d);
    soCamera->heightAngle.setValue(FM_PI/4.0F);

    fHEPVis_SoSDL_Viewer.setCamera(soCamera);
  }
  virtual ~PageViewer(){}
private:
  PageViewer(const PageViewer& aFrom)
  : SoSDL_Viewer(aFrom) {}
  PageViewer& operator=(const PageViewer&) { return *this;}
public:
  bool setParameter(const std::string& aWhat,const std::string& aValue) {
    if(aWhat=="deltaRotation") {
      float value;
      if(!Lib::smanip::tofloat(aValue,value)) return false;
      fHEPVis_SoSDL_Viewer.setDefaultDeltaRotation(value);
    } else if(aWhat=="deltaTranslation") {
      float value;
      if(!Lib::smanip::tofloat(aValue,value)) return false;
      fHEPVis_SoSDL_Viewer.setDefaultDeltaTranslation(value);

    } else if(aWhat=="showCursor") {
      bool value;
      if(!Lib::smanip::tobool(aValue,value)) return false;
      SDL_ShowCursor(value?1:00);

    } else if(aWhat=="showOptionsRegion") {
      bool value;
      if(!Lib::smanip::tobool(aValue,value)) return false;
      SoRegion* soRegion = fHEPVis_SoSDL_Viewer.optionRegion();
      if(soRegion) soRegion->visible.setValue(value?TRUE:FALSE);
    } else {
      return false;
    }
    return true;
  }
};

}

#endif
