//////////////////////////////////////////////////////////////////////////////
/// Pre SWIG /////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

%rename(pck) Slash::UI::IInterpreter::package; //package is a java keyword.

%ignore Slash::Data::IArray::vector();
// Avoid a clash with a finalize() produced by SWIG :
%ignore Slash::UI::IUI::finalize();

%rename(file_remove) Slash::file::remove(const std::string&);

#ifdef SWIGPYTHON
// FIXME : lxplus 64 bits. The code generated has a call to :
//   set(std::vector<unsigned int,std::allocator<unsigned int > > const &,
//    std::vector<unsigned long long,std::allocator<Slash::uint64 > > const &);
// and NOT : 
//   set(std::vector<unsigned int,std::allocator<unsigned int > > const &,
//    std::vector<Slash::uint64,std::allocator<Slash::uint64 > > const &);
// and then does not compile.

%ignore Slash::Core::IValue::set(const std::vector<unsigned int>&,
                                 const std::vector<Slash::uint64>&);
%ignore Slash::Core::IValue::set(const std::vector<unsigned int>&,
                                 const std::vector<Slash::int64>&);
#endif
//////////////////////////////////////////////////////////////////////////////
/// Code to wrap /////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

//FIXME : see the #ifdef SWIG in Slash/All.h
//        there are problems due to the fact that
//        IObject, IArray exists in two different places
%include <Slash/All.h>


#if defined(SWIGJAVA) || defined(SWIGTCL)
//FIXME : despite the upper All.h, SWIG java and Tcl needs that !
%include <Slash/Core/IManager.h>
%include <Slash/Core/ISession.h>
%include <Slash/UI/IUI.h>
%include <Slash/UI/IWidget.h>
%include <Slash/Data/IProcessor.h>
#endif

//////////////////////////////////////////////////////////////
/// Post SWIG ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////

// Put helpers as extensions :

%extend Slash::Core::IManager {
  Slash::Data::IProcessor* cast_Lib_IProcessor() {
    return Lib_Cast(*self,Slash::Data::IProcessor);
  }
}

%extend Slash::Core::ISession {
  std::string parameterValue(const std::string& aKey) {
    std::string s;
    self->parameterValue(aKey,s);
    return s;
  }
}

%extend Slash::Data::IProcessor {
/*void types() {
    data_collect(*self,"Lib_Type");
    data_dump(*self);
  }*/
  void collect(const std::string& aWhat,const std::string& aCuts = "") {
    std::vector<std::string> ss;
    ss.push_back(aWhat);
    ss.push_back(aCuts);
    self->execute("collect",ss);
  }
  void visualize(const std::string& where = "@current@") {
    std::vector<std::string> ss;
    ss.push_back("where="+where);
    self->execute("visualize",ss);
  }
  int number() {
    std::vector<std::string> ss;
    std::string s = self->execute("number",ss);
    int number;
    if(!Lib::smanip::toint(s,number)) return 0;
    return number;  
  }
  std::vector<std::string> values(const std::string& aWhat) {
    std::vector<std::string> ss;
    ss.push_back(aWhat);
    return Lib::smanip::words(self->execute("values",ss),"\n");
  }
  void dump(const std::string& aFormat = "table") {
    std::vector<std::string> ss;
    ss.push_back(aFormat);
    self->execute("dump",ss);
  }
  void filter(const std::string& aWhat,
                          const std::string& aFilter = "") {
    std::vector<std::string> ss;
    ss.push_back(aWhat);
    ss.push_back(aFilter);
    self->execute("filter",ss);
  }
  void set(const std::string& aWhat,
                       const std::string& aValue) {
    std::vector<std::string> ss;
    ss.push_back(aWhat);
    ss.push_back(aValue);
    self->execute("set",ss);
  }
  void destroy() {
    std::vector<std::string> ss;
    self->execute("destroy",ss);
  }
}

%define LIB_CAST(aClass)
  aClass* cast_##aClass () {
    return Lib_Cast(*self,aClass);
  }
%enddef

%extend Slash::Core::IManager {
  LIB_CAST(Slash::UI::IUI);
  LIB_CAST(Slash::UI::IScriptManager);
}

%extend Slash::Core::ISession {
  void setColor(double aR,double aG,double aB) {
    std::string s;
    Lib::smanip::printf(s,128,"%g %g %g",aR,aG,aB);
    self->setParameter("modeling.color",s);
  }
  void setColor(const std::string& aColor) {
    self->setParameter("modeling.color",aColor);
  }
  Slash::Data::IProcessor* accessorManager() {
    return Slash::accessorManager(*self);
  }
  Slash::UI::IScriptManager* scriptManager() {
    return Slash::scriptManager(*self);
  }
  Slash::UI::IUI_Manager* uiManager() {
    return Slash::uiManager(*self);
  }
  Slash::UI::IUI* ui() {
    Slash::UI::IUI_Manager* mgr = Slash::uiManager(*self);
    if(!mgr) return 0;
    return mgr->find("default");
  }
  Slash::UI::IUI* createUI(const std::string& a_file = "@empty@",bool a_show = true,const std::string& a_tk = "") {
    Slash::UI::IUI_Manager* mgr = Slash::uiManager(*self);
    if(!mgr) return 0;
    std::vector<std::string> args;
    return mgr->create(a_tk,a_file,args,a_show);
  }
  bool loadInterpreter(const std::string& aInterp,bool aQuiet = false) {
    Slash::UI::IScriptManager* sm = Slash::scriptManager(*self);
    if(!sm) return false;
    Slash::UI::IInterpreter* interp = sm->findInterpreter(aInterp);
    if(!interp) return false;
    return interp->load(aQuiet);
  }

}

%extend Slash::UI::IUI {
  std::string parameterValue(const std::string& aKey) {
    std::string s;
    self->parameterValue(aKey,s);
    return s;
  }
  bool setTreeItems(const std::string& aTreeWidget,const std::string& aFile) {
    std::vector<std::string> lines;
    if(!Lib::fmanip::read(aFile,lines)) return false;
    self->setParameter(aTreeWidget+".items",Lib::smanip::tostring(lines,"\n"));
    return true;
  }
  void showHelp(const std::string& aFile) {
    self->executeScript("DLD","OnX ui_help_file "+aFile);
  }
  void showDialog(const std::string& aFile,
                  const std::string& aDialog) {
    self->executeScript("DLD","OnX ui_show_dialog "+aFile+" "+aDialog);
  } 
}

#ifdef SWIGPYTHON
%pythoncode {
def create_session(a_args = []):
  set_env(Slash_cout())
  session = Slash_create_session(a_args)
  session.loadInterpreter('Python');
  return session
}
#endif

%inline {
std::ostream& Slash_cout() {return std::cout;}
std::ostream& Slash_cerr() {return std::cerr;}
}
