/*
 * Example of an OnX client program.
 */

#include <Slash/Tools/system.h>
#include <Slash/Tools/args.h>

#include <Lib/InternetSocket.h>

#include <OnX/Net/Protocol.h>

#include <iostream>
#include <cstdlib>

int main(int aArgc,char** aArgv){

  int ret_status = EXIT_SUCCESS;

 {Slash::args args(aArgc,aArgv);

  int verbose = 0;
  if(args.is_arg("-verbose")) verbose = 1;

  std::string host;
  if(!args.find("-host",host)) {
    std::cout << "argument -host=<string> not found."
              << std::endl;
    return EXIT_FAILURE;
  }

  unsigned int port;
  if(!args.find<unsigned int>("-port",port)) port = 50701;
  
  Lib::InternetSocket client(std::cout,host,port,verbose);

  if(!client.connect()) {
    std::cout << "OnX_client : "
              << " can't connect to server."
              << std::endl;
    return EXIT_FAILURE;
  }

  if(!client.send_string("")) { //default GUI driver.
    std::cout << "OnX_client : "
              << " can't send empty string to server."
              << std::endl;
    return EXIT_FAILURE;
  }

 {std::string OSC_HOME_DIR;
  if(!Slash::getenv("OSC_HOME_DIR",OSC_HOME_DIR)) {
    std::cout << "OnX_client :"
          << " Environment variable OSC_HOME_DIR not defined."
          << " You have to setup OpenScientist by sourcing "
          << " the setup script, found in the distribution,"
          << " corresponding to your kind of shell."
          << std::endl;
    return EXIT_FAILURE;
  }
  std::string fs = Slash::file_separator();
  std::string ONXROOT = OSC_HOME_DIR + fs + "Resources" + fs + "OnX";        
  if(!Slash::putenv("ONXROOT",ONXROOT)) {
    std::cout << "OnX_client : "
              << " can't putenv ONXROOT."
              << std::endl;
    return EXIT_FAILURE;
  }
  //std::string GUI("$ONXROOT/examples/OnX/hello_world.onx");
  std::string GUI("$ONXROOT/tests/onx_tests/main.onx");
  //std::string GUI = ONXROOT+fs+"tests"+fs+"onx_tests"+fs+"main.onx";        
  if(!client.send_string(GUI)) {
    std::cout << "OnX_client : "
              << " can't send string to server."
              << std::endl;
    return EXIT_FAILURE;
  }}

  // Treat events coming from the OnX server :
  std::string s;
  std::string what;
  std::string cmd;
  while(1) {
    std::cout << "debug : go to waitString..." << std::endl;
    if(!client.wait_string(s)) {
      return EXIT_FAILURE;
    }
    std::cout << "debug : from server " << s << std::endl;

    if(!OnX_protocol(s,what,cmd)) {
      std::cout << "OnX_client : "
                << " bad syntax in " << Slash::sout(s) 
                << std::endl;
      client.disconnect();
      ret_status = EXIT_FAILURE;
      break;
    }

    if(what==ONX_PROTOCOL_FILE) {
      std::string file = Slash::file_name(cmd);
      client.send_string(ONX_PROTOCOL_FILE+":"+file);
      if(!client.send_file(file)) {
        std::cout << "OnX_client : "
                  << " can't send file " << Slash::sout(file) << " to server."
                  << std::endl;
        ret_status = EXIT_FAILURE;
        break;
      }
    } else if(what==ONX_PROTOCOL_EXIT) {
      break;

    } else {
       std::cout << "OnX_client :"
                 << " receive event : " << Slash::sout(s) << " from server."
                 << std::endl;
    }
  }

  }

  return ret_status;
}
