// this :
#include "Accessors.h"

// AIDA :
#include <AIDA/IPlotter.h>
#include <AIDA/IPlotterRegion.h>
#include <AIDA/IBaseHistogram.h>
#include <AIDA/IConstants.h>

// Slash :
#include <Slash/Core/ISession.h>
#include <Slash/Data/IProcessor.h>
#include <Slash/Data/IIterator.h>
#include <Slash/Tools/AIDA.h>

// amanip :
#include <amanip/create.h>

// Lib :
#include <Lib/Property.h>
#include <Lib/Value.h>
#include <Lib/Manager.h>
#include <Lib/smanip.h>
#include <Lib/Out.h>
#include <Lib/Cast.h>
#include <Lib/Arguments.h>

// HEPVis :
#include <HEPVis/SbPlottableThings.h> //Interfaces

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
OnXLab::SbObject_Accessor::SbObject_Accessor(Slash::Core::ISession& aSession)
:Lib::BaseAccessor(aSession.printer())
,fSession(aSession)
,fName("SbObject")
{
  addProperty("name",Lib::Property::STRING);
  addProperty("title",Lib::Property::STRING);
}
OnXLab::SbObject_Accessor::~SbObject_Accessor(){}
std::string OnXLab::SbObject_Accessor::name() const{return fName;}

Slash::Data::IIterator* OnXLab::SbObject_Accessor::iterator(){return 0;}

Slash::Core::IValue* OnXLab::SbObject_Accessor::findValue(Slash::Data::IAccessor::Data aData,const std::string& aName,void*){
  SbPlottableObject* object = (SbPlottableObject*)aData;

  if(aName=="name") {
    return new Lib::Value(object->getName());
  } else if(aName=="title") {
    return new Lib::Value(object->getLabel());
  } else {
    return new Lib::Value();
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
OnXLab::SbFunction1D_Accessor::SbFunction1D_Accessor(Slash::Core::ISession& aSession)
:Lib::BaseAccessor(aSession.printer())
,fSession(aSession)
,fName("SbFunction1D")
{
  addProperty("name",Lib::Property::STRING);
  addProperty("title",Lib::Property::STRING);
  addProperty("dimension",Lib::Property::INTEGER);
  addProperty("#params",Lib::Property::INTEGER);
  addProperty("codelet",Lib::Property::STRING);
  addProperty("address",Lib::Property::POINTER);
  addProperty("params",Lib::Property::STRING);
}

OnXLab::SbFunction1D_Accessor::~SbFunction1D_Accessor(){}
std::string OnXLab::SbFunction1D_Accessor::name() const{return fName;}
Slash::Data::IIterator* OnXLab::SbFunction1D_Accessor::iterator(){return 0;}

Slash::Core::IValue* OnXLab::SbFunction1D_Accessor::findValue(Slash::Data::IAccessor::Data aData,const std::string& aName,void*){
  SbPlottableFunction1D* object = (SbPlottableFunction1D*)aData;

  if(!object->isValid()) return new Lib::Value();

  //NOTE : special usage of cast() :
  std::string scast = std::string(Sb_data)+"::AIDA::IFunction";
  AIDA::IFunction* data = (AIDA::IFunction*)object->cast(scast.c_str());
  if(!data) {
    fSession.cout() << "OnXLab::SbFunction1D_Accessor::findValue :"
                    << " cast(\"data::AIDA::IFunction\") failed."
                    << std::endl;
    return new Lib::Value();
  }

  if(aName=="name") {
    return new Lib::Value(object->getName());
  } else if(aName=="title") {
    return new Lib::Value(data->title());
  } else if(aName=="dimension") {
    return new Lib::Value(data->dimension());
  } else if(aName=="#params") {
    return new Lib::Value(data->numberOfParameters());
  } else if(aName=="codelet") {
    return new Lib::Value(data->codeletString());
  } else if(aName=="address") {
    return new Lib::Value((void*)data);
  } else if(aName=="params") {
    std::vector<std::string> names = data->parameterNames();
    std::vector<double> vals = data->parameters();
    std::string value;
    unsigned int number = names.size();
    for(unsigned int index=0;index<number;index++){
      std::string s;
      Lib::smanip::printf(s,32,"%g",vals[index]);
      value += names[index] + " = " + s + ";";
    }
    return new Lib::Value(value);
  } else {
    return new Lib::Value();
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
OnXLab::SbFunction2D_Accessor::SbFunction2D_Accessor(Slash::Core::ISession& aSession)
:Lib::BaseAccessor(aSession.printer())
,fSession(aSession)
,fName("SbFunction2D")
{
  addProperty("name",Lib::Property::STRING);
  addProperty("title",Lib::Property::STRING);
  addProperty("dimension",Lib::Property::INTEGER);
  addProperty("#params",Lib::Property::INTEGER);
  addProperty("codelet",Lib::Property::STRING);
  addProperty("address",Lib::Property::POINTER);
  addProperty("params",Lib::Property::STRING);
}

OnXLab::SbFunction2D_Accessor::~SbFunction2D_Accessor(){}
std::string OnXLab::SbFunction2D_Accessor::name() const{return fName;}
Slash::Data::IIterator* OnXLab::SbFunction2D_Accessor::iterator(){return 0;}

Slash::Core::IValue* OnXLab::SbFunction2D_Accessor::findValue(Slash::Data::IAccessor::Data aData,const std::string& aName,void*){
  SbPlottableFunction2D* object = (SbPlottableFunction2D*)aData;

  if(!object->isValid()) return new Lib::Value();

  //NOTE : special usage of cast() :
  std::string scast = std::string(Sb_data)+"::AIDA::IFunction";
  AIDA::IFunction* data = (AIDA::IFunction*)object->cast(scast.c_str());
  if(!data) {
    fSession.cout() << "OnXLab::SbFunction2D_Accessor::findValue :"
                    << " cast(\"data::AIDA::IFunction\") failed."
                    << std::endl;
    return new Lib::Value();
  }

  if(aName=="name") {
    return new Lib::Value(object->getName());
  } else if(aName=="title") {
    return new Lib::Value(data->title());
  } else if(aName=="dimension") {
    return new Lib::Value(data->dimension());
  } else if(aName=="#params") {
    return new Lib::Value(data->numberOfParameters());
  } else if(aName=="codelet") {
    return new Lib::Value(data->codeletString());
  } else if(aName=="address") {
    return new Lib::Value((void*)data);
  } else if(aName=="params") {
    std::vector<std::string> names = data->parameterNames();
    std::vector<double> vals = data->parameters();
    std::string value;
    unsigned int number = names.size();
    for(unsigned int index=0;index<number;index++){
      std::string s;
      Lib::smanip::printf(s,32,"%g",vals[index]);
      value += names[index] + " = " + s + ";";
    }
    return new Lib::Value(value);
  } else {
    return new Lib::Value();
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
OnXLab::SbBin1D_Accessor::SbBin1D_Accessor(Slash::Core::ISession& aSession)
:Lib::BaseAccessor(aSession.printer())
,fSession(aSession)
,fName("SbBin1D")
{
  addProperty("index",Lib::Property::INTEGER);
  addProperty("height",Lib::Property::DOUBLE);
  addProperty("error",Lib::Property::DOUBLE);
  addProperty("entries",Lib::Property::INTEGER);
  addProperty("lowerEdge",Lib::Property::DOUBLE);
  addProperty("upperEdge",Lib::Property::DOUBLE);
  addProperty("name",Lib::Property::STRING);
}
OnXLab::SbBin1D_Accessor::~SbBin1D_Accessor(){}
std::string OnXLab::SbBin1D_Accessor::name() const{return fName;}
Slash::Data::IIterator* OnXLab::SbBin1D_Accessor::iterator(){return 0;}

Slash::Core::IValue* OnXLab::SbBin1D_Accessor::findValue(Slash::Data::IAccessor::Data aData,const std::string& aName,void* aTag){
  SbPlottableBins1D* object = (SbPlottableBins1D*)aData;
  if(!object->isValid()) {
    Lib::Out out(fSession.printer());
    out << "OnXLab::SbBin1D_Accessor::value :"
        << " data object associated to plotted representation " 
        << "is no more valid." << Lib::endl;
    return new Lib::Value(); 
  }
  // aTag is a char* (see SoPlotter).
  std::string infos((char*)aTag);
  std::vector<int> vals;
  if(!Lib::smanip::toints(infos,vals)) return new Lib::Value();
  if(vals.size()!=1) return new Lib::Value();
  int i = vals[0];
  if(aName=="index") {
    return new Lib::Value(i);
  } else if(aName=="height") {
    return new Lib::Value((double)object->getBinSumOfWeights(i));
  } else if(aName=="error") {
    return new Lib::Value((double)object->getBinBar(i));
  } else if(aName=="entries") {
    return new Lib::Value(object->getBinNumberOfEntries(i));
  } else if(aName=="lowerEdge") {
    return new Lib::Value((double)object->getBinLowerEdge(i));
  } else if(aName=="upperEdge") {
    return new Lib::Value((double)object->getBinUpperEdge(i));
  } else if(aName=="name") {
    return new Lib::Value(std::string(object->getName()));
  } 
  return new Lib::Value();
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
OnXLab::SbBin2D_Accessor::SbBin2D_Accessor(Slash::Core::ISession& aSession)
:Lib::BaseAccessor(aSession.printer())
,fSession(aSession)
,fName("SbBin2D")
{
  addProperty("indexX",Lib::Property::INTEGER);
  addProperty("indexY",Lib::Property::INTEGER);
  addProperty("height",Lib::Property::DOUBLE);
  addProperty("error",Lib::Property::DOUBLE);
  addProperty("entries",Lib::Property::INTEGER);
  addProperty("lowerEdgeX",Lib::Property::DOUBLE);
  addProperty("upperEdgeX",Lib::Property::DOUBLE);
  addProperty("lowerEdgeY",Lib::Property::DOUBLE);
  addProperty("upperEdgeY",Lib::Property::DOUBLE);
  addProperty("name",Lib::Property::STRING);
}

OnXLab::SbBin2D_Accessor::~SbBin2D_Accessor(){}
std::string OnXLab::SbBin2D_Accessor::name() const{return fName;}
Slash::Data::IIterator* OnXLab::SbBin2D_Accessor::iterator(){return 0;}

Slash::Core::IValue* OnXLab::SbBin2D_Accessor::findValue(Slash::Data::IAccessor::Data aData,const std::string& aName,void* aTag){
  SbPlottableBins2D* object = (SbPlottableBins2D*)aData;
  if(!object->isValid()) {
    Lib::Out out(fSession.printer());
    out << "OnXLab::SbBin2D_Accessor::value :"
        << " data object associated to plotted representation " 
        << "is no more valid." << Lib::endl;
    return new Lib::Value(); 
  }
  // aTag is a char* (see SoPlotter).
  std::string infos((char*)aTag);
  std::vector<int> vals;
  if(!Lib::smanip::toints(infos,vals)) return new Lib::Value();
  if(vals.size()!=2) return new Lib::Value();
  int i = vals[0];
  int j = vals[1];
  if(aName=="indexX") {
    return new Lib::Value(i);
  } else if(aName=="indexY") {
    return new Lib::Value(j);
  } else if(aName=="height") {
    return new Lib::Value((double)object->getBinSumOfWeights(i,j));
  } else if(aName=="error") {
    return new Lib::Value((double)object->getBinBar(i,j));
  } else if(aName=="entries") {
    return new Lib::Value(object->getBinNumberOfEntries(i,j));
  } else if(aName=="lowerEdgeX") {
    return new Lib::Value((double)object->getBinLowerEdgeX(i));
  } else if(aName=="upperEdgeX") {
    return new Lib::Value((double)object->getBinUpperEdgeX(i));
  } else if(aName=="lowerEdgeY") {
    return new Lib::Value((double)object->getBinLowerEdgeY(j));
  } else if(aName=="upperEdgeY") {
    return new Lib::Value((double)object->getBinUpperEdgeY(j));
  } else if(aName=="name") {
    return new Lib::Value(std::string(object->getName()));
  } 
  return new Lib::Value();
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// AIDA :
#include <AIDA/ITree.h>
#include <AIDA/IManagedObject.h>

// BatchLab :
#include <BatchLab/Core/MemoryTree.h>

OnXLab::ManagedObjectAccessor::ManagedObjectAccessor(Slash::Core::ISession& aSession)
:Lib::BaseAccessor(aSession.printer())
,fSession(aSession)
,fName("ManagedObject")
{
  addProperty("name",Lib::Property::STRING);
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
  addProperty("type",Lib::Property::STRING,20);
#endif
  addProperty("address",Lib::Property::POINTER);
}
OnXLab::ManagedObjectAccessor::~ManagedObjectAccessor(){}
std::string OnXLab::ManagedObjectAccessor::name() const{return fName;}

namespace OnXLab {
  class ManagedObjectIterator : public Slash::Data::IIterator {
  public: //Slash::Data::IIterator
    virtual Slash::Data::IAccessor::Data object() {
      if(fIterator==fList.end()) return 0;
      return *fIterator;
    }
    virtual void next() { ++fIterator;}
    virtual void* tag() { return 0;}
  public:
    ManagedObjectIterator(const std::vector<AIDA::IManagedObject*>& aList):fList(aList){
      fIterator = fList.begin();
    }
  private:
    std::vector<AIDA::IManagedObject*> fList; // Local copy.
    std::vector<AIDA::IManagedObject*>::iterator fIterator;
  };
}

Slash::Data::IIterator* OnXLab::ManagedObjectAccessor::iterator(){
  // Returned object should be deleted by the receiver.
  BatchLab::MemoryTree* memoryTree = 
    Lib_findManager(fSession,"MemoryTree",BatchLab::MemoryTree);
  if(!memoryTree) {
    Lib::Out out(fSession.printer());
    out << "OnXLab::ManagedObjectAccessor::iterator :"
        << " MemoryTree not found." << Lib::endl;
    return 0;
  }
  std::vector<AIDA::IManagedObject*> vec = memoryTree->allObjects();
  return new ManagedObjectIterator(vec);
}

Slash::Core::IValue* OnXLab::ManagedObjectAccessor::findValue(Slash::Data::IAccessor::Data aData,const std::string& aName,void*){
  AIDA::IManagedObject* object = (AIDA::IManagedObject*)aData;
  if(aName=="name") {
    return new Lib::Value(object->name());
#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
  } else if(aName=="type") {
    return new Lib::Value(object->type());
#endif
  } else if(aName=="address") {
    return new Lib::Value((void*)object);
  } else {
    return new Lib::Value();
  }
}
void OnXLab::ManagedObjectAccessor::beginVisualize() {}
void OnXLab::ManagedObjectAccessor::endVisualize() {}
void OnXLab::ManagedObjectAccessor::visualize(Slash::Data::IAccessor::Data aData,void*){
  std::string where;
 {const std::vector<std::string>& eargs = executeArguments();
  if(eargs.size()) {
    Lib::Arguments args(eargs);
    args.find("where",where);
  }}

  AIDA::IAnalysisFactory* aida = Slash::find_AIDA(fSession);
  if(!aida) return;
  AIDA::IPlotter* plotter = amanip::create_plotter(*aida,where);        
  if(!plotter) return;
  AIDA::IPlotterRegion& region = plotter->currentRegion();
  AIDA::IManagedObject* object = (AIDA::IManagedObject*)aData;
  if(AIDA::IBaseHistogram* h = Lib_Cast(*object,AIDA::IBaseHistogram)) {
    region.plot(*h);
  } else if(AIDA::IFunction* f = Lib_Cast(*object,AIDA::IFunction)) {
    region.plot(*f);
  }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
OnXLab::FunctionAccessor::FunctionAccessor(Slash::Core::ISession& aSession)
:Lib::BaseAccessor(aSession.printer())
,fSession(aSession)
,fName("Function")
{
  addProperty("name",Lib::Property::STRING);
  addProperty("title",Lib::Property::STRING);
  addProperty("dimension",Lib::Property::INTEGER);
  addProperty("#params",Lib::Property::INTEGER);
  addProperty("codelet",Lib::Property::STRING);
  addProperty("address",Lib::Property::POINTER);
  addProperty("params",Lib::Property::STRING);
}
OnXLab::FunctionAccessor::~FunctionAccessor(){}
std::string OnXLab::FunctionAccessor::name() const{return fName;}

namespace OnXLab {
  class FunctionIterator : public Slash::Data::IIterator {
  public: //Slash::Data::IIterator
    virtual Slash::Data::IAccessor::Data object() {
      if(fIterator==fList.end()) return 0;
      return *fIterator;
    }
    virtual void next() { ++fIterator;}
    virtual void* tag() { return 0;}
  public:
    FunctionIterator(const std::vector<AIDA::IFunction*>& aList):fList(aList){
      fIterator = fList.begin();
    }
  private:
    std::vector<AIDA::IFunction*> fList; // Local copy.
    std::vector<AIDA::IFunction*>::iterator fIterator;
  };
}

Slash::Data::IIterator* OnXLab::FunctionAccessor::iterator(){
  BatchLab::MemoryTree* memoryTree = 
    Lib_findManager(fSession,"MemoryTree",BatchLab::MemoryTree);
  if(!memoryTree) {
    Lib::Out out(fSession.printer());
    out << "OnXLab::FunctionAccessor::iterator :"
        << " MemoryTree not found." << Lib::endl;
    return 0;
  }
  std::vector<AIDA::IManagedObject*> vec = memoryTree->allObjects();
  std::vector<AIDA::IFunction*> funcs;
  for(unsigned int index=0;index<vec.size();index++) {
    AIDA::IFunction* obj = Lib_Cast(*(vec[index]),AIDA::IFunction);
    if(obj) funcs.push_back(obj);
  }
  return new FunctionIterator(funcs);
}
Slash::Core::IValue* OnXLab::FunctionAccessor::findValue(Slash::Data::IAccessor::Data aData,const std::string& aName,void*){
  AIDA::IFunction* object = (AIDA::IFunction*)aData;
  if(aName=="name") {
    AIDA::IManagedObject* mobject = Lib_Cast(*object,AIDA::IManagedObject);
    return new Lib::Value(mobject?mobject->name():"none");
  } else if(aName=="title") {
    return new Lib::Value(object->title());
  } else if(aName=="dimension") {
    return new Lib::Value(object->dimension());
  } else if(aName=="#params") {
    return new Lib::Value(object->numberOfParameters());
  } else if(aName=="codelet") {
    return new Lib::Value(object->codeletString());
  } else if(aName=="address") {
    return new Lib::Value((void*)object);
  } else if(aName=="params") {
    std::vector<std::string> names = object->parameterNames();
    std::vector<double> vals = object->parameters();
    std::string value;
    unsigned int number = names.size();
    for(unsigned int index=0;index<number;index++){
      std::string s;
      Lib::smanip::printf(s,32,"%g",vals[index]);
      value += names[index] + " = " + s + ";";
    }
    return new Lib::Value(value);
  } else {
    return new Lib::Value();
  }
}
void OnXLab::FunctionAccessor::endVisualize() {}
void OnXLab::FunctionAccessor::beginVisualize() {}
void OnXLab::FunctionAccessor::visualize(Slash::Data::IAccessor::Data aData,void*){
  std::string where;
 {const std::vector<std::string>& eargs = executeArguments();
  if(eargs.size()) {
    Lib::Arguments args(eargs);
    args.find("where",where);
  }}

  AIDA::IAnalysisFactory* aida = Slash::find_AIDA(fSession);
  if(!aida) return;
  AIDA::IPlotter* plotter = amanip::create_plotter(*aida,where);        
  if(!plotter) return;
  AIDA::IPlotterRegion& region = plotter->currentRegion();
  AIDA::IFunction* object = (AIDA::IFunction*)aData;
  region.plot(*object);
}

