// this :
#include <OnXLab/Core/TreeFactory.h>

#include <Slash/Core/IRelation.h>
#include <Slash/Core/IRelationManager.h>
#include <Slash/Tools/Managers.h>
#include <Slash/Tools/UI.h>

#include <Lib/Debug.h>
#include <Lib/Cast.h>
#include <Lib/Messages.h>

#include <Slash/UI/IUI.h>

#include <BatchLab/Core/BaseTree.h> //FIXME
#include <BatchLab/Core/MemoryTree.h> //FIXME

OnXLab::TreeFactory::TreeFactory(Slash::Core::ISession& aSession)
:BatchLab::TreeFactory(aSession){
  Lib::Debug::increment("OnXLab::TreeFactory");
}

OnXLab::TreeFactory::~TreeFactory(){
  Lib::Debug::decrement("OnXLab::TreeFactory");
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

AIDA::ITree* OnXLab::TreeFactory::create_mem(){
  AIDA::ITree* tree =  BatchLab::TreeFactory::create();
  if(!tree) return 0;
  setUIRelation(*tree);
  /*
  AIDA::ITree* memoryTree = 
    Slash_findManager(fSession,"MemoryTree",AIDA::ITree);
  if(memoryTree) return memoryTree;
  BatchLab::MemoryTree* tree = new BatchLab::MemoryTree(fSession,"MemoryTree");
  tree->setVerboseLevel(fSession.verboseLevel());
  fSession.addManager(tree);
  //WARNING : the UI relation is done once for the mem tree,
  //          then the UI should be here at first call.
  */
  return tree;
  
}

AIDA::ITree* OnXLab::TreeFactory::create(){return create_mem();}

AIDA::ITree* OnXLab::TreeFactory::createTree(const std::string&){return create_mem();}

#if (AIDA_VERSION_MAJOR<=3) && (AIDA_VERSION_MINOR<=2) 
#else
AIDA::ITree* OnXLab::TreeFactory::createTree(
 const std::string& aStoreName
,const std::string& aStoreType
,int aMode
,const std::string& aOptions
){
  AIDA::ITree* tree =  
    BatchLab::TreeFactory::createTree(aStoreName,aStoreType,aMode,aOptions);
  if(!tree) return 0;
  setUIRelation(*tree);

  //FIXME : we should be able to pass through AIDA :
  BatchLab::BaseTree* labTree = Lib_Cast(*tree,BatchLab::BaseTree);
  if(labTree) labTree->emitUpdate();

  return tree;
}
AIDA::ITree* OnXLab::TreeFactory::createNamedTree(
 const std::string& aName
,const std::string& aStoreName
,const std::string& aStoreType
,int aMode
,const std::string& aOptions
){
  AIDA::ITree* tree =  
    BatchLab::TreeFactory::createNamedTree
      (aName,aStoreName,aStoreType,aMode,aOptions);
  if(!tree) return 0;
  setUIRelation(*tree);

  //FIXME : we should be able to pass through AIDA :
  BatchLab::BaseTree* labTree = Lib_Cast(*tree,BatchLab::BaseTree);
  if(labTree) labTree->emitUpdate();

  return tree;
}
#endif
AIDA::ITree* OnXLab::TreeFactory::create(
 const std::string& aStoreName
,const std::string& aStoreType
,bool aReadOnly
,bool aCreateNew
,const std::string& aOptions
){
  AIDA::ITree* tree =  
    BatchLab::TreeFactory::create
      (aStoreName,aStoreType,aReadOnly,aCreateNew,aOptions);
  if(!tree) return 0;
  setUIRelation(*tree);

  //FIXME : we should be able to pass through AIDA :
  BatchLab::BaseTree* labTree = Lib_Cast(*tree,BatchLab::BaseTree);
  if(labTree) labTree->emitUpdate();

  return tree;
}

namespace OnXLab {
  class AIDA_ITree_IUI_Relation : public virtual Slash::Core::IRelation {
  public: // IRelation :
    virtual std::string name() const {return fName;}
    virtual void* cast(const std::string& aTo) const {
      if(aTo=="OnXLab::AIDA_ITree_IUI_Relation")  {
        return Lib_SCast(OnXLab::AIDA_ITree_IUI_Relation);
      } else if(aTo=="Slash::Core::IRelation") {
        return Lib_SCast(Slash::Core::IRelation);
      } else {
        return 0;
      }
    }
    virtual bool schedule() {
      if(fScheduled) return true;
      std::string value;
      if(!fUI.session().parameterValue("OnXLab.tree",value)) {
        std::ostream& out = fUI.session().cout();
        out << "OnXLab::AIDA_ITree_IUI_Relation::schedule :"
            << " session parameter OnXLab.tree not found."
            << std::endl;
        return false;
      }
      if(value.empty()) {
        std::ostream& out = fUI.session().cout();
        out << "OnXLab::AIDA_ITree_IUI_Relation::schedule :"
            << " OnXLab.tree variable is empty."
            << std::endl;
        return false;
      }
      //::printf("debug : OnXLab::TreeUI_Relation::notify : widget \"%s\"\n",
      //    value.c_str());
      if(value!="none") {
        if(!fUI.notify(new Lib::UpdateMessage(value,this))) return false;
      }
      fScheduled = true;
      return true;
    }
    virtual void unSchedule() { 
      fScheduled = false;
    }
    virtual void* first() const { return &fTree;}
    virtual void* second() const { return &fUI;}
  public:
    AIDA_ITree_IUI_Relation(AIDA::ITree& aTree,Slash::UI::IUI& aUI)
    :fName("OnXLab::AIDA_ITree_IUI_Relation")
    ,fScheduled(false)
    ,fTree(aTree),fUI(aUI){
      Lib::Debug::increment("OnXLab::AIDA_ITree_IUI_Relation");
      //printf("debug : OnXLab::AIDA_ITree_IUI_Relation::AIDA_ITree_IUI_Relation : %ld\n",this);
    }
    virtual ~AIDA_ITree_IUI_Relation(){
      //printf("debug : OnXLab::AIDA_ITree_IUI_Relation::~AIDA_ITree_IUI_Relation : %ld\n",this);
      Lib::Debug::decrement("OnXLab::AIDA_ITree_IUI_Relation");
    }
  private:
    std::string fName;
    bool fScheduled;
    AIDA::ITree& fTree;
    Slash::UI::IUI& fUI;
  };
}

void OnXLab::TreeFactory::setUIRelation(AIDA::ITree& aTree){
  Slash::Core::IRelationManager* mgr = Slash::relationManager(fSession);
  if(!mgr) {
    //std::ostream& out = fSession.cout();
    //out << "OnXLab::TreeFactory::setUIRelation :"
    //    << " RelationManager not found."
    //    << std::endl;
    return;
  }
  Slash::UI::IUI* ui = Slash::find_UI(fSession);
  if(!ui) {
    std::ostream& out = fSession.cout();
    //out << "OnXLab::TreeFactory::setUIRelation :"
    //    << " UI not found."
    //    << std::endl;
    return;
  }
  mgr->add(new OnXLab::AIDA_ITree_IUI_Relation(aTree,*ui));
  //FIXME : aTree emitUpdate about its creation ?
}
