// this :
#include <OnXLab/Inventor/PlotterManager.h>

// Slash :
#include <Slash/Core/ISession.h>
#include <Slash/Core/IWriter.h>
#include <Slash/UI/IUI.h>
#include <Slash/UI/IWidget.h>
#include <Slash/Tools/Managers.h>
#include <Slash/Tools/UI.h>
#include <Slash/Tools/args.h>
#include <Slash/Tools/system.h>
#include <Slash/Tools/file.h>

// Lib:
#include <Lib/Cast.h>

#include <OnXLab/Inventor/Plotter.h>

#include <HEPVis/nodekits/SoPage.h>

#include <OnXLab/Inventor/SimplePlotter_onx.h>

OnXLab::PlotterManager::PlotterManager(Slash::Core::ISession& aSession)
:fName("OnXLab_PlotterManager")
,fSession(aSession)
{}

OnXLab::PlotterManager::~PlotterManager(){}

std::string OnXLab::PlotterManager::name() const{return fName;}

void* OnXLab::PlotterManager::cast(const std::string& aClass) const{
       if_Lib_SCast(OnXLab::PlotterManager)
  else if_Lib_SCast(Slash::UI::IPlotterManager)
  else if_Lib_SCast(Slash::Core::IManager)
  else return 0;
}

bool OnXLab::PlotterManager::setArguments(const std::vector<std::string>& aArgs){fArgs = aArgs;return true;}

Slash::UI::IPlotter* OnXLab::PlotterManager::create(const std::string& aName){
  Slash::UI::IUI* ui = loadUI();
  if(!ui) {
    std::ostream& out = fSession.cout();
    out << "OnXLab::PlotterManager::create :"
        << " no UI found." 
        << std::endl;
    return 0;
  }
  Slash::UI::IWidget* w = 0;

  if(aName.size()) {
    if(aName==Slash::s_current()) {
      w = ui->currentWidget();
      if(!w) {
        std::ostream& out = fSession.cout();
        out << "OnXLab::PlotterManager::create :"
          << " no current widget."
          << std::endl;
      }
    } else {
      w = ui->findWidget(aName);
      if(!w) {
        std::ostream& out = fSession.cout();
        out << "OnXLab::PlotterManager::create :"
            << " widget " << Slash::sout(aName) << " not found." 
            << " Try to find a widget named \"Plotter\""
            << " or take the currentWidget."
            << std::endl;
      }
    }
  }
  if(!w) {
    w = ui->findWidget("Plotter");
    if(!w) {
      w = ui->currentWidget();
      if(!w) {
        std::ostream& out = fSession.cout();
        out << "OnXLab::PlotterManager::create :"
            << " no current widget."
            << std::endl;
      }
    }

  }
  if(!w) return 0;

  OnXLab::Plotter* plotter = new OnXLab::Plotter
    (fSession.printer(),*w,*ui,fSession.verboseLevel());
  if(!plotter->isValid()) {
    std::ostream& out = fSession.cout();
    out << "OnXLab::PlotterManager::create :"
        << " OnX::Plotter not valid."
        << std::endl;
    delete plotter;
    return 0;
  }
  return plotter;
}

Slash::UI::IUI* OnXLab::PlotterManager::loadUI(){
  Slash::UI::IUI* ui = Slash::find_UI(fSession,Slash::s_default(),true);
  if(ui) return ui; //done
  Slash::UI::IUI_Manager* mgr = Slash::uiManager(fSession);
  if(!mgr) {
    std::ostream& out = fSession.cout();
    out << "OnXLab::PlotterManager::loadUI :"
        << " can't find the UI_Manager."
        << std::endl;
    return 0;
  }

  std::string driver;

 {// the below should be done by OnXLab::Main (created in AIDA.cxx) !
  Slash::args args(fArgs);
  if(args.is_arg("-thread")) {
    fSession.setParameter("UI_Manager.thread","yes");
  }
  if(args.is_arg("-verbose")) {
    fSession.setVerboseLevel(1);
  }
  driver = Slash::gui_toolkit(args,true);
  if(driver.empty()) {
    Slash::getenv("OSC_GUI_TOOLKIT",driver);
  }}

  bool load_simple_plotter = false;

  std::string file;
  if(Slash::check_getenv
       (fSession.cout(),"ONXLAB_PLOTTER","ONXLABPLOTTER",file)) {
    if(!Slash::file::exists(file)) {
      if(fSession.verboseLevel()) {
        std::ostream& out = fSession.cout();
        out << "OnXLab::PlotterManager::loadUI :"
            << " Environment variable ONXLAB_PLOTTER does not refer"
            << " to an existing file."
            << " Load the embedded SimplePlotter.onx..." 
            << std::endl;
      }
      load_simple_plotter = true;
    }
  } else {
    load_simple_plotter = true;
  }

  if(load_simple_plotter) {
    std::string s = 
      Slash::to_string(SimplePlotter_onx_linen,SimplePlotter_onx_lines);
    Slash::replace(s,"@@double_quote@@","\"");
    Slash::replace(s,"@@back_slash@@","\\");

    ui = mgr->create(driver,"@empty@",fArgs,false);
    if(!ui) {
      std::ostream& out = fSession.cout();
      out << "OnXLab::PlotterManager::loadUI :"
          << " can't create @empty@ UI for driver " << Slash::sout(driver) 
          << "."
          << std::endl;
      return 0;
    }
    ui->load("",s,false);

  } else {
    ui = mgr->create(driver,file,fArgs,false);
    if(!ui) {
      std::ostream& out = fSession.cout();
      out << "OnXLab::PlotterManager::loadUI :"
          << " can't create UI for driver " << Slash::sout(driver) 
          << " and file " << Slash::sout(file) 
          << std::endl;
      return 0;
    }
  }

  return ui;
}
