#ifndef OpenPAW_PAW_Manager_h
#define OpenPAW_PAW_Manager_h

// Inheritance :
#include <Slash/Core/IManager.h>

#include <vector>
#include <map>

#include <Lib/Parameters.h>
#include <Lib/Value.h>
#include <Slash/Tools/Manager.h> //for find_PAW

#include <OpenPAW/Unit.h>
#include <OpenPAW/ZONE.h>

namespace Slash {namespace Core {class ISession;}}
namespace Slash {namespace Core {class IWriter;}}
namespace Slash {namespace Store {class ITree;}}
namespace Lib {class Processor;}

namespace AIDA {
 class IAnalysisFactory;
 class ITree;
 class IPlotter;
 class IPlotterRegion;
 class IFunction;
 class ICloud2D;
 class IHistogram1D;
 class IDataPointSet;
}

namespace OpenPAW {

class ColorEntry {
public:
  ColorEntry();
  ColorEntry(double,double,double);
  ColorEntry(const ColorEntry&);
  ColorEntry& operator=(const ColorEntry&);
public:
  double fRed;
  double fGreen;
  double fBlue;
};

class Palette : public std::vector<int>  {
public:
  Palette();
  Palette(const Palette&);
  Palette(const std::vector<int>&);
  Palette& operator=(const Palette&);
};

class PAW_Manager : public virtual Slash::Core::IManager {
public: //Slash::Core::IManager
  virtual std::string name() const;
  virtual void* cast(const std::string&) const;
public:
  PAW_Manager(Slash::Core::ISession&);
  virtual ~PAW_Manager();
public:
  Slash::Core::ISession& session() const;
  void setAIDA(AIDA::IAnalysisFactory*);
  void set_mem(AIDA::ITree*);
  AIDA::IAnalysisFactory* aida() const;
  AIDA::ITree* mem() const;
  Slash::Store::ITree* mem_store_itree() const; //optimization
public:
  Lib::Parameters<int,OpenPAW::Unit>& units();
  OpenPAW::Unit findUnit(const std::string&) const;
  Lib::Parameters<std::string,Lib::Value>& variables();
  Lib::Parameters<int,std::string>& cuts();
  Lib::Parameters<std::string,std::string>& opts();
  Lib::Parameters<std::string,double>& atts();
  void setUTIT(const std::string&);
  int colFromHCOL() const;
  int colFromFCOL(bool&) const;
  bool paletteFromHCOL(int&) const;
  std::string print_fmt(const std::string&) const;
  void setContourMode(bool);
  void setCurrentTree(const std::string&);
  const std::string& currentTree() const;
  std::string generateFunctionName();
  void setFunctionPoints(int,int,int);
  int NPX() const;
  int NPY() const;
  int NPZ() const;
  void setZONE_K(int);
  int ZONE_K() const;
  void setToRefresh();
  void setDoZONE();
  void setZONE(const ZONE&);
  void createColor(int);
  void setColor(int,double,double,double);
  void setPalette(int,const Palette&);
  void setCurrentPalette(int);
  int currentPalette() const;
  void setTHETA(double);
  void setPHI(double);
  double THETA() const;
  double PHI() const;
  
  Lib::Processor* processor() const;
  void warn_metaf_ps();

#define cat_HPLSET   0
#define cat_IGSET    1
#define cat_HPLOPT   2
#define cat_OPAW_OPT 3

  void showATTs(std::ostream&,int = -1);
  void showOPTs(std::ostream&,int = -1);
  bool filterCuts(const std::string&,std::string&) const;
  bool parseTupleName(const std::string&,std::string&,
                      std::vector<std::string>&);
  int getInfosOptions(std::string&) const; 

  //AIDA :
  AIDA::ITree* findCurrentTree() const;
  void setPlotterLayout(AIDA::IPlotter&);
public:
  bool plotterNext(AIDA::IPlotter&,bool = false);
  void setPlotterTitle(AIDA::IPlotter&,const std::string&);
  void plotterSetFON(AIDA::IPlotter&,const std::string&,const std::string&);

  // Region :
  void regionSetLayout(AIDA::IPlotterRegion&); 
  bool paletteSize(int,int&);
  void regionSetColormap(AIDA::IPlotterRegion&,
                         const std::string&,double,double,int); 
  void regionSetFON(AIDA::IPlotterRegion&,const std::string&,
                    const std::string&);

  // Plotting :
  void plotFunction1D(AIDA::IPlotterRegion&,AIDA::IFunction&,
                      double,double,const std::string&); 
  void plotFunction2D(AIDA::IPlotterRegion&,AIDA::IFunction&,
                      double,double,double,double,double,double,
                      int,int,const std::string&); 
  void plotCloud2D(AIDA::IPlotterRegion&,AIDA::ICloud2D&,const std::string&);
  void plotErrors2D(AIDA::IPlotterRegion&,AIDA::IDataPointSet&,int,double);
  void plotHistogram1D(AIDA::IPlotterRegion&,AIDA::IHistogram1D&,
                       const std::string&,bool); 
  void setHTYP(AIDA::IPlotterRegion&,int,int,double,double); 

  void addLegend(AIDA::IPlotterRegion&,
                 double,double,const std::string&,
                 const std::string&,int,bool); 
  void getFont(const std::string&,std::string&,double&,bool&,bool&);
  void setLabels(int,const std::vector<std::string>&);
public:
  static bool parseName(const std::string&,
                        std::string&,std::vector< std::pair<int,int> >&,
                        Slash::Core::IWriter&);
  static bool isAnOption(const std::vector<std::string>&,
                         const std::string&,const std::string&); 
  static bool hasUnkownOption(const std::vector<std::string>&,
                              const std::string&); 
private:
  void regionInitialize(AIDA::IPlotterRegion&); 
  std::string get_MTYP(int);
  void setAxisLabels(AIDA::IPlotterRegion& aRegion,const std::string&,int,int);
private:
  std::string fName;
  Slash::Core::ISession& fSession;
  AIDA::IAnalysisFactory* fAIDA;
  AIDA::ITree* f_mem;
  Slash::Store::ITree* f_mem_store_itree;
  Lib::Parameters<int,OpenPAW::Unit> fUnits;
  Lib::Parameters<std::string,Lib::Value> fVariables;
  Lib::Parameters<int,std::string> fCuts;
  Lib::Parameters<std::string,std::string> fOPTs;
  Lib::Parameters<std::string,double> fATTs;
  std::string fUTIT;
  bool fContourMode;
  std::string fCurrentTree;
  int fFunctionCounter;
  int fNPX;
  int fNPY;
  int fNPZ;
  int fZONE_K;
  bool fToRefresh;
  bool fDoZONE;
  ZONE fZONE;
  bool f_start_ZONE;
  std::vector<ColorEntry> fColorTable;
  std::vector<Palette> fPalettes;
  int fCurrentPalette;
  std::map<int, std::vector<std::string> > fLabels;
  Lib::Processor* fProcessor;
  double fTHETA;
  double fPHI;
  bool f_warn_metaf_ps;
};

}

typedef Lib::Parameters<std::string,double> ATTs_type;
typedef Lib::Parameters<std::string,std::string> OPTs_type;
typedef Lib::Parameters<int,OpenPAW::Unit> Units_type;

//optimization : have a short name.
#define PAW_Manager_name "opaw"

inline OpenPAW::PAW_Manager* find_PAW(Slash::Core::ISession& aSession) {
  return Slash_findManager_backward(aSession,PAW_Manager_name,OpenPAW::PAW_Manager);
}

#define PALETTE_DEFAULT 1

#define PALETTE_HISTOGRAM_PLOT_COL              11
#define PALETTE_HISTOGRAM_2D_PLOT_CONTOUR_OPT_0 12

#define s_LEGO       "LEGO"
#define s_LEGO_LIGHT "LEGO_LIGHT"
#define s_LEGO_LEVEL "LEGO_LEVEL"
#define s_LEGO_COL   "LEGO_COL"

#define s_SURF               "SURF"
#define s_SURF_CONT          "SURF_CONT"
#define s_SURF_LIGHT         "SURF_LIGHT"
#define s_SURF_LEVEL         "SURF_LEVEL"
#define s_SURF_LEVEL_NO_EDGE "SURF_LEVEL_NO_EDGE"
#define s_SURF_COL           "SURF_COL"

#define s_FUN_C              "C"

#define s_Z          "Z"

#endif
