
#include <Slash/Core/ISession.h>
#include <Slash/UI/IUI.h>

#include <Slash/Tools/Managers.h>

#include <Slash/Tools/system.h>
#include <Slash/Tools/smanip.h>
#include <Slash/Tools/args.h>
#include <Slash/Tools/osc_env.h>

#include <OpenPAW/Version.h>

#include "env.h"

// Embedded code :
#include "opaw_help.h"

#include <iostream>
#include <cstdlib>

int main(int argc,char** argv){

  Slash::args args(argc,argv);

  if(args.is_arg("-version")) {
    std::cout << OPENPAW_VERSION << std::endl;
    return EXIT_SUCCESS;
  }

  if(args.is_arg("-h")) {
    std::vector<std::string> text;
    text.resize(opaw_help_linen);
    for(unsigned int index=0;index<opaw_help_linen;index++) {
      std::string s = opaw_help_lines[index];
      Slash::replace(s,"@@double_quote@@","\"");
      Slash::replace(s,"@@back_slash@@","\\");
      text[index] = s;
    }
    for(unsigned int index=0;index<text.size();index++) {
      std::cout << text[index] << std::endl;
    }
    return EXIT_SUCCESS;
  }

  if(!Slash::osc::set_env(std::cout)) {
    std::cout << "OpenPAW_opaw :"
              << " unable to set OnX environment." << std::endl;
    return EXIT_FAILURE;
  }
  if(!OpenPAW::set_env(std::cout)) {
    std::cout << "OpenPAW_opaw :"
              << " unable to set OnXLab environment." << std::endl;
    return EXIT_FAILURE;
  }

 {std::vector<std::string> files = args.files();
  for(std::vector<std::string>::const_iterator it = files.begin();
      it!=files.end();++it) {
    args.add("-kumac_file",*it,false);
    args.remove(*it); //else it will taken as a .onx file !
  }}

  args.add("-stdout","");

//still needed ?
//if(args.find("-home",val)) {
//  args.remove("-home");
//  if(val=="") {
//    if(Lib::dirmanip::goHome())
//      if(Lib::dirmanip::mkcd("OpenScientist"))
//        if(Lib::dirmanip::mkcd("OpenPAW"))
//          Lib::dirmanip::mkcd("session");
//  } else {
//    Lib::dirmanip::mkcd(Slash::words(val,"/"));
//  }
//}

  Slash::Core::ISession* session = Slash_create_session(args.tovector());
  if(!session) {
    std::cout << "OpenPAW_opaw :"
              << " session not found." << std::endl;
    return EXIT_FAILURE;
  }

  Slash::UI::IUI_Manager* uim = Slash::uiManager(*session);
  if(!uim) {
    std::cout << "OpenPAW_opaw :"
              << " UI_Manager not found." << std::endl;
    delete session;
    return EXIT_FAILURE;
  }


  std::string OSC_HOME_DIR;
  if(!Slash::getenv("OSC_HOME_DIR",OSC_HOME_DIR)) {
    std::cout << "OpenPAW_opaw :"
              << " environment variable OSC_HOME_DIR not defined."
              << std::endl;
    delete session;
    return EXIT_FAILURE;
  }
  std::string file = OSC_HOME_DIR + "/Resources/OpenPAW/scripts/OnX/main.onx";

  Slash::UI::IUI* ui = uim->create("",file,std::vector<std::string>(),true);
  if(!ui) {
    std::cout << "OpenPAW_opaw :"
              << " can't create UI."
              << std::endl;
    delete session;
    return EXIT_FAILURE;
  }

  ui->enableTerminal("OPAW> ","KUIP");
  ui->redirectAskToTerminal(true);
  ui->steer();

  delete session;

  // Some libraries (for exa Coin, gtk) may have declared 
  // C functions with ::atexit. To ease debugging we call
  // explicitly ::exit here.
  ::exit(EXIT_SUCCESS);

  return EXIT_SUCCESS;
}
