#ifndef OpenPAW_Callbacks_Array_h
#define OpenPAW_Callbacks_Array_h

//
//  This file contains inline helper methods.
//

#include <Slash/Core/IWriter.h>
#include <Lib/smanip.h>
#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/Cast.h>

#include <AIDA/ITree.h>
#include <AIDA/IManagedObject.h>

#include <Slash/Data/IArray.h>

//////////////////////////////////////////////////////////////////////////////
inline void print_vector(
 std::ostream& a_out
,const Slash::Data::IArray& aArray
,const std::string& aTitle
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aTitle!="") a_out << aTitle << std::endl;
  if(aArray.dimension()==1) {
    for(unsigned int i=0;i<aArray.orders()[0];i++) {
      std::vector<unsigned int> is(1);
      is[0] = i;
      double value;
      aArray.value(is,value);
      a_out << value << std::endl;
    }
  } else if(aArray.dimension()==2) {
    for(unsigned int j=0;j<aArray.orders()[1];j++) {
      for(unsigned int i=0;i<aArray.orders()[0];i++) {
        std::vector<unsigned int> is(2);
        is[0] = i;
        is[1] = j;
        double value;
        aArray.value(is,value);
        std::string s;
        Lib::smanip::printf(s,32,"%g",value);
        Lib::smanip::justify(s,5);
        a_out << s;
      }
      a_out << std::endl;
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
inline Slash::Data::IArray* find_vector(
 AIDA::ITree& aTree
,const std::string& aName
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  AIDA::IManagedObject* object = aTree.find(aName);
  if(!object) return 0;
  return Lib_Cast(*object,Slash::Data::IArray);
}
//////////////////////////////////////////////////////////////////////////////
inline bool vector_get_cut(
 const Slash::Data::IArray& aArray
,const std::vector< std::pair<int,int> >& aDims
,std::vector< std::pair<unsigned int,unsigned int> >& aCut
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aCut.clear();
  unsigned int dim = aArray.dimension();
  if(aDims.size()) {
    if(aDims.size()!=dim) return false;
    for(unsigned int index = 0;index<dim;index++) {
      int mn = aDims[index].first;
      if(mn==(-1)/*NotFound*/) return false;
      int mx = aDims[index].second;
      if(mx==(-1)/*NotFound*/) return false;
      aCut.push_back( std::pair<unsigned int,unsigned int>(mn,mx));
    }
  } else {
    for(unsigned int index = 0;index<dim;index++) {
      unsigned int mn = 0;
      unsigned int mx = aArray.orders()[index] -1;
      aCut.push_back( std::pair<unsigned int,unsigned int>(mn,mx));
    }
  }
  return true;
}

#endif
