/*
 * There is no KUIP in this class.
 * It maintains the "PAW state".
 */


// this :
#include <OpenPAW/PAW_Manager.h>

#include <Slash/Core/ISession.h>
#include <Slash/Core/IWriter.h>

#include <Slash/Store/ITree.h>
#include <AIDA/ITree.h>

#include <Slash/Tools/smanip.h>

#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/smanip.h>
#include <Lib/mmanip.h>
#include <Lib/Cast.h>
#include <Lib/Processor.h>

#include <cstdlib> //strtol, abs

#include <HEPVis/SbStyles.h> //for SbFont_ROOT_<code>

#include "att_opt.cxx"

#define MAXIMUM(a,b) ((a)>(b)?a:b)
#define NotFound (-1)
#define s_retain "retain=true"

#define REGION_SET_MTYP \
  aRegion.setParameter(field+"markerStyle",get_MTYP((int)fATTs.value("MTYP")));

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

OpenPAW::ColorEntry::ColorEntry()
:fRed(0),fGreen(0),fBlue(0)
{}

OpenPAW::ColorEntry::ColorEntry(double aRed,double aGreen,double aBlue)
:fRed(aRed),fGreen(aGreen),fBlue(aBlue)
{}

OpenPAW::ColorEntry::ColorEntry(const ColorEntry& aFrom)
:fRed(aFrom.fRed)
,fGreen(aFrom.fGreen)
,fBlue(aFrom.fBlue)
{}

OpenPAW::ColorEntry& OpenPAW::ColorEntry::operator=(const ColorEntry& aFrom){
  fRed = aFrom.fRed;fGreen = aFrom.fGreen;fBlue = aFrom.fBlue;
  return *this;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
OpenPAW::Palette::Palette(){
  // Default CERN/PAW palette elements (2,3,4,5,6,7)
  push_back(2); //red in default color table.
  push_back(3); //green in default color table.
  push_back(4); //blue in default color table.
  push_back(5); //yellow in default color table.
  push_back(6); //magenta in default color table.
  push_back(7); //cyan in default color table.
}

OpenPAW::Palette::Palette(const std::vector<int>& aFrom){ *((std::vector<int>*)this) = aFrom;}

OpenPAW::Palette::Palette(const Palette& aFrom):std::vector<int>(aFrom){}

OpenPAW::Palette& OpenPAW::Palette::operator=(const Palette& aFrom){
  *((std::vector<int>*)this) = aFrom;
  return *this;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

OpenPAW::PAW_Manager::PAW_Manager(Slash::Core::ISession& aSession)
:fName(PAW_Manager_name)
,fSession(aSession)
,fAIDA(0)
,f_mem(0)
,f_mem_store_itree(0)
,fUnits(NotFound,Unit())
,fVariables("",Lib::Value())
,fCuts(NotFound,"")
,fOPTs("","")
,fATTs("",0)
,fContourMode(true) //Contour or fill area.
,fCurrentTree("MemoryTree")
,fFunctionCounter(0)
,fNPX(100)
,fNPY(100)
,fNPZ(100)
,fZONE_K(NotFound)
,fToRefresh(true)
,fDoZONE(true)
,fZONE(1,1,1,false)
,f_start_ZONE(false)
,fCurrentPalette(-1)
,fProcessor(0)
,fTHETA(30)
,fPHI(30)
,f_warn_metaf_ps(false)
{
  att_opt_init(fATTs,fOPTs);

  // Default CERN/PAW color table is :
  // 0 white, 1 black, 2 red, 3 green, 4 blue, 5 yellow, 6 magenta, 7 cyan.
  fColorTable.push_back(ColorEntry(1,1,1));
  fColorTable.push_back(ColorEntry(0,0,0));
  fColorTable.push_back(ColorEntry(1,0,0));
  fColorTable.push_back(ColorEntry(0,1,0));
  fColorTable.push_back(ColorEntry(0,0,1));
  fColorTable.push_back(ColorEntry(1,1,0));
  fColorTable.push_back(ColorEntry(1,0,1));
  fColorTable.push_back(ColorEntry(0,1,1));

  // Some default palettes :

  //(red,green,blue,yellow,magenta,cyan)
  setPalette(PALETTE_DEFAULT,Palette());
  //fCurrentPalette(PALETTE_DEFAULT);

  //So that :
  //  H/PLOT 1 COL (studeny)
  //without having done a SET PALETTE reproduce CERN/PAW. 
 {std::vector<int> pal;
  pal.push_back(0); //white
  pal.push_back(5); //yellow
  pal.push_back(7); //cyan
  pal.push_back(3); //green
  pal.push_back(6); //magenta
  pal.push_back(2); //red
  pal.push_back(4); //blue
  pal.push_back(1); //black
  setPalette(PALETTE_HISTOGRAM_PLOT_COL,pal);}

  //CERN/PAW has PALNB in [0,9]. PALNB >= 10 are for OpenPAW internal usage.

  //HI/2D/CONTOUR CHOPT=0 (pawex14)
 {std::vector<int> pal;
  pal.push_back(1); //black
  pal.push_back(2); //red
  pal.push_back(3); //green
  pal.push_back(4); //blue
  setPalette(PALETTE_HISTOGRAM_2D_PLOT_CONTOUR_OPT_0,pal);}

  fProcessor = new Lib::Processor("",fSession.printer());
}

OpenPAW::PAW_Manager::~PAW_Manager(){
  delete fProcessor;

  // Close units :
  const Lib::Parameters<int,OpenPAW::Unit>::List& units = fUnits.list();
  Lib::Parameters<int,OpenPAW::Unit>::ConstIterator it;
  for(it=units.begin();it!=units.end();++it) {
    const OpenPAW::Unit& u = (*it).fValue;
    if(fSession.verboseLevel()) {
      std::ostream& out = fSession.cout();
      out << " OpenPAW::PAW_Manager::~PAW_Manager :" 
          << " close unit " << (*it).fKey 
          << " " << Slash::sout(u.fString) 
          << std::endl;
    }
    if(u.isFILE()) {
      ::fclose(u.fFILE);
    } else {
      fSession.destroyManager(u.fString);
    }
  }
}

std::string OpenPAW::PAW_Manager::name() const {return fName;}

void* OpenPAW::PAW_Manager::cast(const std::string& aTo) const {
  if(aTo=="OpenPAW::PAW_Manager") {
    return Lib_SCast(OpenPAW::PAW_Manager);
  } else if(aTo=="Slash::Core::IManager") {
    return Lib_SCast(Slash::Core::IManager);
  } else {
    return 0;
  }
}

Slash::Core::ISession& OpenPAW::PAW_Manager::session() const {return fSession;}

void OpenPAW::PAW_Manager::setAIDA(AIDA::IAnalysisFactory* aAIDA){fAIDA = aAIDA;}
void OpenPAW::PAW_Manager::set_mem(AIDA::ITree* a_mem) {
  f_mem = a_mem;
  f_mem_store_itree = Lib_Cast(*a_mem,Slash::Store::ITree);
}
AIDA::IAnalysisFactory* OpenPAW::PAW_Manager::aida() const {return fAIDA;}
AIDA::ITree* OpenPAW::PAW_Manager::mem() const {return f_mem;}
Slash::Store::ITree* OpenPAW::PAW_Manager::mem_store_itree() const {return f_mem_store_itree;}

Lib::Processor* OpenPAW::PAW_Manager::processor() const {return fProcessor;}

Lib::Parameters<int,OpenPAW::Unit>& OpenPAW::PAW_Manager::units(){return fUnits;}


OpenPAW::Unit OpenPAW::PAW_Manager::findUnit(const std::string& aString) const{
  const Lib::Parameters<int,OpenPAW::Unit>::List& lunits = fUnits.list();
  Lib::Parameters<int,OpenPAW::Unit>::ConstIterator it;
  for(it=lunits.begin();it!=lunits.end();++it) {
    const OpenPAW::Unit& u = (*it).fValue;
    if(u.fString==aString) return u;
  }
  return OpenPAW::Unit();
}

Lib::Parameters<std::string,Lib::Value>& OpenPAW::PAW_Manager::variables(){return fVariables;}

Lib::Parameters<int,std::string>& OpenPAW::PAW_Manager::cuts(){return fCuts;}

Lib::Parameters<std::string,std::string>& OpenPAW::PAW_Manager::opts(){return fOPTs;}

Lib::Parameters<std::string,double>& OpenPAW::PAW_Manager::atts(){return fATTs;}

void OpenPAW::PAW_Manager::setUTIT(const std::string& aString){fUTIT = aString;}

int OpenPAW::PAW_Manager::colFromHCOL() const {
  // HCOL is of the form <hatch>*1000+<col>.<palette>
  // Here we extract the <col>.
  int HCOL = (int)fATTs.value("HCOL");
  return HCOL - ((int)(HCOL/1000)) * 1000;
}

//FIXME : is it correct ?
int OpenPAW::PAW_Manager::colFromFCOL(bool& a_touched) const { 
  // FCOL is of the form <hatch>*1000+<col>.<palette>
  // Here we extract the <col>.
  int FCOL = (int)fATTs.value("FCOL");
  fATTs.touched("FCOL",a_touched);
  return FCOL - ((int)(FCOL/1000)) * 1000;
}

inline int round_d2i(double aDouble){
  if (aDouble == (double) (int(aDouble))) return int(aDouble);
  else return (aDouble>0.) ? int(aDouble+0.5) : -int(0.5-aDouble);
}

bool OpenPAW::PAW_Manager::paletteFromHCOL(int& aPALNB) const {
  // Look if HCOL = <c>.<p> with p in [1,9]

  double HCOL = fATTs.value("HCOL");
  double rest = HCOL - (int)HCOL;
  if(rest==0) {
    aPALNB = 0;
    return false;
  }
  rest *= 10;
  //NOTE : on Linux, for :
  //  HCOL = 0.3  (PAW_C/opaw_color.kumac)
  //  int p = int(rest);
  //returns 2 !
  int p = round_d2i(rest);
  if(p<=0) {   //for exa if HCOL 0.01
    aPALNB = 0;
    return false;
  }
  aPALNB = p;
  return true;
}

std::string OpenPAW::PAW_Manager::print_fmt(const std::string& aFILE) const {
  std::string stype;
  std::string path,name,suffix;
  if(!Slash::path_name_suffix(aFILE,path,name,suffix)) {
    stype = "GL2PS";
  } else {
    Slash::touppercase(suffix);
    if((suffix=="PS")||(suffix=="EPS")) {
      stype = "GL2PS";
    } else if(suffix=="JPG") {
      stype = "JPEG";
    } else {
      stype = suffix;
    }
  }
  return stype;
}

void OpenPAW::PAW_Manager::setContourMode(bool aMode){fContourMode = aMode;}

void OpenPAW::PAW_Manager::setCurrentTree(const std::string& aString){fCurrentTree = aString;}

const std::string& OpenPAW::PAW_Manager::currentTree() const {return fCurrentTree;}

std::string OpenPAW::PAW_Manager::generateFunctionName(){
  std::string name = "scripted_";
  name += Slash::to<int>(fFunctionCounter);
  fFunctionCounter++;
  return name;
}

void OpenPAW::PAW_Manager::setFunctionPoints(int aNPX,int aNPY,int aNPZ) {
  fNPX = aNPX;
  fNPY = aNPY;
  fNPZ = aNPZ;
}

int OpenPAW::PAW_Manager::NPX() const {return fNPX;}
int OpenPAW::PAW_Manager::NPY() const {return fNPY;}
int OpenPAW::PAW_Manager::NPZ() const {return fNPZ;}

void OpenPAW::PAW_Manager::setZONE_K(int aZONE_K){fZONE_K = aZONE_K;}

int OpenPAW::PAW_Manager::ZONE_K() const {return fZONE_K;}

void OpenPAW::PAW_Manager::setToRefresh(){fToRefresh = true;}

void OpenPAW::PAW_Manager::setDoZONE(){fDoZONE = true;}

void OpenPAW::PAW_Manager::setZONE(const ZONE& aValue){fZONE = aValue;}

void OpenPAW::PAW_Manager::setLabels(int a_id,const std::vector<std::string>& a_v){fLabels[a_id] = a_v;}

void OpenPAW::PAW_Manager::setColor(int aICOL,double aRED,double aGREEN,double aBLUE){
  if(aICOL<0) return;
  createColor(aICOL);
  fColorTable[aICOL].fRed = aRED;
  fColorTable[aICOL].fGreen = aGREEN;
  fColorTable[aICOL].fBlue = aBLUE;
}

void OpenPAW::PAW_Manager::createColor(int aICOL){
  if(aICOL<0) return;
  if(aICOL<(int)fColorTable.size()) return;
  int n = (aICOL+1) - fColorTable.size();
  for(int index=0;index<n;index++) fColorTable.push_back(ColorEntry());
}

void OpenPAW::PAW_Manager::setPalette(int aPALNB,const Palette& aPalette){
  if(aPALNB<0) return;
  if(aPALNB>=(int)fPalettes.size()) {
    int n = (aPALNB+1) - fPalettes.size();
    for(int index=0;index<n;index++) {
      fPalettes.push_back(Palette());
    }
  }
  fPalettes[aPALNB] = aPalette;
}

void OpenPAW::PAW_Manager::setCurrentPalette(int aPALNB){fCurrentPalette = aPALNB;}
int OpenPAW::PAW_Manager::currentPalette() const {return fCurrentPalette;}

void OpenPAW::PAW_Manager::setTHETA(double aTHETA) {fTHETA = aTHETA;}
void OpenPAW::PAW_Manager::setPHI(double aPHI) {fPHI = aPHI;}
double OpenPAW::PAW_Manager::THETA() const {return fTHETA;}
double OpenPAW::PAW_Manager::PHI() const {return fPHI;}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void OpenPAW::PAW_Manager::showATTs(std::ostream& a_out,int a_category) { 
  //print defaults and current values for all attributes.
  const Lib::Parameters<std::string,double>::List& attList = fATTs.list();
  Lib::Parameters<std::string,double>::List::const_iterator it;
  a_out << "+-------------+---------------+---------------+--------------------------------+" << std::endl;
  a_out << "|  Parameter  | Current value | Default value |          Explanation           |" << std::endl;
  a_out << "+-------------+---------------+---------------+--------------------------------+" << std::endl;
  for(it=attList.begin();it!=attList.end();++it) {
    int cat = (*it).fCategory;
    if( (a_category!=NotFound)&&(cat!=a_category)) continue;

    std::string s = (*it).fKey; //Key
    Lib::smanip::justify(s,13,Lib::smanip::middle);
    a_out << "|" << s;
    Lib::smanip::printf(s,32,"%g",(*it).fValue); //value
    Lib::smanip::justify(s,15,Lib::smanip::middle);
    a_out << "|" << s;
    Lib::smanip::printf(s,32,"%g",(*it).fDefault); //default
    Lib::smanip::justify(s,15,Lib::smanip::middle);
    a_out << "|" << s;
    s = (*it).fLabel; //label
    Lib::smanip::justify(s,32,Lib::smanip::left);
    a_out << "|" << s << "|";
    a_out << std::endl;        
  }
  a_out << "+------------+---------------+---------------+--------------------------------+" << std::endl;

}

void OpenPAW::PAW_Manager::showOPTs(std::ostream& a_out,int a_category) { 
  //print defaults and current values for all attributes.
  const Lib::Parameters<std::string,std::string>::List& optList = fOPTs.list();
  Lib::Parameters<std::string,std::string>::List::const_iterator it;
  a_out << "+-------------+-------------+---------------+---------------------------------+" << std::endl;
  a_out << "|   Current   |   Default   |  Alternative  |           Explanation           |" << std::endl;
  a_out << "+-------------+-------------+---------------+---------------------------------+" << std::endl;
  for(it=optList.begin();it!=optList.end();++it) {
    int cat = (*it).fCategory;
    if( (a_category!=NotFound)&&(cat!=a_category)) continue;

    std::string s = (*it).fValue;
    Lib::smanip::justify(s,13,Lib::smanip::middle);
    a_out << "|" << s;
    s = (*it).fDefault;
    Lib::smanip::justify(s,13,Lib::smanip::middle);
    a_out << "|" << s;
    s = (*it).fAlternative;
    Lib::smanip::justify(s,15,Lib::smanip::middle);
    a_out << "|" << s;
    s = (*it).fLabel;
    std::vector<std::string> words;
    Lib::smanip::words(s," ",words);
    unsigned int index;
    std::string s1;
    int l = 0;
    int istart = NotFound;
    for(index=0;index<words.size();index++) {
      if((l+words[index].size()+1)>31) {
        istart = index;
        break;
      }
      s1 += words[index];
      s1 += " ";
      l += words[index].size()+1;
    }
    std::string s2;
    if(istart!=NotFound) {
      for(index=(unsigned int)istart;index<words.size();index++) {
        s2 += words[index];
        s2 += " ";
      }
    }
    Lib::smanip::justify(s1,31,Lib::smanip::left);
    a_out << "| " << s1 << " |" << std::endl;        
    Lib::smanip::strip(s2);
    if(s2!="") {
      Lib::smanip::justify(s2,31,Lib::smanip::left);
      a_out << "|             |             |               | " 
          << s2 << " |" << std::endl;
    }

  }
  a_out << "+------------+---------------+---------------+--------------------------------+" << std::endl;

}

bool OpenPAW::PAW_Manager::filterCuts(const std::string& aString,std::string& aFilter) const {
  std::string s = aString;

  // Treat cuts. replace $<number> with (<cut>)
  while(1) {
    std::string::size_type dollar = s.find('$');
    if(dollar==std::string::npos) break;

    const char* ss = s.c_str()+dollar+1;
    char* sss;
    long v = ::strtol(ss,&sss,10);
    if(ss==sss) { //Syntax error :
      aFilter = "";
      return false;
    }
    unsigned long l = sss - ss;     
    std::string scut = s.substr(dollar+1,l);
    //printf("debug : cut : \"%s\"\n",scut.c_str());      
    if(!fCuts.isKey((int)v)) {
      aFilter = "";
      return false;
    }
    std::string cut = fCuts.value((int)v);
    std::string sbef = s.substr(0,dollar);
    std::string saft = s.substr(dollar+1+l,s.size()-(dollar+1+l));
    //printf("debug : filter2C : bef \"%s\"\n",s.c_str());      
    s = sbef + "(" + cut + ")" + saft;
    //printf("debug : filter2C : aft \"%s\"\n",s.c_str());      
  }

  Lib::smanip::tolowercase(s);
  Lib::smanip::replace(s,"!=",".neq.");
  Lib::smanip::replace(s,"=",".eq.");

  Lib::smanip::replace(s,".neq.","!=");
  Lib::smanip::replace(s,".eq.","==");
  Lib::smanip::replace(s,".gt.",">");
  Lib::smanip::replace(s,".ge.",">=");
  Lib::smanip::replace(s,".lt.","<");
  Lib::smanip::replace(s,".le.","<=");
  Lib::smanip::replace(s,".and.","&&");
  Lib::smanip::replace(s,".or.","||");
  Lib::smanip::replace(s,".not.","!");

  aFilter = s;
  //printf("debug : filter2C : res \"%s\"\n",aFilter.c_str());      
  return true;
}

bool OpenPAW::PAW_Manager::parseTupleName(const std::string& aTNAME,std::string& aNAME,std::vector<std::string>& aVars){
  aVars.clear();
  std::string::size_type ldot = aTNAME.find('.');
  if(ldot==std::string::npos){
    std::ostream& out = fSession.cout();
    out << "OpenPAW::PAW_Manager::parseTupleName :" 
        << " Bad syntax in " << Slash::sout(aTNAME) << "." << std::endl;
    return false;
  }
  aNAME = aTNAME.substr(0,ldot);
  std::string svars = aTNAME.substr(ldot+1,aTNAME.size()-(ldot+1));
  Lib::smanip::words(svars,"%",aVars);
  for(unsigned int index=0;index<aVars.size();index++) {
    std::string s;
    if(!filterCuts(aVars[index],s)) {
      std::ostream& out = fSession.cout();
      out << "OpenPAW::PAW_Manager::parseTupleName :" 
          << " problem treating " << Slash::sout(aVars[index]) 
          << std::endl;
      return false;
    }
    aVars[index] = s;
  }
  return true;
}

int OpenPAW::PAW_Manager::getInfosOptions(std::string& aOptions) const {
  int linen = 0;
  std::string opts;

  // STAT infos :
  std::string NSTA = fOPTs.value("NSTA");
  if(NSTA=="STA") {
    double STAT = fATTs.value("STAT"); 
    //p120 : set STAT [OURMEI]  //default 1111
    // O=1 overflow
    // U=1 underflow
    // R=1 RMS
    // M=1 mean
    // E=1 entries
    // I=1 ID
    int code = (int)STAT;
    int O = code/100000;
    int U = (code - O * 100000)/10000; 
    int R = (code - O * 100000 - U * 10000)/1000; 
    int M = (code - O * 100000 - U * 10000 - R * 1000)/100; 
    int E = (code - O * 100000 - U * 10000 - R * 1000 - M *100)/10; 
    int I = (code - O * 100000 - U * 10000 - R * 1000 - M *100 - E*10)/1; 
    //printf("debug : stat : %d %d %d %d %d %d\n",O,U,R,M,E,I);
    if(I) {opts += "name ";linen++;}
    if(E) {opts += "entries ";linen++;}
    if(M) {opts += "mean ";linen++;}
    if(R) {opts += "rms ";linen++;}
    if(U) {opts += "underflow ";linen++;}
    if(O) {opts += "overflow ";linen++;}
  }

  // FIT infos :
  std::string NFIT = fOPTs.value("NFIT");
  if(NFIT=="FIT") {
    double FIT = fATTs.value("FIT"); 
    // PAW set FIT [CEP] //default 101
    //    C = 1;  print Chisquare/Number of degress of freedom
    //    E = 1;  print errors (if e=1, v must be 1)
    //    P = 1;  print name/values of parameters
    //FIXME : from ROOT TPaveStats.
    //  The parameter mode can be = pcev  (default = 0111)
    //    v = 1;  print name/values of parameters
    //    e = 1;  print errors (if e=1, v must be 1)
    //    c = 1;  print Chisquare/Number of degrees of freedom
    //    p = 1;  print Probability
    //    When "v"=1 is specified, only the non-fixed parameters are shown.
    //    When "v"=2 all parameters are shown.
    //  Example: gStyle->SetOptFit(1011);
    //           print fit probability, parameter names/values and errors.
    int code = (int)FIT;
    int C = code/100;
    int E = (code - C * 100)/10; 
    int P = (code - C * 100 - E * 10)/1; 
    //printf("debug : fit : C %d E %d P %d\n",C,E,P);
    if(C) {
      opts += "fit_quality ";
      opts += "fit_ndf ";
      linen++;
    }
    if(E) {opts += "fit_errors ";linen++;}
    if(P) {opts += "fit_parameters ";linen++;}
  }

  aOptions = opts;

  return linen;
}

//////////////////////////////////////////////////////////////////////////////
// static ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool OpenPAW::PAW_Manager::parseName(
 const std::string& aString
,std::string& aNAME
,std::vector< std::pair<int,int> >& aDims
,Slash::Core::IWriter& aPrinter
){
  //printf("debug : parse_name : \"%s\"\n",aString.c_str());
  aDims.clear();
  std::string::size_type lbrace = aString.find('(');
  if(lbrace!=std::string::npos){
    std::string::size_type rbrace = aString.find(')');
    if( (rbrace==std::string::npos) || (rbrace<lbrace) ) {
      Lib::Out out(aPrinter);
      out << "OpenPAW::PAW_Manager::parseName :" 
          << " Bad syntax in " << Lib::sout(aString) << "." << Lib::endl;
      return false;
    }
    aNAME = aString.substr(0,lbrace);
    std::string sdim = aString.substr(lbrace+1,rbrace-(lbrace+1));
    std::vector<std::string> words;
    Lib::smanip::words(sdim,",",words,true);
    if( (words.size()<=0) || (words.size()>=4)) {
      Lib::Out out(aPrinter);
      out << "OpenPAW::PAW_Manager::parseName :" 
          << " Dimension >=4 not supported." << Lib::endl;
      return false;
    }
    for(unsigned int index=0;index<words.size();index++) {
      int idim;
      if(words[index]=="") {
        aDims.push_back( std::pair<int,int>(NotFound,NotFound)); 
      } else if(!Lib::smanip::toint(words[index],idim)) {
        // Case of limits ?
        std::string::size_type colon = words[index].find(':');
        if(colon==std::string::npos){
          Lib::Out out(aPrinter);
          out << "OpenPAW::PAW_Manager::parseName :" 
              << " bad syntax in " << Lib::sout(words[index]) << "." 
              << Lib::endl;
          return false;
        } else {
          std::string left = words[index].substr(0,colon);
          std::string right = 
            words[index].substr(colon+1,words[index].size()-(colon+1));
          int mn;
          if(left=="") {
            mn = NotFound;
          } else {
            if(!Lib::smanip::toint(left,mn)) {
              Lib::Out out(aPrinter);
              out << "OpenPAW::PAW_Manager::parseName :" 
                  << " can't convert " << Lib::sout(left) << " to long." 
                  << Lib::endl;
              return false;
            }
            mn -= 1;
            if(mn<0) {
              Lib::Out out(aPrinter);
              out << "OpenPAW::PAW_Manager::parseName :" 
                  << " bad limits in " << Lib::sout(words[index]) << "." 
                  << Lib::endl;
              return false;
            }
          }
          int mx;
          if(right=="") {
            mx = NotFound;
          } else {
            if(!Lib::smanip::toint(right,mx)) {
              Lib::Out out(aPrinter);
              out << "OpenPAW::PAW_Manager::parseName :" 
                  << " can't convert " << Lib::sout(right) << " to long." 
                  << Lib::endl;
              return false;
            }
            mx -= 1;
            if(mx<0) {
              Lib::Out out(aPrinter);
              out << "OpenPAW::PAW_Manager::parseName :" 
                  << " bad limits in " << Lib::sout(words[index]) << "." 
                  << Lib::endl;
              return false;
            }
            if((mn!=NotFound)&&(mx<mn)) {
              Lib::Out out(aPrinter);
              out << "OpenPAW::PAW_Manager::parseName :" 
                  << " bad limits in " << Lib::sout(words[index]) << "." 
                  << Lib::endl;
              return false;
            }
          }
          //printf("debug : mn,mx1 : %d %d\n",mn,mx);
          aDims.push_back( std::pair<int,int>(mn,mx)); 
        }       
      } else {
        //printf("debug : mn,mx2 : %d %d\n",idim-1,idim-1);
        aDims.push_back( std::pair<int,int>(idim-1,idim-1)); 
      }
    }
  } else {
    aNAME = aString;
  }
  return true;
}

bool OpenPAW::PAW_Manager::isAnOption(
 const std::vector<std::string>& aOPTS
,const std::string& aEntered
,const std::string& aIsIn
) {
  //  For FUN2 ("CHOPT","BOXSSURF","S") yes
  //  For FUN2 ("CHOPT","BOXSURF","S") no
  // aEntered could contain spaces : " A C"

  // Check if aIsIn is in aEntered.
  // WARNING : an option may be a substring of another one
  // like "S" and "SURF" for FUN2 CHOPT.
  std::string s = aEntered;
  if(s.empty()) s = " ";
  std::vector<std::string> sfound;
  unsigned int index;
  for(index=0;index<aOPTS.size();index++) {
    const std::string& OPT = aOPTS[index];
    if(s.size() && (s.find(OPT)!=std::string::npos) ) {
      sfound.push_back(OPT);
      Lib::smanip::replace(s,OPT,"");
    }
  }
  for(index=0;index<sfound.size();index++) {
    if(aIsIn==sfound[index]) return true;
  }
  return false;
}

bool OpenPAW::PAW_Manager::hasUnkownOption(
 const std::vector<std::string>& aOPTS
,const std::string& aEntered
){
  // WARNING : an option may be a substring of another one
  // like "S" and "SURF" for FUN2 CHOPT.
  std::string s = aEntered;
  std::vector<std::string> sfound;
  unsigned int index;
  for(index=0;index<aOPTS.size();index++) {
    const std::string& OPT = aOPTS[index];
    if(OPT=="") {
      sfound.push_back("");
    } else if(s.size() && (s.find(OPT)!=std::string::npos) ) {
      sfound.push_back(OPT);
      Lib::smanip::replace(s,OPT,"");
    }
  }
  return (s.size()?true:false);
}

//////////////////////////////////////////////////////////////////////////////
/// Over AIDA ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#include <AIDA/ITree.h>
#include <AIDA/IPlotter.h>
#include <AIDA/IPlotterRegion.h>
#include <AIDA/IFunction.h>
#include <AIDA/ICloud2D.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IAnnotation.h>

#include <Lib/Manager.h>
#include <HEPVis/SbStyles.h> //#define only.
#include "RegionHelpers.h"

AIDA::ITree* OpenPAW::PAW_Manager::findCurrentTree() const {
  AIDA::ITree* tree = Lib_findManager(fSession,fCurrentTree,AIDA::ITree);
  if(!tree) {
    std::ostream& out = fSession.cout();
    out << "OpenPAW::PAW_Manager::findCurrentTree :"
        << " tree " << Lib::sout(fCurrentTree) << " not found." 
        << std::endl;
  }
  return tree;
}

void OpenPAW::PAW_Manager::setPlotterLayout(AIDA::IPlotter& aPlotter) {

  if(fToRefresh) {

    aPlotter.setParameter("regionsResizePolicy","PAW_GRID");

    //NOTE : if using an ATT or an OPT here, think to do
    //       a pawManager.setDoZONE(true) when changing
    //       the corresponding att or opt.

    double XSIZ = fATTs.value("XSIZ"); 
    double XMGL = fATTs.value("XMGL");
    double XMGR = fATTs.value("XMGR");
    double XWIN = fATTs.value("XWIN");
    aPlotter.setParameter("pageWidth",Slash::to<double>(XSIZ));
    aPlotter.setParameter("leftMargin",Slash::to<double>(XMGL));
    aPlotter.setParameter("rightMargin",Slash::to<double>(XMGR));
    aPlotter.setParameter("horizontalSpacing",Slash::to<double>(XWIN));

    double YSIZ = fATTs.value("YSIZ"); 
    double YMGL = fATTs.value("YMGL");
    double YMGU = fATTs.value("YMGU");
    double YWIN = fATTs.value("YWIN");
    aPlotter.setParameter("pageHeight",Slash::to<double>(YSIZ));
    aPlotter.setParameter("bottomMargin",Slash::to<double>(YMGL));
    aPlotter.setParameter("topMargin",Slash::to<double>(YMGU));
    aPlotter.setParameter("verticalSpacing",Slash::to<double>(YWIN));

    //reinitialize regions ?
    // Initialize region with PAW defaults :
    /*
    int number = aPlotter.numberOfRegions();
    for(int index=0;index<number;index++) {
      AIDA::IPlotterRegion* region = aPlotter.region(index);
      if(region) regionInitialize(*region);
    }
    */

    fToRefresh = false;
  }

  if(fDoZONE) {

    if(fZONE.SUPER()) {
      // create one new region placed in a grid.
      if(!aPlotter.createRegions(-fZONE.NX(),
                                 -fZONE.NY(),
                                  fZONE.FIRST()-1)) {
        std::ostream& out = fSession.cout();
        out << "plotter_ZONE :" 
            << " can't create a grid region"
            << std::endl;
      } else {
        // Initialize the region with PAW defaults :
        AIDA::IPlotterRegion& region = aPlotter.currentRegion();
        regionInitialize(region);
        //consider that it is a new area. Then next h/plot will not do "next".
        f_start_ZONE = true;
      }
  
    } else {
      if(!aPlotter.createRegions(fZONE.NX(),
                                 fZONE.NY(),
                                 fZONE.FIRST()-1)) {
        std::ostream& out = fSession.cout();
        out << "plotter_ZONE :" 
            << " can't create regions"
            << std::endl;
      } else {
        // Initialize region with PAW defaults :
        int number = aPlotter.numberOfRegions();
        for(int index=0;index<number;index++) {
          AIDA::IPlotterRegion* region = aPlotter.region(index);
          if(region) regionInitialize(*region);
        }
        aPlotter.setCurrentRegionNumber(fZONE.FIRST()-1);
        f_start_ZONE = true;
      }

    }
 
    fZONE_K = NotFound;

    fDoZONE = false;
  }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void OpenPAW::PAW_Manager::regionInitialize(AIDA::IPlotterRegion& aRegion){
  aRegion.setParameter("viewportRegion.visible","TRUE");
  aRegion.setParameter("viewportRegion.horizontalBorder","0");
  aRegion.setParameter("viewportRegion.verticalBorder","0");
  aRegion.setParameter("viewportRegion.horizontalShadow","0");
  aRegion.setParameter("viewportRegion.verticalShadow","0");

  aRegion.setParameter("viewportRegion.showHighlight","FALSE");
  aRegion.setParameter("viewportRegion.sensitiveBorder","FALSE");

  aRegion.setParameter("legendRegionVisible","FALSE");
  aRegion.setParameter("legendAutomated","FALSE");
  aRegion.setParameter("legendRegion.viewportRegion.horizontalShadow","0");
  aRegion.setParameter("legendRegion.viewportRegion.verticalShadow","0");
  aRegion.setParameter("legendRegion.viewportRegion.horizontalBorder","1");
  aRegion.setParameter("legendRegion.viewportRegion.verticalBorder","1");
  aRegion.setParameter("legendRegion.viewportRegion.horizontalOriginOffset","1");
  aRegion.setParameter("legendRegion.viewportRegion.verticalOriginOffset","1");

  std::string s;
  for(int index=0;index<10;index++) {

    //FIXME : do the below as needed.
    std::string field;

    Lib::smanip::printf(field,64,"plotter.rightHatchStyle[%d].",index);
    aRegion.setParameter(field+"color","0 0 0");
    aRegion.setParameter(field+"linePattern",SOLID);
    aRegion.setParameter(field+"lineWidth","1");
    aRegion.setParameter(field+"angle","0.7854");
    aRegion.setParameter(field+"spacing","0.05");
    aRegion.setParameter(field+"offset","0");

    Lib::smanip::printf(field,64,"plotter.leftHatchStyle[%d].",index);
    aRegion.setParameter(field+"color","0 0 0");
    aRegion.setParameter(field+"linePattern",SOLID);
    aRegion.setParameter(field+"lineWidth","1");
    aRegion.setParameter(field+"angle","0.7854");
    aRegion.setParameter(field+"spacing","0.05");
    aRegion.setParameter(field+"offset","0");

    aRegion.setParameter("plotter.errorsStyle["+s+"].visible","FALSE");

  }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool OpenPAW::PAW_Manager::plotterNext(AIDA::IPlotter& aPlotter,bool aSuperPose){
  if(aSuperPose) {
    return true;
  } else {
    if(f_start_ZONE) { //stay in current zone (the first one).
      f_start_ZONE = false;
    } else {
      aPlotter.next();
    }

    if(aPlotter.currentRegionNumber()==0) {
      aPlotter.clearRegions();
    } else {
      aPlotter.currentRegion().clear();        
    }
    return true;
  }
}


void OpenPAW::PAW_Manager::setAxisLabels(AIDA::IPlotterRegion& aRegion,const std::string& a_axis,int a_labels_id,int a_labels_al){ 
  //  SET NDVX 3.15                 | 3 div, lab id 1, 5=center on bin
  // labels_id is 1
  // labels_al is 5
 {std::map<int, std::vector<std::string> >::const_iterator lab_it = 
    fLabels.find(a_labels_id);
  if(lab_it==fLabels.end()) {
    //  SET NDVX 3.05
    aRegion.setParameter("plotter."+a_axis+"Axis.labelsEnforced","FALSE");
  } else {
    aRegion.setParameter("plotter."+a_axis+"Axis.labelsEnforced","TRUE");
    std::string s;
    s += "[";
    const std::vector<std::string>& labels = (*lab_it).second;
    std::vector<std::string>::const_iterator it;
    for(it=labels.begin();it!=labels.end();++it) {
      if(it!=labels.begin()) s += ",";
      s+= "\"";
      s+= *it;
      s+= "\"";
    }
    s += "]";
    aRegion.setParameter("plotter."+a_axis+"Axis.labels",s.c_str());
  }}

  // 8=bottom -> up
  std::string field = "plotter."+a_axis+"Axis.labelsStyle.";
  //if(a_labels_al==1) { // center on tick
  //} else if(a_labels_al==5) { //center on bin
  if(a_labels_al==8) { // 8=bottom -> up
    aRegion.setParameter("plotter."+a_axis+"Axis.labelsStyle.enforced","TRUE");
    //FIXME : handle x, y, z and shape
    aRegion.setParameter(field+"orientationX","0 1 0");
    aRegion.setParameter(field+"orientationY","-1 0 0");
    aRegion.setParameter(field+"horizontalJustification","RIGHT");
  } else {
    aRegion.setParameter
      ("plotter."+a_axis+"Axis.labelsStyle.enforced","FALSE");
  }
}

void OpenPAW::PAW_Manager::regionSetLayout(AIDA::IPlotterRegion& aRegion){

  aRegion.setParameter("viewportRegion.background","TRUE");
  aRegion.setParameter("viewportRegion.backgroundColor","1 1 1");

  aRegion.setParameter("plotter.titleUp","FALSE");
  aRegion.setParameter("plotter.wallStyle.visible","FALSE");
  aRegion.setParameter("plotter.gridStyle.visible","FALSE");
  aRegion.setParameter("plotter.innerFrameStyle.visible","TRUE");
  aRegion.setParameter("plotter.rightAxisVisible","FALSE");
  aRegion.setParameter("plotter.topAxisVisible","FALSE");

  aRegion.setParameter("plotter.colorMappingVisible","FALSE");
  aRegion.setParameter("plotter.colorMappingAxisLabeling","MIN_MAX");

  double YHTI = fATTs.value("YHTI"); 
  double TSIZ = fATTs.value("TSIZ"); 

  aRegion.setParameter("titleRegion.styles[0].color","0 0 0");

  std::string HTIT = fOPTs.value("HTIT");
  if(HTIT=="HTIT") {
    aRegion.setParameter("plotter.titleStyle.visible","TRUE");
    aRegion.setParameter("plotter.titleAutomated","TRUE");
    aRegion.setParameter("plotter.titleUp","FALSE");
    aRegion.setParameter("plotter.titleHeight",Slash::to<double>(TSIZ));
    aRegion.setParameter("plotter.titleToAxis",Slash::to<double>(YHTI));
  } else if(HTIT=="UTIT") {
    if(fUTIT=="") {
      aRegion.setParameter("plotter.titleStyle.visible","FALSE");
    } else {
      std::string PTIT = fOPTs.value("PTIT"); //GB option.
      if(PTIT=="PTIT") {
        aRegion.setParameter("titleRegionVisible","FALSE");
        aRegion.setParameter("plotter.titleStyle.visible","TRUE");
        aRegion.setParameter("plotter.titleUp","FALSE");
        aRegion.setParameter("plotter.titleHeight",
                             Slash::to<double>(TSIZ));
        aRegion.setParameter("plotter.titleToAxis",
                             Slash::to<double>(YHTI));
      } else { //RTIT
        aRegion.setParameter("titleRegionVisible","TRUE");
        aRegion.setParameter("plotter.titleStyle.visible","FALSE");
      }
      aRegion.setParameter("plotter.titleAutomated","FALSE");
      aRegion.setParameter("plotter.title",fUTIT);
    }
  }

  ///////////////////////
  /// Axes //////////////
  ///////////////////////
  aRegion.setParameter("plotter.xAxis.modeling","hplot");
  aRegion.setParameter("plotter.yAxis.modeling","hplot");
  aRegion.setParameter("plotter.zAxis.modeling","hplot");
  aRegion.setParameter("plotter.xAxis.titleStyle.encoding","PAW");
  aRegion.setParameter("plotter.yAxis.titleStyle.encoding","PAW");
  aRegion.setParameter("plotter.zAxis.titleStyle.encoding","PAW");

  aRegion.setParameter("plotter.xAxis.lineStyle.color","0 0 0");
  aRegion.setParameter("plotter.xAxis.ticksStyle.color","0 0 0");
  aRegion.setParameter("plotter.xAxis.labelsStyle.color","0 0 0");
  aRegion.setParameter("plotter.xAxis.titleStyle.color","0 0 0");
  aRegion.setParameter("plotter.xAxis.titleToAxisJustification","RIGHT");
  aRegion.setParameter("plotter.xAxis.magStyle.color","0 0 0");

  aRegion.setParameter("plotter.yAxis.lineStyle.color","0 0 0");
  aRegion.setParameter("plotter.yAxis.ticksStyle.color","0 0 0");
  aRegion.setParameter("plotter.yAxis.labelsStyle.color","0 0 0");
  aRegion.setParameter("plotter.yAxis.titleStyle.color","0 0 0");
  aRegion.setParameter("plotter.yAxis.titleToAxisJustification","RIGHT");
  aRegion.setParameter("plotter.yAxis.magStyle.color","0 0 0");

  aRegion.setParameter("plotter.zAxis.lineStyle.color","0 0 0");
  aRegion.setParameter("plotter.zAxis.ticksStyle.color","0 0 0");
  aRegion.setParameter("plotter.zAxis.labelsStyle.color","0 0 0");
  aRegion.setParameter("plotter.zAxis.titleStyle.color","0 0 0");
  aRegion.setParameter("plotter.zAxis.titleToAxisJustification","RIGHT");
  aRegion.setParameter("plotter.zAxis.magStyle.color","0 0 0");

  //Axis title SIZe
  double ASIZ = fATTs.value("ASIZ");
  aRegion.setParameter
    ("plotter.xAxis.titleHeight",Slash::to<double>(ASIZ));
  aRegion.setParameter
    ("plotter.yAxis.titleHeight",Slash::to<double>(ASIZ));

  //y length of X axis ticks. Def 0.015 
  double XTIC = fATTs.value("XTIC"); 
  aRegion.setParameter
    ("plotter.xAxis.tickLength",Slash::to<double>(XTIC));

  //x length of Y axis ticks. Def 0.015 
  double YTIC = fATTs.value("YTIC"); 
  aRegion.setParameter
    ("plotter.yAxis.tickLength",Slash::to<double>(YTIC));

  //y distance of x tick label to data frame. Def 0.02 
  double YVAL = fATTs.value("YVAL"); 
  aRegion.setParameter
    ("plotter.xAxis.labelToAxis",Slash::to<double>(YVAL));

  //x distance of y tick label to data frame. Def 0.02 
  double XVAL = fATTs.value("XVAL"); 
  aRegion.setParameter
    ("plotter.yAxis.labelToAxis",Slash::to<double>(XVAL));

  //y distance of x title to data frame. Def 0.04  
  double YLAB = fATTs.value("YLAB"); 
  aRegion.setParameter
    ("plotter.xAxis.titleToAxis",Slash::to<double>(YLAB));

  //x distance of y title to data frame. Def 0.07  
  double XLAB = fATTs.value("XLAB"); 
  aRegion.setParameter
    ("plotter.yAxis.titleToAxis",Slash::to<double>(XLAB));

  // axis labels SIZe. Def : 0.28F.
  double VSIZ = fATTs.value("VSIZ"); 
  aRegion.setParameter
    ("plotter.xAxis.labelHeight",Slash::to<double>(VSIZ));
  aRegion.setParameter
    ("plotter.yAxis.labelHeight",Slash::to<double>(VSIZ));
  aRegion.setParameter
    ("plotter.zAxis.labelHeight",Slash::to<double>(VSIZ));
  aRegion.setParameter
    ("plotter.colorMappingAxis.labelHeight",Slash::to<double>(VSIZ));

  if(fOPTs.value("LINX")=="LOGX") {
    aRegion.setParameter("plotter.xAxisLogScale","TRUE");
  } else {
    aRegion.setParameter("plotter.xAxisLogScale","FALSE");
  }
  if(fOPTs.value("LINY")=="LOGY") {
    aRegion.setParameter("plotter.yAxisLogScale","TRUE");
  } else {
    aRegion.setParameter("plotter.yAxisLogScale","FALSE");
  }
  if(fOPTs.value("LINZ")=="LOGZ") {
    aRegion.setParameter("plotter.zAxisLogScale","TRUE");
  } else {
    aRegion.setParameter("plotter.zAxisLogScale","FALSE");
  }

 {double d_NDVX = fATTs.value("NDVX"); 
  if(d_NDVX<0) d_NDVX *= -1; //FIXME : handle negative case.
  int NDVX = (int)d_NDVX;
  if(fZONE.NX()>=2) { 
    //seen with pawex22. Prim-div seems to be divided by the number of x zones.
    int N3 = NDVX/10000;    //See SbAxisHPLOT::paint
    int N2 = (NDVX - N3*10000)/100;    
    int N1 = NDVX - N2*100 - N3*10000;    
    NDVX = N1/fZONE.NX() + N2 * 100 + N3 * 10000;
  }
  aRegion.setParameter("plotter.xAxis.divisions",Lib::smanip::tostring(NDVX));
  int rest = (int)((d_NDVX - NDVX)*100);
  int labels_id = rest/10;
  int labels_al = rest - labels_id*10;
  setAxisLabels(aRegion,"x",labels_id,labels_al);}

 {double d_NDVY = fATTs.value("NDVY"); 
  if(d_NDVY<0) d_NDVY *= -1; //FIXME : handle negative value.
  int NDVY = (int)d_NDVY;
  aRegion.setParameter("plotter.yAxis.divisions",Lib::smanip::tostring(NDVY));
  int rest = (int)((d_NDVY - NDVY)*100);
  int labels_id = rest/10;
  int labels_al = rest - labels_id*10;
  setAxisLabels(aRegion,"y",labels_id,labels_al);}


 {double d_NDVZ = fATTs.value("NDVZ"); 
  if(d_NDVZ<0) d_NDVZ *= -1; //FIXME : handle negative value
  int NDVZ = (int)d_NDVZ;
  aRegion.setParameter("plotter.zAxis.divisions",Lib::smanip::tostring(NDVZ));
  int rest = (int)((d_NDVZ - NDVZ)*100);
  int labels_id = rest/10;
  int labels_al = rest - labels_id*10;
  setAxisLabels(aRegion,"z",labels_id,labels_al);}

  //  BWID  //  1   // Box line WIDth      
  double BWID = fATTs.value("BWID"); 
  BWID = MAXIMUM(int(BWID/2.2),1);  //From graflib/higz/ikernel/islwsc.F
  aRegion.setParameter
    ("plotter.innerFrameStyle.lineWidth",Slash::to<double>(BWID));
  aRegion.setParameter
    ("plotter.xAxis.lineStyle.width",Slash::to<double>(BWID));
  aRegion.setParameter
    ("plotter.yAxis.lineStyle.width",Slash::to<double>(BWID));

  //  PWID  //  1   // Picture line WIDth  

  //  XWID  //  1   // X ticks WIDth       
  double XWID = fATTs.value("XWID"); 
  XWID = MAXIMUM(int(XWID/2.2),1);  //From graflib/higz/ikernel/islwsc.F
  aRegion.setParameter
    ("plotter.xAxis.ticksStyle.width",Slash::to<double>(XWID));

  //  YWID  //  1   // Y ticks WIDth       
  double YWID = fATTs.value("YWID"); 
  YWID = MAXIMUM(int(YWID/2.2),1);  //From graflib/higz/ikernel/islwsc.F
  aRegion.setParameter
    ("plotter.yAxis.ticksStyle.width",Slash::to<double>(YWID));

  //  CWID  // 0.50 // Color map WIDth     

  std::string NGRI = fOPTs.value("NGRI");
  if(NGRI=="NGRI") {
    aRegion.setParameter("plotter.gridStyle.visible","FALSE");
  } else {
    aRegion.setParameter("plotter.gridStyle.visible","TRUE");
  }

  int GRID = (int)fATTs.value("GRID"); 
  // o00p
  // o = 1 = horizontal only.
  // o = 2 = vertical only. 
  // o = else = horizontal and vertical.
  // p = 2 = dashed.
  // p = 3 = dotted.
  // p = 4 = dash-dotted.
 {int pattern = GRID % 10;
  int orientation = GRID / 1000;
  std::string spattern = SbLinePattern_s_solid;
  if(pattern==2) spattern = SbLinePattern_s_dashed;
  else if(pattern==3) spattern = SbLinePattern_s_dotted;
  else if(pattern==4) spattern = SbLinePattern_s_dash_dotted;
  aRegion.setParameter("plotter.gridStyle.linePattern",spattern);
  //FIXME : have a field orientation on SoStyle.
  if(orientation==1)
    aRegion.setParameter("plotter.gridStyle.options","horizontal");
  else if(orientation==2)
    aRegion.setParameter("plotter.gridStyle.options","vertical");
  else
    aRegion.setParameter("plotter.gridStyle.options","");}

  ///////////////////////
  /// infosRegion ///////
  ///////////////////////
 {aRegion.setParameter("infosRegionAttachedToDataFrame","TRUE");

  regionSetFON(aRegion,"infosRegion.textStyle.","CFON");
  
  aRegion.setParameter("infosRegionResizePolicy","ASPECT_RATIO");
  aRegion.setParameter("infosRegion.viewportRegion.horizontalShadow","0");
  aRegion.setParameter("infosRegion.viewportRegion.verticalShadow","0");
  aRegion.setParameter("infosRegion.viewportRegion.horizontalBorder","1");
  aRegion.setParameter("infosRegion.viewportRegion.verticalBorder","1");
  aRegion.setParameter("infosRegion.viewportRegion.horizontalOriginOffset","1");
  aRegion.setParameter("infosRegion.viewportRegion.verticalOriginOffset","1");
  aRegion.setParameter("infosRegionStyle","PAW");
  aRegion.setParameter("infosRegion.viewportRegion.backgroundColor","1 1 1");
  aRegion.setParameter("infosRegion.inBetweenSpaces","10");

  std::string NBAC = fOPTs.value("NBAC");
  if(NBAC=="NBAC") { //default
    aRegion.setParameter("infosRegion.viewportRegion.background","FALSE");
  } else if(NBAC=="BACK") {
    aRegion.setParameter("infosRegion.viewportRegion.background","TRUE");
  }

 {double SMGR = fATTs.value("SMGR"); 
  double SMGU = fATTs.value("SMGU"); 
  std::string s;
  Lib::smanip::printf(s,64,"%g %g",SMGR,SMGU);
  aRegion.setParameter("infosRegionOrigin",s);}

  std::string NSTA = fOPTs.value("NSTA");
  std::string NFIT = fOPTs.value("NFIT");
  if( (NSTA=="STA") || (NFIT=="FIT") )  {
    std::string opts;
    getInfosOptions(opts);
    aRegion.setParameter("infosRegionVisible","TRUE");
    aRegion.setParameter("plotter.infosStyle.options",opts);
  } else {
    aRegion.setParameter("infosRegionVisible","FALSE");
  }}

  // Fonts :
  regionSetFON(aRegion,"plotter.xAxis.labelsStyle.","VFON");
  regionSetFON(aRegion,"plotter.yAxis.labelsStyle.","VFON");
  regionSetFON(aRegion,"plotter.zAxis.labelsStyle.","VFON");
  regionSetFON(aRegion,"plotter.xAxis.titleStyle.","LFON");
  regionSetFON(aRegion,"plotter.yAxis.titleStyle.","LFON");
  regionSetFON(aRegion,"plotter.zAxis.titleStyle.","LFON");
  regionSetFON(aRegion,"plotter.titleStyle.","TFON");
  regionSetFON(aRegion,"titleRegion.styles[0].","GFON");

}

void OpenPAW::PAW_Manager::getFont(
 const std::string& aATT
,std::string& aFont
,double& aScale
,bool& aSmoothing
,bool& aHinting
){
  // ATT value code :
  //      2 : Hershey  
  //    -fp : PAW code. f is the font and p the "precision". From page 129.
  //     fp : ROOT code convention. From TAttText.cxx.

  aSmoothing = true;
  aHinting = true;

  int X_FON = (int)fATTs.value(aATT);

  aScale = 0.016/0.28; // majic number scaling for TTF.

  int FF = X_FON/10; //can be negative.

  //int prec = ::abs(X_FON-FF*10); //FIXME : handle that ?
  
  // HIGZ :
  if(FF==0) {  //X_FON = 2, FF = 0, prec = 2
    aFont = "Hershey";
    aScale = 1;
    return;

  // PAW code :
  } else if(FF==-1) {
    aFont = "TTF/timesi";
    return;
  } else if(FF==-2) {
    aFont = "TTF/timesbd";
    return;
  } else if(FF==-3) {
    aFont = "TTF/timesbi";
    return;

  } else if(FF==-4) {
    aFont = "TTF/Helvetica";
    return;
  } else if(FF==-5) {
    aFont = SbFont_ROOT_52; //ok?
    return;
  } else if(FF==-6) {
    aFont = SbFont_ROOT_62; //ok?
    return;
  } else if(FF==-7) {
    aFont = SbFont_ROOT_72; //ok?
    return;

  } else if(FF==-8) {
    aFont = "TTF/cour";
    return;
  } else if(FF==-9) {
    aFont = "TTF/couri";
    return;
  } else if(FF==-10) {
    aFont = "TTF/courbd";
    return;
  } else if(FF==-11) {
    aFont = "TTF/courbi";
    return;

  } else if(FF==-12) {
    aFont = "TTF/symbol";
    return;
  } else if(FF==-13) {
    aFont = "TTF/Times-Roman";
    return;
  } else if(FF==-14) {
    //ZapfDingbats

  } else if(FF==-15) {
    aFont = "TTF/timesi";
    return;
  } else if(FF==-16) {
    aFont = "TTF/timesbd";
    return;
  } else if(FF==-17) {
    aFont = "TTF/timesbi";
    return;

  } else if(FF==-18) {
    aFont = "TTF/Helvetica";
    return;
  } else if(FF==-19) {
  } else if(FF==-20) {
  } else if(FF==-21) {

  } else if(FF==-22) {
    aFont = "TTF/symbol";
    return;
  } else if(FF==-23) {
    aFont = "TTF/Times-Roman";
    return;
  } else if(FF==-24) {
    //ZapfDingbats

  // ROOT code :
  } else if(FF==1) {
    aFont = SbFont_ROOT_12;
    return;
  } else if(FF==2) {
    aFont = SbFont_ROOT_22;
    return;
  } else if(FF==3) {
    aFont = SbFont_ROOT_32;
    return;
  } else if(FF==4) {
    aFont = SbFont_ROOT_42;
    return;
  } else if(FF==5) {
    aFont = SbFont_ROOT_52;
    return;
  } else if(FF==6) {
    aFont = SbFont_ROOT_62;
    return;
  } else if(FF==7) {
    aFont = SbFont_ROOT_72;
    return;
  } else if(FF==8) {
    aFont = SbFont_ROOT_82;
    return;
  } else if(FF==9) {
    aFont = SbFont_ROOT_92;
    return;
  } else if(FF==10) {
    aFont = SbFont_ROOT_102;
    return;
  } else if(FF==11) {
    aFont = SbFont_ROOT_112;
    return;
  } else if(FF==12) {
    aFont = SbFont_ROOT_122;
    return;
  } else if(FF==13) {
    aFont = SbFont_ROOT_132;
    return;
  } else if(FF==14) {
    aFont = SbFont_ROOT_142;
    return;
  }

  std::ostream& out = fSession.cout();
  out << "OpenPAW::PAW_Manager::getFont :" 
      << " for ATT " << Lib::sout(aATT)
      << " unknown font code " << X_FON << "."
      << " Take Hershey (code 2)."
      << std::endl;

  aFont = "Hershey";
  aScale = 1;

}

void OpenPAW::PAW_Manager::plotterSetFON(AIDA::IPlotter& aPlotter,const std::string& aField,const std::string& aFONT){
  std::string font;
  double scale;
  bool smoothing,hinting;
  getFont(aFONT,font,scale,smoothing,hinting);
  aPlotter.setParameter(aField+"fontName",font);
  aPlotter.setParameter(aField+"smoothing",smoothing?"TRUE":"FALSE");
  aPlotter.setParameter(aField+"hinting",hinting?"TRUE":"FALSE");
  aPlotter.setParameter(aField+"scale",Slash::to<double>(scale));
}

void OpenPAW::PAW_Manager::regionSetFON(AIDA::IPlotterRegion& aRegion,const std::string& aField,const std::string& aFONT){
  std::string font;
  double scale;
  bool smoothing,hinting;
  getFont(aFONT,font,scale,smoothing,hinting);
  aRegion.setParameter(aField+"fontName",font);
  aRegion.setParameter(aField+"smoothing","TRUE");
  aRegion.setParameter(aField+"hinting","TRUE");
  aRegion.setParameter(aField+"scale",Slash::to<double>(scale));
}

std::string OpenPAW::PAW_Manager::get_MTYP(int aMTYP){
  // Documented :
  if(aMTYP==31) 
    return "ASTERISK";
  else if(aMTYP==30)
    return "PENTA_STAR_LINE";
  else if(aMTYP==29) 
    return "PENTA_STAR_FILLED";
  else if(aMTYP==28) 
    return "SWISS_CROSS_LINE";
  else if(aMTYP==27) 
    return "DIAMOND_LINE";
  else if(aMTYP==26) 
    return "TRIANGLE_UP_LINE";
  else if(aMTYP==25) 
    return "SQUARE_LINE";
  else if(aMTYP==24) 
    return "CIRCLE_LINE";
  else if(aMTYP==23) 
    return "TRIANGLE_DOWN_FILLED";
  else if(aMTYP==22) 
    return "TRIANGLE_UP_FILLED";
  else if(aMTYP==21) 
    return "SQUARE_FILLED";
  else if(aMTYP==20) 
    return "CIRCLE_FILLED";

  // Not documented :
  else if(aMTYP==1) 
    return "DOT";
  else if(aMTYP==2)
    return "PLUS";
  else if(aMTYP==3)
    return "ASTERISK";
  else if(aMTYP==4)
    return "CIRCLE_LINE";
  else if(aMTYP==5)
    return "CROSS";
  else if(aMTYP==6)
    return "SWISS_CROSS_FILLED";
  else if(aMTYP==7)
    return "SQUARE_FILLED";
  else if(aMTYP==8)
    return "CIRCLE_FILLED";

  // OpenPAW :
  else if(aMTYP==123) 
    return "TRIANGLE_DOWN_LINE";

  else {
    std::ostream& out = fSession.cout();
    out << "OpenPAW::PAW_Manager::get_MTYP :" 
        << " marker type " << aMTYP << " not yet supported." 
        << " We use dot for it for the moment." 
        << std::endl;
    return "DOT";
  }
}

void OpenPAW::PAW_Manager::setPlotterTitle(AIDA::IPlotter& aPlotter,const std::string& aTitle) {
  if(aTitle.empty()) {
    //aPlotter.setParameter("titleVisible","FALSE");
    //aPlotter.setParameter("title","");  
    //uppers done by the below aPlotter.setTitle(aTitle);
  } else {
    //aPlotter.setParameter("titleVisible","TRUE");
    //aPlotter.setParameter("title",aTitle.c_str());
    //uppers done by the below aPlotter.setTitle(aTitle);
  
    aPlotter.setParameter("titleGeometryAutomated","FALSE");
    aPlotter.setParameter("titleRegion.viewportRegion.visible","FALSE");

    //GSIZ : the height of the title region.
    //       def 0.28 for a page height of 20.
    double GSIZ = fATTs.value("GSIZ"); //def 0.28
    double cooking = 1.5;
    double h = (GSIZ/20.0)*cooking; //in [0,1]
    // let us say that "h" pixels corresponds to two chars width...
    double w = h/2.0 * aTitle.size();
    aPlotter.setParameter("titleRegion.viewportRegion.size",
      Slash::to(w)+" "+Slash::to(h));

    //YGTI : distance form the top of the page to
    //       to the top of title region.
    double YGTI = fATTs.value("YGTI"); //def 1.50
    // 2*x + w = 1
    double x = 0.5*(1-w);
    // YGTI+h+y=1
    double y = 1-h-(YGTI/20.);
    aPlotter.setParameter("titleRegion.viewportRegion.origin",
      Slash::to(x)+" "+Slash::to(y));
  
    //Have some int TBCO = (int)fATTs.value("TBCO"); ?
    //aPlotter.setParameter
    //  ("titleRegion.viewportRegion.backgroundColor","0.8 0.8 0.8");

    plotterSetFON(aPlotter,"titleRegion.styles[0].","GFON");
  }

  aPlotter.setTitle(aTitle); //will enforce also a SoPage.refresh()
}

bool OpenPAW::PAW_Manager::paletteSize(int aPALNB,int& aSize) {
  if((aPALNB<0)||(aPALNB>=(int)fPalettes.size())) {
    aSize = 0;
    return false;
  }
  Palette& pal = fPalettes[aPALNB];
  aSize = pal.size();
  return true;
} 

void OpenPAW::PAW_Manager::regionSetColormap(
 AIDA::IPlotterRegion& aRegion
,const std::string& aField
,double aMin
,double aMax
,int aPALNB
){
  // if palette contains : red, green, blue, yellow, magenta, cyan
  // and aMin, aMax is 0 and 6 then we build :
  //    0 red 1 green 2 blue 3 yellow 4 magenta 5 cyan 6
  if((aPALNB<0)||(aPALNB>=(int)fPalettes.size())) return;
  Palette& pal = fPalettes[aPALNB];
  if(!pal.size()) return;
 
  std::vector<std::string> scols;
 {for(unsigned int index=0;index<pal.size();index++) {
    unsigned char r = (unsigned char)(fColorTable[pal[index]].fRed * 255);
    unsigned char g = (unsigned char)(fColorTable[pal[index]].fGreen * 255);
    unsigned char b = (unsigned char)(fColorTable[pal[index]].fBlue * 255);
    std::string s;
    Lib::smanip::printf(s,64,"#%2x%2x%2x",r,g,b);
    Lib::smanip::replace(s," ","0");
    scols.push_back(s);
  }}

  std::vector< std::pair<double,double> > limits;
  unsigned int ncol = scols.size();
  double dw = (aMax - aMin)/ncol;
 {for(unsigned int index=0;index<ncol;index++) {
    double cmn = aMin + index * dw;
    double cmx = cmn + dw;
    limits.push_back(std::pair<double,double>(cmn,cmx));
  }}

  std::string cmap;
 {for(unsigned int index=0;index<ncol;index++) {
    cmap += Slash::to<double>(limits[index].first);
    cmap += " ";
    cmap += scols[index];
    cmap += " ";
  }}
  cmap += Slash::to<double>(limits[ncol-1].second);

  if(fSession.verboseLevel()) {
    std::ostream& out = fSession.cout();
    out << "OpenPAW::PAW_Manager::regionSetColormap :" 
        << " min= " << aMin
        << " max= " << aMax
        << " PALNB= " << aPALNB
        << " cmap= " << Slash::sout(cmap)
        << std::endl;
  }  

  aRegion.setParameter(aField,cmap);
}

void OpenPAW::PAW_Manager::plotFunction1D(
 AIDA::IPlotterRegion& aRegion
,AIDA::IFunction& aFunction
,double aXLOW
,double aXUP
,const std::string& aModel
){
  // For CERN/PAW, NPX is the number of points between
  // XLOW, XUP but by excluding these !

  if(aXUP<=aXLOW) {
    std::ostream& out = fSession.cout();
    out << "OpenPAW::PAW_Manager::plotFunction1D :"
        << " XUP (" << aXUP << ") <= XLOW (" << aXLOW << ") !"
        << std::endl;
    return;
  }
  if(fNPX<=1) {
    std::ostream& out = fSession.cout();
    out << "OpenPAW::PAW_Manager::plotFunction1D :"
        << " NPX <=1 !"
        << std::endl;
    return;
  }

  double df = (aXUP - aXLOW)/(fNPX-1+2);
  double xmn = aXLOW + df;
  double xmx = aXUP  - df;
  aFunction.annotation().addItem
    ("domain.xAxisNumberOfSteps",Lib::smanip::tostring(fNPX-1));
  aFunction.annotation().addItem
    ("domain.xAxisMinimum",Slash::to<double>(xmn));
  aFunction.annotation().addItem
    ("domain.xAxisMaximum",Slash::to<double>(xmx));

  aRegion.setParameter("plotter.shapeAutomated","FALSE");
  aRegion.setParameter("plotter.shape","XY");

  int number;
  region_field_value(aRegion,"numberOfPlottedFunction1D",number,
    fSession.printer());

  std::string field;
  Lib::smanip::printf(field,64,"plotter.functionStyle[%d].",number);

  aRegion.setParameter(field+"linePattern",SOLID);

  if(aModel==s_FUN_C) {
    //function coming from a /HISTOGRAM/FIT or /VECTOR/FIT.
    // It seems that FCOL, PCLI can change the color.
    // FCOL has the priority over PLCI if it had been touched.
  
    //  FWID  //  1   // Function line WIDth 
    double FWID = fATTs.value("FWID"); 
    int iFWID = MAXIMUM(int(FWID/2.2),1);  //From graflib/higz/ikernel/islwsc.F
    aRegion.setParameter(field+"lineWidth",Slash::to<int>(iFWID));

   {bool touched;
    int FCOL = colFromFCOL(touched);
    if(touched) {
      region_set_COL(aRegion,field,FCOL);
    } else {
      int PLCI = (int)fATTs.value("PLCI");
      region_set_COL(aRegion,field,PLCI);
    }}

    aRegion.setParameter(field+"painting","UNIFORM");

    region_set_DMOD(aRegion,field,(int)fATTs.value("DMOD")); 

    aRegion.plot(aFunction,s_retain); //FIXME : have retain for function ?

  } else if(aModel=="P") {

    // Modeling :
    aRegion.setParameter(field+"modeling",SbModeling_points);

    // Parameters :
    int PMCI = (int)fATTs.value("PMCI");
    region_set_COL(aRegion,field,PMCI);

    region_set_MSCF(aRegion,field,fATTs.value("MSCF"));
    REGION_SET_MTYP

    aRegion.plot(aFunction,s_retain);

  } else if(aModel=="L") {
    std::ostream& out = fSession.cout();
    out << "OpenPAW::PAW_Manager::plotFunction1D :"
        << " option L not yet supported."
        << std::endl;

  } else if(aModel=="*") {
    std::ostream& out = fSession.cout();
    out << "OpenPAW::PAW_Manager::plotFunction1D :"
        << " option * not yet supported."
        << std::endl;
  }

}

void OpenPAW::PAW_Manager::plotFunction2D(
 AIDA::IPlotterRegion& aRegion
,AIDA::IFunction& aFunction
,double aXLOW
,double aXUP
,double aYLOW
,double aYUP
,double aZLOW
,double aZUP
,int aNLEVEL
,int aPALNB
,const std::string& aModel
){
  aRegion.setParameter("plotter.shapeAutomated","FALSE");
  aRegion.setParameter("plotter.shape","XY");

  int number;
  region_field_value(aRegion,"numberOfPlottedFunction2D",number,
    fSession.printer());
  std::string field;
  Lib::smanip::printf(field,64,"plotter.functionStyle[%d].",number);

  //aRegion.setParameter(field+"color","0 0 0");
  aRegion.setParameter(field+"linePattern",SOLID);

  //  FWID  //  1   // Function line WIDth 
  double FWID = fATTs.value("FWID"); 
  int iFWID = MAXIMUM(int(FWID/2.2),1);  //From graflib/higz/ikernel/islwsc.F
  aRegion.setParameter(field+"lineWidth",Slash::to<int>(iFWID));

  if(aModel=="CONT_COL") {

    aRegion.setParameter(field+"modeling",SbModeling_curve);

    if(fOPTs.value("COLL")=="COLL") {
      aRegion.setParameter(field+"painting","BY_LEVEL");
      regionSetColormap(aRegion,field+"colorMapping",0,0,aPALNB);
    } else {
      aRegion.setParameter(field+"painting","BY_VALUE");
      regionSetColormap(aRegion,field+"colorMapping",aZLOW,aZUP,aPALNB);
    }

  } else if(aModel=="CONT_COL_FILLED") {

    aRegion.setParameter(field+"modeling",SbModeling_filled_curve);

    if(fOPTs.value("COLL")=="COLL") {
      aRegion.setParameter(field+"painting","BY_LEVEL");
      regionSetColormap(aRegion,field+"colorMapping",0,0,aPALNB);
    } else {
      aRegion.setParameter(field+"painting","BY_VALUE");
      regionSetColormap(aRegion,field+"colorMapping",aZLOW,aZUP,aPALNB);
    }

  } else if(aModel=="CONT_DMOD") {

    aRegion.setParameter(field+"modeling",SbModeling_curve);

    aRegion.setParameter(field+"painting","UNIFORM");
    region_set_COL(aRegion,field,colFromHCOL());

  }
  region_set_DMOD(aRegion,field,(int)fATTs.value("DMOD")); 

  aRegion.setParameter
    ("plotter.numberOfLevels",Lib::smanip::tostring(aNLEVEL));

  aFunction.annotation().addItem
    ("domain.xAxisNumberOfSteps",Lib::smanip::tostring(fNPX));
  aFunction.annotation().addItem
    ("domain.xAxisMinimum",Slash::to<double>(aXLOW));
  aFunction.annotation().addItem
    ("domain.xAxisMaximum",Slash::to<double>(aXUP));

  aFunction.annotation().addItem
    ("domain.yAxisNumberOfSteps",Lib::smanip::tostring(fNPY));
  aFunction.annotation().addItem
    ("domain.yAxisMinimum",Slash::to<double>(aYLOW));
  aFunction.annotation().addItem
    ("domain.yAxisMaximum",Slash::to<double>(aYUP));

  aRegion.plot(aFunction,s_retain);

}

void OpenPAW::PAW_Manager::plotCloud2D(
 AIDA::IPlotterRegion& aRegion
,AIDA::ICloud2D& aCloud
,const std::string& aModel
){
  int number;
  region_field_value
    (aRegion,"numberOfPlottedPoints2D",number,fSession.printer());
  std::string field;
  Lib::smanip::printf(field,64,"plotter.pointsStyle[%d].",number);

  aRegion.setParameter(field+"linePattern",SOLID);
  //aRegion.setParameter(field+"markerStyle","CROSS");

  if(aModel=="L") {
    // Lines joining points.

    // Modeling :
    aRegion.setParameter(field+"modeling",SbModeling_lines);

    // Parameters :
    aRegion.setParameter(field+"painting","UNIFORM");
    aRegion.setParameter(field+"color","0 0 0");
    region_set_DMOD(aRegion,field,(int)fATTs.value("DMOD")); 

  } else if(aModel=="P") {
    // (Poly)markers.

    // Modeling :
    aRegion.setParameter(field+"modeling",SbModeling_points);

    // Parameters :
    aRegion.setParameter(field+"painting","UNIFORM");
    int PMCI = (int)fATTs.value("PMCI");
    region_set_COL(aRegion,field,PMCI);
    region_set_MSCF(aRegion,field,fATTs.value("MSCF"));
    REGION_SET_MTYP

  } else if(aModel=="C") {
    // Smooth curve joining points.

    // Modeling :
    aRegion.setParameter(field+"modeling",SbModeling_curve);

    // Parameters :
    aRegion.setParameter(field+"painting","UNIFORM");

    //  FWID  //  1   // Function line WIDth 
    double FWID = fATTs.value("FWID"); 
    int iFWID = MAXIMUM(int(FWID/2.2),1);  //From graflib/higz/ikernel/islwsc.F
    aRegion.setParameter(field+"lineWidth",Slash::to<int>(iFWID));

   {bool touched;
    int FCOL = colFromFCOL(touched);
    if(touched) {
      region_set_COL(aRegion,field,FCOL);
    } else {
      int PLCI = (int)fATTs.value("PLCI");
      region_set_COL(aRegion,field,PLCI);
    }}

    region_set_DMOD(aRegion,field,(int)fATTs.value("DMOD")); 

  }

  aRegion.plot(aCloud,s_retain);
}

void OpenPAW::PAW_Manager::plotErrors2D(
 AIDA::IPlotterRegion& aRegion
,AIDA::IDataPointSet& aDPS
,int aMTYP
,double aMSCF
){
  int number;
  region_field_value
    (aRegion,"numberOfPlottedBins1D",number,fSession.printer());

  std::string field;
  Lib::smanip::printf(field,64,"plotter.binsStyle[%d].",number);

  aRegion.setParameter(field+"linePattern",SOLID);

  aRegion.setParameter(field+"visible","TRUE");
  int PMCI = (int)fATTs.value("PMCI");
  region_set_COL(aRegion,field,PMCI);
  aRegion.setParameter(field+"markerStyle",get_MTYP(aMTYP));
  region_set_MSCF(aRegion,field,aMSCF);
  aRegion.setParameter(field+"modeling",SbModeling_points);

  std::string e_field;
  Lib::smanip::printf(e_field,64,"plotter.errorsStyle[%d].",number);

  aRegion.setParameter(e_field+"visible","TRUE");
  aRegion.setParameter(e_field+"markerStyle","PLUS");

  //In CERN/PAW, HCOL change error rep color but not the one of the markers.
  aRegion.setParameter(e_field+"painting","UNIFORM");
  region_set_COL(aRegion,e_field,colFromHCOL());

  aRegion.plot(aDPS,s_retain);
}

void OpenPAW::PAW_Manager::plotHistogram1D(
 AIDA::IPlotterRegion& aRegion
,AIDA::IHistogram1D& aHistogram
,const std::string& aModel
,bool aLoopLineStyle
){
  //printf("debug : region_plot_histogram1D : \"%s\"\n",aModel.c_str());

  aRegion.setParameter("plotter.shapeAutomated","FALSE");
  aRegion.setParameter("plotter.shape","XY");

  int number;
  region_field_value
    (aRegion,"numberOfPlottedBins1D",number,fSession.printer());
  if(number==0) region_set_automated(aRegion,aHistogram);
  std::string field;
  Lib::smanip::printf(field,64,"plotter.binsStyle[%d].",number);

  aRegion.setParameter(field+"linePattern",SOLID);
  //aRegion.setParameter(field+"markerStyle","CROSS");

  std::string e_field;
  Lib::smanip::printf(e_field,64,"plotter.errorsStyle[%d].",number);
  aRegion.setParameter(e_field+"visible","FALSE");

  //  HWID  //  1   // Histogram line WIDth
  double HWID = fATTs.value("HWID"); 
  int iHWID = MAXIMUM(int(HWID/2.2),1);  //From graflib/higz/ikernel/islwsc.F
  aRegion.setParameter(field+"lineWidth",Lib::smanip::tostring(iHWID));

  if(aModel=="C") { // Smooth curve joining points.
    // Modeling :
    aRegion.setParameter(field+"modeling",SbModeling_curve);

    // Parameters :
    aRegion.setParameter(field+"painting","UNIFORM");

    //  FWID  //  1   // Function line WIDth 
    double FWID = fATTs.value("FWID"); 
    int iFWID = MAXIMUM(int(FWID/2.2),1);  //From graflib/higz/ikernel/islwsc.F
    aRegion.setParameter(field+"lineWidth",Slash::to<int>(iFWID));

   {bool touched;
    int FCOL = colFromFCOL(touched);
    if(touched) {
      region_set_COL(aRegion,field,FCOL);
    } else {
      int PLCI = (int)fATTs.value("PLCI");
      region_set_COL(aRegion,field,PLCI);
    }}

    region_set_DMOD(aRegion,field,(int)fATTs.value("DMOD")); 

    aRegion.plot(aHistogram,s_retain);

  } else if(aModel=="L") { // Line joining top middle of bins.
    // Modeling :
    aRegion.setParameter(field+"modeling",SbModeling_lines);

    // Parameters :
    aRegion.setParameter(field+"painting","UNIFORM");
    aRegion.setParameter(field+"color","0 0 0");
    region_set_DMOD(aRegion,field,(int)fATTs.value("DMOD")); 

    aRegion.plot(aHistogram,s_retain);

  } else if(aModel=="E") { // Errors. 
    //FIXME : For a vector error = sqrt(value).
    // Modeling :
    aRegion.setParameter(field+"modeling",SbModeling_points);

    // Parameters :
    aRegion.setParameter(field+"painting","UNIFORM");
    aRegion.setParameter(field+"color","0 0 0");
    region_set_MSCF(aRegion,field,fATTs.value("MSCF"));
    REGION_SET_MTYP

    std::string e_field;
    Lib::smanip::printf(e_field,64,"plotter.errorsStyle[%d].",number);
    aRegion.setParameter(e_field+"visible","TRUE");
    region_set_COL(aRegion,e_field,colFromHCOL());

    // in case markerStyle is SbMarkerPlus)
    aRegion.setParameter(e_field+"lineWidth",Lib::smanip::tostring(iHWID));

    aRegion.plot(aHistogram,s_retain);

  } else if(aModel=="P") { // Points at the top middle of bins.
    // Modeling :
    aRegion.setParameter(field+"modeling",SbModeling_points);

    // Parameters :
    aRegion.setParameter(field+"painting","UNIFORM");
    aRegion.setParameter(field+"color","0 0 0");
    region_set_MSCF(aRegion,field,fATTs.value("MSCF"));
    REGION_SET_MTYP

    aRegion.plot(aHistogram,s_retain);

  } else if(aModel=="*") { // '*' at the top middle of bins.
    // Modeling :
    aRegion.setParameter(field+"modeling",SbModeling_points);

    // Parameters :
    aRegion.setParameter(field+"painting","UNIFORM");
    aRegion.setParameter(field+"color","0 0 0");
    aRegion.setParameter(field+"markerSize","9");
    aRegion.setParameter(field+"markerStyle","ASTERISK");

    aRegion.plot(aHistogram,s_retain);

  } else if(aModel=="B") { // Bounding line of bins reduced to half.
    // Modeling :
    aRegion.setParameter(field+"modeling",SbModeling_bar_chart);
    // Parameters :
    aRegion.setParameter(field+"painting","UNIFORM");
    aRegion.setParameter(field+"color","0 0 0");
    aRegion.setParameter(field+"linePattern",SOLID);
    aRegion.setParameter(field+"barOffset",
      Lib::smanip::tostring(fATTs.value("BARO")));
    aRegion.setParameter(field+"barWidth",
      Lib::smanip::tostring(fATTs.value("BARW")));

    setHTYP(aRegion,number,
      (int)fATTs.value("HTYP"),fATTs.value("BARO"),fATTs.value("BARW"));

    aRegion.plot(aHistogram,s_retain);

  } else { // Top lines of bins. Default rep    
    // Modeling :
    if(fContourMode) { 
      aRegion.setParameter(field+"modeling",SbModeling_top_lines);
      std::string spattern = SOLID;
      //PAW loops on : SOLID, DOTTED, DASHED, DASH_DOTTED.
      if(aLoopLineStyle) {
        int lineStyle = number % 5;
             if(lineStyle==0) spattern = SOLID;
        else if(lineStyle==1) spattern = DASHED;
        else if(lineStyle==2) spattern = DOTTED;
        else if(lineStyle==3) spattern = DASH_DOTTED;
        else if(lineStyle==5) spattern = SOLID;
      } 
      aRegion.setParameter(field+"linePattern",spattern);
    } else {
      aRegion.setParameter(field+"modeling",SbModeling_boxes);
    }
    setHTYP(aRegion,number,
      (int)fATTs.value("HTYP"),fATTs.value("BARO"),fATTs.value("BARW"));

    aRegion.setParameter(field+"painting","UNIFORM");
    region_set_COL(aRegion,field,colFromHCOL());

    aRegion.plot(aHistogram,s_retain);
  }
}

void OpenPAW::PAW_Manager::addLegend(
 AIDA::IPlotterRegion& aRegion
,double aX
,double aY
,const std::string& aTEXT
,const std::string& //aModel
,int aIATT
,bool aSuperpose
){
  // aX,aY are in axes coordinates.
  std::string stext = aTEXT;
  Lib::smanip::replace(stext,"\"","\\\"");

  if(aSuperpose) { //Put all in one SoLegendRegion.

    aRegion.setParameter("legendRegionVisible","TRUE");
    aRegion.setParameter("legendAutomated","FALSE");
    aRegion.setParameter("legendRegionSize","0.5 0.1");
  
    aRegion.setParameter("legendRegionAttachedToInfosRegion","FALSE");
    aRegion.setParameter("legendRegionOriginUnit","AXIS");

    //then the last addLegend will enforce the origin.
   {std::string s;
    Lib::smanip::printf(s,64,"%g %g",aX,aY);
    aRegion.setParameter("legendRegionOrigin",s);}
  
    aRegion.setParameter("legendRegion.viewportRegion.visible","FALSE");

    std::string s = aRegion.parameterValue("legendRegion.text");
    Lib::smanip::strip(s,Lib::smanip::leading,'[');
    Lib::smanip::strip(s,Lib::smanip::trailing,']');
    Lib::smanip::strip(s);
    if(s=="\"\"") s = "";

    std::vector<std::string> words = Lib::smanip::words(s,",");
    unsigned int number = words.size();
    //printf("debug : number %d |%s|\n",number,aTEXT.c_str());
  
    std::string si = Lib::smanip::tostring(number);
    std::string field = "legendRegion.styles["+si+"]."; //SoStyle !
    aRegion.setParameter(field+"color","0 0 0");
    aRegion.setParameter(field+"encoding","PAW");
  
    regionSetFON(aRegion,field,"CFON");

    //if(aModel=="M") {
    // IATT is the marker type. 
    //}
    aRegion.setParameter(field+"markerStyle",get_MTYP(aIATT));
  
    region_set_MSCF(aRegion,field,fATTs.value("MSCF"));
  
    if(number) s += ",";
    s += "\""+stext+"\"";
    aRegion.setParameter("legendRegion.text","["+s+"]");

  } else { // One SoLegendRegion per /GRAPHICS/HPLOT/KEY

    aRegion.setParameter("legendRegionVisible","FALSE");

    std::string sir = aRegion.parameterValue("numberOfLegendRegions");
    if(!sir.size()) return;
    std::string rfield = "legendRegions["+sir+"].";
  
    aRegion.setParameter(rfield+"visible","TRUE");

    double CSIZ = fATTs.value("CSIZ");   //0.28 for a YSIZ of 20
    double cooking = 2.2;
    double h = (CSIZ/20.0)*cooking; //in [0,1]
    // let us say that "h" pixels corresponds to two chars width...
    double w = h/2.0 * aTEXT.size();
    aRegion.setParameter(rfield+"size",Slash::to(w)+" "+Slash::to(h));
  
    // so that marker is at (aX,aY)
    aRegion.setParameter(rfield+"verticalJustification","MIDDLE");
    aRegion.setParameter(rfield+"markerOnLeftBorder","TRUE");
  
    aRegion.setParameter(rfield+"originUnit","AXIS");
  
   {std::string s;
    Lib::smanip::printf(s,64,"%g %g",aX,aY);
    aRegion.setParameter(rfield+"origin",s);}
  
    aRegion.setParameter(rfield+"viewportRegion.visible","FALSE");
  
    std::string s = aRegion.parameterValue(rfield+"text");
    Lib::smanip::strip(s,Lib::smanip::leading,'[');
    Lib::smanip::strip(s,Lib::smanip::trailing,']');
    Lib::smanip::strip(s);
    if(s=="\"\"") s = "";
    std::vector<std::string> words = Lib::smanip::words(s,",");
    unsigned int number = words.size();
    //printf("debug : number %d |%s|\n",number,aTEXT.c_str());

    std::string si = Lib::smanip::tostring(number);
    std::string field = rfield+"styles["+si+"]."; //SoStyle !
    int PMCI = (int)fATTs.value("PMCI");
    region_set_COL(aRegion,field,PMCI);
    aRegion.setParameter(rfield+"textInBlack","TRUE");

    aRegion.setParameter(field+"encoding","PAW");
  
    regionSetFON(aRegion,field,"CFON");

    //if(aModel=="M") {
    // IATT is the marker type. 
    //}
    aRegion.setParameter(field+"markerStyle",get_MTYP(aIATT));
  
    region_set_MSCF(aRegion,field,fATTs.value("MSCF"));

    if(number) s += ",";
    s += "\""+stext+"\"";

    aRegion.setParameter(rfield+"text","["+s+"]");
  }
}

void OpenPAW::PAW_Manager::setHTYP(
 AIDA::IPlotterRegion& aRegion
,int aIndex
,int aHTYP
,double aBARO
,double aBARW
){
  // PAW encoding (manual p 122) :
  // See also pagpri PAW_hatch.
 
  // code between [1,25] are specials. See SoPlotter/PAW_hatch.

  int code = aHTYP;
  if(code==0) code = 255; //No hatching.
  int i = code / 100;
  int j = (code - i * 100)/10;
  int k = code - i * 100 - j * 10;

  //printf("debug : %d : %d %d %d\n",code,i,j,k);

  if(i==0) return;

  // j-hatching on rightHatchStyle :
  // k-hatching on leftHatchStyle :

  std::string sindex;
  Lib::smanip::printf(sindex,32,"%d",aIndex);

  std::string field = "plotter.binsStyle["+sindex+"].";

  if((j==5)&&(k==5))
    aRegion.setParameter(field+"hatching","NONE");
  else if((j!=5)&&(k==5))
    aRegion.setParameter(field+"hatching","RIGHT");
  else if((j==5)&&(k!=5))
    aRegion.setParameter(field+"hatching","LEFT");
  else if((j!=5)&&(k!=5))
    aRegion.setParameter(field+"hatching","LEFT_AND_RIGHT");

  std::string ss;

  // In CERN/PAW, hatch spacing is fixed for the page (not for the region) !
  int NX = 0;
  int NY = 0;
  if(fZONE.SUPER()) {
    NX = -fZONE.NX();
    NY = -fZONE.NY();
  } else {
    NX = fZONE.NX();
    NY = fZONE.NY();
  }

  double spacing = NY * i * 0.004;

  if(j!=5) {
    field = "plotter.rightHatchStyle["+sindex+"].";

    aRegion.setParameter(field+"color","0 0 0");

    Lib::smanip::printf(ss,32,"%g",spacing);
    aRegion.setParameter(field+"spacing",ss);

    double angle = (j==4?45:j*10);
    angle = 180 - angle;
    angle *= Lib::mmanip::pi() / 180;
    Lib::smanip::printf(ss,32,"%g",angle);
    aRegion.setParameter(field+"angle",ss);

    aRegion.setParameter(field+"barOffset",Slash::to<double>(aBARO));
    aRegion.setParameter(field+"barWidth",Slash::to<double>(aBARW));
  }

  if(k!=5) {
    field = "plotter.leftHatchStyle["+sindex+"].";

    aRegion.setParameter(field+"color","0 0 0");

    Lib::smanip::printf(ss,32,"%g",spacing);
    aRegion.setParameter(field+"spacing",ss);

    double angle = (k==4?45:k*10);
    angle *= Lib::mmanip::pi() / 180;
    Lib::smanip::printf(ss,32,"%g",angle);
    aRegion.setParameter(field+"angle",ss);

    aRegion.setParameter(field+"barOffset",Slash::to<double>(aBARO));
    aRegion.setParameter(field+"barWidth",Slash::to<double>(aBARW));
  }

}

void OpenPAW::PAW_Manager::warn_metaf_ps(){
  if(f_warn_metaf_ps) return; //done
  std::ostream& out = fSession.cout();
  out << "pagraf :" 
      << " /GRAPHICS/METAFILE : "
      << " WARNING : ps file will be produced at the /FORTRAN/CLOSE of unit." 
      << std::endl;
  f_warn_metaf_ps = true;
}
