#include <AIDA/IFunctionCatalog.h>
#include <AIDA/IAxis.h>

#include <Slash/Core/ISession.h>
#include <Slash/Core/IWriter.h>
#include <Slash/Tools/AIDA.h>

#include <amanip/find.h>
#include <amanip/create.h>
#include <amanip/region.h>

#include <Lib/Out.h>
#include <Lib/sout.h>

#include <OpenPAW/PAW_Manager.h>

#include <HEPVis/SbStyles.h> //#define only.

#include "RegionHelpers.h"

#define s_retain "retain=true"

//////////////////////////////////////////////////////////////////////////////
inline void region_plot_histogram2D(
 Slash::Core::ISession& aSession
,AIDA::IPlotterRegion& aRegion
,AIDA::IHistogram2D& aHistogram
,const std::string& aModel
,Slash::Core::IWriter& aPrinter
,int aNLEVEL
,int aPALNB
,OpenPAW::PAW_Manager& aPAW_Manager
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int number;
  region_field_value(aRegion,"numberOfPlottedBins2D",number,aPrinter);
  if(number==0) region_set_automated(aRegion,aHistogram);
  std::string field;
  Lib::smanip::printf(field,64,"plotter.binsStyle[%d].",number);

  aRegion.setParameter(field+"linePattern",SOLID);
  //aRegion.setParameter(field+"markerStyle","CROSS");

  std::string e_field;
  Lib::smanip::printf(e_field,64,"plotter.errorsStyle[%d].",number);
  aRegion.setParameter(e_field+"visible","FALSE");

  ATTs_type& atts = aPAW_Manager.atts();
  OPTs_type& opts = aPAW_Manager.opts();
  if( (aModel==s_SURF) || 
      (aModel==s_SURF_CONT) ||
      (aModel==s_SURF_LIGHT) ||
      (aModel==s_SURF_LEVEL) ||
      (aModel==s_SURF_LEVEL_NO_EDGE) ||
      (aModel==s_SURF_COL) ) {

    aRegion.setParameter(field+"modeling",SbModeling_curve);

    if(aModel==s_SURF) {
      aRegion.setParameter(field+"painting","UNIFORM");
      aRegion.setParameter(field+"color","1 1 1");
      aRegion.setParameter(field+"areaStyle","EDGED");
      aRegion.setParameter(field+"lineWidth","2");
      aRegion.setParameter(field+"lightModel",SbLightModel_base_color);

    } else if(aModel==s_SURF_CONT) {
      aRegion.setParameter(field+"painting","UNIFORM");
      aRegion.setParameter(field+"color","1 1 1");
      aRegion.setParameter(field+"areaStyle","EDGED");
      aRegion.setParameter(field+"lineWidth","2");
      aRegion.setParameter(field+"lightModel",SbLightModel_base_color);

      double mn = aHistogram.minBinHeight();
      double mx = aHistogram.maxBinHeight();
      aPAW_Manager.regionSetColormap
        (aRegion,field+"colorMapping",mn,mx,aPALNB);

    } else if(aModel==s_SURF_LIGHT) {
      aRegion.setParameter(field+"painting","UNIFORM");
      aRegion.setParameter(field+"color","1 1 1");

    } else if(aModel==s_SURF_LEVEL) {
      aRegion.setParameter(field+"painting","BY_LEVEL");
      double mn = aHistogram.minBinHeight();
      double mx = aHistogram.maxBinHeight();
      aPAW_Manager.regionSetColormap
        (aRegion,field+"colorMapping",mn,mx,aPALNB);
      aRegion.setParameter(field+"areaStyle","EDGED");
      aRegion.setParameter(field+"lineWidth","2");
      aRegion.setParameter(field+"lightModel",SbLightModel_base_color);

    } else if(aModel==s_SURF_LEVEL_NO_EDGE) {
      aRegion.setParameter(field+"painting","BY_LEVEL");
      double mn = aHistogram.minBinHeight();
      double mx = aHistogram.maxBinHeight();
      aPAW_Manager.regionSetColormap
        (aRegion,field+"colorMapping",mn,mx,aPALNB);
      aRegion.setParameter(field+"lightModel",SbLightModel_base_color);

    } else if(aModel==s_SURF_COL) {
      aRegion.setParameter(field+"painting","BY_VALUE");
      double mn = aHistogram.minBinHeight();
      double mx = aHistogram.maxBinHeight();
      aPAW_Manager.regionSetColormap
        (aRegion,field+"colorMapping",mn,mx,aPALNB);
    }

    aRegion.setParameter("plotter.wallStyle.visible","FALSE");
    aRegion.setParameter("plotter.gridStyle.visible","TRUE");

    aRegion.setParameter("plotter.colorMappingAttached","FALSE");
    double THETA = aPAW_Manager.THETA();
    double PHI = aPAW_Manager.PHI();
    aRegion.setParameter("plotter.theta",Slash::to<float>(THETA));
    aRegion.setParameter("plotter.phi",Slash::to<float>(PHI));
    aRegion.setParameter("plotter.shapeAutomated","FALSE");
    aRegion.setParameter("plotter.shape","XYZ");

    aRegion.plot(aHistogram,s_retain);

  } else if( (aModel==s_LEGO) || 
             (aModel==s_LEGO_LIGHT) ||
             (aModel==s_LEGO_LEVEL) ||
             (aModel==s_LEGO_COL) ) {

    aRegion.setParameter(field+"modeling",SbModeling_boxes);

    if(aModel==s_LEGO) {
      aRegion.setParameter(field+"painting","UNIFORM");
      aRegion.setParameter(field+"color","1 1 1");
      aRegion.setParameter(field+"areaStyle","EDGED");
      aRegion.setParameter(field+"lineWidth","2");
      aRegion.setParameter(field+"lightModel",SbLightModel_base_color);
    } else if(aModel==s_LEGO_LIGHT) {
      aRegion.setParameter(field+"painting","UNIFORM");
      aRegion.setParameter(field+"color","1 1 1");
      aRegion.setParameter(field+"areaStyle","EDGED");
      aRegion.setParameter(field+"lineWidth","2");
      //aRegion.setParameter(field+"lightModel",SbLightModel_base_color);
    } else if(aModel==s_LEGO_LEVEL) {
      aRegion.setParameter(field+"painting","BY_LEVEL");
      double mn = aHistogram.minBinHeight();
      double mx = aHistogram.maxBinHeight();
      aPAW_Manager.regionSetColormap
        (aRegion,field+"colorMapping",mn,mx,aPALNB);
      aRegion.setParameter(field+"areaStyle","EDGED");
      aRegion.setParameter(field+"lineWidth","2");
      aRegion.setParameter(field+"lightModel",SbLightModel_base_color);

    } else if(aModel==s_LEGO_COL) { //OpenPAW specific
      aRegion.setParameter(field+"painting","BY_VALUE");
      double mn = aHistogram.minBinHeight();
      double mx = aHistogram.maxBinHeight();
      aPAW_Manager.regionSetColormap
        (aRegion,field+"colorMapping",mn,mx,aPALNB);
    }

    aRegion.setParameter("plotter.wallStyle.visible","FALSE");
    aRegion.setParameter("plotter.gridStyle.visible","TRUE");

    aRegion.setParameter("plotter.colorMappingAttached","FALSE");
    double THETA = aPAW_Manager.THETA();
    double PHI = aPAW_Manager.PHI();
    aRegion.setParameter("plotter.theta",Slash::to<float>(THETA));
    aRegion.setParameter("plotter.phi",Slash::to<float>(PHI));
    aRegion.setParameter("plotter.shapeAutomated","FALSE");
    aRegion.setParameter("plotter.shape","XYZ");

    aRegion.plot(aHistogram,s_retain);

  } else if(aModel=="TEXT") {

    aRegion.setParameter(field+"modeling",SbModeling_texts);

    aRegion.setParameter(field+"painting","UNIFORM");
    aRegion.setParameter(field+"color","0 0 0");

    aRegion.setParameter("plotter.shapeAutomated","FALSE");
    aRegion.setParameter("plotter.shape","XY");

    aRegion.plot(aHistogram,s_retain);

  } else if(aModel=="BOX") {

    aRegion.setParameter(field+"modeling",SbModeling_wire_boxes);

    aRegion.setParameter(field+"painting","UNIFORM");
    aRegion.setParameter(field+"color","0 0 0");

    aRegion.setParameter("plotter.shapeAutomated","FALSE");
    aRegion.setParameter("plotter.shape","XY");

    aRegion.plot(aHistogram,s_retain);

  } else if(aModel=="COL") {
    // For :
    //   H/PLOT <h2d> COL
    // From CERN/PAW, the palette is set clearly by 
    //   PALETTE <num>
    // and not by :
    //   SET HCOL 0.<num>

    aRegion.setParameter(field+"modeling",SbModeling_solid);

    aRegion.setParameter(field+"painting","BY_VALUE");
    double mn = aHistogram.minBinHeight();
    double mx = aHistogram.maxBinHeight();
    aPAW_Manager.regionSetColormap(aRegion,field+"colorMapping",mn,mx,aPALNB);

    aRegion.setParameter("plotter.shapeAutomated","FALSE");
    aRegion.setParameter("plotter.shape","XY");

    aRegion.plot(aHistogram,s_retain);
  } else if( (aModel=="CONT_COL") || 
             (aModel=="CONT_COL_FILLED") ||
             (aModel=="CONT_DMOD") ) {
    AIDA::IAnnotation& annotation = aHistogram.annotation();
    std::string codelet = annotation.value("opaw.codelet");
    if(codelet=="") {
      Lib::Out out(aPrinter);
      out << "region_plot_histogram2D :" 
          << " histogram with title " << Lib::sout(aHistogram.title())
          << " does not have a function script."
          << " We use content of bins to draw contours."
          << Lib::endl;

      if(aModel=="CONT_COL_FILLED") {
        aRegion.setParameter(field+"modeling",SbModeling_filled_curve);
      } else {
        aRegion.setParameter(field+"modeling",SbModeling_curve);
      }  

      if( (aModel=="CONT_COL") || (aModel=="CONT_COL_FILLED") ){
        if(opts.value("COLL")=="COLL") {
          aRegion.setParameter(field+"painting","BY_LEVEL");
          aPAW_Manager.regionSetColormap
            (aRegion,field+"colorMapping",0,0,aPALNB);
        } else {
          aRegion.setParameter(field+"painting","BY_VALUE");
          double mn = aHistogram.minBinHeight();
          double mx = aHistogram.maxBinHeight();
          aPAW_Manager.regionSetColormap
            (aRegion,field+"colorMapping",mn,mx,aPALNB);
        }
      } else if(aModel=="CONT_DMOD") {
        aRegion.setParameter(field+"painting","UNIFORM");
        region_set_COL(aRegion,field,aPAW_Manager.colFromHCOL());
      }

      region_set_DMOD(aRegion,field,(int)atts.value("DMOD")); 
      aRegion.setParameter
        ("plotter.numberOfLevels",Lib::smanip::tostring(aNLEVEL));

      aRegion.setParameter("plotter.shapeAutomated","FALSE");
      aRegion.setParameter("plotter.shape","XY");

      aRegion.plot(aHistogram,s_retain);

    } else { //codelet!=""

      AIDA::IAnalysisFactory* aida = aPAW_Manager.aida();
      if(!aida) return;
      AIDA::ITree* memoryTree = aPAW_Manager.mem();
      if(!memoryTree) return;

      std::string funcName = aPAW_Manager.generateFunctionName();
      AIDA::IFunction* rep = 
        amanip::create_function
          (*aida,*memoryTree,funcName,codelet,codelet);
      if(!rep) {
        Lib::Out out(aPrinter);
        out << "region_plot_histogram2D :" 
            << " can't create function with codelet " << Lib::sout(codelet) 
            << Lib::endl;
        return;
      }
      
      double XLOW,XUP;
     {bool xauto;
      amanip::parameter(aRegion,"plotter.xAxisAutomated",xauto);
      if(xauto) {
        XLOW = aHistogram.xAxis().lowerEdge();
        XUP = aHistogram.xAxis().upperEdge();
      } else {
        amanip::parameter(aRegion,"plotter.xAxisMinimum",XLOW);
        amanip::parameter(aRegion,"plotter.xAxisMaximum",XUP);
      }}

      double YLOW,YUP;
     {bool yauto;
      amanip::parameter(aRegion,"plotter.yAxisAutomated",yauto);
      if(yauto) {
        YLOW = aHistogram.yAxis().lowerEdge();
        YUP = aHistogram.yAxis().upperEdge();
      } else {
        amanip::parameter(aRegion,"plotter.yAxisMinimum",YLOW);
        amanip::parameter(aRegion,"plotter.yAxisMaximum",YUP);
      }}
      

      double ZLOW = aHistogram.minBinHeight();
      double ZUP = aHistogram.maxBinHeight();
      aPAW_Manager.plotFunction2D(aRegion,*rep,
                                  XLOW,XUP,YLOW,YUP,ZLOW,ZUP,
                                  aNLEVEL,aPALNB,aModel);
    }

  } else {

    // Default is random points.

    aRegion.setParameter(field+"modeling",SbModeling_points);

    aRegion.setParameter(field+"painting","UNIFORM");
    aRegion.setParameter(field+"color","0 0 0");

    aRegion.setParameter(field+"markerStyle","DOT");

    aRegion.setParameter("plotter.shapeAutomated","FALSE");
    aRegion.setParameter("plotter.shape","XY");

    aRegion.plot(aHistogram,s_retain);

  }
}
