#include <AIDA/IPlotterRegion.h>
#include <AIDA/IAnnotation.h>
#include <AIDA/IBaseHistogram.h>

#include <Slash/Core/IWriter.h>
#include <Slash/Tools/smanip.h>

#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/smanip.h>
#include <Lib/mmanip.h>

#define SOLID  "0xFFFF"
#define DASHED "0x00FF"
#define DOTTED "0x0101"
#define DASH_DOTTED "0x1C47"

#define region_MAXIMUM(a,b) ((a)>(b)?a:b)

//////////////////////////////////////////////////////////////////////////////
inline void region_set_COL(
 AIDA::IPlotterRegion& aRegion
,const std::string& aField
,int aCOLOR
) 
//////////////////////////////////////////////////////////////////////////////
// 1 black
// 2 red
// 3 green
// 4 blue
// 5 yellow
// 6 magenta
// 7 cyan
// 8 white
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string scolor;

  // See also pagpri PAW_SbColor.

       if(aCOLOR==0) scolor = "0 0 0"; //black
  else if(aCOLOR==1) scolor = "0 0 0"; //black
  else if(aCOLOR==2) scolor = "1 0 0"; //red
  else if(aCOLOR==3) scolor = "0 1 0"; //green
  else if(aCOLOR==4) scolor = "0 0 1"; //blue
  else if(aCOLOR==5) scolor = "1 1 0"; //yellow
  else if(aCOLOR==6) scolor = "1 0 1"; //magenta
  else if(aCOLOR==7) scolor = "0 1 1"; //cyan
  else if(aCOLOR==8) scolor = "1 1 1"; //white
  else               scolor = "1 1 1"; //white

  aRegion.setParameter(aField+"color",scolor);  
}
//////////////////////////////////////////////////////////////////////////////
inline void region_set_DMOD(
 AIDA::IPlotterRegion& aRegion
,const std::string& aField
,int aDMOD
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aDMOD==2) {
    aRegion.setParameter(aField+"linePattern",DASHED);
  } else if(aDMOD==3) {
    aRegion.setParameter(aField+"linePattern",DOTTED);
  } else if(aDMOD==4) {
    aRegion.setParameter(aField+"linePattern",DASH_DOTTED);
  } else {
    aRegion.setParameter(aField+"linePattern",SOLID);
  }
}
//////////////////////////////////////////////////////////////////////////////
inline bool region_field_value(
 AIDA::IPlotterRegion& aRegion
,const std::string& aField
,int& aValue
,Slash::Core::IWriter& aPrinter
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::string svalue = aRegion.parameterValue("plotter."+aField);
  if(!Slash::to<int>(svalue,aValue)) {
    Lib::Out out(aPrinter);
    out << "plotter_field_value :" 
        << " for field " << Lib::sout(aField) 
        << ", can't convert " << Lib::sout(svalue) << "to long." << Lib::endl;
    aValue = 0;
    return false;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
inline void region_set_MSCF(
 AIDA::IPlotterRegion& aRegion
,const std::string& aField
,double aMSCF
) 
//////////////////////////////////////////////////////////////////////////////
// Default aMSCF is 1 and match markers_9_9.
// FIXME : then no larger size today available.
// Note that with the below : 
//   MSCF = 0.56 -> 5
//   MSCF = 0.78 -> 7
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  int sz = int(9.*aMSCF);
  aRegion.setParameter(aField+"markerSize",Lib::smanip::tostring(sz));
}
//////////////////////////////////////////////////////////////////////////////
inline void region_set_automated(
 AIDA::IPlotterRegion& aRegion
,AIDA::IBaseHistogram& aHistogram
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  //printf("debug : region_set_automated \"%s\"\n",histogram_name(aHistogram).c_str());

  AIDA::IAnnotation& annotation = aHistogram.annotation();

  bool automated = true;
  std::string svalue = annotation.value("opaw.value.automated");
  if(svalue!="") {
    if(!Lib::smanip::tobool(svalue,automated)) automated = true;
  }

  if(aHistogram.dimension()==1) {
    if(automated) {
      aRegion.setParameter("plotter.yAxisAutomated","TRUE");
    } else {
      aRegion.setParameter("plotter.yAxisAutomated","FALSE");
      std::string smn = annotation.value("opaw.value.minimum");
      std::string smx = annotation.value("opaw.value.maximum");
      aRegion.setParameter("plotter.yAxisMinimum",smn);
      aRegion.setParameter("plotter.yAxisMaximum",smx);
    }
  } else if(aHistogram.dimension()==2) {
    if(automated) {
      aRegion.setParameter("plotter.zAxisAutomated","TRUE");
    } else {
      aRegion.setParameter("plotter.zAxisAutomated","FALSE");
      std::string smn = annotation.value("opaw.value.minimum");
      std::string smx = annotation.value("opaw.value.maximum");
      aRegion.setParameter("plotter.zAxisMinimum",smn);
      aRegion.setParameter("plotter.zAxisMaximum",smx);
    }
  }
}

#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
#include <AIDA/IAxis.h>

#ifndef NotFound
#define NotFound (-1)
#endif

//////////////////////////////////////////////////////////////////////////////
inline bool region_setLimits(
 AIDA::IPlotterRegion& aRegion
,AIDA::IHistogram1D& aHistogram
,const std::vector< std::pair<int,int> >& aDims
,Slash::Core::IWriter& aPrinter
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aDims.size()) return true;
  if(aDims.size()!=1) {
    Lib::Out out(aPrinter);
    out << "region_setLimits :" 
        << " bad input dimension."
        << Lib::endl;
    return false;
  }
  int ixmn = aDims[0].first;
  int ixmx = aDims[0].second;
  int xbinn = aHistogram.axis().bins();
  if(ixmn==NotFound) ixmn = 0;
  if(ixmx==NotFound) ixmx = xbinn-1;
  if( (ixmn>ixmx) || (ixmn<0)||(ixmn>=xbinn) || (ixmx<0)||(ixmx>=xbinn) ) {
    Lib::Out out(aPrinter);
    out << "region_setLimits :" 
        << " bad binning "
        << " (xbins=" << xbinn 
        << ",ixmn=" << ixmn << ",ixmx=" << ixmx 
        << ")"
        << Lib::endl;
    return false;
  }
  double XLOW = aHistogram.axis().binLowerEdge(ixmn);
  double XUP = aHistogram.axis().binUpperEdge(ixmx);
  aRegion.setXLimits(XLOW,XUP);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
inline bool region_setLimits(
 AIDA::IPlotterRegion& aRegion
,AIDA::IHistogram2D& aHistogram
,const std::vector< std::pair<int,int> >& aDims
,Slash::Core::IWriter& aPrinter
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!aDims.size()) return true;
  if(aDims.size()!=2) {
    Lib::Out out(aPrinter);
    out << "region_setLimits :" 
        << " bad input dimension."
        << Lib::endl;
    return false;
  }
  int ixmn = aDims[0].first;
  int ixmx = aDims[0].second;
  int xbinn = aHistogram.xAxis().bins();
  if(ixmn==NotFound) ixmn = 0;
  if(ixmx==NotFound) ixmx = xbinn-1;
  int iymn = aDims[1].first;
  int iymx = aDims[1].second;
  int ybinn = aHistogram.yAxis().bins();
  if(iymn==NotFound) iymn = 0;
  if(iymx==NotFound) iymx = ybinn-1;
  if( (ixmn>ixmx) || (ixmn<0)||(ixmn>=xbinn) || (ixmx<0)||(ixmx>=xbinn) ||
      (iymn>iymx) || (iymn<0)||(iymn>=ybinn) || (iymx<0)||(iymx>=ybinn) ) {
    Lib::Out out(aPrinter);
    out << "region_setLimits :" 
        << " bad binning "
        << " (xbins=" << xbinn 
        << ",ixmn=" << ixmn << ",ixmx=" << ixmx 
        << ",ybins=" << ybinn 
        << ",iymn=" << iymn << ",iymx=" << iymx 
        << ")"
        << Lib::endl;
    return false;
  }
  double XLOW = aHistogram.xAxis().binLowerEdge(ixmn);
  double XUP = aHistogram.xAxis().binUpperEdge(ixmx);
  aRegion.setXLimits(XLOW,XUP);
  double YLOW = aHistogram.yAxis().binLowerEdge(iymn);
  double YUP = aHistogram.yAxis().binUpperEdge(iymx);
  aRegion.setYLimits(YLOW,YUP);
  return true;
}

