extern "C" {
  void pafith_(void*);
}

#include <kuip/ksys.h>
#include <kuip/kfor.h>
#include <kuip/kuser.h>
#include <kuip/klink.h>

// AIDA helpers :
#include <Slash/Tools/AIDA.h>
#include <amanip/create.h>

// AIDA :
#include <AIDA/IPlotter.h>
#include <AIDA/IFitter.h>
#include <AIDA/IAxis.h>

#include "KuipHelpers.h"
#include "FindHelpers.h"
#include "RegionHelpers.h"

//////////////////////////////////////////////////////////////////////////////
void pafith_(
 void* aSession
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  Slash::Core::ISession* session = (Slash::Core::ISession*)aSession;

  Slash::Core::IWriter& printer = session->printer();
  Lib::Out out(printer);

  std::string cmd_path = ku_path() ? ku_path() : "";
  //std::string cmd_name = ku_name();
  if(session->verboseLevel()) {
    out << "pafith :" 
        << " execute " << Lib::sout(cmd_path) 
        << Lib::endl;
  }

  OpenPAW::PAW_Manager* pawManager = find_PAW(*session);
  if(!pawManager) return;

  AIDA::IAnalysisFactory* aida = pawManager->aida();
  if(!aida) return;
  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return;

  if(cmd_path=="/HISTOGRAM/FIT") {

    std::string ID = ku_gets();
    std::string FUNC = ku_gets();
    std::string CHOPT = ku_gets();
    //int NP = ku_geti();
    //std::string PAR = ku_gets();
    //std::string STEP = ku_gets();
    //std::string PMIN = ku_gets();
    //std::string PMAX = ku_gets();
    //std::string ERRPAR = ku_gets();

    if(session->verboseLevel()) {
      out << "pafith :" 
          << " " << cmd_path
          << " ID= " << Lib::sout(ID) 
          << " FUNC= " << Lib::sout(FUNC) 
          << " CHOPT= " << Lib::sout(CHOPT) 
          << Lib::endl;
    }

    Lib::smanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    AIDA::IHistogram1D* histogram = histogram1D_find(*pawManager,ID);
    if(!histogram) {
      out << "pafith :" 
          << " can't find histogram " << Lib::sout(ID) 
          << Lib::endl;
      return;
    }
    double XLOW = histogram->axis().lowerEdge();
    double XUP = histogram->axis().upperEdge();

    AIDA::IFitter* fitter = amanip::create_fitter(*aida);
    if(!fitter) {
      out << "pafith :" 
          << " can't create a fitter." << Lib::endl;
      return;
    }

    Lib::smanip::touppercase(FUNC);
    if(FUNC=="E") FUNC = "EHBOOK";

    AIDA::IFitResult* fitResult = fitter->fit(*histogram,FUNC);
    delete fitter;
    if(!fitResult) {
      out << "pafith :" 
          << " fit failed." << Lib::endl;
      return;
    }
    
    AIDA::IFunction& function = fitResult->fittedFunction();

    AIDA::IPlotter* plotter = 
      amanip::create_plotter(*aida,""); //FIXME : pass where.
    if(plotter) {
      pawManager->setPlotterLayout(*plotter);

      pawManager->plotterNext
        (*plotter,OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"S"));

      AIDA::IPlotterRegion& region = plotter->currentRegion();

      pawManager->regionSetLayout(region);

      //if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"0"))

      // by default the histo and the function are drawn :
      std::vector< std::pair<int,int> > dims;  
      dims.push_back( std::pair<int,int>(NotFound,NotFound)); 
      if(region_setLimits(region,*histogram,dims,printer)) {
        pawManager->plotHistogram1D(region,*histogram,"",false); 
      }

      if(function.dimension()==1) {
        pawManager->plotFunction1D(region,function,XLOW,XUP,s_FUN_C);
      }

      

      delete plotter;
    }

    delete fitResult;

  } else {
    out << "pafith :" 
        << " " << cmd_path << " : dummy." << Lib::endl;

  }
}
