extern "C" {
  void pafitv_(void*);
}

#include <kuip/ksys.h>
#include <kuip/kfor.h>
#include <kuip/kuser.h>
#include <kuip/klink.h>

#include <AIDA/IPlotter.h>
#include <AIDA/IDataPointSet.h>
#include <AIDA/IFitter.h>
#include <AIDA/IMeasurement.h>
#include <AIDA/IDataPoint.h>
#include <AIDA/IFitResult.h>

#include <Slash/Core/ISession.h>
#include <Slash/Core/IWriter.h>
#include <Slash/Tools/AIDA.h>

#include <amanip/find.h>
#include <amanip/create.h>

#include <Lib/Out.h>
#include <Lib/sout.h>

#include "ArrayHelpers.h" //FIXME : not AIDA.
#include <OpenPAW/PAW_Manager.h>

#include "KuipHelpers.h"

#define MINIMUM(a,b) ((a)<(b)?a:b)

//////////////////////////////////////////////////////////////////////////////
void pafitv_(
 void* aSession
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  Slash::Core::ISession* session = (Slash::Core::ISession*)aSession;

  Slash::Core::IWriter& printer = session->printer();
  Lib::Out out(printer);

  std::string cmd_path = ku_path() ? ku_path() : "";
  //std::string cmd_name = ku_name();
  if(session->verboseLevel()) {
    out << "pafitv :" 
        << " execute " << Lib::sout(cmd_path) 
        << Lib::endl;
  }

  OpenPAW::PAW_Manager* pawManager = find_PAW(*session);
  if(!pawManager) return;

  AIDA::IAnalysisFactory* aida = pawManager->aida();
  if(!aida) return;
  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return;

  if(cmd_path=="/VECTOR/FIT") {

    std::string X = ku_gets();
    std::string Y = ku_gets();
    std::string EY = ku_gets();
    std::string FUNC = ku_gets();
    std::string CHOPT = ku_gets();
    //int NP = ku_geti();
    //std::string PAR = ku_gets();
    //std::string STEP = ku_gets();
    //std::string PMIN = ku_gets();
    //std::string PMAX = ku_gets();
    //std::string ERRPAR = ku_gets();

    if(session->verboseLevel()) {
      out << "pafitv :" 
          << " " << cmd_path
          << " X= " << Lib::sout(X) 
          << " Y= " << Lib::sout(Y) 
          << " EY= " << Lib::sout(EY) 
          << " FUNC= " << Lib::sout(FUNC) 
          << " CHOPT= " << Lib::sout(CHOPT) 
          << Lib::endl;
    }

    Lib::smanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    std::string XNAME;
    std::vector< std::pair<int,int> > xdims;
    if(!OpenPAW::PAW_Manager::parseName(X,XNAME,xdims,printer)) return;
    Lib::smanip::touppercase(XNAME);
    Slash::Data::IArray* xvec = find_vector(*memoryTree,XNAME);
    if(!xvec) {
      out << "pafitv :" 
          << " can't find vector " << Lib::sout(XNAME) 
          << " in tree " << Lib::sout(memoryTree->storeName()) 
          << Lib::endl;
      return;
    }
    std::vector< std::pair<unsigned int,unsigned int> > xcut;
    if(!vector_get_cut(*xvec,xdims,xcut)) {
      out << "pafitv :" 
            << " mismatch dimension in " << Lib::sout(XNAME) 
          << Lib::endl;
      return;
    }

    std::string YNAME;
    std::vector< std::pair<int,int> > ydims;
    if(!OpenPAW::PAW_Manager::parseName(Y,YNAME,ydims,printer)) return;
    Lib::smanip::touppercase(YNAME);
    Slash::Data::IArray* yvec = find_vector(*memoryTree,YNAME);
    if(!yvec) {
      out << "pafitv :" 
          << " can't find vector " << Lib::sout(YNAME) 
          << " in tree " << Lib::sout(memoryTree->storeName()) 
          << Lib::endl;
      return;
    }
    std::vector< std::pair<unsigned int,unsigned int> > ycut;
    if(!vector_get_cut(*yvec,ydims,ycut)) {
      out << "pafitv :" 
            << " mismatch dimension in " << Lib::sout(YNAME) 
          << Lib::endl;
      return;
    }

    std::vector<double> vx;
   {const std::vector<double>& v = xvec->vector();
    unsigned int vsize = xvec->size();
    for(unsigned int index = 0;index<vsize;index++) {
      if(xvec->accept(index,xcut)) vx.push_back(v[index]);
    }}

    std::vector<double> vy;
   {const std::vector<double>& v = yvec->vector();
    unsigned int vsize = yvec->size();
    for(unsigned int index = 0;index<vsize;index++) {
      if(yvec->accept(index,ycut)) vy.push_back(v[index]);
    }}

    // create temporary not managed dps :
    std::string name = "dataPointSet_" + X + "_" + Y;
    AIDA::IDataPointSet* dps = 
      amanip::create_data_point_set(*aida,*memoryTree,
                          amanip::path_not_managed()+"/"+name,name,2);
    if(!dps) {
      out << "pafitv :" 
          << " can't create a data point set." << Lib::endl;
      return;
    }

    int number = MINIMUM(vx.size(),vy.size());
    for(int index=0;index<number;index++) {
      AIDA::IDataPoint* point = dps->addPoint();
      AIDA::IMeasurement* m;
      m = point->coordinate(0);
      m->setValue(vx[index]);
      m = point->coordinate(1);
      m->setValue(vy[index]);
    }

    double XLOW = dps->lowerExtent(0);
    double XUP = dps->upperExtent(0);

    AIDA::IFitter* fitter = amanip::create_fitter(*aida);
    if(!fitter) {
      out << "pafitv :" 
          << " can't create a fitter." 
          << Lib::endl;
      delete dps;
      return;
    }

    Lib::smanip::touppercase(FUNC);
    if(FUNC=="E") FUNC = "EHBOOK";
    AIDA::IFitResult* fitResult = fitter->fit(*dps,FUNC);
    delete fitter;
    delete dps;
    if(!fitResult) {
      out << "pafitv :" 
          << " fit failed." 
          << Lib::endl;
      return;
    }

    AIDA::IFunction& function = fitResult->fittedFunction();

    AIDA::IPlotter* plotter = 
      amanip::create_plotter(*aida,""); //FIXME : pass where.
    if(plotter) {
      pawManager->setPlotterLayout(*plotter);

      pawManager->plotterNext
        (*plotter,OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"S"));

      AIDA::IPlotterRegion& region = plotter->currentRegion();

      pawManager->regionSetLayout(region);

      if(function.dimension()==1) {
        pawManager->plotFunction1D(region,function,XLOW,XUP,s_FUN_C);
      }
      
      delete plotter;
    }

    delete fitResult;

  } else {
    out << "pafitv :" 
        << " " << cmd_path << " : dummy." << Lib::endl;

  }
}
