extern "C" {
  void pafunc_(void*);
}

#include <kuip/ksys.h>
#include <kuip/kfor.h>
#include <kuip/kuser.h>
#include <kuip/klink.h>

#include <AIDA/IPlotter.h>

#include <Slash/Tools/AIDA.h>

#include <amanip/find.h>
#include <amanip/create.h>

#include <Lib/fmanip.h>
#include <Lib/smanip.h>

#include <OpenPAW/PAW_Manager.h>

#include "KuipHelpers.h"
#include "PlotHistogram2D.h"
#include "FindHelpers.h"

// same as in pahist
#define PALNB_FROM_HCOL \
  int PALNB;\
  if(!pawManager->paletteFromHCOL(PALNB)) PALNB = PALETTE_DEFAULT;

//////////////////////////////////////////////////////////////////////////////
void pafunc_(
 void* aSession
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  Slash::Core::ISession* session = (Slash::Core::ISession*)aSession;

  Slash::Core::IWriter& printer = session->printer();
  Lib::Out out(printer);

  //std::string cmd_name = ku_name();
  std::string cmd_path = ku_path() ? ku_path() : "";
  if(session->verboseLevel()) {
    out << "pafunc :" 
        << " execute " << Lib::sout(cmd_path) 
        << Lib::endl;
  }

  OpenPAW::PAW_Manager* pawManager = find_PAW(*session);
  if(!pawManager) return;

  AIDA::IAnalysisFactory* aida = pawManager->aida();
  if(!aida) return;
  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return;

  if(cmd_path=="/FUNCTION/PLOT") {

    std::string UFUNC = ku_gets();
    double XLOW = ku_getr();
    double XUP = ku_getr();
    std::string CHOPT = ku_gets();

    if(session->verboseLevel()) {
      out << "pafunc :" 
          << " " << cmd_path
          << " UFUNC= " << Lib::sout(UFUNC) 
          << " XLOW= " << XLOW
          << " XUP= " << XUP
          << " CHOPT= " << Lib::sout(CHOPT) 
          << Lib::endl;
    }

    Lib::smanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    std::string funcName = pawManager->generateFunctionName();

    char* path = ku_macro_path();
    std::string paths = path?path:"";

    // managed true because there is no retain mode for function.
    // retain_arrays true in case a vector is deleted whilst
    //  the plotter still visualize the function.
    // BUT then, the execution of the function can't update
    // the vectors, if some are used by the function code.
    AIDA::IFunction* func = create_function_vector
      (funcName,paths,UFUNC,*session,*aida,*memoryTree,true,true);

    if(!func) {
      out << "pafunc :" 
          << " can't create function " << Lib::sout(UFUNC) 
          << Lib::endl;
      return;
    }

    AIDA::IPlotter* plotter = 
      amanip::create_plotter(*aida,""); //FIXME : pass where.
    if(plotter) {
      pawManager->setPlotterLayout(*plotter);

      bool superpose = OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"S");
      if(superpose) {
        // If superposed, CERN/PAW ignores the title of the object.
        func->setTitle(""); 
      }

      pawManager->plotterNext(*plotter,superpose);

      AIDA::IPlotterRegion& region = plotter->currentRegion();

      pawManager->regionSetLayout(region);

      if(func->dimension()==1) {

        bool empty = true;
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"C")) {
          pawManager->plotFunction1D(region,*func,XLOW,XUP,s_FUN_C);
          empty = false;
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"P")) {
          pawManager->plotFunction1D(region,*func,XLOW,XUP,"P");
          empty = false;
        }

        if(empty) {
          pawManager->plotFunction1D(region,*func,XLOW,XUP,s_FUN_C);
        }

      } else {
        out << "pafunc :" 
            << " dimension mismatch."
            << " /FUNCTION/PLOT is for 1D functions"
            << " (" << func->dimension() << " found)."
            << Lib::endl;
      }

      delete plotter;
    }
    
  } else if(cmd_path=="/FUNCTION/FUN2") {

    std::string ID = ku_gets();
    std::string UFUNC = ku_gets();
    int NCX = ku_geti();
    double XMIN = ku_getr();
    double XMAX = ku_getr();
    int NCY = ku_geti();
    double YMIN = ku_getr();
    double YMAX = ku_getr();
    std::string CHOPT = ku_gets();

    if(session->verboseLevel()) {
      out << "pafunc :" 
          << " " << cmd_path
          << " ID= " << Lib::sout(ID) 
          << " UFUNC= " << Lib::sout(UFUNC) 
          << " NCX= " << NCX
          << " XMIN= " << XMIN
          << " XMAX= " << XMAX
          << " NCY= " << NCY
          << " YMIN= " << YMIN
          << " YMAX= " << YMAX
          << " CHOPT= " << Lib::sout(CHOPT) 
          << Lib::endl;
    }

    Lib::smanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    if((NCX<=0)||(XMAX<=XMIN)) {
      out << "pafunc :" 
          << " bad booking values."
          << Lib::endl;
      return;
    }
    if((NCY<=0)||(YMAX<=YMIN)) {
      out << "pafunc :" 
          << " bad booking values."
          << Lib::endl;
      return;
    }

    if(memoryTree->find(ID)) {
      memoryTree->rm(ID); //ok.
      out << " *** WARNING in OpenPAW :"
          << " Already existing histogram replaced : ID="
          << ID 
          << Lib::endl;
    }
    AIDA::IHistogram2D* histogram = 
      amanip::create_histogram_2D(*aida,*memoryTree,ID,UFUNC,
                         NCX,XMIN,XMAX,NCY,YMIN,YMAX);
    if(!histogram) {
      out << "pafunc :" 
            << " can't create histogram " << Lib::sout(ID) << Lib::endl;
      return;
    }

    double dx = (XMAX - XMIN)/NCX;
    double dy = (YMAX - YMIN)/NCY;

    std::string funcName = pawManager->generateFunctionName();

    char* path = ku_macro_path();
    std::string paths = path?path:"";

    // managed false
    // retain_arrays false, so that the execution of the 
    //  function can change the vectors.
    AIDA::IFunction* func = create_function_vector
      (funcName,paths,UFUNC,*session,*aida,*memoryTree,false,false);

    if(!func) {
      out << "pafunc :" 
          << " can't create function " << Lib::sout(UFUNC) 
          << Lib::endl;
      return;
    }

    AIDA::IAnnotation& annotation = histogram->annotation();
    annotation.addItem("opaw.codelet",func->codeletString());

    double epsilon = 0;

    std::vector<double> args(2);
    for(int ibin = 0;ibin<NCX;ibin++) {
      double x = XMIN + (ibin + 0.5 + epsilon) * dx;
      args[0] = x;
      for(int jbin = 0;jbin<NCY;jbin++) {
        double y = YMIN + (jbin + 0.5 + epsilon) * dy;
        args[1] = y;
        double value = func->value(args);
        histogram->fill(x,y,value);
      }
    }
    delete func;

    if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT," ")) {
    } else {

      AIDA::IPlotter* plotter = 
        amanip::create_plotter(*aida,""); //FIXME : pass where.
      if(plotter) {
        pawManager->setPlotterLayout(*plotter);
        
        pawManager->plotterNext
          (*plotter,OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"S"));
  
        AIDA::IPlotterRegion& region = plotter->currentRegion();
  
        pawManager->regionSetLayout(region);
  
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"SURF1")) {
          PALNB_FROM_HCOL
          region_plot_histogram2D
            (*session,region,*histogram,
             s_SURF_LEVEL,printer,0,PALNB,*pawManager);
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"SURF2")) {
          PALNB_FROM_HCOL
          region_plot_histogram2D
            (*session,region,*histogram,
             s_SURF_LEVEL_NO_EDGE,printer,0,PALNB,*pawManager);
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"SURF3")) {
          PALNB_FROM_HCOL
          region_plot_histogram2D
            (*session,region,*histogram,
             s_SURF_CONT,printer,0,PALNB,*pawManager);
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"SURF4")) {
          region_plot_histogram2D
            (*session,region,*histogram,s_SURF_LIGHT,printer,0,0,*pawManager);
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"SURF5")) {
          PALNB_FROM_HCOL
          region_plot_histogram2D
            (*session,region,*histogram,
             s_SURF_COL,printer,0,PALNB,*pawManager);
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_SURF)) {
          region_plot_histogram2D
            (*session,region,*histogram,s_SURF,printer,0,0,*pawManager);
        }

        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"LEGO1")) {
          region_plot_histogram2D
         (*session,region,*histogram,s_LEGO_LIGHT,printer,0,0,*pawManager);
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"LEGO2")) {
          PALNB_FROM_HCOL
          region_plot_histogram2D
         (*session,region,*histogram,s_LEGO_LEVEL,printer,0,PALNB,*pawManager);
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"LEGO3")) {
          PALNB_FROM_HCOL
          region_plot_histogram2D
         (*session,region,*histogram,s_LEGO_COL,printer,0,PALNB,*pawManager);
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_LEGO)) {
          region_plot_histogram2D
            (*session,region,*histogram,s_LEGO,printer,0,0,*pawManager);
        }

        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"TEXT")) {
          region_plot_histogram2D
            (*session,region,*histogram,"TEXT",printer,0,0,*pawManager);
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"CONT")) {
          region_plot_histogram2D
            (*session,region,*histogram,"CONT_DMOD",printer,10,0,*pawManager);
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"BOX")) {
          region_plot_histogram2D
            (*session,region,*histogram,"BOX",printer,0,0,*pawManager);
        }
        if(OpenPAW::PAW_Manager::hasUnkownOption(OPTS,CHOPT)) {
          region_plot_histogram2D
            (*session,region,*histogram,"",printer,0,0,*pawManager);
        }
  
        delete plotter;
      }
    }

  } else if(cmd_path=="/FUNCTION/FUN1") {

    std::string ID = ku_gets();
    std::string UFUNC = ku_gets();
    int NCX = ku_geti();
    double XMIN = ku_getr();
    double XMAX = ku_getr();
    std::string CHOPT = ku_gets();

    if(session->verboseLevel()) {
      out << "pafunc :" 
          << " " << cmd_path
          << " ID= " << Lib::sout(ID) 
          << " UFUNC= " << Lib::sout(UFUNC) 
          << " NCX= " << NCX
          << " XMIN= " << XMIN
          << " XMAX= " << XMAX
          << " CHOPT= " << Lib::sout(CHOPT) 
          << Lib::endl;
    }

    Lib::smanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    if((NCX<=0)||(XMAX<=XMIN)) {
      out << "pafunc :" 
          << " bad booking values."
          << Lib::endl;
      return;
    }

    if(memoryTree->find(ID)) {
      memoryTree->rm(ID); //ok.
      out << " *** WARNING in OpenPAW :"
          << " Already existing histogram replaced : ID="
          << ID 
          << Lib::endl;
    }

    AIDA::IHistogram1D* histogram = 
      amanip::create_histogram_1D
        (*aida,*memoryTree,ID,UFUNC,NCX,XMIN,XMAX);
    if(!histogram) {
      out << "pafunc :" 
            << " can't create histogram " << Lib::sout(ID) << Lib::endl;
      return;
    }

    double dx = (XMAX - XMIN)/NCX;

    std::string funcName = pawManager->generateFunctionName();

    char* path = ku_macro_path();
    std::string paths = path?path:"";

    // managed false
    // retain_arrays false, so that the execution of the 
    //  function can change the vectors.
    AIDA::IFunction* func = create_function_vector
      (funcName,paths,UFUNC,*session,*aida,*memoryTree,false,false);

    if(!func) {
      out << "pafunc :" 
          << " can't create function " << Lib::sout(UFUNC) 
          << Lib::endl;
      return;
    }

    //AIDA::IAnnotation& annotation = histogram->annotation();
    //annotation.addItem("opaw.codelet",func->codeletString());

    //double epsilon = 0.001; // Trick to avoid zero.
    double epsilon = 0;

    std::vector<double> args(1);
    for(int ibin = 0;ibin<NCX;ibin++) {
      double x = XMIN + (ibin + 0.5 + epsilon) * dx;
      args[0] = x;
      double value = func->value(args);
      histogram->fill(x,value);
    }
    delete func;

    if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT," ")) {
    } else {

      AIDA::IPlotter* plotter = 
        amanip::create_plotter(*aida,""); //FIXME : pass where.
      if(plotter) {
        pawManager->setPlotterLayout(*plotter);
  
        pawManager->plotterNext
          (*plotter,OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"S"));
  
        AIDA::IPlotterRegion& region = plotter->currentRegion();
  
        pawManager->regionSetLayout(region);
  
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"C")) {
          pawManager->plotHistogram1D(region,*histogram,"C",false); 
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"E")) {
          pawManager->plotHistogram1D(region,*histogram,"E",false); 
        }
  
        delete plotter;
      }
    }

  } else if(cmd_path=="/FUNCTION/POINTS") {

    int NPX = ku_geti();
    int NPY = ku_geti();
    int NPZ = ku_geti();
    pawManager->setFunctionPoints(NPX,NPY,NPZ);

  } else if(cmd_path=="/FUNCTION/ANGLE") {

    double THETA = ku_getr(); //in degrees
    double PHI = ku_getr();   //in degrees

    pawManager->setTHETA(THETA);
    pawManager->setPHI(PHI);

  } else {
    out << "pafunc :" 
        << " " << cmd_path << " : dummy." << Lib::endl;

  }
}
