//
// Contain HEPVis and Inventor related code.
//


extern "C" {
  void pagpri_(void*);
}

#include <kuip/ksys.h>
#include <kuip/kfor.h>
#include <kuip/kuser.h>
#include <kuip/klink.h>


// Slash :
#include <Slash/Core/ISession.h>
#include <Slash/Core/IWriter.h>
#include <Slash/UI/IUI.h>
#include <Slash/Tools/UI.h>
#include <Slash/Tools/AIDA.h>
#include <Slash/Tools/Inventor.h>
//#include <Slash/Tools/out.h>

// AIDA :
#include <AIDA/ICloud2D.h>
#include <AIDA/IPlotter.h>
#include <AIDA/IPlotterRegion.h>

#include <amanip/find.h>
#include <amanip/create.h>
#include <amanip/smanip.h>

#include "ArrayHelpers.h"

#include <OpenPAW/PAW_Manager.h>
#include "KuipHelpers.h"

#include <HEPVis/SbColors.h>
#include <HEPVis/SbPrimitives.h>
#include <HEPVis/nodekits/SoPlotter.h>

#define MINIMUM(a,b) ((a)<(b)?a:b)
#define MAXIMUM(a,b) ((a)>(b)?a:b)

inline SbColor PAW_SbColor(int aCOLOR) {
  // 1 black
  // 2 red
  // 3 green
  // 4 blue
  // 5 yellow
  // 6 magenta
  // 7 cyan
  // 8 white

       if(aCOLOR==0) return SbColor_black;
  else if(aCOLOR==1) return SbColor_black;
  else if(aCOLOR==2) return SbColor_red;
  else if(aCOLOR==3) return SbColor_green;
  else if(aCOLOR==4) return SbColor_blue;
  else if(aCOLOR==5) return SbColor_yellow;
  else if(aCOLOR==6) return SbColor_magenta;
  else if(aCOLOR==7) return SbColor_cyan;
  else if(aCOLOR==8) return SbColor_white;
  else               return SbColor_white;
}

void pagpri_(void* aSession) { 
  Slash::Core::ISession* session = (Slash::Core::ISession*)aSession;

  std::ostream& out = session->cout();
  Slash::Core::IWriter& printer = session->printer();

  //std::string cmd_name = ku_name();
  std::string cmd_path = ku_path() ? ku_path() : "";
  if(session->verboseLevel()) {
    out << "pagpri :" 
        << " execute " << amanip::sout(cmd_path) 
        << std::endl;
  }

  OpenPAW::PAW_Manager* pawManager = find_PAW(*session);
  if(!pawManager) return;

  AIDA::IAnalysisFactory* aida = pawManager->aida();
  if(!aida) return;
  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return;

  //OPTs_type& opts = pawManager->opts();
  ATTs_type& atts = pawManager->atts();

  //double LWID = atts.value("LWID"); 
  //LWID = MAXIMUM(int(LWID/2.2),1);  //From graflib/higz/ikernel/islwsc.F

  if(cmd_path=="/GRAPHICS/PRIMITIVES/GRAPH") {

    int N = ku_geti();
    std::string X = ku_gets();
    std::string Y = ku_gets();
    std::string CHOPT = ku_gets();

    amanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);
    if(session->verboseLevel()) {
      out << "pagpri :" 
          << " CHOPT " << amanip::sout(CHOPT) 
          << std::endl;
    }

    std::string XNAME;
    std::vector< std::pair<int,int> > xdims;
    if(!OpenPAW::PAW_Manager::parseName(X,XNAME,xdims,printer)) return;
    amanip::touppercase(XNAME);
    Slash::Data::IArray* xvec = find_vector(*memoryTree,XNAME);
    if(!xvec) {
      out << "pagpri :" 
          << " can't find vector " << amanip::sout(XNAME) 
          << " in tree " << amanip::sout(memoryTree->storeName()) 
          << std::endl;
      return;
    }
    std::vector< std::pair<unsigned int,unsigned int> > xcut;
    if(!vector_get_cut(*xvec,xdims,xcut)) {
      out << "pagpri :" 
          << " mismatch dimension in " << amanip::sout(XNAME) 
          << std::endl;
      return;
    }

    std::string YNAME;
    std::vector< std::pair<int,int> > ydims;
    if(!OpenPAW::PAW_Manager::parseName(Y,YNAME,ydims,printer)) return;
    amanip::touppercase(YNAME);
    Slash::Data::IArray* yvec = find_vector(*memoryTree,YNAME);
    if(!yvec) {
      out << "pagpri :" 
          << " can't find vector " << amanip::sout(YNAME) 
          << " in tree " << amanip::sout(memoryTree->storeName()) 
          << std::endl;
      return;
    }
    std::vector< std::pair<unsigned int,unsigned int> > ycut;
    if(!vector_get_cut(*yvec,ydims,ycut)) {
      out << "pagpri :" 
          << " mismatch dimension in " << amanip::sout(YNAME) 
          << std::endl;
      return;
    }

    //std::string name = VNAME + "_rep";
    int xsize = xvec->size();
    int ysize = yvec->size();
    int number = MINIMUM(xsize,N);
    number = MINIMUM(number,ysize);

    // create temporary not managed cloud :
    std::string name = "cloud2D_" + X + "_" + Y;
    AIDA::ICloud2D* cloud = 
      amanip::create_cloud_2D
        (*aida,*memoryTree,amanip::path_not_managed()+"/"+name,"");
    if(!cloud) {
      out << "pagpri :" 
          << " can't create a cloud2D." << std::endl;
      return;
    }

    const std::vector<double>& xv = xvec->vector();
    const std::vector<double>& yv = yvec->vector();
    for(int index=0;index<number;index++) cloud->fill(xv[index],yv[index]);

    AIDA::IPlotter* plotter = 
      amanip::create_plotter(*aida,""); //FIXME : where 
    if(plotter) {
      pawManager->setPlotterLayout(*plotter);

      //If A (axes) are required we go in the next region.
      pawManager->plotterNext
        (*plotter,!OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"A"));

      AIDA::IPlotterRegion& region = plotter->currentRegion();

      pawManager->regionSetLayout(region);

      if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"C"))
        pawManager->plotCloud2D(region,*cloud,"C");
      if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"L"))
        pawManager->plotCloud2D(region,*cloud,"L");
      if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"P"))
        pawManager->plotCloud2D(region,*cloud,"P");

      

      delete plotter;
    }

    // Plotter is in retain mode, we can now get rid of the object.
    delete cloud;

  } else if(cmd_path=="/GRAPHICS/PRIMITIVES/TEXT") {

    double X = ku_getr();
    double Y = ku_getr();
    std::string TEXT = ku_gets();
    double SIZE = ku_getr();
    double ANGLE = ku_getr();
    std::string CHOPT = ku_gets();

    amanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    Slash::UI::IUI* ui = Slash::find_UI(*session);
    if(!ui) return;
    SoPlotter* soPlotter = Slash::find_SoPlotter(*ui,Slash::s_current());
    if(!soPlotter) return;

    //SIZE is in page coordinate YSIZ.
    char JUST = 'L';
    if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"L")) JUST = 'L';
    if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"C")) JUST = 'C';
    if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"R")) JUST = 'R';

    soPlotter->primitivesEnforced.setValue(TRUE);
    soPlotter->addPlottablePrimitive(
      new SbPlottableText(TEXT.c_str(),X,Y,SIZE,ANGLE,JUST,"Hershey",0));

  } else if(cmd_path=="/GRAPHICS/PRIMITIVES/ITX") {

    double X = ku_getr();
    double Y = ku_getr();
    std::string TEXT = ku_gets();

    Slash::UI::IUI* ui = Slash::find_UI(*session);
    if(!ui) return;
    SoPlotter* soPlotter = Slash::find_SoPlotter(*ui,Slash::s_current());
    if(!soPlotter) return;

    double SIZE = atts.value("CHHE"); 
    //SIZE is in page coordinate YSIZ.    

    std::string font;
    double scale;
    bool smoothing,hinting;
    pawManager->getFont("CFON",font,scale,smoothing,hinting);

    soPlotter->primitivesEnforced.setValue(TRUE);
    soPlotter->addPlottablePrimitive(
      new SbPlottableText(TEXT.c_str(),X,Y,SIZE,0,'L',
			  font.c_str(),scale,
                          smoothing?TRUE:FALSE,hinting?TRUE:FALSE));

  } else if(cmd_path=="/GRAPHICS/PRIMITIVES/LABELS") {

    int LABNUM = ku_geti();
    int NLABS = ku_geti();
    std::string CHLABS = ku_gets();

    std::vector<std::string> words;
    Slash::words(CHLABS,",",false,words);

    //printf("debug : \"%s\"\n",CHLABS.c_str());
    //Slash::out(out,words);

    int n = words.size();
    if(NLABS>n) {
      for(int index=n;index<NLABS;index++) words.push_back("");
    } else {
      words.resize(NLABS);
    }

    pawManager->setLabels(LABNUM,words);


  } else if(cmd_path=="/GRAPHICS/PRIMITIVES/BOX") {

    double X1 = ku_getr();
    double X2 = ku_getr();
    double Y1 = ku_getr();
    double Y2 = ku_getr();

    int FAIS = (int)atts.value("FAIS");
    int FACI = (int)atts.value("FACI");
    int FASI = (int)atts.value("FASI");
    int BORD = (int)atts.value("BORD");
    int PLCI = (int)atts.value("PLCI");

    Slash::UI::IUI* ui = Slash::find_UI(*session);
    if(!ui) return;
    SoPlotter* soPlotter = Slash::find_SoPlotter(*ui,Slash::s_current());
    if(!soPlotter) return;

    soPlotter->primitivesEnforced.setValue(TRUE);
    soPlotter->addPlottablePrimitive
      (new SbPlottableBox(X1,Y1,X2,Y2,
         (SbPlottableBox::FillAreaStyle)FAIS,PAW_SbColor(FACI),
          FASI,
          BORD?TRUE:FALSE,PAW_SbColor(PLCI)
      ));

  } else {
    out << "pagpri :" 
        << " " << cmd_path << " : dummy." << std::endl;

  }
}
