extern "C" {
  void pagraf_(void*);
}

#include <kuip/ksys.h>
#include <kuip/kfor.h>
#include <kuip/kuser.h>
#include <kuip/klink.h>

#include <Slash/Core/ISession.h>
#include <Slash/Core/IWriter.h>
#include <Slash/Tools/AIDA.h>

#include <amanip/create.h>

// AIDA :
#include <AIDA/IPlotter.h>

#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/smanip.h>

#include <OpenPAW/PAW_Manager.h>
#include "KuipHelpers.h"

//////////////////////////////////////////////////////////////////////////////
static bool setAtt(
 OpenPAW::PAW_Manager& aPawManager
,ATTs_type& aAtts
,const std::string& aCHATT
,double aVALUE
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  if(aCHATT=="HCOL") { 

    if(aVALUE==0) aVALUE = aAtts.defaultValue(aCHATT);
    aAtts.setValue(aCHATT,aVALUE);

    if(aVALUE<=0) {
      aAtts.setValue("HTYP",0); //No hatching
      aPawManager.setContourMode(true);         //Lines
      aAtts.setValue("HCOL",1); //Black.
    } else if((aVALUE>=1)&&(aVALUE<=999)) {
      aAtts.setValue("HTYP",0); //No hatching.
      aPawManager.setContourMode(true);         //Lines
    } else if((aVALUE>=1001)&&(aVALUE<=1999)) {
      aAtts.setValue("HTYP",0); //No hatching.
      aPawManager.setContourMode(false);        //Fill area.
    } else if(aVALUE==1000) { //???
      aAtts.setValue("HTYP",0); //No hatching.
      aPawManager.setContourMode(true);         //Lines
      aAtts.setValue("HCOL",9); //White.
    }

  } else if(aCHATT=="HTYP") { 

    if(aVALUE==0) aVALUE = aAtts.defaultValue(aCHATT);
    aAtts.setValue(aCHATT,aVALUE);

    aPawManager.setContourMode(true);

  } else if(aCHATT=="NCOL") { 

    if(aVALUE==0) aVALUE = aAtts.defaultValue(aCHATT);

    aAtts.setValue(aCHATT,aVALUE);

    int NCOL = (int)aVALUE;
    if(NCOL>8) {
      // Fill color_table elements [8,NCOL] with 
      // a grey scale from white to black :
      int ncol = NCOL-8;
      if(ncol==1) {
        aPawManager.setColor(9,1,1,1);
      } else {
        for(int index=0;index<ncol;index++) {
          double grey = 1 - ((float)index)/((float)(ncol-1));
          aPawManager.setColor(8+index,grey,grey,grey);
        }
      }
    }

  } else if(aAtts.isKey(aCHATT)) {

    double old_VALUE = aAtts.value(aCHATT);

    if(aVALUE==0) aVALUE = aAtts.defaultValue(aCHATT); //PAW convention.

    aAtts.setValue(aCHATT,aVALUE);

    if( (aCHATT=="XSIZ") ||
        (aCHATT=="XMGL") ||
        (aCHATT=="XMGR") ||
        (aCHATT=="XWIN") ||
        (aCHATT=="YSIZ") ||
        (aCHATT=="YMGL") ||
        (aCHATT=="YMGU") ||
        (aCHATT=="YWIN") ){
      if(aVALUE!=old_VALUE) aPawManager.setToRefresh();
    }

  } else {
    return false; 
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
void pagraf_(
 void* aSession
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  Slash::Core::ISession* session = (Slash::Core::ISession*)aSession;

  Slash::Core::IWriter& printer = session->printer();
  Lib::Out out(printer);

  std::string cmd_path = ku_path() ? ku_path() : "";
  //std::string cmd_name = ku_name();
  if(session->verboseLevel()) {
    out << "pagraf :" 
        << " execute " << Lib::sout(cmd_path) 
        << Lib::endl;
  }

  OpenPAW::PAW_Manager* pawManager = find_PAW(*session);
  if(!pawManager) return;

  AIDA::IAnalysisFactory* aida = pawManager->aida();
  if(!aida) return;

  OPTs_type& opts = pawManager->opts();
  ATTs_type& atts = pawManager->atts();

  if(cmd_path=="/GRAPHICS/VIEWING/ZONE") {

    int NX = ku_geti();
    int NY = ku_geti();
    int IFIRST = ku_geti();
    std::string CHOPT = ku_gets();

    Lib::smanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    pawManager->setZONE
      (OpenPAW::ZONE
         (NX,NY,IFIRST,OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"S")));

    pawManager->setDoZONE();

  } else if(cmd_path=="/GRAPHICS/SET") {

    std::string CHATT = ku_gets();
    double VALUE = ku_getr();

    Lib::smanip::touppercase(CHATT);
    if(CHATT.size()>4) CHATT.resize(4);

    if(CHATT=="SHOW") {  
      std::ostream& out = session->cout();
      out << "+------------------------------------------------------------------------------+" << std::endl;
      out << "|                       HPLSET : Current values in use                         |" << std::endl;
      pawManager->showATTs(out,cat_HPLSET);

      out << "+------------------------------------------------------------------------------+" << std::endl;
      out << "|                        IGSET : Current values in use                         |" << std::endl;
      pawManager->showATTs(out,cat_IGSET);

    } else if(CHATT=="*") { 

      atts.setAllToDefault();
      pawManager->setToRefresh();

      pawManager->setZONE(OpenPAW::ZONE(1,1,1,false));
      pawManager->setDoZONE();

    } else if(CHATT.find('*')!=std::string::npos) { 

      // Find matching attributes :
      const ATTs_type::List& atts_list = atts.list();
      ATTs_type::ConstIterator it;
      for(it=atts_list.begin();it!=atts_list.end();++it) {
        if(Lib::smanip::match((*it).key(),CHATT)) {
          setAtt(*pawManager,atts,(*it).key(),VALUE);
        }
      }
      
    } else if(!setAtt(*pawManager,atts,CHATT,VALUE)) {

      out << "pagraf :" 
          << " " << Lib::sout(CHATT) << " is not an attribute." 
          << Lib::endl;

    }

  } else if(cmd_path=="/GRAPHICS/OPTION") {

    std::string CHOPTN = ku_gets();

    Lib::smanip::touppercase(CHOPTN);
    if(CHOPTN.size()>4) CHOPTN.resize(4);

    if(CHOPTN=="SHOW") {

      std::ostream& out = session->cout();
      out << "+-----------------------------------------------------------------------------+" << std::endl;
      out << "|                      PAW                                                    |" << std::endl;
      pawManager->showOPTs(out,cat_HPLOPT);

      out << "+-----------------------------------------------------------------------------+" << std::endl;
      out << "|                      OPAW                                                   |" << std::endl;
      pawManager->showOPTs(out,cat_OPAW_OPT);


    } else if(CHOPTN=="*") {

      opts.setAllToDefault();

    } else {
      if(CHOPTN=="STAT") CHOPTN = "STA"; //STAT accepted by PAW !
      else if(CHOPTN=="NSTAT") CHOPTN = "NSTA"; //NSTAT accepted by PAW !

      std::string key = opts.keyBoolean(CHOPTN);
      if(key==opts.keyNotFound()) {
        out << "pagraf :" 
            << " " << Lib::sout(CHOPTN) << " is not an option." 
            << Lib::endl;
      } else {

        opts.setValue(key,CHOPTN);
 
        if( (CHOPTN=="NGRI") || (CHOPTN=="GRID")) {
        } else if( (CHOPTN=="BOX") || (CHOPTN=="NBOX")) {
        } else if( (CHOPTN=="HTIT") || (CHOPTN=="UTIT")) {
        } else if( (CHOPTN=="NSTA") || (CHOPTN=="STA")) {
        } else if( (CHOPTN=="NBAR") || (CHOPTN=="BAR")) {
        } else if( (CHOPTN=="COLL") || (CHOPTN=="COLV")) {
        } else if( (CHOPTN=="LINX") || (CHOPTN=="LOGX")) {
        } else if( (CHOPTN=="LINY") || (CHOPTN=="LOGY")) {
        } else if( (CHOPTN=="LINZ") || (CHOPTN=="LOGZ")) {
        } else if( (CHOPTN=="NFIT") || (CHOPTN=="FIT")) {

        //OpenPAW
        } else if( (CHOPTN=="PTIT") || (CHOPTN=="RTIT")) {
        } else if( (CHOPTN=="NBAC") || (CHOPTN=="BACK")) {
 
        } else {
          out << "pagraf :" 
              << " option " << Lib::sout(CHOPTN) 
              << " is a valid PAW one, but not yet supported in OpenPAW." 
              << Lib::endl;
        }

      }
    }

  } else if(cmd_path=="/GRAPHICS/MISC/CLR") {

    AIDA::IPlotter* plotter = 
      amanip::create_plotter(*aida,""); //FIXME : pass where.
    if(plotter) {
      plotter->clearRegions();
      delete plotter;
    }

  } else if(cmd_path=="/GRAPHICS/METAFILE") {

    int LUN = ku_geti();
    int METAFL = ku_geti();


    //METAFL= 4 Appendix E GKS. 
    //METAFL=-111 HIGZ/PostScript (Portrait). 
    //METAFL=-112 HIGZ/PostScript (Landscape). 
    //METAFL=-113 HIGZ/Encapsulated PostScript. 
    //METAFL=-114 HIGZ/PostScript Colour (Portrait). 
    //METAFL=-115 HIGZ/PostScript Colour (Landscape). 
    //METAFL=-777 HIGZ/LaTex Encapsulated. 
    //METAFL=-778 HIGZ/LaTex.

    if( (METAFL!=-111) &&
        (METAFL!=-112) &&
        (METAFL!=-113) &&
        (METAFL!=-114) &&
        (METAFL!=-115) ){
      out << "pagraf :" 
          << " METAFL " << METAFL
          << " is a valid PAW one, but not yet supported in OpenPAW." 
          << Lib::endl;
      return;
    }

    //LUN =-10 output only on metafile opened on unit 10;
    //LUN =  0 output only on screen;
    //LUN = 10 output on both screen and metafile opened on unit 10;

    if(LUN<0) LUN = -LUN;

    Units_type& units = pawManager->units();

    OpenPAW::Unit u = units.value(LUN);
    if(u==units.valueNotFound()) {
      out << "pagraf :" 
          << " unknown unit " << LUN
          << Lib::endl;
      return;
    }

    u.fType = OpenPAW::Unit::METAFILE;
    units.setValue(LUN,u);

    pawManager->warn_metaf_ps();

  } else {
    out << "pagraf :" 
        << " " << cmd_path << " : dummy." << Lib::endl;
  }
}
