extern "C" {
  void pahist_(void*);
}

#include <kuip/ksys.h>
#include <kuip/kfor.h>
#include <kuip/kuser.h>
#include <kuip/klink.h>

#include <AIDA/IPlotter.h>

#include <Lib/DLL.h>
#include <Lib/System.h>

#include "KuipHelpers.h"
#include "FindHelpers.h"
#include "PlotHistogram2D.h"

#define MINIMUM(a,b) ((a)<(b)?a:b)
#define MAXIMUM(a,b) ((a)>(b)?a:b)

#ifndef NotFound
#define NotFound (-1)
#endif

#define PALNB_FROM_HCOL \
  int PALNB;\
  if(!pawManager->paletteFromHCOL(PALNB)) PALNB = PALETTE_DEFAULT;

void pahist_(void* aSession) { 
  Slash::Core::ISession* session = (Slash::Core::ISession*)aSession;

  Slash::Core::IWriter& printer = session->printer();
  std::ostream& out = session->cout();

  //std::string cmd_name = ku_name();
  std::string cmd_path = ku_path() ? ku_path() : "";
  if(session->verboseLevel()) {
    out << "pahist :" 
        << " execute " << Lib::sout(cmd_path) 
        << std::endl;
  }

  OpenPAW::PAW_Manager* pawManager = find_PAW(*session);
  if(!pawManager) return;

  AIDA::IAnalysisFactory* aida = pawManager->aida();
  if(!aida) return;
  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return;

  OPTs_type& opts = pawManager->opts();
  ATTs_type& atts = pawManager->atts();

  if(cmd_path=="/HISTOGRAM/PLOT") {

    std::string NAME = ku_gets();
    std::string CHOPT = ku_gets();

    amanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    std::string ID;
    std::vector< std::pair<int,int> > dims;  
    if(!OpenPAW::PAW_Manager::parseName(NAME,ID,dims,printer)) return;

    AIDA::IBaseHistogram* histogram = baseHistogram_find(*pawManager,ID);
    if(!histogram) {
      out << "pahist :" 
          << " can't find histogram " << Lib::sout(ID) 
          << std::endl;
      return;
    }

    AIDA::IPlotter* plotter = 
      amanip::create_plotter(*aida,""); //FIXME : pass where.
    if(plotter) {
      pawManager->setPlotterLayout(*plotter);

      bool superpose = false;
      if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"S") ||
         OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"+") ){
        superpose = true;
      }
      pawManager->plotterNext(*plotter,superpose);

      if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"K")) {
        pawManager->setZONE_K(plotter->currentRegionNumber());
      }

      AIDA::IPlotterRegion* pregion = 0;
      if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"+")) {
        if(pawManager->ZONE_K()==NotFound) {         
          out << "pahist :" 
              << " option + : ZONE 'K' not set."
              << std::endl;
          
          delete plotter;
          return; 
        } else {
       	  pregion = plotter->region(pawManager->ZONE_K());
          if(!pregion) {
            out << "pahist :" 
                << " can't get region " << pawManager->ZONE_K()
                << std::endl;
            
            delete plotter;
            return; 
          }
        }
      } else {
        pregion = &(plotter->currentRegion());
      }

      AIDA::IPlotterRegion& region = *pregion;

      if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"+")) {
        region.setParameter("plotter.superposeBins","TRUE");
      } else {
        region.setParameter("plotter.superposeBins","FALSE");
      }

      pawManager->regionSetLayout(region);

      if(histogram->dimension()==1) {
        AIDA::IHistogram1D* histogram1D = 
          Lib_Cast(*histogram,AIDA::IHistogram1D);
        if(histogram1D) {

          if(region_setLimits(region,*histogram1D,dims,printer)) {

            bool empty = true;
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"C")) {
              pawManager->plotHistogram1D(region,*histogram1D,"C",false); 
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"L")) {
              pawManager->plotHistogram1D(region,*histogram1D,"L",false); 
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"E")) {
              pawManager->plotHistogram1D(region,*histogram1D,"E",false); 
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"P")) {
              pawManager->plotHistogram1D(region,*histogram1D,"P",false); 
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"*")) {
              pawManager->plotHistogram1D(region,*histogram1D,"*",false); 
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"B")) {
              pawManager->plotHistogram1D(region,*histogram1D,"B",false); 
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT," ")) {
              pawManager->plotHistogram1D(region,*histogram1D,"",false); 
              empty = false;
            }

            if(empty) {
              pawManager->plotHistogram1D(region,*histogram1D,"",false); 
            }

            // If histogram is fitted (with /HISTOGRAM/FIT)
            // BatchLab/BaseFitter/BatchLab_BaseFitter_fit
            // stores the name of the function on the histogram 
            // annotation "fit.function".
            AIDA::IAnnotation& annotation = histogram->annotation();
            std::string func_name = annotation.value("fit.function");
            if(func_name!="") {
              AIDA::IFunction* func = 
                amanip::find_function(*memoryTree,func_name);
              if(!func) {
                out << "pahist :" 
                    << " function of histo fit.function " 
                    << Slash::sout(func_name)
                    << " not found."
                    << std::endl;
              } else {
                int xbinn = histogram1D->axis().bins();
                double XLOW = histogram1D->axis().binLowerEdge(0);
                double XUP = histogram1D->axis().binUpperEdge(xbinn-1);
                pawManager->plotFunction1D(region,*func,XLOW,XUP,s_FUN_C);
              }
            }

          }

        } else {
          out << "pahist :" 
              << " cast problem."
              << std::endl;
        }

      } else if(histogram->dimension()==2) {
        AIDA::IHistogram2D* histogram2D = 
          Lib_Cast(*histogram,AIDA::IHistogram2D);
        if(histogram2D) {

          if(region_setLimits(region,*histogram2D,dims,printer)) {

            bool empty = true;


            // SURF <h2d> ! ! [1,2]
            //is the same than :
            // H/PLOT <h2d> SURF[1,2]  
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"SURF1")) {
              if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
                region.setParameter("plotter.colorMappingVisible","TRUE");
              }
              PALNB_FROM_HCOL
              region_plot_histogram2D
                (*session,region,*histogram2D,
                 s_SURF_LEVEL,printer,0,PALNB,*pawManager);
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"SURF2")) {
              if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
                region.setParameter("plotter.colorMappingVisible","TRUE");
              }
              PALNB_FROM_HCOL
              region_plot_histogram2D
                (*session,region,*histogram2D,s_SURF_LEVEL_NO_EDGE,
                 printer,0,PALNB,*pawManager);
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"SURF3")) {
              if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
                region.setParameter("plotter.colorMappingVisible","TRUE");
              }
              PALNB_FROM_HCOL
              region_plot_histogram2D
                (*session,region,*histogram2D,s_SURF_CONT,
                 printer,0,PALNB,*pawManager);
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"SURF4")) {
              if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
                region.setParameter("plotter.colorMappingVisible","TRUE");
              }
              region_plot_histogram2D
                (*session,region,*histogram2D,s_SURF_LIGHT,
                 printer,0,0,*pawManager);
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"SURF5")) {
              if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
                region.setParameter("plotter.colorMappingVisible","TRUE");
              }
              PALNB_FROM_HCOL
              region_plot_histogram2D(*session,region,*histogram2D,s_SURF_COL,
                                      printer,0,PALNB,*pawManager);
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_SURF)) {
              region_plot_histogram2D
                (*session,region,*histogram2D,s_SURF,printer,0,0,*pawManager);
              empty = false;
            }

            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"LEGO1")) {
              if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
                region.setParameter("plotter.colorMappingVisible","TRUE");
              }
              region_plot_histogram2D
                (*session,region,*histogram2D,
                 s_LEGO_LIGHT,printer,0,0,*pawManager);
              empty = false;
            } 
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"LEGO2")) {
              if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
                region.setParameter("plotter.colorMappingVisible","TRUE");
              }
              PALNB_FROM_HCOL
              region_plot_histogram2D
                (*session,region,*histogram2D,
                 s_LEGO_LEVEL,printer,0,PALNB,*pawManager);
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"LEGO3")) {
              if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
                region.setParameter("plotter.colorMappingVisible","TRUE");
              }
              PALNB_FROM_HCOL
              region_plot_histogram2D
                (*session,region,*histogram2D,
                 s_LEGO_COL,printer,0,PALNB,*pawManager);
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_LEGO)) {
              region_plot_histogram2D
                (*session,region,*histogram2D,s_LEGO,printer,0,0,*pawManager);
              empty = false;
            }

            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"TEXT")) {
              region_plot_histogram2D
                (*session,region,*histogram2D,"TEXT",printer,0,0,*pawManager);
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"CONT")) {
              region_plot_histogram2D
             (*session,region,*histogram2D,"CONT_DMOD",printer,10,0,*pawManager);
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"BOX")) {
              region_plot_histogram2D
                (*session,region,*histogram2D,"BOX",printer,0,0,*pawManager);
              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"COL")) {
              if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
                region.setParameter("plotter.colorMappingVisible","TRUE");
              }
              // HCOL 0.<p> does not seem to be taken into account.
              // But only the last PALETTE <p>.
              // And if no PALETTE command had been issued,
              // then the PALETTE_HISTOGRAM_PLOT_COL is taken.

              int PALNB = pawManager->currentPalette();
              if(PALNB==NotFound) {
                PALNB = PALETTE_HISTOGRAM_PLOT_COL;
              }

              region_plot_histogram2D
              (*session,region,*histogram2D,"COL",printer,0,PALNB,*pawManager);

              empty = false;
            }
            if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT," ")) {
              region_plot_histogram2D
                (*session,region,*histogram2D,"",printer,0,0,*pawManager); 
              empty = false;
            }

            if(empty) {
              region_plot_histogram2D
                (*session,region,*histogram2D,"",printer,0,0,*pawManager); 
            }

          }
        } else {
          out << "pahist :" 
              << " cast problem."
              << std::endl;
        }

      } else {
        out << "pahist :" 
            << " histo of dimension " << histogram->dimension()
            << " not treated."
            << std::endl;
      }

      

      delete plotter;
    }

  } else if(cmd_path=="/HISTOGRAM/FILE") {

    int LUN = ku_geti();
    std::string FNAME = ku_gets();
    /*int LRECL =*/ ku_geti();
    std::string CHOPT = ku_gets();

    amanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    // In Unit : key=LUN, value=Unit.
    Units_type& units = pawManager->units();

    if(LUN==0) {
      // If LUN is 0 the next free logical unit will be used.
      const Lib::Parameters<int,OpenPAW::Unit>::List& lunits = units.list();
      Lib::Parameters<int,OpenPAW::Unit>::ConstIterator it;
      for(it=lunits.begin();it!=lunits.end();++it) {
        LUN = MAXIMUM(LUN,(*it).fKey);
      }
      LUN++;
    }

    if(units.isKey(LUN)) {
      out << "pahist :" 
          << " LUN " << LUN << " already in used by OpenPAW." 
          << std::endl;
      return;
    }

    std::string path,name,suffix;
    if(!Lib::smanip::pathNameSuffix(FNAME,path,name,suffix)) {
      out << "pahist :" 
          << " problem getting path,name,suffix in " << Lib::sout(FNAME)
          << std::endl;
      return;
    }
    amanip::tolowercase(suffix);
    if(suffix=="gz") {
      // aptuple.tnt.gz
      // 01234567890123
      name = name.substr(0,name.size()-3); //rm ".gz"
      suffix = Lib::smanip::suffix(name);
      amanip::tolowercase(suffix);
    }

    bool readOnly = true;
    bool createNew = false;
    if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT," ")) {
      // Existing file is opened (read mode only).      
    }
    if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"N")) {
      // A new file is opened.
      readOnly = false;
      createNew = true;
    }
    if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"U")) {
      // Existing file is opened to be modified.
      readOnly = false;
      createNew = false;
    }
    if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"D")) {
      // Reset lock.
    }

    std::string type;
    std::string opts;
    if( (suffix=="hbook") || (suffix=="rz") ) {
      // Look if BatchLabZebra DLL is around.
      bool zebraAvailable = false;
     {Lib::DLL dll("BatchLabZebra",out,0,true);
      zebraAvailable = dll.isOpened();}
      if(!zebraAvailable) {
        out << "pahist :" 
            << " BatchLabZebra DLL not available"
            << " on this OpenPAW installation."
            << std::endl;
        out << "        " 
            << " We can't work on file " << Lib::sout(FNAME) << "."
            << std::endl;
        return;
      } 
      type = "zebra";
      Lib::smanip::printf(opts,64,"unit=%d",LUN);
    } else {
      type = suffix;

      // If using ROOT streaming, in Rio_ROOT we use
      // a bad trick to write/read annotation on the
      // fOption field.

      //if(type=="root") {
      // We need to save annotations. We can't do that
      // cleanly if using ROOT streaming.
      //opts = "streaming=native";
      //}
    }

    if(units.key
         (OpenPAW::Unit(OpenPAW::Unit::TREE,FNAME))!=units.keyNotFound()) {
      out << "pahist :" 
          << " " << Lib::sout(FNAME) << " already in used by OpenPAW." 
          << std::endl;
      return;
    }

    AIDA::ITreeFactory* treeFactory = aida->createTreeFactory();
    if(treeFactory) {
      AIDA::ITree* tree = 
        treeFactory->create(FNAME,type,readOnly,createNew,opts);
      delete treeFactory;
      if(!tree) {
        out << "pahist :" 
            << " unable to create AIDA tree " << Lib::sout(FNAME)
            << std::endl;
      } else {
        units.add(0,LUN,"",OpenPAW::Unit(OpenPAW::Unit::TREE,FNAME),
                           OpenPAW::Unit(),OpenPAW::Unit());
        pawManager->setCurrentTree(FNAME);
      }
    }

  } else if(cmd_path=="/HISTOGRAM/LIST") {

    std::string CHOPT = ku_gets();

    amanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    AIDA::ITree* tree = pawManager->findCurrentTree();
    if(!tree) {
      out << " Not current tree."
          << std::endl;
      return;
    }
    tree->ls("",false);

  } else if(cmd_path=="/HISTOGRAM/DELETE") {

    std::string ID = ku_gets();

    // The H/DELETE of PAW admit an ID, 0 or * as argument.
    // OpenPAW extends that to the same logic as VECTOR, that is to say
    // an "HLIST". An HLIST being names, that may have wildcards,
    // separated by colon.

    std::string HLIST = ID;
    //amanip::touppercase(HLIST);
    if(HLIST=="0") HLIST = "*";

    std::string::size_type star = HLIST.find('*');
    if(star==std::string::npos){
      std::string NAME = HLIST;
      if(!memoryTree->rm(NAME)) { //ok.
        out << "pahist :" 
            << " unable to remove " << Lib::sout(NAME) 
            << std::endl;
      }
    } else { //wildcard.
      std::vector<std::string> objs = memoryTree->listObjectNames(".",true);
      std::vector<std::string> typs = memoryTree->listObjectTypes(".",true);
      if(typs.size()!=objs.size()) {
        out << "pahist :" 
            << " problem retreiving directory infos."
            << std::endl;
        return;
      }
      unsigned int objn = objs.size();
      for(unsigned int index = 0;index<objn;index++) {
        std::string& name = objs[index];
        std::string& type = typs[index];
        //printf("debug : %s %s\n",name.c_str(),type.c_str());
        if(type.find("IHistogram")!=std::string::npos) {
          if(Lib::smanip::match(name,HLIST)) {
            if(session->verboseLevel()) {
              out << "pahist :" 
                  << Lib::sout(name) << " matches " << Lib::sout(HLIST) 
                  << std::endl;
            }
            if(!memoryTree->rm(name)) { //ok.
              out << "pahist :" 
                  << " unable to remove " << Lib::sout(name) 
                  << " in tree " << memoryTree->storeName()
                  << std::endl;
            }
          }
        }
      }
    }

  } else if(cmd_path=="/HISTOGRAM/2D_PLOT/CONTOUR") {

    std::string ID = ku_gets();
    int NLEVEL = ku_geti();
    std::string CHOPT = ku_gets();
    if(session->verboseLevel()) {
      out << "pavect :" 
          << " " << cmd_path
          << " ID= " << Lib::sout(ID) 
          << " NLEVEL= " << NLEVEL
          << " CHOPT= " << Lib::sout(CHOPT) 
          << std::endl;
    }

    amanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    //std::string PARAM = ku_gets(); //Vector

    std::string NAME;
    std::vector< std::pair<int,int> > dims;  
    if(!OpenPAW::PAW_Manager::parseName(ID,NAME,dims,printer)) return;

    AIDA::IHistogram2D* histogram = histogram2D_find(*pawManager,NAME);
    if(!histogram) {
      out << "pahist :" 
          << " can't find histogram 2D " << Lib::sout(NAME) 
          << std::endl;
      return;
    }

    AIDA::IPlotter* plotter = 
      amanip::create_plotter(*aida,""); //FIXME : pass where.
    if(plotter) {
      pawManager->setPlotterLayout(*plotter);
  
      pawManager->plotterNext
        (*plotter,OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"S"));

      AIDA::IPlotterRegion& region = plotter->currentRegion();

      if(region_setLimits(region,*histogram,dims,printer)) {

        pawManager->regionSetLayout(region);
  
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"3")) {
          //-3 The contour is drawn with filled colour levels.

          PALNB_FROM_HCOL //FIXME : ok ?

          region_plot_histogram2D(*session,region,*histogram,
            "CONT_COL_FILLED",printer,NLEVEL,PALNB,*pawManager);
  
        } else {
          // wire frame :
          //CHOPT 0 has priority over 1.
          if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"0")) {
            //-0 Use colour to distinguish contours.

            // In CERN/PAW The palette seems to be hard coded 
            // to one with 4 elements (1,2,3,4)
            int PALNB = PALETTE_HISTOGRAM_2D_PLOT_CONTOUR_OPT_0;

            std::string old_COLL = opts.value("COLL");
            opts.setValue("COLL","COLV");
            region_plot_histogram2D(*session,region,*histogram,"CONT_COL",
                                     printer,NLEVEL,PALNB,*pawManager);
            opts.setValue("COLL",old_COLL);

          } else { //CHOPT 1,2
            //-1 Use line style to distinguish contours.
            //FIXME -2 Line style and colour are the same for all contours.
            region_plot_histogram2D(*session,region,*histogram,"CONT_DMOD",
                                     printer,NLEVEL,0,*pawManager);
          }
  
        }

      }

      

      delete plotter;
    }

  } else if(cmd_path=="/HISTOGRAM/2D_PLOT/SURFACE") {

    std::string ID = ku_gets();
    double THETA = ku_getr();
    double PHI = ku_getr();
    std::string CHOPT = ku_gets();

    amanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    std::string NAME;
    std::vector< std::pair<int,int> > dims;  
    if(!OpenPAW::PAW_Manager::parseName(ID,NAME,dims,printer)) return;

    AIDA::IHistogram2D* histogram = histogram2D_find(*pawManager,NAME);
    if(!histogram) {
      out << "pahist :" 
          << " can't find histogram 2D " << Lib::sout(NAME) 
          << std::endl;
      return;
    }

    AIDA::IPlotter* plotter = 
      amanip::create_plotter(*aida,""); //FIXME : pass where.
    if(plotter) {
      pawManager->setPlotterLayout(*plotter);

      pawManager->plotterNext
        (*plotter,OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"S"));

      AIDA::IPlotterRegion& region = plotter->currentRegion();

      if(region_setLimits(region,*histogram,dims,printer)) {

        pawManager->regionSetLayout(region);
  
        double old_THETA = pawManager->THETA();
        double old_PHI = pawManager->PHI();

        pawManager->setTHETA(THETA);
        pawManager->setPHI(PHI);

        //if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT," "))
        //if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"1"))
        //if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"2"))
        //if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"3"))
        //if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"4"))

        // SURF <h2d> ! ! [1,2]
        //is the same than :
        // H/PLOT <h2d> SURF[1,2]
  
        //FIXME : -1 show bin edgeds. Do we want that ?
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"1")) {
          if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
            region.setParameter("plotter.colorMappingVisible","TRUE");
          }
          //CERN/PAW doc :
          //-1 Hidden surface algorithm is used and each cell is filled_
          //   with a colour corresponding to the Z value.
          PALNB_FROM_HCOL
          region_plot_histogram2D(*session,region,*histogram,s_SURF_LEVEL,
                                  printer,0,PALNB,*pawManager);
  
        } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"2")) {
          if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
            region.setParameter("plotter.colorMappingVisible","TRUE");
          }
          PALNB_FROM_HCOL
          region_plot_histogram2D(*session,region,*histogram,
                                  s_SURF_LEVEL_NO_EDGE,
                                  printer,0,PALNB,*pawManager);
  
        } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"3")) {
          if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
            region.setParameter("plotter.colorMappingVisible","TRUE");
          }
          PALNB_FROM_HCOL
          region_plot_histogram2D(*session,region,*histogram,
                                  s_SURF_CONT,
                                  printer,0,PALNB,*pawManager);
  
        } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"4")) {
          if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
            region.setParameter("plotter.colorMappingVisible","TRUE");
          }
          region_plot_histogram2D(*session,region,*histogram,s_SURF_LIGHT,
                                  printer,0,0,*pawManager);
  
        } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"5")) {
          if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
            region.setParameter("plotter.colorMappingVisible","TRUE");
          }
          PALNB_FROM_HCOL
          region_plot_histogram2D(*session,region,*histogram,s_SURF_COL,
                                  printer,0,PALNB,*pawManager);
  
        } else {
          region_plot_histogram2D(*session,region,*histogram,s_SURF,
                                  printer,0,0,*pawManager);
        }
  
        pawManager->setTHETA(old_THETA);
        pawManager->setPHI(old_PHI);
      }

      

      delete plotter;
    }

  } else if(cmd_path=="/HISTOGRAM/2D_PLOT/LEGO") {

    std::string ID = ku_gets();
    double THETA = ku_getr();
    double PHI = ku_getr();
    std::string CHOPT = ku_gets();

    amanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    std::string NAME;
    std::vector< std::pair<int,int> > dims;  
    if(!OpenPAW::PAW_Manager::parseName(ID,NAME,dims,printer)) return;

    AIDA::IHistogram2D* histogram = histogram2D_find(*pawManager,NAME);
    if(!histogram) {
      out << "pahist :" 
          << " can't find histogram 2D " << Lib::sout(NAME) 
          << std::endl;
      return;
    }

    AIDA::IPlotter* plotter = 
      amanip::create_plotter(*aida,""); //FIXME : pass where.
    if(plotter) {
      pawManager->setPlotterLayout(*plotter);

      pawManager->plotterNext
        (*plotter,OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"S"));

      AIDA::IPlotterRegion& region = plotter->currentRegion();

      if(region_setLimits(region,*histogram,dims,printer)) {

        pawManager->regionSetLayout(region);
  
        double old_THETA = pawManager->THETA();
        double old_PHI = pawManager->PHI();

        pawManager->setTHETA(THETA);
        pawManager->setPHI(PHI);

        //if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT," "))
        //if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"1"))
        //if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"2"))
        //if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"3"))
        //if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"4"))
  
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"1")) {
          if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
            region.setParameter("plotter.colorMappingVisible","TRUE");
          }
          region_plot_histogram2D
             (*session,region,*histogram,
              s_LEGO_LIGHT,printer,0,0,*pawManager);

        } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"2")) {
          if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
            region.setParameter("plotter.colorMappingVisible","TRUE");
          }
          PALNB_FROM_HCOL
          region_plot_histogram2D
            (*session,region,*histogram,
             s_LEGO_LEVEL,printer,0,PALNB,*pawManager);

        } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"3")) {
          if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,s_Z)) {
            region.setParameter("plotter.colorMappingVisible","TRUE");
          }
          PALNB_FROM_HCOL
          region_plot_histogram2D
            (*session,region,*histogram,
             s_LEGO_COL,printer,0,PALNB,*pawManager);

        } else {
          region_plot_histogram2D
            (*session,region,*histogram,s_LEGO,printer,0,0,*pawManager);
        }
  
        pawManager->setTHETA(old_THETA);
        pawManager->setPHI(old_PHI);
      }

      

      delete plotter;
    }

  } else {
    out << "pahist :" 
        << " " << cmd_path << " : dummy." << std::endl;

  }
}
