extern "C" {
  void pahope_(void*);
}

#include <kuip/ksys.h>
#include <kuip/kfor.h>
#include <kuip/kuser.h>
#include <kuip/klink.h>

// AIDA helpers :
#include <Slash/Tools/AIDA.h>

#include "KuipHelpers.h"
#include "FindHelpers.h"

//////////////////////////////////////////////////////////////////////////////
void pahope_(
 void* aSession
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  Slash::Core::ISession* session = (Slash::Core::ISession*)aSession;

  Slash::Core::IWriter& printer = session->printer();
  Lib::Out out(printer);

  std::string cmd_path = ku_path() ? ku_path() : "";
  std::string cmd_name = ku_name();
  if(session->verboseLevel()) {
    out << "pahope :" 
        << " execute " << Lib::sout(cmd_path) 
        << Lib::endl;
  }

  OpenPAW::PAW_Manager* pawManager = find_PAW(*session);
  if(!pawManager) return;
  AIDA::IAnalysisFactory* aida = pawManager->aida();
  if(!aida) return;
  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return;

  if( (cmd_name=="ADD") || 
      (cmd_name=="SUBTRACT") ||
      (cmd_name=="MULTIPLY") ||
      (cmd_name=="DIVIDE") )  {

    //Add histograms: ID3 = C1*ID1 + C2*ID2.

    std::string ID1 = ku_gets();
    std::string ID2 = ku_gets();
    std::string ID3 = ku_gets();
    double C1 = ku_getr();
    double C2 = ku_getr();

    std::string OPTION = ku_gets();
    Lib::smanip::touppercase(OPTION);
    std::vector<std::string> OPTS;
    kuip_getOptions("OPTION",OPTS);

    AIDA::IBaseHistogram* h1 = baseHistogram_find(*pawManager,ID1);
    if(!h1) {
      out << "pahope :" 
          << " can't find histogram " << Lib::sout(ID1) 
          << Lib::endl;
      return;
    }
    AIDA::IBaseHistogram* h2 = 
      (ID2==ID1 ? h1 : baseHistogram_find(*pawManager,ID2));
    if(!h2) {
      out << "pahope :" 
          << " can't find histogram " << Lib::sout(ID2) 
          << Lib::endl;
      return;
    }
    if(h1->dimension()!=h2->dimension()) {
      out << "pahope :" 
          << " dimension mismatch."
          << Lib::endl;
      return;
    }
    Lib::smanip::replace(ID3,"//PAWC","/");

    if(h1->dimension()==1) {
      AIDA::IHistogram1D* h1_1D = Lib_Cast(*h1,AIDA::IHistogram1D);
      if(!h1_1D) {
        out << "pahope :" 
            << " cast problem."
            << Lib::endl;
        return;
      }
      AIDA::IHistogram1D* h2_1D = Lib_Cast(*h2,AIDA::IHistogram1D);
      if(!h2_1D) {
        out << "pahope :" 
            << " cast problem."
            << Lib::endl;
        return;
      }
      AIDA::IHistogramFactory* histogramFactory = 
        aida->createHistogramFactory(*memoryTree);
      if(!histogramFactory) return;
      std::string stmp1 = "tmp_"+ID1;
      AIDA::IHistogram1D* h1_tmp = 
        histogramFactory->createCopy(amanip::path_not_managed()+"/"+stmp1,*h1_1D);
      if(h1_tmp) {
        h1_tmp->scale(C1);
        std::string stmp2 = "tmp_"+ID2;
        AIDA::IHistogram1D* h2_tmp = 
          histogramFactory->createCopy(amanip::path_not_managed()+"/"+stmp2,*h2_1D);
        if(h2_tmp) {
          h2_tmp->scale(C2);
          AIDA::IHistogram1D* hres = 0;
          if(cmd_name=="ADD") 
            hres = histogramFactory->add(ID3,*h1_tmp,*h2_tmp);
          else if(cmd_name=="SUBTRACT") 
            hres = histogramFactory->subtract(ID3,*h1_tmp,*h2_tmp);
          else if(cmd_name=="MULTIPLY") 
            hres = histogramFactory->multiply(ID3,*h1_tmp,*h2_tmp);
          else if(cmd_name=="DIVIDE") 
            hres = histogramFactory->divide(ID3,*h1_tmp,*h2_tmp);
          if(hres) hres->setTitle(h1_1D->title());
          delete h2_tmp;
        }
        delete h1_tmp;
      }
      delete histogramFactory;

    } else if(h1->dimension()==2) {
      AIDA::IHistogram2D* h1_2D = Lib_Cast(*h1,AIDA::IHistogram2D);
      if(!h1_2D) {
        out << "pahope :" 
            << " cast problem."
            << Lib::endl;
        return;
      }
      AIDA::IHistogram2D* h2_2D = Lib_Cast(*h2,AIDA::IHistogram2D);
      if(!h2_2D) {
        out << "pahope :" 
            << " cast problem."
            << Lib::endl;
        return;
      }
      AIDA::IHistogramFactory* histogramFactory = 
        aida->createHistogramFactory(*memoryTree);
      if(!histogramFactory) return;
      std::string stmp1 = "tmp_"+ID1;
      AIDA::IHistogram2D* h1_tmp = 
        histogramFactory->createCopy(amanip::path_not_managed()+"/"+stmp1,*h1_2D);
      if(h1_tmp) {
        h1_tmp->scale(C1);
        std::string stmp2 = "tmp_"+ID2;
        AIDA::IHistogram2D* h2_tmp = 
          histogramFactory->createCopy(amanip::path_not_managed()+"/"+stmp2,*h2_2D);
        if(h2_tmp) {
          h2_tmp->scale(C2);
          AIDA::IHistogram2D* hres = 0;
          if(cmd_name=="ADD") 
            hres = histogramFactory->add(ID3,*h1_tmp,*h2_tmp);
          else if(cmd_name=="SUBTRACT") 
            hres = histogramFactory->subtract(ID3,*h1_tmp,*h2_tmp);
          else if(cmd_name=="MULTIPLY") 
            hres = histogramFactory->multiply(ID3,*h1_tmp,*h2_tmp);
          else if(cmd_name=="DIVIDE") 
            hres = histogramFactory->divide(ID3,*h1_tmp,*h2_tmp);
          if(hres) hres->setTitle(h1_2D->title());
          delete h2_tmp;
        }
        delete h1_tmp;
      }
      delete histogramFactory;
    }

  } else {
    out << "pahope :" 
        << " " << cmd_path << " : dummy." << Lib::endl;

  }
}
