extern "C" {
  void papict_(void*);
}

#include <kuip/ksys.h>
#include <kuip/kfor.h>
#include <kuip/kuser.h>
#include <kuip/klink.h>

#include <AIDA/IPlotter.h>

#include <Slash/Core/ISession.h>
#include <Slash/Core/IWriter.h>
#include <Slash/Tools/AIDA.h>

#include <amanip/create.h>

#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/smanip.h>
#include <Lib/Manager.h>
#include <Lib/System.h>

#include <OpenPAW/PAW_Manager.h>

#include "KuipHelpers.h"

//////////////////////////////////////////////////////////////////////////////
void papict_(
 void* aSession
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  Slash::Core::ISession* session = (Slash::Core::ISession*)aSession;

  Slash::Core::IWriter& printer = session->printer();
  Lib::Out out(printer);

  std::string cmd_path = ku_path() ? ku_path() : "";
  //std::string cmd_name = ku_name();
  if(session->verboseLevel()) {
    out << "papict :" 
        << " execute " << Lib::sout(cmd_path) 
        << Lib::endl;
  }
  
  OpenPAW::PAW_Manager* pawManager = find_PAW(*session);
  if(!pawManager) return;

  AIDA::IAnalysisFactory* aida = pawManager->aida();
  if(!aida) return;

  ATTs_type& atts = pawManager->atts();

  if(cmd_path=="/PICTURE/IGSET") {

    std::string CHATT = ku_gets();
    double VALUE = ku_getr();
    Lib::smanip::touppercase(CHATT);

    int cat;
    if(CHATT=="SHOW") {  

      std::ostream& out = session->cout();
      out << "+------------------------------------------------------------------------------+" << std::endl;
      out << "|                        IGSET : Current values in use                         |" << std::endl;
      pawManager->showATTs(out,cat_IGSET);

    } else if(CHATT=="*") { 

      atts.setCategoryToDefault(cat_IGSET);

    } else if(!atts.isKey(CHATT)) {

      out << "papict :" 
          << " " << Lib::sout(CHATT) << " is not an attribute." 
          << Lib::endl;

    } else if(!atts.category(CHATT,cat)||(cat!=cat_IGSET)) {

      out << "papict :" 
          << " " << Lib::sout(CHATT) << " is not an IGSET attribute." 
          << Lib::endl;

    } else {

      if(VALUE==0) VALUE = atts.defaultValue(CHATT); //PAW convention.
      atts.setValue(CHATT,VALUE);
 
    }

  } else if(cmd_path=="/PICTURE/PRINT") {

    std::string FILE = ku_gets();
    /*int WIDTH =*/ ku_geti();  // used to scale .gif
    /*int HEIGHT =*/ ku_geti();
    std::string CHOPT = ku_gets();

    Lib::smanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    bool sendToPrinter = false;
    if((FILE==" ")||(FILE=="HIGZPRINTER")) {
      FILE = "paw.ps";
      sendToPrinter = true; 
    }

    std::string stype;

    if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"PSPIX")) {
      stype = "PS";
    } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"GIF")) {
      stype = "GIF";
    } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"JPEG")) {
      stype = "JPEG";
    } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"GL2PS")) {
      stype = "GL2PS";
    } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"GL2PDF")) {
      stype = "GL2PDF";
    } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"GL2SVG")) {
      stype = "GL2SVG";
    } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"GL2TEX")) {
      stype = "GL2TEX";
    } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"GL2PGF")) {
      stype = "GL2PGF";
    } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"HIV")) {
      stype = "HIV";
    } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"IV")) {
      stype = "IV";
    } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"WRL")) {
      stype = "WRL";
    } else if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"POV")) {
      stype = "POV";
    } else {
      stype = pawManager->print_fmt(FILE);
    }

    AIDA::IPlotter* plotter = 
      amanip::create_plotter(*aida,""); //FIXME : pass where.
    if(plotter) {
      plotter->writeToFile(FILE,stype);
      delete plotter;
    }   

    if(sendToPrinter) {
      // If defined HIGZPRINTER should be of the form :
      //   csh> setenv HIGZPRINTER 'lp -dprinter_name paw.ps'
      //    sh> HIGZPRINTER='lp -dprinter_name paw.ps';export HIGZPRINTER
      std::string HIGZPRINTER;
      if(Lib::System::getenv("HIGZPRINTER",HIGZPRINTER)) {
        Lib::System::execute(HIGZPRINTER);
      }
    }

  } else {
    out << "papict :" 
        << " " << cmd_path << " : dummy." << Lib::endl;
  }
}
