extern "C" {
  void pasigm_(void*);

  int sigma_C(const char*,void*); 
}

#include <kuip/ksys.h>
#include <kuip/kfor.h>
#include <kuip/kuser.h>
#include <kuip/klink.h>

#include <Slash/Core/ISession.h>
#include <Slash/Core/IWriter.h>
#include <Slash/Data/IArray.h>
#include <Slash/Store/ITree.h> //FIXME
#include <Slash/Tools/AIDA.h>
#include <Slash/Tools/smanip.h>

#include <amanip/find.h>

#include <Lib/Processor.h>
#include <Lib/Array.h>
#include <Lib/valuemanip.h>

#include <BatchLab/Core/ArrayFactory.h> //FIXME : not AIDA.

#include "ArrayHelpers.h"
#include <OpenPAW/PAW_Manager.h>

//static 
//bool SIGMA(OpenPAW::PAW_Manager&,const std::string&);
//So that Shark find the source code of SIGMA.
extern "C" {
  bool SIGMA(OpenPAW::PAW_Manager&,const std::string&);
}
static Lib::Array<double> IArray2Array(const Slash::Data::IArray&); 
static bool setIArray(Slash::Data::IArray&,
                      const std::vector<unsigned int>&, 
                      const std::vector<double>&); 

//////////////////////////////////////////////////////////////////////////////
int sigma_C(
 const char* aScript
,void* aPAW_Manager
) 
//////////////////////////////////////////////////////////////////////////////
// Called by kuip when doing $SIGMA(<exp>)
// It is used in PAW.cxx to initialize KUIP/SIGMA through a call to kusigm.
// It is the equivalent of CERNLIBG/pawlib/sigma/src/sigmae.F
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  OpenPAW::PAW_Manager* pawManager = (OpenPAW::PAW_Manager*)aPAW_Manager;
  //printf("debug : sigma_C : try to proceed \"%s\"...\n",aScript);
  SIGMA(*pawManager,aScript); 
  //return 1; //Failure
  return 0; //Success
}
//////////////////////////////////////////////////////////////////////////////
void pasigm_(
 void* aSession
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  Slash::Core::ISession* session = (Slash::Core::ISession*)aSession;

  OpenPAW::PAW_Manager* pawManager = find_PAW(*session);
  if(!pawManager) return;

  std::ostream& out = session->cout();

  std::string cmd_path = ku_path() ? ku_path() : "";
  //std::string cmd_name = ku_name();
  if(session->verboseLevel()) {
    out << "pasigm :" 
        << " execute " << Slash::sout(cmd_path) 
        << std::endl;
  }

  if(cmd_path=="/FORTRAN/SIGMA") {

    std::string EXPR;
    int argc = ku_npar();
    for(int index=0;index<argc;index++) EXPR += ku_gets();

    SIGMA(*pawManager,EXPR); 

  } else {
    out << "pasigm :" 
        << " " << cmd_path << " : dummy." << std::endl;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool SIGMA(
 OpenPAW::PAW_Manager& aPawManager
,const std::string& aScript
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aScript.empty()) return true;

  Slash::Core::ISession& session = aPawManager.session();
  std::ostream& out = session.cout();

  AIDA::ITree* memoryTree = aPawManager.mem();
  if(!memoryTree) return false;

  std::string::size_type eq = aScript.find('=');
  if(eq==std::string::npos) {
    out << "SIGMA :" 
        << " no assignement in " << Slash::sout(aScript) << std::endl;
    return false;
  }
  
  std::string assign = aScript.substr(0,eq);
  Slash::strip(assign);
  Slash::touppercase(assign);

  std::string script = aScript.substr(eq+1,aScript.size()-(eq+1));  
  Slash::tolowercase(script);
  Slash::replace(script,'#',','); //For ARRAY(x,y#z)

  Lib::Processor* processor = aPawManager.processor();
  if(!processor) {
    out << "SIGMA :" 
        << " processor not found." << std::endl;
    return false;
  }

  processor->setScript(script);

  std::vector<std::string> names;
  if(!processor->findVariables(names)) {
    out << "SIGMA :" 
        << " can't find variables in " << Slash::sout(aScript) << std::endl;
    return false;
  }
  if(!processor->compile(names)) {
    out << "SIGMA :" 
        << " compilation of " << Slash::sout(aScript) << " failed." 
        << std::endl;
    return false;
  }

  // Get variables :
  Lib::Parameters<std::string,Lib::Value>& variables = 
    aPawManager.variables();

  std::vector<Lib::Value> vals;
  std::vector<unsigned int> orders;
  bool foundOrders = false;
  unsigned int number = names.size();
  for(unsigned int count=0;count<number;count++) {
    std::string name = names[count];
    Slash::touppercase(name);
    Slash::Data::IArray* vector = find_vector(*memoryTree,name);
    if(!vector) {
      if(!variables.isKey(name)) {  // Search in sSIGMA variables :
        out << "SIGMA :" 
            << " can't find variable " << Slash::sout(name) << std::endl;
        return false;
      } else {
        vals.push_back(variables.value(name));
      }
    } else {
      if(!foundOrders) {
        orders = vector->orders();
        foundOrders = true;
      } else {
        if(vector->orders()!=orders) {
          out << "SIGMA :" 
              << " orders mismatch."
              << std::endl;
          return false;
        }
      }
      Lib::Array<double> arr = IArray2Array(*vector);
      vals.push_back(Lib::Value(arr.orders(),arr.vector()));
    }
  }

 {Slash::Data::IArray* vassign = find_vector(*memoryTree,assign);
  //if(vassign) memoryTree->rm(assign);
  delete vassign;
  if(variables.isKey(assign)) variables.remove(assign);}

  Lib::Value var;
  std::string serror;
  if(!processor->execute(vals,var,serror)) {
    out << "SIGMA :" 
        << " execution of " << Slash::sout(aScript) << " failed." 
        << std::endl;
    out << serror << std::endl;
    return false;
  }
  if(serror==Lib_valuemanip_error_div_zero) {
    // See Lib::valuemanip::divide.
    // continue anyway. 
    out << "SIGMA :" 
        << " execution of " << Slash::sout(aScript) 
        << " has divide by zero errors, but continue anyway." 
        << std::endl;
  }

  if(var.type()==Slash::Core::IValue::DOUBLE) {

    variables.add(0,assign,"",var,Lib::Value(),Lib::Value()); 
  } else if(var.type()==Slash::Core::IValue::UNSIGNED_INT) {

    variables.add
      (0,assign,"",Lib::Value((double)var.get_unsigned_int()),
                 Lib::Value(),Lib::Value());    

  } else if(var.type()==Slash::Core::IValue::INT) {

    variables.add(0,assign,"",Lib::Value((double)var.get_int()),
                          Lib::Value(),Lib::Value());    

  } else if(var.type()==Slash::Core::IValue::ARRAY_DOUBLE) {

    std::vector<unsigned int> is;
    const std::vector<double>& stdv = var.get_array_double(is);

    //FIXME
    Slash::Store::ITree* baseTree = aPawManager.mem_store_itree();
    if(!baseTree) return false;

    BatchLab::ArrayFactory factory(session.printer(),*baseTree); //FIXME
    Slash::Data::IArray* vassign = factory.create(assign,assign,is);
    if(!vassign) {
      out << "SIGMA :" 
          << " can't create vector " << Slash::sout(assign) << std::endl;
      return false;
    } 
    if(!setIArray(*vassign,is,stdv)) {
      out << "SIGMA :" 
          << " can't assign vector " << Slash::sout(assign) << std::endl;
      return false;
    }
  } else {
    out << "SIGMA :" 
        << " result not a vector or a double." 
        << std::endl;
    return false;
  }

  return true;
}
//////////////////////////////////////////////////////////////////////////////
static Lib::Array<double> IArray2Array(
 const Slash::Data::IArray& aArray
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Array<double> vec(aArray.orders());
  std::vector<double>& stdvec = vec.vector();
  const std::vector<double>& astdv = aArray.vector();
  int sz = astdv.size();
  for(int index=0;index<sz;index++) stdvec[index] = astdv[index];
  return vec;
}
//////////////////////////////////////////////////////////////////////////////
static bool setIArray(
 Slash::Data::IArray& aArray
,const std::vector<unsigned int>& aIS
,const std::vector<double>& aVec
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aArray.orders()!=aIS) return false;
  std::vector<double>& astdv = aArray.vector();
  int sz = aVec.size();
  for(int index=0;index<sz;index++) astdv[index] = aVec[index];
  return true;
} 
