extern "C" {
  void pavect_(void*);
}

#include <kuip/ksys.h>
#include <kuip/kfor.h>
#include <kuip/kuser.h>
#include <kuip/klink.h>

// AIDA helpers :
#include <Slash/Tools/AIDA.h>
#include <Slash/Tools/smanip.h>
#include <amanip/create.h>

// AIDA :
#include <AIDA/IPlotter.h>
#include <AIDA/ICloud2D.h>

#include "ArrayHelpers.h"

#include "KuipHelpers.h"
#include "FindHelpers.h"

#define MINIMUM(a,b) ((a)<(b)?a:b)
#define MAXIMUM(a,b) ((a)>(b)?a:b)

//////////////////////////////////////////////////////////////////////////////
void pavect_(
 void* aSession
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  Slash::Core::ISession* session = (Slash::Core::ISession*)aSession;

  std::ostream& out = session->cout();

  Slash::Core::IWriter& printer = session->printer();

  //std::string cmd_name = ku_name();
  std::string cmd_path = ku_path() ? ku_path() : "";
  if(session->verboseLevel()) {
    out << "pavect :" 
        << " execute " << Lib::sout(cmd_path) 
        << std::endl;
  }

  OpenPAW::PAW_Manager* pawManager = find_PAW(*session);
  if(!pawManager) return;

  AIDA::IAnalysisFactory* aida = pawManager->aida();
  if(!aida) return;
  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return;

  if(cmd_path=="/VECTOR/DRAW") {

    std::string VNAME = ku_gets();
    std::string ID = ku_gets();
    std::string CHOPT = ku_gets();
    if(session->verboseLevel()) {
      out << "pavect :" 
          << " " << cmd_path
          << " VNAME= " << Lib::sout(VNAME) 
          << " ID= " << Lib::sout(ID) 
          << " CHOPT= " << Lib::sout(CHOPT) 
          << std::endl;
    }

    Lib::smanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    std::string NAME;
    std::vector< std::pair<int,int> > dims;
    if(!OpenPAW::PAW_Manager::parseName(VNAME,NAME,dims,printer)) return;
    Lib::smanip::touppercase(NAME);
    Slash::Data::IArray* vector = find_vector(*memoryTree,NAME);
    if(!vector) {
      out << "pavect :" 
          << " can't find vector " << Lib::sout(NAME) << std::endl;
      return;
    } 
    std::vector< std::pair<unsigned int,unsigned int> > cut;
    if(!vector_get_cut(*vector,dims,cut)) {
      out << "pavect :" 
            << " mismatch dimension in " << Lib::sout(VNAME) 
          << std::endl;
      return;
    }

    const std::vector<double>& v = vector->vector();
    unsigned int vsize = v.size();

    int n = 0;
   {for(unsigned int index = 0;index<vsize;index++) {
      if(vector->accept(index,cut)) n++;
    }}


    //FIXME : must be able to do a region.plot(Slash::Data::IArray).

    // create temporary not managed histo :
    std::string name = VNAME + "_rep";
    AIDA::IHistogram1D* rep = 
      amanip::create_histogram_1D(*aida,*memoryTree,
                         amanip::path_not_managed()+"/"+name,VNAME,n,1,n+1);
    if(!rep) {
      out << "pavect :" 
            << " can't create histogram " << Lib::sout(name) 
          << " in memory tree."
          << std::endl;
      return;
    }

    int i = 0;
    for(unsigned int index = 0;index<vsize;index++) {
      if(vector->accept(index,cut)) {
        rep->fill(i+1.5,v[index]);
        i++;
      }
    }

    AIDA::IPlotter* plotter = 
      amanip::create_plotter(*aida,""); //FIXME : pass where.
    if(plotter) {
      pawManager->setPlotterLayout(*plotter);

      pawManager->plotterNext
        (*plotter,OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"S"));

      AIDA::IPlotterRegion& region = plotter->currentRegion();

      pawManager->regionSetLayout(region);

      bool empty = true;
      if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"C")) {
        pawManager->plotHistogram1D(region,*rep,"C",true); 
        empty = false;
      }
      if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"L")) {
        pawManager->plotHistogram1D(region,*rep,"L",true); 
        empty = false;
      }
      if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"E")) {
        pawManager->plotHistogram1D(region,*rep,"E",true); 
        empty = false;
      }
      if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"P")) {
        pawManager->plotHistogram1D(region,*rep,"P",true); 
        empty = false;
      }
      if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"*")) {
        pawManager->plotHistogram1D(region,*rep,"*",true); 
        empty = false;
      }
      if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"B")) {
        pawManager->plotHistogram1D(region,*rep,"B",true); 
        empty = false;
      }
      if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT," ")) {
        pawManager->plotHistogram1D(region,*rep,"",true); 
        empty = false;
      }

      if(empty) {
        pawManager->plotHistogram1D(region,*rep,"",true); 
      }
      
      delete plotter;
    }

    // Plotter is in retain mode, we can now get rid of the rep.
    delete rep;

  } else if(cmd_path=="/VECTOR/PLOT") {

    std::string VNAME = ku_gets();
    std::string ID = ku_gets();
    std::string CHOPT = ku_gets();

    if(session->verboseLevel()) {
      out << "pavect :" 
          << " " << cmd_path
          << " VNAME= " << Lib::sout(VNAME) 
          << " ID= " << Lib::sout(ID) 
          << " CHOPT= " << Lib::sout(CHOPT) 
          << std::endl;
    }

    Lib::smanip::touppercase(CHOPT);
    std::vector<std::string> OPTS;
    kuip_getOptions("CHOPT",OPTS);

    std::vector<std::string> VNAMEs;
    Slash::words(VNAME,"%",false,VNAMEs);
    if(VNAMEs.size()==2) {
      std::string YNAME = VNAMEs[0];
      Lib::smanip::touppercase(YNAME);
      Slash::Data::IArray* yvec = find_vector(*memoryTree,YNAME);
      if(!yvec) {
        out << "pavect :" 
            << " can't find vector " << Lib::sout(YNAME) << std::endl;
        return;
      } 
      std::string XNAME = VNAMEs[1];
      Lib::smanip::touppercase(XNAME);
      Slash::Data::IArray* xvec = find_vector(*memoryTree,XNAME);
      if(!xvec) {
        out << "pavect :" 
            << " can't find vector " << Lib::sout(XNAME) << std::endl;
        return;
      } 

      int xsize = xvec->size();
      int ysize = yvec->size();
      if(xsize!=ysize) {
        out << "pavect :" 
            << " size mismatch for vectors " 
            << Lib::sout(XNAME) << " and "
            << Lib::sout(YNAME) << "."
            << std::endl;
        return;
      }

      // create temporary not managed cloud :
      std::string name = "cloud2D_" + XNAME + "_" + YNAME;
      AIDA::ICloud2D* cloud = 
        amanip::create_cloud_2D
          (*aida,*memoryTree,amanip::path_not_managed()+"/"+name,"");
      if(!cloud) {
        out << "pavect :" 
            << " can't create a cloud2D." << std::endl;
        return;
      }

      const std::vector<double>& xv = xvec->vector();
      const std::vector<double>& yv = yvec->vector();
      for(int index=0;index<xsize;index++) cloud->fill(xv[index],yv[index]);

      AIDA::IPlotter* plotter = 
        amanip::create_plotter(*aida,""); //FIXME : where 
      if(plotter) {
        pawManager->setPlotterLayout(*plotter);

        pawManager->plotterNext
          (*plotter,OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"S"));
        AIDA::IPlotterRegion& region = plotter->currentRegion();

        pawManager->regionSetLayout(region);

        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"C"))
          pawManager->plotCloud2D(region,*cloud,"C");
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"L"))
          pawManager->plotCloud2D(region,*cloud,"L");
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"P"))
          pawManager->plotCloud2D(region,*cloud,"P");

        

        delete plotter;
      }

       // Plotter is in retain mode, we can now get rid of the object.
      delete cloud;

    } else {

      std::string NAME;
      std::vector< std::pair<int,int> > dims;
      if(!OpenPAW::PAW_Manager::parseName(VNAME,NAME,dims,printer)) return;
      Lib::smanip::touppercase(NAME);
      Slash::Data::IArray* vector = find_vector(*memoryTree,NAME);
      if(!vector) {
        out << "pavect :" 
            << " can't find vector " << Lib::sout(NAME) << std::endl;
        return;
      } 
      std::vector< std::pair<unsigned int,unsigned int> > cut;
      if(!vector_get_cut(*vector,dims,cut)) {
        out << "pavect :" 
              << " mismatch dimension in " << Lib::sout(VNAME) 
            << std::endl;
        return;
      }
  
      const std::vector<double>& v = vector->vector();
      unsigned int vsize = v.size();
  
      double xmn = 0;
      double xmx = 0;
      int n = 0;
      double value;
     {for(unsigned int index = 0;index<vsize;index++) {
        if(vector->accept(index,cut)) {
          value = v[index];
          if(n==0) {
            xmx = xmn = value;
          } else {
            xmn = MINIMUM(xmn,value);
            xmx = MAXIMUM(xmx,value);
          }
          n++;
        }
      }}
  
      if(n<=0) {
        out << "pavect :" 
              << " vector selection is empty." << Lib::sout(VNAME) 
            << std::endl;
        return;
      }
  
      if(xmx==xmn) {
        double value = xmn;
        xmn = value - 1;
        xmx = value + 1;
      } else { 
        double dx = 0.1 * (xmx - xmn);
        xmn -= dx;
        xmx += dx;
      }
       
      if(memoryTree->find(ID)) memoryTree->rm(ID); //ok.
      AIDA::IHistogram1D* rep = 
        amanip::create_histogram_1D(*aida,*memoryTree,ID,VNAME,100,xmn,xmx);
      if(!rep) {
        out << "pavect :" 
              << " can't create histogram " << Lib::sout(ID) 
            << " in memory tree."
            << std::endl;
        return;
      }
  
      for(unsigned int index = 0;index<vsize;index++) {
        if(vector->accept(index,cut)) {
          rep->fill(v[index]);
        }
      }
  
      AIDA::IPlotter* plotter = 
        amanip::create_plotter(*aida,""); //FIXME : pass where.
      if(plotter) {
        pawManager->setPlotterLayout(*plotter);
  
        pawManager->plotterNext
          (*plotter,OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"S"));
  
        AIDA::IPlotterRegion& region = plotter->currentRegion();
  
        pawManager->regionSetLayout(region);
  
        bool empty = true;
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"C")) {
          pawManager->plotHistogram1D(region,*rep,"C",false); 
          empty = false;
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"L")) {
          pawManager->plotHistogram1D(region,*rep,"L",false); 
          empty = false;
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"E")) {
          pawManager->plotHistogram1D(region,*rep,"E",false); 
          empty = false;
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"P")) {
          pawManager->plotHistogram1D(region,*rep,"P",false); 
          empty = false;
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"*")) {
          pawManager->plotHistogram1D(region,*rep,"*",false); 
          empty = false;
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT,"B")) {
          pawManager->plotHistogram1D(region,*rep,"B",false); 
          empty = false;
        }
        if(OpenPAW::PAW_Manager::isAnOption(OPTS,CHOPT," ")) {
          pawManager->plotHistogram1D(region,*rep,"",false); 
          empty = false;
        }

        if(empty) {
          pawManager->plotHistogram1D(region,*rep,"",false); 
        }   
        
  
        delete plotter;
      }
 
    } //end VNAMEs.size() == 2

  } else if(cmd_path=="/VECTOR/HFILL") {

    std::string VNAME = ku_gets();
    std::string ID = ku_gets();

    std::string NAME;
    std::vector< std::pair<int,int> > dims;
    if(!OpenPAW::PAW_Manager::parseName(VNAME,NAME,dims,printer)) return;
    Lib::smanip::touppercase(NAME);
    Slash::Data::IArray* vector = find_vector(*memoryTree,NAME);
    if(!vector) {
      out << "pavect :" 
          << " can't find vector " << Lib::sout(NAME) << std::endl;
      return;
    } 
    std::vector< std::pair<unsigned int,unsigned int> > cut;
    if(!vector_get_cut(*vector,dims,cut)) {
      out << "pavect :" 
            << " mismatch dimension in " << Lib::sout(VNAME) 
          << std::endl;
      return;
    }

    AIDA::IBaseHistogram* histogram = baseHistogram_find(*pawManager,ID);
    if(!histogram) {
      out << "pavect :" 
          << " can't find histogram " << Lib::sout(ID) 
          << std::endl;
      return;
    }

    if(histogram->dimension()!=1) {
      out << "pavect :" 
          << " histogram of dimension " << histogram->dimension()
          << " not yet supported."
          << std::endl;
      return;
    }

    
    AIDA::IHistogram1D* histogram1D = Lib_Cast(*histogram,AIDA::IHistogram1D);

    const std::vector<double>& v = vector->vector();
    unsigned int vsize = v.size();

    for(unsigned int index = 0;index<vsize;index++) {
      if(vector->accept(index,cut)) {
        histogram1D->fill(v[index]);
      }
    }

  } else {
    out << "pavect :" 
        << " " << cmd_path << " : dummy." << std::endl;

  }
}
