extern "C" {
  void pnutil_(void*);
}

#include <kuip/ksys.h>
#include <kuip/kfor.h>
#include <kuip/kuser.h>
#include <kuip/klink.h>

#include <Slash/Core/ISession.h>
#include <Slash/Core/IWriter.h>
#include <Slash/Tools/AIDA.h>

#include <amanip/find.h>
#include <amanip/create.h>
#include <amanip/hprint.h>

#include <Lib/Out.h>
#include <Lib/sout.h>
#include <Lib/smanip.h>
#include <Lib/fmanip.h>
#include <Lib/FileReader.h>
#include <Lib/Time.h>
#include <Lib/cstring.h>

#include "FindHelpers.h"

inline std::string sout(const std::string& a_string){
  return std::string("\"")+a_string+"\"";
}

//////////////////////////////////////////////////////////////////////////////
void pnutil_(
 void* aSession
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{ 
  Slash::Core::ISession* session = (Slash::Core::ISession*)aSession;
  std::ostream& out = session->cout();  

  Slash::Core::IWriter& printer = session->printer();

  std::string cmd_path = ku_path() ? ku_path() : "";
  //std::string cmd_name = ku_name();
  if(session->verboseLevel()) {
    out << "pnutil :" 
        << " execute " << sout(cmd_path) 
        << std::endl;
  }

  OpenPAW::PAW_Manager* pawManager = find_PAW(*session);
  if(!pawManager) return;

  AIDA::IAnalysisFactory* aida = pawManager->aida();
  if(!aida) return;
  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return;

  if(cmd_path=="/NTUPLE/CREATE") {

    std::string IDN = ku_gets();
    std::string TITLE = ku_gets();
    /*int NVAR =*/ ku_geti();
    std::string CHRZPA = ku_gets();
    /*int NPRIME =*/ ku_geti();
    std::string VARLIST = ku_gets();
 
    if(memoryTree->find(IDN)) {
      memoryTree->rm(IDN); //ok.
      out << " *** WARNING in OpenPAW :"
          << " Already existing tuple replaced : ID="
          << IDN 
          << std::endl;
    }
    Lib::smanip::replace(VARLIST,","," ");

    AIDA::ITuple* tuple = 
      amanip::create_tuple(*aida,*memoryTree,IDN,TITLE,VARLIST);
    if(!tuple) {
      out << "pnutil :" 
            << " can't create tuple " << sout(IDN) << std::endl;
      return;
    }

  } else if(cmd_path=="/NTUPLE/READ") {

    std::string IDN = ku_gets();
    std::string FNAME = ku_gets();
    std::string FORMAT = ku_gets();
    std::string OPT = ku_gets();
    /*int NEVENT =*/ ku_geti();
    std::string MATCH = ku_gets();
 
    AIDA::ITuple* tuple = 
      amanip::find_tuple(*memoryTree,IDN,false,out);
    if(!tuple) {
      out << "pnutil :" 
          << " can't find tuple " << sout(IDN) 
          << " in tree " << sout(memoryTree->storeName()) 
          << std::endl;
      return;
    }

    Lib::FileReader reader(FNAME,printer);
    if(!reader.enabled()) {
      out << "pnutil :" 
          << " can't open file " << sout(FNAME) << std::endl;
      return;
    }

    std::vector<double> row(tuple->columns());

#define BUFSIZE 65536
    char buffer[BUFSIZE+1];

    Lib::Time begin = Lib::Time::now();

    unsigned int l;
    while(1) {
      if(!reader.sread(buffer,sizeof(buffer),l)) break; //EOF.
      Lib_strtods(buffer,row,true); //the last true is important.
      if(!tuple->fill(row)) {
        out << "pnutil :" 
            << " problem filling tuple from file" 
            << " " << sout(FNAME) << "."
            << std::endl;
        return;
      }
    }

    Lib::Time end = Lib::Time::elapsed(begin);

    out << " ==>   " << tuple->rows() << " events have been read" 
        << " in " << end.seconds() << " s (elapsed)."
        << std::endl;

  } else if(cmd_path=="/NTUPLE/PRINT") {

    std::string IDN = ku_gets();
 
    AIDA::ITree* tree = 0;
    AIDA::ITuple* tuple = tuple_find(*pawManager,IDN,tree);
    if(!tuple) {
      out << "pnutil :" 
          << " object " << sout(IDN) 
          << " not an AIDA::ITuple."
          << std::endl;
      return;
    }

    amanip::hprint(*tuple,IDN,out);

  } else {
    out << "pnutil :" 
        << " " << cmd_path << " : dummy." << std::endl;

  }
}
