extern "C" {
  int vector_create_C(void*);
  int vector_delete_C(void*);
  int vector_read_C(void*);
  int vector_write_C(void*);
  int vector_input_C(void*);
  int vector_list_C(void*);
  int vector_operation_C(void*);
  int vector_print_C(void*);
  int vector_find_C(const char*,void*);
  int vector_value_C(const char*,double*,void*);
}

#include <kuip/ksys.h>
#include <kuip/kfor.h>
#include <kuip/kuser.h>

// Slash :
#include <Slash/Core/ISession.h>
#include <Slash/Tools/AIDA.h>
#include <Slash/Tools/smanip.h>

#include <amanip/find.h>
#include <amanip/smanip.h>

#include <Lib/smanip.h>
#include <Lib/fmanip.h>
#include <Lib/FortranOut.h>
#include <Lib/FileWriter.h>
#include <Lib/valuemanip.h>
#include <Lib/vmanip.h>

#include <BatchLab/Core/ArrayFactory.h> //FIXME : not AIDA.

#include "ArrayHelpers.h"
#include <OpenPAW/PAW_Manager.h>

#include "KuipHelpers.h"

#define MINIMUM(a,b) ((a)<(b)?a:b)
#define MAXIMUM(a,b) ((a)>(b)?a:b)
#define NotFound (-1)

int vector_create_C(void* aPAW_Manager){
  // From KUIP/kvect.c/kxvcreate
  OpenPAW::PAW_Manager* pawManager = (OpenPAW::PAW_Manager*)aPAW_Manager;

  Slash::Core::ISession& session = pawManager->session();

  if(session.verboseLevel()) {
    std::ostream& out = session.cout();
    out << "vector_create_C : execute." << std::endl;
  }

  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return -1;
  Slash::Store::ITree* baseTree = pawManager->mem_store_itree();
  if(!baseTree) return -1;

  std::string VNAME = ku_gets();
  std::string TYPE = ku_gets();
  std::string VALUES = ku_gets();

  //printf("debug : \"%s\" \"%s\" \"%s\"\n",
  // VNAME.c_str(),TYPE.c_str(),VALUES.c_str());

  if( (TYPE!="I") && (TYPE!="i") && (TYPE!="R") && (TYPE!="r") ) {
    //if implementing, change also the ret value of vector_find_C
    std::ostream& out = session.cout();
    out << "vector_create_C :" 
        << " TYPE " << amanip::sout(TYPE) << " not supported." << std::endl;
    return -1;
  }

  //bool real = ((TYPE=="R") || (TYPE=="r") ? true : false);

  std::string NAME;
  std::vector< std::pair<int,int> > dims;
  if(!OpenPAW::PAW_Manager::parseName
        (VNAME,NAME,dims,session.printer())) return -1;
  amanip::touppercase(NAME);

  // Create :
  if(!dims.size()) {
    dims.push_back( std::pair<int,int>(0,0));
  }
  unsigned int dimension = dims.size();
  std::vector<unsigned int> orders(dimension);
  std::vector< std::pair<unsigned int,unsigned int> > cut;
  for(unsigned int index = 0;index<dimension;index++) {
    //FIXME : not clear if range are accepted in the CREATE command.
    //if(dims[index].first!=dims[index].second) {
    //  out << "vector_create_C :" 
    //            << " Bad format in " << amanip::sout(VNAME) 
    //      << " but continue anyway." << std::endl;
    //}
    orders[index] = dims[index].second+1;
    if(dims[index].second==NotFound) {
      std::ostream& out = session.cout();
      out << "vector_create_C :" 
          << " have a NotFound max." << std::endl;
      return -1;
    }
    cut.push_back(std::pair<unsigned int,unsigned int>(0,dims[index].second));
  }
  BatchLab::ArrayFactory factory(session.printer(),*baseTree); //FIXME

  if(memoryTree->find(NAME)) {
    memoryTree->rm(NAME); //ok.
    std::ostream& out = session.cout();
    out << " *** VECTOR/CREATE " << NAME
        << ": existing vector " << NAME << " replaced"
        << std::endl;
  }

  Slash::Data::IArray* vector = factory.create(NAME,NAME,orders);
  if(!vector) {
    std::ostream& out = session.cout();
    out << "vector_create_C :" 
        << " can't create vector " << amanip::sout(NAME) << "."
        << std::endl;
    return -1;
  }

  // Values :
  std::vector<double> values;
  if(!amanip::todoubles(VALUES,',',values)) {
    std::ostream& out = session.cout();
    out << "vector_create_C :" 
        << " can't convert " << amanip::sout(VALUES) << " to doubles." 
        << std::endl;
  } else {
    if(values.size()) {
      if(!vector->fill(values,&cut)) {
        std::ostream& out = session.cout();
        out << "vector_create_C :" 
            << " vector.fill returns false."
            << std::endl;
      }
    }
  }

  if(session.verboseLevel()) {
    std::ostream& out = session.cout();
    print_vector(out,*vector,std::string("vector ")+NAME+" :");
  }

  return -1;
}

int vector_delete_C(void* aPAW_Manager){
  OpenPAW::PAW_Manager* pawManager = (OpenPAW::PAW_Manager*)aPAW_Manager;

  Slash::Core::ISession& session = pawManager->session();

  if(session.verboseLevel()) {
    std::ostream& out = session.cout();
    out << "vector_delete_C : execute." << std::endl;
  }

  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return -1;
 
  std::string VLIST = ku_gets();

  amanip::touppercase(VLIST);
  std::vector<std::string> V_LIST;
  Slash::words(VLIST,",",false,V_LIST);

  for(std::vector<std::string>::const_iterator it = V_LIST.begin();
      it!=V_LIST.end();++it) {

    std::string::size_type star = (*it).find('*');
    if(star==std::string::npos){
      //Avoid warnings of a memoryTree->rm in case obj is not here.
      AIDA::IManagedObject* mobj = memoryTree->find(*it);
      if(mobj) delete mobj;
    } else { //wildcard.
      std::vector<std::string> objs = memoryTree->listObjectNames();
      std::vector<std::string> typs = memoryTree->listObjectTypes();
      if(typs.size()!=objs.size()) {
        std::ostream& out = session.cout();
        out << "vector_delete_C :" 
            << " problem retreiving directory infos."
            << std::endl;
        return -1;
      }
      unsigned int objn = objs.size();
      for(unsigned int index = 0;index<objn;index++) {
        std::string& name = objs[index];
        std::string& type = typs[index];
        //printf("debug : %s %s\n",name.c_str(),type.c_str());
        if(type=="Slash::Data::IArray") {
          if(Lib::smanip::match(name,*it)) {
            if(session.verboseLevel()) {
              std::ostream& out = session.cout();
              out << "vector_delete_C :" 
                  << amanip::sout(name) << " matches " << amanip::sout(*it) 
                  << std::endl;
            }
            //Avoid warnings of a memoryTree->rm in case obj is not here.
            AIDA::IManagedObject* mobj = memoryTree->find(name);
            if(mobj) delete mobj;
          }
        }
      }
    }

  }

  return -1;
}

int vector_read_C(void* aPAW_Manager){
  OpenPAW::PAW_Manager* pawManager = (OpenPAW::PAW_Manager*)aPAW_Manager;

  Slash::Core::ISession& session = pawManager->session();

  if(session.verboseLevel()) {
    std::ostream& out = session.cout();
    out << "vector_read_C : execute." << std::endl;
  }

  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return -1;
  Slash::Store::ITree* baseTree = pawManager->mem_store_itree();
  if(!baseTree) return -1;

  std::string VLIST = ku_gets(); //FIXME : handle list of vectors.
  std::string FNAME = ku_gets();
  //std::string FORMAT = ku_gets();
  //std::string CHOPT = ku_gets();
  //std::string MATCH = ku_gets();

  amanip::touppercase(VLIST);
  std::vector<std::string> V_LIST;
  Slash::words(VLIST,",",false,V_LIST);

  unsigned int coln = V_LIST.size();
  if(!coln) {
    std::ostream& out = session.cout();
    out << "vector_read_C :" 
        << " VLIST empty." << std::endl;
    return -1;
  }

  std::vector<double> values;
  if(!Lib::fmanip::read(FNAME,values)) {
    std::ostream& out = session.cout();
    out << "vector_read_C :" 
        << " can't read file " << amanip::sout(FNAME) << std::endl;
    return -1;
  }

  if(session.verboseLevel()) {
    std::ostream& out = session.cout();
    out << "vector_read_C :" 
        << " file " << amanip::sout(FNAME) << std::endl;
    for(unsigned int index=0;index<values.size();index++) {
      out << values[index] << std::endl;
    }
  }

  unsigned int rows = values.size()/coln;
  if((rows*coln)!=values.size()) {
    std::ostream& out = session.cout();
    out << "vector_read_C :" 
        << " too much data in file." 
        << " Read only first " << (rows*coln) << " values."
        << std::endl;
  }

  for(unsigned int col=0;col<coln;col++) {
    std::string NAME = V_LIST[col];

    Slash::Data::IArray* vector = find_vector(*memoryTree,NAME);
    if(!vector) { //create it
      std::ostream& out = session.cout();
      out << " *** VECTOR/READ: Unknown vector " 
          << NAME << " is created with length        " 
          << rows << std::endl;
      std::vector<unsigned int> orders(1);
      orders[0] = rows;
      BatchLab::ArrayFactory factory(session.printer(),*baseTree); //FIXME
      vector = factory.create(NAME,NAME,orders);
      if(!vector) {
        std::ostream& out = session.cout();
        out << "vector_read_C :" 
            << " can't create vector " << amanip::sout(NAME) << "."
            << std::endl;
        return -1;
      }
    }

    // brute force :
    std::vector<double> v_items(rows);
    for(unsigned int row=0;row<rows;row++) {
      v_items[row] = values[col+row*coln];
    }

    if(v_items.size()) {
      if(!vector->fill(v_items,0)) {
        std::ostream& out = session.cout();
        out << "vector_read_C :" 
            << " vector.fill returns false."
            << std::endl;
      }
    }
  
    if(session.verboseLevel()) {
      std::ostream& out = session.cout();
      print_vector(out,*vector,std::string("vector ")+NAME+" :");
    }
  }

  return -1;
}

int vector_write_C(void* aPAW_Manager){
  OpenPAW::PAW_Manager* pawManager = (OpenPAW::PAW_Manager*)aPAW_Manager;

  Slash::Core::ISession& session = pawManager->session();

  if(session.verboseLevel()) { 
    std::ostream& out = session.cout();
    out << "vector_wite_C : execute." << std::endl;
  }

  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return -1;

  std::string VLIST = ku_gets();
  std::string FNAME = ku_gets();
  std::string FORMAT = ku_gets();
  std::string CHOPT = ku_gets();

  if(session.verboseLevel()) {
    std::ostream& out = session.cout();
    out << "vector_wite_C :" 
        << " FNAME = " << amanip::sout(FNAME) 
        << " FORMAT = " << amanip::sout(FORMAT) 
        << std::endl;
  }

  amanip::touppercase(CHOPT);
  std::vector<std::string> OPTS;
  kuip_getOptions("CHOPT",OPTS);
  if(session.verboseLevel()) {
    std::ostream& out = session.cout();
    out << "vector_wite_C :" 
        << " CHOPT " << amanip::sout(CHOPT) 
        << std::endl;
  }

  amanip::touppercase(VLIST);
  std::vector<std::string> V_LIST;
  Slash::words(VLIST,",",false,V_LIST);

  unsigned int coln = V_LIST.size();
  if(!coln) {
    std::ostream& out = session.cout();
    out << "vector_write_C :" 
        << " VLIST empty." << std::endl;
    return -1;
  }

  std::vector<Slash::Data::IArray*> as(coln);

  unsigned int rows = 0;
 {for(unsigned int col=0;col<coln;col++) {
    std::string NAME = V_LIST[col];
    Slash::Data::IArray* vector = find_vector(*memoryTree,NAME);
    if(!vector) {
      std::ostream& out = session.cout();
      out << "vector_write_C :" 
          << " can't find vector " << amanip::sout(NAME) 
          << " in tree " << amanip::sout(memoryTree->storeName()) 
          << std::endl;
      return -1;
    }
    as[col] = vector;  
    const std::vector<double>& v = vector->vector();  
    if(col) {
      rows = MINIMUM(rows,v.size());
    } else {
      rows = v.size();
    }
  }}

 {for(unsigned int col=0;col<coln;col++) {
    Slash::Data::IArray* vector = as[col];
    const std::vector<double>& v = vector->vector();  
    if(v.size()!=rows) {
      std::ostream& out = session.cout();
      out << "vector_write_C :" 
          << " mismatch size for vector " << amanip::sout(V_LIST[col]) 
          << ". It has " << v.size() << " entries, whilst "
          << rows << " are expected."
          << " Only " << rows << " are going to be written."
          << std::endl;
    }
  }}

  // brute force :
  std::vector<double> vout(rows*coln);
 {for(unsigned int row=0;row<rows;row++) {
    for(unsigned int col=0;col<coln;col++) {
      Slash::Data::IArray* vector = as[col];
      const std::vector<double>& v = vector->vector();  
      vout[col+row*coln] = v[row];
    }
  }}

  std::string name = Lib::fmanip::name(FNAME);
  Lib::FileWriter fileWriter(name,session.printer());
  Lib::FortranOut f77out(fileWriter,"("+FORMAT+")");
  f77out << vout << Lib::endl;
  
  return -1;
}

int vector_input_C(void* aPAW_Manager){
  OpenPAW::PAW_Manager* pawManager = (OpenPAW::PAW_Manager*)aPAW_Manager;

  Slash::Core::ISession& session = pawManager->session();

  if(session.verboseLevel()) {
    std::ostream& out = session.cout();
    out << "vector_input_C : execute." << std::endl;
  }

  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return -1;

  std::string VNAME = ku_gets();
  std::string VALUES = ku_gets();

  // Values :
  std::vector<double> values;
  if(!amanip::todoubles(VALUES,',',values)) {
    std::ostream& out = session.cout();
    out << "vector_input_C :" 
        << " can't convert " << amanip::sout(VALUES) << " to doubles." 
        << std::endl;
    return -1;
  }
  if(session.verboseLevel()) {
    std::ostream& out = session.cout();
    out << "vector_input_C :" 
        << " VALUES " << std::endl;
    for(unsigned int index=0;index<values.size();index++) {
      out << values[index] << std::endl;
    }
  }

  std::string NAME;
  std::vector< std::pair<int,int> > dims;
  if(!OpenPAW::PAW_Manager::parseName
        (VNAME,NAME,dims,session.printer())) return -1;
  amanip::touppercase(NAME);
  Slash::Data::IArray* vector = find_vector(*memoryTree,NAME);
  if(!vector) {
    std::ostream& out = session.cout();
    out << "vector_input_C :" 
        << " can't find vector " << amanip::sout(NAME) << std::endl;
    return -1;
  } 
  std::vector< std::pair<unsigned int,unsigned int> > cut;
  if(!vector_get_cut(*vector,dims,cut)) {
    std::ostream& out = session.cout();
    out << "vector_input_C :" 
        << " mismatch dimension in " << amanip::sout(NAME) 
        << std::endl;
    return -1;
  }

  if(values.size()) {
    if(!vector->fill(values,&cut)) {
      std::ostream& out = session.cout();
      out << "vector_input_C :" 
          << " vector.fill returns false."
          << std::endl;
    }
  }

  if(session.verboseLevel()) {
    std::ostream& out = session.cout();
    print_vector(out,*vector,std::string("vector ")+NAME+" :");
  }

  return -1;
}

int vector_operation_C(void* aPAW_Manager){
  OpenPAW::PAW_Manager* pawManager = (OpenPAW::PAW_Manager*)aPAW_Manager;

  Slash::Core::ISession& session = pawManager->session();

  std::string cmd_name = ku_name();
  if(session.verboseLevel()) {
    std::ostream& out = session.cout();
    out << "vector_operation_C : execute " << amanip::sout(cmd_name) 
        << std::endl;
  }

  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return -1;

  //FIXME
  Slash::Store::ITree* baseTree = 
    Lib_SafeCast(*memoryTree,Slash::Store::ITree);
  if(!baseTree) return -1;

  BatchLab::ArrayFactory factory(session.printer(),*baseTree);

  //cmd_path==/VECTOR/OPERATIONS/VBIAS
  //cmd_path==/VECTOR/OPERATIONS/VSCALE
  //cmd_path==/VECTOR/OPERATIONS/VADD
  //cmd_path==/VECTOR/OPERATIONS/VSUBTRACT
  //cmd_path==/VECTOR/OPERATIONS/VMULTIPLY
  //cmd_path==/VECTOR/OPERATIONS/VDIVIDE

  if( (cmd_name=="VBIAS") || (cmd_name=="VSCALE") )  {

    std::string VNAME1 = ku_gets();
    double BIAS = ku_getr();
    double SCALE = BIAS;
    std::string VNAME2 = ku_gets();

    std::string NAME1;
    std::vector< std::pair<int,int> > dims1;
    if(!OpenPAW::PAW_Manager::parseName
          (VNAME1,NAME1,dims1,session.printer())) return -1;
    amanip::touppercase(NAME1);
    Slash::Data::IArray* vec1 = find_vector(*memoryTree,NAME1);
    if(!vec1) {
      std::ostream& out = session.cout();
      out << "vector_operation_C :" 
          << " can't find vector " << amanip::sout(NAME1) 
          << " in tree " << amanip::sout(memoryTree->storeName()) 
          << std::endl;
      return -1;
    }
    std::vector< std::pair<unsigned int,unsigned int> > cut1;
    if(!vector_get_cut(*vec1,dims1,cut1)) {
      std::ostream& out = session.cout();
      out << "vector_operation_C :" 
          << " mismatch dimension in " << amanip::sout(VNAME1) 
          << std::endl;
      return -1;
    }

    std::string NAME2;
    std::vector< std::pair<int,int> > dims2;
    if(!OpenPAW::PAW_Manager::parseName
          (VNAME2,NAME2,dims2,session.printer())) return -1;
    amanip::touppercase(NAME2);

    if(cmd_name=="VBIAS") {
      if(NAME2==NAME1) { //operation on itself
        vec1->add(BIAS,&cut1);
      } else {
        if(memoryTree->find(NAME2)) memoryTree->rm(NAME2); //ok.
        Slash::Data::IArray* vector = factory.add(NAME2,*vec1,BIAS);
        if(!vector) {
          std::ostream& out = session.cout();
          out << "vector_operation_C :" 
              << " can't add vector " << amanip::sout(NAME1) << "."
              << std::endl;
          return -1;
        }
      }

    } else if(cmd_name=="VSCALE") {  

      if(NAME2==NAME1) { //operation on itself
        vec1->multiply(SCALE);
      } else {
        if(memoryTree->find(NAME2)) memoryTree->rm(NAME2); //ok.
        Slash::Data::IArray* vector = factory.multiply(NAME2,*vec1,SCALE);
        if(!vector) {
          std::ostream& out = session.cout();
          out << "vector_operation_C :" 
              << " can't scale vector " << amanip::sout(NAME1) << "."
              << std::endl;
          return -1;
        }
      }

    }
    

  } else if( (cmd_name=="VADD") || 
             (cmd_name=="VSUBTRACT") ||
             (cmd_name=="VMULTIPLY") ||
             (cmd_name=="VDIVIDE") )  {

    std::string VNAME1 = ku_gets();
    std::string VNAME2 = ku_gets();
    std::string VNAME3 = ku_gets();

    std::string NAME1;
    std::vector< std::pair<int,int> > dims1;
    if(!OpenPAW::PAW_Manager::parseName
          (VNAME1,NAME1,dims1,session.printer())) return -1;
    amanip::touppercase(NAME1);
    Slash::Data::IArray* vec1 = find_vector(*memoryTree,NAME1);
    if(!vec1) {
      std::ostream& out = session.cout();
      out << "vector_operation_C :" 
          << " can't find vector " << amanip::sout(NAME1) 
          << " in tree " << amanip::sout(memoryTree->storeName()) 
          << std::endl;
      return -1;
    }
    std::vector< std::pair<unsigned int,unsigned int> > cut1;
    if(!vector_get_cut(*vec1,dims1,cut1)) {
      std::ostream& out = session.cout();
      out << "vector_operation_C :" 
          << " mismatch dimension in " << amanip::sout(VNAME1) 
          << std::endl;
      return -1;
    }

    std::string NAME2;
    std::vector< std::pair<int,int> > dims2;
    if(!OpenPAW::PAW_Manager::parseName
          (VNAME2,NAME2,dims2,session.printer())) return -1;
    amanip::touppercase(NAME2);
    Slash::Data::IArray* vec2 = find_vector(*memoryTree,NAME2);
    if(!vec2) {
      std::ostream& out = session.cout();
      out << "vector_operation_C :" 
          << " can't find vector " << amanip::sout(NAME2) 
          << " in tree " << amanip::sout(memoryTree->storeName()) 
          << std::endl;
      return -1;
    }
    std::vector< std::pair<unsigned int,unsigned int> > cut2;
    if(!vector_get_cut(*vec2,dims2,cut2)) {
      std::ostream& out = session.cout();
      out << "vector_operation_C :" 
          << " mismatch dimension in " << amanip::sout(VNAME2) 
          << std::endl;
      return -1;
    }

    std::string NAME3;
    std::vector< std::pair<int,int> > dims3;
    if(!OpenPAW::PAW_Manager::parseName
          (VNAME3,NAME3,dims3,session.printer())) return -1;
    amanip::touppercase(NAME3);

    if(cmd_name=="VADD") {

      if(NAME3==NAME1) { //operation on itself
        // V1 + V2 = V1
        vec1->add(*vec2);
      } else if(NAME3==NAME2) { //operation on itself
        // V1 + V2 = V2
        vec2->add(*vec1);
      } else {
        if(memoryTree->find(NAME3)) memoryTree->rm(NAME3); //ok.
        Slash::Data::IArray* vector = factory.add(NAME3,*vec1,*vec2);
        if(!vector) {
          std::ostream& out = session.cout();
          out << "vector_operation_C :" 
              << " can't add vector " 
              << amanip::sout(NAME1) << " with "
              << amanip::sout(NAME2) << "."
              << std::endl;
          return -1;
        }
      }

    } else if(cmd_name=="VSUBTRACT") {

      if(NAME3==NAME1) { //operation on itself
        // V1 - V2 = V1
        vec1->subtract(*vec2);
      } else if(NAME3==NAME2) { //operation on itself
        // V1 - V2 = V2
        vec2->multiply(-1);
        vec2->add(*vec1);
      } else {
        if(memoryTree->find(NAME3)) memoryTree->rm(NAME3); //ok.
        Slash::Data::IArray* vector = factory.subtract(NAME3,*vec1,*vec2);
        if(!vector) {
          std::ostream& out = session.cout();
          out << "vector_operation_C :" 
              << " can't subtract vector " 
              << amanip::sout(NAME1) << " with "
              << amanip::sout(NAME2) << "."
              << std::endl;
          return -1;
        }
      }

    } else if(cmd_name=="VMULTIPLY") {

      if(NAME3==NAME1) { //operation on itself
        // V1 * V2 = V1
        vec1->multiply(*vec2);
      } else if(NAME3==NAME2) { //operation on itself
        // V1 * V2 = V2
        vec2->multiply(*vec1);
      } else {
        if(memoryTree->find(NAME3)) memoryTree->rm(NAME3); //ok.
        Slash::Data::IArray* vector = factory.multiply(NAME3,*vec1,*vec2);
        if(!vector) {
          std::ostream& out = session.cout();
          out << "vector_operation_C :" 
              << " can't multiply vector " 
              << amanip::sout(NAME1) << " with "
              << amanip::sout(NAME2) << "."
              << std::endl;
          return -1;
        }
      }

    } else if(cmd_name=="VDIVIDE") {

      if(NAME3==NAME1) { //operation on itself
        // V1 / V2 = V1
        vec1->divide(*vec2);
      } else if(NAME3==NAME2) { //operation on itself
        // V1 / V2 = V2
        vec2->invert();
        vec2->multiply(*vec1);
      } else {
        if(memoryTree->find(NAME3)) memoryTree->rm(NAME3); //ok.
        Slash::Data::IArray* vector = factory.divide(NAME3,*vec1,*vec2);
        if(!vector) {
          std::ostream& out = session.cout();
          out << "vector_operation_C :" 
              << " can't divide vector " 
              << amanip::sout(NAME1) << " with "
              << amanip::sout(NAME2) << "."
              << std::endl;
          return -1;
        }
      }

    }

  }

  return -1;
}
namespace OpenPAW {
class PrintRange {    
public:
  unsigned int begin;
  unsigned int end;
  double value;
public:
  void dump(std::ostream& a_out,Slash::Data::IArray& aArray,const std::string& aName,const std::vector<int>& aDigits,bool aDENSE) {
    unsigned int dim = aArray.dimension();
    std::vector<unsigned int> is(dim);
    const std::vector<double>& v = aArray.vector();
    std::string s;

    if(aDENSE) { 

      if(begin==end) {
        // V(end) = <value>

        int index = end;
        aArray.indices(index,is);
        a_out << " " << aName;
        a_out << "("; 
        for(unsigned iaxis=0;iaxis<dim;iaxis++) { 
          Lib::smanip::printf(s,32,"%d",is[iaxis]+1);
          Lib::smanip::justify(s,aDigits[iaxis],Lib::smanip::right);
          if(iaxis) a_out << ",";
          a_out << s;
        }
        a_out << ") = " << v[index] << std::endl;


      } else {
        // V(begin)..(end) = <value>

        a_out << " " << aName;
       {int index = begin;
        aArray.indices(index,is);
        a_out << "("; 
        for(unsigned iaxis=0;iaxis<dim;iaxis++) { 
          Lib::smanip::printf(s,32,"%d",is[iaxis]+1);
          Lib::smanip::justify(s,aDigits[iaxis],Lib::smanip::right);
          if(iaxis) a_out << ",";
          a_out << s;
        }
        a_out << ")..";}

       {int index = end;
        aArray.indices(index,is);
        a_out << "("; 
        for(unsigned iaxis=0;iaxis<dim;iaxis++) { 
          Lib::smanip::printf(s,32,"%d",is[iaxis]+1);
          Lib::smanip::justify(s,aDigits[iaxis],Lib::smanip::right);
          if(iaxis) a_out << ",";
          a_out << s;
        }
        a_out << ") = " << v[index] << std::endl;}

      }

    } else {
      // V(begin) = ...
      // V(  end) = <value>

      if(end!=begin) {
        int index = begin;
        a_out << " " << aName << "("; 
        aArray.indices(index,is);
        for(unsigned iaxis=0;iaxis<dim;iaxis++) { 
          Lib::smanip::printf(s,32,"%d",is[iaxis]+1);
          Lib::smanip::justify(s,aDigits[iaxis],Lib::smanip::right);
          if(iaxis) a_out << ",";
          a_out << s;
        }
        a_out << ") = " << "..." << std::endl;
      }

     {int index = end;
      a_out << " " << aName << "("; 
      aArray.indices(index,is);
      for(unsigned iaxis=0;iaxis<dim;iaxis++) { 
        Lib::smanip::printf(s,32,"%d",is[iaxis]+1);
        Lib::smanip::justify(s,aDigits[iaxis],Lib::smanip::right);
        if(iaxis) a_out << ",";
        a_out << s;
      }
      a_out << ") = " << v[index] << std::endl;}

    }

  }
};
}

int vector_print_C(void* aPAW_Manager){
  OpenPAW::PAW_Manager* pawManager = (OpenPAW::PAW_Manager*)aPAW_Manager;

  Slash::Core::ISession& session = pawManager->session();

  std::ostream& out = session.cout();

  if(session.verboseLevel()) {
    out << "vector_print_C : execute." << std::endl;
  }

  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return -1;

  std::string VNAME = ku_gets();
  int DENSE = ku_geti();

  std::string NAME;
  std::vector< std::pair<int,int> > dims;
  if(!OpenPAW::PAW_Manager::parseName
        (VNAME,NAME,dims,session.printer())) return -1;
  amanip::touppercase(NAME);
  Slash::Data::IArray* vector = find_vector(*memoryTree,NAME);
  if(!vector) {
    out << "vector_print_C :" 
        << " can't find vector " << amanip::sout(NAME) 
        << " in tree " << amanip::sout(memoryTree->storeName()) 
        << std::endl;
    return -1;
  }
  std::vector< std::pair<unsigned int,unsigned int> > cut;
  if(!vector_get_cut(*vector,dims,cut)) {
    out << "vector_print_C :" 
        << " mismatch dimension in " << amanip::sout(VNAME) 
        << std::endl;
    return -1;
  }

  unsigned int dim = vector->dimension();
  std::vector<unsigned int> is(dim);
  const std::vector<double>& v = vector->vector();
  unsigned int vsize = vector->size();

  // First pass to get number of digits for indices :
  std::vector<int> digits;
  digits.resize(dim,1);
  for(unsigned int index = 0;index<vsize;index++) {
    if(vector->accept(index,cut)) {
      vector->indices(index,is);
      for(unsigned iaxis=0;iaxis<dim;iaxis++) { 
        int i = is[iaxis]+1;
        int ndigit = 1+(int)::log10((double)i);
        digits[iaxis] = MAXIMUM(digits[iaxis],ndigit);
      }
    }
  }

  if(DENSE!=0) {
    OpenPAW::PrintRange range;
    bool first = true;
    for(unsigned int index = 0;index<vsize;index++) {
      if(vector->accept(index,cut)) {
        if(first) {
          range.begin = index;
          range.end = index;
          range.value = v[index];
          first = false;
        } else {
          if(v[index]==range.value) {
            range.end = index;
            continue;
          } else { //Value changed :
            range.dump(out,*vector,NAME,digits,DENSE==2?true:false);
            range.begin = index;                    
            range.end = index;
            range.value = v[index];
          }
        }
      }
    }
    if(!first) range.dump(out,*vector,NAME,digits,DENSE==2?true:false);

  } else {

    std::string s;
    for(unsigned int index = 0;index<vsize;index++) {
      if(vector->accept(index,cut)) {
        // V(indices) = <value>
        out << " " << NAME << "("; 
        vector->indices(index,is);
        for(unsigned iaxis=0;iaxis<dim;iaxis++) { 
          Lib::smanip::printf(s,32,"%d",is[iaxis]+1);
          Lib::smanip::justify(s,digits[iaxis],Lib::smanip::right);
          if(iaxis) out << ",";
          out << s;
        }
        out << ") = " << v[index] << std::endl;
      }
    }

  }

  return -1;
}

int vector_list_C(void* aPAW_Manager){
  OpenPAW::PAW_Manager* pawManager = (OpenPAW::PAW_Manager*)aPAW_Manager;

  Slash::Core::ISession& session = pawManager->session();

  std::ostream& out = session.cout();

  if(session.verboseLevel()) {
    out << "vector_list_C : execute." << std::endl;
  }

  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return -1;

  std::vector<std::string> objs = memoryTree->listObjectNames();
  std::vector<std::string> typs = memoryTree->listObjectTypes();
  if(typs.size()!=objs.size()) {
    out << "vector_delete_C :" 
        << " problem retreiving directory infos."
        << std::endl;
    return -1;
  }
  unsigned int objn = objs.size();
  std::string nameh = " Vector Name                          ";
  out << " " << nameh << "Type Length    Dim-1    Dim-2    Dim-3" << std::endl;
  int vecn = 0;
  for(unsigned int index = 0;index<objn;index++) {
    std::string& name = objs[index];
    std::string& type = typs[index];
    //printf("debug : %s %s\n",name.c_str(),type.c_str());
    if(type=="Slash::Data::IArray") {
      Slash::Data::IArray* vector = find_vector(*memoryTree,name);
      if(vector) {

        std::string s;
        out << " ";

        s = name;
        Lib::smanip::justify(s,nameh.size(),Lib::smanip::left);
        out << s;

        s = "R";
        Lib::smanip::justify(s,4,Lib::smanip::right);
        out << s;
        out << " ";

        Lib::smanip::printf(s,32,"%d",vector->size());
        Lib::smanip::justify(s,6,Lib::smanip::right);
        out << s;
        out << "    ";

        if(vector->dimension()>=1) {
          Lib::smanip::printf(s,32,"%d",vector->orders()[0]);
          Lib::smanip::justify(s,5,Lib::smanip::right);
          out << s;
          out << "    ";
        }
        if(vector->dimension()>=2) {
          Lib::smanip::printf(s,32,"%d",vector->orders()[1]);
          Lib::smanip::justify(s,5,Lib::smanip::right);
          out << s;
          out << "    ";
        }
        if(vector->dimension()>=3) {
          Lib::smanip::printf(s,32,"%d",vector->orders()[2]);
          Lib::smanip::justify(s,5,Lib::smanip::right);
          out << s;
          out << "    ";
        }
        out << std::endl;
        vecn++;
      }
    }
  }
  out << " Total of " << vecn << " Vector(s)" << std::endl;
 
  return -1;
}

int vector_find_C(const char* aName,void* aPAW_Manager){
  // The return value should match KUIP/ku_vtype comment :
  // return vector type:
  //  0 : vector name not defined
  //  1 : real vector
  //  2 : integer vector
  //
  OpenPAW::PAW_Manager* pawManager = (OpenPAW::PAW_Manager*)aPAW_Manager;
  Slash::Core::ISession& session = pawManager->session();

  Lib::Parameters<std::string,Lib::Value>& variables = 
    pawManager->variables();

  std::string name = aName;
  if(name=="LOOP") return 0; //Failure

  if(variables.isKey(name)) {
    return 1; //Success
  } else { //Search also in vectors.

    // We pass here when doing :
    //    VEC/CRE v(2) R 0. 0.
    //    x = v(2)
    AIDA::ITree* memoryTree = pawManager->mem();
    if(!memoryTree) return 0;
    amanip::touppercase(name);
    Slash::Data::IArray* vector = find_vector(*memoryTree,name);
    if(!vector) {
      // If uncommenting the below, Riccardo_Fabbri/TestOpenPAW.kumac 
      // issues messages...
      //std::ostream& out = session.cout();
      //out << "vector_find_C :"
      //    << " " << amanip::sout(name) << " not found."
      //    << std::endl;
      return 0; //Failure
    }
    return 1; //Success return real type.

  }
}

int vector_value_C( 
 const char* aName
,double* aValue
,void* aPAW_Manager
){
  OpenPAW::PAW_Manager* pawManager = (OpenPAW::PAW_Manager*)aPAW_Manager;

  Slash::Core::ISession& session = pawManager->session();

  // Look if aName is of the form <name>(<index>)
  std::string name;
  int aindex = 0;
  if(!Lib::smanip::parseArray(aName,name,aindex)) {
    std::ostream& out = session.cout();
    out << "vector_value_C :"
        << " syntax error in " << amanip::sout(std::string(aName))
        << std::endl;
    *aValue = 0;
    return 1; //Failure
  }

  Lib::Parameters<std::string,Lib::Value>& variables = 
    pawManager->variables();

  if(variables.isKey(name)) {
    Lib::Value var = variables.value(name);
    if(var.type()==Slash::Core::IValue::ARRAY_DOUBLE) {
      if(aindex==NotFound) {
        std::ostream& out = session.cout();
        out << "vector_value_C :"
            << " vector variable without an index in " 
            << amanip::sout(std::string(aName))
            << std::endl;
        *aValue = 0;
        return 1; //Failure
      } 

      aindex--; //CERN/PAW = FORTRAN !
      std::vector<unsigned int> is;
      const std::vector<double>& stdv = var.get_array_double(is);
      if(aindex>=(int)stdv.size()) {
        std::ostream& out = session.cout();
        out << "vector_value_C :"
            << " bad index in " << amanip::sout(std::string(aName))
            << ". Size of vector is " << (unsigned int)stdv.size()
            << std::endl;
        *aValue = 0;
        return 1; //Failure
      }
      *aValue = stdv[aindex];
      return 0; //Success

    } 

    double d;
    if(!Lib::valuemanip::to_double(var,d)) { 
      std::ostream& out = session.cout();
      out << "vector_value_C :"
          << " can't convert Lib::Value to a double in " 
          << amanip::sout(std::string(aName))
          << std::endl;
      *aValue = 0;
      return 1; //Failure
    }
    *aValue = d;
    return 0; //Success
  } 

  //Search also in vectors.
  // We pass here when doing :
  //    VEC/CRE v(2) R 0. 0.
  //    x = v(2)
  AIDA::ITree* memoryTree = pawManager->mem();
  if(!memoryTree) return 0;
  amanip::touppercase(name);
  Slash::Data::IArray* vector = find_vector(*memoryTree,name);
  if(!vector) {
    std::ostream& out = session.cout();
    out << "vector_value_C :"
        << " " << amanip::sout(name) << " not found."
        << std::endl;
    *aValue = 0;
    return 1; //Failure
  }
  if(aindex==NotFound) {
    std::ostream& out = session.cout();
    out << "vector_value_C :"
        << " vector variable without an index in " 
        << amanip::sout(std::string(aName))
        << std::endl;
    *aValue = 0;
    return 1; //Failure
  } 

  //FIXME : handle more dimensions.
  if(vector->dimension()!=1) {
    std::ostream& out = session.cout();
    out << "vector_value_C :"
        << " vector dimension > 1 not treated yet in " 
        << amanip::sout(std::string(aName))
        << std::endl;
    *aValue = 0;
    return 1; //Failure
  }
  aindex--; //CERN/PAW = FORTRAN !
  if(aindex>=(int)vector->size()) {
    std::ostream& out = session.cout();
    out << "vector_value_C :"
        << " bad index in " << amanip::sout(std::string(aName))
        << ". Size of vector is " << vector->size()
        << std::endl;
    *aValue = 0;
    return 1; //Failure
  }
  std::vector<unsigned int> is;
  is.push_back(aindex);
  double v;
  if(!vector->value(is,v)) {
    std::ostream& out = session.cout();
    out << "vector_value_C :"
        << " can't get value for IArray in " 
        << amanip::sout(std::string(aName))
        << std::endl;
    *aValue = 0;
    return 1; //Failure
  }
  *aValue = v;
  return 0; //Success
}
