//  September/2009 : G.Barrand : creation.

#ifndef SoGLUT_Viewer_h
#define SoGLUT_Viewer_h 

#include <Inventor/GLUT/SoGLUT.h>

#include <Inventor/SoType.h>

class SbVec2s;
class SbVec3f;
class SbColor;
class SbViewportRegion;

class SoSceneManager;
class SoNode;
class SoGLRenderAction;
class SoNodeSensor;
class SoSensor;
class SoEvent;
class SoGroup;
class SoCamera;
class SoDirectionalLight;

typedef void(*SoGLUT_UnlockCB)(void*);

class SoGLUT_Viewer {
public:
  SoGLUT_Viewer(SoGLUT_Widget* = 0);
  virtual ~SoGLUT_Viewer();
  
  virtual SbBool setSceneGraph(SoNode*);
  virtual SoNode* getSceneGraph() const;
  virtual void setViewportRegion(const SbViewportRegion&);
  virtual const SbViewportRegion& getViewportRegion() const;
  virtual void setBackgroundColor(const SbColor&);
  virtual const SbColor& getBackgroundColor() const;
  virtual void render();
  virtual void setSize(const SbVec2s&);
  virtual void viewAll();
  virtual void setViewing(SbBool);
  virtual void show();
  virtual void setGLRenderAction(SoGLRenderAction*);
  virtual SoGLRenderAction* getGLRenderAction() const;
  virtual void setTitle(const char*);
  virtual void setCamera(SoCamera*);
  virtual SoCamera* getCamera() const;
  virtual void setCameraType(SoType);
  virtual SoType getCameraType(void) const;
  virtual void setHeadlight(SbBool);
  virtual void setAutoClipping(SbBool);
  virtual void setDecoration(SbBool);
  
  virtual SbBool processKeyboard(unsigned char,SbBool& render);
  virtual SbBool processSpecial(int,SbBool& render);
  virtual SbBool processMouse(int,int,int,int,SbBool& render);
  virtual SbBool processMotion(int,int,SbBool& render);
  virtual SbBool processJoystick(unsigned int,int,int,int,SbBool& render);

  virtual void lock(const SbString&,SoGLUT_UnlockCB = 0,void* = 0); 

  void rotateCameraPositionAroundX(float);
  void rotateCameraPositionAroundY(float);
  void rotateCameraAroundUp(float);
  void translateCameraAlongDir(float);
  void translateCameraAlongSide(float);
  void translateCameraAlongUp(float);

  float getDefaultDeltaRotation() const;
  float getDefaultDeltaTranslation() const;
  void setDefaultDeltaRotation(float angle_radian);
  void setDefaultDeltaTranslation(float);

  float getFloor() const;
  void setFloor(float);
  void checkFloor(SbBool);

  SbBool emulateLeftButtonDown(SbBool&);
  SbBool emulateLeftButtonUp(SbBool&);
  //SbBool emulateMouseMotion(SbBool&);

  enum Move {
    ROTATE_RIGHT,
    TRANSLATE_RIGHT,
    ROTATE_LEFT,
    TRANSLATE_LEFT,
    UP,
    DOWN,
    FORWARD,
    BACKWARD
  };
  void updateMotion(Move);
  void startMoveTimer();
  void stopMoveTimer();
  void setMoveTimerInterval(unsigned int interval_msec);

protected:
  virtual SbBool processSpecialLock(int,SbBool& render);
  virtual SbBool processKeyboardLock(unsigned char,SbBool& render);
private:
  //static Uint32 moveTimerCB(Uint32,void*);
private:
  void setCameraPosition(const SbVec3f&);
private:
  SoSceneManager* fSceneManager;
  SoGroup* fSceneGraph;
  SoNode* fUserSceneGraph;
  SoCamera* fCamera;
  SoDirectionalLight* fHeadLight;
  SoType fCameraType;
  float fDefaultDeltaRotation;
  float fDefaultDeltaTranslation;  
  SbBool fCheckFloor;
  float fFloor;
private:
  SbBool fJoy_rotate;
  SbBool fJoy_forback;
  //Move fMove_what;
  //Uint32 fMoveTimerInterval;
  //GLUT_TimerID fMoveTimerID;
  // lock system :
  SbString fPass;
  SbString fRead;
  SoGLUT_UnlockCB fUnlockCB;
  void* fUnlockCB_Data;
};

#endif
