
//
//  This program is and should be standalone so that :
//     UNIX> c++ obuild_subf.cxx 
//      DOS> cl.exe obuild_subf.cxx 
// builds at first shoot.
//

#include <string>
#include <stdlib.h>

#include <vector>

// From Lib. Some code had been changed compared to the 
// Lib version ; see the FIXME.

static bool Lib_System_getenv(const std::string&,std::string&);

static bool Lib_smanip_replace(std::string&,
                               const std::string&,
                               const std::string&);
static bool Lib_smanip_printf(std::string&,int,const char*,...);

static bool Lib_fmanip_read(const std::string&,std::vector<std::string>&);
static bool Lib_fmanip_replace(const std::string&,
                               const std::string&,
                               const std::string&,
                               std::vector<std::string>&);
static bool Lib_fmanip_tmpname(std::string&,
                               const std::string& = "",
                               const std::string& = "",
                               const std::string& = "");
static bool Lib_fmanip_remove(const std::string&);
static bool Lib_fmanip_rename(const std::string&,const std::string&);

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
int main(
 int aArgc
,char** aArgv
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aArgc!=4) {
    fprintf(stderr,"obuild_subf : three arguments expected.\n");
    return EXIT_FAILURE;
  }

  std::string f(aArgv[1]);
  std::string o(aArgv[2]);
  std::string n(aArgv[3]);

  std::vector<std::string> changes;
  if(!Lib_fmanip_replace(f,o,n,changes)) {
    fprintf(stderr,"obuild_subf : failed\n");
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool Lib_System_getenv(
 const std::string& aString
,std::string& aValue
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  const char* env = ::getenv(aString.c_str());
  if(env) {
    aValue = std::string(env?env:"");
    return true;
  } else {
    aValue = "";
    return false;
  }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool Lib_smanip_replace(
 std::string& aString
,const std::string& aOld     
,const std::string& aNew     
)
//////////////////////////////////////////////////////////////////////////////
// return true : some replacement done.
// return false : nothing replaced.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(aOld=="") return false;
  std::string snew;
  std::string::size_type lold = aOld.length();
  bool status = false;
  std::string stmp = aString;
  while(true) {
    std::string::size_type pos = stmp.find(aOld);
    if(pos==std::string::npos){
      snew += stmp;
      break;
    } else {
      snew += stmp.substr(0,pos);
      snew += aNew;
      stmp = stmp.substr(pos+lold,stmp.length()-(pos+lold));
      status = true;
    }
  }
  aString = snew;
  return status;
}
#include <stdarg.h>
//////////////////////////////////////////////////////////////////////////////
bool Lib_smanip_printf(
 std::string& aString
,int aLength
,const char* aFormat
,...
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aString = "";
  if(aLength<0) return false;
  if(!aFormat) return false;
  char* s = new char[aLength+1];
  if(!s) return false;
  s[aLength] = '\0';
  va_list args;
  va_start (args,aFormat);
  vsprintf (s,aFormat,args);
  va_end   (args);
  if(s[aLength]!='\0') {
    delete [] s;
    return false;
  }
  aString = s;
  delete [] s;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#include <string.h> //strlen
#define BUFSIZE 65536
//////////////////////////////////////////////////////////////////////////////
bool Lib_fmanip_read(
 const std::string& aName
,std::vector<std::string>& aText
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aText.clear();
  //FIXME std::string name = fmanip::name(aName);
  std::string name = aName;
  FILE* file = ::fopen(name.c_str(),"rb");
  if(!file) return false;
  char buffer[BUFSIZE+1];
  while(1) {
    if(::fgets(buffer,BUFSIZE,(FILE*)file)==NULL) break; // EOF.
    int l = ::strlen(buffer);
    //  On Windows, editors when saving binary files,
    // put \r\n at place of \n ; we then look for \r\n.
    if( (l>=2) && (buffer[l-2]=='\r') && (buffer[l-1]=='\n') ) {
      buffer[l-2] = '\0';
    } else if( (l>=1) && (buffer[l-1]=='\n') ) {
      buffer[l-1] = '\0';
    }
    aText.push_back(buffer);
  }
  ::fclose(file);
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib_fmanip_replace(
 const std::string& aName
,const std::string& aOld
,const std::string& aNew
,std::vector<std::string>& aChanges
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aChanges.clear();
  std::vector<std::string> text;
  if(!Lib_fmanip_read(aName,text)) return false;
  // Replace :
  bool touched = false;
 {unsigned int textn = text.size();
  for(unsigned index=0;index<textn;index++) {
    std::string old = text[index];
    if(Lib_smanip_replace(text[index],aOld,aNew)) {
      //printf("debug :\"%s\"\n",text[index].c_str());
      aChanges.push_back(old);
      aChanges.push_back(text[index]);
      touched = true;
    }
  }}
  if(!touched) return true;
  // Create temporary file :
  std::string tmpName;
  if(!Lib_fmanip_tmpname(tmpName,".","tmp")) return false; //FIXME
  FILE* file = ::fopen(tmpName.c_str(),"wb");
  if(!file) return false;
 {unsigned int textn = text.size();
  for(unsigned index=0;index<textn;index++) {
    if(::fprintf(file,"%s\n",text[index].c_str())<0) {
      ::fclose(file);
      Lib_fmanip_remove(tmpName);
      return false;
    }
  }}
  ::fclose(file);
  if(!Lib_fmanip_remove(aName)) return false; //FIXME
  if(!Lib_fmanip_rename(tmpName,aName)) return false;
  return true;
}
#include <sys/stat.h>
//////////////////////////////////////////////////////////////////////////////
bool Lib_fmanip_tmpname(
 std::string& aName
,const std::string& aDir
,const std::string& aPrefix
,const std::string& aSuffix
)
//////////////////////////////////////////////////////////////////////////////
//  For UNIX aDir must be a string ended with '/'. Exa : /tmp/ .
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  for(int count=0;count<1000;count++) {
    std::string name;
    Lib_smanip_printf(name,3,"%03d",count);
    std::string tmpnm;
    if(aDir==".") {
      tmpnm += aPrefix;
      tmpnm += name;
      tmpnm += aSuffix;
    } else if(aDir!="") {
      tmpnm += aDir;
      tmpnm += aPrefix;
      tmpnm += name;
      tmpnm += aSuffix;
    } else {
#ifdef WIN32
      std::string env;
      Lib_System_getenv("HOMEDRIVE",env);
      std::string drive = (env!="" ? env : std::string("C:"));
      tmpnm += drive;
      tmpnm += "\\";
      tmpnm += aPrefix;
      tmpnm += name;
      tmpnm += aSuffix;
#else
      tmpnm += "/tmp/";
      tmpnm += aPrefix;
      tmpnm += name;
      tmpnm += aSuffix;
#endif
    }
    struct stat finfo;
    if(::stat(tmpnm.c_str(),&finfo) < 0) {
      // No directory or file found :
      aName = tmpnm; // Good file name.
      return true; 
    }
  }
  // No good name found :
  aName = "";
  return false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib_fmanip_remove(
 const std::string& aName
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return (::remove(aName.c_str()) ==0 ? true : false);
}
//////////////////////////////////////////////////////////////////////////////
bool Lib_fmanip_rename(
 const std::string& aName
,const std::string& aNewName
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return (::rename(aName.c_str(),aNewName.c_str()) == 0 ? true : false);
}
