//  September/2009 : G.Barrand : creation.

// this :
#include <Inventor/GLUT/viewers/SoGLUT_Viewer.h>

#include <Inventor/SbRotation.h>
#include <Inventor/SoSceneManager.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodekits/SoBaseKit.h>
#include <Inventor/sensors/SoNodeSensor.h>
#include <Inventor/actions/SoHandleEventAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/events/SoMouseButtonEvent.h>
#include <Inventor/events/SoLocation2Event.h>

#include <Inventor/GLUT/SbGL.h>

SoGLUT_Viewer::SoGLUT_Viewer(SoGLUT_Widget* aWidget)
:fSceneManager(0)
,fSceneGraph(0)
,fUserSceneGraph(0)
,fCamera(0)
,fHeadLight(0)
,fDefaultDeltaRotation(1.0F * 3.1416F/180.0F) //1 degree
,fDefaultDeltaTranslation(1) 
,fCheckFloor(FALSE)
,fFloor(0)
,fJoy_rotate(true)
,fJoy_forback(true)
//,fMove_what(ROTATE_RIGHT)
//NOTE : we must arrange to have key repeat 
//       at the same rythm than the gamepad.
//,fMoveTimerInterval(30) //msec
//,fMoveTimerID(0)
,fUnlockCB(0)
,fUnlockCB_Data(0)
{
  fSceneManager = new SoSceneManager;

  aWidget->setViewer(this); //Used in the SoGLUT::mainLoop.

  fCameraType = SoOrthographicCamera::getClassTypeId();

  fSceneGraph = new SoGroup;
  fSceneGraph->ref();

  //0
  SoLightModel* lightModel = new SoLightModel;
  lightModel->model.setValue(SoLightModel::PHONG);
  //lightModel->model.setValue(SoLightModel::BASE_COLOR);
  fSceneGraph->addChild(lightModel);

  //1
  fHeadLight = new SoDirectionalLight;
  fHeadLight->ref();
  fHeadLight->color.setValue(1,1,1);
  fHeadLight->direction.setValue(SbVec3f(1,-1,-10));
  fSceneGraph->addChild(fHeadLight);

  //2
  SoOrthographicCamera* camera = new SoOrthographicCamera();
  fSceneGraph->addChild(camera);
  setCamera(camera);

#define USER_SCENE_INDEX 3

  fSceneManager->setWindowSize(SbVec2s(500,500));

  fSceneManager->setBackgroundColor(SbColor(0,0,0));

  fSceneManager->setSceneGraph(fSceneGraph);
}

SoGLUT_Viewer::~SoGLUT_Viewer(){
  /*
  if(fMoveTimerID) {
    GLUT_RemoveTimer(fMoveTimerID);
  }
  */
  if(fUserSceneGraph) {
    //fSceneGraph->removeChild(fUserSceneGraph);
    fUserSceneGraph->unref();
  }
  if(fCamera) fCamera->unref();
  if(fHeadLight) fHeadLight->unref();
  fSceneGraph->removeAllChildren();
  fSceneGraph->unref();
  delete fSceneManager;
}

SbBool SoGLUT_Viewer::setSceneGraph(SoNode* aScene){
  if(aScene==fUserSceneGraph) return TRUE; //Done.

  if(fUserSceneGraph) {
    //fSceneGraph->removeChild(fUserSceneGraph);
    fUserSceneGraph->unref();
  }
  fUserSceneGraph = aScene;
  if(!fUserSceneGraph) return TRUE;

  fUserSceneGraph->ref();

  int number = fSceneGraph->getNumChildren();
  if(number==USER_SCENE_INDEX) {
    fSceneGraph->addChild(fUserSceneGraph);
  } else if(number==(USER_SCENE_INDEX+1)) {
    fSceneGraph->replaceChild(USER_SCENE_INDEX,fUserSceneGraph);
  } else {
    //error.
  }

  // Search for a camera in the user-supplied scenegraph.
  SbBool oldsearch = SoBaseKit::isSearchingChildren();
  SoBaseKit::setSearchingChildren(TRUE);
  SoSearchAction* searchAction = new SoSearchAction;
  searchAction->reset();
  searchAction->setType(SoCamera::getClassTypeId());
  searchAction->apply(fUserSceneGraph);
  SoBaseKit::setSearchingChildren(oldsearch);

  SoCamera* sceneCamera = 0;
  if(searchAction->getPath()) {
    SoFullPath* fullpath = (SoFullPath*)searchAction->getPath();
    SoNode* node = fullpath->getTail();
    if(node->isOfType(SoCamera::getClassTypeId())==FALSE) return FALSE;
    sceneCamera = (SoCamera*)node;
  }
  delete searchAction;

  // Make our own camera if none was available.
  /*
  if (!sceneCamera) {
    sceneCamera = (SoCamera*)fCameraType.createInstance();
    //PRIVATE(this)->deletecamera = TRUE;
  }
  */

  if(sceneCamera) setCamera(sceneCamera);

  //FIXME : Should notify the fSceneManager.
  return TRUE;
}

SoNode* SoGLUT_Viewer::getSceneGraph() const {return fUserSceneGraph;}

void SoGLUT_Viewer::setCamera(SoCamera* aCamera){
  if(aCamera==fCamera) return; //Done.
  if(fCamera) fCamera->unref();
  fCamera = aCamera;
  if(fCamera) fCamera->ref();
}

SoCamera* SoGLUT_Viewer::getCamera() const {return fCamera;}

void SoGLUT_Viewer::setCameraType(SoType aType){fCameraType = aType;}
SoType SoGLUT_Viewer::getCameraType() const {return fCameraType;}

void SoGLUT_Viewer::setViewportRegion(const SbViewportRegion& aNewRegion) { 
  fSceneManager->setViewportRegion(aNewRegion); 
}
const SbViewportRegion& SoGLUT_Viewer::getViewportRegion() const { 
  return fSceneManager->getViewportRegion(); 
}

void SoGLUT_Viewer::setBackgroundColor(const SbColor& aC) { 
  fSceneManager->setBackgroundColor(aC); 
}
const SbColor& SoGLUT_Viewer::getBackgroundColor() const { 
  return fSceneManager->getBackgroundColor(); 
}

void SoGLUT_Viewer::setSize(const SbVec2s& a_size){
  fSceneManager->setWindowSize(a_size);
}

void SoGLUT_Viewer::viewAll(){   
  if(!fCamera) return;
  // No shape under fSceneGraph, take directly fUserSceneGraph.
  if(!fUserSceneGraph) return;
  fCamera->viewAll(fUserSceneGraph,getViewportRegion());
}

void SoGLUT_Viewer::setViewing(SbBool){}

void SoGLUT_Viewer::show(){   
  //SoGLUT::sendRender();
}

void SoGLUT_Viewer::setGLRenderAction(SoGLRenderAction*){}

SoGLRenderAction* SoGLUT_Viewer::getGLRenderAction() const {   
  return fSceneManager->getGLRenderAction();
}

void SoGLUT_Viewer::setTitle(const char*){}

void SoGLUT_Viewer::setHeadlight(SbBool aOnOff) {   
  fHeadLight->on.setValue(aOnOff);
}

void SoGLUT_Viewer::render(){   
  if(!fSceneGraph) return;
  fSceneManager->render();
}

void SoGLUT_Viewer::setAutoClipping(SbBool){}
void SoGLUT_Viewer::setDecoration(SbBool){}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

SbBool SoGLUT_Viewer::processKeyboard(unsigned char a_char,SbBool& aRender){

#ifdef GLUT_DEBUG_EVENT
  printf("debug : processKey : which %d\n",a_char);        
#endif

  if(a_char=='d'){
    if(glutGetModifiers()&GLUT_ACTIVE_SHIFT==GLUT_ACTIVE_SHIFT) {
      updateMotion(SoGLUT_Viewer::TRANSLATE_RIGHT);
    } else {
      updateMotion(SoGLUT_Viewer::ROTATE_RIGHT);
    }
    aRender = TRUE;
    return TRUE;
  } else if(a_char=='a'){
    if(glutGetModifiers()&GLUT_ACTIVE_SHIFT==GLUT_ACTIVE_SHIFT) {
      updateMotion(SoGLUT_Viewer::TRANSLATE_LEFT);
    } else {
      updateMotion(SoGLUT_Viewer::ROTATE_LEFT);
    }
    aRender = TRUE;
    return TRUE;
  } else if(a_char=='w'){
    if(glutGetModifiers()&GLUT_ACTIVE_SHIFT==GLUT_ACTIVE_SHIFT) {
      updateMotion(SoGLUT_Viewer::UP);
    } else {
      updateMotion(SoGLUT_Viewer::FORWARD);
    }
    aRender = TRUE;
    return TRUE;
  } else if(a_char=='s'){
    if(glutGetModifiers()&GLUT_ACTIVE_SHIFT==GLUT_ACTIVE_SHIFT) {
      updateMotion(SoGLUT_Viewer::DOWN);
    } else {
      updateMotion(SoGLUT_Viewer::BACKWARD);
    }
    aRender = TRUE;
    return TRUE;
  }

  aRender = FALSE;
  return FALSE;
}

SbBool SoGLUT_Viewer::processSpecial(int aKey,SbBool& aRender){

#ifdef GLUT_DEBUG_EVENT
  printf("debug : processSpecial : which %d\n",aKey);        
#endif

  if(aKey==GLUT_KEY_RIGHT){
    if(glutGetModifiers()&GLUT_ACTIVE_SHIFT==GLUT_ACTIVE_SHIFT) {
      updateMotion(SoGLUT_Viewer::TRANSLATE_RIGHT);
    } else {
      updateMotion(SoGLUT_Viewer::ROTATE_RIGHT);
    }
    aRender = TRUE;
    return TRUE;
  } else if(aKey==GLUT_KEY_LEFT){
    if(glutGetModifiers()&GLUT_ACTIVE_SHIFT==GLUT_ACTIVE_SHIFT) {
      updateMotion(SoGLUT_Viewer::TRANSLATE_LEFT);
    } else {
      updateMotion(SoGLUT_Viewer::ROTATE_LEFT);
    }
    aRender = TRUE;
    return TRUE;
  } else if(aKey==GLUT_KEY_UP){
    if(glutGetModifiers()&GLUT_ACTIVE_SHIFT==GLUT_ACTIVE_SHIFT) {
      updateMotion(SoGLUT_Viewer::UP);
    } else {
      updateMotion(SoGLUT_Viewer::FORWARD);
    }
    aRender = TRUE;
    return TRUE;
  } else if(aKey==GLUT_KEY_DOWN){
    if(glutGetModifiers()&GLUT_ACTIVE_SHIFT==GLUT_ACTIVE_SHIFT) {
      updateMotion(SoGLUT_Viewer::DOWN);
    } else {
      updateMotion(SoGLUT_Viewer::BACKWARD);
    }
    aRender = TRUE;
    return TRUE;
  }

  aRender = FALSE;
  return FALSE;
}


SbBool SoGLUT_Viewer::processMouse(int a_button,int a_state,int a_x,int a_y,SbBool& aRender){

  // Build SoEvent.
  SoMouseButtonEvent event;
    
  // WARNING : for Coin, BUTTON2 is right button, BUTTON3 is middle button.
  switch(a_button){
  case GLUT_LEFT_BUTTON:
    event.setButton(SoMouseButtonEvent::BUTTON1);
    break;
  case GLUT_RIGHT_BUTTON:
    event.setButton(SoMouseButtonEvent::BUTTON2);
    break;
  case GLUT_MIDDLE_BUTTON:
    event.setButton(SoMouseButtonEvent::BUTTON3);
    break;
  default:
    event.setButton(SoMouseButtonEvent::ANY);
    break;
  }

  if(a_state==GLUT_DOWN)   
    event.setState(SoButtonEvent::DOWN);
  else if(a_state==GLUT_UP) 
    event.setState(SoButtonEvent::UP);
  else
    event.setState(SoButtonEvent::UNKNOWN);
      
  if(glutGetModifiers()&GLUT_ACTIVE_SHIFT==GLUT_ACTIVE_SHIFT) 
    event.setShiftDown(TRUE);
  if(glutGetModifiers()&GLUT_ACTIVE_CTRL==GLUT_ACTIVE_CTRL) 
    event.setCtrlDown(TRUE);
  if(glutGetModifiers()&GLUT_ACTIVE_ALT==GLUT_ACTIVE_ALT) 
    event.setAltDown(TRUE);
  
  //GLUT : origin is top-left with y down (like X11).
  //Inventor :  origin is bottom-left with y up (like OpenGL).
  const SbVec2s& win = getViewportRegion().getWindowSize();
  event.setPosition(SbVec2s(short(a_x),win[1] - short(a_y)));
  
  SbBool status = fSceneManager->processEvent(&event);
  aRender = status;
  return status;
}

SbBool SoGLUT_Viewer::processMotion(int a_x,int a_y,SbBool& aRender){
  // Build SoEvent.
  SoLocation2Event event;
    
  //GLUT : origin is top-left with y down (like X11).
  //Inventor :  origin is bottom-left with y up (like OpenGL).
  const SbVec2s& win = getViewportRegion().getWindowSize();
  event.setPosition(SbVec2s(short(a_x),win[1] - short(a_y)));
    
  SbBool status = fSceneManager->processEvent(&event);
  aRender = status;
  return status;
}

SbBool SoGLUT_Viewer::processJoystick(unsigned int a_button,int a_x,int a_y,int a_z,SbBool& aRender)
{
  /*
  } else if(aEvent.type==GLUT_JOYBUTTONDOWN) {
#ifdef GLUT_DEBUG_EVENT
    printf("debug : GLUT_JOYBUTTON : which %d, button %d, state %d\n",aEvent.jbutton.which,aEvent.jbutton.button,aEvent.jbutton.state);        
#endif

     // on a Logitech Precision the button num = jbutton.button+1

     if(aEvent.jbutton.button==0) { 
       fJoy_rotate = fJoy_rotate?false:true;
       aRender = FALSE;
       return TRUE;
     }
     if(aEvent.jbutton.button==1) { 
       fJoy_forback = fJoy_forback?false:true;
       aRender = FALSE;
       return TRUE;
     }
     if((aEvent.jbutton.button==4)|| //shoot
        (aEvent.jbutton.button==5)) {
       return emulateLeftButtonDown(aRender);
     }

  } else if(aEvent.type==GLUT_JOYBUTTONUP) {

     // on a Logitech Precision the button num = jbutton.button+1

     if((aEvent.jbutton.button==4)|| //shoot
        (aEvent.jbutton.button==5)) {
       return emulateLeftButtonUp(aRender);
     }

  } else if(aEvent.type==GLUT_JOYAXISMOTION) {
#ifdef GLUT_DEBUG_EVENT
    printf("debug : GLUT_JOYAXISMOTION : which %d, axis %d, value %d\n",aEvent.jaxis.which,aEvent.jaxis.axis,aEvent.jaxis.value);        
#endif

    if(aEvent.jaxis.axis==0) {

      if((aEvent.jaxis.value>-8000)&&(aEvent.jaxis.value<8000)) {
        stopMoveTimer();
      } else {
        if(fMoveTimerID) {
          stopMoveTimer();
        } else {
          if(aEvent.jaxis.value>0) {
            fMove_what = (fJoy_rotate?
                            SoGLUT_Viewer::ROTATE_RIGHT:
                            SoGLUT_Viewer::TRANSLATE_RIGHT);
          } else {
            fMove_what = (fJoy_rotate?
                            SoGLUT_Viewer::ROTATE_LEFT:
                            SoGLUT_Viewer::TRANSLATE_LEFT);
          }
          startMoveTimer();
        }            
      }            

      aRender = FALSE;
      return TRUE;

    } else if(aEvent.jaxis.axis==1) {

      if((aEvent.jaxis.value>-8000)&&(aEvent.jaxis.value<8000)) {
        stopMoveTimer();
      } else {
        if(fMoveTimerID) {
          stopMoveTimer();
        } else {
          if(aEvent.jaxis.value>0) {
            fMove_what = (fJoy_forback?
                            SoGLUT_Viewer::BACKWARD:
                            SoGLUT_Viewer::DOWN);
          } else {
            fMove_what = (fJoy_forback?
                            SoGLUT_Viewer::FORWARD:
                            SoGLUT_Viewer::UP);
          }
          startMoveTimer();
        }            
      }            

      aRender = FALSE;
      return TRUE;
    }

  }
  */

  aRender = FALSE;
  return FALSE;
}

void SoGLUT_Viewer::rotateCameraPositionAroundX(float aDeltaAngle){
  // From CoinXt/SoGuiExaminerViewerP::rotXWheelMotion.

  if(!fCamera) return;

  SbVec3f dir;
  fCamera->orientation.getValue().multVec(SbVec3f(0,0,-1),dir);

  SbVec3f focalpoint = 
    fCamera->position.getValue() + fCamera->focalDistance.getValue() * dir;

  fCamera->orientation = 
    SbRotation(SbVec3f(-1,0,0),aDeltaAngle) * fCamera->orientation.getValue();

  fCamera->orientation.getValue().multVec(SbVec3f(0,0,-1),dir);

  SbVec3f pos = focalpoint - fCamera->focalDistance.getValue() * dir;
  setCameraPosition(pos);
}

void SoGLUT_Viewer::rotateCameraPositionAroundY(float aDeltaAngle){
  // From CoinXt/SoGuiExaminerViewerP::rotYWheelMotion.
  if(!fCamera) return;

  SbVec3f dir;
  fCamera->orientation.getValue().multVec(SbVec3f(0,0,-1),dir);

  SbVec3f focalpoint = 
    fCamera->position.getValue() + fCamera->focalDistance.getValue() * dir;

  fCamera->orientation = 
    SbRotation(SbVec3f(0,1,0),aDeltaAngle) * fCamera->orientation.getValue();

  fCamera->orientation.getValue().multVec(SbVec3f(0,0,-1),dir);

  SbVec3f pos = focalpoint - fCamera->focalDistance.getValue() * dir;
  setCameraPosition(pos);
}

void SoGLUT_Viewer::rotateCameraAroundUp(float aDeltaAngle){
  if(!fCamera) return;
  SbVec3f up;
  fCamera->orientation.getValue().multVec(SbVec3f(0,1,0),up);
  fCamera->orientation = 
    SbRotation(up,aDeltaAngle) * fCamera->orientation.getValue();
}

void SoGLUT_Viewer::translateCameraAlongDir(float aDelta){
  if(!fCamera) return;

  SbVec3f dir;
  fCamera->orientation.getValue().multVec(SbVec3f(0,0,-1),dir);

  SbVec3f pos = fCamera->position.getValue() + aDelta * dir;
  setCameraPosition(pos);
}

void SoGLUT_Viewer::translateCameraAlongSide(float aDelta){
  if(!fCamera) return;

  SbVec3f dir;
  fCamera->orientation.getValue().multVec(SbVec3f(0,0,-1),dir);
  SbVec3f up;
  fCamera->orientation.getValue().multVec(SbVec3f(0,1,0),up);

  SbVec3f side = dir.cross(up);

  SbVec3f pos = fCamera->position.getValue() + aDelta * side;
  setCameraPosition(pos);
}

void SoGLUT_Viewer::translateCameraAlongUp(float aDelta){
  if(!fCamera) return;

  SbVec3f dir;
  fCamera->orientation.getValue().multVec(SbVec3f(0,0,-1),dir);
  SbVec3f up;
  fCamera->orientation.getValue().multVec(SbVec3f(0,1,0),up);

  SbVec3f pos = fCamera->position.getValue() + aDelta * up;
  setCameraPosition(pos);
}

float SoGLUT_Viewer::getDefaultDeltaRotation() const {
  return fDefaultDeltaRotation;
}

void SoGLUT_Viewer::setDefaultDeltaRotation(float aValue) {
  fDefaultDeltaRotation = aValue;
}

float SoGLUT_Viewer::getDefaultDeltaTranslation() const {
  return fDefaultDeltaTranslation;
}

void SoGLUT_Viewer::setDefaultDeltaTranslation(float aValue) {
  fDefaultDeltaTranslation = aValue;
}

void SoGLUT_Viewer::setMoveTimerInterval(unsigned int aValue) {
  //fMoveTimerInterval = aValue;
}

void SoGLUT_Viewer::checkFloor(SbBool aValue) {fCheckFloor = aValue;}

float SoGLUT_Viewer::getFloor() const {return fFloor;}

void SoGLUT_Viewer::setFloor(float aValue) {fFloor = aValue;}

void SoGLUT_Viewer::setCameraPosition(const SbVec3f& aPos) {
  if(fCheckFloor==FALSE) {
    fCamera->position.setValue(aPos);
  } else {
    if(aPos[1]>fFloor) fCamera->position.setValue(aPos);
  }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void SoGLUT_Viewer::updateMotion(Move aWhat) {
  float deltaRot = getDefaultDeltaRotation();
  float deltaTrans = getDefaultDeltaTranslation();

  if(aWhat==ROTATE_RIGHT) {
    rotateCameraAroundUp(-deltaRot);
  } else if(aWhat==ROTATE_LEFT) {
    rotateCameraAroundUp(deltaRot);

  } else if(aWhat==TRANSLATE_RIGHT) {
    translateCameraAlongSide(deltaTrans);
  } else if(aWhat==TRANSLATE_LEFT) {
    translateCameraAlongSide(-deltaTrans);

  } else if(aWhat==UP) {
    translateCameraAlongUp(deltaTrans);
  } else if(aWhat==DOWN) {
    translateCameraAlongUp(-deltaTrans);

  } else if(aWhat==FORWARD) {
    translateCameraAlongDir(deltaTrans);
  } else if(aWhat==BACKWARD) {
    translateCameraAlongDir(-deltaTrans);
  }
}

/*
void SoGLUT_Viewer::startMoveTimer(){
  if(fMoveTimerID) GLUT_RemoveTimer(fMoveTimerID);
  fMoveTimerID = GLUT_AddTimer(fMoveTimerInterval,moveTimerCB,this);
}

void SoGLUT_Viewer::stopMoveTimer(){
  if(fMoveTimerID) GLUT_RemoveTimer(fMoveTimerID);
  fMoveTimerID = 0;
}

Uint32 SoGLUT_Viewer::moveTimerCB(Uint32 aInterval,void* aTag){
  SoGLUT_Viewer* This = (SoGLUT_Viewer*)aTag;
  This->updateMotion(This->fMove_what);
  //SoGLUT::setViewerRender();
  SoGLUT::sendRender();
  //return 0; //Stop the timer. (Returning aInterval would continue the timer).
  return aInterval;
}
*/

SbBool SoGLUT_Viewer::emulateLeftButtonDown(SbBool& aRender){
  const SbVec2s& win = getViewportRegion().getWindowSize();
  return processMouse(GLUT_LEFT_BUTTON,GLUT_DOWN,win[0]/2,win[1]/2,aRender);
}
SbBool SoGLUT_Viewer::emulateLeftButtonUp(SbBool& aRender){
  const SbVec2s& win = getViewportRegion().getWindowSize();
  return processMouse(GLUT_LEFT_BUTTON,GLUT_UP,win[0]/2,win[1]/2,aRender);
}

/*
SbBool SoGLUT_Viewer::emulateMouseMotion(SbBool& aRender){
  GLUT_Event event;
  event.type = GLUT_MOUSEMOTION;
  //FIXME event.motion.which = ;
  //FIXME event.motion.state = ;
  const SbVec2s& win = getViewportRegion().getWindowSize();
  event.motion.x = win[0]/2;
  event.motion.y = win[1]/2;    
  //FIXME event.motion.xrel = 0;
  //FIXME event.motion.yrel = 0;
  return processEvent(event,aRender);
}
*/

void SoGLUT_Viewer::lock(const SbString& aPass,SoGLUT_UnlockCB aCB,void* aData){
  fPass = aPass;
  fRead = 0;
  fUnlockCB = aCB;
  fUnlockCB_Data = aData;
}

SbBool SoGLUT_Viewer::processSpecialLock(int,SbBool& aRender){
  aRender = FALSE;
  if(!fPass.getLength()) return FALSE;
  return TRUE;
}

SbBool SoGLUT_Viewer::processKeyboardLock(unsigned char aChar,SbBool& aRender){
  aRender = FALSE;
  if(!fPass.getLength()) return FALSE;

  //printf("debug : processKeyboardLock : %d %c\n",aChar,aChar);        

  if(aChar==13) { //RETURN
    if(fRead==fPass) { //unlock.
      fPass.makeEmpty();
      fRead.makeEmpty();
      if(fUnlockCB) {
        fUnlockCB(fUnlockCB_Data);
        aRender = TRUE;
      }
    } else {
      fRead.makeEmpty();
    }
  } else {
    fRead += aChar;
  }

  return TRUE;
}
