// The scene is a grey floor with a red cube over it.
// At startup, the camera is slightly on top of the floor
// looking at the cube.

#include <Inventor/GLUT/SoGLUT.h>
#include <Inventor/GLUT/viewers/SoGLUT_Viewer.h>

#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoTranslation.h>

#include <math.h>
#ifndef M_PI
#define M_PI       3.1415926535897931160E0
#define M_PI_2     1.5707963267948965580E0  
#endif
#define FM_PI      ((float)M_PI)
#define FM_PI_2    ((float)M_PI_2)
#include <Inventor/nodes/SoPerspectiveCamera.h>

#include <iostream>
#include <cstdlib>
#include <string>

#define MENU_ITEM_QUIT 100

void menu(int a_item){
  if(a_item==MENU_ITEM_QUIT) ::exit(0);
}

#include <sstream>

template <class T>
inline bool to(const std::string& a_s,T& a_v) {
  std::istringstream strm(a_s.c_str());
  std::istringstream::pos_type pb = strm.tellg();
  strm >> a_v;
  std::istringstream::pos_type pe = strm.tellg();
  std::string::size_type sz = pe-pb;
  if(sz!=a_s.size()) {a_v = 0;return false;}
  return true;
}

int main(int argc,char* argv[]) {

  ///////////////////////////////////////////////
  // Init :
  ///////////////////////////////////////////////

  int ww = 500;
  int wh = ww*((float)1050)/((float)1680); //inside window height

  int menubar = 23;
  int caption = 23;

  int wx = 0;
  int wy = menubar + caption;

  bool full_screen = true; //full screen by default.
  if(::getenv("SOGLUT_NOT_FULL_SCREEN")) full_screen = false;

  SoGLUT_Widget* topWidget = 
    SoGLUT::init(argc,argv,argv[0],"SoXt",wx,wy,ww,wh,full_screen);
  if(!topWidget) return 1;

  SoSeparator* sceneGraph = new SoSeparator;
  sceneGraph->ref();

  ///////////////////////////////////////////////
  // Camera (us) :
  ///////////////////////////////////////////////
  // Head light :
 {SoDirectionalLight* soLight = new SoDirectionalLight;
  soLight->color.setValue(1,1,1);
  // Same as Coin<Gui>/So<Gui>Viewer :
  soLight->direction.setValue(SbVec3f(1,-1,-10));
  sceneGraph->addChild(soLight);}

  float d = 100;
  float eye = 7;
  float near = d*0.01F;
  float far = 100*d;
  SoPerspectiveCamera* soCamera = new SoPerspectiveCamera;
  soCamera->viewportMapping.setValue(SoCamera::ADJUST_CAMERA);
  soCamera->position.setValue(0,eye,d); //us at y = eye.
  soCamera->focalDistance.setValue(d);
  soCamera->orientation.setValue(SbRotation(SbVec3f(0,1,0),0));
  soCamera->nearDistance.setValue(near);
  soCamera->farDistance.setValue(far);
  soCamera->heightAngle.setValue(FM_PI/4.0F);
  sceneGraph->addChild(soCamera);

  // Light at ceil :
 {SoPointLight* soLight = new SoPointLight;
  soLight->location.setValue(0,100,0); //y = 100.
  sceneGraph->addChild(soLight);}

  SoLightModel* soLightModel = new SoLightModel;
  soLightModel->model.setValue(SoLightModel::PHONG);
  sceneGraph->addChild(soLightModel);

  ///////////////////////////////////////////////
  // Floor :
  ///////////////////////////////////////////////
 {SoSeparator* floor = new SoSeparator;
  sceneGraph->addChild(floor);
  
  SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue(SbColor(1,1,1));
  floor->addChild(material);

  SoCube* soCube = new SoCube;
  soCube->width.setValue(10*d);
  soCube->height.setValue(0.01F);
  soCube->depth.setValue(10*d);
  floor->addChild(soCube);}

  ///////////////////////////////////////////////
  // Scene :
  ///////////////////////////////////////////////
 {SoSeparator* scene = new SoSeparator;
  sceneGraph->addChild(scene);

 {SoSeparator* sep = new SoSeparator;
  scene->addChild(sep);
  SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue(SbColor(0,1,0));
  sep->addChild(material);
  SoTranslation* trans = new SoTranslation;
  trans->translation.setValue(SbVec3f(-10,10,0));
  sep->addChild(trans);
  SoCube* soCube = new SoCube;
  soCube->width.setValue(10.0F);
  soCube->height.setValue(10.0F);
  soCube->depth.setValue(10.0F);
  sep->addChild(soCube);}

 {SoSeparator* sep = new SoSeparator;
  scene->addChild(sep);
  SoMaterial* material = new SoMaterial;
  material->diffuseColor.setValue(SbColor(1,0,0));
  sep->addChild(material);
  SoTranslation* trans = new SoTranslation;
  trans->translation.setValue(SbVec3f(10,10,0));
  sep->addChild(trans);
  SoCube* soCube = new SoCube;
  soCube->width.setValue(10.0F);
  soCube->height.setValue(10.0F);
  soCube->depth.setValue(10.0F);
  sep->addChild(soCube);}

  } //end scene

  // Viewer :
  SoGLUT_Viewer* viewer = new SoGLUT_Viewer(topWidget);
  viewer->setSceneGraph(sceneGraph);
  viewer->setCamera(soCamera);
  viewer->setHeadlight(FALSE);
  viewer->setDefaultDeltaTranslation(1);
  viewer->show();
    
  viewer->checkFloor(TRUE);

  SoGLUT::show(topWidget);

  //glutCreateMenu(menu);
  //glutAddMenuEntry("Quit",MENU_ITEM_QUIT);
  //glutAttachMenu(GLUT_LEFT_BUTTON);

  SoGLUT::mainLoop();
    
  sceneGraph->unref();

  delete viewer;

  SoGLUT::clear();  
	
  std::cout << "exiting..." << std::endl;

  return 0;
}
