
#include "event.h"

#include <Slash/Tools/args.h>
#include <Slash/Tools/file.h>
#include <Slash/Tools/system.h>
#include <Slash/Tools/fmanip.h>

#include <Lib/InternetSocket.h>

#include <Inventor/SDL/SoSDL_enums.h>

#include <string>
#include <cstdlib>
#include <cstdio>
#include <iostream>
#include <map>

static void initJoysticks(std::vector<SDL_Joystick*>& a_joys){
  int numJoystick = SDL_NumJoysticks();
  //printf("debug : joysticks found : %d\n",numJoystick);
  for(int index=0;index<numJoystick;index++)  {
    //printf("debug : joystick name : %i : %s\n",
    //  index,SDL_JoystickName(index));
    SDL_Joystick* joyStick = SDL_JoystickOpen(index);
    if(joyStick) {
      //printf("debug : num axes : %d\n",SDL_JoystickNumAxes(joyStick));
      //printf("debug : num buttons : %d\n",SDL_JoystickNumButtons(joyStick));
      a_joys.push_back(joyStick);
    } else {
      //printf("debug : can't open joystick\n");
    }
  }
}
static void clearJoysticks(std::vector<SDL_Joystick*>& a_joys) {
  std::vector<SDL_Joystick*>::iterator it;
  for(it=a_joys.begin();it!=a_joys.end();++it) {
    SDL_JoystickClose(*it);
  }
  a_joys.clear();
}

class Tag {
public:
  int m_verbose;
  std::string m_app;
  unsigned int m_index;
  std::string m_slave;
  unsigned int m_port;
  // optional :
  std::string m_master;
  std::string m_URL;
};

static bool after_binding(void* a_tag) {
  Tag* tag = (Tag*)a_tag;

  if(tag->m_app=="none") return true; //clients started "by hand".

  // Create client 
  std::string cmd = tag->m_app;
  Slash::replace(cmd,"@@verbose@@",Slash::to<unsigned int>(tag->m_verbose));
  Slash::replace(cmd,"@@slave@@",tag->m_slave);
  Slash::replace(cmd,"@@port@@",Slash::to<unsigned int>(tag->m_port));
  Slash::replace(cmd,"@@index@@",Slash::to<unsigned int>(tag->m_index));

  // For some apps :
  Slash::replace(cmd,"@@master@@",tag->m_master);
  Slash::replace(cmd,"@@URL@@",tag->m_URL);

  int status = ::system(cmd.c_str());
  if(tag->m_verbose) {
    std::cout << "cmd is \"" << cmd << "\"." << std::endl;
    std::cout << "system status cmd is " << status << "." << std::endl;
  }
  if(status==(-1)) return false;

  return true;
}

class JoyTag {
public:
  std::vector<Lib::BaseSocket*>* m_socks;
  unsigned short m_move_what;
};

typedef std::map<std::string,std::string> curls_t;

bool curl_file(const std::string& a_url,const std::string& a_out_file) {

  //::printf("debug : SoSDL_WILD::curl_file : \"%s\"...\n",a_url.c_str());

  std::string cmd("curl ");
  cmd += " --silent ";
  cmd += " --compressed ";
  cmd += a_url;
  cmd += " --output ";
  cmd += a_out_file;
  if(::system(cmd.c_str())==(-1)) {
    std::cout << "SoSDL_WILD::curl_file :"
              << " call to ::system(\"" << cmd << "\") failed."
              << std::endl;
    ::remove(a_out_file.c_str());
    return false;
  }

  if(!Slash::file::exists(a_out_file)) {
    std::cout << "SoSDL_WILD::curl_file :"
              << " can't get " << a_url << "."
              << std::endl;
    return false;
  }

  std::vector<std::string> txt;
  if(!Slash::file::read_num(a_out_file,3,txt)) {
    std::cout << "SoSDL_WILD::curl_file :"
              << " can't read head of " << a_out_file << "."
              << std::endl;
    ::remove(a_out_file.c_str());
    return false;
  }
  if((txt.size()>=3)&&(txt[2]=="<title>404 Not Found</title>")) {
    std::cout << "SoSDL_WILD::curl_file :"
              << " url \"" << a_url << "\" not found."
              << std::endl;
    ::remove(a_out_file.c_str());
    return false;
  }

  //::printf("debug : SoSDL_WILD::curl_file : \"%s\" end\n",a_url.c_str());

  return true;
}

static bool net_poll_event(std::vector<Lib::BaseSocket*>& aSocks,bool& a_has,SDL_Event* a_event,Lib::BaseSocket*& a_socket) {
  std::vector<Lib::BaseSocket*>::iterator it;
  for(it=aSocks.begin();it!=aSocks.end();++it) {
    Lib::BaseSocket* socket = *it;
    bool have_input;
    if(!socket->is_there_input(have_input)) return false;
    if(have_input) {
      if(!wait_SDL_Event(*socket,*a_event)) {
	std::cout << "SoSDL_WILD::net_poll_event : wait_SDL_event failed"
                  << std::endl;
        return false;
      }
      a_socket = socket;
      a_has = true;
      return true;
    }
  }
  a_socket = 0;
  a_has = false;
  return true;
}

static bool send_SDL_Event(std::vector<Lib::BaseSocket*>& aSocks,const SDL_Event& aEvent) {
  std::vector<Lib::BaseSocket*>::iterator it;
  for(it=aSocks.begin();it!=aSocks.end();++it) {
    if(!send_SDL_Event(*(*it),aEvent)) return false;
  }
  return true;
}

static bool send_mouse_motion(
 std::vector<Lib::BaseSocket*>& aSocks
,unsigned int a_cols,unsigned int a_rows
,unsigned int a_ww,unsigned int a_wh
,float a_fx,float a_fy //a_fy = 0 = top
) {
  // Can't send a SDL_MOUSEMOTION since 
  // SDL_MouseMotionEvent.[x,y] are unsigned (Uint16).
  SDL_Event event;
  event.type = SDL_USEREVENT;
  event.user.type = SDL_USEREVENT;
  event.user.code = WILD_USEREVENT_MOUSE_MOVE;

  unsigned int WILD_ww = a_cols*a_ww;
  unsigned int WILD_wh = a_rows*a_wh;
  unsigned int WILD_x = a_fx * WILD_ww;
  unsigned int WILD_y = a_fy * WILD_wh;

  std::vector<Lib::BaseSocket*>::iterator it = aSocks.begin();
  //row bottom (of screen)  -> top
  //col left -> right
  for(unsigned int row=0;row<a_rows;row++) {
    for(unsigned int col=0;col<a_cols;col++) {
      unsigned int left = col * a_ww;
      //unsigned int right = left + a_ww;
      unsigned int bottom = row * a_wh;
      unsigned int top = bottom+a_wh;
      int x = WILD_x - left;
      int y = WILD_y - (WILD_wh-top);
      //printf("debug : col %d row %d send %d %d\n",col,row,x,y);
      Slash::write_buffer<int>(x,(char*)&(event.user.data1));
      Slash::write_buffer<int>(y,(char*)&(event.user.data2));
      if(!send_SDL_Event(*(*it),event)) return false;

      ++it;
    }
  }
  return true;
}

static bool send_mouse_button(
 std::vector<Lib::BaseSocket*>& aSocks
,unsigned int a_cols,unsigned int a_rows
,unsigned int a_ww,unsigned int a_wh
,float a_fx,float a_fy //a_fy = 0 = top
,bool a_down
) {
  // Can't send a SDL_MOUSEMOTION since 
  // SDL_MouseButtonEvent.[x,y] are unsigned (Uint16).
  SDL_Event event;
  event.type = SDL_USEREVENT;
  event.user.type = SDL_USEREVENT;
  if(a_down) 
    event.user.code = WILD_USEREVENT_MOUSE_BUTTON_DOWN;
  else
    event.user.code = WILD_USEREVENT_MOUSE_BUTTON_UP;

  unsigned int WILD_ww = a_cols*a_ww;
  unsigned int WILD_wh = a_rows*a_wh;
  unsigned int WILD_x = a_fx * WILD_ww;
  unsigned int WILD_y = a_fy * WILD_wh;

  std::vector<Lib::BaseSocket*>::iterator it = aSocks.begin();
  //row bottom (of screen)  -> top
  //col left -> right
  for(unsigned int row=0;row<a_rows;row++) {
    for(unsigned int col=0;col<a_cols;col++) {
      unsigned int left = col * a_ww;
      //unsigned int right = left + a_ww;
      unsigned int bottom = row * a_wh;
      unsigned int top = bottom+a_wh;
      int x = WILD_x - left;
      int y = WILD_y - (WILD_wh-top);
      //printf("debug : col %d row %d send %d %d\n",col,row,x,y);
      Slash::write_buffer<int>(x,(char*)&(event.user.data1));
      Slash::write_buffer<int>(y,(char*)&(event.user.data2));

      if(!send_SDL_Event(*(*it),event)) return false;

      ++it;
    }
  }
  return true;
}

static bool send_key_down(std::vector<Lib::BaseSocket*>& aSocks,SDLKey aKey){
  SDL_Event event;
  event.type = SDL_KEYDOWN;
  event.key.type = SDL_KEYDOWN;
  event.key.keysym.sym = aKey;
  event.key.keysym.mod = KMOD_NONE;
  return send_SDL_Event(aSocks,event);
}

inline void create_empty_file(const std::string& a_string) {
  ::remove(a_string.c_str());
  FILE* file = ::fopen(a_string.c_str(),"wb");  
  ::fclose(file);
}

static bool receive_CURL(SDL_Event& a_event,Lib::BaseSocket& a_socket,curls_t& a_curls){
  std::string url;
  read_SDL_Event_string(a_event,url);

  curls_t::const_iterator it = a_curls.find(url);         
  if(it!=a_curls.end()) { 
    //::printf("debug : SoSDL_WILD : HERE %s %s send...\n",
    //   url.c_str(),(*it).second.c_str());         
    if(!Slash::file::exists((*it).second)) {
      std::cout << "SoSDL_WILD :"
                << " WARNING : for url " << url
                << ", file in the map no more here."
		<< " Not fatal, send an empty file."
                << std::endl;
      //return false;
      // Not fatal for the whole app. Create an empty file.
      create_empty_file((*it).second);
    }

   {SDL_Event event;
    event.user.code = WILD_USEREVENT_FILE;
    if(!fill_SDL_Event_with_file((*it).second.c_str(),event)) {
      std::cout << "SoSDL_WILD : fill SDL_event with file failed." 
                << std::endl;
      return false;
    }
    if(!send_SDL_Event(a_socket,event)) {
      std::cout << "SoSDL_WILD : send_SDL_event failed." 
                << std::endl;
      delete [] ((char*)event.user.data2); 
      return false;
    }
    delete [] ((char*)event.user.data2);}

    //::printf("debug : SoSDL_WILD : url here %s end\n",url.c_str());          

  } else {
    //::printf("debug : SoSDL_WILD : NOT here %s send...001\n",url.c_str());
    std::string local;
    if(!Slash::tmpname("","SoSDL_WILD_","",local)) {
      std::cout << "SoSDL_WILD :"
                << " can't get a tmpname."
                << std::endl;
      return false;
    }

    if(!curl_file(url,local)) {
      std::cout << "SoSDL_WILD :"
                << " WARNING : slave request a file which is not here."
		<< " Not fatal, send an empty file."
                << std::endl;
      //return false;
      // Not fatal for the whole app. Create an empty file.
      create_empty_file(local);
    }

   {SDL_Event event;
    event.user.code = WILD_USEREVENT_FILE;
    if(!fill_SDL_Event_with_file(local,event)) {
      std::cout << "SoSDL_WILD : fill SDL_event with file failed." 
                << std::endl;
      return false;
    }
    if(!send_SDL_Event(a_socket,event)) {
      std::cout << "SoSDL_WILD : send_SDL_event failed." 
                << std::endl;
      delete [] ((char*)event.user.data2);
      return false;
    }
    delete [] ((char*)event.user.data2);}

    a_curls[url] = local;
  }
  return true;
}

static Uint32 s_move_timer_interval = 30; //msec
static SDL_TimerID s_move_timer_ID = 0;

static void stop_move_timer() {
  if(s_move_timer_ID) SDL_RemoveTimer(s_move_timer_ID);
  s_move_timer_ID = 0;
}

static Uint32 move_timer_CB(Uint32 aInterval,void* a_tag){
  JoyTag* tag = (JoyTag*)a_tag;
  //printf("debug : move_timer_CB.\n");
  std::vector<Lib::BaseSocket*>& socks = *(tag->m_socks);
  SDL_Event event;
  event.type = SDL_USEREVENT;
  event.user.type = SDL_USEREVENT;
  event.user.code = WILD_USEREVENT_UPDATE_MOTION;
  Slash::write_buffer<unsigned short>(tag->m_move_what,
                                      (char*)&(event.user.data1));
  event.user.data2 = 0;
  send_SDL_Event(socks,event);
  return aInterval;
}

static void start_move_timer(void* aTag) {
  if(s_move_timer_ID) SDL_RemoveTimer(s_move_timer_ID);
  s_move_timer_ID = SDL_AddTimer(s_move_timer_interval,move_timer_CB,aTag);
}

int main(int argc,char* argv[]) {

  bool is_mac_app = false;
#ifdef __APPLE__
  //std::string contents_dir;
  if(argc>0) {
    std::string prog_dir = Slash::dir_name(argv[0]);
    if(Slash::base_name(prog_dir)=="MacOS") {
      is_mac_app = true;
      //contents_dir = Slash::dir_name(prog_dir);
    }
  }
#endif

  Slash::args args(argc,argv);

  int verbose = 0;
  if(is_mac_app) { 
    if(!Slash::get_env<int>("WILD_VERBOSE",verbose)) verbose = 0;
  } else {
    if(args.is_arg("-verbose")) verbose = 1;
  }

  unsigned int rows;
  if(is_mac_app) { 
    if(!Slash::get_env<unsigned int>("WILD_ROWS",rows)) {
      std::cout << "WILD_ROWS not defined." << std::endl;
      return EXIT_FAILURE;
    }
  } else {
    if(!args.find<unsigned int>("-rows",rows)) {
      std::cout << "-rows=<uint> argument not given." << std::endl;
      return EXIT_FAILURE;
    }
  }

  unsigned int cols;
  if(is_mac_app) { 
    if(!Slash::get_env<unsigned int>("WILD_COLS",cols)) {
      std::cout << "WILD_COLS not defined." << std::endl;
      return EXIT_FAILURE;
    }
  } else {
    if(!args.find<unsigned int>("-cols",cols)) {
      std::cout << "-cols=<uint> argument not given." << std::endl;
      return EXIT_FAILURE;
    }
  }

  std::string app;
  if(is_mac_app) { 
    if(!Slash::getenv("WILD_APP",app)) app = "none";
  } else {
    if(!args.find("-app",app)) app = "none";
  }

  std::vector<std::string> slaves;
  if(app!="none") {
    std::string s;
    if(is_mac_app) { 
      if(!Slash::getenv("WILD_SLAVES",s)) {
        std::cout << "WILD_SLAVES not defined."
                  << std::endl;
        return EXIT_FAILURE;
      }
    } else {
      if(!args.find("-slaves",s)) {
        std::cout << "-slaves=<[host:]> argument not given." << std::endl;
        return EXIT_FAILURE;
      }
    }
    slaves = Slash::words(s,":");
    if(slaves.size()!=(rows*cols)) {
      std::cout << "bad number of slaves given." 
                << " " << int(rows*cols) << " expected"
                << " whilst " <<  (int)slaves.size() << " given."
                << std::endl;
      return EXIT_FAILURE;
    }
  }

  unsigned int start_port = 0;
  if(is_mac_app) { 
    if(!Slash::get_env<unsigned int>("WILD_PORT",start_port)) 
      start_port = 50700;
  } else {
    if(!args.find<unsigned int>("-port",start_port)) start_port = 50700;
  }

  if(verbose) {
    std::cout << "app is \"" << app << "\"." << std::endl;
  }

  // optionals :
  std::string master;
  std::string URL;
  if(is_mac_app) { 
    Slash::getenv("WILD_MASTER",master);
    // For the SoSDL_web_WILD.app and HEPVis_web_WILD_SDL.app apps.
    Slash::getenv("WILD_URL",URL);
  }

  std::vector<Lib::BaseSocket*> socks;

  unsigned int ww = 0;
  unsigned int wh = 0;

 {unsigned int index = 0;
  for(unsigned int row=0;row<rows;row++) {
    for(unsigned int col=0;col<cols;col++) {

      unsigned int port = start_port+index;
      //no host needed for an accept.    
      Lib::InternetSocket* sock = 
        new Lib::InternetSocket(std::cout,"",port,verbose);

      Tag tag;
      tag.m_verbose = verbose;
      tag.m_app = app;
      tag.m_index = index;
      if(app!="none") tag.m_slave = slaves[index];
      tag.m_port = port;
      tag.m_master = master;
      tag.m_URL = URL;

      sock->setAfterBindingFunc(after_binding,&tag);
      if(!sock->accept()) return EXIT_FAILURE;

      socks.push_back(sock);

      // wait the hello string :
      std::string s;
      if(!sock->wait_string(s)) return EXIT_FAILURE;
      if(verbose) {
        std::cout << "receive : " << s << std::endl;
      }

      if(!sock->send_uint(index)) return EXIT_FAILURE;

      if(!sock->send_uint(rows)) return EXIT_FAILURE;
      if(!sock->send_uint(cols)) return EXIT_FAILURE;
      if(!sock->send_uint(row)) return EXIT_FAILURE;
      if(!sock->send_uint(col)) return EXIT_FAILURE;

      if(!sock->wait_uint(ww)) return EXIT_FAILURE;
      if(!sock->wait_uint(wh)) return EXIT_FAILURE;

      index++;
    }
  }}

 {// "front end SDL" to get key events to send to viewers :
  if(SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER|SDL_INIT_JOYSTICK)<0) { 
    std::cout << "SoSDL_WILD :"
              << " couldn't initialize SDL " << SDL_GetError() 
              << std::endl;
    return EXIT_FAILURE;
  }

  //Uint32 flags = SDL_OPENGL;
  Uint32 flags = 0; //must not be full screen.
  float aspect = ((float)(cols*ww))/((float)(rows*wh));

  SDL_Surface* screen = SDL_SetVideoMode(200*aspect,200,0,flags);
  if(screen==NULL) {
    std::cout << "SoSDL_WILD :"
              << " can't set video." 
              << std::endl;
    SDL_Quit();
    return EXIT_FAILURE;
  }
  unsigned int www = screen->w;
  unsigned int wwh = screen->h;

  //SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,16);
  //SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
    
  SDL_EnableUNICODE(1);
  SDL_EnableKeyRepeat(200,10);

  std::vector<SDL_Joystick*> joys;
  initJoysticks(joys);

  bool show_menu = false;
  bool joy_state = false;
  bool joy_rotate = true;
  bool joy_forback = true;
  JoyTag tag;
  tag.m_socks = &socks;
  tag.m_move_what = SOSDL_ROTATE_RIGHT;

  std::string first_slave_file;

  curls_t curls;

  bool to_quit = false;

  while(!to_quit) {
    SDL_Event event;

    //printf("debug : count %d\n",global_count);global_count++;

    bool has;
    Lib::BaseSocket* socket = 0;
    if(!net_poll_event(socks,has,&event,socket)) {
      std::cout << "SoSDL_WILD :"
                << " net_poll_event failed." 
                << std::endl;
      break;
    }
    if(has) {
    } else {
      if(!SDL_PollEvent(&event)) {
        SDL_Delay(30);
        continue;
      }
    }

  //if(!SDL_WaitEvent(&event)) {
  //  std::cout << "SoSDL_WILD :"
  //            << " SDL_WaitEvent failed." 
  //            << std::endl;
  //  break;
  //}

    bool treated = false;

    // must be similar to SoSDL::mainLoop switch.
    switch(event.type) {
    case SDL_USEREVENT:
      treated = true;
      if(event.user.code==WILD_USEREVENT_CURL) {
        if(!receive_CURL(event,*socket,curls)) {
          std::cout << "SoSDL_WILD :"
                    << " receive_CURL failed."
                    << std::endl;
          to_quit = true;
          break;
        }
      } else {
        std::cout << "SoSDL_WILD :"
                  << " receive an SDL_USEREVENT with code " 
                  << event.user.code
                  << std::endl;
      }
      break;

    case SDL_QUIT: //close window in non-fullscreen mode.
      to_quit = true;
      treated = true;
      break;

    case SDL_KEYDOWN:
      if(event.key.keysym.sym==SDLK_ESCAPE) {
        to_quit = true;
      }
      if(event.key.keysym.sym==SDLK_m) {
        show_menu = show_menu?false:true;
      }
      break;
      
    case SDL_MOUSEMOTION:
      if(!send_mouse_motion(socks,
                            cols,rows,ww,wh,
                            float(event.motion.x)/float(www),
                            float(event.motion.y)/float(wwh))) {
        to_quit = true;
      }
      treated = true;
      break;

    case SDL_MOUSEBUTTONDOWN:
      if(!send_mouse_button(socks,
                            cols,rows,ww,wh,
                            float(event.motion.x)/float(www),
                            float(event.motion.y)/float(wwh),true)) {
        to_quit = true;
      }
      treated = true;
      break;
    case SDL_MOUSEBUTTONUP:
      if(!send_mouse_button(socks,
                            cols,rows,ww,wh,
                            float(event.button.x)/float(www),
                            float(event.button.y)/float(wwh),false)) {
        to_quit = true;
      }
      treated = true;
      break;

    case SDL_JOYBUTTONDOWN:
      // on a Logitech Precision the button num = jbutton.button+1
      if(event.jbutton.button==0) { 
        joy_rotate = joy_rotate?false:true;
        treated = true;
      }
      if(event.jbutton.button==1) { 
        joy_forback = joy_forback?false:true;
        treated = true;
      }
      if((event.jbutton.button==4)|| //shoot
         (event.jbutton.button==5)){ 
        //FIXME : if(show_menu) //Exit 
      }
      if((event.jbutton.button==6)|| //mire
         (event.jbutton.button==7)){ 
      }
      if(event.jbutton.button==8) { //quit
        to_quit = true;
        treated = true;
      }
      if(event.jbutton.button==9) { //menu 
        show_menu = show_menu?false:true;
      }
      break;

    case SDL_JOYAXISMOTION:
      if(event.jaxis.axis==0) {
        if((event.jaxis.value>-8000)&&(event.jaxis.value<8000)) {
          joy_state = false;
          if(show_menu) {
          } else {
            stop_move_timer();
          }
        } else {
          if(joy_state) {
            joy_state = false;
            if(show_menu) {
            } else {
              stop_move_timer();
            }
          } else {
            if(event.jaxis.value>0) {
              joy_state = true;
              if(show_menu) {
                if(!send_key_down(socks,SDLK_RIGHT)) {
                  to_quit = true;
                }
              } else {
                tag.m_move_what = 
                  (joy_rotate?SOSDL_ROTATE_RIGHT:SOSDL_TRANSLATE_RIGHT);
                start_move_timer(&tag);
              }
            } else {
              joy_state = true;
              if(show_menu) {
                if(!send_key_down(socks,SDLK_LEFT)) {
                  to_quit = true;
                }
              } else {
                tag.m_move_what = 
                  (joy_rotate?SOSDL_ROTATE_LEFT:SOSDL_TRANSLATE_LEFT);
                start_move_timer(&tag);
              }
            }
          }            
        }            
        treated = true;
      }            
      if(event.jaxis.axis==1) {
        if((event.jaxis.value>-8000)&&(event.jaxis.value<8000)) {
          joy_state = false;
          if(show_menu) {
          } else {
            stop_move_timer();
          }
        } else {
          if(joy_state) {
            joy_state = false;
            if(show_menu) {
            } else {
              stop_move_timer();
            }
          } else {
            if(event.jaxis.value>0) {
              joy_state = true;
              if(show_menu) {
              } else {
                tag.m_move_what = (joy_forback?SOSDL_BACKWARD:SOSDL_DOWN);
                start_move_timer(&tag);
              }
            } else {
              joy_state = true;
              if(show_menu) {
              } else {
                tag.m_move_what = (joy_forback?SOSDL_FORWARD:SOSDL_UP);
                start_move_timer(&tag);
              }
            }
          }            
        }            
        treated = true;
      }
      break;
    default:
      break;			
    }

    if(!treated) {
      if(   (event.type==SDL_KEYDOWN)
          ||(event.type==SDL_KEYUP)
          ||(event.type==SDL_JOYBUTTONDOWN)
          ||(event.type==SDL_JOYBUTTONUP)
	    //||(event.type==SDL_JOYAXISMOTION) 
          ){
        if(!send_SDL_Event(socks,event)) {
          to_quit = true;
        }
      }
    }
  }  

 {for(curls_t::iterator it = curls.begin();it!=curls.end();++it) {
    ::remove((*it).second.c_str());    
  }
  curls.clear();}

 {SDL_Event event;
  event.type = SDL_QUIT;
  if(!send_SDL_Event(socks,event)) {
  }}

  stop_move_timer();

  clearJoysticks(joys);
  SDL_Quit();

  } // end front-end-SDL

  for(std::vector<Lib::BaseSocket*>::iterator it = socks.begin();
      it!=socks.end();++it) {
    (*it)->disconnect();
    delete (*it);    
  }

  if(verbose) std::cout << "SoSDL_WILD : exiting..." << std::endl;

  return EXIT_SUCCESS;
}

#include "Lib/BaseSocket.cxx"
#include "Lib/InternetSocket.cxx"

