#ifndef SoSDL_WILD_camera_h
#define SoSDL_WILD_camera_h 

#include <math.h>
#ifndef M_PI
#define M_PI       3.1415926535897931160E0
#define M_PI_2     1.5707963267948965580E0  
#endif

#define FM_PI      ((float)M_PI)
#define FM_PI_2    ((float)M_PI_2)

//
#define CAMERA_PERS

#ifdef CAMERA_PERS
#include <Inventor/nodes/SoFrustumCamera.h>
#endif

#include "Lib/SoOrthoCamera.h"

inline SoCamera* create_camera(
                       unsigned int a_cols,unsigned int a_rows,
                       unsigned int a_col,unsigned int a_row,
                       float a_aspect,
                       float a_height = 2,
                       float a_near = 1){

  //printf("debug : %d %d %d %d\n",a_rows,a_cols,a_row,a_col);

  //row bottom (of screen) -> top
  //col left -> right

#ifdef CAMERA_PERS
  SoFrustumCamera* soCamera = new SoFrustumCamera;
  soCamera->viewportMapping.setValue(SoCamera::LEAVE_ALONE);

  float angle = FM_PI*0.25F;
  float height = 2.0 * a_near * ::tan(angle*0.5F);

  float width = a_aspect * height;
  //printf("debug : w %g h %g\n",width,height);

  float left = -((float)a_cols)*width*0.5F + ((float)a_col) * width;
  float bottom = -((float)a_rows)*height*0.5F + ((float)a_row) * height;

  //printf("debug : left %g bottom %g\n",left,bottom);

  soCamera->left.setValue(left);
  soCamera->right.setValue(left+width);
  soCamera->bottom.setValue(bottom);
  soCamera->top.setValue(bottom+height);

#else
  SoOrthoCamera* soCamera = new SoOrthoCamera;
  soCamera->height.setValue(a_height);

  float width = a_aspect * a_height;

  //printf("debug : w %g h %g\n",width,height);

  float left = -((float)a_cols)*width*0.5F + ((float)a_col) * width;
  float bottom = -((float)a_rows)*a_height*0.5F + ((float)a_row) * a_height;

  //printf("debug : left %g bottom %g\n",left,bottom);

  soCamera->left.setValue(left);
  soCamera->right.setValue(left+width);
  soCamera->bottom.setValue(bottom);
  soCamera->top.setValue(bottom+a_height);

#endif
  soCamera->nearDistance.setValue(a_near);

  return soCamera;
}

inline SoOrthoCamera* create_camera_ortho(
                       unsigned int a_cols,unsigned int a_rows,
  		       unsigned int a_col,unsigned int a_row,
                       float a_aspect = 1,float a_height = 2){

  SoOrthoCamera* soCamera = new SoOrthoCamera;
  //soCamera->viewportMapping.setValue(SoCamera::LEAVE_ALONE);
  soCamera->height.setValue(a_height);

  float width = a_aspect * a_height;

  float left = -((float)a_cols)*width*0.5F + ((float)a_col) * width;
  float bottom = -((float)a_rows)*a_height*0.5F + ((float)a_row) * a_height;

  soCamera->left.setValue(left);
  soCamera->right.setValue(left+width);
  soCamera->bottom.setValue(bottom);
  soCamera->top.setValue(bottom+a_height);

  return soCamera;
}

#endif
