#ifndef SoSDL_WILD_net_h
#define SoSDL_WILD_net_h 

#include <Lib/BaseSocket.h>

#include <list>

class WILD_Tag {
public:
  WILD_Tag():m_socket(0){}
  Lib::BaseSocket* m_socket;
  std::list<SDL_Event*> m_events;
};

#include "event.h"

#include <cstring>

inline int net_poll_event(SDL_Event* a_event,void* a_tag){
  WILD_Tag* WILD_tag = (WILD_Tag*)a_tag;
  if(WILD_tag->m_events.size()) {
    SDL_Event* event = WILD_tag->m_events.front();
    ::memcpy(a_event,event,sizeof(SDL_Event));
    WILD_tag->m_events.pop_front();
    delete event;
    return 1;
  }
  Lib::BaseSocket* socket = WILD_tag->m_socket;
  bool have_input;
  if(!socket->is_there_input(have_input)) return 0;
  if(!have_input) return 0;
  //std::cout << "have input..." << std::endl;
  //std::cout << "receive : " << s << std::endl;
  if(!wait_SDL_Event(*socket,*a_event)) {
    SoSDL::toQuit();
    return 0;
  }
  return 1;
}

#include <Inventor/SoSceneManager.h>
#include <Inventor/events/SoMouseButtonEvent.h>
#include <Inventor/events/SoLocation2Event.h>
#include <Inventor/SDL/SoSDL.h>

inline SbBool WILD_processEvent(const SDL_Event& a_event){
  /////////////////////////////////
  // WILD events. Treat them here :
  /////////////////////////////////
  if(a_event.type!=SDL_USEREVENT) return FALSE;
  if(a_event.user.code==WILD_USEREVENT_UPDATE_MOTION){
    unsigned short v = 0;
    Slash::read_buffer<unsigned short>((char*)&(a_event.user.data1),v);
    //printf("debug : WILD_USEREVENT_UPDATE_MOTION : %d\n",v);
    SoSDL_Viewer* viewer = SoSDL::getTopLevelWidget()->getViewer();
    viewer->updateMotion((SoSDL::Move)v);
    SoSDL::setViewerRender();
    return TRUE;

  } else if(a_event.user.code==WILD_USEREVENT_MOUSE_MOVE) {
    int x = 0;
    int y = 0;
    Slash::read_buffer<int>((char*)&(a_event.user.data1),x);
    Slash::read_buffer<int>((char*)&(a_event.user.data2),y);
    //::printf("debug : %g %g\n",x,y);
    //x,y could be negative.
    // Build SoEvent.
    SoLocation2Event event;    
    //SDL : origin is top-left with y down.
    //Inventor :  origin is bottom-left with y up (like OpenGL).
    SoSDL_Viewer* viewer = SoSDL::getTopLevelWidget()->getViewer();
    SoSceneManager* sm = viewer->getSceneManager();
    const SbVec2s& win = sm->getViewportRegion().getWindowSize();
    event.setPosition(SbVec2s(x,win[1]-y));    
    sm->processEvent(&event);
    return TRUE;

  } else if((a_event.user.code==WILD_USEREVENT_MOUSE_BUTTON_DOWN) ||
            (a_event.user.code==WILD_USEREVENT_MOUSE_BUTTON_UP) ){
    int x = 0;
    int y = 0;
    Slash::read_buffer<int>((char*)&(a_event.user.data1),x);
    Slash::read_buffer<int>((char*)&(a_event.user.data2),y);
    //::printf("debug : %g %g\n",x,y);
    //x,y could be negative.
    // Build SoEvent.
    SoMouseButtonEvent event;    
    // WARNING : for Coin, BUTTON2 is right button,
    //           BUTTON3 is middle button.
    event.setButton(SoMouseButtonEvent::BUTTON1);
    if(a_event.user.code==WILD_USEREVENT_MOUSE_BUTTON_DOWN)
      event.setState(SoButtonEvent::DOWN);
    else if(a_event.user.code==WILD_USEREVENT_MOUSE_BUTTON_UP)
      event.setState(SoButtonEvent::UP);
    else
      event.setState(SoButtonEvent::UNKNOWN);    
    //SDL : origin is top-left with y down.
    //Inventor :  origin is bottom-left with y up (like OpenGL).
    SoSDL_Viewer* viewer = SoSDL::getTopLevelWidget()->getViewer();
    SoSceneManager* sm = viewer->getSceneManager();
    const SbVec2s& win = sm->getViewportRegion().getWindowSize();
    event.setPosition(SbVec2s(x,win[1]-y));
    sm->processEvent(&event);
    return TRUE;

  }

  return FALSE;
}

inline int treat_event(SDL_Event* a_event,void* a_tag){
  if(net_poll_event(a_event,a_tag)) {
    if(WILD_processEvent(*a_event)==TRUE) return 0; //done.
    return 1;
  }
  int has = SDL_PollEvent(a_event);
  if(has) {
    if(a_event->type==SDL_KEYDOWN) {
      //if(a_event->key.keysym.sym==SDLK_q)
      if(a_event->key.keysym.sym==SDLK_ESCAPE){
        //let the possibility to quit client directly.
        return 1;
      }
    }
    if((a_event->type==SDL_KEYDOWN)||(a_event->type==SDL_KEYUP) ){
      if((a_event->key.keysym.sym==SDLK_RETURN) ||
         (a_event->key.keysym.sym==SDLK_r) ||
         (a_event->key.keysym.sym==SDLK_m) ){
        //some key emulations.
        return 1;
      }
    }
    //else disable device client events.
    //NOTE : even if the client window has not the focus, it
    //       receives the joystick events.
    if(   (a_event->type==SDL_KEYDOWN)
        ||(a_event->type==SDL_KEYUP)
        ||(a_event->type==SDL_JOYBUTTONDOWN)
        ||(a_event->type==SDL_JOYBUTTONUP)
        ||(a_event->type==SDL_JOYAXISMOTION) 
        ){
      //std::cout << "client : bypass " << (int)a_event->type << std::endl;
      return 0;       
    }
  }
  return has;
}

#endif
