//  Mai/2007 : G.Barrand : creation.

#ifndef SoSDL_SoSDL_h
#define SoSDL_SoSDL_h 

#ifdef WIN32
#include <SDL.h>
#else
#include <SDL/SDL.h>
#endif

#ifdef WIN32
//  There is a clash with int typedefs (int8_t,...) found in :
//    CoinGL/include/Inventor/system/inttypes_win32.h
// and :
//    SDL/1.2.11/include/SDL_config_win32.h
#define COIN_NO_INT_TYPEDEFS
#endif

#include <Inventor/SbBasic.h>
#include <Inventor/SbString.h>
#include <Inventor/SbVec2s.h>

#include <Inventor/SDL/SoSDL_enums.h>

#include <vector>

class SoSDL_Viewer;
class SoSDL_AudioSample;

class SoSDL_Widget {
public:
  SoSDL_Widget();
  virtual~ SoSDL_Widget();
  void setViewer(SoSDL_Viewer*); //FIXME
  SoSDL_Viewer* getViewer() const; //FIXME
private:  
  SoSDL_Viewer* fViewer; //FIXME
};

// Have similar APU than other SoGui (SoXt, SoQt, etc...).

class SoSDL {
public:
  //static SbBool init();
  static SoSDL_Widget* init(int video_width,int video_height,Uint32 flags);
  static void show(SoSDL_Widget*);
  static void mainLoop();
  //
  static void clear();
  static SDL_Surface* screen();
  static void playAudioFile(const SbString&);
  static void clearAudio();
  //static bool isShiftDown(const SDL_Event& a_event);
  static void setViewerRender();
  static void toQuit();

  static SoSDL_Widget* getTopLevelWidget();

  static void sendMessage(int,void* = 0,void* = 0);

  static void emulateLeftButtonDown(const SbVec2s&);
  static void emulateLeftButtonUp(const SbVec2s&);
  static void emulateKeyDown(SDLKey);
  static void emulateKeyUp(SDLKey);

  typedef int(*EventFunc)(SDL_Event*,void*);
  static void setEventFunc(EventFunc,void*);

  enum Move {
    ROTATE_RIGHT = SOSDL_ROTATE_RIGHT,
    TRANSLATE_RIGHT = SOSDL_TRANSLATE_RIGHT,
    ROTATE_LEFT = SOSDL_ROTATE_LEFT,
    TRANSLATE_LEFT = SOSDL_TRANSLATE_LEFT,
    UP = SOSDL_UP,
    DOWN = SOSDL_DOWN,
    FORWARD = SOSDL_FORWARD,
    BACKWARD = SOSDL_BACKWARD
  };

  //static void setMoveTimerInterval(unsigned int interval_msec);
private:

  static void initJoysticks();
  static void clearJoysticks();

  // joystick animation :
  static Uint32 moveTimerCB(Uint32,void*);

  static void initAudio();
  static void playAudio(SoSDL_AudioSample&);

  static Uint32 mainTimerCB(Uint32,void*);

  //static void sensorQueueChanged(void*);
  //static Uint32 timerSensorCB(Uint32,void*);
  //static Uint32 delaySensorCB(Uint32,void*);

  static void dumpEvent(SDL_Event& aEvent);
private:
  static SDL_Surface* fScreen;
  static SoSDL_Widget* fWidget;
  static std::vector<SoSDL_AudioSample*>* fAudioSamples;  
  static SDL_TimerID fMainTimerID;
  //static SDL_TimerID fTimerSensorID;
  //static SDL_TimerID fDelaySensorID;
  static std::vector<SDL_Joystick*>* fJoySticks;
  static SbBool fViewer_render;
  static SbBool fToQuit;
  static EventFunc fEventFunc;
  static void* fEventFuncTag;

};

#endif
