//  Mai/2007 : G.Barrand : creation.

// this :
#include <Inventor/SDL/viewers/SoSDL_Viewer.h>

#include <Inventor/SbRotation.h>
#include <Inventor/SoSceneManager.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodekits/SoBaseKit.h>
#include <Inventor/sensors/SoNodeSensor.h>
#include <Inventor/actions/SoHandleEventAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/events/SoMouseButtonEvent.h>
#include <Inventor/events/SoLocation2Event.h>
#include <Inventor/events/SoKeyboardEvent.h>

#include <Inventor/SDL/SbGL.h>
#include <Inventor/SDL/SoJoyAxisEvent.h>
#include <Inventor/SDL/SoJoyButtonEvent.h>

SoSDL_Viewer::SoSDL_Viewer(SoSDL_Widget* aWidget)
:fSceneManager(0)
,fSceneGraph(0)
,fUserSceneGraph(0)
,fCamera(0)
,fHeadLight(0)
,fDefaultDeltaRotation(1.0F * 3.1416F/180.0F) //1 degree
,fDefaultDeltaTranslation(1) 
,fCheckFloor(FALSE)
,fFloor(0)
,fJoy_state(FALSE)
{
  fSceneManager = new SoSceneManager;

  aWidget->setViewer(this); //Used in the SoSDL::mainLoop.

  fCameraType = SoOrthographicCamera::getClassTypeId();

  fSceneGraph = new SoGroup;
  fSceneGraph->ref();

  //0
  SoLightModel* lightModel = new SoLightModel;
  lightModel->model.setValue(SoLightModel::PHONG);
  //lightModel->model.setValue(SoLightModel::BASE_COLOR);
  fSceneGraph->addChild(lightModel);

  //1
  fHeadLight = new SoDirectionalLight;
  fHeadLight->ref();
  fHeadLight->color.setValue(1,1,1);
  fHeadLight->direction.setValue(SbVec3f(1,-1,-10));
  fSceneGraph->addChild(fHeadLight);

  //2
  SoOrthographicCamera* camera = new SoOrthographicCamera();
  fSceneGraph->addChild(camera);
  setCamera(camera);

#define USER_SCENE_INDEX 3

  SDL_Surface* screen = SoSDL::screen();
  if(screen) {
    fSceneManager->setWindowSize(SbVec2s(screen->w,screen->h));
  }

  fSceneManager->setBackgroundColor(SbColor(0,0,0));

  fSceneManager->setSceneGraph(fSceneGraph);
}

SoSDL_Viewer::~SoSDL_Viewer(){
  if(fUserSceneGraph) {
    //fSceneGraph->removeChild(fUserSceneGraph);
    fUserSceneGraph->unref();
  }
  if(fCamera) fCamera->unref();
  if(fHeadLight) fHeadLight->unref();
  fSceneGraph->removeAllChildren();
  fSceneGraph->unref();
  delete fSceneManager;
}

SbBool SoSDL_Viewer::setSceneGraph(SoNode* aScene){
  if(aScene==fUserSceneGraph) return TRUE; //Done.

  if(fUserSceneGraph) {
    //fSceneGraph->removeChild(fUserSceneGraph);
    fUserSceneGraph->unref();
  }
  fUserSceneGraph = aScene;
  if(!fUserSceneGraph) return TRUE;

  fUserSceneGraph->ref();

  int number = fSceneGraph->getNumChildren();
  if(number==USER_SCENE_INDEX) {
    fSceneGraph->addChild(fUserSceneGraph);
  } else if(number==(USER_SCENE_INDEX+1)) {
    fSceneGraph->replaceChild(USER_SCENE_INDEX,fUserSceneGraph);
  } else {
    //error.
  }

  // Search for a camera in the user-supplied scenegraph.
  SbBool oldsearch = SoBaseKit::isSearchingChildren();
  SoBaseKit::setSearchingChildren(TRUE);
  SoSearchAction* searchAction = new SoSearchAction;
  searchAction->reset();
  searchAction->setType(SoCamera::getClassTypeId());
  searchAction->apply(fUserSceneGraph);
  SoBaseKit::setSearchingChildren(oldsearch);

  SoCamera* sceneCamera = 0;
  if(searchAction->getPath()) {
    SoFullPath* fullpath = (SoFullPath*)searchAction->getPath();
    SoNode* node = fullpath->getTail();
    if(node->isOfType(SoCamera::getClassTypeId())==FALSE) return FALSE;
    sceneCamera = (SoCamera*)node;
  }
  delete searchAction;

  // Make our own camera if none was available.
  /*
  if (!sceneCamera) {
    sceneCamera = (SoCamera*)fCameraType.createInstance();
    //PRIVATE(this)->deletecamera = TRUE;
  }
  */

  if(sceneCamera) setCamera(sceneCamera);

  //FIXME : Should notify the fSceneManager.
  return TRUE;
}

SoNode* SoSDL_Viewer::getSceneGraph() const {return fUserSceneGraph;}
void SoSDL_Viewer::setCamera(SoCamera* aCamera){
  if(aCamera==fCamera) return; //Done.
  if(fCamera) fCamera->unref();
  fCamera = aCamera;
  if(fCamera) fCamera->ref();
}

SoCamera* SoSDL_Viewer::getCamera() const {return fCamera;}

void SoSDL_Viewer::saveCamera() {
  if(!fCamera) return;
  fSave_camera_position = fCamera->position.getValue();
  fSave_camera_orientation = fCamera->orientation.getValue();
  //fSave_camera_nearDistance = fCamera->nearDistance.getValue();
  //fSave_camera_farDistance = fCamera->farDistance.getValue();
  //fSave_camera_focalDistance = fCamera->focalDistance.getValue();
}
void SoSDL_Viewer::restoreCamera() {
  if(!fCamera) return;
  fCamera->position.setValue(fSave_camera_position);
  fCamera->orientation.setValue(fSave_camera_orientation);
  //fCamera->nearDistance.setValue(fSave_camera_nearDistance);
  //fCamera->farDistance.setValue(fSave_camera_farDistance);
  //fCamera->focalDistance.setValue(fSave_camera_focalDistance);
}

void SoSDL_Viewer::setCameraType(SoType aType){fCameraType = aType;}
SoType SoSDL_Viewer::getCameraType() const {return fCameraType;}
void SoSDL_Viewer::setViewportRegion(const SbViewportRegion& aNewRegion) { 
  fSceneManager->setViewportRegion(aNewRegion); 
}
const SbViewportRegion& SoSDL_Viewer::getViewportRegion() const { 
  return fSceneManager->getViewportRegion(); 
}
void SoSDL_Viewer::setBackgroundColor(const SbColor& aC) { 
  fSceneManager->setBackgroundColor(aC); 
}
const SbColor& SoSDL_Viewer::getBackgroundColor() const { 
  return fSceneManager->getBackgroundColor(); 
}
void SoSDL_Viewer::setSize(const SbVec2s& a_size) {
  //Must stay dummy since size is the full screen.
  //fSceneManager->setWindowSize(a_size);
}
void SoSDL_Viewer::viewAll(){   
  if(!fCamera) return;
  // No shape under fSceneGraph, take directly fUserSceneGraph.
  if(!fUserSceneGraph) return;
  fCamera->viewAll(fUserSceneGraph,getViewportRegion());
}
void SoSDL_Viewer::setViewing(SbBool){}
void SoSDL_Viewer::show(){SoSDL::setViewerRender();}
void SoSDL_Viewer::setGLRenderAction(SoGLRenderAction* aAction){
  fSceneManager->setGLRenderAction(aAction);
}
SoSceneManager* SoSDL_Viewer::getSceneManager() const {   
  return fSceneManager;
}
SoGLRenderAction* SoSDL_Viewer::getGLRenderAction() const {   
  return fSceneManager->getGLRenderAction();
}
void SoSDL_Viewer::setTitle(const char*){}
void SoSDL_Viewer::setHeadlight(SbBool aOnOff) {   
  fHeadLight->on.setValue(aOnOff);
}
void SoSDL_Viewer::render(){   
  if(!fSceneGraph) return;
  glEnable(GL_DEPTH_TEST); //FIXME : do it through the fSceneManager
  fSceneManager->render();
}
void SoSDL_Viewer::setAutoClipping(SbBool){}
void SoSDL_Viewer::setDecoration(SbBool){}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

SbBool SoSDL_Viewer::processEvent(const SDL_Event& aEvent){
  // translate SDL_Event to an SoEvent and do a fSceneManager->processEvent

  if( (aEvent.type==SDL_MOUSEBUTTONDOWN) || 
      (aEvent.type==SDL_MOUSEBUTTONUP) ) {
    
    // Build SoEvent.
    SoMouseButtonEvent event;
    
    // WARNING : for Coin, BUTTON2 is right button, BUTTON3 is middle button.
    switch(aEvent.button.button){
    case SDL_BUTTON_LEFT:
      event.setButton(SoMouseButtonEvent::BUTTON1);
      break;
    case SDL_BUTTON_RIGHT:
      event.setButton(SoMouseButtonEvent::BUTTON2);
      break;
    case SDL_BUTTON_MIDDLE:
      event.setButton(SoMouseButtonEvent::BUTTON3);
      break;
    default:
      event.setButton(SoMouseButtonEvent::ANY);
      break;
    }
    if(aEvent.button.state==SDL_PRESSED)   
      event.setState(SoButtonEvent::DOWN);
    else if(aEvent.button.state==SDL_RELEASED) 
      event.setState(SoButtonEvent::UP);
    else
      event.setState(SoButtonEvent::UNKNOWN);
      
    SDLMod state = SDL_GetModState();
    if((state & KMOD_LSHIFT) == KMOD_LSHIFT) event.setShiftDown (TRUE);
    if((state & KMOD_RSHIFT) == KMOD_RSHIFT) event.setShiftDown (TRUE);
    if((state & KMOD_LCTRL)  == KMOD_LCTRL)  event.setCtrlDown  (TRUE);
    if((state & KMOD_RCTRL)  == KMOD_RCTRL)  event.setCtrlDown  (TRUE);
    if((state & KMOD_LALT)   == KMOD_LALT)   event.setAltDown   (TRUE);  
    if((state & KMOD_RALT)   == KMOD_RALT)   event.setAltDown   (TRUE);  
    
    //y goes from top to bottom of the window.
    //printf("debug : %d %d\n",aEvent.button.x,aEvent.button.y);

    //SDL : origin is top-left with y down.
    //Inventor :  origin is bottom-left with y up (like OpenGL).

    const SbVec2s& win = getViewportRegion().getWindowSize();
    event.setPosition(SbVec2s(short(aEvent.button.x),
                              win[1] - short(aEvent.button.y)));
    
    return fSceneManager->processEvent(&event);

  } else if(aEvent.type==SDL_MOUSEMOTION) {
    
    // Build SoEvent.
    SoLocation2Event event;
    
    SDLMod state = SDL_GetModState();
    if((state & KMOD_LSHIFT) == KMOD_LSHIFT) event.setShiftDown (TRUE);
    if((state & KMOD_RSHIFT) == KMOD_RSHIFT) event.setShiftDown (TRUE);
    if((state & KMOD_LCTRL)  == KMOD_LCTRL)  event.setCtrlDown  (TRUE);
    if((state & KMOD_RCTRL)  == KMOD_RCTRL)  event.setCtrlDown  (TRUE);
    if((state & KMOD_LALT)   == KMOD_LALT)   event.setAltDown   (TRUE);  
    if((state & KMOD_RALT)   == KMOD_RALT)   event.setAltDown   (TRUE);  
    
    //SDL : origin is top-left with y down.
    //Inventor :  origin is bottom-left with y up (like OpenGL).

    const SbVec2s& win = getViewportRegion().getWindowSize();
    event.setPosition(SbVec2s(short(aEvent.motion.x),
                              win[1] - short(aEvent.motion.y)));
    
    return fSceneManager->processEvent(&event);

  } else if((aEvent.type==SDL_KEYDOWN)||(aEvent.type==SDL_KEYUP)) {

    // Build SoEvent.
    SoKeyboardEvent event;
    
    if(aEvent.type==SDL_KEYDOWN) event.setState(SoButtonEvent::DOWN);
    if(aEvent.type==SDL_KEYUP) event.setState(SoButtonEvent::UP);

    SDLMod state = aEvent.key.keysym.mod;
    if((state & KMOD_LSHIFT) == KMOD_LSHIFT) event.setShiftDown (TRUE);
    if((state & KMOD_RSHIFT) == KMOD_RSHIFT) event.setShiftDown (TRUE);
    if((state & KMOD_LCTRL)  == KMOD_LCTRL)  event.setCtrlDown  (TRUE);
    if((state & KMOD_RCTRL)  == KMOD_RCTRL)  event.setCtrlDown  (TRUE);
    if((state & KMOD_LALT)   == KMOD_LALT)   event.setAltDown   (TRUE);  
    if((state & KMOD_RALT)   == KMOD_RALT)   event.setAltDown   (TRUE);  
    
    //if((aEvent.key.keysym.unicode & 0xFF80)==0) {
    //  char ch = aEvent.key.keysym.unicode & 0x7F;
    //  event.setPrintableCharacter(ch);
    //}

    switch(aEvent.key.keysym.sym) {
    case SDLK_a:event.setKey(SoKeyboardEvent::A);break;
    case SDLK_b:event.setKey(SoKeyboardEvent::B);break;
    case SDLK_c:event.setKey(SoKeyboardEvent::C);break;
    case SDLK_d:event.setKey(SoKeyboardEvent::D);break;
    case SDLK_e:event.setKey(SoKeyboardEvent::E);break;
    case SDLK_f:event.setKey(SoKeyboardEvent::F);break;
    case SDLK_g:event.setKey(SoKeyboardEvent::G);break;
    case SDLK_h:event.setKey(SoKeyboardEvent::H);break;
    case SDLK_i:event.setKey(SoKeyboardEvent::I);break;
    case SDLK_j:event.setKey(SoKeyboardEvent::J);break;
    case SDLK_k:event.setKey(SoKeyboardEvent::K);break;
    case SDLK_l:event.setKey(SoKeyboardEvent::L);break;
    case SDLK_m:event.setKey(SoKeyboardEvent::M);break;
    case SDLK_n:event.setKey(SoKeyboardEvent::N);break;
    case SDLK_o:event.setKey(SoKeyboardEvent::O);break;
    case SDLK_p:event.setKey(SoKeyboardEvent::P);break;
    case SDLK_q:event.setKey(SoKeyboardEvent::Q);break;
    case SDLK_r:event.setKey(SoKeyboardEvent::R);break;
    case SDLK_s:event.setKey(SoKeyboardEvent::S);break;
    case SDLK_t:event.setKey(SoKeyboardEvent::T);break;
    case SDLK_u:event.setKey(SoKeyboardEvent::U);break;
    case SDLK_v:event.setKey(SoKeyboardEvent::V);break;
    case SDLK_w:event.setKey(SoKeyboardEvent::W);break;
    case SDLK_x:event.setKey(SoKeyboardEvent::X);break;
    case SDLK_y:event.setKey(SoKeyboardEvent::Y);break;
    case SDLK_z:event.setKey(SoKeyboardEvent::Z);break;

    case SDLK_HOME:event.setKey(SoKeyboardEvent::HOME);break;
    case SDLK_LEFT:event.setKey(SoKeyboardEvent::LEFT_ARROW);break;
    case SDLK_UP:event.setKey(SoKeyboardEvent::UP_ARROW);break;
    case SDLK_RIGHT:event.setKey(SoKeyboardEvent::RIGHT_ARROW);break;
    case SDLK_DOWN:event.setKey(SoKeyboardEvent::DOWN_ARROW);break;
    case SDLK_PAGEUP:event.setKey(SoKeyboardEvent::PAGE_UP);break;
    case SDLK_PAGEDOWN:event.setKey(SoKeyboardEvent::PAGE_DOWN);break;
    case SDLK_END:event.setKey(SoKeyboardEvent::END);break;
  //case SDLK_:event.setKey(SoKeyboardEvent::PRIOR);break;
  //case SDLK_:event.setKey(SoKeyboardEvent::NEXT);break;

    case SDLK_BACKSPACE:event.setKey(SoKeyboardEvent::BACKSPACE);break;
    case SDLK_RETURN:event.setKey(SoKeyboardEvent::RETURN);break;
    case SDLK_PAUSE:event.setKey(SoKeyboardEvent::PAUSE);break;
    case SDLK_ESCAPE:event.setKey(SoKeyboardEvent::ESCAPE);break;
    case SDLK_PRINT:event.setKey(SoKeyboardEvent::PRINT);break;
  //case SDLK_:event.setKey(SoKeyboardEvent::NUM_LOCK);break;

    case SDLK_SPACE:event.setKey(SoKeyboardEvent::SPACE);break;
  //case SDLK_:event.setKey(SoKeyboardEvent::APOSTROPHE);break;
    case SDLK_COMMA:event.setKey(SoKeyboardEvent::COMMA);break;
    case SDLK_MINUS:event.setKey(SoKeyboardEvent::MINUS);break;
    case SDLK_PERIOD:event.setKey(SoKeyboardEvent::PERIOD);break;
    case SDLK_SLASH:event.setKey(SoKeyboardEvent::SLASH);break;
    case SDLK_SEMICOLON:event.setKey(SoKeyboardEvent::SEMICOLON);break;
    case SDLK_EQUALS:event.setKey(SoKeyboardEvent::EQUAL);break;
    case SDLK_LEFTBRACKET:event.setKey(SoKeyboardEvent::BRACKETLEFT);break;
    case SDLK_BACKSLASH:event.setKey(SoKeyboardEvent::BACKSLASH);break;
    case SDLK_RIGHTBRACKET:event.setKey(SoKeyboardEvent::BRACKETRIGHT);break;
  //case SDLK_:event.setKey(SoKeyboardEvent::GRAVE);break;

  //case SDLK_:event.setKey(SoKeyboardEvent::);break;

    default:event.setKey(SoKeyboardEvent::UNDEFINED);break;
    }
  
    return fSceneManager->processEvent(&event);

  //JoyStick
  } else if(aEvent.type==SDL_JOYBUTTONDOWN) {

    // Build SoEvent.
    SoJoyButtonEvent event;
    event.setState(SoJoyButtonEvent::DOWN);
    event.setButton(aEvent.jbutton.button);

    return fSceneManager->processEvent(&event);

  } else if(aEvent.type==SDL_JOYBUTTONUP) {

    // Build SoEvent.
    SoJoyButtonEvent event;
    event.setState(SoJoyButtonEvent::UP);
    event.setButton(aEvent.jbutton.button);

    return fSceneManager->processEvent(&event);

  } else if(aEvent.type==SDL_JOYAXISMOTION) {

    // Build SoEvent.
    SoJoyAxisEvent event;

    if(aEvent.jaxis.axis==0) {
      if((aEvent.jaxis.value>-8000)&&(aEvent.jaxis.value<8000)) {
        fJoy_state = FALSE;
        event.setState(SoJoyAxisEvent::STOP);    
      } else {
        if(fJoy_state==TRUE) {
          fJoy_state = FALSE;
          event.setState(SoJoyAxisEvent::STOP);    
        } else {
          if(aEvent.jaxis.value>0) {
            fJoy_state = TRUE;
            event.setState(SoJoyAxisEvent::RIGHT);    
          } else {
            fJoy_state = TRUE;
            event.setState(SoJoyAxisEvent::LEFT);    
          }
        }
      }            
    } else if(aEvent.jaxis.axis==1) {
      if((aEvent.jaxis.value>-8000)&&(aEvent.jaxis.value<8000)) {
        fJoy_state = FALSE;
        event.setState(SoJoyAxisEvent::STOP);    
      } else {
        if(fJoy_state==TRUE) {
          fJoy_state = FALSE;
          event.setState(SoJoyAxisEvent::STOP);    
        } else {
          if(aEvent.jaxis.value>0) {
            fJoy_state = TRUE;
            event.setState(SoJoyAxisEvent::DOWN);    
          } else {
            fJoy_state = TRUE;
            event.setState(SoJoyAxisEvent::UP);    
          }
        }
      }            
    }

    return fSceneManager->processEvent(&event);
  }

  return FALSE;
}
void SoSDL_Viewer::rotateCameraPositionAroundX(float aDeltaAngle){
  // From CoinXt/SoGuiExaminerViewerP::rotXWheelMotion.
  if(!fCamera) return;

  SbVec3f dir;
  fCamera->orientation.getValue().multVec(SbVec3f(0,0,-1),dir);

  SbVec3f focalpoint = 
    fCamera->position.getValue() + fCamera->focalDistance.getValue() * dir;

  fCamera->orientation = 
    SbRotation(SbVec3f(-1,0,0),aDeltaAngle) * fCamera->orientation.getValue();

  fCamera->orientation.getValue().multVec(SbVec3f(0,0,-1),dir);

  SbVec3f pos = focalpoint - fCamera->focalDistance.getValue() * dir;
  setCameraPosition(pos);
}
void SoSDL_Viewer::rotateCameraPositionAroundY(float aDeltaAngle){
  // From CoinXt/SoGuiExaminerViewerP::rotYWheelMotion.
  if(!fCamera) return;

  SbVec3f dir;
  fCamera->orientation.getValue().multVec(SbVec3f(0,0,-1),dir);

  SbVec3f focalpoint = 
    fCamera->position.getValue() + fCamera->focalDistance.getValue() * dir;

  fCamera->orientation = 
    SbRotation(SbVec3f(0,1,0),aDeltaAngle) * fCamera->orientation.getValue();

  fCamera->orientation.getValue().multVec(SbVec3f(0,0,-1),dir);

  SbVec3f pos = focalpoint - fCamera->focalDistance.getValue() * dir;
  setCameraPosition(pos);
}
void SoSDL_Viewer::rotateCameraAroundUp(float aDeltaAngle){
  if(!fCamera) return;
  SbVec3f up;
  fCamera->orientation.getValue().multVec(SbVec3f(0,1,0),up);
  fCamera->orientation = 
    SbRotation(up,aDeltaAngle) * fCamera->orientation.getValue();
}
void SoSDL_Viewer::translateCameraAlongDir(float aDelta){
  if(!fCamera) return;

  SbVec3f dir;
  fCamera->orientation.getValue().multVec(SbVec3f(0,0,-1),dir);

  SbVec3f pos = fCamera->position.getValue() + aDelta * dir;
  setCameraPosition(pos);
}
void SoSDL_Viewer::translateCameraAlongSide(float aDelta){
  if(!fCamera) return;

  SbVec3f dir;
  fCamera->orientation.getValue().multVec(SbVec3f(0,0,-1),dir);
  SbVec3f up;
  fCamera->orientation.getValue().multVec(SbVec3f(0,1,0),up);

  SbVec3f side = dir.cross(up);

  SbVec3f pos = fCamera->position.getValue() + aDelta * side;
  setCameraPosition(pos);
}
void SoSDL_Viewer::translateCameraAlongUp(float aDelta){
  if(!fCamera) return;

  SbVec3f dir;
  fCamera->orientation.getValue().multVec(SbVec3f(0,0,-1),dir);
  SbVec3f up;
  fCamera->orientation.getValue().multVec(SbVec3f(0,1,0),up);

  SbVec3f pos = fCamera->position.getValue() + aDelta * up;
  setCameraPosition(pos);
}
float SoSDL_Viewer::getDefaultDeltaRotation() const {
  return fDefaultDeltaRotation;
}
void SoSDL_Viewer::setDefaultDeltaRotation(float aValue) {
  fDefaultDeltaRotation = aValue;
}
float SoSDL_Viewer::getDefaultDeltaTranslation() const {
  return fDefaultDeltaTranslation;
}
void SoSDL_Viewer::setDefaultDeltaTranslation(float aValue) {
  fDefaultDeltaTranslation = aValue;
}
void SoSDL_Viewer::checkFloor(SbBool aValue) {fCheckFloor = aValue;}
float SoSDL_Viewer::getFloor() const {return fFloor;}
void SoSDL_Viewer::setFloor(float aValue) {fFloor = aValue;}
void SoSDL_Viewer::setCameraPosition(const SbVec3f& aPos) {
  if(fCheckFloor==FALSE) {
    fCamera->position.setValue(aPos);
  } else {
    if(aPos[1]>fFloor) fCamera->position.setValue(aPos);
  }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void SoSDL_Viewer::updateMotion(SoSDL::Move aWhat) {
  float deltaRot = getDefaultDeltaRotation();
  float deltaTrans = getDefaultDeltaTranslation();

  if(aWhat==SoSDL::ROTATE_RIGHT) {
    rotateCameraAroundUp(-deltaRot);
  } else if(aWhat==SoSDL::ROTATE_LEFT) {
    rotateCameraAroundUp(deltaRot);

  } else if(aWhat==SoSDL::TRANSLATE_RIGHT) {
    translateCameraAlongSide(deltaTrans);
  } else if(aWhat==SoSDL::TRANSLATE_LEFT) {
    translateCameraAlongSide(-deltaTrans);

  } else if(aWhat==SoSDL::UP) {
    translateCameraAlongUp(deltaTrans);
  } else if(aWhat==SoSDL::DOWN) {
    translateCameraAlongUp(-deltaTrans);

  } else if(aWhat==SoSDL::FORWARD) {
    translateCameraAlongDir(deltaTrans);
  } else if(aWhat==SoSDL::BACKWARD) {
    translateCameraAlongDir(-deltaTrans);
  }
}
