#ifndef amanip_atime_h
#define amanip_atime_h 

#ifdef WIN32
#include <winsock.h>
#include <sys/timeb.h>
#else
#include <sys/time.h>
#endif

namespace amanip {

class atime {
public:
  static atime now() {
#ifdef WIN32
    struct timeb tb;
    ::ftime(&tb);
    return atime((unsigned long)tb.time,(unsigned long)(tb.millitm * 1000));
#else
    struct timeval tval;
    ::gettimeofday(&tval,0);
    return atime((unsigned long)tval.tv_sec,(unsigned long)tval.tv_usec);
#endif
  }
  static atime elapsed(const atime& a_from){
    atime elapsed = now();
    elapsed.m_secs -= a_from.m_secs;
    elapsed.m_micro_secs -= a_from.m_micro_secs;
    return elapsed;
  }
public:
  atime(unsigned long a_secs,unsigned long a_micro_secs)
  :m_secs(a_secs),m_micro_secs(a_micro_secs){}
  atime(const atime& a_from)
  :m_secs(a_from.m_secs),m_micro_secs(a_from.m_micro_secs){}
  atime& operator=(const atime& a_from){
    m_secs = a_from.m_secs;
    m_micro_secs = a_from.m_micro_secs;
    return *this;
  }
  virtual ~atime(){}
public:
  unsigned long seconds() const {return m_secs;}
  unsigned long micro_seconds() const {return m_micro_secs;}
  double value() const {
    return ((double)m_secs) + ((double)m_micro_secs)/1000000.0; 
  }
private:
  unsigned long m_secs;
  unsigned long m_micro_secs;
};

}

#ifdef WIN32
#ifdef interface
#undef interface
#endif
#endif

#endif
