
#include <gtk/gtk.h>

#include <gtkgl/gtkglarea.h>

static void paint(unsigned int width,unsigned int height);
static void resize(GtkWidget*,GdkEventConfigure*,void*);
static void expose(GtkWidget*,GdkEventExpose*,void*);
static void OpenGL_DrawCube();

int main(int aArgc,char** aArgv) {
  int glAttributes[16];
  int i = 0;
  GtkWidget* topWidget;
  GtkWidget* glarea;

  gtk_init(&aArgc, &aArgv);
  
  topWidget = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_default_size(GTK_WINDOW(topWidget),600,600);

  g_signal_connect (G_OBJECT(topWidget),"delete_event",
                    G_CALLBACK(gtk_main_quit),NULL);

  printf("%s\n",gdk_gl_get_info());

  glAttributes[i] = GDK_GL_RGBA; i++;
  glAttributes[i] = GDK_GL_DOUBLEBUFFER; i++;
  glAttributes[i] = GDK_GL_DEPTH_SIZE; i++;
  glAttributes[i] = 1; i++;
/*
  glAttributes[i] = GDK_GL_STENCIL_SIZE; i++;
  glAttributes[i] = 1; i++;
*/
  glAttributes[i] = GDK_GL_NONE; i++;
  glarea = gtk_gl_area_new(glAttributes);    

  gtk_container_add(GTK_CONTAINER(topWidget),glarea);

  // To avoid a crash in case doing a
  // gtk_widget_destroy on a container having a GtkGLArea.
  gdk_gl_context_ref(GTK_GL_AREA(glarea)->glcontext);

  gtk_signal_connect(GTK_OBJECT(glarea), "configure_event",
                      GTK_SIGNAL_FUNC(resize),(gpointer)0);
  gtk_signal_connect(GTK_OBJECT(glarea), "expose_event",
                      GTK_SIGNAL_FUNC(expose),(gpointer)0);

  gtk_widget_show_all(topWidget);

  gtk_main ();

  printf("exit...\n");

  return 0;
}

#ifdef WIN32
#include <windows.h>
#endif

#ifdef APPLE_GL
#include <gl.h>
#include <glu.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#endif

#include <stdio.h>

void resize(GtkWidget* aWidget,GdkEventConfigure* aEvent,void* aData){
  if(gtk_gl_area_make_current(GTK_GL_AREA(aWidget))) {
    unsigned int width = 600;
    unsigned int height = 600;
    width = aWidget->allocation.width;
    height = aWidget->allocation.height;
    //printf("debug : %d %d\n",width,height);
    paint(width,height);
    gtk_gl_area_swapbuffers(GTK_GL_AREA(aWidget));
  }
}

void expose(GtkWidget* aWidget,GdkEventExpose* aEvent,void* aData){
  if(gtk_gl_area_make_current(GTK_GL_AREA(aWidget))) {
    unsigned int width = 600;
    unsigned int height = 600;
    width = aWidget->allocation.width;
    height = aWidget->allocation.height;

    //printf("debug : %d %d\n",width,height);
 
    paint(width,height);
    gtk_gl_area_swapbuffers(GTK_GL_AREA(aWidget));
  }
}

void paint(unsigned int width,unsigned int height) {
  GLint viewport[4];
  double aspect;

  glEnable(GL_DEPTH_TEST);
  //glEnable(GL_SCISSOR_TEST);

  glEnable(GL_LINE_STIPPLE);
  glPolygonMode(GL_FRONT,GL_FILL);
  //  So that lightning is correctly computed 
  // when some scaling is applied.
  glEnable(GL_NORMALIZE);

  glShadeModel(GL_FLAT);
  //glEnable(GL_LIGHTING);
  //glDisable(GL_LIGHT0);

  glViewport(0,0,width,height);
  //glScissor(0,0,width,height);

  // In OpenGL, by default,
  // camera is placed at origin, looking at -z, up along y.
  glGetIntegerv(GL_VIEWPORT,viewport);
  aspect = ((double)viewport[2])/((double)viewport[3]);
  glMatrixMode(GL_PROJECTION); 
  glLoadIdentity();
  // height 2.
  glOrtho(-aspect,aspect,-1,1,-100,100);

  // Better to clear after glViewport and glScissor (else problems with Mesa).
  glClearColor(0.8F,0.8F,0.8F,0);
  glClear(GL_COLOR_BUFFER_BIT);
  glClear(GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  //glScaled(1,1,1);
  // Rotate then translate :
  glTranslated(0,0,-10);
  glRotated(30,0,1,0);
  glRotated(30,1,0,0);

  OpenGL_DrawCube();

  glFinish();
}

static void OpenGL_DrawCube(){
  glLineWidth(4);

  // Front :
  glColor3d(1,0,0);    //red
  glBegin(GL_POLYGON);
  glNormal3d(0,0,1);
  glVertex3d( 0.5,-0.5,0.5);
  glVertex3d( 0.5, 0.5,0.5);
  glVertex3d(-0.5, 0.5,0.5);
  glVertex3d(-0.5,-0.5,0.5);
  glEnd();

  // Back :
  glColor3d(0,1,0);   //green
  glBegin(GL_POLYGON);
  glNormal3d(0,0,-1);
  glVertex3d( 0.5,-0.5,-0.5);
  glVertex3d(-0.5,-0.5,-0.5);
  glVertex3d(-0.5, 0.5,-0.5);
  glVertex3d( 0.5, 0.5,-0.5);
  glEnd();

  glColor3d(1,1,0);   //yellow
  glBegin(GL_LINES);
  glVertex3d( 0.5,-0.5, 0.5);
  glVertex3d( 0.5,-0.5,-0.5);
  glVertex3d( 0.5, 0.5, 0.5);
  glVertex3d( 0.5, 0.5,-0.5);
  glVertex3d(-0.5, 0.5, 0.5);
  glVertex3d(-0.5, 0.5,-0.5);
  glVertex3d(-0.5,-0.5, 0.5);
  glVertex3d(-0.5,-0.5,-0.5);
  glEnd();
}
