
#include <Lib/Printer.h>
#include <Lib/Debug.h>
#include <Lib/System.h>
#include <Lib/Arguments.h>
#include <Lib/Out.h>
#include <Lib/dirmanip.h>
#include <Lib/Upgrader.h>

int main (int aArgc,char** aArgv){
  int verboseLevel = 0;

  Lib::Printer printer;

 {// balance block :

  Lib::Arguments args(aArgc,aArgv);
  std::string sverbose;
  if(args.find("-verbose",sverbose)) {
    if(sverbose.empty()) verboseLevel = 1;
  }

  std::vector<std::string> options;
  options.push_back("-h");
  options.push_back("-verbose");
  options.push_back("-orders");

  Lib::Out out(printer);

  if(!args.knownOptions(options)) {
    out << "Lib_upgrade :"
        << " Some options are unknown." << Lib::endl;
    out << "To have help, type :" << Lib::endl;
    out << " OS> Lib_upgrade -h" << Lib::endl;
    return Lib::System::exitSuccess();
  }

  if(args.isAnArgument("-h")) {
    out << "Lib_upgrade :" << Lib::endl;
    out << "Usage :" << Lib::endl;
    out << " OS> Lib_upgrade <options> -orders=<file> <path>" << Lib::endl;
    out << "Options :" << Lib::endl;
    out << " -verbose" << Lib::endl;
    return Lib::System::exitSuccess();
  }

  std::string orders;
  if(!args.find("-orders",orders)) {
    out << "Lib_upgrade :"
        << " no order file given. Take $HOME/Lib.upgrade." 
        << Lib::endl;
    std::string home = Lib::dirmanip::home();
    orders = home+ "/Lib.upgrade";
  }

  std::string path;
  if(!args.file(path)) path = ".";

  Lib::Upgrader upgrader(printer,verboseLevel,orders);
  upgrader.upgrade(path);

  } //end balance block.

  Lib::Debug::balance(printer);

  if(verboseLevel) {
    Lib::Out out(printer);
    out << "Lib_upgrade :"
        << " Exiting..." << Lib::endl;
  }

  return Lib::System::exitSuccess();
}
