#ifndef obuild_Builder_h
#define obuild_Builder_h

#include <string>
#include <vector>

namespace Slash {namespace Core {class IWriter;}}

namespace Lib {class ItemML;}
namespace Lib {class Out;}

namespace obuild {

class Builder {
public:
  Builder(Slash::Core::IWriter&);
public:
  enum eWhat {build_sh,build_bat};
  bool build(const std::string&,eWhat);
  bool query(const std::string&,const std::string&);
  bool cleanup();
public:
  static bool package_and_version(std::string&,std::string&);
  static bool out_default_obuild(Slash::Core::IWriter&,
                           const std::string& aPackage,
                           const std::string& aVersion,
                           const std::string& aFile); 
  static bool findProjectPackages(Slash::Core::IWriter&,
                                  const std::string&,
                                  std::vector<std::string>&,
                                  std::vector<std::string>&);
  bool new_release();
  bool new_release_file_obuild();
  bool handled(const std::string&); //Used by FileGetter.
private: // main producers :
  bool build_pack(Lib::ItemML&);
  bool build_pack_else();
  bool out_build_sh(Lib::ItemML&);
  bool out_build_library_sh(Lib::ItemML&);
  bool out_build_dll_sh(Lib::ItemML&); 
  bool out_build_application_sh(Lib::ItemML&); 
  bool out_build_darwin_app_sh(Lib::ItemML&);
  bool out_setup_sh(Lib::ItemML&); 
  bool out_setup_csh(Lib::ItemML&); 
  bool out_setup_bat(Lib::ItemML&); 
  bool out_broadcast_install_sh(Lib::ItemML&);  
  bool out_build_Darwin_app_sh();  
  bool out_build_Darwin_release_sh();
  bool out_run_sh(Lib::ItemML&);
  bool out_build_project_sh(Lib::ItemML&);
  bool out_build_release_sh(Lib::ItemML&);
  bool out_checkout_release_sh(Lib::ItemML&);
  bool out_iss(Lib::ItemML&);

  bool out_oxxx_cpp(const std::string&,unsigned int,const char*[],
                    bool = false);
  bool out_install_sh();
  bool out_build_sys_sh();
  bool out_build_shlib_sh();
  bool out_check_app_sh();
private:
  bool out_build_xxx_core_sh(Lib::ItemML&,Lib::Out&,
                              const std::string&,
                              const std::string&,
                              const std::vector<std::string>&,
                              const std::vector<std::string>&);
  bool out_build_xxx_end_sh(Lib::ItemML&,Lib::Out&,
                             const std::string&,
                             const std::string&); 
  bool out_load_library_path_sh(Lib::ItemML&,Lib::Out&,bool&);
  bool out_load_library_path_csh(Lib::ItemML&,Lib::Out&,bool&);
  bool out_run_path_sh(Lib::ItemML&,Lib::Out&,bool&);
  bool out_run_path_csh(Lib::ItemML&,Lib::Out&,bool&);
  bool out_path_append_sh(Lib::ItemML&,Lib::Out&,bool&);
  bool out_path_append_csh(Lib::ItemML&,Lib::Out&,bool&);
  bool out_append(Lib::Out&,const std::string&,
                             const std::string&,
                             bool first,
                             bool dirSep = false) const;
  bool out_append_csh(Lib::Out&,const std::string&,
                                const std::string&,
                                bool first,
                                bool dirSep = false) const;
  bool out_if_done(Lib::ItemML&,Lib::Out&); 
  bool out_commands(Lib::ItemML&,Lib::Out&);
  bool out_check_foreigns(Lib::ItemML&,Lib::Out&);
  bool out_check_env(Lib::ItemML&,Lib::Out&);
  bool out_path_of_used_packages(Lib::ItemML&,Lib::Out&);
  bool out_variables(Lib::ItemML&,Lib::Out&); 
  bool out_variables_csh(Lib::ItemML&,Lib::Out&); 
  bool out_var_pwd(Lib::Out&,const std::string&);
  bool out_begin_pwd(Lib::Out&);
  bool out_end_pwd(Lib::Out&);
  bool out_begin_platform(Lib::Out&,const std::string&,const std::string&); 
  bool out_else_platform(Lib::Out&);
  void out_end_platform(Lib::Out&,const std::string&); 
  bool out_begin_platform_csh(Lib::Out&,const std::string&); 
  void out_else_platform_csh(Lib::Out&); 
  void out_end_platform_csh(Lib::Out&); 
  bool out_platform_signature(Lib::Out&);
  bool out_begin_if(Lib::Out&,const std::string&,
                    const std::string&,
                    const std::string& = "",
                    bool a_csh = false); 
  bool out_begin_if_env(Lib::Out&,const std::string&,
                        const std::string&,
                        const std::string&,
                        bool a_csh,bool& a_treated); 
  bool out_begin_if_exist(Lib::Out&,const std::string&,
                        const std::string&,
                        const std::string&,
                        bool a_csh,bool& a_treated); 
  void out_end_if(Lib::Out&,const std::string&,bool a_csh = false); 
  void out_begin_group(Lib::Out&,const std::string&);
  void out_end_group(Lib::Out&);
  void out_begin_main(Lib::Out&,const std::string&);
  void out_end_main(Lib::Out&);
  bool out_linked_libs(Lib::ItemML&,Lib::Out&);
  bool out_linked_jars(Lib::ItemML&,Lib::Out&,bool&);
  bool out_incs(Lib::ItemML&,const std::string&,Lib::Out&);
  bool out_compilation_macros(Lib::ItemML&,const std::string&,Lib::Out&);
  bool out_options(Lib::ItemML&,const std::string&,Lib::Out&);
  bool out_paths_csh(const std::vector<std::string> aPacks,
                     const std::vector<std::string> aVerss,
                     Lib::Out& aOut); 
  bool out_paths_sh(const std::vector<std::string> aPacks,
                    const std::vector<std::string> aVerss,
                    Lib::Out& aOut);
  bool out_path_sh(const std::string&,const std::string&,Lib::Out& aOut);
  bool out_begin_tool(const std::vector<std::string>&,Lib::Out&);
  bool out_begin_tool(const std::string&,Lib::Out&);
  bool out_tool_item(Lib::ItemML&,
                      const std::string&,
                      const std::string&,
                      bool,
                      Lib::Out&);
  bool out_compile_file(const std::string&,
                        const std::string&,
                        Lib::Out&,bool&);
  bool out_wrap(Lib::ItemML&,Lib::Out&);
  //bool out_begin_build(Lib::ItemML&,Lib::Out&);
  bool out_pre_build(Lib::ItemML&,Lib::Out&);
  bool out_post_build(Lib::ItemML&,Lib::Out&);
  bool out_dll_bat(Lib::Out&,const std::string&,const std::string&,
                   bool = true);
  bool out_template(Lib::Out&,unsigned int,const char* []);
  bool out_template2(Lib::Out&,unsigned int,const char* [],bool = false);

  void out_verbose_on(Lib::Out&,unsigned int = 0);
  void out_verbose_off(Lib::Out&,unsigned int = 0);
  void out_empty(Lib::Out&);
  void out_status(Lib::Out&,unsigned int = 0);
  bool out_check_foreigns_sh(Lib::ItemML&);

  typedef std::vector< std::pair<std::string,std::string> > Subs;

  bool out_save_sh(Lib::Out&,const std::string&,const Subs&);
  bool out_save_bat(Lib::Out&,const std::string&,const Subs&);
  bool out_darwin_app_sh(Lib::ItemML&,Lib::Out&);
  bool out_rmdir(Lib::Out&,const std::string&);
  bool out_mkdir(Lib::Out&,const std::string&);
  bool out_rm(Lib::Out&,const std::string&);
  bool out_cd(Lib::Out&,const std::string&);
  bool out_copy(Lib::Out&,const std::string&,const std::string&,const std::string& options = "");
  bool out_copy_check(Lib::Out&,const std::string&,const std::string&);
  bool out_xcopy(Lib::Out&,const std::string&,const std::string&);
  bool out_move(Lib::Out&,const std::string&,const std::string&);
  void out_cd_obuild(Lib::Out&); 
  void out_cd_bin(Lib::Out&); 
  void out_if_component_stamp(const std::string&,
                              const std::string&,
                              Lib::Out&);
  void out_if_element_stamp(const std::string&,
                            const std::string&,
                            Lib::Out&);
  bool out_release_sh(Lib::Out&,Lib::ItemML&,const std::string&,bool&);
private: //does not produce sh or bat code.
  static std::vector<std::string> tags();
  static std::string path_variable(const std::string&);
  std::string dir_pack(const std::string&);
  std::string dir_bin_variable();
  std::string dir_bin(const std::string&);
  std::string dir_obuild(const std::string&); 
  std::string var_val(const std::string&); 
  std::string cmt(); 
  std::string sep();
  std::string path_sep();
  std::string s_obj();
  bool check_directory(const std::string&); 
  bool getProperties(Lib::ItemML&,
                     const std::string& aProperty,
                     std::vector<std::string>& aPacks,
                     std::vector<std::string>& aVerss,
                     std::vector<std::string>& aPlatforms,
                     std::vector<std::string>& aVerbatims,
                     std::vector<std::string>& aNames,
                     std::vector<std::string>& aIfs,
                     std::vector<std::string>& aWhats,
                     std::vector<std::string>& aGroups,
                     std::vector<std::string>& aValues,
                     bool aCheckFromUse); 
  bool getFiles(Lib::ItemML&,
                std::vector<std::string>&,
                std::vector<std::string>&,bool check_handled = true);
  bool handle_use(Lib::ItemML&); 
  bool handle_version(Lib::ItemML&); 
  bool handle_template(Lib::ItemML&);
  bool rm(const std::string&,const std::string&);

  bool queryByTags(Lib::ItemML&,const std::string&);
  bool query(Lib::ItemML&);
  bool doIt(Lib::ItemML&);
  bool parse_tool(const std::string&,std::string&) const;
  bool guess_linker_language(Lib::ItemML&,
                             const std::vector<std::string>&,
                             std::string&);
  bool done_on(Lib::ItemML&,bool&);
  std::string read_include(unsigned int,const char*[]);
  bool handle_tools(Lib::ItemML&);
  bool replace_at_at(std::string&); 
  void reset_local_flags(const std::string&);
  Lib::ItemML* get_parent_package_ML(Lib::ItemML&);
  std::string if_count();
  std::string exe_count();
  bool check_name(const std::string&);
  bool check_env(Lib::ItemML&); 
  bool build_on_platform(const std::string&,bool&);
private:
  static bool default_obuild(const std::string&,
                             const std::string&,std::string&);
  static std::string towin(const std::string&);
  static std::string tounix(const std::string&);
  static std::string esc_quote(const std::string&);
  static std::string no_quote(const std::string&);
  static std::string quote(const std::string&);
private:
  Slash::Core::IWriter& fPrinter;
  Lib::ItemML* fTopML;
  Lib::ItemML* fPackML;
  std::string fPackage;
  std::string fVersion;
  bool fProduced_setups;
  bool fProduced_platform_exe;
  bool fProduced_find_exe;
  bool fProduced_windef_exe;
  bool fProduced_subs_exe;
  bool fProduced_subf_exe;
  bool fProduced_wrap_exe;
  bool fProduced_Darwin_app;
  bool fProduced_build_sys_sh;
  bool fProduced_build_Darwin_release_sh;
  bool fProduced_build_shlib_sh;
  bool fProduced_check_app_sh;
  std::string fProduced_local_what;
  bool fProduced_local_platform_exe;
  bool fProduced_local_find_exe;
  bool fProduced_local_windef_exe;
  bool fProduced_local_subs_exe;
  bool fProduced_local_subf_exe;
  bool fProduced_local_wrap_exe;
  bool fProduced_local_LD_PATH;
  std::string fProduce_local_exit_sh;
  int fProduced_local_pwd_count;
  int fProduced_local_group_count;
  bool fProduced_local_group_begin;
  int fProduced_local_if_count;
  int fProduced_local_exe_count;
  bool fProduced_local_if_begin;
  int fProduced_local_main_count;
  bool fProduced_local_main_begin;
  eWhat fProdWhat;
  std::string f_sh;
  std::string f_dir;
  std::string f_sep;
  std::string f_cmt;
  std::string f_exe;
  std::string fComponent;
  std::string fCopyRight;
  std::string fPublisher;
  std::string fURL;
};

}

#endif
