#!/bin/sh -f

# File produced by the obuild tool.

#////////////////////////////////////////////////
# Usage :
# ( UNIX> chmod a+x build_obuild.sh )
#   UNIX> ./build_obuild.sh
#
#////////////////////////////////////////////////

verbose_build=no
debug=no
optimize=yes
set_x=no
while test $# -ge 1 ; do
  case $1 in
    -h) 
        echo "Options :"
        echo "  -v : verbose."
        echo "  -d : build in debug mode."
        echo "  -g : build in debug mode."
        echo "  -x : execute script with set -x."
        echo ""
        exit ;;
    -v) verbose_build=yes ;;
    -x) set_x=yes ;;
    -d) debug=yes;optimize=no ;;
    -g) debug=yes;optimize=no ;;
    *) echo "unknown option : $1";exit ;;
   esac
  shift
done

if [ ${set_x} == yes ] ; then
  set -x
fi

if [ `uname` == Darwin ] ; then
  platform=Darwin
  cc=cc
  cxx=c++
  f77=g77
  ar=ar
elif [ `uname` == Linux ] ; then
  platform=Linux
  cc=cc
  cxx=c++
  f77=f77
  ar=ar
else
  platform=""
  cc=""
  cxx=""
  f77=""
  ar=""
fi

if [ "${platform}" == "" ] ; then
  echo "platform variable not defined."
  exit
fi

if [ "${cc}" == "" ] ; then
  echo "cc variable not defined."
  exit
fi
if [ "${cxx}" == "" ] ; then
  echo "cxx variable not defined."
  exit
fi
if [ "${f77}" == "" ] ; then
  echo "f77 variable not defined."
  exit
fi

dir_bin=../${platform}_obuild
# Create the bin directory :
/bin/mkdir -p ${dir_bin}

cc_debug=""
cc_link_debug=""
cxx_debug=""
cxx_link_debug=""
f77_debug=""
f77_link_debug=""
if [ ${debug} == yes ] ; then
  cc_debug=-g
  cc_link_debug=-g
  cxx_debug=-g
  cxx_link_debug=-g
  f77_debug=-g
  f77_link_debug=-g
fi

cc_optimize=""
cc_link_optimize=""
cxx_optimize=""
cxx_link_optimize=""
f77_optimize=""
f77_link_optimize=""
if [ ${optimize} == yes ] ; then
  cc_optimize=-O
  cc_link_optimize=-O
  cxx_optimize=-O
  cxx_link_optimize=-O
  f77_optimize=-O
  f77_link_optimize=-O
fi

if ! test -e './cpp/obuild_platform.cpp' ; then
  echo "Program source obuild_platform.cpp not found."
  exit
fi
obuild_platform_exe="${dir_bin}/obuild_platform.exe"
if ! test -e "${obuild_platform_exe}" ; then
  if [ "${cxx}" == "" ] ; then cxx=c++; fi
  ${cxx} -o "${obuild_platform_exe}" ./cpp/obuild_platform.cpp
  if ! test -e "${obuild_platform_exe}" ; then
    echo "Program obuild_platform.exe not found."
    exit
  fi
fi


cc_macros=""

cxx_macros=""

f77_macros=""

cc_options=""

cxx_options=""

f77_options=""

cc_link_options=""

cxx_link_options=""

f77_link_options=""


cc_includes="-I../source -I../source/expat -I../source/expat/private"

cxx_includes="-I../source -I../source/expat -I../include"

f77_includes=""

/bin/rm -R -f ${dir_bin}/app_obuild
/bin/mkdir -p ${dir_bin}/app_obuild

# pre_exec scripts :
../tools/wrap.sh

# Compile obuild files :
if [ ${verbose_build} == yes ] ; then set -x;fi
${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/obuild.o ../applications/obuild.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/Builder.o ../source/Builder.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/CodeGen.o ../source/CodeGen.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cc} -c ${cc_debug} ${cc_optimize} ${cc_options} ${cc_macros} ${cc_includes} -o ${dir_bin}/app_obuild/dllmain.o ../source/expat/dllmain.c
status=$?;if test $status != 0 ; then exit $status;fi

${cc} -c ${cc_debug} ${cc_optimize} ${cc_options} ${cc_macros} ${cc_includes} -o ${dir_bin}/app_obuild/hashtable.o ../source/expat/hashtable.c
status=$?;if test $status != 0 ; then exit $status;fi

${cc} -c ${cc_debug} ${cc_optimize} ${cc_options} ${cc_macros} ${cc_includes} -o ${dir_bin}/app_obuild/xmlparse.o ../source/expat/xmlparse.c
status=$?;if test $status != 0 ; then exit $status;fi

${cc} -c ${cc_debug} ${cc_optimize} ${cc_options} ${cc_macros} ${cc_includes} -o ${dir_bin}/app_obuild/xmlrole.o ../source/expat/xmlrole.c
status=$?;if test $status != 0 ; then exit $status;fi

${cc} -c ${cc_debug} ${cc_optimize} ${cc_options} ${cc_macros} ${cc_includes} -o ${dir_bin}/app_obuild/xmltok.o ../source/expat/xmltok.c
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/Arguments.o ../source/Lib/Arguments.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/BaseML.o ../source/Lib/BaseML.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/cmanip.o ../source/Lib/cmanip.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/Debug.o ../source/Lib/Debug.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/dirmanip.o ../source/Lib/dirmanip.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/Element.o ../source/Lib/Element.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/FileReader.o ../source/Lib/FileReader.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/FileWriter.o ../source/Lib/FileWriter.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/fmanip.o ../source/Lib/fmanip.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/ItemML.o ../source/Lib/ItemML.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/Out.o ../source/Lib/Out.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/Printer.o ../source/Lib/Printer.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/smanip.o ../source/Lib/smanip.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/sout.o ../source/Lib/sout.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/System.o ../source/Lib/System.cxx
status=$?;if test $status != 0 ; then exit $status;fi

${cxx} -c ${cxx_debug} ${cxx_optimize} ${cxx_options} ${cxx_macros} ${cxx_includes} -o ${dir_bin}/app_obuild/ZipReader.o ../source/Lib/ZipReader.cxx
status=$?;if test $status != 0 ; then exit $status;fi

if [ ${verbose_build} == yes ] ; then set +x;fi

linked_libraries=""

# Link obuild :
objs=`find ${dir_bin}/app_obuild -name '*.o' -print`
/bin/rm -f ${dir_bin}/obuild.exe
if [ ${verbose_build} == yes ] ; then set -x;fi
${cxx} ${cxx_link_debug} ${cxx_link_optimize} ${cxx_link_options} -o ${dir_bin}/obuild.exe ${objs} ${linked_libraries}
status=$?;if test $status != 0 ; then exit $status;fi
if [ ${verbose_build} == yes ] ; then set +x;fi

/bin/rm -R -f ${dir_bin}/app_obuild
obuild_find_exe="${dir_bin}/obuild_find.exe"
/bin/rm -R -f ${obuild_find_exe}

