// this :
#include <Lib/GZipReader.h>

#include <Lib/Out.h>
#include <Lib/Debug.h>

#ifdef LIB_HAS_ZLIB
#include <zlib.h>
#endif

//////////////////////////////////////////////////////////////////////////////
bool Lib::GZipReader::available(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  return true;
#else
  return false;
#endif
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Lib::GZipReader::GZipReader(
 const std::string& aFileName
,Slash::Core::IWriter& aPrinter
)
:fPrinter(aPrinter)
,fFileName(aFileName)
,fFile(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  Lib::Debug::increment("Lib::GZipReader");
#ifdef LIB_HAS_ZLIB
  fFile = ::gzopen(aFileName.c_str(),"rb");
#else
  Lib::Out out(fPrinter);
  out << "Lib::GZipReader::GZipReader :"
      << " Lib package not reconstructed with zlib." << Lib::endl;
#endif
}
//////////////////////////////////////////////////////////////////////////////
Lib::GZipReader::~GZipReader(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  if(fFile) ::gzclose((gzFile*)fFile);
#endif
  Lib::Debug::decrement("Lib::GZipReader");
}
//////////////////////////////////////////////////////////////////////////////
void Lib::GZipReader::disable(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  if(!fFile) return; // Already closed.
  ::gzclose((gzFile*)fFile);
  fFile = 0;
#endif
}
//////////////////////////////////////////////////////////////////////////////
void Lib::GZipReader::enable(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  if(fFile) return; // Already open.
  fFile = ::gzopen(fFileName.c_str(),"rb");
#endif
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::GZipReader::enabled(
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return fFile ? true : false;
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::GZipReader::rewind(
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  if(!fFile) return false;
  ::gzrewind(fFile);
  return true;
#else
  return false;
#endif
}

//////////////////////////////////////////////////////////////////////////////
bool Lib::GZipReader::read(
 void* aBuffer
,unsigned int aLength
,int& aReadLength
) 
//////////////////////////////////////////////////////////////////////////////
// aReadLength = 0 when finished (end of file).
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aReadLength = 0;
#ifdef LIB_HAS_ZLIB
  if(!fFile) return false;
  int len = ::gzread(fFile, aBuffer, aLength);
  if(len==(-1)) {
    Lib::Out out(fPrinter);
    out << "Lib::GZipReader::read :"
        << " error reading in file \"" << fFileName 
        << "\"." << Lib::endl;
    disable();
    return false;
  } else {
    aReadLength = len;
    return true;
  }
#else
  return false;
#endif
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::GZipReader::read(
 void* aBuffer
,unsigned int aLength
,std::string& aString
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  aString = "";
#ifdef LIB_HAS_ZLIB
  if(!fFile) return false;
  char* buffer = ::gzgets(fFile, (char*)aBuffer, aLength);
  if(buffer==Z_NULL) {
    if(gzeof(fFile)==0) { //error
      Lib::Out out(fPrinter);
      out << "Lib::GZipReader::read :"
          << " error reading in file \"" << fFileName 
          << "\"." << Lib::endl;
    }
    disable();
    return false;
  } else {
    size_t l = ::strlen(buffer);
    //  On Windows, editors when saving binary files,
    // put \r\n at place of \n ; we then look for \r\n.
    if( (l>=2) && (buffer[l-2]=='\r') && (buffer[l-1]=='\n') ) {
      buffer[l-2] = '\0';
    } else if( (l>=1) && (buffer[l-1]=='\n') ) {
      buffer[l-1] = '\0';
    }
    aString = buffer;
    return true;
  }
#else
  return false;
#endif
}
//////////////////////////////////////////////////////////////////////////////
bool Lib::GZipReader::sread(
 void* aBuffer
,unsigned int aLength
,unsigned int& aStringLength
) 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
#ifdef LIB_HAS_ZLIB
  if(!fFile) {
    aStringLength = 0;
    return false;
  }
  char* buffer = ::gzgets(fFile, (char*)aBuffer, aLength);
  if(buffer==Z_NULL) {
    if(gzeof(fFile)==0) { //error
      Lib::Out out(fPrinter);
      out << "Lib::GZipReader::read :"
          << " error reading in file \"" << fFileName 
          << "\"." << Lib::endl;
    }
    disable();
    aStringLength = 0;
    return false;
  } else {
    size_t l = ::strlen(buffer);
    //  On Windows, editors when saving binary files,
    // put \r\n at place of \n ; we then look for \r\n.
    if( (l>=2) && (buffer[l-2]=='\r') && (buffer[l-1]=='\n') ) {
      buffer[l-2] = '\0';
      l -= 2;
    } else if( (l>=1) && (buffer[l-1]=='\n') ) {
      buffer[l-1] = '\0';
      l -= 1;
    }
    aStringLength = l;
    return true;
  }
#else
  aStringLength = 0;
  return false;
#endif
}
