
static bool Lib_fmanip_replace_if(
 const std::string& aName
,const std::string& aIf
,const std::string& aOld
,const std::string& aNew
,std::vector<std::string>& aChanges
){
  aChanges.clear();
  std::vector<std::string> text;
  if(!Lib::fmanip::read(aName,text)) return false;
  // Replace :
  bool touched = false;
 {unsigned int textn = text.size();
  for(unsigned index=0;index<textn;index++) {
    std::string old = text[index];
    if(old.find(aIf)!=std::string::npos) {
      if(Lib::smanip::replace(text[index],aOld,aNew)) {
        //printf("debug :\"%s\"\n",text[index].c_str());
        aChanges.push_back(old);
        aChanges.push_back(text[index]);
        touched = true;
      }
    }
  }}
  if(!touched) return true;
  // Create temporary file :
  std::string tmpName;
  if(!Lib::fmanip::tmpname(tmpName)) return false;
  FILE* file = ::fopen(tmpName.c_str(),"wb");
  if(!file) return false;
 {unsigned int textn = text.size();
  for(unsigned index=0;index<textn;index++) {
    if(::fprintf(file,"%s\n",text[index].c_str())<0) {
      ::fclose(file);
      Lib::fmanip::remove(tmpName);
      return false;
    }
  }}
  ::fclose(file);
  if(!Lib::fmanip::remove(aName)) return false;
  if(!Lib::fmanip::rename(tmpName,aName)) return false;
  return true;
}

bool obuild::Builder::new_release() {
  std::string file("new_release");
  std::vector<std::string> lines;
  if(!Lib::fmanip::read(file,lines)) {
    Lib::Out out(fPrinter);
    out << "obuild::Builder::new_release :"
        << " can't read " << Lib::sout(file)
        << Lib::endl;
    return false;
  }

  for(std::vector<std::string>::iterator it = lines.begin();
      it!=lines.end();++it) {
    if((*it).empty()) continue;
    if((*it).size()&&(*it)[0]=='#') continue;
    std::vector<std::string> words = Lib::smanip::words(*it," ");
    if((words.size()!=3)&&(words.size()!=4)) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::new_release :"
          << " bad number of words in " << Lib::sout(*it)
          << Lib::endl;
      return false;
    }
    const std::string& spack = words[0];
    const std::string& sold = words[1];
    const std::string& snew = words[2];
    if(sold==snew) continue;
    std::string sversion;
    if(words.size()==4) sversion = words[3];

    std::string sdir = "."+f_sep+spack;
    if(!Lib::dirmanip::go(sdir)) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::new_release :"
          << " can't go in " << Lib::sout(sdir) << " directory."
          << Lib::endl;
      return false;
    }

    if(!Lib::dirmanip::go("obuild")) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::new_release :"
          << " can't go in the obuild directory"
          << " of the package " << Lib::sout(spack) << "."
          << Lib::endl;
      return false;
    }

    ///////////////////////
    // treat this.obuild :
    ///////////////////////
   {std::vector<std::string> lines2;
    if(!Lib::fmanip::read("this.obuild",lines2)) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::new_release :"
          << " can't read this.obuild."
          << Lib::endl;
      return false;
    }

    std::vector<std::string> new_text;
    bool pack_found = false;
    for(std::vector<std::string>::iterator it2 = lines2.begin();
        it2!=lines2.end();++it2) {
      const std::string& line = (*it2);
      
      bool found = false;
      // <package, <project or <reference :
      if( (line.find("<package")!=std::string::npos) ||
          (line.find("<project")!=std::string::npos) ||
          (line.find("<reference")!=std::string::npos) ){
        std::string::size_type pos = line.find("=\""+spack+"\"");
        if(pos!=std::string::npos) {
          pos = line.find(sold,pos);
          if(pos!=std::string::npos) {
            std::string s(line);
            Lib::smanip::replace(s,sold,snew);
            new_text.push_back(s);
            found = true;
            pack_found = true;
          }
        }
      }
      if(!found) new_text.push_back(line);
    }
    if(!pack_found) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::new_release :"
          << " can't find <package> item in this.obuild "
          << " of package " << Lib::sout(spack) << "."
          << Lib::endl;
      return false;
    }

    if(!Lib::fmanip::write("this.obuild",new_text)) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::new_release :"
          << " can't write this.obuild " 
          << " of package " << Lib::sout(spack) << "."
          << Lib::endl;
      return false;
    } else {
      Lib::Out out(fPrinter);
      out << "this.obuild of package " << Lib::sout(spack) << " done." 
          << Lib::endl;
    }}

    /////////////////////////////
    // treat a Version.h if any :
    /////////////////////////////
    if(sversion.empty()) {
      sversion = ".."+f_sep+spack+f_sep+"Version.h";
      if(!Lib::fmanip::exists(sversion)) {
        sversion = ".."+f_sep+"include"+f_sep+spack+f_sep+"Version.h";
        if(!Lib::fmanip::exists(sversion)) {
          sversion = ".."+f_sep+"include"+f_sep+spack+f_sep+"SbVersion.h";
          if(!Lib::fmanip::exists(sversion)) {
            sversion = "";
          }
        }
      }
    }

    if(sversion.size()) {

      std::vector<unsigned int> old_verss;
      if(!Lib::smanip::version(sold,old_verss) || (old_verss.size()!=3)) {
        Lib::Out out(fPrinter);
        out << "obuild::Builder::new_release :"
            << " " << Lib::sout(sold) << " not a version."
            << Lib::endl;
        return false;
      }

      std::vector<unsigned int> new_verss;
      if(!Lib::smanip::version(snew,new_verss) || (new_verss.size()!=3)) {
        Lib::Out out(fPrinter);
        out << "obuild::Builder::new_release :"
            << " " << Lib::sout(snew) << " not a version."
            << Lib::endl;
        return false;
      }

      std::vector<std::string> lines2;
      if(!Lib::fmanip::read(sversion,lines2)) {
        Lib::Out out(fPrinter);
        out << "obuild::Builder::new_release :"
            << " can't read " << Lib::sout(sversion) << "."
            << Lib::endl;
        return false;
      }
  
      std::vector<std::string> new_text;
      for(std::vector<std::string>::iterator it2 = lines2.begin();
          it2!=lines2.end();++it2) {
        const std::string& line = (*it2);
        if(line.find("MAJOR_VERSION")!=std::string::npos) {
          std::string s(line);
          Lib::smanip::replace(s,Lib::smanip::tostring(old_verss[0]),
                                 Lib::smanip::tostring(new_verss[0]));
          new_text.push_back(s);
        } else if(line.find("MINOR_VERSION")!=std::string::npos) {
          std::string s(line);
          Lib::smanip::replace(s,Lib::smanip::tostring(old_verss[1]),
                                 Lib::smanip::tostring(new_verss[1]));
          new_text.push_back(s);
        } else if(line.find("PATCH_VERSION")!=std::string::npos) {
          std::string s(line);
          Lib::smanip::replace(s,Lib::smanip::tostring(old_verss[2]),
                                 Lib::smanip::tostring(new_verss[2]));
          new_text.push_back(s);
        } else if(line.find("VERSION_VRP")!=std::string::npos) {
          std::string s(line);
          Lib::smanip::replace(s,sold,snew);
          new_text.push_back(s);
        } else if(line.find("VERSION ")!=std::string::npos) {   
          std::string old_dot = 
            Lib::smanip::tostring(old_verss[0])+"."+
            Lib::smanip::tostring(old_verss[1])+"."+
            Lib::smanip::tostring(old_verss[2]);
          std::string new_dot = 
            Lib::smanip::tostring(new_verss[0])+"."+
            Lib::smanip::tostring(new_verss[1])+"."+
            Lib::smanip::tostring(new_verss[2]);
          std::string s(line);
          Lib::smanip::replace(s,old_dot,new_dot);
          new_text.push_back(s);
        } else {
          new_text.push_back(line);
        }
      }
  
      if(!Lib::fmanip::write(sversion,new_text)) {
        Lib::Out out(fPrinter);
        out << "obuild::Builder::new_release :"
            << " can't write " << Lib::sout(sversion)
            << " of package " << Lib::sout(spack) << "."
            << Lib::endl;
        return false;
      } else {
        Lib::Out out(fPrinter);
        out << sversion << " of package " << Lib::sout(spack) << " done." 
            << Lib::endl;
      }
    }

    /////////////////////////////
    // treat a Doxyfile if any :
    /////////////////////////////
    std::string Doxyfile = ".."+f_sep+"Doxygen"+f_sep+"Doxyfile";
    if(Lib::fmanip::exists(Doxyfile)) {
      std::string sif = "PROJECT_NUMBER";
      std::vector<std::string> changes;
      if(!Lib_fmanip_replace_if(Doxyfile,sif,sold,snew,changes)) {
        Lib::Out out(fPrinter);
        out << "obuild::Builder::new_release :"
            << " can't replace in " << Lib::sout(Doxyfile) << " ."
            << Lib::endl;
        return false;
      } else {
        Lib::Out out(fPrinter);
        out << " Doxyfile of package " << Lib::sout(spack) << " done." 
            << Lib::endl;
      }
    }

    if(!Lib::dirmanip::go(".."+f_sep+"..")) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::new_release :"
          << " can't go in ../.."
          << Lib::endl;
      return false;
    }

  }

  return true;
}

//FIXME : rm ?
bool obuild::Builder::new_release_file_obuild() {
  std::string file("new_release");
  std::vector<std::string> lines;
  if(!Lib::fmanip::read(file,lines)) {
    Lib::Out out(fPrinter);
    out << "obuild::Builder::new_release :"
        << " can't read " << Lib::sout(file)
        << Lib::endl;
    return false;
  }

  for(std::vector<std::string>::iterator it = lines.begin();
      it!=lines.end();++it) {
    if((*it).empty()) continue;
    if((*it).size()&&(*it)[0]=='#') continue;
    std::vector<std::string> words = Lib::smanip::words(*it," ");
    if((words.size()!=3)&&(words.size()!=4)) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::new_release :"
          << " bad number of words in " << Lib::sout(*it)
          << Lib::endl;
      return false;
    }
    const std::string& spack = words[0];
    const std::string& sold = words[1];
    const std::string& snew = words[2];
    if(sold==snew) continue;

    std::string sif = "package=\""+spack+"\"";
    std::vector<std::string> changes;

    /*

   {std::string file("osc_obuild/obuild/this.obuild");
    if(!Lib_fmanip_replace_if(file,sif,sold,snew,changes)) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::new_release :"
          << " can't replace in " << Lib::sout(file) << " ."
          << Lib::endl;
      return false;
    }}
   {std::string file("osc_batch/obuild/this.obuild");
    if(!Lib_fmanip_replace_if(file,sif,sold,snew,changes)) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::new_release :"
          << " can't replace in " << Lib::sout(file) << " ."
          << Lib::endl;
      return false;
    }}
   {std::string file("osc_batch/obuild/batch.obuild");
    if(!Lib_fmanip_replace_if(file,sif,sold,snew,changes)) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::new_release :"
          << " can't replace in " << Lib::sout(file) << " ."
          << Lib::endl;
      return false;
    }}

   {std::string file("osc_vis/obuild/this.obuild");
    if(!Lib_fmanip_replace_if(file,sif,sold,snew,changes)) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::new_release :"
          << " can't replace in " << Lib::sout(file) << " ."
          << Lib::endl;
      return false;
    }}
   {std::string file("osc_vis/obuild/vis.obuild");
    if(!Lib_fmanip_replace_if(file,sif,sold,snew,changes)) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::new_release :"
          << " can't replace in " << Lib::sout(file) << " ."
          << Lib::endl;
      return false;
    }}
   {std::string file("osc_vis/obuild/g4.obuild");
    if(!Lib_fmanip_replace_if(file,sif,sold,snew,changes)) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::new_release :"
          << " can't replace in " << Lib::sout(file) << " ."
          << Lib::endl;
      return false;
    }}
    */

  }
  return true;
}

