bool obuild::Builder::out_build_release_sh(Lib::ItemML& aProjectML){
  std::string pj_name;
  if(!aProjectML.attributeValue("name",pj_name)) {
    Lib::Out out(fPrinter);
    out << "obuild::Builder::out_build_release_sh :"
        << " no attribute name on the <project> XML item."
        << Lib::endl;
    return false;
  }

  std::string pj_version;
  if(!aProjectML.attributeValue("version",pj_version)) {
    Lib::Out out(fPrinter);
    out << "obuild::Builder::out_build_release_sh :"
        << " no attribute version on the <project> XML item."
        << Lib::endl;
    return false;
  }

  if(!check_name(pj_name)) {
    Lib::Out out(fPrinter);
    out << "obuild::Builder::out_build_release_sh :"
        << " problem with " << Lib::sout(pj_name)
        << ". Project name should have the syntax of a programming variable." 
        << Lib::endl;
    return false;
  }

  fPackage = pj_name;
  fVersion = pj_version;

/*
  if(fProdWhat==obuild::Builder::build_bat) {
    //copy cygwin_zip.exe under obuild/DOS.
  }
*/

  std::string sfile = "."+f_sep+f_dir+f_sep+"build_release"+f_sh;
  Lib::FileWriter fp(sfile,fPrinter);
  if(!fp.enabled()) {
    Lib::Out out(fPrinter);
    out << "obuild::Builder::out_build_release_sh :" 
        << " can't open file " << Lib::sout(sfile) 
        << Lib::endl;
    return false;
  }
  Lib::Out out(fp);

  reset_local_flags("build_release_sh");

  if(fProdWhat==obuild::Builder::build_sh) {
    out_template(out,build_release_header_sh_linen,
                     build_release_header_sh_lines);
  } else if(fProdWhat==obuild::Builder::build_bat) {
    out_template(out,build_release_header_bat_linen,
                     build_release_header_bat_lines);
  }

  // Build variables (especially the ones in foreigns :
  if(!out_variables(aProjectML,out)) {
    fp.disable();
    Lib::fmanip::remove(sfile);
    return false;
  }

  std::string dbin = var_val(dir_bin_variable());

  out_begin_pwd(out);

  out_cd_bin(out);
  out_rmdir(out,pj_name);
  out_mkdir(out,pj_name);
  out << "cd " << pj_name << Lib::endl;

  out_mkdir(out,pj_version);
  out << "cd " << pj_version << Lib::endl;

  out_mkdir(out,"lib");
  out_mkdir(out,"bin");
  out_mkdir(out,"Resources");
  out_var_pwd(out,"obuild_dir_project");

  //out_end_pwd(out);

  out << Lib::endl;

  bool done_something = false;

  std::string rel_dir = 
    quote(var_val("obuild_dir_project"));

  std::string rel_dir_lib = 
    quote(var_val("obuild_dir_project")+sep()+"lib"+sep()+".");

  std::string rel_dir_bin = 
    quote(var_val("obuild_dir_project")+sep()+"bin"+sep()+".");

  // save version of <package> in a Versions file :
 {std::string file = 
    quote(var_val("obuild_dir_project")+sep()+"Versions");

  out_rm(out,file);

 {std::vector<Lib::ItemML*> list;
  Lib::ItemML::collectByTag(aProjectML,"package",list);
  std::vector<Lib::ItemML*>::iterator it;
  for(it=list.begin();it!=list.end();++it) {
    Lib::ItemML* releaseML = (*it)->findByTag("release");
    if(!releaseML) continue;

    std::string pack_name;
    (*it)->attributeValue("name",pack_name);
    std::string pack_version;
    (*it)->attributeValue("version",pack_version);

    //printf("debug : %s : %s\n",pack_name.c_str(),pack_version.c_str());

    std::string s = "package " + pack_name + " " + pack_version;

    if(fProdWhat==obuild::Builder::build_sh) {
      out << "echo \"" << s << "\" >> " << file << Lib::endl;
    } else if(fProdWhat==obuild::Builder::build_bat) {
      out << "ECHO "   << s << "   >> " << file << Lib::endl;
    }

  }}


 {std::vector<Lib::ItemML*> list;
  Lib::ItemML::collectByTag(aProjectML,"foreign",list);
  Lib::ItemML::unique(list);
  std::vector<Lib::ItemML*>::iterator it;
  for(it=list.begin();it!=list.end();++it) {
    Lib::ItemML* releaseML = (*it)->findByTag("release");
    if(!releaseML) continue;

    std::string name;
    (*it)->attributeValue("name",name);

    GETPROP(*(*it),"native_version",false)
    unsigned int number = vals.size();
    if(number<=0) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::out_build_release_sh :" 
          << " WARNING : foreign " << Lib::sout(name) 
          << " without <native_version>."
          << Lib::endl;
    }

    for(unsigned int index=0;index<number;index++) {
      const std::string& platform = platforms[index];
      BUILD_ON_PLATFORM_FP

      std::string s = "foreign " + name + " " + vals[index];

      std::string spif = if_count();
      bool has_test = out_begin_platform(out,platform,spif);

      if(fProdWhat==obuild::Builder::build_sh) {
        out << "echo \"" << s << "\" >> " << file << Lib::endl;
      } else if(fProdWhat==obuild::Builder::build_bat) {
        out << "ECHO "   << s << "   >> " << file << Lib::endl;
      }

      if(has_test) out_end_platform(out,spif);

    }

  }}

  out << Lib::endl;}

  // handle <release> of <foreign> :
  // (See for example foreign/Python.obuild)
  //FIXME : the same foreign (for exa Python) can come 
  //        from different packages !
 {std::vector<Lib::ItemML*> list;
  Lib::ItemML::collectByTag(aProjectML,"foreign",list);
  Lib::ItemML::unique(list);
  std::vector<Lib::ItemML*>::iterator it;
  for(it=list.begin();it!=list.end();++it) {
    Lib::ItemML* releaseML = (*it)->findByTag("release");
    if(!releaseML) continue;

   {GETPROP_FP(*releaseML,"copy_dll",false)
    unsigned int number = vals.size();
    for(unsigned int index=0;index<number;index++) {
      std::string value = vals[index];
      const std::string& platform = platforms[index]; //"platform"
      BUILD_ON_PLATFORM_FP

      std::string spif = if_count();
      bool has_test = out_begin_platform(out,platform,spif);

      //replace_at_at(value);
      out_copy_check(out,quote(value),towin(rel_dir_bin));

      if(has_test) out_end_platform(out,spif);

      done_something = true;
    }}

   {std::string top_from;
    bool out_something;
    if(!out_release_sh(out,*releaseML,top_from,out_something)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }
    if(out_something) done_something = true;}

  }
  out << Lib::endl;}

  // handle <release> of <package> :
 {std::vector<Lib::ItemML*> list;
  Lib::ItemML::collectByTag(aProjectML,"package",list);
  std::vector<Lib::ItemML*>::iterator it;
  for(it=list.begin();it!=list.end();++it) {

    Lib::ItemML* releaseML = 0;

    // Find first <release> in first level of children.
    // ( <foreign> can have also a <release> ).
   {const std::list<Lib::ItemML*>& list2 = (*it)->children();
    std::list<Lib::ItemML*>::const_iterator it2;
    for(it2=list2.begin();it2!=list2.end();++it2) {
      if((*it2)->tagName()=="release") {
        releaseML = *it2;
        break;
      }
    }}

    if(!releaseML) continue;

    std::string pack_name;
    (*it)->attributeValue("name",pack_name);
    std::string pack_version;
    (*it)->attributeValue("version",pack_version);

    if(!out_path_sh(pack_name,pack_version,out)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }

    //printf("debug : <release> under %s : %s\n",
      //pack_name.c_str(),pack_version.c_str());

   {std::string top_from = var_val(path_variable(pack_name));
    top_from += sep();
    bool out_something;
    if(!out_release_sh(out,*releaseML,top_from,out_something)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }
    if(out_something) done_something = true;}

 }}
  out << Lib::endl;

  // libraries :
 {std::vector<Lib::ItemML*> list;
  Lib::ItemML::collectByTag(aProjectML,"library",list);
  std::vector<Lib::ItemML*>::iterator it;
  for(it=list.begin();it!=list.end();++it) {
    Lib::ItemML& itemML = *(*it);

    if(!handle_template(itemML)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }

    std::string name;
    if(!itemML.attributeValue("","name",name)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }

    Lib::ItemML* packML = get_parent_package_ML(itemML);
    if(!packML) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::out_build_release_sh :" 
          << " <" << itemML.tagName() << "> " << Lib::sout(name)
          << " has no parent <package>."
          << Lib::endl;
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }

    /*
   {std::string platform;
    if(fProdWhat==obuild::Builder::build_sh) {
      platform = "UNIX";
    } else {
      platform = "Windows_NT";
    }
    bool doit;
    if(!done_on(itemML,platform,doit)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }
    if(!doit) continue;}*/

    //printf("debug : %s : %s\n",itemML.tagName().c_str(),name.c_str());

    std::string pack_name;
    packML->attributeValue("name",pack_name);
    std::string pack_version;
    packML->attributeValue("version",pack_version);

    if(!out_path_sh(pack_name,pack_version,out)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }

    std::string from = var_val(path_variable(pack_name));
    from += sep();
    from += dbin;

    std::string sif = if_count();

    bool warn_from = (*it)->isProperty("group")?false:true;

    if(fProdWhat==obuild::Builder::build_sh) {

      if(warn_from) {
        out << "if [ ! -d " << quote(from) << " ] ; then" << Lib::endl;      
        out << "  echo \"WARNING : Directory " << from << " not found.\"" 
              << Lib::endl;
        out << "else" << Lib::endl;
      } else {
        out << "if [ -d " << quote(from) << " ] ; then" << Lib::endl;      
      }

      // archive (if any) (for exa G3Lab) :
      out_copy_check(out,quote(from+sep()+"lib"+name+".a"),rel_dir_lib);

      // Darwin :
      out_copy_check(out,quote(from+sep()+"lib"+name+".dylib"),rel_dir_lib);

      // Linux :
      out_copy_check(out,quote(from+sep()+"lib"+name+".so"),rel_dir_lib);

      // CYGWIN :
      out_copy_check(out,quote(from+sep()+"lib"+name+".dll.a"),rel_dir_lib);
      out_copy_check(out,quote(from+sep()+name+".dll"),rel_dir_bin);

      out << "fi" << Lib::endl;

    } else if(fProdWhat==obuild::Builder::build_bat) {

      from = towin(from);

      if(warn_from) {
        out << "IF NOT EXIST " << quote(from)
            << "  ECHO WARNING : Directory " << from << " not found."
            << "  && GOTO end_if_release_" << sif 
            << Lib::endl;
      } else {
        out << "IF NOT EXIST " << quote(from) 
            << "  GOTO end_if_release_" << sif 
            << Lib::endl;
      }

      // archive (if any) (for exa G3Lab) :
      out_copy_check(out,quote(from+sep()+name+".arc"),towin(rel_dir_lib));

      out_copy_check(out,quote(from+sep()+name+".lib"),towin(rel_dir_lib));

      out_copy_check(out,quote(from+sep()+name+".dll"),towin(rel_dir_bin));

      out << ":end_if_release_" << sif << Lib::endl;
    }

    done_something = true;
  }}
  out << Lib::endl;

  // dlls :
 {std::vector<Lib::ItemML*> list;
  Lib::ItemML::collectByTag(aProjectML,"dll",list);
  std::vector<Lib::ItemML*>::iterator it;
  for(it=list.begin();it!=list.end();++it) {
    Lib::ItemML& itemML = *(*it);

    if(!handle_template(itemML)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }

    std::string name;
    if(!itemML.attributeValue("","name",name)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }

    Lib::ItemML* packML = get_parent_package_ML(itemML);
    if(!packML) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::out_build_release_sh :" 
          << " <" << itemML.tagName() << "> " << Lib::sout(name)
          << " has no parent <package>."
          << Lib::endl;
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }

    /*
   {std::string platform;
    if(fProdWhat==obuild::Builder::build_sh) {
      platform = "UNIX";
    } else {
      platform = "Windows_NT";
    }
    bool doit;
    if(!done_on(itemML,platform,doit)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }
    if(!doit) continue;}*/

    //printf("debug : dll : %s : %s\n",itemML.tagName().c_str(),name.c_str());

    std::vector<std::string> copies;
    itemML.propertyValues("copy",copies);

    std::string pack_name;
    packML->attributeValue("name",pack_name);
    std::string pack_version;
    packML->attributeValue("version",pack_version);

    if(!out_path_sh(pack_name,pack_version,out)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }

    std::string pyname;
    itemML.attributeValue("python_module","name",pyname);

    std::string save_file;
   {Lib::ItemML* saveML = itemML.findByTag("save");
    if(saveML) {
      if(!saveML->propertyValue("file",save_file)) {
        save_file = name+".save";
      }
    }}
    //::printf("debug : save_file : \"%s\"\n",save_file.c_str());

    std::string from = var_val(path_variable(pack_name));
    from += sep();
    from += dbin;

    std::string sif = if_count();

    bool warn_from = (*it)->isProperty("group")?false:true;

    if(fProdWhat==obuild::Builder::build_sh) {

      if(warn_from) {
        out << "if [ ! -d " << quote(from) << " ] ; then" << Lib::endl;
        out << "  echo \"WARNING : Directory " << from << " not found.\"" 
            << Lib::endl;
        out << "else" << Lib::endl;
      } else {
        out << "if [ -d " << quote(from) << " ] ; then" << Lib::endl;
      }

      // Darwin :
      out_copy_check(out,quote(from+sep()+name+".bundle"),rel_dir_bin);

     {unsigned int number = copies.size();
      for(unsigned int index=0;index<number;index++) {
        const std::string& scopy = copies[index];
        out_copy_check(out,quote(from+sep()+scopy+".bundle"),rel_dir_bin);
      }}

      if(pyname.size()) {
        out_copy_check(out,quote(from+sep()+pyname+".so"),rel_dir_bin);
      }

      out_copy_check(out,quote(from+sep()+"lib"+name+".jnilib"),rel_dir_bin);

      // Linux :
      out_copy_check(out,quote(from+sep()+"lib"+name+".so"),rel_dir_bin);

     {unsigned int number = copies.size();
      for(unsigned int index=0;index<number;index++) {
        const std::string& scopy = copies[index];
        out_copy_check(out,quote(from+sep()+"lib"+scopy+".so"),rel_dir_bin);
      }}

      // CYGWIN :
      out_copy_check(out,quote(from+sep()+name+".dll"),rel_dir_bin);

     {unsigned int number = copies.size();
      for(unsigned int index=0;index<number;index++) {
        const std::string& scopy = copies[index];
        out_copy_check(out,quote(from+sep()+scopy+".dll"),rel_dir_bin);
      }}

      if(save_file.size()) {
        out_copy_check(out,quote(from+sep()+save_file),rel_dir_bin);
      }

      out << "fi" << Lib::endl;

    } else if(fProdWhat==obuild::Builder::build_bat) {

      from = towin(from);

      if(warn_from) {
        out << "IF NOT EXIST " << quote(from)
            << "  ECHO WARNING : Directory " << from << " not found."
            << "  && GOTO end_if_release_" << sif 
            << Lib::endl;
      } else {
        out << "IF NOT EXIST " << quote(from)
            << "  GOTO end_if_release_" << sif 
            << Lib::endl;
      }

      out_copy_check(out,quote(from+sep()+name+".lib"),towin(rel_dir_lib));

      out_copy_check(out,quote(from+sep()+name+".dll"),towin(rel_dir_bin));

     {unsigned int number = copies.size();
      for(unsigned int index=0;index<number;index++) {
        const std::string& scopy = copies[index];
        out_copy_check(out,quote(from+sep()+scopy+".dll"),towin(rel_dir_bin));
      }}

      if(pyname.size()) {
        out_copy_check(out,quote(from+sep()+pyname+".pyd"),towin(rel_dir_bin));
      }

      if(save_file.size()) {
        out_copy_check(out,quote(from+sep()+save_file),towin(rel_dir_bin));
      }

      out << ":end_if_release_" << sif << Lib::endl;
    }

    done_something = true;
  }}
  out << Lib::endl;

  // applications :
 {std::vector<Lib::ItemML*> list;
  Lib::ItemML::collectByTag(aProjectML,"application",list);
  std::vector<Lib::ItemML*>::iterator it;
  for(it=list.begin();it!=list.end();++it) {
    Lib::ItemML& itemML = *(*it);

    if(!handle_template(itemML)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }

    std::string name;
    if(!itemML.attributeValue("","name",name)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }
    //printf("debug : application : %s : %s\n",
    //  itemML.tagName().c_str(),name.c_str());

    std::string save_file;
   {Lib::ItemML* saveML = itemML.findByTag("save");
    if(saveML) {
      if(!saveML->propertyValue("file",save_file)) {
        save_file = name+".save";
      }
    }}
    //printf("debug : save_file : \"%s\"\n",save_file.c_str());

    Lib::ItemML* packML = get_parent_package_ML(itemML);
    if(!packML) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::out_build_release_sh :" 
          << " <" << itemML.tagName() << "> " << Lib::sout(name)
          << " has no parent <package>."
          << Lib::endl;
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }

    /*
   {std::string platform;
    if(fProdWhat==obuild::Builder::build_sh) {
      platform = "UNIX";
    } else {
      platform = "Windows_NT";
    }
    bool doit;
    if(!done_on(itemML,platform,doit)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }
    if(!doit) continue;}*/

    std::vector<std::string> copies;
    itemML.propertyValues("copy",copies);

    std::string pack_name;
    packML->attributeValue("name",pack_name);
    std::string pack_version;
    packML->attributeValue("version",pack_version);

    if(!out_path_sh(pack_name,pack_version,out)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }

    std::string from = var_val(path_variable(pack_name));
    from += sep();
    from += dbin;

    std::string sif = if_count();

    bool warn_from = (*it)->isProperty("group")?false:true;

    if(fProdWhat==obuild::Builder::build_sh) {

      if(warn_from) {
        out << "if [ ! -d " << quote(from) << " ] ; then" << Lib::endl;
        out << "  echo \"WARNING : Directory " << from << " not found.\"" 
            << Lib::endl;
        out << "else" << Lib::endl;
      } else {
        out << "if [ -d " << quote(from) << " ] ; then" << Lib::endl;
      }

      out_copy_check(out,quote(from+sep()+name),rel_dir_bin);

     {unsigned int number = copies.size();
      for(unsigned int index=0;index<number;index++) {
        const std::string& scopy = copies[index];
        out_copy_check(out,quote(from+sep()+scopy),rel_dir_bin);
      }}

      if(save_file.size()) {
        out_copy_check(out,quote(from+sep()+save_file),rel_dir_bin);
      }

      out << "fi" << Lib::endl;

    } else if(fProdWhat==obuild::Builder::build_bat) {

      from = towin(from);

      if(warn_from) {
        out << "IF NOT EXIST " << quote(from)
            << "  ECHO WARNING : Directory " << from << " not found."
            << "  && GOTO end_if_release_" << sif 
            << Lib::endl;
      } else {
        out << "IF NOT EXIST " << quote(from)
            << "  GOTO end_if_release_" << sif 
            << Lib::endl;
      }

      out_copy_check(out,quote(from+sep()+name+".exe"),towin(rel_dir_bin));

     {unsigned int number = copies.size();
      for(unsigned int index=0;index<number;index++) {
        const std::string& scopy = copies[index];
        out_copy_check(out,quote(from+sep()+scopy+".exe"),towin(rel_dir_bin));
      }}

      if(save_file.size()) {
        out_copy_check(out,quote(from+sep()+save_file),towin(rel_dir_bin));
      }

      out << ":end_if_release_" << sif << Lib::endl;
    }

    done_something = true;
  }}
  out << Lib::endl;

 {// Treat <release> directly under <project> (used in osc_snovis) :
  std::vector<Lib::ItemML*> list;
  Lib::ItemML::collectByTag(aProjectML,"release",list);
  std::vector<Lib::ItemML*>::iterator it;
  for(it=list.begin();it!=list.end();++it) {
    Lib::ItemML& releaseML = *(*it);
    Lib::ItemML* packML = get_parent_package_ML(releaseML);
    if(packML) continue; //Not under <project>.

/*
    Lib::Out out(fPrinter);
    out << "obuild::Builder::out_build_release_sh :"
        << " do not treat <release> directly under <project>."
        << Lib::endl;

    fp.disable();
    Lib::fmanip::remove(sfile);
    return false;
*/

   {std::string top_from = var_val(path_variable(pj_name));
    top_from += sep();
    bool out_something;
    if(!out_release_sh(out,releaseML,top_from,out_something)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }
    if(out_something) done_something = true;}

  }}

 {// producing sys directory ?
  std::vector<Lib::ItemML*> list;
  Lib::ItemML::collectByTag(aProjectML,"sys_dir",list);
  std::vector<Lib::ItemML*>::iterator it;
  for(it=list.begin();it!=list.end();++it) {
    Lib::ItemML& itemML = *(*it);
    std::string platform;
    itemML.attributeValue("platform",platform);

    BUILD_ON_PLATFORM_FP
    std::string spif = if_count();
    bool has_test = out_begin_platform(out,platform,spif);

    if(fProdWhat==obuild::Builder::build_sh) {
      out_begin_pwd(out);
      out_cd_obuild(out);
      std::string s = "./sh/build_sys"+f_sh;
      out << "if [ -e " << s << " ] ; then" << Lib::endl;
      out << "  " << s << " " << rel_dir << Lib::endl;
      out_status(out,2);
      out << "fi" << Lib::endl;
      out_end_pwd(out);
      out << Lib::endl;
      fProduced_build_sys_sh = true;
    }

    if(has_test) out_end_platform(out,spif);

    break; //take first one.
  }}

  if(fProdWhat==obuild::Builder::build_sh) {
    out_begin_pwd(out);
    out_cd_obuild(out);
    std::string s = "./sh/build_Darwin_release"+f_sh;
    out << "if [ -e " << s << " ] ; then" << Lib::endl;
    out << "  " << s << " " << rel_dir << Lib::endl;
    out_status(out,2);
    out << "fi" << Lib::endl;
    out_end_pwd(out);
    out << Lib::endl;
    fProduced_build_Darwin_release_sh = true;
  }

 {// building zip file :
  std::vector<Lib::ItemML*> list;
  Lib::ItemML::collectByTag(aProjectML,"zip",list);
  std::vector<Lib::ItemML*>::iterator it;
  if(list.size()) {
    if(!out_platform_signature(out)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }
  }
  for(it=list.begin();it!=list.end();++it) {
    Lib::ItemML& itemML = *(*it);

    std::string platform;
    itemML.attributeValue("platform",platform);
    BUILD_ON_PLATFORM_FP

    std::string spif = if_count();
    bool has_test = out_begin_platform(out,platform,spif);

    //std::string sif = if_count();

    // Remove some dir from the install area :
   {const std::list<Lib::Element*> props = itemML.properties();
    std::list<Lib::Element*>::const_iterator it2;
    for(it2=props.begin();it2!=props.end();++it2) {
      if((*it2)->name()!="rm_dir") continue;
      const std::string& value = (*it2)->value();
      if(value.empty()) continue;
      // value is relative to install area :
      std::string from = var_val("obuild_dir_project");
      out_rmdir(out,quote(from+sep()+value));
    }}

    // Remove some file from the install area :
   {const std::list<Lib::Element*> props = itemML.properties();
    std::list<Lib::Element*>::const_iterator it2;
    for(it2=props.begin();it2!=props.end();++it2) {
      if( ((*it2)->name()!="rm_files") &&
          ((*it2)->name()!="rm_apps")  ) continue;
      const std::string& value = (*it2)->value();
      if(value.empty()) continue;

      // from is relative to install area :
      std::string from = var_val("obuild_dir_project");

      //name can contain wildcards.
      std::string p,name,s;
      Lib::smanip::pathNameSuffix(value,p,name,s);
      if(p.size()) from += sep()+p;
  
      //std::string sif = if_count();

      if(fProdWhat==obuild::Builder::build_sh) {

        out << "find" 
            << " " << quote(from) 
            << " -maxdepth 1"
            << " -name '" << name << "'"
            << " -follow"
            << " -exec /bin/rm -f {}"
            << " \\;" 
            << Lib::endl;

      } else if(fProdWhat==obuild::Builder::build_bat) {

        from = towin(from);

        if(((*it2)->name()=="rm_apps")&&
           (name.rfind(".exe")==std::string::npos) ){
          name += ".exe";
        }

        //out << "DEL " 
        //    << " " << quote(from) 
        //    << "\\" << name
        //    << Lib::endl;

        if(name.find('*')==std::string::npos) {
          out << "IF EXIST " << quote(from) << "\\" << name
              << " DEL " << quote(from) << "\\" << name
              << Lib::endl;
        } else {
          out << "FOR %%f IN (" << quote(from) << "\\" << name 
              << ") DO DEL \"%%f\"" 
              << Lib::endl;
        }

      }

    }}

    // zip :
    std::string zip_file;
    if(!itemML.propertyValue("file",zip_file)) {
      zip_file = "@@package@@-@@version@@-@@platform_signature@@${obuild_debug_signature}.zip";
    }
    if(!replace_at_at(zip_file)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }
    if(fProdWhat==obuild::Builder::build_bat) {
      zip_file = towin(zip_file);
    }

    std::string darwin_app_zip_file;
    if(!itemML.propertyValue("darwin_app_file",darwin_app_zip_file)) {
      darwin_app_zip_file = "@@package@@-darwin-app-@@version@@-@@platform_signature@@${obuild_debug_signature}.zip";
    }
    if(!replace_at_at(darwin_app_zip_file)) {
      fp.disable();
      Lib::fmanip::remove(sfile);
      return false;
    }

    std::string zip_what;
    itemML.propertyValue("what",zip_what);
    if(zip_what.empty()) zip_what = fPackage;
    
    out_cd_bin(out);

    out_rm(out,zip_file);
    out_rm(out,darwin_app_zip_file);

    if(fProdWhat==obuild::Builder::build_sh) {

      out << "zip -q -r " << zip_file
          << " " << zip_what
          << " -x '*/darwin_app/*'" // do not put darwin_app.
          << Lib::endl;

      std::string s = fPackage+"/"+fVersion+"/darwin_app";
      out << "if [ -d " << s << " ] ; then" << Lib::endl;
      out << "  zip -q -r " << darwin_app_zip_file << " " << s<< Lib::endl;
      out << "fi" << Lib::endl;

    } else if(fProdWhat==obuild::Builder::build_bat) {

      // zip.exe taken from cygwin.
      //FIXME : should check that zip.exe exists.
      out << "zip -q -r " << zip_file
          << " " << zip_what
          << Lib::endl;

    }

    //out_status(out);
    out_cd_obuild(out);

    if(has_test) out_end_platform(out,spif);
  }}

  out_cd_bin(out);

  if(fProdWhat==obuild::Builder::build_sh) {

    out_rm(out,"obuild_platform");

    out_end_pwd(out);

  } else if(fProdWhat==obuild::Builder::build_bat) {

    out_rm(out,"obuild_platform.exe");

    out_end_pwd(out);

    out << ":obuild_return" << Lib::endl;
    out << "ENDLOCAL" << Lib::endl;
    out << "@ECHO ON" << Lib::endl;
  }

  fp.disable();

  Lib::fmanip::chmod(sfile,Lib::fmanip::USER,
                           Lib::fmanip::ADD,
                           Lib::fmanip::EXECUTE);

  if(!done_something) {
    Lib::fmanip::remove(sfile);
  }

  return true;
}

bool obuild::Builder::out_release_sh(
 Lib::Out& aOut
,Lib::ItemML& aParentML //<project>, <package> or <foreign>
,const std::string& aFromTop //If not empty, must be ended with sep()
,bool& a_done_something
){
  a_done_something = false;

 {std::vector<Lib::ItemML*> items;
  Lib::ItemML::collectByTag(aParentML,"copy_file",items);
  unsigned int number = items.size();
  for(unsigned int index=0;index<number;index++) {
    Lib::ItemML& itemML = *(items[index]);

    std::string platform;
    itemML.attributeValue("platform",platform);
    BUILD_ON_PLATFORM

    std::string opts;
    itemML.propertyValue("options",opts);

    std::string copy_from;
    if(!itemML.propertyValue("from",copy_from)) continue;
    bool from_absolute = false;
   {std::string s;
    itemML.attributeValue("from","absolute",s);
    if(s.size()){
      if(!Lib::smanip::tobool(s,from_absolute)) {
        Lib::Out out(fPrinter);
        out << "obuild::Builder::out_release_sh :" 
            << " in <copy_file>, absolute attribute value " << Lib::sout(s) 
            << " is not a boolean."
            << Lib::endl;
        return false;
      }
    }}
    if(!replace_at_at(copy_from)) return false;
    std::string from;
    if(from_absolute) {
      from = copy_from;
    } else {
      if(aFromTop.empty()) {
        Lib::Out out(fPrinter);
        out << "obuild::Builder::out_release_sh :" 
            << " in <copy_file>, not absolute required, but top dir is empty."
            << Lib::endl;
        return false;
      }
      from = aFromTop+copy_from;
    }

    std::string copy_to;
    if(!itemML.propertyValue("to",copy_to)) continue;
    std::string to = var_val("obuild_dir_project")+sep()+copy_to;
    std::string to_dir = Lib::smanip::dirname(to);

    std::string spif = if_count();
    bool has_test = out_begin_platform(aOut,platform,spif);

    std::string sif = if_count();

    if(fProdWhat==obuild::Builder::build_sh) {

      aOut << "if [ ! -e " << quote(from) << " ] ; then" << Lib::endl;
      aOut << "  echo \"WARNING : File " << from << " not found.\"" 
           << Lib::endl;
      aOut << "else" << Lib::endl;

      out_mkdir(aOut,quote(to_dir));

    } else if(fProdWhat==obuild::Builder::build_bat) {

      from = towin(from);

      aOut << "IF NOT EXIST " << quote(from)
           << "  ECHO WARNING : File " << from << " not found."
           << "  && GOTO end_if_release_" << sif 
           << Lib::endl;

      out_mkdir(aOut,quote(towin(to_dir)));

      to = towin(to);

    }

    out_copy(aOut,quote(from),quote(to),opts);

    if(fProdWhat==obuild::Builder::build_sh) {
      aOut << "fi" << Lib::endl;
    } else if(fProdWhat==obuild::Builder::build_bat) {
      aOut << ":end_if_release_" << sif << Lib::endl;
    }

    if(has_test) out_end_platform(aOut,spif);

    a_done_something = true;
  }}

 {std::vector<Lib::ItemML*> items;
  Lib::ItemML::collectByTag(aParentML,"copy_dir",items);
  unsigned int number = items.size();
  for(unsigned int index=0;index<number;index++) {
    Lib::ItemML& itemML = *(items[index]);

    std::string platform;
    itemML.attributeValue("platform",platform);
    BUILD_ON_PLATFORM

    // Assume a UNIX syntax (with / separator).
    std::string copy_from;
    if(!itemML.propertyValue("from",copy_from)) continue;
    bool from_absolute = false;
   {std::string s;
    itemML.attributeValue("from","absolute",s);
    if(s.size()){
      if(!Lib::smanip::tobool(s,from_absolute)) {
        Lib::Out out(fPrinter);
        out << "obuild::Builder::out_release_sh :" 
            << " in <copy_dir>, absolute attribute value " << Lib::sout(s) 
            << " is not a boolean."
            << Lib::endl;
        return false;
      }
    }}
    if(!replace_at_at(copy_from)) return false;
    std::string from;
    if(from_absolute) {
      from = copy_from;
    } else {
      from = aFromTop+copy_from;
      if(aFromTop.empty()) {
        Lib::Out out(fPrinter);
        out << "obuild::Builder::out_release_sh :" 
            << " in <copy_dir>, not absolute required, but top dir is empty."
            << Lib::endl;
        return false;
      }
    }

    std::string copy_to;
    if(!itemML.propertyValue("to",copy_to)) continue;

    if(!replace_at_at(copy_to)) return false;

    //printf("debug : <release> <copy_dir> %s %s\n",
      //copy_from.c_str(),copy_to.c_str());

     
    std::string to = var_val("obuild_dir_project")+sep()+copy_to;
  
    std::string spif = if_count();
    bool has_test = out_begin_platform(aOut,platform,spif);

    std::string sif = if_count();

    bool warn_from;
   {std::string s;
    itemML.propertyValue("warn",s);
    if(!Lib::smanip::tobool(s,warn_from)) warn_from = true;}

    if(fProdWhat==obuild::Builder::build_sh) {

      if(warn_from) {
        aOut << "if [ ! -d " << quote(from) << " ] ; then" << Lib::endl;
        aOut << "  echo \"WARNING : Directory " << from << " not found.\"" 
             << Lib::endl;
        aOut << "else" << Lib::endl;
      } else {
        aOut << "if [ -d " << quote(from) << " ] ; then" << Lib::endl;
      }

      out_mkdir(aOut,quote(to));

      aOut << "/bin/cp -f -R " << quote(from) << " " << quote(to) << Lib::endl;

      //CYGWIN : to rm warnings :
      //  find <dir> -name CVS -print0 | xargs -0 /bin/rm -R -f
      //but it doesn not work on OSF1.
      //CYGWIN, Darwin : 
      //  find <dir> -name CVS -delete
      //but it doesn not work on Linux and OSF1.

     {std::string spif2 = if_count();
      bool has_test_CYGWIN = out_begin_platform(aOut,"CYGWIN",spif2);

        aOut << "find " << quote(to)
             << " -name CVS -print0 | xargs -0 /bin/rm -R -f" 
             << Lib::endl;

        aOut << "find " << quote(to)
             << " -name .svn -print0 | xargs -0 /bin/rm -R -f" 
             << Lib::endl;

      if(has_test_CYGWIN) if(!out_else_platform(aOut)) return false;

        aOut << "find " << quote(to)
             << " -depth -name CVS -exec /bin/rm -R -f {} \\;" 
             << Lib::endl;

        aOut << "find " << quote(to)
            << " -depth -name .svn -exec /bin/rm -R -f {} \\;" 
            << Lib::endl;

      if(has_test_CYGWIN) out_end_platform(aOut,spif2);}

      aOut << "fi" << Lib::endl;

    } else if(fProdWhat==obuild::Builder::build_bat) {

      from = towin(from);

      if(warn_from) {
        aOut << "IF NOT EXIST " << quote(from)
             << "  ECHO WARNING : Directory " << from << " not found."
             << "  && GOTO end_if_release_" << sif 
             << Lib::endl;
      } else {
        aOut << "IF NOT EXIST " << quote(from)
             << "  GOTO end_if_release_" << sif 
             << Lib::endl;
      }

      // XCOPY needs the "from" basename...
      to += sep()+Lib::smanip::basename(copy_from);
      to = towin(to);

      out_mkdir(aOut,quote(to));

      out_xcopy(aOut,quote(from),quote(to));

      aOut << ":end_if_release_" << sif << Lib::endl;
    }

    if(has_test) out_end_platform(aOut,spif);

    a_done_something = true;
 }}

 {std::vector<Lib::ItemML*> items;
  Lib::ItemML::collectByTag(aParentML,"copy_files",items);
  unsigned int number = items.size();
  for(unsigned int index=0;index<number;index++) {
    Lib::ItemML& itemML = *(items[index]);

    std::string platform;
    itemML.attributeValue("platform",platform);
    BUILD_ON_PLATFORM

    std::string copy_from;
    if(!itemML.propertyValue("from",copy_from)) continue;
    bool from_absolute = false;
   {std::string s;
    itemML.attributeValue("from","absolute",s);
    if(s.size()){
      if(!Lib::smanip::tobool(s,from_absolute)) {
        Lib::Out out(fPrinter);
        out << "obuild::Builder::out_release_sh :" 
            << " in <copy_files>, absolute attribute value " << Lib::sout(s) 
            << " is not a boolean."
            << Lib::endl;
        return false;
      }
    }}
    if(!replace_at_at(copy_from)) return false;
    std::string from;
    if(from_absolute) {
      from = copy_from;
    } else {
      from = aFromTop+copy_from;
      if(aFromTop.empty()) {
        Lib::Out out(fPrinter);
        out << "obuild::Builder::out_release_sh :" 
            << " in <copy_files>, not absolute required, but top dir is empty."
            << Lib::endl;
        return false;
      }
    }

    std::string copy_to;
    if(!itemML.propertyValue("to",copy_to)) continue;

    std::string copy_what;
    if(!itemML.propertyValue("what",copy_what)) continue;

    std::string to = var_val("obuild_dir_project")+sep()+copy_to;

    std::string spif = if_count();
    bool has_test = out_begin_platform(aOut,platform,spif);

    std::string sif = if_count();

    if(fProdWhat==obuild::Builder::build_sh) {

      aOut << "if [ ! -d " << quote(from) << " ] ; then" << Lib::endl;
      aOut << "  echo \"WARNING : Directory " << from << " not found.\"" 
           << Lib::endl;
      aOut << "else" << Lib::endl;

      out_mkdir(aOut,quote(to));

     {std::vector<std::string> words = Lib::smanip::words(copy_what," ");
      unsigned int wordn = words.size();
      for(unsigned int index=0;index<wordn;index++){
        aOut << "find" 
             << " " << quote(from) 
             << " -maxdepth 1"
             << " -name '" << words[index] << "'"
             << " -follow"
             << " -exec /bin/cp -f {} " << quote(to) 
             << " \\;" 
             << Lib::endl;
      }}

      aOut << "fi" << Lib::endl;

    } else if(fProdWhat==obuild::Builder::build_bat) {

      from = towin(from);

      aOut << "IF NOT EXIST " << quote(from)
           << "  ECHO WARNING : Directory " << from << " not found."
           << "  && GOTO end_if_release_" << sif 
           << Lib::endl;

      out_mkdir(aOut,quote(to));

     {std::vector<std::string> words = Lib::smanip::words(copy_what," ");
      unsigned int wordn = words.size();
      for(unsigned int index=0;index<wordn;index++){
        out_xcopy(aOut,quote(from)+sep()+words[index],quote(to));
      }}

      //CVS directories not copied by the upper (hidden files).

      aOut << ":end_if_release_" << sif << Lib::endl;
    }

    if(has_test) out_end_platform(aOut,spif);

    a_done_something = true;
  }}

 {std::vector<Lib::ItemML*> items;
  Lib::ItemML::collectByTag(aParentML,"remove_files",items);
  unsigned int number = items.size();
  for(unsigned int index=0;index<number;index++) {
    Lib::ItemML& itemML = *(items[index]);

    std::string remove_from;
    if(!itemML.propertyValue("from",remove_from)) continue;

    std::string remove_what;
    if(!itemML.propertyValue("what",remove_what)) continue;

    std::string platform;
    itemML.attributeValue("platform",platform);
    BUILD_ON_PLATFORM

    std::string from = var_val("obuild_dir_project")+sep()+remove_from;

    std::string spif = if_count();
    bool has_test = out_begin_platform(aOut,platform,spif);

    std::string sif = if_count();

    if(fProdWhat==obuild::Builder::build_sh) {

      aOut << "if [ ! -d " << quote(from) << " ] ; then" << Lib::endl;
      aOut << "  echo \"WARNING : Directory " << from << " not found.\"" 
           << Lib::endl;
      aOut << "else" << Lib::endl;

     {std::vector<std::string> words = Lib::smanip::words(remove_what," ");
      unsigned int wordn = words.size();
      for(unsigned int index=0;index<wordn;index++){
        aOut << "find" 
             << " " << quote(from) 
             << " -maxdepth 1"
             << " -name '" << words[index] << "'"
             << " -follow"
             << " -exec /bin/rm -f {}"
             << " \\;" 
             << Lib::endl;
      }}

      aOut << "fi" << Lib::endl;

    } else if(fProdWhat==obuild::Builder::build_bat) {

      from = towin(from);

      aOut << "IF NOT EXIST " << quote(from)
           << "  ECHO WARNING : Directory " << from << " not found."
           << "  && GOTO end_if_release_" << sif 
           << Lib::endl;

     {std::vector<std::string> words = Lib::smanip::words(remove_what," ");
      unsigned int wordn = words.size();
      for(unsigned int index=0;index<wordn;index++){
        aOut << "DEL " 
             << " " << quote(from) 
             << "\\" << words[index]
             << Lib::endl;
      }}

      //CVS directories not copied by the upper (hidden files).

      aOut << ":end_if_release_" << sif << Lib::endl;
    }

    if(has_test) out_end_platform(aOut,spif);

    a_done_something = true;
  }}

  return true;
}

bool obuild::Builder::out_checkout_release_sh(Lib::ItemML& aProjectML){
  std::string pj_name;
  if(!aProjectML.attributeValue("name",pj_name)) {
    Lib::Out out(fPrinter);
    out << "obuild::Builder::out_checkout_release_sh :"
        << " no attribute name on the <project> XML item."
        << Lib::endl;
    return false;
  }

  std::string pj_version;
  if(!aProjectML.attributeValue("version",pj_version)) {
    Lib::Out out(fPrinter);
    out << "obuild::Builder::out_checkout_release_sh :"
        << " no attribute version on the <project> XML item."
        << Lib::endl;
    return false;
  }

  if(!check_name(pj_name)) {
    Lib::Out out(fPrinter);
    out << "obuild::Builder::out_checkout_release_sh :"
        << " problem with " << Lib::sout(pj_name)
        << ". Project name should have the syntax of a programming variable." 
        << Lib::endl;
    return false;
  }

  fPackage = pj_name;
  fVersion = pj_version;

  std::string sfile = "."+f_sep+f_dir+f_sep+"checkout_release"+f_sh;
  Lib::FileWriter fp(sfile,fPrinter);
  if(!fp.enabled()) {
    Lib::Out out(fPrinter);
    out << "obuild::Builder::out_checkout_release_sh :" 
        << " can't open file " << Lib::sout(sfile) 
        << Lib::endl;
    return false;
  }
  Lib::Out out(fp);

  reset_local_flags("checkout_release_sh");

  if(fProdWhat==obuild::Builder::build_sh) {
    out_template(out,checkout_release_header_sh_linen,
                     checkout_release_header_sh_lines);
  } else if(fProdWhat==obuild::Builder::build_bat) {
    out_template(out,checkout_release_header_bat_linen,
                     checkout_release_header_bat_lines);
  }

  bool done_something = false;

  out_begin_pwd(out);

  out << "cd .." << Lib::endl;
  out << "cd .." << Lib::endl;
  out << "cd .." << Lib::endl;
  out << Lib::endl;

 {std::vector<Lib::ItemML*> list;
  Lib::ItemML::collectByTag(aProjectML,"package",list);
  std::vector<Lib::ItemML*>::iterator it;
  for(it=list.begin();it!=list.end();++it) {

    std::string pack_name;
    (*it)->attributeValue("name",pack_name);
    std::string pack_version;
    (*it)->attributeValue("version",pack_version);

    std::string pack_rep;
    if(!(*it)->propertyValue("repository",pack_rep)) {
      Lib::Out out(fPrinter);
      out << "obuild::Builder::out_checkout_release_sh :"
          << " WARNING : package " << Lib::sout(pack_name) 
          << " has no <repository>."
          << Lib::endl;      
    }

    Lib::smanip::replace(pack_rep,"@@user@@","anonymous");

    if(fProdWhat==obuild::Builder::build_sh) {
      out_mkdir(out,pack_name);
      out_cd(out,pack_name);
      out << "CVSROOT=" << pack_rep << Lib::endl;
      out << "export CVSROOT" << Lib::endl;
    } else if(fProdWhat==obuild::Builder::build_bat) {
      out_mkdir(out,pack_name);
      out_cd(out,pack_name);
      out << "set CVSROOT=" << pack_rep << Lib::endl;
    }
    out << "cvs co"
        << " -d " << pack_version 
        << " -r " << pack_version 
        << " " << pack_name
        << Lib::endl;
    out << "cd .." << Lib::endl;
    out << Lib::endl;

    done_something = true;
 }}
  out << Lib::endl;

  out_end_pwd(out);

  if(fProdWhat==obuild::Builder::build_sh) {

  } else if(fProdWhat==obuild::Builder::build_bat) {

    out << ":obuild_return" << Lib::endl;
    out << "ENDLOCAL" << Lib::endl;
    out << "@ECHO ON" << Lib::endl;
  }

  fp.disable();

  Lib::fmanip::chmod(sfile,Lib::fmanip::USER,
                           Lib::fmanip::ADD,
                           Lib::fmanip::EXECUTE);

  if(!done_something) {
    Lib::fmanip::remove(sfile);
  }

  return true;
}

bool obuild::Builder::out_build_Darwin_release_sh(){
  if(fProdWhat==obuild::Builder::build_bat) return true;

  std::string sfile = "./sh/build_Darwin_release";
  Lib::FileWriter fp(sfile,fPrinter);
  if(!fp.enabled()) {
    Lib::Out out(fPrinter);
    out << "obuild::out_build_Darwin_release_sh :" 
        << " can't open file " << Lib::sout(sfile) 
        << Lib::endl;
    return false;
  }
  Lib::Out out(fp);

  out_template
    (out,build_Darwin_release_sh_linen,build_Darwin_release_sh_lines);

  fp.disable();

  Lib::fmanip::chmod(sfile,Lib::fmanip::USER,
                           Lib::fmanip::ADD,
                           Lib::fmanip::EXECUTE);

  return true;
}
bool obuild::Builder::out_copy_check(
 Lib::Out& aOut
,const std::string& aFrom
,const std::string& aTo
){
  if(fProdWhat==obuild::Builder::build_sh) {
    aOut << "if [ -e " << aFrom << " ] ; then"
         << "  /bin/cp -f " << aFrom << " " << aTo << " ;"
         << "fi" 
         << Lib::endl;
  } else if(fProdWhat==obuild::Builder::build_bat) {
    std::string sec = exe_count();
    aOut << "IF NOT EXIST " << aFrom << " GOTO end_if_copy_" << sec
         << Lib::endl;
    //NOTE : have the exe_error logic to avoid the message :
    //       1 fichier(s) copie(s)
    aOut << "  COPY " << aFrom << " " << aTo << " 1> exe_error_" << sec
         << Lib::endl;
    aOut << "  @IF ERRORLEVEL 1 @TYPE exe_error_" << sec 
         << " && @GOTO obuild_return" 
         << Lib::endl;
    aOut << "  @IF NOT ERRORLEVEL 1 @DEL exe_error_" << sec 
         << Lib::endl;
    aOut << ":end_if_copy_" << sec << Lib::endl;
  }
  return true;
}

bool obuild::Builder::out_xcopy(
 Lib::Out& aOut
,const std::string& aFrom
,const std::string& aTo
){
  if(fProdWhat==obuild::Builder::build_sh) {
    return false;
  } else if(fProdWhat==obuild::Builder::build_bat) {
    std::string sec = exe_count();
    //CVS directories not copied by the below (/E is for hidden files).
    aOut << "XCOPY /E /Q" 
         << " " << aFrom 
         << " " << aTo 
         << " 1> exe_error_" << sec
         << Lib::endl;
    aOut << "@IF ERRORLEVEL 1 @TYPE exe_error_" << sec
         << " && @GOTO obuild_return" 
         << Lib::endl;
    aOut << "@IF NOT ERRORLEVEL 1 @DEL exe_error_" << sec 
         << Lib::endl;
  }
  return true;
}
