
#include <oroot/TROOT.h>
#include <oroot/TSystem.h>

#include <Slash/Tools/UI.h>

#include <amanip/args.h>

#include <oroot/internal/mem.h>

#include <iostream>

static bool oroot_load(const std::string&,std::ostream&);
static bool oroot_function(const std::string&,const Slash::UI::IInterpreter::Options&,const Slash::UI::IInterpreter::Aliases&,void*);

int main(int a_argc,char** a_argv) {
  oroot::mem::check_class(true);

 {// begin main block :

  amanip::args args(a_argc,a_argv);

  std::vector<std::string> C_files;
 {std::vector<std::string> files = args.files();
  for(std::vector<std::string>::const_iterator it = files.begin();
      it!=files.end();++it) {
    C_files.push_back(*it);
    args.remove(*it); //else OnX::Main will take it as a .onx file !
  }}

 {int argc;
  char** argv;
  args.argcv(argc,argv);  
  bool status = TROOT::initialize(argc,argv,std::cout);
  args.delete_argcv(argc,argv);  
  if(!status) {
    std::cout << "oroot_root :"
              << " can't init." << std::endl;
    return 1;
  }}

  Slash::Core::ISession* session = gROOT->session();

  Slash::UI::IScriptManager* mgr = Slash::scriptManager(*session);
  if(!mgr) {
    std::cout << "oroot_root :"
              << " ScriptManager not found." << std::endl;
    delete gROOT;
    gROOT = 0;
    return 1;
  }

  mgr->addInterpreter("oroot","","","C","","",oroot_function,session,0,0);

  Slash::UI::IUI* ui = Slash::find_UI(*session);
  if(!ui) {
    std::cout << "oroot_root :"
        << " UI not created." << std::endl;
    delete gROOT;
    gROOT = 0;
    return 1;    
  }

  ui->enableTerminal("root [%d] ","oroot");

 {for(std::vector<std::string>::const_iterator it = C_files.begin();
      it!=C_files.end();++it) {
    ui->executeScript("oroot",".x "+*it);
  }}

  ui->steer();

  TROOT::finalize();

  } //end main block.

  oroot::mem::balance(std::cout);

  return 0;
}

#include <oroot/internal/DLL.h>

bool oroot_function(const std::string& aString,const Slash::UI::IInterpreter::Options&,const Slash::UI::IInterpreter::Aliases& aAliases,void* aTag){
  Slash::Core::ISession* session = (Slash::Core::ISession*)aTag;
  std::ostream& out = session->cout();

  //::printf("debug : \"%s\"\n",aString.c_str());

  std::vector<std::string> words;
  Slash::words(aString," ",false,words);

  if((words.size()==1)&&
     ( (words[0]==".q") || (words[0]==".qq") || (words[0]==".qqq") ) ){
    Slash::UI::IUI* ui = Slash::find_UI(*session);
    if(!ui) {
      out << "oroot::executeScript :"
          << " Slash::UI::IUI not found" 
          << std::endl;
      return false;
    }
    ui->exit();
  } else if((words.size()==2)&&(words[0]==".x")) {

    const std::string& file = words[1];
    if(!oroot_load(file,session->cout())) {
      out << "can't load file " << file << "." << std::endl;
      return false;
    }

    std::string nm = Slash::nosuffix(file);

    std::string sf(nm+"_0");
#ifdef WIN32
    std::string plugin = "oroot_"+nm+".dll";
#else
#ifdef __APPLE__
    std::string plugin = "oroot_"+nm+".bundle";
#else
    std::string plugin = "liboroot_"+nm+".so";
#endif
#endif
    oroot::DLL module(plugin,session->cout());
    if(!module.is_opened()) {
      out << "Can't open " << plugin << "." << std::endl;
      return false;
    } else {
      typedef void(*F)();
      F f = (F)module.find(sf);
      if(!f) {
        out << "Function " << sf << " not found" << std::endl;
        return false;
      } else {
        f();
      }
    }
  } else {
    out << "Unknown command " << Slash::sout(aString) << "." << std::endl;
  }

  return true;
}

#include <amanip/system.h>
#include <amanip/file.h>
#include <amanip/smanip.h>

bool oroot_load(const std::string& aPath,std::ostream& a_out){
  //FIXME : std::string name = fmanip::name(aPath);

  std::vector<std::string> text;
  if(!amanip::read(aPath,text)) {
    a_out << "oroot_load :"
          << " can't read " << amanip::sout(aPath) << "." 
          << std::endl;
    return false;
  }

  std::string nm = amanip::nosuffix(aPath);
  
  std::vector<std::string> new_text;

 {bool in_head = false;
  for(std::vector<std::string>::const_iterator it=text.begin();
      it!=text.end();++it) {
    if((*it)=="#ifdef HEAD"){
      in_head = true;
    } else if(in_head && ((*it)=="#else //HEAD")){
      in_head = false;
    } else if((*it)=="#endif //HEAD"){
    } else {
      if(in_head) new_text.push_back(*it);
    }
  }}
  new_text.push_back("");

  new_text.push_back("extern \"C\" {");
  new_text.push_back("#ifdef WIN32");
  new_text.push_back("__declspec(dllexport)");
  new_text.push_back("#endif");
  new_text.push_back("  void "+nm+"_0();");
  new_text.push_back("}");

  new_text.push_back("void "+nm+"_0(){");

 {bool in_core = false;
  for(std::vector<std::string>::const_iterator it=text.begin();
      it!=text.end();++it) {
    if((*it)=="#ifdef HEAD"){
    } else if((*it)=="#else //HEAD"){
      in_core = true;
    } else if((*it)=="#endif //HEAD"){
    } else {
      if(in_core) new_text.push_back(*it);
    }
  }}

  new_text.push_back("}");

  std::string name = "oroot_"+nm;

  if(!amanip::write(name+".cpp",new_text)) {
    a_out << "oroot_load :"
          << " can't write " 
          << amanip::sout(name) << "."
          << std::endl;
    return false;
  }

#ifdef WIN32
  std::string incs = "/I\"%OSC_HOME_DIR%\\Resources\\oroot\"";
  std::string flags = "/nologo /DWIN32 /MD /O2 /W3 /GX /GR";
  amanip::execute
    ("CL.exe "+flags+" /c "+incs+" /Fo"+name+".obj /Tp"+name+".cpp");

  std::string libs = "\"%OSC_HOME_DIR%\\lib\\oroot.lib\"";
  amanip::execute("LINK /nologo /dll /out:"+name+".dll "+name+".obj "+libs);

#else
  std::string incs = "-I\"${OSC_HOME_DIR}\"/Resources/oroot";
  amanip::execute("c++ -c "+incs+" -o "+name+".o "+name+".cpp");

  std::string libs = "-L\"${OSC_HOME_DIR}\"/lib -loroot";
#ifdef __APPLE__
  amanip::execute
    ("c++ -bundle -twolevel_namespace -o "+name+".bundle "+name+".o "+libs);
#else
  amanip::execute("c++ -shared -o lib"+name+".so "+name+".o "+libs);
#endif

#endif

  return true;
}
