// this :
#include <oroot/TCanvas.h>

#include <Slash/Tools/UI.h>

#include <amanip/smanip.h>
#include <amanip/system.h>

#include <AIDA/IAnalysisFactory.h>
#include <AIDA/IPlotterFactory.h>
#include <AIDA/IPlotter.h>
#include <AIDA/IPlotterRegion.h>

#include <oroot/TROOT.h>
#include <oroot/internal/mem.h>

// Embedded code :
//#include "main_onx.h"

TCanvas::TCanvas(const char* name, const char* title)
:m_plotter(0) 
,m_log_x(false)
,m_log_y(false)
,m_log_z(false)
{
  mem_inc("TCanvas");

  Slash::Core::ISession* session = gROOT->session();
  if(!session) return;
  AIDA::IAnalysisFactory* aida = gROOT->aida();
  if(!aida) return;

  Slash::UI::IUI* ui = Slash::find_UI(*session);     
  if(!ui) {
    std::ostream& out = session->cout();
    out << "TCanvas::TCanvas :"
         << "Slash::UI::IUI not found."
        << std::endl;
    return;
  }

//{std::string s = amanip::tostring(main_onx_linen,main_onx_lines);
// Slash::replace(s,"@@double_quote@@","\"");
// Slash::replace(s,"@@back_slash@@","\\");
// ui->load("",s,false);}

  std::string OSC_HOME_DIR;
  if(!amanip::getenv("OSC_HOME_DIR",OSC_HOME_DIR)) {
    std::ostream& out = session->cout();
    out << "TCanvas::TCanvas :"
        << " OSC_HOME_DIR not defined."
        << std::endl;
    return;
  }
  std::string file = OSC_HOME_DIR+"/Resources/oroot/scripts/OnX/canvas.onx";
  if(!ui->load("",file,true)) {
    std::ostream& out = session->cout();
    out << "TCanvas::TCanvas :"
        << " can't load " << file << " file."
        << std::endl;
    return;
  }

  AIDA::IPlotterFactory* pf = aida->createPlotterFactory(0,0);
  if(!pf) {
    std::ostream& out = session->cout();
    out << "TCanvas::TCanvas :"
        << " can't get a AIDA::IPlotterFactory." 
        << std::endl;
    return;
  }

  m_plotter = pf->create(""); //FIXME : give widget name.
  delete pf;
  if(!m_plotter) {
    std::ostream& out = session->cout();
    out << "TCanvas::TCanvas :"
        << " can't get a AIDA::IPlotter." 
        << std::endl;
    return;
  }

  //m_plotter->setTitle("oroot");

  m_plotter->createRegions(1,1,0);

  m_plotter->show();

  gROOT->addCanvas(this); //gROOT manages the TCanvas.
}
TCanvas::~TCanvas() {
  delete m_plotter;
  mem_dec("TCanvas");
}

void TCanvas::Update() {}
TVirtualPad* TCanvas::cd(Int_t subpadnumber) {return 0;}

void TCanvas::SetLogx(Int_t a_value) {m_log_x = a_value==1?true:false;}
void TCanvas::SetLogy(Int_t a_value) {m_log_y = a_value==1?true:false;}
void TCanvas::SetLogz(Int_t a_value) {m_log_z = a_value==1?true:false;}
Int_t TCanvas::GetLogx() const {return m_log_x?1:0;}
Int_t TCanvas::GetLogy() const {return m_log_y?1:0;}
Int_t TCanvas::GetLogz() const {return m_log_z?1:0;}

AIDA::IPlotter* TCanvas::plotter() {return m_plotter;}



