// this :
#include <oroot/TFile.h>

#include <Slash/Core/ISession.h>

#include <AIDA/IAnalysisFactory.h>
#include <AIDA/ITreeFactory.h>
#include <AIDA/ITree.h>

#include <amanip/smanip.h>

#include <oroot/TROOT.h>
#include <oroot/TObject.h>
#include <oroot/internal/mem.h>

TFile::TFile(const char* a_fname):m_tree(0) {
  mem_inc("TFile");

  AIDA::IAnalysisFactory* aida = gROOT->aida();

  AIDA::ITreeFactory* treeFactory = aida->createTreeFactory();
  if(!treeFactory) {
    std::ostream& out = gROOT->session()->cout();
    out << "TFile::TFile :"
        << " can't get an AIDA::ITreeFactory." 
        << std::endl;
    return;
  }

  std::string suffix = amanip::suffix(a_fname);

  m_tree = treeFactory->create(a_fname,suffix,true,false);
  delete treeFactory;
  if(!m_tree) {
    std::ostream& out = gROOT->session()->cout();
    out << "TFile::TFile :"
        << " can't get the AIDA::ITree " << amanip::sout(a_fname) << "." 
        << std::endl;
    return;
  }

}


TFile::~TFile() {
  while(!m_objs.empty()) {
    std::list<TObject*>::iterator it = m_objs.begin();
    TObject* obj  = *it;
    m_objs.erase(it);
    delete obj;
  }

  delete m_tree;
  mem_dec("TFile");
}


#include <AIDA/IManagedObject.h>
#include <AIDA/IHistogram1D.h>
#include <oroot/TH1D.h>

TObject* TFile::Get(const char* a_namecycle) {
  if(!m_tree) return 0;
  AIDA::IManagedObject* mo = m_tree->find(a_namecycle);
  if(!mo) {
    std::ostream& out = gROOT->session()->cout();
    out << "TFile::Get :"
        << " object " << amanip::sout(a_namecycle) << " not found." 
        << std::endl;
    return 0;
  }
  if(AIDA::IHistogram1D* h = 
       (AIDA::IHistogram1D*)mo->cast("AIDA::IHistogram1D")) {
    TObject* obj = new TH1D(*h);
    m_objs.push_back(obj);
    return obj;
  } else {
    std::ostream& out = gROOT->session()->cout();
    out << "TFile::Get :"
        << " object type not yet treated."
        << std::endl;
    return 0;
  }
}

