#ifndef steve_tools_config_h
#define steve_tools_config_h

#include <steve/tools/smanip.h>

#include <cstring> //strtol, strlen
#include <cstdlib> //getenv
#include <cstdio> //fopen

// Used in the xxx-config programs.

namespace steve {
namespace tools {

inline bool version(const std::string& a_string,std::vector<unsigned int>& a_version) {
  // a_string is of the form [v,V][<number><char>]<number>
  // Examples : 0 1.2 v0 v1r2 V1r2p3 v0.1.3
  unsigned int number = a_string.size();
  if(!number) {
    a_version.clear();
    return false;
  }
  char* pos = (char*)a_string.c_str();
  if( ((*pos)=='v') || ((*pos)=='V') ) {
    if(number==1) {
      a_version.clear();
      return false; //a_string is only 'v' or 'V' !
    }
    pos++;
  }
  while(*pos!='\0') {
    char* s;
    long v = ::strtol(pos,&s,10);
    if(s==pos) {
      a_version.clear();
      return false;
    }
    if(v<0) {
      a_version.clear();
      return false;
    }
    a_version.push_back((unsigned int)v);
    pos = s; //Could be at end.
    if(*pos=='\0') break;
    // Next char is then not a digit. Skip it.
    pos++;
  }
  return true;
}

inline std::string remove_versions(const std::string& a_string) {
  // replace "@@top@@/pack/vers" -> "@@top@@/pack".
#ifdef WIN32
  std::string sep = "\\";
#else
  std::string sep = "/";
#endif
  std::string sold = a_string;
  replace(sold,"\"",sep+"@quote@"); //To treat AIDA/v3r2p1"/src
  std::vector<std::string> words = steve::tools::words(sold,sep);
  unsigned int wordn = words.size();
  std::string snew;
  for(unsigned int index=0;index<wordn;index++) {
    std::vector<unsigned int> vers;
    std::string word = words[index];
    if(!version(word,vers)) {
      if(snew.size()) snew += sep;
      snew += words[index];
    }
  }
  replace(snew,sep+"@quote@","\"");
  return snew;
}

inline bool getenv(const std::string& a_string,std::string& a_value){
  const char* env = ::getenv(a_string.c_str());
  if(env) {
    a_value = std::string(env?env:"");
    return true;
  } else {
    a_value = "";
    return false;
  }
}

inline bool read(const std::string& a_file,std::vector<std::string>& a_text){
  a_text.clear();
  //FIXME : should read without a limited buffer.
  FILE* file = ::fopen(a_file.c_str(),"rb");
  if(!file) return false;
  unsigned int BUFSIZE = 65536;
  char* buffer = new char[BUFSIZE+1];
  if(!buffer) {::fclose(file);return false;}
  while(true) {
    if(::fgets(buffer,BUFSIZE,(FILE*)file)==NULL) break; // EOF.
    int l = ::strlen(buffer);
    //  On Windows, editors when saving binary files,
    // put \r\n at place of \n ; we then look for \r\n.
    if( (l>=2) && (buffer[l-2]=='\r') && (buffer[l-1]=='\n') ) {
      buffer[l-2] = '\0';
    } else if( (l>=1) && (buffer[l-1]=='\n') ) {
      buffer[l-1] = '\0';
    }
    a_text.push_back(buffer);
  }
  delete [] buffer;
  ::fclose(file);
  return true;
}

}}

#endif
