
#include <steve/tools/args.h>

#include <steve/util/mem.h>

#include "xml/base_ml.h"
#include "xml/item_ml.h"

#include <map>
#include <fstream>

// templates, embedded code :
#include "wrap_schema_h.h"

//XXX
#define sout(a_string)   ("\""+a_string+"\"")

std::string read_include( int a_linen,const char* a_lines[]) {
  std::string s = steve::tools::to_string(a_linen,a_lines);
  steve::tools::replace(s,"@@double_quote@@","\"");
  steve::tools::replace(s,"@@back_slash@@","\\");
  return s;
}

bool out_template(std::ostream& a_fp,unsigned int a_linen,const char* a_lines[],const std::map<std::string,std::string> &a_at_at) {
  std::string s = read_include(a_linen,a_lines);

  for(std::map<std::string,std::string>::const_iterator it=a_at_at.begin();
      it!=a_at_at.end();++it) {
    steve::tools::replace(s,it->first,it->second);
  }

  a_fp << s;

  return true;
}

bool out_build_cxx(steve::util::item_ml& a_item_ml,std::ostream& a_out) {

  std::string name;
  if(!a_item_ml.attribute_value("name",name)) {
    a_out << "steve_moc :" 
          << " name attribute not found for <schema>." 
          << std::endl;
    return false;
  }

  name = "steve::schema::"+name;

  std::string s_name_under = name;
  steve::tools::replace(s_name_under,"::","_");

  //namespace
  std::vector<std::string> nms = steve::tools::words(s_name_under,"_");
  if(!nms.size()) {
    a_out << "steve_moc :" 
          << " empty name attribute for <schema>." 
          << std::endl;
    return false;
  }

  std::string s_class = nms[nms.size()-1];
  nms.resize(nms.size()-1);

  std::string s_namespace_begin;
  std::string s_namespace_end;
  for(std::vector<std::string>::const_iterator it=nms.begin();
      it!=nms.end();++it) {
    s_namespace_begin += "namespace "+(*it)+" {";
    s_namespace_end += "}";
  }

  std::map<std::string,std::string> at_at;
  at_at["@@name_underscore@@"] = s_name_under;
  at_at["@@namespace_begin@@"] = s_namespace_begin;
  at_at["@@class@@"] = s_class;
  at_at["@@nos_headers@@"] = "";
  at_at["@@data_headers@@"] = "";
  at_at["@@namespace_end@@"] = s_namespace_end;

  std::string sfile = s_name_under+".h";

  std::ofstream fp;
  fp.open(sfile.c_str()); //XXX : check if done.

  if(fp.fail()) {
    a_out << "steve_moc :" 
          << " can't open file " << sout(sfile) 
          << std::endl;
    return false;
  }

  out_template(fp,schema_h_linen,schema_h_lines,at_at);

  //if(!out_build_xxx_core_sh(aItemML,out,"lib",name,files,platforms)) {
  //  fp.disable();
  //  steve::util::fmanip::remove(sfile);
  //  return false;
  //}

  fp.close();
  return true;
}

#include <stdlib.h>
#include <iostream>

int main(int a_argc,char* a_argv[]) {

 {//mem
  steve::tools::args args(a_argc,a_argv);

  std::string file;
  if(!args.file(file)) {
    std::cout << "steve_moc :"
              << " no file given."
              << std::endl;    
    return EXIT_FAILURE;
  }

  std::vector<std::string> tags;
  tags.push_back("steve");
  tags.push_back("schema");

  steve::util::item_ml_factory factory;
  steve::util::base_ml ml(factory);
  ml.set_tags(tags);

  //FileGetter fg(fPrinter);
  //ml.setFileGetter(&fg);

  if(!ml.load_file(file,false,std::cout)) {
    std::cout << "steve_moc :"
              << " problem when loading file " << sout(file) 
              << std::endl;    
  } else {
    steve::util::item_ml* top = ml.top_item();
    if(!top) {
      // Empty file : do nothing.
      std::cout << "steve_moc :"
                << " file " << sout(file) << " is empty."
                << std::endl;    
    } else {

      {const std::list<steve::util::item_ml*>& list = top->children();
       std::list<steve::util::item_ml*>::const_iterator it;
       for(it=list.begin();it!=list.end();++it) {
         const std::string& tagName = (*it)->tag_name();
         if(tagName=="schema") {   

           if(!out_build_cxx(*(*it),std::cout)) {
             //FIXME
           }

         }
       }}


    } 
  } 

  }//mem
  steve::util::mem::balance(std::cout);

  return EXIT_SUCCESS;
}
