/*
  An element is :
    <tag atb1="" atb2="" ...>
      <property atb1="" atb2="" ...> value </property>
      <property atb1="" atb2="" ...> value </property>
      <property atb1="" atb2="" ...> value </property>
      ...
    </tag>          

   element.attribute_value("",<name>,s) 
    retrieve the value of atb <name> of a <tag>
   element.attribute_value(<property>,<name>,s) 
    retrieve the value of an atb <name> of a <propery> of a <tag>

 */

// this :
#include "base_ml.h"

#include <steve/interface/file_getter.h>

#include <steve/tools/file.h>
#include <steve/tools/smanip.h>

#include <steve/util/mem.h>

#include "file_reader.h"
#include "item_ml.h"

#include <stdio.h>
#include <ctype.h> //iscntrl

//XXX
#define sout(a_string)   ("\""+a_string+"\"")

#define MINIMUM(a,b) ((a)<(b)?a:b)

steve::util::base_ml::base_ml(
 item_ml_factory& a_item_ml_factory 
,unsigned int a_verbosity
)
:m_item_ml_factory(a_item_ml_factory)
,m_verbosity(a_verbosity)
,m_top(0) // Used to cleanup in case XML parsing failed.
,m_current(0)
,m_file_getter(0)
,m_out(0)
,m_depth(0)
,m_abort(false)
{
  steve::util::mem::increment("steve::util::base_ml");
}
steve::util::base_ml::base_ml(
 const base_ml& a_from
)
:m_item_ml_factory(a_from.m_item_ml_factory)
,m_verbosity(a_from.m_verbosity)
,m_top(0) // Used to cleanup in case XML parsing failed.
,m_current(0)
,m_file_getter(0)
,m_out(0)
,m_depth(0)
,m_abort(false)
{
  steve::util::mem::increment("steve::util::base_ml");
}
steve::util::base_ml::~base_ml(){
  clear();
  steve::util::mem::decrement("steve::util::base_ml");
}
steve::util::base_ml& steve::util::base_ml::operator=(const base_ml& a_from){
  //private
  return *this;
}
void steve::util::base_ml::set_file_getter(
 steve::interface::file_getter* a_getter
){
  m_file_getter = a_getter;
}
bool steve::util::base_ml::visitEndelement(item_ml&,bool& aKeep){
  aKeep = true;
  return true;
}
void steve::util::base_ml::set_tags(const std::vector<std::string>& a_tags){
  m_tags = a_tags;
}
bool steve::util::base_ml::load_file(
 const std::string& a_file
,bool aCompressed
,std::ostream& a_out
){
  //return read_file(a_file,startelement,endelement,this,a_out);
  if(!read_file(a_file,
               (XML_StartElementHandler)startelement,
               (XML_EndElementHandler)endelement,
               this,aCompressed,a_out)) return false;
  if(m_current) {
    if(!solve_references(*m_current,aCompressed,a_out)) {
      clear();
      return false;
    }
    m_current->set_file(a_file);
  }
  return true;
}
bool steve::util::base_ml::load_string(
 const std::string& a_string
,std::ostream& a_out
){
  return read_string(a_string,
               (XML_StartElementHandler)startelement,
               (XML_EndElementHandler)endelement,
               this,a_out);
}
void steve::util::base_ml::clear(){
  // In case of problem, deleting m_current is not sufficient.
  delete m_top; 
  m_top = 0;
  m_current = 0;
}

bool steve::util::base_ml::solve_references(
 steve::util::item_ml& a_item_ml                      
,bool aCompressed
,std::ostream& a_out
){
  if(a_item_ml.tag_name()=="reference") {
    std::string sref;
    if(a_item_ml.attribute_value("","file",sref)) {
      std::string svalue;
      bool optional;
      if(a_item_ml.attribute_value("","optional",svalue) 
         && tools::to(svalue,optional) ) {
        if(optional) {
          if(!tools::file::exists(sref)) return true;
        }
      }
      if(a_item_ml.children().size()<=0) { // Not done.
        base_ml baseML(m_item_ml_factory,m_verbosity);
        baseML.set_tags(m_tags);
        if(m_file_getter) { //Case of client / server.
          std::string newName;
          if(!m_file_getter->fetch(sref,a_item_ml.attributes(),newName)) 
            return false;
          sref = newName;
          baseML.set_file_getter(m_file_getter);
        }
        if(!baseML.load_file(sref,aCompressed,a_out)) return false;
        steve::util::item_ml* top = baseML.top_item();
        baseML.empty();
        if(!top) return false; //FIXME : return true ? The file could be empty.
        // Put things in the Reference node.
        a_item_ml.add_child(top);
        top->set_parent(&a_item_ml);
      }
    }
  }
 
  // Solve children :
  const std::list<steve::util::item_ml*>& list = a_item_ml.children();
  std::list<steve::util::item_ml*>::const_iterator it;
  for(it=list.begin();it!=list.end();++it) {
    if(!solve_references(*(*it),aCompressed,a_out)) return false;
  }

  return true;
}
bool steve::util::base_ml::read_file(
 const std::string& a_file
,XML_StartElementHandler aStartelement
,XML_EndElementHandler aEndelement
,void* a_tag
,bool aCompressed
,std::ostream& a_out
){
  clear();
  if(!parse_file(a_file,aStartelement,aEndelement,a_tag,aCompressed,a_out)) {
    clear();
    return false;
  }
  return true;
}
bool steve::util::base_ml::read_string(
 const std::string& a_string
,XML_StartElementHandler aStartelement
,XML_EndElementHandler aEndelement
,void* a_tag
,std::ostream& a_out
){
  clear();
  if(!parse_string(a_string,aStartelement,aEndelement,a_tag,a_out)) {
    clear();
    return false;
  }
  return true;
}
const steve::util::item_ml* steve::util::base_ml::top_item() const{
  return m_current;
}
steve::util::item_ml* steve::util::base_ml::top_item(){return m_current;}
void steve::util::base_ml::empty(){
  m_top = 0;
  m_current = 0;
}
bool steve::util::base_ml::is_tag(const std::string& a_string) const {
  unsigned int number = m_tags.size();
  for(unsigned int index=0;index<number;index++) {
    if(a_string==m_tags[index]) return true;
  }
  return false;
}
void steve::util::base_ml::character_data_handler(
 void* a_user_data
,const XML_Char* a_string
,int aLength
){
  std::string s;
  s.resize(aLength);
  unsigned int count = 0;
  char* p = (char*)a_string;
  for (int i = 0; i < aLength; i++, p++) {
    if(!iscntrl(*p)) {
      s[count] = *p;
      count++;
    }
  }  
  if(count) {
    s.resize(count);
    steve::util::base_ml* This = (steve::util::base_ml*)a_user_data;
    This->m_value += s;
  }
}
bool steve::util::base_ml::parse_file(
 const std::string& a_file
,XML_StartElementHandler aStartelement
,XML_EndElementHandler aEndelement
,void* a_tag
,bool aCompressed
,std::ostream& a_out
){
  if(m_verbosity) {
    a_out << "steve::util::base_ml::parse_file :" 
          << " parse file " << sout(a_file) << "..." << std::endl;
  }
  //std::string name = fmanip::name(a_file);
  std::string name = a_file;
/*
  Slash::Core::IReader* reader = 0;
  if(aCompressed) {
    if(!steve::util::ZipReader::available()) {
      a_out << "steve::util::base_ml::parse_file :"
            << " compression not available. " 
            << " check installation of Lib package "
            << " and steve::util::ZipReader in particular. "
            << std::endl;
      return false;
    }
    if(m_verbosity) {
      a_out << "steve::util::base_ml::parse_file :" 
            << " file " << sout(a_file) << " is compressed." << std::endl;
    }
    reader = new steve::util::ZipReader(name,a_out);
  } else {
    // May be compressed anyway :
    if(steve::util::ZipReader::available()) {
      reader = new steve::util::ZipReader(name,a_out);
      if(reader && reader->enabled()) {
        // Seems to be compressed.
      } else {
        delete reader;
        reader = new steve::util::FileReader(name,a_out);
      }
    } else {
      reader = new steve::util::FileReader(name,a_out);
    }
  }
*/
  steve::util::file_reader* reader = 
    new steve::util::file_reader(name,a_out);

  if(!reader || !reader->enabled()) {
    a_out << "steve::util::base_ml::parse_file :" 
          << " can't open file " << a_file << std::endl;
    delete reader;
    return false;
  }
  m_out = &a_out;
  m_depth = 0;
  m_abort = false;
  //char buf[1024 * BUFSIZ];
  char buf[BUFSIZ];
  XML_Parser parser = XML_ParserCreate(NULL);
  XML_SetUserData(parser,a_tag);
  XML_SetElementHandler(parser,aStartelement,aEndelement);
  XML_SetCharacterDataHandler(parser,
              (XML_CharacterDataHandler)steve::util::base_ml::character_data_handler);
  //XML_SetProcessingInstructionHandler(parser,
  //      processingInstructionHandler);
  int done = 0;
  do {
    int len;
    if(!reader->read(buf,sizeof(buf),len)) {
      XML_ParserFree(parser);
      delete reader;
      return false;
    }
    done = len < (int)sizeof(buf) ? 1 : 0;
    if (!XML_Parse(parser, buf, len, done)) {
      a_out << "steve::util::base_ml::parse_file :" 
            << " in file " << sout(a_file)
            << " " << XML_ErrorString(XML_GetErrorCode(parser)) 
            << " at line " << XML_GetCurrentLineNumber(parser) << std::endl;
      XML_ParserFree(parser);
      delete reader;
      return false;
    }
    if(m_abort) {
      XML_ParserFree(parser);
      delete reader;
      return false;
    }
  } while (!done);
  XML_ParserFree(parser);
  delete reader;
  if(m_verbosity) {
    a_out << "steve::util::base_ml::parse_file :" 
          << " parse file " << sout(a_file) << " done." << std::endl;
  }
  return true;
}
bool steve::util::base_ml::parse_string(
 const std::string& a_string
,XML_StartElementHandler aStartelement
,XML_EndElementHandler aEndelement
,void* a_tag
,std::ostream& a_out
){
  m_out = &a_out;
  m_depth = 0;
  m_abort = false;
  XML_Parser parser = XML_ParserCreate(NULL);
  XML_SetUserData(parser,a_tag);
  XML_SetElementHandler(parser,aStartelement,aEndelement);
  XML_SetCharacterDataHandler(parser,
              (XML_CharacterDataHandler)steve::util::base_ml::character_data_handler);
  //XML_SetProcessingInstructionHandler(parser,
  //      processingInstructionHandler);
  char* buf = (char*)a_string.c_str();
  //char* debug = buf;
  std::string::size_type l = a_string.size();
  int done = 0;
  do {
    size_t len = MINIMUM(l,BUFSIZ);
    done = len < BUFSIZ ? 1 : 0;
    if (!XML_Parse(parser, buf, len, done)) {
      a_out << "steve::util::base_ml::parse_string :" 
            << " " << XML_ErrorString(XML_GetErrorCode(parser)) 
            << " at line " << XML_GetCurrentLineNumber(parser) 
            << " at byte index " << (int)XML_GetCurrentByteIndex(parser)
            << std::endl; 
/*    int pos = XML_GetCurrentByteIndex(parser);
      int index;
      for(index=-10;index<10;index++) ::printf("%c",debug[pos+index]);
      ::printf("\n");
      for(index=-10;index<0;index++) ::printf(" ");
      printf("^\n");*/
      XML_ParserFree(parser);
      return false;
    }
    if(m_abort) {
      XML_ParserFree(parser);
      return false;
    }
    buf += len;
    l -= len;
  } while (!done);
  XML_ParserFree(parser);
  return true;
}
void steve::util::base_ml::startelement(
 void* a_user_data
,const XML_Char* a_name
,const XML_Char** a_atbs
){
  steve::util::base_ml* This = (steve::util::base_ml*)a_user_data;
  if(This->m_abort) return; //Do nothing.

  This->m_depth++;
  This->m_value = "";

  std::string name = a_name; //Can't be empty
  if(This->is_tag(name)) {

    if(!This->m_current) {
      if(This->m_depth==1) {
        // Ok. Head.
      } else {
	std::ostream& out(*(This->m_out));
        out << "steve::util::base_ml::startelement :" 
            << " no tag with a depth of " << This->m_depth
            << std::endl; 
        This->m_abort = true;
        return;
      }
    } else {
      int delta = This->m_current->depth() - This->m_depth;
      if(delta>=1) {
	std::ostream& out(*(This->m_out));
        out << "steve::util::base_ml::startelement :" 
            << " for element " << sout(name) 
            << " tag with a delta depth of " << delta
            << std::endl; 
        This->m_abort = true;
        return;
      }
    }

    std::vector<steve::util::item_ml::attribute_t> atbs;
   {const XML_Char** a_atts = a_atbs;
    while((*a_atts)&&(*(a_atts+1))) {
      atbs.push_back(steve::util::item_ml::attribute_t(*a_atts,*(a_atts+1)));
      a_atts+=2;
    }}

    item_ml* parent = This->m_current;
    steve::util::item_ml* element = 
      This->m_item_ml_factory.create(name,atbs,parent);
    if(!element) {
      std::ostream& out(*(This->m_out));
      out << "steve::util::base_ml::startelement :" 
          << " can't create a steve::util::item_ml for tag " << sout(name) 
          << std::endl; 
      This->m_abort = true;
      return;
    }

    //steve::util::out out(*(This->m_out));
    //out << "steve::util::base_ml::startelement :" << steve::util::endl;
    //element->print_xml(*(This->m_out),"debug : ");

    if(parent) parent->add_child(element);

    This->m_current = element;
    element->set_depth(This->m_depth); // Internal only.

    if(!This->m_top) This->m_top = element;

  } else {

    if(!This->m_current) {

      // Can't be in a non-tag without a tag !
      std::ostream& out(*(This->m_out));
      out << "steve::util::base_ml::startelement :" 
          << " for element " << sout(name) 
          << " non-tag without some parent tag."
          << std::endl; 
      This->m_abort = true;
      return;

    } else {

      int delta =  This->m_depth - This->m_current->depth();
      if(delta>1) {

        std::ostream& out(*(This->m_out));
        out << "steve::util::base_ml::startelement :" 
            << " for element " << sout(name) 
            << " grand child of a tag."
            << std::endl; 
        This->m_abort = true;
        return;

      } else if(delta==1) { //ok

        This->m_atbs.clear();
       {const XML_Char** a_atts = a_atbs;
        while((*a_atts)&&(*(a_atts+1))) {
          This->m_atbs.push_back(steve::util::item_ml::attribute_t(*a_atts,*(a_atts+1)));
          a_atts+=2;
        }}

      } else {

        std::ostream& out(*(This->m_out));
        out << "steve::util::base_ml::startelement :" 
            << " for element " << sout(name) 
            << " non-tag with a delta depth of " << delta
            << std::endl; 
        This->m_abort = true;
        return;

      }
    }

  }
}
void steve::util::base_ml::endelement(void* a_user_data,const XML_Char* a_name){
  steve::util::base_ml* This = (steve::util::base_ml*)a_user_data;
  if(This->m_abort) return; //Do nothing.

  if(This->m_current) {

    item_ml* element = This->m_current;
    int delta = This->m_depth - element->depth();
    if(delta==0) { //Back to a tag :
      
      item_ml* parent = element->parent();

      bool keep = false;
      bool cont = This->visitEndelement(*element,keep);
      if(keep) {
        if(parent) {
          This->m_current = parent;
        }
      } else {
        //FIXME : the top could be recreated !
        if(This->m_top==element) This->m_top = 0;

        if(parent) {
          parent->remove_child(element); //delete the element
        } else {
          delete element;
        }

        This->m_current = parent; //parent could be 0 : ok.
      }

      if(!cont) This->m_abort = true;

    } else if(delta==1) { //Back to a child of tag :

      //FIXME : correct m_value ? (Can we pick the one of a sub item ?)
      element->add_property(std::string(a_name),This->m_atbs,This->m_value); 
      //This->m_value = "";

    } else {

      std::ostream& out(*(This->m_out));
      out << "steve::util::base_ml::endelement :" 
          << " problem for element " << sout(std::string(a_name)) 
          << " : delta depth of " << delta
          << std::endl; 
      This->m_abort = true;

    }

  }


  This->m_depth--;
}
