// this :
#include "file_reader.h"

#include <steve/util/mem.h>

#include <stdio.h>

steve::util::file_reader::file_reader(
 const std::string& a_file_name
,std::ostream& a_out
)
:m_out(a_out)
,m_file_name(a_file_name)
,m_FILE(0)
,m_Owner(true)
{
  steve::util::mem::increment("steve::util::file_reader");
  m_FILE = ::fopen(a_file_name.c_str(),"rb");
}
steve::util::file_reader::file_reader(
 void* a_FILE //FILE*
,std::ostream& a_out
)
:m_out(a_out)
,m_FILE(a_FILE)
,m_Owner(false)
{
  steve::util::mem::increment("steve::util::file_reader");
}
steve::util::file_reader::file_reader(const file_reader& a_from)
:m_out(a_from.m_out)
,m_FILE(0)
,m_Owner(false)
{
  //private
  steve::util::mem::increment("steve::util::file_reader");
}
steve::util::file_reader::~file_reader(){
  if(m_Owner && m_FILE) ::fclose((FILE*)m_FILE);
  steve::util::mem::decrement("steve::util::file_reader");
}
steve::util::file_reader& steve::util::file_reader::operator=(
 const file_reader& a_from
){
  //private
  return *this;
}
void steve::util::file_reader::disable() {  
  if(!m_FILE) return; // Already closed.
  if(m_Owner) ::fclose((FILE*)m_FILE);
  m_FILE = 0;
}
void steve::util::file_reader::enable() {
  if(m_FILE) return; // Already open.
  if(m_Owner) m_FILE = ::fopen(m_file_name.c_str(),"rb");
}
bool steve::util::file_reader::enabled() const {return m_FILE ? true : false;}
bool steve::util::file_reader::read(
 void* a_buffer
,unsigned int a_length
,int& a_read_length
){
  a_read_length = 0;
  if(!m_FILE) return false;
  size_t len = ::fread(a_buffer,1,a_length,(FILE*)m_FILE);
  a_read_length = (size_t)len;
  return true;
}
bool steve::util::file_reader::rewind() {
  if(!m_FILE) return false;
  ::rewind((FILE*)m_FILE);
  return true;
}

#include <string.h>

bool steve::util::file_reader::read(
 void* a_buffer
,unsigned int a_length
,std::string& a_string
){
  if(!m_FILE) {
    a_string.resize(0);
    return false;
  }

  char* buffer = (char*)a_buffer;
  if(::fgets(buffer,a_length,(FILE*)m_FILE)==NULL) {
    disable();
    a_string.resize(0);
    return false; //EOF
  }

  size_t l = ::strlen(buffer);
  //  On Windows, editors when saving binary files,
  // put \r\n at place of \n ; we then look for \r\n.
  if( (l>=2) && (buffer[l-2]=='\r') && (buffer[l-1]=='\n') ) {
    buffer[l-2] = '\0';
  } else if( (l>=1) && (buffer[l-1]=='\n') ) {
    buffer[l-1] = '\0';
  }

  a_string = buffer;
  return true;
}

bool steve::util::file_reader::sread(
 void* a_buffer
,unsigned int a_length
,unsigned int& a_stringLength
){
  if(!m_FILE) {
    a_stringLength = 0;
    return false;
  }

  char* buffer = (char*)a_buffer;
  if(::fgets(buffer,a_length,(FILE*)m_FILE)==NULL) {
    disable();
    a_stringLength = 0;
    return false; //EOF
  }

  size_t l = ::strlen(buffer);
  //  On Windows, editors when saving binary files,
  // put \r\n at place of \n ; we then look for \r\n.
  if( (l>=2) && (buffer[l-2]=='\r') && (buffer[l-1]=='\n') ) {
    buffer[l-2] = '\0';
    l -= 2;
  } else if( (l>=1) && (buffer[l-1]=='\n') ) {
    buffer[l-1] = '\0';
    l -= 1;
  }

  a_stringLength = l;
  return true;
}

