#ifndef steve_util_item_ml_h
#define steve_util_item_ml_h 

#include <string>
#include <vector>
#include <list>
#include <ostream>

namespace steve {
namespace util {

class item_ml_factory;

class element;

class item_ml {
public:
  item_ml(const std::string&,item_ml_factory&,item_ml*);
  virtual ~item_ml();
  //virtual void* cast(const std::string&) const;
private:
  item_ml(const item_ml&);
  item_ml& operator=(const item_ml&);
public:
  typedef std::pair<std::string,std::string> attribute_t;
  typedef bool (*exec_function_t)(item_ml&,void*);
public:
  static void collect_by_tag(steve::util::item_ml&,const std::string&,
                           std::vector<steve::util::item_ml*>&);
  static void unique(std::vector<steve::util::item_ml*>&);
public:
  void add_attribute(const std::string&,const std::string&);
  void add_property(const std::string&,
                   const std::vector<attribute_t>&,
                   const std::string&);
  void print(std::ostream&);
  void print_xml(std::ostream&,const std::string& = "");
  bool is_attribute(const std::string&) const;
  bool is_property(const std::string&) const;
  bool property_value(const std::string&,std::string&) const;
  bool property_values(const std::string&,std::vector<std::string>&) const;
  void set_attributes(const std::vector<attribute_t>&);
  bool attribute_value(const std::string&,std::string&) const;
  bool attribute_value(const std::string&,const std::string&,std::string&) const;
  const std::list<element*>& properties() const;
  void set_save_flag(bool);
  bool save_flag() const;
  void set_file(const std::string&);
  std::string file() const;
  bool set_property_value(const std::string&,const std::string&,int = 0);
  bool set_attribute_value(const std::string&,const std::string&,const std::string&,int = 0);
  bool remove_attribute(const std::string&);
  void remove_properties(const std::string&);
  bool property_value_boolean(const std::string&,bool&) const;
  void delete_property(const std::string&);
  void delete_property(steve::util::element*);
  const std::vector<attribute_t>& attributes() const;
  void post_execute(exec_function_t,void*);
  void post_execute_backward(exec_function_t,void*);

  item_ml* parent() const;
  const std::list<item_ml*>& children() const;
  item_ml* find_item(const std::string&) const;
  item_ml* find_item_with_tag(const std::string&,const std::string&) const;
  item_ml* find_by_attribute(const std::string&,const std::string&) const;
  const item_ml* find_by_property_in_same_level(const std::string&,
                                          const std::string&) const;
  steve::util::item_ml* find_by_tag(const std::string&) const;
  void add_child(item_ml*);
  void set_parent(item_ml*);
  void add_after(item_ml*,item_ml*);
  int number_of_children() const;
  bool is_first_child(const item_ml&);
  bool is_last_child(const item_ml&);
  item_ml* previous_child(const item_ml&);
  item_ml* next_child(const item_ml&);
  void remove_child(item_ml*&,bool = true);
  void delete_children();
  item_ml* create_copy(item_ml*);
  enum e_copy_what { ATTRIBUTES, PROPERTIES, CHILDREN, ALL };
  bool copy(const item_ml&,e_copy_what = ALL,bool = true);

  void set_data(void*,void*,int);
  void* get_data1() const;
  void* get_data2() const;
  int get_data_int() const;
  void get_data(void*&,void*&,int&) const;

  // Used by base_ml :
  void set_depth(unsigned int a_depth);
  unsigned int depth() const;

  const std::string& tag_name() const;

  void update_tree(const item_ml&);

  void replace(const std::string&,const std::string&);
  //bool name_att_value(std::string&) const;
  //bool class_att_value(std::string&) const;
  //bool path_att_value(std::string&) const;

  bool has_empty_attribute_value(std::ostream&) const;
private:
  void clear();
private:
  std::string m_tag_name;
  item_ml_factory& m_item_ml_factory;
  item_ml* m_parent;
  std::list<item_ml*> m_children;
  std::list<element*> m_properties;
  std::vector<attribute_t> m_atbs;
  std::string m_file;
  bool m_save_flag;
  void* m_data_1;
  void* m_data_2;
  int m_data_int;
  int m_depth;
};

class item_ml_factory {
public:
  item_ml_factory();
  virtual ~item_ml_factory();
public:
  virtual item_ml* create(const std::string&, //tag_name
                         const std::vector<item_ml::attribute_t>&, //atbs
                         item_ml*); //parent
};

}} //namespace

#endif
