//
// compile with -DDRIVER=[zip,HDF5,...]
//

#define STEVE_FACTORY_H(a_driver) <steve/a_driver/factory.h>

#include STEVE_FACTORY_H(DRIVER)

#define S(aWhat) #aWhat
#define STORE_NAME(a_driver) std::string("steve_exa_1.")+S(a_driver)

#include "event.h" // the event model to work with.

#include "schemas.h" // the "schemas" of the event model.

#include <steve/interface/store.h>

#include <steve/tools/randoms.h>
#include <steve/util/mem.h>

#include <iostream>

int main(int a_argc,char* a_argv[]) {
  //steve::util::mem::set_check_by_class(true);

  std::ostream& printer = std::cout;

  steve::interface::store* steve_store = 
    steve::DRIVER::create_store(printer,STORE_NAME(DRIVER));
  if(!steve_store) {
    std::cout << "Can't create store." << std::endl;
    return 1;
  }

  if(!steve_store->open(steve::interface::store::create)) {
    return 1;
  }

  unsigned int event_number = 50;

  std::cout << "number of events = " << event_number << std::endl;

  steve::tools::random::gauss random;

  // store events :
 {exa::schema::event schema(printer,"event");
  for(unsigned int index=0;index<event_number;index++) {
    exa::event* event = new exa::event(random,index);

    steve_store->reset_refs();
    if(!event->store(*steve_store,schema)) {
      std::cout << "error in event.store()" << std::endl;
      return 1;
    }

    delete event;
  }
  schema.commit();
  schema.clear();} //should be done before store.close().

  // store header with number of events :
 {exa::schema::header schema(printer,"header");
  exa::header header;
  header.setNumberOfevents(event_number);
  if(!header.store(*steve_store,schema)) {
    std::cout << "error in header.store()" << std::endl;
    return 1;
  }
  schema.commit();
  schema.clear();}

  steve_store->close();

  delete steve_store;

  steve::util::mem::balance(std::cout);

  return 0;
}
