
#include "event.h"

#include <steve/helper.h>

#include "schemas.h"

//////////////////////////////////////////////////////////////////////////////
/// hit //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool exa::hit::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  schema_get(::exa::schema::event)
  if(!steve::store<double>(a_store,
                           schema_data_get(hit_m_x),
                           m_x)) return false;
  if(!steve::store<double>(a_store,
                           schema_data_get(hit_m_y),
                           m_y)) return false;
  if(!steve::store<double>(a_store,
                           schema_data_get(hit_m_z),
                           m_z)) return false;
  if(!steve::store_ref<exa::track>(a_store,
                                   schema_data_get(hit_m_track),
                                   m_track)) return false;
  return true;
}
bool exa::hit::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) { 
  schema_get(::exa::schema::event)
  if(!steve::retrieve<double>(a_store,
                              schema_data_get(hit_m_x),
                              a_event,a_is,m_x)) return false;
  if(!steve::retrieve<double>(a_store,
                              schema_data_get(hit_m_y),
                              a_event,a_is,m_y)) return false;
  if(!steve::retrieve<double>(a_store,
                              schema_data_get(hit_m_z),
                              a_event,a_is,m_z)) return false;
  if(!steve::retrieve_ref<exa::track>(a_store,
                                      schema_data_get(hit_m_track),
                                      a_event,a_is,m_track)) return false;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// track ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool exa::track::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  schema_get(::exa::schema::event)
  if(!steve::store<double>(a_store,
                           schema_data_get(track_m_px),
                           m_px)) return false;
  if(!steve::store<double>(a_store,
                           schema_data_get(track_m_py),
                           m_py)) return false;
  if(!steve::store<double>(a_store,
                           schema_data_get(track_m_pz),
                           m_pz)) return false;
  if(!steve::store<exa::hit>(a_store,a_schema,
                        schema_offsets_get(track_m_hits),
                        schema_numbers_get(track_m_hits),
                        m_hits)) return false;
  return true;
}
bool exa::track::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) { 
  schema_get(::exa::schema::event)
  if(!steve::retrieve<double>(a_store,
                              schema_data_get(track_m_px),
                              a_event,a_is,m_px)) return false;
  if(!steve::retrieve<double>(a_store,
                              schema_data_get(track_m_py),
                              a_event,a_is,m_py)) return false;
  if(!steve::retrieve<double>(a_store,
                              schema_data_get(track_m_pz),
                              a_event,a_is,m_pz)) return false;
  if(!steve::retrieve<exa::hit>(a_store,a_schema,
                           schema_numbers_get(track_m_hits),
                           a_event,a_is,m_hits)) return false;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// event ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool exa::event::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  schema_get(::exa::schema::event)
  if(!steve::store<unsigned int>(a_store,
                          schema_data_get(event_m_id),
                          m_id)) return false;
  if(!steve::store<float>(a_store,
                          schema_data_get(event_m_energy),
                          m_energy)) return false;
  if(!steve::store_with_refs<exa::track>(a_store,a_schema,
                             schema_offsets_get(event_m_tracks),
                             schema_numbers_get(event_m_tracks),
                             schema_offsets_get(event_m_tracks_refs),
                             schema_numbers_get(event_m_tracks_refs),
                             schema_data_get(event_m_tracks_refs),
                             m_tracks)) return false;
  return true;
}
bool exa::event::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) {    
  schema_get(::exa::schema::event)
  if(!steve::retrieve<unsigned int>(a_store,
                             schema_data_get(event_m_id),
                             a_event,a_is,m_id)) return false;
  if(!steve::retrieve<float>(a_store,
                             schema_data_get(event_m_energy),
                             a_event,a_is,m_energy)) return false;
  if(!steve::retrieve_with_refs<exa::track>(a_store,a_schema,
                             schema_numbers_get(event_m_tracks),
                             schema_data_get(event_m_tracks_refs),
                             a_event,a_is,m_tracks))return false;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// header ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool exa::header::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  schema_get(::exa::schema::header)
  if(!steve::store<unsigned int>(a_store,
                                schema_data_get(header_m_number_of_events),
                                m_number_of_events)) return false;
  return true;
}
bool exa::header::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) {    
  schema_get(::exa::schema::header)
  if(!steve::retrieve<unsigned int>(a_store,
                                   schema_data_get(header_m_number_of_events),
                                   a_event,a_is,m_number_of_events)) return false;
  return true;
}
