//
// compile with -DDRIVER=[zip,HDF5,...]
//

#define STEVE_FACTORY_H(a_driver) <steve/a_driver/factory.h>

#include STEVE_FACTORY_H(DRIVER)

#define S(aWhat) #aWhat
#define STORE_NAME(a_driver) std::string("steve_exa_2.")+S(a_driver)

#include "event.h" //the event model to work with.

#include "schemas.h" // the "schemas" of the event model.

#include <steve/interface/store.h>
#include <steve/util/mem.h>

#include <iostream>

int main(int a_argc,char* a_argv[]) {
  //steve::util::mem::set_check_by_class(true);

  std::ostream& printer = std::cout;

  steve::interface::store* steve_store = 
    steve::DRIVER::create_store(printer,STORE_NAME(DRIVER));
  if(!steve_store) {
    std::cout << "Can't create store." << std::endl;
    return 1;
  }

  if(!steve_store->open(steve::interface::store::readonly)) {
    return 1;
  }

  // Retrieving events :
  unsigned int number_of_events = 0; 

 {exa::schema::header schema(printer,"header");
  exa::header header;
  steve::path_t is;
  if(!header.retrieve(*steve_store,schema,0,is)) {
    std::cout << "error in header.retrieve()" << std::endl;
    return 1;
  }
  number_of_events = header.numberOfevents();} //end exa::schema::header.

  std::cout << "number of events = " << number_of_events
      << std::endl;

 {exa::schema::event schema(printer,"run");
  for(unsigned int ievent=0;ievent<number_of_events;ievent++) {
    exa::event event;

    steve_store->reset_refs();
    steve::path_t is;
    if(!event.retrieve(*steve_store,schema,ievent,is)) {
      std::cout << "error in event.retrieve()"
                << std::endl;
      return 1;
    } 
    if(event.id()!=ievent) {
      std::cout << " event.m_id anomaly." << std::endl;
    }
    unsigned int tracks_number = event.tracks().size();
    for(unsigned int itrack=0;itrack<tracks_number;itrack++) {
      exa::track* track = event.tracks()[itrack];
      if(track->px()!=itrack) {
        std::cout << " track.m_px anomaly." << std::endl;
      }

      unsigned int hits_number = track->hits().size();
      for(unsigned int ihit=0;ihit<hits_number;ihit++) {
        exa::hit* hit = track->hits()[ihit];
        if(hit->x()!=ihit) {
          std::cout << " hit.m_x anomaly." << std::endl;
        //} else {
        //  std::cout << " hit ok" << std::endl;
        }
        if(hit->track()!=track) {
          std::cout << " hit.m_track anomaly." << std::endl;
        }
      }

    }

  }
  } //end exa::schema.

  steve_store->close();

  delete steve_store;

  steve::util::mem::balance(std::cout);

  return 0;
}
