#ifndef exa_event_h
#define exa_event_h

namespace steve {namespace tools {namespace random {class gauss;}}}

#include <steve/instrument.h> //to help instrumenting a class.

#include <vector>

#include <util/histogram.h>
#include <steve/tools/blob.h>

namespace steve { namespace interface { class streamer;}}

namespace exa { class track;}
namespace exa { class analysis;}

namespace exa {

class time {
public:
  time();
  time(unsigned int a_index_in_hit);
  virtual ~time();
private:
  time(const time&);
  time& operator=(const time&);
public:
  unsigned int get_time() const;
private:
  unsigned int m_time;
//steve:
  STEVE_STREAM_HEADER
};

class hit {
public:
  hit();
  hit(steve::tools::random::gauss& a_random,
      unsigned int a_index_in_track,
      exa::track*);
  virtual ~hit();
private:
  hit(const hit&);
  hit& operator=(const hit&);
public:
  double x() const;
  double y() const;
  double z() const;
  exa::track* track() const;
  std::vector<time*>& times();
  std::vector<double>& vec_double();
private:
  double m_x;
  double m_y;
  double m_z;
  exa::track* m_track;
  std::vector<time*> m_times;
  std::vector<double> m_vec_double;
//steve:
  STEVE_STREAM_HEADER
};

class track {
public:
  track();
  track(steve::tools::random::gauss& a_random,unsigned int a_index_in_event);
  virtual ~track();
private:
  track(const track&);
  track& operator=(const track&);
public:
  double px() const;
  double py() const;
  double pz() const;
  std::vector<hit*>& hits();
private:
  double m_px;
  double m_py;
  double m_pz;
private:
  std::vector<hit*> m_hits;
//steve:
  STEVE_STREAM_HEADER
};


class test_stream {
public:
  test_stream();
  virtual ~test_stream();
private:
  test_stream(const test_stream&);
  test_stream& operator=(const test_stream&);
public:
  void initialize();
  steve::uchar get_uchar() const;
  char get_char() const;
  steve::uint16 get_uint16() const;
  steve::int16 get_int16() const;
  steve::uint32 get_uint32() const;
  steve::int32 get_int32() const;
  steve::uint64 get_uint64() const;
  steve::int64 get_int64() const;
  float get_float() const;
  double get_double() const;
  std::string get_string() const;
  bool is_equal(const test_stream&) const;
private:
  steve::uchar m_uchar;
  char m_char;
  steve::uint16 m_uint16;
  steve::int16 m_int16;
  steve::uint32 m_uint32;
  steve::int32 m_int32;
  steve::uint64 m_uint64;
  steve::int64 m_int64;
  float m_float;
  double m_double;
  std::string m_string;
//steve:
  STEVE_STREAM_HEADER
};


class cell {
public:
  cell();
  cell(unsigned int a_index_in_event);
  virtual ~cell();
  cell(const cell&);
  cell& operator=(const cell&);
  std::string id() const;
  std::string id_vec() const;
private:
  std::string m_id;
  std::string m_id_vec;
//steve:
  STEVE_STREAM_HEADER
};

class event {
public:
  event();
  event(steve::tools::random::gauss& a_random,
        unsigned int a_index_in_run,
        analysis& a_analysis);
  virtual ~event();
  unsigned int id() const;  
  float energy() const;  
  std::vector<track*>& tracks();
  exa::util::histogram_1d& histo();
  steve::tools::blob& blob();
  exa::test_stream& test_stream();
  std::vector<cell>& cells();
private:
  event(const event&);
  event& operator=(const event&);
private:
  unsigned int m_id;
  float m_energy;
  std::vector<track*> m_tracks;
  exa::util::histogram_1d m_histo;
  steve::tools::blob m_blob;
  exa::test_stream m_test_stream;
  std::vector<cell> m_cells;
//steve:
  STEVE_HEADER
  STEVE_STREAM_HEADER
};

class header {
public:
  header();
  virtual ~header();
private:
  header(const header&);
  header& operator=(const header&);
public:
  unsigned int numberOfevents() const;
  void setNumberOfevents(unsigned int);
private:
  unsigned int m_number_of_events;
//steve:
  STEVE_HEADER
  STEVE_STREAM_HEADER
};

class analysis {
public:
  analysis();
  virtual ~analysis();
private:
  analysis(const analysis&);
  analysis& operator=(const analysis&);
public:
  exa::util::histogram_1d m_energy;
//steve:
  STEVE_HEADER
  STEVE_STREAM_HEADER
};

}

#endif
