//
// compile with -DDRIVER=[zip,HDF5,...]
//

#define STEVE_FACTORY_H(a_driver) <steve/a_driver/factory.h>

#include STEVE_FACTORY_H(DRIVER)

#define S(aWhat) #aWhat
#define STORE_NAME(a_driver) std::string("steve_exa_4.")+S(a_driver)

#include "event.h" // the event model to work with.

#include "schemas.h" // the "schemas" of the event model.

#include <util/schemas.h> //histogram schema.

#include <steve/interface/store.h>
#include <steve/interface/store_data.h>

#include <steve/tools/randoms.h>
#include <steve/tools/atime.h>
#include <steve/util/mem.h>

#include <iostream>

int main(int a_argc,char* a_argv[]) {
  //steve::util::mem::set_check_by_class(true);

  std::ostream& printer = std::cout;

  steve::interface::store* steve_store = 
    steve::DRIVER::create_store(printer,STORE_NAME(DRIVER));
  if(!steve_store) {
    std::cout << "Can't create store." << std::endl;
    return 1;
  }

  if(!steve_store->open(steve::interface::store::create)) {
    return 1;
  }

  unsigned int event_number = 50;

  std::cout << "number of events = " << event_number << std::endl;

  exa::analysis analysis;

  steve::tools::random::gauss random;

  // store events :
 {exa::schema::event schema(printer,"run");

  // adapt page size to blob (else storing/retrieving is slow) :
 {std::string sd_name = "data_run_m_blob";
  steve::interface::store_data* sd = schema.find_store_data(sd_name);
  if(!sd) {
    std::cout << " schema \"" << sd_name << "\" not found."
              << std::endl;
  } else {
    sd->set_page_size(500000);
    //sd->set_compression(9); //default=6
  }}

  for(unsigned int index=0;index<event_number;index++) {
    steve::tools::atime time_begin = steve::tools::atime::now();

    exa::event* event = new exa::event(random,index,analysis);

    steve_store->reset_refs();
    if(!event->store(*steve_store,schema)) {
      std::cout << "error in event.store()" << std::endl;
      return 1;
    }

    delete event;

    steve::tools::atime time_elaps = steve::tools::atime::elapsed(time_begin);
    std::cout << " event " << index
              << " time = " << time_elaps.time_string()
              << std::endl;
  }
  schema.commit();
  schema.clear();} //should be done before store.close().

  // store header with number of events :
 {exa::schema::header schema(printer,"header");
  exa::header header;
  header.setNumberOfevents(event_number);
  if(!header.store(*steve_store,schema)) {
    std::cout << "error in header.store()" << std::endl;
    return 1;
  }
  schema.commit();
  schema.clear();}

  // store analysis histogram :
 {exa::schema::analysis schema(printer,"analysis");
  std::cout << "histo " << analysis.m_energy.title() 
            << " entries " << analysis.m_energy.entries()
            << " mean " << analysis.m_energy.mean()
            << " rms " << analysis.m_energy.rms()
            << std::endl;
  if(!analysis.store(*steve_store,schema)) {
    std::cout << "error in analysis.store()" << std::endl;
    return 1;
  }
  schema.commit();
  schema.clear();}

  steve_store->close();

  delete steve_store;

  steve::util::mem::balance(std::cout);

  return 0;
}
