
#include <steve/memory/factory.h>

#include "event.h" // the event model to work with.

#include "schemas.h" // the "schemas" of the event model.

#include <util/schemas.h> //histogram schema.

#include <steve/interface/store.h>
#include <steve/interface/store_data.h>

#include <steve/tools/randoms.h>
#include <steve/tools/atime.h>
#include <steve/tools/smanip.h>
#include <steve/util/mem.h>

#include <iostream>

int main(int a_argc,char* a_argv[]) {
  //steve::util::mem::set_check_by_class(true);


  std::ostream& printer = std::cout;

  steve::interface::store* steve_store = 
    steve::memory::create_store(printer);
  if(!steve_store) {
    std::cout << "Can't create store." << std::endl;\
    return 1;
  }

  ///////////////////////////////////////////////////////////////////////////
  /// store /////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////

 {//begin store

  unsigned int event_number = 50;

  std::cout << "number of events = " << event_number << std::endl;

  exa::analysis analysis;

  steve::tools::random::gauss random;

  // store events :
 {exa::schema::event schema(printer,"run");

  // adapt page size to blob (else storing/retrieving is slow) :
 {std::string sd_name = "data_run_m_blob";
  steve::interface::store_data* sd = schema.find_store_data(sd_name);
  if(!sd) {
    std::cout << " schema \"" << sd_name << "\" not found."
              << std::endl;
  } else {
    sd->set_page_size(500000);
    //sd->set_compression(9); //default=6
  }}

  for(unsigned int index=0;index<event_number;index++) {
    steve::tools::atime time_begin = steve::tools::atime::now();

    exa::event* event = new exa::event(random,index,analysis);

    steve_store->reset_refs();
    if(!event->store(*steve_store,schema)) {
      std::cout << "error in event.store()" << std::endl;
      return 1;
    }

    delete event;

    steve::tools::atime time_elaps = steve::tools::atime::elapsed(time_begin);
    std::cout << " event " << index
              << " time = " << time_elaps.time_string()
              << std::endl;
  }
  schema.commit();
  schema.clear();} //should be done before store.close().

  // store header with number of events :
 {exa::schema::header schema(printer,"header");
  exa::header header;
  header.setNumberOfevents(event_number);
  if(!header.store(*steve_store,schema)) {
    std::cout << "error in header.store()" << std::endl;
    return 1;
  }
  schema.commit();
  schema.clear();}

  // store analysis histogram :
 {exa::schema::analysis schema(printer,"analysis");
  std::cout << "histo " << analysis.m_energy.title() 
            << " entries " << analysis.m_energy.entries()
            << " mean " << analysis.m_energy.mean()
            << " rms " << analysis.m_energy.rms()
            << std::endl;
  if(!analysis.store(*steve_store,schema)) {
    std::cout << "error in analysis.store()" << std::endl;
    return 1;
  }
  schema.commit();
  schema.clear();}

  } //end store

  ///////////////////////////////////////////////////////////////////////////
  /// retrieve //////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////

 {//begin retrieve

  // Retrieving events :
  unsigned int number_of_events = 0; 

 {exa::schema::header schema(printer,"header");
  exa::header header;
  steve::path_t is;
  if(!header.retrieve(*steve_store,schema,0,is)) {
    std::cout << "error in header.retrieve()" << std::endl;
    return 1;
  }
  number_of_events = header.numberOfevents();} //end exa::schema::header.

  std::cout << "number of events = " << number_of_events
      << std::endl;

 {exa::schema::event schema(printer,"run");
  for(unsigned int ievent=0;ievent<number_of_events;ievent++) {
  //for(unsigned int ievent=number_of_events-1;ievent>=0;ievent--) {

    steve::tools::atime time_begin = steve::tools::atime::now();

    exa::event event;

    steve_store->reset_refs();
    steve::path_t is;
    if(!event.retrieve(*steve_store,schema,ievent,is)) {
      std::cout << "error in event.retrieve()"
                << std::endl;
      return 1;
    } 
    if(event.id()!=ievent) {
      std::cout << " event.m_id anomaly." << std::endl;
    }
    unsigned int tracks_number = event.tracks().size();
    for(unsigned int itrack=0;itrack<tracks_number;itrack++) {
      exa::track* track = event.tracks()[itrack];
      if(track->px()!=itrack) {
        std::cout << " track.m_px anomaly." << std::endl;
      }

      unsigned int hits_number = track->hits().size();
      for(unsigned int ihit=0;ihit<hits_number;ihit++) {
        exa::hit* hit = track->hits()[ihit];
        if(hit->x()!=ihit) {
          std::cout << " hit.m_x anomaly." << std::endl;
        //} else {
        //  std::cout << " hit ok" << std::endl;
        }
        if(hit->track()!=track) {
          std::cout << " hit.m_track anomaly." << std::endl;
        }

       {std::vector<double>& vec = hit->vec_double();
        for(unsigned int ivec=0;ivec<vec.size();ivec++) {
          if(vec[ivec]!=(ivec+2)) {
            std::cout << " hit.m_vec_double anomaly." 
                      << " get " << vec[ivec] << ", expected " << (ivec+2)
                      << std::endl;
          }
        }}

        unsigned int times_number = hit->times().size();
        for(unsigned int itime=0;itime<times_number;itime++) {
          exa::time* time = hit->times()[itime];
          if(time->get_time()!=itime) {
            std::cout << " time.m_time anomaly." 
                      << " get " << time->get_time() << ", expected " << itime
                      << std::endl;
          }
        }
      }

    }
    if(event.histo().entries()!=ievent) {
      std::cout << " event.m_histo anomaly." 
                << " " << event.histo().entries() << " " << ievent
                << std::endl;
    }
    if(event.blob().size()!=10) {
      std::cout << " event.m_blob size anomaly." << std::endl;
    } else {
      if(event.blob()[9]!=9) {
        std::cout << " event.m_blob value anomaly." << std::endl;
      }
    }
   {exa::test_stream test;
    test.initialize();
    if(!event.test_stream().is_equal(test)) {
      std::cout << " event.m_test_stream anomaly." << std::endl;
    }}

    unsigned int cells_number = event.cells().size();
    for(unsigned int icell=0;icell<cells_number;icell++) {
      exa::cell& cell = event.cells()[icell];
      std::string s = "exa::cell:";
     {for(unsigned int index=0;index<(icell+1);index++) {
       s += steve::tools::to<unsigned int>(icell+1);
     }}
      if(cell.id()!=s) {
        std::cout << " cell.m_id anomaly." 
            << " get \"" << cell.id() << "\""
            << ", expected \"" << s << "\""
            << std::endl;
      }
      if(cell.id()!=cell.id_vec()) {
        std::cout << " cell.m_id_vec anomaly." 
            << " get \"" << cell.id_vec() << "\""
            << ", expected \"" << cell.id() << "\""
            << std::endl;
      }
    }

    steve::tools::atime time_elaps = steve::tools::atime::elapsed(time_begin);
    std::cout << " time = " << time_elaps.time_string() << std::endl;
  }
  } //end exa::schema.

  // retrieve analysis histogram :
 {exa::analysis analysis;
  exa::schema::analysis schema(printer,"analysis");
  steve::path_t is;
  if(!analysis.retrieve(*steve_store,schema,0,is)) {
    std::cout << "error in analysis.store()" << std::endl;
    return 1;
  }
  std::cout << "histo " << analysis.m_energy.title() 
            << " entries " << analysis.m_energy.entries()
            << " mean " << analysis.m_energy.mean()
            << " rms " << analysis.m_energy.rms()
            << std::endl;}

  } //end retrieve

  delete steve_store;

  steve::util::mem::balance(std::cout);

  return 0;
}
