
#include "event.h"

#include <steve/helper.h>

#include "schemas.h"

#include <util/steve.h> //histogram
#include <steve/core/streamer.h>

//////////////////////////////////////////////////////////////////////////////
/// time //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool exa::time::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  schema_get(::exa::schema::time)
  if(!steve::store<unsigned int>(a_store,
                           schema_data_get(m_time),
                           m_time)) return false;
  return true;
}
bool exa::time::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) { 
  schema_get(::exa::schema::time)
  if(!steve::retrieve<unsigned int>(a_store,
                              schema_data_get(m_time),
                              a_event,a_is,m_time)) return false;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// hit //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool exa::hit::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  schema_get(::exa::schema::hit)
  if(!steve::store<double>(a_store,
                           schema_data_get(m_x),
                           m_x)) return false;
  if(!steve::store<double>(a_store,
                           schema_data_get(m_y),
                           m_y)) return false;
  if(!steve::store<double>(a_store,
                           schema_data_get(m_z),
                           m_z)) return false;
  if(!steve::store_ref<exa::track>(a_store,
                                   schema_data_get(m_track),
                                   m_track)) return false;
  if(!steve::store<exa::time>(a_store,
                        *(schema->schema_m_times()),
                        schema_offsets_get(m_times),
                        schema_numbers_get(m_times),
                        m_times)) return false;
  if(!steve::store<double>(a_store,
                        schema_offsets_get(m_vec_double),
                        schema_numbers_get(m_vec_double),
                        schema_data_get(m_vec_double),
                        m_vec_double)) return false;
  return true;
}
bool exa::hit::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) { 
  schema_get(::exa::schema::hit)
  if(!steve::retrieve<double>(a_store,
                              schema_data_get(m_x),
                              a_event,a_is,m_x)) return false;
  if(!steve::retrieve<double>(a_store,
                              schema_data_get(m_y),
                              a_event,a_is,m_y)) return false;
  if(!steve::retrieve<double>(a_store,
                              schema_data_get(m_z),
                              a_event,a_is,m_z)) return false;
  if(!steve::retrieve_ref<exa::track>(a_store,
                                      schema_data_get(m_track),
                                      a_event,a_is,m_track)) return false;
  if(!steve::retrieve<exa::time>(a_store,
                           *(schema->schema_m_times()),
                           schema_numbers_get(m_times),
                           a_event,a_is,m_times)) return false;
  if(!steve::retrieve<double>(a_store,
                        schema_numbers_get(m_vec_double),
                        schema_data_get(m_vec_double),
                        a_event,a_is,m_vec_double)) return false;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// track ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool exa::track::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  schema_get(::exa::schema::track)
  if(!steve::store<double>(a_store,
                           schema_data_get(m_px),
                           m_px)) return false;
  if(!steve::store<double>(a_store,
                           schema_data_get(m_py),
                           m_py)) return false;
  if(!steve::store<double>(a_store,
                           schema_data_get(m_pz),
                           m_pz)) return false;
  if(!steve::store<exa::hit>(a_store,
                        *(schema->schema_m_hits()),
                        schema_offsets_get(m_hits),
                        schema_numbers_get(m_hits),
                        m_hits)) return false;
  return true;
}
bool exa::track::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) { 
  schema_get(::exa::schema::track)
  if(!steve::retrieve<double>(a_store,
                              schema_data_get(m_px),
                              a_event,a_is,m_px)) return false;
  if(!steve::retrieve<double>(a_store,
                              schema_data_get(m_py),
                              a_event,a_is,m_py)) return false;
  if(!steve::retrieve<double>(a_store,
                              schema_data_get(m_pz),
                              a_event,a_is,m_pz)) return false;
  if(!steve::retrieve<exa::hit>(a_store,
                           *(schema->schema_m_hits()),
                           schema_numbers_get(m_hits),
                           a_event,a_is,m_hits)) return false;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// test_stream //////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
schema_store_retrieve_stream(exa::test_stream,::exa::schema::test_stream)


//////////////////////////////////////////////////////////////////////////////
/// cell //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool exa::cell::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  schema_get(::exa::schema::cell)
  if(!steve::store<std::string>(a_store,
                           schema_data_get(m_id),
                           m_id)) return false;
  if(!steve::store(a_store,
                           schema_offsets_get(m_id_vec),
                           schema_numbers_get(m_id_vec),
                           schema_data_get(m_id_vec),
                           m_id_vec)) return false;
  return true;
}
bool exa::cell::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) { 
  schema_get(::exa::schema::cell)
  if(!steve::retrieve<std::string>(a_store,
                              schema_data_get(m_id),
                              a_event,a_is,m_id)) return false;
  if(!steve::retrieve(a_store,
                           schema_numbers_get(m_id_vec),
                           schema_data_get(m_id_vec),
                           a_event,a_is,m_id_vec)) return false;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// event ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool exa::event::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  schema_get(::exa::schema::event)
  if(!steve::store<unsigned int>(a_store,
                          schema_data_get(m_id),
                          m_id)) return false;
  if(!steve::store<float>(a_store,
                          schema_data_get(m_energy),
                          m_energy)) return false;

  if(!steve::store_with_refs<exa::track>(a_store,
                             *(schema->schema_m_tracks()),
                             schema_offsets_get(m_tracks),
                             schema_numbers_get(m_tracks),
                             schema_offsets_get(m_tracks_refs),
                             schema_numbers_get(m_tracks_refs),
                             schema_data_get(m_tracks_refs),
                             m_tracks)) return false;

  if(!exa::util::store(a_store,
                       *(schema->schema_m_histo()),
                       m_histo.data())) return false;

  if(!steve::store(a_store,
                   schema_offsets_get(m_blob),
                   schema_numbers_get(m_blob),
                   schema_data_get(m_blob),
                   m_blob.size(),m_blob.data())) return false;

  if(!m_test_stream.store(a_store,*(schema->schema_m_test_stream()))) 
    return false;

  if(!steve::store<exa::cell>(a_store,
                             *(schema->schema_m_cells()),
                             schema_offsets_get(m_cells),
                             schema_numbers_get(m_cells),
                             m_cells)) return false;

  return true;
}
bool exa::event::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) {    
  schema_get(::exa::schema::event)
  if(!steve::retrieve<unsigned int>(a_store,
                             schema_data_get(m_id),
                             a_event,a_is,m_id)) return false;
  if(!steve::retrieve<float>(a_store,
                             schema_data_get(m_energy),
                             a_event,a_is,m_energy)) return false;
  if(!steve::retrieve_with_refs<exa::track>(a_store,
                             *(schema->schema_m_tracks()),
                             schema_numbers_get(m_tracks),
                             schema_data_get(m_tracks_refs),
                             a_event,a_is,m_tracks))return false;

  if(!exa::util::retrieve(a_store,
                          *(schema->schema_m_histo()),
                          a_event,a_is,m_histo.data())) return false;

 {steve::uint32 sz;
  char* buffer;
  if(!steve::retrieve(a_store,
                      schema_numbers_get(m_blob),
                      schema_data_get(m_blob),
                      a_event,a_is,sz,buffer)) return false;
  m_blob.assign(sz,buffer);}

  if(!m_test_stream.retrieve
    (a_store,*(schema->schema_m_test_stream()),a_event,a_is)) return false;

  if(!steve::retrieve<exa::cell>(a_store,
                             *(schema->schema_m_cells()),
                             schema_numbers_get(m_cells),
                             a_event,a_is,m_cells))return false;

  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// header ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool exa::header::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  schema_get(::exa::schema::header)
  if(!steve::store<unsigned int>(a_store,
                                schema_data_get(header_m_number_of_events),
                                m_number_of_events)) return false;
  return true;
}
bool exa::header::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) {    
  schema_get(::exa::schema::header)
  if(!steve::retrieve<unsigned int>(a_store,
                                   schema_data_get(header_m_number_of_events),
                                   a_event,a_is,m_number_of_events)) return false;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// analysis /////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool exa::analysis::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  schema_get(::exa::schema::analysis)
  if(!exa::util::store(a_store,
                       *(schema->schema_m_energy()),
                       m_energy.data())) return false;
  return true;
}
bool exa::analysis::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is) {    
  schema_get(::exa::schema::analysis)
  if(!exa::util::retrieve(a_store,
                          *(schema->schema_m_energy()),
                          a_event,a_is,m_energy.data())) return false;
  return true;
}

