// this :
#include <util/axis.h> 

//////////////////////////////////////////////////////////////////////////////
exa::util::axis::axis(
)
:m_offset(0)
,m_number_of_bins(0)
,m_minimum_value(0)
,m_maximum_value(0)
,m_fixed(true)
,m_bin_width(0)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
exa::util::axis::axis(
 const axis& a_axis
)
//:Iaxis()
:m_offset(a_axis.m_offset)
,m_number_of_bins(a_axis.m_number_of_bins)
,m_minimum_value(a_axis.m_minimum_value)
,m_maximum_value(a_axis.m_maximum_value)
,m_fixed(a_axis.m_fixed)
,m_bin_width(a_axis.m_bin_width)
,m_edges(a_axis.m_edges)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
exa::util::axis::~axis(
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
}
//////////////////////////////////////////////////////////////////////////////
exa::util::axis& exa::util::axis::operator=(
 const axis& a_axis
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  m_offset = a_axis.m_offset;
  m_number_of_bins = a_axis.m_number_of_bins;
  m_minimum_value = a_axis.m_minimum_value;
  m_maximum_value = a_axis.m_maximum_value;
  m_fixed = a_axis.m_fixed;
  m_bin_width = a_axis.m_bin_width;
  m_edges = a_axis.m_edges;
  return *this;
}
//////////////////////////////////////////////////////////////////////////////
bool exa::util::axis::configure(
 const std::vector<double>& a_edges
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  // init :
  m_number_of_bins = 0;
  m_minimum_value = 0;
  m_maximum_value = 0;
  m_fixed = true;
  m_bin_width = 0;
  m_edges.clear();
  // setup :
  if(a_edges.size()<=1) return false;
  uint32 number = a_edges.size()-1;
  for(uint32 index=0;index<number;index++) {
    if((a_edges[index]>=a_edges[index+1])) {
      return false;
    }
  }
  m_edges = a_edges;
  m_number_of_bins = number;
  m_minimum_value = a_edges[0];
  m_maximum_value = a_edges[m_number_of_bins];
  m_fixed = false;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool exa::util::axis::configure(
 uint32 a_number
,double a_min
,double a_max
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  // init :
  m_number_of_bins = 0;
  m_minimum_value = 0;
  m_maximum_value = 0;
  m_fixed = true;
  m_bin_width = 0;
  m_edges.clear();
  // setup :
  if(!a_number) return false;
  if(a_max<=a_min) return false;
  m_number_of_bins = a_number;
  m_minimum_value = a_min;
  m_maximum_value = a_max;
  m_bin_width = (a_max - a_min)/ a_number;
  m_fixed = true;
  return true;
}
//////////////////////////////////////////////////////////////////////////////
bool exa::util::axis::is_fixed_binning(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return m_fixed;
}
//////////////////////////////////////////////////////////////////////////////
const std::vector<double>& exa::util::axis::edges(
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return m_edges;
}
//////////////////////////////////////////////////////////////////////////////
double exa::util::axis::lower_edge(
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return m_minimum_value;
}
//////////////////////////////////////////////////////////////////////////////
double exa::util::axis::upper_edge(
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return m_maximum_value;
}
//////////////////////////////////////////////////////////////////////////////
exa::util::uint32 exa::util::axis::bins(
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  return m_number_of_bins;
}
//////////////////////////////////////////////////////////////////////////////
bool exa::util::axis::in_range_to_absolute_index(
 int a_index
,uint32& a_abs
) const
//////////////////////////////////////////////////////////////////////////////
// a_index is given in in-range indexing :
//  - [0,n-1] for in-range bins
//  - underflow_bin for the iaxis underflow bin
//  - overflow_bin for the iaxis overflow bin
// a_abs is the absolute indexing in [0,n+1].
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(a_index==axis::underflow_bin) {
    a_abs = 0;
    return true;
  } else if(a_index==axis::overflow_bin) {
    a_abs = m_number_of_bins+1;
    return true;
  } else if((a_index>=0)&&(a_index<m_number_of_bins)){
    a_abs = a_index + 1;
    return true;
  } else {
    return false;
  }
}
//////////////////////////////////////////////////////////////////////////////
bool exa::util::axis::coord_to_index(
 double a_value
,int& a_index
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if( a_value <  m_minimum_value) {
    a_index = axis::underflow_bin;
    return true;
  } else if( a_value >= m_maximum_value) {
    a_index = axis::overflow_bin;
    return true;
  } else {
    if(m_fixed) {
      a_index = (int)((a_value - m_minimum_value)/m_bin_width);
      return true;
    } else {
      for(uint32 index=0;index<m_number_of_bins;index++) {
        if((m_edges[index]<=a_value)&&(a_value<m_edges[index+1])) {
          a_index = index;
          return true;
        }
      }
      return false;
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
bool exa::util::axis::coord_to_absolute_index(
 double a_value
,uint32& a_index
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if( a_value <  m_minimum_value) {
    a_index = 0;
    return true;
  } else if( a_value >= m_maximum_value) {
    a_index = m_number_of_bins+1;
    return true;
  } else {
    if(m_fixed) {
      a_index = (uint32)((a_value - m_minimum_value)/m_bin_width)+1;
      return true;
    } else {
      for(uint32 index=0;index<m_number_of_bins;index++) {
        if((m_edges[index]<=a_value)&&(a_value<m_edges[index+1])) {
          a_index = index+1;
          return true;
        }
      }
      return false;
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
double exa::util::axis::bin_width(
 int a_bin
) const 
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(a_bin==axis::underflow_bin) {
    return 0; //FIXME return DBL_MAX;
  } else if(a_bin==axis::overflow_bin) {
    return 0; //FIXME return DBL_MAX;
  } else if((a_bin<0) ||(a_bin>=m_number_of_bins)) {
    return 0;
  } else {
    if(m_fixed) {
      return m_bin_width;
    } else {
      return (m_edges[a_bin+1]-m_edges[a_bin]);
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
double exa::util::axis::bin_lower_edge(
 int a_bin
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(a_bin==axis::underflow_bin) {
    return 0; //FIXME return -DBL_MAX;
  } else if(a_bin==axis::overflow_bin) {
    return 0; //FIXME return bin_upper_edge(m_number_of_bins-1);
  } else if((a_bin<0) ||(a_bin>=m_number_of_bins)) {
    return 0;
  } else {
    if(m_fixed) {
      return (m_minimum_value + a_bin * m_bin_width);
    } else {
      return m_edges[a_bin];
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
double exa::util::axis::bin_upper_edge(
 int a_bin
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(a_bin==axis::underflow_bin) {
    return 0; //FIXME bin_lower_edge(0)
  } else if(a_bin==axis::overflow_bin) {
    return 0; //FIXME return DBL_MAX;
  } else if((a_bin<0) ||(a_bin>=m_number_of_bins)) {
    return 0;
  } else {
    if(m_fixed) {
      return (m_minimum_value + (a_bin + 1) * m_bin_width);
    } else {
      return m_edges[a_bin+1];
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
double exa::util::axis::bin_center(
 int a_bin
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(a_bin==axis::underflow_bin) {
    return 0; //FIXME : -INF
  } else if(a_bin==axis::overflow_bin) {
    return 0; //FIXME : +INF
  } else if(a_bin<0) {
    return 0; //FIXME : -INF
  } else if(a_bin>=m_number_of_bins) {
    return 0; //FIXME : +INF
  } else {
    if(m_fixed) {
      return (m_minimum_value + (a_bin + 0.5) * m_bin_width);
    } else {
      return (m_edges[a_bin] + m_edges[a_bin+1])/2.;
    }
  }
}
//////////////////////////////////////////////////////////////////////////////
bool exa::util::axis::is_compatible(
 const axis& a_axis
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(m_number_of_bins!=a_axis.m_number_of_bins) return false;
  if(m_minimum_value!=a_axis.m_minimum_value) return false;
  if(m_maximum_value!=a_axis.m_maximum_value) return false;
  return true;
}
