#ifndef exa_util_axis_h
#define exa_util_axis_h

#include <string>
#include <vector>

namespace exa {
namespace util {

typedef unsigned int uint32;

class axis {
public:
  enum { underflow_bin = -2, overflow_bin = -1 };
public:
  bool is_fixed_binning() const;
  double lower_edge() const;
  double upper_edge() const;
  uint32 bins() const;
  double bin_lower_edge(int) const;
  double bin_upper_edge(int) const;
  double bin_width(int) const;
  double bin_center(int) const;
  bool coord_to_index(double,int&) const;
  bool coord_to_absolute_index(double,uint32&) const;
  bool in_range_to_absolute_index(int,uint32&) const;
public:
  // Partition :
  bool configure(const std::vector<double>&);
  bool configure(uint32,double,double);
  bool is_compatible(const axis&) const;
  // Partition and bin :
  const std::vector<double>& edges() const;
public:
  axis();
  axis(const axis&);
  virtual ~axis();
  axis& operator=(const axis&);
public:
  uint32 m_offset;
  uint32 m_number_of_bins;
  double m_minimum_value;
  double m_maximum_value;
  bool m_fixed;
  // Fixed size bins :
  double m_bin_width;
  // Variable size bins :
  std::vector<double> m_edges;
};

}}

#endif




