#ifndef exa_util_histogram_h
#define exa_util_histogram_h

#include <vector>

#include <util/axis.h>  

//#define EXA_UTIL_VEC_VEC

namespace exa {
namespace util {

class histogram_data {
public:
  histogram_data();
  histogram_data(const histogram_data&);
  virtual ~histogram_data();
public:
  void copy(const histogram_data&);
  inline double& bin_Sxw(uint32 a_bin,uint32 a_axis);
  inline double& bin_Sx2w(uint32 a_bin,uint32 a_axis);
  inline double bin_Sxw(uint32 a_bin,uint32 a_axis) const;
  inline double bin_Sx2w(uint32 a_bin,uint32 a_axis) const;
public:
  // General :
  std::string m_title;
  uint32 m_dimension;
  // Bins :
  uint32 m_bin_number;
  std::vector<uint32> m_bin_entries; //m_bin_number
  std::vector<double> m_bin_Sw;   //m_bin_number
  std::vector<double> m_bin_Sw2;  //m_bin_number
#ifdef EXA_UTIL_VEC_VEC
  std::vector< std::vector<double> > m_bin_Sxw;  //m_bin_number*m_dimension
  std::vector< std::vector<double> > m_bin_Sx2w; //m_bin_number*m_dimension
#else
  std::vector<double> m_bin_Sxw;  //m_bin_number*m_dimension
  std::vector<double> m_bin_Sx2w; //m_bin_number*m_dimension
#endif
  // Axes :
  std::vector<axis> m_axes;
  // Profile :
  bool m_is_profile;
  std::vector<double> m_bin_Svw;
  std::vector<double> m_bin_Sv2w;
  bool m_cut_v;
  double m_min_v;
  double m_max_v;
  // Args :
  std::vector<double> m_doubles;
  std::vector<int> m_ints;
};

class histogram : protected histogram_data {
public:
  histogram();
/*
  histogram(uint32,
            const std::vector<uint32>&,
            const std::vector<double>&,
            const std::vector<double>&,
            bool = false);
  histogram(uint32,
            const std::vector< std::vector<double> >&,
            bool = false);
*/
  histogram(const histogram&);
  histogram(const histogram_data&);
  virtual ~histogram();
  histogram& operator=(const histogram&);
public:
  histogram_data& data();
  const histogram_data& data() const;
public: //to be overriden for profile
  virtual void copy(const histogram&);               
  virtual bool add(const histogram&);
  virtual bool multiply(double); 
  virtual bool subtract(const histogram&);
  virtual double sum_bin_heights() const;
  virtual double sum_all_bin_heights() const;
  virtual double sum_extra_bin_heights() const;
  virtual double min_bin_height() const;
  virtual double max_bin_height() const;
protected:
  // Axes :
  virtual double ith_axis_bin_height(uint32,int) const;
  // Bins :
  virtual bool bin_height(const std::vector<int>&,double&) const;
  virtual bool bin_error(const std::vector<int>&,double&) const;
  // General :
  virtual void clear();
  virtual void base_reset();
  virtual void allocate();
public:
  bool reset();
  bool scale(double);
  bool is_compatible(const histogram&) const;
  bool multiply(const histogram&);
  bool divide(const histogram&);
  bool is_profile() const;
  inline bool offset(const std::vector<int>&,uint32&) const;
  bool offset(const std::vector<double>&,uint32&) const;
  void indices(uint32,std::vector<int>&) const;
  inline bool is_out(uint32) const;
  const exa::util::axis& axis(uint32) const;
  double sw() const;
  double sw2() const;
  uint32 number_of_bins() const;
  double ith_axis_Sxw(uint32) const;
  double ith_axis_Sx2w(uint32) const;
  bool is_valid() const;
  const std::string& title() const;
  uint32 dimension() const;
  uint32 entries() const;
  uint32 all_entries() const;
  //HISTOGRAM_COMMON
  bool set_title(const std::string&);
  uint32 extra_entries() const;
  double equivalent_bin_entries() const;
public:
  virtual void update_fast_getters();
protected:
  bool configure(uint32,
                 const std::vector<uint32>&,
                 const std::vector<double>&,
                 const std::vector<double>&,
                 bool = false);
  bool configure(uint32,
                 const std::vector< std::vector<double> >&,
                 bool = false);
  // Axes :
  uint32 ith_axis_bin_entries(uint32,int) const;
  double ith_axis_mean(uint32) const;
  double ith_axis_rms(uint32) const;
  // Bins :
  bool fill_bin(const std::vector<double>&,double);
  //FIXME bool fill_bin(const std::vector<double>&,double,double);
  bool bin_entries(const std::vector<int>&,uint32&) const;
  bool bin_mean(const std::vector<int>&,std::vector<double>&) const;
  bool bin_rms_axes(const std::vector<int>&,std::vector<double>&) const;
  bool bin_rms_value(const std::vector<int>&,double&) const; //FIXME
protected: //helpers
  // histogram_2d to arg histogram_1d
  bool fill_slice_x(histogram&,int,int) const;
  bool fill_slice_y(histogram&,int,int) const;
  // histogram_3d to arg histogram_2d
  bool fill_slice_xy(histogram&,int,int) const;
  bool fill_slice_yz(histogram&,int,int) const;
  bool fill_slice_xz(histogram&,int,int) const;
};

class profile : public histogram {
public:
  profile();
/*
  profile(uint32,
          const std::vector<uint32>&,
          const std::vector<double>&,
          const std::vector<double>&,
          bool = false);
  profile(uint32,
            const std::vector< std::vector<double> >&,
            bool = false);
*/
public:
  profile(const profile&);
  profile(const histogram_data&);
  virtual ~profile();
  profile& operator=(const profile&);
public:
  virtual void copy(const profile&);               
  virtual bool add(const profile&);
  virtual bool subtract(const profile&);
  //bool multiply(const histogram&);
  //bool divide(const histogram&);
  virtual bool multiply(double); 
  //HISTOGRAM_COMMON
  virtual double sum_bin_heights() const;
  virtual double sum_all_bin_heights() const;
  virtual double sum_extra_bin_heights() const;
  virtual double min_bin_height() const;
  virtual double max_bin_height() const;
protected:
  // Axes :
  virtual double ith_axis_bin_height(uint32,int) const;
  // Bins :
  virtual bool bin_height(const std::vector<int>&,double&) const;
  virtual bool bin_error(const std::vector<int>&,double&) const;
  // General :
  virtual void clear();
  virtual void base_reset();
  virtual void allocate();
protected:
  bool fill_bin(const std::vector<double>&,double,double);
  bool bin_rms_value(const std::vector<int>&,double&) const;
};

#define BASE_1D_HEADER \
public:\
  /* Partition :*/\
  bool coord_to_index(double,int&) const;\
  double mean() const;\
  double rms() const;\
  /* Bins :*/\
  bool bin_entries(int,uint32&) const;\
  virtual bool bin_height(int,double&) const;\
  virtual bool bin_error(int,double&) const;\
  double bin_center(int) const;\
  bool bin_mean(int,double&) const;\
  bool bin_rms(int,double&) const;\
  /* axis :*/\
  const util::axis& axis() const;

class histogram_1d : public histogram {
  BASE_1D_HEADER
public:
  bool fill(double,double = 1.0);
public:
  //FIXME histogram_1d();
  histogram_1d(const std::string&,uint32,double,double);
  histogram_1d(const std::string&,const std::vector<double>&);
  histogram_1d(const histogram_1d&);
  virtual ~histogram_1d();
};

class profile_1d : public profile {
  BASE_1D_HEADER
public:
  bool fill(double,double,double = 1.0);
  bool bin_rms_value(int,double&) const;
public:
  //FIXME profile_1d();
  profile_1d(const std::string&,uint32,double,double);
  profile_1d(const std::string&,uint32,double,double,double,double);
  profile_1d(const std::string&,const std::vector<double>&);
  profile_1d(const std::string&,const std::vector<double>&,double,double);
  profile_1d(const profile_1d&);
  virtual ~profile_1d();
};

#define BASE_2D_HEADER \
public:\
  /* Partition :*/\
  double mean_x() const;\
  double rms_x() const;\
  bool coord_to_index_x(double,int&) const;\
  double mean_y() const;\
  double rms_y() const;\
  bool coord_to_index_y(double,int&) const;\
  bool bin_entries(int,int,uint32&) const;\
  virtual bool bin_height(int,int,double&) const;\
  virtual bool bin_error(int,int,double&) const;\
  double bin_center_x(int) const;\
  double bin_center_y(int) const;\
  bool bin_mean_x(int,int,double&) const;\
  bool bin_mean_y(int,int,double&) const;\
  bool bin_rms_x(int,int,double&) const;\
  bool bin_rms_y(int,int,double&) const;\
  /* Axes :*/\
  const util::axis& x_axis() const;\
  const util::axis& y_axis() const;\
  /* Projection :*/\
  bool bin_entries_x(int,uint32&) const;\
  bool bin_height_x(int,double&) const;\
  bool bin_entries_y(int,uint32&) const;\
  bool bin_height_y(int,double&) const;\
private:\
  virtual void update_fast_getters();\
protected:\
  uint32 m_in_range_entries;\
  double m_in_range_Sw;\
  double m_in_range_Sxw;\
  double m_in_range_Syw;\
  double m_in_range_Sx2w;\
  double m_in_range_Sy2w;

class histogram_2d : public histogram {
  BASE_2D_HEADER
public:
  bool fill(double,double,double = 1.0);
public:
  //FIXME histogram_2d();
  histogram_2d(const std::string&,
               uint32,double,double,
               uint32,double,double);
  histogram_2d(const std::string&,
              const std::vector<double>&,
              const std::vector<double>&);
  histogram_2d(const histogram_2d&);
  virtual ~histogram_2d();
public:
  // Slices :
  histogram_1d* slice_x(int,int) const;
  histogram_1d* slice_y(int,int) const;
  histogram_1d* projection_x() const;
  histogram_1d* projection_y() const;
};

class profile_2d : public profile {
  BASE_2D_HEADER
public:
  bool fill(double,double,double,double = 1.0);
  bool bin_rms_value(int,int,double&) const;
public:
  //FIXME profile_2d();
  profile_2d(const std::string&,
             uint32,double,double,
             uint32,double,double);
  profile_2d(const std::string&,
             uint32,double,double,
             uint32,double,double,
             double,double);
  profile_2d(const std::string&,
            const std::vector<double>&,
            const std::vector<double>&);
  profile_2d(const std::string&,
            const std::vector<double>&,
            const std::vector<double>&,
            double,double);
  profile_2d(const profile_2d&);
  virtual ~profile_2d();
};

#define BASE_3D_HEADER \
public:\
  /* Partition :*/\
  double mean_x() const;\
  double rms_x() const;\
  bool coord_to_index_x(double,int&) const;\
  double mean_y() const;\
  double rms_y() const;\
  bool coord_to_index_y(double,int&) const;\
  double mean_z() const;\
  double rms_z() const;\
  bool coord_to_index_z(double,int&) const;\
  /* ints :*/\
  bool bin_entries(int,int,int,uint32&) const;\
  bool bin_height(int,int,int,double&) const;\
  bool bin_error(int,int,int,double&) const;\
  double bin_center_x(int) const;\
  double bin_center_y(int) const;\
  double bin_center_z(int) const;\
  bool bin_mean_x(int,int,int,double&) const;\
  bool bin_mean_y(int,int,int,double&) const;\
  bool bin_mean_z(int,int,int,double&) const;\
  bool bin_rms_x(int,int,int,double&) const;\
  bool bin_rms_y(int,int,int,double&) const;\
  bool bin_rms_z(int,int,int,double&) const;\
  /* Axes :*/\
  const util::axis& x_axis() const;\
  const util::axis& y_axis() const;\
  const util::axis& z_axis() const;\
  /* Projection :*/\
  bool bin_entries_x(int,uint32&) const;\
  bool bin_entries_y(int,uint32&) const;\
  bool bin_entries_z(int,uint32&) const;\
  bool bin_height_x(int,double&) const;\
  bool bin_height_y(int,double&) const;\
  bool bin_height_z(int,double&) const;\
private:\
  virtual void update_fast_getters();\
protected:\
  uint32 m_in_range_entries;\
  double m_in_range_Sw;\
  double m_in_range_Sxw;\
  double m_in_range_Syw;\
  double m_in_range_Szw;\
  double m_in_range_Sx2w;\
  double m_in_range_Sy2w;\
  double m_in_range_Sz2w;

class histogram_3d : public histogram {
  BASE_3D_HEADER
public:
  bool fill(double,double,double,double = 1.0);
public:
  //FIXME histogram_3d();
  histogram_3d(const std::string&,
              uint32,double,double,
              uint32,double,double,
              uint32,double,double);
  histogram_3d(const std::string&,
              const std::vector<double>&,
              const std::vector<double>&,
              const std::vector<double>&);
  histogram_3d(const histogram_3d&);
  virtual ~histogram_3d();
public:
  // Slices :
  histogram_2d* slice_xy(int,int) const;
  histogram_2d* slice_yz(int,int) const;
  histogram_2d* slice_xz(int,int) const;
};

}}

//////////////////////////////////////////////////////////////////////////////
inline bool exa::util::histogram::is_out(
 uint32 a_offset
) const
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  uint32 offset = a_offset;
  uint32 index;
  for(int iaxis=m_dimension-1;iaxis>=0;iaxis--) { 
    index = offset/m_axes[iaxis].m_offset;
    if(index==0) return true;
    if(index==(m_axes[iaxis].m_number_of_bins+1)) return true;
    offset -= index * m_axes[iaxis].m_offset;
  }
  return false;
}

//////////////////////////////////////////////////////////////////////////////
inline bool exa::util::histogram::offset(
 const std::vector<int>& a_is
,uint32& a_offset
) const
//////////////////////////////////////////////////////////////////////////////
// a_is[iaxis] is given in in-range indexing :
//  - [0,n[iaxis]-1] for in-range bins
//  - underflow_bin for the iaxis underflow bin
//  - overflow_bin for the iaxis overflow bin
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  if(!m_dimension) return false;
  a_offset = 0;
  for(uint32 iaxis=0;iaxis<m_dimension;iaxis++) { 
    uint32 ibin;
    if(!m_axes[iaxis].in_range_to_absolute_index
          (a_is[iaxis],ibin)) return false;
    a_offset += ibin * m_axes[iaxis].m_offset;
  }
  return true;
}
//////////////////////////////////////////////////////////////////////////////
#ifdef EXA_UTIL_VEC_VEC
inline double& exa::util::histogram_data::bin_Sxw(uint32 a_bin,uint32 a_axis) {
  return m_bin_Sxw[a_bin][a_axis];
}
inline double& exa::util::histogram_data::bin_Sx2w(uint32 a_bin,uint32 a_axis) {
  return m_bin_Sx2w[a_bin][a_axis];
}
inline double exa::util::histogram_data::bin_Sxw(uint32 a_bin,uint32 a_axis) const {
  return m_bin_Sxw[a_bin][a_axis];
}
inline double exa::util::histogram_data::bin_Sx2w(uint32 a_bin,uint32 a_axis) const {
  return m_bin_Sx2w[a_bin][a_axis];
}
#else
inline double& exa::util::histogram_data::bin_Sxw(uint32 a_bin,uint32 a_axis) {
  return m_bin_Sxw[a_axis+m_dimension*a_bin];
}
inline double& exa::util::histogram_data::bin_Sx2w(uint32 a_bin,uint32 a_axis) {
  return m_bin_Sx2w[a_axis+m_dimension*a_bin];
}
inline double exa::util::histogram_data::bin_Sxw(uint32 a_bin,uint32 a_axis) const {
  return m_bin_Sxw[a_axis+m_dimension*a_bin];
}
inline double exa::util::histogram_data::bin_Sx2w(uint32 a_bin,uint32 a_axis) const {
  return m_bin_Sx2w[a_axis+m_dimension*a_bin];
}
#endif

#endif




