
#include "histogram.h"

#include <iostream>

static double rflat();
static double rgauss(double,double);
static double rbw(double,double);

//////////////////////////////////////////////////////////////////////////////
template <class T>
bool histogram_print_1d(
 T& a_histogram
)
//////////////////////////////////////////////////////////////////////////////
// A la HPRINT.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::cout << a_histogram.dimension() << a_histogram.title() << std::endl;
 {double hunder;
  if(!a_histogram.bin_height
        (exa::util::axis::underflow_bin,hunder)) return false;
  double hover;
  if(!a_histogram.bin_height
        (exa::util::axis::overflow_bin,hover)) return false;
  std::cout 
    << " * ENTRIES = " << a_histogram.all_entries()
    << " * ALL CHANNELS = " << a_histogram.sum_bin_heights()
    << " * UNDERFLOW = " << hunder
    << " * OVERFLOW = " << hover
    << std::endl;}
  std::cout 
    << " * BIN WID = " << a_histogram.axis().bin_width(0)
    << " * MEAN VALUE = " << a_histogram.mean()
    << " * R . M . S = " << a_histogram.rms()
    << std::endl;

  // Some bins :
  int bins = a_histogram.axis().bins();
 {exa::util::uint32 n;
  double h,e;
  if(!a_histogram.bin_entries(0,n)) return false;
  if(!a_histogram.bin_height(0,h)) return false;
  if(!a_histogram.bin_error(0,e)) return false;
  std::cout 
    << " * ENTRIES[0]   = " 
    << n
    << " * HEIGHT[0] = " 
    << h
    << " * ERROR[0] = "     
    << e
    << std::endl;}

 {exa::util::uint32 n;
  double h,e;
  if(!a_histogram.bin_entries(bins/2,n)) return false;
  if(!a_histogram.bin_height(bins/2,h)) return false;
  if(!a_histogram.bin_error(bins/2,e)) return false;
  std::cout 
    << " * ENTRIES[N/2] = " 
    << n
    << " * HEIGHT[N/2] = " 
    << h
    << " * ERROR[N/2] = " 
    << e
    << std::endl;}

 {exa::util::uint32 n;
  double h,e;
  if(!a_histogram.bin_entries(bins-1,n)) return false;
  if(!a_histogram.bin_height(bins-1,h)) return false;
  if(!a_histogram.bin_error(bins-1,e)) return false;
  std::cout 
    << " * ENTRIES[N-1] = " 
    << n
    << " * HEIGHT[N-1] = " 
    << h
    << " * ERROR[N-1] = " 
    << e
    << std::endl;}

  return true;
}
//////////////////////////////////////////////////////////////////////////////
template <class T>
bool histogram_print_2d(
 T& a_histogram
)
//////////////////////////////////////////////////////////////////////////////
// A la HPRINT.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  std::cout << a_histogram.dimension() << a_histogram.title() << std::endl;
  std::cout 
    << " * ENTRIES = " << a_histogram.all_entries() << std::endl;

  //   6 | 7 | 8
  //  -----------
  //   3 | 4 | 5
  //  -----------
  //   0 | 1 | 2

  double height_0;
  if(!a_histogram.bin_height(exa::util::axis::underflow_bin,
                             exa::util::axis::underflow_bin,
                             height_0)) return false;
  double height_2;
  if(!a_histogram.bin_height(exa::util::axis::overflow_bin,
                             exa::util::axis::underflow_bin,
                             height_2)) return false;
  double height_6;
  if(!a_histogram.bin_height(exa::util::axis::underflow_bin,
                             exa::util::axis::overflow_bin,
                             height_6)) return false;
  double height_8;
  if(!a_histogram.bin_height(exa::util::axis::overflow_bin,
                             exa::util::axis::overflow_bin,
                             height_8)) return false;

  int i,j;

  double height_1 = 0;
  double height_7 = 0;
  for(i=0;i<a_histogram.x_axis().bins();i++){
   {double h;
    if(!a_histogram.bin_height
          (i,exa::util::axis::underflow_bin,h)) return false;
    height_1 += h;}

   {double h;
    if(!a_histogram.bin_height
          (i,exa::util::axis::overflow_bin,h)) return false;
    height_7 += h;}
  }

  double height_3 = 0;
  double height_5 = 0;
  for(j=0;j<a_histogram.y_axis().bins();j++){
   {double h;
    if(!a_histogram.bin_height
          (exa::util::axis::underflow_bin,j,h)) return false;
    height_3 += h;}
   {double h;
    if(!a_histogram.bin_height
          (exa::util::axis::overflow_bin,j,h)) return false;
    height_5 += h;}
  }

  double height_4 = 0;
  for(i=0;i<a_histogram.x_axis().bins();i++){
    for(j=0;j<a_histogram.y_axis().bins();j++){
      double h;
      if(!a_histogram.bin_height(i,j,h)) return false;
      height_4 += h;
    }
  }

  std::cout 
    << " " << height_6 << " " << height_7 << " " << height_8 << std::endl;
  std::cout 
    << " " << height_3 << " " << height_4 << " " << height_5 << std::endl;
  std::cout 
    << " " << height_0 << " " << height_1 << " " << height_2 << std::endl;

  // Some bins :
  int xbins = a_histogram.x_axis().bins();
  int ybins = a_histogram.y_axis().bins();

 {exa::util::uint32 n;
  double h,e;
  if(!a_histogram.bin_entries(0,0,n)) return false;
  if(!a_histogram.bin_height(0,0,h)) return false;
  if(!a_histogram.bin_error(0,0,e)) return false;
  std::cout 
    << " * ENTRIES[0,0]     = " 
    << n
    << " * HEIGHT[0,0] = " 
    << h
    << " * ERROR[0,0] = "     
    << e
    << std::endl;}

 {exa::util::uint32 n;
  double h,e;
  if(!a_histogram.bin_entries(xbins/2,ybins/2,n)) return false;
  if(!a_histogram.bin_height(xbins/2,ybins/2,h)) return false;
  if(!a_histogram.bin_error(xbins/2,ybins/2,e)) return false;
  std::cout 
    << " * ENTRIES[N/2,N/2] = " 
    << n
    << " * HEIGHT[N/2,N/2] = " 
    << h
    << " * ERROR[N/2,N/2] = " 
    << e
    << std::endl;}

 {exa::util::uint32 n;
  double h,e;
  if(!a_histogram.bin_entries(xbins-1,ybins-1,n)) return false;
  if(!a_histogram.bin_height(xbins-1,ybins-1,h)) return false;
  if(!a_histogram.bin_error(xbins-1,ybins-1,e)) return false;
  std::cout 
    << " * ENTRIES[N-1,N-1] = " 
    << n
    << " * HEIGHT[N-1,N-1] = " 
    << h
    << " * ERROR[N-1,N-1] = " 
    << e
    << std::endl;}

  return true;
}
//////////////////////////////////////////////////////////////////////////////
int main(
 int a_argc
,char** a_args
)
//////////////////////////////////////////////////////////////////////////////
//  Book and fill some histos.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  bool print = true;
  if(a_argc==2) {
    std::string s = a_args[1];
    if(s=="-noprint") print = false;
  }

  int entries = 1000000;
 {
   exa::util::histogram_1d* histogram = 
     new exa::util::histogram_1d("Gauss",100,-5,5);
   for(int count=0;count<entries;count++) {
     histogram->fill(rgauss(1,2),1.4);
   }
   if(print) histogram_print_1d<exa::util::histogram_1d>(*histogram);
   delete histogram;
 }

 {
   exa::util::histogram_1d* histogram = 
     new exa::util::histogram_1d("BW",100,-5,5);
   for(int count=0;count<entries;count++) {
     histogram->fill(rbw(0,1),2.3);
   }
   if(print) histogram_print_1d<exa::util::histogram_1d>(*histogram);
   delete histogram;
 }

 {
   exa::util::profile_1d* histogram = 
     new exa::util::profile_1d("Profile",100,-5,5,-2,2);
   for(int count=0;count<entries;count++) {
     histogram->fill(rgauss(1,2),rbw(0,1),1);
   }
   if(print) histogram_print_1d<exa::util::profile_1d>(*histogram);
   delete histogram;
 }

 {
   exa::util::histogram_2d* histogram = 
     new exa::util::histogram_2d("Gauss_BW",100,-5,5,100,-2,2);
   for(int count=0;count<entries;count++) {
     histogram->fill(rgauss(1,2),rbw(0,1),0.8);
   }
   if(print) histogram_print_2d<exa::util::histogram_2d>(*histogram);

  {
    exa::util::histogram_1d* projection = histogram->projection_x();
    if(!projection) return -1;
    projection->set_title("Gauss_BW_projection_x");
    if(print) histogram_print_1d<exa::util::histogram_1d>(*projection);
    delete projection;
  }
   
  {
    exa::util::histogram_1d* projection = histogram->projection_y();
    if(!projection) return -1;
    projection->set_title("Gauss_BW_projection_y");
    if(print) histogram_print_1d<exa::util::histogram_1d>(*projection);
    delete projection;
  }

  {
    exa::util::histogram_1d* slice = histogram->slice_x(40,60);
    if(!slice) return -1;
    slice->set_title("Gauss_BW_slice_x");
    if(print) histogram_print_1d<exa::util::histogram_1d>(*slice);
    delete slice;
  }

  {
    exa::util::histogram_1d* slice = 
      histogram->slice_x(exa::util::axis::underflow_bin,
                         exa::util::axis::underflow_bin);
    if(!slice) return -1;
    slice->set_title("Gauss_BW_slice_x_UNDER");
    if(print) histogram_print_1d<exa::util::histogram_1d>(*slice);
    delete slice;
  }

  {
    exa::util::histogram_1d* slice = 
      histogram->slice_x(exa::util::axis::overflow_bin,
                         exa::util::axis::overflow_bin);
    if(!slice) return -1;
    slice->set_title("Gauss_BW_slice_x_OVER");
    if(print) histogram_print_1d<exa::util::histogram_1d>(*slice);
    delete slice;
  }

  {
    exa::util::histogram_1d* slice = histogram->slice_y(30,50);
    if(!slice) return -1;
    slice->set_title("Gauss_BW_slice_y");
    if(print) histogram_print_1d<exa::util::histogram_1d>(*slice);
    delete slice;
  }

  {
    exa::util::histogram_1d* slice = 
      histogram->slice_y(exa::util::axis::underflow_bin,
                         exa::util::axis::underflow_bin);
    if(!slice) return -1;
    slice->set_title("Gauss_BW_slice_y_UNDER");
    if(print) histogram_print_1d<exa::util::histogram_1d>(*slice);
    delete slice;
  }

  {
    exa::util::histogram_1d* slice = 
      histogram->slice_y(exa::util::axis::overflow_bin,
                         exa::util::axis::overflow_bin);
    if(!slice) return -1;
    slice->set_title("Gauss_BW_slice_y_OVER");
    if(print) histogram_print_1d<exa::util::histogram_1d>(*slice);
    delete slice;
  }

   delete histogram;
 }

  return 0;
}

#include <stdlib.h> //rand
#include <math.h>
#ifndef M_PI
#define M_PI       3.1415926535897931160E0
#define M_PI_2     1.5707963267948965580E0  
#endif
//////////////////////////////////////////////////////////////////////////////
double rflat(
)
//////////////////////////////////////////////////////////////////////////////
// Shoot random numbers according a flat distribution.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  double value  = (double)::rand();
  value /= (double)RAND_MAX;
  return value;
}
//////////////////////////////////////////////////////////////////////////////
double rgauss(
 double aMean
,double aStdDev
)
//////////////////////////////////////////////////////////////////////////////
//  Shoot random numbers according a 
// gaussian distribution of mean 0 and sigma 1.
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  double v1,v2,r,fac;
  do {
    v1 = 2.0 * rflat() - 1.0;
    v2 = 2.0 * rflat() - 1.0;
    r = v1*v1 + v2*v2;
  } while ( r > 1.0 );
  fac = ::sqrt(-2.0*::log(r)/r);
  return (v2 * fac) * aStdDev + aMean;
}
//////////////////////////////////////////////////////////////////////////////
double rbw(
 double aMean
,double aGamma
)
//////////////////////////////////////////////////////////////////////////////
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!//
{
  double rval = 2.0 * rflat() - 1.0;
  double displ = 0.5 * aGamma * ::tan(rval * M_PI_2);
  return aMean + displ;
}
