// this :
#include <steve/HDF5/helper.h>

// steve :
#include <steve/HDF5/T.h>
#include <steve/HDF5/types.h>

// HDF5 :
#include <H5LT.h>

#define steve_failure (-1)

steve::HDF5::helper::helper(
 int a_level
)
//:m_compress(1)
:m_scalar(steve_failure)
//,m_compact(steve_failure)
,m_simple(steve_failure)
//,m_CPT(steve_failure)
{
  if(a_level<0) a_level = 0;
  if(a_level>9) a_level = 9;
  //m_compress = a_level;

  m_scalar = H5Screate(H5S_SCALAR);
  if(m_scalar<0) {
    m_scalar = steve_failure;
  }

/*m_compact = H5Pcreate(H5P_DATASET_CREATE);
  if(m_compact<0) {
    m_compact = steve_failure;
  } else {
    if(m_compress) {
      if(H5Pset_layout(m_compact,H5D_COMPACT)<0) { 
    //if(H5Pset_layout(m_compact,H5D_CONTIGUOUS)<0) { 
        H5Pclose(m_compact);
        m_compact = steve_failure;
      }
    }
  }
*/

  // For write_object :
 {hsize_t sz = 1;
  m_simple = H5Screate_simple(1,&sz,NULL);
  if(m_simple<0) {
     m_simple = steve_failure;
  }}

/*if(m_compress) {
    m_CPT = H5Pcreate(H5P_DATASET_CREATE);
    if(m_CPT<0) {
      m_CPT = steve_failure;
    } else {
      if(H5Pset_layout(m_CPT,H5D_COMPACT)<0) { 
    //if(H5Pset_layout(m_CPT,H5D_CONTIGUOUS)<0) { 
        H5Pclose(m_CPT);
        m_CPT = steve_failure;
      } else {
        hsize_t sz = 1;
        if(H5Pset_chunk(m_CPT,1,&sz)<0) {
          H5Pclose(m_CPT);
          m_CPT = steve_failure;
        } else {
          if(H5Pset_deflate(m_CPT,m_compress)<0) {
            H5Pclose(m_CPT);
            m_CPT = steve_failure;
          }
        }
      }
    }
  } else {
    m_CPT = H5Pcopy(H5P_DEFAULT);
  }
*/
}
steve::HDF5::helper::~helper() {
  if(m_scalar>=0) H5Sclose(m_scalar);
  //if(m_compact>=0) H5Pclose(m_compact);

  if(m_simple>=0) H5Sclose(m_simple);
  //if(m_CPT>=0) H5Pclose(m_CPT);
}
//private:
steve::HDF5::helper::helper(const helper&){}
//private:
steve::HDF5::helper& steve::HDF5::helper::operator=(const helper&) { return *this;}
bool steve::HDF5::helper::write_uint32(
 hid_t a_location
,const std::string& a_name
,uint32 a_data
){
  hid_t dataset = 
//    H5Dcreate(a_location,a_name.c_str(),H5T_STD_U32XX,m_scalar,m_compact);
    H5Dcreate(a_location,a_name.c_str(),H5T_STD_U32XX,m_scalar,H5P_DEFAULT);
  if(dataset<0) return false;

  if(H5Dwrite(dataset,H5T_NATIVE_UINT,H5S_ALL,H5S_ALL,H5P_DEFAULT,&a_data)<0) {
    H5Dclose(dataset);                 
    return false;
  }

  H5Dclose(dataset);                 
  return true;
}
bool steve::HDF5::helper::write_string_buffer(
 hid_t a_location
,const std::string& a_name
,unsigned int a_size
,const char* a_buffer
){
  // From H5LTmake_dataset_string.
  // buffer contains C strings in the form ab0cde00. There is
  // an extra NULL char at end.

  if(!a_size) return false;

  hid_t filetype = string_datatype(a_size);
  if(filetype<0) return false;

  hid_t dataset = 
    H5Dcreate(a_location,a_name.c_str(),filetype,m_scalar,H5P_DEFAULT);
  if(dataset<0) {
    H5Tclose(filetype);
    return false;
  }

  hid_t memtype = filetype;
  if(H5Dwrite(dataset,memtype,H5S_ALL,H5S_ALL,H5P_DEFAULT,a_buffer)<0) {
    H5Dclose(dataset);                 
    H5Tclose(filetype);
    return false;
  }

  H5Dclose(dataset);                 
  H5Tclose(filetype);
  return true;
}

hid_t steve::HDF5::string_datatype(unsigned int a_size) {
  // a_size should include the trailing null char.

  hid_t datatype = H5Tcopy(H5T_C_S1);
  if(datatype<0) return steve_failure;

  if(H5Tset_size(datatype,a_size)<0) {
    H5Tclose(datatype);
    return steve_failure;
  }

  if(H5Tset_strpad(datatype,H5T_STR_NULLTERM)<0) {
    H5Tclose(datatype);
    return steve_failure;
  }

  return datatype;
}
bool steve::HDF5::read_uint32(
 hid_t a_location
,const std::string& a_name
,uint32& a_value
){
  return read_scalar<uint32>(a_location,a_name,H5T_NATIVE_UINT,a_value);
}

bool steve::HDF5::read_cstrings(
 hid_t a_location
,const std::string& a_name
,unsigned int& a_size
,char*& a_buffer
){
  hid_t dataset = H5Dopen(a_location,a_name.c_str());
  if(dataset<0) return false;

  hid_t filetype = H5Dget_type(dataset);
  if(filetype<0) {
    H5Dclose(dataset);
    return false;
  }

  H5T_class_t t_class = H5Tget_class(filetype);
  if(t_class!=H5T_STRING) {
    H5Tclose(filetype);
    H5Dclose(dataset);
    return false;
  }
  
  size_t sz = H5Tget_size(filetype);
  //FIXME
  H5Tclose(filetype);

  // We could have use filetype since, for string, 
  // file type is the same than memory type.
  hid_t memtype = string_datatype(sz);
  if(memtype<0) {
    H5Dclose(dataset);
    return false;
  }

  a_buffer = new char[sz];

  if(!a_buffer) {
    H5Dclose(dataset);
    return false;
  }
  herr_t stat = H5Dread(dataset,memtype,H5S_ALL,H5S_ALL,H5P_DEFAULT,a_buffer);
  if(stat<0) {
    H5Tclose(memtype);
    H5Dclose(dataset);
    delete [] a_buffer;
    return false;
  }

  H5Tclose(memtype);
  H5Dclose(dataset);

  a_size = sz;

  return true;
}
