// this :
#include <steve/HDF5/store.h>

#include <steve/interface/store_data.h>

#include <steve/tools/smanip.h>

#include <steve/HDF5/store_pages.h>
#include <steve/HDF5/retrieve_pages.h>

#include <steve/HDF5/page_xxx.h>

#include <steve/version.h>

//#include <stdlib.h> //div

#define steve_failure (-1)

steve::HDF5::store::store(
 std::ostream& a_writer
,const std::string& a_file_name
,const std::string& a_options
)
:base_store(a_writer,a_file_name,a_options)
,m_file(steve_failure)
{
}

steve::HDF5::store::~store(){
}

//private:
steve::HDF5::store::store(const store& a_from)
:base_store(a_from)
,m_file(steve_failure){}
//private:
steve::HDF5::store& steve::HDF5::store::operator=(const store&) { return *this;}

void* steve::HDF5::store::cast_head() const {
  return (void*)static_cast<const steve::HDF5::store*>(this);
}

bool steve::HDF5::store::open(
 e_mode a_mode
)
{
  if(m_opened) return false;

  switch(a_mode) {
  case create:{
    m_file = 
      H5Fcreate(m_store_name.c_str(),H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
    if(m_file<0) {
      m_out << "steve::HDF5::store::open :"
            << " can't create file " << m_store_name
            << std::endl;
      m_file = steve_failure;
      return false;
    }

   {hid_t header = H5Gcreate(m_file,"steve.header",0);
    if(header<0) {
      m_out << "steve::HDF5::store::open :"
            << " can't create header group in file " << m_store_name
            << std::endl;
      H5Fclose(m_file);
      m_file = steve_failure;
      return false;
    }
    if(!m_helper.write_uint32(header,"major_version",STEVE_MAJOR_VERSION)) {
      m_out << "steve::HDF5::store::open :"
            << " can't write major_version in header of file " 
            << m_store_name
            << std::endl;
      H5Gclose(header);
      H5Fclose(m_file);
      m_file = steve_failure;
      return false;
    }  
    if(!m_helper.write_uint32(header,"minor_version",STEVE_MINOR_VERSION)) {
      m_out << "steve::HDF5::store::open :"
            << " can't write minor_version in header of file " 
            << m_store_name
            << std::endl;
      H5Gclose(header);
      H5Fclose(m_file);
      m_file = steve_failure;
      return false;
    }  
    if(!m_helper.write_uint32(header,"patch_version",STEVE_PATCH_VERSION)) {
      m_out << "steve::HDF5::store::open :"
            << " can't write patch_version in header of file " 
            << m_store_name
            << std::endl;
      H5Gclose(header);
      H5Fclose(m_file);
      m_file = steve_failure;
      return false;
    }  
    H5Gclose(header);}

    //FIXME : reset numbers.m_total, clear pages.

    m_opened = true;
    }return true;
  case readonly:{

    m_file = H5Fopen(m_store_name.c_str(),H5F_ACC_RDONLY, H5P_DEFAULT);
    if(m_file<0) {
      m_out << "steve::HDF5::store::open :"
            << " can't open file " << m_store_name
            << std::endl;
      m_file = steve_failure;
      return false;
    }

    // open the header group :
   {hid_t header = H5Gopen(m_file,"steve.header");
    if(header<0) {
      m_out << "steve::HDF5::store::open :"
            << " can't open header group in file " << m_store_name
            << std::endl;
      H5Fclose(m_file);
      m_file = steve_failure;
      return false;
    }
   {uint32 major_version;
    if(!read_uint32(header,"major_version",major_version)) {
      m_out << "steve::HDF5::store::open :"
            << " can't read major_version in header of file " 
            << m_store_name
            << std::endl;
      H5Gclose(header);
      H5Fclose(m_file);
      m_file = steve_failure;
      return false;
    }}
   {uint32 minor_version;
    if(!read_uint32(header,"minor_version",minor_version)) {
      m_out << "steve::HDF5::store::open :"
            << " can't read minor_version in header of file " 
            << m_store_name
            << std::endl;
      H5Gclose(header);
      H5Fclose(m_file);
      m_file = steve_failure;
      return false;
    }}
   {uint32 patch_version;
    if(!read_uint32(header,"patch_version",patch_version)) {
      m_out << "steve::HDF5::store::open :"
            << " can't read patch_version in header of file " 
            << m_store_name
            << std::endl;
      H5Gclose(header);
      H5Fclose(m_file);
      m_file = steve_failure;
      return false;
    }}
    H5Gclose(header);}

    m_opened = true;
    }return true;
  default:
   {m_out << "steve::HDF5::store::open :"
          << " bad mode."
          << std::endl;
    }break;
  }

  return false;
}

bool steve::HDF5::store::close()
{
  if(!m_opened) return false;

  bool status = true;

  if(H5Fclose(m_file)<0) status = false;
  m_file = steve_failure;

  m_opened = false;
  return status;
}

steve::interface::page* steve::HDF5::store::create_page(
 interface::store_data& a_store_data
,uint32 a_page_size
,uint32 a_compression
) {
  if(!a_store_data.store_pages()) {

    std::string group_name = "steve."+a_store_data.name();

    hid_t hid = H5Gcreate(m_file,group_name.c_str(),0);
    if(hid<0) {
      m_out << "steve::HDF5::store::create_page :"
            << " can't create group " << group_name
            << std::endl;
      return false;
    }
    if(!m_helper.write_uint32(hid,"page_size",a_page_size)) {
      m_out << "steve::HDF5::store::create_page :"
            << " can't write page_size in group " << group_name
            << std::endl;
      H5Gclose(hid);
      return false;
    }  

    a_store_data.set_store_pages(new store_pages(hid));
  }

  store_pages* pgs = (store_pages*)a_store_data.store_pages()->cast_head();

  std::string name = tools::to<uint32>(pgs->count());
  pgs->increment_count();
  //printf("debug : name \"%s\"\n",name.c_str());
  steve::interface::page* pg = 0;


#define if_new_save_page(a_class,a_id_type) \
  if(a_store_data.type()==a_id_type) {\
    pg = new a_class(m_out,pgs->id(),a_page_size,a_compression,name);\
  }

#define if_new_save_page_helper(a_class,a_id_type) \
  if(a_store_data.type()==a_id_type) {\
    pg = new a_class(m_out,pgs->id(),a_page_size,a_compression,name,m_helper);\
  }

  if_new_save_page(page_uchar,steve_id_uchar)
  else if_new_save_page(page_char,steve_id_char)
  else if_new_save_page(page_uint16,steve_id_uint16)
  else if_new_save_page(page_int16,steve_id_int16)
  else if_new_save_page(page_uint32,steve_id_uint32)
  else if_new_save_page(page_int32,steve_id_int32)
  else if_new_save_page(page_uint64,steve_id_uint64)
  else if_new_save_page(page_int64,steve_id_int64)
  else if_new_save_page(page_float,steve_id_float)
  else if_new_save_page(page_double,steve_id_double)

  else if_new_save_page(page_uint32,steve_id_ref)

  else if_new_save_page(page_bool,steve_id_bool)
  else if_new_save_page_helper(page_std_string,steve_id_string)

  else {
    return 0;
  }

#undef if_new_save_page
#undef if_new_save_page_helper

  pgs->push_back(pg);
  return pg;
}

bool steve::HDF5::store::commit_page(
 interface::store_data& a_store_data
,interface::page*& a_page) {
  //printf("debug : steve::HDF5::store::commit_page\n");
  if(!a_page->commit()) {
    m_out << "steve::store::commit_page :"
          << " problem saving page."
          << std::endl;
    return false;
  }

  if(!a_store_data.store_pages()) return false;

  store_pages* pgs = (store_pages*)a_store_data.store_pages()->cast_head();

  if(!pgs) return false;
  pgs->remove(a_page);

  delete a_page;
  a_page = 0; //invalidate the pointer.
  return true;
}

steve::interface::page* steve::HDF5::store::find_page(
 interface::store_data& a_store_data
,uint32 a_index
,uint32& a_index_in_page
)
{ 

  // Look if page is already loaded :
  if(!a_store_data.retrieve_pages()){

    uint32 page_size = 0;

    std::string group_name = "steve."+a_store_data.name();

    hid_t hid = H5Gopen(m_file,group_name.c_str());
    if(hid<0) {
      m_out << "steve::HDF5::store::find_page :"
            << " can't open group " << group_name
            << std::endl;
      return 0;
    }
    if(!read_uint32(hid,"page_size",page_size)) {
      m_out << "steve::HDF5::store::find_page :"
            << " can't read page_size in group " << group_name
            << std::endl;
      H5Gclose(hid);
      return 0;
    }

    a_store_data.set_retrieve_pages(new retrieve_pages(hid,page_size));
  }

  retrieve_pages* pgs = 
    (retrieve_pages*)a_store_data.retrieve_pages()->cast_head();

  uint32 page_size = pgs->page_size();

  uint32 page_index = a_index/page_size;
  a_index_in_page = a_index%page_size;

  //div_t dv = div(a_index,page_size); //optimization.
  //uint32 page_index = (uint32 )dv.quot;
  //a_index_in_page = (uint32 )dv.rem;

  if(page_index==pgs->first) return pgs->second;
  // Erase old page :
  delete pgs->second;
  pgs->first = steve_not_found;
  pgs->second = 0;

  std::string name = tools::to<uint32>(page_index);

  steve::interface::page* pg = 0;


#define if_new_retrieve_page(a_class,a_id_type) \
  if(a_store_data.type()==a_id_type) {\
    pg = new a_class(m_out,pgs->id(),page_size,0,name);\
  }

#define if_new_retrieve_page_helper(a_class,a_id_type) \
  if(a_store_data.type()==a_id_type) {\
    pg = new a_class(m_out,pgs->id(),page_size,0,name,m_helper);\
  }

  if_new_retrieve_page(page_uchar,steve_id_uchar)
  else if_new_retrieve_page(page_char,steve_id_char)
  else if_new_retrieve_page(page_uint16,steve_id_uint16)
  else if_new_retrieve_page(page_int16,steve_id_int16)
  else if_new_retrieve_page(page_uint32,steve_id_uint32)
  else if_new_retrieve_page(page_int32,steve_id_int32)
  else if_new_retrieve_page(page_uint64,steve_id_uint64)
  else if_new_retrieve_page(page_int64,steve_id_int64)
  else if_new_retrieve_page(page_float,steve_id_float)
  else if_new_retrieve_page(page_double,steve_id_double)

  else if_new_retrieve_page(page_uint32,steve_id_ref)

  else if_new_retrieve_page(page_bool,steve_id_bool)
  else if_new_retrieve_page_helper(page_std_string,steve_id_string)

  else {
    return 0;
  }

#undef if_new_retrieve_page
#undef if_new_retrieve_page_helper

  if(!pg->retrieve()) {
    m_out << "steve::HDF5::store::find_page :"
          << " problem retrieving page."
          << std::endl;
    return 0;
  }

  pgs->first = page_index;
  pgs->second = pg;

  return pg;
}
/*
bool steve::HDF5::store::store_number_of_events(
 uint32 a_number
)
{
  if(m_file==steve_failure) return false;

 {hid_t header = H5Gopen(m_file,"steve.header");
  if(header<0) {
    m_out << "steve::HDF5::store::store_number_of_events :"
          << " can't open header group."
          << std::endl;
    return false;
  }
  if(!m_helper.write_uint32(header,"number_of_events",a_number)) {
    m_out << "steve::HDF5::store::store_number_of_events :"
          << " can't write number_of_events."
          << std::endl;
    return false;
  }  
  H5Gclose(header);}

  return true;
}
bool steve::HDF5::store::retrieve_number_of_events(
 uint32& a_number
)
{
  if(m_file==steve_failure) return false;

 {hid_t header = H5Gopen(m_file,"steve.header");
  if(header<0) {
    m_out << "steve::HDF5::store::retrieve_number_of_events :"
          << " can't open header group."
          << std::endl;
    return false;
  }
  if(!read_uint32(header,"number_of_events",a_number)) {
    m_out << "steve::HDF5::store::retrieve_number_of_events :"
          << " can't read retrieve_number_of_events." 
          << std::endl;
    a_number = 0;
    return false;
  }
  H5Gclose(header);}

  return true;
}
*/
