// this :
#include <steve/ROOT/store.h>

#include <steve/tools/smanip.h>

//#include <steve/interface/store_data.h>

#include <TFile.h>
#include <TTree.h>

steve::ROOT::store::store(
 std::ostream& a_out
,const std::string& a_file_name
,const std::string& a_options
)
:base_store(a_out,a_file_name,a_options)
,m_file(0)
,m_write(false)
{
}

steve::ROOT::store::~store(){m_schemas.clear();}

//private:
steve::ROOT::store::store(const store& a_from)
:base_store(a_from)
,m_file(0)
,m_write(false)
{}
//private:
steve::ROOT::store& steve::ROOT::store::operator=(const store&) { return *this;}

void* steve::ROOT::store::cast_head() const {
  return (void*)static_cast<const steve::ROOT::store*>(this);
}

bool steve::ROOT::store::open(
 e_mode a_mode
)
{
  if(m_opened) return false;

  switch(a_mode) {
  case create:{

    int compression = 6;
   {std::string value;
    if(tools::option(m_options,"compression",value)) {
      unsigned int uival;
      if(tools::to<unsigned int>(value,uival)) compression = uival;
    }}

    //printf("debug : compression : %d\n",compression);

    m_file = new TFile(m_store_name.c_str(),"recreate","steve",compression);
    if(!m_file) {
      m_out << "steve::ROOT::store::open :"
            << " can't create file " << m_store_name
            << std::endl;
      m_file = 0;
      return false;
    }

    m_write = true;

/*FIXME : store :
    STEVE_MAJOR_VERSION
    STEVE_MINOR_VERSION
    STEVE_PATCH_VERSION
*/

    //FIXME : reset numbers.m_total, clear pages.

    m_opened = true;
    }return true;
  case readonly:{

    m_file = new TFile(m_store_name.c_str());
    if(!m_file) {
      m_out << "steve::ROOT::store::open :"
            << " can't open file " << m_store_name
            << std::endl;
      m_file = 0;
      return false;
    }

    m_write = false;

/*FIXME  : retrieve
    STEVE_MAJOR_VERSION
    STEVE_MINOR_VERSION
    STEVE_PATCH_VERSION
*/

    m_opened = true;
    }return true;
  default:
   {m_out << "steve::ROOT::store::open :"
          << " bad mode."
          << std::endl;
    }break;
  }

  return false;
}

bool steve::ROOT::store::close()
{
  if(!m_opened) return false;

  if(m_write) {
    std::map<std::string,TTree*>::const_iterator it;
    for(it=m_schemas.begin();it!=m_schemas.end();++it) {
      //it->second->Print();
      it->second->Write();
    }
    m_schemas.clear();
  }

  delete m_file;
  m_file = 0;

  m_opened = false;
  return true;
}

steve::interface::page* steve::ROOT::store::create_page(
 interface::store_data& //a_store_data
,uint32 //a_page_size
,uint32 //a_compression
) {
  return 0;
}

bool steve::ROOT::store::commit_page(
 interface::store_data& //a_store_data
,interface::page*& //a_page
) {
  return false;
}

steve::interface::page* steve::ROOT::store::find_page(
 interface::store_data& //a_store_data
,uint32 //a_index
,uint32& //a_index_in_page
)
{ 
  return 0;
}
bool steve::ROOT::store::bind_to_store(
 const std::string& a_schema
,const std::string& a_class_name
,void* a_pointer
,bool a_fill
) {
  if(!m_file) return false;
  if(!m_write) return false;

  std::map<std::string,TTree*>::iterator  mit(m_schemas.find(a_schema));
  TTree* tree = 0;
  if(mit!=m_schemas.end()) {
    tree = mit->second;
    tree->SetBranchAddress(a_schema.c_str(),a_pointer);
  } else {
    tree = new TTree(a_schema.c_str(),a_schema.c_str());
    if(!tree) {
      m_out << "steve::ROOT::store::bind_to_store :"
            << " can't create TTree " << a_schema
            << " in file " << m_store_name
            << std::endl;
      return false;
    }
    //NOTE : TClonesArray within TClonesArray can't be splitted !
    //int splitlevel = 99; //default.
    int splitlevel = 0; //exa_5 and exa_4 can be compared.
    tree->Branch(a_schema.c_str(),
                 a_class_name.c_str(),a_pointer,8000,splitlevel);
    //dump_branches();
    m_schemas[a_schema] = tree;
  }
  if(a_fill) {
    if(!tree->Fill()<0) return false;
  }  
  return true;
}
bool steve::ROOT::store::bind_to_retrieve(
 const std::string& a_schema
,void* a_pointer
,bool a_get_entry
,steve::uint64 a_entry
) {
  if(!m_file) return false;
  if(m_write) return false;
  std::map<std::string,TTree*>::iterator  mit(m_schemas.find(a_schema));
  TTree* tree = 0;
  if(mit!=m_schemas.end()) {
    tree = mit->second;
    tree->SetBranchAddress(a_schema.c_str(),a_pointer);
  } else {
    tree = (TTree*)m_file->Get(a_schema.c_str());
    if(!tree) {
      m_out << "steve::ROOT::store::bind_to_retrieve :"
            << " can't find TTree " << a_schema
            << " in file " << m_store_name
            << std::endl;
      return false;
    }
    tree->SetBranchAddress(a_schema.c_str(),a_pointer);
    m_schemas[a_schema] = tree;
  }
  if(a_get_entry) {
    Int_t ret = tree->GetEntry(a_entry);
    if(ret<=0) return false;
  }  
  return true;
}

/*
bool steve::ROOT::store::dump_branches() {
  if(!m_tree) return false;
  uint32 level = 0;
  TObjArray* branches = m_tree->GetListOfBranches();
  for(int index=0;index<branches->GetEntries();index++) {
    TBranch* branch = (TBranch*)((*branches)[index]);
    std::string spaces = tools::create(' ',level);
    m_out << spaces 
          << std::string(branch->GetName()) 
          << " split = " << branch->GetSplitLevel() 
          << std::endl;
    dump_branches(*branch,level+1);
  }
  return true;
}
void steve::ROOT::store::dump_branches(
 TBranch& a_branch
,uint32 a_level
) {
  TObjArray* branches = a_branch.GetListOfBranches();
  for(int index=0;index<branches->GetEntries();index++) {
    TBranch* branch = (TBranch*)((*branches)[index]);
    std::string spaces = tools::create(' ',a_level);
    m_out << spaces 
          << std::string(branch->GetName()) 
          << " split = " << branch->GetSplitLevel() 
          << std::endl;
    dump_branches(*branch,a_level+1);
  }
}
*/
