// this :
#include <steve/core/ascii_buffer.h>

#include <steve/tools/realloc.h>

#include <stdio.h> //sprintf
#include <stdlib.h> //strtol,strtod

#define MAXIMUM(a,b) ((a)>(b)?(a):(b)) 

steve::core::ascii_buffer::ascii_buffer(
 std::ostream& a_out
,uint32 a_buffer_size
,char* a_buffer
)
:base_buffer(a_out,a_buffer_size,a_buffer)
{}
steve::core::ascii_buffer::~ascii_buffer()
{}

//private:
steve::core::ascii_buffer::ascii_buffer(const ascii_buffer& a_buffer)
:base_buffer(a_buffer){}
steve::core::ascii_buffer& steve::core::ascii_buffer::operator=(const ascii_buffer&){return *this;}

#define MAX_SPRINTF_SIZE 256

//////////////////////////////////////////////////////////////////////////////
/// writers //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#define WRITE(a_type,a_format)\
bool steve::core::ascii_buffer::write(a_type a_v) {\
  if((m_pos+MAX_SPRINTF_SIZE)>m_max) \
    if(!expand(MAXIMUM(2*m_size,m_size+MAX_SPRINTF_SIZE))) return false;\
  int sz = ::sprintf(m_pos,a_format,a_v);\
  if(sz<0) return false;\
  if((sz+1)>MAX_SPRINTF_SIZE) return false;\
  m_pos += sz+1;\
  return true;\
}

WRITE(uchar,"%u")
WRITE(char,"%d")
WRITE(uint16,"%u")
WRITE(int16,"%d")
WRITE(uint32,"%u")
WRITE(int32,"%d")
WRITE(uint64,STEVE_UINT64_FORMAT)
WRITE(int64,STEVE_INT64_FORMAT)
WRITE(float,"%g")
WRITE(double,"%g")

#undef WRITE

bool steve::core::ascii_buffer::write(bool a_v) {
  if((m_pos+MAX_SPRINTF_SIZE)>m_max)
    if(!expand(MAXIMUM(2*m_size,m_size+MAX_SPRINTF_SIZE))) return false;
  int sz = ::sprintf(m_pos,"%u",a_v?1:0);
  if(sz<0) return false;
  if((sz+1)>MAX_SPRINTF_SIZE) return false;
  m_pos += sz+1;
  return true;
}

bool steve::core::ascii_buffer::write(const std::string& a_v) {
  uint32 sz = a_v.size()+1;
  if((m_pos+sz)>m_max)
    if(!expand(MAXIMUM(2*m_size,m_size+sz))) return false;
  ::memcpy(m_pos,a_v.c_str(),sz);
  m_pos += sz;
  return true;
}

#define WRITE_ARRAY(a_type)\
bool steve::core::ascii_buffer::write_array(uint32 a_n,const a_type* a_a) {\
  for(uint32 i=0;i<a_n;i++) {\
    if(!write(a_a[i])) return false;\
  }\
  return true;\
}

WRITE_ARRAY(uchar)
WRITE_ARRAY(char)
WRITE_ARRAY(uint16)
WRITE_ARRAY(int16)
WRITE_ARRAY(uint32)
WRITE_ARRAY(int32)
WRITE_ARRAY(uint64)
WRITE_ARRAY(int64)
WRITE_ARRAY(float)
WRITE_ARRAY(double)
WRITE_ARRAY(bool)

#undef WRITE_ARRAY

//////////////////////////////////////////////////////////////////////////////
/// readers //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#define READ(a_type,a_ret_type,a_func)\
bool steve::core::ascii_buffer::read(a_type& a_v) {\
  char* end;\
  a_ret_type v = a_func(m_pos,&end,10);\
  if(end==m_pos) return false;\
  if((end+1)>m_max) return false;\
  m_pos = end+1;\
  a_v = (a_type)v;\
  return true;\
}

READ(uchar,long,::strtol)
READ(char,long,::strtol)
READ(uint16,long,::strtol)
READ(int16,long,::strtol)
READ(uint32,long,::strtol)
READ(int32,long,::strtol)

#ifdef WIN32
READ(uint64,__int64,::_strtoi64)
READ(int64,__int64,::_strtoi64)
#else

#ifdef STEVE_INT64_AS_LONG
READ(uint64,long,::strtol)
READ(int64,long,::strtol)
#else //32bits platform :
READ(uint64,long long,::strtoll)
READ(int64,long long,::strtoll)
#endif

#endif

#undef READ

bool steve::core::ascii_buffer::read(bool& a_v) {
  uchar v;
  bool status = read(v);
  a_v = (v==1?true:false);
  return status;
}

bool steve::core::ascii_buffer::read(float& a_v){
  char* end;
#ifdef WIN32
  double v = ::strtod(m_pos,&end);
  a_v = (float)v;
#else
  a_v = ::strtof(m_pos,&end);
#endif
  if(end==m_pos) return false;
  if((end+1)>m_max) return false;
  m_pos = end+1;
  return true;
}
bool steve::core::ascii_buffer::read(double& a_v){
  char* end;
  a_v = ::strtod(m_pos,&end);
  if(end==m_pos) return false;
  if((end+1)>m_max) return false;
  m_pos = end+1;
  return true;
}

/*
bool steve::core::ascii_buffer::read(std::string& a_v) {
  char* pos = m_pos;
  char* begin = m_pos;
  steve::diff_pointer_t sz = remaining();
  for(uint32 index=0;index<sz;index++,pos++) {
    if((*pos)=='\0') {
      a_v.assign(begin);     
      begin = pos+1;
      m_pos = begin;
      return true; //stop at first string found.
    }
  }
  a_v.clear();
  return false;
}
*/

#define READ_ARRAY(a_type)\
bool steve::core::ascii_buffer::read_array(uint32& a_n,a_type*& a_a){\
  /* assume that the whole remaining buffer containes {xxx0}*/\
  /* a_a is allocated by this method. */\
  a_n = 0;\
  uint32 a_size = 100;\
  a_a = new a_type[a_size];\
  if(!a_a) return false;\
  char* pos = m_pos;\
  char* begin = m_pos;\
  steve::diff_pointer_t sz = remaining();\
  for(uint32 index=0;index<sz;index++,pos++) {\
    if((*pos)=='\0') {\
      m_pos = begin;\
      a_type v;\
      if(!read(v)) {\
        delete [] a_a;\
        return false;\
      }\
      if((a_n+1)>a_size) {\
        uint32 new_a_size = 2*a_size;\
        if(!tools::realloc<a_type>(a_a,new_a_size,a_size)) return false;\
        a_size = new_a_size;\
      }\
      a_a[a_n] = v;\
      a_n++;\
      begin = pos+1;\
    }\
  }\
  /* FIXME : return false if not ending with null char. */\
  return true;\
}

READ_ARRAY(uchar)
READ_ARRAY(char)
READ_ARRAY(uint16)
READ_ARRAY(int16)
READ_ARRAY(uint32)
READ_ARRAY(int32)
READ_ARRAY(uint64)
READ_ARRAY(int64)
READ_ARRAY(float)
READ_ARRAY(double)
READ_ARRAY(bool)

#undef READ_ARRAY

bool steve::core::ascii_buffer::read_array(std::vector<std::string>& a_a){
  /* assume that the whole remaining buffer containes {xxx0}*/
  /* a_a is allocated by this method. */
  a_a.clear();
  char* pos = m_pos;
  char* begin = m_pos;
  steve::diff_pointer_t sz = remaining();
  for(uint32 index=0;index<sz;index++,pos++) {
    if((*pos)=='\0') {
      a_a.push_back(begin);     
      begin = pos+1;
      m_pos = begin;
    }
  }
  //FIXME : return false if not ending with null char.
  return true;
}
