// this :
#include <steve/core/base_buffer.h>

#include <steve/util/mem.h>
#include <steve/tools/realloc.h>

#define steve_assert(e) \
  if (!(e)) {\
    m_out << "steve_assert : " << #e << " violated at line " \
          << __LINE__ << " of \"" << __FILE__ << "\"." << std::endl;\
    return;\
  }


steve::core::base_buffer::base_buffer(
 std::ostream& a_out
,uint32 a_buffer_size
,char* a_buffer
)
:m_out(a_out)
,m_verbose(0)
,m_size(0)
,m_buffer(0)
,m_pos(0)
,m_max(0)
{
  util::mem::increment("steve::core::base_buffer");
  // Before using the buffer make sure some assumptions are true
  steve_assert(sizeof(short) == 2)
  steve_assert(sizeof(int) == 4)
// commented lines below in view of support for longlong on 32 bits machines
//#ifdef R__B64
//   steve_assert(sizeof(long) == 8)
//#else
//   steve_assert(sizeof(long) == 4)
//#endif
  steve_assert(sizeof(float) == 4)
  steve_assert(sizeof(double) == 8)

  if(a_buffer) {
    m_size  = a_buffer_size;
    m_buffer = a_buffer;
  } else {
    uint32 MINIMAL = 128;
    m_size  = (a_buffer_size < MINIMAL) ? MINIMAL : a_buffer_size;   
    m_buffer = new char[m_size];
    if(!m_buffer) {
      m_out << "steve::core::base_buffer::buffer : can't alloc " 
            << m_size << " bytes." << std::endl;
      m_size = 0;
    }
  }

  m_pos = m_buffer;
  m_max = m_buffer + m_size;
}

steve::core::base_buffer::~base_buffer() {
  delete [] m_buffer;
  util::mem::decrement("steve::core::base_buffer");
}

//private:
steve::core::base_buffer::base_buffer(const base_buffer& a_buffer)
:m_out(a_buffer.m_out),m_verbose(0){}
steve::core::base_buffer& steve::core::base_buffer::operator=(const base_buffer&){return *this;}

void steve::core::base_buffer::assign(uint32 a_size,char* a_buffer) {
  delete [] m_buffer;
  m_size = a_size;
  m_buffer = a_buffer;
  m_pos = m_buffer;
  m_max = m_buffer + m_size;
}
void steve::core::base_buffer::clear(bool a_delete) {
  if(a_delete) delete [] m_buffer;
  m_size = 0;
  m_buffer = 0;
  m_pos = 0;
  m_max = 0;
}
char* steve::core::base_buffer::data() const { 
  return m_buffer;
}
steve::uint32 steve::core::base_buffer::size() const { 
  return m_size;
}
steve::diff_pointer_t steve::core::base_buffer::used() const { 
  return (diff_pointer_t)(m_pos - m_buffer);
}
steve::diff_pointer_t steve::core::base_buffer::remaining() const { 
  return (diff_pointer_t)(m_max - m_pos);
}
char* steve::core::base_buffer::position() const {
  return m_pos;
}
char* steve::core::base_buffer::max_position() const {
  return m_max;
}
bool steve::core::base_buffer::expand(uint32 a_new_size) {
  diff_pointer_t len = used();

  if(!steve::tools::realloc<char>(m_buffer,a_new_size,m_size)) {
    m_out << "steve::core::base_buffer::expand : can't realloc "
          << a_new_size * sizeof(char) << " bytes." 
          << std::endl;
    m_size = 0;
    m_pos = 0;
    m_max = 0;
    return false;
  }
  m_size = a_new_size;
  m_pos  = m_buffer + len;
  m_max  = m_buffer + m_size;
  return true;
}
