// this :
#include <steve/core/base_schema.h>

#include <steve/interface/data.h>
#include <steve/interface/store_data_visitor.h>

#include <steve/tools/smanip.h>
#include <steve/tools/sprintf.h>

#include <steve/core/stat_visitor.h>
#include <steve/core/offsets.h>
#include <steve/core/numbers.h>

steve::core::base_schema::base_schema(
 std::ostream& a_out
,const std::string& a_name
)
:m_out(a_out)
,m_name(a_name)
{
}
steve::core::base_schema::~base_schema(){
  clear();
}

//private:
steve::core::base_schema::base_schema(const base_schema& a_from)
:m_out(a_from.m_out)
{}
//private:
steve::core::base_schema& steve::core::base_schema::operator=(const base_schema&) { return *this;}

steve::interface::offsets* steve::core::base_schema::create_offsets(
 const std::string& a_name
,uint32 a_page_size
,uint32 a_compression
)
{
 {std::vector<interface::offsets*>::const_iterator it;
  for(it=m_os.begin();it!=m_os.end();++it) {
    if((*it)->store_data().name()==a_name) {
      m_out << "steve::base_schema::create_offsets :"
            << " " << a_name << " already exists."
            << std::endl;
      return 0;
    }
  }}

  steve::core::offsets* item = 
    new steve::core::offsets(m_out,a_name,a_page_size,a_compression);
  m_os.push_back(item);
  return item;
}
steve::interface::numbers* steve::core::base_schema::create_numbers(
 const path_os_t& a_path_os
,const std::string& a_name
,uint32 a_page_size
,uint32 a_compression
){
 {std::vector<interface::numbers*>::const_iterator it;
  for(it=m_ns.begin();it!=m_ns.end();++it) {
    if((*it)->store_data().name()==a_name) {
      m_out << "steve::base_schema::create_numbers :"
            << " " << a_name << " already exists."
            << std::endl;
      return 0;
    }
  }}

  steve::core::numbers* item = 
    new steve::core::numbers
      (m_out,a_path_os,a_name,a_page_size,a_compression);
  m_ns.push_back(item);
  return item;
}

void steve::core::base_schema::add_schema(steve::interface::schema* a_schema){
  m_schemas.push_back(a_schema);
}

bool steve::core::base_schema::visit_store_data(
 interface::store_data_visitor& a_visitor
,bool a_visit_sub_schemas
)
{
 {std::vector<interface::offsets*>::const_iterator it;
  for(it=m_os.begin();it!=m_os.end();++it) {
    if(!a_visitor.visit((*it)->store_data())) return false;
  }}
 {std::vector<interface::numbers*>::const_iterator it;
  for(it=m_ns.begin();it!=m_ns.end();++it) {
    if(!a_visitor.visit((*it)->store_data())) return false;
  }}
 {std::vector< interface::data* >::const_iterator it;
  for(it=m_datas.begin();it!=m_datas.end();++it) {
    if(!a_visitor.visit((*it)->store_data())) return false;
  }}

  if(a_visit_sub_schemas) {
    std::vector<steve::interface::schema*>::iterator it;
    for(it=m_schemas.begin();it!=m_schemas.end();++it) {
      if(!(*it)->visit_store_data(a_visitor)) return false;    
    }
  }

  return true;
}

namespace steve {
namespace core {
class commit_visitor : public virtual interface::store_data_visitor {
public:
  virtual bool visit(interface::store_data& a_store_data) {
    if(a_store_data.store_pages()) {
      interface::pages* pgs = a_store_data.store_pages();
      if(!pgs) return true; //it can be ok. (No data stored for this column).
      if(!pgs->commit()) {
        m_out << "steve::commit_visitor::visit :"
              << " problem committing pages."
              << std::endl;
        return false;
      }
    }
    return true;
  }
public:
  commit_visitor(std::ostream& a_out):m_out(a_out){}
  virtual ~commit_visitor(){}
private:
  commit_visitor(const commit_visitor& a_from):m_out(a_from.m_out){}
  commit_visitor& operator=(const commit_visitor&){ return *this;}
private:
  std::ostream& m_out;
};
}}

bool steve::core::base_schema::commit()
{
  commit_visitor v(m_out);
  if(!visit_store_data(v)) {
    m_out << "steve::core::store::close :"
          << " something went wrong in visitor."
          << std::endl;
    return false;
  }

 {std::vector<steve::interface::schema*>::iterator it;
  for(it=m_schemas.begin();it!=m_schemas.end();++it) {
    if(!(*it)->commit()) return false;    
  }}

  return true;
}

namespace steve {
namespace core {
class close_visitor : public virtual interface::store_data_visitor {
public:
  virtual bool visit(interface::store_data& a_store_data) {
    interface::pages* pgs = 0;
    if(a_store_data.store_pages()) {
      pgs = a_store_data.store_pages();
    } else if(a_store_data.retrieve_pages()) {
      pgs = a_store_data.retrieve_pages();
    }
    if(!pgs) return true; //it can be ok.
    if(!pgs->close()) {
      m_out << "steve::close_visitor::visit :"
            << " problem closing pages."
            << std::endl;
      return false;
    }
    return true;
  }
public:
  close_visitor(std::ostream& a_out):m_out(a_out){}
  virtual ~close_visitor(){}
private:
  close_visitor(const close_visitor& a_from):m_out(a_from.m_out){}
  close_visitor& operator=(const close_visitor&){ return *this;}
private:
  std::ostream& m_out;
};
}}

bool steve::core::base_schema::clear(){
  bool status = true;

  close_visitor v(m_out);
  if(!visit_store_data(v,false)) {
    m_out << "steve::core::store::clear :"
          << " something went wrong in visitor."
          << std::endl;
    status = false;
  }

 {std::vector<interface::offsets*>::const_iterator it;
  for(it=m_os.begin();it!=m_os.end();++it) delete (*it);
  m_os.clear();}
 {std::vector<interface::numbers*>::const_iterator it;
  for(it=m_ns.begin();it!=m_ns.end();++it) delete (*it);
  m_ns.clear();}
 {std::vector< interface::data* >::const_iterator it;
  for(it=m_datas.begin();it!=m_datas.end();++it) delete (*it);
  m_datas.clear();}

 {std::vector<steve::interface::schema*>::iterator it;
  for(it=m_schemas.begin();it!=m_schemas.end();++it) delete *it;
  m_schemas.clear();}

  return status;
}

//bool steve::core::base_schema::clear_store_data(){
//  return clear();
//}

namespace steve {
namespace core {
class find_visitor : public virtual interface::store_data_visitor {
public:
  virtual bool visit(interface::store_data& a_store_data) {
    if(a_store_data.name()==m_name) {
      m_store_data = &a_store_data;
      return false; //stop
    }
    return true;
  }
public:
  find_visitor(const std::string& a_name):m_name(a_name),m_store_data(0){}
  virtual ~find_visitor(){}
private:
  find_visitor(const find_visitor&){}
  find_visitor& operator=(const find_visitor&){ return *this;}
public:
  std::string m_name;
  steve::interface::store_data* m_store_data;  
};
}}

steve::interface::store_data* steve::core::base_schema::find_store_data(
 const std::string& a_name
) const
{
  find_visitor v(a_name);
  base_schema& self = const_cast<base_schema&>(*this);
  if(!self.visit_store_data(v)) return v.m_store_data;

 {std::vector<steve::interface::schema*>::const_iterator it;
  for(it=m_schemas.begin();it!=m_schemas.end();++it) {
    steve::interface::store_data* sd = (*it)->find_store_data(a_name);
    if(sd) return sd;
  }}

  return 0;
}

bool steve::core::base_schema::print_stats(std::ostream& a_out) const{
  stat_visitor v;
  base_schema& self = const_cast<base_schema&>(*this);
  if(!self.visit_store_data(v)) return false;

 {std::vector<steve::interface::schema*>::const_iterator it;
  for(it=m_schemas.begin();it!=m_schemas.end();++it) {
    if(!(*it)->visit_store_data(v)) return false;
  }}

 {std::string s;
  tools::sprintf(s,12,"%d",v.bytes());
  tools::justify(s,12,tools::right);
  a_out << "bytes total " << s << std::endl;}

 {std::string s;
  tools::sprintf(s,12,"%d",v.bytes_data());
  tools::justify(s,12,tools::right);
  a_out << "bytes data  " << s << std::endl;}

 {std::string s;
  tools::sprintf(s,12,"%d",v.bytes_numbers());
  tools::justify(s,12,tools::right);
  a_out << "bytes nums  " << s << std::endl;}

 {std::string s;
  tools::sprintf(s,12,"%d",v.bytes_offsets());
  tools::justify(s,12,tools::right);
  a_out << "bytes offs  " << s << std::endl;}

 {std::string s;
  tools::sprintf(s,12,"%d",v.bytes_refs());
  tools::justify(s,12,tools::right);
  a_out << "bytes refs  " << s << std::endl;}

  uint32 total = 
    v.bytes_data()+v.bytes_numbers()+v.bytes_offsets()+v.bytes_refs();
  if(total!=v.bytes()) {
    a_out << "bytes anomaly." << std::endl;
  }

  return true;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// basic type specific :
#include <steve/core/data.h>

steve::interface::data* steve::core::base_schema::create_data(
 const path_os_t& a_path_os
,const std::string& a_name
,const std::string& a_id_type
,uint32 a_page_size
,uint32 a_compression
){
 {std::vector<interface::data*>::const_iterator it;
  for(it=m_datas.begin();it!=m_datas.end();++it) {
    if((*it)->store_data().name()==a_name) {
      m_out << "steve::base_schema::create_data :"
            << " " << a_name << " already exists."
            << std::endl;
      return 0;
    }
  }}

  steve::interface::data* item = 0;

#define if_new_data(a__type,a__id_type) \
  if(a_id_type==a__id_type) {\
    item = new steve::core::data<a__type>\
      (m_out,a_path_os,a_name,a__id_type,a_page_size,a_compression);\
    m_datas.push_back(item);\
    return item;\
  }

  if_new_data(bool,steve_id_bool)
  else if_new_data(char,steve_id_char)
  else if_new_data(uchar,steve_id_uchar)
  else if_new_data(uint16,steve_id_uint16)
  else if_new_data(int16,steve_id_int16)
  else if_new_data(uint32,steve_id_uint32)
  else if_new_data(int32,steve_id_int32)
  else if_new_data(uint64,steve_id_uint64)
  else if_new_data(int64,steve_id_int64)
  else if_new_data(float,steve_id_float)
  else if_new_data(double,steve_id_double)
  else if_new_data(std::string,steve_id_string)

  else if_new_data(ref,steve_id_ref)

  else { 
    m_out << "steve::base_schema::create_data :"
          << " unknow type " << a_id_type
          << std::endl;
    return 0;
  }

#undef if_new_data

  return 0;
}

