// this :
#include <steve/core/base_store.h>

#include <steve/interface/data.h>
#include <steve/interface/store_data_visitor.h>

#include <steve/core/numbers.h>
#include <steve/core/offsets.h>

steve::core::base_store::base_store(
 std::ostream& a_out
,const std::string& a_store_name
,const std::string& a_options
)
:m_out(a_out)
,m_store_name(a_store_name)
,m_options(a_options)
,m_opened(false)
,m_verbosity(0)
,m_ref_counter(0)
{
}
steve::core::base_store::~base_store(){}

//private:
steve::core::base_store::base_store(const base_store& a_from)
:m_out(a_from.m_out)
{}
//private:
steve::core::base_store& steve::core::base_store::operator=(const base_store&) { return *this;}

void steve::core::base_store::set_verbosity(uint32 a_verbosity) {
  m_verbosity = a_verbosity;
}
steve::uint32 steve::core::base_store::verbosity() const {
  return m_verbosity;
}

bool steve::core::base_store::opened() const { return m_opened;}
std::string steve::core::base_store::name() const { return m_store_name;}

std::ostream& steve::core::base_store::printer() const {return m_out;}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool steve::core::base_store::create_store_ref(
 void* a_pointer
,ref& a_ref
)
// used in steve::helper::store_refs.
{
  a_ref = m_ref_counter;
  m_store_refs[a_pointer] = a_ref;
  m_ref_counter++;
  return true;
}
bool steve::core::base_store::find_store_ref(
 void* a_pointer
,ref& a_ref
)
// used in steve::helper::store_ref.
{
  std::map<void*,ref>::iterator  mit(m_store_refs.find(a_pointer));
  if(mit==m_store_refs.end()) {
    m_out << "steve::core::base_store::find_store_ref :"
          << " ref not found." 
          //<< " for data " 
          //<< a_data.store_data().name()
          << std::endl;
    return false;
  }
  a_ref = mit->second;
  return true;
}

bool steve::core::base_store::create_retrieve_ref(
 ref a_ref
,void* a_pointer
)
// used in steve::helper::retrieve_refs.
{
  m_retrieve_refs[a_ref] = a_pointer;
  return true;
}
bool steve::core::base_store::find_retrieve_ref(
 ref a_ref
,void*& a_pointer
)
// used in steve::helper::store_ref.
{
  std::map<ref,void*>::iterator  mit(m_retrieve_refs.find(a_ref));
  if(mit==m_retrieve_refs.end()) {
    m_out << "steve::core::base_store::find_retrieve_ref :"
          << " ref not found." 
          //<< " for data " 
          //<< a_data.store_data().name()
          << std::endl;
    return false;
  }
  a_pointer = mit->second;
  return true;
}
bool steve::core::base_store::reset_refs() {
  m_store_refs.clear();
  m_retrieve_refs.clear();
  return true;
}
