// this :
#include <steve/core/binary_swap_buffer.h>

#include <string.h> //memcpy

#include "io_swap.h"

steve::core::binary_swap_buffer::binary_swap_buffer(
 std::ostream& a_out
,uint32 a_buffer_size
,char* a_buffer
)
: core::base_buffer(a_out,a_buffer_size,a_buffer)
{}
steve::core::binary_swap_buffer::~binary_swap_buffer()
{}

//private:
steve::core::binary_swap_buffer::binary_swap_buffer(const binary_swap_buffer& a_buffer)
: core::base_buffer(a_buffer){}
steve::core::binary_swap_buffer& steve::core::binary_swap_buffer::operator=(const binary_swap_buffer&){return *this;}

//////////////////////////////////////////////////////////////////////////////
/// writers //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool steve::core::binary_swap_buffer::write(uchar x) {
  if(m_pos+sizeof(uchar)>m_max) if(!expand(2*m_size)) return false;
  *m_pos++ = x;
  return true;
}
bool steve::core::binary_swap_buffer::write(char x){ 
  return write((uchar)x); 
}
bool steve::core::binary_swap_buffer::write(uint16 x){
  if(m_pos+sizeof(uint16)>m_max) if(!expand(2*m_size)) return false;
  return swap::to(m_out,m_max,m_pos,x);
}
bool steve::core::binary_swap_buffer::write(int16 x){ 
  return write((uint16)x);
}
bool steve::core::binary_swap_buffer::write(uint32 x){
  if(m_pos+sizeof(uint32)>m_max) if(!expand(2*m_size)) return false;
  return swap::to(m_out,m_max,m_pos,x);
}
bool steve::core::binary_swap_buffer::write(int32 x){ 
  return write((uint32)x); 
}
bool steve::core::binary_swap_buffer::write(uint64 x){
  if(m_pos+sizeof(uint64)>m_max) if(!expand(2*m_size)) return false;
  return swap::to(m_out,m_max,m_pos,x);
}
bool steve::core::binary_swap_buffer::write(int64 x){ 
  return write((uint64)x); 
}
bool steve::core::binary_swap_buffer::write(float x){
  if(m_pos+sizeof(float)>m_max) if(!expand(2*m_size)) return false;
  return swap::to(m_out,m_max,m_pos,x);
}
bool steve::core::binary_swap_buffer::write(double x) {
  if(m_pos+sizeof(double)>m_max) if(!expand(2*m_size)) return false;
  return swap::to(m_out,m_max,m_pos,x);
}
bool steve::core::binary_swap_buffer::write(bool x){ 
  return write((uchar)(x?1:0)); 
}

#define MAXIMUM(a,b) ((a)>(b)?(a):(b)) 

/*
bool steve::core::binary_swap_buffer::write(const std::string& a_v) {
  uint32 sz = a_v.size();
  if(!write(sz)) return false;
  if((m_pos+sz)>m_max)
    if(!expand(MAXIMUM(2*m_size,m_size+sz))) return false;
  ::memcpy(m_pos,a_v.c_str(),sz);
  m_pos += sz;
  return true;
}
*/

#define WRITE_ARRAY(a_type)\
bool steve::core::binary_swap_buffer::write_array(uint32 a_n,const a_type* a_a) {\
  if(!a_n) return true;\
  uint32 l = a_n * sizeof(a_type);\
  if((m_pos+l)>m_max) if(!expand(MAXIMUM(2*m_size,m_size+l))) return false;\
  for(uint32 i=0;i<a_n;i++) {\
    if(!swap::to(m_out,m_max,m_pos,a_a[i])) return false;\
  }\
  return true;\
}

WRITE_ARRAY(uchar)
WRITE_ARRAY(char)
WRITE_ARRAY(uint16)
WRITE_ARRAY(int16)
WRITE_ARRAY(uint32)
WRITE_ARRAY(int32)
WRITE_ARRAY(uint64)
WRITE_ARRAY(int64)
WRITE_ARRAY(float)
WRITE_ARRAY(double)

#undef WRITE_ARRAY

bool steve::core::binary_swap_buffer::write_array(uint32 a_n,const bool* a_a) {
  if(!a_n) return true;
  uint32 l = a_n * sizeof(uchar);
  if((m_pos+l)>m_max) if(!expand(MAXIMUM(2*m_size,m_size+l))) return false;
  for(uint32 i=0;i<a_n;i++) {
    if(!swap::to(m_out,m_max,m_pos,(uchar)(a_a[i]?1:0))) return false;
  }
  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// readers //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool steve::core::binary_swap_buffer::read(uchar& x){
  return swap::from(m_out,m_max,m_pos,&x);
}
bool steve::core::binary_swap_buffer::read(char& x){
  return swap::from(m_out,m_max,m_pos,&x);
}
bool steve::core::binary_swap_buffer::read(uint16& x){
  return swap::from(m_out,m_max,m_pos,&x);
}
bool steve::core::binary_swap_buffer::read(int16& x){ 
  return swap::from(m_out,m_max,m_pos,&x);
}
bool steve::core::binary_swap_buffer::read(uint32& x){
  return swap::from(m_out,m_max,m_pos,&x);
}
bool steve::core::binary_swap_buffer::read(int32& x){ 
  return swap::from(m_out,m_max,m_pos,&x);
}
bool steve::core::binary_swap_buffer::read(uint64& x){
  return swap::from(m_out,m_max,m_pos,&x);
}
bool steve::core::binary_swap_buffer::read(int64& x){ 
  return swap::from(m_out,m_max,m_pos,&x);
}
bool steve::core::binary_swap_buffer::read(float& x){
  return swap::from(m_out,m_max,m_pos,&x);
}
bool steve::core::binary_swap_buffer::read(double& x){
  return swap::from(m_out,m_max,m_pos,&x);
}
bool steve::core::binary_swap_buffer::read(bool& x){
  uchar v;
  bool status = swap::from(m_out,m_max,m_pos,&v);
  x = (v==1?true:false);
  return status;
}

/*
bool steve::core::binary_swap_buffer::read(std::string& a_v) {
  uint32 sz = 0;
  if(!read(sz)) return false;
  a_v.resize(sz);  
  ::memcpy((char*)a_v.c_str(),m_pos,sz);
  m_pos += sz;
  return true;
}
*/

#define READ_ARRAY(a_type)\
bool steve::core::binary_swap_buffer::read_array(uint32& a_n,a_type*& a_a){\
  /* assume that the whole remaining buffer containes a set      */\
  /* of a_type data with each of the same size = sizeof(a_type). */\
  /* a_a is allocated by this method. */\
  a_n = remaining()/sizeof(a_type);\
  if(!a_n) {\
    a_a = 0;\
    return true;\
  }\
  a_a = new a_type[a_n];\
  for(uint32 i=0;i<a_n;i++) {\
    if(!swap::from(m_out,m_max,m_pos,a_a+i)) return false;\
  }\
  return true;\
}

READ_ARRAY(uchar)
READ_ARRAY(char)
READ_ARRAY(uint16)
READ_ARRAY(int16)
READ_ARRAY(uint32)
READ_ARRAY(int32)
READ_ARRAY(uint64)
READ_ARRAY(int64)
READ_ARRAY(float)
READ_ARRAY(double)

#undef READ_ARRAY

bool steve::core::binary_swap_buffer::read_array(uint32& a_n,bool*& a_a){
  /* assume that the whole remaining buffer containes a set      */
  /* of a_type data with each of the same size = sizeof(a_type). */
  /* a_a is allocated by this method. */
  a_n = remaining()/sizeof(uchar);
  if(!a_n) {
    a_a = 0;
    return true;
  }
  a_a = new bool[a_n];
  for(uint32 i=0;i<a_n;i++) {
    uchar v;
    if(!swap::from(m_out,m_max,m_pos,&v)) return false;
    *(a_a+i) = (v==1?true:false);
  }
  return true;
}


