#ifndef steve_core_io_no_swap_h
#define steve_core_io_no_swap_h

#include <steve/typedefs.h>

#include <string.h> //memcpy

namespace steve {
namespace core {
namespace no_swap {

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#define STEVE_TO_CHECK_EOB(n,cmt) \
  if((buf+n)>eob) {\
    a_out << "steve::core::no_swap : " << cmt << " : "\
          << " try to access out of buffer " << n << " bytes "\
          << " (buf=" << (void*)buf \
          << ", eob=" << (void*)eob << ")." << std::endl;\
    return false;}

inline bool to(std::ostream& a_out,const char* eob,char *&buf,uchar x) {
  STEVE_TO_CHECK_EOB(sizeof(uchar),"uchar")
  *buf = x;
  buf += sizeof(uchar);
  return true;
}

inline bool to(std::ostream& a_out,const char* eob,char *&buf,char x) { return to(a_out,eob,buf,(uchar)x);}

inline bool to(std::ostream& a_out,const char* eob,char*& buf,uint32 x) {
  STEVE_TO_CHECK_EOB(sizeof(uint32),"uint32")
  ::memcpy(buf,&x,sizeof(uint32));
  buf += sizeof(uint32);
  return true;
}

inline bool to(std::ostream& a_out,const char* eob,char *&buf,int32 x) { return to(a_out,eob,buf,(uint32)x);}

inline bool to(std::ostream& a_out,const char* eob,char *&buf,uint16 x) {
  STEVE_TO_CHECK_EOB(sizeof(uint16),"uint16")
  ::memcpy(buf, &x, sizeof(uint16));
  buf += sizeof(uint16);
  return true;
}


inline bool to(std::ostream& a_out,const char* eob,char*& buf,int16 x) { 
  return to(a_out,eob,buf,(uint16)x); 
}

inline bool to(std::ostream& a_out,const char* eob,char*& buf,uint64 x){
  STEVE_TO_CHECK_EOB(8,"uint64")
  if (sizeof(uint64) == 8) {
    ::memcpy(buf, &x, 8);
  } else {
    buf[0] = 0;
    buf[1] = 0;
    buf[2] = 0;
    buf[3] = 0;
    ::memcpy(buf+4, &x, 4);
  }
  buf += 8;
  return true;
}

inline bool to(std::ostream& a_out,const char* eob,char*& buf,int64 x)   { 
  return to(a_out,eob,buf,(uint64)x); 
}

inline bool to(std::ostream& a_out,const char* eob,char*& buf,float x) {
  STEVE_TO_CHECK_EOB(sizeof(float),"float")
  ::memcpy(buf,&x,sizeof(float));
  buf += sizeof(float);
  return true;
}

inline bool to(std::ostream& a_out,const char* eob,char *&buf,double x) {
  STEVE_TO_CHECK_EOB(sizeof(double),"double")
  ::memcpy(buf,&x,sizeof(double));
  buf += sizeof(double);
  return true;
}

#undef STEVE_TO_CHECK_EOB

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#define STEVE_FROM_CHECK_EOB(n,v,cmt) \
  if((buf+n)>eob) {\
    *x = v;\
    a_out << "steve::core::no_swap : " << cmt << " : "\
          << " try to access out of buffer " << n << " bytes"\
          << " (buf=" << (void*)buf \
          << ", eob=" << (void*)eob << ")." << std::endl;\
    return false;}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,uchar* x) {
  STEVE_FROM_CHECK_EOB(sizeof(uchar),0,"uchar")
  *x = *buf;
  buf += sizeof(uchar);
  return true;
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,char* x) { 
  return from(a_out,eob,buf,(uchar*)x); 
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,uint16* x) {
  STEVE_FROM_CHECK_EOB(sizeof(uint16),0,"uint16")
  ::memcpy(x,buf,sizeof(uint16));
  buf += sizeof(uint16);
  return true;
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,int16* x) { 
  return from(a_out,eob,buf,(uint16*)x); 
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,uint32* x) {
  STEVE_FROM_CHECK_EOB(sizeof(uint32),0,"uint32")
  ::memcpy(x,buf,sizeof(uint32));
  buf += sizeof(uint32);
  return true;
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,int32* x) { 
  return from(a_out,eob,buf,(uint32*)x); 
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,uint64 *x){
  STEVE_FROM_CHECK_EOB(8,0,"uint64")
  if (sizeof(uint64) == 8) {
    ::memcpy(x, buf, 8);
  } else {
    ::memcpy(x, buf+4, 4);
  }
  buf += 8;
  return true;
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,int64* x) { 
  return from(a_out,eob,buf,(uint64*)x); 
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,float* x) {
  STEVE_FROM_CHECK_EOB(sizeof(float),0,"float")
  ::memcpy(x, buf, sizeof(float));
  buf += sizeof(float);
  return true;
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,double* x) {
  STEVE_FROM_CHECK_EOB(sizeof(double),0,"double")
  ::memcpy(x, buf, sizeof(double));
  buf += sizeof(double);
  return true;
}


#undef STEVE_FROM_CHECK_EOB

}}}

#endif
