#ifndef steve_core_io_swap_h
#define steve_core_io_swap_h

#include <steve/typedefs.h>

namespace steve {
namespace core {
namespace swap {

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#define STEVE_TO_CHECK_EOB(n,cmt) \
  if((buf+n)>eob) {\
    a_out << "steve::core::swap::to : " << cmt << " : "\
          << " try to access out of buffer " << n << " bytes "\
          << " (buf=" << (void*)buf \
          << ", eob=" << (void*)eob << ")." << std::endl;\
    return false;}

inline bool to(std::ostream& a_out,const char* eob,char *&buf,uchar x) {
  STEVE_TO_CHECK_EOB(sizeof(uchar),"uchar")
  *buf = x;
  buf += sizeof(uchar);
  return true;
}

inline bool to(std::ostream& a_out,const char* eob,char *&buf,char x) { return to(a_out,eob,buf,(uchar)x);}

inline bool to(std::ostream& a_out,const char* eob,char*& buf,uint32 x) {
  STEVE_TO_CHECK_EOB(sizeof(uint32),"uint32")
  char* sw = (char*)&x;
  buf[0] = sw[3];
  buf[1] = sw[2];
  buf[2] = sw[1];
  buf[3] = sw[0];
  buf += sizeof(uint32);
  return true;
}

inline bool to(std::ostream& a_out,const char* eob,char *&buf,int32 x) { return to(a_out,eob,buf,(uint32)x);}

inline bool to(std::ostream& a_out,const char* eob,char *&buf,uint16 x) {
  STEVE_TO_CHECK_EOB(sizeof(uint16),"uint16")
  char* sw = (char*)&x;
  buf[0] = sw[1];
  buf[1] = sw[0];
  buf += sizeof(uint16);
  return true;
}


inline bool to(std::ostream& a_out,const char* eob,char*& buf,int16 x) { 
  return to(a_out,eob,buf,(uint16)x); 
}

inline bool to(std::ostream& a_out,const char* eob,char*& buf,uint64 x){
  STEVE_TO_CHECK_EOB(8,"uint64")
  char* sw = (char*)&x;
  if (sizeof(uint64) == 8) {
    buf[0] = sw[7];
    buf[1] = sw[6];
    buf[2] = sw[5];
    buf[3] = sw[4];
    buf[4] = sw[3];
    buf[5] = sw[2];
    buf[6] = sw[1];
    buf[7] = sw[0];
  } else {
    buf[0] = 0;
    buf[1] = 0;
    buf[2] = 0;
    buf[3] = 0;
    buf[4] = sw[3];
    buf[5] = sw[2];
    buf[6] = sw[1];
    buf[7] = sw[0];
  }
  buf += 8;
  return true;
}

inline bool to(std::ostream& a_out,const char* eob,char*& buf,int64 x)   { 
  return to(a_out,eob,buf,(uint64)x); 
}

inline bool to(std::ostream& a_out,const char* eob,char*& buf,float x) {
  STEVE_TO_CHECK_EOB(sizeof(float),"float")
  char* sw = (char*)&x;
  buf[0] = sw[3];
  buf[1] = sw[2];
  buf[2] = sw[1];
  buf[3] = sw[0];
  buf += sizeof(float);
  return true;
}

inline bool to(std::ostream& a_out,const char* eob,char *&buf,double x) {
  STEVE_TO_CHECK_EOB(sizeof(double),"double")
  char* sw = (char*)&x;
  buf[0] = sw[7];
  buf[1] = sw[6];
  buf[2] = sw[5];
  buf[3] = sw[4];
  buf[4] = sw[3];
  buf[5] = sw[2];
  buf[6] = sw[1];
  buf[7] = sw[0];
  buf += sizeof(double);
  return true;
}

#undef STEVE_TO_CHECK_EOB

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#define STEVE_FROM_CHECK_EOB(n,v,cmt) \
  if((buf+n)>eob) {\
    *x = v;\
    a_out << "steve::core::swap::from : " << cmt << " : "\
          << " try to access out of buffer " << n << " bytes"\
          << " (buf=" << (void*)buf \
          << ", eob=" << (void*)eob << ")." << std::endl;\
    return false;}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,uchar* x) {
  STEVE_FROM_CHECK_EOB(sizeof(uchar),0,"uchar")
  *x = *buf;
  buf += sizeof(uchar);
  return true;
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,char* x) { 
  return from(a_out,eob,buf,(uchar*)x); 
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,uint16* x) {
  STEVE_FROM_CHECK_EOB(sizeof(uint16),0,"uint16")
  char* sw = (char*)x;
  sw[0] = buf[1];
  sw[1] = buf[0];
  buf += sizeof(uint16);
  return true;
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,int16* x) { 
  return from(a_out,eob,buf,(uint16*)x); 
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,uint32* x) {
  STEVE_FROM_CHECK_EOB(sizeof(uint32),0,"uint32")
  char* sw = (char*)x;
  sw[0] = buf[3];
  sw[1] = buf[2];
  sw[2] = buf[1];
  sw[3] = buf[0];
  buf += sizeof(uint32);
  return true;
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,int32* x) { 
  return from(a_out,eob,buf,(uint32*)x); 
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,uint64 *x){
  STEVE_FROM_CHECK_EOB(8,0,"uint64")
  char* sw = (char*)x;
  if (sizeof(uint64) == 8) {
    sw[0] = buf[7];
    sw[1] = buf[6];
    sw[2] = buf[5];
    sw[3] = buf[4];
    sw[4] = buf[3];
    sw[5] = buf[2];
    sw[6] = buf[1];
    sw[7] = buf[0];
  } else {
    sw[0] = buf[7];
    sw[1] = buf[6];
    sw[2] = buf[5];
    sw[3] = buf[4];
  }
  buf += 8;
  return true;
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,int64* x) { 
  return from(a_out,eob,buf,(uint64*)x); 
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,float* x) {
  STEVE_FROM_CHECK_EOB(sizeof(float),0,"float")
  char* sw = (char*)x;
  sw[0] = buf[3];
  sw[1] = buf[2];
  sw[2] = buf[1];
  sw[3] = buf[0];
  buf += sizeof(float);
  return true;
}

inline bool from(std::ostream& a_out,const char* eob,char*& buf,double* x) {
  STEVE_FROM_CHECK_EOB(sizeof(double),0,"double")
  char* sw = (char*)x;
  sw[0] = buf[7];
  sw[1] = buf[6];
  sw[2] = buf[5];
  sw[3] = buf[4];
  sw[4] = buf[3];
  sw[5] = buf[2];
  sw[6] = buf[1];
  sw[7] = buf[0];
  buf += sizeof(double);
  return true;
}


#undef STEVE_FROM_CHECK_EOB

}}}

#endif
