// this :
#include <steve/core/stat_visitor.h>

#include <steve/tools/smanip.h>
#include <steve/tools/sprintf.h>

#include <steve/core/numbers.h>
#include <steve/core/offsets.h>

bool steve::core::stat_visitor::visit(interface::store_data& a_store_data) {
  m_bytes += a_store_data.store_bytes();
  if(dynamic_cast<numbers*>(&a_store_data)) {
    m_bytes_numbers += a_store_data.store_bytes();
  } else if(dynamic_cast<offsets*>(&a_store_data)) {
    m_bytes_offsets += a_store_data.store_bytes();
  } else if(a_store_data.type()==steve_id_ref) {
    m_bytes_refs += a_store_data.store_bytes();
  } else {
    m_bytes_data += a_store_data.store_bytes();
  }
  return true;
}

steve::core::stat_visitor::stat_visitor()
:m_bytes(0)
,m_bytes_data(0)
,m_bytes_numbers(0)
,m_bytes_offsets(0)
,m_bytes_refs(0)
{}
steve::core::stat_visitor::~stat_visitor(){}

steve::core::stat_visitor::stat_visitor(const stat_visitor& a_from)
:m_bytes(0)
,m_bytes_data(0)
,m_bytes_numbers(0)
,m_bytes_offsets(0)
,m_bytes_refs(0)
{}

steve::core::stat_visitor& steve::core::stat_visitor::operator=(const stat_visitor&){ return *this;}

void steve::core::stat_visitor::reset() {
  m_bytes = 0;
  m_bytes_data = 0;
  m_bytes_numbers = 0;
  m_bytes_offsets = 0;
  m_bytes_refs = 0;
}

steve::uint32 steve::core::stat_visitor::bytes() const { return m_bytes;}
steve::uint32 steve::core::stat_visitor::bytes_data() const { return m_bytes_data;}
steve::uint32 steve::core::stat_visitor::bytes_numbers() const { return m_bytes_offsets;}
steve::uint32 steve::core::stat_visitor::bytes_offsets() const { return m_bytes_numbers;}
steve::uint32 steve::core::stat_visitor::bytes_refs() const { return m_bytes_refs;}


void steve::core::stat_visitor::print(std::ostream& a_out) const{

 {std::string s;
  tools::sprintf(s,12,"%d",bytes());
  tools::justify(s,12,tools::right);
  a_out << "bytes total " << s << std::endl;}

 {std::string s;
  tools::sprintf(s,12,"%d",bytes_data());
  tools::justify(s,12,tools::right);
  a_out << "bytes data  " << s << std::endl;}

 {std::string s;
  tools::sprintf(s,12,"%d",bytes_numbers());
  tools::justify(s,12,tools::right);
  a_out << "bytes nums  " << s << std::endl;}

 {std::string s;
  tools::sprintf(s,12,"%d",bytes_offsets());
  tools::justify(s,12,tools::right);
  a_out << "bytes offs  " << s << std::endl;}

 {std::string s;
  tools::sprintf(s,12,"%d",bytes_refs());
  tools::justify(s,12,tools::right);
  a_out << "bytes refs  " << s << std::endl;}

  uint32 total = 
    bytes_data()+bytes_numbers()+bytes_offsets()+bytes_refs();
  if(total!=bytes()) {
    a_out << "bytes anomaly." << std::endl;
  }
}
