// this :
#include <steve/core/streamer.h>

#include <steve/interface/store.h>

#include <steve/util/mem.h>

#include <steve/core/platform.h>
#include <steve/core/binary_no_swap_buffer.h>
#include <steve/core/binary_swap_buffer.h>

steve::core::streamer::streamer(
 interface::store& a_store
)
:m_store(a_store)
,m_buffer(0)
{
  util::mem::increment("steve::core::streamer");
  if(steve::core::is_little_endian()) { //We do not byte swap on intels.
    m_buffer = new steve::core::binary_no_swap_buffer(a_store.printer());
  } else {
    m_buffer = new steve::core::binary_swap_buffer(a_store.printer());
  }
}
steve::core::streamer::~streamer()
{
  delete m_buffer;
  util::mem::decrement("steve::core::streamer");
}

//private:
steve::core::streamer::streamer(const streamer& a_from)
:m_store(a_from.m_store){}
steve::core::streamer& steve::core::streamer::operator=(const streamer&){return *this;}

steve::interface::store& steve::core::streamer::store() { return m_store;}

void steve::core::streamer::assign(uint32 a_size,char* a_buffer)
{
  m_buffer->assign(a_size,a_buffer);
}
void steve::core::streamer::clear(bool a_delete) {
  m_buffer->clear(a_delete);
}
char* steve::core::streamer::data() const { 
  return m_buffer->data();
}
steve::uint32 steve::core::streamer::size() const { 
  return m_buffer->used();
}

//////////////////////////////////////////////////////////////////////////////
/// writers //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#define WRITE(a_type)\
bool steve::core::streamer::write(a_type a_v) {\
  return m_buffer->write(a_v);\
}

WRITE(uchar)
WRITE(char)
WRITE(uint16)
WRITE(int16)
WRITE(uint32)
WRITE(int32)
WRITE(uint64)
WRITE(int64)
WRITE(float)
WRITE(double)

#undef WRITE

bool steve::core::streamer::write(bool a_v) {
  return m_buffer->write((uchar)(a_v?1:0));
}
/*XXX
bool steve::core::streamer::write(int a_v) {
  return m_buffer->write(int32(a_v));
}
bool steve::core::streamer::write(unsigned int a_v) {
  return m_buffer->write(uint32(a_v));
}
*/
bool steve::core::streamer::write(const std::string& a_v) {
  uint32 sz = a_v.size();
  if(!m_buffer->write(sz)) return false;
  //FIXME : optimize :
  typedef std::string::const_iterator it_t;
  for(it_t it=a_v.begin();it!=a_v.end();++it) {
    if(!m_buffer->write(*it)) return false;
  }
  return true;
}

bool steve::core::streamer::write(uint32 a_size,char* a_buffer) { 
  if(!m_buffer->write(a_size)) return false;
  //FIXME : optimize with memcpy.
 {for(uint32 index=0;index<a_size;index++) {
    if(!m_buffer->write(a_buffer[index])) return false;
  }}
  return true;
}


//////////////////////////////////////////////////////////////////////////////
/// readers //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#define READ(a_type)\
bool steve::core::streamer::read(a_type& a_v) {\
  return m_buffer->read(a_v);\
}

READ(uchar)
READ(char)
READ(uint16)
READ(int16)
READ(uint32)
READ(int32)
READ(uint64)
READ(int64)
READ(float)
READ(double)

#undef READ

bool steve::core::streamer::read(bool& a_v){
  uchar v;
  bool status = m_buffer->read(v);
  a_v = (v==0?false:true);
  return status;
}
/*XXX
bool steve::core::streamer::read(int& a_v){
  int32 v;
  bool status = m_buffer->read(v);
  a_v = int(v);
  return status;
}
bool steve::core::streamer::read(unsigned int& a_v){
  uint32 v;
  bool status = m_buffer->read(v);
  a_v = (unsigned int)v;
  return status;
}
*/

bool steve::core::streamer::read(std::string& a_v) {
  uint32 sz;
  if(!m_buffer->read(sz)) return false;
  a_v.resize(sz);
  //FIXME : optimize :
  typedef std::string::iterator it_t;
  for(it_t it=a_v.begin();it!=a_v.end();++it) {
    if(!m_buffer->read(*it)) return false;
  }
  return true;
}

bool steve::core::streamer::read(uint32& a_size,char*& a_buffer) { 
  // this method allocate a_buffer.
  if(!m_buffer->read(a_size)) return false;
  if(!a_size) {a_buffer = 0;return true;}
  a_buffer = new char[a_size];
 {for(uint32 index = 0;index<a_size;index++) {
    if(!m_buffer->read(a_buffer[index])) return false;
  }}
  return true;
}

