// this :
#include <steve/mysql/buffer.h>

#include <steve/tools/realloc.h>

#include <stdio.h> //sprintf
#include <stdlib.h> //strtol,strtod

#define MAXIMUM(a,b) ((a)>(b)?(a):(b)) 

steve::mysql::buffer::buffer(
 std::ostream& a_out
,uint32 a_buffer_size
,char* a_buffer
)
: core::base_buffer(a_out,a_buffer_size,a_buffer)
{}

steve::mysql::buffer::~buffer(){}

//private:
steve::mysql::buffer::buffer(const buffer& a_buffer)
: core::base_buffer(a_buffer){}
steve::mysql::buffer& steve::mysql::buffer::operator=(const buffer&){return *this;}

#define MAX_SPRINTF_SIZE 256

//////////////////////////////////////////////////////////////////////////////
/// writers //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#define WRITE(a_type,a_format)\
bool steve::mysql::buffer::write(a_type a_v) {\
  if((m_pos+MAX_SPRINTF_SIZE)>m_max) \
    if(!expand(MAXIMUM(2*m_size,m_size+MAX_SPRINTF_SIZE))) return false;\
  int sz = ::sprintf(m_pos,a_format,a_v);\
  if(sz<0) return false;\
  if(sz>MAX_SPRINTF_SIZE) return false;\
  m_pos += sz;\
  return true;\
}

WRITE(uchar,"%u")
WRITE(char,"%d")
WRITE(uint16,"%u")
WRITE(int16,"%d")
WRITE(uint32,"%u")
WRITE(int32,"%d")
WRITE(uint64,STEVE_UINT64_FORMAT)
WRITE(int64,STEVE_INT64_FORMAT)
WRITE(float,"%g")
WRITE(double,"%g")

#undef WRITE

bool steve::mysql::buffer::write(const std::string& a_v) {
  uint32 sz = a_v.size();
  if((m_pos+sz)>m_max)
    if(!expand(MAXIMUM(2*m_size,m_size+sz))) return false;
  ::memcpy(m_pos,a_v.c_str(),sz);
  m_pos += sz;
  return true;
}

bool steve::mysql::buffer::write_null() {
  uint32 sz = 1;
  if((m_pos+sz)>m_max)
    if(!expand(MAXIMUM(2*m_size,m_size+sz))) return false;
  *m_pos = '\0';
  m_pos += sz;
  return true;
}

//WARNING : dummies :
bool steve::mysql::buffer::write(bool a_v) {
  m_out << "steve::mysql::buffer::write(bool) dummy." << std::endl;
  return false;
}

#define WRITE_ARRAY(a_type)\
bool steve::mysql::buffer::write_array(uint32 a_n,const a_type* a_a) {\
  m_out << "steve::mysql::buffer::write_array() dummy." << std::endl;\
  return false;\
}

WRITE_ARRAY(uchar)
WRITE_ARRAY(char)
WRITE_ARRAY(uint16)
WRITE_ARRAY(int16)
WRITE_ARRAY(uint32)
WRITE_ARRAY(int32)
WRITE_ARRAY(uint64)
WRITE_ARRAY(int64)
WRITE_ARRAY(float)
WRITE_ARRAY(double)
WRITE_ARRAY(bool)

#undef WRITE_ARRAY

//////////////////////////////////////////////////////////////////////////////
/// readers //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#define READ(a_type)\
bool steve::mysql::buffer::read(a_type&) {\
  m_out << "steve::mysql::buffer::read() dummy." << std::endl;\
  return false;\
}


READ(uchar)
READ(char)
READ(uint16)
READ(int16)
READ(uint32)
READ(int32)
READ(uint64)
READ(int64)
READ(float)
READ(double)
READ(bool)

READ(std::string)

#define READ_ARRAY(a_type)\
bool steve::mysql::buffer::read_array(uint32&,a_type*&) {\
  m_out << "steve::mysql::buffer::read_array() dummy." << std::endl;\
  return false;\
}

READ_ARRAY(uchar)
READ_ARRAY(char)
READ_ARRAY(uint16)
READ_ARRAY(int16)
READ_ARRAY(uint32)
READ_ARRAY(int32)
READ_ARRAY(uint64)
READ_ARRAY(int64)
READ_ARRAY(float)
READ_ARRAY(double)
READ_ARRAY(bool)
