// this :
#include <steve/mysql/store.h>

#include <steve/interface/store_data.h>

#include <steve/version.h>

#include <steve/tools/sql.h>
#include <steve/tools/smanip.h>

#include <steve/core/store_pages.h>
#include <steve/core/retrieve_pages.h>

#include <steve/mysql/page_xxx.h>
#include <steve/mysql/helper.h>

steve::mysql::store::store(
 std::ostream& a_out
,const std::string& a_db
,const std::string& a_options
)
:base_store(a_out,a_db,a_options)
,m_mysql(0)
,m_connection(0)
{
}

steve::mysql::store::~store(){
}

//private:
steve::mysql::store::store(const store& a_from)
:base_store(a_from)
,m_mysql(0)
,m_connection(0)
{}
//private:
steve::mysql::store& steve::mysql::store::operator=(const store&) { return *this;}

void* steve::mysql::store::cast_head() const {
  return (void*)static_cast<const steve::mysql::store*>(this);
}

bool steve::mysql::store::open(
 e_mode a_mode
)
{
  if(m_opened) return false;

  std::vector<std::string> words = tools::words(m_store_name," ");
  if(words.size()<3) {
    m_out << "steve::mysql::store::open :"
          << " for database name, at least threee fields expected " 
          << tools::sout("host database user [passowrd]")
          << " (Given " << tools::sout(m_store_name) << ")."
          << std::endl;
    return false;
  }

  m_mysql = mysql_init(NULL);
  if(m_mysql==NULL) {
    m_out << "steve::mysql::store::open :"
          << " unable to initialize mysql."
          << std::endl;
    return false;
  }

  std::string db = words[1];

  switch(a_mode) {
  case create:{

    //FIXME : DROP, CREATE DATABASE m_store_name here ?

    std::string host = words[0];
    std::string user = words[2];
    std::string password = (words.size()==4 ? words[3] : "");
    
    unsigned int port = 3306;
    
    m_connection = mysql_real_connect(m_mysql,
                                      host.c_str(),
                                      user.c_str(),
                                      password.c_str(),
                                      db.c_str(),
                                      port,0,0);
    if(!m_connection) {
      m_out << "steve::mysql::store::open :"
            << " " << std::string(mysql_error(m_mysql)) 
            << std::endl;
      mysql_close(m_mysql);
      m_mysql = 0;
      return false;
    }

    if(m_verbosity) {
      m_out << "steve::mysql::store::open :"
            << " connected to host " << tools::sout(host)
            << " base " << tools::sout(db)
            << " user " << tools::sout(user)
            << std::endl;
    }

    if(!drop_tables()) {
      mysql_close(m_mysql);
      m_mysql = 0;
      return false;
    }

    // to be not sync table should be MyISAM 
    // (and not InnoDB which is full sync).

   {std::string header = "header";
    if(!do_query(m_out,m_connection,"DROP TABLE "+header,false)) {
      mysql_close(m_mysql);
      m_mysql = 0;
      return false;
    }
   {std::string query;
    query  = "CREATE TABLE "+header+" (";
    query += "number_of_events INT UNSIGNED";
/*
    query += ",";
    query += "STEVE_MAJOR_VERSION INT UNSIGNED";
    query += ",";
    query += "STEVE_MINOR_VERSION INT UNSIGNED";
    query += ",";
    query += "STEVE_PATCH_VERSION INT UNSIGNED";
*/
    query += ")";
    if(!do_query(m_out,m_connection,query)) {
      mysql_close(m_mysql);
      m_mysql = 0;
      return false;
    }}}

/*
   {if(!do_query(m_out,m_connection,"DROP TABLE versions",false)) {
      mysql_close(m_mysql);
      m_mysql = 0;
      return false;
    }
   {std::string query;
    query  = "CREATE TABLE versions (";
    query += "STEVE_MAJOR_VERSION INT UNSIGNED";
    query += ",";
    query += "STEVE_MINOR_VERSION INT UNSIGNED";
    query += ",";
    query += "STEVE_PATCH_VERSION INT UNSIGNED";
    query += ")";
    if(!do_query(m_out,m_connection,query)) {
      mysql_close(m_mysql);
      m_mysql = 0;
      return false;
    }}
    if(!store_uint32(m_out,m_connection,
                     "versions","STEVE_MAJOR_VERSION",STEVE_MAJOR_VERSION)) {
      mysql_close(m_mysql);
      m_mysql = 0;
      return false;
    }
    if(!store_uint32(m_out,m_connection,
                     "versions","STEVE_MINOR_VERSION",STEVE_MINOR_VERSION)) {
      mysql_close(m_mysql);
      m_mysql = 0;
      return false;
    }
    if(!store_uint32(m_out,m_connection,
                     "versions","STEVE_PATCH_VERSION",STEVE_PATCH_VERSION)) {
      mysql_close(m_mysql);
      m_mysql = 0;
      return false;
    }}
*/

    //FIXME : reset numbers.m_total, clear pages.

    m_opened = true;
    }return true;
  case readonly:{

    std::string host = words[0];
    std::string db = words[1];
    std::string user = words[2];
    std::string password = (words.size()==4 ? words[3] : "");
    
    unsigned int port = 3306;
    
    m_connection = mysql_real_connect(m_mysql,
                                      host.c_str(),
                                      user.c_str(),
                                      password.c_str(),
                                      db.c_str(),
                                      port,0,0);
    if(!m_connection) {
      m_out << "steve::mysql::store::open :"
            << " " << std::string(mysql_error(m_mysql)) 
            << std::endl;
      mysql_close(m_mysql);
      m_mysql = 0;
      return false;
    }

    if(m_verbosity) {
      m_out << "steve::mysql::store::open :"
            << " connected to host " << tools::sout(host)
            << " base " << tools::sout(db)
            << " user " << tools::sout(user)
            << std::endl;
    }

    // to be not sync table should be MyISAM 
    // (and not InnoDB which is full sync).

    //if(m_verbosity) ls();

/*
   {uint32 version;
    if(!retrieve_uint32(m_out,m_connection,
                       "header","STEVE_MAJOR_VERSION",version)) {
      printf("debug : major error\n");
      mysql_close(m_mysql);
      m_mysql = 0;
      return false;
    }
    printf("debug : major %d\n",version);
    }
   {uint32 version;
    if(!retrieve_uint32(m_out,m_connection,
                       "header","STEVE_MINOR_VERSION",version)) {
      mysql_close(m_mysql);
      m_mysql = 0;
      return false;
    }
    printf("debug : minor %d\n",version);
    }
   {uint32 version;
    if(!retrieve_uint32(m_out,m_connection,
                       "header","STEVE_PATCH_VERSION",version)) {
      mysql_close(m_mysql);
      m_mysql = 0;
      return false;
    }
    printf("debug : patch %d\n",version);
    }
*/

    m_opened = true;
    }return true;
  default:
   {m_out << "steve::mysql::store::open :"
          << " bad mode."
          << std::endl;
    }break;
  }

  return false;
}

bool steve::mysql::store::close()
{
  if(!m_opened) return false;

  bool status = true;

  mysql_close(m_mysql);
  m_mysql = 0;
  m_connection = 0;

  m_opened = false;
  return true;
}

steve::interface::page* steve::mysql::store::create_page(
 interface::store_data& a_store_data
,uint32 a_page_size
,uint32 //a_compression
) {
  if(!m_connection) return 0;

  std::string table = a_store_data.name();
  tools::replace(table,".","_"); //SQL does not want a . as separato.

  if(!a_store_data.store_pages()) {

    std::string header = table+"_header";
    if(!do_query(m_out,m_connection,"DROP TABLE "+header,false)) {
      return 0;
    }

   {std::string query;
    query = "CREATE TABLE "+header+" (page_size INT)";
    if(!do_query(m_out,m_connection,query)) {
      return 0;
    }}

    if(!store_uint32(m_out,m_connection,
                     header,"page_size",a_page_size)) {
      return 0;
    }

    a_store_data.set_store_pages(new core::store_pages());
  }

  core::store_pages* pgs = 
    (core::store_pages*)a_store_data.store_pages()->cast_head();
  uint32 page_index = pgs->count();
  std::string page_name = table+"_"+tools::to<uint32>(page_index);
  pgs->increment_count();

 {std::string type;
  if(!steve::tools::to_sql(a_store_data.type(),type)) {
    m_out << "steve::mysql::store::create_page :"
          << " unknown type " << a_store_data.type()
          << std::endl;
    return 0;
  }
  std::string query;
  query = "CREATE TABLE "+page_name+" (value "+type+")";
  if(!do_query(m_out,m_connection,query)) return 0;}

  steve::interface::page* pg = 0;

#define if_new_save_page(a_class,a_id_type) \
  if(a_store_data.type()==a_id_type) {\
    pg = new a_class(m_out,m_connection,a_page_size,page_index,page_name);\
  }

  if_new_save_page(page_uchar,steve_id_uchar)
  else if_new_save_page(page_char,steve_id_char)
  else if_new_save_page(page_uint16,steve_id_uint16)
  else if_new_save_page(page_int16,steve_id_int16)
  else if_new_save_page(page_uint32,steve_id_uint32)
  else if_new_save_page(page_int32,steve_id_int32)
  else if_new_save_page(page_uint64,steve_id_uint64)
  else if_new_save_page(page_int64,steve_id_int64)
  else if_new_save_page(page_float,steve_id_float)
  else if_new_save_page(page_double,steve_id_double)

  else if_new_save_page(page_uint32,steve_id_ref)

  else if_new_save_page(page_bool,steve_id_bool)
  else if_new_save_page(page_std_string,steve_id_string)

  else {
    return 0;
  }

#undef if_new_save_page

  pgs->push_back(pg);

  return pg;

}

bool steve::mysql::store::commit_page(
 interface::store_data& a_store_data
,interface::page*& a_page
) {
  //printf("debug : steve::mysql::store::commit_page\n");
  if(!a_page->commit()) {
    m_out << "steve::store::commit_page :"
          << " problem saving page."
          << std::endl;
    return false;
  }

  if(!a_store_data.store_pages()) return false;

  core::store_pages* pgs = 
    (core::store_pages*)a_store_data.store_pages()->cast_head();

  if(!pgs) return false;
  pgs->remove(a_page);

  delete a_page;
  a_page = 0; //invalidate the pointer.
  return true;
}

steve::interface::page* steve::mysql::store::find_page(
 interface::store_data& a_store_data
,uint32 a_index
,uint32& a_index_in_page
)
{ 
  if(!m_connection) return 0;

  // Look if page is already loaded :
  if(!a_store_data.retrieve_pages()){

    //uint32 page_size = 10000;//FIXME

    std::string table = a_store_data.name();
    tools::replace(table,".","_");
    std::string header = table+"_header";

    uint32 page_size;
    if(!retrieve_uint32(m_out,m_connection,header,"page_size",page_size)) {
      return false;
    }

    a_store_data.set_retrieve_pages(new core::retrieve_pages(page_size));
  }

  core::retrieve_pages* pgs = 
    (core::retrieve_pages*)a_store_data.retrieve_pages()->cast_head();

  uint32 page_size = pgs->page_size();

  uint32 page_index = a_index/page_size;
  a_index_in_page = a_index%page_size;

  //div_t dv = div(a_index,page_size); //optimization.
  //uint32 page_index = (uint32 )dv.quot;
  //a_index_in_page = (uint32)dv.rem;

  if(page_index==pgs->first) return pgs->second;

  // Erase old page :
  delete pgs->second;
  pgs->first = steve_not_found;
  pgs->second = 0;

  std::string table = a_store_data.name();
  tools::replace(table,".","_");

  std::string page_name = table+"_"+tools::to<uint32>(page_index);

  steve::interface::page* pg = 0;

#define if_new_retrieve_page(a_class,a_id_type) \
  if(a_store_data.type()==a_id_type) {\
    pg = new a_class(m_out,m_connection,page_size,page_index,page_name);\
  }

  if_new_retrieve_page(page_uchar,steve_id_uchar)
  else if_new_retrieve_page(page_char,steve_id_char)
  else if_new_retrieve_page(page_uint16,steve_id_uint16)
  else if_new_retrieve_page(page_int16,steve_id_int16)
  else if_new_retrieve_page(page_uint32,steve_id_uint32)
  else if_new_retrieve_page(page_int32,steve_id_int32)
  else if_new_retrieve_page(page_uint64,steve_id_uint64)
  else if_new_retrieve_page(page_int64,steve_id_int64)
  else if_new_retrieve_page(page_float,steve_id_float)
  else if_new_retrieve_page(page_double,steve_id_double)

  else if_new_retrieve_page(page_uint32,steve_id_ref)

  else if_new_retrieve_page(page_bool,steve_id_bool)
  else if_new_retrieve_page(page_std_string,steve_id_string)

  else {
    return 0;
  }

#undef if_new_retrieve_page

  if(!pg->retrieve()) {
    m_out << "steve::mysql::store::find_page :"
          << " problem retrieving page."
          << std::endl;
    return 0;
  }

  pgs->first = page_index;
  pgs->second = pg;

  return pg;
}
/*
bool steve::mysql::store::store_number_of_events(
 uint32 a_number
)
{
  if(!m_connection) return false;
  if(!do_query(m_out,m_connection,"DELETE from header",false)) 
    return false;
  return store_uint32(m_out,m_connection,
                      "header","number_of_events",a_number);
}
bool steve::mysql::store::retrieve_number_of_events(
 uint32& a_number
)
{
  if(!m_connection) return false;
  return retrieve_uint32(m_out,m_connection,
                         "header","number_of_events",a_number);
}
*/
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void steve::mysql::store::dump_result(MYSQL_RES* a_result) const {

  unsigned int num_fields = mysql_num_fields(a_result);
  my_ulonglong num_rows = mysql_num_rows(a_result);

  m_out << "number of fields : " << num_fields << std::endl;

  m_out << "number of rows : " << (unsigned long)num_rows << std::endl;

  MYSQL_ROW row;
  while( ( row = mysql_fetch_row(a_result) ) != NULL )  {
    for(uint32 i=0;i<num_fields;i++ ) {
      m_out << " " << std::string(row[i]?row[i]:"NULL");
    }
    m_out << std::endl;
  }
}

bool steve::mysql::store::ls() const {
  if(!m_connection) return false;

  // Get the list of the db tables
  MYSQL_RES* result = mysql_list_tables(m_connection,NULL);
  if(!result) {
    m_out << "steve::mysql::store::ls :"
          << " " << std::string(mysql_error(m_mysql)) 
          << std::endl;
    return false;
  }

  m_out << "tables : " << std::endl;

  dump_result(result);
  
  mysql_free_result(result);

  return true;
}
bool steve::mysql::store::drop_tables() const {
  if(!m_connection) return false;

  std::vector<std::string> tables;

 {MYSQL_RES* result = mysql_list_tables(m_connection,NULL);
  if(!result) {
    m_out << "steve::mysql::store::drop_tables :"
          << " " << std::string(mysql_error(m_mysql)) 
          << std::endl;
    return false;
  }
  MYSQL_ROW row;
  while( ( row = mysql_fetch_row(result) ) != NULL )  {    
    if(row[0]!=NULL) tables.push_back(std::string(row[0]));
  }
  mysql_free_result(result);}

 {uint32 number = tables.size();
  for(uint32 index=0;index<number;index++) {
    if(!do_query(m_out,m_connection,
                 "DROP TABLE "+tables[index],false)) return false;
  }}

  return true;
}
