// this :
#include <steve/sqlite/store.h>

#include <steve/interface/store_data.h>

#include <steve/version.h>

#include <steve/tools/smanip.h>

#include <steve/tools/sql.h>

#include <steve/core/store_pages.h>
#include <steve/core/retrieve_pages.h>

#include <steve/sqlite/page_xxx.h>
#include <steve/sqlite/helper.h>

steve::sqlite::store::store(
 std::ostream& a_out
,const std::string& a_db
,const std::string& a_options
)
:base_store(a_out,a_db,a_options)
,m_sqlite(0)
{
}

steve::sqlite::store::~store(){
}

//private:
steve::sqlite::store::store(const store& a_from)
:base_store(a_from)
,m_sqlite(0)
{}
//private:
steve::sqlite::store& steve::sqlite::store::operator=(const store&) { return *this;}

void* steve::sqlite::store::cast_head() const {
  return (void*)static_cast<const steve::sqlite::store*>(this);
}

bool steve::sqlite::store::open(
 e_mode a_mode
)
{
  if(m_opened) return false;

  switch(a_mode) {
  case create:{

    if(sqlite3_open(m_store_name.c_str(),&m_sqlite)!=SQLITE_OK) {
      m_out << "steve::sqlite::store::open :"
            << " " << std::string(sqlite3_errmsg(m_sqlite)) 
            << std::endl;
      m_sqlite = 0;
      return false;
    }

    if(!do_query(m_out,m_sqlite,"PRAGMA synchronous = OFF")) {
      sqlite3_close(m_sqlite);
      m_sqlite = 0;
      return false;
    }

    if(!drop_tables(m_out,m_sqlite)) {
      sqlite3_close(m_sqlite);
      m_sqlite = 0;
      return false;
    }

   {std::string header = "header";
    if(!do_query(m_out,m_sqlite,"DROP TABLE "+header,false)) {
      sqlite3_close(m_sqlite);
      m_sqlite = 0;
      return false;
    }
   {std::string query;
    query  = "CREATE TABLE "+header+" (";
    query += "number_of_events INT UNSIGNED";
    query += ")";
    if(!do_query(m_out,m_sqlite,query)) {
      sqlite3_close(m_sqlite);
      m_sqlite = 0;
      return false;
    }}}

    //FIXME if(m_verbosity) ls();

    //FIXME : reset numbers.m_total, clear pages.

    m_opened = true;
    }return true;
  case readonly:{

    if(sqlite3_open(m_store_name.c_str(),&m_sqlite)!=SQLITE_OK) {
      m_out << "steve::sqlite::store::open :"
            << " " << std::string(sqlite3_errmsg(m_sqlite)) 
            << std::endl;
      m_sqlite = 0;
      return false;
    }

    if(!do_query(m_out,m_sqlite,"PRAGMA synchronous = OFF")) {
      sqlite3_close(m_sqlite);
      m_sqlite = 0;
      return false;
    }

    //if(m_verbosity) ls();

    m_opened = true;
    }return true;
  default:
   {m_out << "steve::sqlite::store::open :"
          << " bad mode."
          << std::endl;
    }break;
  }

  return false;
}

bool steve::sqlite::store::close()
{
  if(!m_opened) return false;

  bool status = true;

  if(sqlite3_close(m_sqlite)!=SQLITE_OK) {
    m_out << "steve::sqlite::store::open :"
          << " " << std::string(sqlite3_errmsg(m_sqlite)) 
          << std::endl;
    status = false;
  }

  m_sqlite = 0;

  m_opened = false;
  return true;
}

steve::interface::page* steve::sqlite::store::create_page(
 interface::store_data& a_store_data
,uint32 a_page_size
,uint32 //a_compression
) {
  if(!m_sqlite) return 0;

  std::string table = a_store_data.name();
  tools::replace(table,".","_"); //SQL does not want a . as separato.

  if(!a_store_data.store_pages()) {

    std::string header = table+"_header";
    if(!do_query(m_out,m_sqlite,"DROP TABLE "+header,false)) {
      return 0;
    }

   {std::string query;
    query = "CREATE TABLE "+header+" (page_size INT)";
    if(!do_query(m_out,m_sqlite,query)) {
      return 0;
    }}

    if(!store_uint32(m_out,m_sqlite,
                     header,"page_size",a_page_size)) {
      return 0;
    }

    a_store_data.set_store_pages(new core::store_pages());
  }

  core::store_pages* pgs = 
    (core::store_pages*)a_store_data.store_pages()->cast_head();
  uint32 page_index = pgs->count();
  std::string page_name = table+"_"+tools::to<uint32>(page_index);
  pgs->increment_count();

 {std::string type;
  if(!steve::tools::to_sql(a_store_data.type(),type)) {
    m_out << "steve::sqlite::store::create_page :"
          << " unknown type " << a_store_data.type()
          << std::endl;
    return 0;
  }
  std::string query;
  query = "CREATE TABLE "+page_name+" (value "+type+")";
  if(!do_query(m_out,m_sqlite,query)) return 0;}

  steve::interface::page* pg = 0;

#define if_new_save_page(a_class,a_id_type) \
  if(a_store_data.type()==a_id_type) {\
    pg = new a_class(m_out,m_sqlite,a_page_size,page_index,page_name);\
  }

  if_new_save_page(page_uchar,steve_id_uchar)
  else if_new_save_page(page_char,steve_id_char)
  else if_new_save_page(page_uint16,steve_id_uint16)
  else if_new_save_page(page_int16,steve_id_int16)
  else if_new_save_page(page_uint32,steve_id_uint32)
  else if_new_save_page(page_int32,steve_id_int32)
  else if_new_save_page(page_uint64,steve_id_uint64)
  else if_new_save_page(page_int64,steve_id_int64)
  else if_new_save_page(page_float,steve_id_float)
  else if_new_save_page(page_double,steve_id_double)

  else if_new_save_page(page_uint32,steve_id_ref)

  else if_new_save_page(page_bool,steve_id_bool)
  else if_new_save_page(page_std_string,steve_id_string)

  else {
    return 0;
  }

#undef if_new_save_page

  pgs->push_back(pg);

  return pg;
}

bool steve::sqlite::store::commit_page(
 interface::store_data& a_store_data
,interface::page*& a_page
) {
  //printf("debug : steve::sqlite::store::commit_page\n");
  if(!a_page->commit()) {
    m_out << "steve::store::commit_page :"
          << " problem saving page."
          << std::endl;
    return false;
  }

  if(!a_store_data.store_pages()) return false;

  core::store_pages* pgs = 
    (core::store_pages*)a_store_data.store_pages()->cast_head();

  if(!pgs) return false;
  pgs->remove(a_page);

  delete a_page;
  a_page = 0; //invalidate the pointer.
  return true;
}

steve::interface::page* steve::sqlite::store::find_page(
 interface::store_data& a_store_data
,uint32 a_index
,uint32& a_index_in_page
)
{ 
  if(!m_sqlite) return 0;

  // Look if page is already loaded :
  if(!a_store_data.retrieve_pages()){

    //uint32 page_size = 10000;//FIXME

    std::string table = a_store_data.name();
    tools::replace(table,".","_");
    std::string header = table+"_header";

    uint32 page_size;
    if(!retrieve_uint32(m_out,m_sqlite,header,"page_size",page_size)) {
      return false;
    }

    a_store_data.set_retrieve_pages(new core::retrieve_pages(page_size));
  }

  core::retrieve_pages* pgs = 
    (core::retrieve_pages*)a_store_data.retrieve_pages()->cast_head();

  uint32 page_size = pgs->page_size();

  uint32 page_index = a_index/page_size;
  a_index_in_page = a_index%page_size;

  //div_t dv = div(a_index,page_size); //optimization.
  //uint32 page_index = (uint32 )dv.quot;
  //a_index_in_page = (uint32)dv.rem;

  if(page_index==pgs->first) return pgs->second;

  // Erase old page :
  delete pgs->second;
  pgs->first = steve_not_found;
  pgs->second = 0;

  std::string table = a_store_data.name();
  tools::replace(table,".","_");

  std::string page_name = table+"_"+tools::to<uint32>(page_index);

  steve::interface::page* pg = 0;

#define if_new_retrieve_page(a_class,a_id_type) \
  if(a_store_data.type()==a_id_type) {\
    pg = new a_class(m_out,m_sqlite,page_size,page_index,page_name);\
  }

  if_new_retrieve_page(page_uchar,steve_id_uchar)
  else if_new_retrieve_page(page_char,steve_id_char)
  else if_new_retrieve_page(page_uint16,steve_id_uint16)
  else if_new_retrieve_page(page_int16,steve_id_int16)
  else if_new_retrieve_page(page_uint32,steve_id_uint32)
  else if_new_retrieve_page(page_int32,steve_id_int32)
  else if_new_retrieve_page(page_uint64,steve_id_uint64)
  else if_new_retrieve_page(page_int64,steve_id_int64)
  else if_new_retrieve_page(page_float,steve_id_float)
  else if_new_retrieve_page(page_double,steve_id_double)

  else if_new_retrieve_page(page_uint32,steve_id_ref)

  else if_new_retrieve_page(page_bool,steve_id_bool)
  else if_new_retrieve_page(page_std_string,steve_id_string)

  else {
    return 0;
  }

#undef if_new_retrieve_page

  if(!pg->retrieve()) {
    m_out << "steve::sqlite::store::find_page :"
          << " problem retrieving page."
          << std::endl;
    return 0;
  }

  pgs->first = page_index;
  pgs->second = pg;

  return pg;
}
/*
bool steve::sqlite::store::store_number_of_events(
 uint32 a_number
)
{
  if(!m_sqlite) return false;
  if(!do_query(m_out,m_sqlite,"DELETE from header",false)) return false;
  return store_uint32(m_out,m_sqlite,
                      "header","number_of_events",a_number);
}
bool steve::sqlite::store::retrieve_number_of_events(
 uint32& a_number
)
{
  if(!m_sqlite) return false;
  return retrieve_uint32(m_out,m_sqlite,
                         "header","number_of_events",a_number);
}
*/
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool steve::sqlite::store::ls() const {
/*
  if(!m_sqlite) return false;

  // Get the list of the db tables
  MYSQL_RES* result = sqlite_list_tables(m_sqlite,NULL);
  if(!result) {
    m_out << "steve::sqlite::store::ls :"
          << " " << std::string(sqlite_error(m_sqlite)) 
          << std::endl;
    return false;
  }

  m_out << "tables : " << std::endl;

  dump_result(result);
  
  sqlite_free_result(result);

  return true;
*/
  return false;
}
