// this :
#include <steve/util/mem.h>

#include <list>

#include <string.h> //strcmp

static bool s_check_by_class = false;
static int s_count = 0;
static std::list< std::pair<std::string,int> >* s_list = 0;

void steve::util::mem::set_check_by_class(bool aEnable){
  s_check_by_class = aEnable;
}
bool steve::util::mem::check_by_class(){return s_check_by_class;}
void steve::util::mem::increment(){s_count++;}
void steve::util::mem::decrement(){s_count--;}
void steve::util::mem::increment(const char* a_class){
  s_count++;
  if(s_check_by_class) {
    if(!s_list) s_list = new std::list< std::pair<std::string,int> >;
    std::list< std::pair<std::string,int> >::iterator it;
    for(it=s_list->begin();it!=s_list->end();++it) {
      if(!::strcmp((*it).first.c_str(),a_class)) {
        (*it).second++;
        return;
      }
    }
    s_list->push_back(std::pair<std::string,int>(std::string(a_class),1));
  }
}
void steve::util::mem::increment(const std::string& a_class){
  s_count++;
  if(s_check_by_class) {
    if(!s_list) s_list = new std::list< std::pair<std::string,int> >;
    std::list< std::pair<std::string,int> >::iterator it;
    for(it=s_list->begin();it!=s_list->end();++it) {
      if((*it).first==a_class) {
        (*it).second++;
        return;
      }
    }
    s_list->push_back(std::pair<std::string,int>(a_class,1));
  }
}
void steve::util::mem::decrement(const char* a_class){
  s_count--;
  if(s_check_by_class) {
    if(!s_list) s_list = new std::list< std::pair<std::string,int> >;
    std::list< std::pair<std::string,int> >::iterator it;
    for(it=s_list->begin();it!=s_list->end();++it) {
      if(!::strcmp((*it).first.c_str(),a_class)) {
        (*it).second--;
        return;
      }
    }
    s_list->push_back(std::pair<std::string,int>(std::string(a_class),-1));
  }
}
void steve::util::mem::decrement(const std::string& a_class){
  s_count--;
  if(s_check_by_class) {
    if(!s_list) s_list = new std::list< std::pair<std::string,int> >;
    std::list< std::pair<std::string,int> >::iterator it;
    for(it=s_list->begin();it!=s_list->end();++it) {
      if((*it).first==a_class) {
        (*it).second--;
        return;
      }
    }
    s_list->push_back(std::pair<std::string,int>(a_class,-1));
  }
}
steve::util::mem::mem(){increment();}
steve::util::mem::~mem(){decrement();}

//private:
steve::util::mem::mem(const mem&){}
//private:
steve::util::mem& steve::util::mem::operator=(const mem&) { return *this;}

void steve::util::mem::reset(){
  s_count = 0;
  if(s_list) {
    delete s_list;
    s_list = 0;
  }
}
void steve::util::mem::balance(std::ostream& a_out){
  if(s_count) {
    a_out << "steve::util::mem::balance :"
          << " bad global object balance : " << s_count 
          << std::endl;
    if(s_check_by_class) {
      a_out << "steve::util::mem::balance :"
            << " check by class was enabled." 
            << std::endl;
    } else {
      a_out << "steve::util::mem::balance :"
            << " check by class was disabled." 
            << std::endl;
    }
  }
  if(s_list) {
    std::list< std::pair<std::string,int> >::iterator it;
    for(it=s_list->begin();it!=s_list->end();++it) {
      if((*it).second) {
        a_out << "steve::util::mem::balance :"
              << " for class " << (*it).first 
              << ", bad object balance : " << (*it).second
              << std::endl;
      }
    }
    delete s_list;
    s_list = 0;
  }
}
