// this :
#include <steve/zip/reader.h>

#include <steve/tools/smanip.h>
#include <steve/util/mem.h>

#include <zip.h>
#include <unzip.h>

#define WRITEBUFFERSIZE (16384)

#define SIZE_NAME 2048

steve::zip::reader::reader(std::ostream& a_out)
:m_out(a_out)
,m_file(0)
{
  steve::util::mem::increment("steve::zip::reader");
}
steve::zip::reader::~reader() {
  m_map.clear();
  steve::util::mem::decrement("steve::zip::reader");
}

//private:
steve::zip::reader::reader(const reader& a_from)
:m_out(a_from.m_out)
,m_file(0)
{}
//private:
steve::zip::reader& steve::zip::reader::operator=(const reader&) { return *this;}

bool steve::zip::reader::open(const std::string& a_file) {
  m_map.clear();
  if(m_file) return false; //done.

  //m_out << "steve::zip::reader::open :"
  //    << " load " << a_file << std::endl;

  m_file = unzOpen(a_file.c_str());
  if(!m_file) {
    m_out << "steve::zip::reader::open :"
          << " error opening " << a_file << std::endl;
    return false;
  }

/*
  unz_global_info gi;
  int err = unzGetGlobalInfo(m_file,&gi);
  if(err!=UNZ_OK) {
    m_out << "steve::zip::reader::open :"
          << " error in file " << a_file << std::endl;
    unzClose(m_file);
    m_file = 0;
    return false;
  }*/

  m_file_name = a_file;

  //m_out << "steve::zip::reader::open :"
  //      << " number of entries " << int(gi.number_entry) 
  //      << std::endl;

  build_map();

  return true;
}
bool steve::zip::reader::close() {
  if(!m_file) return true; //done.

 {int err = unzClose(m_file);
  if(err!=UNZ_OK) {
    m_out << "steve::zip::reader::close :"
          << " error in closing file " << m_file_name << std::endl;
    return false;
  }}

  m_file_name = "";
  m_file = 0;
  m_map.clear();

  return true;
}
bool steve::zip::reader::read_entry(
 const std::string& a_name
,unsigned int& a_size
,char*& a_buffer
){
  if(!m_file) {
    m_out << "steve::zip::reader::read_entry :"
          << " file not opened."
          << std::endl;
    return false;
  }

  //m_out << "steve::zip::reader::read_entry :"
  //      << " load " << a_name << std::endl;

/*
  //NOTE : the below is to slow if searching at end of file.
  //       It slow down the per event time when looping on events.
 {int err = unzLocateFile(m_file,a_name.c_str(),1);
  if(err==UNZ_END_OF_LIST_OF_FILE) {
    m_out << "steve::zip::reader::read_entry :"
          << " entry " << tools::sout(a_name) << " not found."
          << std::endl;
    unzClose(m_file);
    m_file = 0;
    return false;
  }
  if(err!=UNZ_OK) {
    m_out << "steve::zip::reader::read_entry :"
          << " error in file " << m_file_name << std::endl;
    unzClose(m_file);
    m_file = 0;
    return false;
  }}
*/
  //NOTE : then we pass through a map.
  //       With that the time per event stay constant.
 {std::map<std::string,file_pos_t>::iterator  mit(m_map.find(a_name));
  if(mit==m_map.end()) {
    m_out << "steve::zip::reader::read_entry :"
          << " entry " << tools::sout(a_name) << " not found."
          << std::endl;
    unzClose(m_file);
    m_file = 0;
    return false;
  } else {
    unz_file_pos file_pos;
    file_pos.pos_in_zip_directory = mit->second.first;
    file_pos.num_of_file = mit->second.second;
   {int err = unzGoToFilePos(m_file,&file_pos);
    if(err!=UNZ_OK) {
      m_out << "steve::zip::reader::read_entry :"
            << " can't set file pos for entry " << tools::sout(a_name) << "."
            << std::endl;
      unzClose(m_file);
      m_file = 0;
      return false;
    }}
  }}  

 {int err = unzOpenCurrentFilePassword(m_file,NULL);
  if(err!=UNZ_OK) {
    m_out << "steve::zip::reader::read_entry :"
          << " error in file " << m_file_name << std::endl;
    unzClose(m_file);
    m_file = 0;
    return false;
  }}
  
  unz_file_info file_info;
 {int err = unzGetCurrentFileInfo(m_file,&file_info,NULL,0,NULL,0,NULL,0);
  if(err!=UNZ_OK) {
    m_out << "steve::zip::reader::read_entry :"
          << " error in file " << m_file_name << std::endl;
    unzClose(m_file);
    m_file = 0;
    return false;
  }}

  uLong size_buf = file_info.uncompressed_size;
  if(size_buf) {

    a_buffer = new char[size_buf];
    if(!a_buffer) {
      m_out << "steve::zip::reader::read_entry :"
            << " can't alloc buffer." 
            << std::endl;
      unzClose(m_file);
      m_file = 0;
      return false;
    }

    int nread = unzReadCurrentFile(m_file,a_buffer,size_buf);
    if(nread<0) {
      m_out << "steve::zip::reader::read_entry :"
            << " error in file " << m_file_name << std::endl;
      delete [] a_buffer;
      unzClose(m_file);
      m_file = 0;
      return false;
    }
    a_size = nread;
  
    //if(nread>0) {
    //  for(int i=0;i<nread;i++) printf("%c",buf[i]);
    //  printf("\n");
    //}
  
  } else {
    a_size = 0;
    a_buffer = 0;
    //it is ok.
  }

 {int err = unzCloseCurrentFile(m_file);
  if(err!=UNZ_OK) {
    m_out << "steve::zip::reader::read_entry :"
        << " error in file " << m_file_name << std::endl;
    unzClose(m_file);
    m_file = 0;
    return false;
  }}

  return true;
}

bool steve::zip::reader::read_entry(
 const std::string& a_name
,std::string& a_value
){
  unsigned int sz = 0;
  char* buffer;
  if(!read_entry(a_name,sz,buffer)) return false;
  if(!sz) {
    a_value = "";
    return true;
  }
  if(buffer[sz-1]!='\0') {
    m_out << "steve::zip::reader::read_entry :"
          << " buffer not a C string."
          << std::endl;
    delete [] buffer;
    a_value = "";
    return false;
  }
  a_value = std::string(buffer);
  delete [] buffer;
  return true;
}
bool steve::zip::reader::enabled() const {
  return m_file ? true : false;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool steve::zip::reader::start(uint32& a_number) {
  if(!m_file) {
    m_out << "steve::zip::reader::start :"
          << " file not opened."
          << std::endl;
    return false;
  }

 {int err = unzGoToFirstFile(m_file);
  if(err!=UNZ_OK) {
    m_out << "steve::zip::reader::start :"
          << " error in file " << m_file_name << std::endl;
    unzClose(m_file);
    m_file = 0;
    return false;
  }}

  unz_global_info gi;
 {int err = unzGetGlobalInfo(m_file,&gi);
  if(err!=UNZ_OK) {
    m_out << "steve::zip::reader::start :"
          << " error in file " << m_file_name << std::endl;
    unzClose(m_file);
    m_file = 0;
    return false;
  }}

  a_number = (uint32)gi.number_entry;

  //out << "steve::zip::reader::start :"
  //    << " number of entries " << a_number << steve::util::endl;

  return true;
}
bool steve::zip::reader::next(bool& a_end) {
  if(!m_file) {
    m_out << "steve::zip::reader::next :"
          << " file not opened."
          << std::endl;
    return false;
  }

  // go to next :
 {int err = unzGoToNextFile(m_file);
  if(err==UNZ_END_OF_LIST_OF_FILE) {
    a_end = true;
    return true; //it is ok.
  }
  if(err!=UNZ_OK) {
    m_out << "steve::zip::reader::next :"
          << " error in file " << m_file_name << std::endl;
    unzClose(m_file);
    m_file = 0;
    return false;
  }}

  a_end = false;
  return true;
}
bool steve::zip::reader::current_name(std::string& a_name) {
  if(!m_file) {
    m_out << "steve::zip::reader::current_name :"
          << " file not opened."
          << std::endl;
    return false;
  }

 {char file_in_zip[SIZE_NAME];
  int err = unzGetCurrentFileInfo
    (m_file,NULL,file_in_zip,sizeof(file_in_zip),NULL,0,NULL,0);
  if(err!=UNZ_OK) {
    m_out << "steve::zip::reader::current_name :"
          << " error in file " << m_file_name << std::endl;
    unzClose(m_file);
    m_file = 0;
    return false;
  }
  a_name.assign(file_in_zip);}

  //out << a_name << steve::util::endl;

  return true;
}
bool steve::zip::reader::current_read(unsigned int& a_size,char*& a_buffer) {
  if(!m_file) {
    m_out << "steve::zip::reader::current_read :"
          << " file not opened."
          << std::endl;
    return false;
  }

 {int err = unzOpenCurrentFilePassword(m_file,NULL);
  if(err!=UNZ_OK) {
    m_out << "steve::zip::reader::current_read :"
          << " error in file " << m_file_name << std::endl;
    unzClose(m_file);
    m_file = 0;
    return false;
  }}

  unz_file_info file_info;
 {int err = unzGetCurrentFileInfo(m_file,&file_info,NULL,0,NULL,0,NULL,0);
  if(err!=UNZ_OK) {
    m_out << "steve::zip::reader::current_read :"
          << " error in file " << m_file_name << std::endl;
    unzClose(m_file);
    m_file = 0;
    return false;
  }}

  uLong size_buf = file_info.uncompressed_size;
  if(size_buf) {

    a_buffer = new char[size_buf];
    if(!a_buffer) {
      m_out << "steve::zip::reader::current_read :"
            << " can't alloc buffer." 
            << std::endl;
      unzClose(m_file);
      m_file = 0;
      return false;
    }

    int nread = unzReadCurrentFile(m_file,a_buffer,size_buf);
    if(nread<0) {
      m_out << "steve::zip::reader::current_read :"
            << " error in file " << m_file_name << std::endl;
      delete [] a_buffer;
      unzClose(m_file);
      m_file = 0;
      return false;
    }
    a_size = nread;
  
    //if(nread>0) {
    //  for(int i=0;i<nread;i++) printf("%c",buf[i]);
    //  printf("\n");
    //}
  
  } else {
    a_size = 0;
    a_buffer = 0;
    //it is ok.
  }

 {int err = unzCloseCurrentFile(m_file);
  if(err!=UNZ_OK) {
    m_out << "steve::zip::reader::current_read :"
          << " error in file " << m_file_name << std::endl;
    unzClose(m_file);
    m_file = 0;
    return false;
  }}

  return true;
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
bool steve::zip::reader::build_map() {
  if(!m_file) return false;

  uint32 number;
  if(!start(number)) return false;
  if(!number) return true;

  unz_file_pos file_pos;
  std::string name;

  bool end = false;
  while(!end) {
    if(!current_name(name)) return false;

   {int err = unzGetFilePos(m_file,&file_pos);
    if(err!=UNZ_OK) {
      unzClose(m_file);
      m_file = 0;
      return false;
    }}

    m_map[name] = file_pos_t(file_pos.pos_in_zip_directory,
                             file_pos.num_of_file);

/*
    printf("debug : %s %lu %lu\n",name.c_str(),
                                  file_pos.pos_in_zip_directory,
                                  file_pos.num_of_file);
*/

    if(!next(end)) return false;
  }

  return true;
}
