// this :
#include <steve/zip/store.h>

#include <steve/interface/store_data.h>

#include <steve/tools/smanip.h>

#include <steve/core/store_pages.h>
#include <steve/core/retrieve_pages.h>

#include <steve/zip/page_xxx.h>

#include <steve/version.h>

//#include <stdlib.h> //div

#define steve_failure (-1)

steve::zip::store::store(
 std::ostream& a_out
,const std::string& a_file_name
,const std::string& a_options
)
:base_store(a_out,a_file_name,a_options)
,m_writer(a_out)
,m_reader(a_out)
{
}

steve::zip::store::~store(){
}

//private:
steve::zip::store::store(const store& a_from)
:base_store(a_from)
,m_writer(a_from.m_out)
,m_reader(a_from.m_out)
{}
//private:
steve::zip::store& steve::zip::store::operator=(const store&) { return *this;}

void* steve::zip::store::cast_head() const {
  return (void*)static_cast<const steve::zip::store*>(this);
}

bool steve::zip::store::open(
 e_mode a_mode
)
{
  if(m_opened) return false;

  switch(a_mode) {
  case create:{
    if(!m_writer.open(m_store_name)) {
      m_out << "steve::zip::store::open :"
            << " can't create file " << m_store_name
            << std::endl;
      return false;
    }

    //FIXME store VERSION
    //FIXME : reset numbers.m_total, clear pages.

    m_opened = true;
    }return true;
  case readonly:{
    if(!m_reader.open(m_store_name)) {
      m_out << "steve::zip::store::open :"
            << " can't open file " << m_store_name
            << std::endl;
      return false;
    }

    //FIXME retrieve VERSION

    m_opened = true;
    }return true;
  default:{
    m_out << "steve::zip::store::open :"
          << " bad mode."
          << std::endl;
    }break;
  }

  return false;
}

bool steve::zip::store::close()
{
  if(!m_opened) return false;

  bool status = true;
 
  m_writer.disable();
  if(!m_reader.close()) status = false;

  m_opened = false;
  return status;
}

steve::interface::page* steve::zip::store::create_page(
 interface::store_data& a_store_data
,uint32 a_page_size
,uint32 a_compression
) {

  std::string group_name = "steve."+a_store_data.name();

  if(!a_store_data.store_pages()) {

   {if(!m_writer.open_entry(group_name+".page_size",a_compression)) return 0;
    std::string s = tools::to<uint32>(a_page_size);
    if(!m_writer.write(s.size()+1,s.c_str())) return 0;
    if(!m_writer.close_entry()) return 0;}

    a_store_data.set_store_pages(new core::store_pages());
  }

  core::store_pages* pgs = 
    (core::store_pages*)a_store_data.store_pages()->cast_head();

  std::string name = group_name +"."+tools::to<uint32>(pgs->count());
  pgs->increment_count();
  //printf("debug : name \"%s\"\n",name.c_str());
  steve::interface::page* pg = 0;


#define if_new_save_page(a_class,a_id_type) \
  if(a_store_data.type()==a_id_type) {\
    pg = \
      new a_class(m_out,m_writer,m_reader,a_page_size,a_compression,name);\
  }

  if_new_save_page(page_uchar,steve_id_uchar)
  else if_new_save_page(page_char,steve_id_char)
  else if_new_save_page(page_uint16,steve_id_uint16)
  else if_new_save_page(page_int16,steve_id_int16)
  else if_new_save_page(page_uint32,steve_id_uint32)
  else if_new_save_page(page_int32,steve_id_int32)
  else if_new_save_page(page_uint64,steve_id_uint64)
  else if_new_save_page(page_int64,steve_id_int64)
  else if_new_save_page(page_float,steve_id_float)
  else if_new_save_page(page_double,steve_id_double)

  else if_new_save_page(page_uint32,steve_id_ref)

  else if_new_save_page(page_bool,steve_id_bool)
  else if_new_save_page(page_std_string,steve_id_string)

  else {
    return 0;
  }

#undef if_new_save_page

  pgs->push_back(pg);
  return pg;
}

bool steve::zip::store::commit_page(
 interface::store_data& a_store_data
,interface::page*& a_page) {
  if(!a_page->commit()) {
    m_out << "steve::store::commit_page :"
          << " problem saving page."
          << std::endl;
    return false;
  }

  if(!a_store_data.store_pages()) return false;

  core::store_pages* pgs = 
    (core::store_pages*)a_store_data.store_pages()->cast_head();

  if(!pgs) return false;
  pgs->remove(a_page);

  delete a_page;
  a_page = 0; //invalidate the pointer.
  return true;
}

steve::interface::page* steve::zip::store::find_page(
 interface::store_data& a_store_data
,uint32 a_index
,uint32& a_index_in_page
)
{ 
  if(!m_reader.enabled()) return 0;

  // Look if page is already loaded :
  if(!a_store_data.retrieve_pages()){

    uint32 page_size = 0;

    std::string group_name = "steve."+a_store_data.name(); //duplicated below.

   {std::string s;
    if(!m_reader.read_entry(group_name+".page_size",s)) return 0;
    if(!tools::to<uint32>(s,page_size)) return 0;}

    a_store_data.set_retrieve_pages(new core::retrieve_pages(page_size));
  }

  core::retrieve_pages* pgs = 
    (core::retrieve_pages*)a_store_data.retrieve_pages()->cast_head();

  uint32 page_size = pgs->page_size();

  uint32 page_index = a_index/page_size;
  a_index_in_page = a_index%page_size;

  //div_t dv = div(a_index,page_size); //optimization.
  //uint32 page_index = (uint32 )dv.quot;
  //a_index_in_page = (uint32)dv.rem;

  if(page_index==pgs->first) return pgs->second;

  // Erase old page :
  delete pgs->second;
  pgs->first = steve_not_found;
  pgs->second = 0;

  std::string group_name = "steve."+a_store_data.name();
  std::string name = group_name+"."+tools::to<uint32>(page_index);

  steve::interface::page* pg = 0;

#define if_new_retrieve_page(a_class,a_id_type) \
  if(a_store_data.type()==a_id_type) {\
    pg = new a_class(m_out,m_writer,m_reader,page_size,0,name);\
  }

  if_new_retrieve_page(page_uchar,steve_id_uchar)
  else if_new_retrieve_page(page_char,steve_id_char)
  else if_new_retrieve_page(page_uint16,steve_id_uint16)
  else if_new_retrieve_page(page_int16,steve_id_int16)
  else if_new_retrieve_page(page_uint32,steve_id_uint32)
  else if_new_retrieve_page(page_int32,steve_id_int32)
  else if_new_retrieve_page(page_uint64,steve_id_uint64)
  else if_new_retrieve_page(page_int64,steve_id_int64)
  else if_new_retrieve_page(page_float,steve_id_float)
  else if_new_retrieve_page(page_double,steve_id_double)

  else if_new_retrieve_page(page_uint32,steve_id_ref)

  else if_new_retrieve_page(page_bool,steve_id_bool)
  else if_new_retrieve_page(page_std_string,steve_id_string)

  else {
    return 0;
  }

#undef if_new_retrieve_page

  if(!pg->retrieve()) {
    m_out << "steve::zip::store::find_page :"
          << " problem retrieving page."
          << std::endl;
    return 0;
  }

  pgs->first = page_index;
  pgs->second = pg;

  return pg;
}
/*
bool steve::zip::store::store_number_of_events(
 uint32 a_number
)
{
  if(!m_writer.enabled()) return false;

  uint32 compression = 9; //FIXME
  if(!m_writer.open_entry("steve.number_of_events",compression)) return false;
  std::string s = tools::to<uint32>(a_number);
  if(!m_writer.write(s.size()+1,s.c_str())) return false;
  if(!m_writer.close_entry()) return false;

  return true;
}
bool steve::zip::store::retrieve_number_of_events(
 uint32& a_number
)
{
  if(!m_reader.enabled()) return false;
  std::string s;
  if(!m_reader.read_entry("steve.number_of_events",s)) return false;
  if(!tools::to<uint32>(s,a_number)) return false;
  return true;
}
*/
