// this :
#include <steve/zip/writer.h>

#include <steve/util/mem.h>
#include <steve/tools/smanip.h>

#include <zip.h>

steve::zip::writer::writer(std::ostream& a_error_out)
:m_err(a_error_out)
,m_file(0)
,m_entry_opened(false)
{
  steve::util::mem::increment("steve::zip::writer");
}

bool steve::zip::writer::open(const std::string& a_file_name) {
  if(m_file) return false;

  m_fileName = a_file_name;

  //m_file is a zipFile.
  m_file = zipOpen(m_fileName.c_str(),APPEND_STATUS_CREATE);
  if(!m_file) {
    m_err << "steve::zip::writer::open :"
	  << "error opening " << tools::sout(m_fileName)
          << std::endl;
    return false;
  }
  return true;
}
steve::zip::writer::~writer() {
  if(m_file) {
    if(zipClose((zipFile)m_file,NULL)!=ZIP_OK) {
      m_err << "steve::zip::writer::~writer :"
	    << " closing zip file " << tools::sout(m_fileName) << " failed."
            << std::endl;
    }
  }
  steve::util::mem::decrement("steve::zip::writer");
}

//private:
steve::zip::writer::writer(const writer& a_from)
:m_err(a_from.m_err)
,m_file(0)
,m_entry_opened(false)
{}
//private:
steve::zip::writer& steve::zip::writer::operator=(const writer&) { return *this;}

void steve::zip::writer::disable() {
  if(!m_file) return; // Already closed.
  if(zipClose((zipFile)m_file,NULL)!=ZIP_OK) {
    m_err << "steve::zip::writer::disable :"
	  << " closing zip file " << tools::sout(m_fileName) << " failed."
          << std::endl;
  }
  m_file = 0;
  m_entry_opened = false;
}
void steve::zip::writer::enable() {
  if(m_file) return; // Already open.
  m_file = zipOpen(m_fileName.c_str(),APPEND_STATUS_CREATE);
  if(!m_file) {
    m_err << "steve::zip::writer::enable :"
	  << "error opening " << tools::sout(m_fileName)
          << std::endl;
  }
  m_entry_opened = false;
}
bool steve::zip::writer::enabled() const {
  return m_file ? true : false;
}
bool steve::zip::writer::write(
 unsigned int a_size
,const char* a_buffer
) const {
  if(!m_file) return false;
  if(!a_size) return true;

  int size_write = a_size;
  if(zipWriteInFileInZip((zipFile)m_file,a_buffer,size_write)<0) {
    m_err << "steve::zip::writer::close :"
	  << " error writing in file " << tools::sout(m_fileName) 
          << std::endl;
    writer& self = const_cast<writer&>(*this);
    self.disable();
    return false;
  }
  return true;
}
bool steve::zip::writer::open_entry(
 const std::string& a_string
,unsigned int a_compression
){
  if(!m_file) return false;
  if(a_compression>9) return false;

  if(m_entry_opened) return false; //Already opened. Do a close_entry first.

  zip_fileinfo zi;

  zi.tmz_date.tm_sec = 0;
  zi.tmz_date.tm_min = 0;
  zi.tmz_date.tm_hour = 0;
  zi.tmz_date.tm_mday = 0;
  zi.tmz_date.tm_mon = 0;
  zi.tmz_date.tm_year = 0;
  //filetime(filenameinzip,&zi.tmz_date,&zi.dosDate);

  zi.dosDate = 0;
  zi.internal_fa = 0;
  zi.external_fa = 0;

  int err = 
    //zipOpenNewFileInZip3(zf,file_in_zip.c_str(),&zi,
    zipOpenNewFileInZip((zipFile)m_file,a_string.c_str(),&zi,
      NULL,0,NULL,0,NULL /* comment*/,
      Z_DEFLATED,a_compression);
  if(err!=ZIP_OK) {
    m_err << "steve::zip::writer::open :"
	  << " can't open entry " << tools::sout(a_string)
	  << " in file " << tools::sout(m_fileName) 
          << std::endl;
    writer& self = const_cast<writer&>(*this);
    self.disable();
    return false;
  }

  m_entry_opened = true;
  return true;
}
bool steve::zip::writer::close_entry() {
  if(!m_file) return false;
  if(!m_entry_opened) return false;
  if(zipCloseFileInZip((zipFile)m_file)!=ZIP_OK) {
    m_err << "steve::zip::writer::close :"
	  << " can't close entry in file" << tools::sout(m_fileName) 
          << std::endl;
    writer& self = const_cast<writer&>(*this);
    self.disable();
    return false;
  }
  m_entry_opened = false;
  return true;
}
