#ifndef steve_HDF5_page_xxx_h
#define steve_HDF5_page_xxx_h

// inheritance :
#include <steve/HDF5/page.h>

#include <steve/HDF5/T.h>
#include <steve/HDF5/types.h>
#include <steve/HDF5/helper.h>
#include <steve/core/ascii_buffer.h>

#include <steve/tools/smanip.h>

#include <string.h> //memcpy

namespace steve {
namespace HDF5 {

#define class_page_xxx(a_class,a_type,a_file,a_mem)\
class a_class : public HDF5::page< a_type > {\
public: /*interface::page :*/\
  virtual bool add(const a_type& a_value,uint32& a_size) {\
    if(m_index>=m_size) return false;\
    m_data[m_index] = a_value;\
    m_index++;\
    a_size = sizeof(a_type);\
    return true;\
  }\
  virtual bool add(uint32 a_size,const a_type* a_values,\
                   uint32& a_items_size) {\
    if((m_index+a_size)>m_size) return false;\
    a_type* apos = (a_type*)a_values;\
    a_type* mpos = (a_type*)(m_data+m_index);\
    for(uint32 index=0;index<a_size;index++,apos++,mpos++) *mpos = *apos;\
    m_index += a_size;\
    a_items_size = a_size*sizeof(a_type);\
    return true;\
  }\
  virtual bool value(uint32 a_index,a_type& a_value) const {\
    if(a_index>=m_size) return false;\
    a_value = m_data[a_index];\
    return true;\
  }\
  virtual bool values(uint32 a_index,uint32 a_size,a_type* a_values) const {\
    if((a_index+a_size)>m_size) return false;\
    ::memcpy(a_values,m_data+a_index,a_size*sizeof(a_type));\
    return true;\
  }\
  virtual bool commit() const {\
    if(!m_index) {\
      m_out << "steve::HDF5::" << #a_class << "::commit :"\
            << " attempt to store empty page for "\
            << " " << tools::sout(m_name)	  \
            << std::endl;\
      return false;/*do not save empty pages.*/\
    }\
    unsigned int sz = m_size;\
    if(m_index<m_size) { sz = m_index;}\
    if(!write_array< a_type >(m_id,m_name,\
                              a_file,a_mem,m_compression,\
                              sz,m_data)) {\
      m_out << "steve::HDF5::" << #a_class << "::commit :"\
            << " problem when writing array for"\
            << " " << tools::sout(m_name)	\
            << std::endl;\
      return false;\
    }\
    return true;\
  }\
  virtual bool retrieve() {\
    unsigned int sz = m_size;\
    if(!read_array< a_type >(m_id,m_name,a_mem,sz,m_data)) {\
      m_out << "steve::HDF5::" << #a_class << "::commit :"\
            << " problem when reading array for"\
            << " " << tools::sout(m_name)	\
            << std::endl;\
      return false;\
    }\
    if(!sz) return false;\
    return true;\
  }\
public:\
  a_class(std::ostream& a_out,		\
          hid_t a_id,uint32 a_size,uint32 a_compression,\
          const std::string& a_name)\
  : HDF5::page< a_type >(a_id,a_size,a_compression,a_name)\
  ,m_out(a_out)\
  {\
    m_data = new a_type[m_size];\
  }\
  virtual ~a_class(){\
    delete [] m_data;\
  }\
protected:\
  a_class(const a_class& a_from)\
  : HDF5::page< a_type >(a_from)\
  ,m_out(a_from.m_out)\
  {}\
private:\
  a_class& operator=(const a_class&) { return *this;}\
protected:\
  std::ostream& m_out;	\
  a_type* m_data;\
};


class_page_xxx(page_uchar,uchar,H5T_STD_U8XX,H5T_NATIVE_UCHAR)
class_page_xxx(page_char,char,H5T_STD_I8XX,H5T_NATIVE_CHAR)
class_page_xxx(page_uint32,uint32,H5T_STD_U32XX,H5T_NATIVE_UINT)
class_page_xxx(page_int32,int32,H5T_STD_I32XX,H5T_NATIVE_INT)
class_page_xxx(page_uint16,uint16,H5T_STD_U16XX,H5T_NATIVE_USHORT)
class_page_xxx(page_int16,int16,H5T_STD_I16XX,H5T_NATIVE_SHORT)

#ifdef STEVE_INT64_AS_LONG
class_page_xxx(page_uint64,uint64,H5T_STD_U64XX,H5T_NATIVE_ULONG)
#else //32bits platform :
class_page_xxx(page_uint64,uint64,H5T_STD_U64XX,H5T_NATIVE_ULLONG)
#endif

#ifdef STEVE_INT64_AS_LONG
class_page_xxx(page_int64,int64,H5T_STD_I64XX,H5T_NATIVE_LONG)
#else //32bits platform :
class_page_xxx(page_int64,int64,H5T_STD_I64XX,H5T_NATIVE_LLONG)
#endif

class_page_xxx(page_float,float,H5T_IEEE_F32XX,H5T_NATIVE_FLOAT)
class_page_xxx(page_double,double,H5T_IEEE_F64XX,H5T_NATIVE_DOUBLE)

class page_bool : public page_uchar {
public: //interface::page :
  virtual bool add(const bool& a_value,uint32& a_size) {
    return page_uchar::add(a_value?1:0,a_size);
  }
  virtual bool value(uint32 a_index,bool& a_value) const { 
    uchar value;
    if(!page_uchar::value(a_index,value)) return false;
    a_value = value==1?true:false;
    return true;
  }
public:
  page_bool(std::ostream& a_out,
            hid_t a_id,uint32 a_size,uint32 a_compression,
            const std::string& a_name)
  :page_uchar(a_out,a_id,a_size,a_compression,a_name)
  {}  
  virtual ~page_bool(){}
private:
  page_bool(const page_bool& a_from): page_uchar(a_from){}
  page_bool& operator=(const page_bool&) { return *this;}
};

//NOTE : optimizzation : 
//  We avoid the std::vector<std::string> when storing 
//  because it induces an extra copy of the added string.
class page_std_string : public HDF5::page<std::string> {
public: //interface::page :
  virtual bool add(const std::string& a_value,uint32& a_size) {
    if(m_index>=m_size) return false;    
    if(!m_buffer->write(a_value)) return false;
    m_index++;
    a_size = (uint32)a_value.size();
    return true;
  }
  virtual bool add(uint32 a_size,const std::string* a_values,
                   uint32& a_items_size) {
    std::string* pos = (std::string*)a_values;
    for(uint32 index=0;index<a_size;index++,pos++) {
      if(!add(*pos,a_items_size)) return false;
    }
    return true;
  }
  virtual bool value(uint32 a_index,std::string& a_value) const { 
    if(a_index>=m_data.size()) return false;    
    a_value = m_data[a_index];
    return true;
  }
  virtual bool values(uint32 a_index,
                      uint32 a_size,std::string* a_values) const {
    std::string* pos = (std::string*)a_values;
    for(uint32 index=0;index<a_size;index++,pos++) {
      if(!value(a_index+index,*pos)) return false;
    }
    return true;
  }
  virtual bool commit() const {
    page_std_string& self = const_cast<page_std_string&>(*this);
    if(!self.m_helper.write_string_buffer\
          (m_id,m_name,(uint32)m_buffer->used(),m_buffer->data())) {
      m_out << "steve::HDF5::page_std_string::commit :"
            << " problem when writing array for" 
            << " " << tools::sout(m_name)
            << std::endl;
      return false;
    }
    delete self.m_buffer;
    self.m_buffer = 0;
    return true;
  }
  virtual bool retrieve() {
    unsigned int sz;
    char* buffer;
    if(!read_cstrings(m_id,m_name,sz,buffer)) {
      m_out << "steve::HDF5::page_std_string::retrieve :"
            << " problem when reading array for"
            << " " << tools::sout(m_name)
            << std::endl;
      return false;
    }
    m_buffer->assign(sz,buffer);
    if(!m_buffer->read_array(m_data)) {
      m_out << "steve::HDF5::page_std_string::retrieve :"
            << " problem when reading array for"
            << " " << tools::sout(m_name)
            << std::endl;
      return false;
    }
    delete m_buffer;
    m_buffer = 0;
    return true;
  }
public:
  page_std_string(std::ostream& a_out,
                  hid_t a_id,
                  uint32 a_size,
                  uint32 a_compression,
                  const std::string& a_name,
                  helper& a_helper)
  : HDF5::page<std::string>(a_id,a_size,a_compression,a_name)
  ,m_out(a_out)
  ,m_helper(a_helper)
  ,m_buffer(0)
  {
    m_buffer = new core::ascii_buffer(m_out);
  }  
  virtual ~page_std_string(){
    delete m_buffer;
  }
private:
  page_std_string(const page_std_string& a_from)
  : HDF5::page<std::string>(a_from)
  ,m_out(a_from.m_out)
  ,m_helper(a_from.m_helper)
  ,m_buffer(0)
  {}
  page_std_string& operator=(const page_std_string&) { return *this;}
private:
  std::ostream& m_out;
  helper& m_helper;
  core::ascii_buffer* m_buffer;
  std::vector<std::string> m_data;
};

}}

#endif
