#ifndef steve_data_h
#define steve_data_h

// Inheritance :
#include <steve/interface/data.h>
#include <steve/interface/data_T.h>
#include <steve/core/offset.h>

#include <steve/core/store_data.h>

namespace steve {
namespace core {

template <class T>
class data 
:public virtual interface::data 
,public virtual interface::data_T<T> 
,public core::offset {
public: //interface::data
  virtual interface::store_data& store_data() {
    return m_data;
  }
  //typedef data<T> data_T_t;
  virtual void* cast_interface_data_T() const {
    return (void*)static_cast< const interface::data_T<T>* >(this);
  }
public: //interface::data_T<T>
  virtual bool add(interface::store& a_store,
                   const T& a_value) { return m_data.add(a_store,a_value);}
  virtual bool add(interface::store& a_store,
                   uint32 a_size,const T* a_values) { 
    //used mainly for blob char columns (exa_4).
    return m_data.add(a_store,a_size,a_values);
  }
  virtual bool find_data(interface::store& a_store,
                         uint32 a_event,const path_t& a_is,T& a_value) {
    uint32 off = 0;
    if(!a_is.size()) {
      off = a_event;
    } else {
      if(!find_offset(a_store,a_event,a_is,off)) {
        m_out << "steve::data::find_data :"
              << " find_offset() failed."
              << std::endl;
        return false;
      }
    }
    if(!m_data.value(a_store,off,a_value)) {
      m_out << "steve::data::find_data :"
            << " error in m_data.value() for "
            << m_data.name()
            << std::endl;
      return false;
    } 
    return true;
  }
  virtual bool find_data(interface::store& a_store,
                         uint32 a_event,const path_t& a_begin_is,
                         uint32 a_size,T* a_values) {
    //used mainly for blob char columns (exa_4).
    uint32 off = 0;
    if(!a_begin_is.size()) {
      off = a_event;
    } else {
      if(!find_offset(a_store,a_event,a_begin_is,off)) {
        m_out << "steve::data::find_data :"
              << " find_offset() failed."
              << std::endl;
        return false;
      }
    }
    if(!m_data.values(a_store,off,a_size,a_values)) {
      m_out << "steve::data::find_data :"
            << " error in m_data.value() for "
            << m_data.name()
            << std::endl;
      return false;
    }
    return true;
  }
public:
  data(std::ostream& a_out,
       const path_os_t& a_path_os,
       const std::string& a_name,
       const std::string& a_T,
       uint32 a_page_size,
       uint32 a_compression)
  : core::offset(a_path_os)
  ,m_data(a_out,a_name,a_T,a_page_size,a_compression)
  ,m_out(a_out)
  {}
  virtual ~data(){}
private:
  data(const data& a_from)
  : core::offset(a_from)
  ,m_data(a_from.m_data)
  ,m_out(a_from.m_out)
  {}
  data& operator=(const data&){ return *this;}
public:
  core::store_data<T> m_data;
private:
  std::ostream& m_out;
};

}}

#endif
