#ifndef steve_core_store_data_h
#define steve_core_store_data_h

// inheritance :
#include <steve/interface/store_data.h>

#include <steve/interface/page.h>
#include <steve/interface/page_T.h>
#include <steve/interface/store.h>
#include <steve/interface/pages.h>

namespace steve {
namespace core {

template <class T>
class store_data : public virtual interface::store_data 
{
public: //interface::store_data
  virtual const std::string& name() const { return m_name;}
  virtual std::string name() { return m_name;}
  virtual const std::string& type() const { return m_T;}
  virtual std::string type() { return m_T;}
  virtual bool set_store_pages(interface::pages* a_pages,
                               bool a_delete = true) {
    m_store_pages = a_pages; 
    m_delete_store_pages = a_delete;
    return true;
  }
  virtual interface::pages* store_pages() const { 
    return m_store_pages;
  }
  virtual bool set_retrieve_pages(interface::pages* a_pages,
                                  bool a_delete = true) {
    m_retrieve_pages = a_pages;
    m_delete_retrieve_pages = a_delete;
    return true;
  }
  virtual interface::pages* retrieve_pages() const { 
    return m_retrieve_pages;
  }
  virtual steve::uint32 entries() const {
    return m_entries;
  } 
  virtual steve::uint32 store_bytes() const {
    return m_store_bytes;
  } 
  virtual steve::uint32 number_of_pages() const {
    if(m_store_pages) {
      return m_store_pages->number_of_pages();
    }
    if(m_retrieve_pages) {
      return m_retrieve_pages->number_of_pages();
    }
    return 0;
  }
  virtual steve::uint32 page_size() const {
    return m_page_size;
  }
  virtual steve::uint32 compression() const {
    return m_compression;
  }
  virtual bool set_page_size(steve::uint32 a_size) {
    m_page_size = a_size;
    return true;
  }
  virtual bool set_compression(steve::uint32 a_compression) {
    m_compression = a_compression;
    return true;
  }
private:
  typedef store_data<T> store_data_t;
public:
  store_data(std::ostream& a_out,
             const std::string& a_name,
             const std::string& a_T,
             uint32 a_page_size,
             uint32 a_compression)
  :m_out(a_out)
  ,m_name(a_name)
  ,m_T(a_T)
  ,m_page_size(a_page_size)
  ,m_compression(a_compression)
  ,m_entries(0)
  ,m_store_bytes(0)
  ,m_current_page(0)  
  ,m_store_pages(0)  
  ,m_delete_store_pages(true)
  ,m_retrieve_pages(0)  
  ,m_delete_retrieve_pages(true)
  {
  }
  virtual ~store_data(){
    if(m_delete_store_pages) delete m_store_pages;
    if(m_delete_retrieve_pages) delete m_retrieve_pages;
  }
protected:
  store_data(const store_data& a_from)
  :m_out(a_from.m_out)
  ,m_name(a_from.m_name)
  ,m_T(a_from.m_T)
  ,m_page_size(a_from.m_page_size)
  ,m_compression(a_from.m_compression)
  ,m_entries(0)
  ,m_store_bytes(0)
  ,m_current_page(0)  
  ,m_store_pages(0)  
  ,m_delete_store_pages(true)
  ,m_retrieve_pages(0)  
  ,m_delete_retrieve_pages(true)
  {}
private:
  store_data& operator=(const store_data&){ return *this;}
public:
  virtual bool add(interface::store& a_store,const T& a_value) {
    if(!m_current_page) {
      if(!new_page(a_store)) return false;
    }
    if(m_current_page->is_full()) {
      interface::page* bp = 
        (interface::page*)m_current_page->cast_base_page();
      if(!bp) return false;
      if(!a_store.commit_page(*this,bp)) return false;
      m_current_page = 0;
      if(!new_page(a_store)) return false;
    }
    uint32 sz;
    if(!m_current_page->add(a_value,sz)) return false;
    m_store_bytes += sz;
    m_entries++;
    return true;
  }
  virtual bool add(interface::store& a_store,uint32 a_size,const T* a_values) {
    //used mainly for blob char columns (exa_4).
    //if(!a_size) return true;
    uint32 rest = a_size;
    T* pos = (T*)a_values;      
    while(true) {
      if(!rest) return true;
      //something to store :
      if(!m_current_page) {
        if(!new_page(a_store)) return false;
      }
      uint32 remain = m_current_page->remain();
      if(!remain) {
        m_out << "steve::store_data::add :"
            << " attempt to use a non committed full page."
            << std::endl;
        return false;
      }
      if(rest<remain) {
        uint32 sz;
        if(!m_current_page->add(rest,pos,sz)) return false;
        m_store_bytes += sz;
        m_entries += rest;
        return true; //stop
      } else {
        uint32 sz;
        if(!m_current_page->add(remain,pos,sz)) return false;
        m_store_bytes += sz;
        m_entries += remain;
        pos += remain;
        rest -= remain;
        // commit page :
        interface::page* bp = 
         (interface::page*)m_current_page->cast_base_page();
        if(!bp) return false;
        if(!a_store.commit_page(*this,bp)) return false;
        m_current_page = 0;
        //continue
      }
    }
  }
public:
  bool value(interface::store& a_store,uint32 a_index,T& a_value) const { 
    uint32 index_in_page;
    interface::page_T<T>* page = find_page(a_store,a_index,index_in_page);
    if(!page) {
      m_out << "steve::store_data::value :"
            << " find_page() failed."
            << std::endl;
      return false;
    }
    return page->value(index_in_page,a_value);
  }
  bool values(interface::store& a_store,uint32 a_begin_index,
              uint32 a_size,T* a_values) const { 
    //used mainly for blob char columns (exa_4).

    //if(!a_size) return true;
    store_data_t& self = const_cast<store_data_t&>(*this);

/*
    uint32 index_in_page;
    interface::page_T<T>* page = 0;
 
    // get page size (need a page) :
    interface::page* bd = 
      a_store.find_page(self,a_begin_index,index_in_page);
    if(!bd) return false;
    page = (interface::page_T<T>*)bd->cast_interface_page_T();
    uint32 page_size = page->page_size();
    T* pos = (T*)a_values;
    if(!page->value(index_in_page,*pos)) return false;
    pos++;

    //FIXME : optimize : ask directly to the page a bunch of data.
    for(uint32 index=1;index<a_size;index++,pos++) {
      if((index_in_page+1)>=page_size) { //change page.
        interface::page* bd = 
          a_store.find_page(self,a_begin_index+index,index_in_page);
        if(!bd) return false;
        page = (interface::page_T<T>*)bd->cast_interface_page_T();
      } else { //stay in page
        index_in_page++;
      }
      if(!page->value(index_in_page,*pos)) return false;
    }
*/

    uint32 index_in_page;
    interface::page* bd = 
      a_store.find_page(self,a_begin_index,index_in_page);
    if(!bd) return false;
    interface::page_T<T>* page =
      (interface::page_T<T>*)bd->cast_interface_page_T();
    uint32 page_size = page->page_size();

    uint32 index = 0;
    T* pos = (T*)a_values;
    uint32 rest = a_size;
    while(true) {
      if(!rest) return true;
      uint32 remain = page_size-index_in_page;
      if(!remain) {
        m_out << "steve::store_data::values :"
              << " attempt to use a page at end."
              << std::endl;
        return false;
      }
      if(rest<remain) {
        if(!page->values(index_in_page,rest,pos)) return false;
        return true;
      } else {
        if(!page->values(index_in_page,remain,pos)) return false;
        pos += remain;
        rest -= remain;        
        index += remain;
        if(!rest) return true;
        // get a new page :
        interface::page* bd = 
          a_store.find_page(self,a_begin_index+index,index_in_page);
        if(!bd) return false;
        page = (interface::page_T<T>*)bd->cast_interface_page_T();
      }
    }

/*
*/

    return true;
  }
private:
  bool new_page(interface::store& a_store) {
    interface::page* bp = 
      a_store.create_page(*this,m_page_size,m_compression);
    if(!bp) return false;
    interface::page_T<T>* pg = 
      (interface::page_T<T>*)bp->cast_interface_page_T();
    if(!pg) return false;
    m_current_page = pg;
    return true;
  }
  interface::page_T<T>* find_page(interface::store& a_store,
                                  uint32 a_index,
                                  uint32& a_index_in_page) const { 
    store_data_t& self = const_cast<store_data_t&>(*this);
    interface::page* bd = 
      a_store.find_page(self,a_index,a_index_in_page);
    if(!bd) {
      m_out << "steve::store_data::find_page :"
            << " a_store.find_page() failed."
            << std::endl;
      return 0;
    }
    return (interface::page_T<T>*)bd->cast_interface_page_T();
  }
private:
  std::ostream& m_out;
private:
  std::string m_name;
  std::string m_T;
  uint32 m_page_size;
  uint32 m_compression;
  uint32 m_entries; //FIXME : uint64 ?
  uint32 m_store_bytes; //FIXME : uint64 ?
  interface::page_T<T>* m_current_page;
  interface::pages* m_store_pages;
  bool m_delete_store_pages;
  interface::pages* m_retrieve_pages;
  bool m_delete_retrieve_pages;
};

}}

#endif
