#ifndef steve_core_streamer_T_h
#define steve_core_streamer_T_h

#include <steve/interface/streamer.h>
#include <steve/interface/store.h>

#include <vector>

namespace steve {
namespace stream {

//////////////////////////////////////////////////////////////////////////////
/// store ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

template <class T>
bool out(interface::streamer& a_streamer,const std::vector<T>& a_value) { 
  steve::uint32 number = a_value.size();
  if(!a_streamer.write(number)) return false;
  typedef typename std::vector<T>::const_iterator it_t;
  for(it_t it=a_value.begin();it!=a_value.end();++it) {
    if(!a_streamer.write(*it)) return false;
  }
  return true;
}

template <class T>
bool out_with_refs(interface::streamer& a_streamer,
                  const std::vector<T*>& a_value) { 
  steve::uint32 number = a_value.size();
  // we must create the refs before writing the objects :
  typedef typename std::vector<T*>::const_iterator it_t;
 {std::vector<ref> refs(number);
  uint32 index = 0;
  for(it_t it=a_value.begin();it!=a_value.end();++it,index++) {
    ref ref;
    if(!a_streamer.store().create_store_ref(*it,ref)) return false;
    refs[index] = ref;
  }
  if(!out<ref>(a_streamer,refs)) return false;}
  // then we process the objects :
 {for(it_t it=a_value.begin();it!=a_value.end();++it) {
    if(!(*it)->stream_out(a_streamer)) return false;
  }}
  return true;
}

template <class T>
bool out(interface::streamer& a_streamer,const std::vector<T*>& a_value) { 
  steve::uint32 number = a_value.size();
  if(!a_streamer.write(number)) return false;
  typedef typename std::vector<T*>::const_iterator it_t;
  for(it_t it=a_value.begin();it!=a_value.end();++it) {
    if(!(*it)->stream_out(a_streamer)) return false;
  }
  return true;
}

template <class T>
bool out2(interface::streamer& a_streamer,const std::vector<T>& a_value) { 
  steve::uint32 number = a_value.size();
  if(!a_streamer.write(number)) return false;
  typedef typename std::vector<T>::const_iterator it_t;
  for(it_t it=a_value.begin();it!=a_value.end();++it) {
    if(!(*it).stream_out(a_streamer)) return false;
  }
  return true;
}

template <class T>
bool out3(interface::streamer& a_streamer,
                bool(*a_func)(interface::streamer&,const T&),
                const std::vector<T>& a_value) { 
  steve::uint32 number = a_value.size();
  if(!a_streamer.write(number)) return false;
  typedef typename std::vector<T>::const_iterator it_t;
  for(it_t it=a_value.begin();it!=a_value.end();++it) {
    if(!a_func(a_streamer,*it)) return false;
  }
  return true;
}

template <class T>
bool out_ref(interface::streamer& a_streamer,T* a_value) { 
  ref ref = 0;
  if(!a_streamer.store().find_store_ref(a_value,ref)) return false;
  return a_streamer.write(ref);
}

inline bool out(interface::streamer& a_streamer,const std::string& a_value) { 
  steve::uint32 number = a_value.size();
  if(!a_streamer.write(number)) return false;
  typedef std::string::const_iterator it_t;
  for(it_t it=a_value.begin();it!=a_value.end();++it) {
    if(!a_streamer.write(*it)) return false;
  }
  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// retrieve /////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

template <class T>
bool in(interface::streamer& a_streamer,std::vector<T>& a_value) { 
  steve::uint32 number;
  if(!a_streamer.read(number)) return false;
  a_value.resize(number);
  typedef typename std::vector<T>::iterator it_t;
  for(it_t it=a_value.begin();it!=a_value.end();++it) {
    if(!a_streamer.read(*it)) return false;
  }
  return true;
}

template <class T>
bool in(interface::streamer& a_streamer,std::vector<T*>& a_value) { 
  steve::uint32 number;
  if(!a_streamer.read(number)) return false;
  a_value.resize(number);
  typedef typename std::vector<T*>::iterator it_t;
  for(it_t it=a_value.begin();it!=a_value.end();++it) {
    (*it) = new T();
    if(!(*it)->stream_in(a_streamer)) return false;
  }
  return true;
}

template <class T>
bool in_with_refs(interface::streamer& a_streamer,
                           std::vector<T*>& a_value) { 

  std::vector<ref> refs;
  if(!in<ref>(a_streamer,refs)) return false;
  steve::uint32 number = refs.size();
  a_value.resize(number);
  typedef typename std::vector<T*>::iterator it_t;
 {for(it_t it=a_value.begin();it!=a_value.end();++it) (*it) = new T();}
 {uint32 index = 0;
  for(it_t it=a_value.begin();it!=a_value.end();++it,index++) {
    if(!a_streamer.store().create_retrieve_ref(refs[index],*it)) return false;
  }}
 {for(it_t it=a_value.begin();it!=a_value.end();++it) {
    if(!(*it)->stream_in(a_streamer)) return false;
  }}
  return true;
}

template <class T>
bool in2(interface::streamer& a_streamer,std::vector<T>& a_value) { 
  steve::uint32 number;
  if(!a_streamer.read(number)) return false;
  a_value.resize(number);
  typedef typename std::vector<T>::iterator it_t;
  for(it_t it=a_value.begin();it!=a_value.end();++it) {
    if(!(*it).stream_in(a_streamer)) return false;
  }
  return true;
}

template <class T>
bool in3(interface::streamer& a_streamer,
                  bool(*a_func)(interface::streamer&,T&),
                  std::vector<T>& a_value) { 
  steve::uint32 number;
  if(!a_streamer.read(number)) return false;
  a_value.resize(number);
  typedef typename std::vector<T>::iterator it_t;
  for(it_t it=a_value.begin();it!=a_value.end();++it) {
    if(!a_func(a_streamer,*it)) return false;
  }
  return true;
}

template <class T>
bool in_ref(interface::streamer& a_streamer,T*& a_value) { 
  ref ref = 0;
  if(!a_streamer.read(ref)) return false;
  void* pointer;
  if(!a_streamer.store().find_retrieve_ref(ref,pointer)) return false;
  a_value = (T*)pointer;
  return true;
}

inline bool in(interface::streamer& a_streamer,std::string& a_value) { 
  steve::uint32 number;
  if(!a_streamer.read(number)) return false;
  a_value.resize(number);
  typedef std::string::iterator it_t;
  for(it_t it=a_value.begin();it!=a_value.end();++it) {
    if(!a_streamer.read(*it)) return false;
  }
  return true;
}

}}

#endif
