#ifndef steve_helper_h
#define steve_helper_h

#include <steve/typedefs.h>
#include <steve/interface/numbers.h>
#include <steve/interface/offsets.h>
#include <steve/interface/data.h>
#include <steve/interface/store_data.h>
#include <steve/interface/data_T.h>
#include <steve/interface/store.h>
#include <steve/interface/schema.h>

#include <vector>

namespace steve {

//////////////////////////////////////////////////////////////////////////////
/// store ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

// to store basic types :
template <class T>
bool store(interface::store& a_store,
           steve::interface::data& a_data,
           const T& a_value) { 
  steve::interface::data_T<T>* d = 
    (steve::interface::data_T<T>*)a_data.cast_interface_data_T();
  if(!d) return false;
  if(!d->add(a_store,a_value)) return false;
  return true;
}

// to store refs :
template <class T>
bool store_ref(interface::store& a_store,
               steve::interface::data& a_data,
               T* a_value) { 
  steve::interface::data_T<ref>* d = 
    (steve::interface::data_T<ref>*)a_data.cast_interface_data_T();
  if(!d) return false;
  ref ref = 0;
  if(!a_store.find_store_ref(a_value,ref)) return false;
  if(!d->add(a_store,ref)) return false;
  return true;
}

// to store vector of basic types :
template <class T>
bool store(interface::store& a_store,
           steve::interface::offsets& a_offsets,
           steve::interface::numbers& a_numbers,
           steve::interface::data& a_data,
           const std::vector<T>& a_value) { 
  if(!a_offsets.add(a_store,a_numbers.total())) return false;
  if(!a_numbers.add(a_store,a_value.size())) return false;

  steve::interface::data_T<T>* d = 
    (steve::interface::data_T<T>*)a_data.cast_interface_data_T();
  if(!d) return false;

  if(a_value.size()) {
    if(!d->add(a_store,a_value.size(),&(a_value[0]))) return false;
  }

  return true;
}

// to store vector of vector of basic types 
// with sub vector having the same size :
template <class T>
bool store(interface::store& a_store,
           steve::interface::offsets& a_offsets,
           steve::interface::numbers& a_numbers,
           steve::interface::data& a_data,
           const std::vector< std::vector<T> >& a_value) { 

  unsigned int sz;
  if(a_value.size()) {
    sz = a_value.size()*a_value[0].size;
  } else {
    sz = 0;
  }

  if(!a_offsets.add(a_store,a_numbers.total())) return false;
  if(!a_numbers.add(a_store,sz)) return false;

  steve::interface::data_T<T>* d = 
    (steve::interface::data_T<T>*)a_data.cast_interface_data_T();
  if(!d) return false;

  typedef typename std::vector< std::vector<T> >::const_iterator it_t;
  for(it_t it=a_value.begin();it!=a_value.end();++it) {
    std::vector<T>& vec = *it;
    typedef typename std::vector<T>::const_iterator it2_t;
    for(it2_t it2=vec.begin();it2!=vec.end();++it2) {
      if(!d->add(a_store,*it2)) return false;
    }
  }

  return true;
}

// to store vector of ref :
template <class T>
bool store_refs(interface::store& a_store,
                steve::interface::offsets& a_offsets,
                steve::interface::numbers& a_numbers,
                steve::interface::data& a_data,
                const std::vector<T*>& a_value) { 
  if(!a_offsets.add(a_store,a_numbers.total())) return false;
  if(!a_numbers.add(a_store,a_value.size())) return false;

  steve::interface::data_T<ref>* d = 
    (steve::interface::data_T<ref>*)a_data.cast_interface_data_T();
  if(!d) return false;

  std::vector<ref> refs(a_value.size());

  typedef typename std::vector<T*>::const_iterator it_t;
  uint32 index = 0;
  for(it_t it=a_value.begin();it!=a_value.end();++it,index++) {
    ref ref;
    if(!a_store.create_store_ref(*it,ref)) return false;
    refs[index] = ref;
  }

  if(refs.size()) {
    if(!d->add(a_store,refs.size(),&(refs[0]))) return false;
  }

  return true;
}

// to store vector of instances* :
template <class T>
bool store(interface::store& a_store,
           steve::interface::schema& a_schema,
           steve::interface::offsets& a_offsets,
           steve::interface::numbers& a_numbers,
           const std::vector<T*>& a_vector) { 
  if(!a_offsets.add(a_store,a_numbers.total())) return false;
  if(!a_numbers.add(a_store,a_vector.size())) return false;
  typedef typename std::vector<T*>::const_iterator it_t;
  for(it_t it=a_vector.begin();it!=a_vector.end();++it) {
    if(!(*it)->store(a_store,a_schema)) return false;
  }
  return true;
}

// to store vector of instances :
template <class T>
bool store(interface::store& a_store,
           steve::interface::schema& a_schema,
           steve::interface::offsets& a_offsets,
           steve::interface::numbers& a_numbers,
           const std::vector<T>& a_vector) { 
  if(!a_offsets.add(a_store,a_numbers.total())) return false;
  if(!a_numbers.add(a_store,a_vector.size())) return false;
  typedef typename std::vector<T>::const_iterator it_t;
  for(it_t it=a_vector.begin();it!=a_vector.end();++it) {
    if(!(*it).store(a_store,a_schema)) return false;
  }
  return true;
}

// to store vector of instances without a store method :
template <class T>
bool store(interface::store& a_store,
           steve::interface::schema& a_schema,
           steve::interface::offsets& a_offsets,
           steve::interface::numbers& a_numbers,
           bool(*a_func)(interface::store&,steve::interface::schema&,const T&),
           const std::vector<T>& a_vector) { 
  if(!a_offsets.add(a_store,a_numbers.total())) return false;
  if(!a_numbers.add(a_store,a_vector.size())) return false;
  typedef typename std::vector<T>::const_iterator it_t;
  for(it_t it=a_vector.begin();it!=a_vector.end();++it) {
    if(!a_func(a_store,a_schema,*it)) return false;
  }
  return true;
}

template <class T>
bool store_with_refs(interface::store& a_store,
                     steve::interface::schema& a_schema,
                     steve::interface::offsets& a_offsets,
                     steve::interface::numbers& a_numbers,
                     steve::interface::offsets& a_ref_offsets,
                     steve::interface::numbers& a_ref_numbers,
                     steve::interface::data& a_ref_data,
                     const std::vector<T*>& a_value) { 
  if(!steve::store_refs<T>
        (a_store,a_ref_offsets,a_ref_numbers,a_ref_data,a_value)) return false;
  if(!steve::store<T>(a_store,a_schema,
                      a_offsets,a_numbers,a_value)) return false;
  return true;
}

// to store std::string :
inline bool store(steve::interface::store& a_store,
           steve::interface::offsets& a_offsets,
           steve::interface::numbers& a_numbers,
           steve::interface::data& a_data,
           const std::string& a_value) { 
  if(!a_offsets.add(a_store,a_numbers.total())) return false;
  if(!a_numbers.add(a_store,a_value.size())) return false;

  steve::interface::data_T<char>* d = 
    (steve::interface::data_T<char>*)a_data.cast_interface_data_T();
  if(!d) return false;

  if(!d->add(a_store,a_value.size(),a_value.c_str())) return false;

  return true;
}

// to store blob :
inline bool store(steve::interface::store& a_store,
           steve::interface::offsets& a_offsets,
           steve::interface::numbers& a_numbers,
           steve::interface::data& a_data,
           steve::uint32 a_size,char* a_buffer) {
  if(!a_offsets.add(a_store,a_numbers.total())) return false;
  if(!a_numbers.add(a_store,a_size)) return false;

  steve::interface::data_T<char>* d = 
    (steve::interface::data_T<char>*)a_data.cast_interface_data_T();
  if(!d) return false;

  if(!d->add(a_store,a_size,a_buffer)) return false;

  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// retrieve /////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

// to retrieve basic types :
template <class T>
bool retrieve(interface::store& a_store,
              steve::interface::data& a_data,
              uint32 a_event,
              const path_t& a_is,
              T& a_value) { 
  steve::interface::data_T<T>* d = 
    (steve::interface::data_T<T>*)a_data.cast_interface_data_T();
  if(!d) return false;
  return d->find_data(a_store,a_event,a_is,a_value);
}

// to retrieve refs :
template <class T>
bool retrieve_ref(interface::store& a_store,
              steve::interface::data& a_data,
              uint32 a_event,
              const path_t& a_is,
              T*& a_value) { 
  steve::interface::data_T<ref>* d = 
    (steve::interface::data_T<ref>*)a_data.cast_interface_data_T();
  if(!d) return false;
  ref ref;
  if(!d->find_data(a_store,a_event,a_is,ref)) return false;
  void* pointer;
  if(!a_store.find_retrieve_ref(ref,pointer)) return false;
  a_value = (T*)pointer;
  return true;
}

// to retrieve vector of basic types :
template <class T>
bool retrieve(interface::store& a_store,
              steve::interface::numbers& a_numbers,
              steve::interface::data& a_data,
              uint32 a_event,
              const path_t& a_is,
              std::vector<T>& a_value) { 
  steve::uint32 number;
  if(!a_numbers.find_value(a_store,a_event,a_is,number)) return false;
  a_value.resize(number);

  steve::interface::data_T<T>* d = 
    (steve::interface::data_T<T>*)a_data.cast_interface_data_T();
  if(!d) return false;

  path_t is = a_is;
  is.push_back(0);
  uint32 index_end = a_is.size();

  if(number) {
    is[index_end] = 0;
    T* p = (T*)&(a_value[0]);
    if(!d->find_data(a_store,a_event,is,number,p)) return false;
  }

  return true;
}

// to retrieve vector of vector of basic types 
// with sub vector having the same size :
template <class T>
bool retrieve(interface::store& a_store,
              steve::interface::numbers& a_numbers,
              steve::interface::data& a_data,
              uint32 a_event,
              const path_t& a_is,
              unsigned int a_sub_size,
              std::vector< std::vector<T> >& a_value) { 

  steve::uint32 number; //overall size.
  if(!a_numbers.find_value(a_store,a_event,a_is,number)) return false;

  if(a_sub_size>number) return false;
  unsigned int nvec = number/a_sub_size;
  if(nvec*a_sub_size!=number) return false;

  a_value.resize(nvec);

  steve::interface::data_T<T>* d = 
    (steve::interface::data_T<T>*)a_data.cast_interface_data_T();
  if(!d) return false;

  path_t is = a_is;
  is.push_back(0);
  uint32 index_end = a_is.size();

  uint32 index = 0;
  typedef typename std::vector< std::vector<T> >::iterator it_t;
  for(it_t it=a_value.begin();it!=a_value.end();++it) {
    std::vector<T>& vec = *it;
    vec.resize(a_sub_size);
    typedef typename std::vector<T>::iterator it2_t;
    for(it2_t it2=vec.begin();it2!=vec.end();++it2) {
      is[index_end] = index;
      if(!d->find_data(a_store,a_event,is,*it2)) return false;
      index++;
    }
  }

  return true;
}

// to retrieve vector of refs :
template <class T>
bool retrieve_refs(interface::store& a_store,
                   steve::interface::data& a_data,
                   uint32 a_event,
                   const path_t& a_is,
                   const std::vector<T*>& a_value) { 

  steve::interface::data_T<ref>* d = 
    (steve::interface::data_T<ref>*)a_data.cast_interface_data_T();
  if(!d) return false;

  path_t is = a_is;
  is.push_back(0);
  uint32 index_end = a_is.size();

  std::vector<ref> refs(a_value.size());
  if(a_value.size()) {
    is[index_end] = 0;
    ref* p = (ref*)&(refs[0]);
    if(!d->find_data(a_store,a_event,is,a_value.size(),p)) return false;
  }

  uint32 index = 0;
  typedef typename std::vector<T*>::const_iterator it_t;
  for(it_t it=a_value.begin();it!=a_value.end();++it,++index) {
    is[index_end] = index;
    if(!a_store.create_retrieve_ref(refs[index],a_value[index])) return false;
  }

  return true;
}

// to retrieve vector of instances* :
template <class T>
bool retrieve_new(interface::store& a_store,
                  steve::interface::numbers& a_numbers,
                  uint32 a_event,
                  const path_t& a_is,
                  std::vector<T*>& a_vector) { 
  steve::uint32 number;
  if(!a_numbers.find_value(a_store,a_event,a_is,number)) return false;
  a_vector.resize(number);
  for(steve::uint32 index=0;index<number;index++) a_vector[index] = new T();
  return true;
}

// to retrieve vector of instances* without taking into account refs :
template <class T>
bool retrieve(interface::store& a_store,
              steve::interface::schema& a_schema,
              steve::interface::numbers& a_numbers,
              uint32 a_event,
              const path_t& a_is,
              std::vector<T*>& a_vector) { 
  if(!retrieve_new(a_store,a_numbers,a_event,a_is,a_vector)) return false;

  path_t is = a_is;
  is.push_back(0);
  uint32 index_end = a_is.size();

  uint32 index = 0;
  typedef typename std::vector<T*>::iterator it_t;
  for(it_t it=a_vector.begin();it!=a_vector.end();++it,++index) {
    is[index_end] = index;
    if(!(*it)->retrieve(a_store,a_schema,a_event,is)) return false;
  }

  return true;
}

// to retrieve vector of instances :
template <class T>
bool retrieve_new(interface::store& a_store,
                  steve::interface::numbers& a_numbers,
                  uint32 a_event,
                  const path_t& a_is,
                  std::vector<T>& a_vector) { 
  steve::uint32 number;
  if(!a_numbers.find_value(a_store,a_event,a_is,number)) return false;
  a_vector.resize(number);
  return true;
}

// to retrieve vector of instances without taking into account refs :
template <class T>
bool retrieve(interface::store& a_store,
              steve::interface::schema& a_schema,
              steve::interface::numbers& a_numbers,
              uint32 a_event,
              const path_t& a_is,
              std::vector<T>& a_vector) { 
  if(!retrieve_new(a_store,a_numbers,a_event,a_is,a_vector)) return false;

  path_t is = a_is;
  is.push_back(0);
  uint32 index_end = a_is.size();

  uint32 index = 0;
  typedef typename std::vector<T>::iterator it_t;
  for(it_t it=a_vector.begin();it!=a_vector.end();++it,++index) {
    is[index_end] = index;
    if(!(*it).retrieve(a_store,a_schema,a_event,is)) return false;
  }

  return true;
}

// to retrieve vector of instances having no retrieve method
// an without taking into account refs :
template <class T>
bool retrieve(interface::store& a_store,
              steve::interface::schema& a_schema,
              steve::interface::numbers& a_numbers,
              bool(*a_func)(interface::store&,steve::interface::schema&,
                            steve::uint32,const steve::path_t&,T&),
              uint32 a_event,
              const path_t& a_is,
              std::vector<T>& a_vector) { 
  if(!retrieve_new(a_store,a_numbers,a_event,a_is,a_vector)) return false;

  path_t is = a_is;
  is.push_back(0);
  uint32 index_end = a_is.size();

  uint32 index = 0;
  typedef typename std::vector<T>::iterator it_t;
  for(it_t it=a_vector.begin();it!=a_vector.end();++it,++index) {
    is[index_end] = index;
    if(!a_func(a_store,a_schema,a_event,is,*it)) return false;
  }

  return true;
}

// to retrieve vector of instances and build refs :
template <class T>
bool retrieve_with_refs(interface::store& a_store,
              steve::interface::schema& a_schema,
              steve::interface::numbers& a_numbers,
              steve::interface::data& a_ref_data,
              uint32 a_event,
              const path_t& a_is,
              std::vector<T*>& a_vector) { 
  if(!retrieve_new(a_store,a_numbers,a_event,a_is,a_vector)) return false;

  if(!retrieve_refs(a_store,a_ref_data,a_event,a_is,a_vector)) 
    return false;

  path_t is = a_is;
  is.push_back(0);
  uint32 index_end = a_is.size();

  uint32 index = 0;
  typedef typename std::vector<T*>::iterator it_t;
  for(it_t it=a_vector.begin();it!=a_vector.end();++it,++index) {
    is[index_end] = index;
    if(!(*it)->retrieve(a_store,a_schema,a_event,is)) return false;
  }

  return true;
}

// to retrieve std::string :
inline bool retrieve(steve::interface::store& a_store,
              steve::interface::numbers& a_numbers,
              steve::interface::data& a_data,
              steve::uint32 a_event,
              const steve::path_t& a_is,
              std::string& a_value) { 
  steve::uint32 number;
  if(!a_numbers.find_value(a_store,a_event,a_is,number)) return false;
  a_value.resize(number);

  steve::interface::data_T<char>* d = 
    (steve::interface::data_T<char>*)a_data.cast_interface_data_T();
  if(!d) return false;

  path_t is = a_is;
  is.push_back(0);
  uint32 index_end = a_is.size();

  is[index_end] = 0;
  char* p = (char*)a_value.c_str();
  if(!d->find_data(a_store,a_event,is,number,p)) return false;

  return true;
}

// to retrieve blob :
inline bool retrieve(steve::interface::store& a_store,
              steve::interface::numbers& a_numbers,
              steve::interface::data& a_data,
              steve::uint32 a_event,
              const steve::path_t& a_is,
              steve::uint32& a_size,char*& a_buffer) { 
  // this method allocate a_buffer.
  if(!a_numbers.find_value(a_store,a_event,a_is,a_size)) return false;
  if(!a_size) { a_buffer = 0;return true;} // it is ok.
  a_buffer = new char[a_size];

  steve::interface::data_T<char>* d = 
    (steve::interface::data_T<char>*)a_data.cast_interface_data_T();
  if(!d) return false;

  path_t is = a_is;
  is.push_back(0);
  uint32 index_end = a_is.size();

  is[index_end] = 0;
  if(!d->find_data(a_store,a_event,is,a_size,a_buffer)) return false;

  return true;
}

}

#endif
