#ifndef steve_mysql_helper_h
#define steve_mysql_helper_h

#include <steve/tools/smanip.h>

#include <mysql/mysql.h>

#include <vector>

namespace steve {
namespace mysql {

bool do_query(
 std::ostream& a_out
,MYSQL* a_connection
,const char* a_string
,bool a_fatal = true
){
  if(mysql_query(a_connection,a_string)) {
    if(a_fatal) {
      a_out << "do_query :"
            << " for query " << tools::sout(a_string) //FIXME : may be huge !
            << " : " << std::string(mysql_error(a_connection))
            << std::endl;
      return false;
    }
  }
  MYSQL_RES* result = mysql_store_result(a_connection);
  if(result==NULL) return true;
  mysql_free_result(result);
  return true;
}

bool do_query(
 std::ostream& a_out
,MYSQL* a_connection
,const std::string& a_string
,bool a_fatal = true
){
  return do_query(a_out,a_connection,a_string.c_str(),a_fatal);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool store_uint32(
 std::ostream& a_out
,MYSQL* a_connection
,const std::string& a_table
,const std::string& a_what
,uint32 a_value
){
  std::string query = "INSERT INTO "+a_table+" ("+a_what+") VALUES (";
  std::string s = tools::to<uint32>(a_value);
  query += s;
  query += ")";
  return do_query(a_out,a_connection,query);
}

bool retrieve_uint32(
 std::ostream& a_out
,MYSQL* a_connection
,const std::string& a_table
,const std::string& a_what
,uint32& a_value
){
  std::string query;
  query  = "SELECT "+a_what+" FROM "+a_table;
  if(mysql_query(a_connection,query.c_str())) {
    a_out << "retrieve_uint32 :"
          << " for query " << tools::sout(query)
          << " : " << std::string(mysql_error(a_connection))
          << std::endl;
    return false;
  }

  MYSQL_RES* result = mysql_store_result(a_connection);
  if(result==NULL) return false;

  unsigned int num_fields = mysql_num_fields(result);
  my_ulonglong rows = mysql_num_rows(result);
  if((num_fields!=1)||(rows!=1)) {
    mysql_free_result(result);
    return false;
  }

  MYSQL_ROW row = mysql_fetch_row(result);
  if((row==NULL) || (row[0]==NULL) || !tools::to<uint32>(row[0],a_value)) {
    mysql_free_result(result);
    return false;
  }

  mysql_free_result(result);
  return true;
}

bool write_strings(
 std::ostream& a_out
,MYSQL* a_connection
,const std::string& a_name
,const std::vector<std::string>& a_array
){
  std::string query = "INSERT INTO "+a_name+" (value) VALUES ";
  uint32 number = a_array.size();
  bool anomalies = false;
  for(uint32 index=0;index<number;index++) {
    if(index) query += ",";
    query += "(";
    query += "'";
    query += a_array[index];
    query += "')";
    if(a_array[index].size()>65535) anomalies = true;
  }
  if(anomalies) {
    a_out << "write_strings :"
          << " some string lengths are greter than 65535."
          << ". Too much for BLOB type. MEDUIMBLOB or LONGBLOB needed."
          << std::endl;
  }
  return do_query(a_out,a_connection,query);
}

bool read_strings(
 std::ostream& a_out
,MYSQL* a_connection
,const std::string& a_name
,std::vector<std::string>& a_array
){
  std::string query;
  query  = "SELECT value FROM "+a_name;
  if(mysql_query(a_connection,query.c_str())) {
    a_out << "read_strings :"
          << " for query " << tools::sout(query)
          << " : " << std::string(mysql_error(a_connection))
          << std::endl;
    return false;
  }

  MYSQL_RES* result = mysql_store_result(a_connection);
  if(result==NULL) return false;

  unsigned int num_fields = mysql_num_fields(result);
  my_ulonglong rows = mysql_num_rows(result);
  if(num_fields!=1) {
    mysql_free_result(result);
    return false;
  }
  if(!rows) {
    a_array.clear();
    mysql_free_result(result);
    return true; // it is ok.
  }

  a_array.resize(rows);

 {MYSQL_ROW row;
  uint32 index = 0;
  while( ( row = mysql_fetch_row(result) ) != NULL ) {
    if(row[0]==NULL) {
      mysql_free_result(result);
      return false;
    }
    a_array[index].assign(row[0]);
    index++;
  }}

  mysql_free_result(result);
  return true;
}  
  
}}

#endif
