#ifndef steve_mysql_page_xxx_h
#define steve_mysql_page_xxx_h

// inheritance :
#include <steve/mysql/page.h>

#include <steve/mysql/T.h>
#include <steve/mysql/helper.h>
#include <steve/mysql/buffer.h>

namespace steve {
namespace mysql {

#define class_page_xxx(a_class,a_type)\
class a_class : public mysql::page< a_type > {\
public: /*interface::page :*/\
  virtual bool add(const a_type& a_value,uint32& a_size) {\
    if(m_index>=m_size) return false;\
    if(m_index) {\
      if(!m_buffer->write(",")) return false;\
    } else {\
      if(!m_buffer->write("INSERT INTO ")) return false;\
      if(!m_buffer->write(m_name)) return false;\
      if(!m_buffer->write(" (value) VALUES ")) return false;\
    }\
    if(!m_buffer->write("(")) return false;\
    if(!m_buffer->write(a_value)) return false;\
    if(!m_buffer->write(")")) return false;\
    m_index++;\
    a_size = sizeof(a_type);\
    return true;\
  }\
  virtual bool add(uint32 a_size,const a_type* a_values,\
                   uint32& a_items_size) {\
    for(uint32 index=0;index<a_size;index++) {\
      if(!add(a_values[index],a_items_size)) return false;\
    }\
    return true;\
  }\
  virtual bool value(uint32 a_index,a_type& a_value) const {\
    if(a_index>=m_size) return false;\
    a_value = m_data[a_index];\
    return true;\
  }\
  virtual bool values(uint32 a_index,uint32 a_size,a_type* a_values) const {\
    for(uint32 index=0;index<a_size;index++) {\
      if(!value(a_index+index,a_values[index])) return false;\
    }\
    return true;\
  }\
  virtual bool commit() const {\
    m_buffer->write_null();\
    if(!do_query(m_out,m_connection,m_buffer->data())) {\
      m_out << "steve::mysql::" << #a_class << "::commit :"\
            << " problem when writing array for"\
            << " " << tools::sout(m_name)\
            << std::endl;\
      return false;\
    }\
    a_class& self = const_cast<a_class&>(*this);\
    delete self.m_buffer;\
    self.m_buffer = 0;\
    return true;\
  }\
  virtual bool retrieve() {\
    unsigned int sz = m_size;\
    if(!read_array< a_type >(m_out,m_connection,m_name,\
                             sz,m_data)) {\
      m_out << "steve::mysql::" << #a_class << "::commit :"\
            << " problem when reading array for"\
            << " " << tools::sout(m_name)\
            << std::endl;\
      return false;\
    }\
    if(!sz) return false;\
    return true;\
  }\
public:\
  a_class(std::ostream& a_out,		\
          MYSQL* a_connection,uint32 a_size,uint32 a_index,\
          const std::string& a_name)\
  : mysql::page< a_type >(a_connection,a_size,a_index,a_name)\
  ,m_out(a_out)\
  ,m_buffer(0)\
  ,m_page_offset(0)\
  {\
    m_page_offset = m_page_index*m_size;\
    m_data = new a_type[m_size];\
    m_buffer = new mysql::buffer(m_out);\
  }\
  virtual ~a_class(){\
    delete [] m_data;\
    delete m_buffer;\
  }\
protected:\
  a_class(const a_class& a_from)\
  : mysql::page< a_type >(a_from)\
  ,m_out(a_from.m_out)\
  ,m_buffer(0)\
  ,m_page_offset(0)\
  {}\
private:\
  a_class& operator=(const a_class&) { return *this;}\
protected:\
  std::ostream& m_out;	\
  a_type* m_data;\
  mysql::buffer* m_buffer;\
  uint32 m_page_offset;\
};


class_page_xxx(page_uchar,uchar)
class_page_xxx(page_char,char)
class_page_xxx(page_uint32,uint32)
class_page_xxx(page_int32,int32)
class_page_xxx(page_uint16,uint16)
class_page_xxx(page_int16,int16)
class_page_xxx(page_uint64,uint64)
class_page_xxx(page_int64,int64)
class_page_xxx(page_float,float)
class_page_xxx(page_double,double)

class page_bool : public page_uchar {
public: //interface::page :
  virtual bool add(const bool& a_value,uint32& a_size) {
    return page_uchar::add(a_value?1:0,a_size);
  }
  virtual bool value(uint32 a_index,bool& a_value) const { 
    uchar value;
    if(!page_uchar::value(a_index,value)) return false;
    a_value = value==1?true:false;
    return true;
  }
public:
  page_bool(std::ostream& a_out,
            MYSQL* a_connection,uint32 a_size,uint32 a_index,
            const std::string& a_name)
  :page_uchar(a_out,a_connection,a_size,a_index,a_name)
  {}  
  virtual ~page_bool(){}
private:
  page_bool(const page_bool& a_from): page_uchar(a_from){}
  page_bool& operator=(const page_bool&) { return *this;}
};

class page_std_string : public mysql::page<std::string> {
public: //interface::page :
  virtual bool add(const std::string& a_value,uint32& a_size) {
    if(m_index>=m_size) return false;    
    m_data.push_back(a_value);
    m_index++;
    a_size = (uint32)a_value.size();
    return true;
  }
  virtual bool add(uint32 a_size,const std::string* a_values,
                   uint32& a_items_size) {
    std::string* pos = (std::string*)a_values;
    for(uint32 index=0;index<a_size;index++,pos++) {
      if(!add(*pos,a_items_size)) return false;
    }
    return true;
  }
  virtual bool value(uint32 a_index,std::string& a_value) const { 
    if(a_index>=m_data.size()) return false;    
    a_value = m_data[a_index];
    return true;
  }
  virtual bool values(uint32 a_index,
                      uint32 a_size,std::string* a_values) const {
    std::string* pos = (std::string*)a_values;
    for(uint32 index=0;index<a_size;index++,pos++) {
      if(!value(a_index+index,*pos)) return false;
    }
    return true;
  }
  virtual bool commit() const {
    page_std_string& self = const_cast<page_std_string&>(*this);
    if(!write_strings(m_out,m_connection,
                      m_name,m_data)) {
      m_out << "steve::mysql::page_std_string::commit :"
            << " problem when writing array for" 
            << " " << tools::sout(m_name)
            << std::endl;
      return false;
    }
    self.m_data.clear();
    return true;
  }
  virtual bool retrieve() {
    if(!read_strings(m_out,m_connection,m_name,m_data)) {
      m_out << "steve::mysql::page_std_string::retrieve :"
            << " problem when reading array for"
            << " " << tools::sout(m_name)
            << std::endl;
      return false;
    }
    return true;
  }
public:
  page_std_string(std::ostream& a_out,
                  MYSQL* a_connection,
                  uint32 a_size,
                  uint32 a_index,
                  const std::string& a_name)
  : mysql::page<std::string>(a_connection,a_size,a_index,a_name)
  ,m_out(a_out)
  {}  
  virtual ~page_std_string(){}
private:
  page_std_string(const page_std_string& a_from)
  : mysql::page<std::string>(a_from)
  ,m_out(a_from.m_out)
  {}
  page_std_string& operator=(const page_std_string&) { return *this;}
private:
  std::ostream& m_out;
  std::vector<std::string> m_data;
};

}}

#endif
