#ifndef steve_schema_h
#define steve_schema_h

// inheritance :
#include <steve/interface/schema.h>

#include <steve/interface/store.h>
#include <steve/interface/offsets.h>
#include <steve/interface/numbers.h>
#include <steve/interface/data.h>
#include <steve/path_os.h>

#define schema_header(a_class)\
public:\
  virtual void* cast_head() const;\
public:\
  a_class(std::ostream& a_printer,	\
          const std::string& a_namespace,\
          steve::path_os_t a_head = steve::path_os_t());\
  virtual ~a_class();\
private:\
  a_class(const a_class&);\
  a_class& operator=(const a_class&);

#define schema_imp(a_class)\
a_class::~a_class(){}\
a_class::a_class(const a_class& a_from):base_schema(a_from){}\
a_class& a_class::operator=(const a_class&){return *this;}\
void* a_class::cast_head() const {\
  return (void*)static_cast<const a_class*>(this);\
}

#define schema_constructor(a_class)\
a_class::a_class(\
 std::ostream& a_printer	\
,const std::string& a_namespace\
,steve::path_os_t a_head\
)\
:base_schema(a_printer,a_namespace)

#define schema_nos_header(a_name)\
public:\
  steve::interface::offsets* steve_offsets_##a_name() const;\
  steve::interface::numbers* steve_numbers_##a_name() const;\
private:\
  steve::interface::offsets* m_steve_offsets_##a_name;\
  steve::interface::numbers* m_steve_numbers_##a_name;


#define schema_offsets_new(a_name)\
 {std::string name = std::string("offsets_")+a_namespace+"_"+#a_name;\
  m_steve_offsets_##a_name = create_offsets(name,10000,6);\
  if(!m_steve_offsets_##a_name) {\
    m_out << "schema : can't create offsets " << name << std::endl;\
    return;\
  }}

#define schema_nos_method(a_schema,a_name)\
  steve::interface::offsets* a_schema::steve_offsets_##a_name() const {\
    return m_steve_offsets_##a_name;\
  }\
  steve::interface::numbers* a_schema::steve_numbers_##a_name() const {\
    return m_steve_numbers_##a_name;\
  }

#define schema_data_header(a_name)\
public:\
  steve::interface::data* steve_data_##a_name() const;\
private:\
  steve::interface::data* m_steve_data_##a_name;

#define schema_data_new(a_path,a_name,a_type) \
 {std::string name = std::string("data_")+a_namespace+"_"+#a_name;\
  m_steve_data_##a_name = create_data(a_path,name,a_type,10000,6);\
  if(!m_steve_data_##a_name) {\
    m_out << "schema : can't create data " << name << std::endl;\
    return;\
  }}

#define schema_data_method(a_schema,a_name)\
  steve::interface::data* a_schema::steve_data_##a_name() const {\
    return m_steve_data_##a_name;\
  }

#define schema_numbers_new(a_path,a_name) \
 {std::string name = std::string("numbers_")+a_namespace+"_"+#a_name;\
  m_steve_numbers_##a_name = create_numbers(a_path,name,10000,6);\
  if(!m_steve_numbers_##a_name) {\
    m_out << "schema : can't create numbers " << name << std::endl;\
    return;\
  }}

// used in the <class>::[store,retrieve] methods :
#define schema_get(a_class) a_class* schema = (a_class*)a_schema.cast_head();
#define schema_data_get(a_name)    *(schema->steve_data_##a_name())
#define schema_offsets_get(a_name) *(schema->steve_offsets_##a_name())
#define schema_numbers_get(a_name) *(schema->steve_numbers_##a_name())

#define schema_store_retrieve_stream(a__type,a__schema)\
bool a__type::store(\
 steve::interface::store& a_store\
,steve::interface::schema& a_schema\
) const {\
  schema_get(a__schema)\
  steve::core::streamer streamer(a_store);\
  if(!stream_out(streamer)) return false;\
  return steve::store(a_store,\
                      schema_offsets_get(m_blob),\
                      schema_numbers_get(m_blob),\
                      schema_data_get(m_blob),\
                      streamer.size(),streamer.data());\
}\
bool a__type::retrieve(\
 steve::interface::store& a_store\
,steve::interface::schema& a_schema\
,steve::uint32 a_event\
,const steve::path_t& a_is\
) {\
  schema_get(a__schema)\
  steve::uint32 sz;\
  char* buffer;\
  if(!steve::retrieve(a_store,\
                      schema_numbers_get(m_blob),\
                      schema_data_get(m_blob),\
                      a_event,a_is,sz,buffer)) return false;\
  steve::core::streamer streamer(a_store);\
  streamer.assign(sz,buffer); /*streamer get ownership of buffer.*/\
  if(!stream_in(streamer)) return false;\
  return true;\
}

#endif
