#ifndef steve_sqlite_T_h
#define steve_sqlite_T_h

#include <steve/sqlite/helper.h>

#include <sqlite3.h>

namespace steve {
namespace sqlite {

template <class T>
bool write_array(
 std::ostream& a_out
,::sqlite3* a_sqlite
,const std::string& a_name
,unsigned int a_size
,const T a_array[]
){
  if(!do_query(a_out,a_sqlite,"BEGIN")) return false; //optimize.

  std::string query = "INSERT INTO "+a_name+" (value) VALUES (?)";

  sqlite3_stmt* stmt;
  const char* tail = 0;
  if(sqlite3_prepare(a_sqlite,query.c_str(),-1,&stmt,&tail)!=SQLITE_OK) {
    a_out << "write_array<T> :"
          << " sqlite3_prepare failed :"
          << " " << std::string(sqlite3_errmsg(a_sqlite)) 
          << std::endl;
    return false;
  }

  for(uint32 index=0;index<a_size;index++) {
    write_col(stmt,1,a_array[index]);
    while(sqlite3_step(stmt)!=SQLITE_DONE) {}
    if(sqlite3_reset(stmt)!=SQLITE_OK) {
      a_out << "write_array<T> :"
            << " sqlite3_reset failed :"
            << " " << std::string(sqlite3_errmsg(a_sqlite)) 
            << std::endl;
      //return false;
      break;
    }
  }

  if(sqlite3_finalize(stmt)!=SQLITE_OK) {
    a_out << "write_array<T> :"
          << " sqlite3_finalize failed :"
          << " " << std::string(sqlite3_errmsg(a_sqlite)) 
          << std::endl;
    return false;
  }

  if(!do_query(a_out,a_sqlite,"COMMIT")) return false; //optimize.

  return true;
}

template <class T>
bool read_array_stmt(
 std::ostream& a_out
,::sqlite3* a_sqlite
,const std::string& a_name
,unsigned int& a_size
,T*& a_array
){
  // It is assumed that a_size contains in input 
  // the page size and that a_array is allocated with a_size data.
  std::string query;
  query  = "SELECT value FROM "+a_name;

  sqlite3_stmt* stmt;
  const char* tail = 0;
  if(sqlite3_prepare(a_sqlite,query.c_str(),-1,&stmt,&tail)!=SQLITE_OK) {
    a_out << "read_array_stmt :"
          << " sqlite3_prepare failed :"
          << " " << std::string(sqlite3_errmsg(a_sqlite)) 
          << std::endl;
    return false;
  }

  uint32 count = 0;
  while(sqlite3_step(stmt)==SQLITE_ROW) {
    int coln = sqlite3_column_count(stmt);
    if(coln!=1) return false;
    T v;
    read_col(stmt,0,v);
    if(count>a_size) return false;
    a_array[count] = v;
    count++;    
  }

  if(sqlite3_finalize(stmt)!=SQLITE_OK) {
    a_out << "read_array_stmt :"
          << " sqlite3_finalize failed :"
          << " " << std::string(sqlite3_errmsg(a_sqlite)) 
          << std::endl;
    return false;
  }

  a_size = count;

  if(!count) a_array = 0;

  return true;
}
  
}}

#endif
