#ifndef steve_sqlite_helper_h
#define steve_sqlite_helper_h

#include <steve/tools/smanip.h>

#include <sqlite3.h>

#include <vector>

namespace steve {
namespace sqlite {

bool do_query(
 std::ostream& a_out
,::sqlite3* a_sqlite
,const char* a_string
,bool a_fatal = true
){
  char* errmsg = 0;
  if(sqlite3_exec(a_sqlite,a_string,NULL,0,&errmsg)!=SQLITE_OK) {
    if(a_fatal) {
      a_out << "do_query :"
            << " for query " << tools::sout(a_string) //FIXME : may be huge !
            << " : " << std::string(errmsg)
            << std::endl;
      sqlite3_free(errmsg);
      return false;
    } else {
      sqlite3_free(errmsg);
    }
  }
  return true;
}

bool do_query(
 std::ostream& a_out
,::sqlite3* a_sqlite
,const std::string& a_string
,bool a_fatal = true
){
  return do_query(a_out,a_sqlite,a_string.c_str(),a_fatal);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool store_uint32(
 std::ostream& a_out
,::sqlite3* a_sqlite
,const std::string& a_table
,const std::string& a_what
,uint32 a_value
){
  std::string query = "INSERT INTO "+a_table+" ("+a_what+") VALUES (";
  std::string s = tools::to<uint32>(a_value);
  query += s;
  query += ")";
  return do_query(a_out,a_sqlite,query);
}

static int retrieve_callback(
 void* a_tag
,int a_argn
,char** a_args
,char** a_cols
){
  typedef std::vector< std::vector<uint32> > retrieve_rows;
  retrieve_rows* a_rows = (retrieve_rows*)a_tag;
  std::vector<uint32> row(a_argn);
  for(int i=0; i<a_argn; i++){
    if(a_args[i]==NULL) return 1;
    if(!tools::to<uint32>(a_args[i],row[i])) return 1;
  }
  a_rows->push_back(row);
  return 0;
}

bool retrieve_uint32(
 std::ostream& a_out
,::sqlite3* a_sqlite
,const std::string& a_table
,const std::string& a_what
,uint32& a_value
){
  std::string query;
  query  = "SELECT "+a_what+" FROM "+a_table;
  typedef std::vector< std::vector<uint32> > retrieve_rows;
  retrieve_rows rrows;
  char* errmsg = 0;
  if(sqlite3_exec(a_sqlite,query.c_str(),
                  retrieve_callback,&rrows,&errmsg)!=SQLITE_OK) {
    a_out << "retrieve_uint32 :"
          << " for query " << tools::sout(query)
          << " : " << std::string(errmsg)
          << std::endl;
    return false;
  }

  uint32 rows = rrows.size();
  if(rows!=1) return false;
  uint32 num_fields = rrows[0].size();
  if(num_fields!=1) return false;

  a_value = rrows[0][0];

  return true;
}

bool write_strings(
 std::ostream& a_out
,::sqlite3* a_sqlite
,const std::string& a_name
,const std::vector<std::string>& a_array
){
  if(!do_query(a_out,a_sqlite,"BEGIN")) return false; //optimize.
  uint32 number = a_array.size();
  for(uint32 index=0;index<number;index++) {
    if(a_array[index].size()>65535) {
      a_out << "write_strings :"
            << " some string lengths are greter than 65535."
            << ". Too much for BLOB type. MEDUIMBLOB or LONGBLOB needed."
            << std::endl;
    }

    std::string query = "INSERT INTO "+a_name+" (value) VALUES ";
    query += "(";
    query += "'";
    query += a_array[index];
    query += "')";
    if(!do_query(a_out,a_sqlite,query)) return false;
  }
  if(!do_query(a_out,a_sqlite,"COMMIT")) return false; //optimize.
  return true;
}

bool read_strings(
 std::ostream& a_out
,::sqlite3* a_sqlite
,const std::string& a_name
,std::vector<std::string>& a_array
){
  std::string query;
  query  = "SELECT value FROM "+a_name;

  sqlite3_stmt* stmt;
  const char* tail = 0;
  if(sqlite3_prepare(a_sqlite,query.c_str(),-1,&stmt,&tail)!=SQLITE_OK) {
    a_out << "read_strings :"
          << " sqlite3_prepare failed :"
          << " " << std::string(sqlite3_errmsg(a_sqlite)) 
          << std::endl;
    return false;
  }

  a_array.clear();

  while(sqlite3_step(stmt)==SQLITE_ROW) {
    int coln = sqlite3_column_count(stmt);
    if(coln!=1) return false;
    char* txt = (char*)sqlite3_column_text(stmt,0);
    //printf("debug : \"%s\"\n",txt);
    a_array.push_back(txt);
  }

  if(sqlite3_finalize(stmt)!=SQLITE_OK) {
    a_out << "read_strings :"
          << " sqlite3_finalize failed :"
          << " " << std::string(sqlite3_errmsg(a_sqlite)) 
          << std::endl;
    return false;
  }

  return true;
}

bool drop_tables(std::ostream& a_out,::sqlite3* a_sqlite){
  // get list of tables :
  std::vector<std::string> tables;

 {std::string query;
  query  = "SELECT name FROM sqlite_master WHERE type='table'";
  sqlite3_stmt* stmt;
  const char* tail = 0;
  if(sqlite3_prepare(a_sqlite,query.c_str(),-1,&stmt,&tail)!=SQLITE_OK) {
    a_out << "read_strings :"
          << " sqlite3_prepare failed :"
          << " " << std::string(sqlite3_errmsg(a_sqlite)) 
          << std::endl;
    return false;
  }
  while(sqlite3_step(stmt)==SQLITE_ROW) {
    int coln = sqlite3_column_count(stmt);
    if(coln!=1) return false;
    char* txt = (char*)sqlite3_column_text(stmt,0);
    //printf("debug : \"%s\"\n",txt);
    tables.push_back(txt);
  }
  if(sqlite3_finalize(stmt)!=SQLITE_OK) {
    a_out << "read_strings :"
          << " sqlite3_finalize failed :"
          << " " << std::string(sqlite3_errmsg(a_sqlite)) 
          << std::endl;
    return false;
  }}

  // drop tables :
 {uint32 number = tables.size();
  for(uint32 index=0;index<number;index++) {
    if(!do_query(a_out,a_sqlite,
                 "DROP TABLE "+tables[index],false)) return false;
  }}

  return true;
}
  
//////////////////////////////////////////////////////////////////////////////
/// used in T.h //////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool write_col(sqlite3_stmt* a_stmt,int a_col,uchar a_v) {
  sqlite3_bind_int(a_stmt,a_col,a_v); //FIXME
  return true;
}
bool write_col(sqlite3_stmt* a_stmt,int a_col,char a_v) {
  sqlite3_bind_int(a_stmt,a_col,a_v); //FIXME
  return true;
}
bool write_col(sqlite3_stmt* a_stmt,int a_col,uint16 a_v) {
  sqlite3_bind_int(a_stmt,a_col,a_v); //FIXME
  return true;
}
bool write_col(sqlite3_stmt* a_stmt,int a_col,int16 a_v) {
  sqlite3_bind_int(a_stmt,a_col,a_v); //FIXME
  return true;
}
bool write_col(sqlite3_stmt* a_stmt,int a_col,uint32 a_v) {
  sqlite3_bind_int(a_stmt,a_col,a_v);
  return true;
}
bool write_col(sqlite3_stmt* a_stmt,int a_col,int32 a_v) {
  sqlite3_bind_int(a_stmt,a_col,a_v);
  return true;
}
bool write_col(sqlite3_stmt* a_stmt,int a_col,uint64 a_v) {
  sqlite3_bind_int64(a_stmt,a_col,a_v);
  return true;
}
bool write_col(sqlite3_stmt* a_stmt,int a_col,int64 a_v) {
  sqlite3_bind_int64(a_stmt,a_col,a_v);
  return true;
}
bool write_col(sqlite3_stmt* a_stmt,int a_col,double a_v) {
  sqlite3_bind_double(a_stmt,a_col,a_v);
  return true;
}
bool write_col(sqlite3_stmt* a_stmt,int a_col,float a_v) {
  sqlite3_bind_double(a_stmt,a_col,a_v);
  return true;
}

/*
  const void *sqlite3_column_blob(sqlite3_stmt*, int iCol);
  int sqlite3_column_bytes(sqlite3_stmt*, int iCol);
  int sqlite3_column_bytes16(sqlite3_stmt*, int iCol);
  sqlite3_int64 sqlite3_column_int64(sqlite3_stmt*, int iCol);
  const unsigned char *sqlite3_column_text(sqlite3_stmt*, int iCol);
  const void *sqlite3_column_text16(sqlite3_stmt*, int iCol);
  int sqlite3_column_type(sqlite3_stmt*, int iCol);
  sqlite3_value *sqlite3_column_value(sqlite3_stmt*, int iCol);
*/
bool read_col(sqlite3_stmt* a_stmt,int a_col,uchar& a_v) {
  a_v = (uchar)sqlite3_column_int(a_stmt,a_col); //FIXME
  return true;
}
bool read_col(sqlite3_stmt* a_stmt,int a_col,char& a_v) {
  a_v = (char)sqlite3_column_int(a_stmt,a_col); //FIXME
  return true;
}
bool read_col(sqlite3_stmt* a_stmt,int a_col,uint16& a_v) {
  a_v = (uint16)sqlite3_column_int(a_stmt,a_col); //FIXME
  return true;
}
bool read_col(sqlite3_stmt* a_stmt,int a_col,int16& a_v) {
  a_v = (int16)sqlite3_column_int(a_stmt,a_col); //FIXME
  return true;
}
bool read_col(sqlite3_stmt* a_stmt,int a_col,uint32& a_v) {
  a_v = (uint32)sqlite3_column_int(a_stmt,a_col);
  return true;
}
bool read_col(sqlite3_stmt* a_stmt,int a_col,int32& a_v) {
  a_v = (int32)sqlite3_column_int(a_stmt,a_col);
  return true;
}
bool read_col(sqlite3_stmt* a_stmt,int a_col,uint64& a_v) {
  a_v = (uint64)sqlite3_column_int64(a_stmt,a_col);
  return true;
}
bool read_col(sqlite3_stmt* a_stmt,int a_col,int64& a_v) {
  a_v = (int64)sqlite3_column_int64(a_stmt,a_col);
  return true;
}
bool read_col(sqlite3_stmt* a_stmt,int a_col,double& a_v) {
  a_v = sqlite3_column_double(a_stmt,a_col);
  return true;
}
bool read_col(sqlite3_stmt* a_stmt,int a_col,float& a_v) {
  a_v = (float)sqlite3_column_double(a_stmt,a_col); //FIXME
  return true;
}

}}

#endif
