#ifndef steve_tools_args_h
#define steve_tools_args_h

#include "smanip.h"

namespace steve {
namespace tools {

class args {
public:
  typedef std::pair<std::string,std::string> arg;
public:
  args(){}
  args(int a_argc,char* a_argv[]){
    for(int index=0;index<a_argc;index++) {
      std::vector<std::string> ws;
      words(a_argv[index],"=",false,ws);
      if(ws.size()==1) {
        m_args.push_back(arg(ws[0],"")); 
      } else if(ws.size()>=2) {
        m_args.push_back(arg(ws[0],ws[1])); 
      }
    }
  }
  /*
  args(const std::vector<std::string>& a_args){add(a_args);}
  args(const std::vector<arg>& a_args):m_args(a_args){}
  args(const std::string& a_args,const std::string& a_sep = " ",bool a_strip = false){
    std::vector<std::string> args;
    words(a_args,a_sep,false,args);  
    add(args,a_strip);
  }
  */
  virtual ~args(){}
  args(const args& a_from):m_args(a_from.m_args){}
  args& operator=(const args& a_from){m_args = a_from.m_args;return *this;}
public:
  /*
  inline const std::vector<arg>& get_args() const {return m_args;}
  */

  inline bool is_arg(const std::string& a_string) const {
    for(std::vector<arg>::const_iterator it = m_args.begin();
        it!=m_args.end();++it) {
      if((*it).first==a_string) return true;
    }
    return false;
  }
  unsigned int number() const {return m_args.size();}
  inline bool find(const std::string& a_string,std::string& a_value) const {
    for(std::vector<arg>::const_iterator it = m_args.begin();
        it!=m_args.end();++it) {
      if((*it).first==a_string) {
        a_value = (*it).second;
        return true;
      }
    }
    a_value = "";
    return false;
  }
  template <class aT>
  inline bool find(const std::string& a_string,aT& a_value) const {
    std::string s;
    if(!find(a_string,s)) {a_value = 0;return false;}
    return to<aT>(s,a_value);
  }
  /*
  inline std::vector<std::string> tovector() const {
    // Return a vector of string <name=value>
    std::vector<std::string> vec;
    for(std::vector<arg>::const_iterator it = m_args.begin();
        it!=m_args.end();++it) {
      std::string s;
      if((*it).second.empty()) {
        s = (*it).first;
      } else {
        s = (*it).first;
        s += "=";
        s += (*it).second;
      }
      vec.push_back(s);
    }
    return vec;
  }

  inline bool add(const std::string& a_key,const std::string& a_value,bool a_override = true){
    if(a_override) {
      for(std::vector<arg>::iterator it = m_args.begin();
          it!=m_args.end();++it) {
        if((*it).first==a_key) {
          (*it).second = a_value;
          return true;
        }
      }
    }
    if(a_key.empty()) return false;
    m_args.push_back(arg(a_key,a_value)); 
    return true;
  }

  inline void add(const std::vector<std::string>& a_args,bool a_strip = false) {
    for(std::vector<std::string>::const_iterator it = a_args.begin();
        it!=a_args.end();++it) {
      std::vector<std::string> ws;
      words((*it),"=",false,ws);
      if(ws.size()==1) {
        if(a_strip) {
          m_args.push_back(arg(strp(ws[0]),"")); 
        } else {
          m_args.push_back(arg(ws[0],"")); 
        }
      } else if(ws.size()>=2) {
        if(a_strip) {
          m_args.push_back(arg(strp(ws[0]),strp(ws[1]))); 
        } else {
          m_args.push_back(arg(ws[0],ws[1])); 
        }
      }
    }
  }

  inline int remove(const std::string& a_key){
    unsigned int nbeg = m_args.size();
    for(std::vector<arg>::iterator it = m_args.begin();it!=m_args.end();) {
      if(a_key==(*it).first) {
        it = m_args.erase(it);
      } else {
        ++it;
      }
    }
    return nbeg - m_args.size();
  }
  */  

  inline bool last(std::string& a_key,std::string& a_value) const {
    a_key = "";
    a_value = "";
    if(m_args.empty()) return false;
    a_key = m_args.back().first;
    a_value = m_args.back().second;
    return true;
  }

  inline bool file(std::string& a_file) const {
    std::string slast;
    std::string s;
    if((m_args.size()>1) //first arg is the program name !
       && last(slast,s) 
       && (slast.find('-')!=0) 
       && (s.empty()) ) {
      a_file = slast; //Last argument is not an option.
      return true;
    } else {
      a_file = "";
      return false;
    }
  }
  /*  
  inline std::vector<std::string> files(bool a_skip_first = true) const {
    // Get the serie of trailing args not beginning with '-'
    // and without a value (not of the form [-]xxx=yyy). 
    // Note that an argument like that in between arguments
    // not like that is NOT taken into account. 
    std::vector<std::string> files;
    if(m_args.empty()) return files;
    std::vector<arg>::const_iterator it = m_args.begin();
    if(a_skip_first) it++;
    for(;it!=m_args.end();++it) {
      if( ((*it).first.find('-')==0) || (*it).second.size() ) {
        files.clear();
      } else {
        files.push_back((*it).first);
      }
    }
    return files;
  }  
  inline bool argcv(int& a_argc,char**& a_argv) const {
    // If using with :
    //   int argc;
    //   char** argv;
    //   args.argcv(argc,argv);  
    // you can delete with :
    //   args.delete_argcv(argc,argv);
    if(m_args.empty()) {a_argc = 0;a_argv = 0;return true;}
    typedef char* cstring;
    cstring* av = new cstring[m_args.size()];
    if(!av) {a_argc = 0;a_argv = 0;return false;}
    a_argv = av;
    for(std::vector<arg>::const_iterator it = m_args.begin();
        it!=m_args.end();++it,av++) {
      std::string::size_type lf = (*it).first.length();
      std::string::size_type ls = (*it).second.length();
      std::string::size_type sz = 0;
      if(ls) {
        sz = lf + 1 + ls;
      } else {
        sz = lf;
      }
      char* p = new char[sz+1];
      if(!p) {a_argc = 0;a_argv = 0;return false;} //some delete are lacking.
      *av = p;
     {char* pf = (char*)(*it).first.c_str();
      for(std::string::size_type i=0;i<lf;i++) {*p++ = *pf++;}
      *p = 0;}
      if(ls) {*p = '=';p++;}
     {char* ps = (char*)(*it).second.c_str();
      for(std::string::size_type i=0;i<ls;i++) {*p++ = *ps++;}
      *p = 0;}
    }
    a_argc = (int)m_args.size();
    return true;
  }
  inline void delete_argcv(int& a_argc,char**& a_argv) const {
    for(int index=0;index<a_argc;index++) delete [] a_argv[index];
    delete [] a_argv;
    a_argc = 0;
    a_argv = 0;
  }
  inline bool known_options(const std::vector<std::string>& a_opts) const {
    for(std::vector<arg>::const_iterator it = m_args.begin();
        it!=m_args.end();++it) {
      if((*it).first.find('-')==0) { //find '-' at first pos.
        bool found = false;
        for(std::vector<std::string>::const_iterator it2 = a_opts.begin();
          it2!=a_opts.end();++it2) {
          if((*it).first==(*it2)) {
            found = true;
            break;        
          }
        }
        if(!found) return false;
      }
    }
    return true;
  }

  inline void files_at_end(bool a_skip_first = true) {
    // reorder to have "file" arguments at end.
    if(m_args.empty()) return;
    std::vector<arg> args;
    if(a_skip_first) args.push_back(*(m_args.begin()));
    //first pass :
   {std::vector<arg>::const_iterator it = m_args.begin();
    if(a_skip_first) it++;
    for(;it!=m_args.end();++it) {
      if( ((*it).first.find('-')==0) || (*it).second.size() ) {
        args.push_back(*it);
      }
    }}
    //second pass :
   {std::vector<arg>::const_iterator it = m_args.begin();
    if(a_skip_first) it++;
    for(;it!=m_args.end();++it) {
      if( ((*it).first.find('-')==0) || (*it).second.size() ) {
      } else {
        args.push_back(*it);
      }
    }}
    m_args = args;
  }
  */
private:
  std::vector<arg> m_args;
};


}}

#endif
