#ifndef steve_zip_page_xxx_h
#define steve_zip_page_xxx_h

// inheritance :
#include <steve/zip/page.h>

#include <steve/interface/buffer.h>

#include <steve/core/binary_no_swap_buffer.h>
#include <steve/core/platform.h>
#include <steve/core/binary_swap_buffer.h>
#include <steve/core/ascii_buffer.h>

#include <steve/tools/smanip.h>

#include <vector>

#include <string.h> //memcpy

namespace steve {
namespace zip {

#define class_page_xxx(a_class,a_type)\
class a_class : public zip::page< a_type > {\
public: /*interface::page :*/\
  virtual bool add(const a_type& a_value,uint32& a_item_size) {\
    if(m_index>=m_size) return false;\
    if(!m_buffer->write(a_value)) return false;\
    m_index++;\
    a_item_size = sizeof(a_type);\
    return true;\
  }\
  virtual bool add(uint32 a_size,const a_type* a_values,\
                   uint32& a_items_size) {\
    if((m_index+a_size)>m_size) return false;\
    if(!m_buffer->write_array(a_size,a_values)) return false;\
    m_index += a_size;\
    a_items_size = a_size*sizeof(a_type);\
    return true;\
  }\
  virtual bool value(uint32 a_index,a_type& a_value) const {\
    if(a_index>=m_data_size) {\
      m_out << "steve::zip::" << #a_class << "::value :"\
            << " failed : a_index(" << a_index << ") >=m_data_size(" << m_data_size << ")."\
            << std::endl;\
      return false;\
    }\
    a_value = *(m_data+a_index);\
    return true;\
  }\
  virtual bool values(uint32 a_index,uint32 a_size,a_type* a_values) const {\
    if((a_index+a_size)>m_data_size) {\
      m_out << "steve::zip::" << #a_class << "::value :"\
            << " failed : a_index+a_size(" << (a_index+a_size) << ") >=m_data_size(" << m_data_size << ")."\
            << std::endl;\
      return false;\
    }\
    ::memcpy(a_values,m_data+a_index,a_size*sizeof(a_type));\
    return true;\
  }\
  virtual bool commit() const {\
    if(!m_writer.open_entry(m_name,m_compression)) return false;\
    if(!m_writer.write((uint32)m_buffer->used(),\
                       m_buffer->data())) return false;\
    if(!m_writer.close_entry()) return false;\
    a_class& self = const_cast<a_class&>(*this);\
    delete self.m_buffer;\
    self.m_buffer = 0;\
    return true;\
  }\
  virtual bool retrieve() {\
    unsigned int sz;\
    char* buffer;\
    if(!m_reader.read_entry(m_name,sz,buffer)) return false;\
    m_buffer->assign(sz,buffer);\
    m_data_size = 0;\
    m_data = 0;\
    if(!m_buffer->read_array(m_data_size,m_data)) {\
      m_out << "steve::zip::" << #a_class << "::commit :"\
            << " problem when reading array for"\
            << " " << tools::sout(m_name)	\
            << std::endl;\
      return false;\
    }\
    delete m_buffer;\
    m_buffer = 0;\
    /*if(!m_data_size) return false;*/\
    return true;\
  }\
public:\
  a_class(std::ostream& a_out,	\
          zip::writer& a_writer,\
          zip::reader& a_reader,\
          uint32 a_size,uint32 a_compression,const std::string& a_name)\
  : zip::page< a_type >(a_writer,a_reader,a_size,a_compression,a_name)\
  ,m_out(a_out)\
  ,m_buffer(0)\
  ,m_data_size(0)\
  ,m_data(0)\
  ,m_binary(true)\
  {\
    if(m_binary) {\
      if(core::is_little_endian()) {\
        /*We do not byte swap on intels.*/\
        m_buffer = new core::binary_no_swap_buffer(m_out);\
      } else {\
        m_buffer = new core::binary_swap_buffer(m_out);\
      }\
    } else {\
      m_buffer = new core::ascii_buffer(m_out);\
    }\
  }\
  virtual ~a_class(){\
    delete [] m_data;\
    delete m_buffer;\
  }\
protected:\
  a_class(const a_class& a_from)\
  : zip::page< a_type >(a_from)\
  ,m_out(a_from.m_out)\
  ,m_buffer(0)\
  ,m_data_size(0)\
  ,m_data(0)\
  ,m_binary(true)\
  {}\
private:\
  a_class& operator=(const a_class&) { return *this;}\
protected:\
  std::ostream& m_out;	\
  interface::buffer* m_buffer;\
  uint32 m_data_size;\
  a_type* m_data;\
  bool m_binary;\
};


class_page_xxx(page_uchar,uchar)
class_page_xxx(page_char,char)
class_page_xxx(page_uint32,uint32)
class_page_xxx(page_int32,int32)
class_page_xxx(page_uint16,uint16)
class_page_xxx(page_int16,int16)
class_page_xxx(page_uint64,uint64)
class_page_xxx(page_int64,int64)
class_page_xxx(page_float,float)
class_page_xxx(page_double,double)
class_page_xxx(page_bool,bool)

//NOTE : optimizzation : 
//  We avoid the std::vector<std::string> when storing 
//  because it induces an extra copy of the added string.
class page_std_string : public zip::page<std::string> {
public: //interface::page :
  virtual bool add(const std::string& a_value,uint32& a_size) {
    if(m_index>=m_size) return false;    
    if(!m_buffer->write(a_value)) return false;
    m_index++;
    a_size = (uint32)a_value.size();
    return true;
  }
  virtual bool add(uint32 a_size,const std::string* a_values,
                   uint32& a_items_size) {
    std::string* pos = (std::string*)a_values;
    for(uint32 index=0;index<a_size;index++,pos++) {
      if(!add(*pos,a_items_size)) return false;
    }
    return true;
  }
  virtual bool value(uint32 a_index,std::string& a_value) const { 
    if(a_index>=m_data.size()) return false;    
    a_value = m_data[a_index];
    return true;
  }
  virtual bool values(uint32 a_index,
                      uint32 a_size,std::string* a_values) const {
    std::string* pos = (std::string*)a_values;
    for(uint32 index=0;index<a_size;index++,pos++) {
      if(!value(a_index+index,*pos)) return false;
    }
    return true;
  }
  virtual bool commit() const {
    if(!m_writer.open_entry(m_name,m_compression)) return false;
    if(!m_writer.write((uint32)m_buffer->used(),
                       m_buffer->data())) return false;
    if(!m_writer.close_entry()) return false;
    page_std_string& self = const_cast<page_std_string&>(*this);
    delete self.m_buffer;
    self.m_buffer = 0;
    return true;
  }
  virtual bool retrieve() {
    unsigned int sz;
    char* buffer;
    if(!m_reader.read_entry(m_name,sz,buffer)) return false;
    m_buffer->assign(sz,buffer);
    if(!m_buffer->read_array(m_data)) {
      m_out << "steve::zip::page_std_string::commit :"
            << " problem when reading array for"
            << " " << tools::sout(m_name)
            << std::endl;
      return false;
    }
    delete m_buffer;
    m_buffer = 0;
    return true;
  }
public:
  page_std_string(std::ostream& a_out,
                  writer& a_writer,
                  reader& a_reader,
                  uint32 a_size,
                  uint32 a_compression,
                  const std::string& a_name)
  : zip::page<std::string>(a_writer,a_reader,a_size,a_compression,a_name)
  ,m_out(a_out)
  ,m_buffer(0)
  {
    m_buffer = new core::ascii_buffer(m_out);
  }  
  virtual ~page_std_string(){
    delete m_buffer;
  }
private:
  page_std_string(const page_std_string& a_from)
  : zip::page<std::string>(a_from)
  ,m_out(a_from.m_out)
  ,m_buffer(0)
  {}
  page_std_string& operator=(const page_std_string&) { return *this;}
private:
  std::ostream& m_out;
  core::ascii_buffer* m_buffer;
  std::vector<std::string> m_data;
};

}}

#endif
