// this :
#include <Exp/Event.h>

#include <Exp/Analysis.h>

#include <steve/tools/smanip.h>
#include <steve/util/mem.h>

//////////////////////////////////////////////////////////////////////////////
/// Time /////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Exp::Time::Time()
:fTime(0)
,fRandom(false)
{
  steve::util::mem::increment("Exp::Time");
}
Exp::Time::Time(Params& a_params,
     Exp::uint32 a_indexInRun,
     Exp::uint32 a_indexInEvent,
     Exp::uint32 a_indexInTrack,
     Exp::uint32 a_indexInHit)
:fTime(0)
{
  steve::util::mem::increment("Exp::Time");
  fRandom = a_params.m_event_random;
  if(a_params.m_event_random) {
    fTime = (Exp::int32)(100.0*a_params.m_random_flat->shoot());
  } else {
    fTime = a_indexInHit;
  }
}
Exp::Time::~Time(){
  steve::util::mem::decrement("Exp::Time");
}
Exp::Time::Time(const Time&){}
Exp::Time& Exp::Time::operator=(const Time&){return *this;}
Exp::int32 Exp::Time::time() const { return fTime;}
bool Exp::Time::is_random() const { return fRandom;}

//////////////////////////////////////////////////////////////////////////////
/// Hit //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Exp::Hit::Hit()
:fX(0)
,fY(0)
,fZ(0)
,ARRAY_STAR_INIT(fTimes,Exp::Time)
#ifdef EXP_REF
,REF_INIT(fTrack,0)
#endif
,fRandom(false)
{
  steve::util::mem::increment("Exp::Hit");
}
Exp::Hit::Hit(Params& a_params,
    Exp::Track& a_track,
    Exp::uint32 a_indexInRun,
    Exp::uint32 a_indexInEvent,
    Exp::uint32 a_indexInTrack)
:fX(0)
,fY(0)
,fZ(0)
,ARRAY_STAR_INIT(fTimes,Exp::Time)
#ifdef EXP_REF
,REF_INIT(fTrack,&a_track)
#endif
{
  steve::util::mem::increment("Exp::Hit");
  fRandom = a_params.m_event_random;
  if(a_params.m_event_random) {
    fX = a_params.m_random_gauss->shoot();
    fY = a_params.m_random_gauss->shoot();
    fZ = a_params.m_random_gauss->shoot();
  } else {
    fX = a_indexInRun;
    fY = a_indexInEvent;
    fZ = a_indexInTrack;
  } 

 {// Create times on hit :
  Exp::uint32 number_of_times = a_params.m_event_fixed?10:a_indexInTrack+2;
   for(Exp::uint32 index=0;index<number_of_times;index++) {
     ARRAY_STAR_ADD(fTimes,index,Exp::Time(a_params,a_indexInRun,a_indexInEvent,a_indexInTrack,index));
     fVec.push_back(index+2);
  }}
}
Exp::Hit::~Hit(){
  ARRAY_STAR_DELETE(fTimes,Exp::Time)
  steve::util::mem::decrement("Exp::Hit");
}
Exp::Hit::Hit(const Hit&){}
Exp::Hit& Exp::Hit::operator=(const Hit&){return *this;}
double Exp::Hit::x() const { return fX;}
double Exp::Hit::y() const { return fY;}
double Exp::Hit::z() const { return fZ;}
bool Exp::Hit::is_random() const { return fRandom;}
std::vector<double> Exp::Hit::vec() const { return fVec;}
ARRAY_STAR_METHODS(Exp::Hit,fTimes,Exp::Time,number_of_times,time_at)
#ifdef EXP_REF
REF_METHOD(Exp::Hit,fTrack,Exp::Track,track)
#endif

//////////////////////////////////////////////////////////////////////////////
/// Track ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Exp::Track::Track() //Used when retrieving.
:fPx(0)
,fPy(0)
,fPz(0)
,ARRAY_STAR_INIT(fHits,Exp::Hit)
,fRandom(false)
{
  steve::util::mem::increment("Exp::Track");
}
Exp::Track::Track(Params& a_params,
             Exp::uint32 a_indexInRun,Exp::uint32 a_indexInEvent)
:fPx(0)
,fPy(0)
,fPz(0)
,ARRAY_STAR_INIT(fHits,Exp::Hit)
{
  steve::util::mem::increment("Exp::Track");
  fRandom = a_params.m_event_random;
  if(a_params.m_event_random) {
    fPx = a_params.m_random_gauss->shoot();
    fPy = a_params.m_random_gauss->shoot();
    fPz = a_params.m_random_gauss->shoot();
  } else {
    fPx = a_indexInRun;
    fPy = a_indexInEvent;
    fPz = fPx+fPy;
  }
 {// Create hits on track :
  Exp::uint32 number_of_hits = a_params.m_event_fixed?100:a_indexInEvent+1;
  for(Exp::uint32 index=0;index<number_of_hits;index++) {
    ARRAY_STAR_ADD(fHits,index,Exp::Hit(a_params,*this,a_indexInRun,a_indexInEvent,index));
  }}
}  
Exp::Track::~Track(){
  ARRAY_STAR_DELETE(fHits,Exp::Hit)
  steve::util::mem::decrement("Exp::Track");
}
Exp::Track::Track(const Track&){}
Exp::Track& Exp::Track::operator=(const Track&){return *this;}
double Exp::Track::px() const { return fPx;}
double Exp::Track::py() const { return fPy;}
double Exp::Track::pz() const { return fPz;}
bool Exp::Track::is_random() const { return fRandom;}
ARRAY_STAR_METHODS(Exp::Track,fHits,Exp::Hit,number_of_hits,hit_at)

//////////////////////////////////////////////////////////////////////////////
/// Cell /////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Exp::Cell::Cell() //Used when retrieving.
#ifdef EXP_STRING
:fID("")
,fID_Vec("")
#endif
{
  steve::util::mem::increment("Exp::Cell");
}
Exp::Cell::Cell(const Cell& a_from)
#ifdef EXP_STRING
:fID(a_from.fID)
,fID_Vec(a_from.fID_Vec)
#endif
{
  steve::util::mem::increment("Exp::Cell");
}
Exp::Cell& Exp::Cell::operator=(const Cell& a_from)
{
#ifdef EXP_STRING
  fID = a_from.fID;
  fID_Vec = a_from.fID_Vec;
#endif
  return *this;
}
Exp::Cell::Cell(Params& a_params,
           Exp::uint32 a_indexInRun,Exp::uint32 a_indexInEvent)
{
  steve::util::mem::increment("Exp::Cell");
#ifdef EXP_STRING
  fID = "Exp::Cell:";
 {for(Exp::uint32 index=0;index<(a_indexInEvent+1);index++) {
     fID += steve::tools::to<Exp::uint32>(a_indexInEvent+1);
  }}
  fID_Vec = fID;
#endif
}  
Exp::Cell::~Cell(){
  steve::util::mem::decrement("Exp::Cell");
}
#ifdef EXP_STRING
std::string Exp::Cell::id() const { return fID;}
std::string Exp::Cell::id_vec() const { return fID_Vec;}
#endif

//////////////////////////////////////////////////////////////////////////////
/// Event ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Exp::Event::Event() //Used when retrieving.
:fAnalysis(0)
,fID(0)
,fEnergy(0)
,fHisto("",0,0,0)
,ARRAY_STAR_INIT(fTracks,Exp::Track)
,ARRAY_INIT(fCells,Exp::Cell)
{
  steve::util::mem::increment("Exp::Event");
}
Exp::Event::Event(Params& a_params,
                  Exp::uint32 a_indexInRun,Analysis* a_analysis)
:fAnalysis(a_analysis)
,fID(0)
,fEnergy(0)
,fHisto("Event_fHisto",100,-10,10)
,ARRAY_STAR_INIT(fTracks,Exp::Track)
,ARRAY_INIT(fCells,Exp::Cell)
{
  steve::util::mem::increment("Exp::Event");
  fID = a_indexInRun;
  fEnergy = a_params.m_random_gauss->shoot();
  if(a_analysis) a_analysis->fEnergy.fill(fEnergy);

 {// Create tracks on event :
  Exp::uint32 number_of_tracks = a_params.m_event_fixed?100:a_indexInRun+1;
  for(Exp::uint32 index=0;index<number_of_tracks;index++) {
    ARRAY_STAR_ADD(fTracks,index,Exp::Track(a_params,a_indexInRun,index));
  }}

 {// Create cells in event :
  Exp::uint32 number_of_cells = a_params.m_event_fixed?100:a_indexInRun+1;
  for(Exp::uint32 index=0;index<number_of_cells;index++) {
    ARRAY_ADD(fCells,index,Exp::Cell(a_params,a_indexInRun,index));
  }}

 {// Fill fHisto :
  for(Exp::uint32 index=0;index<a_indexInRun;index++) {
    fHisto.fill(a_params.m_random_gauss->shoot());
  }}

 {fBlob.resize(10);
  for(Exp::uint32 index=0;index<10;index++) {
    fBlob[index] = (char)index;
  }}

  fObject.initialize();
  
}  
Exp::Event::~Event(){
  ARRAY_STAR_DELETE(fTracks,Exp::Track)
  ARRAY_DELETE(fCells,Exp::Cell)
  steve::util::mem::decrement("Exp::Event");
}
Exp::Event::Event(const Event& a_from)
:fAnalysis(0)
,fHisto("",0,0,0)
{}
Exp::Event& Exp::Event::operator=(const Event&){return *this;}

unsigned int Exp::Event::id() const { return fID;}
float Exp::Event::energy() const { return fEnergy;}
HCL::Histogram1D& Exp::Event::histo() { return fHisto;}
steve::tools::blob& Exp::Event::blob() { return fBlob;}
Exp::Object& Exp::Event::object() { return fObject;}
ARRAY_STAR_METHODS(Exp::Event,fTracks,Exp::Track,number_of_tracks,track_at)
ARRAY_METHODS(Exp::Event,fCells,Exp::Cell,number_of_cells,cell_at)


//////////////////////////////////////////////////////////////////////////////
/// Header ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Exp::Header::Header() //Used when retrieving.
:fNumberOfEvents(0)
{
  steve::util::mem::increment("Exp::Header");
}
Exp::Header::~Header(){
  steve::util::mem::decrement("Exp::Header");
}
Exp::Header::Header(const Header& a_from)
:fNumberOfEvents(0)
{}
Exp::Header& Exp::Header::operator=(const Header&){return *this;}

unsigned int Exp::Header::numberOfEvents() const { return fNumberOfEvents;}
void Exp::Header::setNumberOfEvents(unsigned int a_value) { 
 fNumberOfEvents = a_value;
}

//////////////////////////////////////////////////////////////////////////////
/// Object ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Exp::Object::Object() //Used when retrieving.
:f_uchar(0)
,f_char(0)
,f_uint16(0)
,f_int16(0)
,f_uint32(0)
,f_int32(0)
,f_uint64(0)
,f_int64(0)
,f_float(0)
,f_double(0)
,f_string("")
{
  steve::util::mem::increment("Exp::Object");
}
void Exp::Object::initialize() {
  f_uchar = 1;
  f_char = 2;
  f_uint16 = 3;
  f_int16 = 4;
  f_uint32 = 5;
  f_int32 = 6;
  f_uint64 = 7;
  f_int64 = 8;
  f_float = 9;
  f_double = 10;
  f_string = "11";
}
Exp::Object::Object(const Object& a_from) //private
:f_string(a_from.f_string)
{
  steve::util::mem::increment("Exp::Object");
}
Exp::Object& Exp::Object::operator=(const Object& a_from) { return *this;}
Exp::Object::~Object(){
  steve::util::mem::decrement("Exp::Object");
}
steve::uchar Exp::Object::get_uchar() const { return f_uchar;}
char Exp::Object::get_char() const { return f_char;}
steve::uint16 Exp::Object::get_uint16() const { return f_uint16;}
steve::int16 Exp::Object::get_int16() const { return f_int16;}
steve::uint32 Exp::Object::get_uint32() const { return f_uint32;}
steve::int32 Exp::Object::get_int32() const { return f_int32;}
steve::uint64 Exp::Object::get_uint64() const { return f_uint64;}
steve::int64 Exp::Object::get_int64() const { return f_int64;}
float Exp::Object::get_float() const { return f_float;}
double Exp::Object::get_double() const { return f_double;}
std::string Exp::Object::get_string() const { return f_string;}

bool Exp::Object::is_equal(const Object& a_from) const {
  if(f_uchar!=a_from.f_uchar) return false;
  if(f_char!=a_from.f_char) return false;
  if(f_uint16!=a_from.f_uint16) return false;
  if(f_int16!=a_from.f_int16) return false;
  if(f_uint32!=a_from.f_uint32) return false;
  if(f_int32!=a_from.f_int32) return false;
  if(f_uint64!=a_from.f_uint64) return false;
  if(f_int64!=a_from.f_int64) return false;
  if(f_float!=a_from.f_float) return false;
  if(f_double!=a_from. f_double) return false;
  if(f_string!=a_from.f_string) return false;
  return true;
}
