
#include <Exp/Event.h>

#include <steve/helper.h>
#include <steve/core/streamer.h> //for Object.

#include <Exp/Schema.h>
#include <Exp/Event_Schema.h>
#include <HCL/helper.h> //for Event.fHisto.
#include <Exp/Header_Schema.h>

//////////////////////////////////////////////////////////////////////////////
/// Time /////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool Exp::Time::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  EXP_SCHEMA
  if(!steve::store<Exp::int32>(a_store,
                               EXP_DATA(Time_fTime),
                               fTime)) return false;
  if(!steve::store<bool>(a_store,
                         EXP_DATA(Time_fRandom),
                         fRandom)) return false;
  return true;
}
bool Exp::Time::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) { 
  EXP_SCHEMA
  if(!steve::retrieve<Exp::int32>(a_store,
                                  EXP_DATA(Time_fTime),
                                  a_event,a_is,fTime)) return false;
  if(!steve::retrieve<bool>(a_store,
                            EXP_DATA(Time_fRandom),
                            a_event,a_is,fRandom)) return false;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// Hit //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool Exp::Hit::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  EXP_SCHEMA
  if(!steve::store<double>(a_store,
                           EXP_DATA(Hit_fX),
                           fX)) return false;
  if(!steve::store<double>(a_store,
                           EXP_DATA(Hit_fY),
                           fY)) return false;
  if(!steve::store<double>(a_store,
                           EXP_DATA(Hit_fZ),
                           fZ)) return false;
  if(!steve::store<double>(a_store,
                           EXP_OFFSETS(Hit_fVec),
                           EXP_NUMBERS(Hit_fVec),
                           EXP_DATA(Hit_fVec),
                           fVec)) return false;
  if(!steve::store<bool>(a_store,
                         EXP_DATA(Hit_fRandom),
                         fRandom)) return false;
#ifdef EXP_REF
  if(!steve::store_ref<Exp::Track>(a_store,
                              EXP_DATA(Hit_fTrack),
                              fTrack)) return false;
#endif
  if(!steve::store<Exp::Time>(a_store,a_schema,
                         EXP_OFFSETS(Hit_fTimes),
                         EXP_NUMBERS(Hit_fTimes),
                         fTimes)) return false;
  return true;
}
bool Exp::Hit::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) { 
  EXP_SCHEMA
  if(!steve::retrieve<double>(a_store,
                              EXP_DATA(Hit_fX),
                              a_event,a_is,fX)) return false;
  if(!steve::retrieve<double>(a_store,
                              EXP_DATA(Hit_fY),
                              a_event,a_is,fY)) return false;
  if(!steve::retrieve<double>(a_store,
                              EXP_DATA(Hit_fZ),
                              a_event,a_is,fZ)) return false;
  if(!steve::retrieve<double>(a_store,
                              EXP_NUMBERS(Hit_fVec),
                              EXP_DATA(Hit_fVec),
                              a_event,a_is,fVec)) return false;
  if(!steve::retrieve<bool>(a_store,
                            EXP_DATA(Hit_fRandom),
                            a_event,a_is,fRandom)) return false;
#ifdef EXP_REF
  if(!steve::retrieve_ref<Exp::Track>(a_store,
                              EXP_DATA(Hit_fTrack),
                              a_event,a_is,fTrack)) return false;
#endif
  if(!steve::retrieve<Exp::Time>(a_store,a_schema,
                            EXP_NUMBERS(Hit_fTimes),
                            a_event,a_is,fTimes)) return false;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// Track ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool Exp::Track::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  EXP_SCHEMA
  if(!steve::store<double>(a_store,
                           EXP_DATA(Track_fPx),
                           fPx)) return false;
  if(!steve::store<double>(a_store,
                           EXP_DATA(Track_fPy),
                           fPy)) return false;
  if(!steve::store<double>(a_store,
                           EXP_DATA(Track_fPz),
                           fPz)) return false;
  if(!steve::store<Exp::Hit>(a_store,a_schema,
                        EXP_OFFSETS(Track_fHits),
                        EXP_NUMBERS(Track_fHits),
                        fHits)) return false;
  if(!steve::store<bool>(a_store,
                         EXP_DATA(Track_fRandom),
                         fRandom)) return false;
  return true;
}
bool Exp::Track::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) { 
  EXP_SCHEMA
  if(!steve::retrieve<double>(a_store,
                              EXP_DATA(Track_fPx),
                              a_event,a_is,fPx)) return false;
  if(!steve::retrieve<double>(a_store,
                              EXP_DATA(Track_fPy),
                              a_event,a_is,fPy)) return false;
  if(!steve::retrieve<double>(a_store,
                              EXP_DATA(Track_fPz),
                              a_event,a_is,fPz)) return false;
  if(!steve::retrieve<Exp::Hit>(a_store,a_schema,
                           EXP_NUMBERS(Track_fHits),
                           a_event,a_is,fHits)) return false;
  if(!steve::retrieve<bool>(a_store,
                            EXP_DATA(Track_fRandom),
                            a_event,a_is,fRandom)) return false;
  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// Cell /////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool Exp::Cell::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  EXP_SCHEMA
#ifdef EXP_STRING
  if(!steve::store<std::string>(a_store,
                                EXP_DATA(Cell_fID),
                                fID)) return false;
  if(!steve::store(a_store,
                   EXP_OFFSETS(Cell_fID_Vec),
                   EXP_NUMBERS(Cell_fID_Vec),
                   EXP_DATA(Cell_fID_Vec),
                   fID_Vec)) return false;
#endif
  return true;
}
bool Exp::Cell::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) { 
  EXP_SCHEMA
#ifdef EXP_STRING
  if(!steve::retrieve<std::string>(a_store,
                                   EXP_DATA(Cell_fID),
                                   a_event,a_is,fID)) return false;
  if(!steve::retrieve(a_store,
                      EXP_NUMBERS(Cell_fID_Vec),
                      EXP_DATA(Cell_fID_Vec),
                      a_event,a_is,fID_Vec)) return false;
#endif
  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// Event ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

//#define EVENT_THIS_BLOB

#ifndef EVENT_THIS_BLOB

bool Exp::Event::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  //FIXME steve::core::schema_streamer streamer(a_store,a_schema);
  //FIXME if(!stream_out(streamer)) return false;

  EXP_SCHEMA
  if(!steve::store<Exp::uint32>(a_store,
                          EXP_DATA(Event_fID),
                          fID)) return false;
  if(!steve::store<float>(a_store,
                          EXP_DATA(Event_fEnergy),
                          fEnergy)) return false;

  if(!steve::HCL::store(a_store,exp_schema->f_Event_fHisto,
                        fHisto.getData())) return false;

  if(!steve::store(a_store,
                   EXP_OFFSETS(Event_fBlob),
                   EXP_NUMBERS(Event_fBlob),
                   EXP_DATA(Event_fBlob),
                   fBlob.size(),fBlob.data())) return false;

  if(!fObject.store(a_store,a_schema)) return false;

#ifdef EXP_REF
  if(!steve::store_with_refs<Exp::Track>(a_store,a_schema,
                                    EXP_OFFSETS(Event_fTracks),
                                    EXP_NUMBERS(Event_fTracks),
                                    EXP_OFFSETS(Event_fTracks_ref),
                                    EXP_NUMBERS(Event_fTracks_ref),
                                    EXP_DATA(Event_fTracks_ref),
                                    fTracks)) return false;
#else
  if(!steve::store<Exp::Track>(a_store,a_schema,
                          EXP_OFFSETS(Event_fTracks),
                          EXP_NUMBERS(Event_fTracks),
                          fTracks)) return false;
#endif
  if(!steve::store<Exp::Cell>(a_store,a_schema,
                         EXP_OFFSETS(Event_fCells),
                         EXP_NUMBERS(Event_fCells),
                         fCells)) return false;
  return true;
}
bool Exp::Event::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) {    
  EXP_SCHEMA
  if(!steve::retrieve<Exp::uint32>(a_store,
                             EXP_DATA(Event_fID),
                             a_event,a_is,fID)) return false;
  if(!steve::retrieve<float>(a_store,
                             EXP_DATA(Event_fEnergy),
                             a_event,a_is,fEnergy)) return false;
 {HCL::HistogramData data;
  if(!steve::HCL::retrieve(a_store,exp_schema->f_Event_fHisto,
                           a_event,a_is,data)) return false;
  fHisto.copy(data);}

  steve::uint32 sz;
  char* buffer;
  if(!steve::retrieve(a_store,
                      EXP_NUMBERS(Event_fBlob),
                      EXP_DATA(Event_fBlob),
                      a_event,a_is,sz,buffer)) return false;
  fBlob.assign(sz,buffer);

  if(!fObject.retrieve(a_store,a_schema,a_event,a_is)) return false;

#ifdef EXP_REF
  if(!steve::retrieve_with_refs<Exp::Track>(a_store,a_schema,
                                       EXP_NUMBERS(Event_fTracks),
                                       EXP_DATA(Event_fTracks_ref),
                                       a_event,a_is,fTracks))return false;
#else
  if(!steve::retrieve<Exp::Track>(a_store,a_schema,
                             EXP_NUMBERS(Event_fTracks),
                             a_event,a_is,fTracks))return false;
#endif
  if(!steve::retrieve<Exp::Cell>(a_store,a_schema,
                            EXP_NUMBERS(Event_fCells),
                            a_event,a_is,fCells)) return false;
  return true;
}

#else //EVENT_THIS_BLOB

// put each  Event instance in a blob.

bool Exp::Event::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  EXP_SCHEMA

  steve::core::streamer streamer(a_store);

  if(!stream_out(streamer)) return false;

  //printf("debug : blob size %d\n",streamer.size());
  
  return steve::store(a_store,
                      EXP_OFFSETS(Event_this_blob),
                      EXP_NUMBERS(Event_this_blob),
                      EXP_DATA(Event_this_blob),
                      streamer.size(),streamer.data());
}
bool Exp::Event::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) { 
  EXP_SCHEMA

  steve::uint32 sz = 0;
  char* buffer = 0;
  if(!steve::retrieve(a_store,
                      EXP_NUMBERS(Event_this_blob),
                      EXP_DATA(Event_this_blob),
                      a_event,a_is,sz,buffer)) return false;

  steve::core::streamer streamer(a_store);
  streamer.assign(sz,buffer); //streamer get the ownership of buffer.

  if(!stream_in(streamer)) return false;

  return true;
}

#endif //EVENT_THIS_BLOB

//////////////////////////////////////////////////////////////////////////////
/// Header ////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

bool Exp::Header::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  HEADER_SCHEMA
  if(!steve::store<Exp::uint32>(a_store,
                                HEADER_DATA(Header_fNumberOfEvents),
                                fNumberOfEvents)) return false;
  return true;
}
bool Exp::Header::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) {    
  HEADER_SCHEMA
  if(!steve::retrieve<Exp::uint32>(a_store,
                                   HEADER_DATA(Header_fNumberOfEvents),
                                   a_event,a_is,fNumberOfEvents)) return false;
  return true;
}

bool Exp::Object::store(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
) const { 
  EXP_SCHEMA

  steve::core::streamer streamer(a_store);

  if(!stream_out(streamer)) return false;

  return steve::store(a_store,
                   EXP_OFFSETS(Event_fObject),
                   EXP_NUMBERS(Event_fObject),
                   EXP_DATA(Event_fObject),
                   streamer.size(),streamer.data());
}
bool Exp::Object::retrieve(
 steve::interface::store& a_store
,steve::interface::schema& a_schema
,steve::uint32 a_event
,const steve::path_t& a_is
) { 
  EXP_SCHEMA

  steve::uint32 sz;
  char* buffer;
  if(!steve::retrieve(a_store,
                      EXP_NUMBERS(Event_fObject),
                      EXP_DATA(Event_fObject),
                      a_event,a_is,sz,buffer)) return false;

  steve::core::streamer streamer(a_store);
  streamer.assign(sz,buffer); //streamer get ownership of buffer.

  if(!stream_in(streamer)) return false;

  return true;
}

