
#include <Exp/Event.h>

#include <steve/interface/streamer.h>
#include <steve/core/streamer_T.h>

#include <HCL/stream.h>

//#define CLASS_VERSION 1

bool Exp::Object::stream_out(steve::interface::streamer& a_streamer) const { 
  //steve::uint16 version = CLASS_VERSION;
  //if(!a_streamer.write(version)) return false;

  if(!a_streamer.write(f_uchar)) return false;
  if(!a_streamer.write(f_char)) return false;
  if(!a_streamer.write(f_uint16)) return false;
  if(!a_streamer.write(f_int16)) return false;
  if(!a_streamer.write(f_uint32)) return false;
  if(!a_streamer.write(f_int32)) return false;
  if(!a_streamer.write(f_uint64)) return false;
  if(!a_streamer.write(f_int64)) return false;
  if(!a_streamer.write(f_float)) return false;
  if(!a_streamer.write(f_double)) return false;
  if(!a_streamer.write(f_string)) return false;
  return true;
}
bool Exp::Object::stream_in(steve::interface::streamer& a_streamer) { 
  //steve::uint16 version;
  //if(!a_streamer.read(version)) return false;
  //if(version!=CLASS_VERSION) return false;

  if(!a_streamer.read(f_uchar)) return false;
  if(!a_streamer.read(f_char)) return false;
  if(!a_streamer.read(f_uint16)) return false;
  if(!a_streamer.read(f_int16)) return false;
  if(!a_streamer.read(f_uint32)) return false;
  if(!a_streamer.read(f_int32)) return false;
  if(!a_streamer.read(f_uint64)) return false;
  if(!a_streamer.read(f_int64)) return false;
  if(!a_streamer.read(f_float)) return false;
  if(!a_streamer.read(f_double)) return false;
  if(!a_streamer.read(f_string)) return false;
  return true;
}

bool Exp::Time::stream_out(steve::interface::streamer& a_streamer) const { 
  //steve::uint16 version = CLASS_VERSION;
  //if(!a_streamer.write(version)) return false;

  if(!a_streamer.write(fTime)) return false;
  if(!a_streamer.write(fRandom)) return false;
  return true;
}
bool Exp::Time::stream_in(steve::interface::streamer& a_streamer) { 
  //steve::uint16 version;
  //if(!a_streamer.read(version)) return false;
  //if(version!=CLASS_VERSION) return false;

  if(!a_streamer.read(fTime)) return false;
  if(!a_streamer.read(fRandom)) return false;
  return true;
}

bool Exp::Hit::stream_out(steve::interface::streamer& a_streamer) const { 
  //steve::uint16 version = CLASS_VERSION;
  //if(!a_streamer.write(version)) return false;

  if(!a_streamer.write(fX)) return false;
  if(!a_streamer.write(fY)) return false;
  if(!a_streamer.write(fZ)) return false;
  if(!steve::stream::out<double>(a_streamer,fVec)) return false;
  if(!a_streamer.write(fRandom)) return false;
  if(!steve::stream::out<Time>(a_streamer,fTimes)) return false;
#ifdef EXP_REF
  if(!steve::stream::out_ref<Track>(a_streamer,fTrack)) return false;
#endif
  return true;
}
bool Exp::Hit::stream_in(steve::interface::streamer& a_streamer) { 
  //steve::uint16 version;
  //if(!a_streamer.read(version)) return false;
  //if(version!=CLASS_VERSION) return false;

  if(!a_streamer.read(fX)) return false;
  if(!a_streamer.read(fY)) return false;
  if(!a_streamer.read(fZ)) return false;
  if(!steve::stream::in<double>(a_streamer,fVec)) return false;
  if(!a_streamer.read(fRandom)) return false;
  if(!steve::stream::in<Time>(a_streamer,fTimes)) return false;
#ifdef EXP_REF
  if(!steve::stream::in_ref<Track>(a_streamer,fTrack)) return false;
#endif
  return true;
}

bool Exp::Track::stream_out(steve::interface::streamer& a_streamer) const { 
  //steve::uint16 version = CLASS_VERSION;
  //if(!a_streamer.write(version)) return false;

  if(!a_streamer.write(fPx)) return false;
  if(!a_streamer.write(fPy)) return false;
  if(!a_streamer.write(fPz)) return false;
  if(!steve::stream::out<Hit>(a_streamer,fHits)) return false;
  if(!a_streamer.write(fRandom)) return false;
  return true;
}
bool Exp::Track::stream_in(steve::interface::streamer& a_streamer) { 
  //steve::uint16 version;
  //if(!a_streamer.read(version)) return false;
  //if(version!=CLASS_VERSION) return false;

  if(!a_streamer.read(fPx)) return false;
  if(!a_streamer.read(fPy)) return false;
  if(!a_streamer.read(fPz)) return false;
  if(!steve::stream::in<Hit>(a_streamer,fHits)) return false;
  if(!a_streamer.read(fRandom)) return false;
  return true;
}

bool Exp::Cell::stream_out(steve::interface::streamer& a_streamer) const { 
  //steve::uint16 version = CLASS_VERSION;
  //if(!a_streamer.write(version)) return false;

#ifdef EXP_STRING
  if(!a_streamer.write(fID)) return false;
  if(!a_streamer.write(fID_Vec)) return false;
#endif
  return true;
}
bool Exp::Cell::stream_in(steve::interface::streamer& a_streamer) { 
  //steve::uint16 version;
  //if(!a_streamer.read(version)) return false;
  //if(version!=CLASS_VERSION) return false;

#ifdef EXP_STRING
  if(!a_streamer.read(fID)) return false;
  if(!a_streamer.read(fID_Vec)) return false;
#endif
  return true;
}

bool Exp::Header::stream_out(steve::interface::streamer& a_streamer) const { 
  //steve::uint16 version = CLASS_VERSION;
  //if(!a_streamer.write(version)) return false;

  if(!a_streamer.write(fNumberOfEvents)) return false;
  return true;
}
bool Exp::Header::stream_in(steve::interface::streamer& a_streamer) { 
  //steve::uint16 version;
  //if(!a_streamer.read(version)) return false;
  //if(version!=CLASS_VERSION) return false;

  if(!a_streamer.read(fNumberOfEvents)) return false;
  return true;
}

bool Exp::Event::stream_out(steve::interface::streamer& a_streamer) const { 
  //steve::uint16 version = CLASS_VERSION;
  //if(!a_streamer.write(version)) return false;

  if(!a_streamer.write(fID)) return false;
  if(!a_streamer.write(fEnergy)) return false;

  if(!steve::HCL::out(a_streamer,fHisto.getData())) return false;

  if(!a_streamer.write(fBlob.size(),fBlob.data())) return false;
  if(!fObject.stream_out(a_streamer)) return false;

#ifdef EXP_REF
  if(!steve::stream::out_with_refs<Track>(a_streamer,fTracks)) return false;
#else
  if(!steve::stream::out<Track>(a_streamer,fTracks)) return false;
#endif

  if(!steve::stream::out2<Cell>(a_streamer,fCells)) return false;
  return true;
}
bool Exp::Event::stream_in(steve::interface::streamer& a_streamer) { 
  //steve::uint16 version;
  //if(!a_streamer.read(version)) return false;
  //if(version!=CLASS_VERSION) return false;

  if(!a_streamer.read(fID)) return false;
  if(!a_streamer.read(fEnergy)) return false;

  if(!steve::HCL::in(a_streamer,fHisto.getData())) return false;

 {steve::uint32 sz;
  char* buffer;
  if(!a_streamer.read(sz,buffer)) return false;
  fBlob.assign(sz,buffer);}

  if(!fObject.stream_in(a_streamer)) return false;

#ifdef EXP_REF
  if(!steve::stream::in_with_refs<Track>(a_streamer,fTracks)) return false;
#else
  if(!steve::stream::in<Track>(a_streamer,fTracks)) return false;
#endif

  if(!steve::stream::in2<Cell>(a_streamer,fCells)) return false;
  return true;
}

/*
#define DUMMY_BUFFER(a_class)\
bool Exp::a_class::stream_out(steve::interface::streamer& a_streamer) const {return false;}\
bool Exp::a_class::stream_in(steve::interface::streamer& a_streamer) { return false;}
*/

