#ifndef Exp_instrument_h
#define Exp_instrument_h

#include <steve/typedefs.h>
#include <steve/path.h>

namespace steve {namespace interface {class store;}}
namespace steve {namespace interface {class schema;}}
namespace steve {namespace interface {class streamer;}}

//////////////////////////////////////////////////////////////////////////////
/// ClassDef /////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#ifdef STEVE_USE_ROOT
#include <TObject.h>
#include <TClonesArray.h>
#include <TRef.h>
#endif

#ifdef STEVE_USE_STEVE
#define ClassDef(a_class,a_version) \
public:\
  bool store(steve::interface::store&,steve::interface::schema&) const;\
  bool retrieve(steve::interface::store&,steve::interface::schema&,\
                steve::uint32,const steve::path_t&);\
public:\
  bool stream_out(steve::interface::streamer& a_streamer) const;\
  bool stream_in(steve::interface::streamer& a_streamer);
#endif

//////////////////////////////////////////////////////////////////////////////
/// ARRAY_STAR ///////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#ifdef STEVE_USE_STEVE
#define ARRAY_STAR_FIELD(a_field,a_class) std::vector<a_class*> a_field
#endif
#ifdef STEVE_USE_ROOT
#define ARRAY_STAR_FIELD(a_field,a_class) TClonesArray a_field
#endif

#ifdef STEVE_USE_STEVE
#define ARRAY_STAR_INIT(a_field,a_class) a_field()
#endif
#ifdef STEVE_USE_ROOT
#define ARRAY_STAR_INIT(a_field,a_class) a_field(#a_class)
#endif

#ifdef STEVE_USE_STEVE
#define ARRAY_STAR_ADD(a_array,a_index,a_item) a_array.push_back(new a_item)
#endif
#ifdef STEVE_USE_ROOT
#define ARRAY_STAR_ADD(a_array,a_index,a_item) new(a_array[a_index]) a_item
#endif

#ifdef STEVE_USE_STEVE
#define ARRAY_STAR_METHODS(a_class,a_field,a_field_class,a_number,a_at)\
  unsigned int a_class::a_number() const { return a_field.size();}\
  a_field_class* a_class::a_at(unsigned int a_index) { return a_field[a_index];}
#endif
#ifdef STEVE_USE_ROOT
#define ARRAY_STAR_METHODS(a_class,a_field,a_field_class,a_number,a_at)\
  unsigned int a_class::a_number() const { return a_field.GetEntries();}\
  a_field_class* a_class::a_at(unsigned int a_index) { return (a_field_class*)a_field[a_index];}
#endif

#ifdef STEVE_USE_STEVE
#define ARRAY_STAR_DELETE(a_field,a_class)\
   {std::vector<a_class*>::const_iterator it;\
    for(it=a_field.begin();it!=a_field.end();++it) delete (*it);\
    a_field.clear();}
#endif
#ifdef STEVE_USE_ROOT
#define ARRAY_STAR_DELETE(a_field,a_class)
#endif

//////////////////////////////////////////////////////////////////////////////
/// ARRAY ///////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#ifdef STEVE_USE_STEVE
#define ARRAY_FIELD(a_field,a_class) std::vector<a_class> a_field
#endif
#ifdef STEVE_USE_ROOT
#define ARRAY_FIELD(a_field,a_class) TClonesArray a_field
#endif

#ifdef STEVE_USE_STEVE
#define ARRAY_INIT(a_field,a_class) a_field()
#endif
#ifdef STEVE_USE_ROOT
#define ARRAY_INIT(a_field,a_class) a_field(#a_class)
#endif

#ifdef STEVE_USE_STEVE
#define ARRAY_ADD(a_array,a_index,a_item) a_array.push_back(a_item)
#endif
#ifdef STEVE_USE_ROOT
#define ARRAY_ADD(a_array,a_index,a_item) new(a_array[a_index]) a_item
#endif

#ifdef STEVE_USE_STEVE
#define ARRAY_METHODS(a_class,a_field,a_field_class,a_number,a_at)\
  unsigned int a_class::a_number() const { return a_field.size();}\
  a_field_class* a_class::a_at(unsigned int a_index) { return &(a_field[a_index]);}
#endif
#ifdef STEVE_USE_ROOT
#define ARRAY_METHODS(a_class,a_field,a_field_class,a_number,a_at)\
  unsigned int a_class::a_number() const { return a_field.GetEntries();}\
  a_field_class* a_class::a_at(unsigned int a_index) { return (a_field_class*)a_field[a_index];}
#endif

#ifdef STEVE_USE_STEVE
#define ARRAY_DELETE(a_field,a_class)\
    a_field.clear();
#endif
#ifdef STEVE_USE_ROOT
#define ARRAY_DELETE(a_field,a_class)
#endif

//////////////////////////////////////////////////////////////////////////////
/// REF //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#ifdef STEVE_USE_STEVE
#define REF_FIELD(a_field,a_class) a_class* a_field
#endif
#ifdef STEVE_USE_ROOT
#define REF_FIELD(a_field,a_class) TRef a_field
#endif

#ifdef STEVE_USE_STEVE
#define REF_INIT(a_field,a_pointer) a_field(a_pointer)
#endif
#ifdef STEVE_USE_ROOT
#define REF_INIT(a_field,a_pointer) a_field((TObject*)a_pointer)
#endif

#ifdef STEVE_USE_STEVE
#define REF_METHOD(a_class,a_field,a_field_class,a_method)\
  a_field_class* a_class::a_method() const { return a_field;}
#endif
#ifdef STEVE_USE_ROOT
#define REF_METHOD(a_class,a_field,a_field_class,a_method)\
  a_field_class* a_class::a_method() const { return (a_field_class*)a_field.GetObject();}
#endif

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#endif
